# This file is part of the SMUML toolset.
# Copyright (C) 2006-2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.


# Author: Tommi Junttila

"""
State machine action language flattener.

Flattens the 'for', 'if', 'while', 'do', break', and 'continue'
constructs appearing in the effects of the transitions in
a UML StateMachine.
Can add new pseudostates in the state machine.
Does not flatten the state hierarchy.
"""

import sys

import coral

import actlang
import Sim
import Utils.Models
import Utils.StateMachines

UML14 = coral.lang.UML14

def _qn(e):
    # Just an abbreviation
    return Utils.Models.get_qualified_name(e)

def _mcname(e):
    # Just an abbreviation
    return Utils.Models.get_metaclassname(e)



def _copy_states(state, state_map = None):
    """
    Copy the StateVertex 'state' and all its descendants.
    Returns a map from the original states to the new ones.
    """
    assert state.oclIsKindOf(UML14.StateVertex)
    if state_map is None:
        state_map = {}

    if state.oclIsKindOf(UML14.Pseudostate):
        new_state = UML14.Pseudostate()
        new_state.kind = state.kind
    elif state.oclIsKindOf(UML14.FinalState):
        new_state = UML14.FinalState()
    elif state.oclIsKindOf(UML14.SimpleState):
        new_state = UML14.SimpleState()
    elif state.oclIsKindOf(UML14.CompositeState):
        for subvertex in state.subvertex:
            _copy_states(subvertex, state_map)
        new_state = UML14.CompositeState()
        for subvertex in state.subvertex:
            new_state.subvertex.add(state_map[subvertex])
        new_state.isConcurrent = state.isConcurrent
        new_state.isRegion = state.isRegion
    else:
        raise RuntimeError("State type not handled: " + _mcname(state))

    # Copy attributes common to all state vertices
    new_state.isSpecification = state.isSpecification
    new_state.name = state.name

    # Copy attributes common to all states
    if state.oclIsKindOf(UML14.State):
        if state.entry:
            # TO DO: Make a copy of entry action
            raise RuntimeError("Entry actions are not supported")
        if state.exit:
            # TO DO: Make a copy of exit action
            raise RuntimeError("Exit actions are not supported")
        if state.entry:
            # TO DO: Make a copy of do activity
            raise RuntimeError("Do activities are not supported")
        if state.internalTransition:
            raise RuntimeError("Internal transitions are not supported")
        # Copy deferrable events
        for e in state.deferrableEvent:
            new_state.deferrableEvent.add(e)

    assert not state in state_map
    state_map[state] = new_state
    return state_map




        

def _copy_guard(g):
    """
    An auxiliary routine making a copy of the transition guard 'g'.
    """
    if g is None:
        return None
    assert g.oclIsKindOf(UML14.Guard)
    new_g = UML14.Guard()
    new_g.name = g.name
    new_g.expression.body = g.expression.body
    assert not g.expression.language
    assert not g.expression.xmi_extension
    return new_g




class _CFG:
    """
    A class for Control Flow Graphs.
    """
    class Node:
        def __init__(self, label = None):
            assert label is None or type(label) is str
            self.label = label
            self.incoming = []
            self.outgoing = []
        def is_exit(self):
            return self.outgoing == []
    class Transition:
        def __init__(self, source, target, guard, code):
            """
            'code' is either None, an action language expr or statement,
            or a list of action language expressions and statements
            """
            assert isinstance(source, _CFG.Node)
            assert isinstance(target, _CFG.Node)
            assert (guard is None) or (type(guard) is str and str != "")
            assert((code is None) or
                   isinstance(code, actlang.expr.Expr) or
                   isinstance(code, actlang.statement.Statement) or
                   (type(code) is list))
            self.source = source
            source.outgoing.append(self)
            self.target = target
            target.incoming.append(self)
            self.guard = guard
            self.code = []
            if code is None:
                pass
            elif isinstance(code, actlang.expr.Expr):
                self.code.append(code)
            elif isinstance(code, actlang.statement.Statement):
                self.code.append(code)
            elif type(code) is list:
                self.code.extend(code)
            else:
                raise RuntimeError("Internal error")
        def get_guard(self):
            return self.guard
        def has_guard(self):
            return self.guard is not None
        def nof_send_statements(self):
            n = 0
            for stmt in self.code:
                if isinstance(stmt, actlang.statement.SendStatement):
                    n += 1
            return n
        def has_send_statements(self):
            """
            Does the transition have any send statements in its effect?
            """
            return self.nof_send_statements() > 0
        def has_statements(self):
            """
            Does the transition have an effect?
            """
            return self.code != []
    def __init__(self):
        """
        Create an empty CFG with one node.
        """
        source = self.Node()
        self.nodes = [source]
        self.source = source
        self.labeled_exits = {}
    def get_source(self):
        return self.source
    def set_source(self, node):
        self.source = node
    def has_exit_label(self, label):
        return label in self.labeled_exits
    def get_labeled_exits(self, label):
        if label in self.labeled_exits:
            return self.labeled_exits[label]
        return []
    def remove_exit_labels(self, label):
        if label in self.labeled_exits:
            for node in self.labeled_exits[label]:
                node.label = None
            del self.labeled_exits[label]
    def add_node(self, label = None):
        node = self.Node(label)
        self.nodes.append(node)
        if label:
            if not label in self.labeled_exits:
                self.labeled_exits[label] = []
            self.labeled_exits[label].append(node)
        return node
    def add_transition(self, source, target, guard, code):
        """
        Source shall not be a labeled exit node.
        """
        assert source.label is None
        _CFG.Transition(source, target, guard, code)
    def get_unlabeled_exits(self):
        return [n for n in self.nodes if ((not n.outgoing) and (not n.label))]
    def concatenate(self, other):
        unlabeled_exits = self.get_unlabeled_exits()
        if not unlabeled_exits:
            return
        mapping = {}
        for other_node in other.nodes:
            mapping[other_node] = self.add_node(other_node.label)
        for other_node in other.nodes:
            for t in other_node.outgoing:
                self.add_transition(mapping[t.source],
                                    mapping[t.target],
                                    t.guard,
                                    t.code[:])
        for old_exit in unlabeled_exits:
            self.add_transition(old_exit, mapping[other.source], None, None)
    def insert(self, other):
        mapping = {}
        for other_node in other.nodes:
            mapping[other_node] = self.add_node(other_node.label)
        for other_node in other.nodes:
            for t in other_node.outgoing:
                self.add_transition(mapping[t.source],
                                    mapping[t.target],
                                    t.guard,
                                    t.code[:])
        return mapping
    
    def simplify(self,
                 opt_single_statement = False,
                 opt_atmost_one_send_statement = False):
        """
        Perform some simplifications in the CFG.

        If 'opt_single_statement' is True,
        then each transition will have only one statement.

        If 'opt_atmost_one_send_statement' is True,
        then each transition will have at most one send statement.
        """
        assert type(opt_single_statement) is bool
        assert type(opt_atmost_one_send_statement) is bool
        changes = True
        while changes:
            changes = False
            for node in self.nodes:
                if len(node.outgoing) == 1:
                    assert not node.label
                    single_t = node.outgoing[0]
                    if single_t.target is node:
                        # Self-loop, do nothing
                        continue
                    if((single_t.guard is None) and
                       (not single_t.code)):
                        # Redirect incoming transitions to the target of
                        # single_t
                        for t in node.incoming:
                            t.target = single_t.target
                            t.target.incoming.append(t)
                        if node is self.source:
                            self.source = single_t.target
                        single_t.target.incoming.remove(single_t)
                        node.incoming = []
                        node.outgoing = []
                        changes = True
                        continue
                    if((single_t.guard is None) and
                       (len(node.incoming) == 1) and
                       (node is not self.source)):
                        incoming_t = node.incoming[0]
                        if(opt_atmost_one_send_statement and
                           incoming_t.has_send_statements() and
                           single_t.has_send_statements()):
                            continue
                        if(opt_single_statement and
                           incoming_t.has_statements() and
                           single_t.has_statements()):
                            continue
                        incoming_t.code.extend(single_t.code)
                        single_t.code = []
                        changes = True
                        continue
        # Remove isolated nodes that are not the source or a labeled exit
        self.nodes = [n for n in self.nodes if
                      ((n is self.source) or n.label or
                       n.incoming or n.outgoing)]



    def print_dotty(self, f):
        """
        Print the control flow graph in the dotty format to 'f',
        where 'f' is either a file handle or the name of the output file.
        """
        assert (type(f) is file) or (type(f) is str and f != "")
        if type(f) is str:
            fs = open(f, "w")
        else:
            fs = f
        fs.write("digraph CFG {\n")
        for n in self.nodes:
            fs.write("n"+str(id(n)))
            if n.label:
                fs.write(" [label=\""+n.label+"\"]")
            else:
                fs.write(" [label=\"\"]")
            if n is self.source:
                fs.write(" [color=blue]")
            fs.write(";\n")
            for t in n.outgoing:
                fs.write("n"+str(id(t.source))+" -> n"+str(id(t.target))+" ")
                fs.write("[label=\"")
                if t.guard:
                    fs.write("["+t.guard+"]")
                if t.code:
                    fs.write("/")
                    for s in t.code:
                        if isinstance(s, actlang.expr.Expr):
                            fs.write(" "+str(s)+";")
                        elif isinstance(s, actlang.statement.Statement):
                            fs.write(" "+str(s))
                        else:
                            raise RuntimeError("Internal error")
                fs.write("\"];\n")
        fs.write("}\n")
        if type(f) is str:
            fs.close()




def _stmt_to_cfg(stmt, my_label = None):
    assert(isinstance(stmt, actlang.statement.Statement) or
           isinstance(stmt, actlang.expr.Expr))
    assert (my_label is None) or (my_label and type(my_label) is str)

    if isinstance(stmt, actlang.statement.EmptyStatement):
        # Nothing to do
        return _CFG()


    elif isinstance(stmt, actlang.statement.WhileStatement):
        condition = str(stmt.api_getCondition())
        
        cfg = _stmt_to_cfg(stmt.api_getBody())
        body_exits = cfg.get_unlabeled_exits()

        body_breaks = cfg.get_labeled_exits("break ")
        cfg.remove_exit_labels("break ")
        body_continues = cfg.get_labeled_exits("continue ")
        cfg.remove_exit_labels("continue ")
        
        new_entry = cfg.add_node()
        new_exit = cfg.add_node()
        cfg.add_transition(new_entry, cfg.get_source(), condition, None)
        cfg.add_transition(new_entry, new_exit, "!("+condition+")", None)
        for body_exit in body_exits:
            cfg.add_transition(body_exit, new_entry, None, None)
        for body_continue in body_continues:
            cfg.add_transition(body_continue, new_entry, None, None)
        for body_break in body_breaks:
            cfg.add_transition(body_break, new_exit, None, None)
        cfg.set_source(new_entry)
        return cfg


    elif isinstance(stmt, actlang.statement.DoStatement):
        condition = str(stmt.api_getCondition())
        
        cfg = _stmt_to_cfg(stmt.api_getBody())
        body_exits = cfg.get_unlabeled_exits()

        body_breaks = cfg.get_labeled_exits("break ")
        cfg.remove_exit_labels("break ")
        body_continues = cfg.get_labeled_exits("continue ")
        cfg.remove_exit_labels("continue ")
        if my_label:
            body_continues.extend(cfg.get_labeled_exits("continue "+my_label))
            cfg.remove_exit_labels("continue "+my_label)

        new_test = cfg.add_node()
        new_exit = cfg.add_node()
        cfg.add_transition(new_test, cfg.get_source(), condition, None)
        cfg.add_transition(new_test, new_exit, "!("+condition+")", None)
        for body_exit in body_exits:
            cfg.add_transition(body_exit, new_test, None, None)
        for body_continue in body_continues:
            cfg.add_transition(body_continue, new_test, None, None)
        for body_break in body_breaks:
            cfg.add_transition(body_break, new_exit, None, None)
        return cfg


    elif isinstance(stmt, actlang.statement.ExprListForStatement):
        cfg = _CFG()
        for init_stmt in stmt.api_getInitialization():
            cfg.concatenate(_stmt_to_cfg(init_stmt))
        init_exits = cfg.get_unlabeled_exits()

        loop_entry = cfg.add_node()
        for init_exit in init_exits:
            cfg.add_transition(init_exit, loop_entry, None, None)

        condition = str(stmt.api_getCondition())

        body_cfg = _stmt_to_cfg(stmt.api_getBody())
        body_cfg.remove_exit_labels("continue ")
        body_cfg_exits = body_cfg.get_unlabeled_exits()

        update_cfg = _CFG()
        for update_stmt in stmt.api_getUpdatePart():
            update_cfg.concatenate(_stmt_to_cfg(update_stmt))

        m = body_cfg.insert(update_cfg)

        for body_exit in body_cfg_exits:
            body_cfg.add_transition(body_exit, m[update_cfg.source],
                                    None, None)

        body_cfg.remove_exit_labels("continue ")
        body_cfg_exits = body_cfg.get_unlabeled_exits()
        body_breaks = body_cfg.get_labeled_exits("break ")
        body_cfg.remove_exit_labels("break ")

        m = cfg.insert(body_cfg)
        #body_cfg_exits = body_cfg.get_unlabeled_exits()
        for body_exit in body_cfg_exits:
            cfg.add_transition(m[body_exit], loop_entry, None, None)
        cfg.add_transition(loop_entry, m[body_cfg.source],
                           str(condition), None)

        new_exit = cfg.add_node()
        cfg.add_transition(loop_entry, new_exit,
                           "!("+str(condition)+")", None)
        for body_break in body_breaks:
            cfg.add_transition(m[body_break], new_exit, None, None)

        return cfg


    elif isinstance(stmt, actlang.statement.LocalVarForStatement):
        raise RuntimeError("Local variable declarations are not supported: "+
                           str(stmt))


    elif isinstance(stmt, actlang.statement.IfStatement):
        condition = str(stmt.api_getCondition())
        
        cfg_true = _stmt_to_cfg(stmt.api_getTrueStatement())
        cfg_false = _stmt_to_cfg(stmt.api_getFalseStatement())

        cfg = _CFG()
        cfg_true_map = cfg.insert(cfg_true)
        cfg_false_map = cfg.insert(cfg_false)
        
        cfg.add_transition(cfg.source, cfg_true_map[cfg_true.source],
                           condition, None)
        cfg.add_transition(cfg.source, cfg_false_map[cfg_false.source],
                           "!("+condition+")", None)

        return cfg


    elif isinstance(stmt, actlang.statement.Block):
        # Simply concatenate the CFGs for the statements in the block
        stmts = stmt.api_getBlockStatements()
        assert len(stmts) >= 1
        cfg = _CFG()
        for stmt in stmts:
            sub_cfg = _stmt_to_cfg(stmt)
            cfg.concatenate(sub_cfg)
        return cfg


    elif isinstance(stmt, actlang.statement.LabeledStatement):
        label = stmt.api_getLabelName()
        assert label
        cfg = _stmt_to_cfg(stmt.api_getContainedStatement(), my_label = label)
        if cfg.has_exit_label("break "+label):
            cfg.remove_exit_labels("break "+label)
        if cfg.has_exit_label("continue "+label):
            continue_nodes = cfg.get_labeled_exits("continue "+label)
            cfg.remove_exit_labels("continue "+label)
            for continue_node in continue_nodes:
                cfg.add_transition(continue_node, cfg.source, None, None)
        return cfg


    elif isinstance(stmt, actlang.statement.BreakStatement):
        label = stmt.api_getLabelName()
        if label is None:
            label = ""
        cfg = _CFG()
        new_exit = cfg.add_node("break "+label)
        cfg.add_transition(cfg.source, new_exit, None, None)
        return cfg


    elif isinstance(stmt, actlang.statement.ContinueStatement):
        label = stmt.api_getLabelName()
        if label is None:
            label = ""
        cfg = _CFG()
        new_exit = cfg.add_node("continue "+label)
        cfg.add_transition(cfg.source, new_exit, None, None)
        return cfg


    elif isinstance(stmt, (actlang.expr.SimpleAssignmentExpr,
                           actlang.expr.PostfixExpr)):
        cfg = _CFG()
        source = cfg.get_source()
        target = cfg.add_node()
        cfg.add_transition(source, target, None, stmt)
        return cfg


    elif isinstance(stmt, (actlang.expr.SimpleAssignmentExpr,
                           actlang.expr.PostfixExpr,
                           actlang.expr.PrefixExpr,
                           actlang.statement.ExprStatement,
                           actlang.statement.AssertStatement,
                           actlang.statement.SendStatement)):
        cfg = _CFG()
        source = cfg.get_source()
        target = cfg.add_node()
        cfg.add_transition(source, target, None, stmt)
        return cfg


    elif isinstance(stmt, actlang.statement.LocalVarDeclarationStatement):
        raise RuntimeError("Local variable declarations are not supported: "+
                           str(stmt))

    else:
        print stmt.__class__
        raise RuntimeError("Cannot handle statement '"+str(stmt)+"'")



def _stmts_to_cfg(stmts):
    """
    Conver a list (concatenation) of statements into the corresponding CFG.
    """
    assert type(stmts) is list
    assert stmts != []

    cfg = _CFG()
    for (index,stmt) in enumerate(stmts):
        sub_cfg = _stmt_to_cfg(stmt)
        if sub_cfg.labeled_exits:
            raise RuntimeError("Statement has unbound labels ("+
                               (",".join(sub_cfg.labeled_exits))+"):\n"+
                               str(stmt))
        #sub_cfg.print_dotty("CFG_sub"+str(index)+".dot")
        cfg.concatenate(sub_cfg)
        #cfg.print_dotty("CFG_ssub"+str(index)+".dot")
    return cfg




def flatten(sm,
            opt_single_statement = False,
            opt_atmost_one_send_statement = False,
            stdout = sys.stdout, stderr = sys.stderr):
    """
    Flattens the 'if', 'for', 'while', 'do while', 'break' and 'continue'
    constructs appearing in the effects of the transitions in
    the UML StateMachine 'sm'.
    Can add new pseudostates in the state machine.
    Does NOT flatten the state hierarchy.
    
    If 'opt_single_statement' is True,
    then each transition will have only one statement in its effect.

    If 'opt_atmost_one_send_statement' is True,
    then each transition will have at most one send statement in its effect.
    """
    assert sm.oclIsKindOf(UML14.StateMachine)
    assert type(opt_single_statement) is bool
    assert type(opt_atmost_one_send_statement) is bool

    assert sm.top
    assert not sm.top.isConcurrent
    if not sm.top.oclIsKindOf(UML14.CompositeState):
        raise RuntimeError("State machines with non-composite top states not supported")
    if sm.top.oclIsKindOf(UML14.SubmachineState):
        raise RuntimeError("State machines with submachineState top states not supported")
    if sm.top.entry:
        raise RuntimeError("State machines with top states with entry actions not supported")
    if sm.top.exit:
        raise RuntimeError("State machines with top states with exit actions not supported")
    if sm.top.doActivity:
        raise RuntimeError("State machines with top states with do activity not supported")

    context = sm.context
    assert context
    assert context.oclIsKindOf(UML14.Class)
    
    # Create a new state machine
    new_sm = UML14.StateMachine()

    # Copy the states
    state_map = _copy_states(sm.top)
    new_sm.top = state_map[sm.top]
    new_sm.top.stateMachine = new_sm

    _flatten_effects(new_sm,
                     sm.top,
                     state_map,
                     opt_single_statement,
                     opt_atmost_one_send_statement)

    return new_sm


def _flatten_effects(new_sm,
                     state,
                     state_map,
                     opt_single_statement = False,
                     opt_atmost_one_send_statement = False):
    """
    Flatten action language control flow structures
    (while, if, for, do, break, continue)
    of transition effects.
    """
    assert new_sm.oclIsKindOf(UML14.StateMachine)
    assert state.oclIsKindOf(UML14.StateVertex)
    assert type(state_map) is dict
    assert type(opt_single_statement) is bool
    assert type(opt_atmost_one_send_statement) is bool

    new_state = state_map[state]
    
    # Recurse on substates
    if state.oclIsKindOf(UML14.CompositeState):
        for subvertex in state.subvertex:
            _flatten_effects(new_sm,
                             subvertex,
                             state_map,
                             opt_single_statement,
                             opt_atmost_one_send_statement)

    for t in state.outgoing:
        if not t.effect:
            # No effect -> nothing to flatten -> just copy the transition
            new_t = UML14.Transition()
            state_map[t.source].outgoing.add(new_t)
            state_map[t.target].incoming.add(new_t)
            new_t.trigger = t.trigger
            new_t.guard = _copy_guard(t.guard)
            new_sm.transitions.add(new_t)
            continue

        assert(t.effect.oclIsKindOf(UML14.Action) or
               t.effect.oclIsKindOf(UML14.UninterpretedAction))
        assert t.effect.recurrence.body == ""
        assert t.effect.script.body != ""
        assert t.effect.target.body == ""
        
        # Parse the transition effect Jumbala code
        try:
            tree = actlang.grammar.parse(str(t.effect.script.body))
        except actlang.errors.Error, e:
            raise RuntimeError("Syntax error on the effect "+
                               "'"+str(t.effect.script.body)+"' "
                               "of transition "+_qn(t)+": "+str(e))
        # Prune away empty statements
        stmts = [stmt for stmt in tree.api_getTopStatements() if
                 not isinstance(stmt, actlang.statement.EmptyStatement)]

        if not stmts:
            # No statements -> nothing to flatten -> just copy the transition
            new_t = UML14.Transition()
            state_map[t.source].outgoing.add(new_t)
            state_map[t.target].incoming.add(new_t)
            new_t.trigger = t.trigger
            new_t.guard = _copy_guard(t.guard)
            new_sm.transitions.add(new_t)
            continue

        assert len(stmts) > 0

        # Make the flattened control flow graph (CFG)
        cfg = _stmts_to_cfg(stmts)
        #cfg.print_dotty("CFG.dot")
        cfg.simplify(opt_single_statement = opt_single_statement,
                     opt_atmost_one_send_statement = opt_atmost_one_send_statement)
        #cfg.print_dotty("CFG2.dot")
        assert len(cfg.nodes) >= 2
        assert not cfg.get_source().is_exit()


        # Create a new choice Pseudostate for each node in the flattened CFG.
        # In order to preserve the execution order of exit actions, transition
        # effect, and entry actions, the choice Pseudostates are located
        # in the least common ancestor non-concurrent composite state of
        # the source and target state vertices (i.e. the container of
        # the transition).

        # Compute container
        container = Utils.StateMachines.get_least_common_ancestor(state_map[t.source], state_map[t.target])
        assert container.oclIsKindOf(UML14.CompositeState)
        while container.isConcurrent:
            assert container.container
            assert container.container.oclIsKindOf(UML14.CompositeState)
            container = container.container
        assert container.oclIsKindOf(UML14.CompositeState)

        if((len(cfg.get_source().outgoing) > 1) or
           (len(cfg.get_source().outgoing) == 1 and
            cfg.get_source().outgoing[0].has_guard() and
            new_t.guard and new_t.guard.expression.body)):
            new_source = cfg.add_node()
            cfg.add_transition(new_source, cfg.get_source(), None, None)
            cfg.set_source(new_source)

        # Create a new choice pseudostate for each non-source, non-exit
        # node in the CFG
        cfg_node_to_choise_state_map = {}
        for node in cfg.nodes:
            if node is cfg.get_source():
                continue
            if node.is_exit():
                continue
            new_vertex = UML14.Pseudostate()
            new_vertex.kind = "choice"
            container.subvertex.add(new_vertex)
            cfg_node_to_choise_state_map[node] = new_vertex


        for cfg_node in cfg.nodes:
            if cfg_node is cfg.get_source():
                assert len(cfg_node.outgoing) == 1
                cfg_t = cfg_node.outgoing[0]

                new_t = UML14.Transition()
                # Compute source
                uml_source = state_map[t.source]
                # Compute target
                cfg_target = cfg_t.target
                if cfg_target.is_exit():
                    uml_target = state_map[t.target]
                else:
                    uml_target = cfg_node_to_choise_state_map[cfg_target]
                # Make trigger
                new_t.trigger = t.trigger
                if cfg_t.has_guard():
                    assert not t.guard
                    new_t.guard = UML14.Guard()
                    new_t.guard.expression.body = cfg_t.get_guard()
                # Make effect
                e = ""
                for s in cfg_t.code:
                    if isinstance(s, actlang.expr.Expr):
                        e += " "+str(s)+";"
                    elif isinstance(s, actlang.statement.Statement):
                        e += " "+str(s)
                    else:
                        raise RuntimeError("Internal error")
                if e != "":
                    new_t.effect = UML14.UninterpretedAction()
                    new_t.effect.script.body = e
                # Install transition
                uml_source.outgoing.add(new_t)
                uml_target.incoming.add(new_t)
                new_sm.transitions.add(new_t)
            else:
                for cfg_t in cfg_node.outgoing:
                    new_t = UML14.Transition()
                    # Compute source state
                    uml_source = cfg_node_to_choise_state_map[cfg_node]
                    # Compute target state
                    cfg_target = cfg_t.target
                    if cfg_target.is_exit():
                        uml_target = state_map[t.target]
                    else:
                        uml_target = cfg_node_to_choise_state_map[cfg_target]
                    # Make guard
                    if cfg_t.has_guard():
                        new_t.guard = UML14.Guard()
                        new_t.guard.expression.body = cfg_t.get_guard()
                    # Make effect
                    e = ""
                    for s in cfg_t.code:
                        if isinstance(s, actlang.expr.Expr):
                            e += " "+str(s)+";"
                        elif isinstance(s, actlang.statement.Statement):
                            e += " "+str(s)
                        else:
                            raise RuntimeError("Internal error")
                    if e != "":
                        new_t.effect = UML14.UninterpretedAction()
                        new_t.effect.script.body = e
                    # Install transition
                    uml_source.outgoing.add(new_t)
                    uml_target.incoming.add(new_t)
                    new_sm.transitions.add(new_t)
 




#
# Just some testing code
#
if __name__ == 'TEST__main__':
    c = "x += 1; y = x / 3; send s() to x; assert(true); send y() to z; for(i=0, j=2; i<10; i++) a: while(i<3) {if(i>4?b:new A()) continue a; else if(k) j--; else break a; k++;}"
    #c = "a: {x=0; y = x+3; if(new A().x == 3) continue a; } k = 1;"
    #c = "a: while(i) b: for(a.j=k++;j<7;j++) {if(i<j) break a; if(k) break; }"
    #c = "a: do {i = i+1; if(i) continue a; if(j) break a; k++; } while(true);"
    #c = "int[10] x;" # = {1,2,3,4};"

    try:
        tree = actlang.grammar.parse(c)
    except actlang.errors.Error, e:
        raise RuntimeError("Syntax error on '"+c+"': "+str(e))
    # Prune away empty statements
    stmts = [stmt for stmt in tree.api_getTopStatements() if
             not isinstance(stmt, actlang.statement.EmptyStatement)]

    #print "\n".join([str(s.__class__)+": "+str(s) for s in stmts])

    cfg = _stmts_to_cfg(stmts)
    cfg.print_dotty("CFG.dot")
    cfg.simplify(opt_single_statement = False,
                 opt_atmost_one_send_statement = True)
    cfg.print_dotty("CFG2.dot")
