#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# actions.py
#
# 2007-09-21  Jori Dubrovin:  Limiting comment length in NuSMV code
# 2007-09-05  Jori Dubrovin:  Fixed a module name
# 2007-06-20  Jori Dubrovin:  New choose expression mapping
# 2007-03-15  Jori Dubrovin:  Added an error message
# 2007-03-02  Jori Dubrovin:  Adding choose to element map
# 2006-12-18  Jori Dubrovin:  Adding nondeterministic choose
# 2006-12-05  Jori Dubrovin:  Fixed cast error handling
# 2006-08-22  Jori Dubrovin:  Added modules, fixed arithmetic
# 2006-08-16  Jori Dubrovin:  Added checks for run-time errors
# 2006-07-13  Jori Dubrovin:  Initial version
#

"""Translation of Jumbala to NuSMV."""

import exceptions

import actlang
import errors
import options
from datatypes import nullType, intType, boolType
from smvutils import smvAnd, smvOr, smvWord, smvWordType
from smvutils import codeDefine, codeTrans, codeAtMostOne


_chooseIntMax = 0  # Max number of choices in choose_int

def processActions(actionString, activity):
    """Apply the effect of Jumbala actions.

    The necessary expressions are added to the activity context via
    the given activity.  Any field read or write accesses and signal
    transmissions are added to the activity.

    Param actionString: the actions as a string containing zero or
    more Jumbala statements..

    Param activity: the Activity giving the current valuations of
    fields and receiving the changes.
    """
    tree = actlang.grammar.parse(actionString)
    # Extract statements other than plain semicolon
    stmts = [stmt for stmt in tree.api_getTopStatements() if
             not isinstance(stmt, actlang.statement.EmptyStatement)]
    sp = _StatementProcessor(activity)
    for stmt in stmts:
        sp.processStatement(stmt)

def processExpr(exprString, desiredType, activity):
    """Generate a side-effect free expression.

    As a side effect, the necessary expressions are added to the
    activity context and any field read accesses are added to the
    activity.

    Param exprString: a string containing the side-effect free Jumbala
    expression.

    Param desiredType: the expected type of the expression.  Must not
    be the null type.  If the expression cannot be cast to this type,
    an InputError is raised.

    Param activity: the Activity containing the current valuations of
    fields.

    Return value: expression identifier string valid in the activity
    context.
    """
    tree = actlang.grammar.parse('value = (' + exprString + '\n);')
    if len(tree.api_getTopStatements()) > 1:
        raise errors.InputError('invalid expression ' + exprString)
    et = _ExprTranslator(activity)
    # Translate the right hand side of the assignment 'value = (...)'
    stmt = tree.api_getTopStatements()[0]
    (type_, exprId) = et.translate(stmt.api_getExpr().api_getRhsExpr())
    if not type_.isCastableTo(desiredType):
        raise errors.InputError('invalid type for expression ' + exprString)
    return et.cast(desiredType, type_, exprId)


def _shallowStr(node, depth=3):
    """Return a short string representation of a Jumbala syntax tree node.

    Param node: a SyntaxNode.

    Param depth: maximum recursion depth.

    Return value: string representation of node.
    """
    if depth > 0:
        return node.api_recurseStr(_shallowStr, depth-1)
    else:
        return '..'


class _StatementProcessor:
    """A processor of parsed Jumbala statements in an activity context."""

    def __init__(self, activity):
        """Construct a statement processor.

        Param activity: the Activity giving valuations of fields.  The
        effects of statements are made to this.
        """
        self.__activity = activity
        self.__et = _ExprTranslator(activity)

    def processStatement(self, stmt):
        """Apply the effect of a Statement."""
        if isinstance(stmt, actlang.statement.ExprStatement):
            expr = stmt.api_getExpr()
            if isinstance(expr, actlang.expr.SimpleAssignmentExpr):
                self.__processAssignment(expr)
            else:
                raise errors.NotSupportedError('unsupported statement ' +
                                               str(stmt))
        elif isinstance(stmt, actlang.statement.SendStatement):
            self.__processSendStatement(stmt)
        elif isinstance(stmt, actlang.statement.AssertStatement):
            self.__processAssertStatement(stmt)
        else:
            raise errors.NotSupportedError('unsupported statement ' +
                                           str(stmt))

    def __processSendStatement(self, stmt):
        """Apply the effect of a SendStatement."""
        try:
            signal = self.__activity.getSignalForName(stmt.api_getSignalName())
        except exceptions.LookupError:
            raise errors.InputError('unknown signal in ' + str(stmt))
        if signal.isExternal():
            raise errors.InputError('illegal signal in ' + str(stmt))
        # Translate arguments before target to keep consistent with
        # the left-to-right reading order.
        args = [self.__et.translate(argExpr) for
                argExpr in stmt.api_getArgList()]
        targetExpr = stmt.api_getTarget()
        (targetType, targetId) = self.__et.translate(targetExpr)
        if not targetType.isReferenceType():
            raise errors.InputError('invalid target reference in ' + str(stmt))
        targetClass = targetType.getClass()
        self.__activity.addError(targetId + '.error')
        self.__activity.addError(targetId + '.isNull')
        paramTypes = signal.getParamTypes()
        if len(args) != len(paramTypes):
            raise errors.InputError('wrong number of arguments in ' +
                                    str(stmt))
        argIds = []
        for ((argType, exprId), paramType) in zip(args, paramTypes):
            if not argType.isSubtypeOf(paramType):
                raise errors.InputError('invalid argument type in ' +
                                        str(stmt))
            exprId = self.__et.cast(paramType, argType, exprId)
            self.__activity.addError(exprId + '.error')
            argIds.append(exprId)
        self.__activity.sendSignal(signal, targetClass, targetId, argIds)

    def __processAssertStatement(self, stmt):
        """Apply the effect of an AssertStatement."""
        (type_, exprId) = self.__et.translate(stmt.api_getExpr())
        if type_ is not boolType:
            raise errors.InputError('invalid type in assertion ' + str(stmt))
        self.__activity.addError(exprId + '.error')
        self.__activity.addAssertion(exprId + '.value')

    def __processAssignment(self, expr):
        """Apply the effect of a SimpleAssignmentExpr."""
        lhsExpr = expr.api_getLhsExpr()
        rhsExpr = expr.api_getRhsExpr()
        if isinstance(lhsExpr, actlang.expr.PrimaryFieldAccessExpr):
            refExpr = lhsExpr.api_getReferenceExpr()
            (refType, refId) = self.__et.translate(refExpr)
            fieldName = lhsExpr.api_getFieldName()
        elif isinstance(lhsExpr, actlang.names.SimpleName):
            (refType, refId) = self.__activity.getThis()
            fieldName = lhsExpr.api_getParts()[0]
        elif isinstance(lhsExpr, actlang.names.QualifiedName):
            refExpr = lhsExpr.api_getQualifyingName()
            (refType, refId) = self.__et.translate(refExpr)
            fieldName = lhsExpr.api_getParts()[-1]
        else:
            raise errors.NotSupportedError('unsupported left-hand side ' +
                                           str(lhsExpr) + ' in assignment')
        if not refType.isReferenceType():
            raise errors.InputError('invalid reference in ' + str(lhsExpr))
        refClass = refType.getClass()
        try:
            field = refClass.getFieldForName(fieldName)
        except exceptions.LookupError:
            raise errors.InputError('unknown field ' + str(fieldName))
        self.__activity.addError(refId + '.error')
        self.__activity.addError(refId + '.isNull')
        if isinstance(rhsExpr, actlang.expr.InstanceCreationExpr):
            className = rhsExpr.api_getTypeName()
            args = rhsExpr.api_getArgList()
            #$$ TODO
            print 'skipping instance creation ' + str(expr)
        else:
            (rhsType, rhsId) = self.__et.translate(rhsExpr)
            if not rhsType.isSubtypeOf(field.getType()):
                raise errors.InputError('type mismatch in assignment ' +
                                        str(expr))
            exprId = self.__et.cast(field.getType(), rhsType, rhsId)
            self.__activity.addError(exprId + '.error')
            self.__activity.writeField(field, refId, exprId + '.value',
                                       _shallowStr(expr))


class _ExprTranslator:
    """A translator for expressions in an activity context."""
    
    def __init__(self, activity):
        """Construct an expression translator.

        Param activity: the Activity giving valuations of fields.
        """
        self.__activity = activity

    def cast(self, toType, fromType, fromExprId):
        """Generate a cast.

        The given types must be castable to each other.

        Param toType: the type to which the expression is cast.  Must
        not be the null type unless fromType is also the null type.

        Param fromType: type of the original expression.

        Param fromExprId: the original expression id.

        Return value: expression id giving the value in toType.
        """
        if toType is fromType:
            return fromExprId
        # The only case we need to handle so far is '(ClassType) null'.
        assert fromType is nullType and toType.isReferenceType()
        moduleName = toType.getClass().smvCastFromNullExprModule()
        exprId = self.__activity.addExpr(moduleName=moduleName,
                                         moduleArgs=[fromExprId],
                                         comment='() null',
                                         chooseKey=None)
        return exprId

    def translate(self, expr):
        """Translate a side-effect free expression.

        Param expr: the Expr to translate.

        Return value: (type_, exprId), where type_ is the Type of the
        expression and exprId is an expression identifier.
        """
        # Remove parentheses
        while isinstance(expr, actlang.expr.ParenExpr):
            expr = expr.api_getInnerExpr()
        if isinstance(expr, actlang.expr.ThisExpr):
            # A special case
            return self.__activity.getThis()
        if isinstance(expr, actlang.names.SimpleName):
            r = self.__translateSimpleName(expr)
        elif isinstance(expr, actlang.names.QualifiedName):
            r = self.__translateQualifiedName(expr)
        elif isinstance(expr, actlang.expr.IntLiteralExpr):
            r = self.__translateIntLiteral(expr)
        elif isinstance(expr, actlang.expr.BoolLiteralExpr):
            r = self.__translateBoolLiteral(expr)
        elif isinstance(expr, actlang.expr.NullExpr):
            r = self.__translateNull()
        elif isinstance(expr, actlang.expr.UnaryPlusExpr):
            r = self.__translateUnaryPlus(expr)
        elif isinstance(expr, actlang.expr.UnaryMinusExpr):
            r = self.__translateUnaryMinus(expr)
        elif isinstance(expr, actlang.expr.BitNotExpr):
            r = self.__translateBitNot(expr)
        elif isinstance(expr, actlang.expr.LogicalNotExpr):
            r = self.__translateLogicalNot(expr)
        elif isinstance(expr, actlang.expr.InfixExpr):
            r = self.__translateInfix(expr)
        elif isinstance(expr, actlang.expr.TernaryCondExpr):
            r = self.__translateTernaryCond(expr)
        elif isinstance(expr, actlang.expr.CondAndExpr):
            r = self.__translateCondAnd(expr)
        elif isinstance(expr, actlang.expr.CondOrExpr):
            r = self.__translateCondOr(expr)
        elif isinstance(expr, actlang.expr.PrimaryFieldAccessExpr):
            r = self.__translatePrimaryFieldAccess(expr)
        elif isinstance(expr, actlang.expr.NameMethodInvocationExpr):
            r = self.__translateNameMethodInvocation(expr)
        else:
            raise errors.NotSupportedError('unsupported expression ' +
                                           str(expr))
        (type_, moduleName, moduleArgs, chooseKey) = r
        exprId = self.__activity.addExpr(moduleName=moduleName,
                                         moduleArgs=moduleArgs,
                                         comment=_shallowStr(expr),
                                         chooseKey=chooseKey)
        return (type_, exprId)

    # The following functions take an expression of a suitable kind
    # and return a tuple (type_, moduleName, moduleArgs, chooseKey),
    # where type_ is the static Type of the expression, moduleName is
    # the module that realizes the expression, moduleArgs contains the
    # arguments as a list of strings, and chooseKey is the string used
    # to identify each invocation of one of the $$SMUML.choose*
    # methods.  If the expression is not a choose method invocation,
    # chooseKey is None.

    def __translateSimpleName(self, expr):
        # As long as there are no local variables or static fields, a
        # simple name 'x' is treated as 'this.x'.
        (refType, refId) = self.__activity.getThis()
        fieldName = expr.api_getParts()[0]
        return self.__genFieldAccess(refType, refId, fieldName)

    def __translateQualifiedName(self, expr):
        (refType, refId) = self.translate(expr.api_getQualifyingName())
        fieldName = expr.api_getParts()[-1]
        return self.__genFieldAccess(refType, refId, fieldName)

    def __translatePrimaryFieldAccess(self, expr):
        refExpr = expr.api_getReferenceExpr()
        fieldName = expr.api_getFieldName()
        (refType, refId) = self.translate(refExpr)
        return self.__genFieldAccess(refType, refId, fieldName)

    def __genFieldAccess(self, refType, refId, fieldName):
        """A general function for generating field access expressions."""
        if not refType.isReferenceType():
            raise errors.InputError('invalid reference for ' + str(fieldName))
        refClass = refType.getClass()
        try:
            field = refClass.getFieldForName(fieldName)
        except exceptions.LookupError:
            raise errors.InputError('unknown field ' + str(fieldName))
        valuation = self.__activity.readField(field, refId)
        return (field.getType(), field.smvAccessExprModule(),
                [refId, valuation], None)

    def __translateIntLiteral(self, expr):
        return (intType, 'IntLiteralExpr',
                [intType.smvValue(expr.api_getValue())], None)

    def __translateBoolLiteral(self, expr):
        return (boolType, 'BoolLiteralExpr',
                [boolType.smvValue(expr.api_getValue())], None)

    def __translateNull(self):
        return (nullType, 'NullExpr', [], None)

    def __translateUnaryPlus(self, expr):
        (opType, opId) = self.translate(expr.api_getOperand())
        if opType is not intType:
            raise errors.InputError('wrong operand type in ' + str(expr))
        return (intType, 'IntUnaryPlusExpr', [opId], None)

    def __translateUnaryMinus(self, expr):
        (opType, opId) = self.translate(expr.api_getOperand())
        if opType is not intType:
            raise errors.InputError('wrong operand type in ' + str(expr))
        return (intType, 'IntUnaryMinusExpr', [opId], None)

    def __translateBitNot(self, expr):
        (opType, opId) = self.translate(expr.api_getOperand())
        if opType is not intType:
            raise errors.InputError('wrong operand type in ' + str(expr))
        return (intType, 'IntNotExpr', [opId], None)

    def __translateLogicalNot(self, expr):
        (opType, opId) = self.translate(expr.api_getOperand())
        if opType is not boolType:
            raise errors.InputError('wrong operand type in ' + str(expr))
        return (boolType, 'BoolNotExpr', [opId], None)

    def __translateInfix(self, expr):
        (leftType, leftId) = self.translate(expr.api_getLeftOperand())
        (rightType, rightId) = self.translate(expr.api_getRightOperand())
        operator = expr.api_getOperator()
        if isinstance(operator, actlang.binaryop.Plus):
            if leftType is not intType or rightType is not intType:
                raise errors.InputError('wrong operand type in ' + str(expr))
            type_ = intType
            moduleName = 'IntPlusExpr'
        elif isinstance(operator, actlang.binaryop.NumericalOperator):
            if leftType is not intType or rightType is not intType:
                raise errors.InputError('wrong operand type in ' + str(expr))
            type_ = intType
            if isinstance(operator, actlang.binaryop.Multiply):
                moduleName = 'IntMultiplyExpr'
            elif isinstance(operator, actlang.binaryop.Divide):
                moduleName = 'IntDivideExpr'
            elif isinstance(operator, actlang.binaryop.Remainder):
                moduleName = 'IntRemainderExpr'
            elif isinstance(operator, actlang.binaryop.Minus):
                moduleName = 'IntMinusExpr'
            elif isinstance(operator, actlang.binaryop.ShiftLeft):
                moduleName = 'IntShiftLeftExpr'
            elif isinstance(operator, actlang.binaryop.SignedShiftRight):
                moduleName = 'IntSignedShiftRightExpr'
            elif isinstance(operator, actlang.binaryop.UnsignedShiftRight):
                moduleName = 'IntUnsignedShiftRightExpr'
            else:
                assert False  # No such operator
        elif isinstance(operator, actlang.binaryop.NumericalBooleanOperator):
            if leftType is intType and rightType is intType:
                type_ = intType
                prefix = 'Int'
            elif leftType is boolType and rightType is boolType:
                type_ = boolType
                prefix = 'Bool'
            else:
                raise errors.InputError('incompatible operand types in ' +
                                   str(expr))
            if isinstance(operator, actlang.binaryop.BitLogicalAnd):
                moduleName = prefix + 'AndExpr'
            elif isinstance(operator, actlang.binaryop.BitLogicalXor):
                moduleName = prefix + 'XorExpr'
            elif isinstance(operator, actlang.binaryop.BitLogicalOr):
                moduleName = prefix + 'OrExpr'
            else:
                assert False  # No such operator
        elif isinstance(operator, actlang.binaryop.NumericalComparison):
            if leftType is not intType or rightType is not intType:
                raise errors.InputError('wrong operand type in ' + str(expr))
            type_ = boolType
            if isinstance(operator, actlang.binaryop.LessThan):
                moduleName = 'IntLessThanExpr'
            elif isinstance(operator, actlang.binaryop.GreaterThan):
                moduleName = 'IntGreaterThanExpr'
            elif isinstance(operator, actlang.binaryop.LessOrEqual):
                moduleName = 'IntLessOrEqualExpr'
            elif isinstance(operator, actlang.binaryop.GreaterOrEqual):
                moduleName = 'IntGreaterOrEqualExpr'
            else:
                assert False  # No such operator
        elif isinstance(operator, actlang.binaryop.EqualityOperator):
            if not leftType.isCastableTo(rightType):
                raise errors.InputError('incompatible operand types in ' +
                                        str(expr))
            if leftType.isSubtypeOf(rightType):
                # Cast the left operand to the type of the right operand.
                leftId = self.cast(rightType, leftType, leftId)
                leftType = rightType
            elif rightType.isSubtypeOf(leftType):
                # Cast the right operand to the type of the left operand.
                rightId = self.cast(leftType, rightType, rightId)
                rightType = leftType
            else:
                # Should not happen as long as there are no interface types.
                assert False
            type_ = boolType
            if isinstance(operator, actlang.binaryop.Equal):
                moduleName = 'EqualExpr'
            elif isinstance(operator, actlang.binaryop.NotEqual):
                moduleName = 'NotEqualExpr'
            else:
                assert False  # No such operator
        else:
            assert False  # No such operator
        return (type_, moduleName, [leftId, rightId], None)

    def __translateTernaryCond(self, expr):
        (condType, condId) = self.translate(expr.api_getCondition())
        (trueType, trueId) = self.translate(expr.api_getTrueExpr())
        (falseType, falseId) = self.translate(expr.api_getFalseExpr())
        if condType is not boolType:
            raise errors.InputError('wrong condition type in ' + str(expr))
        if trueType.isSubtypeOf(falseType):
            # Cast the true operand to the type of the false operand.
            trueId = self.cast(falseType, trueType, trueId)
            trueType = falseType
        elif falseType.isSubtypeOf(trueType):
            # Cast the false operand to the type of the true operand.
            falseId = self.cast(trueType, falseType, falseId)
            falseType = trueType
        else:
            raise errors.InputError('incompatible operand types in ' +
                                    str(expr))
        if trueType is nullType or trueType.isReferenceType():
            moduleName = 'TernaryCondReferenceExpr'
        else:
            moduleName = 'TernaryCondPrimitiveExpr'
        return (trueType, moduleName, [condId, trueId, falseId], None)

    def __translateCondAnd(self, expr):
        (leftType, leftId) = self.translate(expr.api_getLeftOperand())
        (rightType, rightId) = self.translate(expr.api_getRightOperand())
        if leftType is not boolType or rightType is not boolType:
            raise errors.InputError('wrong operand type in ' + str(expr))
        return (boolType, 'CondAndExpr', [leftId, rightId], None)

    def __translateCondOr(self, expr):
        (leftType, leftId) = self.translate(expr.api_getLeftOperand())
        (rightType, rightId) = self.translate(expr.api_getRightOperand())
        if leftType is not boolType or rightType is not boolType:
            raise errors.InputError('wrong operand type in ' + str(expr))
        return (boolType, 'CondOrExpr', [leftId, rightId], None)

    def __translateNameMethodInvocation(self, expr):
        nameParts = expr.api_getInvocationName().api_getParts()
        args = expr.api_getArgList()
        if (nameParts == ['$$SMUML', 'choose_boolean'] and len(args) >= 3 and
            isinstance(args[0], actlang.expr.StringLiteralExpr) and
            isinstance(args[2], actlang.expr.IntLiteralExpr)):
            return self.__translateChooseBool(args[0].api_getValue(),
                                              str(args[2].api_getValue()))
        elif (nameParts == ['$$SMUML', 'choose_int'] and len(args) >= 3 and
              isinstance(args[0], actlang.expr.StringLiteralExpr) and
              isinstance(args[2], actlang.expr.IntLiteralExpr)):
            return self.__translateChooseInt(args[0].api_getValue(),
                                             str(args[2].api_getValue()))
        else:
            raise errors.NotSupportedError('unsupported method invocation ' +
                                           str(expr))

    def __translateChooseBool(self, valueString, chooseKey):
        """Translate a nondeterministic choose method for boolean type.

        Param valueString: a string containing the values among which
        to choose, as a comma-separated list.

        Param chooseKey: the string used as a key to identify the
        expression.

        Return value: (type_, moduleName, moduleArgs, isChoose).
        """
        has = {False: False, True: False}
        for v in [s.strip() for s in valueString.split(',')]:
            if v == 'false':
                has[False] = True
            elif v == 'true':
                has[True] = True
            else:
                raise errors.InputError('invalid argument for ' +
                                        'choose_boolean: ' + valueString)
        if has[False] and has[True]:
            return (boolType, 'ChooseBoolExpr', [], chooseKey)
        elif has[False]:
            return (boolType, 'BoolLiteralExpr', [boolType.smvValue(False)],
                    chooseKey)
        elif has[True]:
            return (boolType, 'BoolLiteralExpr', [boolType.smvValue(True)],
                    chooseKey)
        else:
            assert False

    def __translateChooseInt(self, valueString, chooseKey):
        """Translate a nondeterministic choose method for int type.

        Param valueString: a string containing the values among which
        to choose, as a comma-separated list.

        Param chooseKey: the string used as a key to identify the
        expression.

        Return value: (type_, moduleName, moduleArgs, isChoose).
        """
        args = []
        for v in [s.strip() for s in valueString.split(',')]:
            try:
                args.append(int(v))
            except exceptions.ValueError:
                raise errors.InputError('invalid argument for ' +
                                        'choose_int: ' + valueString)
        global _chooseIntMax
        _chooseIntMax = max(_chooseIntMax, len(args))
        return (intType, 'ChooseIntExpr#' + str(len(args)),
                [intType.smvValue(arg) for arg in args], chooseKey)


def codeCommon():
    """Return NuSMV code for common expressions."""
    def smvIsNegative(intExpr):
        """Return a NuSMV expression that is TRUE if an int is negative."""
        msb = str(options.opt.int_bits-1)
        return ('bool((' + intExpr + ')[' + msb + ':' + msb + '])')
    c = ['',
         '',
         '-'*70,
         '-- Common expressions',
         '-'*70]
    c += ['',
          'MODULE IntLiteralExpr(constant)',
          '  DEFINE',
          '    value := constant;',
          '    error := FALSE;',
          '',
          'MODULE BoolLiteralExpr(constant)',
          '  DEFINE',
          '    value := constant;',
          '    error := FALSE;',
          '',
          'MODULE NullExpr()',
          '  DEFINE',
          '    value := 0;  -- irrelevant',
          '    error := FALSE;',
          '    isNull := TRUE;',
          '',
          'MODULE ThisExpr(instance)',
          '  DEFINE',
          '    value := instance;',
          '    error := FALSE;',
          '    isNull := FALSE;',
          '',
          'MODULE IntUnaryPlusExpr(operand)',
          '  DEFINE',
          '    value := operand.value;',
          '    error := operand.error;',
          '',
          'MODULE IntUnaryMinusExpr(operand)',
          '  DEFINE',
          '    value := ' + intType.smvValue(0) + ' - operand.value;',
          '    error := operand.error;',
          '',
          'MODULE IntNotExpr(operand)',
          '  DEFINE',
          '    value := !operand.value;',
          '    error := operand.error;',
          '',
          'MODULE BoolNotExpr(operand)',
          '  DEFINE',
          '    value := !operand.value;',
          '    error := operand.error;',
          '',
          'MODULE IntPlusExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value := leftOp.value + rightOp.value;',
          '    error := leftOp.error | rightOp.error;',
          '',
          'MODULE IntMultiplyExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value := leftOp.value * rightOp.value;',
          '    error := leftOp.error | rightOp.error;',
          '',
          'MODULE IntDivideExpr(leftOp, rightOp)',
          '  DEFINE',
          '    leftNegative := ' + smvIsNegative('leftOp.value') + ';',
          '    rightNegative := ' + smvIsNegative('rightOp.value') + ';',
          '    leftAbs :=',
          '      case',
          '        leftNegative : ' + intType.smvValue(0) + ' - leftOp.value;',
          '        1 : leftOp.value;',
          '      esac;',
          '    rightAbs :=',
          '      case',
          '        rightNegative : ' +
          intType.smvValue(0) + ' - rightOp.value;',
          '        1 : rightOp.value;',
          '      esac;',
          '    quotient :=',
          '      case',
          '        rightAbs = ' + intType.smvValue(0) + ' : ' +
          intType.smvValue(0) + ';  -- error',
          '        1 : leftAbs / rightAbs;',
          '      esac;',
          '    value :=',
          '      case',
          '        (leftNegative xor rightNegative) : ' +
          intType.smvValue(0) + ' - quotient;',
          '        1 : quotient;',
          '      esac;',
          '    error := leftOp.error | rightOp.error | (rightAbs = ' +
          intType.smvValue(0) + ');',
          '',
          'MODULE IntRemainderExpr(leftOp, rightOp)',
          '  DEFINE',
          '    leftNegative := ' + smvIsNegative('leftOp.value') + ';',
          '    rightNegative := ' + smvIsNegative('rightOp.value') + ';',
          '    leftAbs :=',
          '      case',
          '        leftNegative : ' + intType.smvValue(0) + ' - leftOp.value;',
          '        1 : leftOp.value;',
          '      esac;',
          '    rightAbs :=',
          '      case',
          '        rightNegative : ' +
          intType.smvValue(0) + ' - rightOp.value;',
          '        1 : rightOp.value;',
          '      esac;',
          '    remainder :=',
          '      case',
          '        rightAbs = ' + intType.smvValue(0) + ' : ' +
          intType.smvValue(0) + ';  -- error',
          '        1 : leftAbs mod rightAbs;',
          '      esac;',
          '    value :=',
          '      case',
          '        leftNegative : ' + intType.smvValue(0) + ' - remainder;',
          '        1 : remainder;',
          '      esac;',
          '    error := leftOp.error | rightOp.error | (rightAbs = ' +
          intType.smvValue(0) + ');',
          '',
          'MODULE IntMinusExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value := leftOp.value - rightOp.value;',
          '    error := leftOp.error | rightOp.error;',
          '',
          'MODULE IntShiftLeftExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value :=',
          '      case']
    for i in xrange(options.opt.int_bits):
        c += ['        rightOp.value = ' + intType.smvValue(i) +
              ' : leftOp.value << ' + str(i) + ';']
    c += ['        1 : ' + intType.smvValue(0) + ';',
          '      esac;',
          '    error := leftOp.error | rightOp.error | ' +
          smvIsNegative('rightOp.value') + ';',
          '',
          'MODULE IntSignedShiftRightExpr(leftOp, rightOp)',
          '  DEFINE',
          '    sign :=',
          '      case',
          '        ' + smvIsNegative('leftOp.value') + ' : ' +
          intType.smvValue(-1) + ';',
          '        1 : ' + intType.smvValue(0) + ';',
          '      esac;',
          '    value :=',
          '      case',
          '        rightOp.value = ' + intType.smvValue(0) +
          ' : leftOp.value;']
    for i in xrange(1, options.opt.int_bits-1):
        c += ['        rightOp.value = ' + intType.smvValue(i) +
              ' : (leftOp.value >> ' + str(i) + ') | (sign << ' +
              str(options.opt.int_bits - i) + ');']
    c += ['        1 : sign;',
          '      esac;',
          '    error := leftOp.error | rightOp.error | ' +
          smvIsNegative('rightOp.value') + ';',
          '',
         'MODULE IntUnsignedShiftRightExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value :=',
          '      case']
    for i in xrange(options.opt.int_bits):
        c += ['        rightOp.value = ' + intType.smvValue(i) +
              ' : leftOp.value >> ' + str(i) + ';']
    c += ['        1 : ' + intType.smvValue(0) + ';',
          '      esac;',
          '    error := leftOp.error | rightOp.error | ' +
          smvIsNegative('rightOp.value') + ';',
          '',
          'MODULE IntAndExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value := leftOp.value & rightOp.value;',
          '    error := leftOp.error | rightOp.error;',
          '',
          'MODULE IntXorExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value := leftOp.value xor rightOp.value;',
          '    error := leftOp.error | rightOp.error;',
          '',
          'MODULE IntOrExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value := leftOp.value | rightOp.value;',
          '    error := leftOp.error | rightOp.error;',
          '',
          'MODULE BoolAndExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value := leftOp.value & rightOp.value;',
          '    error := leftOp.error | rightOp.error;',
          '',
          'MODULE BoolXorExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value := leftOp.value xor rightOp.value;',
          '    error := leftOp.error | rightOp.error;',
          '',
          'MODULE BoolOrExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value := leftOp.value | rightOp.value;',
          '    error := leftOp.error | rightOp.error;']
    # Perform signed compare by adding INT_MIN to each operand and
    # then performing unsigned compare.
    left = ('(leftOp.value xor ' +
            intType.smvValue((-1) << (options.opt.int_bits-1)) + ')')
    right = ('(rightOp.value xor ' +
            intType.smvValue((-1) << (options.opt.int_bits-1)) + ')')
    c += ['',
          'MODULE IntLessThanExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value := (' + left + ' < ' + right + ');',
          '    error := leftOp.error | rightOp.error;',
          '',
          'MODULE IntGreaterThanExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value := (' + left + ' > ' + right + ');',
          '    error := leftOp.error | rightOp.error;',
          '',
          'MODULE IntLessOrEqualExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value := (' + left + ' <= ' + right + ');',
          '    error := leftOp.error | rightOp.error;',
          '',
          'MODULE IntGreaterOrEqualExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value := (' + left + ' >= ' + right + ');',
          '    error := leftOp.error | rightOp.error;',
          '',
          'MODULE EqualExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value := (leftOp.value = rightOp.value);',
          '    error := leftOp.error | rightOp.error;',
          '',
          'MODULE NotEqualExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value := (leftOp.value != rightOp.value);',
          '    error := leftOp.error | rightOp.error;']
    c += ['',
          'MODULE TernaryCondPrimitiveExpr(condition, trueOp, falseOp)',
          '  DEFINE',
          '    value :=',
          '      case',
          '        condition.value : trueOp.value;',
          '        1 : falseOp.value;',
          '      esac;',
          '    error := condition.error |',
          '      case',
          '        condition.value : trueOp.error;',
          '        1 : falseOp.error;',
          '      esac;',
          '',
          'MODULE TernaryCondReferenceExpr(condition, trueOp, falseOp)',
          '  DEFINE',
          '    value :=',
          '      case',
          '        condition.value : trueOp.value;',
          '        1 : falseOp.value;',
          '      esac;',
          '    error := condition.error |',
          '      case',
          '        condition.value : trueOp.error;',
          '        1 : falseOp.error;',
          '      esac;',
          '    isNull :=',
          '      case',
          '        condition.value : trueOp.isNull;',
          '        1 : falseOp.isNull;',
          '      esac;',
          '',
          'MODULE CondAndExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value := leftOp.value & rightOp.value;',
          '    error := leftOp.error | (leftOp.value & rightOp.error);',
          '',
          'MODULE CondOrExpr(leftOp, rightOp)',
          '  DEFINE',
          '    value := leftOp.value | rightOp.value;',
          '    error := leftOp.error | (!leftOp.value & rightOp.error);']
    c += ['',
          'MODULE ChooseBoolExpr()',
          '  IVAR',
          '    value : boolean;',
          '  DEFINE',
          '    error := FALSE;']
    n = _chooseIntMax
    while n > 0:
        args = ['arg' + str(i) for i in xrange(n)]
        c += ['',
              'MODULE ChooseIntExpr#' + str(n) + '(' + ', '.join(args) + ')',
              '  IVAR',
              '    value : ' + intType.smvTypeDefinition() + ';',
              '  DEFINE',
              '    error := FALSE;']
        c += codeTrans(smvOr(['(value = ' + arg + ')' for arg in args]))
        n -= 1
    return c
