#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# activities.py
#
# 2007-10-02  Jori Dubrovin:  New semantics for choose exprs with common keys
# 2007-07-17  Jori Dubrovin:  Fixed addExpr invocation
# 2007-06-20  Jori Dubrovin:  New choose expression mapping
# 2007-03-02  Jori Dubrovin:  Adding choose to element map
# 2006-12-18  Jori Dubrovin:  Adding nondeterministic choose
# 2006-11-25  Jori Dubrovin:  Implementing events
# 2006-11-15  Jori Dubrovin:  Initial version
#

"""Determining the side effects of transitions."""

import errors

class ActivityContext:
    """An activity context.

    An activity context stores all the expressions evaluated in the
    class and all field valuations resulting from assignments.

    There is a one-to-one correspondence between ActiveClasses and
    ActivityContexts.
    """

    def __init__(self, class_):
        """Construct an activity context.

        Param class_: the associated ActiveClass.
        """
        self.__class = class_
        self.__exprs = []       # list of expression definition tuples
        self.__valuations = []  # list of valuation definition tuples
        self.__thisExprId = self.addExpr('ThisExpr', ['this'], 'this', False)

    def getClass(self):
        """Return the ActiveClass associated with self."""
        return self.__class

    def smvThisExprId(self):
        """Return the expression id for the 'this' expression."""
        return self.__thisExprId

    def addExpr(self, moduleName, moduleArgs, comment, unique):
        """Add an expression to self.

        Param moduleName: name of the module that realizes the expression.

        Param moduleArgs: module arguments as a list of strings.
        These are NuSMV expressions valid in the activity context.

        Param comment: a free-form explanation for the expression
        instance.

        Param unique: False iff it is ok to reuse a previous
        expression with the same module name and arguments.

        Return value: an expression id string.
        """
        if not unique:
            for (exprId, name, args, c) in self.__exprs:
                # Try to find an already added expression with the
                # same module and arguments
                if name == moduleName and args == moduleArgs:
                    return exprId
        exprId = 'expr' + str(len(self.__exprs) + 1)
        self.__exprs += [(exprId, moduleName, moduleArgs, comment)]
        return exprId

    def addValuation(self, field, reference, newValue,
                     oldValuation, comment):
        """Add a valuation to self.

        A valuation contains the values of all instances of a field
        after a write operation to the field.  The field may be owned
        by any class.  The returned valuation is defined as a variable
        in the module for the class associated with self.  The
        returned valuation may not be a new one, if addValuation has
        already been called with equivalent arguments.

        Param field: the Field for which the valuation is obtained.

        Param reference: a NuSMV expression that gives the object
        reference used in the write operation.

        Param newValue: a NuSMV expression that gives the value
        written in the write operation.

        Param oldValuation: the valuation id of the field before the
        write operation.

        Param comment: a free-form comment string explaining the
        valuation variable.

        Return value: a valuation identifier as a string.
        """
        for (v, f, r, n, o, c) in self.__valuations:
            if (f is field and r == reference and
                n == newValue and o == oldValuation):
                # Avoid defining a new identical valuation.
                return v
        v = 'valuation' + str(len(self.__valuations) + 1)
        self.__valuations += [(v, field, reference, newValue,
                               oldValuation, comment)]
        return v

    def code(self):
        """Generate code for the activity context.

        The code is placed in the module for the associated class.

        Return value: NuSMV code as a list of strings.
        """
        return self.__codeExprs() + self.__codeValuations()

    def __codeExprs(self):
        """Generate code for all expressions in the activity context.

        Return value: NuSMV code as a list of strings.
        """
        c = ['  -- Side-effect free Jumbala expressions.',
             '  VAR']
        for (exprId, moduleName, moduleArgs, comment) in self.__exprs:
            c += ['    ' + exprId + ' : ' + moduleName +
                  '(' + ', '.join(moduleArgs) + ');  -- ' + comment]
        return c

    def __codeValuations(self):
        """Generate code for defining the valuations in self.

        Return value: NuSMV code as a list of strings.
        """
        c = ['  -- Valuations',
             '  VAR']
        for (valuationId, field, reference, newValue,
             oldValuationId, comment) in self.__valuations:
            c += ['    ' + valuationId + ' : ' + field.smvValuationModule() +
                  '(' + reference + ', ' + newValue + ', ' + oldValuationId +
                  ');  -- ' + comment]
        return c


class Activity:
    """An activity that reads and/or changes the global configuration.

    An activity consists of read accesses from fields, write accesses
    to fields, and signal transmissions to object.  Additionally, an
    activity may result in a Jumbala run-time error or assertion error
    when executed.  Changes to the active state of a state machine are
    not part of an activity.

    If an activity is associated with a transition, then it consists
    of information about fields read by the guard, fields updated when
    reading signal parameters, fields accessed by the transition
    effect, signals sent by the transition effect, assertions in the
    effect, and possible Jumbala run-time errors caused by evaluating
    the guard or effect.  In this case, one Activity object
    corresponds to one Transition, not one transition instance.
    """

    class ChooseInstance:
        pass

    def __init__(self, context, isConstant=False):
        """Create an empty activity.

        Param context: the ActivityContext in which the activity occurs.

        Param isConstant: True if this activity is not allowed to
        access 'this' or read any fields.
        """
        self.__context = context
        self.__isConstant = isConstant
        self.__reads = []
        self.__readForField = {}   # map Field to Read
        self.__writes = []
        self.__writeForField = {}  # map Field to Write
        self.__sends = []
        self.__sendForClass = {}   # map target ActiveClass to Send
        self.__errors = []         # list of NuSMV expressions
        self.__assertions = []     # list of NuSMV expressions
        self.__chooseInstance = {} # map choose key string to ChooseInstance

    def getFieldReads(self):
        """Return the list of Reads from fields."""
        return self.__reads[:]

    def getFieldWrites(self):
        """Return the list of Writes to fields."""
        return self.__writes[:]

    def getSends(self):
        """Return the list of signal Sends."""
        return self.__sends[:]

    def getErrors(self):
        """Return Jumbala run-time errors associated with self.

        Assertion errors are not included in the list.

        Return value: a list of NuSMV expressions.  If any of them
        evaluates to TRUE, then execution of self results in a
        run-time error (unless an assertion is violated before that).
        """
        return self.__errors[:]

    def getAssertions(self):
        """Return assertions associated with self.

        Return value: a list of NuSMV expressions.  If any of them
        evaluates to FALSE, then execution of self results in an
        assertion violation (unless a run-time error occurs before
        that).
        """
        return self.__assertions[:]

    def getThis(self):
        """Return the type and expression for 'this'.

        Return value: (type_, exprId), where type_ is the Type of the
        ActiveClass corresponding to the activity context, and exprId
        is the expression identifier giving the 'this' reference.
        """
        if self.__isConstant:
            raise errors.InputError("cannot access 'this' from the context")
        return (self.__context.getClass().getType(),
                self.__context.smvThisExprId())

    def getSignalForName(self, signalName):
        """Return the Signal with the given name.

        If there are no such Signals, raise a LookupError.  If there
        are several, raise an InputError.
        """
        return self.__context.getClass().getSignalForName(signalName)

    def getChooseExprIds(self):
        """Return the choose expressions used in this activity.

        Return value: a dictionary from choose key strings to
        expression identifier strings.
        """
        return dict([(k, self.__chooseInstance[k].exprId) for
                     k in self.__chooseInstance])

    def clearChooseExprIds(self):
        """Clear the choose expressions added so far in this activity."""
        self.__chooseInstance = {}

    def readField(self, field, refExprId):
        """Add a read from a field.

        Param field: the Field that is read.

        Param refExprId: an expression identifier (valid in the
        activity context) giving the reference expression.

        Return value: a valuation identifier giving the values of the
        field.
        """
        if self.__isConstant:
            raise errors.InputError("cannot access fields from the context")
        if field in self.__readForField:
            read = self.__readForField[field]
        else:
            read = Read(self.__context, field)
            self.__reads.append(read)
            self.__readForField[field] = read
        read.addRead(refExprId)
        if field in self.__writeForField:
            return self.__writeForField[field].smvValuation()
        else:
            return 'global'

    def writeField(self, field, refExprId, newValue, comment):
        """Add a write to a field.

        Param field: the Field that is written.

        Param refExprId: an expression identifier (in the activity
        context) giving the reference expression.

        Param newValue: a NuSMV expression (valid in the activity
        context) giving the written value.

        Param comment: a free-form comment string.
        """
        if field in self.__writeForField:
            write = self.__writeForField[field]
        else:
            write = Write(self.__context, field)
            self.__writes.append(write)
            self.__writeForField[field] = write
        write.addWrite(refExprId, newValue, comment)

    def sendSignal(self, signal, targetClass, targetExprId, paramExprIds):
        """Add a signal transmission.

        Raises an InputError if a signal to targetClass has already
        been added.

        Param signal: the Signal being sent.

        Param targetClass: the ActiveClass of the target.

        Param targetExprId: an expression identifier (in the activity
        context) giving the reference expression for the signal
        target.

        Param paramExprIds: a list of expression identifiers giving
        the argument values for signal parameters.  The number and
        types must match.
        """
        if targetClass in self.__sendForClass:
            raise errors.InputError('multiple sends to a class')
        send = Send(self.__context, signal,
                    targetClass, targetExprId, paramExprIds)
        self.__sends.append(send)
        self.__sendForClass[targetClass] = send
        targetClass.getEventQueue().addReceivedSignal(signal)

    def addExpr(self, moduleName, moduleArgs, comment, chooseKey):
        """Add an expression to the activity context.

        Param moduleName: name of the module that realizes the expression.

        Param moduleArgs: module arguments (NuSMV expressions valid in
        the activity context) as a list of strings.

        Param comment: explanation for the expression instance.

        Param chooseKey: a string identifying a special choose method
        invocation.  None iff the added expression is not a choose
        method invocation.  In this case, it is ok to reuse a previous
        expression with the same module name and arguments.

        Return value: an expression id string.
        """
        if chooseKey is None:
            return self.__context.addExpr(moduleName, moduleArgs,
                                          comment, unique=False)
        if chooseKey in self.__chooseInstance:
            ci = self.__chooseInstance[chooseKey]
            if moduleName != ci.moduleName or moduleArgs != ci.moduleArgs:
                # A previous choose method invocation has the same key
                # but different type or parameters.
                raise errors.InputError(
                    'inconsistent choose expressions')
            exprId = ci.exprId
        else:
            exprId = self.__context.addExpr(moduleName, moduleArgs,
                                            comment, unique=True)
            ci = Activity.ChooseInstance()
            ci.exprId = exprId
            ci.moduleName = moduleName
            ci.moduleArgs = moduleArgs
            self.__chooseInstance[chooseKey] = ci
        return exprId

    def addError(self, error):
        """Add a Jumbala run-time error.

        Param error: a NuSMV expression valid in the activity context
        that evaluates to TRUE in case of an error.
        """
        self.__errors.append(error)

    def addAssertion(self, assertion):
        """Add an assertion.

        Param assertion: a NuSMV expression valid in the activity
        context that evaluates to FALSE in case of an assertion
        violation.
        """
        self.__assertions.append(assertion)


class Access:
    """Base class for reading or writing a field, possibly several times."""

    def __init__(self, context, field):
        """Initialize the access.

        Param context: the ActivityContext in which the execution occurs.

        Param field: the Field being read or written.
        """
        self._context = context
        self._field = field
        self.__refExprIds = []

    def getField(self):
        """Return the Field being read or written."""
        return self._field

    def isLocal(self):
        """Return True if all accesses occur through the reference this."""
        if not self.__refExprIds:
            return True
        if len(self.__refExprIds) > 1:
            return False
        return (self.__refExprIds[0] == self._context.smvThisExprId())

    def _add(self, refExprId):
        """Add an access through a given reference expression identifier."""
        if refExprId not in self.__refExprIds:
            self.__refExprIds.append(refExprId)

    def smvRefExprIds(self):
        """Return the expression identifiers used as references.

        The expressions are valid in the context of the acting class.

        Return value: a list of expression identifiers as strings.
        """
        return self.__refExprIds[:]


class Read(Access):
    """The act of reading the value of a field, possibly several times."""

    def __init__(self, context, field):
        """Create an empty read access object.

        Param context: the ActivityContext in which the execution occurs.

        Param field: the Field being read.
        """
        Access.__init__(self, context, field)

    def addRead(self, refExprId):
        """Add a read through a given reference expression identifier."""
        Access._add(self, refExprId)


class Write(Access):
    """The act of writing the value of a field, possibly several times."""

    def __init__(self, context, field):
        """Create an empty write access object.

        Param context: the ActivityContext in which the execution occurs.

        Param field: the Field being written.
        """
        Access.__init__(self, context, field)
        self.__valuation = 'global'

    def addWrite(self, refExprId, newValue, comment):
        """Add a write to the field.

        Param refExprId: an expression identifier through which the
        writing occurs.

        Param newValue: a NuSMV expression giving the new value.

        Param comment: a free-form comment string
        """
        Access._add(self, refExprId)
        v = self._context.addValuation(self._field, refExprId + '.value',
                                       newValue, self.__valuation, comment)
        self.__valuation = v

    def smvValuation(self):
        """Return the valuation id giving the resulting valuation.

        The valuation contains values for all the instances of the
        field after they have been written.
        """
        return self.__valuation


class Send:
    """A signal transmission."""

    def __init__(self, context, signal,
                 targetClass, targetExprId, paramExprIds):
        """Create an initialized signal transmission.

        Param context: the ActivityContext in which the execution occurs.

        Param signal: the Signal being sent.

        Param targetClass: the ActiveClass of the target.

        Param targetExprId: an expression identifier (in the activity
        context) giving the reference expression for the signal
        target.

        Param paramExprIds: a list of expression identifiers giving
        the argument values.
        """
        self.__context = context
        self.__signal = signal
        self.__targetClass = targetClass
        self.__targetExprId = targetExprId
        self.__paramExprIds = paramExprIds

    def getTargetClass(self):
        """Return the ActiveClass of the target object."""
        return self.__targetClass

    def getSignal(self):
        """Return the Signal sent."""
        return self.__signal

    def isLocal(self):
        """Return True if the sending occurs through the reference this."""
        return (self.__targetExprId == self.__context.smvThisExprId())

    def smvTargetExprId(self):
        """Return the expression identifier used as a reference.

        Return value: an expression identifier (in the activity
        context) as a string.
        """
        return self.__targetExprId

    def smvParamExprIds(self):
        """Return the expression identifiers for parameter values.

        Return value: a list of expression identifiers (in the
        activity context) as strings.
        """
        return self.__paramExprIds[:]
