#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# classes.py
#
# 2007-07-17  Jori Dubrovin:  Fixed attribute initialization
# 2007-06-20  Jori Dubrovin:  New choose expression mapping
# 2007-03-08  Jori Dubrovin:  Adding states to element map
# 2007-03-02  Jori Dubrovin:  Adding choose to element map
# 2006-12-18  Jori Dubrovin:  Fixed Field.__owner
# 2006-12-05  Jori Dubrovin:  Fixed cast error handling
# 2006-08-31  Jori Dubrovin:  Added encodeStateExclusion
# 2006-08-16  Jori Dubrovin:  Added checks for run-time errors
# 2006-07-28  Jori Dubrovin:  Initial version
#

"""Translation of UML classes to NuSMV."""

import exceptions

import actions
import activities
import errors
import events
import statemachines
import datatypes
import options
from coralutils import uml, umlSort
from smvutils import smvAnd, smvOr, smvWord, smvWordType
from smvutils import codeDefine, codeTrans, codeAtMostOne
from smvutils import generateId


class ActiveClass:
    """Representation of an active UML class."""

    def __init__(self, umlClass, id_, model):
        """Construct an active class.

        Param umlClass: the underlying uml.Class.

        Param id_: an identification string.

        Param model: the Model owning the class.
        """
        assert umlClass.oclIsKindOf(uml.Class)
        self.__umlClass = umlClass
        self.__id = id_
        self.__model = model
        self.__fields = []                    # all Fields of the class
        self.__fieldForUmlAttribute = {}      # map uml.Attribute to Field
        self.__fieldForUmlAssociationEnd = {} # map uml.AssociationEnd to Field
        self.__objects = []                   # all ActiveObjects of the class
        self.__objectForUml = {}              # map uml.Object to ActiveObject
        if (len(umlClass.behavior) != 1 or
            not umlClass.behavior[0].oclIsKindOf(uml.StateMachine)):
            raise errors.NotSupportedError('class ' + str(umlClass) +
                                           'does not have exactly one ' +
                                           'state machine')
        self.__machine = statemachines.StateMachine(umlClass.behavior[0], self)
        self.__queue = events.EventQueue(self)
        self.__context = activities.ActivityContext(self)
        self.__type = datatypes.ReferenceType(self)

    def getStateMachine(self):
        """Return the StateMachine of self."""
        return self.__machine

    def getEventQueue(self):
        """Return the EventQueue of self."""
        return self.__queue

    def getActivityContext(self):
        """Return the ActivityContext corresponding to self."""
        return self.__context

    def getType(self):
        """Return the ReferenceType corresponding to self."""
        return self.__type

    def getFields(self):
        """Return all Fields."""
        return self.__fields[:]

    def getObjects(self):
        """Return all ActiveObjects of this class."""
        return self.__objects[:]

    def getObjectForUml(self, umlObject):
        """Return the ActiveObject corresponding to a given uml.Object.

        If there is no such object of this class, raise a LookupError.
        """
        return self.__objectForUml[umlObject]

    def getName(self):
        """Return the name of the class as a string."""
        return str(self.__umlClass.name)

    def getFieldForName(self, fieldName):
        """Return the Field with the given name.

        If there are no such Fields, raise a LookupError.  If there
        are several, raise an InputError.
        """
        fields = [f for f in self.__fields if f.getName() == fieldName]
        if len(fields) > 1:
            raise errors.InputError("ambiguous field '" + fieldName +
                                    "' in " + str(self.__umlClass))
        return fields[0]

    def getFieldForUmlAttribute(self, umlAttribute):
        """Return the Field corresponding to a uml.Attribute.

        If there are no such Fields, raise a LookupError.
        """
        return self.__fieldForUmlAttribute[umlAttribute]

    def getFieldForUmlAssociationEnd(self, umlAssociationEnd):
        """Return the Field corresponding to a uml.AssociationEnd.

        If there are no such Fields, raise a LookupError.
        """
        return self.__fieldForUmlAssociationEnd[umlAssociationEnd]

    def getSignalForUml(self, umlSignal):
        """Return the global Signal for a given uml.Signal.

        If there are no such Signals, raise a LookupError.
        """
        return self.__model.getSignalForUml(umlSignal)

    def getSignalForName(self, signalName):
        """Return the Signal with the given name.

        If there are no such Signals, raise a LookupError.  If there
        are several, raise an InputError.
        """
        return self.__model.getSignalForName(signalName)

    def addObject(self, umlObject):
        """Add a new ActiveObject to self.

        Param umlObject: the underlying uml.Object, or None if the
        object is not initially alive.
        """
        i = len(self.__objects) + 1
        id_ = 'i' + str(i) + '#' + self.__id
        obj = ActiveObject(umlObject, id_, i, self)
        self.__objects.append(obj)
        if umlObject:
            assert not umlObject in self.__objectForUml
            self.__objectForUml[umlObject] = obj

    def processObjects(self):
        """Add the remaining objects to self."""
        # Determine the minumum number of instances of self according
        # to global options.
        n = options.opt.instances
        umlName = str(self.__umlClass.name)
        for s in options.opt.specific_instances:
            (name, count) = s.split(':')
            if name == umlName:
                n = int(count)
        # Add uninitialized objects until the minimum amount is
        # reached.
        while len(self.__objects) < n:
            self.addObject(None)

    def processFields(self):
        """Add all Fields to self.

        All classes must have been added to the model already.
        """
        self.__addAttributes()
        self.__addAssociations()

    def __addAttributes(self):
        """Find all attributes and add them as Fields."""
        for umlFeature in umlSort(self.__umlClass.feature):
            if umlFeature.oclIsKindOf(uml.Attribute):
                id_ = generateId('a', umlFeature.name,
                                 [f.smvId() for f in self.__fields])
                f = AttributeField(umlFeature, id_, self)
                self.__fields.append(f)
                assert not umlFeature in self.__fieldForUmlAttribute
                self.__fieldForUmlAttribute[umlFeature] = f

    def __addAssociations(self):
        """Find all navigable associations and add them as Fields.

        All classes must have been added to the model already.
        """
        for umlThisEnd in umlSort(self.__umlClass.association):
            umlAssociation = umlThisEnd.__parent__()
            ends = [e for e in umlSort(umlAssociation.connection) if
                    e is not umlThisEnd and e.isNavigable]
            for umlEnd in ends:
                assert umlEnd.oclIsKindOf(uml.AssociationEnd)
                try:
                    endClass = self.__model.getClassForUml(umlEnd.participant)
                except exceptions.LookupError:
                    raise errors.InputError('dangling association end ' +
                                            str(umlEnd))
                id_ = generateId('l', umlEnd.name,
                                 [f.smvId() for f in self.__fields])
                f = AssociationField(umlEnd, endClass.getType(), id_, self)
                self.__fields.append(f)
                assert not umlEnd in self.__fieldForUmlAssociationEnd
                self.__fieldForUmlAssociationEnd[umlEnd] = f

    def processInitialConfiguration(self):
        """Determine the field values for initially alive objects."""
        for o in self.__objects:
            o.processInitialConfiguration()

    def processActivities(self):
        """Generate the global effects of the transitions in this class."""
        for t in self.__machine.getTransitions():
            t.processActivity()

    def elementMap(self):
        """Generate a mapping of the class to UML.

        Return value: mapping code as a list of strings.
        """
        classXmiId = str(self.__umlClass.__XMIid__())
        c = []
        for o in self.__objects:
            line = '--$Map Object ' + o.smvVariable() + ' ' + classXmiId
            x = o.getXmiId()
            if x:
                line += ' ' + x
            c += [line]
        for s in self.__machine.getVertices():
            c += ['--$Map StateVertex ' + classXmiId + ' ' +
                  s.getXmiId() + ' ' + s.smvId()]
        for t in self.__machine.getTransitions():
            c += ['--$Map Trans ' + classXmiId + ' ' +
                  t.getXmiId() + ' ' + t.smvId()]
            exprIds = t.getGuardChooseExprIds()
            chooseKeys = exprIds.keys()
            chooseKeys.sort()
            for k in chooseKeys:
                c += ['--$Map TransGuardChoose ' + classXmiId + ' ' +
                      t.getXmiId() + ' ' + exprIds[k] + '.value ' + k]
            exprIds = t.getEffectChooseExprIds()
            chooseKeys = exprIds.keys()
            chooseKeys.sort()
            for k in chooseKeys:
                c += ['--$Map TransEffectChoose ' + classXmiId + ' ' +
                      t.getXmiId() + ' ' + exprIds[k] + '.value ' + k]
        return c

    def smvId(self):
        """Return the id string."""
        return self.__id

    def smvAlive(self):
        """Return the predicate Alive."""
        return 'alive'

    def smvCreated(self):
        """Return the predicate Created."""
        return 'created'

    def smvDestroyed(self):
        """Return the predicate Destroyed."""
        return 'destroyed'

    def smvModule(self):
        """Return the module name for self."""
        return 'Class#' + self.__id

    def smvWithinGlobalResources(self):
        """Return the global WithinResources predicate."""
        return 'global.' + self.__model.smvWithinResources()

    def smvCastFromNullExprModule(self):
        """Return the module name for the cast null -> type of self."""
        return 'CastNullTo#' + self.__id

    def code(self):
        """Generate NuSMV code for the class.

        Return value: NuSMV code as a list of strings.
        """
        c = ['',
             '',
             '-'*70,
             '-- Class ' + self.__id,
             '-'*70,
             '',
             'MODULE ' + self.smvModule() + '(this, global)']
        c += ['  -- Alive/dead status',
              '  VAR',
              '    ' + self.smvAlive() + ' : boolean;',
              '  ASSIGN',
              '    next(' + self.smvAlive() + ') :=',
              '      case',
              '        ' + self.smvCreated() + ' : TRUE;',
              '        ' + self.smvDestroyed() + ' : FALSE;',
              '        1 : ' + self.smvAlive() + ';',
              '      esac;']
        c += codeDefine(self.smvCreated(), 'FALSE')
        c += codeDefine(self.smvDestroyed(), 'FALSE')
        for f in self.__fields:
            c += f.codeInClass()
        c += self.__queue.codeInClass()
        c += self.__machine.code()
        c += self.__context.code()
        c += self.__codeCastModules()
        c += self.__queue.codeQElementModule()
        for f in self.__fields:
            c += f.codeModules()
        return c

    def __codeCastModules(self):
        """Generate NuSMV modules for cast expressions.

        Return value: NuSMV code as a list of strings.
        """
        c = ['',
             '-- Cast expression modules for class ' + self.__id + '.',
             'MODULE ' + self.smvCastFromNullExprModule() + '(operand)']
        c += codeDefine('value', self.__type.smvDefaultValue())
        c += codeDefine('error', 'operand.error')
        c += codeDefine('isNull', 'TRUE')
        return c


class ActiveObject:
    """An active UML object."""

    def __init__(self, umlObject, id_, number, class_):
        """Construct an active object.

        Param umlObject: the underlying uml.Object, or None if this
        object is not initially alive.

        Param id_: an id string.

        Param number: the integer number of this object.

        Param class_: the ActiveClass of the object.
        """
        assert umlObject is None or umlObject.oclIsKindOf(uml.Object)
        self.__umlObject = umlObject
        self.__id = id_
        self.__number = number
        self.__class = class_

    def getClass(self):
        """Return the ActiveClass of self."""
        return self.__class

    def getXmiId(self):
        """Return the XMI id of the underlying uml.Object as a string.

        If there is no underlying uml.Object, return None.
        """
        if self.__umlObject:
            return str(self.__umlObject.__XMIid__())
        else:
            return None

    def processInitialConfiguration(self):
        """Determine the initial field values for self."""
        # self.__fieldValues maps Fields of the class to NuSMV
        # expressions that give the initial values and that are
        # evaluated in the main module.
        self.__fieldValues = {}
        if self.__umlObject:
            for umlSlot in umlSort(self.__umlObject.slot):
                try:
                    f = self.__class.getFieldForUmlAttribute(umlSlot.attribute)
                except exceptions.LookupError:
                    raise errors.InputError('unknown attribute for slot ' +
                                            str(umlSlot))
                if f in self.__fieldValues:
                    raise errors.InputError('attribute value ' + str(umlSlot) +
                                            ' initialized twice')
                # Treat the name of the slot's value as an action
                # language expression.
                body = str(umlSlot.value.name)
                context = self.__class.getActivityContext()
                activity = activities.Activity(context, isConstant=True)
                exprId = actions.processExpr(body, f.getType(), activity)
                self.__fieldValues[f] = (self.smvVariable() + '.' +
                                         exprId + '.value')
            for umlLinkEnd in umlSort(self.__umlObject.linkEnd):
                umlLink = umlLinkEnd.__parent__()
                for umlOtherEnd in umlSort(umlLink.connection):
                    if (umlOtherEnd is not umlLinkEnd and
                        umlOtherEnd.associationEnd.isNavigable):
                        # The uml.LinkEnd umlOtherEnd is an instance
                        # of a uml.AssociationEnd that is navigable
                        # from the class of self.  There should be a
                        # field corresponding the uml.AssociationEnd,
                        # and an object at umlOtherEnd.  Initialize
                        # the field with a reference to the object.
                        (f, obj) = self.__getLink(umlOtherEnd)
                        if f in self.__fieldValues:
                            if options.opt.sloppy_input:
                                print ('warning: link ' + str(umlOtherEnd) +
                                       ' initialized twice')
                            else:
                                raise errors.InputError('link ' +
                                                        str(umlOtherEnd) +
                                                        ' initialized twice')
                        self.__fieldValues[f] = obj.smvReference()
        for f in self.__class.getFields():
            if f not in self.__fieldValues:
                # Uninitialized fields get the default initial value
                # of the field.
                self.__fieldValues[f] = (self.smvVariable() + '.' +
                                         f.smvInitialValue())

    def __getLink(self, umlEnd):
        """Return a pair (field, object) for a given uml.LinkEnd."""
        umlAssoEnd = umlEnd.associationEnd
        try:
            f = self.__class.getFieldForUmlAssociationEnd(umlAssoEnd)
        except exceptions.LookupError:
            raise errors.InputError('unknown link end ' + str(umlEnd))
        try:
            obj = f.getType().getClass().getObjectForUml(umlEnd.instance)
        except exceptions.LookupError:
            raise errors.InputError('dangling link end ' + str(umlEnd))
        return (f, obj)

    def smvId(self):
        """Return the id string."""
        return self.__id

    def smvVariable(self):
        """Return the variable name for self in the main module."""
        return 'obj#' + self.__id

    def smvReference(self):
        """Return a NuSMV expression giving the instance number of self.

        The expression can be evaluated in any module.
        """
        refValueCount = len(self.__class.getObjects()) + 1
        return smvWord(self.__number, refValueCount)

    def smvFieldValue(self, field, valuation):
        """Return a NuSMV expression for the value of a field in self.

        Param field: the Field.

        Param valuation: a valuation id or 'global', giving the
        valuation from which the field is taken.  If this is None,
        then the field is taken from the 'this' object.

        Return value: a NuSMV expression as a string.
        """
        s = ''
        if valuation:
            s += valuation + '.'
        s += self.smvVariable() + '.' + field.smvVariable()
        return s

    def code(self):
        """Generate code for defining and initializing the object.

        The returned code belongs to the main module.

        Return value: NuSMV code as a list of strings.
        """
        c = ['  -- Object ' + str(self.__number) +
             ' of class ' + self.__class.smvId()]
        if self.__umlObject is None:
            initAlive = 'FALSE'
        else:
            initAlive = 'TRUE'
        c += ['  VAR',
              '    ' + self.smvVariable() + ' : ' + self.__class.smvModule() +
              '(' + self.smvReference() + ', self);',
              '  ASSIGN',
              '    init(' + self.smvVariable() + '.' +
              self.__class.smvAlive() + ') := ' + initAlive + ';']
        for f in self.__class.getFields():
            c += ['    init(' + self.smvFieldValue(f, None) + ') := ' +
                  self.__fieldValues[f] + ';']
        return c


class Field:
    """A base class for fields.

    A field of a class represents either a uml.Attribute of the class
    or a uml.AssociationEnd navigable from the class.
    """

    def __init__(self, id_, owner):
        """Construct a field.

        Param id_: an id string.

        Param owner: the owning ActiveClass.
        """
        self.__id = id_
        self._owner = owner
        self._type = None  # self._type must be set in the subclass.

    def getOwner(self):
        """Return the owning ActiveClass."""
        return self._owner

    def getType(self):
        """Return the Type of the field."""
        return self._type

    def smvInitialValue(self):
        """Return a NuSMV expression giving the initial value of self.

        The expression is valid in the context of the owning class.
        """
        return 'initial#' + self.__id

    def smvId(self):
        """Return the id string."""
        return self.__id

    def smvVariable(self):
        """Return the variable name of self in the owning class module.

        The same name is also used inside valuations.
        """
        return 'field#' + self.__id

    def smvWritten(self):
        """Return the predicate Written."""
        return 'written#' + self.__id

    def smvAccessExprModule(self):
        """Return the NuSMV module name for a field access expression."""
        return 'FieldAccessExpr#' + self._owner.smvId() + '#' + self.__id

    def smvValuationModule(self):
        """Return the NuSMV module name for a valuation of self."""
        return 'Valuation#' + self._owner.smvId() + '#' + self.__id

    def codeInClass(self):
        """Generate code for defining the field.

        The variable for the field is defined, and the frame condition
        (if not written, value stays the same).  Initial values and
        other cases are determined elsewhere.  The returned code
        belongs to the module of the class.

        Return value: NuSMV code as a list of strings.
        """
        c = ['  VAR',
             '    ' + self.smvVariable() + ' : ' +
             self._type.smvTypeDefinition() + ';']
        c += codeTrans('(' + smvAnd(['!' + self.smvWritten(),
                                     '!' + self._owner.smvDestroyed()]) +
                       ') -> (next(' + self.smvVariable() + ') = ' +
                       self.smvVariable() + ')')
        c += codeTrans(self._owner.smvDestroyed() +
                       ' -> (next(' + self.smvVariable() + ') = ' +
                       self.smvInitialValue() + ')')
        c += codeDefine(self.smvInitialValue(), self._smvInitial())
        return c

    def _smvInitial(self):  #pragma: no cover
        """Return a NuSMV expression for the initial value of the field."""
        raise exceptions.NotImplementedError('Pure virtual method')

    def codeModules(self):
        """Generate code for modules specific to self.

        Return value: NuSMV code as a list of strings.
        """
        c = ['',
             '-- Field access modules for ' + self.__id + '.']
        c += ['',
              'MODULE ' + self.smvAccessExprModule() + '(refExpr, valuation)',
              '  DEFINE',
              '    value :=',
              '      case']
        for obj in self._owner.getObjects():
            c += ['        refExpr.value = ' + obj.smvReference() + ' : ' +
                  obj.smvFieldValue(self, 'valuation') + ';']
        c += ['        1 : ' + self._type.smvDefaultValue() + ';  -- error',
              '      esac;']
        c += codeDefine('error', 'refExpr.error | refExpr.isNull')
        if self._type.isReferenceType():
            c += codeDefine('isNull', '(value = ' +
                            self._type.smvDefaultValue() + ')')
        c += ['',
              'MODULE ' + self.smvValuationModule() + '(ref, new, default)',
              '  DEFINE']
        for obj in self._owner.getObjects():
            c += ['    ' + obj.smvFieldValue(self, None) + ' :=',
                  '      case',
                  '        ref = ' + obj.smvReference() + ' : new;',
                  '        1 : ' + obj.smvFieldValue(self, 'default') + ';',
                  '      esac;']
        return c


class AttributeField(Field):
    """A field represented as an uml.Attribute."""

    def __init__(self, umlAttribute, id_, owner):
        """Construct a field from an attribute.

        Param umlAttribute: a uml.Attribute.

        Param id_: an id string.

        Param owner: the owning ActiveClass.
        """
        Field.__init__(self, id_, owner)
        self.__umlAttribute = umlAttribute
        try:
            self._type = datatypes.resolvePrimitiveType(umlAttribute.type)
        except exceptions.LookupError:
            raise errors.NotSupportedError('unsupported type for attribute ' +
                                           str(umlAttribute))
        for r in umlAttribute.multiplicity.range:
            if r.lower != 1 or r.upper != 1:
                raise errors.NotSupportedError('unsupported multiplicity ' +
                                               'for attribute ' +
                                               str(umlAttribute))

    def _smvInitial(self):
        if self.__umlAttribute.initialValue.body:
            body = str(self.__umlAttribute.initialValue.body)
            activity = activities.Activity(self._owner.getActivityContext(),
                                           isConstant=True)
            exprId = actions.processExpr(body, self._type, activity)
            return exprId + '.value'
        else:
            return self._type.smvDefaultValue()

    def getName(self):
        """Return the UML name as a string."""
        return str(self.__umlAttribute.name)


class AssociationField(Field):
    """A field represented in UML as a navigable association end."""

    def __init__(self, umlEnd, type_, id_, owner):
        """Construct a field from an association end.

        Param umlEnd: a uml.AssociationEnd.

        Param type_: the ReferenceType of the field.

        Param id_: an id string.

        Param owner: the owning ActiveClass.
        """
        Field.__init__(self, id_, owner)
        self.__umlEnd = umlEnd
        self._type = type_
        for r in umlEnd.multiplicity.range:
            if r.upper != 1:
                raise errors.NotSupportedError('unsupported multiplicity ' +
                                               'for association end ' +
                                               str(umlEnd))

    def _smvInitial(self):
        return self._type.smvDefaultValue()

    def getName(self):
        """Return the UML name as a string."""
        return str(self.__umlEnd.name)
