#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# datatypes.py
#
# 2006-12-04  Jori Dubrovin:  Using optparse for options
# 2006-07-30  Jori Dubrovin:  Initial version
#

"""Type definitions for UML to NuSMV."""

import exceptions

import errors
import options
from smvutils import smvAnd, smvOr, smvWord, smvWordType


class Type:
    """A base class for the type system."""

    def __init__(self):
        pass

    def isSubtypeOf(self, other):
        """Return True if self is a subtype of the other Type."""
        return (self is other)  # Default implementation

    def isCastableTo(self, other):
        """Return True if self is castable to the other Type."""
        if self.isSubtypeOf(other) or other.isSubtypeOf(self):
            return True
        return False

    def isReferenceType(self):
        """Return True if self is a reference type."""
        return False  # Default implementation

    def smvTypeDefinition(self):  #pragma: no cover
        """Return the NuSMV type corresponding to self."""
        raise exceptions.NotImplementedError('Pure virtual method')

    def smvDefaultValue(self):  #pragma: no cover
        """Return NuSMV code for the default value of this type."""
        raise exceptions.NotImplementedError('Pure virtual method')


class _NullType(Type):
    """The null type."""

    def __init__(self):
        Type.__init__(self)

    def __str__(self):
        return '<null type>'

    def isSubtypeOf(self, other):
        return (self is other or other.isReferenceType())

    def smvDefaultValue(self):
        assert False  # No values

nullType = _NullType()


class PrimitiveType(Type):
    """A base class for primitive types."""

    def __init__(self):
        Type.__init__(self)


class _IntType(PrimitiveType):
    """The Jumbala int type."""

    def __init__(self):
        PrimitiveType.__init__(self)

    def __str__(self):
        return 'int'

    def smvTypeDefinition(self):
        return smvWordType(1L << options.opt.int_bits)

    def smvDefaultValue(self):
        return self.smvValue(0)

    def smvValue(self, value):
        """Return a NuSMV expression for a given Python value.

        Param value: an integer between -2**31 and 2**31-1.
        """
        count = 1L << options.opt.int_bits
        if value >= 0:  # Implement 2's complement with unsigned words
            n = value
        else:
            n = value + count
        if n < 0 or n >= count:
            raise errors.NotSupportedError('too narrow int to hold value ' +
                                           str(value))
        return smvWord(n, count)

intType = _IntType()


class _BoolType(PrimitiveType):
    """The Jumbala boolean type."""

    def __init__(self):
        PrimitiveType.__init__(self)

    def __str__(self):
        return 'boolean'

    def smvTypeDefinition(self):
        return 'boolean'

    def smvDefaultValue(self):
        return 'FALSE'

    def smvValue(self, value):
        """Return a NuSMV expression for a given Python value.

        Param value: False or True.
        """
        return {False: 'FALSE', True: 'TRUE'}[value]

boolType = _BoolType()


class ReferenceType(Type):
    """A reference type.

    The only reference types are references to active classes.  There
    is a one-to-one correspondence between ReferenceTypes and
    ActiveClasses.
    """

    def __init__(self, class_):
        """Construct a reference type.

        Param class_: the ActiveClass to which self refers.
        """
        Type.__init__(self)
        self.__class = class_

    def isReferenceType(self):
        return True

    def getClass(self):
        """Return the corresponding ActiveClass."""
        return self.__class

    def smvTypeDefinition(self):
        refValueCount = len(self.__class.getObjects()) + 1
        # This does not work if there are 0 objects of the class.
        return smvWordType(refValueCount)

    def smvDefaultValue(self):
        refValueCount = len(self.__class.getObjects()) + 1
        # null value is the zero word
        return smvWord(0, refValueCount)


def resolvePrimitiveType(umlType):
    """Return a PrimitiveType corresponding to a given uml.Classifier."""
    if str(umlType.name).lower() in ['bool', 'boolean']:
        return boolType
    if str(umlType.name).lower() in ['int', 'integer']:
        return intType
    raise errors.InputError('unknown datatype ' + str(umlType))
