#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# events.py
#
# 2007-05-04  Jori Dubrovin:  Simplified definitions
# 2007-03-08  Jori Dubrovin:  Fixed defer queue flush on implicit consumption
# 2006-12-04  Jori Dubrovin:  Fixed a signal type problem
# 2006-11-21  Jori Dubrovin:  Initial version
#

"""Translation of UML events, event queues, and signals to NuSMV."""

import exceptions

import datatypes
import errors
import options
from coralutils import uml, umlSort
from smvutils import smvAnd, smvOr, smvWord, smvWordType
from smvutils import codeDefine, codeTrans, codeAtMostOne


class EventQueue:
    """Representation of an event queue in an active UML class.

    There is a one-to-one correspondence between ActiveClass objects
    and EventQueue objects.
    """

    class Item:
        """An item in an element in the event queue.

        An item holds either a signal index or a signal parameter value.
        """
        pass

    def __init__(self, class_):
        """Construct an event queue.

        Param class_: the associated ActiveClass.
        """
        self.__class = class_
        self.__receivedSignals = []  # Signals sent to this queue
        self.__triggerSignals = []   # Signals that trigger transitions
        self.__relevantSignals = []  # union of the above two
        self.__items = []  # Items for signal parameters and signal indices

    def getQueueSize(self):
        """Return the event queue size."""
        n = options.opt.queue_size
        umlName = self.__class.getName()
        for s in options.opt.specific_queue_size:
            (name, size) = s.split(':')
            if name == umlName:
                n = int(size)
        return n

    def getRelevantSignals(self):
        """Return the relevant Signals received as a list.

        A signal is relevant if it appears as a trigger in a
        transition in the associated state machine, or if there is
        globally at least one transition that sends the signal to the
        associated class.

        All relevant signals must have been added.
        """
        return self.__relevantSignals[:]

    def smvQElementModule(self):
        """Return the module name for a queue element."""
        return 'QElement#' + self.__class.smvId()

    def smvQVar(self, index):
        """Return the variable name for the queue element with the given index.

        The variable lives in the module of the class.

        Queue indices run from zero to (queue size - 1).
        """
        return 'queue' + str(index)

    def smvDispatch(self, signal):
        """Return the predicate Dispatch.

        Defined for all relevant signals.
        """
        return 'dispatch#' + signal.smvId()

    def smvDispatchSys(self):
        """Return the predicate DispatchSys."""
        return 'dispatchSys'

    def smvDispatchExt(self):
        """Return the predicate DispatchExt."""
        return 'dispatchExt'

    def smvDispatchAny(self):
        """Return the predicate DispatchAny."""
        return 'dispatchAny'

    def smvDefer(self, signal):
        """Return the predicate Defer.

        Defined for all received system signals.
        """
        return 'defer#' + signal.smvId()

    def smvDeferSys(self):
        """Return the predicate DeferSys."""
        return 'deferSys'

    def smvEnabledSys(self):
        """Return the predicate EnabledSys."""
        return 'enabledSys'

    def smvAnyEnabled(self, signal):
        """Return the predicate AnyEnabled.

        Defined for all received system signals.
        """
        return 'anyEnabled#' + signal.smvId()

    def smvImplicitConsumeSys(self):
        """Return the predicate ImplicitConsumeSys."""
        return 'implicitConsumeSys'

    def smvDequeue(self):
        """Return the predicate Dequeue."""
        return 'dequeue'

    def smvInputQueueEmpty(self):
        """Return the predicate InputQueueEmpty."""
        return 'inputQueueEmpty'

    def smvQueueFull(self):
        """Return the predicate QueueFull."""
        return 'queueFull'

    def smvQueueOverflow(self):
        """Return the predicate QueueOverflow."""
        return 'queueOverflow'

    def smvSentTo(self):
        """Return the predicate SentTo."""
        return 'sentTo'

    def smvSignalIndex(self, signal):
        """Return an expression for the signal number in this queue.

        Defined for all received system signals.
        """
        return 'sig#' + signal.smvId()

    def smvReception(self, obj, signal, params):
        """Generate a NuSMV expression for signal reception.

        processReceptions must have been called.

        The returned expression can be evaluated within any class, and
        it is true if and only if the signal received by self is the
        given signal with the given parameter values.

        Param obj: the ActiveObject (whose class must be the class
        associated with self) receiving the signal.

        Param signal: the desired Signal or None, denoting the fact
        that no signal is received.

        Param params: a list of expression identifiers giving the
        values for signal parameters.

        Return value: a NuSMV expression as a string.
        """
        assert obj.getClass() is self.__class
        objDot = 'global.' + obj.smvVariable() + '.'
        if signal is None:
            return ('(' + objDot + 'newEvent.signal = ' +
                    objDot + 'noEvent.signal)')
        sigIndex = objDot + self.smvSignalIndex(signal)
        s = ['(' + objDot + 'newEvent.signal = ' + sigIndex + ')']
        for item in self.__receiveItems:
            if item.signal is signal:
                s += ['(' + objDot + 'newEvent.' + item.id + ' = ' +
                      params[item.paramIndex] + '.value)']
        return '(' + smvAnd(s) + ')'

    def smvSignalParamValues(self, signal):
        """Return NuSMV expressions giving values for signal parameters.

        The expressions give the values for all parameters of the
        given signal, assuming that it is the signal of the current
        event.  The expressions must be evaluated in module of the
        associated class.

        This can be called as soon as addTriggerSignal has been called
        for the signal.

        Param signal: the Signal.

        Return value: a list of NuSMV expressions.
        """
        return ['currEvent.' + item.id for
                item in self.__items if item.signal is signal]

    def addReceivedSignal(self, signal):
        """Register the reception of a signal from a send statement."""
        if signal not in self.__receivedSignals:
            self.__receivedSignals.append(signal)
            if signal not in self.__relevantSignals:
                self.__relevantSignals.append(signal)

    def addTriggerSignal(self, signal):
        """Register a trigger in the associated state machine."""
        if signal not in self.__triggerSignals:
            self.__triggerSignals.append(signal)
            if signal not in self.__relevantSignals:
                self.__relevantSignals.append(signal)
            # Add the Items corresponding to signal parameters.
            p = 0
            for paramType in signal.getParamTypes():
                item = EventQueue.Item()
                item.id = 'param' + str(p) + '#' + signal.smvId()
                item.smvTypeDefinition = paramType.smvTypeDefinition()
                item.smvDefaultValue = paramType.smvDefaultValue()
                item.signal = signal
                item.paramIndex = p
                self.__items.append(item)
                p += 1

    def processReceptions(self):
        """Process the signals that are sent to self.

        All received signals must have been added.
        """
        # Signal type must be able to hold any received signal plus
        # nosignal.
        sigCount = len(self.__receivedSignals) + 1
        sigItem = EventQueue.Item()
        sigItem.id = 'signal'
        if sigCount == 1:
            # The only value of the signal type is nosignal.
            sigItem.smvTypeDefinition = None
            sigItem.smvDefaultValue = '0'  # Could be anything.
        else:
            sigItem.smvTypeDefinition = smvWordType(sigCount)
            sigItem.smvDefaultValue = smvWord(0, sigCount)
        sigItem.signal = None
        sigItem.paramIndex = None
        self.__items.append(sigItem)
        self.__receiveItems = [item for item in self.__items if
                               item is sigItem or
                               item.signal in self.__receivedSignals]

    def codeQElementModule(self):
        """Generate the queue element module for the associated class.

        Return value: NuSMV code as a list of strings.
        """
        c = ['',
             '-- An element of the event queue.',
             'MODULE ' + self.smvQElementModule() +
             '(index, aheadFree, behind, owner)',
             '  -- index is the integer index of the element.',
             '  -- aheadFree is true iff the element in front of this one',
             '  -- is empty (false if this is the first element).',
             '  -- behind is the elements behind this one.',
             '  -- owner is the containing object.']
        for item in self.__receiveItems:
            typeDef = item.smvTypeDefinition
            if typeDef is None:
                # A singleton type.  Make item a constant because
                # NuSMV does not properly support types with just a
                # single value.
                c += codeDefine(item.id, 'owner.noEvent.' + item.id)
            else:
                c += ['  VAR',
                      '    ' + item.id + ' : ' + typeDef + ';']
        c += ['  -- thisFree is true iff this element is empty.',
              '  -- behindFree is true iff the element behind this one',
              '  -- is empty (true if this is the last element).',
              '  -- newHere is true iff this element should receive',
              '  -- the new event (or noEvent if there is no new event).',
              '  -- remove is true iff an event is being consumed from the',
              '  -- queue from a position that is not behind this element.',
              '  DEFINE',
              '    thisFree := (signal = owner.noEvent.signal);',
              '    behindFree := (behind.signal = owner.noEvent.signal);',
              '    newHere := ' + smvAnd(['!aheadFree', 'behindFree',
                                          '(owner.' + self.smvDequeue() +
                                          ' <-> !thisFree)']) + ';',
              '    remove := ' + smvAnd(['owner.' + self.smvDequeue(),
                                         '(owner.queuePos <= index)']) + ';',
              '  ASSIGN']
        for item in self.__receiveItems:
            if item.smvTypeDefinition is not None:
                c += ['    init(' + item.id + ') := owner.noEvent.' +
                      item.id + ';',
                      '    next(' + item.id + ') :=',
                      '      case',
                      '        owner.' + self.__class.smvDestroyed() +
                      ' : owner.noEvent.' + item.id + ';',
                      '        newHere : owner.newEvent.' + item.id + ';',
                      '        remove : behind.' + item.id + ';',
                      '        1 : ' + item.id + ';',
                      '      esac;']
        return c

    def codeInClass(self):
        """Generate code for event handling.

        The code is placed in the module for the associated class.

        Return value: NuSMV code as a list of strings.
        """
        return (self.__codeQueue() +
                self.__codeEvents() +
                self.__codeDispatching())

    def __codeQueue(self):
        """Generate code for the event queue.

        This goes to the code module of the class.

        Return value: NuSMV code as a list of strings.
        """
        c = ['  -- Event queue.',
             '  -- The frontmost element is queue0.',
             '  VAR']
        size = self.getQueueSize()
        for i in xrange(size):
            if i == 0:
                aheadFree = 'FALSE'
            else:
                aheadFree = ('(' + self.smvQVar(i-1) +
                             '.signal = noEvent.signal)')
            if i == size-1:
                behind = 'noEvent'
            else:
                behind = self.smvQVar(i+1)
            c += ['    ' + self.smvQVar(i) + ' : ' +
                  self.smvQElementModule() + '(' + smvWord(i, size+1) + ', ' +
                  aheadFree + ', ' + behind + ', self);']
        # There are size+1 distinct values for queuePos.
        c += ['  -- queuePos points to the front element of the input',
              '  -- queue in the event queue.  The elements',
              '  -- whose index is less than queuePos contain a deferred',
              '  -- event.  The input queue consists of the elements whose',
              '  -- index is greater or equal to queuePos.',
              '  VAR',
              '    queuePos : ' + smvWordType(size+1) + ';',
              '  ASSIGN',
              '    init(queuePos) := ' + smvWord(0, size+1) + ';',
              '    next(queuePos) :=',
              '      case',
              '        ' + smvOr([self.smvEnabledSys(), self.smvDispatchExt(),
                                  self.__class.smvDestroyed()]) + ' : ' +
              smvWord(0, size+1) + ';',
              '        ' + self.smvDeferSys() + ' : queuePos + ' +
              smvWord(1, size+1) + ';',
              '        1 : queuePos;',
              '      esac;']
        if size > 0:
            c += codeDefine(self.smvQueueFull(),
                            '!(' + self.smvQVar(size-1) +
                            '.signal = noEvent.signal)')
        else:
            c += codeDefine(self.smvQueueFull(), 'TRUE')
        c += codeDefine(self.smvInputQueueEmpty(),
                        '(currEvent.signal = noEvent.signal)')
        return c

    def __codeEvents(self):
        """Generate code for determining the current and received event.

        This goes to the code module of the class.

        Return value: NuSMV code as a list of strings.
        """
        size = self.getQueueSize()
        c = ['  -- The current event (next event to be dispatched).',
             '  DEFINE']
        # All items are defined for currEvent and noEvent.  Only
        # receive-items are defined for newEvent and queue elements.
        # Receive-items are the signal item and parameters for signals
        # that are received.
        for item in self.__items:
            # Nondeterministic external signal parameters:
            # IVAR currEvent.param2#$extsig : ...
            c += ['    currEvent.' + item.id + ' :=',
                  '      case']
            if item in self.__receiveItems:
                for i in xrange(size):
                    c += ['        queuePos = ' + smvWord(i, size+1) + ' : ' +
                          self.smvQVar(i) + '.' + item.id + ';']
            c += ['        1 : noEvent.' + item.id + ';',
                  '      esac;']
        c += ['  -- An empty event.',
              '  DEFINE']
        for item in self.__items:
            c += ['    noEvent.' + item.id + ' := ' +
                  item.smvDefaultValue + ';']
        c += ['  -- newEvent is the event being received by this object.']
        for item in self.__receiveItems:
            typeDef = item.smvTypeDefinition
            if typeDef is None:  # A singleton type
                c += codeDefine('newEvent.' + item.id, 'noEvent.' + item.id)
            else:
                # newEvent is not really a nondeterministic choice, but
                # its definition is scattered to different modules.
                c += ['  IVAR',
                      '    newEvent.' + item.id + ' : ' + typeDef + ';']
            if item.signal is not None:
                # Clear all non-applicable signal parameter values in
                # newEvent.
                idx = self.smvSignalIndex(item.signal)
                c += codeTrans('!(newEvent.signal = ' + idx +
                               ') -> (newEvent.' + item.id +
                               ' = noEvent.' + item.id + ')')
        # A frame condition: if no one sends a signal to us,
        # then clear newEvent.  Other cases are covered by the
        # sender.
        c += codeTrans('!' + self.smvSentTo() +
                       ' -> (newEvent.signal = noEvent.signal)')
        c += ['  -- queueOverflow is true if a signal is being sent to',
              '  -- this object when the queue is already full.']
        c += codeDefine(self.smvQueueOverflow(),
                        smvAnd([self.smvQueueFull(),
                                '!(newEvent.signal = noEvent.signal)',
                                '!' + self.smvDequeue()]))
        return c

    def __codeDispatching(self):
        """Generate code for event dispatching from the event queue.

        This goes to the code module of the class.

        Return value: NuSMV code as a list of strings.
        """
        c = ['  -- Event dispatching.']
        machine = self.__class.getStateMachine()
        sigCount = len(self.__receivedSignals) + 1
        k = 0
        for signal in self.__receivedSignals:
            k += 1
            c += codeDefine(self.smvSignalIndex(signal), smvWord(k, sigCount))
            c += ['  IVAR',
                  '    ' + self.smvDispatch(signal) + ' : boolean;']
            c += codeTrans('!(currEvent.signal = ' +
                           self.smvSignalIndex(signal) + ') -> !' +
                           self.smvDispatch(signal))
            c += codeDefine(self.smvAnyEnabled(signal),
                            smvAnd([self.smvDispatch(signal),
                                    machine.smvAnyFeasible(signal)]))
            c += codeDefine(self.smvDefer(signal),
                            smvAnd([self.smvDispatch(signal),
                                    '!' + machine.smvAnyFeasible(signal),
                                    machine.smvAnyDeferring(signal)]))
        for signal in self.__relevantSignals:
            if signal.isExternal():
                c += ['  IVAR',
                      '    ' + self.smvDispatch(signal) + ' : boolean;']
                # An external signal can be dispatched only if it
                # fires a transition.
                c += codeTrans('!' + machine.smvAnyFeasible(signal) +
                               ' -> !' + self.smvDispatch(signal))
            if (signal not in self.__receivedSignals and
                not signal.isExternal()):
                # This signal triggers transitions but cannot appear
                # in the event queue.
                c += codeDefine(self.smvDispatch(signal), 'FALSE')
        c += codeDefine(self.smvDispatchSys(),
                        smvOr([self.smvDispatch(e) for
                               e in self.__receivedSignals]))
        c += codeDefine(self.smvEnabledSys(),
                        smvOr([self.smvAnyEnabled(e) for
                               e in self.__receivedSignals]))
        c += codeDefine(self.smvDeferSys(),
                        smvOr([self.smvDefer(e) for
                               e in self.__receivedSignals]))
        c += ['  -- dequeue is true when an event is removed from the input',
              '  -- queue and not added to the deferred queue.']
        c += codeDefine(self.smvDequeue(), smvAnd([self.smvDispatchSys(),
                                                   '!' + self.smvDeferSys()]))
        c += codeDefine(self.smvImplicitConsumeSys(),
                        smvAnd([self.smvDequeue(),
                                '!' + self.smvEnabledSys()]))
        if options.opt.check_implicit_consumption:
            c += ['  LTLSPEC',
                  '    G (' + self.__class.smvWithinGlobalResources() +
                  ' -> !' + self.smvImplicitConsumeSys() + ');']
        c += codeAtMostOne('extDispatches',
                           [self.smvDispatch(e) for
                            e in self.__relevantSignals if e.isExternal()])
        c += codeDefine(self.smvDispatchExt(), 'extDispatches.atLeastOne')
        c += codeAtMostOne('anyDispatches',
                           [self.smvDispatchSys(), self.smvDispatchExt()])
        c += codeDefine(self.smvDispatchAny(), 'anyDispatches.atLeastOne')
        c += codeTrans(machine.smvInRtc() + ' -> !' + self.smvDispatchAny())
        return c


class Signal:
    """Representation of a UML signal."""

    def __init__(self, umlSignal, id_, model):
        """Construct a signal.

        Param umlSignal: the underlying uml.Signal.

        Param id_: an identification string.

        Param model: the Model owning the signal.
        """
        assert umlSignal.oclIsKindOf(uml.Signal)
        self.__umlSignal = umlSignal
        self.__id = id_
        self.__model = model

    def getName(self):
        """Return the name of the signal as a string."""
        return str(self.__umlSignal.name)

    def isExternal(self):
        """Return True if self is an external (ATO) signal."""
        return self.getName().startswith('$')

    def isInUmlSignals(self, umlSignals):
        """Return True if self is in a given list of uml.Signals."""
        return (self.__umlSignal in umlSignals)

    def getParamTypes(self):
        """Return a list of Types of signal parameters."""
        return self.__paramTypes[:]

    def smvId(self):
        """Return the id string."""
        return self.__id

    def processParams(self):
        """Add all signal parameters to self.

        All ActiveClasses must already be added to the model.
        """
        self.__paramTypes = []
        # umlSort must not be used to read Signal.feature.
        for umlFeature in self.__umlSignal.feature:
            if umlFeature.oclIsKindOf(uml.Attribute):
                self.__paramTypes.append(self.__resolveParamType(umlFeature))

    def __resolveParamType(self, umlAttribute):
        """Return the Type of a parameter represented as a uml.Attribute."""
        try:
            type_ = self.__model.getClassForUml(umlAttribute.type).getType()
        except exceptions.LookupError:
            try:
                type_ = datatypes.resolvePrimitiveType(umlAttribute.type)
            except exceptions.LookupError:
                raise errors.NotSupportedError('unsupported type for ' +
                                               'signal attribute ' +
                                               str(umlAttribute))
        for r in umlAttribute.multiplicity.range:
            if r.upper != 1:
                raise errors.NotSupportedError('unsupported multiplicity ' +
                                               'for signal attribute ' +
                                               str(umlAttribute))
        return type_
