#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# models.py
#
# 2007-06-20  Jori Dubrovin:  New choose expression mapping
# 2007-04-01  Jori Dubrovin:  Strictly static queue-full predicate
# 2007-03-08  Jori Dubrovin:  Adding states to element map
# 2006-12-05  Jori Dubrovin:  Transition instance logging
# 2006-11-03  Jori Dubrovin:  Initial version
#

"""NuSMV encoding of an entire UML model."""

import exceptions

import actions
import classes
import events
import errors
import options
from coralutils import uml, umlSort
from smvutils import smvAnd, smvOr, smvWord, smvWordType
from smvutils import codeDefine, codeTrans, codeAtMostOne
from smvutils import generateId
from smvutils import codeAtMostOneModules


class Model:
    """A behavioral UML model."""

    def __init__(self, umlModel):
        """Construct a model.

        Param umlModel: the underlying uml.Model.
        """
        assert umlModel.oclIsKindOf(uml.Model)
        self.__classes = []           # all ActiveClasses in self
        self.__classForUml = {}       # map uml.Class to ActiveClass
        self.__signals = []           # all Signals in self
        self.__signalForUml = {}      # map uml.Signal to Signal

        self.__addNestedClassifiers(umlModel)
        self.__addNestedObjects(umlModel)
        for class_ in self.__classes:
            class_.processObjects()
        for signal in self.__signals:
            signal.processParams()
        for class_ in self.__classes:
            class_.processFields()
        for class_ in self.__classes:
            class_.processInitialConfiguration()
        for class_ in self.__classes:
            class_.processActivities()
        for class_ in self.__classes:
            class_.getEventQueue().processReceptions()
        self.__transitionInstances = self.__orderTransitionInstances()

    def __addNestedClassifiers(self, umlElem):
        """Add all classes and signals in the given uml.ModelElement."""
        if umlElem.oclIsKindOf(uml.Class):
            self.__addClass(umlElem)
        if umlElem.oclIsKindOf(uml.Signal):
            self.__addSignal(umlElem)
        if umlElem.oclIsKindOf(uml.Namespace):
            for e in umlSort(umlElem.ownedElement):
                self.__addNestedClassifiers(e)

    def __addClass(self, umlClass):
        """Add a new ActiveClass for the given uml.Class to self."""
        if not umlClass.isActive:
            if options.opt.sloppy_input:
                print ('warning: treating passive class ' + str(umlClass) +
                       ' as active')
            else:
                raise errors.NotSupportedError('passive class ' +
                                               str(umlClass) +
                                               ' not supported')
        id_ = generateId('c', umlClass.name,
                         [class_.smvId() for class_ in self.__classes])
        class_ = classes.ActiveClass(umlClass, id_, self)
        self.__classes.append(class_)
        assert not umlClass in self.__classForUml
        self.__classForUml[umlClass] = class_

    def __addSignal(self, umlSignal):
        """Add a new Signal for the given uml.Signal to self."""
        id_ = generateId('e', umlSignal.name,
                         [signal.smvId() for signal in self.__signals])
        signal = events.Signal(umlSignal, id_, self)
        self.__signals.append(signal)
        assert not umlSignal in self.__signalForUml
        self.__signalForUml[umlSignal] = signal

    def __addNestedObjects(self, umlElem):
        """Add all objects in the given uml.ModelElement."""
        if umlElem.oclIsKindOf(uml.Object):
            self.__addUmlObject(umlElem)
        if umlElem.oclIsKindOf(uml.Namespace):
            for e in umlSort(umlElem.ownedElement):
                self.__addNestedObjects(e)
        if umlElem.oclIsKindOf(uml.Instance):
            for e in umlSort(umlElem.ownedInstance):
                self.__addNestedObjects(e)

    def __addUmlObject(self, umlObject):
        """Add a new ActiveObject for the given uml.Object to self.

        All classes must have already been added.
        """
        umlClassifier = umlObject.classifier[0]
        try:
            class_ = self.getClassForUml(umlClassifier)
        except exceptions.LookupError:
            raise errors.InputError('unknown class for object ' +
                                    str(umlObject))
        class_.addObject(umlObject)

    def getClassForUml(self, umlClassifier):
        """Return the ActiveClass corresponding to the given uml.Classifier.

        If there is no such class, raise a LookupError.
        """
        return self.__classForUml[umlClassifier]

    def getSignalForName(self, signalName):
        """Return the Signal with the given name.

        If there are no such Signals, raise a LookupError.  If there
        are several, raise an InputError.
        """
        signals = [e for e in self.__signals if e.getName() == signalName]
        if len(signals) > 1:
            raise errors.InputError("ambiguous signal name '" +
                                    signalName + "'")
        return signals[0]

    def getSignalForUml(self, umlSignal):
        """Return the global Signal for a given uml.Signal.

        If there are no such Signals, raise a LookupError.
        """
        return self.__signalForUml[umlSignal]

    def smvWithinResources(self):
        """Return the predicate WithinResources."""
        return 'withinResources'

    def smvDeadlock(self):
        """Return the predicate Deadlock."""
        return 'deadlock'

    def smvIdle(self):
        """Return the predicate Idle."""
        return 'idle'

    def smvWriteUntil(self, actor, t, target, f):
        """Return the predicate WriteUntil."""
        return '#'.join(['writeUntil', actor.smvId(), t.smvId(),
                        target.smvId(), f.smvId()])

    def smvSendUntil(self, actor, t, target):
        """Return the predicate SendUntil."""
        return '#'.join(['sendUntil', actor.smvId(), t.smvId(),
                         target.smvId()])

    def code(self):
        """Generate code for the model.

        Return value: NuSMV code as a list of strings.
        """
        c = ['',
             '',
             '-'*70,
             '-- UML model',
             '-'*70]
        for class_ in self.__classes:
            c += class_.code()
        # We want the main module to appear last for easier
        # readability.  Nevertheless, codeAtMostOneModules must be
        # called after codeMainModule.
        main = self.__codeMainModule()
        c += actions.codeCommon()
        c += codeAtMostOneModules()
        c += main
        return c

    def __codeMainModule(self):
        """Generate code for the main module containing all objects.

        Return value: NuSMV code as a list of strings.
        """
        c = ['',
             '',
             '-'*70,
             '-- Main module',
             '-'*70,
             '',
             'MODULE main']
        objects = [obj for
                   class_ in self.__classes for obj in class_.getObjects()]
        for obj in objects:
            c += obj.code()
        c += self.__codeTransitionInstances()
        c += ['  -- withinResources stays true as long as no queue overflows.',
              '  VAR',
              '    ' + self.smvWithinResources() + ' : boolean;',
              '  ASSIGN',
              '    init(' + self.smvWithinResources() + ') := TRUE;',
              '    next(' + self.smvWithinResources() + ') := ' +
              smvAnd([self.smvWithinResources()] +
                     ['!' + obj.smvVariable() + '.' +
                      obj.getClass().getEventQueue().smvQueueOverflow() for
                      obj in objects]) + ';']
        c += codeDefine(self.smvDeadlock(),
                        smvAnd(['!' + obj.smvVariable() + '.' +
                                obj.getClass().getStateMachine().smvReady() for
                                obj in objects]))
        sched = [obj.smvVariable() + '.' +
                 obj.getClass().getStateMachine().smvScheduled() for
                 obj in objects]
        if options.opt.encode_interleaving:
            c += codeAtMostOne('scheduler', sched)
            c += codeDefine(self.smvIdle(), '!scheduler.atLeastOne')
        else:
            c += codeDefine(self.smvIdle(), '!(' + smvOr(sched) + ')')
        c += codeTrans('!' + self.smvDeadlock() + ' -> !' + self.smvIdle())
        if options.opt.check_deadlock:
            c += ['  LTLSPEC',
                  '    G (' + self.smvWithinResources() + ' -> !' +
                  self.smvDeadlock() + ');']
        return c

    def __codeTransitionInstances(self):
        """Generate code for transition instances.

        Return value: NuSMV code as a list of strings.
        """
        written = {}  # map (object, field) to predicate WriteUntil
        sentTo = {}   # map object to predicate SendUntil
        c = ['  -- Transition instances']
        for (actor, t) in self.__transitionInstances:
            class_ = actor.getClass()
            actorDot = actor.smvVariable() + '.'
            if not options.opt.encode_interleaving:
                for read in t.getActivity().getFieldReads():
                    f = read.getField()
                    for target in f.getOwner().getObjects():
                        c += codeTrans(written.get((target, f), 'FALSE') +
                                       ' -> !' + actorDot +
                                       t.smvRead(target, f))
            for write in t.getActivity().getFieldWrites():
                f = write.getField()
                for target in f.getOwner().getObjects():
                    writeUntil = self.smvWriteUntil(actor, t, target, f)
                    c += codeDefine(writeUntil,
                                    smvOr([written.get((target, f), 'FALSE'),
                                           actorDot + t.smvWrite(target, f)]))
                    written[(target, f)] = writeUntil
            if (not options.opt.encode_interleaving and
                not options.opt.allow_queue_overdraft):
                if (not t.isCompletion() and
                    not t.getTriggerSignal().isExternal()):
                    # Forbid firing a transition that reads the event
                    # queue if the queue may have been written to.
                    if options.opt.encode_static:
                        full = 'TRUE'  # static overapproximation
                    else:
                        full = actorDot + class_.getEventQueue().smvQueueFull()
                    c += codeTrans('(' + smvAnd([sentTo.get(actor, 'FALSE'),
                                                 full]) +
                                   ') -> !' + actorDot + t.smvFire())
            for send in t.getActivity().getSends():
                for target in send.getTargetClass().getObjects():
                    if not options.opt.encode_interleaving:
                        c += codeTrans(sentTo.get(target, 'FALSE') + ' -> !' +
                                       actorDot + t.smvSend(target))
                    sendUntil = self.smvSendUntil(actor, t, target)
                    c += codeDefine(sendUntil,
                                    smvOr([sentTo.get(target, 'FALSE'),
                                           actorDot + t.smvSend(target)]))
                    sentTo[target] = sendUntil
        for class_ in self.__classes:
            for obj in class_.getObjects():
                objDot = obj.smvVariable() + '.'
                for f in class_.getFields():
                    c += codeDefine(objDot + f.smvWritten(),
                                    written.get((obj, f), 'FALSE'))
                c += codeDefine(objDot + class_.getEventQueue().smvSentTo(),
                                sentTo.get(obj, 'FALSE'))
        return c

    def __orderTransitionInstances(self):
        """Generate a total order of all transition instances.

        Return value: a list of all possible pairs (obj, t), where obj
        is an ActiveObject and t is a Transition in the state machine
        of the class of the object.
        """
        transitionInstances = []
        for class_ in self.__classes:
            objects = class_.getObjects()
            for t in class_.getStateMachine().getTransitions():
                transitionInstances += zip(objects, [t]*len(objects))
        return transitionInstances

    def elementMap(self, modelFilename):
        """Generate a mapping of model elements to UML.

        The mapping matches variables of the NuSMV input file with
        elements of the original UML model.  The mapping is coded as
        text lines that NuSMV sees as comments, so they can be
        embedded in the NuSMV input file.

        Param modelFilename: path to the XMI model file as a string.

        Return value: mapping code as a list of strings.
        """
        # Element map format:
        # --$Map Model <model file name>
        # --$Map Object <NuSMV variable name for the object>
        #         <XMI id of the class of the object>
        #         [XMI id of the object if it is in initial configuration]
        # ...
        # --$Map StateVertex <XMI id of the class containing the vertex>
        #         <XMI id of the vertex>
        #         <NuSMV id of the vertex>
        # --$Map Trans <XMI id of the class containing the transition>
        #         <XMI id of the transition>
        #         <NuSMV id of the transition>
        # --$Map TransGuardChoose
        #         <XMI id of the class containing the transition>
        #         <NuSMV id of the transition>
        #         <NuSMV variable name for a choose expression in guard>
        #         <identifying integer for the choose expression>
        # --$Map TransEffectChoose
        #         <XMI id of the class containing the transition>
        #         <NuSMV id of the transition>
        #         <NuSMV variable name for a choose expression in effect>
        #         <identifying integer for the choose expression>
        # ...
        # --$Map TransInstance <NuSMV variable name for the object>
        #         <NuSMV id of the transition>
        #
        # The transition instances are in the static E-step firing order.
        c = ['',
             '',
             '-'*70,
             '-- Mapping for tracing NuSMV variables back to UML elements',
             '-'*70,
             '',
             '--$Map Model ' + modelFilename]
        for class_ in self.__classes:
            c += class_.elementMap()
        for (actor, t) in self.__transitionInstances:
            c += ['--$Map TransInstance ' + actor.smvVariable() + ' ' +
                  t.smvId()]
        c += ['--$Map End']
        return c
