#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# smvutils.py
#
# 2006-12-04  Jori Dubrovin:  Using optparse for options
# 2006-10-30  Jori Dubrovin:  Initial version
#

"""Helper functions for producing NuSMV code."""


_atMostOneMax = 0  # Max number of arguments seen in codeAtMostOne

def smvAnd(exprList):
    """Return a NuSMV expression that is a conjunction of given expressions.

    The given expressions must be properly parenthesized as necessary,
    but the returned expression will not be parenthesized.

    Param exprList: a list of NuSMV expressions as strings.

    Return value: the conjunction expression as a string.
    """
    if exprList:
        return ' & '.join(exprList)
    else:
        return 'TRUE'

def smvOr(exprList):
    """Return a NuSMV expression that is a disjunction of given expressions.

    The given expressions must be properly parenthesized as necessary,
    but the returned expression will not be parenthesized.

    Param exprList: a list of NuSMV expressions as strings.

    Return value: the disjunction expression as a string.
    """
    if exprList:
        return ' | '.join(exprList)
    else:
        return 'FALSE'

def smvWord(value, valueCount):
    """Return a NuSMV word constant definition.

    Param value: a nonnegative integer value.

    Param valueCount: the type of the constant is the smallest word
    type that can hold at least this many distinct values.
    """
    return '0d' + str(_log2ceil(valueCount)) + '_' + str(value)

def smvWordType(valueCount):
    """Return a NuSMV word type definition.

    Param valueCount: the returned type can hold at least this many
    distinct values.
    """
    return 'word[' + str(_log2ceil(valueCount)) + ']'

def _log2ceil(n):
    """Return max(1, ceiling(log2(n)))."""
    width = 1
    while (1L << width) < n:
        width += 1
    return width

def codeDefine(symbol, expr):
    """Generate code that defines symbols.

    Param symbol: the NuSMV symbol to define.

    Param expr: a NuSMV expression whose value is assigned to symbol.

    Return value: NuSMV code as a list of strings.
    """
    c = ['  DEFINE',
         '    ' + symbol + ' := ' + expr + ';']
    return c

def codeTrans(constraint):
    """Generate code with a TRANS constraint.

    Param constraint: a NuSMV expression as strings.

    Return value: NuSMV code as a list of strings.
    """
    c = ['  TRANS',
         '    ' + constraint + ';']
    return c

def codeAtMostOne(variable, exprs):
    """Generate code that declares at-most-one constraints.

    Param variable: a variable name as a string.

    Param exprs: a list of expressions as strings, at most one of
    which must be true.

    Return value: NuSMV code as a list of strings.
    """
    # A global variable is not too elegant but should work.
    global _atMostOneMax
    _atMostOneMax = max(_atMostOneMax, len(exprs))
    c = ['  VAR',
         '    ' + variable + ' : AtMostOne' + str(len(exprs)) +
         '(' + ', '.join(exprs) + ');']
    return c

def codeAtMostOneModules():
    """Generate a required number of AtMostOneN modules.

    Return value: NuSMV code as a list of strings.
    """
    c = ['',
         '',
         '-'*70,
         '-- AtMostOne modules',
         '-'*70,
         '',
         '-- The AtMostOne modules are used to constrain a number of',
         '-- bits such that exactly zero or one of them can be true',
         '-- at the same time.']
    n = _atMostOneMax
    while n > 1:
        c += ['',
              'MODULE AtMostOne' + str(n) + '(' +
              ', '.join(['b' + str(i) for i in xrange(1, n+1)]) + ')']
        c += codeAtMostOne('child', ['b' + str(i) for i in xrange(1, n)])
        c += ['  TRANS',
              '    !(child.atLeastOne & b' + str(n) + ');',
              '  DEFINE',
              '    atLeastOne := child.atLeastOne | b' + str(n) + ';']
        n -= 1
    c += ['',
          'MODULE AtMostOne1(b1)',
          '  DEFINE',
          '    atLeastOne := b1;',
          '',
          'MODULE AtMostOne0()',
          '  DEFINE',
          '    atLeastOne := FALSE;']
    return c

def generateId(prefix, default, reserved):
    """Return a unique string identifier.

    The identifier can be used as a part of a NuSMV variable name.  If
    the global option printNames is False, the identifier contains
    only the given prefix followed by digits.

    Param prefix: a prefix string for the identifier.

    Param default: the preferred string without the prefix.

    Param reserved: an iterable whose keys are the identifiers already in use.

    Return value: a string in <prefix>[A-Za-z0-9$_]+ that is not in reserved.
    """
    stem = prefix
    if True:  # Skip this to generate numbers instead of names
        for c in default:
            if c in ('ABCDEFGHIJKLMNOPQRSTUVWXYZ' +
                     'abcdefghijklmnopqrstuvwxyz' +
                     '0123456789$'):
                stem += str(c)
            else:
                stem += '_'
    n = 0
    s = stem
    while s == prefix or s in reserved:
        n += 1
        s = stem + str(n)
    return s
