#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# statemachines.py
#
# 2007-06-20  Jori Dubrovin:  New choose expression mapping
# 2007-05-07  Jori Dubrovin:  Optimized linear Enter predicate
# 2007-05-04  Jori Dubrovin:  Added linear Enter predicate
# 2007-04-01  Jori Dubrovin:  Using simpler constraint for FireTriggered
# 2007-03-08  Jori Dubrovin:  Adding states to element map
# 2007-03-02  Jori Dubrovin:  Adding choose to element map
# 2006-12-04  Jori Dubrovin:  Fixed enabledness condition
# 2006-10-30  Jori Dubrovin:  Initial version
#

"""NuSMV encoding of state machines and their control logic."""

import exceptions

import actions
import activities
import datatypes
import errors
import options
from coralutils import uml, umlSort
from smvutils import smvAnd, smvOr
from smvutils import codeDefine, codeTrans, codeAtMostOne
from smvutils import generateId


class StateMachine:
    """A behavioral state machine."""

    def __init__(self, umlMachine, class_):
        """Construct a state machine.

        Param umlMachine: the underlying uml.StateMachine.

        Param class_: the owning ActiveClass.
        """
        assert umlMachine.oclIsKindOf(uml.StateMachine)
        self.__class = class_
        self.__regions = []           # all Regions in self
        self.__regionForUml = {}      # map uml.StateVertex to Region
        self.__vertices = []          # all Vertices in self
        self.__vertexForUml = {}      # map uml.StateVertex to Vertex
        self.__transitions = []       # all Transitions in self
        self.__transitionForUml = {}  # map uml.Transition to Transition

        if umlMachine.top.container:
            raise errors.InputError('top region of state machine ' +
                                    str(umlMachine) + ' has a container')
        self.__topRegion = self.__addRegion(umlMachine.top)
        self.__addDescendants(self.__topRegion)
        for umlTransition in umlSort(umlMachine.transitions):
            self.__addTransition(umlTransition)
        topInitial = [s for
                      s in self.__topRegion.getSubvertices() if s.isInitial()]
        if len(topInitial) != 1:
            raise errors.InputError('state machine ' + str(umlMachine) +
                                    ' does not have exactly one initial state')
        for s in self.__vertices:
            s.checkSanity()
        for t in self.__transitions:
            t.processControl()

    def __addDescendants(self, region):
        """Add all descendant vertices and regions of a Region to self."""
        for umlVertex in region.getUmlSubvertices():
            s = self.__addVertex(umlVertex)
            for umlRegion in s.getUmlRegions():
                r = self.__addRegion(umlRegion)
                self.__addDescendants(r)

    def __addRegion(self, umlRegion):
        """Add a new Region for the given uml.StateVertex to self.

        Return value: the new Region.
        """
        id_ = generateId('r', umlRegion.name,
                         [r.smvId() for r in self.__regions])
        r = Region(umlRegion, id_, self)
        self.__regions.append(r)
        assert not umlRegion in self.__regionForUml
        self.__regionForUml[umlRegion] = r
        return r

    def __addVertex(self, umlVertex):
        """Add a new Vertex for the given uml.StateVertex to self.

        Return value: the new Vertex.
        """
        id_ = generateId('s', umlVertex.name,
                         [s.smvId() for s in self.__vertices])
        s = Vertex(umlVertex, id_, self)
        self.__vertices.append(s)
        assert not umlVertex in self.__vertexForUml
        self.__vertexForUml[umlVertex] = s
        return s

    def __addTransition(self, umlTransition):
        """Add a new Transition for the given uml.Transition to self.

        Return value: the new Transition.
        """
        id_ = generateId('t', umlTransition.name,
                         [t.smvId() for t in self.__transitions])
        t = Transition(umlTransition, id_, self)
        self.__transitions.append(t)
        assert not umlTransition in self.__transitionForUml
        self.__transitionForUml[umlTransition] = t
        return t

    def getClass(self):
        """Return the owning ActiveClass."""
        return self.__class

    def getVertices(self):
        """Return all Vertices in self."""
        return self.__vertices[:]

    def getTransitions(self):
        """Return all transitions in self."""
        return self.__transitions[:]

    def getRegionForUml(self, umlRegion):
        """Return the Region (owned by self) for a given uml.StateVertex."""
        return self.__regionForUml[umlRegion]

    def getVertexForUml(self, umlVertex):
        """Return the Vertex (owned by self) for a given uml.StateVertex."""
        return self.__vertexForUml[umlVertex]

    def getTransitionForUml(self, umlTransition):
        """Return the Transition (owned by self) for a given uml.Transition."""
        return self.__transitionForUml[umlTransition]

    def smvInPseudostate(self):
        """Return the predicate InPseudostate."""
        return 'inPseudostate'

    def smvInRtc(self):
        """Return the predicate InRtc."""
        return 'inRtc'

    def smvReady(self):
        """Return the predicate Ready."""
        return 'ready'

    def smvScheduled(self):
        """Return the predicate Scheduled."""
        return 'scheduled'

    def smvAnyFeasible(self, signal):
        """Return the predicate AnyFeasible."""
        return 'anyFeasible#' + signal.smvId()

    def smvAnyDeferring(self, signal):
        """Return the predicate AnyDeferring."""
        return 'anyDeferring#' + signal.smvId()

    def code(self):
        """Generate NuSMV code for the state machine.

        The code goes to the module of the associated class.

        Return value: NuSMV code as a list of strings.
        """
        c = ['  -- State machine']
        signals = self.__class.getEventQueue().getRelevantSignals()
        c += codeDefine(self.smvInPseudostate(),
                        smvOr([s.smvActive() for
                               s in self.__vertices if s.isPseudostate()]))
        c += codeDefine(self.smvInRtc(),
                        smvOr([self.smvInPseudostate()] +
                              [s.smvCompleted() for
                               s in self.__vertices if not s.isPseudostate()]))
        ready = ([self.smvInRtc()] +
                 ['!' + self.__class.getEventQueue().smvInputQueueEmpty()] +
                 [self.smvAnyFeasible(e) for e in signals if e.isExternal()])
        c += codeDefine(self.smvReady(), smvOr(ready))

        pseudoFire = [t.smvFire() for t in self.__transitions if
                      t.getSource().isPseudostate()]
        busyFire = [t.smvFire() for t in self.__transitions if
                    not t.getSource().isPseudostate() and t.isCompletion()]
        quiesce = [s.smvQuiesce() for s in self.__vertices if
                   s.isCompletionSensitive()]
        if options.opt.encode_interleaving:
            c += codeAtMostOne('pseudoCompletion', pseudoFire)
            c += codeAtMostOne('busyCompletion', busyFire + quiesce)
            c += codeDefine('consumeCompletion',
                            smvOr(['pseudoCompletion.atLeastOne',
                                   'busyCompletion.atLeastOne']))
        else:
            c += codeDefine('consumeCompletion',
                            smvOr(pseudoFire + busyFire + quiesce))
        c += codeDefine(self.smvScheduled(),
                        smvOr(['consumeCompletion',
                               self.__class.getEventQueue().smvDispatchAny()]))

        vertices = [s for s in self.__topRegion.getSubvertices() if
                    not s.isPseudostate()]
        for e in signals:
            # This is analogous to the definitions of
            # DescendantFeasible and DescendantDeferring in
            # Vertex.__codeSignalSpecific.
            c += codeDefine(self.smvAnyFeasible(e),
                            smvOr([smvOr([s.smvFeasibleTriggered(e),
                                          s.smvDescendantFeasible(e)]) for
                                   s in vertices]))
            c += codeDefine(self.smvAnyDeferring(e),
                            smvOr([smvOr([s.smvDeferring(e),
                                          s.smvDescendantDeferring(e)]) for
                                   s in vertices]))
        for s in self.__vertices:
            c += s.code()
        for r in self.__regions:
            c += r.code()
        for t in self.__transitions:
            c += t.code()
        return c


class Region:
    """A region in a state machine."""

    def __init__(self, umlRegion, id_, machine):
        """Construct a region.

        Param umlRegion: the underlying uml.StateVertex.

        Param id_: an id string.

        Param machine: the StateMachine owning the region.
        """
        assert umlRegion.oclIsKindOf(uml.StateVertex)
        self.__umlRegion = umlRegion
        self.__id = id_
        self.__machine = machine
        self.__containedTransitions = []  # Initially empty
        self.__breakingTransitions = []  # Initially empty
        if not umlRegion.oclIsKindOf(uml.CompositeState):
            raise errors.InputError('region ' + str(umlRegion) +
                                    ' is not composite')
        if umlRegion.isConcurrent:
            raise errors.InputError('region ' + str(umlRegion) +
                                    ' is concurrent')
        if umlRegion.oclIsKindOf(uml.SubmachineState):
            raise errors.NotSupportedError('submachine state ' +
                                           str(umlRegion) +
                                           ' is not supported')
        if umlRegion.internalTransition:
            raise errors.NotSupportedError('region ' + str(umlRegion) +
                                           ' has an internal transition')
        if umlRegion.entry:
            raise errors.NotSupportedError('region ' + str(umlRegion) +
                                           ' has an entry activity')
        if umlRegion.exit:
            raise errors.NotSupportedError('region ' + str(umlRegion) +
                                           ' has an exit activity')
        if umlRegion.doActivity:
            raise errors.NotSupportedError('region ' + str(umlRegion) +
                                           ' has a do activity')
        if not umlRegion.container or umlRegion.container.isConcurrent:
            # self.__umlRegion is top or a region in a concurrent
            # composite state.  Therefore it must not have incoming or
            # outgoing transitions.
            if umlRegion.incoming:
                raise errors.NotSupportedError('region ' + str(umlRegion) +
                                               ' has an incoming transition')
            if umlRegion.outgoing:
                raise errors.NotSupportedError('region ' + str(umlRegion) +
                                               ' has an outgoing transition')

    def getUmlSubvertices(self):
        """Return a list of direct subvertices.

        Return value: a list of uml.StateVertex.
        """
        return umlSort(self.__umlRegion.subvertex)

    def getSubvertices(self):
        """Return a list of direct subvertices.

        Return value: a list of Vertex.
        """
        return [self.__machine.getVertexForUml(s) for
                s in self.getUmlSubvertices()]

    def getInitial(self):
        """Return the initial pseudostate (Vertex) in self."""
        initial = [s for s in self.getSubvertices() if s.isInitial()]
        if len(initial) != 1:
            raise errors.InputError('region ' + str(self.__umlRegion()) +
                                    ' does not have exactly one ' +
                                    'initial pseudostate')
        return initial[0]

    def getContainingState(self):
        """Return the containing composite state (Vertex)."""
        assert not self.isTop()
        if self.__umlRegion.container.isConcurrent:
            return self.__machine.getVertexForUml(self.__umlRegion.container)
        else:
            return self.__machine.getVertexForUml(self.__umlRegion)

    def isTop(self):
        """Return True if self is the top region of a state machine."""
        if self.__umlRegion.container:
            return False
        return True

    def addContainedTransition(self, t):
        """Add a Transition that whose lca(source, target) is self."""
        self.__containedTransitions += [t]

    def addBreakingTransition(self, t):
        """Add a Transition that that breaks in self."""
        self.__breakingTransitions += [t]

    def smvId(self):
        """Return the id string."""
        return self.__id

    def smvBreakIn(self):
        """Return the predicate BreakIn."""
        return 'breakIn#' + self.__id

    def smvFinished(self):
        """Return a NuSMV expression for the finished status of self.

        A region is finished if one of its direct final substates is active
        and quiescent.
        """
        return '(' + smvOr([s.smvActive() for
                            s in self.getSubvertices() if s.isFinal()]) + ')'

    def code(self):
        """Generate code for region specific control logic.

        Return value: NuSMV code as a list of strings.
        """
        c = ['  -- Region ' + self.__id]
        if self.__breakingTransitions:
            sub = [s.smvEnter() for
                   s in self.getSubvertices() if not s.isInitial()]
            cont = [t.smvFire() for t in self.__containedTransitions]
            breakIn = smvAnd(['(' + smvOr(sub) + ')',
                              '!(' + smvOr(cont) + ')'])
        else:
            breakIn = 'FALSE'
        c += codeDefine(self.smvBreakIn(), breakIn)
        return c


class Vertex:
    """A state vertex in a state machine."""

    def __init__(self, umlVertex, id_, machine):
        """Construct a state vertex.

        Param umlVertex: a uml.StateVertex.

        Param id_: an id string.

        Param machine: the StateMachine owning the state vertex.
        """
        assert umlVertex.oclIsKindOf(uml.StateVertex)
        self.__umlVertex = umlVertex
        self.__id = id_
        self.__machine = machine
        self.__enteringTransitions = []  # Initially empty
        self.__escapingTransitions = []  # Initially empty
        if not (umlVertex.oclIsKindOf(uml.SimpleState) or
                umlVertex.oclIsKindOf(uml.FinalState) or
                (umlVertex.oclIsKindOf(uml.CompositeState) and
                 not umlVertex.oclIsKindOf(uml.SubmachineState)) or
                (umlVertex.oclIsKindOf(uml.Pseudostate) and
                 umlVertex.kind in ['initial', 'choice'])):
            raise errors.NotSupportedError('unsupported state vertex ' +
                                           str(umlVertex))
        if umlVertex.oclIsKindOf(uml.State):
            if umlVertex.internalTransition:
                raise errors.NotSupportedError('state ' + str(umlVertex) +
                                               ' has an internal transition')
            if umlVertex.entry:
                raise errors.NotSupportedError('state ' + str(umlVertex) +
                                               ' has an entry activity')
            if umlVertex.exit:
                raise errors.NotSupportedError('state ' + str(umlVertex) +
                                               ' has an exit activity')
            if umlVertex.doActivity:
                raise errors.NotSupportedError('state ' + str(umlVertex) +
                                               ' has a do activity')

    def getXmiId(self):
        """Return the XMI id of the underlying uml.StateVertex as a string."""
        return str(self.__umlVertex.__XMIid__())

    def checkSanity(self):
        """Perform sanity checks in the model and raise if errors found."""
        try:
            for umlTransition in self.__umlVertex.incoming:
                t = self.__machine.getTransitionForUml(umlTransition)
                if self.isInitial():
                    raise errors.InputError('initial pseudostate ' +
                                            str(self.__umlVertex) +
                                            ' has an incoming transition')
            for umlTransition in self.__umlVertex.outgoing:
                t = self.__machine.getTransitionForUml(umlTransition)
                if self.isFinal():
                    raise errors.InputError('final state ' +
                                            str(self.__umlVertex) +
                                            ' has an outgoing transition')
                if self.isPseudostate() and not t.isCompletion():
                    raise errors.InputError('pseudostate ' +
                                            str(self.__umlVertex) +
                                            ' has an outgoing ' +
                                            'signal-triggered transition')
        except exceptions.LookupError:
            raise errors.InputError('state ' + str(self.__umlVertex) +
                                    ' has a dangling transition')

    def isPseudostate(self):
        """Return True if self is a pseudostate."""
        return self.__umlVertex.oclIsKindOf(uml.Pseudostate)

    def isInitial(self):
        """Return True if self is an initial pseudostate."""
        return (self.__umlVertex.oclIsKindOf(uml.Pseudostate) and
                self.__umlVertex.kind == 'initial')

    def isFinal(self):
        """Return True if self is a final state."""
        return self.__umlVertex.oclIsKindOf(uml.FinalState)

    def getUmlRegions(self):
        """Return the direct subregions.

        Return value: a list of uml.StateVertex.
        """
        if self.__umlVertex.oclIsKindOf(uml.CompositeState):
            if self.__umlVertex.isConcurrent:
                return umlSort(self.__umlVertex.subvertex)
            else:
                return [self.__umlVertex]
        else:
            return []

    def getRegions(self):
        """Return the direct subregions.

        Return value: a list of Regions.
        """
        return [self.__machine.getRegionForUml(umlRegion) for
                umlRegion in self.getUmlRegions()]

    def getSubvertices(self):
        """Return the direct subvertices in the direct subregions.

        Return value: a list of Vertex.
        """
        sub = []
        for r in self.getRegions():
            sub.extend(r.getSubvertices())
        return sub

    def getContainingRegion(self):
        """Return the containing Region."""
        return self.__machine.getRegionForUml(self.__umlVertex.container)

    def getContainingState(self):
        """Return the containing composite state (Vertex)."""
        return self.getContainingRegion().getContainingState()

    def getOutgoing(self):
        """Return a list of outgoing Transitions."""
        return [self.__machine.getTransitionForUml(umlTransition) for
                umlTransition in umlSort(self.__umlVertex.outgoing)]

    def isCompletionSensitive(self):
        """Return True if self is completion sensitive.

        A state vertex is completion sensitive if it is a
        non-pseudostate and has outgoing completion transitions.
        """
        if self.isPseudostate():
            return False
        for t in self.getOutgoing():
            if t.isCompletion():
                return True
        return False

    def defers(self, signal):
        """Return True if self declares the given Signal deferrable."""
        if self.isPseudostate():
            return False
        deferrableUmlSignals = [umlEvent.signal for
                                umlEvent in self.__umlVertex.deferrableEvent]
        return signal.isInUmlSignals(deferrableUmlSignals)

    def addEnteringTransition(self, t):
        """Add a Transition that enters self."""
        self.__enteringTransitions += [t]

    def addEscapingTransition(self, t):
        """Add a Transition that escapes self.

        A transition t escapes a state vertex s if t exits s and t
        does not exit any proper ancestor of s.
        """
        self.__escapingTransitions += [t]

    def smvId(self):
        """Return the id string."""
        return self.__id

    def smvActive(self):
        """Return the predicate Active."""
        return 'active#' + self.__id

    def smvEnter(self):
        """Return the predicate Enter."""
        return 'enter#' + self.__id

    def smvExit(self):
        """Return the predicate Exit."""
        return 'exit#' + self.__id

    def smvQuiescent(self):
        """Return the predicate Quiescent."""
        return 'quiescent#' + self.__id

    def smvCompleted(self):
        """Return the predicate Completed."""
        return 'completed#' + self.__id

    def smvCanQuiesce(self):
        """Return the predicate CanQuiesce."""
        return 'canQuiesce#' + self.__id

    def smvQuiesce(self):
        """Return the predicate Quiesce."""
        return 'quiesce#' + self.__id

    def smvDeferring(self, signal):
        """Return the predicate Deferring."""
        return 'deferring#' + self.__id + '#' + signal.smvId()

    def smvFireTriggered(self, signal):
        """Return the predicate FireTriggered."""
        return 'fireTriggered#' + self.__id + '#' + signal.smvId()

    def smvFeasibleTriggered(self, signal):
        """Return the predicate FeasibleTriggered."""
        return 'feasibleTriggered#' + self.__id + '#' + signal.smvId()

    def smvPriorityEnabled(self, signal):
        """Return the predicate PriorityEnabled."""
        return 'priorityEnabled#' + self.__id + '#' + signal.smvId()

    def smvDescendantFeasible(self, signal):
        """Return the predicate DescendantFeasible."""
        return 'descendantFeasible#' + self.__id + '#' + signal.smvId()

    def smvDescendantDeferring(self, signal):
        """Return the predicate DescendantDeferring."""
        return 'descendantDeferring#' + self.__id + '#' + signal.smvId()

    def code(self):
        """Generate code for state vertex specific control logic.

        Return value: NuSMV code as a list of strings.
        """
        c = ['  -- State vertex ' + self.__id]
        if self.getContainingRegion().isTop() and self.isInitial():
            initActive = self.__machine.getClass().smvAlive()
            enter = self.__machine.getClass().smvCreated()
            assert not self.__enteringTransitions
        else:
            initActive = 'FALSE'
            if options.opt.old_enter:
                enter = smvOr([t.smvFire() for
                               t in self.__enteringTransitions])
            elif self.isInitial():
                enter = smvAnd([self.getContainingState().smvEnter(),
                                '!' + self.getContainingRegion().smvBreakIn()])
            else:
                enter = smvOr([t.smvFire() for
                               t in self.__machine.getTransitions() if
                               t.getTarget() is self] +
                              [r.smvBreakIn() for
                               r in self.getRegions()])
        c += ['  VAR',
              '    ' + self.smvActive() + ' : boolean;',
              '  ASSIGN',
              '    init(' + self.smvActive() + ') := ' + initActive + ';',
              '    next(' + self.smvActive() + ') := ' +
              smvOr([self.smvEnter(),
                     '(' + smvAnd([self.smvActive(),
                                   '!' + self.smvExit()]) + ')']) + ';']
        exitReasons = [t.smvFire() for t in self.__escapingTransitions]
        if self.getContainingRegion().isTop():
            exitReasons += [self.__machine.getClass().smvDestroyed()]
        else:
            exitReasons += [self.getContainingState().smvExit()]
        c += codeDefine(self.smvEnter(), enter)
        c += codeDefine(self.smvExit(),
                        'exitReason#' + self.__id + '.atLeastOne')
        c += codeAtMostOne('exitReason#' + self.__id, exitReasons)
        if not self.isPseudostate():
            if self.isCompletionSensitive():
                c += ['  VAR',
                      '    ' + self.smvQuiescent() + ' : boolean;',
                      '  ASSIGN',
                      '    init(' + self.smvQuiescent() + ') := FALSE;',
                      '    next(' + self.smvQuiescent() + ') := ' +
                      smvAnd(['!' + self.smvExit(),
                              '(' + smvOr([self.smvQuiescent(),
                                           self.smvQuiesce()]) + ')']) + ';']
                c += ['  IVAR',
                      '    ' + self.smvQuiesce() + ' : boolean;']
                canQuiesce = smvAnd(['!' + self.__machine.smvInPseudostate(),
                                     self.smvCompleted()] +
                                    ['!' + t.smvEvalGuard() for
                                     t in self.getOutgoing() if
                                     t.isCompletion()])
                # Could be optimized.
                c += codeDefine(self.smvCanQuiesce(), canQuiesce)
                c += codeTrans(self.smvQuiesce() + ' -> ' +
                               self.smvCanQuiesce())
                compl = smvAnd([self.smvActive(), '!' + self.smvQuiescent()] +
                               [r.smvFinished() for r in self.getRegions()])
            else:
                compl = 'FALSE'
            c += codeDefine(self.smvCompleted(), compl)
            queue = self.__machine.getClass().getEventQueue()
            for signal in queue.getRelevantSignals():
                c += self.__codeSignalSpecific(signal)
        return c

    def __codeSignalSpecific(self, e):
        """Generate code for vertex and signal specific control logic.

        Param e: a Signal.

        Return value: NuSMV code as a list of strings.
        """
        assert not self.isPseudostate()
        c = []
        transitions = [t for t in self.getOutgoing() if
                       t.getTriggerSignal() is e]
        c += codeDefine(self.smvFireTriggered(e),
                        smvOr([t.smvFire() for t in transitions]))
        c += codeDefine(self.smvFeasibleTriggered(e),
                        smvAnd([self.smvActive(),
                                '!' + self.smvDescendantFeasible(e),
                                '!' + self.smvDescendantDeferring(e),
                                '(' + smvOr([t.smvEvalGuard() for
                                             t in transitions]) + ')']))
        queue = self.__machine.getClass().getEventQueue()
        c += codeDefine(self.smvPriorityEnabled(e),
                        smvAnd([queue.smvDispatch(e),
                                self.smvFeasibleTriggered(e)]))
        if self.defers(e):
            deferring = self.smvActive()
        else:
            deferring = 'FALSE'
        c += codeDefine(self.smvDeferring(e), deferring)
        vertices = [s for s in self.getSubvertices() if
                    not s.isPseudostate()]
        c += codeDefine(self.smvDescendantFeasible(e),
                        smvOr([smvOr([s.smvFeasibleTriggered(e),
                                      s.smvDescendantFeasible(e)]) for
                               s in vertices]))
        c += codeDefine(self.smvDescendantDeferring(e),
                        smvOr([smvOr([s.smvDeferring(e),
                                      s.smvDescendantDeferring(e)]) for
                               s in vertices]))
        #c += codeTrans(self.smvFireTriggered(e) + ' -> ' +
        #               self.smvPriorityEnabled(e))
        #c += codeTrans(self.smvPriorityEnabled(e) + ' -> ' + self.smvExit())
        #
        # The constraints above work even for simultaneous firing in
        # orthogonal regions.  The one below does not, but is simpler.
        c += codeTrans(self.smvFireTriggered(e) + ' <-> ' +
                       self.smvPriorityEnabled(e))
        return c


class Transition:
    """A transition in a state machine."""

    def __init__(self, umlTransition, id_, machine):
        """Construct a transition.

        Param umlTransition: a uml.Transition.

        Param id_: an id string.

        Param machine: the StateMachine owning the transition.
        """
        assert umlTransition.oclIsKindOf(uml.Transition)
        self.__umlTransition = umlTransition
        self.__id = id_
        self.__machine = machine
        umlTrigger = self.__umlTransition.trigger
        if umlTrigger and not umlTrigger.oclIsKindOf(uml.SignalEvent):
            raise errors.NotSupportedError('unsupported trigger ' +
                                           str(umlTrigger))

    def getXmiId(self):
        """Return the XMI id of the underlying uml.Transition as a string."""
        return str(self.__umlTransition.__XMIid__())

    def getSource(self):
        """Return the source Vertex."""
        try:
            return self.__machine.getVertexForUml(self.__umlTransition.source)
        except exceptions.LookupError:
            raise errors.InputError('dangling source in transition ' +
                                    str(self.__umlTransition))

    def getTarget(self):
        """Return the target Vertex."""
        try:
            return self.__machine.getVertexForUml(self.__umlTransition.target)
        except exceptions.LookupError:
            raise errors.InputError('dangling target in transition ' +
                                    str(self.__umlTransition))

    def isCompletion(self):
        """Return True if self is a completion transition."""
        if self.__umlTransition.trigger:
            return False
        return True

    def getTriggerSignal(self):
        """Return the triggering Signal or None."""
        if self.isCompletion():
            return None
        umlTrigger = self.__umlTransition.trigger
        try:
            return self.__machine.getClass().getSignalForUml(umlTrigger.signal)
        except exceptions.LookupError:
            raise errors.InputError('unknown trigger in transition ' +
                                    str(self.__umlTransition))

    def getTriggerParamFields(self):
        """Return the Fields to which the signal parameters are read."""
        signal = self.getTriggerSignal()
        if signal is None:
            return []
        class_ = self.__machine.getClass()
        try:
            # umlSort must not be used for trigger.parameter.
            fields = [class_.getFieldForName(umlParam.name) for
                      umlParam in self.__umlTransition.trigger.parameter]
        except exceptions.LookupError:
            raise errors.InputError('unknown field in the signal parameter ' +
                                    'list of transition ' +
                                    str(self.__umlTransition))
        types = signal.getParamTypes()
        if len(types) != len(fields):
            raise errors.InputError('signal parameter list length mismatch '
                                    'in transition ' +
                                    str(self.__umlTransition))
        for (f, type_) in zip(fields, types):
            if fields.count(f) != 1:
                raise errors.InputError('duplicate field in the signal ' +
                                        'parameter list of transition ' +
                                        str(self.__umlTransition))
            if not type_.isSubtypeOf(f.getType()):
                raise errors.InputError('signal parameter type mismatch '
                                        'in transition ' +
                                        str(self.__umlTransition))
        return fields

    def getActivity(self):
        """Return the Activity of the transition.

        processActivity must have been called.
        """
        return self.__activity

    def getGuardChooseExprIds(self):
        """Return expression identifiers for choose expressions in the guard.

        processActivity must have been called.

        Return value: a mapping from choose key strings to expression
        identifiers in the context of the owning class.
        """
        return dict(self.__guardChooseExprIds)

    def getEffectChooseExprIds(self):
        """Return expression identifiers for choose expressions in the effect.

        processActivity must have been called.

        Return value: a mapping from choose key strings to expression
        identifiers in the context of the owning class.
        """
        return self.__activity.getChooseExprIds()

    def processControl(self):
        """Determine how the state machine control logic is affected by self.

        This is done by notifying all Vertices that are entered or
        exited when self is fired.
        """
        # Determine the least common ancestor (lca) of source
        # and target state vertices
        sourcePath = [self.getSource()]
        while not sourcePath[0].getContainingRegion().isTop():
            sourcePath.insert(0, sourcePath[0].getContainingState())
        targetPath = [self.getTarget()]
        while not targetPath[0].getContainingRegion().isTop():
            targetPath.insert(0, targetPath[0].getContainingState())
        while (len(sourcePath) >= 2 and len(targetPath) >= 2 and
               sourcePath[1].getContainingRegion() is
               targetPath[1].getContainingRegion()):
            sourcePath = sourcePath[1:]
            targetPath = targetPath[1:]
        assert (sourcePath[0].getContainingRegion() is
                targetPath[0].getContainingRegion())
        sourcePath[0].getContainingRegion().addContainedTransition(self)
        # Now sourcePath (targetPath) is a list of nested state
        # vertices from the lca to the source (target).  Every state
        # vertex in sourcePath (plus more, possibly) is exited by
        # self.  Every state vertex in targetPath (plus more,
        # possibly) is entered by self.
        sourcePath[0].addEscapingTransition(self)
        # Walk the path from the lca to target and add all state
        # vertices to the enters set.  Also add the initial state of
        # every region that is orthogonal to the path.
        while targetPath:
            s = targetPath[0]
            s.addEnteringTransition(self)
            for r in s.getRegions():
                if (len(targetPath) >= 2 and
                    r is targetPath[1].getContainingRegion()):
                    # r is a region such that the target state is a
                    # nested substate of r.  Firing self breaks in r.
                    r.addBreakingTransition(self)
                else:
                    # r is a region orthogonal to the target state
                    # vertex or a direct subregion of the target.
                    # Firing self enters the initial pseudostate in r.
                    r.getInitial().addEnteringTransition(self)
            targetPath = targetPath[1:]
        assert s is self.getTarget()

    def processActivity(self):
        """Process the signal parameters, guard, and effect of self."""
        # This function sets self.__activity and self.__guardExprId.
        class_ = self.__machine.getClass()
        context = class_.getActivityContext()
        queue = class_.getEventQueue()
        # By default, the activity of self does nothing.
        self.__activity = activities.Activity(context)
        signal = self.getTriggerSignal()
        if signal is not None:
            # Notify the event queue.
            queue.addTriggerSignal(signal)
            # Update the activity to reflect the effect of reading the
            # signal parameter values to the fields of this object.
            paramValues = queue.smvSignalParamValues(signal)
            for (f, value) in zip(self.getTriggerParamFields(), paramValues):
                if signal.isExternal():
                    raise errors.NotSupportedError('external signal ' +
                                                   'parameters not supported')
                self.__activity.writeField(f, context.smvThisExprId(),
                                           value, 'signal parameter')
        umlGuard = self.__umlTransition.guard
        if umlGuard:
            # Evaluate the guard in a context where the signal
            # parameters are already assigned to fields.
            guard = str(umlGuard.expression.body)
            self.__guardExprId = actions.processExpr(guard, datatypes.boolType,
                                                     self.__activity)
            self.__activity.addError(self.__guardExprId + '.error')
        else:
            self.__guardExprId = None
        self.__guardChooseExprIds = self.__activity.getChooseExprIds()
        self.__activity.clearChooseExprIds()
        umlEffect = self.__umlTransition.effect
        if umlEffect:
            # Update the activity to reflect the action language
            # statements in the effect of self.  As a side effect,
            # notify classes about the signals that are sent to them.
            effect = str(umlEffect.script.body)
            actions.processActions(effect, self.__activity)

    def smvId(self):
        """Return the id string."""
        return self.__id

    def smvFire(self):
        """Return the predicate Fire."""
        return 'fire#' + self.__id

    def smvEvalGuard(self):
        """Return the predicate EvalGuard."""
        return 'evalGuard#' + self.__id

    def smvEnabled(self):
        """Return the predicate Enabled."""
        return 'enabled#' + self.__id

    def smvAssertion(self):
        """Return the predicate Assertion."""
        return 'assertion#' + self.__id

    def smvRuntimeError(self):
        """Return the predicate RuntimeError."""
        return 'runtimeError#' + self.__id

    def smvRead(self, obj, field):
        """Return the predicate Read."""
        return '#'.join(['read', self.__id, obj.smvId(), field.smvId()])

    def smvWrite(self, obj, field):
        """Return the predicate Write."""
        return '#'.join(['write', self.__id, obj.smvId(), field.smvId()])

    def smvSend(self, obj):
        """Return the predicate Send."""
        return 'send#' + self.__id + '#' + obj.smvId()

    def code(self):
        """Generate code for the transition.

        Return value: NuSMV code as a list of strings.
        """
        c = ['  -- Transition ' + self.__id]
        c += self.__codeControl()
        c += self.__codeActivity()
        return c

    def __codeControl(self):
        """Generate code for transition specific control logic.

        Return value: NuSMV code as a list of strings.
        """
        c = []
        c += ['  IVAR',
              '    ' + self.smvFire() + ' : boolean;']
        if self.isCompletion():
            source = self.getSource()
            if source.isPseudostate():
                enabled = smvAnd([source.smvActive(), self.smvEvalGuard()])
            else:
                enabled = smvAnd(['!' + self.__machine.smvInPseudostate(),
                                  source.smvCompleted(), self.smvEvalGuard()])
                # Could be optimized.
            c += codeDefine(self.smvEnabled(), enabled)
            c += codeTrans(self.smvFire() + ' -> ' + self.smvEnabled())
        else:
            c += codeTrans(self.smvFire() + ' -> ' + self.smvEvalGuard())
            # Could be optimized.
        return c

    def __codeActivity(self):
        """Generate code for the activity of the transition.

        Return value: NuSMV code as a list of strings.
        """
        c = []
        # Define the Guard predicate.
        if self.__guardExprId is None:
            c += codeDefine(self.smvEvalGuard(), 'TRUE')
        else:
            c += codeDefine(self.smvEvalGuard(),
                            smvOr([self.__guardExprId + '.value',
                                   self.__guardExprId + '.error']))
        # Define the Read predicate for every instance of all fields
        # read by self.
        for read in self.__activity.getFieldReads():
            f = read.getField()
            for target in f.getOwner().getObjects():
                if read.isLocal() or not options.opt.encode_static:
                    readingTarget = ['(' + e + '.value = ' +
                                     target.smvReference() + ')' for
                                     e in read.smvRefExprIds()]
                    c += codeDefine(self.smvRead(target, f),
                                    smvAnd([self.smvFire(),
                                            '(' + smvOr(readingTarget) + ')']))
                else:
                    c += codeDefine(self.smvRead(target, f), self.smvFire())
        # Define the Write predicate for every instance of all fields
        # written by self.
        for write in self.__activity.getFieldWrites():
            f = write.getField()
            for target in f.getOwner().getObjects():
                if write.isLocal() or not options.opt.encode_static:
                    writingTarget = ['(' + e + '.value = ' +
                                     target.smvReference() + ')' for
                                     e in write.smvRefExprIds()]
                    c += codeDefine(self.smvWrite(target, f),
                                    smvAnd([self.smvFire(),
                                            '(' + smvOr(writingTarget) + ')']))
                else:
                    c += codeDefine(self.smvWrite(target, f), self.smvFire())
                # If Write is true, then the next value of target.f
                # must be equal to the value of target.f in the
                # valuation obtained by executing the actions of the
                # transition.
                c += codeTrans(self.smvWrite(target, f) + ' -> (next(' +
                               target.smvFieldValue(f, 'global') + ') = ' +
                               target.smvFieldValue(f, write.smvValuation()) +
                               ')')
                # Could be optimized in some cases by replacing
                # valuation by expression.
        for send in self.__activity.getSends():
            for target in send.getTargetClass().getObjects():
                toTarget = ('(' + send.smvTargetExprId() + '.value = ' +
                            target.smvReference() + ')')
                queue = send.getTargetClass().getEventQueue()
                reception = queue.smvReception(target, send.getSignal(),
                                               send.smvParamExprIds())
                noReception = queue.smvReception(target, None, [])
                if send.isLocal() or not options.opt.encode_static:
                    # Send(self, target) is true iff self is fired and
                    # the reference expression in the send statement
                    # evaluates to target.  Send(self, target) implies
                    # reception of the signal in target.
                    c += codeDefine(self.smvSend(target),
                                    smvAnd([self.smvFire(), toTarget]))
                    c += codeTrans(self.smvSend(target) + ' -> ' + reception)
                else:
                    # Send(self, target) is true iff self is fired,
                    # regardless of the reference expression in the
                    # send statement.  Dynamically, if the reference
                    # expression evaluates to target, then Send(self,
                    # target) implies reception of the signal in
                    # target.  If not, then Send(self, target) implies
                    # that target does not receive any signal.
                    c += codeDefine(self.smvSend(target), self.smvFire())
                    c += codeTrans('(' + smvAnd([self.smvSend(target),
                                                 toTarget]) +
                                   ') -> ' + reception)
                    c += codeTrans('(' + smvAnd([self.smvSend(target),
                                                 '!' + toTarget]) +
                                   ') -> ' + noReception)
        withinResources = self.__machine.getClass().smvWithinGlobalResources()
        if self.__activity.getAssertions():
            c += codeDefine(self.smvAssertion(),
                            smvAnd(self.__activity.getAssertions()))
            if options.opt.check_assertions:
                c += ['  LTLSPEC',
                      '    G ((' + smvAnd([withinResources, self.smvFire()]) +
                      ') -> ' + self.smvAssertion() + ');']
        if self.__activity.getErrors():
            c += codeDefine(self.smvRuntimeError(),
                            smvOr(self.__activity.getErrors()))
            if options.opt.check_runtime_errors:
                c += ['  LTLSPEC',
                      '    G ((' + smvAnd([withinResources, self.smvFire()]) +
                      ') -> !' + self.smvRuntimeError() + ');']
        return c
