"""
Some auxiliary helpers for action language manipulation.

Author: Tommi Junttila

Copyright (c) 2006-2007 Helsinki University of Technology
Released under the GNU General Public License version 2
"""

import sets
import sys

import actlang


def remove_assertions(code):
    """
    Remove assertion statements from the (possibly compound)
    Jumbala statement 'code'.
    Return the resulting code as a string.
    """
    assert type(code) is str

    def rec(tree):
        if tree.__class__ == actlang.statement.AssertStatement:
            return ";"
        return tree.api_recurseStr(rec)

    try:
        root = actlang.grammar.parse(code)
    except actlang.errors.Error, e:
        if False:
            debug_filename = "DEBUG.jmb"
            f = open(debug_filename, 'w')
            f.write(code)
            f.close()
        raise e
    
    stmts = root.api_getTopStatements()
    result = "\n".join([rec(stmt) for stmt in stmts])
    return result


def condition_assertions(code, c):
    """
    Condition assertion statements in the (possibly compound)
    Jumbala statement 'code' with condition 'c', i.e. rewrite
    'assert(e);' to 'assert(!(c) || (e));'.
    Return the resulting code as a string.
    """
    assert type(code) is str
    assert code != ""
    assert type(c) is str
    assert c != ""

    def rec(tree):
        if tree.__class__ == actlang.statement.AssertStatement:
            return "assert(!("+c+") || ("+rec(tree.api_getExpr())+"));"
        return tree.api_recurseStr(rec)

    try:
        root = actlang.grammar.parse(code)
    except actlang.errors.Error, e:
        if True:
            debug_filename = "DEBUG.jmb"
            f = open(debug_filename, 'w')
            f.write(code)
            f.close()
        raise e
    
    stmts = root.api_getTopStatements()
    result = "\n".join([rec(stmt) for stmt in stmts])
    return result




def stmt_contextifier(stmt_str,
                      this_str,
                      class_names,
                      global_function_names,
                      global_variable_names):
    """
    Replaces all explicit and implicit references to 'this' in
    'stmt_str' with 'this_str'.
    'stmt_str' and 'this_str' must be strings.
    'class_names', 'global_function_names', and 'global_variable_names'
    must be lists, dictionaries, or sets containing the names
    of the classes, global functions, and global variables, respectively.
    Return statements are rewritten so that they stop the excution of the code
    and provide the returned value through the static methods of
    the built-in $$$reader class.
    'stmt_str' shall not contain any identifiers starting with the dollar sign
    except for the attribute and method accesses in the built-in
    class $$SMUML of the form $$SMUML.xxx.

    As an example,
    stmt_contextifier("y = f() + this.f() + A.f() + c.f() + g.f();", "c",
                      ["A"],["f"],["g"])
    returns "c.y = f() + c.f() + A.f() + c.c.f() + g.f();".
    while
    stmt_contextifier("if(x) {return 1;} else {return y;}", "c", [], [], [])
    returns the code
    $pseudocontext: {
      if (c.x) {
        $$$reader.read(1); break $pseudocontext;
      } else {
        $$$reader.read(c.y); break $pseudocontext;
      }
    }
    """
    assert type(stmt_str) is str
    assert type(this_str) is str
    assert((type(class_names) is list) or
           (type(class_names) is dict) or
           isinstance(class_names, sets.Set))
    assert((type(global_function_names) is list) or
           (type(global_function_names) is dict) or
           isinstance(global_function_names, sets.Set))
    assert((type(global_variable_names) is list) or
           (type(global_variable_names) is dict) or
           isinstance(global_variable_names, sets.Set))

    def rec(tree):
        #print tree.__class__, str(tree)
        if isinstance(tree, (actlang.statement.AssertStatement
                             ,actlang.statement.Block
                             ,actlang.statement.ExprStatement
                             ,actlang.statement.IfStatement
                             ,actlang.statement.LabeledStatement
                             ,actlang.statement.SendStatement
                             ,actlang.statement.WhileStatement
                             ,actlang.expr.ArrayAccessExpr
                             ,actlang.expr.ArrayCreationExpr
                             ,actlang.expr.BitNotExpr
                             ,actlang.expr.BoolLiteralExpr
                             ,actlang.expr.CastExpr
                             ,actlang.expr.CompoundAssignmentExpr
                             ,actlang.expr.CondAndExpr
                             ,actlang.expr.CondOrExpr
                             ,actlang.expr.InfixExpr
                             ,actlang.expr.InstanceCreationExpr
                             ,actlang.expr.InstanceOfExpr
                             ,actlang.expr.IntLiteralExpr
                             ,actlang.expr.LogicalNotExpr
                             ,actlang.expr.NameMethodInvocationExpr
                             ,actlang.expr.NullExpr
                             ,actlang.expr.PrimaryMethodInvocationExpr
                             ,actlang.expr.SimpleAssignmentExpr
                             ,actlang.expr.ParenExpr
                             ,actlang.expr.PostfixExpr
                             ,actlang.expr.PrefixExpr
                             ,actlang.expr.PrimaryFieldAccessExpr
                             ,actlang.expr.StringLiteralExpr
                             ,actlang.expr.TernaryCondExpr
                             ,actlang.expr.UnaryMinusExpr
                             ,actlang.expr.UnaryPlusExpr
                             ,actlang.expr.VarInitializer
                             ,actlang.binaryop.BinaryOperator
                             )
                      ):
            return tree.api_recurseStr(rec)
        elif isinstance(tree, actlang.expr.ThisExpr):
            return this_str
        elif isinstance(tree, actlang.names.SimpleName):
            n = str(tree)
            if((n in class_names) or
               (n in global_function_names) or
               (n in global_variable_names)):
                return n
            return this_str+"."+n
        elif isinstance(tree, actlang.names.QualifiedName):
            parts = tree.api_getParts()
            if((parts[0] in class_names) or
               (parts[0] in global_function_names) or
               (parts[0] in global_variable_names)):
                return str(tree)
            return this_str+"."+str(tree)
        elif isinstance(tree, actlang.statement.ReturnStatement):
            expr_str = rec(tree.api_getExpr())
            return "$$$reader.read("+expr_str+"); break $pseudocontext;"
        print tree.__class__
        raise RuntimeError("Not supported in statement contextifier: "+
                           str(tree))

    stmt_str = "$pseudocontext: {"+stmt_str+"}"
    try:
        root = actlang.grammar.parse(stmt_str)
    except actlang.errors.Error, e:
        if True:
            debug_filename = "DEBUG.jmb"
            f = open(debug_filename, 'w')
            f.write(stmt_str)
            f.close()
        raise e
    
    stmts = root.api_getTopStatements()
    #print "\n".join([str(stmt) for stmt in stmts])
    contextified = "\n".join([rec(stmt) for stmt in stmts])
    #print contextified
    return contextified


#if __name__ == '__main__':
#    s = "y = f() + this.f() + A.f() + c.f() + g.f(); if(g.x) {return y;} else {g = null;} while(true) {x = x+1;}"
#    s = "if(x) {return 1;} else {return y;}"
#    #s = "y = this.A.foo(x,(a[this.y]+5)?true:\"MUU\")++;"
#    #print s
#    s2 = stmt_contextifier(s, "c", ["A"],["f"],["g"])
#    s2 = stmt_contextifier(s, "c", [],[],[])
#    root = actlang.grammar.parse(s2)
#    print s2





def __choose_augmenter(ast):
    """
    The actual code for choose augmenters below.
    """
    def rec(tree):
        if tree.__class__ == actlang.expr.NameMethodInvocationExpr:
            new_args = [rec(arg) for arg in tree.api_getArgList()]
            name = str(tree.api_getInvocationName())
            if(len(new_args) == 4 and
               (name == "$$SMUML.choose_boolean" or
                name == "$$SMUML.choose_int")):
                return name+"("+(", ".join(new_args))+", __id)"
            return name+"("+(", ".join(new_args))+")"

        return tree.api_recurseStr(rec)

    return rec(ast)


def choose_augmenter_stmt(code):
    """
    An internal helper function for the simulator.
    Given a (possibly compound) Jumbala statement 'code',
    augment each method call to $$SMUML.choose_boolean or
    $$SMUML.choose_int having 4 arguments to have an extra argument '__id'.
    """
    assert type(code) is str

    try:
        root = actlang.grammar.parse(code)
    except actlang.errors.Error, e:
        if False:
            debug_filename = "DEBUG.jmb"
            f = open(debug_filename, 'w')
            f.write(code)
            f.close()
        raise RuntimeError("Error while parsing a Jumbala statement "+
                           "in the simulator, consult '"+debug_filename+"' "+
                           "for the erroneous program code: "+str(e))
    stmts = root.api_getTopStatements()
    augmented = "\n".join([__choose_augmenter(stmt) for stmt in stmts])
    return augmented



def choose_augmenter_expr(code):
    """
    An internal helper function for the simulator.
    Given a (possibly compound) Jumbala expression 'code',
    augment each method call to $$SMUML.choose_boolean or
    $$SMUML.choose_int having 4 arguments to have an extra argument '__id'.
    """
    assert type(code) is str

    try:
        root = actlang.grammar.parse("assert("+str(code)+");")
    except actlang.errors.Error, e:
        if False:
            debug_filename = "DEBUG.jmb"
            f = open(debug_filename, 'w')
            f.write(code)
            f.close()
        raise RuntimeError("Error while parsing a Jumbala expression "+
                           "in the simulator, consult '"+debug_filename+"' "+
                           "for the erroneous program code: "+str(e))
    stmts = root.api_getTopStatements()
    assert len(stmts) == 1
    stmt = stmts[0]
    assert isinstance(stmt, actlang.statement.AssertStatement)
    return __choose_augmenter(stmt.api_getExpr())





def is_expr_side_effect_free(expr):
    """
    Check whether the Jumbala expression 'expr' is side-effect free.
    """
    assert isinstance(expr, actlang.expr.Expr)

    if isinstance(expr, (actlang.names.SimpleName
                         ,actlang.names.QualifiedName)):
        return True

    if isinstance(expr, actlang.expr.StatementExpr):
        return False
    
    if isinstance(expr, actlang.expr.ParenExpr):
        return is_expr_side_effect_free(expr.api_getInnerExpr())

    if isinstance(expr, (actlang.expr.IntLiteralExpr
                         ,actlang.expr.BoolLiteralExpr
                         ,actlang.expr.NullExpr
                         ,actlang.expr.StringLiteralExpr
                         ,actlang.expr.ThisExpr)):
        return True
    
    if isinstance(expr, (actlang.expr.CastExpr
                         ,actlang.expr.TypeCastExpr
                         ,actlang.expr.NameCastExpr)):
        return is_expr_side_effect_free(expr.api_getOperand())

    if isinstance(expr, (actlang.expr.IncrementOrDecrementExpr
                         ,actlang.expr.PrefixExpr
                         ,actlang.expr.PostfixExpr)):
        return False
    
    if isinstance(expr, (actlang.expr.UnaryPlusExpr
                         ,actlang.expr.UnaryMinusExpr
                         ,actlang.expr.BitNotExpr
                         ,actlang.expr.LogicalNotExpr
                         ,actlang.expr.InstanceOfExpr)):
        return is_expr_side_effect_free(expr.api_getOperand())    

    if isinstance(expr, actlang.expr.InfixExpr):
        if not is_expr_side_effect_free(expr.api_getLeftOperand()):
            return False
        if not is_expr_side_effect_free(expr.api_getRightOperand()):
            return False
        return True

    if isinstance(expr, actlang.expr.TernaryCondExpr):
        if not is_expr_side_effect_free(expr.api_getCondition()):
            return False
        if not is_expr_side_effect_free(expr.api_getTrueExpr()):
            return False
        if not is_expr_side_effect_free(expr.api_getFalseExpr()):
            return False
        return True
    
    if isinstance(expr, (actlang.expr.CondAndExpr
                         ,actlang.expr.CondOrExpr)):
        if not is_expr_side_effect_free(expr.api_getLeftOperand()):
            return False
        if not is_expr_side_effect_free(expr.api_getRightOperand()):
            return False
        return True

    if isinstance(expr, (actlang.expr.SimpleAssignmentExpr
                         ,actlang.expr.CompoundAssignmentExpr)):
        return False

    if isinstance(expr, actlang.expr.PrimaryFieldAccessExpr):
        if not is_expr_side_effect_free(expr.api_getReferenceExpr()):
            return False
        return True

    if isinstance(expr, actlang.expr.SuperFieldAccessExpr):
        return True

    if isinstance(expr, actlang.expr.ArrayAccessExpr):
        if not is_expr_side_effect_free(expr.api_getArrayExpr()):
            return False
        if not is_expr_side_effect_free(expr.api_getIndexExpr()):
            return False
        return True

    if isinstance(expr, (actlang.expr.MethodInvocationExpr,
                         actlang.expr.PrimaryMethodInvocationExpr,
                         actlang.expr.SuperMethodInvocationExpr)):
        for arg in expr.api_getArgList():
            if not is_expr_side_effect_free(arg):
                return False
        raise RuntimeError("Method invocation expressions not handled")
   
    if isinstance(expr, actlang.expr.NameMethodInvocationExpr):
        for arg in expr.api_getArgList():
            if not is_expr_side_effect_free(arg):
                return False
        n = str(node.api_getInvocationName())
        if n == "$$SMUML.choose_int":
            return True
        elif n == "$$SMUML.choose_boolean":
            return True
        raise RuntimeError("Method invocation expressions not handled")
   
    if isinstance(expr, (actlang.expr.InstanceCreationExpr,
                         actlang.expr.ArrayCreationExpr)):
        return False

    #print expr.__class__
    raise RuntimeError("Not implemented: side-effect freedom test for: "+
                       str(expr))


import coral
import PySMT
import Utils
UML14 = coral.meta.UML14

def _qn(e):
    # Just an abbreviation
    return Utils.Models.get_qualified_name(e)


def _guard_to_circuit(sim, uml_transition, c, context):
    assert isinstance(sim, Sim.Core.Simulator)
    assert uml_transition.oclIsKindOf(UML14.Transition)
    assert isinstance(c, PySMT.Circuit)
    assert type(context) is dict

    if not(uml_transition.guard and
           uml_transition.guard.expression and
           uml_transition.guard.expression.body):
        # No guard -> implicitly true -> return the constant true gate
        return c.add_true()
    guard_str = str(uml_transition.guard.expression.body)

    # Get the UML Class of the StateMachine containing 'uml_transition'
    uml_class = uml_transition
    while not uml_class.oclIsKindOf(UML14.Class):
        uml_class = uml_class.__parent__()
    class_def = sim.get_class_def(uml_class)

    # TO DO: Get the trigger and its parameters


    try:
        root = actlang.grammar.parse("assert("+str(guard_str)+");")
    except actlang.errors.Error, e:
        raise RuntimeError("Error while parsing the guard '"+
                           guard_str+"' of the transition '"+
                           _qn(uml_transition)+"': "+str(e))
    stmts = root.api_getTopStatements()
    assert len(stmts) == 1
    stmt = stmts[0]
    assert isinstance(stmt, actlang.statement.AssertStatement)
    guard_ast = stmt.api_getExpr()

    SMT_objref_type = PySMT.get_int()
    SMT_objref_null = c.add_int_const(0)
    SMT_objref_this = c.add_int_const(1)

    def _trans(expr):
        #
        # The actual recursive translation
        #
        assert isinstance(expr, actlang.expr.Expr)

        def _err():
            raise RuntimeError("Illegal guard '"+guard_str+"' "+
                               "in the transition '"+_qn(uml_transition)+"': "+
                               "expression '"+str(expr)+"' not handled")

        def _ns(s):
            raise RuntimeError(s+" not yet supported in the Jumbala-to-SMT translation")

        if isinstance(expr, actlang.names.SimpleName):
            name = expr.api_getLastPart()
            assert type(name) is str
            # Resolve: try context class attributes
            var_defs = class_def.get_variable_defs()
            if name in var_defs:
                var_def = var_defs[name]
                sim_type = var_def.get_type()
                uif = None
                if (class_def, name) in context:
                    uif = context[(class_def, name)]
                else:
                    if isinstance(sim_type, Sim.Core.SimType_Boolean):
                        uif = PySMT.new_uif([SMT_objref_type],
                                            PySMT.get_bool())
                    elif isinstance(sim_type, Sim.Core.SimType_Int):
                        uif = PySMT.new_uif([SMT_objref_type],
                                            PySMT.get_sint(32))
                    elif isinstance(sim_type,
                                    Sim.Core.SimType_ObjectRef):
                        uif = PySMT.new_uif([SMT_objref_type],
                                            SMT_objref_type)
                    else:
                        raise RuntimeError("Internal error")
                    context[(class_def, name)] = uif
                return (sim_type, c.add_uif(uif, SMT_objref_this))
            raise RuntimeError("Cannot bind variable '"+name+"' in the Class "+
                               _qn(class_def.get_uml_class()))
            
        elif isinstance(expr, actlang.names.QualifiedName):
            (qual_type, qual_gate) = _trans(expr.api_getQualifyingName())
            assert isinstance(qual_type, Sim.Core.SimType_ObjectRef)
            assert qual_type.get_target_name()
            qual_class_def = sim.get_class_def(qual_type.get_target_name())

            name = expr.api_getLastPart()
            assert type(name) is str
            # Resolve: try context class attributes
            var_defs = qual_class_def.get_variable_defs()
            if name in var_defs:
                var_def = var_defs[name]
                sim_type = var_def.get_type()
                uif = None
                if (qual_class_def, name) in context:
                    uif = context[(qual_class_def, name)]
                else:
                    if isinstance(sim_type, Sim.Core.SimType_Boolean):
                        uif = PySMT.new_uif([SMT_objref_type],
                                            PySMT.get_bool())
                    elif isinstance(sim_type, Sim.Core.SimType_Int):
                        uif = PySMT.new_uif([SMT_objref_type],
                                            PySMT.get_sint(32))
                    elif isinstance(sim_type,
                                    Sim.Core.SimType_ObjectRef):
                        uif = PySMT.new_uif([SMT_objref_type],
                                            SMT_objref_type)
                    else:
                        raise RuntimeError("Internal error")
                    context[(qual_class_def, name)] = uif
                return (sim_type, c.add_uif(uif, qual_gate))
            raise RuntimeError("Cannot bind variable '"+name+"' in the Class "+
                               _qn(qual_class_def.get_uml_class()))

        elif isinstance(expr, actlang.expr.StatementExpr):
            _err()
    
        elif isinstance(expr, actlang.expr.ParenExpr):
            return _trans(expr.api_getInnerExpr())

        elif isinstance(expr, actlang.expr.IntLiteralExpr):
            return (Sim.Core.SimType_Int(),
                    c.add_sint_const(32, expr.api_getValue()))

        elif isinstance(expr, actlang.expr.BoolLiteralExpr):
            if expr.api_getValue():
                return (Sim.Core.SimType_Boolean(), c.add_true())
            return (Sim.Core.SimType_Boolean(), c.add_false())
            
        elif isinstance(expr, actlang.expr.NullExpr):
            return (Sim.Core.SimType_ObjectRef(), SMT_objref_null)

        elif isinstance(expr, actlang.expr.StringLiteralExpr):
            _err()

        elif isinstance(expr, actlang.expr.ThisExpr):
            return (Sim.Core.SimType_ObjectRef(class_def.get_name()),
                    SMT_objref_this)

        elif isinstance(expr, (actlang.expr.CastExpr
                               ,actlang.expr.TypeCastExpr
                               ,actlang.expr.NameCastExpr)):
            _err()

        elif isinstance(expr, (actlang.expr.IncrementOrDecrementExpr
                               ,actlang.expr.PrefixExpr
                               ,actlang.expr.PostfixExpr)):
            _err()
    
        elif isinstance(expr, actlang.expr.UnaryPlusExpr):
            return _trans(expr.api_getOperand())

        elif isinstance(expr, actlang.expr.UnaryMinusExpr):
            (sim_type, g) = _trans(expr.api_getOperand())
            return (sim_type, c.add_neg(g))

        elif isinstance(expr, actlang.expr.BitNotExpr):
            (sim_type, g) = _trans(expr.api_getOperand())
            return (sim_type, c.add_bitwise_not(g))

        elif isinstance(expr, actlang.expr.LogicalNotExpr):
            (sim_type, g) = _trans(expr.api_getOperand())
            return (sim_type, c.add_not(g))

        elif isinstance(expr, actlang.expr.InstanceOfExpr):
            _err()

        elif isinstance(expr, actlang.expr.InfixExpr):
            (lt, lg) = _trans(expr.api_getLeftOperand())
            (rt, rg) = _trans(expr.api_getRightOperand())
            op = expr.api_getOperator()
            if isinstance(op, actlang.binaryop.LessThan):
                assert isinstance(lt, Sim.Core.SimType_Int)
                assert isinstance(rt, Sim.Core.SimType_Int)
                return (Sim.Core.SimType_Boolean(), c.add_lt(lg, rg))

            elif isinstance(op, actlang.binaryop.LessOrEqual):
                assert isinstance(lt, Sim.Core.SimType_Int)
                assert isinstance(rt, Sim.Core.SimType_Int)
                return (Sim.Core.SimType_Boolean(), c.add_le(lg, rg))

            elif isinstance(op, actlang.binaryop.Equal):
                assert ((isinstance(lt, Sim.Core.SimType_Int) and
                         isinstance(rt, Sim.Core.SimType_Int)) or
                        (isinstance(lt, Sim.Core.SimType_Boolean) and
                         isinstance(rt, Sim.Core.SimType_Boolean)))
                return (Sim.Core.SimType_Boolean(), c.add_eq(lg, rg))

            elif isinstance(op, actlang.binaryop.NotEqual):
                assert ((isinstance(lt, Sim.Core.SimType_Int) and
                         isinstance(rt, Sim.Core.SimType_Int)) or
                        (isinstance(lt, Sim.Core.SimType_Boolean) and
                         isinstance(rt, Sim.Core.SimType_Boolean)))
                return (Sim.Core.SimType_Boolean(), c.add_neq(lg, rg))

            elif isinstance(op, actlang.binaryop.GreaterOrEqual):
                assert isinstance(lt, Sim.Core.SimType_Int)
                assert isinstance(rt, Sim.Core.SimType_Int)
                return (Sim.Core.SimType_Boolean(), c.add_ge(lg, rg))

            elif isinstance(op, actlang.binaryop.GreaterThan):
                assert isinstance(lt, Sim.Core.SimType_Int)
                assert isinstance(rt, Sim.Core.SimType_Int)
                return (Sim.Core.SimType_Boolean(), c.add_gt(lg, rg))

            elif isinstance(op, actlang.binaryop.Plus):
                assert isinstance(lt, Sim.Core.SimType_Int)
                assert isinstance(rt, Sim.Core.SimType_Int)
                return (lt, c.add_plus(lg, rg))
            
            elif isinstance(op, actlang.binaryop.Times):
                assert isinstance(lt, Sim.Core.SimType_Int)
                assert isinstance(rt, Sim.Core.SimType_Int)
                return (lt, c.add_times(lg, rg))

            elif isinstance(op, actlang.binaryop.Divide):
                assert isinstance(lt, Sim.Core.SimType_Int)
                assert isinstance(rt, Sim.Core.SimType_Int)
                _ns("Division")

            elif isinstance(op, actlang.binaryop.Remainder):
                assert isinstance(lt, Sim.Core.SimType_Int)
                assert isinstance(rt, Sim.Core.SimType_Int)
                _ns("Remainder")

            elif isinstance(op, actlang.binaryop.Minus):
                assert isinstance(lt, Sim.Core.SimType_Int)
                assert isinstance(rt, Sim.Core.SimType_Int)
                return (lt, c.add_minus(lg, rg))

            elif isinstance(op, (actlang.binaryop.ShiftLeft
                                 ,actlang.binaryop.SignedShiftRight
                                 ,actlang.binaryop.UnsignedShiftRight)):
                assert isinstance(lt, Sim.Core.SimType_Int)
                assert isinstance(rt, Sim.Core.SimType_Int)
                _ns("Shift operators")

            elif isinstance(op, actlang.binaryop.BitLogicalAnd):
                if isinstance(lt, Sim.Core.SimType_Boolean):
                    assert isinstance(rt, Sim.Core.SimType_Boolean)
                    return (lt, c.add_and(lg, rg))
                assert isinstance(lt, Sim.Core.SimType_Int)
                assert isinstance(rt, Sim.Core.SimType_Int)
                return (lt, c.add_bitwise_and(lg, rg))
            
            elif isinstance(op, actlang.binaryop.BitLogicalXor):
                if isinstance(lt, Sim.Core.SimType_Boolean):
                    assert isinstance(rt, Sim.Core.SimType_Boolean)
                    return (lt, c.add_xor(lg, rg))
                assert isinstance(lt, Sim.Core.SimType_Int)
                assert isinstance(rt, Sim.Core.SimType_Int)
                return (lt, c.add_bitwise_xor(lg, rg))

            elif isinstance(op, actlang.binaryop.BitLogicalOr):
                if isinstance(lt, Sim.Core.SimType_Boolean):
                    assert isinstance(rt, Sim.Core.SimType_Boolean)
                    return (lt, c.add_or(lg, rg))
                assert isinstance(lt, Sim.Core.SimType_Int)
                assert isinstance(rt, Sim.Core.SimType_Int)
                return (lt, c.add_bitwise_or(lg, rg))

            _ns("Operator "+str(op))

        elif isinstance(expr, actlang.expr.TernaryCondExpr):
            (ct, cg) = _trans(expr.api_getCondition())
            (tt, tg) = _trans(expr.api_getTrueExpr())
            (ft, fg) = _trans(expr.api_getFalseExpr())
            assert isinstance(ct, Sim.Core.SimType_Boolean)
            assert ((isinstance(tt, Sim.Core.SimType_Int) and
                     isinstance(ft, Sim.Core.SimType_Int)) or
                    (isinstance(tt, Sim.Core.SimType_Boolean) and
                     isinstance(ft, Sim.Core.SimType_Boolean)))
            return (tt, c.add_ite(cg, tg, fg))
    
        elif isinstance(expr, actlang.expr.CondOrExpr):
            (lt, lg) = _trans(expr.api_getLeftOperand())
            (rt, rg) = _trans(expr.api_getRightOperand())
            assert isinstance(lt, Sim.Core.SimType_Boolean)
            assert isinstance(rt, Sim.Core.SimType_Boolean)
            return (lt, c.add_or(lg, rg))

        elif isinstance(expr, actlang.expr.CondAndExpr):
            (lt, lg) = _trans(expr.api_getLeftOperand())
            (rt, rg) = _trans(expr.api_getRightOperand())
            assert isinstance(lt, Sim.Core.SimType_Boolean)
            assert isinstance(rt, Sim.Core.SimType_Boolean)
            return (lt, c.add_and(lg, rg))

        elif isinstance(expr, (actlang.expr.SimpleAssignmentExpr
                               ,actlang.expr.CompoundAssignmentExpr)):
            _err()

        elif isinstance(expr, actlang.expr.PrimaryFieldAccessExpr):
            (rt, rg) = _trans(expr.api_getReferenceExpr())
            assert False

        elif isinstance(expr, actlang.expr.SuperFieldAccessExpr):
            _err()

        elif isinstance(expr, actlang.expr.ArrayAccessExpr):
            _err()

        elif isinstance(expr, (actlang.expr.MethodInvocationExpr,
                               actlang.expr.PrimaryMethodInvocationExpr,
                               actlang.expr.SuperMethodInvocationExpr)):
            _err()
   
        elif isinstance(expr, actlang.expr.NameMethodInvocationExpr):
            n = str(node.api_getInvocationName())
            if n == "$$SMUML.choose_int":
                #return c.add_sint_var()
                _err()
            elif n == "$$SMUML.choose_boolean":
                #return True
                _err()
            trans_args = [_trans(arg) for arg in expr.api_getArgList()]
            _err()
   
        elif isinstance(expr, (actlang.expr.InstanceCreationExpr,
                               actlang.expr.ArrayCreationExpr)):
            _err()

        #print expr.__class__
        raise RuntimeError("Not implemented: SMT translation for: "+str(expr))

    #print str(guard_ast)
    return _trans(guard_ast)[1]



import Sim.Core
if __name__ == '__main__':
    #
    # Load project
    #
    P = coral.loadDocument(sys.argv[1])

    #
    # Find the model
    #
    models = [e for e in P.contents() if e.oclIsKindOf(UML14.Model)]
    if len(models) != 1:
        raise "The project contains "+str(len(models))+\
              " models instead of one - aborting!"
    model = models[0]

    sim =  Sim.Core.Simulator(model)


    #
    # Find all state machines in the model
    #
    state_machines = [sm for sm in model.__getAllParts__() \
                      if sm.oclIsKindOf(UML14.StateMachine)]

    for sv in Utils.Models.get_elements_of_type(model, UML14.State):
        uml_cts = Utils.StateMachines.get_outgoing_completion_transitions(sv)
        if not uml_cts:
            continue
        circuit = PySMT.Circuit()
        context = {}
        gates = []
        for uml_ct in uml_cts:
            gates.append(_guard_to_circuit(sim, uml_ct, circuit, context))
        eq = circuit.add_or(gates)
        #print eq.get_type()
        #print circuit.get_type(eq)
        #print context

        f = open("DEBUG2.dot", "w")
        circuit.to_dot(f)
        f.close()

        if circuit.is_constant_true(eq) and len(gates) >= 2:
            print "MYYYYYYYYYYYYY"
            for uml_t in uml_cts:
                if uml_t.guard:
                    print " ",str(uml_t.guard.expression.body)
                else:
                    print " -"
            sys.exit(0)


        circuit.constrain(eq, False)

        result = circuit.solve_yices()
        if result == "sat":
            print "Warning: state '"+_qn(sv)+"': "+\
                  "cannot quarantee that at least one of the outgoing "+\
                  "completion transitions is always enabled, the guards are: "
            for uml_ct in uml_cts:
                print " ",str(uml_ct.guard.expression.body)
            if False:
                f = open("DEBUG.dot", "w")
                circuit.to_dot(f)
                f.close()
        elif result == "unsat":
            if False:
                f = open("DEBUG.dot", "w")
                circuit.to_dot(f)
                f.close()
            if True:
                print "State '"+_qn(sv)+"': "+\
                      "can quarantee that at least one of the outgoing "+\
                      "completion transitions is always enabled, "+\
                      "the guards are: "
                for uml_ct in uml_cts:
                    if uml_ct.guard:
                        print " ",str(uml_ct.guard.expression.body)
                    else:
                        print " -"
                if len(gates) >= 2: sys.exit(0)
        else:
            assert False

    for sv in Utils.Models.get_elements_of_type(model, UML14.Pseudostate):
        assert sv.outgoing
        circuit = PySMT.Circuit()
        context = {}
        gates = []
        for uml_t in sv.outgoing:
            gates.append(_guard_to_circuit(sim, uml_t, circuit, context))
        eq = circuit.add_not(circuit.add_or(gates))
        #print eq.get_type()
        #print c.get_type(eq)
        #print context

        circuit.constrain(eq, True)

        result = circuit.solve_yices()
        if result == "sat":
            print "Warning: pseudo-state '"+_qn(sv)+"': "+\
                  "cannot quarantee that at least one of the outgoing "+\
                  "transitions is always enabled, the guards are:"
            for uml_t in sv.outgoing:
                print " ",str(uml_t.guard.expression.body)
            f = open("DEBUG.dot", "w")
            circuit.to_dot(f)
            f.close()
        elif result == "unsat":
            if False:
                print "pseudo-state '"+_qn(sv)+"': "+\
                      "Can quarantee that at least one of the outgoing "+\
                      "transitions is always enabled, the guards are:"
                for uml_t in sv.outgoing:
                    if uml_t.guard:
                        print " ",str(uml_t.guard.expression.body)
                    else:
                        print " -"
                f = open("DEBUG.dot", "w")
                circuit.to_dot(f)
                f.close()
        else:
            assert False
