# This file is part of the SMUML toolset.
# Copyright (C) 2005-2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# Authors: Tommi Junttila and Timo Latvala

"""
The core of the SMUML UML simulator
"""


import os
import sys
from sets import Set
import random

import coral

import actlang

from Sim import GlobalState
from Sim import Trace
from Sim.Object import Object
import Sim.Error
from Sim.Error import SimError
from Sim.Message import Message
import Utils.String
import Utils.Models
from Utils.Models import get_qualified_name
from Utils.Models import get_elements_of_type
import Utils.StateMachines
import Utils.UML14

import Sim.ActLangUtils


UML14 = coral.lang.UML14
DIA20 = coral.lang.XMIDI20

#constant magic string, global name for action language object array
al_objs="$$$ObjectTracker.objects"


class ObservedAction:
    """
    An abstract base class for actions that can be observed by the simulator
    while the action language virtual machine executes code.
    """
    pass

class OA_SignalAction(ObservedAction):
    receiver_id = -1
    signal_num = -1
    parameters = []

    def __init__(self, receiver_id, signal_num):
        self.receiver_id = receiver_id
        self.signal_num = signal_num
        self.parameters = []

    def add_parameter(self, param):
        assert isinstance(param, SimValue)
        self.parameters.append(param.get_copy())

    def _print(self):
        s =  "Object " + str(self.receiver_id)
        s += " received signal #" + str(self.signal_num)
        s += " with parameters ["+\
             (",".join([p.get_str() for p in self.parameters]))+"]"
        print s
        
class OA_CreateAction(ObservedAction):
    object_id = -1
    class_num = -1

    def __init__(self, object_id, class_num):
        self.object_id = object_id
        self.class_num = class_num
    def _print(self):
        print "Object " + str(self.object_id) + " of class #" + str(self.class_num) + " created"





#
#
# Default hooks for providing values for the non-deterministic choose functions
#
#
def _interactive_choose_boolean_hook(dummy, sim, obj, uml_transition, where,
                                     index):
    """
    The default hook for querying Booleans for the choose_boolean function.
    Asks the user for the value.
    """
    assert dummy is None
    assert isinstance(sim, Sim.Core.Simulator)
    assert isinstance(obj, Sim.Core.Object)
    assert (uml_transition is None) or uml_transition.oclIsKindOf(UML14.Transition)
    assert where in ["guard","effect",None]
    assert type(index) is int

    print "Evaluating",
    if uml_transition:
        print "the "+where+" of transition "+get_qualified_name(uml_transition),
    else:
        print "an expression",
    print "in Object "+str(obj.idnum)

    choice_num = 0
    while True:
        print "A random Boolean is required, please enter 0 or 1: ",
        answer = sys.stdin.readline()
        if Utils.String.IsInt(answer):
            choice_num = int(answer)
            if choice_num >= 0 and choice_num <= 1:
                break
    return choice_num == 1

def _interactive_choose_int_hook(dummy, sim, obj, uml_transition, where,
                                 index, int_list):
    """
    The default hook for querying integers for the choose_int function.
    Asks the user for the value.
    """
    assert dummy is None
    assert isinstance(sim, Sim.Core.Simulator)
    assert isinstance(obj, Sim.Core.Object)
    assert (uml_transition is None) or uml_transition.oclIsKindOf(UML14.Transition)
    assert where in ["guard","effect",None]
    assert type(index) is int

    print "Evaluating",
    if uml_transition:
        print "the "+where+" of transition "+get_qualified_name(uml_transition),
    else:
        print "an expression",
    print "in Object "+str(obj.idnum)

    choice_num = 0
    while True:
        print "Please select a number in "+str(int_list)+": ",
        answer = sys.stdin.readline()
        if Utils.String.IsInt(answer):
            choice_num = int(answer)
            if choice_num in int_list:
                break
    return choice_num



#
#
#
#
#
class ActionGenerator:
    """
    A container class for methods generating action language.
    """

    #List of ActionGeneratorClasses
    #classes = []
    #UML Statemachine Transition -> action methodname map
    #uml_transition_action_method_map = {}

    #UML Statemachine Transition -> action guardname map
    #uml_transition_guard_method_map = {}

    def __init__(self):
        self.classes = []
        self._uml_class_to_algen_class = {}
        self.uml_transition_guard_method_map = {}
        self.uml_transition_action_method_map = {}


    def add_uml_class(self, uml_class, definition):
        assert uml_class.oclIsKindOf(UML14.Class)
        assert isinstance(definition, ClassDef)
        newclass = ActionGeneratorClass(self, uml_class, definition)
        self.classes.append(newclass)
        self._uml_class_to_algen_class[uml_class] = newclass
        return newclass


    def get_algen_class(self, uml_class):
        """
        Get the ActionGeneratorClass corresponding to UML Class 'uml_class'.
        """
        assert uml_class.oclIsKindOf(UML14.Class)
        return self._uml_class_to_algen_class[uml_class]


    def gen_uml_base_class(self, signals):
        """
        'signals' should be the list of UML Signals appearing in the model.
        """
        assert type(signals) is list
        prog = ""
        prog += "class $$$Uml_Base {\n"
        prog += "  static boolean $$run_constructors = false;\n"
        prog += "  static int __nextid = 0;\n"
        prog += "  int __id=0;\n"
        prog += "  static native void $$$signal_received(int id, int signal_num);\n"
        prog += "  static native void $$$signal_add_bool_parameter(boolean param);\n"
        prog += "  static native void $$$signal_add_int_parameter(int param); \n"
        prog += "  static native void $$$signal_add_ref_parameter(int param); \n"
        prog += "  static native void $$$object_created(int id, int class_num); \n"

        # Generate signal reception methods
        for (signal_num, signal) in enumerate(signals):
            assert signal.oclIsKindOf(UML14.Signal)
            prog += "  void $$signal_" + str(signal.name) + "("
            lsep = ""
            for feature in signal.feature:
                assert feature.oclIsKindOf(UML14.Attribute)
                vdef = _VariableDef_from_Attribute(feature)
                prog += lsep + vdef.get_type().al_type()+" "+vdef.get_name()
                lsep = ","
            prog += ") { $$$signal_received(__id," + str(signal_num) + "); "
            for feature in signal.feature:
                assert feature.oclIsKindOf(UML14.Attribute)
                vdef = _VariableDef_from_Attribute(feature)
                if isinstance(vdef.get_type(), SimType_Int):
                    prog += "$$$signal_add_int_parameter(" + vdef.get_name() + "); "
                elif isinstance(vdef.get_type(), SimType_Boolean):
                    prog += "$$$signal_add_bool_parameter(" + vdef.get_name() + "); "
                elif isinstance(vdef.get_type(), SimType_ObjectRef):
                    prog += "if("+vdef.get_name()+"!=null) {$$$signal_add_ref_parameter(" + vdef.get_name() + ".__id);} else {$$$signal_add_ref_parameter(-1);} "
                else:
                    raise "Internal error: Signal parameter type not supported"
            prog += "}\n"

        prog += "}\n"

        return prog


    def gen_action_code(self, indent, uml_class_to_class_index_map,
                        add_sim_code):
        """
        If 'add_sim_code' is True, adds the simulator specific code, too.
        """
        assert isinstance(indent, Utils.String.Indent)
        prog = ""
        for c in self.classes:
            prog += c.generate(indent, uml_class_to_class_index_map,
                               add_sim_code)
        return prog

    def get_guard_method_name(self, uml_transition):
        """
        Get the name of the method that contains the code for
        the guard of the UML Transition 'uml_transition'.
        """
        assert uml_transition.oclIsKindOf(UML14.Transition)
        return self.uml_transition_guard_method_map[uml_transition]

    def get_effect_method_name(self, uml_transition):
        """
        Get the name of the method that contains the code for
        the effect of the UML Transition 'uml_transition'.
        """
        assert uml_transition.oclIsKindOf(UML14.Transition)
        return self.uml_transition_action_method_map[uml_transition]


class ActionGeneratorMethod:
    name = None
    #A SimType
    return_type = None
    #name, simtype pairs
    parameterlist = {}
    body = None


    def __init__(self, name, return_type, body):
        """
        New generator for method with name 'name'.
        'return_type' should be either None (a void method) or a SimType.
        The body of the method should be a string.
        Method parameters should be added with add_parameter function.
        """
        assert type(name) is str and name != ""
        assert return_type == None or isinstance(return_type, SimType)
        assert body and type(body) is str
        self.name = name
        self.return_type = return_type
        self.parameterlist = []
        self.body = body

    
    def add_parameter(self, name, t):
        """
        Add a new parameter for the method.
        The name 'name' should be a string different from parameters
        defined earlier.
        The type 't' of the parameter should be a SimType.
        """
        assert type(name) is str and name != ""
        assert isinstance(t, SimType)
        assert not (name in [p[0] for p in self.parameterlist])
        self.parameterlist.append((name, t))


    def generate(self, indent):
        """
        Generate the action language method.
        Returns a string.
        """
        assert isinstance(indent, Utils.String.Indent)
        prog = ""
        if self.return_type != None:
            prog += str(indent) + self.return_type.al_type() +" "
        else:
            prog += str(indent) + "void "
            
        prog += self.name

        prog += "("
        sep = ""
        for (name,t) in self.parameterlist:
            prog += sep + t.al_type() + " " + name
            sep = ","
        prog += ") {\n"

        indent.push()
        for l in self.body.split('\n'):
            prog += str(indent) + l +"\n"
        indent.pop()
        
        prog += str(indent) + "}\n"

        return prog

class ActionGeneratorClass:
    #ActionGenerator
    generator = None
    #name of the class
    uml_class = None
    name = None
    # simulator ClassDef
    definition = None
    #name -> (SimType,model_default_value_str) pairs
    variables = {}
    #name, method objects pairs
    methods = {}
	
    def __init__(self, generator, uml_class, definition):
        assert uml_class.oclIsKindOf(UML14.Class)
        assert isinstance(definition, ClassDef)
        Utils.UML14.check_name(uml_class)
        self.generator = generator
        self.uml_class = uml_class
        self.name = uml_class.name
        self.variables = {}
        self.methods = {}
        self.__init_vars(definition)
        self.__init_methods(definition)

    def __init_vars(self,definition):
        assert isinstance(definition, ClassDef)
        for (name, vdef) in definition.get_variable_defs().iteritems():
            simtype = None
            if isinstance(vdef.get_type(), SimType_ObjectRef):
                uml_ref = vdef.get_uml_reference()
                if uml_ref.oclIsKindOf(UML14.AssociationEnd):
                    simtype = SimType_ObjectRef(str(uml_ref.participant.name))
                    self.variables[name]=(simtype,vdef.model_default_value_str)
                elif uml_ref.oclIsKindOf(UML14.Attribute):
                    simtype = SimType_ObjectRef(str(uml_ref.type.name))
                    self.variables[name]=(simtype,vdef.model_default_value_str)
                else:
                    raise "Unknown UML variable"
            else:
                self.variables[name] = (vdef.get_type(),
                                        vdef.model_default_value_str)


    def __init_methods(self, definition):
        assert isinstance(definition, ClassDef)
        for method_def in definition.methods:
            uml_method = method_def.get_uml_method()
            assert uml_method.body
            assert uml_method.body.body
            body = str(uml_method.body.body)
            if Simulator._ASSERTION_CONDITION:
                body = Sim.ActLangUtils.condition_assertions(body, Simulator._ASSERTION_CONDITION)
            method = ActionGeneratorMethod(method_def.name,
                                           method_def.return_type,
                                           body)
            for (name,t) in method_def.param_list:
                method.add_parameter(name, t)
            self.add_method(method_def.name, method)

        #
	# Generate methods for transition guards and effects
        # of the state machine of the class
        #
        uml_sm = definition.get_uml_state_machine()
        if not uml_sm:
            return 
        for (trans_num,trans) in enumerate(uml_sm.transitions):
            assert trans.oclIsKindOf(UML14.Transition)

            #
            # Jumbala code for the transition guard
            #
            guard_name = "$guard"+str(trans_num)
            self.generator.uml_transition_guard_method_map[trans] = guard_name + "()"
            # First a comment for debugging purposes...
            code =  "// The guard of transition '" + str(trans.name) + "'\n"
            code += "// from '" + sm_get_qualified_name(trans.source) + "'\n"
            code += "// to '" + sm_get_qualified_name(trans.target) + "'\n"
            # And then the actual code
            if trans.guard:
                assert trans.guard.expression
                assert trans.guard.expression.body
                code += "return " + Sim.ActLangUtils.choose_augmenter_expr(str(trans.guard.expression.body))+";" #, "guard:"+str(trans.__XMIid__())) + ";"
            else:
                code += "return true;"
            method = ActionGeneratorMethod(guard_name, SimType_Boolean(), code)
            self.add_method(guard_name, method)


            #
            # Jumbala code for the transition effect
            #
            effect_name = "$effect"+str(trans_num)
            self.generator.uml_transition_action_method_map[trans] = effect_name + "()"
            # First a comment for debugging purposes...
            code =  "// The effect of transition '" + str(trans.name) + "'\n"
            code += "// from '" + sm_get_qualified_name(trans.source) + "'\n"
            code += "// to '" + sm_get_qualified_name(trans.target) + "'\n"
            # And then the actual code
            if trans.effect:
                assert trans.effect.script
                assert trans.effect.script.body
                body = str(trans.effect.script.body)
                if Simulator._ASSERTION_CONDITION:
                    body = Sim.ActLangUtils.condition_assertions(body, Simulator._ASSERTION_CONDITION)
                code += Sim.ActLangUtils.choose_augmenter_stmt(body)
            else:
                code += "return;"
            method = ActionGeneratorMethod(effect_name, None, code);
            self.add_method(effect_name, method)



    def add_variable(self, name, t):
        assert False
        #self.variables[name] = t


    def add_method(self, name, method):
        """
        Add a new ActionGeneratorMethod 'method' with name 'name'.
        """
        assert name and isinstance(name, str) and name != ""
        assert method and isinstance(method, ActionGeneratorMethod)
        assert not self.methods.has_key(name)
        self.methods[name] = method 


    def remove_method(self, name):
        """
        Remove the ActionGeneratorMethod with name 'name'.
        """
        assert name and isinstance(name, str) and name != ""
        assert self.methods.has_key(name)
        del self.methods[name]


    def generate(self, indent, uml_class_to_class_index_map, add_sim_code):
        """
        Generate action language code for the class.
        If 'add_sim_code' is True, adds the simulator specific code, too.
        """
        assert isinstance(indent, Utils.String.Indent)
        prog = str(indent)+"class "+str(self.name)
        if add_sim_code:
            prog += " extends $$$Uml_Base"
        prog += " {\n"
        indent.push()

        # Constructor
        prog += str(indent)+str(self.name)+"() {\n"
        indent.push()
        if add_sim_code:
            prog += str(indent)+"__id = __nextid++;\n"
            prog += str(indent)+"$$$ObjectTracker.add(__id,this);\n"
            prog += str(indent)+"$$$object_created(__id,"+\
                    str(uml_class_to_class_index_map[self.uml_class])+");\n"
        indent.pop()
        prog += str(indent)+"}\n"

        # Class variable definitions
        for name,(t,model_default_value_str) in self.variables.iteritems():
            prog += str(indent) + t.al_type() + " " + name;
            if model_default_value_str:
                prog += " = ($$run_constructors)?("+model_default_value_str+"):("
                if isinstance(t, SimType_Boolean):
                    prog += "false"
                elif isinstance(t, SimType_Int):
                    prog += "0"
                elif isinstance(t, SimType_ObjectRef):
                    prog += "null"
                else:
                    assert False
                prog += ")"
            prog += ";\n"  

        for name,method in self.methods.iteritems():
            prog += method.generate(indent)
        indent.pop()
        prog += "}\n"
        return prog



#
# Some auxiliary functions
#
    
def sm_print_parent_path(path):
    print "::".join([s.name for s in path])
   


def sm_get_qualified_name(stateVertex):
    assert stateVertex.oclIsKindOf(UML14.StateVertex)
    path = Utils.StateMachines.get_parent_path(stateVertex)
    name = ""
    sep = ""
    for s in path:
        name += sep
        sep = "::"
        if s.name:
            name += str(s.name)
        else:
            name += "<<anonymous>>"
    return name





def sm_validate_legal_orthogonality(stateMachine):
    """
    Validate that the orthogonal regions in the state machine
    are not interested in the same signals.
    """
    assert stateMachine.oclIsKindOf(UML14.StateMachine)
    assert stateMachine.top
    result,conflict = _sm_validate_legal_orthogonality(stateMachine.top)

    if result:
        return True
    (signal,sv1,sv2) = conflict

    raise SimError("State machine '" + str(stateMachine.name) + "': orthogonal states '" + sm_get_qualified_name(sv1) + "' and '" + sm_get_qualified_name(sv2) + "' are interested in the same signal '" + str(signal.name) + "'")
    
    return False


def _sm_validate_legal_orthogonality(stateVertex):
    assert stateVertex.oclIsKindOf(UML14.StateVertex)
    signal_set = {}

    if stateVertex.oclIsKindOf(UML14.CompositeState):
        if stateVertex.isConcurrent:
            for subvertex in stateVertex.subvertex:
                (subresult,signal_set2) = _sm_validate_legal_orthogonality(subvertex)
                if subresult == False:
                    return (False, signal_set2)
                for (signal,where) in signal_set2.iteritems():
                    if signal_set.has_key(signal):
                        return (False, (signal,signal_set[signal],where))
                    signal_set[signal] = where
        else:
            for subvertex in stateVertex.subvertex:
                (subresult,signal_set2) = _sm_validate_legal_orthogonality(subvertex)
                if subresult == False:
                    return (False, signal_set2)
                for (signal,where) in signal_set2.iteritems():
                    signal_set[signal] = where
                    
    if stateVertex.oclIsKindOf(UML14.SubmachineState):
        raise "Submachines not yet handled"

    if stateVertex.oclIsKindOf(UML14.State):
        for e in stateVertex.deferrableEvent:
            if e.oclIsKindOf(UML14.SignalEvent):
                signal_set[e.signal] = stateVertex
    if stateVertex.oclIsKindOf(UML14.State):
        for t in stateVertex.internalTransition:
            if t.trigger and t.trigger.oclIsKindOf(UML14.SignalEvent):
                signal_set[t.trigger.signal] = stateVertex
    for t in stateVertex.outgoing:
        if t.trigger and t.trigger.oclIsKindOf(UML14.SignalEvent):
            signal_set[t.trigger.signal] = stateVertex

    return (True, signal_set)



class SM_StateNode:
    """
    A node in a UML state machine state configuration tree.
    """
    # Either a statemachine or a vertex (state or pseudostate)
    #uml_reference = None
    # Children, a list of SM_StateNode
    #children = []
    # Are the child states contained concurrent?
    #isConcurrent = False
    #Is the state initial
    #isInitial = False

    def __init__(self):
        self.uml_reference = None
        self.children = []
        self.isConcurrent = False
        self.isInitial = False


    def __hash__(self):
        h = [hash(str(self.uml_reference))]
        for child in self.children:
            h.append(hash(child))
        return hash(str(h))


    def __cmp__(self, other):
        assert isinstance(other, SM_StateNode)
        v = cmp(self.uml_reference, other.uml_reference)
        if v != 0:
            return v
        v = cmp(len(self.children),len(other.children))
        if v != 0:
            return v
        for c1,c2 in zip(self.children, other.children):
            v = cmp(c1, c2)
            if v != 0:
                return v
        return 0

    def _print(self, indent):
        if self.uml_reference.name:
            print str(indent) + self.uml_reference.name
        else:
            print str(indent) + "<<anonymous>>"
        indent.push()
        for c in self.children:
            c._print(indent)
        indent.pop()

    def get_uml_reference(self):
        """
        Return the UML StateVertex to which this state configuration node
        corresponds.
        """
        assert self.uml_reference
        return self.uml_reference

    def get_leaf_state_names(self):
        l = []
        self._get_leaf_state_names(l)
        return l

    def _get_leaf_state_names(self, l):
        if self.children and self.children != []:
            for c in self.children:
                c._get_leaf_state_names(l)
        else:
            if self.uml_reference.name:
                l.append(str(self.uml_reference.name))
            else:
                l.append("<<anonymous>>")


    def copy(self):
        """
        Return a copy of this state configuration tree.
        """
        copied_state_node = SM_StateNode()
        copied_state_node.uml_reference = self.uml_reference
        copied_state_node.isConcurrent = self.isConcurrent
        copied_state_node.isInitial = self.isInitial
        copied_state_node.children = []
        for c in self.children:
            copied_state_node.children.append(c.copy())
        return copied_state_node


    def find(self, state):
        """
        Check if state is part of this StateNode
        """
        if self.uml_reference == state:
            return state
        for s in self.children:
            if s.find(state):
                return state
        return None


    def replace(self, old, new):
        """Replace a subtree old with new, return new complete state"""

        if False:
            print "r Replacing"
            old._print("r   ")
            print "r with"
            new._print("r   ")
            print "r in"
            self._print("r   ")

        assert self.find(old.uml_reference)
        if self==old:
            return new
        stack=[self]
        while stack != []:
            state=stack.pop()
            for n in range(len(state.children)):
                if state.children[n]==old:
                    state.children[n]=new
                    return self
                stack.append(state.children[n])
        assert False                   



    def find_signals(self):
        """Returns a list of signals that appear as triggers from this state
        and its children recursively"""
        # Simulate the set data structure with map in Python 2.3.5
        signal_map = {}
        self._find_signals(signal_map)
        return signal_map.keys();

    #
    # The real signal finding code
    #
    def _find_signals(self, signal_map):
        # Handle normal outgoing transitions
        for transition in self.uml_reference.outgoing:
            if not transition.trigger:
                continue
            if not transition.trigger.oclIsKindOf(UML14.SignalEvent):
                raise SimError("Internal error: only signal events supported")
            signal_map[transition.trigger.signal] = True

        if self.uml_reference.oclIsKindOf(UML14.SubmachineState):
            raise SimError("Submachine states not yet supported")

        if self.uml_reference.oclIsKindOf(UML14.State):
            # Handle internal transitions
            for transition in self.uml_reference.internalTransition:
                if not transition.trigger:
                    continue
                if not transition.trigger.oclIsKindOf(UML14.SignalEvent):
                    raise SimError("Internal error: only signal events supported")
                signal_map[transition.trigger.signal] = True
            # Handle deferrable events
            for e in self.uml_reference.deferrableEvent:
                assert e.oclIsKindOf(UML14.SignalEvent)
                signal_map[e.signal] = True
                
        for child in self.children:
            child._find_signals(signal_map)

        

    # The following code should be moved somewhere else?
    def evaluate_guard(self, object, transition, message,
                       simulator, global_state):
        assert (message is None) or isinstance(message, Message)
        # Remember the current working state
        old_working_global_state = simulator.working_global_state
        # Set global_state as the working_global state
        simulator.set_working_global_state(global_state)

        #
        # If the triggering signal event has attributes,
        # copy the signal event attributes to the corresponding object
        # attributes
        #
        if transition.trigger and transition.trigger.parameter:
            assert message and message.parameters
            assert message.signal == transition.trigger.signal
            if len(message.signal.feature)!=len(transition.trigger.parameter):
                raise SimError("Incorrect number of parameters in the trigger"+
                               " of transition '"+transition.name+"'")
            prog = ""
            for (param,value) in zip(transition.trigger.parameter,
                                     message.parameters):
                assert param.oclIsKindOf(UML14.Parameter)
                assert isinstance(value, SimValue)
                prog += "((" + object.get_class_def().get_name() + ")" + al_objs + "[" + str(object.idnum)+"])." + str(param.name) + "=" + value.get_act_str() +";\n"
            # Execute the code, don't catch vm exceptions
            simulator._execute_statement_in_vm(prog, False)

        # The code to evaluate the guard
        expr = "(("+object.get_class_def().get_name()+")"+\
               al_objs+"["+str(object.idnum)+"])."+\
               simulator.generator.get_guard_method_name(transition)
        result = simulator._evaluate_boolean_expr_in_vm(expr, False)

        #
        # Restore the original working global state
        #
        if old_working_global_state:
            simulator.set_working_global_state(old_working_global_state)

        return result



    def find_enabled(self, object, message, simulator, global_state,
                     opt_do_evaluate_guards = True):
        """
        Find the transitions that are enabled in the Object 'object'
        in the GlobalState 'global_state' when 'message'
        is the current Message in the signal queue/pool
        (if message == None, only completion transitions are considered).
        Returns a set of sets of transitions.
        """
        assert type(opt_do_evaluate_guards) is bool
        assert not message or isinstance(message, Message)
        assert not opt_do_evaluate_guards or isinstance(object, Object)
        assert not opt_do_evaluate_guards or isinstance(simulator, Simulator)
        assert not opt_do_evaluate_guards or isinstance(global_state, GlobalState.GlobalState)
        assert self.uml_reference
        uml_state = self.uml_reference

        #
        # A temporary constraint check
        #
        if uml_state.oclIsKindOf(UML14.State) and uml_state.internalTransition:
            raise SimError("Internal transitions not yet handled")


        #
        # Base case: initial pseudostates
        #
        if (uml_state.oclIsKindOf(UML14.Pseudostate) and
            uml_state.kind == "initial"):
            assert self.children == []
            
            enabled_transitions = []

            assert len(uml_state.outgoing) == 1
            transition = uml_state.outgoing[0]
            assert not transition.trigger

            if transition.guard and opt_do_evaluate_guards:
                if not self.evaluate_guard(object, transition, None,
                                           simulator, global_state):
                    raise SimError("The guard '"+str(transition.guard.expression.body)+"' of the outgoing transition from the initial pseudostate '" + sm_get_qualified_name(uml_state) + "' evaluates to false")

            enabled_transitions.append([("completion",transition)])
            return enabled_transitions



        #
        # Base case: choice pseudostates
        #
        if (uml_state.oclIsKindOf(UML14.Pseudostate) and
            uml_state.kind == "choice"):
            assert self.children == []
            
            enabled_transitions = []

            #
            # First find enabled completion transitions
            #
            has_completion_transitions = False
            has_enabled_completion_transitions = False
            for t in uml_state.outgoing:
                if t.trigger:
                    raise SimError("Ill-formed model: choice pseudostate '" + sm_get_qualified_name(uml_state) + "' has an outgoing transition with a trigger")
                has_completion_transitions = True
                if t.guard and opt_do_evaluate_guards:
                    if not self.evaluate_guard(object, t, None,
                                               simulator, global_state):
                        continue
                has_enabled_completion_transitions = True
                enabled_transitions.append([("completion",t)])

            #
            # It is an error if there are no completion transitions leaving
            # the state
            #
            if not has_completion_transitions: # or not has_enabled_completion_transitions:
                raise SimError("There are no completion transitions leaving from choice pseudostate '" + sm_get_qualified_name(uml_state) + "'")

            #
            # It is an error if there are no enabled completion transitions
            # leaving the state
            #
            if not has_enabled_completion_transitions:
                raise SimError("None of the completion transitions leaving from the choice pseudostate '" + sm_get_qualified_name(uml_state) + "' is enabled in the current state of object " + str(object.idnum))

            return enabled_transitions



        #
        # Base case: final states
        # A final state is not supposed to have outgoing transitions,
        # internal transitions, or deferrable events.
        # Nothing is ever enabled.
        #
        if uml_state.oclIsKindOf(UML14.FinalState):
            assert not uml_state.outgoing or uml_state.outgoing == []
            assert not uml_state.internalTransition
            assert not uml_state.deferrableEvent
            return []



        #
        # Base case: simple states
        #
        if uml_state.oclIsKindOf(UML14.SimpleState):
            assert self.children == []

            enabled_transitions = []

            #
            # First find enabled completion transitions
            #
            has_completion_transitions = False
            for t in uml_state.outgoing:
                if t.trigger:
                    # Not a completion transition
                    continue
                has_completion_transitions = True
                if t.guard and opt_do_evaluate_guards:
                    if not self.evaluate_guard(object, t, None,
                                               simulator, global_state):
                        continue
                enabled_transitions.append([("completion",t)])

            #
            # It is an error if there are completion transitions leaving
            # the state but none of them is enabled
            #
            if has_completion_transitions and enabled_transitions == []:
                raise SimError("There are completion transitions leaving from state '" + sm_get_qualified_name(uml_state) + "' but none of them is enabled in the current state of object " + str(object.idnum))

            #
            # Completion transitions have priority over others
            #
            if enabled_transitions != []:
                return enabled_transitions

            # Only interested in completion transitions -> return
            if not message:
                return enabled_transitions
            
            #
            # Handle signal triggered transitions
            #
            for t in uml_state.outgoing:
                if not t.trigger:
                    continue
                assert t.trigger.oclIsKindOf(UML14.SignalEvent)
                if t.trigger.signal != message.get_signal():
                    continue
                if t.guard and opt_do_evaluate_guards:
                    if not self.evaluate_guard(object, t, message,
                                               simulator, global_state):
                        continue
                enabled_transitions.append([("signal", t)])

            if enabled_transitions != []:
                return enabled_transitions

            # Nothing enabled so far, try deferring
            assert not uml_state.internalTransition
            for e in uml_state.deferrableEvent:
                assert e.oclIsKindOf(UML14.SignalEvent)
                if e.signal == message.get_signal():
                    # Event can be deferred
                    enabled_transitions.append([("defer", None)])
                    return enabled_transitions

            return enabled_transitions



        #
        # Inductive case: submachine states
        #
        if uml_state.oclIsKindOf(UML14.SubmachineState):
            raise SimError("Submachine states not yet handled: '" + sm_get_qualified_name(uml_state) + "'")



        #
        # Inductive case: composite states
        #
        if uml_state.oclIsKindOf(UML14.CompositeState):
            assert not(uml_state.isConcurrent and len(uml_state.subvertex) != len(self.children))
            assert not(uml_state.isConcurrent and len(self.children) <= 1)
            assert not(not uml_state.isConcurrent and len(self.children) != 1)

            #
            # First find the enabled sets of children
            #
            enabled_sets_sets = []
            for child in self.children:
                enabled_sets = child.find_enabled(object, message, simulator, global_state, opt_do_evaluate_guards)
                if enabled_sets != []:
                    enabled_sets_sets.append(enabled_sets)
                    
            if len(enabled_sets_sets) == 1:
                #
                # Only one child had enabled transitions
                #
                return enabled_sets_sets[0]

            elif len(enabled_sets_sets) >= 2:
                #
                # More than one child had enabled transitions
                #

                # Filter out completion transitions that are fired one by one
                completion_sets = []
                remaining_sets_sets = []
                for sets in enabled_sets_sets:
                    remaining_sets = []
                    for set in sets:
                        if len(set) == 1:
                            (reason, t) = set[0]
                            if reason == "completion":
                                assert not t.trigger
                                completion_sets.append(set)
                            else:
                                assert reason == "signal"
                                assert t
                                assert t.trigger
                                remaining_sets.append(set)
                        else:
                            assert len(set) > 1
                            for t in set:
                                assert t.trigger
                            remaining_sets.append(set)
                    if remaining_sets != []:
                        remaining_sets_sets.append(remaining_sets)
                
                if completion_sets != []:
                    return completion_sets
                
                # Handle non-completion transitions
                if len(remaining_sets_sets) >= 1:
                    result_set = []
                    seen_signals = {}
                    for s1 in remaining_sets_sets:
                        seen_signals2 = {}
                        for s2 in s1:
                            for (r,t) in s2:
                                if seen_signals.has_key(t.trigger.signal):
                                    raise SimError("Orthogonal regions have transitions with the same signal trigger")
                                seen_signals2[t.trigger.signal] = True
                                result_set.append([(r,t)])
                        for s in seen_signals2.keys():
                            seen_signals[s] = True
                    return result_set

                return []

            
            #
            # No children had an enabled transition,
            # check if this composite state has some
            #
            assert enabled_sets_sets == []

            #
            # Find and handle completion transitions from a composite state
            #
            # Find out whether all regions are in a final state
            all_in_final = True
            if uml_state.isConcurrent:
                # iterate over regions
                for child in self.children:
                    assert child.uml_reference.oclIsKindOf(UML14.CompositeState)
                    assert len(child.children) == 1
                    if not child.children[0].uml_reference.oclIsKindOf(UML14.FinalState):
                        all_in_final = False
                        break
            else:
                assert len(self.children) == 1
                if not self.children[0].uml_reference.oclIsKindOf(UML14.FinalState):
                    all_in_final = False
            # If all regions were in final states, try firing completion
            # transitons
            if all_in_final:
                # Find all enabled completion transitions
                completion_sets = []
                has_completion_transitions = False
                for t in uml_state.outgoing:
                    if t.trigger:
                        continue  # Not a completion transition
                    has_completion_transitions = True
                    if t.guard and opt_do_evaluate_guards:
                        if not self.evaluate_guard(object, t, None,
                                                   simulator, global_state):
                            continue
                    completion_sets.append([("completion", t)])

                #
                # It is an error if there are completion transitions but none
                # of them is enabled
                #
                if has_completion_transitions and completion_sets == []:
                    raise SimError("None of the completion transitions leaving from the composite state '" + sm_get_qualified_name(uml_state) + "' is enabled in the current state of object " + str(object.idnum))
                #
                # Completion transitions have priority over others
                #
                if has_completion_transitions:
                    return completion_sets

            # Only interested in completion transitions -> return
            if not message:
                return []

            #
            # Handle normal signal event transitions from a composite state
            #
            enabled_sets = []
            if message:
                for t in uml_state.outgoing:
                    if not t.trigger:
                        continue
                    if message.get_signal() != t.trigger.signal:
                        continue
                    if t.guard and opt_do_evaluate_guards:
                        if not self.evaluate_guard(object, t, message,
                                                   simulator, global_state):
                            continue
                    enabled_sets.append([("signal", t)])

            if enabled_sets != []:
                return enabled_sets
            
            # Nothing enabled, try deferring
            assert not uml_state.internalTransition
            for e in uml_state.deferrableEvent:
                assert e.oclIsKindOf(UML14.SignalEvent)
                if e.signal == message.get_signal():
                    # Event can be deferred
                    enabled_sets.append([("defer", None)])
                    return enabled_sets

            return enabled_sets

        #end of "if uml_state.oclIsKindOf(UML14.CompositeState):"


        #
        # Other states and pseudostates are not yet supported
        #
        raise SimError("Currently unsupported state type: "+str(uml_state))




    def find_state(self, uml_state):
        # Find the substate that corresponds to uml_state
        if self.uml_reference == uml_state:
            return self
        for child in self.children:
            found_state = child.find_state(uml_state)
            if found_state:
                return found_state
        return None
                       

    def fire(self, object, transitions, simulator, global_state):
        """Fire enabled transitions"""

        assert len(transitions) == 1

        new_state = self.copy()

        # A list of Sim.Trace.TraceAction
        action_list = []

        #
        # Consume event when firing transition with proper signal event trigger
        #
        reason = None
        for (why,transition) in transitions:
            if why == "completion" or why == "ATO" or why == "defer":
                assert not reason or reason == why
                reason = why
                continue
            elif why == "signal":
                assert not reason or reason == why
                reason = why
                assert transition.trigger
                assert transition.trigger.signal == object.sm_input_queue[0].get_signal()
                continue
            else:
                raise SimError("Unknown reason to fire a transition")
        assert reason

        if reason == "signal":
            consumed_event = object.sm_input_queue[0]
        
        if reason == "signal":
            object.consume()
            # Update action list
            action_list.append(Trace.TA_Consumption(object.idnum,
                                                    consumed_event))

        if reason == "defer":
            object.sm_defer_queue.append(object.sm_input_queue[0].copy())
            object.consume()
            return new_state, action_list

        for (why,transition) in transitions:
            assert why == "signal" or why == "completion" or why == "ATO"
            assert transition

            #
            # Reset the list of observed actions
            #
            simulator.reset_observed_actions()

            # Update action list
            action_list.append(Trace.TA_ExecuteTransition(object.idnum,
                                                          transition))

            uml_source = transition.source
            source = new_state.find_state(uml_source)
            assert source
            uml_target = transition.target
            lca_path = Utils.StateMachines.get_least_common_ancestor_path(uml_source,uml_target)
            assert (lca_path[-1].isConcurrent == False or
                    uml_target == lca_path[-1])
            source_parent_path = Utils.StateMachines.get_parent_path(uml_source)
            target_parent_path = Utils.StateMachines.get_parent_path(uml_target)
            #sm_print_parent_path(source_parent_path)
            #sm_print_parent_path(target_parent_path)
            #sm_print_parent_path(lca_path)

            # Remove the common prefix from source_parent_path and
            # target_parent_path
            for i in range(0, len(lca_path)):
                source_parent_path.pop(0)
                target_parent_path.pop(0)
            #sm_print_parent_path(source_parent_path)
            #sm_print_parent_path(target_parent_path)

            #
            # Handle signal parameter reception
            #
            if transition.trigger and transition.trigger.parameter:
                assert transition.trigger.oclIsKindOf(UML14.SignalEvent)
                if reason == "ATO":
                    raise SimError("ATO signals with parameters not yet implemented")
                assert reason == "signal"
                # Remind me that I should eventually check whether
                # signal parameters should be read before or after
                # executing the exit actions...
                #print "INTERNAL WARNING: check something..."
                assert len(consumed_event.signal.feature) == len(transition.trigger.parameter)
                # The code to assign the local variables with signal params
                prog = ""
                for (param,value) in zip(transition.trigger.parameter,consumed_event.parameters):
                    assert param.oclIsKindOf(UML14.Parameter)
                    assert isinstance(value, SimValue)
                    prog += "((" + object.get_class_def().get_name() + ")" + al_objs + "[" + str(object.idnum)+"])." + str(param.name) + "=" + value.get_act_str() + ";\n"
                # Execute the code, don't catch vm exceptions
                simulator._execute_statement_in_vm(prog, False)

            
            # Execute the exit actions of source subtree
            uml_source_root = source_parent_path[0]
            source_root = new_state.find_state(uml_source_root)
            source_root.execute_exit_actions(object, "Dummy env placeholder")

            # Execute the transition action
            if transition.effect:
                # The code to execute the action
                prog = "(("+object.get_class_def().get_name()+")"+al_objs+\
                       "["+str(object.idnum)+"])."+\
                       simulator.generator.get_effect_method_name(transition)+\
                       ";\n"
                # Execute it
                simulator.action_vm.initialize(actlang.grammar.parse(prog).translate(simulator.action_env))
                try:
                    simulator.action_vm.run()
                except actlang.errors.AssertionError_, orig:
                    e = "An assertion violation occurred when "+\
                        "object "+str(object.idnum)+" executes "+\
                        "transition '"+get_qualified_name(transition)+"' "+\
                        "from '"+get_qualified_name(transition.source)+"' "+\
                        "to '"+get_qualified_name(transition.target)+"' "+\
                        "with effect '"+str(transition.effect.script.body)+"'"
                    raise Sim.Error.ActionLanguageError(orig, e)
                except actlang.errors.Error, orig:
                    e = "An error occurred when object "+str(object.idnum)+\
                        " executes transition '"+get_qualified_name(transition)+"'"+\
                        " from '"+get_qualified_name(transition.source)+"'"+\
                        " to '"+get_qualified_name(transition.target)+"'"+\
                        " with effect '"+str(transition.effect.script.body)+"'"
                    raise Sim.Error.ActionLanguageError(orig, e)


            # Initialize the new subtree for target
            uml_target_root = target_parent_path[0]
            target_root = SM_StateNode_from_uml_path(target_parent_path)

            #source_root._print("sr ")
            #target_root._print("tr ")
            
            # Replace the source subtree with target subtree
            new_state = new_state.replace(source_root, target_root)

            # Execute the entry actions of target subtree
            target_root.execute_entry_actions(object, "Dummy env placeholder")


            action_list2 = simulator._execute_observed_actions(global_state,
                                                               object)
            action_list.extend(action_list2)

        # Reset action list observed by the simulator
        simulator.reset_observed_actions()

        #
        # Flush defer queue #if state was changed
        #
        if object.sm_defer_queue != []: # and cmp(self, new_state) != 0:
            for e in object.sm_input_queue:
                object.sm_defer_queue.append(e)
            object.sm_input_queue = object.sm_defer_queue
            object.sm_defer_queue = []

        return new_state, action_list


    def execute_entry_actions(self, object, env):
        """Execute the entry actions of this state and its children"""
        uml_state = self.uml_reference
        if uml_state.oclIsKindOf(UML14.State) and uml_state.entry:
            raise SimError("Entry actions not yet handled")
        for child in self.children:
            child.execute_entry_actions(object, env)


    def execute_exit_actions(self, object, env):
        """Execute the exit actions of the children and this state"""
        uml_state = self.uml_reference
        for child in self.children:
            child.execute_exit_actions(object, env)
        if uml_state.oclIsKindOf(UML14.State) and uml_state.exit:
            raise SimError("Exit actions not yet handled")


def SM_StateNode_from_uml_state(uml_state):
    assert (uml_state.oclIsKindOf(UML14.StateMachine) or
            uml_state.oclIsKindOf(UML14.State) or
            uml_state.oclIsKindOf(UML14.Pseudostate))

    if uml_state.oclIsKindOf(UML14.SubmachineState):
        raise "Submachine states not yet handled"

    node = SM_StateNode()
    node.uml_reference = uml_state
    if uml_state.oclIsKindOf(UML14.Pseudostate) and uml_state.kind=="initial":
        node.isInitial = True
    else:
        node.isInitial = False
    node.children = []
    #perform default entry
    if uml_state.oclIsKindOf(UML14.CompositeState):
        node.isConcurrent = uml_state.isConcurrent
        if node.isConcurrent:
            for s in uml_state.subvertex:
                node.children.append(SM_StateNode_from_uml_state(s))
        else:
            ##find the initial state
            for s in uml_state.subvertex:
                if s.oclIsKindOf(UML14.Pseudostate) and s.kind=="initial":
                    node.children.append(SM_StateNode_from_uml_state(s))
            if node.children==[]:
                raise SimError("Composite state " + sm_get_qualified_name(uml_state) + " without an initial state")
            if len(node.children) > 1:
                raise SimError("Composite state " + sm_get_qualified_name(uml_state) + " with multiple initial states")
    else:
        node.isConcurrent = False

    return node


def SM_StateNode_from_uml_path(uml_path):
    assert uml_path != []
    
    if len(uml_path) == 1:
        return SM_StateNode_from_uml_state(uml_path[0])

    uml_state = uml_path[0]

    assert uml_state.oclIsKindOf(UML14.CompositeState)

    if uml_state.oclIsKindOf(UML14.SubmachineState):
        raise SimError("Submachine states not yet handled")

    uml_path = uml_path[1:len(uml_path)]
    
    node = SM_StateNode()
    node.uml_reference = uml_state
    if uml_state.oclIsKindOf(UML14.Pseudostate) and uml_state.kind=="initial":
        node.isInitial = True
    else:
        node.isInitial = False
    #perform default entry
    if uml_state.oclIsKindOf(UML14.CompositeState):
        node.isConcurrent = uml_state.isConcurrent
        if node.isConcurrent:
            if uml_path == []:
                # Find initial states of regions
                for s in uml_state.subvertex:
                    node.children.append(SM_StateNode_from_uml_state(s))
            else:
                for s in uml_state.subvertex:
                    if s == uml_path[0]:
                        child = SM_StateNode_from_uml_path(uml_path)
                    else:
                        child = SM_StateNode_from_uml_state(s)
                    node.children.append(child)
        else:
            node.children.append(SM_StateNode_from_uml_path(uml_path))
    else:
        node.isConcurrent = False
        assert uml_path == []
    return node


#
# Types of attributes supported in Simulator
#
class SimType:
    def _name(self):
        raise "Abstract class error"
    def al_type(self):
        raise "Abstract class error"
    def compare(self, other):
        assert isinstance(other, SimType)
        return cmp(self._name(), other._name())

class SimType_Int(SimType):
    def _name(self):
        return "Int"
    def al_type(self):
        return "int"

class SimType_Boolean(SimType):
    def _name(self):
        return "Boolean"
    def al_type(self):
        return "boolean"

class SimType_ObjectRef(SimType):
    name=None
    def __init__(self, name = None):
        """
        'name' is the name of the class to which this is a reference type.
        If 'name' is None, then this is a reference type to a 'generic Object'
        or 'void *' in C-terms.
        """
        assert (name is None) or (type(name) is str)
        self.name = name
    def _name(self):
        s = "ObjectRef"
        if self.name != None:
            s += " "+str(self.name)
        return s
    def al_type(self):
        if self.name == None:
            return "$$$Uml_Base"
        return self.name	
    def compare(self, other):
        assert isinstance(other, SimType)
        if (self.name is None) and isinstance(other, SimType_ObjectRef):
            return 0
        if isinstance(other, SimType_ObjectRef) and (other.name is None):
            return 0
        return cmp(self._name(), other._name())
    def get_target_name(self):
        assert self.name
        assert type(self.name) is str
        return self.name

simtype_map = {
    "boolean" : SimType_Boolean,
    "int" : SimType_Int,
}


class SimValue:
    def get_type(self):
        raise "Abstract class error"
    def get_value(self):
        raise "Abstract class error"
    def set_value(self, v):
        raise "Abstract class error"
    def set_default_value(self):
        raise "Abstract class error"
    def is_defined(self):
        raise "Abstract class error"
    def get_str(self):
        raise "Abstract class error"
    def __str__(self):
        return self.get_str()
    def get_act_str(self):
        raise "Abstract class error"
    def set_value_from_al_str(self,str):
        raise "Abstract class error"
    def get_copy(self):
        raise "Abstract class error"
    def compare(self, other):
        raise "Abstract class error"

def simValue_from_simType(sim_type):
    assert isinstance(sim_type, SimType)
    if isinstance(sim_type, SimType_Int):
        return SimValue_Int(None)
    if isinstance(sim_type, SimType_Boolean):
        return SimValue_Boolean(None)
    if isinstance(sim_type, SimType_ObjectRef):
        return SimValue_ObjectRef(None)
    raise "Unsupported SimType " + sim_type._name()

class SimValue_Boolean(SimValue):
    v = None
    def __init__(self, v):
        self.set_value(v)
    def get_type(self):
        return SimType_Boolean()
    def get_value(self):
        return self.v
    def set_value(self, v):
        assert (v == None) or (v == False) or (v == True)
        self.v = v
    def set_default_value(self):
        self.set_value(None) #self.v = False
    def is_defined(self):
        return self.v != None
    def get_str(self):
        return str(self.v)
    def get_act_str(self):
        if self.v == True:
            return "true"
        if self.v == False:
            return "false"
        raise SimError("Cannot convert SimValue_Boolean '"+str(s)+"' to Jumbala")
    def set_value_from_al_str(self, s):
        if s == "true":
            self.v = True
        elif s == "false":
            self.v = False
        else:
            raise SimError("Bad initialization value for Boolean SimValue: "+str(s))
    def get_copy(self):
        return SimValue_Boolean(self.v)
    def compare(self, other):
        assert self.get_type().compare(other.get_type()) == 0
        return cmp(self.v, other.v)


class SimValue_Int(SimValue):
    v = None
    def __init__(self, v):
        self.v = v
    def get_type(self):
        return SimType_Int()
    def get_value(self):
        return self.v
    def set_value(self, v):
        self.v = v
    def set_default_value(self):
        self.set_value(None) #self.v = 0
    def is_defined(self):
        return self.v != None
    def get_str(self):
        return str(self.v)
    def get_act_str(self):
        if Utils.String.IsInt(self.v):
            return self.get_str()
        raise SimError("Cannot convert SimValue_Int '"+str(s)+"' to Jumbala")
    def set_value_from_al_str(self, s):
        self.v = int(s)
    def get_copy(self):
        return SimValue_Int(self.v)
    def compare(self, other):
        assert self.get_type().compare(other.get_type()) == 0
        return cmp(self.v, other.v)


class SimValue_ObjectRef(SimValue):
    v = None
    def __init__(self, v):
        self.v = v
    def get_type(self):
        return SimType_ObjectRef()
    def get_value(self):
        return self.v
    def set_value(self, v):
        assert isinstance(v, Object) or v == "null" or v == None 
        self.v = v
    def set_default_value(self):
        self.set_value(None)
    def is_defined(self):
        return self.v != None
    def get_str(self):
        #return str(self.v)
        if isinstance(self.v, Object):
            return "ref to Object " + str(self.v.idnum)
        elif self.v == "null":
            return "null"
        else:
            return str(self.v)
    def get_act_str(self):
        if isinstance(self.v, Object):
            #return "(" +self.v.get_class_def().get_name() + ")" + al_objs+"["+str(self.v.idnum)+"]"
            return "(" + self.v.get_class_def().get_name() + ")$$sim_objects[" + str(self.v.idnum) + "]"
        elif self.v == "null":
            return "null"
        else:
            assert False
    def get_copy(self):
        return SimValue_ObjectRef(self.v)
    def compare(self, other):
        """
        Compare the referenced types and the idnum of the referenced objects.
        """
        result = self.get_type().compare(other.get_type())
        if result != 0:
            return result
        return cmp(self.get_str(), other.get_str())


class VariableDef:
    # Link to the corresponding UML property/association end
    #uml_reference = None
    # name
    #name = None
    # a SimType
    #t = None
    # a string or None
    #model_default_value_str = None

    def __init__(self, uml_reference, name, t, model_default_value_str):
        """
        'uml_reference' should be an UML Attribute or AssociationEnd
        to which this variable corresponds to.
        't' should be the SimType of the variable.
        """
        assert (uml_reference.oclIsKindOf(UML14.Attribute) or
                uml_reference.oclIsKindOf(UML14.AssociationEnd))
        assert isinstance(name, str)
        assert name != ""
        assert isinstance(t, SimType)
        assert (model_default_value_str == None or
                isinstance(model_default_value_str, str))

        self.uml_reference = uml_reference
        self.name = name
        self.t = t
        self.model_default_value_str = model_default_value_str

    def get_uml_reference(self):
        """
        Return the UML source of this variable,
        i.e., an Attribute or an AssociationEnd.
        """
        assert self.uml_reference
        return self.uml_reference
        
    def get_type(self):
        """
        Get the SimType of this variable.
        """
        return self.t

    def get_name(self):
        """
        Get the name of this variable.
        """
        return self.name
                           
    def _print(self, indent = Utils.String.Indent()):
        print str(indent) + self.t._name() + " " + self.name,
        if self.model_default_value_str:
            print "(model default: "+self.model_default_value_str+")",
        print "(UML ref: " + str(self.uml_reference) + ")"


def _SimType_from_UML_class(c):
    assert c.oclIsKindOf(UML14.Classifier)
    assert c.name
    name = str(c.name)
    if c.oclIsKindOf(UML14.DataType) and (name in simtype_map):
        t = simtype_map[name]()
    else:
        Utils.UML14.check_name(c)
        t = SimType_ObjectRef(name)
    return t


def _VariableDef_from_Attribute(attr):
    assert attr.oclIsKindOf(UML14.Attribute)

    if attr.multiplicity.range and (attr.multiplicity.range.lower[0] != 1 or
                                    attr.multiplicity.range.upper[0] != 1):
        raise "Attribute with multiplicity other than 1..1 not yet handled"
    
    # Name resolution
    name = str(attr.name)
    if not name or name == "":
        raise "Unnamed attributes not supported"

    # Type resolution
    t_class = attr.type
    assert t_class
    t = _SimType_from_UML_class(t_class)
#    t = None
#    if t_class.name and simtype_map.has_key(str(t_class.name)):
#        t = simtype_map[str(t_class.name)]()
#    else:
#        t = SimType_ObjectRef()
#        #raise "Cannot yet handle attribute of type class " + str(t_class.name)

    # Default value resolution
    model_default_value_str = None
    if attr.initialValue and attr.initialValue.body:
        model_default_value_str = str(attr.initialValue.body)
        
    return VariableDef(attr, name, t, model_default_value_str)



class MethodDef:
    # Link to the corresponding UML Method
    #uml_reference = None
    # name
    #name = None
    # a SimType
    #return_type = None
    # a list of (name,SimType)
    #param_list = None

    def __init__(self, uml_method):
        assert uml_method
        assert uml_method.oclIsKindOf(UML14.Method)
        Utils.UML14.check_name(uml_method)

        self.uml_method = uml_method
        self.uml_reference = uml_method
        self.name = str(uml_method.name)
        self.return_type = None
        self.param_list = []

        if uml_method.ownerScope != "instance":
            raise SimError("Method owner scope '"+str(uml_method.ownerScope)+"' not supported")
        
        for param in uml_method.parameter:
            assert param.oclIsKindOf(UML14.Parameter)
            assert param.name and str(param.name) != ""
            assert param.name[0] != "$"
            assert param.type
            assert param.defaultValue
            if param.defaultValue.body:
                raise SimError("Method attribute default values not yet implemented")
            if param.kind == "in":
                self.param_list.append((param.name,_SimType_from_UML_class(param.type)))
            elif param.kind == "return":
                assert not self.return_type
                self.return_type = _SimType_from_UML_class(param.type)
            else:
                raise SimError("Operation parameter kind '"+str(param.kind)+"' not supported")
        print self.param_list


    def get_uml_method(self):
        assert self.uml_method
        return self.uml_method
    
    def get_return_type(self):
        return self.return_type

    def get_name(self):
        return str(self.get_uml_method().name)
                           
    def get_as_str(self, sep = ""):
        s = ""
        if self.return_type:
            s += sep + self.return_type._name()
        else:
            s += sep + "void"
        s += " " + self.get_name() + "("
        lsep = ""
        for (name,t) in self.param_list:
            s += lsep + t._name() + " " + name
            lsep = ","
        s += ")"
        return s

    def _hash_string(self):
        s = self.get_name()
        for (name,t) in self.param_list:
            s += "#" + t._name()
        return str(s)

    def __hash__(self):
        return hash(self._hash_string())

    def __cmp__(self, other):
        """
        Two MethodDefs are same iff they have the same name and
        all the parameters match by types.
        """
        return cmp(self._hash_string(), other._hash_string())

    def _print(self, sep = ""):
        print self.get_as_str(sep)



class SignalDef:
    """
    A correspondent for UML Signals.
    Parameter types are resolved to SimTypes, though.
    """
    def __init__(self, uml_signal):
        assert uml_signal.oclIsKindOf(UML14.Signal)
        assert Utils.UML14.check_signal_name(uml_signal)
        
        self.uml_signal = uml_signal
        self.params = []
        for uml_parameter in uml_signal.feature:
            assert uml_parameter.oclIsKindOf(UML14.Attribute)
            self.params.append(_VariableDef_from_Attribute(uml_parameter))

    def get_uml_signal(self):
        """
        Get the corresponding UML Signal.
        """
        return self.uml_signal

    def get_name(self):
        """
        Get the name of the signal.
        """
        return str(self.uml_signal.name)

    def get_nof_parameters(self):
        """
        Get the number of parameters of this Signal.
        """
        return len(self.params)

    def get_parameter_type(self, i):
        """
        Get the SimType of the i:th parameter.
        """
        return self.params[i].get_type()

    def is_external(self):
        """
        Is this an external signal?
        """
        return self.get_name().startswith("$")

    def __str__(self):
        s = ""
        if self.is_external():
            s += "external "
        s += "Signal "+get_qualified_name(self.uml_signal)+"("
        s += ",".join([p.get_type()._name() for p in self.params])
        s += ")"
        return s



class ClassDef:
    """
    Data structure for storing Class related information in the simulator.
    """
    # Link to the UML class
    #uml_reference = None
    # Derived: Link to the UML state machine (if any) describing
    # the behavior of the class
    #uml_state_machine = None
    # Variable defintions corresponding to attributes and associations
    #_variable_defs = {}
    # Set of MethodDefs
    # methods = {}
    
    def __init__(self, uml_class):
        """
        Create a new ClassDef corresponding to the UML Class 'uml_class'.
        """
        assert uml_class.oclIsKindOf(UML14.Class)
        Utils.UML14.check_name(uml_class)
        
        self.uml_reference = uml_class
        self._uml_class = uml_class
        self.uml_state_machine = None
        self._variable_defs = {}
        self.methods = Set()
        
        #
        # Generalizations not yet handled
        #
        if uml_class.generalization:
            raise SimError("Class "+get_qualified_name(uml_class)+
                           ": Generalizations not yet handled")

        #
        # Behavioral element reference resolution
        #
        if uml_class.behavior:
            assert len(uml_class.behavior) == 1
            self.uml_state_machine = uml_class.behavior[0]
            assert self.uml_state_machine.oclIsKindOf(UML14.StateMachine)
            assert sm_validate_legal_orthogonality(self.uml_state_machine)


        #
        # Find out attribute, association, and operation definitions
        #
        (attrs, assocs) = Utils.UML14.find_attributes_and_associations(uml_class)

        # Make VariableDefs corresponding to attributes
        for (name, attr) in attrs.iteritems():
            vdef = _VariableDef_from_Attribute(attr)
            # Duplicate names should have been catched earlier
            assert not self._variable_defs.has_key(vdef.get_name())
            self._variable_defs[vdef.get_name()] = vdef

        # Find variables corresponding to associations
        for (name, (other_assoc_end, assoc_end)) in assocs.iteritems():
            vdef = self._VariableDef_from_AssociationEnd(assoc_end)
            if not vdef:
                continue
            # Duplicate names should have been catched earlier
            assert not self._variable_defs.has_key(vdef.get_name())
            self._variable_defs[vdef.get_name()] = vdef

        # Find methods
        for method in [m for m in uml_class.feature if
                       m.oclIsKindOf(UML14.Method)]:
            Utils.UML14.check_name(method)
            if method.name == uml_class.name:
                raise SimError("Constructors not yet handled");
            method_def = MethodDef(method)
            if method_def in self.methods:
                raise SimError("Method '"+str(method.name)+"' defined twice "+
                               "with same parameter types in class "+
                               get_qualified_name(uml_class))
            self.methods.add(method_def)

        # Check that everything in features has been handled
        for feat in uml_class.feature:
            if feat.oclIsKindOf(UML14.Attribute):
                pass
            elif feat.oclIsKindOf(UML14.Method):
                pass
            else:
                raise SimError("Class feature not handled: " + str(feat))



    def _VariableDef_from_AssociationEnd(self, assoc_end):
        assert assoc_end.oclIsKindOf(UML14.AssociationEnd)
        assoc = assoc_end.association
        assert assoc.oclIsKindOf(UML14.Association)
        assert len(assoc.connection) == 2
        other_assoc_end = assoc.connection[0]
        if assoc_end == other_assoc_end:
            other_assoc_end = assoc.connection[1]
        assert other_assoc_end != assoc_end

        # Name resolution
        name = str(other_assoc_end.name)
        if not name:
            if other_assoc_end.isNavigable:
                sys.stderr.write("Warning: a navigable association end with "
                                 "no name: "+Utils.Models.get_qualified_name(other_assoc_end)+"\n")
            return None
        if not other_assoc_end.isNavigable:
            sys.stderr.write("Warning: ignoring a non-navigable association "
                             "end with name: '"+Utils.Models.get_qualified_name(other_assoc_end)+"'\n")
            return None


        if other_assoc_end.multiplicity.range:
            if other_assoc_end.multiplicity.range.lower[0] > 1 or \
                   other_assoc_end.multiplicity.range.upper[0] != 1:
                raise SimError("Association with multiplicities other than 0..1 and 1..1 not yet supported")
        if assoc_end.aggregation != "none" or other_assoc_end.aggregation != "none":
            raise SimError("Aggregation and composition associations not yet handled")
        if assoc_end.targetScope != "instance" or other_assoc_end.targetScope != "instance":
            raise SimError("Association end target scopes other than 'instance' not yet handled")
        
        # Type "resolution"
        assert assoc_end.participant
        assert other_assoc_end.participant
        t = SimType_ObjectRef(str(other_assoc_end.participant.name))

        # No default value
        model_default_value_str = None

        return VariableDef(other_assoc_end, name, t, model_default_value_str)


    def get_uml_class(self):
        """
        Get the UML Class to which this ClassDef corresponds to.
        """
        assert self._uml_class
        assert self._uml_class.oclIsKindOf(UML14.Class)
        return self._uml_class


    def get_uml_state_machine(self):
        """
        Return the UML StateMachine of this object (if any).
        """
        return self.uml_state_machine


    def get_name(self):
        """
        Get the name of the UML Class to which this ClassDef corresponds to.
        """
        return str(self.get_uml_class().name)

    
    def get_variable_defs(self):
        """
        Get the dictionary from the names of the variables in this class
        to the corresponding VariableDef objects.
        WARINING: You should NOT modify this dictionary or the VariableDef
        objects referenced by it.
        """
        return self._variable_defs


    def _print(self, indent):
        print str(indent) + "class " + self.get_name()
        indent.push()
        for (vname,vdef) in self._variable_defs.iteritems():
            vdef._print(indent)
        for m in self.methods:
            m._print(indent)
        if self.uml_state_machine:
            print str(indent) + "uml_state_machine = " + str(self.uml_state_machine)
        indent.pop()




class Simulator:
    # The model
    # uml_model = None

    # A mapping from uml_model's transitions' xmiids to themselves
    xmiid_to_uml_transition = {}
    
    # UML class -> ClassDef
    uml_class_to_class_def_map = {}

    # A list of global states
    trace = None

    # The list of classes defined in the model
    #uml_classes_in_model = []
    # map from UML class to index in classes_in_model
    #uml_class_to_uml_class_index = {}

    # The list of signals defined in the model
    # uml_signals = []
    # signal_defs = []
    # signal_name_to_signal_def = {}

    #Static environment for action language interpreter
    #action_env = None

    #Virtual machine for action language interpreter
    #action_vm = None

    #Action generator
    #generator = None
 
    # The "current" state
    #working_global_state = None

    # A constant static flag
    # The name of the assertion condition
    _ASSERTION_CONDITION = "$$SMUML.evaluate_assertions"
    _ASSERTION_CONDITION_ATTR = "evaluate_assertions"


    def __init__(self,
                 uml_model,
                 #deployment_diagram,
                 opt_evaluate_jumbala_assertions = True):
        """
        Initialize a new simulator for the UML Model 'uml_model'.
        
        Action language assert statements are evaluated iff
        the 'opt_evaluate_jumbala_assertions' flag is True.
        """
        assert uml_model.oclIsKindOf(UML14.Model)
        assert type(opt_evaluate_jumbala_assertions) is bool

        # The model
        self.uml_model = uml_model

        # Options
        self.opt_evaluate_jumbala_assertions = opt_evaluate_jumbala_assertions

        # Signal related data structures
        self.uml_signals = []
        self.signal_defs = []
        self.signal_name_to_signal_def = {}
        self.uml_signal_to_signal_def = {}

        self.trace = Trace.Trace()

        self.uml_classes_in_model = []
        self.uml_class_to_uml_class_index = {}

        self.working_global_state = None

        # Set default choose function query hooks
        self.set_default_choose_boolean_hook()
        self.set_default_choose_int_hook()

        self.reset_observed_actions()

        # Find transitions in the model
        uml_transitions = get_elements_of_type(uml_model, UML14.Transition)
        # Build the reverse map from XMIid to models's transitions
        for uml_t in uml_transitions:
            self.xmiid_to_uml_transition[uml_t.__XMIid__()] = uml_t


        #
        # Find all the signals defined in the model
        #
        self.uml_signals = get_elements_of_type(uml_model, UML14.Signal)
        self.uml_signals.sort(lambda a,b: cmp(str(a.name),str(b.name)))
        for uml_signal in self.uml_signals:
            signal_name = str(uml_signal.name)
            signal_def = SignalDef(uml_signal)
            self.signal_defs.append(signal_def)
            assert not signal_name in self.signal_name_to_signal_def
            self.signal_name_to_signal_def[signal_name] = signal_def
            self.uml_signal_to_signal_def[uml_signal] = signal_def

        #
        # Find class definitions in model
        #
        self.uml_classes_in_model = get_elements_of_type(uml_model,
                                                         UML14.Class)

        for index,uml_class in enumerate(self.uml_classes_in_model):
            self.uml_class_to_uml_class_index[uml_class] = index
        # build corresponding ClassDefs
        self.uml_class_to_class_def_map = {}
        self.name_to_class_def_map = {}
        for c in self.uml_classes_in_model:
            assert c.name
            if self.uml_class_to_class_def_map.has_key(c):
                raise "Should not happen: Class "+str(c.name)+" defined twice"
            class_def = ClassDef(c)
            self.uml_class_to_class_def_map[c] = class_def
            self.name_to_class_def_map[str(c.name)] = class_def

        #
        # Initalize the action language generator
        #
        self.generator = ActionGenerator()
        for (uml_class,class_def) in self.uml_class_to_class_def_map.iteritems():
           self.generator.add_uml_class(uml_class, class_def)
            
        # Obtain a static environment and a virtual machine for action language
        self.action_env = actlang.basetypes.getInitialEnvironment()
        self.action_vm = actlang.basetypes.getInitialMachine()

        # Get the static part of the program with constructor code included
        (prog,hooks) = self._get_static_Jumbala_code()

        # Debugging
        if False:
            debug_filename = "DEBUG_static.jmb"
            f = open(debug_filename, 'w')
            f.write(prog)
            f.close()

        # Set hooks for native thingies
        self.action_env.setNativeHooks(hooks)

        # Parse the program into a syntax tree.
        try:
            tree = actlang.grammar.parse(str(prog))
        except actlang.errors.Error, e:
            debug_filename = "DEBUG.jmb"
            f = open(debug_filename, 'w')
            f.write(prog)
            f.close()
            print e
            print "Consult "+debug_filename+" for the erroneous program code"
            sys.exit(-1)

        # Translate the syntax tree into internal code.
        try:
            main = tree.translate(self.action_env)
        except actlang.errors.StaticError, e:
            debug_filename = "DEBUG.jmb"
            f = open(debug_filename, 'w')
            f.write(prog)
            f.close()
            print e
            print "Consult "+debug_filename+" for the erroneous program code"
            sys.exit(-1)

        # Load the code into the virtual machine.
        try:
            self.action_vm.initialize(main)
        except actlang.errors.StaticError, e:
            print e
            assert False
        # Run the top-level statements.
        try:
            self.action_vm.run()
        except actlang.errors.RuntimeError_, e:
            print e
            assert False


        #
        # Build initial global state
        #
        pseudo_initial_global_state = GlobalState.GlobalState()

        # A Model Objects to the initial global state objects
        uml_object_to_object_map = {}

        # Find all Objects in the deployment diagram
        #assert deployment_diagram.oclIsKindOf(DIA20.Diagram)
        #assert deployment_diagram.semanticModel.typeInfo=="DeploymentDiagram"
        #uml_objects = []
        #for e in deployment_diagram.contained:
        #    if e.semanticModel.element.oclIsKindOf(UML14.Object):
        #        uml_objects.append(e.semanticModel.element)
        # Find all Objects in the Model
        uml_objects = get_elements_of_type(uml_model, UML14.Object)
        # Sort Objects according their XMI identifiers
        Utils.Models.xmi_sort(uml_objects)
        # Instantiate their skeletons in the initial global state
        for uml_object in uml_objects:
            assert len(uml_object.classifier) == 1
            uml_classifier = uml_object.classifier[0]
            assert uml_classifier.oclIsKindOf(UML14.Classifier)
            # Create the object in the initial global state
            obj = Object(self.uml_class_to_class_def_map[uml_classifier],
                         uml_object,
                         self.trace.get_next_free_oid())
            pseudo_initial_global_state.objects.append(obj)
            # Update object_map
            uml_object_to_object_map[uml_object] = obj
        # Initialize the objects
        for o in pseudo_initial_global_state.objects:
            o._initialize_from_uml_object(uml_object_to_object_map)


        
        prog = ""

        # Copy simulator state to action language objects but
        # do not copy variable values
        prog += self.copy_simstate_to_action(pseudo_initial_global_state,
                                             True, False)

        # Forget things
        self.trace = Trace.Trace()
        initial_global_state = GlobalState.GlobalState()

        
        # Reset action list observed by the simulator
        self.reset_observed_actions()

        #parse
        try:
            self.action_vm.initialize(actlang.grammar.parse(prog).translate(self.action_env))
        except actlang.errors.StaticError, e:
            print e
            assert False

        #print "running initialisation...\n"    
        #run    
        try:
            self.action_vm.run()
        except actlang.errors.RuntimeError_, e:
            print e
            assert False

        #print self.action_env    

        # Execute the "new" and possible "send" actions 
        actions = self._execute_observed_actions(initial_global_state, None)

        # Reset action list observed by the simulator
        self.reset_observed_actions()

        self.copy_state_from_action_to_sim(initial_global_state)

        self.trace.add_global_state(initial_global_state)

        #
        # Find the possibly new idnums of the objects that were
        # in the original initial configuration
        #
        for (index,o) in enumerate(pseudo_initial_global_state.objects):
            assert index == o.idnum

            expr = "$$sim_objects["+str(index)+"].__id"
            idnum = self._evaluate_integer_expr_in_vm(expr, True)
            assert idnum >= 0
            assert idnum < len(initial_global_state.objects)
            initial_global_state.objects[idnum].set_uml_object(o.get_uml_object())

        #
        # Reset the action language environment.
        # Initialize it with the static code (class declarations)
        # without constructor codes
        #
        
        # Obtain a static environment and a virtual machine for action language
        self.action_env = actlang.basetypes.getInitialEnvironment()
        self.action_vm = actlang.basetypes.getInitialMachine()

        # Get the static part of the program without constructor code
        (prog,hooks) = self._get_static_Jumbala_code()

        # Debugging
        if False:
            debug_filename = "DEBUG.jmb"
            f = open(debug_filename, 'w')
            f.write(prog)
            f.close()

        # Parse the program into a syntax tree.
        try:
            tree = actlang.grammar.parse(str(prog))
        except actlang.errors.Error, e:
            print e
            print "Consult "+debug_filename+" for the erroneous program code"
            sys.exit(-1)

        # Set hooks for native thingies
        self.action_env.setNativeHooks(hooks)

        # Translate the syntax tree into internal code.
        try:
            main = tree.translate(self.action_env)
        except actlang.errors.StaticError, e:
            print e
            print "Consult "+debug_filename+" for the erroneous program code"
            sys.exit(-1)
        # Load the code into the virtual machine.
        try:
            self.action_vm.initialize(main)
        except actlang.errors.StaticError, e:
            print e
            assert False
        # Run the top-level statements.
        try:
            self.action_vm.run()
        except actlang.errors.RuntimeError_, e:
            print e
            print "Consult "+debug_filename+" for the erroneous program code"
            sys.exit(-1)


    def get_uml_model(self):
        """
        Get the UML Model this instance is simulating.
        """
        assert self.uml_model
        return self.uml_model


    def get_class_defs(self):
        """
        Get the ClassDefs corresponding to the UML Classes in the model.
        Thou shall NOT modify the returned ClassDefs.
        """
        return self.uml_class_to_class_def_map.values()


    def get_class_def(self, x):
        """
        Get the ClassDefs corresponding to the name or UML Class 'x'
        in the model.
        Thou shall NOT modify the returned ClassDef.
        """
        if type(x) is str:
            return self.name_to_class_def_map[x]
        assert x.oclIsKindOf(UML14.Class)
        return self.uml_class_to_class_def_map[x]


    def get_signal_defs(self):
        """
        Get the SignalDefs corresponding to the UML Signals in the Model.
        Thou shall NOT modify the list nor the SignalDefs in it.
        """
        return self.signal_defs


    def get_signal_def(self, x):
        """
        Get the SignalDef corresponding to the UML Signal 'x' or
        of the signal with name 'x' in case 'x' is a string.
        Thou shall NOT modify the returned SignalDef.
        Raises a RuntimeError if the model has no signal with that name.
        """
        if type(x) is str:
            if not self.signal_name_to_signal_def.has_key(x):
                raise RuntimeError("The model has no signal with name '"+x+"'")
            return self.signal_name_to_signal_def[x]
        assert x.oclIsKindOf(UML14.Signal)
        return self.uml_signal_to_signal_def[x]



    # The hooks for querying values for non-deterministic choose functions
    # They should be only manipulated by using the methods below
    # By default they are linked to interactive query functions
    # For the arguments given to the hooks, see
    # _interactive_choose_boolean_hook and _interactive_choose_int_hook
    __choose_boolean_hook = None
    __choose_boolean_hook_first_param = None
    __choose_int_hook = None
    __choose_int_hook_first_param = None

    def set_default_choose_boolean_hook(self):
        self.set_choose_boolean_hook(_interactive_choose_boolean_hook, None)
    def set_choose_boolean_hook(self, hook, first_param):
        self.__choose_boolean_hook = hook
        self.__choose_boolean_hook_first_param = first_param

    def set_default_choose_int_hook(self):
        self.set_choose_int_hook(_interactive_choose_int_hook, None)
    def set_choose_int_hook(self, hook, first_param):
        self.__choose_int_hook = hook
        self.__choose_int_hook_first_param = first_param



    #
    # Action language hook functions
    # Will call the real hooks __choose_boolean_hook and __choose_int_hook
    # Internal use only
    #
    def __ALHooks_choose_boolean(self, vm, l, t, index, info, oid):
        assert isinstance(vm, actlang.machine.VirtualMachine)
        assert type(info) is str
        assert type(index) is int
        assert self.working_global_state
        assert type(oid) is int
        assert 0 <= oid and oid < len(self.working_global_state.objects)

        obj = self.working_global_state.objects[oid]
        xmiid = None
        place = None
        if info.startswith("guard:"):
            xmiid = info[len("guard:"):]
            place = "guard"
        elif info.startswith("effect:"):
            xmiid = info[len("effect:"):]
            place = "effect"
        else:
            xmiid = None
            place = None

        uml_transition = None
        if xmiid:
            uml_transition = self.xmiid_to_uml_transition[xmiid]

        l = l.split(',')
        l = map(lambda a:a.strip(),l)
        assert len(l) == 2 and ("true" in l) and ("false" in l)
        return self.__choose_boolean_hook(self.__choose_boolean_hook_first_param,
                                          self, obj, uml_transition, place,
                                          index)

    def __ALHooks_choose_int(self, vm, l, t, index, info, oid):
        assert isinstance(vm, actlang.machine.VirtualMachine)
        assert type(info) is str
        assert type(index) is int
        assert self.working_global_state
        assert type(oid) is int
        assert 0 <= oid and oid < len(self.working_global_state.objects)

        obj = self.working_global_state.objects[oid]
        xmiid = None
        place = None
        if info.startswith("guard:"):
            xmiid = info[len("guard:"):]
            place = "guard"
        elif info.startswith("effect:"):
            xmiid = info[len("effect:"):]
            place = "effect"
        else:
            xmiid = None
            place = None

        uml_transition = None
        if xmiid:
            uml_transition = self.xmiid_to_uml_transition[xmiid]

        l = l.split(',')
        assert len(l) > 0
        int_list = []
        for s in l:
            assert Utils.String.IsInt(s)
            int_list.append(int(s))
        return self.__choose_int_hook(self.__choose_int_hook_first_param,
                                      self, obj, uml_transition, place, index,
                                      int_list)

    #
    # The hooks for passing signal reception and object creation information
    # from the Jumbala interpreter to the simulator
    #
    def reset_observed_actions(self):
        self.__ALHooks_observed_actions = []
    def get_observed_actions(self):
        return self.__ALHooks_observed_actions
    def __ALHooks_object_created(self, vm, obj_id, class_num):
        """Trap for object creation actions"""
        assert isinstance(vm, actlang.machine.VirtualMachine)
        self.__ALHooks_observed_actions.append(OA_CreateAction(obj_id,
                                                               class_num))
    def __ALHooks_signal_received(self, vm, obj_id, signal_num):
        """Trap for signal receive actions"""
        assert isinstance(vm, actlang.machine.VirtualMachine)
        self.__ALHooks_observed_actions.append(OA_SignalAction(obj_id,
                                                               signal_num))
    # These hooks should be called immediately after __ALHooks_signal_received
    # to pass the signal parameters to the simulator
    def __ALHooks_signal_add_bool_parameter(self, vm, param):
        """Adds a Boolean parameter to the last received signal."""
        assert isinstance(vm, actlang.machine.VirtualMachine)
        assert type(param) is bool
        action = self.__ALHooks_observed_actions[-1]
        assert isinstance(action, OA_SignalAction)
        action.add_parameter(SimValue_Boolean(param))
    def __ALHooks_signal_add_int_parameter(self, vm, param):
        """Adds an integer parameter to the last received signal."""
        assert isinstance(vm, actlang.machine.VirtualMachine)
        assert type(param) is int
        action = self.__ALHooks_observed_actions[-1]
        assert isinstance(action, OA_SignalAction)
        action.add_parameter(SimValue_Int(param))
    def __ALHooks_signal_add_ref_parameter(self, vm, param):
        """Adds an object reference parameter to the last received signal.
        Reference is saved as an integer and resolved later."""
        assert isinstance(vm, actlang.machine.VirtualMachine)
        action = self.__ALHooks_observed_actions[-1]
        assert isinstance(action, OA_SignalAction)
        action.add_parameter(SimValue_Int(param))


    def _execute_observed_actions(self, global_state, obj):
        """
        Execute the signal send and object creation actions observed
        during the last execution of Jumbala code.
        'obj' is the object in which the Jumbala code was executed.
        Returns a list of corresponding TraceActions.
        """
        assert isinstance(global_state, GlobalState.GlobalState)

        if obj is None:
            object_idnum = "env"
        else:
            assert isinstance(obj, Object)
            object_idnum = obj.idnum
        
        action_list = []
        #
        # Handle the actions observed during the action code execution
        #
        for oa in self.get_observed_actions():
            if isinstance(oa, OA_SignalAction):
                signal = self.uml_signals[oa.signal_num]
                # Resolve object reference parameters
                resolved_parameters = []
                for uml_parameter,sim_parameter in zip(signal.feature,
                                                       oa.parameters):
                    assert uml_parameter.oclIsKindOf(UML14.Attribute)
                    assert isinstance(sim_parameter, SimValue)
                    vdef = _VariableDef_from_Attribute(uml_parameter)
                    if isinstance(vdef.get_type(), SimType_ObjectRef):
                        assert isinstance(sim_parameter, SimValue_Int);
                        if sim_parameter.get_value() == -1:
                            resolved_parameters.append(SimValue_ObjectRef("null"))
                        else:
                            resolved_parameters.append(SimValue_ObjectRef(global_state.objects[sim_parameter.get_value()]))
                    else:
                        resolved_parameters.append(sim_parameter.get_copy())
                # Create message
                message = Message(signal, resolved_parameters)
                # Add the message to the receiver's input queue
                global_state.objects[oa.receiver_id].sm_input_queue.append(message)
                # Augment action list
                action_list.append(Trace.TA_SignalSend(object_idnum, oa.receiver_id, message))

            elif isinstance(oa, OA_CreateAction):
                assert oa.object_id == len(global_state.objects)
                # Create the new object and add it to global_state
                new_obj = Object(self.uml_class_to_class_def_map[self.uml_classes_in_model[oa.class_num]], None, self.trace.get_next_free_oid())
                assert new_obj.idnum == oa.object_id
                global_state.objects.append(new_obj)
                # Augment action list
                action_list.append(Trace.TA_CreateObject(object_idnum, oa.object_id, self.uml_classes_in_model[oa.class_num]))

            else:
                raise SimError("Internal error: unknown observed action type")

        return action_list



    def _get_static_Jumbala_code(self, add_sim_code = True):
        """
        Get the static part of the Jumbala code (class definitions).
        If 'add_sim_code' is True, adds the simulator specific classes, too.
        Returns the pair (prog,hooks), where
        prog contains the Jumbala code and
        hooks contains the native methods hooks.
        """
        prog = ""
        hooks = {}

        if add_sim_code:
            prog += "class $$SMUML {\n"
            prog += "  static native boolean choose_boolean(String l, String t, int index, String info, int oid);\n"
            prog += "  static native int choose_int(String l, String t, int index, String info, int oid);\n"
            if self._ASSERTION_CONDITION:
                prog += "  static boolean "+self._ASSERTION_CONDITION_ATTR+\
                        " = "
                if self.opt_evaluate_jumbala_assertions:
                    prog += "true"
                else:
                    prog += "false"
                prog += ";\n"
            prog += "}\n"
            hooks['$$SMUML.choose_boolean#String#String#int#String#int'] = self.__ALHooks_choose_boolean
            hooks['$$SMUML.choose_int#String#String#int#String#int'] = self.__ALHooks_choose_int

        if add_sim_code:
            # Code for the UML base Jumbala class
            prog += self.generator.gen_uml_base_class(self.uml_signals)

        if add_sim_code:
            # Add hooks for Uml base class observing actions
            hooks['$$$Uml_Base.$$$signal_received#int#int'] = self.__ALHooks_signal_received
            hooks['$$$Uml_Base.$$$signal_add_bool_parameter#boolean'] = self.__ALHooks_signal_add_bool_parameter
            hooks['$$$Uml_Base.$$$signal_add_int_parameter#int'] = self.__ALHooks_signal_add_int_parameter
            hooks['$$$Uml_Base.$$$signal_add_ref_parameter#int'] = self.__ALHooks_signal_add_ref_parameter
            hooks['$$$Uml_Base.$$$object_created#int#int'] = self.__ALHooks_object_created

        if add_sim_code:
            # Add code and hooks that enables us to read values from
            # the action language model inside the Jumbala virtual machine.
            prog += "class $$$reader {\n"
            prog += "  static native void read(boolean v);\n"
            prog += "  static native void read(int v);\n"
            prog += "  static void read($$$Uml_Base o) {read((o == null)?-1:o.__id);}\n"
            prog += "}\n"
            hooks['$$$reader.read#boolean'] = self.__ALReader_read_boolean_hook
            hooks['$$$reader.read#int'] = self.__ALReader_read_integer_hook

        if add_sim_code:
            # Declare an array for the objects.
            # Each object of a subclass of $$$Uml_Base will insert itself here
            # by calling 'add' in its constructor.
            # In this way the simulator will be able to access the objects
            # created during the execution of an action language statement.
            prog += "class $$$ObjectTracker {\n"
            prog += "  static $$$Uml_Base[] objects = null;\n"
            prog += "  static int length = 0;\n"
            prog += "  static void add(int index, $$$Uml_Base o) {\n"
            prog += "    assert index == length;\n"
            prog += "    $$$Uml_Base[] new_objects = new $$$Uml_Base[length+1];\n"
            prog += "    for(int i = 0; i < length; i++) {\n"
            prog += "      new_objects[i] = objects[i];\n"
            prog += "    }\n"
            prog += "    new_objects[length++] = o;\n"
            prog += "    objects = new_objects;\n"
            prog += "  }\n"
            prog += "}\n"

        # Generate Jumbala code for each UML class
        indent = Utils.String.Indent()
        prog += self.generator.gen_action_code(indent, self.uml_class_to_uml_class_index, add_sim_code)

        if add_sim_code:
            # An array for the objects inserted by the simulator.
            prog += "$$$Uml_Base[] $$sim_objects = null;\n"

        return (str(prog), hooks)

    def get_al_generator(self):
        """
        Get the action language generator object.
        """
        return self.generator



    def copy_simstate_to_action(self,
                                global_state,
                                run_constructors = False,
                                assume_fully_defined_state = True):
        """
        Generate code to copy the state of the simulator objects to the
        action language objects.
        """
        assert type(run_constructors) is bool
        assert type(assume_fully_defined_state) is bool
        prog = ""
        # Reset Uml_Base
        prog += "$$$Uml_Base.__nextid = 0;\n"
        prog += "$$$ObjectTracker.objects = null;\n"
        prog += "$$$ObjectTracker.length = 0;\n"
        if run_constructors:
            prog += "$$$Uml_Base.$$run_constructors = true;\n"
        else:
            prog += "$$$Uml_Base.$$run_constructors = false;\n"
        prog += "$$sim_objects = new $$$Uml_Base["+\
                str(len(global_state.objects))+"];\n"
        # Create action lang objects                            
        for (index,o) in enumerate(global_state.objects):
            assert index == o.idnum
            prog += "$$sim_objects["+str(index)+"] = new "+\
                    o.get_class_def().get_name()+"();\n"
            #prog += o.get_class_def().get_name()+" $$o"+str(o.idnum)+" = new "+o.get_class_def().get_name()+"();\n"
            #prog += "$$$Uml_Base $$o"+str(o.idnum)+" = new "+o.get_class_def().get_name()+"();\n"

        # Assign values to objects' variables
        for (index,o) in enumerate(global_state.objects):
            assert index == o.idnum
            for name,value in o.variable_values.iteritems():
                if value.is_defined():
                    prog += "((" + o.get_class_def().get_name() + ")$$sim_objects["+str(index)+"])."+str(name)+" = "+value.get_act_str()+";\n"
                    #prog += "$$o"+str(o.idnum)+"."+str(name)+" = "+value.get_act_str()+";\n"
                    #prog += "((" + o.get_class_def().get_name() + ")" + al_objs + "[" + str(o.idnum)+"])." + str(name) + " = " + value.get_act_str() +";\n"
                else:
                    assert assume_fully_defined_state == False

        prog += "$$$Uml_Base.$$run_constructors = true;\n"

        # Debugging
        if False:
            debug_filename = "DEBUG2.jmb"
            f = open(debug_filename, 'w')
            f.write(prog)
            f.close()
        
        return prog
    

    
    #
    # Native reader methods for the action language
    #
    def __ALReader_reset(self):
        self.__ALReader_value = None
        self.__ALReader_values = []
    def __ALReader_read_boolean_hook(self, vm, v):
        """Action language hook for reading Boolean values."""
        assert isinstance(vm, actlang.machine.VirtualMachine)
        assert type(v) is bool
        self.__ALReader_value = v
        self.__ALReader_values.append(v)
    def __ALReader_read_integer_hook(self, vm, v):
        """Action language hook for reading integer values."""
        assert isinstance(vm, actlang.machine.VirtualMachine)
        assert type(v) is int
        self.__ALReader_value = v
        self.__ALReader_values.append(v)

    def _evaluate_boolean_expr_in_vm(self, expr_str, catch_and_exit = False):
        """
        The second argument tells what to do in the case of an error.
        """
        assert type(expr_str) is str
        #print "Evaluating boolean expression in vm: "+expr_str
        self.__ALReader_reset()
        prog = "$$$reader.read("+expr_str+");\n"
        self._execute_statement_in_vm(prog, catch_and_exit)
        assert isinstance(self.__ALReader_value, bool)
        return self.__ALReader_value


    def _evaluate_integer_expr_in_vm(self, expr_str, catch_and_exit = False):
        """
        The second argument tells what to do in the case of an error.
        """
        assert type(expr_str) is str
        #print "Evaluating integer expression in vm: "+expr_str
        self.__ALReader_reset()
        prog = "$$$reader.read("+expr_str+");\n"
        self._execute_statement_in_vm(prog, catch_and_exit)
        assert isinstance(self.__ALReader_value, int)
        return self.__ALReader_value


    def _execute_statement_in_vm(self,
                                 stmt_str,
                                 catch_exceptions_and_exit = False):
        """
        Execute the statement 'stmt_str' in the action language environment
        and vm, possibly modifying them.
        
        If 'catch_exceptions_and_exit' is True,
        possible parsing and runtime errors are caught and
        the execution of the simulator is terminated with as assertion error.
        """
        assert type(stmt_str) is str
        assert type(catch_exceptions_and_exit) is bool
        #print "Executing statement in vm: "+stmt_str
        if catch_exceptions_and_exit:
            try:
                self.action_vm.initialize(actlang.grammar.parse(stmt_str).translate(self.action_env))
            except actlang.errors.StaticError, e:
                print e
                assert False
            try:
                self.action_vm.run()
            except actlang.errors.RuntimeError_, e:
                print e
                assert False
        else:
            self.action_vm.initialize(actlang.grammar.parse(stmt_str).translate(self.action_env))
            self.action_vm.run()



    def copy_state_from_action_to_sim(self, global_state):
        """
        Copy the values of local variables of each object in the global state
        from action language virtual machine to the simulator.
        """
        # Read objects' attribute values
        stmts = ""
        for o in global_state.objects:
            class_def = o.get_class_def()
            for (name,vdef) in class_def.get_variable_defs().iteritems():
                stmts += "$$$reader.read((("+class_def.get_name()+")"+\
                         al_objs+"["+str(o.idnum)+"])."+\
                         str(vdef.get_name())+");\n"
        self.__ALReader_reset()
        self._execute_statement_in_vm(stmts)

        # Assign values to objects' attributes
        index = 0
        for o in global_state.objects:
            class_def = o.get_class_def()
            for (name,vdef) in class_def.get_variable_defs().iteritems():
                v = self.__ALReader_values[index]
                if isinstance(vdef.get_type(), SimType_Boolean):
                    assert type(v) is bool
                    o.variable_values[vdef.get_name()].set_value(v)

                elif isinstance(vdef.get_type(), SimType_Int):
                    assert type(v) is int
                    o.variable_values[vdef.get_name()].set_value(v)

                elif isinstance(vdef.get_type(), SimType_ObjectRef):
                    assert type(v) is int
                    assert -1 <= v and v < len(global_state.objects)
                    if v == -1:
                        o.variable_values[vdef.get_name()].set_value("null")
                    else:
                        o.variable_values[vdef.get_name()].set_value(global_state.objects[v])
                        
                else:
                    raise "Internal error: unknown variable SimType"
                index += 1


    def __copy_state_from_action_to_sim_old(self, global_state):
        """
        Copy the values of local variables of each object in the global state
        from action language virtual machine to the simulator.
        """
        # Assign values to objects' variables
        for o in global_state.objects:
            class_def = o.get_class_def()
            stmts = ""
            for (name,vdef) in class_def.get_variable_defs().iteritems():
                stmts += "$$$reader.read((("+class_def.get_name()+")"+\
                         al_objs+"["+str(o.idnum)+"])."+\
                         str(vdef.get_name())+");\n"
            self.__ALReader_reset()
            self._execute_statement_in_vm(stmts)
            assert len(self.__ALReader_values) == \
                   len(class_def.get_variable_defs())
            index = 0
            for (name,vdef) in class_def.get_variable_defs().iteritems():
                v = self.__ALReader_values[index]
                if isinstance(vdef.get_type(), SimType_Boolean):
                    assert type(v) is bool
                    o.variable_values[vdef.get_name()].set_value(v)

                elif isinstance(vdef.get_type(), SimType_Int):
                    assert type(v) is int
                    o.variable_values[vdef.get_name()].set_value(v)

                elif isinstance(vdef.get_type(), SimType_ObjectRef):
                    assert type(v) is int
                    assert -1 <= v and v < len(global_state.objects)
                    if v == -1:
                        o.variable_values[vdef.get_name()].set_value("null")
                    else:
                        o.variable_values[vdef.get_name()].set_value(global_state.objects[v])
                        
                else:
                    raise "Internal error: unknown variable SimType"
                index += 1


    def __copy_state_from_action_to_sim_older(self, global_state):
        """
        Copy the values of local variables of each object in the global state
        from action language virtual machine to the simulator.
        """
        # Assign values to objects' variables
        for o in global_state.objects:
            class_def = o.get_class_def()
            for (name,vdef) in class_def.get_variable_defs().iteritems():
                if isinstance(vdef.get_type(), SimType_Boolean):
                    expr = "(("+class_def.get_name()+")"+\
                           al_objs+"["+str(o.idnum)+"])."+str(vdef.get_name())
                    v = self._evaluate_boolean_expr_in_vm(expr, True)
                    o.variable_values[vdef.get_name()].set_value(v)

                elif isinstance(vdef.get_type(), SimType_Int):
                    expr = "(("+class_def.get_name()+")"+\
                           al_objs+"["+str(o.idnum)+"])."+str(vdef.get_name())
                    v = self._evaluate_integer_expr_in_vm(expr, True)
                    o.variable_values[vdef.get_name()].set_value(v)

                elif isinstance(vdef.get_type(), SimType_ObjectRef):
                    vref = "(("+class_def.get_name()+")"+\
                           al_objs+"["+str(o.idnum)+"])."+str(vdef.get_name())
                    expr = "("+vref+"==null)?-1:"+vref+".__id"
                    v = self._evaluate_integer_expr_in_vm(expr, True)
                    if v == -1:
                        o.variable_values[vdef.get_name()].set_value("null")
                    else:
                        o.variable_values[vdef.get_name()].set_value(global_state.objects[v])
                        
                else:
                    raise "Internal error: unknown variable SimType"
                

    def api_print_trace(self, f = sys.stdout, indent = Utils.String.Indent()):
        """
        Print the trace in the file 'f'.
        """
        self.trace.output(f, indent)


    def trace_get_last_global_state(self):
        """
        Get the last GlobalState in the trace.
        Thou shall NOT modify the GlobalState directly.
        """
        assert self.trace.length() >= 1
        return self.trace.get_last_global_state()


    def step(self):
        """Randomly select an enabled transition and execute it"""
        assert self.trace.length() >= 1
        #assert len(self.trace) >= 1
        current_global_state = self.trace.get_last_global_state()
        #current_global_state = self.trace[-1]
        next_global_state = current_global_state.api_copy()
        #next_global_state.api_print("$ ")

        #compute transitions with a source in the current state
        enabled=[]
        all_enabled_sets = []
        disabled_objects = []
        for o in next_global_state.objects:
            # Find the set of transitions that are enabled
            event = None
            if o.sm_input_queue != []:
                event = o.sm_input_queue[0]
            disabled = True
            enabled_sets = o.sm_state.find_enabled(o, event, self, current_global_state)
            if enabled_sets != []:
                disabled = False
            for set in enabled_sets:
                all_enabled_sets.append((o, set))
            if disabled and o.sm_input_queue != []: disabled_objects.append(o)

        if all_enabled_sets == [] and disabled_objects == []:
            print "No enabled objects"
            return False
        #randomly select which transition set to fire
        assert len(all_enabled_sets)+len(disabled_objects) > 0
        pivot=random.randint(0,len(all_enabled_sets)+len(disabled_objects)-1)
        theTransitions = None
        if(pivot<len(all_enabled_sets)):
            theTransitions = all_enabled_sets[pivot]
        else:
            disabled_objects[pivot-len(all_enabled_sets)].consume()
            print 10*"<"
            print "Implicitly consumed a signal from the object ", disabled_objects[pivot-len(all_enabled_sets)].name 
            print 10*">"
            return True
 
        o,set = theTransitions
        print 10*"<"
        print "Enabled transition set:"
        for transition in set:
            print "  from:",; print sm_get_qualified_name(transition.source)
            print "  to:  ",; print sm_get_qualified_name(transition.target)
        print 10*">"
        o.sm_state,action_list = o.sm_state.fire(o, set, self, next_global_state)
        # An ad-hoc thingie to remove consumed signals
#        if set[0].trigger:
#            assert set[0].trigger.signal == o.sm_input_queue[0].signal
#            o.sm_input_queue.pop(0)

        self.trace.add_global_state(next_global_state)
        #self.trace.append(next_global_state)
        return True


    #
    # Copy the global state into the action language
    #
    def set_working_global_state(self, global_state):
        """
        Set the GlobalState 'global_state' as the working global state.
        """
        assert isinstance(global_state, GlobalState.GlobalState)
        # copy simulator global state to action language objects
        self.working_global_state = global_state
        prog = self.copy_simstate_to_action(global_state) 
        #print prog
        try:
            self.action_vm.initialize(actlang.grammar.parse(prog).translate(self.action_env))
        except actlang.errors.StaticError, e:
            print e
            assert False
        try:
            self.action_vm.run()
        except actlang.errors.RuntimeError_, e:
            print e
            assert False

    def get_working_global_state(self):
        """
        Get the working global state.
        Yes, the real one, not a copy, i.e. modifying it will have
        the expected consequences.
        """
        assert self.working_global_state
        return self.working_global_state

    def get_enabled_in_working_global_state(self):
        """
        Get the enabled transitions and implicit consumption transitions
        in the working global state.
        """
        assert self.working_global_state
        all_enabled_sets = []
        implicit_consumption_objects = []
        for o in self.working_global_state.objects:
            # Find the set of transitions that are enabled
            event = None
            if o.sm_input_queue != []:
                event = o.sm_input_queue[0]
            enabled_sets = o.sm_state.find_enabled(o, event, self, self.working_global_state)
            for set in enabled_sets:
                all_enabled_sets.append((o, set))
            # Only disabled objects that have signals in the queue are appended
            if enabled_sets == [] and o.sm_input_queue != []:
                implicit_consumption_objects.append(o)
        return all_enabled_sets, implicit_consumption_objects


    def execute_transition_set_in_working_global_state(self, transition_set_num):
        assert self.working_global_state

        all_enabled_sets, implicit_consumption_objects = self.get_enabled_in_working_global_state()
    
        assert transition_set_num >= 0
        assert transition_set_num < len(all_enabled_sets)

        theTransitions = all_enabled_sets[transition_set_num]
        o,set = theTransitions
        o.sm_state,action_list = o.sm_state.fire(o, set, self, self.working_global_state)
        self.copy_state_from_action_to_sim(self.working_global_state)

        return action_list


    def execute_implicit_consumption_in_working_global_state(self, implicit_consumption_num):
        assert self.working_global_state

        all_enabled_sets, implicit_consumption_objects = self.get_enabled_in_working_global_state()
    
        assert implicit_consumption_num >= 0
        assert implicit_consumption_num < len(implicit_consumption_objects)

        o = implicit_consumption_objects[implicit_consumption_num]
        message = o.sm_input_queue[0]
        o.consume()

        return [Trace.TA_ImplicitConsumption(o.idnum, message)]

    
    #ct = float(0)
    def execute_code_in_object(self, object_num, code, return_type):
        """
        Execute Jumbala statement 'code' in the 'object_num'th object in
        the working global state (possibly modifying it).
        If 'return_type' is not None, it should be a SimType,
        'code' should contain a return statement of the corresponding type,
        and a SimValue of SimType will be returned.
        If an error occurs while parsing or executing 'code',
        an exception will be raised and the working global state will
        remain unmodified.
        The 'code' shall not have any variable names, labels etc starting with
        the dollar sign except for the method and attribute access
        of form '$$SMUML.xxx'.
        Currently this method can be bit inefficient.
        """
        #print >> sys.stderr, code
        #start = os.times()[0]
        if True:
            v = self._execute_code_in_object_new(object_num,
                                                 code,
                                                 return_type)
        else:
            v = self._execute_code_in_object_old(object_num,
                                                 code,
                                                 return_type)
        #end = os.times()[0]
        #self.ct += end - start
        #print >> sys.stderr, end-start, self.ct
        return v

    def _execute_code_in_object_new(self, object_num, code, return_type):
        """
        Execute Jumbala statement 'code' in the 'object_num'th object in
        the working global state (possibly modifying it).
        If 'return_type' is not None, it should be a SimType,
        'code' should contain a return statement of the corresponding type,
        and a SimValue of SimType will be returned.
        If an error occurs while parsing or executing 'code',
        an exception will be raised and the working global state will
        remain unmodified.
        Not as inefficient as the old version.
        """
        assert self.working_global_state
        assert (type(object_num) is int) or (type(object_num) is long)
        assert object_num >= 0
        assert object_num < self.working_global_state.get_nof_objects()
        assert type(code) is str
        assert (return_type is None) or isinstance(return_type, SimType)
        return_value = None

        # Get the object and the UML Class that it is an instance of
        obj = self.working_global_state.get_object(object_num)
        assert isinstance(obj, Object)
        class_def = obj.get_class_def()
        assert isinstance(class_def, ClassDef)
        uml_class = class_def.get_uml_class()
        assert uml_class.oclIsKindOf(UML14.Class)

        # Copy global state from simulator to action language vm
        self.set_working_global_state(self.working_global_state)

        # Self-explanatory
        self.reset_observed_actions()

        #print code
        code = Sim.ActLangUtils.choose_augmenter_stmt(code)
        context = "(("+str(uml_class.name)+")$$sim_objects["+str(object_num)+"])"
        code = Sim.ActLangUtils.stmt_contextifier(code, context,
                                                  ["$$SMUML"], [], [])
        #print code

        # Execute the temporary method and handle its return value
        try:
            self.__ALReader_reset()
            self._execute_statement_in_vm(code)
            if return_type == None:
                assert self.__ALReader_value is None
                return_value = None
            elif isinstance(return_type, SimType_Boolean):
                assert isinstance(self.__ALReader_value, bool)
                return_value = SimValue_Boolean(self.__ALReader_value)
            elif isinstance(return_type, SimType_Int):
                assert isinstance(self.__ALReader_value, int)
                return_value = SimValue_Int(self.__ALReader_value)
            elif isinstance(return_type, SimType_ObjectRef):
                assert isinstance(self.__ALReader_value, int)
                returned_objid = self.__ALReader_value
                # The objid is resolved to a real object after the possible
                # object creations are lifted from the action language vm to
                # the simulator
            else:
                raise RuntimeError("Internal error: "+
                                   "Something not implemented")
        except actlang.errors.Error, e:
            self.reset_observed_actions()
            # Pass the error to the caller
            raise e

        # Execution may have resulted in creating new objects or
        # sending signals, insert them to the working global state
        self._execute_observed_actions(self.working_global_state, obj)

        # Resolve the object of the returned objid
        if isinstance(return_type, SimType_ObjectRef):
            assert -1 <= returned_objid
            assert returned_objid < self.working_global_state.get_nof_objects()
            if returned_objid == -1:
                return_value = SimValue_ObjectRef("null")
            else:
                obj = self.working_global_state.get_object(returned_objid)
                return_value = SimValue_ObjectRef(obj)

        # Self-explanatory
        self.reset_observed_actions()

        # Copy the changes in attribute values of objects to
        # the working global state
        self.copy_state_from_action_to_sim(self.working_global_state)

        # End of story
        return return_value



    def _execute_code_in_object_old(self, object_num, code, return_type):
        """
        Execute Jumbala statement 'code' in the 'object_num'th object in
        the working global state (possibly modifying it).
        If 'return_type' is not None, it should be a SimType,
        'code' should contain a return statement of the corresponding type,
        and a SimValue of SimType will be returned.
        If an error occurs while parsing or executing 'code',
        an exception will be raised and the working global state will
        remain unmodified.
        Currently this method is quite inefficient.
        """
        assert self.working_global_state
        assert isinstance(object_num, int) or isinstance(object_num, long)
        assert object_num >= 0
        assert object_num < self.working_global_state.get_nof_objects()
        assert type(code) is str
        assert (return_type is None) or isinstance(return_type, SimType)
        return_value = None

        # Get the object and the UML Class that it is an instance of
        obj = self.working_global_state.get_object(object_num)
        assert isinstance(obj, Object)
        class_def = obj.get_class_def()
        assert isinstance(class_def, ClassDef)
        uml_class = class_def.get_uml_class()
        assert uml_class.oclIsKindOf(UML14.Class)

        # Save the original action language environment and vm
        saved_env = self.action_env
        saved_vm = self.action_vm

        # Create a temporary action language environment and vm
        self.action_env = actlang.basetypes.getInitialEnvironment()
        self.action_vm = actlang.basetypes.getInitialMachine()

        # Add a temporary method in which 'code' is executed to
        # the action language generator class corresponding to the UML Class
        method_name = "$$exec"
        algen_method = ActionGeneratorMethod(method_name,
                                             return_type,
                                             Sim.ActLangUtils.choose_augmenter_stmt(code))
        self.generator.get_algen_class(uml_class).add_method(method_name,
                                                             algen_method)

        # Get the static part of the Jumbala code for the UML Model
        (prog,hooks) = self._get_static_Jumbala_code()

        # Set hooks for the native thingies
        self.action_env.setNativeHooks(hooks)

        try:
            self.action_vm.initialize(actlang.grammar.parse(prog).translate(self.action_env))
        except actlang.errors.StaticError, e:
            # Remove the temporary method from the action language generator
            self.generator.get_algen_class(uml_class).remove_method(method_name)
            # Restore the original action language environment and vm etc
            self.action_env = saved_env
            self.action_vm = saved_vm
            self.reset_observed_actions()
            # Pass the error to the caller
            raise e
        try:
            self.action_vm.run()
        except actlang.errors.RuntimeError_, e:
            # Remove the temporary method from the action language generator
            self.generator.get_algen_class(uml_class).remove_method(method_name)
            # Restore the original action language environment and vm etc
            self.action_env = saved_env
            self.action_vm = saved_vm
            self.reset_observed_actions()
            # Pass the error to the caller
            raise e

        # Copy global state from simulator to action language vm
        self.set_working_global_state(self.working_global_state)

        # Self-explanatory
        self.reset_observed_actions()

        # Execute the temporary method and handle its return value
        try:
            al_method_call = "(("+str(uml_class.name)+")$$sim_objects["+\
                             str(object_num)+"])."+method_name+"()"
            if return_type == None:
                self._execute_statement_in_vm(al_method_call+";")
                return_value = None
            elif isinstance(return_type, SimType_Boolean):
                v = self._evaluate_boolean_expr_in_vm(al_method_call)
                return_value = SimValue_Boolean(v)
            elif isinstance(return_type, SimType_Int):
                v = self._evaluate_integer_expr_in_vm(al_method_call)
                return_value = SimValue_Int(v)
            elif isinstance(return_type, SimType_ObjectRef):
                al_stmt = "$$$Uml_Base $$temp = "+al_method_call+";\n"
                self._execute_statement_in_vm(al_stmt)
                al_expr = "($$temp==null)?-1:$$temp.__id"
                returned_objid = self._evaluate_integer_expr_in_vm(al_expr)
                # The objid is resolved to a real object after the possible
                # object creations are lifted from the action language vm to
                # the simulator
            else:
                assert False and "Internal error: Something not implemented"
        except actlang.errors.Error, e:
            # Remove the temporary method from the action language generator
            self.generator.get_algen_class(uml_class).remove_method(method_name)
            # Restore the original action language environment and vm etc
            self.action_env = saved_env
            self.action_vm = saved_vm
            self.reset_observed_actions()
            # Pass the error to the caller
            raise e

        # Execution may have resulted in creating new objects or
        # sending signals, insert them to the working global state
        self._execute_observed_actions(self.working_global_state, obj)

        # Resolve the object of the returned objid
        if isinstance(return_type, SimType_ObjectRef):
            assert -1 <= returned_objid
            assert returned_objid < self.working_global_state.get_nof_objects()
            if returned_objid == -1:
                return_value = SimValue_ObjectRef("null")
            else:
                obj = self.working_global_state.get_object(returned_objid)
                return_value = SimValue_ObjectRef(obj)

        # Self-explanatory
        self.reset_observed_actions()

        # Copy the changes in attribute values of objects to
        # the working global state
        self.copy_state_from_action_to_sim(self.working_global_state)

        # Remove the temporary method from the action language generator
        self.generator.get_algen_class(uml_class).remove_method(method_name)

        # Restore the original action language environment and vm
        self.action_env = saved_env
        self.action_vm = saved_vm

        # Copy global state from simulator to the original action language vm
        self.set_working_global_state(self.working_global_state)

        # End of story
        return return_value



    def interactive_step(self):
        """
        Ask the user to select an enabled transition in the last global state
        in the trace and execute it.
        """
        assert self.trace
        assert self.trace.length() >= 1
        #assert len(self.trace) >= 1
        current_global_state = self.trace.get_last_global_state()
        next_global_state = current_global_state.api_copy()

        # copy simulator state to action language objects
        prog = self.set_working_global_state(next_global_state) 

        #compute transitions with a source in the current state
        enabled=[]
        all_enabled_sets = []
        implicit_consumption_objects = []
        for o in next_global_state.objects:
            if not o.sm_state:
                # A passive object
                continue

            # Find the set of transitions that are enabled
            event = None
            if o.sm_input_queue != []:
                event = o.sm_input_queue[0]
            enabled_sets = o.sm_state.find_enabled(o, event, self,
                                                   next_global_state)

            if enabled_sets == []:
                # Object disabled
                if o.sm_input_queue != []:
                    # has signals in the queue => implcit consumption possible
                    implicit_consumption_objects.append(o)
            else:
                # There is a transition enabled in the object
                for set in enabled_sets:
                    all_enabled_sets.append((o, set))


        # Print enabled transition sets
        transition_set_num = 1
        for enabled_set in all_enabled_sets:
            object,transitions = enabled_set
            print "Action " +  str(transition_set_num) + ": execute transition set on object " + str(object.idnum) + " (" + object.name + ")"
            #print "Enabled transition set " + str(transition_set_num) + ":",
            #print "object " + object.name
            transition_set_num = transition_set_num + 1
            for (why,transition) in transitions:
                if why == "defer":
                    print "  defer event"
                    continue
                print "  from:   ",; print sm_get_qualified_name(transition.source)
                print "  to:     ",; print sm_get_qualified_name(transition.target)
                if transition.trigger:
                    assert transition.trigger.oclIsKindOf(UML14.SignalEvent)
                    print "  trigger:",
                    print transition.trigger.signal.name
                if transition.guard:
                    assert transition.guard.oclIsKindOf(UML14.Guard)
                    assert transition.guard.expression
                    assert transition.guard.expression.body
                    print "  guard: " + Utils.String.length_limited_string(str(transition.guard.expression.body), 60)
                if transition.effect:
                    print "  effect: " + Utils.String.length_limited_string(str(transition.effect.script.body), 60)
 
        # Print objects that can do implicit consumption
        for o in implicit_consumption_objects:
            print "Action " + str(transition_set_num) + ": implicitly consume signal " + o.sm_input_queue[0].signal.name + " on object " + str(o.idnum) + " (" + o.name + ")"
            transition_set_num = transition_set_num + 1

        #
        # Get user's choice
        #
        choice_num = 0
        command = "execute"
        while True:
            print "Please select an action, 0 to send a signal to an object, or 'b' to backtrack one step, 'q' to quit: ",
            answer = sys.stdin.readline()
            if Utils.String.IsInt(answer):
                choice_num = int(answer)
                if choice_num >= 0 and choice_num <= len(all_enabled_sets)+len(implicit_consumption_objects):
                    break
            elif answer == "b\n":
                if self.trace.length() <= 1:
                    #if len(self.trace) <= 1:
                    print "Cannot backtrack further"
                else:
                    command = "backtrack"
                    break
            elif answer == "q\n":
                command = "quit"
                break

        if command == "quit":
            return False
        
        if command == "backtrack":
            self.trace.backtrack_one_step()
            #self.trace.pop()
            return True

        assert command == "execute"
            
        if choice_num > 0 and choice_num <= len(all_enabled_sets):
            theTransitions = all_enabled_sets[choice_num-1]
            o,set = theTransitions
            o.sm_state,action_list = o.sm_state.fire(o, set, self, next_global_state)
            self.copy_state_from_action_to_sim(next_global_state)
            for action in action_list:
                self.trace.add_action(action)

        elif choice_num > len(all_enabled_sets):
            o = implicit_consumption_objects[choice_num-len(all_enabled_sets)-1]
            message = o.sm_input_queue[0]
            o.consume()
            self.trace.add_action(Trace.TA_ImplicitConsumption(o.idnum, message))

        elif choice_num == 0:
            #
            # The user acts as an environment and sends a signal to an object
            #
            print "The available signals are:"
            signal_num = 1
            for signal in self.uml_signals:
                print " " + str(signal_num) + ". " + signal.name
                signal_num += 1

            signal_sum = 0
            while True:
                print "Select a signal to send from environment: ",
                answer = sys.stdin.readline()
                if Utils.String.IsInt(answer):
                    signal_num = int(answer)
                    if(signal_num >= 1 and
                       signal_num <= len(self.uml_signals)):
                        break

            print "Active objects are:"
            object_num = 1
            for object in next_global_state.objects:
                print " " + str(object_num) + ". " + object.name
                object_num += 1
            while True:
                print "Select an object: ",
                answer = sys.stdin.readline()
                if Utils.String.IsInt(answer):
                    object_num = int(answer)
                    if(object_num >= 1 and
                       object_num <= len(next_global_state.objects)):
                        break

            message = Message(self.uml_signals[signal_num-1],[])
            self.trace.add_action(Trace.TA_SignalSend("env", next_global_state.objects[object_num-1].idnum, message))
            
            next_global_state.objects[object_num-1].sm_input_queue.append(message)

        else:
            raise "Internal error"

        self.trace.add_global_state(next_global_state)
        return True
