"""
A state machine hierachy flattener

Author: Tommi Junttila

Copyright (c) 2005-2007 Helsinki University of Technology
Released under the GNU General Public License version 2
"""

import sys
import sets

import coral

import Sim
import Sim.Core
from Sim.Error import SimError
from Sim.StateConfiguration import StateConfiguration
from Sim.StateConfiguration import StateConfiguration_from_uml_state
import Utils.DependencyAnalyzer
import Utils.Models
import Utils.String
from Utils.Models import get_qualified_name

UML14 = coral.lang.UML14

def _mcname(element):
    """Get the name of the metaclass of the Coral Element 'element'."""
    return Utils.Models.get_metaclassname(element)





def flatten(sm,
            opt_use_partial_order_reduction = True,
            stdout = sys.stdout, stderr = sys.stderr, verbose_level = 1):
    """
    Returns a flattened version of the UML1.4 state machine 'sm'.
    Can raise RuntimeErrors.
    """
    assert sm.oclIsKindOf(UML14.StateMachine)

    verbwrt = Utils.String.Writer(stdout)
    def _verb(s):
        verbwrt.ln(s)

    assert sm.top
    if not sm.top.oclIsKindOf(UML14.CompositeState):
        raise RuntimeError("Not supported: state machines with "
                           "non-composite top states")
    if sm.top.oclIsKindOf(UML14.SubmachineState):
        raise RuntimeError("Not supported: state machines with "
                           "submachineState top states")
    if sm.top.entry:
        raise RuntimeError("Not supported: state machines with "
                           "top states having entry actions not supported")
    if sm.top.exit:
        raise RuntimeError("Not supported: state machines with "
                           "top states having exit actions")

    if verbose_level >= 1:
        _verb("Flattening the state machine '"+get_qualified_name(sm)+"'")

    if verbose_level >= 1:
        stats = Utils.StateMachines.Statistics(sm)
        _verb("The state machine has")
        _verb(" %5d state vertices" % stats.get_nof_state_vertices())
        _verb(" %5d concurrent composite states" %
              stats.get_nof_concurrent_composite_states())
        _verb(" %5d nonconcurrent composite states" %
              stats.get_nof_nonconcurrent_composite_states())
        _verb(" %5d simple states" % stats.get_nof_simple_states())
        _verb(" %5d final states" % stats.get_nof_final_states())
        _verb(" %5d initial pseudostates" %
              stats.get_nof_initial_pseudostates())
        _verb(" %5d choice pseudostates" % stats.get_nof_choice_pseudostates())
        _verb(" %5d transitions" % stats.get_nof_transitions())

    context_class = sm.context
    assert context_class
    assert context_class.oclIsKindOf(UML14.Class)
    context_model = context_class.namespace
    while not context_model.oclIsKindOf(UML14.Model):
        context_model = context_model.namespace

    # The set of states in the flattened state machine
    sconfs = sets.Set()

    # Make the initial state configuration
    first_sconf = StateConfiguration_from_uml_state(sm.top)

    # Map each StateConfiguration in 'sconfs' to a list of
    # (source_st, target_st, trigger, guard, effect)
    outgoing_transitions = {}
    incoming_transitions = {}
    # Map each StateConfiguration into a list of UML SignalEvents that are deferrable
    deferrable_signal_events = {}

    sconfs.add(first_sconf)
    incoming_transitions[first_sconf] = []

    # Unprocessed sconfs
    sconf_queue = [first_sconf]

    while sconf_queue != []:
        sconf = sconf_queue.pop()

        # Initialize data structure for outgoing transitions
        assert sconf not in outgoing_transitions
        outgoing_transitions[sconf] = []
        assert sconf in incoming_transitions

        if verbose_level >= 2:
            _verb("Generating successors for")
            verbwrt.push()
            sconf.output(verbwrt)
            _verb(sconf.isStable())
            _verb(sconf.isInCompoundTransition())

        if not sconf.is_stable():
            actions = sconf.find_enabled(None, None, None)
            assert actions
            for (r,g,t) in actions:
                if r == 'completion':
                    (new_sconf,exit_actions,entry_actions) = sconf.fire(t)

                    if new_sconf not in sconfs:
                        assert new_sconf not in incoming_transitions
                        incoming_transitions[new_sconf] = []
                        sconf_queue.append(new_sconf)
                        sconfs.add(new_sconf)

                    tr = (sconf, new_sconf, None, g,
                          _make_effect(t.effect,exit_actions,entry_actions))
                    outgoing_transitions[sconf].append(tr)
                    incoming_transitions[new_sconf].append(tr)

                elif r == 'quiesce':
                    assert not sconf.is_in_compound_transition()
                    assert t.oclIsKindOf(UML14.State)
                    new_sconf = sconf.quiesce(t)
                    
                    if new_sconf not in sconfs:
                        assert new_sconf not in incoming_transitions
                        incoming_transitions[new_sconf] = []
                        sconf_queue.append(new_sconf)
                        sconfs.add(new_sconf)

                    tr = (sconf, new_sconf, None, g, None)
                    outgoing_transitions[sconf].append(tr)
                    incoming_transitions[new_sconf].append(tr)

                else:
                    raise RuntimeError("Internal error")
                
                if verbose_level >= 2:
                    verbwrt.ln("Successor:")
                    verbwrt.push()
                    new_sconf.output(verbwrt)
                    verbwrt.pop()
            

        else:
            #
            # Handle transitions with triggers
            #

            #
            # Find signals that are
            # (i) used as triggers in transitions leaving
            #     this state configuration, or
            # (ii) deferrable in the states in this state configuration.
            #
            signals = sconf.find_signals()

            #
            # Build transitions and successors in the flattened state machine
            #
            for signal in signals:
                # Find potentially enabled 'signal'-triggered transitions
                # and whether 'signal' should be deferred if none of them
                # is actually enabled.
                actions = sconf.find_enabled(signal, None, None)

                #
                # Create successor state configurations for each
                # 'signal'-triggered transition that is potentially enabled
                #
                deferring_added = False
                for (r,g,t) in actions:
                    if r == 'defer':
                        assert t.oclIsKindOf(UML14.SignalEvent)
                        if not deferring_added:
                            deferring_added = True
                            if sconf not in deferrable_signal_events:
                                deferrable_signal_events[sconf] = []
                            deferrable_signal_events[sconf].append(t)

                    elif r == 'signal':
                        (new_sconf,exit_actions,entry_actions) = sconf.fire(t)

                        if new_sconf not in sconfs:
                            assert new_sconf not in incoming_transitions
                            incoming_transitions[new_sconf] = []
                            sconf_queue.append(new_sconf)
                            sconfs.add(new_sconf)
                        tr = (sconf, new_sconf, t.trigger, g,
                              _make_effect(t.effect,
                                           exit_actions,
                                           entry_actions))
                        outgoing_transitions[sconf].append(tr)
                        incoming_transitions[new_sconf].append(tr)

                    else:
                        raise RuntimeError("Internal error")

                    if verbose_level >= 2:
                        verbwrt.ln("Successor:")
                        verbwrt.push()
                        new_sconf.output(verbwrt)
                        verbwrt.pop()

        if verbose_level >= 2:
            verbwrt.pop()

    #
    # Create the new, flattened state machine
    #
    flat_sm = UML14.StateMachine()
    flat_sm.name = sm.name
    flat_sm.top = UML14.CompositeState()
    flat_sm.top.stateMachine = flat_sm
    flat_sm.top.isConcurrent = False
    
    #
    # Add states corresponding to the reachable state configurations
    # into the flattened state machine
    #
    sconf_to_flat_uml_state = {}
    for sconf in sconfs:
        uml_state = _make_flat_state(sconf)
        flat_sm.top.subvertex.append(uml_state)
        sconf_to_flat_uml_state[sconf] = uml_state

    #
    # Add deferable events
    #
    for (sconf,uml_state) in sconf_to_flat_uml_state.iteritems():
        if sconf not in deferrable_signal_events:
            continue
        for defer_event in deferrable_signal_events[sconf]:
            uml_state.deferrableEvent.append(defer_event)

    #
    # Add the transitions into the flattened state machine
    #
    for (sconf,uml_state) in sconf_to_flat_uml_state.iteritems():
        for (source,target,trigger,guard,effect) in outgoing_transitions[sconf]:
            assert source is sconf
            uml_t = UML14.Transition()
            uml_t.source = uml_state
            uml_t.target = sconf_to_flat_uml_state[target]
            uml_t.trigger = trigger
            if guard and (guard != 'true'):
                uml_t.guard = UML14.Guard()
                uml_t.guard.transition = uml_t
                uml_t.guard.expression = UML14.BooleanExpression()
                uml_t.guard.expression.body = guard
            uml_t.effect = effect
            flat_sm.transitions.append(uml_t)

    #
    # Print the statistics of the flattened state machine
    #
    if verbose_level >= 1:
        stats = Utils.StateMachines.Statistics(flat_sm)
        _verb("The flattened state machine has")
        _verb(" %5d state vertices" % stats.get_nof_state_vertices())
        _verb(" %5d concurrent composite states" %
              stats.get_nof_concurrent_composite_states())
        _verb(" %5d nonconcurrent composite states" %
              stats.get_nof_nonconcurrent_composite_states())
        _verb(" %5d simple states" % stats.get_nof_simple_states())
        _verb(" %5d final states" % stats.get_nof_final_states())
        _verb(" %5d initial pseudostates" %
              stats.get_nof_initial_pseudostates())
        _verb(" %5d choice pseudostates" % stats.get_nof_choice_pseudostates())
        _verb(" %5d transitions" % stats.get_nof_transitions())

    return flat_sm










def _make_flat_state(sconf):
    assert isinstance(sconf, StateConfiguration)
    assert sconf.uml_sv
    assert sconf.uml_sv.oclIsKindOf(UML14.CompositeState)
    assert len(sconf.children) == 1

    child_sconf = sconf.children[0]

    if child_sconf.is_initial():
        # The unique initial state configuration 
        flat_state = UML14.Pseudostate()
        flat_state.kind = "initial"
        flat_state.name = ",".join(sconf.getLeafStateNames())
        return flat_state

    if child_sconf.uml_sv.oclIsKindOf(UML14.FinalState):
        # A final state configuration
        flat_state = UML14.FinalState()
        flat_state.name = ",".join(sconf.getLeafStateNames())
        return flat_state

    if sconf.is_in_compound_transition():
        flat_state = UML14.Pseudostate()
        flat_state.kind = "choice"
        flat_state.name = ",".join(sconf.getLeafStateNames())
        return flat_state

    # Default case
    flat_state = UML14.SimpleState()
    flat_state.name = ",".join(sconf.getLeafStateNames())
    return flat_state


def _linearize_entry_actions(entry_actions_tree):
    linearized = []
    for (action,subactions) in entry_actions_tree:
        linearized.append(action)
        linearized.extend(_linearize_entry_actions(subactions))
    return linearized


def _linearize_exit_actions(exit_actions_tree):
    linearized = []
    for (action,subactions) in exit_actions_tree:
        linearized.extend(_linearize_exit_actions(subactions))
        linearized.append(action)
    return linearized


def _check_entryexit_action_disjointness(context_model,
                                         context_class,
                                         tree,
                                         left_tree_nodes):
    assert context_model
    assert context_class
    assert context_class.oclIsKindOf(UML14.Class)
    for (action,subtree) in tree:
        if action.state:
            assert action.state.entry == action
            action_type = "entry"
            action_state = action.state
        elif action.stateExit:
            assert action.stateExit.exit == action
            action_type = "exit"
            action_state = action.stateExit
        else:
            assert False
        action_info = Utils.DependencyAnalyzer.analyze_statement(context_model, context_class, str(action.script.body))
        for action2 in left_tree_nodes:
            #print " Compare '"+str(action)+"' with '"+str(action2)+"'"
            if action2.state:
                assert action2.state.entry == action2
                action2_type = "entry"
                action2_state = action2.state
            elif action2.stateExit:
                assert action2.stateExit.exit == action2
                action2_type = "exit"
                action2_state = action2.stateExit
            else:
                assert False
            assert action_type == action2_type
            action2_info = Utils.DependencyAnalyzer.analyze_statement(context_model, context_class, str(action2.script.body))
            #print action_info
            #print action2_info
            bad = False
            # Can a1 and a2 potentially send to the same object?
            action_receiving_classes = sets.Set([c for (signal,c) in action_info.signals_sent])
            action2_receiving_classes = sets.Set([c for (signal,c) in action2_info.signals_sent])
            if action_receiving_classes.intersection(action2_receiving_classes):
                print "Warning: Cannot quarantee the independence of "+\
                      "concurrent "+action_type+" actions "+\
                      "'"+str(action.script.body)+"' "+\
                      "(in state '"+str(action_state.name)+"') and "+\
                      "'"+str(action2.script.body)+"' "+\
                      "(in state '"+str(action2_state.name)+"') "+\
                      "in the state machine of "+\
                      "class '"+str(context_class.name)+"'. "+\
                      "Possibly sending signals to the same object."

            action_read = action_info.this_read.union(action_info.other_read)
            action2_read = action2_info.this_read.union(action2_info.other_read)
            action_write = action_info.this_write.union(action_info.other_write)
            action2_write = action2_info.this_write.union(action2_info.other_write)
            # Can a1 and a2 write to a same place?
            bad = action_write.intersection(action2_write)
            if bad:
                print "Warning: Cannot quarantee the independence of concurrent "+action_type+" actions '"+str(action.script.body)+"' (in state '"+str(action_state.name)+"') and '"+str(action2.script.body)+"' (in state '"+str(action2_state.name)+"') in the state machine of class '"+str(context_class.name)+"'. Possibly writing to the same locations: "+str([str(c.name)+"::"+str(e.name) for (c,e) in bad])
            # Can a1 write something that a2 can read?
            bad = action_write.intersection(action2_read)
            if bad:
                print "Warning: Cannot quarantee the independence of concurrent "+action_type+" actions '"+str(action.script.body)+"' (in state '"+str(action_state.name)+"') and '"+str(action2.script.body)+"' (in state '"+str(action2_state.name)+"') in the state machine of class '"+str(context_class.name)+"'. First possibly writes to a location read by the second: "+str([str(c.name)+"::"+str(e.name) for (c,e) in bad])
            bad = action2_write.intersection(action_read)
            if bad:
                print "Warning: Cannot quarantee the independence of concurrent "+action_type+" actions '"+str(action2.script.body)+"' (in state '"+str(action_state.name)+"') and '"+str(action.script.body)+"' (in state '"+str(action2_state.name)+"') in the state machine of class '"+str(context_class.name)+"'. First possibly writes to a location read by the second: "+str([str(c.name)+"::"+str(e.name) for (c,e) in bad])
        left_tree_nodes = _check_entryexit_action_disjointness(context_model,
                                                               context_class,
                                                               subtree,
                                                               left_tree_nodes)
        left_tree_nodes.append(action)
    return left_tree_nodes


def _copy_effect(effect):
    if not effect:
        return None

    if effect.oclIsKindOf(UML14.Action):
        #print effect.recurrence
        #print effect.script
        #print effect.target
        assert not effect.recurrence.body
        assert effect.script.body
        assert not effect.target.body
        new_effect = UML14.UninterpretedAction()
        new_effect.script.body = effect.script.body
        return new_effect

    if effect.oclIsKindOf(UML14.UninterpretedAction):
        #print effect.recurrence
        #print effect.script
        #print effect.target
        assert not effect.recurrence.body
        assert effect.script.body
        assert not effect.target.body
        new_effect = UML14.UninterpretedAction()
        new_effect.script.body = effect.script.body
        return new_effect

    raise RuntimeError("Unsupported effect action type '"+_mcname(effect)+"'")



def _make_effect(effect,exit_actions_tree,entry_actions_tree):

    exit_actions = _linearize_exit_actions(exit_actions_tree)
    entry_actions = _linearize_entry_actions(entry_actions_tree)

    if not effect:
        if exit_actions == [] and entry_actions == []:
            return None
        new_effect = UML14.UninterpretedAction()
        new_effect.script.body = ""
        for a in exit_actions:
            new_effect.script.body += str(a.script.body)
        for a in entry_actions:
            new_effect.script.body += str(a.script.body)
        return new_effect

    if effect.oclIsKindOf(UML14.Action):
        #print effect.recurrence
        #print effect.script
        #print effect.target
        assert not effect.recurrence.body
        assert effect.script.body
        assert not effect.target.body
        new_effect = UML14.UninterpretedAction()
        new_effect.script.body = ""
        for a in exit_actions:
            new_effect.script.body += str(a.script.body)
        new_effect.script.body += effect.script.body
        for a in entry_actions:
            new_effect.script.body += str(a.script.body)
        return new_effect

    if effect.oclIsKindOf(UML14.UninterpretedAction):
        #print effect.recurrence
        #print effect.script
        #print effect.target
        assert not effect.recurrence.body
        assert effect.script.body
        assert not effect.target.body
        new_effect = UML14.UninterpretedAction()
        new_effect.script.body = ""
        for a in exit_actions:
            new_effect.script.body += str(a.script.body)
        new_effect.script.body = effect.script.body
        for a in entry_actions:
            new_effect.script.body += str(a.script.body)
        return new_effect

    raise RuntimeError("Unsupported effect action type '"+_mcname(effect)+"'")
