"""
Classes and functions related to global configurations
in the SMUML UML simulator

Author: Tommi Junttila

Copyright (c) 2006-2007 Helsinki University of Technology
Released under the GNU General Public License version 2
"""

import sys
import types
import coral
import Utils.String
import Sim
import Sim.Core
from Sim.Object import Object

class GlobalState:
    """
    The class for global states in the UML simulator.
    """

    # The objects in the global state
    # objects = []


    def __init__(self):
        self.objects = []

    def get_nof_objects(self):
        """
        Get the number of objects in the global state.
        """
        return len(self.objects)

    def get_object(self, object_num):
        """
        Get the 'object_num'th object in the global state.
        Required: 0 <= 'object_num' < self.get_nof_objects()
        """
        assert 0 <= object_num and object_num < self.get_nof_objects()
        return self.objects[object_num]

    def compare(self, other):
        """
        Compare this GlobalState to the argument GlobalState 'other'.
        Returns a negative number, zero, or a strictly positive number,
        depending whether this GlobalState is 'smaller' than, equal to,
        or 'greater' than 'other', respectively.
        """
        assert isinstance(other, GlobalState)
        result = cmp(self.get_nof_objects(), other.get_nof_objects())
        if result != 0:
            return result
        for o1,o2 in zip(self.objects, other.objects):
            result = o1.compare(o2)
            if result != 0:
                return result
        return 0

    def api_print(self, f = sys.stdout, indent = None):
        """
        Prints the global state in file 'f'.
        'indent' should be None or an instance of Utils.String.Indent.
        """
        assert type(f) is file
        assert indent == None or isinstance(indent, Utils.String.Indent)
        if indent is None:
            indent = Utils.String.Indent()
        for o in self.objects:
            o.write(f, indent)
    

    def api_copy(self):
        """
        Return a copy of the global state
        """

        # The new global state
        new_global_state = GlobalState()

        # A map from my objects to new objects
        object_map = {}

        # Allocate new objects
        for o in self.objects:
            new_object = Object(o.get_class_def(), o.get_uml_object())
            new_global_state.objects.append(new_object)
            object_map[o] = new_object

        # Copy objects
        for o in self.objects:
            o._copy_to(object_map);

        return new_global_state
