"""
Classes and routines related to messages in the SMUML UML simulator

Authors: Tommi Junttila

Copyright (c) 2006-2007 Helsinki University of Technology
Released under the GNU General Public License version 2
"""

import coral
import Sim.Core
from Sim.Error import SimError

UML14 = coral.lang.UML14

class _MessageIdGenerator:
    def __init__(self):
        self.next_id = 0L
    def get_next_id(self):
        self.next_id += 1
        return self.next_id

_message_id_generator = _MessageIdGenerator()

#
# An instance of signal event (signal with instanciated parameters)
#
class Message:
    """
    A signal message.
    Each message has a reference to the UML Signal that it is
    'an instance of', and a list of parameter values (SimValues).
    In addition, each message has an identifier (long int),
    two messages represent the same message if they have the same
    identifier.
    """

    def __init__(self, signal, parameters, idnum = None):
        """
        Create a message with UML Signal 'signal'.
        'parameters' should be a list of SimValues whose types
        match the signal attributes.
        If 'idnum' is None, a new identifier (a long integer) is allocated
        for the message, otherwise the identifier is 'idnum'.
        """
        assert signal
        assert signal.oclIsKindOf(UML14.Signal)
        assert type(parameters) is list

        self.signal = signal            
        # Copy parameters with a type check
        self.parameters = []
        if len(parameters) != len(signal.feature):
            raise SimError("Illegal number of parameters to a signal")
        for uml_parameter,sim_parameter in zip(signal.feature,parameters):
            assert uml_parameter.oclIsKindOf(UML14.Attribute)
            assert isinstance(sim_parameter, Sim.Core.SimValue)
            vdef = Sim.Core._VariableDef_from_Attribute(uml_parameter)
            if vdef.get_type().__class__ != sim_parameter.get_type().__class__:
                raise SimError("Signal parameter type mismatch")
            self.parameters.append(sim_parameter.get_copy())
        if idnum == None:
            self.idnum = _message_id_generator.get_next_id()
        else:
            self.idnum = idnum


    def __str__(self):
        s = str(self.signal.name)+\
            "("+(",".join([p.get_str() for p in self.parameters]))+")"
        #s += "[id="+str(self.idnum)+"]"
        return s


    def copy(self):
        """
        Create a copy of the message.
        The identifier of the message is also copied.
        """
        new_msg = Message(self.signal, self.parameters, self.idnum)
        return new_msg


    def get_signal(self):
        """
        Return the UML Signal of the message.
        """
        assert self.signal
        assert self.signal.oclIsKindOf(UML14.Signal)
        return self.signal


    def get_parameter(self, i):
        """
        Return (a copy of) the i:th parameter of this Message.
        """
        assert type(i) is int
        assert 0 <= i and i < len(self.parameters)
        return self.parameters[i].get_copy()


    def compare(self, other):
        """
        Compare this Message to the argument Message 'other'.
        Returns a negative number, zero, or a strictly positive number,
        depending whether this Message is 'smaller' than, equal to,
        or 'greater' than 'other', respectively.
        The message identifier is not considered.
        """
        assert isinstance(other, Message)
        result = cmp(str(self.get_signal().__XMIid__()),
                     str(other.get_signal().__XMIid__()))
        if result != 0:
            return result
        assert len(self.parameters) ==  len(other.parameters)
        for p1,p2 in zip(self.parameters, other.parameters):
            assert isinstance(p1, Sim.Core.SimValue)
            assert isinstance(p2, Sim.Core.SimValue)
            result = p1.compare(p2)
            if result != 0:
                return result
        return 0
