"""
Classes and routines related to global configuration objects
in the SMUML UML simulator

Authors: Tommi Junttila and Timo Latvala

Copyright (c) 2005-2007 Helsinki University of Technology
Released under the GNU General Public License version 2
"""

import sys
import coral
import Utils
import Sim.Core

UML14 = coral.lang.UML14

class Object:
    """
    A class for representing Objects in a GlobalState of the simulator.
    """

    def __init__(self, definition, uml_object = None, idnum = -1):
        """
        'definition' is the ClassDef of which this Object is an instance.
        If 'uml_object' is not None, then it should be the UML Object to
        which this Object corresponds to.
        """
        assert isinstance(definition, Sim.Core.ClassDef)
        assert (uml_object is None) or uml_object.oclIsKindOf(UML14.Object)
        assert type(idnum) is int

        self._class_def = definition
        self._uml_object = uml_object
        self.name = None
        self.variable_values = {}
        self.sm_state = None
        self.sm_input_queue = []
        self.sm_defer_queue = []
        self.idnum = idnum
        
        # Resolve a name for the object
        if self._uml_object:
            assert self._uml_object.name
            self.name = str(self._uml_object.name)
        else:
            self.name = "<anonymous>"

        # Instanciate and set default values for local variables
        for (vname,vdef) in self.get_class_def().get_variable_defs().iteritems():
            value = Sim.Core.simValue_from_simType(vdef.get_type())
            assert vdef.get_name() == vname
            assert not self.variable_values.has_key(vdef.name)
            self.variable_values[vdef.name] = value

        # Behavioral element initialization
        uml_state_machine = self.get_class_def().uml_state_machine
        if uml_state_machine:
            self.sm_state = self._init_stateMachine(uml_state_machine)


    def _copy_to(self, object_map):
        """
        Copy object's information to the object object_map[self].
        'object_map' should be a dictionary from objects in the GlobalState
        of self to objects in some other target GlobalState.
        """
        assert type(object_map) is dict
        assert self in object_map
        destination_object = object_map[self];
        assert destination_object._class_def == self._class_def;
        assert destination_object._uml_object == self._uml_object;

        destination_object.idnum = self.idnum

        destination_object.name = self.name
        
        # Copy local variable values
        destination_object.variable_values = {}
        for n,v in self.variable_values.iteritems():
            destination_object.variable_values[n] = v.get_copy()
            # Object references must be modified according to object_map
            if isinstance(v.get_type(), Sim.Core.SimType_ObjectRef) and v.is_defined():
                if v.get_value() == "null":
                    destination_object.variable_values[n].set_value("null")
                else:
                    destination_object.variable_values[n].set_value(object_map[v.get_value()])

        # Copy state machine related information
        if self.sm_state:
            destination_object.sm_state = self.sm_state.copy()
            destination_object.sm_input_queue = []
            for event in self.sm_input_queue:
                destination_object.sm_input_queue.append(event.copy())
            for event in self.sm_defer_queue:
                destination_object.sm_defer_queue.append(event.copy())


    def get_class_def(self):
        """
        Get the ClassDef of which this Object is an instance.
        """
        assert isinstance(self._class_def, Sim.Core.ClassDef)
        return self._class_def
    

    def get_uml_object(self):
        """
        Get the UML Object to which this Object corresponds to
        (None if there is no such correspondence).
        """
        assert((self._uml_object is None) or
               self._uml_object.oclIsKindOf(UML14.Object))
        return self._uml_object


    def set_uml_object(self, uml_object):
        """
        Set the UML Object to which this Object corresponds to.
        Should not usually be used, use the constructor instead if possible.
        """
        assert uml_object.oclIsKindOf(UML14.Object)
        self._uml_object = uml_object
        assert uml_object.name
        self.name = str(uml_object.name)


    def get_uml_name(self, default = "<anonymous>"):
        """
        Get the name of the UML Object to which this Object corresponds to,
        'default' if no such correspondence exists.
        """
        if self._uml_object:
            assert self._uml_object.name
            return str(self._uml_object.name)
        else:
            return default


    def get_variable_value(self, var_name):
        """
        Get the SimValue of the variable named 'var_name'.
        WARNING: you should not usually modify the returned SimValue,
        doing it will modify this Object, too.
        """
        assert isinstance(var_name, str)
        if not self.variable_values.has_key(var_name):
            raise RuntimeError("Object "+self.get_uml_name()+"::"+
                               self.get_class_def().get_name()+
                               " has no variable '"+var_name+"'")
        return self.variable_values[var_name]


    def get_state_machine_configuration(self):
        """
        Get a copy of the state machine configuration of this object.
        """
        if self.sm_state is None:
            raise RuntimeError("Object "+self.get_uml_name()+"::"+
                               self.get_class_def().get_name()+
                               " has no state machine")
        return self.sm_state.copy()


    def set_state_machine_configuration(self, sm_configuration):
        """
        Set the state machine configuration of this object to (a copy of)
        'sm_configuration'.
        """
        assert isinstance(sm_configuration, Sim.Core.SM_StateNode)
        assert Utils.StateMachines.get_state_machine(sm_configuration.get_uml_reference()) == self.get_class_def().uml_state_machine
        if self.sm_state is None:
            raise RuntimeError("Object "+self.get_uml_name()+"::"+
                               self.get_class_def().get_name()+
                               " has no state machine")
        self.sm_state = sm_configuration.copy()


    def get_input_queue_length(self):
        """
        Get the number of messages in the input queue of this object.
        """
        return len(self.sm_input_queue)


    def get_input_queue_message(self, i):
        """
        Get the i:th Message in the input queue of this object.
        WARNING: You should not usually modify the returned Message,
        doing so will modify this object as well.
        """
        assert isinstance(i, int)
        assert 0 <= i
        assert i < self.get_input_queue_length()
        return self.sm_input_queue[i]


    def get_defer_queue_length(self):
        """
        Get the number of messages in the defer queue of this object.
        """
        return len(self.sm_defer_queue)


    def get_defer_queue_message(self, i):
        """
        Get the i:th Message in the defer queue of this object.
        WARNING: You should not usually modify the returned Message,
        doing so will modify this object as well.
        """
        assert isinstance(i, int)
        assert 0 <= i
        assert i < self.get_defer_queue_length()
        return self.sm_defer_queue[i]


    def _initialize_from_uml_object(self, uml_object_to_object_map):
        """
        Initialize this Object based on the corresponding UML Object.
        Required: self.get_uml_object() != None
        uml_object_to_object_map should provide a map from
        existing UML Objects to Objects.
        """
        assert self.get_uml_object() != None
        uml_object = self.get_uml_object()
        
        #
        # Initialize attribute values given in the slots of the object
        #
        for slot in uml_object.slot:
            assert slot.attribute
            assert slot.attribute.name
            attr_name = str(slot.attribute.name)
            assert attr_name in self.variable_values

            assert slot.value
            assert len(slot.value.classifier) == 1
            assert slot.value.classifier[0].oclIsKindOf(UML14.DataType)
            t = slot.value.classifier[0]
            if str(t.name) == "int":
                assert isinstance(self.get_variable_value(attr_name),
                                  Sim.Core.SimValue_Int)
                assert Utils.String.IsInt(str(slot.value.name))
                value = Sim.Core.SimValue_Int(int(str(slot.value.name)))
            elif str(t.name) == "boolean":
                assert isinstance(self.get_variable_value(attr_name),
                                  Sim.Core.SimValue_Boolean)
                assert str(slot.value.name) in ["false","true"]
                value = Sim.Core.SimValue_Boolean(str(slot.value.name)=="true")
            else:
                raise RuntimeError("DataType '"+str(t.name)+"' not handled")
            self.variable_values[attr_name] = value
        #value1 = object1.variable_values[name2]
        #assert isinstance(value1, Sim.Core.SimValue_ObjectRef)
        #value1.set_value(object2)

        #
        # Initialize object reference according to links
        #
        for link in uml_object.ownedLink:
            self._init_from_link(link, uml_object_to_object_map)
            

    def _init_stateMachine(self, sm):
        """
        Find the initial state configuration of a state machine.
        """
        assert sm.oclIsKindOf(UML14.StateMachine)
        if not sm.top.oclIsKindOf(UML14.CompositeState):
            raise "State machines with non-composite top states not supported"
        if sm.top.oclIsKindOf(UML14.SubmachineState):
            raise "State machines with submachineState top states not supported"
        if sm.top.entry:
            raise "State machines with top states having entry actions not supported"

        # Find the initial state
        return Sim.Core.SM_StateNode_from_uml_state(sm.top)
        

    def _init_from_link(self, link, uml_object_to_object_map):
        """
        Instantiate the object reference attributes from a link.
        """
        assert link.oclIsKindOf(UML14.Link)

        # Find the LinkEnds
        assert len(link.connection) == 2
        link_end1 = link.connection[0]
        link_end2 = link.connection[1]
        assert link_end1.oclIsKindOf(UML14.LinkEnd)
        assert link_end2.oclIsKindOf(UML14.LinkEnd)

        # LinkEnds should be associated with AssociationEnds
        if not link_end1.associationEnd or not link_end2.associationEnd:
            raise "A link end with no association end found"
        assert link_end1.associationEnd.oclIsKindOf(UML14.AssociationEnd)
        assert link_end2.associationEnd.oclIsKindOf(UML14.AssociationEnd)

        # Get the link end objects
        uml_object1 = link_end1.instance
        uml_object2 = link_end2.instance
        assert uml_object1
        assert uml_object2
        object1 = uml_object_to_object_map[uml_object1]
        object2 = uml_object_to_object_map[uml_object2]

        # Get the local variable names from the association ends
        name1 = str(link_end1.associationEnd.name)
        name2 = str(link_end2.associationEnd.name)

        # Set the values
        if name2:
            if link_end2.associationEnd.isNavigable:
                value1 = object1.variable_values[name2]
                assert isinstance(value1, Sim.Core.SimValue_ObjectRef)
                value1.set_value(object2)
            else:
                sys.stderr.write("Warning: ignoring a link end with a named, non-navigable association end: '"+Utils.Models.get_qualified_name(link_end2.associationEnd)+"'\n")
        if name1:
            if link_end1.associationEnd.isNavigable:
                value2 = object2.variable_values[name1]
                assert isinstance(value2, Sim.Core.SimValue_ObjectRef)
                value2.set_value(object1)
            else:
                sys.stderr.write("Warning: ignoring a link end with a named, non-navigable association end: '"+Utils.Models.get_qualified_name(link_end1.associationEnd)+"'\n")


    def consume(self):
        """
        Consume the first message from the input queue of the object.
        """
        assert len(self.sm_input_queue) >= 1
        self.sm_input_queue.pop(0)
    

    def compare(self, o):
        """
        Compare this Object to the argument Object 'o'.
        Returns a negative number, zero, or a strictly positive number,
        depending whether this Object is 'smaller' than, equal to,
        or 'greater' than 'o', respectively.
        """
        assert isinstance(o, Object)
        # Compare idnums. Should this be done?
        result = cmp(self.idnum, o.idnum)
        if result != 0:
            return result
        # Compare classes of which the objects are instances of.
        result = cmp(self.get_class_def(), o.get_class_def())
        if result != 0:
            return result
        # Compare class attribute values
        class_def = self.get_class_def()
        var_names = class_def.get_variable_defs().keys()
        for var_name in var_names:
            v1 = self.get_variable_value(var_name)
            v2 = o.get_variable_value(var_name)
            assert isinstance(v1, Sim.Core.SimValue)
            assert isinstance(v2, Sim.Core.SimValue)
            result = v1.get_type().compare(v2.get_type())
            if result != 0:
                return result
            result = v1.compare(v2)
            if result != 0:
                return result
        # Compare state machine configuration
        assert (self.sm_state is None) == (o.sm_state is None)
        if self.sm_state:
            result = cmp(self.sm_state, o.sm_state)
            if result != 0:
                return result
        # Compare state machine input queue contents
        result = cmp(len(self.sm_input_queue), len(o.sm_input_queue))
        if result != 0:
            return result
        for (m1,m2) in zip(self.sm_input_queue, o.sm_input_queue):
            result = m1.compare(m2)
            if result != 0:
                return result
        # Compare state machine defer queue contents
        result = cmp(len(self.sm_defer_queue), len(o.sm_defer_queue))
        if result != 0:
            return result
        for (m1,m2) in zip(self.sm_defer_queue, o.sm_defer_queue):
            result = m1.compare(m2)
            if result != 0:
                return result
        # Yes, the objects are the same!
        return 0


    def write(self, f, indent):
        """
        Print the state of the Object to the file f.
        'indent' should be an instance of Utils.String.Indent.
        """
        assert isinstance(f, file)
        assert isinstance(indent, Utils.String.Indent)

        f.write(str(indent) + "Object " + str(self.idnum) + " " +
                self.get_uml_name() + "::" + self.get_class_def().get_name() +
                "\n")
        indent.push()
        for (n,v) in self.variable_values.iteritems():            
            f.write(str(indent) + n + " = " + v.get_str() + "\n")
        if self.get_class_def().uml_state_machine:
            # Print the input queue contents
            f.write(str(indent) + "sm_input_queue = [")
            f.write(",".join([str(msg) for msg in self.sm_input_queue]))
            f.write("]\n")
            # Print the defer queue contents
            f.write(str(indent) + "sm_defer_queue = [")
            f.write(",".join([str(msg) for msg in self.sm_defer_queue]))
            f.write("]\n")

            # Print the state configuration
            f.write(str(indent) + "sm_state =\n")
            indent.push()
            self.sm_state._print(indent)
            indent.pop()
        indent.pop()


