import sys
import coral
import Sim

UML14 = coral.lang.UML14

#
#
# Print a UML state machine in dotty format
# The resulting graph may be too complicated for dotty if the state machine
#  has hierarchy and triggers/guards/effects on transitions.
# The transitions leaving and entering the same composite state cause
#  problems, too.
#
# Author: Tommi Junttila
#
#

def SMToDot(file, sm):
    assert sm.oclIsKindOf(UML14.StateMachine)
    assert sm.top

    file.write("digraph SM {\n")
    file.write("   graph [fontname=Helvetica];\n")
    file.write("   node [fontname=Helvetica, shape=box];\n")
    file.write("   edge [fontname=Helvetica];\n")
    file.write("   compound=true;\n")

    _print_states(file, sm.top, "   ")
    _print_transitions(file, sm.top, "   ")
    file.write("}\n")


def _escape_str(s):
    res = ""
    for c in s:
        if c == '"':
            res += "\\\""
        else:
            res += c
    return res

def _dotty_id(vertex):
    assert vertex.oclIsKindOf(UML14.StateVertex)
    return "\"" + str(vertex.__XMIid__()) + "\""

def _dotty_label(vertex):
    assert vertex.oclIsKindOf(UML14.StateVertex)
    s = str(vertex.name)
    if vertex.oclIsKindOf(UML14.State):
        # Print deferrable events as part of the vertex name
        for e in vertex.deferrableEvent:
            assert e.oclIsKindOf(UML14.SignalEvent)
            s += " defer("+str(e.signal.name)+")"
    return s

def _print_states(f, vertex, sep):
    assert vertex.oclIsKindOf(UML14.StateVertex)
    
    if vertex.oclIsKindOf(UML14.SubmachineState):
        assert False and "SubmachineStates not yet handled"

    elif vertex.oclIsKindOf(UML14.CompositeState):
        f.write(sep + "subgraph \"cluster" + str(vertex.__XMIid__()) + "\" {\n")
        f.write(sep + "   label=\""+_dotty_label(vertex)+"\";\n")
        for subvertex in vertex.subvertex:
            _print_states(f, subvertex, sep + "   ")
        f.write(sep + "}\n")

    elif (vertex.oclIsKindOf(UML14.Pseudostate) and vertex.kind == "initial"):
        f.write(sep + _dotty_id(vertex) +
                " [shape=circle, fillcolor=white, style=filled, height=.1]" +
                " [label=\""+_dotty_label(vertex)+"\"];\n")

    elif (vertex.oclIsKindOf(UML14.Pseudostate) and vertex.kind == "choice"):
        f.write(sep + _dotty_id(vertex) +
                " [shape=diamond, fillcolor=white, style=filled, height=.1]" +
                " [label=\""+_dotty_label(vertex)+"\"];\n")

    elif vertex.oclIsKindOf(UML14.FinalState):
        f.write(sep + _dotty_id(vertex) +
                " [shape=doublecircle,fillcolor=white,style=filled,height=.2]"+
                " [label=\""+_dotty_label(vertex)+"\"];\n")

    elif vertex.oclIsKindOf(UML14.SimpleState):
        f.write(sep + _dotty_id(vertex) +
                " [label=\""+_dotty_label(vertex)+"\"];\n")

    else:
        print vertex
        assert False and "Internal error: Vertex type not yet handled"

    return


def _print_transitions(f, vertex, sep):
    assert vertex.oclIsKindOf(UML14.StateVertex)
    
    if vertex.oclIsKindOf(UML14.CompositeState):
        for subvertex in vertex.subvertex:
            _print_transitions(f, subvertex, sep)

    if vertex.oclIsKindOf(UML14.State):
        for t in vertex.internalTransition:
            pass
    for t in vertex.outgoing:
        assert t.source
        assert t.target

        f.write(sep)
        if t.source.oclIsKindOf(UML14.CompositeState):
            source = _find_noncomposite_state(t.source)
            f.write(_dotty_id(source))
        else:
            f.write(_dotty_id(t.source))
        f.write(" -> ")
        if t.target.oclIsKindOf(UML14.CompositeState):
            target = _find_noncomposite_state(t.target)
            f.write(_dotty_id(target))
        else:
            f.write(_dotty_id(t.target))

        f.write(" [")

        if t.source.oclIsKindOf(UML14.CompositeState):
            f.write("ltail=\"cluster" + str(t.source.__XMIid__()) + "\", ")

        if t.target.oclIsKindOf(UML14.CompositeState):
            f.write("lhead=\"cluster" + str(t.target.__XMIid__()) + "\", ")

        f.write("label=\"")
        if t.trigger:
            assert t.trigger.oclIsKindOf(UML14.SignalEvent)
            f.write(_escape_str(str(t.trigger.signal.name)))
            param_names = [str(n.name) for n in t.trigger.parameter]
            s = "("+(",".join(param_names))+")"
            f.write(s)
        if t.guard:
            assert t.guard.expression
            assert t.guard.expression.body
            f.write("['"+_escape_str(str(t.guard.expression.body))+"']")
        if t.effect:
            assert t.effect.script
            assert t.effect.script.body
            f.write("/'"+_escape_str(str(t.effect.script.body))+"'")
        f.write("\"")

        f.write("];\n")

def _find_noncomposite_state(vertex):
    assert vertex.oclIsKindOf(UML14.StateVertex)
    
    if not vertex.oclIsKindOf(UML14.CompositeState):
        return vertex
    
    for subvertex in vertex.subvertex:
        state = _find_noncomposite_state(subvertex)
        if state:
            return state

    return None

