"""
Classes and routines related to state machine configurations
in the SMUML UML simulator

Author: Tommi Junttila

Copyright (c) 2006-2007 Helsinki University of Technology
Released under the GNU General Public License version 2
"""

import sys
import types
import sets

import coral

import Sim

import Utils.StateMachines
from Utils.StateMachines import has_outgoing_completion_transitions
from Utils.StateMachines import get_outgoing_completion_transitions

import Utils.String
import Utils.Models

def _mcname(element):
    """Return the metaclass name of Coral Element 'element'."""
    return Utils.Models.get_metaclassname(element)

def _qn(element):
    return Utils.Models.get_qualified_name(element)


UML14 = coral.lang.UML14

class StateConfiguration:
    """
    A node in a UML state machine state configuration tree.
    """

    def __init__(self, uml_sv):
        """
        Create a state configuration tree node corresponding to the
        UML StateVertex 'uml_sv'.
        Default entry completion is not performed.
        """
        assert uml_sv.oclIsKindOf(UML14.StateVertex)
        self.uml_sv = uml_sv
        self.children = []
        self.can_quiesce = (self.uml_sv.oclIsKindOf(UML14.State) and
                            has_outgoing_completion_transitions(self.uml_sv))
        # This is only relevant for states that can quiesce
        self.is_quiescent = False


    def __hash__(self):
        s = str(self.is_quiescent) + " " + str(self.uml_sv)
        h = [hash(s)]
        for child in self.children:
            h.append(hash(child))
        return hash(str(h))


    def __cmp__(self, other):
        if not isinstance(other, StateConfiguration):
            return cmp(id(self), id(other))
        v = cmp(self.is_quiescent, other.is_quiescent)
        if v != 0:
            return v
        v = cmp(self.uml_sv, other.uml_sv)
        if v != 0:
            return v
        v = cmp(len(self.children),len(other.children))
        if v != 0:
            return v
        for c1,c2 in zip(self.children, other.children):
            v = cmp(c1, c2)
            if v != 0:
                return v
        return 0


    def __str__(self):
        s = "["
        if self.can_quiesce:
            if self.is_quiescent:
                s += "q "
            else:
                s += "b "
        if self.uml_sv.name:
            s += str(self.uml_sv.name)+" "
        else:
            s += "<<anon>> "
        s += "<"+(",".join([str(c) for c in self.children]))+">"
        s += "]"
        return s


    def output(self, writer = Utils.String.Writer(sys.stdout)):
        """
        Print the state configuration.
        """
        assert isinstance(writer, Utils.String.Writer)
        s = ""
        if self.can_quiesce:
            if self.is_quiescent:
                s += "q "
            else:
                s += "b "
        if self.uml_sv.name:
            s += str(self.uml_sv.name)
        else:
            s += "<<anon>>"
        writer.ln(s)
        writer.push()
        for c in self.children:
            c.output(writer)
        writer.pop()


    def get_leaf_state_configurations(self):
        """
        Get the list of leaf StateConfigurations in the StateConfiguration
        (inclusive).
        """
        if self.children:
            assert self.uml_sv.oclIsKindOf(UML14.CompositeState)
            l = []
            for c in self.children:
                l.extend(c.get_leaf_state_configurations())
            return l
        assert not self.uml_sv.oclIsKindOf(UML14.CompositeState)
        return [self]
        

    def getLeafStateNames(self):
        """
        Get the list of leaf state names in this StateConfiguration
        (inclusive).
        """
        leaf_statenames = []
        for sconf in self.get_leaf_state_configurations():
            if sconf.uml_sv.name:
                leaf_statenames.append(str(sconf.uml_sv.name))
            else:
                leaf_statenames.append("<<anon>>")
        return leaf_statenames


    def copy(self):
        """
        Return a deep copy of the state configuration.
        """
        copied_sconf = StateConfiguration(self.uml_sv)
        copied_sconf.can_quiesce = self.can_quiesce
        copied_sconf.is_quiescent = self.is_quiescent
        copied_sconf.children = []
        for c in self.children:
            copied_sconf.children.append(c.copy())
        return copied_sconf

    
    def is_initial(self):
        """
        Is this an initial pseudostate?
        """
        assert self.uml_sv
        return (self.uml_sv.oclIsKindOf(UML14.Pseudostate) and
                self.uml_sv.kind=="initial")


    def is_concurrent_composite(self):
        """
        Is this a concurrent composite state?
        """
        assert self.uml_sv
        return (self.uml_sv.oclIsKindOf(UML14.CompositeState) and
                self.uml_sv.isConcurrent and
                len(self.children) > 1)
        return False


    def is_stable(self):
        """
        Check whether the state configuration is stable, meaning that
        (i) there are no pseudostates, and
        (ii) all completion sensitive simple and composite states
             are quiescent.
        """
        assert self.uml_sv

        if self.uml_sv.oclIsKindOf(UML14.CompositeState):
            all_children_stable = True
            all_children_in_final = True
            if self.uml_sv.isConcurrent:
                # iterate over regions
                for r in self.children:
                    assert r.uml_sv.oclIsKindOf(UML14.CompositeState)
                    assert len(r.children) == 1
                    assert not r.uml_sv.outgoing
                    assert not r.uml_sv.incoming
                    assert not r.is_quiescent
                    child = r.children[0]
                    all_children_stable = (all_children_stable and child.is_stable())
                    all_children_in_final = (all_children_in_final and child.uml_sv.oclIsKindOf(UML14.FinalState))
            else:
                assert len(self.children) == 1
                child = self.children[0]
                all_children_stable = (all_children_stable and child.is_stable())
                all_children_in_final = (all_children_in_final and child.uml_sv.oclIsKindOf(UML14.FinalState))

            if all_children_in_final:
                assert all_children_stable
                return not (self.can_quiesce and not self.is_quiescent)

            # Not all children are in final states
            return all_children_stable
            
        elif self.uml_sv.oclIsKindOf(UML14.Pseudostate):
            return False
        
        elif self.uml_sv.oclIsKindOf(UML14.SimpleState):
            return not (self.can_quiesce and not self.is_quiescent)
            
        elif self.uml_sv.oclIsKindOf(UML14.FinalState):
            assert not self.can_quiesce
            assert not self.is_quiescent
            return True

        raise RuntimeError("Inernal error: something not implemented")



    def is_in_rtc(self):
        """
        Is the state configuration in a run-to-completion step?
        """
        return not self.is_stable()


    
    def is_in_compound_transition(self):
        """
        Is the state configuration in a compound transition?
        That is, does it contain a pseudostate?
        """
        return len(self.get_pseudostate_configurations()) > 0
    


    def get_pseudostate_configurations(self):
        """
        Returns the Set of StateConfigurations corresponding to pseudostates
        within this StateConfiguration (inclusive).
        """
        uml_sv = self.uml_sv
        assert uml_sv
        assert not uml_sv.oclIsKindOf(UML14.SubmachineState)

        if uml_sv.oclIsKindOf(UML14.CompositeState):
            pseudostate_confs = sets.Set()
            for c in self.children:
                pseudostate_confs.update(c.get_pseudostate_configurations())
            return pseudostate_confs
        elif uml_sv.oclIsKindOf(UML14.FinalState):
            return sets.Set()
        elif uml_sv.oclIsKindOf(UML14.SimpleState):
            return sets.Set()
        elif uml_sv.oclIsKindOf(UML14.Pseudostate):
            return sets.Set([self])
        raise RuntimeError("StateVertex type '"+mcname(uml_sv)+"' "
                           "not handled");


    def substitute(self, old, new):
        """
        Replace the sub-state configuration 'old'
        within this state configuration with the state confguration 'new'.
        Return the resulting state configuration.
        Does not copy anything but mutates self -> self should not be
        included in any Set or dictionary.
        """
        assert isinstance(old, StateConfiguration)
        assert isinstance(new, StateConfiguration)

        if False:
            writer = Utils.String.Writer(sys.stdout, "r ")
            writer.ln("Substituting")
            writer.push()
            old.output(writer)
            writer.pop()
            writer.ln("with")
            writer.push()
            new.output(writer)
            writer.pop()
            writer.ln("in")
            writer.push()
            self.output(writer)
            writer.pop()

        assert self.contains_uml_state(old.uml_sv)
        return self._substitute(old, new)

    def _substitute(self, old, new):
        """
        The real code for substitute.
        """
        assert isinstance(self, StateConfiguration)
        assert isinstance(old, StateConfiguration)
        assert isinstance(new, StateConfiguration)

        if self is old:
            return new
        for i in range(len(self.children)):
            self.children[i] = self.children[i]._substitute(old, new)
        return self


    def contains_uml_state(self, uml_state):
        """
        Check if uml_state occurs in this state configuration.
        """
        return self.find_state(uml_state) != None
    

    def find_state(self, uml_state):
        """
        Find the sub state configuration (including self) that corresponds to
        the UML StateVertex uml_state.
        Return None is no such substate was found.
        """
        assert uml_state
        assert uml_state.oclIsKindOf(UML14.StateVertex)
        if self.uml_sv == uml_state:
            return self
        for child in self.children:
            found_state = child.find_state(uml_state)
            if found_state:
                return found_state
        return None


    def find_signals(self):
        """
        Returns the set of signals that appear as triggers or
        are deferrable in this state configuration.
        """
        signal_set = sets.Set()
        self._find_signals(signal_set)
        return signal_set

    def _find_signals(self, signal_set):
        """
        The real code for find_signals.
        """
        # Handle normal outgoing transitions
        for transition in self.uml_sv.outgoing:
            if not transition.trigger:
                continue
            if not transition.trigger.oclIsKindOf(UML14.SignalEvent):
                raise RuntimeError("'"+_mcname(transition.trigger)+"' "
                                   "not supported as a transition trigger")
            signal_set.add(transition.trigger.signal)

        if self.uml_sv.oclIsKindOf(UML14.SubmachineState):
            raise RuntimeError("Submachine states not yet supported")

        if self.uml_sv.oclIsKindOf(UML14.State):
            # Handle internal transitions
            for transition in self.uml_sv.internalTransition:
                if not transition.trigger:
                    continue
                if not transition.trigger.oclIsKindOf(UML14.SignalEvent):
                    raise RuntimeError("'"+_mcname(transition.trigger)+"' "
                                       "not supported as a transition trigger")
                signal_set.add(transition.trigger.signal)
            # Handle deferrable events
            for e in self.uml_sv.deferrableEvent:
                if not e.oclIsKindOf(UML14.SignalEvent):
                    raise RuntimeError("'"+_mcname(e)+"' "
                                       "not supported as a deferrable event")
                signal_set.add(e.signal)

        # Recursively do the same for all substates
        for child in self.children:
            child._find_signals(signal_set)


    def _gather_entry_actions_inclusive(self):
        entry_actions = []
        for c in self.children:
            c_entry_actions = c._gather_entry_actions_inclusive()
            if c_entry_actions != []:
                entry_actions.extend(c_entry_actions)
        if (self.uml_sv.oclIsKindOf(UML14.State) and
            self.uml_sv.entry and
            self.uml_sv.entry.script.body):
            entry_actions = [(self.uml_sv.entry,entry_actions)]
        return entry_actions

    def _gather_entry_actions_exclusive(self):
        entry_actions = []
        for c in self.children:
            c_entry_actions = c._gather_entry_actions_inclusive()
            if c_entry_actions != []:
                entry_actions.extend(c_entry_actions)
        return entry_actions

    def _gather_exit_actions_inclusive(self):
        exit_actions = []
        for c in self.children:
            c_exit_actions = c._gather_exit_actions_inclusive()
            if c_exit_actions != []:
                exit_actions.extend(c_exit_actions)
        if (self.uml_sv.oclIsKindOf(UML14.State) and
            self.uml_sv.exit and
            self.uml_sv.exit.script.body):
            exit_actions = [(self.uml_sv.exit,exit_actions)]
        return exit_actions

    def _gather_exit_actions_exclusive(self):
        exit_actions = []
        for c in self.children:
            c_exit_actions = c._gather_exit_actions_inclusive()
            if c_exit_actions != []:
                exit_actions.extend(c_exit_actions)
        return exit_actions



    def quiesce(self, uml_state):
        """
        Return the state configuration resulting when
        the completion sensitive UML State 'uml_state'
        in the state configuration is made quiescent.
        """
        assert uml_state.oclIsKindOf(UML14.StateVertex)
        new_sconf = self.copy()
        subsconf = new_sconf.find_state(uml_state)
        if not subsconf:
            raise RuntimeError("Internal error")
        if subsconf.is_quiescent:
            raise RuntimeError("Internal error")
        if not subsconf.can_quiesce:
            raise RuntimeError("Internal error")
        subsconf.is_quiescent = True
        return new_sconf



    def _find_enabled_1(self,
                        guard_evaluator = None,
                        guard_evaluator_params = None):
        """
        Find enabled completion transitions originating from pseudostates.
        """
        enabled = []

        for child in self.children:
            enabled.extend(child._find_enabled_1(guard_evaluator,
                                                 guard_evaluator_params))
        if enabled:
            return enabled

        uml_state = self.uml_sv

        # Not a pseudo state, return
        if not uml_state.oclIsKindOf(UML14.Pseudostate):
            return []

        outgoing_completion = get_outgoing_completion_transitions(uml_state)

        if guard_evaluator:
            for t in outgoing_completion:
                if t.guard:
                    assert t.guard.expression
                    assert t.guard.expression.body
                    if guard_evaluator(guard_evaluator_params, None,
                                       str(t.guard.expression.body)):
                        enabled.append(('completion',None,t))
                else:
                    enabled.append(('completion',None,t))
        else:
            # No guard evaluator
            for t in outgoing_completion:
                if t.guard:
                    assert t.guard.expression
                    assert t.guard.expression.body
                    enabled.append(('completion',
                                    str(t.guard.expression.body),
                                    t))
                else:
                    enabled.append(('completion', 'true', t))

        if not enabled:
            raise RuntimeError("None of the completion transitions "
                               "is enabled for pseudostate: "+_qn(uml_state))

        return enabled

                        


    def _find_enabled_2(self,
                        guard_evaluator = None,
                        guard_evaluator_params = None):
        """
        Find enabled completion transitions originating from
        states that have outgoing completion transitions and
        are ready to consume the implicit consumption event.
        If no completion transition is enabled for such a state,
        a 'quiescing' event is returned.
        """
        enabled = []

        for child in self.children:
            enabled.extend(child._find_enabled_2(guard_evaluator,
                                                 guard_evaluator_params))
        if enabled:
            return enabled

        uml_state = self.uml_sv

        # Not a state, no completion events
        if not uml_state.oclIsKindOf(UML14.State):
            return []

        # Final states cannot have outgoin transitions
        if uml_state.oclIsKindOf(UML14.FinalState):
            assert not uml_state.outgoing
            return []
        
        # No outgoing completion transitions?
        outgoing_completion = get_outgoing_completion_transitions(uml_state)
        if not outgoing_completion:
            return []

        # Already quiescent?
        if self.is_quiescent:
            return []

        # Find out whether all regions are in a final state
        if uml_state.oclIsKindOf(UML14.CompositeState):
            assert not(uml_state.isConcurrent and
                       len(uml_state.subvertex) != len(self.children))
            assert not(uml_state.isConcurrent and len(self.children) <= 1)
            assert not(not uml_state.isConcurrent and
                       len(self.children) != 1)
            all_in_final = True
            if uml_state.isConcurrent:
                # iterate over regions
                for child in self.children:
                    assert child.uml_sv.oclIsKindOf(UML14.CompositeState)
                    assert len(child.children) == 1
                    if not child.children[0].uml_sv.oclIsKindOf(UML14.FinalState):
                        all_in_final = False
                        break
            else:
                assert len(self.children) == 1
                if not self.children[0].uml_sv.oclIsKindOf(UML14.FinalState):
                    all_in_final = False
            if not all_in_final:
                return []

        if guard_evaluator:
            for t in outgoing_completion:
                if t.guard:
                    assert t.guard.expression
                    assert t.guard.expression.body
                    if guard_evaluator(guard_evaluator_params, None,
                                       str(t.guard.expression.body)):
                        enabled.append(('completion',None,t))
                else:
                    enabled.append(('completion',None,t))
            if not enabled:
                enabled.append('quiesce', None, uml_state)
            return enabled

        # No guard evaluator
        guards = []
        transs = []
        for t in outgoing_completion:
            if t.guard:
                assert t.guard.expression
                assert t.guard.expression.body
                guards.append(str(t.guard.expression.body))
            else:
                guards.append('true')
            transs.append(t)
        enabled = [('completion',g,t) for (g,t) in zip(guards, transs)]
        if 'true' in guards:
            return enabled
        g = " && ".join(["!("+str(g)+")" for g in guards])
        enabled.append(('quiesce', g, uml_state))
        return enabled
                        


    def _find_enabled_3(self,
                        signal,
                        guard_evaluator = None,
                        guard_evaluator_params = None):
        """
        Find 'signal'-triggered transitions that are enabled in this
        state configuration.
        """
        assert signal
        assert signal.oclIsKindOf(UML14.Signal)

        uml_state = self.uml_sv
        assert uml_state.oclIsKindOf(UML14.State)
        assert self.is_stable()

        enabled = []

        #
        # Internal transitions are not yet handled
        #
        if uml_state.internalTransition:
            raise RuntimeError("Internal transitions are not supported")

        # Final states cannot have outgoing transitions
        if uml_state.oclIsKindOf(UML14.FinalState):
            assert not uml_state.outgoing
            return []

        for child in self.children:
            enabled.extend(child._find_enabled_3(signal,
                                                 guard_evaluator,
                                                 guard_evaluator_params))
            
        # Already deferred at a deeper level?
        if len([r for (r,g,t) in enabled if r == 'defer']) > 0:
            return enabled

        if guard_evaluator:
            if enabled:
                return enabled
            for t in uml_state.outgoing:
                if not t.trigger:
                    continue
                assert t.trigger.oclIsKindOf(UML14.SignalEvent)
                if t.trigger.signal != signal:
                    continue
                if t.guard:
                    assert t.guard.expression
                    assert t.guard.expression.body
                    if guard_evaluator(guard_evaluator_params,
                                       t.trigger,
                                       str(t.guard.expression.body)):
                        enabled.append(('signal',None,t))
                else:
                    enabled.append(('signal',None,t))

            # Nothing enabled so far, try deferring
            for e in uml_state.deferrableEvent:
                assert e.oclIsKindOf(UML14.SignalEvent)
                if e.signal == signal:
                    # Event can be deferred
                    enabled.append(('defer', None, e))

            return enabled

        # No guard evaluator
        # Already consumed at a deeper level?
        if len([True for (r,g,t) in enabled if (g == 'true')]) > 0:
            return enabled

        guards = []
        transs = []
        for t in uml_state.outgoing:
            if not t.trigger:
                continue
            assert t.trigger.oclIsKindOf(UML14.SignalEvent)
            if t.trigger.signal != signal:
                continue
            transs.append(t)

        guards = []
        for t1 in transs:
            new_guard = "true"
            if t1.guard:
                new_guard = str(t1.guard.expression.body)
            trigger1 = t1.trigger
            for (r2,g2,t2) in enabled:
                assert r2 == 'signal'
                orig_guard2 = "true"
                if t2.guard:
                    orig_guard2 = str(t2.guard.expression.body)
                trigger2 = t2.trigger
                assert len(trigger1.parameter) == len(trigger2.parameter)
                param_rename = {}
                for (p1,p2) in zip(trigger1.parameter, trigger2.parameter):
                    assert p1.name
                    assert p2.name
                    if str(p1.name) != str(p2.name):
                        param_rename[(str(p2.name))] = str(p1.name)
                if not param_rename:
                    new_guard += " && !("+orig_guard2+")"
                else:
                    #tmp = _attribute_rename(orig_guard2, param_rename)
                    new_guard += " && !("+tmp+")"
                    raise RuntimeError("Symbolic flattening with signal "
                                       "parameters not yet supported")
            guards.append(new_guard)

        for (g,t) in zip(guards, transs):
            enabled.append(('signal',g,t))

        # Check whewther the signal can be deferred
        defer_guard = " && ".join(["!("+str(g)+")" for g in guards])
        for e in uml_state.deferrableEvent:
            assert e.oclIsKindOf(UML14.SignalEvent)
            if e.signal == signal:
                # Signal can be deferred
                enabled.append(('defer', defer_guard, e))

        return enabled



    def find_enabled(self,
                     signal,
                     guard_evaluator = None,
                     guard_evaluator_params = None):
        """
        """
        if self.is_in_compound_transition():
            return self._find_enabled_1(guard_evaluator,
                                        guard_evaluator_params)
        if self.is_in_rtc():
            return self._find_enabled_2(guard_evaluator,
                                        guard_evaluator_params)

        assert signal
        assert signal.oclIsKindOf(UML14.Signal)
        return self._find_enabled_3(signal,
                                    guard_evaluator,
                                    guard_evaluator_params)





    def fire(self, uml_transition):
        """
        Returns the state tree obtained from this by firing
        the UML Transition 'uml_transition'.
        """
        assert uml_transition
        assert uml_transition.oclIsKindOf(UML14.Transition)
        assert uml_transition.source
        assert uml_transition.target
        assert self.contains_uml_state(uml_transition.source)

        new_state_tree = self.copy()

        uml_lca_path = Utils.StateMachines.get_least_common_ancestor_path(uml_transition.source, uml_transition.target)
        assert len(uml_lca_path) >= 1
        uml_root = uml_lca_path[len(uml_lca_path)-1]
        assert uml_root.oclIsKindOf(UML14.CompositeState)
        root = new_state_tree.find_state(uml_root)
        assert root
        
        # Gather the tree of executed exit actions
        if uml_root.isConcurrent:
            # The exit action of the root has to be included
            exit_actions = root._gather_exit_actions_inclusive()
        else:
            # The exit action of the root is not included
            exit_actions = root._gather_exit_actions_exclusive()
        

        # Build the new, default entry completed sub-tree
        target_parent_path = Utils.StateMachines.get_parent_path(uml_transition.target)
        for i in range(0, len(uml_lca_path)-1):
            target_parent_path.pop(0)
        if False:
            print uml_lca_path
            print uml_transition.source
            print uml_transition.target
            print target_parent_path
        assert target_parent_path[0] == root.uml_sv
        new_root = StateConfiguration_from_uml_path(target_parent_path)
        new_state_tree = new_state_tree.substitute(root, new_root)
        new_root = new_state_tree.find_state(new_root.uml_sv)
        assert new_root

        # Gather the tree of executed entry actions
        if new_root.uml_sv.isConcurrent:
            # The entry action of the root has to be included
            entry_actions = new_root._gather_entry_actions_inclusive()
        else:
            # The entry action of the root is not included
            entry_actions = new_root._gather_entry_actions_exclusive()
        
        return (new_state_tree,exit_actions,entry_actions)



def StateConfiguration_from_uml_state(uml_state):
    """
    Build the sub state configuration corresponding to UML State 'uml_state'.
    Performs default entry completion downwards for composite states.
    """
    assert uml_state
    assert uml_state.oclIsKindOf(UML14.StateVertex)

    if uml_state.oclIsKindOf(UML14.SubmachineState):
        assert False and "Submachine states not yet handled"

    st = StateConfiguration(uml_state)
    if uml_state.oclIsKindOf(UML14.CompositeState):
        if uml_state.isConcurrent:
            for s in uml_state.subvertex:
                assert s.oclIsKindOf(UML14.CompositeState)
                st.children.append(StateConfiguration_from_uml_state(s))
        else:
            # Find the initial state
            for s in uml_state.subvertex:
                if s.oclIsKindOf(UML14.Pseudostate) and s.kind=="initial":
                    st.children.append(StateConfiguration_from_uml_state(s))
            if st.children == []:
                raise RuntimeError("Composite state " +
                                   sm_get_qualified_name(uml_state) +
                                   " without an initial state")
            if len(st.children) > 1:
                raise RuntimeError("Composite state " +
                                   sm_get_qualified_name(uml_state) +
                                   " with multiple initial states")            
    return st



def StateConfiguration_from_uml_path(uml_path):
    assert type(uml_path) is list
    assert len(uml_path) > 0

    if len(uml_path) == 1:
        return StateConfiguration_from_uml_state(uml_path[0])

    uml_state = uml_path[0]
    assert uml_state.oclIsKindOf(UML14.CompositeState)

    if uml_state.oclIsKindOf(UML14.SubmachineState):
        raise RuntimeError("Submachine states not yet handled")

    uml_path = uml_path[1:]

    st = StateConfiguration(uml_state)
    if uml_state.oclIsKindOf(UML14.CompositeState):
        if uml_state.isConcurrent:
            if uml_path == []:
                # Peform default entry completion with initial states
                for uml_region in uml_state.subvertex:
                    st.children.append(StateConfiguration_from_uml_state(uml_region))
            else:
                for uml_region in uml_state.subvertex:
                    if uml_region == uml_path[0]:
                        # Follow the path
                        child = StateConfiguration_from_uml_path(uml_path)
                    else:
                        # Peform default entry completion with initial state
                        child = StateConfiguration_from_uml_state(uml_region)
                    st.children.append(child)
        else:
            st.children.append(StateConfiguration_from_uml_path(uml_path))
        return st

    assert uml_path == []

    return st

