"""
Classes and functions related to traces in the SMUML UML simulator

Author: Tommi Junttila

Copyright (c) 2005-2007 Helsinki University of Technology
Released under the GNU General Public License version 2
"""

import sys
import types
import coral
import Sim
import Utils
from Utils.Models import get_qualified_name

UML14 = coral.meta.UML14
DIA20 = coral.meta.XMIDI20

indent = "  "

def _from_idnum_to_string(idnum):
    if idnum == "env":
        return "env"
    if isinstance(idnum, types.IntType):
        return str(idnum)
    assert False


class TraceAction:
    """
    Abstract base class for actions that can appear in traces.
    """
    pass


#
# 
#
class TA_ExecuteTransition(TraceAction):
    """
    Trace action denoting that an object executed a transition.
    """
    # Executing object's idnum
    executor_idnum = None
    # The UML transition
    transition = None

    def __init__(self, executor_idnum, transition):
        assert isinstance(executor_idnum, types.IntType)
        assert transition.oclIsKindOf(UML14.Transition)
        self.executor_idnum = executor_idnum
        self.transition = transition

    def __str__(self):
        s =  "Object " + str(self.executor_idnum)
        s += " executed transition '" + str(self.transition.name) + "'"
        return s


#
# Action for signal sending
#
class TA_SignalSend(TraceAction):
    """
    Trace action denoting that an object send a message to another object.
    """
    # Sender object's idnum or "env" for the environment
    sender_idnum = None
    # Receiver object's idnum or "env" for the environment
    receiver_idnum = None
    # The message
    message = None

    def __init__(self, sender_idnum, receiver_idnum, message):
        assert isinstance(message, Sim.Message.Message)
        self.sender_idnum = sender_idnum
        self.receiver_idnum = receiver_idnum
        self.message = message

    def __str__(self):
        s =  "Object " + _from_idnum_to_string(self.sender_idnum)
        s += " sent " + str(self.message)
        s += " to object "  + _from_idnum_to_string(self.receiver_idnum)
        return s


#
# Action for explicit consumption
#
class TA_Consumption(TraceAction):
    """
    Trace action denoting that an object explicitly consumed
    the first message in its input queue.
    """
    # Member variables:
    ## Object's idnum
    # object_idnum = None
    ## The message that is consumed
    # message

    def __init__(self, object_idnum, message):
        assert isinstance(message, Sim.Message.Message)
        self.object_idnum = object_idnum
        self.message = message

    def __str__(self):
        s =  "Object " + str(self.object_idnum)
        s += " consumed " + str(self.message)
        return s


#
# Action for implicit consumption
#
class TA_ImplicitConsumption(TraceAction):
    """
    Trace action denoting that an object implicitly consumed
    the first message in its input queue.
    """
    # Member variables:
    ## Object's idnum
    # object_idnum = None
    ## The message that is consumed
    # message

    def __init__(self, object_idnum, message):
        assert isinstance(message, Sim.Message.Message)
        self.object_idnum = object_idnum
        self.message = message

    def __str__(self):
        s =  "Object " + str(self.object_idnum)
        s += " implicitly consumed " + str(self.message)
        return s


#
# Action for object creation
#
class TA_CreateObject(TraceAction):
    """
    Trace action denoting that an object created a new object.
    """
    # Member variables:
    ## Creating object's idnum or "env" for the environment
    # creator_idnum = None
    ## Created object's idnum
    # created_idnum = None
    # created_uml_class
    
    def __init__(self, creator_idnum, created_idnum, uml_class):
        self.creator_idnum = creator_idnum
        self.created_idnum = created_idnum
        assert uml_class.oclIsKindOf(UML14.Class)
        self.created_uml_class = uml_class

    def __str__(self):
        s =  "Object " + _from_idnum_to_string(self.creator_idnum)
        s += " created object "  + _from_idnum_to_string(self.created_idnum)
        return s





class Trace:
    # A list of global states
    # states = []

    # actions[i] describes the actions (of type TraceAction) taken
    # between state i and i+1
    # actions = []
    
    # The next free object identifier array
    # next_free_oid_stack[i] keeps track of next_free_oid between states
    # for backtracking purposes

    def __init__(self):
        self.states = []
        self.actions = []
        self.next_free_oid = 0
        self.next_free_oid_stack = []

    def length(self):
        assert len(self.states) == len(self.actions)
        assert len(self.next_free_oid_stack) == len(self.states)
        return len(self.states)

    def get_next_free_oid(self):
        oid = self.next_free_oid
        self.next_free_oid += 1
        return oid

    def get_global_state(self, index):
        assert (type(index) is int) or (type(index) is long)
        assert 0 <= index and index < self.length()
        return self.states[index]

    def get_last_global_state(self):
        assert self.length() >= 1
        return self.states[-1]

    def add_global_state(self, global_state):
        assert isinstance(global_state, Sim.GlobalState.GlobalState)
        self.states.append(global_state)
        self.actions.append([])
        self.next_free_oid_stack.append(self.next_free_oid)
        assert len(self.states) == len(self.actions)
        assert len(self.next_free_oid_stack) == len(self.states)

    def add_action(self, action):
        assert isinstance(action, TraceAction)
        assert self.length() >= 1
        assert len(self.states) == len(self.actions)
        self.actions[self.length()-1].append(action)

    def backtrack_one_step(self):
        assert len(self.states) == len(self.actions)
        assert self.length() > 1
        self.states.pop()
        self.actions.pop()
        self.next_free_oid_stack.pop()
        self.next_free_oid = self.next_free_oid_stack[len(self.next_free_oid_stack)-1]
        # Reset the action sequence after the last state
        self.actions[len(self.states)-1] = []

    def output(self, f, indent):
        assert type(f) is file
        assert isinstance(indent, Utils.String.Indent)
        for i in xrange(0, self.length()):
            print >> f, str(indent) + "State", i
            indent.push()
            self.states[i].api_print(f, indent)
            for action in self.actions[i]:
                print >> f, "Action: "+str(action)
            indent.pop()

