#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# actions.py
#
# 2007-10-02  Jori Dubrovin:  New semantics for choose exprs with common keys
# 2007-08-27  Jori Dubrovin:  Suboco 1.00
# 2007-06-20  Jori Dubrovin:  New choose expression mapping
# 2007-03-15  Jori Dubrovin:  Added an error message
# 2007-03-02  Jori Dubrovin:  Adding choose to element map
# 2006-12-18  Jori Dubrovin:  Adding nondeterministic choose
# 2006-12-05  Jori Dubrovin:  Fixed cast error handling
# 2006-08-22  Jori Dubrovin:  Added modules, fixed arithmetic
# 2006-08-16  Jori Dubrovin:  Added checks for run-time errors
# 2006-07-13  Jori Dubrovin:  Initial version
#

"""Encoding of Jumbala."""

import exceptions

import actlang
import errors
import options
from datatypes import nullType, intType, boolType


def processActions(actionString, activity):
    """Apply the effect of Jumbala actions.

    The necessary expressions are added to the activity context via
    the given activity.  Any field read or write accesses and signal
    transmissions are added to the activity.

    Param actionString: the actions as a string containing zero or
    more Jumbala statements..

    Param activity: the Activity giving the current valuations of
    fields and receiving the changes.
    """
    tree = actlang.grammar.parse(actionString)
    # Extract statements other than plain semicolon
    stmts = [stmt for stmt in tree.api_getTopStatements() if
             not isinstance(stmt, actlang.statement.EmptyStatement)]
    sp = _StatementProcessor(activity)
    for stmt in stmts:
        sp.processStatement(stmt)

def processExpr(exprString, desiredType, activity):
    """Generate a side-effect free expression.

    As a side effect, any field read accesses are added to the
    activity.

    Param exprString: a string containing the side-effect free Jumbala
    expression.

    Param desiredType: the expected type of the expression.  Must not
    be the null type.  If the expression cannot be cast to this type,
    an InputError is raised.

    Param activity: the Activity containing the current valuations of
    fields.

    Return value: tuple (PySMT Gate for the value of the expression,
    PySMT Gate for the error flag).
    """
    tree = actlang.grammar.parse('value = (' + exprString + '\n);')
    if len(tree.api_getTopStatements()) > 1:
        raise errors.InputError('invalid expression ' + exprString)
    et = _ExprTranslator(activity)
    # Translate the right hand side of the assignment 'value = (...)'
    stmt = tree.api_getTopStatements()[0]
    (type_, smtValue, smtError) = et.translate(
        stmt.api_getExpr().api_getRhsExpr())
    if not type_.isCastableTo(desiredType):
        raise errors.InputError('invalid type for expression ' + exprString)
    return et.cast(desiredType, type_, smtValue, smtError)


class _StatementProcessor:
    """A processor of parsed Jumbala statements in an activity context."""

    def __init__(self, activity):
        """Construct a statement processor.

        Param activity: the Activity giving valuations of fields.  The
        effects of statements are made to this.
        """
        self.__activity = activity
        self.__et = _ExprTranslator(activity)

    def processStatement(self, stmt):
        """Apply the effect of a Statement."""
        if isinstance(stmt, actlang.statement.ExprStatement):
            expr = stmt.api_getExpr()
            if isinstance(expr, actlang.expr.SimpleAssignmentExpr):
                self.__processAssignment(expr)
            else:
                raise errors.NotSupportedError('unsupported statement ' +
                                               str(stmt))
        elif isinstance(stmt, actlang.statement.SendStatement):
            self.__processSendStatement(stmt)
        elif isinstance(stmt, actlang.statement.AssertStatement):
            self.__processAssertStatement(stmt)
        else:
            raise errors.NotSupportedError('unsupported statement ' +
                                           str(stmt))

    def __processSendStatement(self, stmt):
        """Apply the effect of a SendStatement."""
        try:
            signal = self.__activity.getSignalForName(stmt.api_getSignalName())
        except exceptions.LookupError:
            raise errors.InputError('unknown signal in ' + str(stmt))
        if signal.isExternal():
            raise errors.InputError('illegal signal in ' + str(stmt))
        args = [self.__et.translate(argExpr) for
                argExpr in stmt.api_getArgList()]
        targetExpr = stmt.api_getTarget()
        (targetType, smtTarget, smtError) = self.__et.translate(targetExpr)
        if not targetType.isReferenceType():
            raise errors.InputError('invalid target reference in ' + str(stmt))
        targetClass = targetType.getClass()
        self.__activity.addRuntimeError(smtError)
        ac = self.__activity.getCircuit()
        isNull = ac.add_eq(smtTarget, targetType.smtDefaultValue(ac))
        self.__activity.addRuntimeError(isNull)
        paramTypes = signal.getParamTypes()
        if len(args) != len(paramTypes):
            raise errors.InputError('wrong number of arguments in ' +
                                    str(stmt))
        smtArgs = []
        for ((argType, smtArg, smtError), paramType) in zip(args, paramTypes):
            if not argType.isSubtypeOf(paramType):
                raise errors.InputError('invalid argument type in ' +
                                        str(stmt))
            (smtArg, smtError) = self.__et.cast(paramType, argType,
                                                smtArg, smtError)
            self.__activity.addRuntimeError(smtError)
            smtArgs.append(smtArg)
        self.__activity.sendSignal(signal, targetClass, smtTarget, smtArgs)

    def __processAssertStatement(self, stmt):
        """Apply the effect of an AssertStatement."""
        (type_, smtValue, smtError) = self.__et.translate(stmt.api_getExpr())
        if type_ is not boolType:
            raise errors.InputError('invalid type in assertion ' + str(stmt))
        self.__activity.addRuntimeError(smtError)
        self.__activity.addAssertion(smtValue)

    def __processAssignment(self, expr):
        """Apply the effect of a SimpleAssignmentExpr."""
        lhsExpr = expr.api_getLhsExpr()
        rhsExpr = expr.api_getRhsExpr()
        if isinstance(lhsExpr, actlang.expr.PrimaryFieldAccessExpr):
            refExpr = lhsExpr.api_getReferenceExpr()
            (refType, smtRef, smtRefError) = self.__et.translate(refExpr)
            fieldName = lhsExpr.api_getFieldName()
        elif isinstance(lhsExpr, actlang.names.SimpleName):
            (refType, smtRef, smtRefError) = self.__activity.getThis()
            fieldName = lhsExpr.api_getParts()[0]
        elif isinstance(lhsExpr, actlang.names.QualifiedName):
            refExpr = lhsExpr.api_getQualifyingName()
            (refType, smtRef, smtRefError) = self.__et.translate(refExpr)
            fieldName = lhsExpr.api_getParts()[-1]
        else:
            raise errors.NotSupportedError('unsupported left-hand side ' +
                                           str(lhsExpr) + ' in assignment')
        if not refType.isReferenceType():
            raise errors.InputError('invalid reference in ' + str(lhsExpr))
        refClass = refType.getClass()
        try:
            field = refClass.getFieldForName(fieldName)
        except exceptions.LookupError:
            raise errors.InputError('unknown field ' + str(fieldName))
        self.__activity.addRuntimeError(smtRefError)
        ac = self.__activity.getCircuit()
        isNull = ac.add_eq(smtRef, refType.smtDefaultValue(ac))
        self.__activity.addRuntimeError(isNull)
        if isinstance(rhsExpr, actlang.expr.InstanceCreationExpr):
            className = rhsExpr.api_getTypeName()
            args = rhsExpr.api_getArgList()
            #$$ TODO
            print 'skipping instance creation ' + str(expr)
        else:
            (rhsType, smtRhs, smtRhsError) = self.__et.translate(rhsExpr)
            if not rhsType.isSubtypeOf(field.getType()):
                raise errors.InputError('type mismatch in assignment ' +
                                        str(expr))
            (smtRhs, smtRhsError) = self.__et.cast(field.getType(), rhsType,
                                                   smtRhs, smtRhsError)
            self.__activity.addRuntimeError(smtRhsError)
            self.__activity.writeField(field, smtRef, smtRhs)


class _ExprTranslator:
    """A translator for expressions in an activity context."""
    
    def __init__(self, activity):
        """Construct an expression translator.

        Param activity: the Activity giving valuations of fields.
        """
        self.__activity = activity

    def cast(self, toType, fromType, smtValue, smtError):
        """Generate a cast.

        The given types must be castable to each other.

        Param toType: the type to which the expression is cast.  Must
        not be the null type unless fromType is also the null type.

        Param fromType: type of the original expression.

        Param smtValue: a PySMT Gate for the value being cast.

        Param smtError: a PySMT Gate for the initial error flag.

        Return value: pair (PySMT Gate for the new value, PySMT Gate
        for the cast error OR'd with smtError).
        """
        if toType is fromType:
            return (smtValue, smtError)
        # The only case we need to handle so far is '(ClassType) null'.
        assert fromType is nullType and toType.isReferenceType()
        ac = self.__activity.getCircuit()
        return (toType.smtDefaultValue(ac), smtError)

    def translate(self, expr):
        """Translate a side-effect free expression.

        Param expr: the Expr to translate.

        Return value: tuple (Type of the expression, PySMT Gate for
        the value, PySMT Gate for the error flag).
        """
        # Remove parentheses
        while isinstance(expr, actlang.expr.ParenExpr):
            expr = expr.api_getInnerExpr()
        if isinstance(expr, actlang.expr.ThisExpr):
            return self.__activity.getThis()
        if isinstance(expr, actlang.names.SimpleName):
            return self.__translateSimpleName(expr)
        elif isinstance(expr, actlang.names.QualifiedName):
            return self.__translateQualifiedName(expr)
        elif isinstance(expr, actlang.expr.IntLiteralExpr):
            return self.__translateIntLiteral(expr)
        elif isinstance(expr, actlang.expr.BoolLiteralExpr):
            return self.__translateBoolLiteral(expr)
        elif isinstance(expr, actlang.expr.NullExpr):
            return self.__translateNull()
        elif isinstance(expr, actlang.expr.UnaryPlusExpr):
            return self.__translateUnaryPlus(expr)
        elif isinstance(expr, actlang.expr.UnaryMinusExpr):
            return self.__translateUnaryMinus(expr)
        elif isinstance(expr, actlang.expr.BitNotExpr):
            return self.__translateBitNot(expr)
        elif isinstance(expr, actlang.expr.LogicalNotExpr):
            return self.__translateLogicalNot(expr)
        elif isinstance(expr, actlang.expr.InfixExpr):
            return self.__translateInfix(expr)
        elif isinstance(expr, actlang.expr.TernaryCondExpr):
            return self.__translateTernaryCond(expr)
        elif isinstance(expr, actlang.expr.CondAndExpr):
            return self.__translateCondAnd(expr)
        elif isinstance(expr, actlang.expr.CondOrExpr):
            return self.__translateCondOr(expr)
        elif isinstance(expr, actlang.expr.PrimaryFieldAccessExpr):
            return self.__translatePrimaryFieldAccess(expr)
        elif isinstance(expr, actlang.expr.NameMethodInvocationExpr):
            return self.__translateNameMethodInvocation(expr)
        else:
            raise errors.NotSupportedError('unsupported expression ' +
                                           str(expr))

    # The following functions take an expression of a suitable kind
    # and, like translate(), return a tuple (type_, smtValue,
    # smtError).

    def __translateSimpleName(self, expr):
        # As long as there are no local variables or static fields, a
        # simple name 'x' is treated as 'this.x'.
        (refType, smtRef, smtError) = self.__activity.getThis()
        fieldName = expr.api_getParts()[0]
        return self.__genFieldAccess(refType, smtRef, smtError, fieldName)

    def __translateQualifiedName(self, expr):
        (refType, smtRef, smtError) = self.translate(
            expr.api_getQualifyingName())
        fieldName = expr.api_getParts()[-1]
        return self.__genFieldAccess(refType, smtRef, smtError, fieldName)

    def __translatePrimaryFieldAccess(self, expr):
        refExpr = expr.api_getReferenceExpr()
        fieldName = expr.api_getFieldName()
        (refType, smtRef, smtError) = self.translate(refExpr)
        return self.__genFieldAccess(refType, smtRef, smtError, fieldName)

    def __genFieldAccess(self, refType, smtRef, smtRefError, fieldName):
        """A general function for generating field access expressions."""
        if not refType.isReferenceType():
            raise errors.InputError('invalid reference for ' + str(fieldName))
        refClass = refType.getClass()
        try:
            field = refClass.getFieldForName(fieldName)
        except exceptions.LookupError:
            raise errors.InputError('unknown field ' + str(fieldName))
        smtValue = self.__activity.readField(field, smtRef)
        ac = self.__activity.getCircuit()
        isNull = ac.add_eq(smtRef, refType.smtDefaultValue(ac))
        smtError = ac.add_or(smtRefError, isNull)
        return (field.getType(), smtValue, smtError)

    def __translateIntLiteral(self, expr):
        ac = self.__activity.getCircuit()
        return (intType, intType.smtValue(ac, expr.api_getValue()),
                ac.add_false())

    def __translateBoolLiteral(self, expr):
        ac = self.__activity.getCircuit()
        return (boolType, boolType.smtValue(ac, expr.api_getValue()),
                ac.add_false())

    def __translateNull(self):
        ac = self.__activity.getCircuit()
        return (nullType, nullType.smtDefaultValue(ac), ac.add_false())

    def __translateUnaryPlus(self, expr):
        (opType, smtOp, smtError) = self.translate(expr.api_getOperand())
        if opType is not intType:
            raise errors.InputError('wrong operand type in ' + str(expr))
        return (intType, smtOp, smtError)

    def __translateUnaryMinus(self, expr):
        ac = self.__activity.getCircuit()
        (opType, smtOp, smtError) = self.translate(expr.api_getOperand())
        if opType is not intType:
            raise errors.InputError('wrong operand type in ' + str(expr))
        return (intType, ac.add_neg(smtOp), smtError)

    def __translateBitNot(self, expr):
        ac = self.__activity.getCircuit()
        (opType, smtOp, smtError) = self.translate(expr.api_getOperand())
        if opType is not intType:
            raise errors.InputError('wrong operand type in ' + str(expr))
        return (intType, ac.add_bitwise_not(smtOp), smtError)

    def __translateLogicalNot(self, expr):
        ac = self.__activity.getCircuit()
        (opType, smtOp, smtError) = self.translate(expr.api_getOperand())
        if opType is not boolType:
            raise errors.InputError('wrong operand type in ' + str(expr))
        return (boolType, ac.add_not(smtOp), smtError)

    def __translateInfix(self, expr):
        ac = self.__activity.getCircuit()
        (leftType, smtLeft, smtLeftError) = self.translate(
            expr.api_getLeftOperand())
        (rightType, smtRight, smtRightError) = self.translate(
            expr.api_getRightOperand())
        smtError = ac.add_or(smtLeftError, smtRightError)  # Default error flag
        operator = expr.api_getOperator()
        if isinstance(operator, actlang.binaryop.Plus):
            if leftType is not intType or rightType is not intType:
                raise errors.InputError('wrong operand type in ' + str(expr))
            type_ = intType
            moduleName = 'IntPlusExpr'
            smtValue = ac.add_plus(smtLeft, smtRight)
        elif isinstance(operator, actlang.binaryop.NumericalOperator):
            if leftType is not intType or rightType is not intType:
                raise errors.InputError('wrong operand type in ' + str(expr))
            type_ = intType
            if isinstance(operator, actlang.binaryop.Multiply):
                smtValue = ac.add_times(smtLeft, smtRight)
            elif isinstance(operator, actlang.binaryop.Divide):
                raise errors.NotSupportedError(
                    'Operator / not supported in SMT')
            elif isinstance(operator, actlang.binaryop.Remainder):
                raise errors.NotSupportedError(
                    'Operator % not supported in SMT')
            elif isinstance(operator, actlang.binaryop.Minus):
                smtValue = ac.add_minus(smtLeft, smtRight)
            elif isinstance(operator, actlang.binaryop.ShiftLeft):
                raise errors.NotSupportedError(
                    'Operator << not supported in SMT')
            elif isinstance(operator, actlang.binaryop.SignedShiftRight):
                raise errors.NotSupportedError(
                    'Operator >> not supported in SMT')
            elif isinstance(operator, actlang.binaryop.UnsignedShiftRight):
                raise errors.NotSupportedError(
                    'Operator >>> not supported in SMT')
            else:
                assert False  # No such operator
        elif isinstance(operator, actlang.binaryop.NumericalBooleanOperator):
            if leftType is intType and rightType is intType:
                type_ = intType
                if isinstance(operator, actlang.binaryop.BitLogicalAnd):
                    smtValue = ac.add_bitwise_and(smtLeft, smtRight)
                elif isinstance(operator, actlang.binaryop.BitLogicalXor):
                    smtValue = ac.add_bitwise_xor(smtLeft, smtRight)
                elif isinstance(operator, actlang.binaryop.BitLogicalOr):
                    smtValue = ac.add_bitwise_or(smtLeft, smtRight)
                else:
                    assert False  # No such operator
            elif leftType is boolType and rightType is boolType:
                type_ = boolType
                if isinstance(operator, actlang.binaryop.BitLogicalAnd):
                    smtValue = ac.add_and(smtLeft, smtRight)
                elif isinstance(operator, actlang.binaryop.BitLogicalXor):
                    smtValue = ac.add_not(ac.add_iff(smtLeft, smtRight))
                elif isinstance(operator, actlang.binaryop.BitLogicalOr):
                    smtValue = ac.add_or(smtLeft, smtRight)
                else:
                    assert False  # No such operator
            else:
                raise errors.InputError('incompatible operand types in ' +
                                        str(expr))
        elif isinstance(operator, actlang.binaryop.NumericalComparison):
            if leftType is not intType or rightType is not intType:
                raise errors.InputError('wrong operand type in ' + str(expr))
            type_ = boolType
            if isinstance(operator, actlang.binaryop.LessThan):
                smtValue = ac.add_lt(smtLeft, smtRight)
            elif isinstance(operator, actlang.binaryop.GreaterThan):
                smtValue = ac.add_gt(smtLeft, smtRight)
            elif isinstance(operator, actlang.binaryop.LessOrEqual):
                smtValue = ac.add_le(smtLeft, smtRight)
            elif isinstance(operator, actlang.binaryop.GreaterOrEqual):
                smtValue = ac.add_ge(smtLeft, smtRight)
            else:
                assert False  # No such operator
        elif isinstance(operator, actlang.binaryop.EqualityOperator):
            if not leftType.isCastableTo(rightType):
                raise errors.InputError('incompatible operand types in ' +
                                        str(expr))
            type_ = boolType
            if leftType.isSubtypeOf(rightType):
                # Cast the left operand to the type of the right operand.
                (smtLeft, smtLeftError) = self.cast(rightType, leftType,
                                                    smtLeft, smtLeftError)
                leftType = rightType
            elif rightType.isSubtypeOf(leftType):
                # Cast the right operand to the type of the left operand.
                (smtRight, smtRightError) = self.cast(leftType, rightType,
                                                      smtRight, smtRightError)
                rightType = leftType
            else:
                # Should not happen as long as there are no interface types.
                assert False
            smtError = ac.add_or(smtLeftError, smtRightError)
            if leftType is boolType:
                smtValue = ac.add_iff(smtLeft, smtRight)
            else:
                smtValue = ac.add_eq(smtLeft, smtRight)
            if isinstance(operator, actlang.binaryop.NotEqual):
                smtValue = ac.add_not(smtValue)
            else:
                assert isinstance(operator, actlang.binaryop.Equal)
        else:
            assert False  # No such operator
        return (type_, smtValue, smtError)

    def __translateTernaryCond(self, expr):
        ac = self.__activity.getCircuit()
        (condType, smtCond, smtCondError) = self.translate(
            expr.api_getCondition())
        (trueType, smtTrue, smtTrueError) = self.translate(
            expr.api_getTrueExpr())
        (falseType, smtFalse, smtFalseError) = self.translate(
            expr.api_getFalseExpr())
        if condType is not boolType:
            raise errors.InputError('wrong condition type in ' + str(expr))
        if trueType.isSubtypeOf(falseType):
            # Cast the true operand to the type of the false operand.
            (smtTrue, smtTrueError) = self.cast(falseType, trueType,
                                                smtTrue, smtTrueError)
            trueType = falseType
        elif falseType.isSubtypeOf(trueType):
            # Cast the false operand to the type of the true operand.
            (smtFalse, smtFalseError) = self.cast(trueType, falseType,
                                                  smtFalse, smtFalseError)
            falseType = trueType
        else:
            raise errors.InputError('incompatible operand types in ' +
                                    str(expr))
        if trueType is nullType or trueType.isReferenceType():
            moduleName = 'TernaryCondReferenceExpr'
        else:
            moduleName = 'TernaryCondPrimitiveExpr'
        smtValue = ac.add_ite(smtCond, smtTrue, smtFalse)
        smtError = ac.add_or(smtCondError,
                             ac.add_ite(smtCond, smtTrueError, smtFalseError))
        return (trueType, smtValue, smtError)

    def __translateCondAnd(self, expr):
        ac = self.__activity.getCircuit()
        (leftType, smtLeft, smtLeftError) = self.translate(
            expr.api_getLeftOperand())
        (rightType, smtRight, smtRightError) = self.translate(
            expr.api_getRightOperand())
        if leftType is not boolType or rightType is not boolType:
            raise errors.InputError('wrong operand type in ' + str(expr))
        smtError = ac.add_or(smtLeftError,
                             ac.add_and(smtLeft, smtRightError))
        return (boolType, ac.add_and(smtLeft, smtRight), smtError)

    def __translateCondOr(self, expr):
        ac = self.__activity.getCircuit()
        (leftType, smtLeft, smtLeftError) = self.translate(
            expr.api_getLeftOperand())
        (rightType, smtRight, smtRightError) = self.translate(
            expr.api_getRightOperand())
        if leftType is not boolType or rightType is not boolType:
            raise errors.InputError('wrong operand type in ' + str(expr))
        smtError = ac.add_or(smtLeftError,
                             ac.add_and(ac.add_not(smtLeft), smtRightError))
        return (boolType, ac.add_or(smtLeft, smtRight), smtError)

    def __translateNameMethodInvocation(self, expr):
        nameParts = expr.api_getInvocationName().api_getParts()
        args = expr.api_getArgList()
        if (nameParts == ['$$SMUML', 'choose_boolean'] and len(args) >= 3 and
            isinstance(args[0], actlang.expr.StringLiteralExpr) and
            isinstance(args[2], actlang.expr.IntLiteralExpr)):
            return self.__translateChooseBool(args[0].api_getValue(),
                                              str(args[2].api_getValue()))
        elif (nameParts == ['$$SMUML', 'choose_int'] and len(args) >= 3 and
              isinstance(args[0], actlang.expr.StringLiteralExpr) and
              isinstance(args[2], actlang.expr.IntLiteralExpr)):
            return self.__translateChooseInt(args[0].api_getValue(),
                                             str(args[2].api_getValue()))
        else:
            raise errors.NotSupportedError('unsupported method invocation ' +
                                           str(expr))

    def __translateChooseBool(self, valueString, chooseKey):
        """Translate a nondeterministic choose method for boolean type.

        Param valueString: a string containing the values among which
        to choose, as a comma-separated list.

        Param chooseKey: the string used as a key to identify the
        expression.

        Return value: (type_, smtValue, smtError).
        """
        ac = self.__activity.getCircuit()
        has = {False: False, True: False}
        for v in [s.strip() for s in valueString.split(',')]:
            if v == 'false':
                has[False] = True
            elif v == 'true':
                has[True] = True
            else:
                raise errors.InputError('invalid argument for ' +
                                        'choose_boolean: ' + valueString)
        if self.__activity.hasChoose(chooseKey):
            if self.__activity.getChooseSignature(chooseKey) != has:
                raise errors.InputError('inconsistent choose expressions')
            smtValue = self.__activity.getChooseValue(chooseKey)
        else:
            if has[False] and has[True]:
                smtValue = ac.add_var()
            elif has[False]:
                smtValue = ac.add_false()
            elif has[True]:
                smtValue = ac.add_true()
            else:
                assert False
            self.__activity.addChoose(chooseKey, smtValue, has)
        return (boolType, smtValue, ac.add_false())

    def __translateChooseInt(self, valueString, chooseKey):
        """Translate a nondeterministic choose method for int type.

        Param valueString: a string containing the values among which
        to choose, as a comma-separated list.

        Param chooseKey: the string used as a key to identify the
        expression.

        Return value: (type_, smtValue, smtError).
        """
        ac = self.__activity.getCircuit()
        args = []
        for v in [s.strip() for s in valueString.split(',')]:
            try:
                args.append(int(v))
            except exceptions.ValueError:
                raise errors.InputError('invalid argument for ' +
                                        'choose_int: ' + valueString)
        if self.__activity.hasChoose(chooseKey):
            if self.__activity.getChooseSignature(chooseKey) != args:
                raise errors.InputError('inconsistent choose expressions')
            smtValue = self.__activity.getChooseValue(chooseKey)
        else:
            smtValue = intType.smtVariable(ac)
            self.__activity.addConstraint(
                ac.add_or(*[ac.add_eq(smtValue, intType.smtValue(ac, arg)) for
                            arg in args]))
            self.__activity.addChoose(chooseKey, smtValue, args)
        return (intType, smtValue, ac.add_false())
