#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# activities.py
#
# 2007-10-02  Jori Dubrovin:  New semantics for choose exprs with common keys
# 2007-09-01  Jori Dubrovin:  Suboco 1.02
# 2007-08-27  Jori Dubrovin:  Suboco 1.00
# 2007-07-17  Jori Dubrovin:  Fixed addExpr invocation
# 2007-06-20  Jori Dubrovin:  New choose expression mapping
# 2007-03-02  Jori Dubrovin:  Adding choose to element map
# 2006-12-18  Jori Dubrovin:  Adding nondeterministic choose
# 2006-11-25  Jori Dubrovin:  Implementing events
# 2006-11-15  Jori Dubrovin:  Initial version
#

"""Determining the side effects of transitions."""

import errors


class Activity:
    """An activity that reads and/or changes the global configuration.

    An activity consists of read accesses from fields, write accesses
    to fields, and signal transmissions to objects.  Additionally, an
    activity may result in a Jumbala run-time error or assertion error
    when executed.  Changes to the active state of a state machine are
    not part of an activity.

    If an activity is associated with a transition, then it consists
    of information about fields read by the guard, fields updated when
    reading signal parameters, fields accessed by the transition
    effect, signals sent by the transition effect, assertions in the
    effect, and possible Jumbala run-time errors caused by evaluating
    the guard or effect.  In this case, one Activity object
    corresponds to one Transition, not one transition instance.
    """

    def __init__(self, class_, circuit, isConstant=False):
        """Create an empty activity.

        Param class_: the ActiveClass in which the activity occurs.

        Param circuit: the PySMT Circuit used for storing intermediate
        values.

        Param isConstant: True if this activity is not allowed to
        access 'this' or read any fields.
        """
        self.__class = class_
        self.__circuit = circuit
        self.__isConstant = isConstant
        self.__reads = []
        self.__readForField = {}        # map Field to Read
        self.__writes = []
        self.__writeForField = {}       # map Field to Write
        self.__sends = []
        self.__sendForClass = {}        # map target ActiveClass to Send
        self.__errors = []              # list of Gates
        self.__assertions = []          # list of Gates
        self.__constraints = []         # list of Gates
        self.__valueForChoose = {}      # map choose key string to Gate
        self.__signatureForChoose = {}  # map choose key string to signature

    def getCircuit(self):
        """Return the activity circuit."""
        return self.__circuit

    def getFieldReads(self):
        """Return the list of Reads from fields."""
        return self.__reads[:]

    def getFieldWrites(self):
        """Return the list of Writes to fields."""
        return self.__writes[:]

    def getSends(self):
        """Return the list of signal Sends."""
        return self.__sends[:]

    def getRuntimeErrors(self):
        """Return Jumbala run-time errors associated with self.

        Assertion errors are not included in the list.

        Return value: a list of PySMT Gates.  If any of them evaluates
        to true, then execution of self results in a run-time error
        (unless an assertion is violated before that).
        """
        return self.__errors[:]

    def getAssertions(self):
        """Return assertions associated with self.

        Return value: a list of PySMT gates.  If any of them evaluates
        to false, then execution of self results in an assertion
        violation (unless a run-time error occurs before that).
        """
        return self.__assertions[:]

    def getConstraints(self):
        """Return constraints associated with self as a list of PySMT Gates."""
        return self.__constraints[:]

    def getThis(self):
        """Return the expression 'this'.

        Return value: (Type of the ActiveClass corresponding to the
        activity, PySMT Gate giving the reference to this, PySMT Gate
        giving the error flag (always false)).
        """
        if self.__isConstant:
            raise errors.InputError("cannot access 'this' from the context")
        return (self.__class.getType(),
                self.__class.varSmtThisRef(self.__circuit, None),
                self.__circuit.add_false())

    def getSignalForName(self, signalName):
        """Return the Signal with the given name.

        If there are no such Signals, raise a LookupError.  If there
        are several, raise an InputError.
        """
        return self.__class.getSignalForName(signalName)

    def getChooseMap(self):
        """Return the choose expressions used in this activity.

        Return value: a dictionary from choose key strings to PySMT
        Gates.
        """
        return dict(self.__valueForChoose)

    def clearChooseMap(self):
        """Clear the choose expressions added so far in this activity."""
        self.__valueForChoose = {}
        self.__signatureForChoose = {}

    def hasChoose(self, chooseKey):
        """Return True if a choose with the given key has been added."""
        return chooseKey in self.__valueForChoose

    def getChooseValue(self, chooseKey):
        """Return the PySMT Gate attached to the given key."""
        return self.__valueForChoose[chooseKey]

    def getChooseSignature(self, chooseKey):
        """Return the signature attached to the given key."""
        return self.__signatureForChoose[chooseKey]

    def addChoose(self, chooseKey, smtValue, signature):
        """Add a (key, value) pair to the choose map.

        Param chooseKey: a choose key string.

        Param smtValue: a PySMT Gate whose value is the value chosen
        for the key.
        """
        self.__valueForChoose[chooseKey] = smtValue
        self.__signatureForChoose[chooseKey] = signature

    def readField(self, field, smtReference):
        """Add a read from a field.

        Param field: the Field that is read.

        Param smtReference: a PySMT gate giving the reference
        expression.

        Return value: a PySMT gate giving the value of the field.
        """
        if self.__isConstant:
            raise errors.InputError("cannot access fields from the context")
        if field in self.__readForField:
            read = self.__readForField[field]
        else:
            read = Read(self, field)
            self.__reads.append(read)
            self.__readForField[field] = read
        read.addRead(smtReference)
        if field in self.__writeForField:
            write = self.__writeForField[field]
        else:
            write = Write(self, field)
        ac = self.__circuit
        class_ = field.getOwner()
        smtValue = field.getType().smtDefaultValue(ac)
        for obj in class_.getObjects():
            smtValue = ac.add_ite(
                ac.add_eq(smtReference, obj.smtReference(ac)),
                write.smtValueIn(obj),
                smtValue)
        return smtValue

    def writeField(self, field, smtReference, smtValue):
        """Add a write to a field.

        Param field: the Field that is written.

        Param smtReference: a PySMT Gate for the reference expression.

        Param smtValue: a PySMT Gate for the written value.
        """
        if field in self.__writeForField:
            write = self.__writeForField[field]
        else:
            write = Write(self, field)
            self.__writes.append(write)
            self.__writeForField[field] = write
        write.addWrite(smtReference, smtValue)

    def sendSignal(self, signal, targetClass, smtTarget, smtArgs):
        """Add a signal transmission.

        Raises an InputError if a signal to targetClass has already
        been added.

        Param signal: the Signal being sent.

        Param targetClass: the ActiveClass of the target.

        Param smtTarget: a PySMT Gate giving the reference for the
        signal target.

        Param smtArgs: a list of PySMT Gates giving the argument
        values for signal parameters.  The number and types must
        match.
        """
        if targetClass in self.__sendForClass:
            raise errors.InputError('multiple sends to a class')
        send = Send(self, signal, targetClass, smtTarget, smtArgs)
        self.__sends.append(send)
        self.__sendForClass[targetClass] = send
        targetClass.getEventQueue().addReceivedSignal(signal)

    def addRuntimeError(self, smtError):
        """Add a Jumbala run-time error.

        Param smtError: a PySMT Gate that evaluates to true in case of
        an error.
        """
        self.__errors.append(smtError)

    def addAssertion(self, smtAssertion):
        """Add an assertion.

        Param assertion: a PySMT Gate that evaluates to false in case
        of an assertion violation.
        """
        self.__assertions.append(smtAssertion)

    def addConstraint(self, smtConstraint):
        """Add a constraint.

        Param assertion: a PySMT Gate that must evaluate to true if
        this is a valid activity.
        """
        self.__constraints.append(smtConstraint)


class Access:
    """Base class for reading or writing a field, possibly several times."""

    def __init__(self, activity, field):
        """Initialize the access.

        Param activity: the Activity in which the execution occurs.

        Param field: the Field being read or written.
        """
        self._activity = activity
        self._field = field
        self.__smtReferences = []

    def getField(self):
        """Return the Field being read or written."""
        return self._field

    def isLocal(self):
        """Return True if all accesses occur through the reference this."""
        (type_, smtThis, smtError) = self._activity.getThis()
        for ref in self.__smtReferences:
            if ref is not smtThis:
                return False
        return True

    def _add(self, smtReference):
        """Add an access through a reference given as a PySMT gate."""
        if smtReference not in self.__smtReferences:
            self.__smtReferences.append(smtReference)

    def smtReferences(self):
        """Return the list of PySMT gates used as references."""
        return self.__smtReferences[:]


class Read(Access):
    """The act of reading the value of a field, possibly several times."""

    def __init__(self, activity, field):
        """Create an empty read access object.

        Param activity: the Activity in which the execution occurs.

        Param field: the Field being read.
        """
        Access.__init__(self, activity, field)

    def addRead(self, smtReference):
        """Add a read through a reference given as a PySMT Gate."""
        Access._add(self, smtReference)


class Write(Access):
    """The act of writing the value of a field, possibly several times."""

    def __init__(self, activity, field):
        """Create an empty write access object.

        Param activity: the Activity in which the execution occurs.

        Param field: the Field being written.
        """
        Access.__init__(self, activity, field)
        self.__valueIn = {}
        ac = activity.getCircuit()
        for obj in field.getOwner().getObjects():
            self.__valueIn[obj] = field.varSmtFieldValue(ac, obj)

    def addWrite(self, smtReference, smtValue):
        """Add a write to the field.

        Param smtReference: a PySMT Gate for the reference expression.

        Param smtValue: a PySMT Gate for the written value.
        """
        Access._add(self, smtReference)
        ac = self._activity.getCircuit()
        for obj in self._field.getOwner().getObjects():
            isObj = ac.add_eq(smtReference, obj.smtReference(ac))
            self.__valueIn[obj] = ac.add_ite(isObj, smtValue,
                                             self.__valueIn[obj])

    def smtValueIn(self, obj):
        """Return a PySMT Gate for the written value in the given object."""
        return self.__valueIn[obj]


class Send:
    """A signal transmission."""

    def __init__(self, activity, signal, targetClass, smtTarget, smtArgs):
        """Create a signal transmission.

        Param activity: the Activity in which the execution occurs.

        Param signal: the Signal being sent.

        Param targetClass: the ActiveClass of the target.

        Param smtTarget: a PySMT Gate giving the reference to the
        signal target.

        Param smtArgs: a list of PySMT Gates giving the argument
        values.
        """
        self.__activity = activity
        self.__signal = signal
        self.__targetClass = targetClass
        self.__smtTarget = smtTarget
        self.__smtArgs = smtArgs

    def getTargetClass(self):
        """Return the ActiveClass of the target object."""
        return self.__targetClass

    def getSignal(self):
        """Return the Signal sent."""
        return self.__signal

    def isLocal(self):
        """Return True if the sending occurs through the reference this."""
        (type_, smtThis, smtError) = self.__activity.getThis()
        return (self.__smtTarget is smtThis)

    def smtTarget(self):
        """Return the target reference as a PySMT Gate."""
        return self.__smtTarget

    def smtArgs(self):
        """Return the argument values as a list of PySMT Gates."""
        return self.__smtArgs[:]
