#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# classes.py
#
# 2007-12-16  Jori Dubrovin:  Cleaned up options, changed defaults
# 2007-11-22  Jori Dubrovin:  Added real enums
# 2007-08-27  Jori Dubrovin:  Suboco 1.00
# 2007-07-17  Jori Dubrovin:  Fixed attribute initialization
# 2007-06-20  Jori Dubrovin:  New choose expression mapping
# 2007-03-08  Jori Dubrovin:  Adding states to element map
# 2007-03-02  Jori Dubrovin:  Adding choose to element map
# 2006-12-18  Jori Dubrovin:  Fixed Field.__owner
# 2006-12-05  Jori Dubrovin:  Fixed cast error handling
# 2006-08-31  Jori Dubrovin:  Added encodeStateExclusion
# 2006-08-16  Jori Dubrovin:  Added checks for run-time errors
# 2006-07-28  Jori Dubrovin:  Initial version
#

"""Encoding of UML classes."""

import exceptions

import actions
import activities
import errors
import events
import statemachines
import datatypes
import options
import smtutils
from coralutils import uml, umlSort


class ActiveClass:
    """Representation of an active UML class."""

    def __init__(self, umlClass, id_, model):
        """Construct an active class.

        Param umlClass: the underlying uml.Class.

        Param id_: an identification string.

        Param model: the Model owning the class.
        """
        assert umlClass.oclIsKindOf(uml.Class)
        self.__umlClass = umlClass
        self.__id = id_
        self.__model = model
        self.__fields = []                    # all Fields of the class
        self.__fieldForUmlAttribute = {}      # map uml.Attribute to Field
        self.__fieldForUmlAssociationEnd = {} # map uml.AssociationEnd to Field
        self.__objects = []                   # all ActiveObjects of the class
        self.__objectForUml = {}              # map uml.Object to ActiveObject
        if (len(umlClass.behavior) != 1 or
            not umlClass.behavior[0].oclIsKindOf(uml.StateMachine)):
            raise errors.NotSupportedError('class ' + str(umlClass) +
                                           'does not have exactly one ' +
                                           'state machine')
        self.__machine = statemachines.StateMachine(umlClass.behavior[0], self)
        self.__queue = events.EventQueue(self)
        self.__type = datatypes.ReferenceType(self)
        self.__initialCircuit = smtutils.LabeledCircuit()

    def __str__(self):
        return self.__id

    def getModel(self):
        """Return the Model that owns self."""
        return self.__model

    def getStateMachine(self):
        """Return the StateMachine of self."""
        return self.__machine

    def getEventQueue(self):
        """Return the EventQueue of self."""
        return self.__queue

    def getType(self):
        """Return the ReferenceType corresponding to self."""
        return self.__type

    def getInitialCircuit(self):
        """Return the LabeledCircuit used for initial configuration."""
        return self.__initialCircuit

    def getFields(self):
        """Return all Fields."""
        return self.__fields[:]

    def getObjects(self):
        """Return all ActiveObjects of this class."""
        return self.__objects[:]

    def getObjectForUml(self, umlObject):
        """Return the ActiveObject corresponding to a given uml.Object.

        If there is no such object of this class, raise a LookupError.
        """
        return self.__objectForUml[umlObject]

    def getName(self):
        """Return the name of the class as a string."""
        return str(self.__umlClass.name)

    def getFieldForName(self, fieldName):
        """Return the Field with the given name.

        If there are no such Fields, raise a LookupError.  If there
        are several, raise an InputError.
        """
        fields = [f for f in self.__fields if f.getName() == fieldName]
        if len(fields) > 1:
            raise errors.InputError("ambiguous field '" + fieldName +
                                    "' in " + str(self.__umlClass))
        return fields[0]

    def getFieldForUmlAttribute(self, umlAttribute):
        """Return the Field corresponding to a uml.Attribute.

        If there are no such Fields, raise a LookupError.
        """
        return self.__fieldForUmlAttribute[umlAttribute]

    def getFieldForUmlAssociationEnd(self, umlAssociationEnd):
        """Return the Field corresponding to a uml.AssociationEnd.

        If there are no such Fields, raise a LookupError.
        """
        return self.__fieldForUmlAssociationEnd[umlAssociationEnd]

    def getSignalForUml(self, umlSignal):
        """Return the global Signal for a given uml.Signal.

        If there are no such Signals, raise a LookupError.
        """
        return self.__model.getSignalForUml(umlSignal)

    def getSignalForName(self, signalName):
        """Return the Signal with the given name.

        If there are no such Signals, raise a LookupError.  If there
        are several, raise an InputError.
        """
        return self.__model.getSignalForName(signalName)

    def addObject(self, umlObject):
        """Add a new ActiveObject to self.

        Param umlObject: the underlying uml.Object, or None if the
        object is not initially alive.
        """
        i = len(self.__objects) + 1
        id_ = 'i' + str(i) + '#' + self.__id
        obj = ActiveObject(umlObject, id_, i, self)
        self.__objects.append(obj)
        if umlObject:
            assert not umlObject in self.__objectForUml
            self.__objectForUml[umlObject] = obj

    def processObjects(self):
        """Add the remaining objects to self."""
        if False:
            # Determine the minumum number of instances of self according
            # to global options.
            n = options.opt.instances
            umlName = str(self.__umlClass.name)
            for s in options.opt.specific_instances:
                (name, count) = s.split(':')
                if name == umlName:
                    n = int(count)
            # Add uninitialized objects until the minimum amount is
            # reached.
            while len(self.__objects) < n:
                self.addObject(None)

    def processFields(self):
        """Add all Fields to self.

        All classes must have been added to the model already.
        """
        self.__addAttributes()
        self.__addAssociations()

    def __addAttributes(self):
        """Find all attributes and add them as Fields."""
        for umlFeature in umlSort(self.__umlClass.feature):
            if umlFeature.oclIsKindOf(uml.Attribute):
                id_ = smtutils.generateId('a', umlFeature.name,
                                          [f.smvId() for f in self.__fields])
                f = AttributeField(umlFeature, id_, self)
                self.__fields.append(f)
                assert not umlFeature in self.__fieldForUmlAttribute
                self.__fieldForUmlAttribute[umlFeature] = f

    def __addAssociations(self):
        """Find all navigable associations and add them as Fields.

        All classes must have been added to the model already.
        """
        for umlThisEnd in umlSort(self.__umlClass.association):
            umlAssociation = umlThisEnd.__parent__()
            ends = [e for e in umlSort(umlAssociation.connection) if
                    e is not umlThisEnd and e.isNavigable]
            for umlEnd in ends:
                assert umlEnd.oclIsKindOf(uml.AssociationEnd)
                try:
                    endClass = self.__model.getClassForUml(umlEnd.participant)
                except exceptions.LookupError:
                    raise errors.InputError('dangling association end ' +
                                            str(umlEnd))
                id_ = smtutils.generateId('l', umlEnd.name,
                                          [f.smvId() for f in self.__fields])
                f = AssociationField(umlEnd, endClass.getType(), id_, self)
                self.__fields.append(f)
                assert not umlEnd in self.__fieldForUmlAssociationEnd
                self.__fieldForUmlAssociationEnd[umlEnd] = f

    def processInitialConfiguration(self):
        """Determine the field values for initially alive objects."""
        for obj in self.__objects:
            obj.processInitialConfiguration()

    def processActivities(self):
        """Generate the global effects of the transitions in this class."""
        for t in self.__machine.getTransitions():
            t.processActivity()

    def elementMap(self):#$$
        """Generate a mapping of the class to UML.

        Return value: mapping code as a list of strings.
        """
        classXmiId = str(self.__umlClass.__XMIid__())
        c = []
        for o in self.__objects:
            line = '--$Map Object ' + o.smvVariable() + ' ' + classXmiId
            x = o.getXmiId()
            if x:
                line += ' ' + x
            c += [line]
        for s in self.__machine.getVertices():
            c += ['--$Map StateVertex ' + classXmiId + ' ' +
                  s.getXmiId() + ' ' + s.smvId()]
        for t in self.__machine.getTransitions():
            c += ['--$Map Trans ' + classXmiId + ' ' +
                  t.getXmiId() + ' ' + t.smvId()]
            exprIds = t.getGuardChooseExprIds()
            chooseKeys = exprIds.keys()
            chooseKeys.sort()
            for k in chooseKeys:
                c += ['--$Map TransGuardChoose ' + classXmiId + ' ' +
                      t.getXmiId() + ' ' + exprIds[k] + '.value ' + k]
            exprIds = t.getEffectChooseExprIds()
            chooseKeys = exprIds.keys()
            chooseKeys.sort()
            for k in chooseKeys:
                c += ['--$Map TransEffectChoose ' + classXmiId + ' ' +
                      t.getXmiId() + ' ' + exprIds[k] + '.value ' + k]
        return c

    def smvId(self):
        """Return the id string."""
        return self.__id

    def smtAlive(self, obj):
        """Return the predicate Alive."""
        mc = self.__model.getCircuit()
        label = (0, obj, 'Alive')
        if label not in mc:
            mc[label] = mc.add_var()
        return mc[label]

    def smtNextAlive(self, obj):
        """Return the next-state predicate Alive."""
        mc = self.__model.getCircuit()
        label = (1, obj, 'Alive')
        if label not in mc:
            mc[label] = mc.add_var()
        return mc[label]

    def smtCreated(self, obj):
        """Return the predicate Created."""
        mc = self.__model.getCircuit()
        return mc.add_false()

    def smtDestroyed(self, obj):
        """Return the predicate Destroyed."""
        mc = self.__model.getCircuit()
        return mc.add_false()

    def smtThisRef(self, obj):
        """Return the function ThisRef."""
        mc = self.__model.getCircuit()
        label = (0, obj, 'ThisRef')
        if label not in mc:
            mc[label] = obj.smtReference(mc)
        return mc[label]

    def varSmtThisRef(self, circuit, obj):
        """Return a PySMT variable Gate for the function ThisRef."""
        label = (0, obj, 'ThisRef')
        if label not in circuit:
            circuit[label] = self.__type.smtVariable(circuit)
        return circuit[label]

    def encode(self, obj):
        """Generate an SMT encoding for an object of the class.

        Constraints for the initial configuration, transition
        relation, configuration specification, and transition
        specification are added to the model circuit.

        Param obj: an ActiveObject of class self.
        """
        obj.encode()
        self.__queue.encode(obj)
        self.__machine.encode(obj)

    def defineInitialConfiguration(self):
        """Instantiate the initial circuit of self in the model circuit."""
        mc = self.__model.getCircuit()
        # Define in the model circuit all gates that are referred to
        # in self.__initialCircuit.
        self.__model.defineFieldValues()
        mc.addLabeledCircuit(self.__initialCircuit)


class ActiveObject:
    """An active UML object."""

    def __init__(self, umlObject, id_, number, class_):
        """Construct an active object.

        Param umlObject: the underlying uml.Object, or None if this
        object is not initially alive.

        Param id_: an id string.

        Param number: the integer number of this object.

        Param class_: the ActiveClass of the object.
        """
        assert umlObject is None or umlObject.oclIsKindOf(uml.Object)
        self.__umlObject = umlObject
        self.__id = id_
        self.__number = number
        self.__class = class_

    def __str__(self):
        return self.__id

    def getClass(self):
        """Return the ActiveClass of self."""
        return self.__class

    def getXmiId(self):
        """Return the XMI id of the underlying uml.Object as a string.

        If there is no underlying uml.Object, return None.
        """
        if self.__umlObject:
            return str(self.__umlObject.__XMIid__())
        else:
            return None

    def processInitialConfiguration(self):
        """Determine the initial field values for self."""
        # smtConstraint maps Fields of the class of self to PySMT
        # Gates in the initial circuit that must evaluate to true in
        # the initial configuration.
        smtConstraint = {}
        initializer = {}  # map Field to Jumbala initializer
        ic = self.__class.getInitialCircuit()
        if self.__umlObject:
            for umlSlot in umlSort(self.__umlObject.slot):
                try:
                    f = self.__class.getFieldForUmlAttribute(umlSlot.attribute)
                except exceptions.LookupError:
                    raise errors.InputError('unknown attribute for slot ' +
                                            str(umlSlot))
                if f in initializer:
                    raise errors.InputError('attribute value ' + str(umlSlot) +
                                            ' initialized twice')
                # Treat the name of the slot's value as an action
                # language expression.
                initializer[f] = str(umlSlot.value.name)
        # Parse the Jumbala initializer for all fields that have one.
        for f in self.__class.getFields():
            if f in initializer:
                init = initializer[f]
            else:
                init = f.getDefaultInitializer()
            if init is not None:
                activity = activities.Activity(self.__class, ic,
                                               isConstant=True)
                (smtValue, smtError) = actions.processExpr(
                    init, f.getType(), activity)
                # smtError is not checked.
                if activity.getChooseMap():
                    raise errors.NotSupportedError(
                        'choose expressions in initializers not supported')
                assert not activity.getRuntimeErrors()
                assert not activity.getAssertions()
                smtConstraint[f] = ic.add_and(
                    ic.add_eq(f.varSmtFieldValue(ic, self), smtValue),
                    *activity.getConstraints())
        # Handle links between objects in the initial configuration.
        if self.__umlObject:
            for umlLinkEnd in umlSort(self.__umlObject.linkEnd):
                umlLink = umlLinkEnd.__parent__()
                for umlOtherEnd in umlSort(umlLink.connection):
                    if (umlOtherEnd is not umlLinkEnd and
                        umlOtherEnd.associationEnd.isNavigable):
                        # The uml.LinkEnd umlOtherEnd is an instance
                        # of a uml.AssociationEnd that is navigable
                        # from the class of self.  There should be a
                        # field corresponding the uml.AssociationEnd,
                        # and an object at umlOtherEnd.  Initialize
                        # the field with a reference to the object.
                        (f, obj) = self.__getLink(umlOtherEnd)
                        if f in smtConstraint:
                            if options.opt.sloppy_input:
                                print ('warning: link ' + str(umlOtherEnd) +
                                       ' initialized twice')
                            else:
                                raise errors.InputError('link ' +
                                                        str(umlOtherEnd) +
                                                        ' initialized twice')
                        smtConstraint[f] = ic.add_eq(
                            f.varSmtFieldValue(ic, self), obj.smtReference(ic))
        for f in self.__class.getFields():
            if f not in smtConstraint:
                # Uninitialized fields get the default initial value
                # of the field.
                smtConstraint[f] = ic.add_eq(f.varSmtFieldValue(ic, self),
                                             f.getType().smtDefaultValue(ic))
            ic[0, self, 'FieldInitialized', f] = smtConstraint[f]

    def __getLink(self, umlEnd):
        """Return a pair (field, object) for a given uml.LinkEnd."""
        umlAssoEnd = umlEnd.associationEnd
        try:
            f = self.__class.getFieldForUmlAssociationEnd(umlAssoEnd)
        except exceptions.LookupError:
            raise errors.InputError('unknown link end ' + str(umlEnd))
        try:
            obj = f.getType().getClass().getObjectForUml(umlEnd.instance)
        except exceptions.LookupError:
            raise errors.InputError('dangling link end ' + str(umlEnd))
        return (f, obj)

    def smvId(self):
        """Return the id string."""
        return self.__id

    def smtFieldInitialized(self, field):
        """Return the predicate FieldInitialized."""
        mc = self.__class.getModel().getCircuit()
        label = (0, self, 'FieldInitialized', field)
        if label not in mc:
            self.__class.defineInitialConfiguration()
        return mc[label]

    def smtReference(self, circuit):
        """Return a PySMT gate giving the instance number of self."""
        refValueCount = len(self.__class.getObjects()) + 1
        return circuit.addEnumConst(refValueCount, self.__number)

    def encode(self):
        """Generate constraints for object lifetime."""
        model = self.__class.getModel()
        mc = model.getCircuit()
        if self.__umlObject is None:
            initAlive = mc.add_false()
        else:
            initAlive = mc.add_true()
        model.addInit(mc.add_iff(self.__class.smtAlive(self), initAlive))
        model.addTrans(mc.add_iff(self.__class.smtNextAlive(self), mc.add_or(
            self.__class.smtCreated(self),
            mc.add_and(mc.add_not(self.__class.smtDestroyed(self)),
                       self.__class.smtAlive(self)))))
        for f in self.__class.getFields():
            model.addInit(self.smtFieldInitialized(f))


class Field:
    """A base class for fields.

    A field of a class represents either a uml.Attribute of the class
    or a uml.AssociationEnd navigable from the class.
    """

    def __init__(self, id_, owner):
        """Construct a field.

        Param id_: an id string.

        Param owner: the owning ActiveClass.
        """
        self.__id = id_
        self._owner = owner
        self._type = None  # self._type must be set in the subclass.

    def __str__(self):
        return self.__id

    def getOwner(self):
        """Return the owning ActiveClass."""
        return self._owner

    def getType(self):
        """Return the Type of the field."""
        return self._type

    def smvId(self):
        """Return the id string."""
        return self.__id

    def varSmtFieldValue(self, circuit, obj):
        """Return a variable Gate for the function FieldValue."""
        label = (0, obj, 'FieldValue', self)
        if label not in circuit:
            circuit[label] = self._type.smtVariable(circuit)
        return circuit[label]

    def smtFieldValue(self, obj):
        """Return the function FieldValue."""
        mc = self._owner.getModel().getCircuit()
        label = (0, obj, 'FieldValue', self)
        if label not in mc:
            mc[label] = self._type.smtVariable(mc)
        return mc[label]

    def smtNextFieldValue(self, obj):
        """Return the next-state function FieldValue."""
        mc = self._owner.getModel().getCircuit()
        label = (1, obj, 'FieldValue', self)
        if label not in mc:
            mc[label] = self._type.smtVariable(mc)
        return mc[label]

    def smtRetain(self, obj):
        """Return the predicate Retain."""
        mc = self._owner.getModel().getCircuit()
        label = (0, obj, 'Retain', self)
        if label not in mc:
            mc[label] = mc.add_eq(self.smtFieldValue(obj),
                                  self.smtNextFieldValue(obj))
        return mc[label]

    def getDefaultInitializer(self):  #pragma: no cover
        """Return a Jumbala expression for the initial value of the field.

        Return value; a Jumbala expression as a string or None if
        there is no default initializer.
        """
        raise exceptions.NotImplementedError('Pure virtual method')


class AttributeField(Field):
    """A field represented as an uml.Attribute."""

    def __init__(self, umlAttribute, id_, owner):
        """Construct a field from an attribute.

        Param umlAttribute: a uml.Attribute.

        Param id_: an id string.

        Param owner: the owning ActiveClass.
        """
        Field.__init__(self, id_, owner)
        self.__umlAttribute = umlAttribute
        try:
            self._type = datatypes.resolvePrimitiveType(umlAttribute.type)
        except exceptions.LookupError:
            raise errors.NotSupportedError('unsupported type for attribute ' +
                                           str(umlAttribute))
        for r in umlAttribute.multiplicity.range:
            if r.lower != 1 or r.upper != 1:
                raise errors.NotSupportedError('unsupported multiplicity ' +
                                               'for attribute ' +
                                               str(umlAttribute))

    def getDefaultInitializer(self):
        if self.__umlAttribute.initialValue.body:
            return str(self.__umlAttribute.initialValue.body)
        return None

    def getName(self):
        """Return the UML name as a string."""
        return str(self.__umlAttribute.name)


class AssociationField(Field):
    """A field represented in UML as a navigable association end."""

    def __init__(self, umlEnd, type_, id_, owner):
        """Construct a field from an association end.

        Param umlEnd: a uml.AssociationEnd.

        Param type_: the ReferenceType of the field.

        Param id_: an id string.

        Param owner: the owning ActiveClass.
        """
        Field.__init__(self, id_, owner)
        self.__umlEnd = umlEnd
        self._type = type_
        for r in umlEnd.multiplicity.range:
            if r.upper != 1:
                raise errors.NotSupportedError('unsupported multiplicity ' +
                                               'for association end ' +
                                               str(umlEnd))

    def getDefaultInitializer(self):
        return None

    def getName(self):
        """Return the UML name as a string."""
        return str(self.__umlEnd.name)
