#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# datatypes.py
#
# 2007-11-22  Jori Dubrovin:  Added real enums
# 2007-08-27  Jori Dubrovin:  Suboco 1.00
# 2006-12-04  Jori Dubrovin:  Using optparse for options
# 2006-07-30  Jori Dubrovin:  Initial version
#

"""Type definitions for SMT encoding."""

import exceptions

import errors
import options


class Type:
    """A base class for the type system."""

    def __init__(self):
        pass

    def isSubtypeOf(self, other):
        """Return True if self is a subtype of the other Type."""
        return (self is other)  # Default implementation

    def isCastableTo(self, other):
        """Return True if self is castable to the other Type."""
        if self.isSubtypeOf(other) or other.isSubtypeOf(self):
            return True
        return False

    def isReferenceType(self):
        """Return True if self is a reference type."""
        return False  # Default implementation

    def smtVariable(self, circuit):  #pragma: no cover
        """Return a PySMT variable Gate corresponding to self.

        Param circuit: the LabeledCircuit in which to instantiate the
        gate.

        Return value: a Gate.
        """
        raise exceptions.NotImplementedError('Pure virtual method')

    def smtDefaultValue(self, circuit):  #pragma: no cover
        """Return a PySMT Gate corresponding the default value of self.

        Param circuit: the LabeledCircuit in which to instantiate the
        gate.

        Return value: a constant Gate.
        """
        raise exceptions.NotImplementedError('Pure virtual method')


class _NullType(Type):
    """The null type."""

    def __init__(self):
        Type.__init__(self)

    def __str__(self):
        return '<null type>'

    def isSubtypeOf(self, other):
        return (self is other or other.isReferenceType())

    def smtVariable(self, circuit):
        return circuit.addEnumVar(1)

    def smtDefaultValue(self, circuit):
        return circuit.addEnumConst(1, 0)

nullType = _NullType()


class PrimitiveType(Type):
    """A base class for primitive types."""

    def __init__(self):
        Type.__init__(self)


class _IntType(PrimitiveType):
    """The Jumbala int type."""

    def __init__(self):
        PrimitiveType.__init__(self)

    def __str__(self):
        return 'int'

    def smtVariable(self, circuit):
        return circuit.add_sint_var(options.opt.int_bits)

    def smtDefaultValue(self, circuit):
        return self.smtValue(circuit, 0)

    def smtValue(self, circuit, value):
        """Return a PySMT Gate for a given Python value.

        Param circuit: the LabeledCircuit in which to instantiate the
        gate.

        Param value: an integer between -2**31 and 2**31-1.

        Return value: a constant Gate.
        """
        half = 1L << (options.opt.int_bits - 1)
        if value < -half or value >= half:
            raise errors.NotSupportedError('too narrow int to hold value ' +
                                           str(value))
        return circuit.add_sint_const(options.opt.int_bits, value)

intType = _IntType()


class _BoolType(PrimitiveType):
    """The Jumbala boolean type."""

    def __init__(self):
        PrimitiveType.__init__(self)

    def __str__(self):
        return 'boolean'

    def smtVariable(self, circuit):
        return circuit.add_var()

    def smtDefaultValue(self, circuit):
        return self.smtValue(circuit, False)

    def smtValue(self, circuit, value):
        """Return a PySMT Gate for a given Python value.

        Param circuit: the LabeledCircuit in which to instantiate the
        gate.

        Param value: False or True.

        Return value: a constant Gate.
        """
        if value:
            return circuit.add_true()
        else:
            return circuit.add_false()

boolType = _BoolType()


class ReferenceType(Type):
    """A reference type.

    The only reference types are references to active classes.  There
    is a one-to-one correspondence between ReferenceTypes and
    ActiveClasses.
    """

    def __init__(self, class_):
        """Construct a reference type.

        Param class_: the ActiveClass to which self refers.
        """
        Type.__init__(self)
        self.__class = class_

    def isReferenceType(self):
        return True

    def getClass(self):
        """Return the corresponding ActiveClass."""
        return self.__class

    def smtVariable(self, circuit):
        refValueCount = len(self.__class.getObjects()) + 1
        return circuit.addEnumVar(refValueCount)

    def smtDefaultValue(self, circuit):
        refValueCount = len(self.__class.getObjects()) + 1
        # null value is the zero word
        return circuit.addEnumConst(refValueCount, 0)


def resolvePrimitiveType(umlType):
    """Return a PrimitiveType corresponding to a given uml.Classifier."""
    if str(umlType.name).lower() in ['bool', 'boolean']:
        return boolType
    if str(umlType.name).lower() in ['int', 'integer']:
        return intType
    raise errors.InputError('unknown datatype ' + str(umlType))
