#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# events.py
#
# 2007-12-16  Jori Dubrovin:  Cleaned up options, changed defaults
# 2007-11-22  Jori Dubrovin:  Added real enums
# 2007-11-22  Jori Dubrovin:  Parameterized queue contents
# 2007-11-21  Jori Dubrovin:  Message tags
# 2007-11-20  Jori Dubrovin:  Encoding options for queues and states
# 2007-08-27  Jori Dubrovin:  Suboco 1.00
# 2007-05-04  Jori Dubrovin:  Simplified definitions
# 2007-03-08  Jori Dubrovin:  Fixed defer queue flush on implicit consumption
# 2006-12-04  Jori Dubrovin:  Fixed a signal type problem
# 2006-11-21  Jori Dubrovin:  Initial version
#

"""Encoding of UML events, event queues, and signals."""

import exceptions

import PySMT
import datatypes
import errors
import options
import smtutils
from coralutils import uml, umlSort

class EventQueue:
    """Representation of an event queue in an active UML class.

    There is a one-to-one correspondence between ActiveClass objects
    and EventQueue objects.
    """

    class Item:
        """An item in an element in the event queue.

        An item holds either a signal index or a signal parameter value.
        """

        def __str__(self):
            return self.id

    def __init__(self, class_):
        """Construct an event queue.

        Param class_: the associated ActiveClass.
        """
        self.__class = class_
        self.__receivedSignals = []  # Signals sent to this queue
        self.__triggerSignals = []   # Signals that trigger transitions
        self.__relevantSignals = []  # union of the above two
        self.__items = []  # Items for signal parameters and signal indices

    def getQueueSize(self):
        """Return the event queue size."""
        n = options.opt.queue_size
        umlName = self.__class.getName()
        for s in options.opt.specific_queue_size:
            (name, size) = s.split(':')
            if name == umlName:
                n = int(size)
        return n

    def getRelevantSignals(self):
        """Return the relevant Signals received as a list.

        A signal is relevant if it appears as a trigger in a
        transition in the associated state machine, or if there is
        globally at least one transition that sends the signal to the
        associated class.

        All relevant signals must have been added.
        """
        return self.__relevantSignals[:]

    def smtQueuePos(self, obj):
        """Return the function QueuePos."""
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'QueuePos')
        if label not in mc:
            mc[label] = self.__addPosVar(mc)
        return mc[label]

    def smtNextQueuePos(self, obj):
        """Return the next-state function QueuePos."""
        mc = self.__class.getModel().getCircuit()
        label = (1, obj, 'QueuePos')
        if label not in mc:
            mc[label] = self.__addPosVar(mc)
        return mc[label]

    def uifTagInterpreter(self, item):
        """Return an uninterpreted function that interprets tags.

        processReceptions must have been called.

        Return value: a PySMT uninterpreted function that takes an
        unbounded integer tag that corresponds to a message and
        returns the signal index or parameter value (depending on the
        given Item) carried by the message.
        """
        assert options.opt.queue_contents == 'tags'
        assert item in self.__receiveItems
        assert item is not self.__tagItem
        if item not in self.__tagUif:
            smtType = self.smtNoEvent(item).get_type()
            self.__tagUif[item] = PySMT.new_uif([PySMT.get_int()], smtType)
        return self.__tagUif[item]

    def uifQueuedEvent(self, obj, item):
        """Return an uninterpreted function that holds queued events.

        processReceptions must have been called.

        Return value: a PySMT uninterpreted function that takes a time
        step and a queue position and returns the message tag, signal
        index or parameter value (depending on the given Item) in the
        queue of the given object.  The position may be up to one
        greater than the last element in the queue.
        """
        assert options.opt.queue_indexing == 'parameter'
        assert item in self.__receiveItems
        if (obj, item) not in self.__queueUif:
            posType = self.smtQueuePos(obj).get_type()
            itemType = self.smtNoEvent(item).get_type()
            self.__queueUif[obj, item] = PySMT.new_uif(
                [PySMT.get_int(), posType], itemType)
        return self.__queueUif[obj, item]

    def smtNewEvent(self, obj, item):
        """Return the function NewEvent."""
        assert item in self.__receiveItems
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'NewEvent', item)
        if label not in mc:
            if (options.opt.queue_contents == 'tags' and
                item is not self.__tagItem):
                smtTag = self.smtNewEvent(obj, self.__tagItem)
                mc[label] = mc.add_uif(self.uifTagInterpreter(item), smtTag)
            else:
                mc[label] = item.smtVariable(mc)
        return mc[label]

    def smtCurrEvent(self, obj, item):
        """Return the function CurrEvent."""
        assert item in self.__items
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'CurrEvent', item)
        if label not in mc:
            size = self.getQueueSize()
            # Nondeterministic external signal parameters:
            # IVAR currEvent.param2#$extsig : ...
            value = self.smtNoEvent(item)
            if item in self.__receiveItems:
                if options.opt.queue_indexing == 'parameter':
                    value = self.smtQueuedEventVarPos(
                        obj, self.smtQueuePos(obj), item)
                else:
                    for i in xrange(size-1, -1, -1):
                        value = mc.add_ite(
                            mc.add_eq(self.smtQueuePos(obj),
                                      self.__addPosConst(mc, i)),
                            self.smtQueuedEvent(obj, i, item),
                            value)
            mc[label] = value
        return mc[label]

    def varSmtCurrEventParam(self, circuit, obj, signal, paramIndex):
        """Return a variable Gate for the function CurrEvent.

        addTriggerSignal must have been called for the given Signal.
        """
        for item in self.__items:
            if item.signal is signal and item.paramIndex == paramIndex:
                label = (0, obj, 'CurrEvent', item)
                if label not in circuit:
                    circuit[label] = item.smtVariable(circuit)
                return circuit[label]
        assert False

    def defineCurrEventParams(self, obj):
        """Instantiate CurrEvent Gates in the model circuit."""
        for item in self.__items:
            void = self.smtCurrEvent(obj, item)

    def smtQueuedEvent(self, obj, pos, item):
        """Return the function QueuedEvent."""
        assert item in self.__receiveItems
        mc = self.__class.getModel().getCircuit()
        if options.opt.queue_indexing == 'parameter':
            smtPos = self.__addPosConst(mc, pos)
            return self.smtQueuedEventVarPos(obj, smtPos, item)
        if options.opt.queue_contents == 'tags' and item is not self.__tagItem:
            smtTag = self.smtQueuedEvent(obj, pos, self.__tagItem)
            return mc.add_uif(self.uifTagInterpreter(item), smtTag)
        label = (0, obj, 'QueuedEvent', pos, item)
        if label not in mc:
            mc[label] = item.smtVariable(mc)
        return mc[label]

    def smtNextQueuedEvent(self, obj, pos, item):
        """Return the next-state function QueuedEvent."""
        assert item in self.__receiveItems
        mc = self.__class.getModel().getCircuit()
        if options.opt.queue_indexing == 'parameter':
            smtPos = self.__addPosConst(mc, pos)
            return self.smtNextQueuedEventVarPos(obj, smtPos, item)
        if options.opt.queue_contents == 'tags' and item is not self.__tagItem:
            smtTag = self.smtNextQueuedEvent(obj, pos, self.__tagItem)
            return mc.add_uif(self.uifTagInterpreter(item), smtTag)
        label = (1, obj, 'QueuedEvent', pos, item)
        if label not in mc:
            mc[label] = item.smtVariable(mc)
        return mc[label]

    def smtQueuedEventVarPos(self, obj, smtPos, item):
        """Return the function QueuedEvent for non-constant queue position."""
        assert options.opt.queue_indexing == 'parameter'
        assert item in self.__receiveItems
        mc = self.__class.getModel().getCircuit()
        if options.opt.queue_contents == 'tags' and item is not self.__tagItem:
            smtTag = self.smtQueuedEventVarPos(obj, smtPos, self.__tagItem)
            return mc.add_uif(self.uifTagInterpreter(item), smtTag)
        smtTime = self.__class.getModel().smtTime()
        return mc.add_uif(self.uifQueuedEvent(obj, item), smtTime, smtPos)

    def smtNextQueuedEventVarPos(self, obj, smtPos, item):
        """Return the function QueuedEvent for non-constant queue position."""
        assert options.opt.queue_indexing == 'parameter'
        assert item in self.__receiveItems
        mc = self.__class.getModel().getCircuit()
        if options.opt.queue_contents == 'tags' and item is not self.__tagItem:
            smtTag = self.smtQueuedEventVarPos(obj, smtPos, self.__tagItem)
            return mc.add_uif(self.uifTagInterpreter(item), smtTag)
        smtTime = self.__class.getModel().smtNextTime()
        return mc.add_uif(self.uifQueuedEvent(obj, item), smtTime, smtPos)

    def smtNoEvent(self, item):
        """Return an item of the empty event."""
        assert item in self.__items
        mc = self.__class.getModel().getCircuit()
        return item.smtDefaultValue(mc)

    def smtDispatch(self, obj, signal):
        """Return the predicate Dispatch.

        Defined for all relevant signals.
        """
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'Dispatch', signal)
        if label not in mc:
            if signal in self.__receivedSignals or signal.isExternal():
                mc[label] = mc.add_var()
            else:
                # This signal triggers transitions but cannot appear
                # in the event queue.
                mc[label] = mc.add_false()
        return mc[label]

    def smtDispatchSys(self, obj):
        """Return the predicate DispatchSys."""
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'DispatchSys')
        if label not in mc:
            mc[label] = mc.add_or(*[self.smtDispatch(obj, e) for
                                    e in self.__receivedSignals])
        return mc[label]

    def smtDispatchExt(self, obj):
        """Return the predicate DispatchExt."""
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'DispatchExt')
        if label not in mc:
            mc[label] = mc.addAtLeastOne(*[self.smtDispatch(obj, e) for
                                           e in self.__relevantSignals if
                                           e.isExternal()])
        return mc[label]

    def smtDispatchAny(self, obj):
        """Return the predicate DispatchAny."""
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'DispatchAny')
        if label not in mc:
            mc[label] = mc.addAtLeastOne(self.smtDispatchSys(obj),
                                         self.smtDispatchExt(obj))
        return mc[label]

    def smtDefer(self, obj, signal):
        """Return the predicate Defer.

        Defined for all received system signals.
        """
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'Defer', signal)
        if label not in mc:
            machine = self.__class.getStateMachine()
            mc[label] = mc.add_and(
                self.smtDispatch(obj, signal),
                mc.add_not(machine.smtAnyFeasible(obj, signal)),
                machine.smtAnyDeferring(obj, signal))
        return mc[label]

    def smtDeferSys(self, obj):
        """Return the predicate DeferSys."""
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'DeferSys')
        if label not in mc:
            mc[label] = mc.add_or(*[self.smtDefer(obj, e) for
                                    e in self.__receivedSignals])
        return mc[label]

    def smtEnabledSys(self, obj):
        """Return the predicate EnabledSys."""
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'EnabledSys')
        if label not in mc:
            machine = self.__class.getStateMachine()
            mc[label] = mc.add_or(
                *[mc.add_and(self.smtDispatch(obj, e),
                             machine.smtAnyFeasible(obj, e)) for
                  e in self.__receivedSignals])
        return mc[label]

    def smtImplicitConsumeSys(self, obj):
        """Return the predicate ImplicitConsumeSys."""
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'ImplicitConsumeSys')
        if label not in mc:
            mc[label] = mc.add_and(
                self.smtDequeue(obj), mc.add_not(self.smtEnabledSys(obj)))
        return mc[label]

    def smtDequeue(self, obj):
        """Return the predicate Dequeue."""
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'Dequeue')
        if label not in mc:
            mc[label] = mc.add_and(self.smtDispatchSys(obj),
                                   mc.add_not(self.smtDeferSys(obj)))
        return mc[label]

    def smtInputQueueEmpty(self, obj):
        """Return the predicate InputQueueEmpty."""
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'InputQueueEmpty')
        if label not in mc:
            if options.opt.queue_contents == 'tags':
                item = self.__tagItem
            else:
                item = self.__signalItem
            mc[label] = mc.add_eq(self.smtCurrEvent(obj, item),
                                  self.smtNoEvent(item))
        return mc[label]

    def smtQueueFull(self, obj):
        """Return the predicate QueueFull."""
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'QueueFull')
        if label not in mc:
            size = self.getQueueSize()
            if size > 0:
                if options.opt.queue_contents == 'tags':
                    item = self.__tagItem
                else:
                    item = self.__signalItem
                mc[label] = mc.add_neq(self.smtQueuedEvent(obj, size-1, item),
                                       self.smtNoEvent(item))
            else:
                mc[label] = mc.add_true()
        return mc[label]

    def smtNoReception(self, obj):
        """Return the predicate NoReception."""
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'NoReception')
        if label not in mc:
            if options.opt.queue_contents == 'tags':
                item = self.__tagItem
            else:
                item = self.__signalItem
            mc[label] = mc.add_eq(self.smtNewEvent(obj, item),
                                  self.smtNoEvent(item))
        return mc[label]

    def smtSignalIndex(self, signal):
        """Return an expression for the signal number in this queue.

        Defined for all received system signals.
        """
        mc = self.__class.getModel().getCircuit()
        sigCount = len(self.__receivedSignals) + 1
        return mc.addEnumConst(sigCount,
                               self.__receivedSignals.index(signal) + 1)

    def smtReception(self, obj, signal, smtParams):
        """Generate a PySMT Gate for signal reception.

        processReceptions must have been called.

        The returned gate evaluates to true if and only if the signal
        received by self is the given signal with the given parameter
        values.

        Param obj: the ActiveObject (whose class must be the class
        associated with self) receiving the signal.

        Param signal: the desired Signal.

        Param smtParams: a list Gates giving the values for signal
        parameters.

        Return value: a boolean PySMT Gate in the model circuit.
        """
        assert obj.getClass() is self.__class
        mc = self.__class.getModel().getCircuit()
        rec = []
        for item in self.__receiveItems:
            if item is self.__signalItem:
                rec.append(mc.add_eq(self.smtNewEvent(obj, item),
                                     self.smtSignalIndex(signal)))
            elif item.signal is signal:
                rec.append(mc.add_eq(self.smtNewEvent(obj, item),
                                     smtParams[item.paramIndex]))
        return mc.add_and(*rec)

    def addReceivedSignal(self, signal):
        """Register the reception of a signal from a send statement."""
        if signal not in self.__receivedSignals:
            self.__receivedSignals.append(signal)
            if signal not in self.__relevantSignals:
                self.__relevantSignals.append(signal)

    def addTriggerSignal(self, signal):
        """Register a trigger in the associated state machine."""
        if signal not in self.__triggerSignals:
            self.__triggerSignals.append(signal)
            if signal not in self.__relevantSignals:
                self.__relevantSignals.append(signal)
            # Add the Items corresponding to signal parameters.
            for (index, paramType) in enumerate(signal.getParamTypes()):
                item = EventQueue.Item()
                item.id = 'param' + str(index) + '#' + signal.smvId()
                item.smtVariable = paramType.smtVariable
                item.smtDefaultValue = paramType.smtDefaultValue
                item.signal = signal
                item.paramIndex = index
                self.__items.append(item)

    def processReceptions(self):
        """Process the signals that are sent to self.

        All received signals and trigger signals must have been added.
        """
        # Signal type must be able to hold any received signal plus
        # nosignal.
        sigCount = len(self.__receivedSignals) + 1
        self.__signalItem = EventQueue.Item()
        self.__signalItem.id = 'signal'
        self.__signalItem.smtVariable = (
            lambda circuit: circuit.addEnumVar(sigCount))
        self.__signalItem.smtDefaultValue = (
            lambda circuit: circuit.addEnumConst(sigCount, 0))
        self.__signalItem.signal = None
        self.__signalItem.paramIndex = None
        self.__items.append(self.__signalItem)
        self.__receiveItems = [item for item in self.__items if
                               item is self.__signalItem or
                               item.signal in self.__receivedSignals]
        if options.opt.queue_contents == 'tags':
            self.__tagItem = EventQueue.Item()
            self.__tagItem.id = 'tag'
            self.__tagItem.smtVariable = (
                lambda circuit: circuit.add_int_var())
            self.__tagItem.smtDefaultValue = (
                lambda circuit: circuit.add_int_const(0))
            self.__tagItem.signal = None
            self.__tagItem.paramIndex = None
            self.__items.append(self.__tagItem)
            self.__receiveItems.append(self.__tagItem)
            # self.__tagUif maps an Item to the PySMT uninterpreted
            # function that extracts the item value from a tag.
            self.__tagUif = {}
        if options.opt.queue_indexing == 'parameter':
            # self.__queueUif maps a pair (ActiveObject, Item) to the
            # PySMT uninterpreted function that extracts queued
            # events.
            self.__queueUif = {}
        if options.opt.qpos_type == 'integer':
            self.__addPosVar = lambda circuit: circuit.add_int_var()
            self.__addPosConst = lambda circuit, i: circuit.add_int_const(i)
        else:
            posBits = smtutils.log2ceil(self.getQueueSize() + 1)
            self.__addPosVar = lambda circuit: circuit.add_uint_var(posBits)
            self.__addPosConst = (
                lambda circuit, i: circuit.add_uint_const(posBits, i))

    def encode(self, obj):
        """Encode queue and events."""
        self.__encodeQueue(obj)
        self.__encodeEvents(obj)
        self.__encodeDispatching(obj)

    def __encodeQueue(self, obj):
        """Encode queue logic."""
        model = self.__class.getModel()
        mc = model.getCircuit()
        size = self.getQueueSize()
        slotFree = {-1: mc.add_false(), size: mc.add_true()}
        if options.opt.queue_contents == 'tags':
            queueItems = [self.__tagItem]
            item = self.__tagItem
        else:
            queueItems = self.__receiveItems
            item = self.__signalItem
        for i in xrange(size):
            slotFree[i] = mc.add_eq(self.smtQueuedEvent(obj, i, item),
                                    self.smtNoEvent(item))
        for i in xrange(size):
            # behindFree is true iff the element behind this one is
            # empty (true if this is the last element).  newHere is
            # true iff this element should receive the new event (or
            # noEvent if there is no new event).  remove is true iff
            # an event is being consumed from the queue from a
            # position that is not behind this element.
            newHere = mc.add_and(mc.add_not(slotFree[i-1]), slotFree[i+1],
                                 mc.add_iff(self.smtDequeue(obj),
                                            mc.add_not(slotFree[i])))
            remove = mc.add_and(self.smtDequeue(obj),
                                mc.add_le(self.smtQueuePos(obj),
                                          self.__addPosConst(mc, i)))
            for item in queueItems:
                model.addInit(mc.add_eq(self.smtQueuedEvent(obj, i, item),
                                        self.smtNoEvent(item)))
                if i == size-1:
                    behindItem = self.smtNoEvent(item)
                else:
                    behindItem = self.smtQueuedEvent(obj, i+1, item)
                itemNext = self.smtQueuedEvent(obj, i, item)
                itemNext = mc.add_ite(remove, behindItem, itemNext)
                itemNext = mc.add_ite(newHere,
                                      self.smtNewEvent(obj, item), itemNext)
                itemNext = mc.add_ite(self.__class.smtDestroyed(obj),
                                      self.smtNoEvent(item), itemNext)
                model.addTrans(mc.add_eq(
                    self.smtNextQueuedEvent(obj, i, item), itemNext))
        if options.opt.queue_indexing == 'parameter':
            # Fix the "last+1"st element in the queue to noevent.
            for item in queueItems:
                model.addInit(mc.add_eq(self.smtQueuedEvent(obj, size, item),
                                        self.smtNoEvent(item)))
                model.addTrans(mc.add_eq(self.smtNextQueuedEvent(obj, size,
                                                                 item),
                                         self.smtNoEvent(item)))
        # QueuePos points to the front element of the input queue in
        # the event queue.  The elements whose index is less than
        # QueuePos contain a deferred event.  The input queue consists
        # of the elements whose index is greater or equal to QueuePos.
        # There are size+1 distinct values for QueuePos.
        model.addInit(mc.add_eq(self.smtQueuePos(obj),
                                self.__addPosConst(mc, 0)))
        posNext = self.smtQueuePos(obj)
        posNext = mc.add_ite(self.smtDeferSys(obj),
                             mc.add_plus(self.smtQueuePos(obj),
                                         self.__addPosConst(mc, 1)),
                             posNext)
        posNext = mc.add_ite(mc.add_or(self.smtEnabledSys(obj),
                                       self.smtDispatchExt(obj),
                                       self.__class.smtDestroyed(obj)),
                             self.__addPosConst(mc, 0),
                             posNext)
        model.addTrans(mc.add_eq(self.smtNextQueuePos(obj), posNext))
        # Explicit limits for QueuePos:
        #  model.addTrans(mc.add_ge(self.smtNextQueuePos(obj),
        #                           self.__addPosConst(mc, 0)))
        #  model.addTrans(mc.add_le(self.smtNextQueuePos(obj),
        #                           self.__addPosConst(mc, size)))
        # Forbid queue overflow.
        model.addTrans(mc.add_or(mc.add_not(self.smtQueueFull(obj)),
                                 self.smtNoReception(obj),
                                 self.smtDequeue(obj)))

    def __encodeEvents(self, obj):
        """Encode local event handling."""
        model = self.__class.getModel()
        mc = model.getCircuit()
        if options.opt.queue_contents == 'tags':
            smtTag = self.smtNoEvent(self.__tagItem)
            for item in self.__receiveItems:
                if item is not self.__tagItem:
                    model.addInit(mc.add_eq(
                        mc.add_uif(self.uifTagInterpreter(item), smtTag),
                        self.smtNoEvent(item)))
        else:
            for item in self.__receiveItems:
                if item.signal is not None:
                    # Clear all non-applicable signal parameter values in
                    # newEvent.
                    model.addTransImpl(
                        mc.add_neq(self.smtNewEvent(obj, self.__signalItem),
                                   self.smtSignalIndex(item.signal)),
                        mc.add_eq(self.smtNewEvent(obj, item),
                                  self.smtNoEvent(item)))

    def __encodeDispatching(self, obj):
        """Encode event dispatching from queue."""
        model = self.__class.getModel()
        mc = model.getCircuit()
        machine = self.__class.getStateMachine()
        for signal in self.__receivedSignals:
            isCurr = mc.add_eq(self.smtCurrEvent(obj, self.__signalItem),
                               self.smtSignalIndex(signal))
            model.addTransImpl(self.smtDispatch(obj, signal), isCurr)
        for signal in self.__relevantSignals:
            if signal.isExternal():
                # An external signal can be dispatched only if it
                # fires a transition.
                model.addTransImpl(self.smtDispatch(obj, signal),
                                   machine.smtAnyFeasible(obj, signal))
        if options.opt.check_implicit_consumption:
            model.addTransSpec(mc.add_not(self.smtImplicitConsumeSys(obj)))
        model.addTrans(mc.addAtMostOne(*[self.smtDispatch(obj, e) for
                                         e in self.__relevantSignals if
                                         e.isExternal()]))
        model.addTrans(mc.addAtMostOne(self.smtDispatchSys(obj),
                                       self.smtDispatchExt(obj)))
        model.addTrans(mc.addAtMostOne(self.smtDispatchAny(obj),
                                       machine.smtInRtc(obj)))


class Signal:
    """Representation of a UML signal."""

    def __init__(self, umlSignal, id_, model):
        """Construct a signal.

        Param umlSignal: the underlying uml.Signal.

        Param id_: an identification string.

        Param model: the Model owning the signal.
        """
        assert umlSignal.oclIsKindOf(uml.Signal)
        self.__umlSignal = umlSignal
        self.__id = id_
        self.__model = model

    def __str__(self):
        return self.getName()

    def getName(self):
        """Return the name of the signal as a string."""
        return str(self.__umlSignal.name)

    def isExternal(self):
        """Return True if self is an external (ATO) signal."""
        return self.getName().startswith('$')

    def isInUmlSignals(self, umlSignals):
        """Return True if self is in a given list of uml.Signals."""
        return (self.__umlSignal in umlSignals)

    def getParamTypes(self):
        """Return a list of Types of signal parameters."""
        return self.__paramTypes[:]

    def smvId(self):
        """Return the id string."""
        return self.__id

    def processParams(self):
        """Add all signal parameters to self.

        All ActiveClasses must already be added to the model.
        """
        self.__paramTypes = []
        # umlSort must not be used to read Signal.feature.
        for umlFeature in self.__umlSignal.feature:
            if umlFeature.oclIsKindOf(uml.Attribute):
                self.__paramTypes.append(self.__resolveParamType(umlFeature))

    def __resolveParamType(self, umlAttribute):
        """Return the Type of a parameter represented as a uml.Attribute."""
        try:
            type_ = self.__model.getClassForUml(umlAttribute.type).getType()
        except exceptions.LookupError:
            try:
                type_ = datatypes.resolvePrimitiveType(umlAttribute.type)
            except exceptions.LookupError:
                raise errors.NotSupportedError('unsupported type for ' +
                                               'signal attribute ' +
                                               str(umlAttribute))
        for r in umlAttribute.multiplicity.range:
            if r.upper != 1:
                raise errors.NotSupportedError('unsupported multiplicity ' +
                                               'for signal attribute ' +
                                               str(umlAttribute))
        return type_
