#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# models.py
#
# 2007-11-22  Jori Dubrovin:  Parameterized queue contents
# 2007-09-01  Jori Dubrovin:  Suboco 1.02
# 2007-08-27  Jori Dubrovin:  Suboco 1.00
# 2007-06-20  Jori Dubrovin:  New choose expression mapping
# 2007-04-01  Jori Dubrovin:  Strictly static queue-full predicate
# 2007-03-08  Jori Dubrovin:  Adding states to element map
# 2006-12-05  Jori Dubrovin:  Transition instance logging
# 2006-11-03  Jori Dubrovin:  Initial version
#

"""PySMT encoding of an entire UML model.

Pass 1.  Construct the objects corresponding to UML classifiers and
state machine elements.

Pass 2.  Process the model (methods named process*), add objects,
construct the initial circuit (ic) for each class and the activity
circuit (ac) for each state machine (methods named varSmt*).  The
initial circuit of a class contains the encoding of Jumbala
initializer for each field of each object of that class.  The activity
circuit of a state machine contains the encoding of Jumbala guard and
effect for each transition in that state machine.

Pass 3.  Make the model circuit (methods named encode* and smt*).  The
model circuit (mc) contains the transition relation, the initial
condition, the configuration specification (state invariant to be
checked) and the transition specification (an invariant to be checked
that depends on input functions).  Labels in the model circuit have
the form (time, object, name, ...), where time is 0 for state
functions, input functions, and defined functions, and time is 1 for
next-state functions.  Instantiate the initial circuit for each class
and the activity circuit for each object (methods named define*).

Pass 4.  Unroll the transition relation by instatiating the model
circuit for time steps 0 to bound.  Constrain the initial condition to
hold at time 0, the transition relation to hold at time 0 to bound-1,
and either the configuration specification not to hold at time bound
or the transition specification not to hold at time bound-1.
"""

import exceptions

import actions
import classes
import events
import errors
import options
import smtutils
from coralutils import uml, umlSort


class Model:
    """A behavioral UML model."""

    def __init__(self, umlModel):
        """Construct a model.

        Param umlModel: the underlying uml.Model.
        """
        assert umlModel.oclIsKindOf(uml.Model)
        self.__classes = []           # all ActiveClasses in self
        self.__classForUml = {}       # map uml.Class to ActiveClass
        self.__signals = []           # all Signals in self
        self.__signalForUml = {}      # map uml.Signal to Signal
        self.__addNestedClassifiers(umlModel)
        self.__addNestedObjects(umlModel)
        self.__process()
        self.__encode()

    def getClassForUml(self, umlClassifier):
        """Return the ActiveClass corresponding to the given uml.Classifier.

        If there is no such class, raise a LookupError.
        """
        return self.__classForUml[umlClassifier]

    def getSignalForName(self, signalName):
        """Return the Signal with the given name.

        If there are no such Signals, raise a LookupError.  If there
        are several, raise an InputError.
        """
        signals = [e for e in self.__signals if e.getName() == signalName]
        if len(signals) > 1:
            raise errors.InputError("ambiguous signal name '" +
                                    signalName + "'")
        return signals[0]

    def getSignalForUml(self, umlSignal):
        """Return the global Signal for a given uml.Signal.

        If there are no such Signals, raise a LookupError.
        """
        return self.__signalForUml[umlSignal]

    def getCircuit(self):
        """Return the model circuit."""
        return self.__circuit

    def __addNestedClassifiers(self, umlElem):
        """Add all classes and signals in the given uml.ModelElement."""
        if umlElem.oclIsKindOf(uml.Class):
            self.__addClass(umlElem)
        if umlElem.oclIsKindOf(uml.Signal):
            self.__addSignal(umlElem)
        if umlElem.oclIsKindOf(uml.Namespace):
            for e in umlSort(umlElem.ownedElement):
                self.__addNestedClassifiers(e)

    def __addClass(self, umlClass):
        """Add a new ActiveClass for the given uml.Class to self."""
        if not umlClass.isActive:
            if options.opt.sloppy_input:
                print ('warning: treating passive class ' + str(umlClass) +
                       ' as active')
            else:
                raise errors.NotSupportedError('passive class ' +
                                               str(umlClass) +
                                               ' not supported')
        id_ = smtutils.generateId(
            'c', umlClass.name, [class_.smvId() for class_ in self.__classes])
        class_ = classes.ActiveClass(umlClass, id_, self)
        self.__classes.append(class_)
        assert not umlClass in self.__classForUml
        self.__classForUml[umlClass] = class_

    def __addSignal(self, umlSignal):
        """Add a new Signal for the given uml.Signal to self."""
        id_ = smtutils.generateId(
            'e', umlSignal.name, [signal.smvId() for signal in self.__signals])
        signal = events.Signal(umlSignal, id_, self)
        self.__signals.append(signal)
        assert not umlSignal in self.__signalForUml
        self.__signalForUml[umlSignal] = signal

    def __addNestedObjects(self, umlElem):
        """Add all objects in the given uml.ModelElement."""
        if umlElem.oclIsKindOf(uml.Object):
            self.__addUmlObject(umlElem)
        if umlElem.oclIsKindOf(uml.Namespace):
            for e in umlSort(umlElem.ownedElement):
                self.__addNestedObjects(e)
        if umlElem.oclIsKindOf(uml.Instance):
            for e in umlSort(umlElem.ownedInstance):
                self.__addNestedObjects(e)

    def __addUmlObject(self, umlObject):
        """Add a new ActiveObject for the given uml.Object to self.

        All classes must have already been added.
        """
        umlClassifier = umlObject.classifier[0]
        try:
            class_ = self.getClassForUml(umlClassifier)
        except exceptions.LookupError:
            raise errors.InputError('unknown class for object ' +
                                    str(umlObject))
        class_.addObject(umlObject)

    def __process(self):
        """Process the model to prepare for creating the model circuit."""
        # processObjects must precede most other calls because it
        # determines the domains of reference types.
        # processActivities must precede processReceptions because it
        # registers the signals received by classes.  processFields
        # must precede processInitialConfiguration and
        # processActivities because it adds the fields to classes.
        for class_ in self.__classes:
            class_.processObjects()
        for signal in self.__signals:
            signal.processParams()
        for class_ in self.__classes:
            class_.processFields()
        for class_ in self.__classes:
            class_.processInitialConfiguration()
        for class_ in self.__classes:
            class_.processActivities()
        for class_ in self.__classes:
            class_.getEventQueue().processReceptions()

    def __encode(self):
        """Encode the model and create the model circuit."""
        self.__circuit = smtutils.LabeledCircuit()
        mc = self.__circuit
        self.__initConstraints = []
        self.__transConstraints = []
        self.__confSpecs = []
        self.__transSpecs = []
        # Define FieldValue for all (ActiveObject, Field)-pairs.  This
        # could be postponed to defineFieldValues.
        for class_ in self.__classes:
            for obj in class_.getObjects():
                for f in class_.getFields():
                    void = f.smtFieldValue(obj)
        for class_ in self.__classes:
            for obj in class_.getObjects():
                class_.encode(obj)
        self.__transitionInstances = self.__orderTransitionInstances()
        self.__encodeScheduling()
        self.__encodeSteps()
        self.__defineTraceFunctions()
        mc[0, None, 'Init'] = mc.add_and(*self.__initConstraints)
        mc[0, None, 'Trans'] = mc.add_and(*self.__transConstraints)
        mc[0, None, 'ConfSpec'] = mc.add_and(*self.__confSpecs)
        mc[0, None, 'TransSpec'] = mc.add_and(*self.__transSpecs)

    def __defineTraceFunctions(self):
        """Define the functions used in extractTrace."""
        for class_ in self.__classes:
            machine = class_.getStateMachine()
            queue = class_.getEventQueue()
            for obj in class_.getObjects():
                void = queue.smtDeferSys(obj)
                void = queue.smtImplicitConsumeSys(obj)
            for s in machine.getVertices():
                if s.isCompletionSensitive():
                    for obj in class_.getObjects():
                        void = s.smtQuiesce(obj)
        for (obj, t) in self.__transitionInstances:
            void = t.smtFire(obj)
            for key in t.getGuardChooseKeys():
                void = t.smtGuardChoose(obj, key)
            for key in t.getEffectChooseKeys():
                void = t.smtEffectChoose(obj, key)

    def defineFieldValues(self):
        """Define all FieldValue functions (not next-state)."""
        # Already done in self.__encode().
        pass

    def smtTime(self):
        """Return a variable Gate that represents the current time step."""
        mc = self.__circuit
        label = (0, None, 'Time')
        if label not in mc:
            mc[label] = mc.add_int_var()
        return mc[label]

    def smtNextTime(self):
        """Return a variable Gate that represents the next time step."""
        mc = self.__circuit
        label = (1, None, 'Time')
        if label not in mc:
            mc[label] = mc.add_int_var()
        return mc[label]

    def smtDeadlock(self):
        """Return the predicate Deadlock."""
        mc = self.__circuit
        label = (0, None, 'Deadlock')
        if label not in mc:
            mc[label] = mc.add_and(
                *[mc.add_not(class_.getStateMachine().smtReady(obj)) for
                  class_ in self.__classes for obj in class_.getObjects()])
        return mc[label]

    def addInit(self, smtConstraint):
        """Add a constraint for the initial configuration.

        Param smtConstraint: a boolean PySMT Gate in the model circuit
        that must evaluate to true in the initial configuration.
        """
        self.__initConstraints.append(smtConstraint)

    def addTrans(self, smtConstraint):
        """Add a constraint to the transition relation.

        Param smtConstraint: a boolean PySMT Gate in the model circuit that
        must evaluate to true in a valid step.
        """
        self.__transConstraints.append(smtConstraint)

    def addTransImpl(self, smtLhs, smtRhs):
        """Add a constraint of the form a -> b to the transition relation.

        Param smtLhs, smtRhs: boolean PySMT Gates in the model
        circuit.  The implication smtLhs -> smtRhs must be true in a
        valid step.
        """
        self.addTrans(self.__circuit.add_implies(smtLhs, smtRhs))

    def addTransSpec(self, smtSpec):
        """Add a transition specification.

        Param smtSpec: a boolean PySMT Gate in the model circuit that
        evaluates to false in a step that violates the specification.
        """
        self.__transSpecs.append(smtSpec)

    def addConfSpec(self, smtSpec):
        """Add a configuration specification.

        Param smtSpec: a boolean PySMT Gate in the model circuit that
        evaluates to false in a configuration that violates the
        specification.  The gate must not refer to any next-state
        functions.
        """
        self.__confSpecs.append(smtSpec)

    def __encodeScheduling(self):
        """
        """
        mc = self.__circuit
        sched = [class_.getStateMachine().smtScheduled(obj) for
                 class_ in self.__classes for obj in class_.getObjects()]
        if options.opt.encode_interleaving:
            self.addTrans(mc.addAtMostOne(*sched))
        notIdle = mc.addAtLeastOne(*sched)
        self.addTransImpl(mc.add_not(self.smtDeadlock()), notIdle)
        if options.opt.check_deadlock:
            self.addConfSpec(mc.add_not(self.smtDeadlock()))

    def __encodeSteps(self):
        """
        """
        written = {}  # map (object, field) to predicate WriteUntil
        sentTo = {}   # map object to predicate SendUntil
        mc = self.__circuit
        for class_ in self.__classes:
            for obj in class_.getObjects():
                for f in class_.getFields():
                    written[obj, f] = mc.add_false()
                sentTo[obj] = mc.add_false()
        for (actor, t) in self.__transitionInstances:
            class_ = actor.getClass()
            if not options.opt.encode_interleaving:
                for read in t.getActivity().getFieldReads():
                    f = read.getField()
                    for target in f.getOwner().getObjects():
                        readTarget = t.smtRead(actor, target, f)
                        self.addTransImpl(written[target, f],
                                          mc.add_not(readTarget))
            for write in t.getActivity().getFieldWrites():
                f = write.getField()
                for target in f.getOwner().getObjects():
                    writeTarget = t.smtWrite(actor, target, f)
                    written[target, f] = mc.add_or(written[target, f],
                                                   writeTarget)
            if (not options.opt.encode_interleaving and
                not options.opt.allow_queue_overdraft):
                if (not t.isCompletion() and
                    not t.getTriggerSignal().isExternal()):
                    # Forbid firing a transition that reads the event
                    # queue if the queue may have been written to.
                    if options.opt.encode_static:
                        full = mc.add_true()  # static overapproximation
                    else:
                        full = class_.getEventQueue().smtQueueFull(actor)
                    self.addTransImpl(mc.add_and(sentTo[actor], full),
                                      mc.add_not(t.smtFire(actor)))
            for send in t.getActivity().getSends():
                for target in send.getTargetClass().getObjects():
                    sendTarget = t.smtSend(actor, target)
                    if not options.opt.encode_interleaving:
                        self.addTransImpl(sentTo[target],
                                          mc.add_not(sendTarget))
                    sentTo[target] = mc.add_or(sentTo[target], sendTarget)
        # Add frame constraints
        for class_ in self.__classes:
            for obj in class_.getObjects():
                for f in class_.getFields():
                    self.addTransImpl(mc.add_not(written[obj, f]),
                                      f.smtRetain(obj))
                self.addTransImpl(mc.add_not(sentTo[obj]),
                                  class_.getEventQueue().smtNoReception(obj))

    def __orderTransitionInstances(self):
        """Generate a total order of all transition instances.

        Return value: a list of all possible pairs (obj, t), where obj
        is an ActiveObject and t is a Transition in the state machine
        of the class of the object.
        """
        transitionInstances = []
        for class_ in self.__classes:
            objects = class_.getObjects()
            for t in class_.getStateMachine().getTransitions():
                transitionInstances += zip(objects, [t]*len(objects))
        return transitionInstances

    def elementMap(self, modelFilename):#$$
        """Generate a mapping of model elements to UML.

        The mapping matches variables of the NuSMV input file with
        elements of the original UML model.  The mapping is coded as
        text lines that NuSMV sees as comments, so they can be
        embedded in the NuSMV input file.

        Param modelFilename: path to the XMI model file as a string.

        Return value: mapping code as a list of strings.
        """
        # Element map format:
        # --$Map Model <model file name>
        # --$Map Object <NuSMV variable name for the object>
        #         <XMI id of the class of the object>
        #         [XMI id of the object if it is in initial configuration]
        # ...
        # --$Map StateVertex <XMI id of the class containing the vertex>
        #         <XMI id of the vertex>
        #         <NuSMV id of the vertex>
        # --$Map Trans <XMI id of the class containing the transition>
        #         <XMI id of the transition>
        #         <NuSMV id of the transition>
        # --$Map TransGuardChoose
        #         <XMI id of the class containing the transition>
        #         <NuSMV id of the transition>
        #         <NuSMV variable name for a choose expression in guard>
        #         <identifying integer for the choose expression>
        # --$Map TransEffectChoose
        #         <XMI id of the class containing the transition>
        #         <NuSMV id of the transition>
        #         <NuSMV variable name for a choose expression in effect>
        #         <identifying integer for the choose expression>
        # ...
        # --$Map TransInstance <NuSMV variable name for the object>
        #         <NuSMV id of the transition>
        #
        # The transition instances are in the static E-step firing order.
        c = ['',
             '',
             '-'*70,
             '-- Mapping for tracing NuSMV variables back to UML elements',
             '-'*70,
             '',
             '--$Map Model ' + modelFilename]
        for class_ in self.__classes:
            c += class_.elementMap()
        for (actor, t) in self.__transitionInstances:
            c += ['--$Map TransInstance ' + actor.smvVariable() + ' ' +
                  t.smvId()]
        c += ['--$Map End']
        return c

    def unroll(self, bound):
        """Generate a BMC problem with the given bound.

        Return value: a PySMT Circuit.
        """
        problem = smtutils.LabeledCircuit()
        for k in xrange(bound+2):
            problem[k, None, 'Time'] = problem.add_int_const(k)
        for k in xrange(bound+1):
            def mapLabel(label):
                label = list(label)
                label[0] += k
                return tuple(label)
            problem.addLabeledCircuit(self.__circuit, mapLabel)
        problem.constrain(problem[0, None, 'Init'], True)
        for k in xrange(bound):
            problem.constrain(problem[k, None, 'Trans'], True)
        bad = problem.add_not(problem[bound, None, 'ConfSpec'])
        if bound > 0:
            bad = problem.add_or(
                bad, problem.add_not(problem[bound-1, None, 'TransSpec']))
        problem.constrain(bad, True)
        return problem

    def extractTrace(self, modelFilename, problem):
        """Convert a satisfiable circuit to a generic SMUML trace.

        Param modelFilename: path and file name of the model as a string.

        Param problem: a LabeledCircuit representing a BMC problem for
        which the latest solve() has returned 'sat'.

        Return value: generic trace as a string.
        """
        csStateInstances = []  # all pairs (object, completion sensitive state)
        objects = []  # all objects
        for class_ in self.__classes:
            classObjs = class_.getObjects()
            objects += classObjs
            for s in class_.getStateMachine().getVertices():
                if s.isCompletionSensitive():
                    csStateInstances += zip(classObjs, [s]*len(classObjs))
        trace = ['Model ' + modelFilename]
        trace += ['#' + line for line in options.comment]
        for obj in objects:
            if obj.getXmiId():
                trace.append('InitialConfiguration object ' + obj.smvId() +
                             ' umlobject ' + obj.getXmiId())
        k = 0
        while (k+1, None, 'Trans') in problem:  # While k < max time
            trace += ['# Time ' + str(k)]
            for (obj, t) in self.__transitionInstances:
                for key in t.getGuardChooseKeys():
                    value = problem.get_value(
                        problem[k, obj, 'GuardChoose', t, key])
                    trace.append('GuardChooseValue' +
                                 ' object ' + obj.smvId() +
                                 ' transition ' + t.getXmiId() +
                                 ' index ' + key +
                                 ' value ' + str(int(value)))
            for obj in objects:
                if problem.get_value(problem[k, obj, 'DeferSys']):
                    trace.append('EventDefer object ' + obj.smvId())
                if problem.get_value(problem[k, obj, 'ImplicitConsumeSys']):
                    trace.append('ImplicitConsumption object ' + obj.smvId())
            for (obj, s) in csStateInstances:
                if problem.get_value(problem[k, obj, 'Quiesce', s]):
                    # Quiescing not supported in traces
                    trace.append('# Quiesce object ' + obj.smvId() +
                                  ' state ' + s.getXmiId())
            for (obj, t) in self.__transitionInstances:
                if problem.get_value(problem[k, obj, 'Fire', t]):
                    for key in t.getEffectChooseKeys():
                        value = problem.get_value(
                            problem[k, obj, 'EffectChoose', t, key])
                        trace.append('EffectChooseValue' +
                                     ' object ' + obj.smvId() +
                                     ' transition ' + t.getXmiId() +
                                     ' index ' + key +
                                     ' value ' + str(int(value)))
                    trace.append('TransitionExecution object ' +
                                 obj.smvId() + ' transition ' + t.getXmiId())
            k += 1
        trace.append('')  # newline at the end
        return '\n'.join(trace)
