#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# options.py
#
# 2007-12-16  Jori Dubrovin:  Cleaned up options, changed defaults
# 2007-11-22  Jori Dubrovin:  Added real enums
# 2007-11-22  Jori Dubrovin:  Parameterized queue contents
# 2007-11-21  Jori Dubrovin:  Message tags
# 2007-11-20  Jori Dubrovin:  Encoding options for queues and states
# 2007-09-05  Jori Dubrovin:  Suboco 1.03
# 2007-09-01  Jori Dubrovin:  Suboco 1.02
# 2007-08-27  Jori Dubrovin:  Suboco 1.00
# 2007-05-04  Jori Dubrovin:  Added linear Enter predicate
# 2007-02-05  Jori Dubrovin:  Version 1.00
# 2006-12-04  Jori Dubrovin:  Using optparse for options
# 2006-08-31  Jori Dubrovin:  Added encodeStateExclusion
# 2006-08-16  Jori Dubrovin:  Added checks for run-time errors
# 2006-07-30  Jori Dubrovin:  Initial version
#

"""Global options."""

import time
import optparse


opt = None
comment = None

def initializeOptions(version):
    """Read command line and initialize the options structure.

    Param version: Suboco version number as a string.

    Return value: infilename.
    """
    p = optparse.OptionParser(usage='python %prog [options] infile.xmi',
                              version='Suboco ' + version)
    p.add_option('--sloppy-input', action='store_true', default=False,
                 help='accept inconsistent UML models')
    p.add_option('--check-deadlock', action='store_true', default=False,
                 help='check deadlocks')
    p.add_option('--check-implicit-consumption', action='store_true',
                 default=False,
                 help='check implicit consumption of events')
    p.add_option('--check-assertions', action='store_true', default=False,
                 help='check Jumbala assertion errors')
    p.add_option('--check-runtime-errors', action='store_true', default=False,
                 help='check Jumbala run-time errors')
    p.add_option('--trace-file', type='string', default=None, metavar='FILE',
                 help='trace output file name [default: "infile.trace"]')
    p.add_option('--dot-file', type='string', default=None, metavar='FILE',
                 help='dot output file name [default: <none>]')
    p.add_option('--yices', type='string', default='yices', metavar='FILE',
                 help='path to Yices solver binary [default: "yices"]')
    p.add_option('--cvc3', type='string', default=None, metavar='FILE',
                 help='path to CVC3 solver binary [default: <none>]')
    p.add_option('--mathsat', type='string', default=None, metavar='FILE',
                 help='path to MathSAT solver binary [default: <none>]')
    p.add_option('--stp', type='string', default=None, metavar='FILE',
                 help='path to STP solver binary [default: <none>]')
    p.add_option('--z3', type='string', default=None, metavar='FILE',
                 help='path to Z3 solver binary [default: <none>]')
    p.add_option('--format', type='string', default='native', metavar='FORMAT',
                 help='solver format: "native" or "smtlib" [default: native]')
    #p.add_option('--normalize-circuit', action='store_true', default=False,
    #             help='restrict the use of PySMT gates (not recommended)')
    p.add_option('--min-bound', type='int', default=0, metavar='N',
                 help='minimum BMC problem bound [default: 0]')
    p.add_option('--max-bound', type='int', default=999999, metavar='N',
                 help='maximum BMC problem bound [default: 999999]')
    p.add_option('--int-bits', type='int', default=32, metavar='N',
                 help='number of bits in the int type [default: 32]')
    # Not used
    #p.add_option('--instances', type='int', default=0, metavar='N',
    #             help='number of instances per class [default: 0]')
    #p.add_option('--specific-instances', action='append', default=[],
    #             metavar='CLASSNAME:N',
    #             help='number of instances of a specific class')
    p.add_option('--queue-size', type='int', default=2, metavar='N',
                 help='size of event queues [default: 2]')
    p.add_option('--specific-queue-size', action='append', default=[],
                 metavar='CLASSNAME:N',
                 help='size of event queue of a specific class')
    p.add_option('--allow-queue-overdraft', action='store_true', default=False,
                 help='allow exceeding queue capacity in step semantics')
    p.add_option('--encode-interleaving', action='store_true', default=False,
                 help='encode interleaving execution semantics')
    p.add_option('--encode-static', action='store_true', default=False,
                 help='encode static step semantics [default: dynamic steps]')
    p.add_option('--enum-type', type='string', default='enum',
                 metavar='ENCODING',
                 help='encoding of enumerations: "enum" or "bitvector" '
                 '[default: "enum"]')
    p.add_option('--enter', type='string', default='linear',
                 metavar='ENCODING',
                 help='encoding for enter/exit predicates: '
                 '"old", "trim", or "linear" [default: "linear"]')
    p.add_option('--qpos-type', type='string', default='bitvector',
                 metavar='ENCODING',
                 help='queue position type: "bitvector" or "integer" '
                 '[default: "bitvector"]')
    p.add_option('--queue-indexing', type='string', default='parameter',
                 metavar='ENCODING',
                 help='queue index encoding: "constant" or "parameter" '
                 '[default: "parameter"]')
    p.add_option('--queue-contents', type='string', default='tags',
                 metavar='ENCODING',
                 help='encoding of queue contents: "messages" or "tags" '
                 '[default: "tags"]')
    # Could add: --encode-state-exclusion
    global opt
    (opt, args) = p.parse_args()
    if len(args) != 1:
        p.error('wrong number of arguments')
    if opt.int_bits < 2 or opt.int_bits > 32:
        p.error('value of --int-bits must be between 2 and 32')
    if opt.format not in ['native', 'smtlib']:
        p.error('--format must be native or smtlib')
    if opt.enter not in ['old', 'trim', 'linear']:
        p.error('--enter must be old, trim, or linear')
    if opt.qpos_type not in ['bitvector', 'integer']:
        p.error('--qpos-type must be bitvector or integer')
    if opt.queue_indexing not in ['constant', 'parameter']:
        p.error('--queue-indexing must be constant or parameter')
    if opt.queue_contents not in ['messages', 'tags']:
        p.error('--queue-contents must be messages or tags')
    if opt.enum_type not in ['enum', 'bitvector']:
        p.error('--enum-type must be enum or bitvector')
    makeGenericComment(version, args[0])
    return args[0]

def makeGenericComment(version, inFilename):
    global comment
    comment = [' File generated by Suboco ' + version,
               ' Input file: ' + inFilename,
               ' Created on ' + time.ctime(time.time()),
               ' Options:']
    optionNames = opt.__dict__.keys()
    optionNames.sort()
    for n in optionNames:
        comment.append('   ' + n + ': ' + str(opt.__dict__[n]))
    comment.append('')
