#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# smtutils.py
#
# 2007-12-16  Jori Dubrovin:  Cleaned up options, changed defaults
# 2007-11-22  Jori Dubrovin:  Added real enums
# 2007-11-22  Jori Dubrovin:  Parameterized queue contents
# 2007-11-20  Jori Dubrovin:  Encoding options for queues and states
# 2007-08-30  Jori Dubrovin:  Suboco 1.01
# 2007-08-27  Jori Dubrovin:  Suboco 1.00
# 2007-07-26  Jori Dubrovin:  Initial version
#

"""Helper functions for PySMT interface."""

import PySMT

import options


# _enumType maps an integer to a PySMT enum type with that many literals.
# _enumValues maps an integer to a list of the literals.
_enumType = {}
_enumValues = {}

def log2ceil(n):
    """Return max(0, ceiling(log2(n)))."""
    width = 1  #$$ should be 0 but PySMT has a limitation.
    while (1L << width) < n:
        width += 1
    return width

class LabeledCircuit(PySMT.Circuit):
    """A Circuit that supports named gates."""

    def __init__(self):
        """Construct an empty labeled circuit."""
        PySMT.Circuit.__init__(self)
        self.__labeled = {}
        if options.opt.dot_file:
            self.__labelOfRealGate = {}

    def __contains__(self, label):
        return (label in self.__labeled)

    def __getitem__(self, label):
        return self.__labeled[label]

    def __setitem__(self, label, gate):
        self.__labeled[label] = gate
        if options.opt.dot_file:
            real = self._get_real_gate(gate)  #$$ not nice to hack
            if real not in self.__labelOfRealGate:
                self.__labelOfRealGate[real] = label

    def getLabels(self):
        return self.__labeled.keys()

    def getLabeledGates(self):
        return self.__labeled.values()

    def addLabeledCircuit(self, other, mapLabel=None):
        """Instantiate another labeled circuit in self.

        For a gate g in other with label l: if self contains a gate g'
        with label mapLabel(l), then g must be a variable gate, and it
        is mapped to g'.  Otherwise g is mapped to a new gate in self
        with label mapLabel(l).

        Param other: a LabeledCircuit.

        Param mapLabel: a function that takes a label and returns a
        new label.  The default is the identity function.
        """
        if mapLabel is None:
            mapLabel = lambda label: label
        inputMap = {}
        tracked = {}
        for label in other.getLabels():
            newLabel = mapLabel(label)
            if newLabel in self:
                inputMap[other[label]] = self[newLabel]
            else:
                tracked[newLabel] = other[label]
        inst = self.add_subcircuit(other, inputMap, tracked.values())
        for newLabel in tracked:
            self[newLabel] = inst[tracked[newLabel]]

    def addAtMostOne(self, *children):
        """Add and return an at most one gate.

        Param children: a list of boolean Gates.

        Return value: a boolean Gate that evaluates to true if and
        only if zero or one of the given gates evaluate to true.
        """
        (al1, al2) = self.__atLeastOneTwo(*children)
        return self.add_not(al2)

    def addAtLeastOne(self, *children):
        """Add and return an at least one gate.

        This is functionally the same as add_or, but this reuses the
        intermediate gates produced by calling addAtMostOne with the
        same arguments.
        """
        (al1, al2) = self.__atLeastOneTwo(*children)
        return al1

    def __atLeastOneTwo(self, *children):
        """Add gates for at least one and at least two.

        Return value: pair of Gates (al1, al2), where al1 is true iff
        at least one child is true, and al2 is true iff at least two
        children are true.
        """
        if len(children) == 0:
            return (self.add_false(), self.add_false())
        (al1, al2) = self.__atLeastOneTwo(*children[:-1])
        al2 = self.add_or(al2, self.add_and(al1, children[-1]))
        al1 = self.add_or(al1, children[-1])
        return (al1, al2)

    def addEnumVar(self, domainSize):
        """Add a variable whose domain contains [0, domainSize-1]."""
        if options.opt.enum_type == 'enum':
            return self.add_enum_var(
                self.addEnumConst(domainSize, 0).get_type())
        else:
            return self.add_uint_var(log2ceil(domainSize))

    def addEnumConst(self, domainSize, value):
        """Add a constant with the type of self.addEnumVar(domainSize)."""
        if options.opt.enum_type == 'enum':
            if domainSize not in _enumType:
                _enumValues[domainSize] = (
                    ['enum' + str(i) + 'of' + str(domainSize) for
                     i in xrange(domainSize)])
                _enumType[domainSize] = PySMT.new_enum(_enumValues[domainSize])
            return self.add_enum_const(_enumType[domainSize],
                                       _enumValues[domainSize][value])
        else:
            return self.add_uint_const(log2ceil(domainSize), value)

    def to_dot(self, f):
        """
        Print the circuit in the graphviz dotty format to
        the file object f.

        This function hacks through the PySMT interface.
        """
        assert options.opt.dot_file
        import random
        ordered = {'ITE': ['if', 'then', 'else'],
                   'LE': ['left', 'right'],
                   'LT': ['left', 'right'],
                   'MINUS': ['left', 'right']}
        f.write('digraph problem {\n')
        f.write('concentrate=true;\n')
        f.write('ranksep=2.0;\n')
        #f.write('ratio=1.0;\n')
        #f.write('mclimit=5.0;\n')
        #f.write('nslimit=5.0;\n')
        #labelOf = dict([(self._get_real_gate(self.__labeled[label]), label)
        #                for label in self.__labeled])
        for gate in self._gates:
            if gate is None:
                continue
            param_str = ''
            if gate.opparams != None:
                param_str = str(gate.opparams)
            i = gate.get_index()
            f.write('n' + str(i) + ' ')
            f.write('[label="')
            if gate in self.__labelOfRealGate:
                parts = [str(t) for t in self.__labelOfRealGate[gate]]
                f.write('(' + ',\\n'.join(parts) + ')\\n')
            f.write(#'g' + str(i) + ' ' +
                str(gate.operand) + param_str)
            if gate.is_constrained():
                f.write(' '+str(gate.constrained_value))
            f.write('"];\n')
            color = '"' + ','.join([
                str(random.uniform(0.0, 1.0))[:5],  # hue
                str(random.uniform(0.4, 1.0))[:5],  # saturation
                str(random.uniform(0.3, 0.9))[:5],  # brightness
                ]) + '"'
            for (index, child) in enumerate(gate.children):
                f.write('n' + str(i ) +' -> n' + str(child.get_index()) + ' [')
                if gate.operand in ordered:
                    f.write('label="' + ordered[gate.operand][index] + '", ')
                    #f.write('taillabel=true, ')
                    #       'labeldistance=100, ')
                #f.write('minlen=3, ')
                f.write('color=' + color + ', ')
                f.write('labelfontcolor=' + color + ', ')
                #f.write('weight=' + str(3.0/len(gate.children))[:5] + ', ')
                f.write('dir=forward];\n')
        f.write('}\n')

##  def add_iff(self, child1, child2):
##      if options.opt.normalize_circuit:
##          if self._get_real_gate(child1).operand == 'NOT':
##              return self.add_not(self.add_iff(self.add_not(child1), child2))
##          if self._get_real_gate(child2).operand == 'NOT':
##              return self.add_not(self.add_iff(child1, self.add_not(child2)))
##      return PySMT.Circuit.add_iff(self, child1, child2)
        
##  def add_or(self, *children):
##      if options.opt.normalize_circuit:
##          return self.add_not(self.add_and(*[self.add_not(child) for
##                                             child in children]))
##      return PySMT.Circuit.add_or(self, *children)

##  def add_ite(self, if_child, then_child, else_child):
##      if options.opt.normalize_circuit:
##          if self._get_real_gate(if_child).operand == 'NOT':
##              return self.add_ite(self.add_not(if_child),
##                                  else_child, then_child)
##          if self._get_real_gate(then_child).operand == 'NOT':
##              return self.add_not(self.add_ite(if_child,
##                                               self.add_not(then_child),
##                                               self.add_not(else_child)))
##      return PySMT.Circuit.add_ite(self, if_child, then_child, else_child)

##  def add_uint_var(self, n):
##      if options.opt.normalize_circuit:
##          if n == 1:
##              return self.add_var()
##      return PySMT.Circuit.add_uint_var(self, n)

##  def add_uint_const(self, n, c):
##      if options.opt.normalize_circuit:
##          if n == 1:
##              return [self.add_false(), self.add_true()][c]
##      return PySMT.Circuit.add_uint_const(self, n, c)

##  def add_lt(self, child1, child2):
##      if options.opt.normalize_circuit:
##          if self._get_real_gate(child1).is_boolean():
##              return self.add_and(self.add_not(child1), child2)
##      return PySMT.Circuit.add_lt(self, child1, child2)

##  def add_le(self, child1, child2):
##      if options.opt.normalize_circuit:
##          if self._get_real_gate(child1).is_boolean():
##              return self.add_implies(child1, child2)
##      return PySMT.Circuit.add_le(self, child1, child2)

##  def add_plus(self, *children):
##      if options.opt.normalize_circuit:
##          if self._get_real_gate(children[0]).is_boolean():
##              v = self.add_false()
##              for child in children:
##                  v = self.add_neq(v, child)
##              return v
##      return PySMT.Circuit.add_plus(self, *children)


def generateId(prefix, default, reserved):
    """Return a unique string identifier.

    The identifier can be used as a part of a NuSMV variable name.  If
    the global option printNames is False, the identifier contains
    only the given prefix followed by digits.

    Param prefix: a prefix string for the identifier.

    Param default: the preferred string without the prefix.

    Param reserved: an iterable whose keys are the identifiers already in use.

    Return value: a string in <prefix>[A-Za-z0-9$_]+ that is not in reserved.
    """
    stem = prefix
    if True:  # Skip this to generate numbers instead of names
        for c in default:
            if c in ('ABCDEFGHIJKLMNOPQRSTUVWXYZ' +
                     'abcdefghijklmnopqrstuvwxyz' +
                     '0123456789$'):
                stem += str(c)
            else:
                stem += '_'
    n = 0
    s = stem
    while s == prefix or s in reserved:
        n += 1
        s = stem + str(n)
    return s
