#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# statemachines.py
#
# 2007-12-16  Jori Dubrovin:  Cleaned up options, changed defaults
# 2007-11-20  Jori Dubrovin:  Encoding options for queues and states
# 2007-09-01  Jori Dubrovin:  Suboco 1.02
# 2007-08-27  Jori Dubrovin:  Suboco 1.00
# 2007-06-20  Jori Dubrovin:  New choose expression mapping
# 2007-05-07  Jori Dubrovin:  Optimized linear Enter predicate
# 2007-05-04  Jori Dubrovin:  Added linear Enter predicate
# 2007-04-01  Jori Dubrovin:  Using simpler constraint for FireTriggered
# 2007-03-08  Jori Dubrovin:  Adding states to element map
# 2007-03-02  Jori Dubrovin:  Adding choose to element map
# 2006-12-04  Jori Dubrovin:  Fixed enabledness condition
# 2006-10-30  Jori Dubrovin:  Initial version
#

"""Encoding of state machines and their control logic."""

import exceptions

import actions
import activities
import datatypes
import errors
import options
import smtutils
from coralutils import uml, umlSort


class StateMachine:
    """A behavioral state machine."""

    def __init__(self, umlMachine, class_):
        """Construct a state machine.

        Param umlMachine: the underlying uml.StateMachine.

        Param class_: the owning ActiveClass.
        """
        assert umlMachine.oclIsKindOf(uml.StateMachine)
        self.__class = class_
        self.__regions = []           # all Regions in self
        self.__regionForUml = {}      # map uml.StateVertex to Region
        self.__vertices = []          # all Vertices in self
        self.__vertexForUml = {}      # map uml.StateVertex to Vertex
        self.__transitions = []       # all Transitions in self
        self.__transitionForUml = {}  # map uml.Transition to Transition
        self.__activityCircuit = smtutils.LabeledCircuit()
        # Inputs of activity circuit:
        # (0, obj=None, 'ThisRef')
        # (0, obj, 'FieldValue', field)
        # (0, obj=None, 'CurrEvent', item)
        # (0, obj=None, 'Fire', trans)
        # Outputs of activity circuit:
        # (0, obj=None, 'EvalGuard', transition)
        # (0, actor=None, 'Read', transition, target, field)
        # (0, actor=None, 'Write', transition, target, field)
        # (0, actor=None, 'Valuation', transition, target, field)
        # (0, actor=None, 'Send', transition, target)
        # (0, actor=None, 'SendExact', transition, target)
        # (0, actor=None, 'SendArg', transition, targetClass, paramIndex)
        # (0, actor=None, 'ActivityConstraints', transition)
        # (0, actor=None, 'Assertions', transition)
        # (0, actor=None, 'RuntimeErrors', transition)
        if umlMachine.top.container:
            raise errors.InputError('top region of state machine ' +
                                    str(umlMachine) + ' has a container')
        self.__topRegion = self.__addRegion(umlMachine.top)
        self.__addDescendants(self.__topRegion)
        for umlTransition in umlSort(umlMachine.transitions):
            self.__addTransition(umlTransition)
        topInitial = [s for
                      s in self.__topRegion.getSubvertices() if s.isInitial()]
        if len(topInitial) != 1:
            raise errors.InputError('state machine ' + str(umlMachine) +
                                    ' does not have exactly one initial state')
        for s in self.__vertices:
            s.checkSanity()
        for t in self.__transitions:
            t.processControl()

    def __addDescendants(self, region):
        """Add all descendant vertices and regions of a Region to self."""
        for umlVertex in region.getUmlSubvertices():
            s = self.__addVertex(umlVertex)
            for umlRegion in s.getUmlRegions():
                r = self.__addRegion(umlRegion)
                self.__addDescendants(r)

    def __addRegion(self, umlRegion):
        """Add a new Region for the given uml.StateVertex to self.

        Return value: the new Region.
        """
        id_ = smtutils.generateId('r', umlRegion.name,
                                  [r.smvId() for r in self.__regions])
        r = Region(umlRegion, id_, self)
        self.__regions.append(r)
        assert not umlRegion in self.__regionForUml
        self.__regionForUml[umlRegion] = r
        return r

    def __addVertex(self, umlVertex):
        """Add a new Vertex for the given uml.StateVertex to self.

        Return value: the new Vertex.
        """
        id_ = smtutils.generateId('s', umlVertex.name,
                                  [s.smvId() for s in self.__vertices])
        s = Vertex(umlVertex, id_, self)
        self.__vertices.append(s)
        assert not umlVertex in self.__vertexForUml
        self.__vertexForUml[umlVertex] = s
        return s

    def __addTransition(self, umlTransition):
        """Add a new Transition for the given uml.Transition to self.

        Return value: the new Transition.
        """
        id_ = smtutils.generateId('t', umlTransition.name,
                                  [t.smvId() for t in self.__transitions])
        t = Transition(umlTransition, id_, self)
        self.__transitions.append(t)
        assert not umlTransition in self.__transitionForUml
        self.__transitionForUml[umlTransition] = t
        return t

    def getClass(self):
        """Return the owning ActiveClass."""
        return self.__class

    def getVertices(self):
        """Return all Vertices in self."""
        return self.__vertices[:]

    def getTransitions(self):
        """Return all transitions in self."""
        return self.__transitions[:]

    def getRegionForUml(self, umlRegion):
        """Return the Region (owned by self) for a given uml.StateVertex."""
        return self.__regionForUml[umlRegion]

    def getVertexForUml(self, umlVertex):
        """Return the Vertex (owned by self) for a given uml.StateVertex."""
        return self.__vertexForUml[umlVertex]

    def getTransitionForUml(self, umlTransition):
        """Return the Transition (owned by self) for a given uml.Transition."""
        return self.__transitionForUml[umlTransition]

    def getActivityCircuit(self):
        """Return the activity circuit."""
        return self.__activityCircuit

    def smtInPseudostate(self, obj):
        """Return the predicate InPseudostate."""
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'InPseudostate')
        if label not in mc:
            mc[label] = mc.add_or(*[s.smtActive(obj) for
                                    s in self.__vertices if s.isPseudostate()])
        return mc[label]

    def smtInRtc(self, obj):
        """Return the predicate InRtc."""
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'InRtc')
        if label not in mc:
            mc[label] = mc.add_or(
                self.smtInPseudostate(obj),
                *[s.smtCompleted(obj) for
                  s in self.__vertices if not s.isPseudostate()])
        return mc[label]

    def smtReady(self, obj):
        """Return the predicate Ready."""
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'Ready')
        if label not in mc:
            queue = self.__class.getEventQueue()
            signals = queue.getRelevantSignals()
            mc[label] = mc.add_or(self.smtInRtc(obj),
                                  mc.add_not(queue.smtInputQueueEmpty(obj)),
                                  *[self.smtAnyFeasible(obj, e) for
                                    e in signals if e.isExternal()])
        return mc[label]

    def smtScheduled(self, obj):
        """Return the predicate Scheduled."""
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'Scheduled')
        if label not in mc:
            pseudoFire = [t.smtFire(obj) for t in self.__transitions if
                          t.getSource().isPseudostate()]
            busyFire = [t.smtFire(obj) for t in self.__transitions if
                        not t.getSource().isPseudostate() and t.isCompletion()]
            quiesce = [s.smtQuiesce(obj) for s in self.__vertices if
                       s.isCompletionSensitive()]
            queue = self.__class.getEventQueue()
            mc[label] = mc.add_or(mc.addAtLeastOne(*pseudoFire),
                                  mc.addAtLeastOne(*busyFire + quiesce),
                                  queue.smtDispatchAny(obj))
        return mc[label]

    def smtAnyFeasible(self, obj, signal):
        """Return the predicate AnyFeasible."""
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'AnyFeasible', signal)
        if label not in mc:
            # This is analogous to the definition of
            # DescendantFeasible.
            mc[label] = mc.add_or(
                *[mc.add_or(s.smtFeasibleTriggered(obj, signal),
                            s.smtDescendantFeasible(obj, signal)) for
                  s in self.__topRegion.getSubvertices() if
                  not s.isPseudostate()])
        return mc[label]

    def smtAnyDeferring(self, obj, signal):
        """Return the predicate AnyDeferring."""
        mc = self.__class.getModel().getCircuit()
        label = (0, obj, 'AnyDeferring', signal)
        if label not in mc:
            # This is analogous to the definition of
            # DescendantDeferring.
            mc[label] = mc.add_or(
                *[mc.add_or(s.smtDeferring(obj, signal),
                            s.smtDescendantDeferring(obj, signal)) for
                  s in self.__topRegion.getSubvertices() if
                  not s.isPseudostate()])
        return mc[label]

    def encode(self, obj):
        """Encode state machine logic."""
        model = self.__class.getModel()
        mc = model.getCircuit()
        if options.opt.encode_interleaving:
            pseudoFire = [t.smtFire(obj) for t in self.__transitions if
                          t.getSource().isPseudostate()]
            busyFire = [t.smtFire(obj) for t in self.__transitions if
                        not t.getSource().isPseudostate() and t.isCompletion()]
            quiesce = [s.smtQuiesce(obj) for s in self.__vertices if
                       s.isCompletionSensitive()]
            model.addTrans(mc.addAtMostOne(*pseudoFire))
            model.addTrans(mc.addAtMostOne(*busyFire + quiesce))
        for s in self.__vertices:
            s.encode(obj)
        for t in self.__transitions:
            t.encodeControl(obj)
            t.encodeActivity(obj)

    def defineActivities(self, actor):
        """Instantiate the activity circuit in the model circuit."""
        mc = self.__class.getModel().getCircuit()
        def mapLabel(label):
            if label[1] is not None:
                return label
            label = list(label)
            label[1] = actor
            return tuple(label)
        # Define in the model circuit all gates that are referred to
        # in the activity circuit.
        for t in self.__transitions:
            void = t.smtFire(actor)
        void = self.__class.smtThisRef(actor)
        self.__class.getEventQueue().defineCurrEventParams(actor)
        self.__class.getModel().defineFieldValues()
        mc.addLabeledCircuit(self.__activityCircuit, mapLabel)


class Region:
    """A region in a state machine."""

    def __init__(self, umlRegion, id_, machine):
        """Construct a region.

        Param umlRegion: the underlying uml.StateVertex.

        Param id_: an id string.

        Param machine: the StateMachine owning the region.
        """
        assert umlRegion.oclIsKindOf(uml.StateVertex)
        self.__umlRegion = umlRegion
        self.__id = id_
        self.__machine = machine
        self.__containedTransitions = []  # Initially empty
        self.__breakingTransitions = []   # Initially empty
        if not umlRegion.oclIsKindOf(uml.CompositeState):
            raise errors.InputError('region ' + str(umlRegion) +
                                    ' is not composite')
        if umlRegion.isConcurrent:
            raise errors.InputError('region ' + str(umlRegion) +
                                    ' is concurrent')
        if umlRegion.oclIsKindOf(uml.SubmachineState):
            raise errors.NotSupportedError('submachine state ' +
                                           str(umlRegion) +
                                           ' is not supported')
        if umlRegion.internalTransition:
            raise errors.NotSupportedError('region ' + str(umlRegion) +
                                           ' has an internal transition')
        if umlRegion.entry:
            raise errors.NotSupportedError('region ' + str(umlRegion) +
                                           ' has an entry activity')
        if umlRegion.exit:
            raise errors.NotSupportedError('region ' + str(umlRegion) +
                                           ' has an exit activity')
        if umlRegion.doActivity:
            raise errors.NotSupportedError('region ' + str(umlRegion) +
                                           ' has a do activity')
        if not umlRegion.container or umlRegion.container.isConcurrent:
            # self.__umlRegion is top or a region in a concurrent
            # composite state.  Therefore it must not have incoming or
            # outgoing transitions.
            if umlRegion.incoming:
                raise errors.NotSupportedError('region ' + str(umlRegion) +
                                               ' has an incoming transition')
            if umlRegion.outgoing:
                raise errors.NotSupportedError('region ' + str(umlRegion) +
                                               ' has an outgoing transition')

    def __str__(self):
        return self.__id

    def getUmlSubvertices(self):
        """Return a list of direct subvertices.

        Return value: a list of uml.StateVertex.
        """
        return umlSort(self.__umlRegion.subvertex)

    def getSubvertices(self):
        """Return a list of direct subvertices.

        Return value: a list of Vertex.
        """
        return [self.__machine.getVertexForUml(s) for
                s in self.getUmlSubvertices()]

    def getInitial(self):
        """Return the initial pseudostate (Vertex) in self."""
        initial = [s for s in self.getSubvertices() if s.isInitial()]
        if len(initial) != 1:
            raise errors.InputError('region ' + str(self.__umlRegion) +
                                    ' does not have exactly one ' +
                                    'initial pseudostate')
        return initial[0]

    def getContainingState(self):
        """Return the containing composite state (Vertex)."""
        assert not self.isTop()
        if self.__umlRegion.container.isConcurrent:
            return self.__machine.getVertexForUml(self.__umlRegion.container)
        else:
            return self.__machine.getVertexForUml(self.__umlRegion)

    def isTop(self):
        """Return True if self is the top region of a state machine."""
        if self.__umlRegion.container:
            return False
        return True

    def addContainedTransition(self, t):
        """Add a Transition that whose lca(source, target) is self."""
        self.__containedTransitions += [t]

    def addBreakingTransition(self, t):
        """Add a Transition that that breaks in self."""
        self.__breakingTransitions += [t]

    def smvId(self):
        """Return the id string."""
        return self.__id

    def smtBreakIn(self, obj):
        """Return the predicate BreakIn."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, obj, 'BreakIn', self)
        if label not in mc:
            if self.__breakingTransitions:
                sub = [s.smtEnter(obj) for
                       s in self.getSubvertices() if not s.isInitial()]
                cont = [t.smtFire(obj) for t in self.__containedTransitions]
                mc[label] = mc.add_and(mc.add_or(*sub),
                                       mc.add_not(mc.add_or(*cont)))
            else:
                mc[label] = mc.add_false()
        return mc[label]

    def smtFinished(self, obj):
        """Return the predicate Finished.

        A region is finished if one of its direct final substates is active
        and quiescent.
        """
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, obj, 'Finished', self)
        if label not in mc:
            mc[label] = mc.add_or(*[s.smtActive(obj) for
                                    s in self.getSubvertices() if s.isFinal()])
        return mc[label]


class Vertex:
    """A state vertex in a state machine."""

    def __init__(self, umlVertex, id_, machine):
        """Construct a state vertex.

        Param umlVertex: a uml.StateVertex.

        Param id_: an id string.

        Param machine: the StateMachine owning the state vertex.
        """
        assert umlVertex.oclIsKindOf(uml.StateVertex)
        self.__umlVertex = umlVertex
        self.__id = id_
        self.__machine = machine
        self.__enteringTransitions = []  # Initially empty
        self.__escapingTransitions = []  # Initially empty
        if not (umlVertex.oclIsKindOf(uml.SimpleState) or
                umlVertex.oclIsKindOf(uml.FinalState) or
                (umlVertex.oclIsKindOf(uml.CompositeState) and
                 not umlVertex.oclIsKindOf(uml.SubmachineState)) or
                (umlVertex.oclIsKindOf(uml.Pseudostate) and
                 umlVertex.kind in ['initial', 'choice'])):
            raise errors.NotSupportedError('unsupported state vertex ' +
                                           str(umlVertex))
        if umlVertex.oclIsKindOf(uml.State):
            if umlVertex.internalTransition:
                raise errors.NotSupportedError('state ' + str(umlVertex) +
                                               ' has an internal transition')
            if umlVertex.entry:
                raise errors.NotSupportedError('state ' + str(umlVertex) +
                                               ' has an entry activity')
            if umlVertex.exit:
                raise errors.NotSupportedError('state ' + str(umlVertex) +
                                               ' has an exit activity')
            if umlVertex.doActivity:
                raise errors.NotSupportedError('state ' + str(umlVertex) +
                                               ' has a do activity')

    def __str__(self):
        return self.__id

    def getXmiId(self):
        """Return the XMI id of the underlying uml.StateVertex as a string."""
        return str(self.__umlVertex.__XMIid__())

    def checkSanity(self):
        """Perform sanity checks in the model and raise if errors found."""
        try:
            for umlTransition in self.__umlVertex.incoming:
                t = self.__machine.getTransitionForUml(umlTransition)
                if self.isInitial():
                    raise errors.InputError('initial pseudostate ' +
                                            str(self.__umlVertex) +
                                            ' has an incoming transition')
            for umlTransition in self.__umlVertex.outgoing:
                t = self.__machine.getTransitionForUml(umlTransition)
                if self.isFinal():
                    raise errors.InputError('final state ' +
                                            str(self.__umlVertex) +
                                            ' has an outgoing transition')
                if self.isPseudostate() and not t.isCompletion():
                    raise errors.InputError('pseudostate ' +
                                            str(self.__umlVertex) +
                                            ' has an outgoing ' +
                                            'signal-triggered transition')
        except exceptions.LookupError:
            raise errors.InputError('state ' + str(self.__umlVertex) +
                                    ' has a dangling transition')

    def isPseudostate(self):
        """Return True if self is a pseudostate."""
        return self.__umlVertex.oclIsKindOf(uml.Pseudostate)

    def isInitial(self):
        """Return True if self is an initial pseudostate."""
        return (self.__umlVertex.oclIsKindOf(uml.Pseudostate) and
                self.__umlVertex.kind == 'initial')

    def isFinal(self):
        """Return True if self is a final state."""
        return self.__umlVertex.oclIsKindOf(uml.FinalState)

    def getUmlRegions(self):
        """Return the direct subregions.

        Return value: a list of uml.StateVertex.
        """
        if self.__umlVertex.oclIsKindOf(uml.CompositeState):
            if self.__umlVertex.isConcurrent:
                return umlSort(self.__umlVertex.subvertex)
            else:
                return [self.__umlVertex]
        else:
            return []

    def getRegions(self):
        """Return the direct subregions.

        Return value: a list of Regions.
        """
        return [self.__machine.getRegionForUml(umlRegion) for
                umlRegion in self.getUmlRegions()]

    def getSubvertices(self):
        """Return the direct subvertices in the direct subregions.

        Return value: a list of Vertex.
        """
        sub = []
        for r in self.getRegions():
            sub.extend(r.getSubvertices())
        return sub

    def getContainingRegion(self):
        """Return the containing Region."""
        return self.__machine.getRegionForUml(self.__umlVertex.container)

    def getContainingState(self):
        """Return the containing composite state (Vertex)."""
        return self.getContainingRegion().getContainingState()

    def getOutgoing(self):
        """Return a list of outgoing Transitions."""
        return [self.__machine.getTransitionForUml(umlTransition) for
                umlTransition in umlSort(self.__umlVertex.outgoing)]

    def isCompletionSensitive(self):
        """Return True if self is completion sensitive.

        A state vertex is completion sensitive if it is a
        non-pseudostate and has outgoing completion transitions.
        """
        if self.isPseudostate():
            return False
        for t in self.getOutgoing():
            if t.isCompletion():
                return True
        return False

    def defers(self, signal):
        """Return True if self declares the given Signal deferrable."""
        if self.isPseudostate():
            return False
        deferrableUmlSignals = [umlEvent.signal for
                                umlEvent in self.__umlVertex.deferrableEvent]
        return signal.isInUmlSignals(deferrableUmlSignals)

    def addEnteringTransition(self, t):
        """Add a Transition that enters self."""
        self.__enteringTransitions += [t]

    def addEscapingTransition(self, t):
        """Add a Transition that escapes self.

        A transition t escapes a state vertex s if t exits s and t
        does not exit any proper ancestor of s.
        """
        self.__escapingTransitions += [t]

    def smvId(self):
        """Return the id string."""
        return self.__id

    def smtActive(self, obj):
        """Return the predicate Active."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, obj, 'Active', self)
        if label not in mc:
            mc[label] = mc.add_var()
        return mc[label]

    def smtNextActive(self, obj):
        """Return the next-state predicate Active."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (1, obj, 'Active', self)
        if label not in mc:
            mc[label] = mc.add_var()
        return mc[label]

    def smtEnter(self, obj):
        """Return the predicate Enter."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, obj, 'Enter', self)
        if label not in mc:
            if self.getContainingRegion().isTop() and self.isInitial():
                mc[label] = self.__machine.getClass().smtCreated(obj)
                assert not self.__enteringTransitions
            else:
                if options.opt.enter == 'old':
                    mc[label] = mc.add_or(*[t.smtFire(obj) for
                                            t in self.__enteringTransitions])
                elif self.isInitial():
                    mc[label] = mc.add_and(
                        self.getContainingState().smtEnter(obj),
                        mc.add_not(self.getContainingRegion().smtBreakIn(obj)))
                else:
                    enter = ([t.smtFire(obj) for
                              t in self.__machine.getTransitions() if
                              t.getTarget() is self] +
                             [r.smtBreakIn(obj) for r in self.getRegions()])
                    mc[label] = mc.add_or(*enter)
        return mc[label]

    def smtEnterOnly(self, obj):
        """Return the predicate EnterOnly."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, obj, 'EnterOnly', self)
        if label not in mc:
            if self.getContainingRegion().isTop() and self.isInitial():
                mc[label] = self.__machine.getClass().smtCreated(obj)
                assert not self.__enteringTransitions
            else:
                mc[label] = mc.add_or(*[t.smtFire(obj) for
                                        t in self.__enteringTransitions if
                                        t not in self.__escapingTransitions])
        return mc[label]

    def smtExit(self, obj):
        """Return the predicate Exit."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, obj, 'Exit', self)
        if label not in mc:
            exitReasons = ([self.smtExitOnly(obj)] +
                           [t.smtFire(obj) for
                            t in self.__escapingTransitions if
                            t in self.__enteringTransitions])
            mc[label] = mc.addAtLeastOne(*exitReasons)
        return mc[label]

    def smtExitOnly(self, obj):
        """Return the predicate ExitOnly."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, obj, 'ExitOnly', self)
        if label not in mc:
            if self.getContainingRegion().isTop():
                exitOnlyReasons = [self.__machine.getClass().smtDestroyed(obj)]
            else:
                exitOnlyReasons = [self.getContainingState().smtExit(obj)]
            exitOnlyReasons += [t.smtFire(obj) for
                                t in self.__escapingTransitions if
                                t not in self.__enteringTransitions]
            mc[label] = mc.addAtLeastOne(*exitOnlyReasons)
        return mc[label]

    def smtQuiescent(self, obj):
        """Return the predicate Quiescent."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, obj, 'Quiescent', self)
        if label not in mc:
            mc[label] = mc.add_var()
        return mc[label]

    def smtNextQuiescent(self, obj):
        """Return the next-state predicate Quiescent."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (1, obj, 'Quiescent', self)
        if label not in mc:
            mc[label] = mc.add_var()
        return mc[label]

    def smtCompleted(self, obj):
        """Return the predicate Completed."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, obj, 'Completed', self)
        if label not in mc:
            assert not self.isPseudostate()
            if self.isCompletionSensitive():
                mc[label] = mc.add_and(
                    self.smtActive(obj),
                    mc.add_not(self.smtQuiescent(obj)),
                    *[r.smtFinished(obj) for r in self.getRegions()])
            else:
                mc[label] = mc.add_false()
        return mc[label]

    def smtQuiesce(self, obj):
        """Return the predicate Quiesce."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, obj, 'Quiesce', self)
        if label not in mc:
            mc[label] = mc.add_var()
        return mc[label]

    def smtDeferring(self, obj, signal):
        """Return the predicate Deferring."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, obj, 'Deferring', self, signal)
        if label not in mc:
            if self.defers(signal):
                mc[label] = self.smtActive(obj)
            else:
                mc[label] = mc.add_false()
        return mc[label]

    def smtFeasibleTriggered(self, obj, signal):
        """Return the predicate FeasibleTriggered."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, obj, 'FeasibleTriggered', self, signal)
        if label not in mc:
            mc[label] = mc.add_and(
                self.smtActive(obj),
                mc.add_not(self.smtDescendantFeasible(obj, signal)),
                mc.add_not(self.smtDescendantDeferring(obj, signal)),
                mc.add_or(*[t.smtEvalGuard(obj) for
                            t in self.getOutgoing()
                            if t.getTriggerSignal() is signal]))
        return mc[label]

    def smtDescendantFeasible(self, obj, signal):
        """Return the predicate DescendantFeasible."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, obj, 'DescendantFeasible', self, signal)
        if label not in mc:
            mc[label] = mc.add_or(
                *[mc.add_or(s.smtFeasibleTriggered(obj, signal),
                            s.smtDescendantFeasible(obj, signal)) for
                  s in self.getSubvertices() if not s.isPseudostate()])
        return mc[label]

    def smtDescendantDeferring(self, obj, signal):
        """Return the predicate DescendantDeferring."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, obj, 'DescendantDeferring', self, signal)
        if label not in mc:
            mc[label] = mc.add_or(
                *[mc.add_or(s.smtDeferring(obj, signal),
                            s.smtDescendantDeferring(obj, signal)) for
                  s in self.getSubvertices() if not s.isPseudostate()])
        return mc[label]

    def encode(self, obj):
        """Generate constraints for state vertex specific control logic."""
        model = self.__machine.getClass().getModel()
        mc = model.getCircuit()
        if self.getContainingRegion().isTop() and self.isInitial():
            model.addInit(mc.add_iff(self.smtActive(obj),
                                     self.__machine.getClass().smtAlive(obj)))
        else:
            model.addInit(mc.add_not(self.smtActive(obj)))
        if options.opt.enter == 'trim':
            model.addTrans(mc.add_iff(
                self.smtNextActive(obj),
                mc.add_or(self.smtEnterOnly(obj),
                          mc.add_and(self.smtActive(obj),
                                     mc.add_not(self.smtExitOnly(obj))))))
        else:
            model.addTrans(mc.add_iff(
                self.smtNextActive(obj),
                mc.add_or(self.smtEnter(obj),
                          mc.add_and(self.smtActive(obj),
                                     mc.add_not(self.smtExit(obj))))))
        if self.getContainingRegion().isTop():
            exitOnlyReasons = [self.__machine.getClass().smtDestroyed(obj)]
        else:
            exitOnlyReasons = [self.getContainingState().smtExit(obj)]
        exitOnlyReasons += [t.smtFire(obj) for
                            t in self.__escapingTransitions if
                            t not in self.__enteringTransitions]
        model.addTrans(mc.addAtMostOne(*exitOnlyReasons))
        exitReasons = ([self.smtExitOnly(obj)] +
                       [t.smtFire(obj) for
                        t in self.__escapingTransitions if
                        t in self.__enteringTransitions])
        model.addTrans(mc.addAtMostOne(*exitReasons))
        if not self.isPseudostate():
            if self.isCompletionSensitive():
                model.addInit(mc.add_not(self.smtQuiescent(obj)))
                model.addTrans(mc.add_iff(self.smtNextQuiescent(obj),
                                          mc.add_and(
                    mc.add_not(self.smtExit(obj)),
                    mc.add_or(self.smtQuiescent(obj), self.smtQuiesce(obj)))))
                canQuiesce = mc.add_and(
                    mc.add_not(self.__machine.smtInPseudostate(obj)),
                    self.smtCompleted(obj),
                    *[mc.add_not(t.smtEvalGuard(obj)) for
                      t in self.getOutgoing() if t.isCompletion()])
                # Could be optimized.
                model.addTransImpl(self.smtQuiesce(obj), canQuiesce)
            queue = self.__machine.getClass().getEventQueue()
            for signal in queue.getRelevantSignals():
                fireTriggered = mc.add_or(*[t.smtFire(obj) for
                                            t in self.getOutgoing() if
                                            t.getTriggerSignal() is signal])
                priorityEnabled = mc.add_and(
                    queue.smtDispatch(obj, signal),
                    self.smtFeasibleTriggered(obj, signal))
                # c += codeTrans(self.smvFireTriggered(e) + ' -> ' +
                #                self.smvPriorityEnabled(e))
                # c += codeTrans(self.smvPriorityEnabled(e) + ' -> ' +
                #                self.smvExit())
                # The constraints above work even for simultaneous
                # firing in orthogonal regions.  The one below does
                # not, but is simpler.
                model.addTrans(mc.add_iff(fireTriggered, priorityEnabled))


class Transition:
    """A transition in a state machine."""

    def __init__(self, umlTransition, id_, machine):
        """Construct a transition.

        Param umlTransition: a uml.Transition.

        Param id_: an id string.

        Param machine: the StateMachine owning the transition.
        """
        assert umlTransition.oclIsKindOf(uml.Transition)
        self.__umlTransition = umlTransition
        self.__id = id_
        self.__machine = machine
        umlTrigger = self.__umlTransition.trigger
        if umlTrigger and not umlTrigger.oclIsKindOf(uml.SignalEvent):
            raise errors.NotSupportedError('unsupported trigger ' +
                                           str(umlTrigger))

    def __str__(self):
        return self.__id

    def getXmiId(self):
        """Return the XMI id of the underlying uml.Transition as a string."""
        return str(self.__umlTransition.__XMIid__())

    def getSource(self):
        """Return the source Vertex."""
        try:
            return self.__machine.getVertexForUml(self.__umlTransition.source)
        except exceptions.LookupError:
            raise errors.InputError('dangling source in transition ' +
                                    str(self.__umlTransition))

    def getTarget(self):
        """Return the target Vertex."""
        try:
            return self.__machine.getVertexForUml(self.__umlTransition.target)
        except exceptions.LookupError:
            raise errors.InputError('dangling target in transition ' +
                                    str(self.__umlTransition))

    def isCompletion(self):
        """Return True if self is a completion transition."""
        if self.__umlTransition.trigger:
            return False
        return True

    def getTriggerSignal(self):
        """Return the triggering Signal or None."""
        if self.isCompletion():
            return None
        umlTrigger = self.__umlTransition.trigger
        try:
            return self.__machine.getClass().getSignalForUml(umlTrigger.signal)
        except exceptions.LookupError:
            raise errors.InputError('unknown trigger in transition ' +
                                    str(self.__umlTransition))

    def getTriggerParamFields(self):
        """Return the Fields to which the signal parameters are read."""
        signal = self.getTriggerSignal()
        if signal is None:
            return []
        class_ = self.__machine.getClass()
        try:
            # umlSort must not be used for trigger.parameter.
            fields = [class_.getFieldForName(umlParam.name) for
                      umlParam in self.__umlTransition.trigger.parameter]
        except exceptions.LookupError:
            raise errors.InputError('unknown field in the signal parameter ' +
                                    'list of transition ' +
                                    str(self.__umlTransition))
        types = signal.getParamTypes()
        if len(types) != len(fields):
            raise errors.InputError('signal parameter list length mismatch '
                                    'in transition ' +
                                    str(self.__umlTransition))
        for (f, type_) in zip(fields, types):
            if fields.count(f) != 1:
                raise errors.InputError('duplicate field in the signal ' +
                                        'parameter list of transition ' +
                                        str(self.__umlTransition))
            if not type_.isSubtypeOf(f.getType()):
                raise errors.InputError('signal parameter type mismatch '
                                        'in transition ' +
                                        str(self.__umlTransition))
        return fields

    def getActivity(self):
        """Return the Activity of the transition.

        processActivity must have been called.
        """
        return self.__activity

    def getGuardChooseKeys(self):
        """Return the choose keys used in the guard of self.

        processActivity must have been called.

        Return value: a list of choose keys.
        """
        return self.__guardChooseKeys[:]

    def getEffectChooseKeys(self):
        """Return the choose keys used in the effect of self.

        processActivity must have been called.

        Return value: a list of choose keys.
        """
        return self.__effectChooseKeys[:]

    def processControl(self):
        """Determine how the state machine control logic is affected by self.

        This is done by notifying all Vertices that are entered or
        exited when self is fired.
        """
        # Determine the least common ancestor (lca) of source
        # and target state vertices
        sourcePath = [self.getSource()]
        while not sourcePath[0].getContainingRegion().isTop():
            sourcePath.insert(0, sourcePath[0].getContainingState())
        targetPath = [self.getTarget()]
        while not targetPath[0].getContainingRegion().isTop():
            targetPath.insert(0, targetPath[0].getContainingState())
        while (len(sourcePath) >= 2 and len(targetPath) >= 2 and
               sourcePath[1].getContainingRegion() is
               targetPath[1].getContainingRegion()):
            sourcePath = sourcePath[1:]
            targetPath = targetPath[1:]
        assert (sourcePath[0].getContainingRegion() is
                targetPath[0].getContainingRegion())
        sourcePath[0].getContainingRegion().addContainedTransition(self)
        # Now sourcePath (targetPath) is a list of nested state
        # vertices from the lca to the source (target).  Every state
        # vertex in sourcePath (plus more, possibly) is exited by
        # self.  Every state vertex in targetPath (plus more,
        # possibly) is entered by self.
        sourcePath[0].addEscapingTransition(self)
        # Walk the path from the lca to target and add all state
        # vertices to the enters set.  Also add the initial state of
        # every region that is orthogonal to the path.
        while targetPath:
            s = targetPath[0]
            s.addEnteringTransition(self)
            for r in s.getRegions():
                if (len(targetPath) >= 2 and
                    r is targetPath[1].getContainingRegion()):
                    # r is a region such that the target state is a
                    # nested substate of r.  Firing self breaks in r.
                    r.addBreakingTransition(self)
                else:
                    # r is a region orthogonal to the target state
                    # vertex or a direct subregion of the target.
                    # Firing self enters the initial pseudostate in r.
                    r.getInitial().addEnteringTransition(self)
            targetPath = targetPath[1:]
        assert s is self.getTarget()

    def processActivity(self):
        """Process the signal parameters, guard, and effect of self."""
        # This function sets self.__activity.
        class_ = self.__machine.getClass()
        ac = self.__machine.getActivityCircuit()
        queue = class_.getEventQueue()
        # By default, the activity of self does nothing.
        self.__activity = activities.Activity(class_, ac)
        signal = self.getTriggerSignal()
        if signal is not None:
            # Notify the event queue.
            queue.addTriggerSignal(signal)
            # Update the activity to reflect the effect of reading the
            # signal parameter values to the fields of this object.
            fields = self.getTriggerParamFields()
            if fields and signal.isExternal():
                raise errors.NotSupportedError('external signal ' +
                                               'parameters not supported')
            for (i, f) in enumerate(fields):
                smtParamValue = queue.varSmtCurrEventParam(ac, None, signal, i)
                self.__activity.writeField(f,
                                           class_.varSmtThisRef(ac, None),
                                           smtParamValue)
        umlGuard = self.__umlTransition.guard
        if umlGuard:
            # Evaluate the guard in a context where the signal
            # parameters are already assigned to fields.
            guard = str(umlGuard.expression.body)
            (smtValue, smtError) = actions.processExpr(
                guard, datatypes.boolType, self.__activity)
            self.__activity.addRuntimeError(smtError)
            evalGuard = ac.add_or(smtValue, smtError)
        else:
            evalGuard = ac.add_true()
        ac[0, None, 'EvalGuard', self] = evalGuard
        guardChooseMap = self.__activity.getChooseMap()
        self.__activity.clearChooseMap()
        umlEffect = self.__umlTransition.effect
        if umlEffect:
            # Update the activity to reflect the action language
            # statements in the effect of self.  As a side effect,
            # notify classes about the signals that are sent to them.
            effect = str(umlEffect.script.body)
            actions.processActions(effect, self.__activity)
        effectChooseMap = self.__activity.getChooseMap()
        self.__guardChooseKeys = guardChooseMap.keys()
        self.__guardChooseKeys.sort()
        for key in self.__guardChooseKeys:
            ac[0, None, 'GuardChoose', self, key] = guardChooseMap[key]
        self.__effectChooseKeys = effectChooseMap.keys()
        self.__effectChooseKeys.sort()
        for key in self.__effectChooseKeys:
            ac[0, None, 'EffectChoose', self, key] = effectChooseMap[key]
        # Define the Read predicate for every instance of all fields
        # read by self.
        for read in self.__activity.getFieldReads():
            f = read.getField()
            for target in f.getOwner().getObjects():
                if read.isLocal() or not options.opt.encode_static:
                    mayRead = [ac.add_eq(smtRef, target.smtReference(ac)) for
                               smtRef in read.smtReferences()]
                else:
                    mayRead = [ac.add_true()]
                ac[0, None, 'Read', self, target, f] = ac.add_and(
                    self.varSmtFire(ac, None), ac.add_or(*mayRead))
        # Define the Write predicate for every instance of all fields
        # written by self.
        for write in self.__activity.getFieldWrites():
            f = write.getField()
            for target in f.getOwner().getObjects():
                if write.isLocal() or not options.opt.encode_static:
                    mayWrite = [ac.add_eq(smtRef, target.smtReference(ac)) for
                                smtRef in write.smtReferences()]
                else:
                    mayWrite = [ac.add_true()]
                ac[0, None, 'Write', self, target, f] = ac.add_and(
                    self.varSmtFire(ac, None), ac.add_or(*mayWrite))
                ac[0, None, 'Valuation', self, target, f] = (
                    write.smtValueIn(target))
                # Could be optimized in some cases by replacing
                # valuation by expression.
        for send in self.__activity.getSends():
            for target in send.getTargetClass().getObjects():
                toTarget = ac.add_eq(send.smtTarget(), target.smtReference(ac))
                if send.isLocal() or not options.opt.encode_static:
                    # Send(self, target) is true iff self is fired and
                    # the reference expression in the send statement
                    # evaluates to target.  Send(self, target) implies
                    # reception of the signal in target.
                    maySendToTarget = toTarget
                    isExact = ac.add_true()
                else:
                    # Send(self, target) is true iff self is fired,
                    # regardless of the reference expression in the
                    # send statement.  Dynamically, if the reference
                    # expression evaluates to target, then Send(self,
                    # target) implies reception of the signal in
                    # target.  If not, then Send(self, target) implies
                    # that target does not receive any signal.
                    maySendToTarget = ac.add_true()
                    isExact = toTarget
                ac[0, None, 'Send', self, target] = ac.add_and(
                    self.varSmtFire(ac, None), maySendToTarget)
                ac[0, None, 'SendExact', self, target] = isExact
            for (i, smtArg) in enumerate(send.smtArgs()):
                ac[0, None, 'SendArg', self, send.getTargetClass(), i] = smtArg
        ac[0, None, 'ActivityConstraints', self] = ac.add_and(
            *self.__activity.getConstraints())
        ac[0, None, 'Assertions', self] = ac.add_and(
            *self.__activity.getAssertions())
        ac[0, None, 'RuntimeErrors', self] = ac.add_or(
            *self.__activity.getRuntimeErrors())

    def smvId(self):
        """Return the id string."""
        return self.__id

    def smtActivityConstraints(self, actor):
        """Return the predicate ActivityConstraints."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, actor, 'ActivityConstraints', self)
        if label not in mc:
            self.__machine.defineActivities(actor)
        return mc[label]

    def smtAssertions(self, actor):
        """Return the predicate Assertions."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, actor, 'Assertions', self)
        if label not in mc:
            self.__machine.defineActivities(actor)
        return mc[label]

    def smtRuntimeErrors(self, actor):
        """Return the predicate RuntimeErrors."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, actor, 'RuntimeErrors', self)
        if label not in mc:
            self.__machine.defineActivities(actor)
        return mc[label]

    def smtEvalGuard(self, actor):
        """Return the predicate EvalGuard."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, actor, 'EvalGuard', self)
        if label not in mc:
            self.__machine.defineActivities(actor)
        return mc[label]

    def smtRead(self, actor, target, field):
        """Return the predicate Read."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, actor, 'Read', self, target, field)
        if label not in mc:
            self.__machine.defineActivities(actor)
        return mc[label]

    def smtWrite(self, actor, target, field):
        """Return the predicate Write."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, actor, 'Write', self, target, field)
        if label not in mc:
            self.__machine.defineActivities(actor)
        return mc[label]

    def smtValuation(self, actor, target, field):
        """Return the function Valuation."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, actor, 'Valuation', self, target, field)
        if label not in mc:
            self.__machine.defineActivities(actor)
        return mc[label]

    def smtSend(self, actor, target):
        """Return the predicate Send."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, actor, 'Send', self, target)
        if label not in mc:
            self.__machine.defineActivities(actor)
        return mc[label]

    def smtSendExact(self, actor, target):
        """Return the predicate SendExact."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, actor, 'SendExact', self, target)
        if label not in mc:
            self.__machine.defineActivities(actor)
        return mc[label]

    def smtSendArg(self, actor, targetClass, paramIndex):
        """Return the function SendArg."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, actor, 'SendArg', self, targetClass, paramIndex)
        if label not in mc:
            self.__machine.defineActivities(actor)
        return mc[label]

    def smtGuardChoose(self, actor, key):
        """Return the function GuardChoose."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, actor, 'GuardChoose', self, key)
        if label not in mc:
            self.__machine.defineActivities(actor)
        return mc[label]

    def smtEffectChoose(self, actor, key):
        """Return the function EffectChoose."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, actor, 'EffectChoose', self, key)
        if label not in mc:
            self.__machine.defineActivities(actor)
        return mc[label]

    def smtFire(self, actor):
        """Return the predicate Fire."""
        mc = self.__machine.getClass().getModel().getCircuit()
        label = (0, actor, 'Fire', self)
        if label not in mc:
            mc[label] = mc.add_var()
        return mc[label]

    def varSmtFire(self, circuit, actor):
        """Return a variable Gate for the predicate Fire."""
        label = (0, actor, 'Fire', self)
        if label not in circuit:
            circuit[label] = circuit.add_var()
        return circuit[label]

    def encodeControl(self, actor):
        """Generate constraints for transition specific control logic."""
        model = self.__machine.getClass().getModel()
        mc = model.getCircuit()
        if self.isCompletion():
            source = self.getSource()
            if source.isPseudostate():
                enabled = mc.add_and(source.smtActive(actor),
                                     self.smtEvalGuard(actor))
            else:
                enabled = mc.add_and(
                    mc.add_not(self.__machine.smtInPseudostate(actor)),
                    source.smtCompleted(actor),
                    self.smtEvalGuard(actor))
                # Could be optimized.
            model.addTransImpl(self.smtFire(actor), enabled)
        else:
            model.addTransImpl(self.smtFire(actor), self.smtEvalGuard(actor))
            # Could be optimized.

    def encodeActivity(self, actor):
        """Encode constraints for the results of the activity of self."""
        model = self.__machine.getClass().getModel()
        mc = model.getCircuit()
        model.addTrans(self.smtActivityConstraints(actor))
        for write in self.__activity.getFieldWrites():
            f = write.getField()
            for target in f.getOwner().getObjects():
                # If Write is true, then the next value of target.f
                # must be equal to the value of target.f in the
                # valuation obtained by executing the actions of the
                # transition.
                fixNext = mc.add_eq(f.smtNextFieldValue(target),
                                    self.smtValuation(actor, target, f))
                model.addTransImpl(self.smtWrite(actor, target, f), fixNext)
        for send in self.__activity.getSends():
            targetClass = send.getTargetClass()
            queue = targetClass.getEventQueue()
            signal = send.getSignal()
            for target in targetClass.getObjects():
                smtArgs = [self.smtSendArg(actor, targetClass, i) for
                           i in xrange(len(signal.getParamTypes()))]
                smtSend = self.smtSend(actor, target)
                smtExact = self.smtSendExact(actor, target)
                model.addTransImpl(mc.add_and(smtSend, smtExact),
                                   queue.smtReception(target, signal, smtArgs))
                model.addTransImpl(mc.add_and(smtSend, mc.add_not(smtExact)),
                                   queue.smtNoReception(target))
        if options.opt.check_assertions:
            model.addTransSpec(mc.add_implies(self.smtFire(actor),
                                              self.smtAssertions(actor)))
        if options.opt.check_runtime_errors:
            model.addTransSpec(mc.add_implies(
                self.smtFire(actor), mc.add_not(self.smtRuntimeErrors(actor))))
