# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Generic recursive traversal of action language expressions.

Author: Heikki Tauriainen
"""

import actlang

# Exceptions for signaling unregistered callbacks.

class UnregisteredCallbackError(Exception):
    """Exception raised on attempts to access unregistered callbacks."""

    def __init__(self, expr):
        """Initialize an UnregisteredCallbackError."""
        self.expr = expr

    def __str__(self):
        """Return the string representation of the error."""
        return "no callback defined for expression `%s' of type %s" \
               % (str(self.expr), self.expr.__class__.__name__)

class UnregisteredPreCallbackError(UnregisteredCallbackError):
    """Exception raised on attempts to access unregistered pre-callbacks."""
    pass

class UnregisteredPostCallbackError(UnregisteredCallbackError):
    """Exception raised on attempts to access unregistered post-callbacks."""
    pass

# Exception for preventing recursive traversal of the subexpressions of an
# expression.

class Return(Exception):
    """Exception for making traversal of an expression return immediately."""

    def __init__(self, value=None):
        """Initialize a Return exception with a return value."""
        self.value = value

# Predefined default callbacks.

def raise_exception_pre(expr, *args, **kwargs):
    """Raise an UnregisteredPreCallbackError."""
    raise UnregisteredPreCallbackError(expr)

def raise_exception_post(expr, *args, **kwargs):
    """Raise an UnregisteredPostCallbackError."""
    raise UnregisteredPostCallbackError(expr)

def ignore(expr, *args, **kwargs):
    """Do nothing."""
    return None

# The ExpressionTraverser class.

class ExpressionTraverser:
    """Traverse the syntax tree of an action language expression."""

    # Dictionary of functions for accessing subexpressions of action
    # language expression objects (in syntactic left-to-right order).
    __accessors = {
        actlang.expr.ArrayAccessExpr :
        (actlang.expr.ArrayAccessExpr.api_getArrayExpr,
         actlang.expr.ArrayAccessExpr.api_getIndexExpr),

        actlang.expr.ArrayCreationExpr :
        (actlang.expr.ArrayCreationExpr.api_getDimensions,),

        actlang.expr.ArrayInitializer :
        (actlang.expr.ArrayInitializer.api_getComponentInitializers,),

        actlang.expr.BitNotExpr :
        (actlang.expr.BitNotExpr.api_getOperand,),

        actlang.expr.BoolLiteralExpr :
        (),

        actlang.expr.CompoundAssignmentExpr :
        (actlang.expr.CompoundAssignmentExpr.api_getLhsExpr,
         actlang.expr.CompoundAssignmentExpr.api_getRhsExpr),

        actlang.expr.CondAndExpr :
        (actlang.expr.CondAndExpr.api_getLeftOperand,
         actlang.expr.CondAndExpr.api_getRightOperand),

        actlang.expr.CondOrExpr :
        (actlang.expr.CondOrExpr.api_getLeftOperand,
         actlang.expr.CondOrExpr.api_getRightOperand),

        actlang.expr.ExprInitializer :
        (actlang.expr.ExprInitializer.api_getExpr,),

        actlang.expr.InfixExpr :
        (actlang.expr.InfixExpr.api_getLeftOperand,
         actlang.expr.InfixExpr.api_getRightOperand),

        actlang.expr.InstanceCreationExpr :
        (actlang.expr.InstanceCreationExpr.api_getArgList,),

        actlang.expr.InstanceOfExpr :
        (actlang.expr.InstanceOfExpr.api_getOperand,),

        actlang.expr.IntLiteralExpr :
        (),

        actlang.expr.LogicalNotExpr :
        (actlang.expr.LogicalNotExpr.api_getOperand,),

        actlang.expr.NameCastExpr :
        (actlang.expr.NameCastExpr.api_getTypeExpr,
         actlang.expr.NameCastExpr.api_getOperand),

        actlang.expr.NameMethodInvocationExpr :
        (actlang.expr.NameMethodInvocationExpr.api_getArgList,),

        actlang.expr.NullExpr :
        (),

        actlang.expr.ParenExpr :
        (actlang.expr.ParenExpr.api_getInnerExpr,),

        actlang.expr.PrimaryFieldAccessExpr :
        (actlang.expr.PrimaryFieldAccessExpr.api_getReferenceExpr,),

        actlang.expr.PrimaryMethodInvocationExpr :
        (actlang.expr.PrimaryMethodInvocationExpr.api_getReferenceExpr,
         actlang.expr.PrimaryMethodInvocationExpr.api_getArgList),

        actlang.expr.SimpleAssignmentExpr :
        (actlang.expr.SimpleAssignmentExpr.api_getLhsExpr,
         actlang.expr.SimpleAssignmentExpr.api_getRhsExpr),

        actlang.expr.StringLiteralExpr :
        (),

        actlang.expr.SuperFieldAccessExpr :
        (),

        actlang.expr.SuperMethodInvocationExpr :
        (actlang.expr.SuperMethodInvocationExpr.api_getArgList,),

        actlang.expr.TernaryCondExpr :
        (actlang.expr.TernaryCondExpr.api_getCondition,
         actlang.expr.TernaryCondExpr.api_getTrueExpr,
         actlang.expr.TernaryCondExpr.api_getFalseExpr),

        actlang.expr.ThisExpr :
        (),

        actlang.expr.TypeCastExpr :
        (actlang.expr.TypeCastExpr.api_getOperand,),

        actlang.expr.UnaryMinusExpr :
        (actlang.expr.UnaryMinusExpr.api_getOperand,),

        actlang.expr.UnaryPlusExpr :
        (actlang.expr.UnaryPlusExpr.api_getOperand,),

        actlang.names.SimpleName:
        (),

        actlang.names.QualifiedName:
        ()
        }        
    
    def __init__(self, default_pre_callback=raise_exception_pre, default_post_callback=raise_exception_post):
        """Initialize an ExpressionTraverser with default callbacks.

        Arguments:
            default_pre_callback, -- Default callback functions to use
            default_post_callback    during expression traversal for action
                                     language expression classes with no
                                     explicitly registered callback
                                     functions.  The following pre-defined
                                     default callbacks are provided for
                                     convenience:
                                       raise_exception_pre:
                                         Abort the expression traversal by
                                         raising an
                                         UnregisteredPreCallbackException.
                                       raise_exception_post:
                                         Abort the expression traversal by
                                         raising an
                                         UnregisteredPostCallbackException.
                                       ignore
                                         Do nothing.
        """
        if not hasattr(default_pre_callback, "__call__"):
            raise TypeError("expected a callable object as first argument")
        if not hasattr(default_post_callback, "__call__"):
            raise TypeError("expected a callable object as second argument")
        self.__pre_callbacks = {"default" : default_pre_callback}
        self.__post_callbacks = {"default" : default_post_callback}

    def register_pre_callback(self, cls, callback):
        """Associate a callback function with action language expression(s).
        
        Arguments:
            cls      -- An action language expression class or an iterable
                        of classes.  (If the argument is an iterable, the
                        given callback function will be associated with all
                        the classes in the iterable.)  Also the following
                        special values are recognized:
                            "entry"   -- (Re)register an additional
                                         pre-callback function to be called
                                         immediately after recursing to an
                                         expression, before applying the
                                         pre-callback registered with the
                                         type of the expression.  (This
                                         callback can be used to collect
                                         code to be run on entry to an
                                         expression.)
                            "default" -- Change the default pre-callback to
                                         be used for all classes with no
                                         explicitly registered callbacks.
            callback -- Callback function accepting an actlang.expr.Expr
                        object as its first parameter.  See documentation
                        for __call__ for more information.
        """
        self.__register_callback(cls, callback, self.__pre_callbacks)

    def register_post_callback(self, cls, callback):
        """Associate a callback function with action language expression(s).

        Arguments:
            cls      -- An action language expression class or an iterable
                        of classes.  (If the argument is an iterable, the
                        given callback function will be associated with all
                        the classes in the iterable.)  Also the following
                        special values are recognized:
                            "default" -- Change the default post-callback
                                         to be used for all classes with no
                                         explicitly registered callbacks.
                            "exit"    -- (Re)register an additional
                                         post-callback function to be called
                                         immediately before backtracking
                                         from an expression, after applying
                                         the post-callback registered with
                                         the type of the expression.  (This
                                         callback can be used to collect
                                         code to be run when exiting from
                                         an expression.)
            callback -- Callback function accepting an actlang.expr.Expr
                        object as its first parameter and a list as the
                        second parameter.  See documentation for __call__
                        for more information.
        """
        self.__register_callback(cls, callback, self.__post_callbacks)

    def __register_callback(self, cls, callback, callbacks):
        """Associate a callback function with action language expression(s).

        Arguments:
           cls       -- An action language expression class or an iterable
                        of classes.  (If the argument is an iterable, the
                        given callback function will be associated with all
                        the classes in the iterable.)
           callback  -- Callback function.
           callbacks -- A dictionary of callbacks in which to associate
                        the class with the callback.
        """
        if not hasattr(callback, "__call__"):
            raise TypeError("expected a callable object as second argument")
        if hasattr(cls, "__iter__"):
            for c in cls:
                callbacks[c] = callback
        else:
            callbacks[cls] = callback

    def has_pre_callback(self, cls):
        """Test whether there is a pre-callback associated with a class.

        Argument:
            cls -- An action language expression class.
        """
        return cls in self.__pre_callbacks

    def has_post_callback(self, cls):
        """Test whether there is a post-callback associated with a class.

        Argument:
            cls -- An action language expression class.
        """
        return cls in self.__post_callbacks

    def unregister_pre_callback(self, cls):
        """Unregister a pre-callback function.

        Argument:
           cls -- An action language expression class or an iterable of
                  classes.  (If the argument is an iterable, callbacks
                  will be unregistered from all the classes in the
                  iterable.)  Unregistering a pre-callback for a class
                  will revert the ExpressionTraverser to using the
                  default pre-callback for the class.
        """
        self.__unregister_callback(self, cls, self.__pre_callbacks)

    def unregister_post_callback(self, cls):
        """Unregister a post-callback function.

        Argument:
           cls -- An action language expression class or an iterable of
                  classes.  (If the argument is an iterable, callbacks
                  will be unregistered from all the classes in the
                  iterable.)  Unregistering a post-callback for a class
                  will revert the ExpressionTraverser to using the
                  default post-callback for the class.
        """
        self.__unregister_callback(self, cls, self.__post_callbacks)

    def __unregister_callback(self, cls, callbacks):
        """Unregister a callback function.

        Arguments:
            cls       -- An action language expression class or an iterable
                         of classes.  (If the argument is an iterable,
                         callbacks will be unregistered from all the classes
                         in the iterable.)
            callbacks -- A dictionary of callbacks.
        """
        if cls == "default":
            raise ValueError("cannot unregister default callback")
        if hasattr(cls, "__iter__"):
            for c in cls:
                if c in callbacks:
                    del callbacks[c]
        else:
            if cls in callbacks:
                del callbacks[cls]

    def __call__(self, expr, *args, **kwargs):
        """Traverse the syntax tree of an action language expression.

        Arguments:
            expr   -- An actlang.expr.Expr object.
            args   -- Optional arguments to be passed to each callback
                      function.
            kwargs -- Optional keyword arguments to be passed to each
                      callback function.

        This function will traverse the syntax tree of the action language
        expression `expr' in depth-first order (from "left to right"
        according to the syntactic structure of the expression), applying
        user-defined callback functions to each subexpression in the
        syntax tree of the expression.

        There are two types of callback functions:

            1. Pre-callbacks, which are called at each node in the
               abstract syntax tree of `expr' before recursing to the
               expression's subexpressions.  Every pre-callback function
               will receive the expression `expr' as its first parameter
               followed by any optional arguments in `args' and `kwargs'.

            2. Post-callbacks, which are called at each node in the
               abstract syntax tree of `expr' after recursively traversing
               all of the expression's subexpressions.  Every post-callback
               function will receive the following parameters (in order):
                 - The action language expression `expr'.
                 - For all post-callbacks except for the special "exit"
                   post-callback, a list of values which would be returned
                   by recursive calls to `__call__' on the immediate
                   subexpressions of `expr', generally ordered in the
                   syntactic left-to-right order of the subexpressions.
                   The length and contents of this list depend on type of
                   `expr': for example, if
                     `isinstance(expr, actlang.expr.InfixExpr) == True'
                   holds, the list consists of two elements corresponding
                   to the values which would be returned by `__call__' when
                   applied to the left- and right-hand side subexpressions
                   of `expr' in this order.  (In general, the length of
                   the list depends on the number of API functions for
                   accessing `expr''s immediate subexpressions; see the
                   documentation for actlang.expr.)  If a function used to
                   access subexpressions of `expr' returns a sequence (for
                   example, if `expr' is of type
                       actlang.expr.NameMethodInvocationExpr,
                   for which the API function api_getArgList() returns a
                   sequence of subexpressions instead of a single
                   subexpression), the corresponding element in the value
                   returned by this function will be a list.  For the
                   special "exit" post-callback, see below.
                 - The optional arguments in `args' and `kwargs'.

        If no pre- (post-)callback has been explicitly registered with
        the type of `expr', the default pre- (post-)callback function
        will be used.  The "entry" ("exit") callback (if registered) will
        be called before (respectively after) any type-specific or default
        callbacks.

        If all callback functions applied to the subexpressions of `expr'
        return normally and no callback applied to `expr' itself raises an
        exception, either, this function will have the value returned by
        the post-callback function applied to `expr' as its return value.
        If registered, the special "exit" post-callback called after the
        type-specific callback will receive this value as its second
        parameter.  (Note that the return value of the "exit" callback
        itself is thus normally ignored; see below for instructions on how
        to change this behavior.)

        Optionally, any callback function (including the "exit" post-
        callback) can raise a Return exception to modify the value which
        would (in effect) be returned by this function when applied
        to a subexpression of `expr'.  The Return exception can be
        initialized with an optional value which should act as the new
        return value.  Raising a Return exception from within a pre-
        callback function will also prevent the syntax tree traversal
        from recursing into any subexpression of the expression given to
        the pre-callback function as the first parameter.

        If a Return exception is raised in a pre-callback or a post-
        callback function other than the "exit" post-callback, the "exit"
        callback will still get called appropriately also in this case,
        i.e., before the syntax tree traversal backtracks from the
        expression given to the callback function as a parameter.  (The
        "exit" callback will receive the return value associated with the
        exception as its second parameter in this case.)  A further Return
        exception raised in the "exit" callback can be still used to
        modify the return value as usual.
        """
        if not isinstance(expr, actlang.expr.Expr):
            raise TypeError \
                  ("an actlang.expr.Expr object expected as first argument")

        # Perform syntax tree traversal using a stack of triples.  The first
        # element of each triple indicates whether the iterator in the second
        # element of the triple refers to a sequence of subexpression accessor
        # functions associated with a particular expression type (in which case
        # the first element of the triple is of type actlang.expr.Expr), or to
        # a sequence of subexpressions returned by some subexpression accessor
        # function (in which case the first element of the triple has the value
        # None).  The third element of the triple is used to collect return
        # values of post-callback functions (or the values associated with
        # Return exceptions raised by callback functions).
        (expr, subexpr_accessors, result) = (None, (expr,).__iter__(), [])
        expr_stack = [(expr, subexpr_accessors, result)]
        while True:
            try:
                try:
                    # begin processing next subexpression or sequence of
                    # subexpressions
                    if expr != None:
                        expr = subexpr_accessors.next()(expr)
                    else:
                        expr = subexpr_accessors.next()
                    if isinstance(expr, actlang.expr.Expr):
                        # object to process is an language expression `expr'
                        # apply the "entry" callback (if registered) to `expr'
                        try:
                            self.__pre_callbacks["entry"] \
                                (expr, *args, **kwargs)
                        except KeyError:
                            pass
                        # apply the class-specific pre-callback, or the default
                        # pre-callback to `expr'
                        try:
                            callback = self.__pre_callbacks[expr.__class__]
                        except KeyError:
                            callback = self.__pre_callbacks["default"]
                        callback(expr, *args, **kwargs)
                        (subexpr_accessors, result) = \
                            (self.__accessors[expr.__class__].__iter__(), [])
                    elif isinstance(expr, (list, tuple)):
                        # object to process is a sequence of subexpressions
                        (expr, subexpr_accessors, result) = \
                            (None, expr.__iter__(), [])
                    else:
                        raise TypeError \
                                  ("object to process is of unexpected type")
                    expr_stack.append((expr, subexpr_accessors, result))
                    continue
                except StopIteration:
                    # all subexpressions processed normally
                    expr_stack.pop()
                    if expr != None:
                        # if `expr' is a real expression, apply the
                        # class-specific post-callback, or the default
                        # post-callback to `expr'
                        try:
                            callback  = self.__post_callbacks[expr.__class__]
                        except KeyError:
                            callback = self.__post_callbacks["default"]
                        retval = callback(expr, result, *args, **kwargs)
                    else:
                        retval = result
            except Return, e:
                # pre- or post-callback raised a Return exception
                retval = e.value
            # backtrack
            if expr != None:
                # if `expr' is a real expression, apply the "exit" callback (if
                # registered) to `expr'
                try:
                    self.__post_callbacks["exit"] \
                        (expr, retval, *args, **kwargs)
                except KeyError:
                    pass
                except Return, e:
                    # "exit" callback raised a Return exception, change the
                    # value to be propagated back to upper level
                    retval = e.value
            if not expr_stack:
                return retval[0]
            (expr, subexpr_accessors, result) = expr_stack[-1]
            result.append(retval)

# -----------------------------------------------------------------------------
# Translate action language expressions into Python.

# Helper function for truncating the values of unbounded Python integer
# expressions to 32 bits.

def __truncateTo32Bits(e):
    """Helper function for translating Jumbala expressions into Python."""
    return "(lambda e: ((e & 0X80000000) and " \
           "[(e & 0XFFFFFFFF) - 0X100000000L] or " \
           "[e & 0XFFFFFFFF])[0])(%s)" % e

def __createTranslator():

    def infix(expr, subexprs):
        op = expr.api_getOperator()

        # Translate the integer arithmetic operators to Python, following the
        # Jumbala action language semantics (following actlang/icode.py).
        if isinstance(op, actlang.binaryop.Divide):
            return __truncateTo32Bits \
                   ("(lambda a, b: b == 0 and (1 // 0) or ((a < 0) == (b < 0) "
                    "and [a // b] or [-(a // b)])[0])(%s, %s)"
                    % (subexprs[0], subexprs[1]))
        elif isinstance(op, actlang.binaryop.Remainder):
            return __truncateTo32Bits \
                   ("(lambda a, b: b == 0 and (1 %% 0) or ((a < 0) == (b < 0) "
                    "and [a %% b] or [a %% -b])[0])(%s, %s)"
                    % (subexprs[0], subexprs[1]))
        elif isinstance(op, actlang.binaryop.ShiftLeft):
            return __truncateTo32Bits \
                   ("(lambda a, b: b < 0 and (1 << -1) or "
                    "(b >= 32 and [0] or [long(a) << b])[0])(%s, %s)"
                    % (subexprs[0], subexprs[1]))
        elif isinstance(op, actlang.binaryop.SignedShiftRight):
            return __truncateTo32Bits \
                   ("(lambda a, b: b < 0 and (1 >> -1) or "
                    "(b >= 32 and [a < 0 and -1 or 0] or [a >> b])[0])(%s, %s)"
                    % (subexprs[0], subexprs[1]))
        elif isinstance(op, actlang.binaryop.UnsignedShiftRight):
            return __truncateTo32Bits \
                   ("(lambda a, b: b < 0 and (1 >> -1) or "
                    "(b >= 32 and [0] or "
                    "[a < 0 and (a + 0X100000000L) or a >> b])[0])(%s, %s)"
                    % (subexprs[0], subexprs[1]))
        elif isinstance(op, (actlang.binaryop.Equal,
                             actlang.binaryop.NumericalComparison)):
            return ("%s %s %s" % (subexprs[0], str(op), subexprs[1]))
        else: # op in ("+", "-", "*", "&", "|", "^")
            return __truncateTo32Bits \
                   ("%s %s %s" % (subexprs[0], str(op), subexprs[1]))

    callbacks = (
        (actlang.expr.BoolLiteralExpr,
         lambda expr, *rest: (expr.api_getValue() and "True") or "False"),
        (actlang.expr.IntLiteralExpr,
         lambda expr, *rest:
         str(eval(__truncateTo32Bits(expr.api_getValue())))),
        (actlang.names.SimpleName, lambda expr, *rest: str(expr)),
        (actlang.expr.ParenExpr,
         lambda expr, subexprs: "(" + subexprs[0] + ")"),
        (actlang.expr.TernaryCondExpr,
         lambda expr, subexprs:
         "((%s and [%s]) or [%s])[0]" % tuple(subexprs)),
        (actlang.expr.UnaryPlusExpr,
         lambda expr, subexprs: "(+%s)" % tuple(subexprs)),
        (actlang.expr.LogicalNotExpr,
         lambda expr, subexprs: "(!%s)" % tuple(subexprs)),
        (actlang.expr.UnaryMinusExpr,
         lambda expr, subexprs: __truncateTo32Bits("-" + subexprs[0])),
        (actlang.expr.BitNotExpr,
         lambda expr, subexprs: __truncateTo32Bits("~" + subexprs[0])),
        (actlang.expr.CondAndExpr,
         lambda expr, subexprs: "(%s and %s)" % tuple(subexprs)),
        (actlang.expr.CondOrExpr,
         lambda expr, subexprs: "(%s or %s)" % tuple(subexprs)),
        (actlang.expr.InfixExpr, infix))
    translator = ExpressionTraverser(ignore)
    for (cls, f) in callbacks:
        translator.register_post_callback(cls, f)
    return translator

# Function object for translating an action language expression into a Python
# expression, preserving the semantics of the action language (in particular,
# the wrapping semantics on signed integer overflow).

__translate = __createTranslator()
del __createTranslator

def actlang_to_python(expr):
    """Translate a side-effect free Jumbala expression into Python.

    Argument:
        expr -- An actlang.expr.Expr object.

    Returns a string representing the translation of `expr' into Python
    using the semantics of the Jumbala action language.
    """
    return __translate(expr)

# -----------------------------------------------------------------------------
# Convert a subexpression in an expression into a location and vice versa.

class __Found(Exception):
    """Exception for reporting return values."""

    def __init__(self, value):
        self.value = value

def __createExprFinder():
    """Create a function object for finding a subexpr. in an expression."""

    def recurse(expr, target_expr, path):
        if expr == target_expr:
            raise __Found(tuple(path))
        path.append(0)

    def backtrack(expr, result, target_expr, path):
        path.pop()
        if path:
            path[-1] += 1

    return ExpressionTraverser(recurse, backtrack)

__find_subexpr_in_expr = __createExprFinder()
del __createExprFinder

def subexpression_to_path(expr, subexpr):
    """Identify the location of a subexpression in an expression.

    Arguments:
        expr    -- An actlang.expr.Expr object.
        subexpr -- An actlang.expr.Expr object, which should be a
                   subexpression of `expr' (i.e., an object referred to
                   in the action language expression tree rooted at
                   `expr').

    If `subexpr' is a subexpression of `expr', returns a tuple of integers
    which identifies a location of `subexpr' in `expr'.  (If `expr'
    contains multiple references to `subexpr', one of the locations of
    these references is returned.)  This tuple can be converted back to the
    action language expression `subexpr' by calling
    `path_to_subexpression'.

    Raises a ValueError if `subexpr' is not a subexpression of `expr'.
    """
    if not isinstance(expr, actlang.expr.Expr):
        raise TypeError \
                  ("an actlang.expr.Expr object expected as first argument")
    if not isinstance(subexpr, actlang.expr.Expr):
        raise TypeError \
                  ("an actlang.expr.Expr object expected as second argument")
    try:
        __find_subexpr_in_expr(expr, subexpr, [])
    except __Found, e:
        return e.value
    else:
        raise ValueError("subexpression not found in expression")

def __createPathWalker():
    """Create a function object for traversing a path in a syntax tree."""
    
    def recurse(expr, path):
        if path[0] > 0:
            path[0] -= 1
            raise Return()
        else:
            path.pop(0)
            if not path:
                raise __Found(expr)

    def raiseException(expr, x, path):
        raise ValueError("invalid location")

    return ExpressionTraverser(recurse, raiseException)

__walk_path_in_expr = __createPathWalker()
del __createPathWalker

def path_to_subexpression(expr, path):
    """Find a subexpression corresponding to a location.

    Arguments:
        expr -- An actlang.expr.Expr object.
        path -- A sequence of integers identifying a subexpression of
                `expr' (as returned, for example, by
                `subexpression_to_path').

    If `path' identifies a valid location in the syntax tree of `expr',
    returns the action language expression (as an actlang.expr.Expr object)
    corresponding to that location.  This action language expression can
    be converted back to a location by calling `subexpression_to_path'.

    Raises a ValueError if `path' does not correspond to a valid location
    in the syntax tree of `expr'.
    """
    if not isinstance(expr, actlang.expr.Expr):
        raise TypeError \
                  ("an actlang.expr.Expr object expected as first argument")
    if not isinstance(path, (list, tuple)):
        raise TypeError("a sequence of integers expected as second argument")
    for p in path:
        if not isinstance(p, int):
            raise ValueError("invalid location")
    p = [0]
    p.extend(path)
    try:
        __walk_path_in_expr(expr, p)
    except __Found, e:
        return e.value
    else:
        raise ValueError("invalid location")

# -----------------------------------------------------------------------------
# Substitute expressions for variables in an expression.

def __createSubstitutor():
    """Generate a func. object for substituting variables for expressions."""
    
    # Helper function for detecting whether any substitutions were made in a
    # subexpression.

    def checkForRewrittenSubexprs(expr, subexprs):
        for (e, r) in subexprs:
            if r:
                return
        raise Return((expr, False))

    # Post-callbacks.  Each callback takes an actlang.expr.Expr object, a list
    # of subresults, and a dictionary of replacements and returns a pair of the
    # form
    #     (<expr>, <rewritten>)
    # where <expr> is the result of substituting variables in <expr> using
    # replacements from the replacement dictionary, and <rewritten> is a
    # Boolean value telling whether any substitutions were actually made in
    # the expression.

    def identifier(expr, subexprs, replacements):
        try:
            return (replacements[str(expr)], True)
        except KeyError:
            return (expr, False)

    def method(expr, subexprs, replacements):
        checkForRewrittenSubexprs(expr, subexprs[0])
        args = [s[0] for s in subexprs[0]]
        return (expr.__class__ \
                    (expr.api_getInvocationName(),
                     [s[0] for s in subexprs[0]],
                     (0, 0)),
                True)

    def defaultWithLineSpan(expr, subexprs, replacements):
        checkForRewrittenSubexprs(expr, subexprs)
        args = [s[0] for s in subexprs]
        args.append((0, 0))
        return (expr.__class__(*args), True)

    def defaultWithoutLineSpan(expr, subexprs, replacements):
        checkForRewrittenSubexprs(expr, subexprs)
        return (expr.__class__(*[s[0] for s in subexprs]), True)

    def infix(expr, subexprs, replacements):
        checkForRewrittenSubexprs(expr, subexprs)
        return (actlang.expr.InfixExpr \
                    (subexprs[0][0], subexprs[1][0], expr.api_getOperator()),
                True)

    e = actlang.expr
    n = actlang.names
    post_callbacks = \
        (((e.BoolLiteralExpr, e.IntLiteralExpr, n.QualifiedName,
           e.PrimaryFieldAccessExpr, e.ThisExpr, e.InstanceCreationExpr),
          lambda expr, *rest: (expr, False)),
         (n.SimpleName, identifier),
         (e.NameMethodInvocationExpr, method),
         ((e.ParenExpr, e.UnaryPlusExpr, e.UnaryMinusExpr, e.BitNotExpr,
           e.LogicalNotExpr, e.SimpleAssignmentExpr), defaultWithLineSpan),
         ((e.CondAndExpr, e.CondOrExpr, e.TernaryCondExpr),
          defaultWithoutLineSpan),
         (e.InfixExpr, infix))
    substitutor = ExpressionTraverser(ignore)
    for (cls, f) in post_callbacks:
        substitutor.register_post_callback(cls, f)
    return substitutor

__substitute = __createSubstitutor()
del __createSubstitutor

def substitute_exprs(expr, replacements):
    """Substitute expression(s) for variables in an expression.

    Arguments:
        expr         -- An actlang.expr.Expr object.
        replacements -- A mapping from strings (names of variables referred
                        to in `expr') to actlang.expr.Expr objects.

    Returns a pair consisting of an actlang.expr.Expr object and a truth
    value telling whether any substitution was actually made in the
    expression.  The returned action language expression represents `expr'
    with each occurrence of a variable (represented by an an
    actlang.names.SimpleName object in the syntax tree of `expr') whose
    name is included as a key in `replacements' with the corresponding
    subexpression in `replacements'.  No substitution is applied to
    variables not included as keys in this dictionary.

    This function currently supports only a subset of action language
    expressions (mainly those used with model abstraction).
    """
    if not isinstance(expr, actlang.expr.Expr):
        raise TypeError \
                  ("an actlang.expr.Expr object expected as first argument")
    if not isinstance(replacements, dict):
        raise TypeError("a dictionary expected as second argument")
    for (k, v) in replacements.iteritems():
        if not isinstance(k, str) or not isinstance(v, actlang.expr.Expr):
            raise ValueError \
                      ("the dictionary given as second argument must "
                       "represent a mapping from strings to actlang.expr.Expr "
                       "objects")
    return __substitute(expr, replacements)
