"""
Small utilities for Coral models.

Author: Tommi Junttila

Copyright (c) 2006-2007 Helsinki University of Technology
Released under the GNU General Public License version 2
"""

import sys
import coral
import Utils.String

def xmi_sort(element_list):
    """
    Sort the list of Coral Elements according to the XMI ids.
    """
    assert type(element_list) is list
    element_list.sort(lambda x,y: cmp(str(x.__XMIid__()),str(y.__XMIid__())))


def get_qualified_name(element, anon_name="<<anon>>"):
    """
    Returns the qualified name of a Coral Element 'element'.
    If an element in the path does not have a name,
    'anon_name' is used instead.
    """
    assert isinstance(element, coral.kernel.Element)
    if element.__hasSlotByName__("name") and element.name:
        name = str(element.name)
    else:
        name = anon_name
    parent = element.__parent__()
    if parent:
        prefix = get_qualified_name(parent, anon_name) + "::"
    else:
        prefix = ""
    return prefix + name


def get_elements_of_type(element, element_type):
    """
    Get the list of Elements owned by the Element 'element'
    (including the 'element' itself) that are of type 'element_type'.
    """
    assert isinstance(element, coral.kernel.Element)
    return [e for e in element.__getAllParts__()
            if e.oclIsKindOf(element_type)]


def get_metaclassname(element):
    """
    Get the name of the metaclass of the Element 'element'.
    """
    assert isinstance(element, coral.kernel.Element)
    return str(element.__type__().getName())


def get_non_null_properties(element):
    """
    Return a list of non-{null,empty,False} non-anon properties of
    the Coral Element 'element'.
    """
    assert isinstance(element, coral.kernel.Element)
    props = []
    pi = element.__type__().__properties__()
    while True:
        try:
            prop = pi.next()
            prop_name = prop.getName()
            if str(prop_name).startswith('__coral_anon'):
                continue
            value = element.__getattr__(prop_name)
            if not value:
                continue
            props.append(prop)
        except StopIteration:
            break
    return props


def dump(element,
         indent = Utils.String.Indent(),
         f = sys.stdout,
         recursively = True):
    """
    Print the Coral Element 'element' in the file 'f'.
    If 'recursively' is True, then all the owned Elements are printed
    recursively, too.
    Mainly for debugging purposes.
    """
    assert isinstance(element, coral.kernel.Element)
    assert isinstance(indent, Utils.String.Indent)
    assert isinstance(f, file)
    assert isinstance(recursively, bool)
    
    f.write(str(indent) + str(element)+"\n")

    props = []
    pi = element.__type__().__properties__()
    while True:
        try:
            prop = pi.next()
            prop_name = str(prop.getName())
            if prop_name.startswith("__coral_anon"):
                continue
            props.append(prop)
        except StopIteration:
            break
    props.sort(lambda a,b: cmp(str(a.getName()),str(b.getName())))

    indent.push()

    # Iterate over properties, printing 'em recursively
    for prop in props:
        prop_name = prop.getName()
        value = element.__getattr__(prop_name)
        # Print composite, non-primitive type properties
        if(prop.isComposition() and
           isinstance(prop.getTarget(), coral.ElementDefinition)):
            f.write(str(indent)+str(prop_name)+" = ")
            if prop.isOrdered():
                f.write("[\n")
            else:
                f.write("{\n")
            indent.push()
            try:
                for v in value:
                    if recursively:
                        dump(v, indent)
                    else:
                        f.write(str(indent)+str(v)+"\n")
            except TypeError:
                if recursively:
                    dump(value, indent)
                else:
                    f.write(str(indent)+str(v)+"\n")
            indent.pop()
            if prop.isOrdered():
                f.write(str(indent)+"]\n")
            else:
                f.write(str(indent)+"}\n")
        else:
            try:
                f.write(str(indent)+str(prop_name)+" = "+str(value)+"\n")
            except UnicodeEncodeError:
                f.write(str(indent)+str(prop_name)+
                        " = <<<SOMETHINGUNPRINTABLE>>>\n")
    indent.pop()

