"""
Some small utility functions for UML 1.4 state machines.

Author: Tommi Junttila

Copyright (c) 2006-2007 Helsinki University of Technology
Released under the GNU General Public License version 2
"""

import sets
import coral
import Utils.Models

UML14 = coral.lang.UML14


def get_outgoing_completion_transitions(sv):
    """
    Get the completion transitions leaving the StateVertex 'sv'.
    """
    assert sv.oclIsKindOf(UML14.StateVertex)
    return [t for t in sv.outgoing if not t.trigger]


def has_outgoing_completion_transitions(sv):
    """
    Check whether the StateVertex 'sv' has outgoing completion transitions.
    """
    return len(get_outgoing_completion_transitions(sv)) > 0


def get_outgoing_triggered_transitions(sv):
    """
    Get the triggered transitions leaving the StateVertex 'sv'.
    """
    assert sv.oclIsKindOf(UML14.StateVertex)
    return [t for t in sv.outgoing if t.trigger]


def has_outgoing_triggered_transitions(sv):
    """
    Check whether the StateVertex 'sv' has outgoing triggered transitions.
    """
    return len(get_outgoing_triggered_transitions(sv)) > 0


def get_interesting_signals(sm):
    """
    Get the set of Signals that the StateMachine 'sm' is interested in,
    i.e. those occurring as (normal or internal) transition triggers or
    in deferrable events.
    """
    assert sm.oclIsKindOf(UML14.StateMachine)

    signals = sets.Set()

    states = Utils.Models.get_elements_of_type(sm, UML14.State)
    for state in states:
        for e in state.deferrableEvent:
            if e.oclIsKindOf(UML14.SignalEvent) and e.signal:
                signals.add(e.signal)

    transitions = Utils.Models.get_elements_of_type(sm, UML14.Transition)
    for t in transitions:
        if (t.trigger and t.trigger.oclIsKindOf(UML14.SignalEvent) and
            t.trigger.signal):
            signals.add(t.trigger.signal)
                
    return signals


def get_parent_path(sv):
    """
    Get the 'parent path' [top,...,sv] from the top state of
    the StateMachine containing the StateVertex 'sv' to 'sv' itself.
    """
    assert sv.oclIsKindOf(UML14.StateVertex)
    path = []
    while sv:
        path.append(sv)
        assert((sv.container == None) ==
               (sv.oclIsKindOf(UML14.State) and sv.stateMachine != None))
        sv = sv.container
    path.reverse()
    return path


def get_state_machine(sv):
    """
    Get the containing StateMachine of the StateVertex 'sv'.
    """
    assert sv.oclIsKindOf(UML14.StateVertex)
    while True:
        if sv.container:
            assert sv.stateMachine == None
            sv = sv.container
            continue
        assert sv.oclIsKindOf(UML14.State)
        assert sv.stateMachine
        return sv.stateMachine


def get_least_common_ancestor_path(sv1, sv2):
    """
    Get the path from top to the least common ancestor of the
    StateVertices 'sv1' and 'sv2' located in the same StateMachine.
    Neither 'sv1' nor 'sv2' can be the top vertex.
    """
    assert sv1.oclIsKindOf(UML14.StateVertex)
    assert sv2.oclIsKindOf(UML14.StateVertex)
    assert sv1.container
    assert sv2.container
    path1 = get_parent_path(sv1)
    path2 = get_parent_path(sv2)
    assert path1[0] == path2[0]
    lca_path = []
    i = 0
    while i < len(path1) and i < len(path2):
        if path1[i] != path2[i]:
            break
        lca_path.append(path1[i])
        i = i + 1
    if lca_path[-1] == sv1 or lca_path[-1] == sv2:
        lca_path.pop()
    assert lca_path
    return lca_path


def get_least_common_ancestor(sv1, sv2):
    """
    Get the least common ancestor of the StateVertices 'sv1' and 'sv2'
    located in the same StateMachine.
    Neither 'sv1' nor 'sv2' can be the top vertex.
    """
    lca_path = get_least_common_ancestor_path(sv1, sv2)
    return lca_path[-1]


def is_completion_sensitive(sv):
    """
    Check whether the UML StateVertex 'sv' is completion sensitive.
    Currently 'sv' is completion sensitive if it is a UML State
    (not a Pseudostate) that has outgoing completion transitions
    and it cannot be deduced that at least one of them is always enabled
    when the state 'sv' is active (and completed if 'sv' is a CompositeState).
    """
    assert sv.oclIsKindOf(UML14.StateVertex)
    if sv.oclIsKindOf(UML14.Pseudostate):
        return False
    if sv.oclIsKindOf(UML14.StubState):
        raise RuntimeError("StubStates are not supported")
    if sv.oclIsKindOf(UML14.SynchState):
        raise RuntimeError("SynchStates are not supported")
    if sv.oclIsKindOf(UML14.State):
        for t in sv.outgoing:
            if t.trigger:
                continue
            if not t.guard:
                # Missing guard is interpreted as 'true'
                continue
            assert t.guard.expression
            assert t.guard.expression.body
            if str(t.guard.expression.body) == 'true':
                continue
            # There is an outgoing completion transition whose guard is not
            # equal to true -> sv is completion sensitive
            return True
        return False
    raise RuntimeError("Internal error")


class Statistics:
    """
    A class for computing some statistics of state machines.
    Usage: create an object with e.g. o = Statistics(sm) and
    then call the methods of o to receive up-to-date information of sm.
    """
    def __init__(self, sm):
        assert sm.oclIsKindOf(UML14.StateMachine)
        self.sm = sm

    def get_nof_state_vertices(self):
        es = Utils.Models.get_elements_of_type(self.sm, UML14.StateVertex)
        return len(es)
    
    def get_nof_simple_states(self):
        es = Utils.Models.get_elements_of_type(self.sm, UML14.SimpleState)
        return len(es)
    
    def get_nof_nonconcurrent_composite_states(self):
        es = Utils.Models.get_elements_of_type(self.sm, UML14.CompositeState)
        es = [e for e in es if not e.isConcurrent]
        return len(es)

    def get_nof_concurrent_composite_states(self):
        es = Utils.Models.get_elements_of_type(self.sm, UML14.CompositeState)
        es = [e for e in es if e.isConcurrent]
        return len(es)
    
    def get_nof_final_states(self):
        es = Utils.Models.get_elements_of_type(self.sm, UML14.FinalState)
        return len(es)
    
    def get_nof_choice_pseudostates(self):
        es = Utils.Models.get_elements_of_type(self.sm, UML14.Pseudostate)
        es = [e for e in es if e.kind == "choice"]
        return len(es)
    
    def get_nof_initial_pseudostates(self):
        es = Utils.Models.get_elements_of_type(self.sm, UML14.Pseudostate)
        es = [e for e in es if e.kind == "initial"]
        return len(es)
    
    def get_nof_transitions(self):
        es = Utils.Models.get_elements_of_type(self.sm, UML14.Transition)
        return len(es)
