"""
Small miscellaneous utilities.

Author: Tommi Junttila

Copyright (c) 2007 Helsinki University of Technology
Released under the GNU General Public License version 2
"""

import sys

class Indent:
    def __init__(self, prefix = "", step = 2):
        self.prefix = prefix
        self.step = step
        self.i = 0;
    def __str__(self):
        return self.prefix+(" "*self.i)
    def push(self):
        self.i += self.step
    def pop(self):
        assert self.i >= self.step
        self.i -= self.step


class Writer:
    def __init__(self, f, prefix = ""):
        """
        Initialize a new writer for the file f.
        If 'f' is None, then the writes are ignored.
        """
        assert (f is None) or (type(f) is file)
        assert type(prefix) is str
        self.f = f
        self.prefix = prefix
        self.indent = 0
        self.indent_stack = []
    def push(self, amount = 2):
        """
        Add amount of indentation.
        If 'amount' is an int, then that many white spaces are added;
        otherwise 'amount' should be a str and the length of 'amount'
        white space is added.
        """
        assert type(amount) is int or type(amount) is str
        if type(amount) is int:
            assert amount >= 0
            self.indent += amount
            self.indent_stack.append(amount)
        elif type(amount) is str:
            self.indent += len(amount)
            self.indent_stack.append(len(amount))
        else:
            assert False
    def pop(self):
        """
        Go back to the previous indentation level.
        """
        assert self.indent_stack
        self.indent -= self.indent_stack.pop()
        assert self.indent >= 0
    def ln(self, s):
        """
        Print an indented line into the file.
        """
        if self.f:
            self.f.write(self.prefix+(" "*self.indent)+str(s)+"\n")
    def wr(self, s):
        """
        Write an unindented string to the file.
        """
        if self.f:
            self.f.write(str(s))
    

def IsInt(s):
    """
    Does the given string 's' represent an integer?
    """
    ok = True
    try:
        num = int(s)
    except ValueError:
        ok = False
    return ok


def length_limited_string(s, l):
    """
    Return the 'l' first letters in the string 's',
    followed by '...' if 's' was longer than 'l' letters.
    """
    assert isinstance(s, str)
    assert isinstance(l, int)
    assert l > 0
    tmp = s[0:l]
    if len(s) > l:
        tmp += "..."
    return tmp
