"""
Small utilities for UML 1.4 Models

Author: Tommi Junttila

Copyright (c) 2006-2007 Helsinki University of Technology
Released under the GNU General Public License version 2
"""

import sys
import re
from sets import Set

import coral

import Utils.Models

UML14 = coral.lang.UML14


def _qn(e):
    # Just an abbreviation
    return Utils.Models.get_qualified_name(e)

def _mcname(e):
    # Just an abbreviation
    return Utils.Models.get_metaclassname(e)

def _err(e, s):
    # Just an abbreviation
    raise RuntimeError(_mcname(e)+" '"+_qn(e)+"': "+s)

def _warn(e, s):
    # Just an abbreviation
    sys.stderr.write("Warning: "+_mcname(e)+" '"+_qn(e)+"': "+s+"\n")



def _warn_skipping_property(element, prop_name):
    print "WARNING: skipping property '"+prop_name+"' "+\
          "in "+_mcname(element)+" "+_qn(element)
    #print "WARNING: skipping property '"+prop_name+"' in '"+str(tau_element)+"'"

def _warn_skipping_ownedMember(element, ownedMember):
    print "WARNING: skipping ownedMember '"+str(ownedMember)+"' "+\
          "in "+_mcname(element)+" "+_qn(element)
    #s = "WARNING: skipping ownedMember '"+str(ownedMember)+"'"
    #s += " in '"+str(tau_element)+"'"
    #print s


def _check_prop_handling(element, handled, ignored, parent_pointers):
    value = True
    for prop in Utils.Models.get_non_null_properties(element):
        prop_name = prop.getName()
        if prop_name in handled:
            pass # Already handled
        elif prop_name in ignored:
            pass # Silently ignored
        elif prop_name in parent_pointers:
            assert element.__getattr__(prop_name) == \
                   element.__parent__()
        else:
            _warn_skipping_property(element, prop_name)
            value = False
    return value


def _cph(element, handled, ignored, parent_pointers):
    """A short-cut for _check_prop_handling."""
    return _check_prop_handling(element, handled, ignored, parent_pointers)



def find_attributes_and_associations(uml_class):
    """
    Find the Attributes and navigable Associations of the Class 'uml_class'.
    Returns a pair of maps, the first being the result of
    find_attributes(uml_class) and the second that of
    find_associations(uml_class).
    Raises a RuntimeError if a name clash or other obvious error
    is encountered.
    Does not perform a thorough well-formedness check, though.
    """
    assert uml_class.oclIsKindOf(UML14.Class)

    attrs = find_attributes(uml_class)
    assocs = find_associations(uml_class)

    for assoc_name in assocs.keys():
        if attrs.has_key(assoc_name):
            _err(uml_class, "Both attribute and association called '"+
                 assoc_name+"'")

    return (attrs, assocs)



def find_attributes(uml_class):
    """
    Find the Attributes of the Class 'uml_class'.
    Returns a map from names to Attributes.
    Raises a RuntimeError if multiple Attributes have the same name.
    Does not perform a thorough well-formedness check, though.
    """
    assert uml_class.oclIsKindOf(UML14.Class)

    attr_map = {}

    for attr in [f for f in uml_class.feature
                 if f.oclIsKindOf(UML14.Attribute)]:
        assert attr.name
        # Check that the name of the Attribute does not appear in other
        # Attribute or Association in this Class
        if attr_map.has_key(str(attr.name)):
            _err(uml_class, "More than attribute called '"+str(attr.name)+"'")
        attr_map[str(attr.name)] = attr
            
    return attr_map



def get_other_association_end(association_end):
    """
    Given a UML AssociationEnd 'association_end',
    return a the other end of the associated Association.
    """
    assert association_end.oclIsKindOf(UML14.AssociationEnd)
    association = association_end.association
    if len(association.connection) != 2:
        _err(association, "an association with "+
             str(len(association.connection))+
             " AssociationEnds instead of two")
    other_association_end = association.connection[0]
    if other_association_end == association_end:
        other_association_end = association.connection[1]
    if other_association_end == association_end:
        _err(association, "an association with the same AssociationEnd")
    return other_association_end



def find_associations(uml_class):
    """
    Find the Associations that are navigable from the Class 'uml_class'.
    Returns a map from names to pairs of AssociationEnds.
    Raises a RuntimeError if multiple navigable AssociationEnds have
    the same name or if an obviously malformed Association is encountered.
    Does not perform a thorough well-formedness check, though.
    """
    assert uml_class.oclIsKindOf(UML14.Class)

    assoc_map = {}

    for assocend in uml_class.association:
        assert assocend.oclIsKindOf(UML14.AssociationEnd)
        assoc = assocend.association
        other_assocend = get_other_association_end(assocend)
        assert other_assocend != assocend
        if not (assocend.isNavigable or other_assocend.isNavigable):
            _warn(assoc, "An association with no navigable end, ignored")
        if not other_assocend.isNavigable:
            continue
        if not other_assocend.name:
            _err(other_assocend,
                 "An AssociationEnd is navigable but has no name")
        if assoc_map.has_key(str(other_assocend.name)):
            _err(uml_class, "More than navigable association end"+
                 "called '"+str(other_assocend.name)+"'")
        assoc_map[str(other_assocend.name)] = (other_assocend, assocend)

    return assoc_map



def _multiplicity_str(m):
    """
    Return the Multiplicity 'm' as a string of form 'l..u', where
    l is an integer and u is an integer or '*',
    except when m.range is None, the string 'default' is returned.
    """
    assert m.oclIsKindOf(UML14.Multiplicity)
    if not m.range:
        _cph(m, [], [], [])
        return "default"
    assert len(m.range.lower) == 1
    assert len(m.range.upper) == 1
    l = int(m.range.lower[0])
    assert l >= 0
    u = int(m.range.upper[0])
    s = ""
    if u == -1:
        s = str(l)+"..*"
    else:
        assert u >= 1
        if l > u:
            _err(m, "Malformed multiplicity")
        s = str(l)+".."+str(u)
    _cph(m, ['range'], [], [])
    return s


def check_name(e, raise_exception = True):
    """
    Check whether the name of the UML Element e is SMUML-legal.
    If it is not, raise an error or return False, depending on whether
    raise_exception is True.
    """
    assert isinstance(e, coral.kernel.Element)
    if e.oclIsKindOf(UML14.Signal):
        return check_signal_name(e, raise_exception)
    match = re.compile("[a-zA-Z_][a-zA-Z0-9_]*").match(str(e.name))
    if((not match) or (match.end() != len(e.name)) or
       str(e.name) == 'boolean' or str(e.name) == 'int'):
        if raise_exception:
            raise RuntimeError(_mcname(e)+" "+_qn(e)+
                               ": Illegal name '"+str(e.name)+"'")
        else:
            return False
    return True


def check_signal_name(signal, raise_exception = True):
    """
    Check whether the name of the UML Signal 'signal' is SMUML-legal.
    If it is not, raise an error or return False, depending on whether
    raise_exception is True.
    """
    assert signal.oclIsKindOf(UML14.Signal)
    match = re.compile("\$?[a-zA-Z_][a-zA-Z0-9_]*").match(str(signal.name))
    if((not match) or (match.end() != len(signal.name)) or
       str(signal.name) == 'boolean' or str(signal.name) == 'int'):
        if raise_exception:
            raise RuntimeError(_mcname(e)+" "+_qn(signal)+
                               ": Illegal name '"+str(signal.name)+"'")
        else:
            return False
    return True


def _validate_legal_orthogonality(sv):
    """
    Returns a pair (tsignals, dsignals), where
    - tsignals is a map from signals to sets of transitions (i) leaving the
      StateVertex 'sv' or any of its descendants and (ii) having the signal
      as the trigger, and
    - dsignals is a map from signals to sets of substates of 'sv' (including
      'sv' itself) in which the signal is deferrable
    Raises an error if two transitions whose sources are orthogonal have
    the same signal as the trigger.
    """
    assert sv.oclIsKindOf(UML14.StateVertex)

    tsignals = {}
    dsignals = {}

    # Traverse children and check that transitions leaving orthogonal states
    # have disjoint signal event triggers
    if sv.oclIsKindOf(UML14.CompositeState):
        if sv.isConcurrent:
            for subvertex in sv.subvertex:
                (cts, cds) = _validate_legal_orthogonality(subvertex)
                for signal in cts.keys():
                    if tsignals.has_key(signal):
                        for t1 in tsignals[signal]:
                            for t2 in cts[signal]:
                                raise RuntimeError("Transitions '"+_qn(t1)+"'"+
                                                   " and '"+_qn(t2)+"' "+
                                                   "leaving otrhogonal states"+
                                                   " have the same signal as "+
                                                   " trigger")
                    else:
                        tsignals[signal] = Set(cts[signal])
                for signal in cds.keys():
                    if not dsignals.has_key(signal):
                        dsignals[signal] = Set()
                    dsignals[signal] = dsignals[signal].union(cds[signal])
        else:
            for subvertex in sv.subvertex:
                (cts, cds) = _validate_legal_orthogonality(subvertex)
                for signal in cts.keys():
                    if not tsignals.has_key(signal):
                        tsignals[signal] = Set()
                    tsignals[signal] = tsignals[signal].union(cts[signal])
                for signal in cds.keys():
                    if not dsignals.has_key(signal):
                        dsignals[signal] = Set()
                    dsignals[signal] = dsignals[signal].union(cds[signal])
      

    if sv.oclIsKindOf(UML14.SubmachineState):
        _err(sv, "Submachines not yet handled")

    # Handle deferrable events
    if sv.oclIsKindOf(UML14.State):
        for event in sv.deferrableEvent:
            if not event.oclIsKindOf(UML14.SignalEvent):
                _err(sv, "Only SignalEvents can be deferrable, "+
                     "not "+_mcname(event))
            assert event.signal
            if str(event.signal.name).startswith('$'):
                _err(sv, "external signals cannot be deferrable")
            if not dsignals.has_key(event.signal):
                dsignals[event.signal] = Set()
            dsignals[event.signal].add(sv)

    # Handle internal transitions
    if sv.oclIsKindOf(UML14.State):
        for t in sv.internalTransition:
            _err(sv, "Internal transitions not yet handled")
            if t.trigger and t.trigger.oclIsKindOf(UML14.SignalEvent):
                assert t.trigger.signal
                assert False and "Not yet implemented"

    for t in sv.outgoing:
        if t.trigger and t.trigger.oclIsKindOf(UML14.SignalEvent):
            assert t.trigger.signal
            if not tsignals.has_key(t.trigger.signal):
                tsignals[t.trigger.signal] = Set()
            tsignals[t.trigger.signal].add(t)

    return (tsignals, dsignals)



def _check_StateVertex(s, sm):
    assert s.oclIsKindOf(UML14.StateVertex)
    assert sm.oclIsKindOf(UML14.StateMachine)

    if s.visibility != 'public':
        _err(s, "visibility should be 'public'")

    # Handle deferrable events
    if s.oclIsKindOf(UML14.State):
        for event in s.deferrableEvent:
            if not event.oclIsKindOf(UML14.SignalEvent):
                _err(sv, "Only SignalEvents can be deferrable, "+
                     "not "+_mcname(event))
            assert event.signal
            assert event.signal.name
            #assert event.namespace == sm.context
            if str(event.signal.name).startswith('$'):
                _err(s, "external signal events cannot be deferrable")


    if s.__type__() == UML14.CompositeState:
        # Handle children
        for subvertex in s.subvertex:
            _check_StateVertex(subvertex, sm)

        _cph(s, ['deferrableEvent','incoming','outgoing','subvertex',
                 'visibility'],['isConcurrent','name'],
             ['container','stateMachine'])
        return

    elif s.__type__() == UML14.SimpleState:
        if not s.incoming:
            _err(s, "A SimpleState with no incoming transitions")
        _cph(s, ['deferrableEvent','incoming','outgoing','visibility'],
             ['name'], ['container'])
        return

    elif s.__type__() == UML14.FinalState:
        if not s.incoming:
            _warn(s, "a FinalState with no incoming transitions")
        if s.outgoing:
            _warn(s, "a FinalState cannot have outgoing transitions")
        if s.deferrableEvent:
            _warn(s, "a FinalState cannot have deferrable events")
        _cph(s, ['deferrableEvent','incoming','outgoing','visibility'],
             ['name'], ['container'])
        return

    elif s.__type__() == UML14.Pseudostate:
        for t in s.outgoing:
            if t.trigger:
                _err(s, "A transition leaving this Pseudostate has a trigger")
        if s.kind == "initial":
            if s.incoming:
                _err(s, "An initial state with incoming transitions")
            if len(s.outgoing) != 1:
                _err(s, "An initial state with "+len(s.outgoing)+" outgoing "+
                     "transitions instead of one")
            for t in s.outgoing:
                if t.guard:
                    _err(s, "A transition leaving an initial state has "+
                         "a guard")
            _cph(s, ['incoming','kind','outgoing','visibility'],
                 ['name'], ['container'])
        elif s.kind == "choice":
            if not s.incoming:
                _err(s, "A choice Pseudostate has no incoming transitions")
            if not s.outgoing:
                _err(s, "A choice Pseudostate has no outgoing transitions")
            _cph(s, ['incoming','kind','outgoing','visibility'],
                 ['name'], ['container'])
        else:
            _err(s, "Unsupported Pseudostate kind '"+str(s.kind)+"'")
            _cph(s, [], [], ['container'])
        return

    _err(s, "Unsupported StateVertex type '"+_mcname(s)+"'")



def _check_StateMachine(sm):
    assert sm.oclIsKindOf(UML14.StateMachine)

    if sm.context != sm.namespace:
        _err(sm, "the context of a StateMachine is not equal to its namespace")
    if not sm.context.oclIsKindOf(UML14.Class):
        _err(sm, "StateMachines should be owned by Classes, not by '"+
             _mcname(sm.context)+"'")

    # name of the state machine is irrelevant, skip

    for t in sm.transitions:
        assert t.source
        assert t.target

        # Check trigger
        if t.trigger:
            if not t.trigger.oclIsKindOf(UML14.SignalEvent):
                _err(t, "Only SignalEvents are allowed as triggers")
            #if t.trigger.namespace != sm.context:
            #    _err(t, "Trigger should be defined in the same context as "+
            #         "the StateMachine")
            _check_trigger(t, sm.context)

        # Check guard
        if t.guard:
            if not t.guard.expression:
                _err(t.guard, "A guard with no expression")
            if not t.guard.expression.oclIsKindOf(UML14.BooleanExpression):
                _err(t.guard, "Only BooleanExpressions allowed as guards")
            if not t.guard.expression.body:
                _err(t.guard, "A guard with empty expression body")
            res = None
            try:
                res = _analyze_expr(str(t.guard.expression.body), sm.context)
            except actlang.errors.Error, exception:
                print t.guard.expression.body
                print exception.__class__
                _err(t.guard, str(exception))
            except RuntimeError, exception:
                print t.guard.expression.body
                print exception.__class__
                _err(t.guard, str(exception))
            assert res
            if res.t != 'boolean':
                _err(t.guard, "A non-boolean guard")
            if not res.is_side_effect_free:
                _err(t.guard, "The guard is not side-effect free")
            _cph(t.guard.expression, ['body'], [], [])
            if t.guard.visibility != 'public':
                _err(t.guard, "visibility should be 'public'")
            _cph(t.guard, ['expression','visibility'],
                 ['name'], ['transition'])

        # Check effect
        if t.effect:
            if not (t.effect.oclIsKindOf(UML14.UninterpretedAction) or
                    t.effect.oclIsKindOf(UML14.Action)):
                _err(t, "Only Actions and UninterpretedActions are allowed "+
                     "as effects, not '"+_mcname(t.effect)+"'")
            if not t.effect.script.body:
                _err(t, "An effect with an empty script body")
            _cph(t.effect.script, ['body'], [], [])
            _cph(t.effect.recurrence, [], [], [])
            _cph(t.effect.target, [], [], [])

            try:
                _analyze_stmt(str(t.effect.script.body), sm.context)
            except Exception, exception:
                _err(t, "Error in the effect code: "+str(exception))

            #_warn(t, "Effects not yet properly validated")
            if t.effect.visibility != 'public':
                _err(t.effect, "visibility should be 'public'")
            _cph(t.effect, ['recurrence','script','target','visibility'],
                 ['name'], ['transition'])

        if t.visibility != 'public':
            _err(t, "visibility should be 'public'")
        _cph(t, ['effect','guard','source','target','trigger','visibility'],
             ['name'], ['stateMachine'])


    if not sm.top:
        _err(sm, "a StateMachine does not have a top state")
    if not sm.top.oclIsKindOf(UML14.CompositeState):
        _err(sm, "the top state of a StateMachine is not a CompositeState")

    _check_StateVertex(sm.top, sm)

    if sm.visibility != 'public':
        _err(sm, "visibility should be 'public'")

    _validate_legal_orthogonality(sm.top)

    _cph(sm, ['context','name','top','transitions','visibility'], [],
         ['namespace'])



def _check_DataType(e):
    assert e.oclIsKindOf(UML14.DataType)
    parent = e.__parent__()
    #if not (parent.__type__() == UML14.Model or
    #        parent.__type__() == UML14.Package):
    if not (parent.__type__() in [UML14.Model, UML14.Package]):
        _err(e, "DataTypes must be owned by the Model or by a Package, "+
             "not by '"+_mcname(parent)+"'")
    if e.oclIsKindOf(UML14.Primitive):
        if e.name in ['boolean', 'int']:
            # The predefined primitive types
            if e.visibility != 'public':
                _err(e, "visibility should be 'public'")
            # typedFeature contains model elements that reference e, ignore
            _cph(e, ['name','visibility'],
                 ['instance','typedFeature','typedParameter'],
                 ['namespace'])
            return
        raise RuntimeError("Unsupported Primitive DataType "+_qn(e))
    else:
        raise RuntimeError("Unsupported DataType "+_qn(e))
    _cph(e, ['name'], ['instance','typedParameter'], ['namespace'])



def _check_DataValue(dv):
    assert dv.oclIsKindOf(UML14.DataValue)

    if dv.visibility != 'public':
        _err(ev, "visibility should be 'public'")

    if not dv.classifier:
        _err(dv, "no classifier in a DataValue")
    if len(dv.classifier) > 1:
        _err(dv, "multiple classifiers defined in a DataValue")
    classifier = dv.classifier[0]
    
    if not classifier.oclIsKindOf(UML14.DataType):
        _err(dv, "the classifier of a DataValue should be a 'DataType', "+
             "not '"+_mcname(classifier)+"'")
        
    if classifier.name == 'int':
        if not Utils.String.IsInt(str(dv.name)):
            _err(dv, "not an integer value")
        v = int(str(dv.name))
        if not (-pow(2,31) <= v and v <= pow(2,31)-1):
            _err(dv, "the value is not in range [-2^31,2^31-1]")
    elif classifier.name == 'bool':
        if not str(dv.name) in ['false','true']:
            _err(dv, "not a boolean value (false or true)")
    else:
        _err(dv, "not yet implemented: validation of DataValues of type "+
             "'"+str(classifier.name)+"'")

    _cph(dv, ['classifier','name','visibility'],
         ['attributeLink'], ['namespace','owner'])



def _check_Signal(e):
    assert e.oclIsKindOf(UML14.Signal)

    check_signal_name(e)

    if e.visibility != 'public':
        _err(e, "visibility should be 'public'")

    for feat in e.feature:
        if feat.oclIsKindOf(UML14.Attribute):
            attr = feat
            if str(e.name).startswith('$'):
                _err(e, "external signals cannot have attributes")
            if attr.changeability != 'changeable':
                _err(attr, "changeability should be 'changeable'")
            if attr.ordering != 'unordered':
                _err(attr, "ordering should be 'unordered'")
            if attr.ownerScope != 'instance':
                _err(attr, "ownerScope should be 'instance'")
            if attr.targetScope != 'instance':
                _err(attr, "targetScope should be 'instance'")
            if attr.visibility != 'public':
                _err(attr, "visibility should be 'public'")
            if not attr.type:
                _err(attr, "An Attribute with no type")
            m = _multiplicity_str(attr.multiplicity)
            if attr.type.__type__() == UML14.Class:
                if not(m == "default" or m == "0..1"):
                    _err(attr, "Illegal multiplicity '"+m+"' "+
                         "for a Class-typed (i.e. object reference) "+
                         "signal attribute")
            elif attr.type.__type__() == UML14.DataType:
                if not(m == "default" or m == "1..1"):
                    _err(attr, "Illegal multiplicity '"+m+"' "+
                         "for a DataType signal attribute")
            if attr.initialValue and attr.initialValue.body:
                _err(attr, "Signal attributes should not have initialValues")
            _cph(attr,
                 ['changeability','initialValue','multiplicity','name',
                  'ordering','ownerScope','targetScope','type','visibility'],
                 [], ['owner'])
        else:
            _err(e, "a Signal cannot have a feature '"+_mcname(feat)+"'")

    # e.occurrence gives the SignalEvents using the signal, ignore
    _cph(e, ['feature','name','visibility'], ['occurrence'], ['namespace'])


def _check_trigger(t, context):
    assert t.oclIsKindOf(UML14.Transition)
    assert context.oclIsKindOf(UML14.Class)
    if not t.trigger:
        return
    e = t.trigger
    if not e.oclIsKindOf(UML14.SignalEvent):
        _err(t, "illegal transition trigger type '"+_mcname(e)+"'")

    # e.name is irrelevant, ignore

    if not e.signal:
        _err(e, "no signal defined")

    # e.transition gives the transitions in which signal event is used, ignore

    # e.state gives the states in which signal event is used (defer), ignore

    if len(e.parameter) != len(e.signal.feature):
        _err(e, "The number of parameters does not match the number of "+
             "attribues in the Signal '"+_qn(e.signal)+"'")

    (class_attrs,class_assocs) = find_attributes_and_associations(context)

    seen_parameter_names = {}
    for (p_index, p) in enumerate(e.parameter):
        if p.defaultValue:
            if p.defaultValue.body:
                _err(p, "SignalEvent parameters cannot have defaultValues")
            _cph(p.defaultValue, [], [], [])
        if p.kind != 'in':
            _err(p, "kind should be 'in'")
        if p.visibility != 'public':
            _err(p, "visibility should be 'public'")
        if not p.name:
            _err(p, "unnamed SignalEvent parameters are not allowed")
        if p.type:
            if p.type is not e.signal.feature[p_index].type:
                print p.type
                print e.signal.feature[p_index].type
                _err(e, "the type of the parameter number "+str(p_index)+
                     " does not match that of the Signal '"+_qn(e.signal)+"'")
        if p.name:
            # Parameter names must be disjoint
            p_name = str(p.name)
            if p_name in seen_parameter_names:
                _err(p, "Multiple parameters with name '"+p_name+"'")
            seen_parameter_names[p_name] = p
            # If the parameter has a name, then it must correspond to
            # a attribute or a navigable association end of the context class
            if p_name in class_attrs:
                # Check that types match
                attr = class_attrs[p_name]
                if attr.type != e.signal.feature[p_index].type:
                    _err(p, "The parameter type does not match the signal "+
                         "attribute type")
                if (_multiplicity_str(attr.multiplicity) !=
                    _multiplicity_str(e.signal.feature[p_index].multiplicity)):
                    _err(p, "The parameter multiplicity does not match that "+
                         "of the signal attribute")
                #_cph(attr,['multiplicity','type'],['name'],[])
            elif p_name in class_assocs:
                # Check that types match
                (targetend,sourceend) = class_assocs[p_name]
                if targetend.participant != e.signal.feature[p_index].type:
                    _err(p, "The parameter type does not match the signal "+
                         "attribute type")
                if (_multiplicity_str(targetend.multiplicity) !=
                    _multiplicity_str(e.signal.feature[p_index].multiplicity)):
                    _err(p, "The parameter multiplicity does not match that "+
                         "of the signal attribute")
            else:
                _err(p, "No attribute or navigable association end in "+
                     "the class '"+str(e.namespace.name)+"' "+
                     "matching the parameter")
        _cph(p, ['defaultValue','kind','name','type','visibility'],
             [], ['event'])

    if e.visibility != 'public':
        _err(e, "visibility should be 'public'")

    _cph(e, ['name','parameter','signal','visibility'],
         ['state','transition'], ['namespace'])



def _check_SignalEvent(e):
    assert e.oclIsKindOf(UML14.SignalEvent)

    return
    #SignalEvents should be owned by a Class
    if (not e.namespace) or (not e.namespace.oclIsKindOf(UML14.Class)):
        _err(e, "a SignalEvent should be owned by a Class")

    # e.name is irrelevant, ignore

    if not e.signal:
        _err(e, "no signal defined")

    # e.transition gives the transitions in which signal event is used, ignore

    # e.state gives the states in which signal event is used (defer), ignore

    (class_attrs,class_assocs) = find_attributes_and_associations(e.namespace)

    if len(e.parameter) != len(e.signal.feature):
        _err(e, "The number of parameters does not match the number of "+
             "attribues in the Signal '"+_qn(e.signal)+"'")

    seen_parameter_names = {}
    for (p_index, p) in enumerate(e.parameter):
        assert p.defaultValue
        if p.defaultValue.body:
            _err(p, "SignalEvent parameters cannot have defaultValues")
        _cph(p.defaultValue, [], [], [])
        if p.kind != 'in':
            _err(p, "kind should be 'in'")
        if p.visibility != 'public':
            _err(p, "visibility should be 'public'")
        if p.name:
            # Parameter names must be disjoint
            p_name = str(p.name)
            if p_name in seen_parameter_names:
                _err(p, "Multiple parameters with name '"+p_name+"'")
            seen_parameter_names[p_name] = p
            # If the parameter has a name, then it must correspond to
            # a attribute or a navigable association end of the context class
            if p_name in class_attrs:
                # Check that types match
                attr = class_attrs[p_name]
                if attr.type != e.signal.feature[p_index].type:
                    _err(p, "The parameter type does not match the signal "+
                         "attribute type")
                if (_multiplicity_str(attr.multiplicity) !=
                    _multiplicity_str(e.signal.feature[p_index].multiplicity)):
                    _err(p, "The parameter multiplicity does not match that "+
                         "of the signal attribute")
                #_cph(attr,['multiplicity','type'],['name'],[])
            elif p_name in class_assocs:
                # Check that types match
                (targetend,sourceend) = class_assocs[p_name]
                if targetend.participant != e.signal.feature[p_index].type:
                    _err(p, "The parameter type does not match the signal "+
                         "attribute type")
                if (_multiplicity_str(targetend.multiplicity) !=
                    _multiplicity_str(e.signal.feature[p_index].multiplicity)):
                    _err(p, "The parameter multiplicity does not match that "+
                         "of the signal attribute")
            else:
                _err(p, "No attribute or navigable association end in "+
                     "the class '"+str(e.namespace.name)+"' "+
                     "matching the parameter")
        _cph(p, ['defaultValue','kind','name','visibility'],
             [], ['event'])

    if e.visibility != 'public':
        _err(e, "visibility should be 'public'")

    _cph(e, ['name','parameter','signal','visibility'],
         ['state','transition'], ['namespace'])



def _check_Class(e):
    assert e.oclIsKindOf(UML14.Class)

    check_name(e)

    has_statemachine = False
    for oe in e.ownedElement:
        if oe.__type__() == UML14.Comment:
            pass # Comments are allowed
        elif oe.__type__() == UML14.Association:
            # Associations are handled below
            pass
        elif oe.__type__() == UML14.Signal:
            _check_Signal(oe)
        elif oe.__type__() == UML14.SignalEvent:
            _check_SignalEvent(oe)
        elif oe.__type__() == UML14.StateMachine:
            if has_statemachine:
                _err(e, "a Class contains more than one StateMachine")
            if not (oe in e.behavior):
                err(e, "a StateMachine in ownedElement but not in behavior")
            has_statemachine = True
            _check_StateMachine(oe)
        else:
            _err(e, "a Class cannot contain a "+_mcname(oe))

    if has_statemachine and not e.isActive:
        _err(e, "a Class contains a StateMachine but is not active")
    if not has_statemachine and e.isActive:
        _err(e, "a Class is active but does not contain a StateMachine")


    # Get the attributes and navigable associations of the class
    (attrs, assocs) = find_attributes_and_associations(e)

    # Check attributes
    for (attrname, attr) in attrs.iteritems():
        check_name(attr)
        if attr.changeability != 'changeable':
            _err(attr, "changeability should be 'changeable'")
        if attr.ordering != 'unordered':
            _err(attr, "ordering should be 'unordered'")
        if attr.ownerScope != 'instance':
            _err(attr, "ownerScope should be 'instance'")
        if attr.targetScope != 'instance':
            _err(attr, "targetScope should be 'instance'")
        if attr.visibility != 'public':
            _err(attr, "visibility should be 'public'")

        if not attr.type:
            _err(attr, "an Attribute with no type")
        if attr.type.__type__() == UML14.Class:
            _err(attr, "an attribute should not have a Class as a type, "+
                 "use associations to model object references")
        if attr.type.__type__() != UML14.Primitive:
            _err(attr, "illegal attribute type '"+_mcname(attr.type)+"'")

        m = _multiplicity_str(attr.multiplicity)
        if not(m == 'default' or m == "1..1"):
            _err(attr, "illegal attribute multiplicity '"+m+"'")

        if attr.initialValue.body:
            res = None
            try:
                res = _analyze_expr(str(attr.initialValue.body), e)
            except Exception, exception:
                _err(attr, "Illegal initial value: "+str(exception))
            assert res
            if not res.is_constant:
                _err(attr, "The initial value is not a constant expression")
            if not res.is_side_effect_free:
                _err(attr, "The initial value is not side-effect free")
            if res.t != str(attr.type.name):
                _err(attr, "The type '"+res.t+"' of the initial value "
                     "expression does not match the type '"+
                     _qn(attr.type)+"' of the attribute")
            # Check that the type of the initialValue matches 'type'
            #_warn(attr, "initial values of Attributes not "+
            #      "yet properly validated")
            #print attr.initialValue.body
            #print "TODO: "+_qn(attr)
        _cph(attr,
             ['changeability','initialValue','multiplicity','name',
              'ordering','ownerScope','targetScope','type','visibility'],
             ['attributeLink'], ['owner'])

    # Check associations
    for (targetname, (targetend, sourceend)) in assocs.iteritems():
        assert targetend.isNavigable
        assoc = targetend.association
        assert len(assoc.connection) == 2
        if sourceend.isNavigable:
            _err(assoc, "An association with both ends navigable")

        # Name of the association does not matter, skip
        
        if assoc.visibility != 'public':
            _err(assoc, "visibility should be 'public'")

        if sourceend.visibility != 'public':
            _err(sourceend, "visibility should be 'public'")
        if targetend.visibility != 'public':
            _err(targetend, "visibility should be 'public'")

        if sourceend.aggregation != 'none':
            _err(sourceend, "aggregation should be 'none'")
        if targetend.aggregation != 'none':
            _err(targetend, "aggregation should be 'none'")

        if sourceend.name:
            check_name(sourceend)
            if not sourceend.isNavigable:
                _warn(sourceend, "A non-navigable AssociationEnd with a name")
        else:
            if sourceend.isNavigable:
                _err(sourceend,
                     "An AssociationEnd is navigable but has no name")

        assert targetend.isNavigable
        assert targetend.name
        check_name(targetend)

        if sourceend.changeability != 'changeable':
            _err(sourceend, "changeability should be 'changeable'")
        if targetend.changeability != 'changeable':
            _err(targetend, "changeability should be 'changeable'")

        if sourceend.targetScope != 'instance':
            _err(sourceend, "targetScope should be 'instance'")
        if targetend.targetScope != 'instance':
            _err(targetend, "targetScope should be 'instance'")

        if sourceend.ordering != 'unordered':
            _err(sourceend, "ordering should be 'unordered'")
        if targetend.ordering != 'unordered':
            _err(targetend, "ordering should be 'unordered'")

        m = _multiplicity_str(targetend.multiplicity)
        if not m in ['default', '0..1']:
            _err(targetend, "multiplicity of a navigable association end "+
                 "must be '0..1', not '"+m+"'")
        source_m = _multiplicity_str(sourceend.multiplicity)
        if not source_m in ['default', '0..*']:
            _warn(sourceend, "The multiplicity '"+source_m+
              "' of a non-navigable AssociatioEnd ignored")

        assert sourceend.participant == e
        if targetend.participant.__type__() != UML14.Class:
            _err(targetend,
                 "an AssociationEnd with a non-Class participant")

        _cph(assoc, ['connection','visibility'], ['link','name'],
             ['namespace'])
        _cph(sourceend,
             ['aggregation','association','changeability','isNavigable',
              'multiplicity','name','ordering','participant',
              'targetScope','visibility'],
             ['linkEnd'],[])
        _cph(targetend,
             ['aggregation','association','changeability','isNavigable',
              'multiplicity','name','ordering','participant',
              'targetScope','visibility'],
             ['linkEnd'],[])



    for behavior in e.behavior:
        if behavior.__type__() == UML14.StateMachine:
            if not (behavior in e.ownedElement):
                err(e, "A StateMachine in behavior but not in ownedElement")
            # Already handled above, skip
            continue
        _err(e, "a Class cannot contain a behavior "+_mcname(behavior))


    for feature in e.feature:
        if feature.__type__() == UML14.Attribute:
            # Already handled above
            continue
        else:
            _err(e, "a Class cannot contain a feature "+_mcname(feature))

    
    for instance in e.instance:
        assert instance.oclIsKindOf(UML14.Object)

    if e.visibility != 'public':
        _err(e, "visibility should be 'public'")

    # typedFeature contains signal references to this Class, ignored
    
    _cph(e,
         ['association','behavior','feature','instance','isActive','name',
          'ownedElement','visibility'],
         ['typedFeature','typedParameter'],
         ['namespace'])



def _check_Package(e):
    assert e.oclIsKindOf(UML14.Package)

    # names of Packages do not matter as Packages have
    # no scoping role currently
    #check_name(e)

    if e.visibility != 'public':
        _err(e, "visibility should be 'public'")
    for oe in e.ownedElement:
        if oe.__type__() == UML14.Comment:
            pass # Comments are allowed
        elif oe.__type__() == UML14.Package:
            _check_Package(oe)
        elif oe.__type__() == UML14.Signal:
            _check_Signal(oe)
        elif oe.__type__() == UML14.SignalEvent:
            _check_SignalEvent(oe)
        elif oe.oclIsKindOf(UML14.DataType):
            _check_DataType(oe)
        elif oe.__type__() == UML14.DataValue:
            _check_DataValue(oe)
        elif oe.__type__() == UML14.Class:
            _check_Class(oe)
        elif oe.__type__() == UML14.Association:
            # Associations are handled when their contexts (e.g. class)
            # are handled
            pass
        elif oe.__type__() == UML14.Object:
            _check_Object(oe)
        elif oe.__type__() == UML14.Link:
            _check_Link(oe)
        else:
            _err(e, "A Package cannot contain a "+_mcname(oe))

    _cph(e, ['ownedElement','visibility'], ['name'], ['namespace'])





#
#
# Validate initial configuration related stuff: Objects and Links
#
#
def _check_Object(obj):
    assert obj.oclIsKindOf(UML14.Object)

    check_name(obj)

    if obj.visibility != 'public':
        _err(obj, "visibility should be 'public'")

    if not obj.classifier:
        _err(obj, "an Object must have a classifier")
    if len(obj.classifier) > 1:
        _err(obj, "an Object cannot have multiple classifiers")
    classifier = obj.classifier[0]

    seen_assoc_ends = {}
    for link_end in obj.linkEnd:
        if not link_end.link:
            _err(link_end, "LinkEnd has no link")
        _check_Link(link_end.link)
        assert link_end.associationEnd
        assoc_end = link_end.associationEnd
        other_assoc_end = get_other_association_end(assoc_end)
        if other_assoc_end.isNavigable:
            if other_assoc_end in seen_assoc_ends:
                _err(link, "the reference attribute (association) '"+
                     str(assoc_end.name)+"' is already defined for "+
                     "the object '"+_qn(obj)+"' in the Link '"+
                     _qn(seen_assoc_ends[other_assoc_end])+"'")
            seen_assoc_ends[other_assoc_end] = link_end.link

    for link in obj.ownedLink:
        _check_Link(link)

    seen_attributes = Set()
    for alink in obj.slot:
        assert alink.__type__() == UML14.AttributeLink
        if alink.visibility != 'public':
            _err(alink, "visibility should be public")
        if not alink.attribute:
            _err(alink, "AttributeLink has no attribute")
        if alink.attribute in seen_attributes:
            _err(obj, "the initial value of the Attribute "+
                 "'"+_qn(alink.attribute)+"' defined multiple times")
        seen_attributes.add(alink.attribute)
        if alink.attribute.owner != classifier:
            _err(alink, "the attribute '"+_qn(alink.attribute)+"' "+
                 "in a slot of an object is not owned by "+
                 "the class '"+_qn(obj.classifier)+"' of the object")
        if not alink.value:
            _err(alink, "AttributeLink has no value")
        if alink.value.classifier[0] != alink.attribute.type:
            _err(alink, "the value of the AttributeLink is of different type "+
                 "than its attribute: '"+_qn(alink.value.classifier[0])+
                 "' vs '"+_qn(alink.attribute.type)+"'")
        _cph(alink, ['attribute','value','visibility'], ['name'], ['instance'])


    for oi in obj.ownedInstance:
        if oi.__type__() == UML14.Comment:
            pass # Comments are allowed
        elif oi.__type__() == UML14.DataValue:
            _check_DataValue(oi)
        else:
            _err(model, "An Object cannot own a "+_mcname(oi))

    _cph(obj, ['classifier','name','ownedInstance','ownedLink',
               'slot','visibility'],
         ['linkEnd'],['namespace'])



def _check_Link(link):
    assert link.oclIsKindOf(UML14.Link)

    # names of links do not matter, skip...

    if link.visibility != 'public':
        _err(link, "visibility should be public")

    if not link.association:
        _err(link, "no association in a link")

    for link_end in link.connection:
        if link_end.visibility != 'public':
            _err(link_end, "visibility should be 'public'")
        assoc_end = link_end.associationEnd
        if not assoc_end:
            _err(link_end, "no associationEnd in a LinkEnd")
        if not link_end.instance:
            _err(link_end, "no instance in LinkEnd")
        _cph(link_end, ['associationEnd','instance','visibility'],
             [], ['link'])

    _cph(link, ['association','connection','visibility'],
         ['name'], ['owner_at_Link'])





#
#
# Validate the Model
#
#
def _check_Model(model):
    assert model.oclIsKindOf(UML14.Model)
    if model.visibility != 'public':
        _err(model, "visibility should be 'public'")
    for oe in model.ownedElement:
        if oe.__type__() == UML14.Comment:
            pass # Comments are allowed
        elif oe.__type__() == UML14.Package:
            _check_Package(oe)
        elif oe.__type__() == UML14.Signal:
            _check_Signal(oe)
        elif oe.__type__() == UML14.SignalEvent:
            _check_SignalEvent(oe)
        elif oe.oclIsKindOf(UML14.DataType):
            _check_DataType(oe)
        elif oe.__type__() == UML14.DataValue:
            _check_DataValue(oe)
        elif oe.__type__() == UML14.Class:
            _check_Class(oe)
        elif oe.__type__() == UML14.Association:
            # Associations are handled when their contexts (e.g. class)
            # are handled
            pass
        elif oe.__type__() == UML14.Object:
            _check_Object(oe)
        elif oe.__type__() == UML14.Link:
            _check_Link(oe)
        else:
            _err(model, "A Model cannot contain a "+_mcname(oe))

    # Check that signal names are disjoint
    signals = Utils.Models.get_elements_of_type(model, UML14.Signal)
    signal_names = {}
    for signal in signals:
        signal_name = str(signal.name)
        if signal_name in signal_names:
            _err(signal, "Signal with the same name defined also in "+
                 _qn(signal_names[signal_name]))
        signal_names[signal_name] = signal

    # Check that Class names are disjoint
    class_names = {}
    for c in Utils.Models.get_elements_of_type(model, UML14.Class):
        class_name = str(c.name)
        if class_name in class_names:
            _err(c, "Class with the same name defined also in '"+
                 _qn(class_names[class_name])+"'")
        class_names[class_name] = c

    _cph(model, ['ownedElement','visibility'], ['name'], [])






def validate_project(project):
    """
    Validate that the UML 1.4 Model in the Coral Project 'project' is
    a valid SMUML UML model.
    Raises an error if this is not the case.
    Returns the model.
    """
    assert isinstance(project, coral.kernel.Project)

    models = [m for m in project.contents() if m.oclIsKindOf(UML14.Model)]
    if len(models) != 1:
        raise RuntimeError('Coral Project contains ' + str(len(models)) +
                           ' models instead of one')
    model = models[0]

    _check_Model(model)

    return model


#if __name__ == '__main__':
#    if len(sys.argv) != 2:
#        print "Usage "+sys.argv[0]+" uml_model.xmi"
#        sys.exit(-1)
#    project = coral.loadDocument(sys.argv[1])
#
#    validate_project(project)


def get_model(element):
    """
    Returns the UML Model to which 'element' belongs.
    """
    assert isinstance(element, coral.kernel.Element)
    while not element.oclIsKindOf(UML14.Model):
        element = element.__parent__()
    return element




import actlang

class _AnalyzeResult:
    def __init__(self,
                 t = 'void',
                 is_constant = False,
                 is_side_effect_free = False):
        assert (type(t) is str) or t.oclIsKindOf(UML14.Class)
        assert type(is_constant) is bool
        assert type(is_side_effect_free) is bool
        self.t = t
        self.is_constant = is_constant
        self.is_side_effect_free = is_side_effect_free


def _analyze_expr(expr, context):
    """
    Can raise exceptions.
    """
    assert type(expr) is str
    assert context.oclIsKindOf(UML14.Class)


    # Build the parse tree 
    tmp = "x = "+expr+";"
    tree = actlang.grammar.parse(tmp)
    assert len(tree.api_getTopStatements()) == 1
    ast_node = tree.api_getTopStatements()[0]
    assert isinstance(ast_node, actlang.statement.ExprStatement)
    assert isinstance(ast_node.api_getExpr(),
                      actlang.expr.SimpleAssignmentExpr)
    ast_node = ast_node.api_getExpr().api_getRhsExpr()
    return _analyze(ast_node, context)



def _analyze_stmt(stmt, context):
    """
    Can raise exceptions.
    """
    assert type(stmt) is str
    assert context.oclIsKindOf(UML14.Class)


    # Build the parse tree 
    tree = actlang.grammar.parse(stmt)
    for ast_node in tree.api_getTopStatements():
        r = _analyze(ast_node, context)
    return _AnalyzeResult()


def _match_types(t1, t2):
    if (type(t1) is str) and (type(t2) is str):
        return t1 == t2
    if (type(t1) is str) and isinstance(t2, coral.kernel.Element):
        return (t1 == 'object') and t2.oclIsKindOf(UML14.Class)
    if isinstance(t1, coral.kernel.Element) and (type(t2) is str):
        return t1.oclIsKindOf(UML14.Class) and (t2 == 'object')
    if(isinstance(t1, coral.kernel.Element) and
       isinstance(t2, coral.kernel.Element)):
        return t1 is t2
    return False

def _err2(s):
    raise RuntimeError(s)
    
def _analyze(node, context):
    if isinstance(node, actlang.statement.EmptyStatement):
        return _AnalyzeResult(is_constant = True,
                              is_side_effect_free = True)
    elif isinstance(node, actlang.expr.ParenExpr):
        return _analyze(node.api_getInnerExpr(), context)
    elif isinstance(node, actlang.expr.IntLiteralExpr):
        return _AnalyzeResult(t = 'int',
                              is_constant = True,
                              is_side_effect_free = True)
    elif isinstance(node, actlang.expr.BoolLiteralExpr):
        return _AnalyzeResult(t = 'boolean',
                              is_constant = True,
                              is_side_effect_free = True)
    elif isinstance(node, actlang.expr.NullExpr):
        return _AnalyzeResult(t = 'object',
                              is_constant = True,
                              is_side_effect_free = True)
    elif isinstance(node, actlang.expr.StringLiteralExpr):
        return _AnalyzeResult(t = 'string',
                              is_constant = True,
                              is_side_effect_free = True)
    elif isinstance(node, actlang.expr.ThisExpr):
        return _AnalyzeResult(t = context,
                              is_constant = True,
                              is_side_effect_free = True)
    elif isinstance(node, actlang.names.SimpleName):
        name = str(node)
        (attrs,assocs) = find_attributes_and_associations(context)
        if name in attrs:
            attr = attrs[name]
            if not attr.type:
                _err2("Attribute '"+name+"' of Class '"+
                      _qn(context)+"' has no type")
            if(attr.type.oclIsKindOf(UML14.Primitive) and
               attr.type.name == 'boolean'):
                return _AnalyzeResult(t = 'boolean',
                                      is_side_effect_free = True)
            if(attr.type.oclIsKindOf(UML14.Primitive) and
               attr.type.name == 'int'):
                return _AnalyzeResult(t = 'int',
                                      is_side_effect_free = True)
            print attr.type
            _err2("Attribute '"+name+"' of Class '"+
                  _qn(context)+"' is of unsupported type '"+
                  _qn(attr.type)+"'")
        if name in assocs:
            (assoc_end, dummy) = assocs[name]
            if not assoc_end.participant:
                _err(assoc_end, "An association end with no participant")
            if not assoc_end.participant.oclIsKindOf(UML14.Class):
                _err(assoc_end, "An association end with a non-Class "
                     "participant '"+_qn(assoc_end.participant)+"'")
            return _AnalyzeResult(t = assoc_end.participant,
                                  is_side_effect_free = True)
        _err2("The Class '"+_qn(context)+"' has no attribute "
              "or association end named '"+name+"'")
    elif isinstance(node, actlang.names.QualifiedName):
        q = _analyze(node.api_getQualifyingName(), context)
        if not isinstance(q.t, coral.kernel.Element):
            _err2("The qualifying part is not of reference "
                  "type: "+str(node))
        context2 = q.t
        name = node.api_getLastPart()
        (attrs,assocs) = find_attributes_and_associations(context2)
        if name in attrs:
            attr = attrs[name]
            if not attr.type:
                _err2("Attribute '"+name+"' of Class '"+
                      _qn(context2)+"' has no type")
            if(attr.type.oclIsKindOf(UML14.Primitive) and
               attr.type.name == 'boolean'):
                return _AnalyzeResult(t = 'boolean',
                                      is_side_effect_free = True)
            if(attr.type.oclIsKindOf(UML14.Primitive) and
               attr.type.name == 'int'):
                return _AnalyzeResult(t = 'int',
                                      is_side_effect_free = True)
            _err2("Attribute '"+name+"' of Class '"+_qn(context2)+"' "
                  "is of unsupported type '"+_qn(attr.type)+"'")
        if name in assocs:
            (assoc_end, dummy) = assocs[name]
            if not assoc_end.participant:
                _err(assoc_end, "An association end with no participant")
            if not assoc_end.participant.oclIsKindOf(UML14.Class):
                _err(assoc_end, "An association end with a non-Class "
                     "participant '"+_qn(assoc_end.participant)+"'")
            return _AnalyzeResult(t = assoc_end.participant,
                                  is_side_effect_free = True)
        _err2("The Class '"+_qn(context2)+"' has no attribute "
              "or association end named '"+name+"'")
    elif isinstance(node, actlang.expr.UnaryPlusExpr):
        c = _analyze(node.api_getOperand(), context)
        if c.t == 'int':
            return c
        _err2("Cannot apply unary plus to a non-integer expression: "+
              str(node))
    elif isinstance(node, actlang.expr.UnaryMinusExpr):
        c = _analyze(node.api_getOperand(), context)
        if c.t == 'int':
            return c
        _err2("Cannot apply unary minus to a non-integer expression: "+
              str(node))

    elif isinstance(node, actlang.expr.LogicalNotExpr):
        c = _analyze(node.api_getOperand(), context)
        if c.t == 'boolean':
            return c
        _err2("Cannot apply logical not to a non-boolean expression: "+
              str(node))

    elif isinstance(node, actlang.expr.BitNotExpr):
        c = _analyze(node.api_getOperand(), context)
        if c.t == 'int':
            return c
        _err2("Cannot apply bit-wise not to a non-integer expression: "+
              str(node))

    elif isinstance(node, (actlang.expr.CastExpr
                           ,actlang.expr.TypeCastExpr
                           ,actlang.expr.NameCastExpr)):
        _err2("Cast expressions not supported: "+str(node))
    
    elif isinstance(node, actlang.expr.InstanceOfExpr):
        _err2("Instance-of expressions not supported: "+str(node))

    elif isinstance(node, actlang.expr.NameMethodInvocationExpr):
        n = node.api_getInvocationName()
        if str(n) == "$$SMUML.choose_int":
            return _AnalyzeResult(t = 'int',
                              is_constant = False,
                              is_side_effect_free = True)
        if str(n) == "$$SMUML.choose_boolean":
            return _AnalyzeResult(t = 'boolean',
                              is_constant = False,
                              is_side_effect_free = True)
        _err2("Method invocations are not supported: "+str(node))

    elif isinstance(node, actlang.expr.InstanceCreationExpr):
        class_name = node.api_getTypeName()
        class_name = str(class_name)
        if node.api_getArgList():
            _err2("Arguments in 'new' statements are not supported: "+
                  str(node))
        model = get_model(context)
        classes = Utils.Models.get_elements_of_type(model, UML14.Class)
        classes = [c for c in classes if str(c.name) == class_name]
        if not classes:
            _err2("There is no Class '"+class_name+"' in the Model")
        return _AnalyzeResult(t = classes[0])

    elif isinstance(node, actlang.expr.CondOrExpr):
        l = _analyze(node.api_getLeftOperand(), context)
        r = _analyze(node.api_getRightOperand(), context)
        is_constant = l.is_constant and r.is_constant
        is_side_effect_free = l.is_side_effect_free and r.is_side_effect_free
        if l.t != 'boolean':
            _err2("Cannot apply || to an expression of type '"+
                  l.t+"': "+str(node.api_getLeftOperand()))
        if r.t != 'boolean':
            _err2("Cannot apply || to an expression of type '"+
                  r.t+"': "+str(node.api_getRightOperand()))
        return _AnalyzeResult(t = 'boolean',
                              is_constant = is_constant,
                              is_side_effect_free=is_side_effect_free)

    elif isinstance(node, actlang.expr.CondAndExpr):
        l = _analyze(node.api_getLeftOperand(), context)
        r = _analyze(node.api_getRightOperand(), context)
        is_constant = l.is_constant and r.is_constant
        is_side_effect_free = l.is_side_effect_free and r.is_side_effect_free
        if l.t != 'boolean':
            _err2("Cannot apply && to an expression of type '"+
                  l.t+"': "+str(node.api_getLeftOperand()))
        if r.t != 'boolean':
            _err2("Cannot apply && to an expression of type '"+
                  r.t+"': "+str(node.api_getRightOperand()))
        return _AnalyzeResult(t = 'boolean',
                              is_constant = is_constant,
                              is_side_effect_free=is_side_effect_free)

    elif isinstance(node, actlang.expr.TernaryCondExpr):
        c = _analyze(node.api_getCondition(), context)
        t = _analyze(node.api_getTrueExpr(), context)
        f = _analyze(node.api_getFalseExpr(), context)
        is_constant = (c.is_constant and t.is_constant and f.is_constant)
        is_side_effect_free = (c.is_side_effect_free and
                               t.is_side_effect_free and
                               f.is_side_effect_free)
        if c.t != 'boolean':
            _err2("The condition part is not boolean: "+str(node))
        #if t.t != f.t:
        #    raise RuntimeError("The types of the then and else parts "
        #                       "are not the same: "+str(node))
        return _AnalyzeResult(t = t.t, is_constant = is_constant,
                              is_side_effect_free = is_side_effect_free)

    elif isinstance(node, actlang.expr.InfixExpr):
        r1 = _analyze(node.api_getLeftOperand(), context)
        r2 = _analyze(node.api_getRightOperand(), context)
        op = node.api_getOperator()
        is_constant = r1.is_constant and r2.is_constant
        is_side_effect_free = r1.is_side_effect_free and r2.is_side_effect_free
        if (type(r1.t) is str) and (type(r2.t) is str):
            if r1.t == 'int' and r2.t == 'int':
                if isinstance(op, (actlang.binaryop.NumericalComparison,
                                   actlang.binaryop.EqualityOperator)):
                    return _AnalyzeResult(t = 'boolean',
                                          is_constant = is_constant,
                                          is_side_effect_free =
                                          is_side_effect_free)
                if isinstance(op, (actlang.binaryop.Plus,
                                   actlang.binaryop.Minus,
                                   actlang.binaryop.Multiply,
                                   actlang.binaryop.Divide,
                                   actlang.binaryop.BitLogicalOr,
                                   actlang.binaryop.BitLogicalAnd,
                                   actlang.binaryop.BitLogicalXor)):
                    return _AnalyzeResult(t = 'int',
                                          is_constant = is_constant,
                                          is_side_effect_free =
                                          is_side_effect_free)
            elif(r1.t == 'boolean' and r2.t == 'boolean'):
                if isinstance(op, (actlang.binaryop.EqualityOperator)):
                    return _AnalyzeResult(t = 'boolean',
                                          is_constant = is_constant,
                                          is_side_effect_free =
                                          is_side_effect_free)
            elif(r1.t == 'object' and r2.t == 'object'):
                if isinstance(op, (actlang.binaryop.EqualityOperator)):
                    return _AnalyzeResult(t = 'boolean',
                                          is_constant = is_constant,
                                          is_side_effect_free =
                                          is_side_effect_free)
        if((isinstance(r1.t, coral.kernel.Element) and
            isinstance(r2.t, coral.kernel.Element)) or
           (type(r1.t) is str and r1.t == 'object' and
            isinstance(r2.t, coral.kernel.Element)) or
           (isinstance(r1.t, coral.kernel.Element) and
            type(r2.t) is str and r2.t == 'object')):
            if _match_types(r1.t, r2.t):
                if isinstance(op, actlang.binaryop.EqualityOperator):
                    return _AnalyzeResult(t = 'boolean',
                                          is_constant = is_constant,
                                          is_side_effect_free =
                                          is_side_effect_free)
        _err2("Cannot apply '"+str(op)+"' to expressions of types "
              "'"+str(r1.t)+"' and '"+str(r2.t)+"'")

    elif isinstance(node, actlang.expr.IncrementOrDecrementExpr):
        r = _analyze(node.api_getOperand(), context)
        if r.t != 'int':
            _err2("Pre- and postfix increment and decrement are "
                  "only supported for integers: "+str(node))
        r.is_side_effect_free = False
        return r

    elif isinstance(node, actlang.statement.ExprStatement):
        return _analyze(node.api_getExpr(), context)

    elif isinstance(node, actlang.statement.AssertStatement):
        r = _analyze(node.api_getExpr(), context)
        if not r.is_side_effect_free:
            _err2("The assert expression is not side-effect free: "+str(node))
        return _AnalyzeResult()

    elif isinstance(node, actlang.expr.SimpleAssignmentExpr):
        l = _analyze(node.api_getLhsExpr(), context)
        r = _analyze(node.api_getRhsExpr(), context)
        if not l.is_side_effect_free:
            _err2("The left-hand side of the assignment is "
                  "not side-effect free: "+str(node))
        if not _match_types(l.t, r.t):
            _err2("The types of the left- and right-hand side do not match: "+
                  str(node))
        if isinstance(node.api_getRhsExpr(),
                      actlang.expr.InstanceCreationExpr):
            # r = new A() is allowed
            return _AnalyzeResult(t = r.t)
        if not r.is_side_effect_free:
            _err2("The right-hand side of the assignment is not "
                  "side-effect free: "+str(node))
        return _AnalyzeResult(t = r.t, is_constant = r.is_constant)
    
    elif isinstance(node, actlang.statement.SendStatement):
        signal_name = node.api_getSignalName()
        target_r = _analyze(node.api_getTarget(), context)
        if not target_r.is_side_effect_free:
            _err2("The target of a send statement is not side-effect free: "+
                  str(node))
        for (arg_index,arg_ast) in enumerate(node.api_getArgList()):
            r = _analyze(arg_ast, context)
            if not r.is_side_effect_free:
                _err2("The signal parameter expressions are not "
                      "side-effect free: "+str(node))
            # TO DO: math with signal parameter types
        return _AnalyzeResult()

    elif isinstance(node, actlang.statement.Block):
        for stmt in node.api_getBlockStatements():
            _analyze(stmt, context)
        return _AnalyzeResult()


    elif isinstance(node, actlang.statement.LabeledStatement):
        _analyze(node.api_getContainedStatement(), context)
        return _AnalyzeResult()

    elif isinstance(node, actlang.statement.BreakStatement):
        return _AnalyzeResult()

    elif isinstance(node, actlang.statement.ContinueStatement):
        return _AnalyzeResult()
        
    elif isinstance(node, actlang.statement.IfStatement):
        c = _analyze(node.api_getCondition(), context)
        if not c.is_side_effect_free:
            _err2("The condition part is not side-effect free: "+str(node))
        t = _analyze(node.api_getTrueStatement(), context)
        f = _analyze(node.api_getFalseStatement(), context)
        return _AnalyzeResult()

    elif isinstance(node, actlang.statement.LocalVarForStatement):
        _err2("Local variable declarations are not supported: "+str(node))

    elif isinstance(node, actlang.statement.DoStatement):
        c = _analyze(node.api_getCondition(), context)
        if not c.is_side_effect_free:
            _err2("The condition part is not side-effect free: "+str(node))

        _analyze(node.api_getBody(), context)
        return _AnalyzeResult()

    elif isinstance(node, actlang.statement.WhileStatement):
        c = _analyze(node.api_getCondition(), context)
        if not c.is_side_effect_free:
            _err2("The condition part is not side-effect free: "+str(node))

        _analyze(node.api_getBody(), context)
        return _AnalyzeResult()

    elif isinstance(node, actlang.statement.ExprListForStatement):
        for init_stmt in node.api_getInitialization():
            if not isinstance(init_stmt, actlang.expr.SimpleAssignmentExpr):
                _err2("Unsupported initialization part: "+str(node))
            l = _analyze(init_stmt.api_getLhsExpr(), context)
            r = _analyze(init_stmt.api_getRhsExpr(), context)
            if not (l.is_side_effect_free and r.is_side_effect_free):
                _err2("Unsupported initialization part: "+str(node))

        r = _analyze(node.api_getCondition(), context)
        if not r.is_side_effect_free:
            _err2("The condition part is not side-effect free: "+str(node))
   
        for update_stmt in node.api_getUpdatePart():
            _analyze(update_stmt, context)

        _analyze(node.api_getBody(), context)
        return _AnalyzeResult()


    #print node.__class__
    _err2("Unsuported action language expression/statement: "+str(node))
