# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Automatic generation of operations for abstract data type libraries.

This module provides functionality for generating operations for
abstract types using external SMT solvers.

Author: Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

from copy import deepcopy
import subprocess
import actlang
import PySMT
from abstraction import error
from abstraction.typelibrary import TypeLibrary
from abstraction.cartesize import cartesize
from abstraction.typehierarchy import ConsistencyError
from Utils import ExpressionTraverser

# A dictionary of integer operations available in the Jumbala action language
# together with the symbols of the operations for various output languages.
# The dictionary associates an action language class with a tuple containing
# the following components (in order)
#    1. the string "operator" or "test" telling whether the operation is a
#       function returning an integer or a Boolean, respectively.
#    2. the arity of the operation.
# Every remaining element in the tuple gives the symbol (if any) for the
# operation in a specific output language.  The following output languages are
# currently defined:
#    3. symbol for the operation in Python
#    4. symbol for the operation in Jumbala
#    5. symbol for the operation in the QF_UFBV[32] logic of the Satisfiability
#       Modulo Theories library (<http://combination.cs.uiowa.edu/smtlib/>).
# Given a key `k' in the dictionary, the components of the tuple associated
# with `symbols[k]' can be referred to by the symbolic constants
#    SYMBOL_TYPE, SYMBOL_ARITY, SYMBOL_TOKEN_PYTHON, SYMBOL_TOKEN_JUMBALA, and
#    SYMBOL_TOKEN_SMTLIB_QF_UFBV32,
# respectively.  For example, `symbols[k][SYMBOL_ARITY]' gives the arity of the
# operation associated with `k'.
#
# Except for the languages Python and Jumbala, the symbols defined for the
# operations in this dictionary are not referenced in any way in the generic
# code for generating abstract types.  The tuples in this dictionary may be
# freely extended with additional elements to make the translation of Jumbala
# action language objects into solver input languages more convenient (for an
# example, see the function SMTLIB.translateQFUFBV32).

_symbols = \
    { actlang.expr.UnaryPlusExpr : ("operator", 1, "+", "+", "+"),
      actlang.expr.UnaryMinusExpr : ("operator", 1, "-", "-", "bvneg"),
      actlang.expr.BitNotExpr : ("operator", 1, "~", "~", "bvnot"),
      actlang.expr.LogicalNotExpr : ("test", 1, "not ", "!", "not"),
      actlang.binaryop.Plus : ("operator", 2, "+", "+", "bvadd"),
      actlang.binaryop.Minus : ("operator", 2, "-", "-", "bvsub"),
      actlang.binaryop.Multiply : ("operator", 2, "*", "*", "bvmul"),
      actlang.binaryop.Divide : ("operator", 2, "/", "/", "/"),
      actlang.binaryop.Remainder : ("operator", 2, "%", "%", "%"),
      actlang.binaryop.BitLogicalAnd : ("operator", 2, "&", "&",
                                              "bvand"),
      actlang.binaryop.BitLogicalOr : ("operator", 2, "|", "|", "bvor"),
      actlang.binaryop.BitLogicalXor : ("operator", 2, "^", "^",
                                              "bvxor"),
      actlang.binaryop.ShiftLeft : ("operator", 2, "<<", "<<", "<<"),
      actlang.binaryop.SignedShiftRight : ("operator", 2, ">>", ">>",
                                                 ">>"),
      actlang.binaryop.UnsignedShiftRight : ("operator", 2, ">>>", ">>>",
                                                   ">>>"),
      actlang.expr.CondAndExpr : ("test", 2, "and", "&&", "and"),
      actlang.expr.CondOrExpr : ("test", 2, "or", "||", "or"),
      actlang.binaryop.Equal : ("test", 2, "==", "==", "="),
      actlang.binaryop.NotEqual : ("test", 2, "!=", "!=", "!="),
      actlang.binaryop.LessThan : ("test", 2, "<", "<", "bvslt"),
      actlang.binaryop.LessOrEqual : ("test", 2, "<=", "<=", "bvsleq"),
      actlang.binaryop.GreaterThan : ("test", 2, ">", ">", "bvsgt"),
      actlang.binaryop.GreaterOrEqual : ("test", 2, ">=", ">=",
                                         "bvsgeq") }

(SYMBOL_TYPE, SYMBOL_ARITY, SYMBOL_TOKEN_PYTHON, SYMBOL_TOKEN_JUMBALA,
 SYMBOL_TOKEN_SMTLIB_QF_UFBV32) = range(0, 5)

def _generateOperation(solver, operator, domain, concretization, operations):
    """Define the abstraction of an operation on integers.

    Arguments:
        solver         -- An object supporting the method `unsat', which
                          takes an actlang.expr.Expr object representing
                          a Boolean expression in the action language as
                          a parameter and returns a Boolean value.  To
                          ensure that the abstract operations generated
                          by the `generateOperation' functions are safe
                          overapproximations, the `unsat' method should
                          return True only if the input expression is
                          unsatisfiable.  See the YICES_SMTLIB_QF_UFBV32
                          class for an example.
        operator       -- Action language class corresponding to the
                          operation.
        domain         -- The abstract domain as a mapping from symbolic
                          values into integers.
        concretization -- A mapping from values in the (numeric) abstract
                          domain to Boolean action language expressions
                          representing their concretizations.
        operations     -- A dictionary in which to store the definition
                          of the operation.
    """
    arity = _symbols[operator][SYMBOL_ARITY]
    #print _symbols[operator][SYMBOL_TOKEN_JUMBALA], arity
    if not arity in operations:
        operations[arity] = {}
    is_test = (_symbols[operator][SYMBOL_TYPE] == "test")
    dfn = {}
    operations[arity][_symbols[operator][SYMBOL_TOKEN_JUMBALA]] = dfn
    try:
        invalid = domain["INVALID"]
    except KeyError:
        invalid = None
    # iterate over all tuples in DOMAIN^{arity} x RANGE (= "product")
    # where DOMAIN collects the integers representing abstract
    # values, and RANGE is either DOMAIN, or ["true", "false"]
    product = []
    for i in xrange(0, arity):
        product.append(tuple(domain.values()))
    if is_test:
        product.append((True, False))
    else:
        product.append(tuple(domain.values()))
    ids = \
        (actlang.names.SimpleName("arg1", (0, 0)),
         actlang.names.SimpleName("arg2", (0, 0)),
         actlang.names.SimpleName("result", (0, 0)))
    # construct a dictionary of concretization expressions renamed with
    # the variables in ids
    renamed_expressions = {}
    for value in domain.itervalues():
        for identifier in ids:
            renamed_expressions[(value, identifier)] = \
                ExpressionTraverser.substitute_exprs \
                    (concretization[value], {"x" : identifier})[0]
    if is_test:
        # constraint for the result of the concrete (test) operation; an
        # abstractable test operation should always be a test between two
        # integers
        if issubclass(operator, (actlang.binaryop.EqualityOperator,
                                 actlang.binaryop.NumericalComparison)):
            # constraint: "arg1 op arg2" for (concrete) integers arg1, arg2
            conc_constraints = actlang.expr.InfixExpr \
                                   (ids[0], ids[1], operator((0, 0)))
            neg_conc_constraints = actlang.expr.LogicalNotExpr \
                                       (conc_constraints, (0, 0))
        else:
            raise ValueError \
                      ("unsupported operator `%s'"
                       % _symbols[operator][SYMBOL_TOKEN_JUMBALA])
    else:
        # constraint for the result of the concrete (arithmetic) operation
        # "(op arg1) == result"   or  "(arg2 op arg2) == result" depending on
        # the arity of the operation
        if operator in (actlang.expr.UnaryPlusExpr,
                        actlang.expr.UnaryMinusExpr,
                        actlang.expr.BitNotExpr):
            conc_constraints = operator(ids[0], (0, 0))
        elif issubclass(operator,
                        (actlang.binaryop.NumericalBooleanOperator,
                         actlang.binaryop.NumericalOperator,
                         actlang.binaryop.Plus)):
            conc_constraints = \
                actlang.expr.InfixExpr(ids[0], ids[1], operator((0, 0)))
        else:
            raise ValueError \
                      ("unsupported operator `%s'"
                       % _symbols[operator][SYMBOL_TOKEN_JUMBALA])
        conc_constraints = actlang.expr.InfixExpr \
                               (ids[2], conc_constraints,
                                actlang.binaryop.Equal((0, 0)))

    for t in cartesize(*product):
        args = tuple(t[0:arity])
        if invalid in args:
            dfn[args] = (is_test and [True, False]) or [invalid]
            continue
        if t[-1] == invalid:
            continue
        if not args in dfn:
            dfn[args] = []
        # constraints for the arguments of the abstract operation
        abst_constraints = renamed_expressions[(args[0], ids[0])]
        for i in xrange(1, arity):
            abst_constraints = actlang.expr.CondAndExpr \
                                  (abst_constraints,
                                   renamed_expressions[(args[i], ids[i])])
        if is_test:
            # determine whether the concrete test should be constrained to
            # evaluate to true or false
            conc = (t[-1] and conc_constraints) or neg_conc_constraints
        else:
            conc = conc_constraints
            # constraint on the value of the abstract result variable
            abst_constraints = \
                actlang.expr.CondAndExpr \
                (abst_constraints, renamed_expressions[(t[-1], ids[2])])

        if not solver.unsat(actlang.expr.CondAndExpr(abst_constraints, conc)):
            dfn[args].append(t[-1])

# -----------------------------------------------------------------------------

class PySmt:
    """Class for checking SAT on Jumbala expressions using PySMT circuits."""

    # Tuple of names of supported SMT solvers.
    SMT_SOLVERS = ("CVC3", "MathSAT", "STP", "Yices", "Z3")

    def __init__(self, solver, path, stdout=None, stderr=None):
        """Initialize a PySmt object.

        Objects of this class can be given as the `solver' parameter for
        the `smtlib_typegen.generateOperations' function.

        Arguments:
            solver -- A name of one of the supported SMT solvers
                      ("CVC3", "MathSAT", "STP", "Yices", "Z3").
            path   -- Full path name of an executable file for the
                      chosen solver.
            stdout -- Unused.
            stderr -- If not None, use this stream for error messages
                      about failures to translate action language
                      expressions into PySMT circuits.
        """
        s = solver.lower()
        if s == "cvc3":
            self.solver = PySMT.CVC3(path)
        elif s == "mathsat":
            self.solver = PySMT.MathSAT(path)
        elif s == "stp":
            self.solver = PySMT.STP(path)
        elif s == "yices":
            self.solver = PySMT.Yices(path)
        elif s == "z3":
            self.solver = PySMT.Z3(path)
        else:
            raise ValueError("unknown solver `%s'" % solver)
        self.stderr = stderr

    def unsat(self, e):
        """Test the unsatisfiability of a Boolean expression.

        Argument:
            e -- The Boolean expression as an actlang.expr.Expr object.

        Returns:
            True if the expression can be successfully translated into a
            PySMT circuit, and the SMT solver used by the PySmt object
            reports the expression to be unsatisfiable, otherwise False.
        """
        circuit = PySMT.Circuit()
        var_gates = {}
        extra_constraints = []
        try:
            try:
                root_gate = \
                    self.__translatePySmt \
                        (e, circuit, var_gates, extra_constraints)
                if extra_constraints:
                    extra_constraints.append(root_gate)
                    root_gate = circuit.add_and(*extra_constraints)
            except ExpressionTraverser.UnregisteredCallbackError, e:
                raise error.Error \
                          ("translation to PySMT circuit failed: %s" % str(e))
            circuit.constrain(root_gate, True)
            result = circuit.solve(self.solver, produce_model=False)
            if result == "error":
                raise error.Error("failed to solve circuit")
            return result == "unsat"
        except error.Error, e:
            if self.stderr:
                s = error.getErrorStream()
                error.setErrorStream(self.stderr)
                try:
                    error.warn(str(e))
                finally:
                    error.setErrorStream(s)
            raise

    # Callbacks for translating action language expressions into PySMT
    # circuits.

    def __createPySmtTranslator():
        """Build a function object for translating Jumbala expressions."""

        def simpleName(expr, subexprs, circuit, var_gates, *rest):
            expr = str(expr)
            try:
                return var_gates[expr]
            except KeyError:
                gate = circuit.add_sint_var(32)
                var_gates[expr] = gate
                return gate

        def infixExpr(expr, subexprs, circuit, vars, extra_constraints, *rest):
            op = expr.api_getOperator()
            if isinstance(op, actlang.binaryop.Equal):
                return circuit.add_eq(*subexprs)
            elif isinstance(op, actlang.binaryop.NotEqual):
                return circuit.add_neq(*subexprs)
            elif isinstance(op, actlang.binaryop.BitLogicalAnd):
                return circuit.add_bitwise_and(*subexprs)
            elif isinstance(op, actlang.binaryop.BitLogicalOr):
                return circuit.add_bitwise_or(*subexprs)
            elif isinstance(op, actlang.binaryop.BitLogicalXor):
                return circuit.add_bitwise_xor(*subexprs)
            elif isinstance(op, actlang.binaryop.GreaterOrEqual):
                return circuit.add_ge(*subexprs)
            elif isinstance(op, actlang.binaryop.GreaterThan):
                return circuit.add_gt(*subexprs)
            elif isinstance(op, actlang.binaryop.LessOrEqual):
                return circuit.add_le(*subexprs)
            elif isinstance(op, actlang.binaryop.LessThan):
                return circuit.add_lt(*subexprs)
            elif isinstance(op, actlang.binaryop.Plus):
                return circuit.add_plus(*subexprs)
            elif isinstance(op, actlang.binaryop.Minus):
                return circuit.add_minus(*subexprs)
            elif isinstance(op, actlang.binaryop.Multiply):
                return circuit.add_times(*subexprs)
            elif isinstance(op, actlang.binaryop.ShiftLeft):
                # Left shift operator.  The value to shift is in subexprs[0],
                # the shift count in subexprs[1].  The shift count is assumed
                # to be nonnegative.
                result = circuit.add_sint_var(32)
                zero = circuit.add_sint_const(32, 0)
                # shift count >= 0; x << 0 == x; x >> y == 0 (y>=32)
                extra_constraints.extend \
                    ([circuit.add_ge(subexprs[1], zero),
                      circuit.add_implies
                          (circuit.add_eq(subexprs[1], zero),
                           circuit.add_eq(result, subexprs[0])),
                      circuit.add_implies
                          (circuit.add_ge
                               (subexprs[1],
                                circuit.add_sint_const(32, 32)),
                           circuit.add_eq(result, zero))])
                for i in xrange(1, 32):
                    conj = []
                    # for j in [0..i):
                    #    j'th bit of result = 0
                    mask = circuit.add_sint_const(32, (2 ** i) - 1)
                    conj.append \
                        (circuit.add_eq
                             (circuit.add_bitwise_and(result, mask),
                              zero))
                    # for j in [i..32):
                    #    j'th bit of result = j-i'th bit of value to be shifted
                    for j in xrange(i, 32):
                        conj.append \
                            (circuit.add_iff
                                 (circuit.add_eq
                                      (circuit.add_bitwise_and
                                           (result,
                                            circuit.add_sint_const
                                                (32,
                                                 j < 31 and 2 ** j
                                                 or -(2 ** 31))),
                                       zero),
                                  circuit.add_eq
                                      (circuit.add_bitwise_and
                                           (subexprs[0],
                                            circuit.add_sint_const
                                                (32, 2 ** (j - i))),
                                       zero)))
                    extra_constraints.append \
                        (circuit.add_implies
                             (circuit.add_eq
                                  (subexprs[1], circuit.add_sint_const(32, i)),
                              circuit.add_and(*conj)))
                return result
            elif isinstance(op, (actlang.binaryop.SignedShiftRight,
                                 actlang.binaryop.UnsignedShiftRight)):
                # Right shift operators.  The value to shift is in subexprs[0],
                # the shift count in subexprs[1].  The shift count is assumed
                # to be nonnegative.
                result = circuit.add_sint_var(32)
                zero = circuit.add_sint_const(32, 0)
                # shift count >= 0; x >> 0 == x >>> 0 == x
                extra_constraints.extend \
                    ([circuit.add_ge(subexprs[1], zero),
                      circuit.add_implies
                          (circuit.add_eq(subexprs[1], zero),
                           circuit.add_eq(result, subexprs[0]))])
                if isinstance(op, actlang.binaryop.SignedShiftRight):
                    # x >= 0 => x >> y == 0  (y >= 32)
                    # x < 0  => x >> y == -1 (y >= 32)
                    extra_constraints.append \
                        (circuit.add_implies
                             (circuit.add_ge
                                  (subexprs[1],
                                   circuit.add_sint_const(32, 32)),
                              circuit.add_eq
                                  (result,
                                   circuit.add_ite
                                       (circuit.add_ge(subexprs[0], zero),
                                        zero,
                                        circuit.add_sint_const(32, -1)))))
                else:
                    # x >>> y == 0 (y >= 32)
                    extra_constraints.append \
                        (circuit.add_implies
                             (circuit.add_ge
                                  (subexprs[1],
                                   circuit.add_sint_const(32, 32)),
                              circuit.add_eq(result, zero)))
                for i in xrange(1, 32):
                    conj = []
                    # for j in [0..32-i):
                    #    j'th bit of result = j+i'th bit of value to be shifted
                    for j in xrange(0, 32 - i):
                        conj.append \
                            (circuit.add_iff
                                 (circuit.add_eq
                                      (circuit.add_bitwise_and
                                           (result,
                                            circuit.add_sint_const
                                                (32, 2 ** j)),
                                       zero),
                                  circuit.add_eq
                                      (circuit.add_bitwise_and
                                           (subexprs[0],
                                            circuit.add_sint_const
                                                (32,
                                                 (i + j) < 31 and 2 ** (i + j)
                                                 or -(2 ** 31))),
                                       zero)))
                    # for j in [32-i..32):
                    #    j'th bit of result = sign bit of value to be shifted
                    #                         (signed shift)
                    #                       = 0 (unsigned shift)
                    mask = circuit.add_sint_const(32, ~((2 ** (32 - i)) - 1))
                    if isinstance(op, actlang.binaryop.SignedShiftRight):
                        conj.append \
                            (circuit.add_eq
                                 (circuit.add_bitwise_and(result, mask),
                                  circuit.add_ite
                                      (circuit.add_lt(subexprs[0], zero),
                                       mask,
                                       zero)))
                    else:
                        conj.append \
                            (circuit.add_eq
                                 (circuit.add_bitwise_and(result, mask),
                                  zero))
                    extra_constraints.append \
                        (circuit.add_implies
                             (circuit.add_eq
                                  (subexprs[1], circuit.add_sint_const(32, i)),
                              circuit.add_and(*conj)))
                return result
            elif isinstance(op, (actlang.binaryop.Divide,
                                 actlang.binaryop.Remainder)):
                # Division and remainder operators.  The dividend is in
                # subexprs[0], divisor in subexprs[1].  The divisor/modulus is
                # assumed to be nonzero.
                quot = circuit.add_sint_var(32) # result of integer division
                rem = circuit.add_sint_var(32)  # result of remainder operation
                zero = circuit.add_sint_const(32, 0)
                extra_constraints.extend([
                    # remainder == dividend - quotient * divisor
                    # [Jumbala action language specification v1.00, Sect. 9.9]
                    circuit.add_eq \
                        (rem,
                         circuit.add_minus
                             (subexprs[0],
                              circuit.add_times(quot, subexprs[1]))),

                    # dividend == -2^31 and divisor == -1
                    # ==>  quotient == -2^31 (overflow)
                    # [Jumbala action language specification v1.00, Sect. 9.9]
                    circuit.add_implies \
                        (circuit.add_and
                             (circuit.add_eq
                                  (subexprs[0],
                                   circuit.add_sint_const(32, -2147483648)),
                              circuit.add_eq
                                  (subexprs[1],
                                   circuit.add_sint_const(32, -1))),
                         circuit.add_eq
                             (quot, circuit.add_sint_const(32, -2147483648))),

                    # |dividend| < |divisor|  ==>  quotient == 0
                    circuit.add_implies \
                        (circuit.add_or
                             (circuit.add_and
                                  (circuit.add_ge(subexprs[0], zero),
                                   circuit.add_ge(subexprs[1], zero),
                                   circuit.add_lt(subexprs[0], subexprs[1])),
                              circuit.add_and
                                  (circuit.add_lt(subexprs[0], zero),
                                   circuit.add_ge(subexprs[1], zero),
                                   circuit.add_gt
                                       (subexprs[0],
                                        circuit.add_neg(subexprs[1]))),
                              circuit.add_and
                                  (circuit.add_ge(subexprs[0], zero),
                                   circuit.add_lt(subexprs[1], zero),
                                   circuit.add_gt
                                       (circuit.add_neg(subexprs[0]),
                                        subexprs[1])),
                              circuit.add_and
                                  (circuit.add_lt(subexprs[0], zero),
                                   circuit.add_lt(subexprs[1], zero),
                                   circuit.add_gt(subexprs[0], subexprs[1]))),
                         (circuit.add_eq(quot, zero))),

                    # remainder is either equal to zero, or has the same sign
                    # as the dividend
                    # [Jumbala action language specification v1.00, Sect. 9.9]
                    circuit.add_or \
                        (circuit.add_eq(rem, zero),
                         circuit.add_iff
                             (circuit.add_gt(rem, zero),
                              circuit.add_gt(subexprs[0], zero))),

                    # dividend < 0, divisor > 0  or  dividend > 0, divisor < 0
                    # ==> quotient <= 0
                    circuit.add_implies \
                        (circuit.add_or
                             (circuit.add_and
                                  (circuit.add_lt(subexprs[0], zero),
                                   circuit.add_gt(subexprs[1], zero)),
                              circuit.add_and
                                  (circuit.add_gt(subexprs[0], zero),
                                   circuit.add_lt(subexprs[1], zero))),
                         circuit.add_le(quot, zero)),

                    # dividend > 0, divisor > 0  or  dividend < 0, divisor < 0
                    #                                and (dividend != -2^31 or
                    #                                     divisor != -1)
                    # ==> quotient >= 0
                    circuit.add_implies \
                        (circuit.add_or
                             (circuit.add_and
                                  (circuit.add_gt(subexprs[0], zero),
                                   circuit.add_gt(subexprs[1], zero)),
                              circuit.add_and
                                  (circuit.add_lt(subexprs[0], zero),
                                   circuit.add_lt(subexprs[1], zero),
                                   circuit.add_or
                                       (circuit.add_neq
                                            (subexprs[0],
                                             circuit.add_sint_const
                                                 (32, -2147483648)),
                                        circuit.add_neq
                                            (subexprs[1],
                                             circuit.add_sint_const
                                                 (32, -1))))),
                         circuit.add_ge(quot, zero)),
                                            
                    # |remainder| < |divisor|
                    # [Jumbala action language specification v1.00, Sect. 9.9]
                    circuit.add_implies \
                        (circuit.add_and
                             (circuit.add_ge(rem, zero),
                              circuit.add_ge(subexprs[1], zero)),
                         circuit.add_lt(rem, subexprs[1])),
                    circuit.add_implies \
                        (circuit.add_and
                             (circuit.add_lt(rem, zero),
                              circuit.add_ge(subexprs[1], zero)),
                         circuit.add_gt(rem, circuit.add_neg(subexprs[1]))),
                    circuit.add_implies \
                        (circuit.add_and
                             (circuit.add_ge(rem, zero),
                              circuit.add_lt(subexprs[1], zero)),
                         circuit.add_gt(circuit.add_neg(rem), subexprs[1])),
                    circuit.add_implies \
                        (circuit.add_and
                             (circuit.add_lt(rem, zero),
                              circuit.add_lt(subexprs[1], zero)),
                         circuit.add_gt(rem, subexprs[1])),

                    # |quotient| <= |dividend|
                    circuit.add_implies \
                        (circuit.add_and
                             (circuit.add_ge(quot, zero),
                              circuit.add_ge(subexprs[0], zero)),
                         circuit.add_le(quot, subexprs[0])),
                    circuit.add_implies \
                        (circuit.add_and
                             (circuit.add_lt(quot, zero),
                              circuit.add_ge(subexprs[0], zero)),
                         circuit.add_ge(quot, circuit.add_neg(subexprs[0]))),
                    circuit.add_implies \
                        (circuit.add_and
                             (circuit.add_ge(quot, zero),
                              circuit.add_lt(subexprs[0], zero)),
                         circuit.add_ge(circuit.add_neg(quot), subexprs[0])),
                    circuit.add_implies \
                        (circuit.add_and
                             (circuit.add_lt(quot, zero),
                              circuit.add_lt(subexprs[0], zero)),
                         circuit.add_ge(quot, subexprs[0])),

                    # divisor != 0
                    circuit.add_neq(subexprs[1], zero)])
                return (isinstance(op, actlang.binaryop.Divide) and quot) \
                       or rem

        callbacks = \
            ((actlang.expr.BoolLiteralExpr,
              lambda expr, subexprs, circuit, *rest:
                  (expr.api_getValue() and circuit.add_true()
                   or circuit.add_false())),
             (actlang.expr.IntLiteralExpr,
              lambda expr, subexprs, circuit, *rest:
                  circuit.add_sint_const(32, expr.api_getValue())),
             (actlang.names.SimpleName, simpleName),
             ((actlang.expr.ParenExpr, actlang.expr.UnaryPlusExpr),
              lambda expr, subexprs, *rest: subexprs[0]),
             (actlang.expr.TernaryCondExpr,
              lambda expr, subexprs, circuit, *rest:
                  circuit.add_ite(*subexprs)),
             (actlang.expr.InfixExpr, infixExpr),
             (actlang.expr.CondAndExpr,
              lambda expr, subexprs, circuit, *rest:
                  circuit.add_and(*subexprs)),
             (actlang.expr.CondOrExpr,
              lambda expr, subexprs, circuit, *rest:
                  circuit.add_or(*subexprs)),
             (actlang.expr.LogicalNotExpr,
              lambda expr, subexprs, circuit, *rest:
                  circuit.add_not(*subexprs)),
             (actlang.expr.BitNotExpr,
              lambda expr, subexprs, circuit, *rest:
                  circuit.add_bitwise_not(*subexprs)),
             (actlang.expr.UnaryMinusExpr,
              lambda expr, subexprs, circuit, *rest:
                  circuit.add_neg(*subexprs)))

        translator = \
            ExpressionTraverser.ExpressionTraverser(ExpressionTraverser.ignore)
        for (cls, f) in callbacks:
            translator.register_post_callback(cls, f)
        return translator

    # Function object for translating action language expressions into PySMT
    # circuits.
    #
    # The function object takes the following parameters:
    #    1. An actlang.expr.Expr object (the expression to be translated).
    #    2. A PySMT.Circuit object.
    #    3. A dictionary mapping variable names into the corresponding gates.
    #    4. A list of circuit gates representing constraints on additional
    #       gates which were introduced during the translation.
    #
    # The function object returns a gate representing the root gate in the
    # circuit obtained from the translation.

    __translatePySmt = __createPySmtTranslator()
    del __createPySmtTranslator

class SMTLIB:
    """Class with helper functions used with SMT-LIB formats.

    This class provides a method for translating Jumbala expressions
    into the QF_UFBV[32] logic of the SMT-LIB Satisfiability Modulo
    Theories library.
    """

    # Callbacks for translating action language expressions into the logic
    # QF_UFBV[32].

    def __createQFUFBV32Translator():
        """Build a function object for translating Jumbala expressions."""

        def intLiteral(expr, *rest):
            val = expr.api_getValue()
            if val >= 0:
                return "bv%d" % val
            return "(bvneg bv%d)" % -val

        def simpleName(expr, subexprs, varnames, *rest):
            varnames.add(str(expr))
            return str(expr)

        def infixExpr(expr, subexprs, varnames, extra_constraints):
            op = _symbols[expr.api_getOperator().__class__] \
                 [SYMBOL_TOKEN_SMTLIB_QF_UFBV32]
            if op in ("<<", ">>>"):
                # Shift left and unsigned shift right operators.  The value to
                # shift is in subexprs[0], the shift count in subexprs[1].  The
                # shift count is assumed to be nonnegative.
                varnames.update("s")
                qf_ufbv32_op = (op == "<<" and "left") or "right"
                for i in xrange(0, 32):
                    extra_constraints.append \
                        ("(implies (= %s bv%d) (= s (shift_%s0 %s %d)))"
                         % (subexprs[1], i, qf_ufbv32_op, subexprs[0], i))
                extra_constraints.extend([
                    "(bvsgeq %s bv0)" % subexprs[1],
                    "(implies (bvsgeq %s bv32) (= s bv0))" % subexprs[1]])
                return "s"
            elif op == ">>":
                # Signed shift right operator.  The value to shift is in
                # subexprs[0], the shift count in subexprs[1].  The shift count
                # is assumed to be nonnegative.
                varnames.update("s")
                for i in xrange(0, 32):
                    extra_constraints.append \
                        ("(implies (= %s bv%d) "
                         "(if_then_else (bvslt %s bv0) "
                         "(= s (shift_right1 %s %d)) "
                         "(= s (shift_right0 %s %d))))"
                         % (subexprs[1], i, subexprs[0],
                            subexprs[0], i, subexprs[0], i))
                extra_constraints.extend([
                    "(bvsgeq %s bv0)" % subexprs[1],
                    "(implies (bvsgeq %s bv32) "
                    "(if_then_else (bvslt %s bv0) "
                    "(= s (bvneg bv1)) (= s bv0)))"
                    % (subexprs[1], subexprs[0])])
                return "s"
            elif op in ("/", "%"):
                # Division and remainder operators.  The dividend is in
                # subexprs[0], divisor in subexprs[1].  The divisor/modulus is
                # assumed to be nonzero.
                quot = "d%d" % len(varnames) # result of integer division
                rem = "r%d" % len(varnames)  # result of remainder operation
                varnames.update((quot, rem))
                extra_constraints.extend([
                    # remainder == dividend - quotient * divisor
                    # [Jumbala action language specification v1.00, Sect. 9.9]
                    "(= %s (bvsub %s (bvmul %s %s)))"
                    % (rem, subexprs[0], quot, subexprs[1]),

                    # dividend == -2^31 and divisor == -1
                    # ==>  quotient == -2^31 (overflow)
                    # [Jumbala action language specification v1.00, Sect. 9.9]
                    "(implies (and (= %s (bvneg bv2147483648)) "
                    "(= %s (bvneg bv1))) (= %s (bvneg bv2147483648)))"
                    % (subexprs[0], subexprs[1], quot),

                    # |dividend| < |divisor|  ==>  quotient == 0
                    "(implies (or "
                    "(and (bvsgeq %s bv0) (bvsgeq %s bv0) (bvslt %s %s)) "
                    "(and (bvslt %s bv0) (bvsgeq %s bv0) "
                    "(bvsgt %s (bvneg %s))) "
                    "(and (bvsgeq %s bv0) (bvslt %s bv0) "
                    "(bvsgt (bvneg %s) %s)) "
                    "(and (bvslt %s bv0) (bvslt %s bv0) (bvsgt %s %s)))"
                    "(= %s bv0))"
                    % (subexprs[0], subexprs[1],
                       subexprs[0], subexprs[1],
                       subexprs[0], subexprs[1],
                       subexprs[0], subexprs[1],
                       subexprs[0], subexprs[1],
                       subexprs[0], subexprs[1],
                       subexprs[0], subexprs[1],
                       subexprs[0], subexprs[1],
                       quot),

                    # remainder is either equal to zero, or has the same sign
                    # as the dividend
                    # [Jumbala action language specification v1.00, Sect. 9.9]
                    "(or (= %s bv0) (iff (bvsgt %s bv0) (bvsgt %s bv0)))"
                    % (rem, rem, subexprs[0]),

                    # dividend < 0, divisor > 0  or  dividend > 0, divisor < 0
                    # ==> quotient <= 0
                    "(implies (or (and (bvslt %s bv0) (bvsgt %s bv0)) "
                    "(and (bvsgt %s bv0) (bvslt %s bv0))) (bvsleq %s bv0))"
                    % (subexprs[0], subexprs[1], subexprs[0], subexprs[1],
                       quot),
                    # dividend > 0, divisor > 0  or  dividend < 0, divisor < 0
                    #                                and (dividend != -2^31 or
                    #                                     divisor != -1)
                    # ==> quotient >= 0
                    "(implies (or (and (bvsgt %s bv0) (bvsgt %s bv0)) "
                    "(and (bvslt %s bv0) (bvslt %s bv0) "
                    "(not (and (= %s (bvneg bv2147483648)) "
                    "(= %s (bvneg bv1)))))) (bvsgeq %s bv0))"
                    % (subexprs[0], subexprs[1],
                       subexprs[0], subexprs[1],
                       subexprs[0], subexprs[1],
                       quot),

                    # |remainder| < |divisor|
                    # [Jumbala action language specification v1.00, Sect. 9.9]
                    "(implies (and (bvsgeq %s bv0) (bvsgeq %s bv0)) "
                    "(bvslt %s %s))"
                    % (rem, subexprs[1], rem, subexprs[1]),
                    "(implies (and (bvslt %s bv0) (bvsgeq %s bv0)) "
                    "(bvsgt %s (bvneg %s)))"
                    % (rem, subexprs[1], rem, subexprs[1]),
                    "(implies (and (bvsgeq %s bv0) (bvslt %s bv0)) "
                    "(bvsgt (bvneg %s) %s))"
                    % (rem, subexprs[1], rem, subexprs[1]),
                    "(implies (and (bvslt %s bv0) (bvslt %s bv0)) "
                    "(bvsgt %s %s))"
                    % (rem, subexprs[1], rem, subexprs[1]),

                    # |quotient| <= |dividend|
                    "(implies (and (bvsgeq %s bv0) (bvsgeq %s bv0)) "
                    "(bvsleq %s %s))"
                    % (quot, subexprs[0], quot, subexprs[0]),
                    "(implies (and (bvslt %s bv0) (bvsgeq %s bv0)) "
                    "(bvsgeq %s (bvneg %s)))"
                    % (quot, subexprs[0], quot, subexprs[0]),
                    "(implies (and (bvsgeq %s bv0) (bvslt %s bv0)) "
                    "(bvsgeq (bvneg %s) %s))"
                    % (quot, subexprs[0], quot, subexprs[0]),
                    "(implies (and (bvslt %s bv0) (bvslt %s bv0)) "
                    "(bvsgeq %s %s))"
                    % (quot, subexprs[0], quot, subexprs[0]),

                    # divisor != 0
                    "(not (= %s bv0))" % subexprs[1]])
                return (op == "/" and quot) or rem
            elif op == "!=":
                return "(not (= %s))" % " ".join(subexprs)
            else:
                return "(%s %s)" % (op, " ".join(subexprs))

        callbacks = (
            (actlang.expr.BoolLiteralExpr,
             lambda expr, *rest: (expr.api_getValue() and "true") or "false"),
            (actlang.expr.IntLiteralExpr, intLiteral),
            (actlang.names.SimpleName, simpleName),
            ((actlang.expr.ParenExpr, actlang.expr.UnaryPlusExpr),
             lambda expr, subexprs, *rest: subexprs[0]),
            (actlang.expr.TernaryCondExpr,
             lambda expr, subexprs, *rest:
             "(if_then_else %s %s %s)" % tuple(subexprs)),
            (actlang.expr.InfixExpr, infixExpr),
            ((actlang.expr.CondAndExpr, actlang.expr.CondOrExpr,
              actlang.expr.LogicalNotExpr, actlang.expr.BitNotExpr,
              actlang.expr.UnaryMinusExpr),
             lambda expr, subexprs, *rest:
             "(%s %s)" % (_symbols[expr.__class__] \
                          [SYMBOL_TOKEN_SMTLIB_QF_UFBV32],
                          " ".join(subexprs))))

        translator = \
            ExpressionTraverser.ExpressionTraverser(ExpressionTraverser.ignore)
        for (cls, f) in callbacks:
            translator.register_post_callback(cls, f)
        return translator

    # Function object for translating action language expressions into the
    # logic QF_UFBV[32].
    #
    # The function object takes the following parameters:
    #    1. An actlang.expr.Expr object (the expression to be translated).
    #    2. A set to be populated (during the translation) with names of
    #       uninterpreted functions which occur in the formula obtained from
    #       the translation.
    #    3. A list to be populated with formulas in the logic QF_UFBV[32]
    #       representing additional constraints on uninterpreted functions
    #       which were introduced in the formula during the translation.
    #
    # The function object returns a string representing the formula obtained
    # from the translation.

    translateQFUFBV32 = __createQFUFBV32Translator()
    del __createQFUFBV32Translator

class YICES_SMTLIB_QF_UFBV32(SMTLIB):
    """Wrapper class for the Yices SMT solver using the logic QF_UFBV[32]."""

    def __init__(self, path, stdout=None, stderr=None):
        """Initialize an object of the wrapper class.

        Arguments:
            path   -- Full path name of the Yices executable.
            stdout -- Unused.
            stderr -- If not None, use this stream for error messages
                      about failures to translate action language
                      expressions into the logic QF_UFBV[32].
        """
        self.path = path
        self.stderr = stderr

    def unsat(self, e):
        """Test the unsatisfiability of a Boolean expression using Yices.

        Argument:
            e -- The Boolean expression as an actlang.expr.Expr object.

        Returns:
            True if the expression can be successfully translated into the
            logic QF_UFBV[32] and Yices reports the expression to be
            unsatisfiable, otherwise False.
        """
        varnames = set()
        extra_constraints = []
        try:
            formula = SMTLIB.translateQFUFBV32(e, varnames, extra_constraints)
            if extra_constraints:
                formula = "(and %s %s)" % (formula,
                                           " ".join(extra_constraints))
        except ExpressionTraverser.UnregisteredCallbackError, e:
            if self.stderr:
                s = error.getErrorStream()
                error.setErrorStream(self.stderr)
                try:
                    error.warn \
                        ("translation to QF_UFBV[32] logic failed: %s"
                         % str(e))
                finally:
                    error.setErrorStream(s)
            raise
        b = "(benchmark b :logic QF_UFBV[32] :extrafuns (%s) :formula %s)" \
            % (" ".join([("(%s BitVec[32])" % v) for v in varnames]), formula)
        try:
            child = \
                subprocess.Popen \
                    ([self.path, "-smt"], stdin=subprocess.PIPE,
                     stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            child.stdin.write(b)
            child.stdin.close()
            result = child.stdout.readline().replace("\n", "")
            status = child.wait()
        except IOError:
            status = 1
        if status:
            msg = ["failed to execute Yices"]
            msg.extend([line for line in child.stderr])
            if len(msg) > 1:
                msg[0] = msg[0] + ":"
            raise IOError("\n    ".join(msg))
        return result == "unsat"

# -----------------------------------------------------------------------------

def generateOperations(typelibrary, operations, solver, model_runtime_exceptions=False, stdout=None, stderr=None, verbose=False):
    """Update a type library with automatically generated op. definitions.

    Arguments:
        typelibrary              -- An abstraction.typelibrary.TypeLibrary
                                    object [note (1)].
        operations               -- An iterable of operations to generate.
                                    Each element in the iterable should be
                                    either of the form
                                        (<type name>, <operation symbol>,
                                         <arity>)
                                    or
                                        ("coercion", <name of source type>,
                                         <name of target type>, <back-end>),
                                    where
                                      - <type name>, <name of source type>,
                                        and <name of target type> are names
                                        of types defined in `typelibrary',
                                      - <operation symbol> is an operator or
                                        test symbol in the Jumbala action
                                        language,
                                      - <arity> is the arity of the operator
                                        or test, and
                                      - <back-end> is one of the strings
                                        "python" or "Jumbala".
        solver                   -- An object supporting the method `unsat'
                                    which can be called with an
                                    actlang.expr.Expr object (representing a
                                    Boolean expression) as a parameter.  To
                                    guarantee that the generated abstract
                                    operations safely overapproximate the
                                    concrete ones, this method should return
                                    True only if (not necessarily if) the
                                    parameter expression is unsatisfiable.
                                    For example, instances of the
                                    smtlib_typegen.YICES_SMTLIB_QF_UFBV32
                                    class support the required interface.
        model_runtime_exceptions -- If True, the definition of each
                                    operation whose evaluation (on concrete
                                    integer operands) may result in a runtime
                                    exception on some operand combinations
                                    (division/remainder by zero, bit shifts
                                    by a negative number of bits) is extended
                                    such that the return values defined for
                                    the abstract operation will include the
                                    special value INVALID on those abstract
                                    operands whose concretizations cover the
                                    corresponding invalid concrete integer
                                    operand combinations.  The domains of
                                    types in the type library will be
                                    augmented with the INVALID value if
                                    necessary.
        stdout                   -- Unused.
        stderr                   -- If not None, use this stream for
                                    messages.
        verbose                  -- If True, display verbose output (if
                                    also `stderr' is not None).
    
    On success, the function updates `typelibrary' with automatically
    generated definitions for the given operations.  Should the generation
    of operations fail (e.g., due to inconsistencies in the type library),
    a TypeError or a ValueError (or any exception raised by the `unsat'
    method of the `solver' object) will be raised.  In these cases
    `typelibrary' will not be updated.

    (1) For all types referred to in the specified operations, the type
        library should define a coercion from the type "int" to each type
        for the "Jumbala" back-end.  Automatic definition of coercions
        from/to primitive types of the action language is not supported
        except for the possibility to fill in definitions of coercions from
        integers for the "python" back-end.
    """
    if not isinstance(typelibrary, TypeLibrary):
        raise TypeError("a type library expected as first argument")
    if not hasattr(operations, "__iter__"):
        raise TypeError \
                  ("an iterable of operations expected as second argument")
    if not hasattr(solver, "unsat") or not hasattr(solver.unsat, "__call__"):
        raise TypeError("a solver object expected as third argument")

    operations_to_generate = {}
    coercions_to_generate = set()
    typenames = set()
    for op in operations:
        if op[0] == "coercion":
            (fromtype, totype, backend) = (op[1], op[2], op[3])
            if not fromtype in typelibrary.typenames:
                raise ValueError \
                          ("coercion from undefined type `%s' requested"
                           % fromtype)
            if not totype in typelibrary.typenames:
                raise ValueError \
                          ("coercion to undefined type `%s' requested"
                           % totype)
            if not backend in ("Jumbala", "python"):
                raise ValueError \
                          ("coercion from type `%s' to type `%s' requested "
                           "for an unsupported back-end `%s'"
                           % (fromtype, totype, backend))
            if fromtype in typelibrary.STD_TYPES \
                    and not (fromtype == "int" and backend == "python"):
                raise ValueError \
                          ("cannot generate coercion from type `%s' to type "
                           "`%s' for back-end `%s' automatically"
                           % (fromtype, totype, backend))
            if totype in typelibrary.STD_TYPES:
                raise ValueError \
                          ("cannot generate coercions to primitive types")
            coercions_to_generate.add((fromtype, totype, backend))
            for t in (fromtype, totype):
                if not t in typelibrary.STD_TYPES:
                    typenames.add(t)
        else:
            (typename, symbol, arity) = (op[0], op[1], op[2])
            if typename in typelibrary.STD_TYPES:
                raise ValueError \
                          ("cannot define operations for primitive types")
            if not typename in typelibrary.typenames:
                raise ValueError \
                          ("`%s' is not a type defined in the type library"
                           % typename)
            if not typelibrary.domain[typename]:
                raise ValueError \
                          ("cannot define operations for a type `%s' with "
                           "empty domain" % typename)
            try:
                ops = operations_to_generate[typename]
            except KeyError:
                ops = set()
                operations_to_generate[typename] = ops
            typenames.add(typename)
            for (op, data) in _symbols.iteritems():
                if (symbol == data[SYMBOL_TOKEN_JUMBALA]
                        and arity == data[SYMBOL_ARITY]):
                    ops.add((op, data[SYMBOL_TYPE], symbol, arity))

    # Determine which types (should) contain the special value INVALID in their
    # domain
    types_including_invalid_value = \
        set([t for (t, ops) in operations_to_generate.iteritems()
             if "INVALID" in typelibrary.domain[t]
             or (model_runtime_exceptions and
                 frozenset([op[2] for op in ops])
                     .intersection(("/", "<<", ">>", ">>>")))])
    deps = [totype for (fromtype, totype, backend) in coercions_to_generate
            if fromtype in types_including_invalid_value]
    while deps:
        type = deps.pop()
        if not type in types_including_invalid_value:
            types_including_invalid_value.add(type)
            deps.extend \
                ([totype for (fromtype, totype, backend)
                  in coercions_to_generate if fromtype == type])

    # Ensure that the requested coercions will not make the ordering between
    # types in the type library cyclic.
    ordering = deepcopy(typelibrary.type_ordering)
    try:
        for (fromtype, totype, backend) in coercions_to_generate:
            ordering.orderTypes(fromtype, totype)
    except ConsistencyError:
        raise ValueError \
                  ("requested coercions would make the ordering between types "
                   "cyclic")

    if stderr:
        error.setErrorStream(stderr)

    # Mapping from type names into lists of Boolean predicates representing the
    # concretizations of the values in the domain of each abstract type.
    concretizations = {}

    new_lib = TypeLibrary()

    # Add types to the "new" type library, build the concretizations of
    # each abstract value of each abstract type, and define coercions from
    # type "int" for the "python" back-end if requested.
    for t in typenames:
        domain = []
        numdomain = []
        for (name, value) in typelibrary.domain[t].iteritems():
            domain.append(name)
            numdomain.append(value)
        new_lib.addType(t, domain, numdomain)

        # Extend the domain of the type with the special value INVALID if
        # necessary.
        if t in types_including_invalid_value and not "INVALID" in domain:
            new_lib.addDomainElements(t, "INVALID")

        domain = new_lib.domain[t].keys()
        numdomain = new_lib.numdomain[t].keys()

        new_lib.addCoercion("int", t)

        # the type library should (at the very least) define a coercion from
        # integers to the abstract type `t' for the Jumbala backend
        try:
            coercion = typelibrary.coercions[t]["int"]["Jumbala"]
        except KeyError:
            raise ValueError \
                      ("type library does not define a coercion from `int' to "
                       "type `%s' for the Jumbala back-end" % t)

        new_lib.coercions[t]["int"]["Jumbala"] = coercion
        if ("int", t, "python") in coercions_to_generate:
            python_coercion = []
            new_lib.coercions[t]["int"]["python"] = python_coercion
        else:
            python_coercion = None

        # Variables for handling definitions of abstraction mappings
        testcases_returning_abstract_value = {}
        for value in numdomain:
            testcases_returning_abstract_value[value] = set()
        actlang_tests = []
        explicated_tests = []

        # Generate Jumbala action language expressions (and Python code if
        # requested) from the user-defined abstraction mapping from integers to
        # the type `t'.
        for (testcase, (test, retvals)) in enumerate(coercion):
            # The abstraction mapping is defined as a list of tests
            #     test_0 : return_values_0
            #     test_1 : return_values_1
            #     ...
            # Update the mapping from abstract domain values to the set of test
            # cases whose return values include a particular domain element.
            if not retvals:
                raise ValueError \
                          ("abstraction mapping for type `%s' is not "
                           "completely defined" % t)
            for val in retvals:
                testcases_returning_abstract_value[val].add(testcase)
            # Translate the test into an action language expression object.
            try:
                actlang_exp = new_lib.parseJumbalaCoercionTestTemplate(test)
            except Exception, e:
                raise ValueError \
                          ("invalid abstraction mapping defined for type `%s':"
                           "%s" % (t, str(e)))
            # Generate Python code for the test if necessary.
            if python_coercion != None:
                python_coercion.append \
                    ((ExpressionTraverser.actlang_to_python(actlang_exp)
                      .replace("x", "<>"),
                      retvals))
            # Explicate the semantics of the current test case.  That is, if
            # the coercion is defined as a sequence of Boolean expressions
            #     test_0 : ...
            #     test_1 : ...
            #     test_2 : ...
            #     ...
            #     test_n : ...
            # then, test_1 will be (in effect) replaced with
            #     (not test_0) and test_1,
            # and test_i (2 <= i <= n) will be replaced with
            #     not (test_0 or test_1 or ... or test_{i-1}) and test_i to
            # model the intended semantics by which the test expressions are
            # evaluated sequentially.
            if testcase == 0:
                # first test case
                explicated_tests.append(actlang_exp)
            elif testcase == 1:
                # second test case
                explicated_tests.append \
                    (actlang.expr.CondAndExpr \
                         (actlang.expr.LogicalNotExpr(actlang_tests[0],
                                                      (0, 0)),
                          actlang_exp))
            else:
                e = None
                for f in actlang_tests:
                    e = (e and actlang.expr.CondOrExpr(e, f)) or f
                explicated_tests.append \
                    (actlang.expr.CondAndExpr \
                         (actlang.expr.LogicalNotExpr(e, (0, 0)), actlang_exp))
            actlang_tests.append(actlang_exp)

        # Add the Python coercion (if defined) to the type library.
        if python_coercion != None:
            new_lib.defineCoercion("int", t, "python", python_coercion, False)

        # (Try to) check the well-definedness of the abstraction mapping. 
        if stderr:
            e = actlang.expr.BoolLiteralExpr(False, (0, 0))
            for test in actlang_tests:
                e = actlang.expr.CondOrExpr(e, test)
            if not solver.unsat(actlang.expr.LogicalNotExpr(e, (0, 0))):
                error.warn \
                    ("type `%s': abstraction mapping may not be "
                     "well-defined" % t)

        # Define the concretization of each abstract value.  The concretization
        # of a given abstract value is the logical disjunction of all
        # predicates (i.e., tests in a coercion definition using the semantics
        # explicated above) whose evaluation to true for some concrete value
        # (integer) may result in the integer being abstracted to the given
        # abstract value.  For example, if the coercion is defined as
        #     test_1 : value_1
        #     test_2 : value_1, value_2
        # then the concretization of value_1 will be "test_1 or test_2" (with
        # test_1 and test_2 explicated as above), and the concretization of
        # value_2 will be "test_2".  If the domain of the abstract type
        # contains a value to which no integer will ever be coerced (i.e., the
        # value does not appear on the right hand side of any test in the
        # coercion definition), the concretization of this value will be
        # "false".
        concretization = {}
        concretizations[t] = concretization
        for (value, testcases) in \
                testcases_returning_abstract_value.iteritems():
            e = None
            for testcase in testcases:
                e = (e
                     and actlang.expr.CondOrExpr
                             (e, explicated_tests[testcase])) \
                    or explicated_tests[testcase]
            if not e:
                e = actlang.expr.BoolLiteralExpr(False, (0, 0))
            concretization[value] = e

    # Generate definitions of operators and tests
    for (t, operations) in operations_to_generate.iteritems():
        domain = new_lib.domain[t].keys()
        numdomain = new_lib.numdomain[t].keys()
        concretization = concretizations[t]
        if verbose and stderr:
            error.inform("generating operations for type `%s'" % t)
        for (op, op_type, op_sym, op_arity) in operations:
            if verbose and stderr:
                stderr.write \
                    ("    generating operator `%s' of arity %d\n"
                     % (op_sym, op_arity))
            if op_type == "operator":
                ops = new_lib.operators[t]
            else:
                ops = new_lib.tests[t]
            _generateOperation \
                (solver, op, new_lib.domain[t], concretization, ops)
            # if requested, model the runtime exceptions caused by some
            # arithmetic operations (division/remainder by zero, shifts by a
            # negative number of bit positions)
            if model_runtime_exceptions \
                    and op in (actlang.binaryop.Divide,
                               actlang.binaryop.Remainder,
                               actlang.binaryop.ShiftLeft,
                               actlang.binaryop.SignedShiftRight,
                               actlang.binaryop.UnsignedShiftRight):
                dom = [numdomain]
                if op in (actlang.binaryop.Divide, actlang.binaryop.Remainder):
                    # Test that the value 0 is included in the
                    # concretization of some abstract value.
                    f = actlang.expr.InfixExpr \
                            (actlang.names.SimpleName("x", (0, 0)),
                             actlang.expr.IntLiteralExpr(0, (0, 0)),
                             actlang.binaryop.Equal((0, 0)))
                    error_msg = \
                        "the value 0 has no corresponding abstract value in " \
                        "the domain of type `%s'" % t
                else:
                    # Test that there is at least one abstract value which
                    # includes negative values in its concretization.
                    f = actlang.expr.InfixExpr \
                            (actlang.names.SimpleName("x", (0, 0)),
                             actlang.expr.IntLiteralExpr(0, (0, 0)),
                             actlang.binaryop.LessThan((0, 0)))
                    error_msg = \
                        "no abstract value in the domain of type `%s' " \
                        "represents any negative integers" % t
                values_causing_runtime_exceptions = []
                for (value, c) in concretization.iteritems():
                    e = actlang.expr.CondAndExpr(f, c)
                    if not solver.unsat(e):
                        values_causing_runtime_exceptions.append(value)
                if not values_causing_runtime_exceptions:
                    raise ValueError(error_msg)
                dom.append(values_causing_runtime_exceptions)
                invalid = new_lib.domain[t]["INVALID"]
                for args in cartesize(*dom):
                    retvals = ops[op_arity][op_sym][tuple(args)]
                    if not invalid in retvals:
                        retvals.append(invalid)
            # ensure that the abstract operation is well-defined for all
            # inputs
            dom = op_arity * [numdomain]
            for args in cartesize(*dom):
                retvals = ops[op_arity][op_sym][tuple(args)]
                if not retvals:
                    if stderr:
                        error.warn \
                            ("type `%s': no admissible return values for "
                             "operation `%s(%s)', using the maximal range"
                             % (t, op_sym,
                                ", ".join([new_lib.numdomain[t][i]
                                           for i in args])))
                    if op_type == "operator":
                        retvals = numdomain[:]
                    else:
                        retvals = [True, False]
                    ops[op_arity][op_sym][tuple(args)] = retvals

    # Generate user-requested coercions between user-defined abstract types.
    coercion_dfns = {}
    for (fromtype, totype, backend) in coercions_to_generate:
        if fromtype in new_lib.STD_TYPES \
               or not backend in ("python", "Jumbala"):
            continue
        if stderr and verbose:
            error.inform \
                ("generating coercion from type `%s' to type `%s' for "
                 "back-end `%s'" % (fromtype, totype, backend))
        try:
            new_lib.defineCoercion \
                (fromtype, totype, backend, coercion_dfns[(fromtype, totype)],
                 False)
        except KeyError:
            dfn = []
            for (fromvalname, fromval) in new_lib.domain[fromtype].iteritems():
                if fromvalname == "INVALID":
                    retvals = [new_lib.domain[totype]["INVALID"]]
                else:
                    retvals = []
                    c = concretizations[fromtype][fromval]
                    for toval in new_lib.domain[totype].itervalues():
                        e = actlang.expr.CondAndExpr \
                                (c, concretizations[totype][toval])
                        if not solver.unsat(e):
                            retvals.append(toval)
                    if not retvals:
                        if stderr:
                            error.warn \
                                ("no admissible return values for mapping "
                                 "`%s:%s' to a value of type `%s', using the "
                                 "maximal range"
                                 % (fromtype, fromvalname, totype))
                        retvals = new_lib.domain[totype].values()
                dfn.append(("<> == <%s>" % fromvalname, retvals))
            coercion_dfns[(fromtype, totype)] = dfn
            new_lib.defineCoercion(fromtype, totype, backend, dfn, False)

    # Copy the defined operations to the original type library...
    for t in operations_to_generate.iterkeys():
        # update domains because some domains may have been extended with the
        # INVALID value
        typelibrary.domain[t] = new_lib.domain[t]
        typelibrary.numdomain[t] = new_lib.numdomain[t]
    # ... operators and tests ...
    for (source_operations, target_operations) in \
            ((new_lib.operators, typelibrary.operators),
             (new_lib.tests, typelibrary.tests)):
        for (typename, arities) in source_operations.iteritems():
            target_arities = target_operations[typename]
            for (arity, syms) in arities.iteritems():
                try:
                    target_dfns = target_arities[arity]
                except KeyError:
                    target_dfns = {}
                    target_arities[arity] = target_dfns
                for (symbol, dfn) in syms.iteritems():
                    target_dfns[symbol] = dfn
    # ... coercions
    for (totype, fromtypes) in new_lib.coercions.iteritems():
        target_fromtypes = typelibrary.coercions[totype]
        for (fromtype, backends) in fromtypes.iteritems():
            if not fromtype in target_fromtypes:
                typelibrary.addCoercion(fromtype, totype)
            target_backends = target_fromtypes[fromtype]
            for (backend, dfn) in backends.iteritems():
                target_backends[backend] = dfn
