# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Facilities for recording constraints on the types of attributes in a model.

This module provides a class for representing associations between
abstract attributes (or expressions) and their types, and, in
particular, constructing such associations from external type
association files and project files containing abstracted models.

Author: Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

import actlang
import coral
import error
from Utils.Models import get_elements_of_type
from Utils.ExpressionTraverser import path_to_subexpression

_UML = coral.lang.UML14

class InputFormatError(error.Error):
    """Class for reporting parse errors."""
    pass

class TypeAssociation:
    """A class representing associations between attributes and their types.

    The principal functionality provided by the class is to initialize
    type associations from the contents of a type association file or
    from the tagged values included in an abstracted UML model.  When
    working with abstracted models, this class can also be used to record
    the correspondence between action language subexpressions in concrete
    and abstract models.

    A type association file consists of zero or more lines of the form
        import <filename>
    followed by zero or more lines of the form
        <UUID> <constraint-operator> <type-name>
    where
        <filename>            is the path name of a type specification file
                              to be used for building an abstract type
                              library (if the file name begins or ends with
                              white space, enclose it in single or double
                              quotes),
        <UUID>                is the UUID of the attribute,
        <constraint-operator> is one of the strings "==", "<=", or ">="
                              (excluding the quotes) separated from <UUID>
                              by white space, and
        <type-name>           is the name of the abstract type of the
                              attribute with UUID <UUID> (assumed to be
                              defined in one of the imported abstract type
                              specification files).

    Every TypeAssociation object has the following attributes:
        libfiles
            A list of names of library files "import"ed in the input (or
            listed in the SMUML_typelibrary tagged value of an abstracted
            UML model).
        constraints
            A dictionary mapping UUID's of attributes to their associated
            constraints.  For an attribute with UUID `id',
            `constraints[id]' returns a list of constraints associated with
            the attribute.  Every constraint is a triple
                (<operator>, <type name>, <lineno>)
            where
                <operator>  is one of "==", "<=", and ">=",
                <type name> is a type name, and
                <lineno>    records (if the TypeAssociation object was
                            initialized from a type association file) the
                            number of the line in the input in which the
                            constraint was defined.  If the object was
                            initialized from an abstracted UML model, each
                            `lineno' will have the value -1).
        expr_correspondence
            If the object was initialized from a type association file,
            this attribute contains an empty dictionary.  If the object
            was initialized from an abstract UML model, this attribute
            represents a mapping from action language expressions in the
            concrete model to information about corresponding action
            language expressions in the abstract model.  More precisely,
            given an action language expression (an actlang.expr.Expr)
            object `e' from (a collection of parsed action language
            expressions in) a concrete model s.t. `e' is a key in this
            dictionary, `expr_correspondence[e]' is an object (of class
            TypeAssociation.ExprCorrespondenceData) with the following
            attributes:
                expr       an actlang.expr.Expr object representing the
                           abstracted action language expression
                           corresponding to `e',
                typename   is the abstract type of the abstracted
                           expression, and
                flattened  is a truth value telling whether the abstracted
                           expression is "flattened" or not (i.e., whether
                           the subexpressions of the concrete expression
                           `e' have corresponding abstract expressions in
                           the `expr_correspondence' dictionary).
            Not all concrete expressions necessarily have corresponding
            abstract expressions in the dictionary (this is possible if one
            of the abstract expressions corresponding to another concrete
            expression which contains the concrete expression `e' as a
            subexpression has  been flattened).  All "top-level"
            subexpressions (i.e., expressions which are not part of a
            larger subexpression, such as guards of state machine
            transitions and top-level expressions in statements in the
            effects of transitions) in the concrete model are nevertheless
            guaranteed to have corresponding abstract expressions.
    """

    def __init__(self):
        """Initialize an empty type association."""
        self.libfiles = []
        self.constraints = {}
        self.expr_correspondence = {}
        
    def read(self, f):
        """Initialize a type association structure from a file.

        Argument:
            f -- An open file object.

        This function will initialize `self.libfiles' and
        `self.constraints' from a type association file.
        `self.correspondence' will be initialized to an empty dictionary.

        An InputFormatError will be raised if the input does not match
        the expected syntax.
        """
        self.__init__()
        lineno = 0
        in_header = True
        for line in f:
            lineno += 1
            pos = line.find('#')
            if pos != -1:
                line = line[:pos]
            line = line.strip()
            if not line:
                continue
            if in_header:
                tokens = [s.strip() for s in line.split(None, 1)]
                keyword = tokens[0]
                if keyword != "import":
                    # assume end of header listing type specification files to
                    # be imported
                    in_header = False
                else:
                    # expect the line to be of the form 'import <filename>',
                    # where <filename> may be optionally enclosed in single or
                    # double quotes
                    if len(tokens) < 2:
                        raise InputFormatError \
                              ("%d:type specification file name expected"
                               % lineno)
                    filename = tokens[1]
                    if TypeAssociation._TypeAssociation__quoted(filename):
                        filename = filename[1:-1]
                    self.libfiles.append(filename)
                    continue
            prefix_and_typename = [s.strip() for s in line.rsplit(None, 1)]
            if len(prefix_and_typename) != 2:
                raise InputFormatError("%d:syntax error" % lineno)
            typename = prefix_and_typename[1]
            if typename[0:2] in ("==", "<=", ">="):
                (op, typename) = (typename[0:2], typename[2:])
                xmi_id = prefix_and_typename[0]
            else:
                tokens = [s.strip() for s in
                          prefix_and_typename[0].rsplit(None, 1)]
                if len(tokens) != 2:
                    raise InputFormatError("%d:syntax error" % lineno)
                (xmi_id, op) = tuple(tokens)
            if not op in ("==", "<=", ">="):
                raise InputFormatError \
                      ("%d:constraint operator expected before type name"
                       % lineno)
            if not xmi_id in self.constraints:
                self.constraints[xmi_id] = []
            self.constraints[xmi_id].append((op, typename, lineno))

    def __quoted(s):
        """Test whether a string starts and ends with matching quotes."""
        return len(s) > 1 and ((s[0] == '"' and s[-1] == '"')
                               or (s[0] == "'" and s[-1] == "'"))

    __quoted = staticmethod(__quoted)

    def initFromModel(self, model, concrete_stmt_dict):
        """Initialize a type association structure from an abstract model.

        Arguments:
            model              -- An abstract UML model (a
                                  coral.lang.ElementPtr object satisfying
                                  model.oclIsKindOf(coral.lang.UML14.Model)
                                    == True).  The model should also
                                  satisfy the additional constraints
                                  described below.
            concrete_stmt_dict -- A dictionary mapping UUID's of guards
                                  and effects of transitions in state
                                  machines in `model' to lists of
                                  actlang.statement.Statement objects
                                  representing action language code in the
                                  concrete model from which the abstract
                                  model `model' was obtained.  It is
                                  assumed that (i) the UUID's of guards
                                  and effects of transitions are the same
                                  in the concrete and abstract models,
                                  (ii) the number of action language
                                  statements in a guard or effect of a
                                  particular transition is the same in
                                  both models (guards are thought to
                                  consist of a single statement, see
                                  below), and (iii) corresponding
                                  statements in the concrete and abstract
                                  model are of the same type and contain
                                  the same number of top-level expressions.
                                  The dictionary is also subject to the
                                  following restrictions:
                                    * Each statement in the list of
                                      statements associated with the UUID
                                      of a guard or effect of a transition
                                      must be parsed without constructing
                                      any actlang.Expr.ParenExprs.  (Use
                                      the generateParenExpr=False parameter
                                      for `actlang.grammar.parse'.)
                                    * The list of statements associated in
                                      this dictionary with each non-empty
                                      guard of a transition must consist
                                      of a single
                                      actlang.statement.AssertStatement
                                      with the guard expression as the
                                      assertion expression.

        The additional requirements for the abstracted model `model' are
        (this information is mainly useful for the purpose of examining
        abstract model files):
          * `model.ownedElement' should contain three UML TagDefinitions
            named "SMUML_typelibrary", "SMUML_typeassignment", and
            "SMUML_expr_data".
          * `model.taggedValue' should include exactly one UML
            TaggedValue `v' satisfying
                str(v.type.name) == "SMUML_typelibrary"
            such that str(v.dataValue) is a semicolon-separated string of
            type library file names (with literal ;'s and \'s escaped by
            preceding them with \).  These file names represent the type
            library files used to define abstract types and operations.
            (Warning!  This method will not try to check whether the
            file names actually refer to valid type libary files; this
            is the responsibility of the user of the TypeAssociation
            object.)
          * `model.taggedValue' should include exactly one UML
            TaggedValue `w' satisfying
                str(w.type.name) == "SMUML_typeassignment"
            such that str(w.dataValue) is a semicolon-separated list of
            constraints of the form <UUID>==<type name>, where <UUID>
            is the UUID is an attribute in the model, and <type name> is
            a name for the type of the attribute (a type defined in one
            of the type library files listed in the tagged value
            "SMUML_typelibrary").  (Warning!  This method will not
            validate that <UUID> (<type name>) is actually a valid
            UUID of a model attribute (a type name defined in a type
            library); this is the responsibility of the user of the
            TypeAssociation object.)
          * A guard or an effect of a transition in `model' may include
            at most one tagged value `d' satisfying
                str(d.type.name) == "SMUML_expr_data"
            such that str(d.dataValue) is a string which can be
            generated from the non-terminal <data> using the following
            grammar:
               <data> ::= <stmt-data> `|' <data>
                        | <*empty*>
               <stmt-data> ::= <toplev-expr-data> `;' <stmt-data>
                             | <*empty*>
               <toplev-expr-data> ::= <expr-data> `!' <toplev-expr-data>
                                    | <*empty*>
               <expr-data> ::= <location> `.' <location> `.' <type name>
                                   `.' <flattened>
            where
               nonterminals are enclosed in < >,
               literals are enclosed in ` ',
               there should not be any white space between terminals or
                   nonterminals,
               <location>  is a (possibly empty) comma-separated list of
                           non-negative integers,
               <type name> is a name of an abstract type (which should
                           be a type defined in one of the type library
                           files listed in the "SMUML_typelibrary" tagged
                           value; however, no validation is actually
                           applied to the type name), and
               <flattened> is either "0" or "1".

            The number of <stmt-data> segments in <data> should match the
            number of statements in the guard or effect of the transition
            (guards of transitions are assumed to consist of a single
            statement).  The number of <toplev-expr-data> segments in the
            i'th <stmt-data> segment of <data> should be equal to the
            number of the top-level statements in the i'th statement in
            the guard or effect of the transition.  The two locations in
            each <expr-data> segment in the j'th <toplev-expr-data>
            segment of the i'th <stmt-data> segment of <data> should
            identify a pair of corresponding concrete and abstract
            subexpressions in the j'th top-level action language
            expression (counted from left to right) of the i'th statement
            of the guard or effect of the transition.  (All indices start
            from 0.)  The additional fields in each <expr-data> segment
            tell the name of the abstract type of the abstract
            subexpression, and whether the abstract subexpression is
            flattened ("1") or not ("0").

        Raises a TypeError or a ValueError if `model' or
        `concrete_stmt_dict' does not conform to the above requirements.
        In this case the TypeAssociation object will retain its old state.
        Otherwise, replaces the contents of `self.libfiles',
        `self.constraints' and `self.expr_correspondence' with information
        from `model'.
        """
        if not isinstance(model, coral.kernel.ElementPtr) or \
                not model.oclIsKindOf(_UML.Model):
            raise TypeError("a coral.lang.UML14.Model expected as argument")

        tagvars = \
            {"SMUML_typelibrary" : False, "SMUML_typeassignment" : False,
             "SMUML_expr_data" : False}

        for e in model.ownedElement.select \
                (lambda e: e.oclIsKindOf(_UML.TagDefinition)
                           and str(e.name) in tagvars):
            if tagvars[str(e.name)]:
                raise ValueError \
                          ("model contains multiple definitions of tag `%s'"
                           % tag)
            tagvars[str(e.name)] = True
        for (tag, present) in tagvars.iteritems():
            if not present:
                raise ValueError \
                          ("model does not contain a definition for tag `%s'"
                           % tag)

        libfiles = []
        constraints = {}
        expr_correspondence = {}
        for v in model.taggedValue:
            typename = str(v.type.name)
            if typename == "SMUML_typelibrary":
                if libfiles:
                    raise ValueError \
                              ("model contains multiple instances of tagged "
                               "value of type `SMUML_typelibrary'")
                name = []
                escape = False
                for c in str(v.dataValue):
                    if c == "\\" and not escape:
                        escape = True
                    elif c == ';' and not escape:
                        libfiles.append("".join(name))
                        name = []
                    else:
                        escape = False
                        name.append(c)
                libfiles.append("".join(name))
            elif typename == "SMUML_typeassignment":
                if constraints:
                    raise ValueError \
                              ("model contains multiple instances of tagged "
                               "value of type `SMUML_typeassignment'")
                for a in str(v.dataValue).split(";"):
                    zz = a.split("==")
                    if len(zz) != 2:
                        raise ValueError \
                                  ("error parsing tagged value "
                                   "`SMUML_typeassignment' in model")
                    (uuid, typename) = zz
                    try:
                        constraints[uuid].append(("==", typename, -1))
                    except KeyError:
                        constraints[uuid] = [("==", typename, -1)]

        for m in get_elements_of_type(model, _UML.StateMachine):
            for t in m.transitions:
                for elem in ("guard", "effect"):
                    try:
                        element = getattr(t, elem)
                    except AttributeError:
                        continue
                    if not element:
                        continue
                    uuid = element.__XMIid__()
                    try:
                        concrete_statements = concrete_stmt_dict[uuid]
                    except KeyError:
                        raise ValueError \
                                  ("transition with UUID %s has a non-empty "
                                   "%s, whose UUID is not found in the "
                                   "dictionary given as argument"
                                   % (t.__XMIid__(), elem))
                    if elem == "guard":
                        code = "assert (%s);" % str(element.expression.body)
                    else:
                        code = str(element.script.body)
                    try:
                        abstract_statements = \
                            actlang.grammar.parse \
                                (code, generateParenExpr=False) \
                                    .api_getTopStatements()
                    except actlang.errors.Error:
                        raise ValueError \
                                  ("error parsing action language code in %s "
                                   "of transition with UUID %s in the given "
                                   "model" % (elem, t.__XMIid__()))
                    if len(concrete_statements) != len(abstract_statements):
                        raise ValueError \
                                  ("number of statements associated with %s "
                                   "of transition with UUID %s in the given "
                                   "model differs from the number of "
                                   "statements associated with the "
                                   "corresponding element in the dictionary"
                                   "given as argument"
                                   % (elem, t.__XMIid__()))
                    try:
                        for v in element.taggedValue.select \
                                (lambda v:
                                     str(v.type.name) == "SMUML_expr_data"):
                            if element in expr_correspondence:
                                raise ValueError \
                                          ("model contains multiple "
                                           "instances of tagged value "
                                           "`SMUML_expr_data'")
                            self.__processStmtCorrespondence \
                                (str(v.dataValue), concrete_statements,
                                 abstract_statements, expr_correspondence)
                    except ValueError, e:
                        raise ValueError \
                                  (str(e)
                                   + " in %s of transition with UUID %s"
                                     % (elem, t.__XMIid__()))

        self.__init__()
        self.libfiles = libfiles
        self.constraints = constraints
        self.expr_correspondence = expr_correspondence

    class ExprCorrespondenceData:
        """Information about correspondence between expressions.

        Each object of this class has the following attributes:
            expr       an actlang.expr.Expr object representing an
                       abstracted action language expression,
            typename   the name of the abstract type of the abstracted
                       expression, and
            flattened  a truth value telling whether the abstracted
                       expression is "flattened" or not.
        """

        def __init__(self, expr, typename, flattened):
            """Initialize an ExprCorrespondenceData object."""
            self.expr = expr
            self.typename = typename
            self.flattened = flattened

    def __processStmtCorrespondence(self, data, concrete_statements, abstract_statements, expr_correspondence):
        """Initialize correspondence between expressions in statements.

        Arguments:
            data                -- A string to be decoded for information
                                   on the correspondence between action
                                   language expressions (more precisely,
                                   a string generated from the nonterminal
                                   <data> using the grammar described in
                                   the documentation for initFromModel).
            concrete_statements -- A list of actlang.statement.Statement
                                   objects to scan for the expressions to
                                   associate with expressions contained in
                                   statements in `abstract_statements'.
            abstract_statements -- A list of actlang.statement.Statement
                                   objects.  The length of this list
                                   should be equal to that of
                                   `concrete_statements'; furthermore,
                                   the types of statements in the statement
                                   lists should match, and each pair of
                                   corresponding statements should have the
                                   same number of top-level expressions.
            expr_correspondence -- A dictionary from actlang.expr.Expr
                                   objects (expressions contained in
                                   statements in `concrete_statements') to
                                   triples
                                       (<expr>, <type name>, <flattened>)
                                   where <expr> in an actlang.expr.Expr
                                   object (an expression contained in a
                                   statement in `abtract_statements'),
                                   <type name> is the name of the abstract
                                   type of <expr>, and <flattened> is a
                                   Boolean value telling whether <expr> is
                                   a flattened abstract expression or not.

        This method populates the `expr_correspondence' dictionary by
        associating (sub)expressions contained in statements of
        `concrete_statements' with information about corresponding
        subexpessions in statements of `abstract_statements'.

        Raises a ValueError if an inconsistency is detected.
        """
        stmts_data = data.split("|")
        if len(stmts_data) != len(concrete_statements):
            raise ValueError \
                      ("number of statements (%d) listed in the data value of "
                       "tagged value `SMUML_expr_data' does not match the "
                       "expected one (%d)"
                       % (len(stmts_data), len(concrete_statements)))
        for (i, (cs, as, stmt_data)) in enumerate \
                (zip(concrete_statements, abstract_statements, stmts_data)):
            if type(cs) != type(as):
                raise ValueError \
                          ("statement type mismatch between concrete "
                           "and abstract model (statement %d)" % i)
            if isinstance(cs, (actlang.statement.AssertStatement,
                               actlang.statement.ExprStatement)):
                toplev_expr_data = stmt_data.split(";")
                if toplev_expr_data == [""]:
                    toplev_expr_data = []
                if len(toplev_expr_data) != 1:
                    raise ValueError \
                              ("number of top-level expressions (%d) "
                               "associated with statement %d in data value of "
                               "tagged value `SMUML_expr_data' does not match "
                               "the expected one (1)"
                               % (len(toplev_expr_data), i))
                try:
                    self.__processExprCorrespondence \
                        (toplev_expr_data[0], cs.api_getExpr(),
                         as.api_getExpr(), expr_correspondence)
                except ValueError, e:
                    raise ValueError \
                              (str(e)
                               + " (in top-level expression 0 of statement %d)"
                                 % i)
            elif isinstance(cs, actlang.statement.EmptyStatement):
                toplev_expr_data = stmt_data.split(";")
                if toplev_expr_data != [""]:
                    raise ValueError \
                              ("number of top-level expressions (%d) "
                               "associated with statement %d in data value of "
                               "tagged value `SMUML_expr_data' does not match "
                               "the expected one (0)"
                               % (len(toplev_expr_data), i))
            elif isinstance(cs, actlang.statement.SendStatement):
                toplev_expr_data = stmt_data.split(";")
                c_args = cs.api_getArgList()
                if len(toplev_expr_data) != len(c_args) + 1:
                    raise ValueError \
                              ("number of top-level expressions (%d) "
                               "associated with statement %d in data value of "
                               "tagged value `SMUML_expr_data' does not match "
                               "the expected one (%d)"
                               % (len(toplev_expr_data), i, len(c_args) + 1))
                a_args = as.api_getArgList()
                if len(c_args) != len(a_args):
                    raise ValueError \
                              ("parameter number mismatch in `send' statement "
                               "(statement %d) between concrete and abstract "
                               "model" % i)
                try:
                    for (j, (data, ca, aa)) in enumerate \
                            (zip(toplev_expr_data, c_args, a_args)):
                        self.__processExprCorrespondence \
                            (data, ca, aa, expr_correspondence)
                    j = len(toplev_expr_data) - 1
                    self.__processExprCorrespondence \
                        (toplev_expr_data[-1], cs.api_getTarget(),
                         as.api_getTarget(), expr_correspondence)
                except ValueError, e:
                    raise ValueError \
                              (str(e)
                               + " (in top-level expression %d of statement "
                                 "%d)" % (j, i))

    def __processExprCorrespondence(self, toplevel_expr_data, toplevel_concrete_expr, toplevel_abstract_expr, expr_correspondence):
        """Initialize correspondence between expressions.

        Arguments:
            toplevel_expr_data     -- A string to be decoded for
                                      information on the correspondence
                                      between subexpressions of
                                      `toplevel_concrete_expr' and
                                      `toplevel_abstract_expr' (more
                                      precisely, a string generable from
                                      the nonterminal <toplev-expr-data>
                                      using the grammar defined in the
                                      documentation of initFromModel).
            toplevel_concrete_expr -- An actlang.expr.Expr object
                                      representing a "concrete" action
                                      language expression.
            toplevel_abstract_expr -- An actlang.expr.Expr object
                                      representing an "abstract" action
                                      language expression.
            expr_correspondence    -- A dictionary mapping actlang.expr.Expr
                                      objects to triples (see the
                                      documentation of
                                      __processStmtCorrespondence).

        This method populates the `expr_correspondence' dictionary by
        associating subexpressions of `toplevel_concrete_expr' with
        information about corresponding subexpessions of
        `toplevel_abstract_expr'.

        Raises a ValueError if an inconsistency is detected.
        """
        for subexpr_data in toplevel_expr_data.split("!"):
            expr_data = subexpr_data.split(".")
            if len(expr_data) != 4:
                raise ValueError \
                          ("error parsing data value of tagged value "
                           "`SMUML_expr_data'")
            (conc_loc, abst_loc, typename, flattened) = expr_data
            try:
                conc_loc = \
                    conc_loc and tuple(map(int, conc_loc.split(","))) or ()
                concrete_expr = \
                    path_to_subexpression(toplevel_concrete_expr, conc_loc)
            except ValueError:
                raise ValueError \
                          ("invalid location `%s' of concrete subexpression "
                           "in data value of tagged value `SMUML_expr_data'"
                           % conc_loc)
            try:
                abst_loc = \
                    abst_loc and tuple(map(int, abst_loc.split(","))) or ()
                abstract_expr = \
                    path_to_subexpression(toplevel_abstract_expr, abst_loc)
            except ValueError:
                raise ValueError \
                          ("invalid location `%s' of abstract subexpression "
                           "in data value of tagged value `SMUML_expr_data'"
                           % abst_loc)
            try:
                data = expr_correspondence[concrete_expr]
                if data != (abstract_expr, typename, flattened):
                    raise ValueError \
                              ("conflicting expression correspondence "
                               "information for concrete expression at "
                               "location `%s' in data value of tagged value "
                               "`SMUML_expr_data'" % conc_loc)
            except KeyError:
                expr_correspondence[concrete_expr] = \
                    TypeAssociation.ExprCorrespondenceData \
                        (abstract_expr, typename,
                         (flattened == "1" and True) or False)
