# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Partial ordering between types according to their precision.

This module provides a class for representing partial orders between
types in an abstract type library.

Author: Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

import error

class ConsistencyError(error.Error):
    """An exception class for reporting fatal consistency errors."""
    pass

class TypeNode:
    """A node in the type hierarchy.

    Each TypeNode object has the following attributes:
        name          -- Name of the type associated with the node.
        more_abstract -- A list of TypeNodes representing types
                         "immediately more abstract" than the type
                         associated with the node.
        more_precise  -- A list of TypeNodes representing types
                         "immediately more precise" than the type
                         associated with the node.
    """
    def __init__(self, name):
        self.name = name
        self.more_abstract = []
        self.more_precise = []

    def findLessAbstractNodes(self):
        """Find all nodes representing strictly more precise types."""
        result = set()
        self.collectReachableNodes(lambda x: x.more_precise, result)
        return result

    def findMoreAbstractNodes(self):
        """Find all nodes representing strictly more abstract types."""
        result = set()
        self.collectReachableNodes(lambda x: x.more_abstract, result)
        return result

    def collectReachableNodes(self, succ, result):
        """Find the nodes reachable from the node.

        Arguments:
            succ   -- Successor function between nodes.
            result -- A set for collecting the result.

        A node is reachable from itself only if it can be reached by at
        least one application of `succ'.
        """
        for n in succ(self):
            if not n in result:
                result.add(n)
                n.collectReachableNodes(succ, result)

class TypeHierarchy:
    """Partial ordering between types defined in a type library."""

    (CMP_LT, CMP_LE, CMP_EQ, CMP_GE, CMP_GT, CMP_IC) = range(0, 6)
    __s = ('<', '<=', '==', '>=', '>', 'X')

    def __init__(self):
        # Mapping from names of types known to the type hierarchy to their
        # corresponding nodes in the hierarchy.
        self.type_nodes = {}

    def __contains__(self, x):
        """Test whether a given type name is known to the hierarchy."""
        if isinstance(x, TypeNode):
            return x in self.type_nodes.values()
        else:
            return x in self.type_nodes

    def __getitem__(self, typename):
        """Get the type node corresponding to a type name."""
        return self.type_nodes[typename]

    def compare(self, x1, x2):
        """Compare two types by their relative precision.

        Arguments:
            x1, x2 -- Type names or type nodes in the type hierarchy.

        Returns:
            TypeHierarchy.CMP_LT  if (the type corresponding to) x1 is
                                  strictly less abstract than (the type
                                  corresponding to) x2
            TypeHierarchy.CMP_EQ  if (the type corresponding to) x1 and
                                  (the type corresponding to) x2 are equal
            TypeHierarchy.CMP_GT  if (the type corresponding to) x1 is
                                  strictly more abstract than (the type
                                  corresponding to) x2
            TypeHierarchy.CMP_IC  if (the type corresponding to) x1 and
                                  (the type corresponding to) x2 are
                                  incomparable
        """
        return self.__compare \
               (self.__normalize(x1), self.__normalize(x2),
                lambda x: x.more_abstract)

    def __normalize(self, x):
        """Return the type node corresponding to the argument.

        Argument:
            x -- A type name or a type node in the type hierarchy.

        If `x' is a type name, returns the type node corresponding to the
        name in the type hierarchy.  If `x' is a type node in the type
        hierarchy, returns `x' itself.
        """
        if isinstance(x, TypeNode):
            if x in self:
                return x
            else:
                raise ValueError \
                      ("argument is not a node in the type hierarchy")
        else:
            return self.type_nodes[x]

    def __compare(self, n1, n2, succ):
        """Determine the order of two nodes in the type hierarchy.

        Arguments:
            n1, n2 -- Nodes in the type hierarchy.
            succ   -- The successor function.  `succ(n)' (for n in n1,
                      n2) should return the successor nodes of n in the
                      ordering between types.

        Returns:
            TypeHierarchy.CMP_LT  if n1 != n2 and n1 is reachable from n2
                                  in the ordering
            TypeHierarchy.CMP_EQ  if n1 and n2 are the same node
            TypeHierarchy.CMP_GT  if n1 != n2 and n2 is reachable from n1
                                  in the ordering
            TypeHierarchy.CMP_IC  if neither of the two nodes is
                                  reachable from the other node
        """
        if n1 == n2:
            return self.CMP_EQ
        if TypeHierarchy._TypeHierarchy__isReachable(n1, n2, succ, set()):
            return self.CMP_LT
        elif TypeHierarchy._TypeHierarchy__isReachable(n2, n1, succ, set()):
            return self.CMP_GT
        return self.CMP_IC

    def join(self, *types):
        """Find the most precise type that is more abstract than given types.

        Argument:
            types -- One or more type names or type nodes, or a list of type
                     names or type nodes.

        Returns:
            The name of the type which is the least upper bound of the
            given types in the abstraction ordering, or None if such a
            bound does not exist.
        """
        if len(types) == 1 and hasattr(types, "__iter__"):
            types = types[0]
        n = self.__findBound \
            ([self.__normalize(t) for t in types], lambda x: x.more_abstract)
        return (n and n.name) or None

    def meet(self, *types):
        """Find the most abstract type that is more precise than given types.

        Argument:
            types -- One or more type names or type nodes, or a list of type
                     names or type nodes.

        Returns:
            The name of the type which is the greatest lower bound of the
            given types in the abstraction ordering, or None if such a
            bound does not exist.
        """
        if len(types) == 1 and hasattr(types, "__iter__"):
            types = types[0]
        n = self.__findBound \
            ([self.__normalize(t) for t in types], lambda x: x.more_precise)
        return (n and n.name) or None

    def __findBound(self, nodes, succ):
        """Determine the "least" node reachable from a set of type nodes.

        Arguments:
            nodes -- A list of type nodes.
            succ  -- The successor function.  `succ(n)' (for n in nodes)
                     should return the successor nodes of n in the
                     ordering between types.  The ordering is assumed to
                     be acyclic.

        Returns the "least" node which is reachable from all of the given
        nodes in the type hierarchy, or None if there is no such node.
        """
        if not nodes:
            return None
        result = set(self.type_nodes.values())
        for n in nodes:
            reach = set([n])
            n.collectReachableNodes(succ, reach)
            result.intersection_update(reach)
            if not result:
                return None
        to_remove = set()
        for n1 in result:
            for n2 in result:
                if n1 != n2 and self.__compare(n1, n2, succ) == self.CMP_LT:
                    to_remove.add(n2)
        result.difference_update(to_remove)
        if len(result) == 1:
            return result.pop()
        return self.__findBound(result, succ)

    def addType(self, t):
        """Add a new type to the precision ordering.

        Argument:
            t -- A type name.

        Returns a TypeNode corresponding to the new type added to the
        type hierarchy.
        """
        if not isinstance(t, str):
            raise TypeError("expected string as argument")
        if t in self.type_nodes:
            raise ValueError("type `%s' is already defined" % t)
        node = TypeNode(t)
        self.type_nodes[t] = node
        return node

    def removeType(self, t):
        """Remove a type from the precision ordering.

        Argument:
            t -- A type name or a type node in the type hierarchy.
        """
        node = self.__normalize(t)
        for n in self.type_nodes.itervalues():
            while node in n.more_abstract:
                n.more_abstract.remove(node)
            while node in n.more_precise:
                n.more_precise.remove(node)
        del self.type_nodes[node.name]

    def orderTypes(self, x1, x2):
        """Introduce a relationship between the precision of two types.

        Arguments:
            x1, x2 -- Names of two types (or two type nodes), the first
                      one of which is "less abstract" than the second.
                      The type names must have been previously introduced
                      to the type hierarchy by using addType.

        Raises a ConsistencyError if the introduction of the relationship
        would make the ordering relation cyclic.  In this case the type
        hierarchy will retain its old state.
        """
        n1 = self.__normalize(x1)
        n2 = self.__normalize(x2)
        n1.more_abstract.append(n2)
        path = []
        if TypeHierarchy._TypeHierarchy__isReachable \
               (n1, n1, lambda x: x.more_abstract, set(), path):
            n1.more_abstract.remove(n2)
            raise ConsistencyError \
                  ("precision ordering between types must be acyclic: %s"
                   % " < ".join(n.name for n in path))
        n2.more_precise.append(n1)

    def unorderTypes(self, x1, x2):
        """Remove a relationship between the precision of two types.

        Arguments:
            x1, x2 -- Names of two types (or two type nodes), the first
                      one of which is "less abstract" than the second.
        """
        n1 = self.__normalize(x1)
        n2 = self.__normalize(x2)
        while n2 in n1.more_abstract:
            n1.more_abstract.remove(n2)
        while n1 in n2.more_precise:
            n2.more_precise.remove(n1)

    def __isReachable(n1, n2, succ, visited, path=None):
        """Test whether a node is reachable from another in a type hierarchy.

        Arguments:
            n1, n2  -- Two type nodes.
            succ    -- Successor function.
            visited -- A set of nodes which have been processed.
            path    -- An optional list for storing a path between the type
                       nodes.

        Returns True if the `n2' is reachable from `n1' using the ordering
        induced by `succ' and False otherwise.  A node is reachable from
        itself only if it can be reached by at least one application of
        `succ'.
        """
        if path != None:
            path.append(n1)
        visited.add(n1)
        for n in succ(n1):
            if n == n2:
                if path != None:
                    path.append(n2)
                return True
            if not n in visited \
               and TypeHierarchy._TypeHierarchy__isReachable \
               (n, n2, succ, visited, path):
                return True
        if path != None:
            path.pop()
        return False

    __isReachable = staticmethod(__isReachable)

    def lessAbstractTypes(self, x):
        """Find the types strictly "less abstract" than a given type.

        Argument:
            x -- A type name or a node in the type hierarchy.

        Returns a list of type names as a result.
        """
        result = set()
        self.__normalize(x).collectReachableNodes \
            (lambda x: x.more_precise, result)
        return [n.name for n in result]

    def moreAbstractTypes(self, x):
        """Find the types strictly "more abstract" than a given type.

        Argument:
            x -- A type name or a node in the type hierarchy.

        Returns a list of type names as a result.
        """
        result = set()
        self.__normalize(x).collectReachableNodes \
            (lambda x: x.more_abstract, result)
        return [n.name for n in result]
