# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Facilities for type inference.

This module provides operations for collecting and solving constraints
on the types of model attributes using type inference systems, and
abstract model generation.

Type inference systems are represented using TypeInferenceSystem objects.
These objects encapsulate operations and data for type inference (in
particular, a "type inference graph").  A type inference graph contains
type inference nodes (each of which is associated with a type name and a
list of constraints) connected to each other with directed edges
representing the value flow between the nodes.  A node in the type
inference graph may be of the following types (each of which is
implemented as a subclass of InferenceNode):

    TypeNode         Nodes representing constant type values which
                     may not change during type inference.  Every
                     data type in the type library used for model
                     abstraction (and every class defined in a model)
                     will have a corresponding TypeNode in the type
                     inference graph.  Using these nodes, a node in
                     the type inference graph can be assigned a type
                     from the type library by constraining the type
                     associated with the node to be equal to the type
                     associated with some TypeNode in the graph.

    ExpressionNode   ExpressionNodes are used to associate types with
                     action language expressions and their
                     subexpressions.

    AttributeNode    Intuitively, AttributeNodes represent the static
                     storage for attributes of classes and signals,
                     and association ends of classes (which can be
                     seen as class attributes of reference type).  As
                     type inference is a static analysis method, there
                     is only one AttributeNode for each attribute of
                     each class (and thus no distinction is made between
                     the attributes of the different instances of the
                     class, as is usual for static analysis).

                     Different types of attributes can be distinguished
                     by inspecting the types of the corresponding
                     AttributeNodes.  The AttributeNode subclasses
                     ObjectReferenceNode and SignalParameterNode are
                     used to represent attributes corresponding to
                     association ends and signal parameters,
                     respectively.

The value flow in the type inference graph represents the relationships
between different nodes in the graph.  For example, the value flow in an
assignment statement `x = y + z;' of the Jumbala action language is directed
from the AttributeNodes representing the variables `y' and `z' (which are
assumed to resolve to attributes of some class in the context of the
statement) via an ExpressionNode representing the addition expression to the
AttributeNode which represents the variable `x' in the statement's context.

See the documentation of TypeInferenceSystem for information on how to work
with type inference systems.

Author: Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

import sys
import copy
import actlang
import coral
import abstraction
import Utils.Models
import Utils.UML14
from Utils import ExpressionTraverser
_UML = coral.lang.UML14

class InputFormatError(abstraction.error.Error):
    """Exception for reporting errors while reading input."""
    pass

class ResolvementError(abstraction.error.Error):
    """Exception for reporting errors while resolving names."""
    pass

class TypeInferenceError(abstraction.error.Error):
    """Exception for reporting failures during type inference.

    Every object of the class has the attribute `failure', which is a
    TypeInferenceSystem.DetectInconsistency object whose type can be
    examined further to obtain the reason for the error.
    """
    def __init__(self, failure):
        self.failure = failure
        self.reason = "type constraints cannot be satisfied"

class RewriteError(abstraction.error.Error):
    """Exception for reporting failures during model rewriting."""

    def __init__(self, msg=None):
        if msg == None:
            msg = "error during model rewriting"
        abstraction.error.Error.__init__(self, msg)

class UndefinedOperationError(RewriteError):
    """Exception for reporting undefined operations during model rewriting."""

    def __init__(self, typename, arity, operation):
        """Initialize an UndefinedOperationError."""
        RewriteError.__init__ \
            (self,
             "missing or incomplete definition of operation `%s' of arity %d "
             "for type `%s'" % (operation, arity, typename))
        self.typename = typename
        self.arity = arity
        self.operation = operation

class UndefinedCoercionError(RewriteError):
    """Exception for reporting undefined coercions during model rewriting."""

    def __init__(self, from_type, to_type):
        """Initialize an UndefinedCoercionError."""
        RewriteError.__init__ \
            (self,
             "missing or incomplete definition of coercion from type `%s' to "
             "type `%s'" % (from_type, to_type))
        self.from_type = from_type
        self.to_type = to_type

# -----------------------------------------------------------------------------
# Utility functions for accessing the Coral UML model.

def _resolveIdentifier(name, context):
    """Resolve an action language identifier in a given context.

    Arguments:
        name    -- A string or a stringifiable object (the name).
        context -- Context (a Coral object corresponding to an UML
                   Classifier) from which to start searching (via
                   associations in the UML model) for the name.

    Returns a Coral object representing the attribute or association
    end to which the name resolves in the given context, or raises a
    ResolvementError if the name does not match the name of any attribute
    or (navigable) association end in the given context.
    """
    if not isinstance(name, str):
        name = str(name)
    if not context.oclIsKindOf(_UML.Classifier):
        raise TypeError \
              ("coral.lang.UML14.Classifier required as second argument")
    original_name = name
    while True:
        pos = name.find('.')
        if pos == -1:
            for f in context.feature:
                if name == str(f.name):
                    return f
            (prefix, name) = (name, "")
        else:
            (prefix, name) = (name[0:pos], name[pos+1:])
        for a_end in context.association:
            try:
                a_other_end = _getOtherAssociationEnd(a_end)
            except Error, e:
                raise ResolvementError(e.reason)
            if a_other_end.isNavigable:
                if str(a_other_end.name) == prefix:
                    if pos == -1:
                        return a_other_end
                    context = a_other_end.participant
                    break
        else:
            raise ResolvementError \
                      ("could not resolve identifier `%s'" % original_name)

def _getOtherAssociationEnd(association_end):
    """Find the other end of a 1..1 association in a UML model.

    Argument:
        association_end -- An association end as a
                           coral.lang.UML14.AssociationEnd object.

    Returns a Coral object representing the other end of the association.
    """
    a_ends = association_end.association.connection
    if len(a_ends) != 2:
        raise Error("only 1..1 associations between classes are supported")
    if a_ends[0] == association_end:
        return a_ends[1]
    else:
        return a_ends[0]

# -----------------------------------------------------------------------------
# Utility functions for action language parser.

def _parseJumbalaStatement(statement):
    """Parse an action language statement."""
    return actlang.grammar.parse(statement, generateParenExpr=False) \
               .api_getTopStatements()

# -----------------------------------------------------------------------------
# Pseudotypes for type inference.  The _UNRESTRICTED pseudotype is used as a
# pseudotype "more precise" than any other type; the _UNRESTRICTED_PRIMITIVE
# pseudotype (which is "less precise" than _UNRESTRICTED) is used to represent
# any (possibly abstract) integer type or the boolean type, and the
# _UNRESTRICTED_REFERENCE pseudotype (also "less precise" than _UNRESTRICTED,
# but incomparable with _UNRESTRICTED_PRIMITIVE) represents any reference type.
#
# The _UNRESTRICTED_INT pseudotype is used to constrain nodes whose concrete
# types are integers, but which may nevertheless remain unconstrained during
# type inference.  (In type inference, this type is treated as less precise
# than _UNRESTRICTED_PRIMITIVE, but more precise than the type 'int'.)  The
# _NULL type represents the null type, which is "less precise" than the
# _UNRESTRICTED_REFERNECE pseudotype, and incomparable with any type which is
# less or equally precise as the _UNRESTRICTED_PRIMITIVE pseudotype.
#
# The complete partial ordering of (pseudo)types is as follows (the precision
# increases from top to bottom).  <any class type> represents any class defined
# in a model used to initialize a type inference system;
# <any abstract integer type> represents any abstract type defined in the
# library of abstract types attached to the type inference system.  (Two types
# matching <any class type> are always considered to be incomparable; two types
# matching <any abstract integer type> are comparable iff they are comparable
# in the type library attached to the type inference system).
#
#                                  <any abstract integer type>
#                                            |
#   <any class type>                        int
#     |          \                           |
#     |         _NULL      boolean    _UNRESTRICTED_INT
#     \           |           |              |
#  _UNRESTRICTED_REFERENCE    _UNRESTRICTED_PRIMITIVE
#               \             /
#                _UNRESTRICTED

_UNRESTRICTED = "<SMUML_unrestricted_pseudotype>"
_UNRESTRICTED_INT = "<SMUML_unrestricted_int_pseudotype>"
_UNRESTRICTED_PRIMITIVE = "<SMUML_unrestricted_primitive_pseudotype>"
_UNRESTRICTED_REFERENCE = "<SMUML_unrestricted_reference_pseudotype>"
_NULL = "<null_type>"

# -----------------------------------------------------------------------------
# Type constraints.

class TypeConstraint:
    """Class for representing type constraints in a type inference graph.

    The class defines the constants
        TypeConstraint.CMP_LE, TypeConstraint.CMP_EQ and
        TypeConstraint.CMP_GE
    as aliases for the corresponding constants defined in the class
    abstraction.typehierarchy.TypeHierarchy.

    Furthermore, every TypeConstraint object has the following attributes:
        constrainee -- An InferenceNode whose associated type is to be
                       constrained.
        relation    -- One of the values TypeConstraint.CMP_LE,
                       TypeConstraint.CMP_EQ, TypeConstraint.CMP_GE.
        constrainer -- The constraining InferenceNode.
        is_external -- Flag for distinguishing constrains obtained from
                       a model from "external" constraints (such as those
                       obtained from a type association file).
    """

    (CMP_LE, CMP_EQ, CMP_GE) \
        = (abstraction.typehierarchy.TypeHierarchy.CMP_LE,
           abstraction.typehierarchy.TypeHierarchy.CMP_EQ,
           abstraction.typehierarchy.TypeHierarchy.CMP_GE)

    def __init__(self, constrainee, relation, constrainer, is_external=False):
        """Initialize a constraint between two nodes in a type inference graph.

        Arguments:
            constrainee -- The node whose type should be constrained.
            relation    -- One of the strings "<=", "==", or ">=", or
                           one of the values TypeConstraint.CMP_LE,
                           TypeConstraint.CMP_EQ, or TypeConstraint.CMP_GE.
            constrainer -- The node acting as the constraining node.
            is_external -- A Boolean for distinguishing between "internal"
                           and "external" constraints.

        Returns the initialized TypeConstraint object.
        """
        if not isinstance(constrainee, InferenceNode):
            raise TypeError("InferenceNode expected as first argument")
        self.constrainee = constrainee
        if isinstance(relation, int):
            if relation in (TypeConstraint.CMP_LE, TypeConstraint.CMP_EQ,
                            TypeConstraint.CMP_GE):
                self.relation = relation
            else:
                raise ValueError("invalid second argument (%d)" % relation)
        elif isinstance(relation, str):
            try:
                self.relation = {"<=" : TypeConstraint.CMP_LE,
                                 "==" : TypeConstraint.CMP_EQ,
                                 ">=" : TypeConstraint.CMP_GE }[relation]
            except ValueError:
                raise ValueError("invalid second argument (%s)" % relation)
        else:
            raise TypeError("second argument is of incorrect type")
        if not isinstance(constrainer, InferenceNode):
            raise TypeError("InferenceNode expected as third argument")
        self.constrainer = constrainer
        if not isinstance(is_external, bool):
            raise TypeError("boolean expected as fourth argument")
        self.is_external = is_external

    def test(self, type_ordering):
        """Test whether the constraint holds.

        Argument:
            type_ordering -- An abstraction.typehierarchy.TypeHierarchy
                             object defining the ordering between
                             types.

        Returns True if the constraint holds and False otherwise.
        """
        if not hasattr(type_ordering, "compare"):
            raise TypeError("argument must support the compare method")
        c = type_ordering.compare \
            (self.constrainee.inferred_type, self.constrainer.inferred_type)
        if c == TypeConstraint.CMP_EQ:
            return True
        if self.relation == TypeConstraint.CMP_LE:
            return c == abstraction.typehierarchy.TypeHierarchy.CMP_LT
        elif self.relation == TypeConstraint.CMP_GE:
            return c == abstraction.typehierarchy.TypeHierarchy.CMP_GT
        return False

    def __str__(self):
        """Return the string representation of a type constraint."""
        return "[%s] type(%s) %s type(%s)" \
               % (self.is_external and "External" or "From model",
                  str(self.constrainee),
                  { TypeConstraint.CMP_LE : "<=",
                    TypeConstraint.CMP_EQ : "==",
                    TypeConstraint.CMP_GE : ">=" }[self.relation],
                  str(self.constrainer))

# -----------------------------------------------------------------------------
# Nodes in the type inference system.

class InferenceNode:
    """Base class for nodes in a type inference graph.

    Every InferenceNode object has the following attributes:
        inferred_type    -- Name of the data type associated with the node
                            after type inference.
        type_constraints -- A list of TypeConstraints representing the
                            constraints on the type of the node (initially
                            empty).
        value_flow       -- A list of InferenceNode objects representing
                            those expressions whose value directly depends
                            on the value of the expression associated with
                            the node (initially empty).
    """
    
    def __init__(self):
        """Initialize a node in a type inference graph."""
        self.inferred_type = _UNRESTRICTED
        self.type_constraints = []
        self.value_flow = []

    def addTypeConstraint(self, relation, node, is_external=False):
        """Add a constraint on the type of the node.

        Arguments:
            relation    -- A constraint relation identifier in any of the
                           forms supported by TypeConstraint.__init__.
            node        -- A node whose associated type should constrain
                           the type of `self'.
            is_external -- Flag for distinguishing constraints obtained
                           from a model from "external" constraints.

        Returns the TypeConstraint object representing the new constraint.
        """
        c = TypeConstraint(self, relation, node, is_external)
        self.type_constraints.append(c)
        return c

    def findReachableNodes(self, filt=lambda node: True, prune=lambda node: False):
        """Find all nodes reachable from the node via value flow.

        A node is reachable from itself only if there exists a non-empty
        path from the node to itself in the value flow relation.

        Arguments:
            filt   -- An optional function for filtering the nodes to
                      be included in the result.  The function should
                      accept an InferenceNode as a parameter and return
                      True if the node should be included in the result.
                      By default, all reachable nodes are returned.
            prune  -- An optional function for pruning the search.  The
                      function should accept an InferenceNode as a
                      parameter and return True if the search should
                      not be extended to the descendants of the node.
                      By default, no pruning is applied.

        Returns the set of nodes reachable from the given node (without
        passing through nodes for which `prune' returns True) which
        satisfy the condition `filt'.
        """
        result = set()
        node_stack = [self]
        visited = set(node_stack)
        while node_stack:
            node = node_stack.pop()
            if not prune(node):
                for succ in node.value_flow:
                    if not succ in visited:
                        visited.add(succ)
                        node_stack.append(succ)
                        if filt(succ):
                            result.add(succ)
        return result

class ConstantNode:
    """Class for representing "constant" nodes in a type inference graph.

    The type of "constant" node is not allowed to change during type
    inference.
    """

    pass

class TypeNode(InferenceNode, ConstantNode):
    """Class for nodes representing types.

    TypeNodes are treated as constants during type inference.
    """

    def __init__(self, typename, is_external=False):
        """Initialize a node representing a type in the type library.

        Arguments:
            typename    -- Name of the type which the node represents.
            is_external -- Flag for distinguishing types (classes) defined
                           in a model from types defined in an abstract
                           type library.

        Returns the initialized TypeNode object.
        """
        if not isinstance(typename, str):
            raise TypeError("string expected as first argument")
        if not isinstance(is_external, bool):
            raise TypeError("boolean expected as second argument")
        InferenceNode.__init__(self)
        self.inferred_type = typename
        self.is_external = is_external

    def __str__(self):
        """Return the string representation of a type node."""
        return "<Type %s>" % self.inferred_type

class RewriteInfo:
    """A class for storing data on abstracted action language expressions.

    Each object of the class has the following attributes:
        expression -- An action language expression representing the
                      abstract version of some action language expression
                      in the concrete model.  (In practice, a RewriteInfo
                      object is an attribute of an ExpressionNode object,
                      whose `actlang_expr' attribute gives the
                      corresponding concrete expression.)
        flattened  -- A truth value telling whether `expression' is
                      flattened or not.  (The expression is said to be
                      flattened if its subexpressions cannot be matched
                      with corresponding subexpressions of the concrete
                      action language expression from which it was
                      obtained.)
        typename   -- Name of the abstract type of `expression'.
    """

    def __init__(self, expression=None, flattened=False, typename=None):
        """Initialize a RewriteInfo object.

        Arguments:
            expression -- An action language expression representing the
                          abstracted version of some concrete action
                          language expression.
            flattened  -- Whether the expression is flattened or not.
            typename   -- Name of the abstract type of the expression.
        """
        self.expression = expression
        self.flattened = flattened
        self.typename = typename

class ExpressionNode(InferenceNode):
    """Class for nodes representing action language expressions.

    In addition to the attributes common to all InferenceNodes, an
    ExpressionNode has the following additional attributes:
        model_type   -- Name of the type of the expression as determined
                        from a model.
        actlang_expr -- An action language expression associated with
                        the node.
        rewrite_info -- A RewriteInfo object (the contents of this
                        object will be initialized while rewriting the
                        model).
        coral_object -- The Coral object which "contains" the action
                        language expression represented by the node.
                        This can be either a state machine transition
                        (in which case `actlang_expr' is a subexpression
                        in the guard or effect of the transition), or a
                        model attribute (in which case `actlang_expr' is
                        a subexpression of the attribute's initializer
                        expression).
        stmt_id      -- Unused if `coral_object' does not refer to a
                        transition in a state machine.  Otherwise
                        `stmt_id' is an integer with the following
                        interpretation (see also the example below):
                          - if `stmt_id' is negative, the ExpressionNode
                            represents a (sub)expression of the guard of
                            the transition referred to by `coral_object',
                            and
                          - if `stmt_id' is nonnegative, it identifies an
                            action language statement in the effect of
                            the transition referred to by `coral_object'.
        stmt_expr_id -- Unused if `coral_object' does not refer to a
                        transition in a state machine, or if `stmt_id' is
                        negative.  Otherwise, `stmt_expr_id' identifies a
                        top-level expression in the action language
                        statement identified by `stmt_id' (in the effect
                        of the transition referred to by `coral_object');
                        see the example below.
        expr_id      -- An integer which identifies `actlang_expr'
                        uniquely among the subexpressions of the maximal
                        action language expression which contains
                        `actlang_expr' as a subexpression (see the example
                        below).

    For example, given a state machine transition with the effect
        x = 1; y = 3; assert(x > 0); send Signal(x + y, 12345) to object;
    the ExpressionNodes corresponding to the expressions in the last
    statement of the could have the values
                     stmt_id   stmt_expr_id   expr_id
            [x + y]     3           0            0
            [x]         3           0            1
            [y]         3           0            2
            [12345]     3           1            0
            [object]    3           2            0
    The numbering of statements (top-level expressions occurring in a
    statement) follows the order in which the statements (expressions)
    occur from left to right in the transition's effect (statement).
    The first statement in the effect (top-level expression in the
    statement) has the number 0.  The subexpressions of an expression
    are numbered following a depth-first ("left to right") preorder
    traversal of the abstract syntax tree of the expression.
    """

    __nodecount = 0

    def __init__(self, actlang_expr, coral_object, stmt_id=-1, stmt_expr_id=0,
                 expr_id=0):
        """Initialize a node representing an action language expression.

        Arguments:
            actlang_expr  -- An action language expression to be associated
                             with the node.
            coral_object  -- A Coral object which "contains" the action
                             language expression `actlang_expr'.  This can
                             be either a state machine transition (in which
                             case `actlang_expr' is a subexpression in the
                             guard or effect of the transition), or a model
                             attribute (in which case `actlang_expr' is a
                             subexpression of the attribute's initializer
                             expression).
            stmt_id       -- Identifier of an action language statement in
                             the effect of the transition referred to by
                             `coral_object' (if applicable).
            stmt_expr_id  -- Identifier of a top-level expression in the
                             action language statement identified by
                             `stmt_id' (if applicable).
            expr_id       -- Identifier of a subexpression in the maximal
                             action language expression which contains
                             `actlang_expr' as a subexpression.

        Returns the initialized ExpressionNode object.
        """
        if not isinstance(actlang_expr, actlang.expr.Expr):
            raise TypeError("actlang.expr.Expr expected as first argument")
        if not isinstance(coral_object, coral.kernel.ElementPtr):
            raise TypeError \
                  ("coral.kernel.ElementPtr expected as second argument")
        if not coral_object.oclIsKindOf(_UML.Transition) \
               and not coral_object.oclIsKindOf(_UML.Attribute):
            raise ValueError \
                  ("coral.lang.UML14.Transition or coral.lang.UML14.Attribute "
                   "expected as second argument")
        if not isinstance(stmt_id, int) or not isinstance(stmt_expr_id, int) \
           or not isinstance(expr_id, int):
            raise TypeError \
                  ("integer expected as third, fourth and fifth argument")
        InferenceNode.__init__(self)
        self.model_type = _UNRESTRICTED
        self.actlang_expr = actlang_expr
        self.rewrite_info = RewriteInfo()
        self.coral_object = coral_object
        self.stmt_id = stmt_id
        self.stmt_expr_id = stmt_expr_id
        self.expr_id = expr_id
        self.__id = ExpressionNode.__nodecount
        ExpressionNode.__nodecount += 1

    def __str__(self):
        """Return the string representation of an ExpressionNode."""
        if self.coral_object.oclIsKindOf(_UML.Transition):
            return '<Expr %d:%s:%s.%d:"%s">' \
                   % (self.__id,
                      self.coral_object.__XMIid__(),
                      ((self.stmt_id < 0 and "guard")
                       or ("effect.%d.%d" % (self.stmt_expr_id,
                                             self.stmt_id))),
                      self.expr_id,
                      str(self.actlang_expr))
        else:
            return '<Expr %d:%s:init.%d:"%s">' \
                   % (self.__id, self.coral_object.__XMIid__(),
                      self.expr_id,
                      str(self.actlang_expr))

class AttributeNode(InferenceNode):
    """Base class for nodes representing attributes or association ends.

    In addition to the attributes common to all InferenceNodes, an
    AttributeNode has the following additional attributes:
        coral_element          -- A Coral object representing the class
                                  or signal attribute or association end
                                  in a UML model.
        model_type             -- Name of the type of the attribute or
                                  the association end as determined from
                                  the model.
        is_abstractable        -- A Boolean flag for determining whether
                                  it is possible to assign the attribute
                                  a more abstract type during type
                                  inference.
        initializer_expression -- An action language expression
                                  representing the initial value
                                  expression for the attribute
                                  (initialized to the empty string if the
                                  attribute does not have an initializer).
    """
    
    def __init__(self, element):
        """Initialize a node for an attribute or an association end.

        Arguments:
            element -- A Coral object corresponding to the attribute or
                       the association end.

        Raises an InputFormatError if the initializer expression for the
        attribute is neither empty nor a valid expression in the Jumbala
        action language.
        """
        InferenceNode.__init__(self)
        if not isinstance(element, coral.kernel.ElementPtr):
            raise TypeError \
                  ("coral.kernel.ElementPtr expected as first argument")
        self.coral_element = element
        self.is_abstractable = False
        self.initializer_expression = ""
        if element.oclIsKindOf(_UML.Attribute):
            typename = str(element.type.name)
            if typename in ("int", "boolean"):
                self.model_type = typename
            else:
                self.model_type = Utils.Models.get_qualified_name(element.type)
            # Abstraction is supported only for integer attributes
            if self.model_type == "int":
                self.is_abstractable = True
            if element.initialValue and str(element.initialValue.body).strip():
                # Associate the attribute with its initializer expression in
                # the action language.
                code = "assert(" + str(element.initialValue.body) + ");"
                try:
                    stmts = _parseJumbalaStatement(code)
                    if len(stmts) != 1:
                        raise actlang.errors.Error()
                except actlang.errors.Error, e:
                    raise InputFormatError \
                          ("initial value of attribute `%s' is not a valid "
                           "action language expression" % i)
                self.initializer_expression = stmts[0].api_getExpr()
            elif self.model_type == "int":
                self.initializer_expression = \
                    actlang.expr.IntLiteralExpr(0, (0, 0))
        elif element.oclIsKindOf(_UML.AssociationEnd):
            self.model_type = \
                Utils.Models.get_qualified_name(element.participant)
        else:
            raise ValueError \
                  ("argument does not represent a "
                   "coral.lang.UML14.Attribute or a "
                   "coral.lang.UML14.AssociationEnd")

    def __str__(self):
        """Return the string representation of the AttributeNode."""
        return '<Attr "%s">' \
               % Utils.Models.get_qualified_name(self.coral_element)

class ObjectReferenceNode(AttributeNode):
    """Class for representing attributes modeled as association ends."""

    def __init__(self, element):
        """Initialize an ObjectReferenceNode.

        Argument:
            element -- A Coral object representing an association end.

        Returns the initialized ObjectReferenceNode.
        """
        AttributeNode.__init__(self, element)
        if not element.oclIsKindOf(_UML.AssociationEnd):
            raise ValueError \
                  ("argument does not represent a "
                   "coral.lang.UML14.AssociationEnd")

    def __str__(self):
        """Return the string representation of the ObjectReferenceNode."""
        return '<Ref "%s">' % str(self.coral_element.name)

class SignalParameterNode(AttributeNode):
    """Class for representing attributes of signals.

    Arguments:
        element  -- A Coral object representing a signal attribute.

    In addition to the attributes common to all AttributeNodes,
    SignalParameterNodes have the following additional attributes:
        signal   -- The signal to which the signal parameter is associated
                    as a Coral object.
        position -- Position of the signal in the list of parameters of
                    the SignalParameterNode's associated signal.  The first
                    parameter has the position 0.

    Returns the initialized SignalParameterNode.
    """
    def __init__(self, element):
        AttributeNode.__init__(self, element)
        self.signal = element.__parent__()
        if not element.oclIsKindOf(_UML.Attribute) \
               or not self.signal.oclIsKindOf(_UML.Signal):
            raise ValueError("argument does not represent a signal attribute")
        for (i, f) in enumerate(self.signal.feature.select \
                                (lambda x: x.oclIsKindOf(_UML.Attribute))):
            if f is element:
                self.position = i
                return
        else:
            raise ValueError \
                  ("unexpected: could not find position of signal parameter")

    def __str__(self):
        """Return the string representation of the SignalParameterNode."""
        return '<Parm %s.%d:"%s">' \
               % (str(self.signal.name), self.position,
                  str(self.coral_element.name))

# -----------------------------------------------------------------------------
# Type inference system.

class TypeInferenceSystem:
    """Class for representing type inference systems."""

    # -------------------------------------------------------------------------
    # Initialization

    def __init__(self, model, typelibrary=None):
        """Initialize a type inference system from a UML model.

        Arguments:
            model       -- A Coral object representing a UML model.
            typelibrary -- An optional abstraction.typelibrary.TypeLibrary
                           object to be associated with the type inference
                           system.  The type library object should be
                           finalizable.

        Returns a newly initialized TypeInferenceSystem.  Raises an
        InputFormatError if the model or the type library does not satisfy
        certain consistency requirements.
        """
        if not isinstance(model, coral.kernel.ElementPtr) \
               or not model.oclIsKindOf(_UML.Model):
            raise TypeError \
                  ("coral.lang.UML14.Model expected as first argument")
        if typelibrary != None \
               and not isinstance(typelibrary,
                                  abstraction.typelibrary.TypeLibrary):
            raise TypeError \
                  ("abstraction.typelibrary.TypeLibrary expected as second "
                   "argument")

        # Model.
        self.__model = model
        # State machines in the model.
        self.__statemachines = \
            Utils.Models.get_elements_of_type(model, _UML.StateMachine)
        # Names of classes defined in the model.
        self.__model_class_names = []
        # Mapping from instantiated objects in the model to lists of triples
        # (UML AttributeLink, AttributeNode, initial value) consisting of
        # an AttributeLink referring to an integer type model attribute, an
        # AttributeNode corresponding to the attribute in the type inference
        # system, and the integer value referred to by the AttributeLink
        # object.
        self.__abstractable_object_attrlinks = {}
        # Mapping from UML class or signal attributes, and association ends
        # to their corresponding AttributeNode objects.
        self.__attribute_nodes = {}
        # Mapping from signal names to lists of AttributeNode objects
        # corresponding to their parameters.
        self.__sigparm_nodes = {}
        # Mapping from type names to their corresponding TypeNode objects.
        self.__type_nodes = {}
        # Type library.
        self.__typelibrary = copy.deepcopy(typelibrary) \
                             or abstraction.typelibrary.TypeLibrary()
        # Nodes in the type inference system graph.
        self.__nodes = set()
        # Mapping from action language expressions (obtained by parsing
        # action language code in the UML model) to their corresponding
        # ExpressionNode objects.
        self.__expr_node = {}
        # Mapping from guards and effects of transitions in the UML model
        # to lists of action language statements representing the
        # corresponding code.
        self.__uml_to_statement = {}
        # Mapping from guards and effects of transitions in the UML model to
        # lists of action language statements representing the code obtained
        # by model rewriting.
        self.__uml_to_rewritten_statement = {}

        # check that the name of the model is either empty or a valid Jumbala
        # identifier to ensure the validity of fully qualified names of classes
        # and identifiers
        if str(model.name) and not Utils.UML14.check_name(model, False):
            raise InputFormatError \
                      ("name of model with UUID %s (`%s') is not a valid "
                       "identifier in the Jumbala action language"
                       % (model.__XMIid__(), str(model.name)))
        # Initialize `self.__type_nodes' with type nodes corresponding to the
        # types defined in the type library.
        for t in self.__typelibrary.typenames:
            # all non-primitive types defined in the type library are
            # "external"; these are all abstract integer types
            n = TypeNode(t, not t in self.__typelibrary.STD_TYPES)
            self.__type_nodes[t] = n
            self.__nodes.add(n)
        classes = Utils.Models.get_elements_of_type(model, _UML.Class)
        # types for classes defined in the model
        for c in classes:
            if not Utils.UML14.check_name(c, False):
                raise InputFormatError \
                          ("name of class with UUID %s (`%s') is not a valid "
                           "identifier in the Jumbala action language"
                           % (c.__XMIid__(), str(c.name)))
            classname = Utils.Models.get_qualified_name(c)
            if self.__type_nodes.has_key(classname):
                raise InputFormatError \
                          ("multiply defined class or data type `%s'"
                           % classname)
            self.__model_class_names.append(classname)
            n = TypeNode(classname) # class names defined in the model are
                                    # "non-external" types
            self.__type_nodes[classname] = n
            self.__nodes.add(n)
            # Register the class name in the partial ordering between types
            self.__typelibrary.type_ordering.addType(classname)
        # Add type nodes for the pseudotypes used for type inference
        for t in (_UNRESTRICTED, _UNRESTRICTED_INT, _UNRESTRICTED_PRIMITIVE,
                  _UNRESTRICTED_REFERENCE, _NULL):
            if t in self.__type_nodes:
                raise InputFormatError \
                          ("unexpected: cannot augment type library with "
                           "pseudotypes: type `%s' has already been defined"
                           % t)
            n = TypeNode(t)
            self.__type_nodes[t] = n
            self.__nodes.add(n)
            self.__typelibrary.type_ordering.addType(t)
        self.__orderPseudoTypesInTypeLibrary()

        # For every class in the model and for every attribute and (navigable)
        # association end defined in the class, associate a new AttributeNode
        # object with the attribute/association end in the dictionary
        # `self.__attribute_nodes'.
        for c in classes:
            identifiers = set()
            for f in c.feature:
                if f.oclIsKindOf(_UML.Attribute):
                    name = str(f.name)
                    if not Utils.UML14.check_name(f, False):
                        raise InputFormatError \
                                  ("name of attribute with UUID %s (`%s') is "
                                   "not a valid identifier in the Jumbala "
                                   "action language" % (f.__XMIid__(), name))
                    if name in identifiers:
                        raise InputFormatError \
                                  ("multiply defined class attribute `%s'"
                                   % Utils.Models.get_qualified_name(f))
                    identifiers.add(name)
                    a = _resolveIdentifier(f.name, c)
                    self.__registerAttribute(AttributeNode(a))

            for a_end in c.association:
                try:
                    a_other_end = _getOtherAssociationEnd(a_end)
                except Error, e:
                    raise InputFormatError(e.reason)
                if not a_other_end.isNavigable:
                    continue
                name = str(a_other_end.name)
                if not Utils.UML14.check_name(a_other_end, False):
                    raise InputFormatError \
                              ("name of association end with UUID %s (`%s') "
                               "is not a valid identifier in the Jumbala "
                               "action language"
                               % (a_other_end.__XMIid__(), name))
                if name in identifiers:
                    raise InputFormatError \
                              ("multiply defined attribute or association "
                               "end `%s::%s'"
                               % (Utils.Models.get_qualified_name(c), name))
                identifiers.add(name)
                self.__registerAttribute(ObjectReferenceNode(a_other_end))

        # Do the same for all signals defined in the model.
        signals = Utils.Models.get_elements_of_type(model, _UML.Signal)
        for s in signals:
            signame = str(s.name)
            if not Utils.UML14.check_signal_name(s, False):
                raise InputFormatError \
                          ("name of signal with UUID %s (`%s') is not a valid "
                           "identifier in the Jumbala action language"
                           % (s.__XMIid__(), signame))
            if self.__sigparm_nodes.has_key(signame):
                raise InputFormatError \
                          ("multiply defined signal `%s'" % signame)
            parms = []
            self.__sigparm_nodes[signame] = parms
            for f in s.feature.select(lambda x: x.oclIsKindOf(_UML.Attribute)):
                n = SignalParameterNode(f)
                self.__registerAttribute(n)
                parms.append(n)

        # Add type constraints from attribute initializer expressions.
        for (a, n) in self.__attribute_nodes.iteritems():
            if n.initializer_expression:
                info = self.__ConstraintCollectionInfo(self, a.owner, a)
                info.stmt_id = -1
                self.__addExpressionConstraints(n.initializer_expression, info)

        # Add type constraints from the model.
        for m in self.__statemachines:
            for s in m.top.subvertex:
                try:
                    if s.oclIsKindOf(_UML.Pseudostate):
                        if not str(s.kind) in ('initial', 'choice'):
                            raise InputFormatError \
                                      ("state is an unsupported kind of "
                                       "pseudostate (`%s')" % str(s.kind))
                    elif not (s.oclIsKindOf(_UML.SimpleState)
                              or s.oclIsKindOf(_UML.FinalState)):
                        raise InputFormatError \
                                  ("state is not a UML SimpleState")
                    state_has_completion_transitions = False
                    state_has_triggered_transitions = False
                    for t in s.outgoing:
                        try:
                            if hasattr(t, 'trigger') and t.trigger:
                                state_has_triggered_transitions = True
                                if not t.trigger.oclIsKindOf(_UML.SignalEvent):
                                    raise InputFormatError \
                                              ("triggering event of transition"
                                               " is not a UML SignalEvent")
                                signame = str(t.trigger.signal.name)
                                try:
                                    nodes = self.__sigparm_nodes[signame]
                                except KeyError:
                                    raise InputFormatError \
                                          ("trigger of transition refers to an"
                                           " unknown signal (`%s')" % signame)
                                if len(nodes) != len(t.trigger.parameter):
                                    raise InputFormatError \
                                              ("number of parameters for "
                                               "triggering SignalEvent (%d) "
                                               "does not match the number of "
                                               "attributes defined for signal "
                                               "`%s' (%d)"
                                               % (len(t.trigger.parameter),
                                                  signame,
                                                  len(nodes)))
                                # Add type constraints for the variables to
                                # which the signal parameters get assigned on
                                # the reception of a signal.
                                # (8 Mar 2007) Require that the types of the
                                # signal parameters be equal to guarantee that
                                # the simulator will handle reception of
                                # signals with abstract parameters correctly.
                                for (i, p) in enumerate(t.trigger.parameter):
                                    if hasattr(p, "name") and p.name:
                                        n = self.__attribute_nodes \
                                                [_resolveIdentifier \
                                                     (str(p.name), m.context)]
                                        n.addTypeConstraint("==", nodes[i])
                                        nodes[i].value_flow.append(n)
                            else:
                                state_has_completion_transitions = True
                            if hasattr(t, 'guard') and t.guard:
                                self.__addGuardOrEffectConstraints \
                                    (t, str(t.guard.expression.body), t.guard,
                                     m.context, True)
                            if hasattr(t, 'effect') and t.effect:
                                if not t.effect.oclIsKindOf \
                                       (_UML.UninterpretedAction):
                                    raise InputFormatError \
                                              ("effect of transition is not an"
                                               " UML UninterpretedAction")
                                self.__addGuardOrEffectConstraints \
                                    (t, str(t.effect.script.body), t.effect,
                                     m.context)
                        except abstraction.error.Error, e:
                            raise InputFormatError \
                                      (e.reason
                                       + "\n    while processing transition "
                                         "`%s' (UUID %s) to state `%s' (UUID "
                                         "%s)"
                                       % (str(t.name), t.__XMIid__(),
                                          str(t.target.name),
                                          t.target.__XMIid__()))
                    # Do not accept models with mixed completion and
                    # signal-triggered transitions (not strictly necessary for
                    # model abstraction, but may be required for analyzing
                    # error traces)
                    if state_has_completion_transitions \
                           and state_has_triggered_transitions:
                        raise InputFormatError \
                                  ("state has both outgoing completion and "
                                   "signal-triggered transitions: run the "
                                   "state machine hierarchy flattener on the "
                                   "model first")
                except abstraction.error.Error, e:
                    raise InputFormatError \
                              (e.reason
                               + "\n    while processing state `%s' (UUID %s) "
                                 "of state machine `%s' (UUID %s)\n"
                               % (str(s.name), s.__XMIid__(), str(m.name),
                                  m.__XMIid__()))

        # Collect the integer DataValues in slots of instantiated objects in
        # the model.
        for o in Utils.Models.get_elements_of_type(self.__model, _UML.Object):
            try:
                datavalues = []
                self.__abstractable_object_attrlinks[o] = datavalues
                for link in o.slot:
                    if not link.oclIsKindOf(_UML.AttributeLink):
                        raise InputFormatError \
                                  ("element with UUID %s is not a UML "
                                   "AttributeLink" % link.__XMIid__())
                    try:
                        try:
                            node = self.__attribute_nodes[link.attribute]
                        except KeyError:
                            raise InputFormatError \
                                      ("reference to an unknown attribute")
                        value = link.value
                        if not value.oclIsKindOf(_UML.DataValue):
                            raise InputFormatError \
                                      ("value of link is not a UML DataValue")
                        try:
                            owner = value.owner
                            # the owner of every DataValue must be a UML
                            # Object, Class, Model, or Package
                            if owner.oclIsKindOf(_UML.Object):
                                owned_elements = owner.ownedInstance
                            elif owner.oclIsKindOf(_UML.Class) \
                                     or owner.oclIsKindOf(_UML.Model) \
                                     or owner.oclIsKindOf(_UML.Package):
                                owned_elements = owner.ownedElement
                            else:
                                raise InputFormatError \
                                          ("owner of DataValue is not a UML "
                                           "Object, Class, Model or Package")
                            if not value in owned_elements:
                                raise InputFormatError \
                                          ("owner of DataValue is "
                                           "inconsistent (no reference back "
                                           "to the DataValue from its "
                                           "supposed owner)")
                            if node.is_abstractable:
                                # all abstractable attributes are assumed to be
                                # of type int
                                classifier = value.classifier
                                if len(classifier) != 1:
                                    raise InputFormatError \
                                              ("ambiguous classifier for "
                                               "DataValue")
                                datatype = classifier[0]
                                if datatype != link.attribute.type:
                                    raise InputFormatError \
                                              ("classifier of DataValue does "
                                               "not match the type of the "
                                               "AttributeLink's attribute")
                                if not datatype.oclIsKindOf(_UML.Primitive) \
                                       or str(datatype.name) != "int":
                                    raise InputFormatError \
                                              ("classifier of DataValue does "
                                               "not refer to a UML Primitive "
                                               "named `int'")
                                try:
                                    val = long(str(value.name))
                                    if val < -2147483648 or val > 2147483647:
                                        raise ValueError()
                                except ValueError:
                                    raise InputFormatError \
                                              ("name for DataValue (`%s') is "
                                               "not a valid integer constant "
                                               "in the Jumbala action language"
                                               % str(value.name))
                                datavalues.append((link, node, val))
                        except InputFormatError, e:
                            raise InputFormatError \
                                      ("%s\n    in DataValue with UUID %s"
                                       % (str(e), value.__XMIid__()))
                    except InputFormatError, e:
                        raise InputFormatError \
                                  ("%s\n    in AttributeLink with UUID %s"
                                   % (str(e), link.__XMIid__()))
            except InputFormatError, e:
                raise InputFormatError \
                          ("unexpected: %s\n    in Object with UUID %s"
                           % (str(e), o.__XMIid__()))

        # Initialize the model type of all inference graph nodes corresponding
        # to action language expressions by running the type inference
        # procedure once.
        try:
            self.solveConstraints()
        except TypeInferenceError:
            raise InputFormatError("model is not type-consistent")
        for n in self.__expr_node.itervalues():
            n.model_type = n.inferred_type

    def __orderPseudoTypesInTypeLibrary(self):
        """Order pseudotypes used for type inference.

        This method constructs the partial ordering between pseudotypes,
        primitive types, abstract integer types defined in a type library,
        and reference types (classes) defined in the model.
        """
        # Basic ordering of pseudotypes and primitive types
        for (t1, t2) in ((_UNRESTRICTED, _UNRESTRICTED_PRIMITIVE),
                         (_UNRESTRICTED, _UNRESTRICTED_REFERENCE),
                         (_UNRESTRICTED_REFERENCE, _NULL),
                         (_UNRESTRICTED_PRIMITIVE, "boolean"),
                         (_UNRESTRICTED_PRIMITIVE, _UNRESTRICTED_INT),
                         (_UNRESTRICTED_INT, "int")):
            self.__typelibrary.type_ordering.orderTypes(t1, t2)
        # Ordering for reference types defined in the model
        for t in self.__model_class_names:
            self.__typelibrary.type_ordering.orderTypes \
                (_UNRESTRICTED_REFERENCE, t)
            self.__typelibrary.type_ordering.orderTypes(_NULL, t)

    def __registerAttribute(self, node):
        """Add an AttributeNode to the type inference system.

        Argument:
            node -- An AttributeNode object.
        """
        if not node.is_abstractable:
            node.addTypeConstraint("==", self.__type_nodes[node.model_type])
        else:
            node.addTypeConstraint(">=", self.__type_nodes[_UNRESTRICTED_INT])
        self.__attribute_nodes[node.coral_element] = node
        self.__nodes.add(node)

    class __ConstraintCollectionInfo:
        """Information for collecting expression constraints."""

        def __init__(self, tis, context, transition):
            self.nodes = tis._TypeInferenceSystem__nodes
            self.type_nodes = tis._TypeInferenceSystem__type_nodes
            self.attribute_nodes = tis._TypeInferenceSystem__attribute_nodes
            self.expr_node = tis._TypeInferenceSystem__expr_node
            self.context = context
            self.transition = transition
            self.stmt_id = 0
            self.stmt_expr_id = 0
            self.expr_id = 0
            self.lvalue_exprs = set()
            self.side_effect_exprs = set()
            self.node = None
            self.node_stack = []

    def __addGuardOrEffectConstraints(self, transition, code, model_element,
                                      context, is_guard=False):
        """Add type constraints from guards and effects of transitions.

        Arguments:
            transition    -- A transition in a UML state machine as a Coral
                             object.
            code          -- A piece of Jumbala code as a string (the guard
                             or effect of a transition in a UML state
                             machine).
            model_element -- The Coral object with which to associate a list
                             of statements (in the dictionary
                             `self.__uml_to_statement') generated by the
                             action language parser from the code.
            context       -- Context to be used for resolving names of
                             identifiers in the code.
            is_guard      -- Tells whether `code' represents a guard instead
                             of an effect of a transition.
        """
        if is_guard: # transform a guard expression for Jumbala parser
            code = "assert(" + code + ");"
        try:
            stmts = _parseJumbalaStatement(code)
        except actlang.errors.Error, e:
            raise InputFormatError(str(e))
        self.__uml_to_statement[model_element] = stmts
        info = self.__ConstraintCollectionInfo(self, context, transition)
        if is_guard:
            info.stmt_id = -1
            info.stmt_expr_id = 0
            self.__addExpressionConstraints(stmts[0].api_getExpr(), info)
            return
        for (info.stmt_id, stmt) in enumerate(stmts):
            info.lvalue_exprs = set()
            info.side_effect_exprs = set()
            info.stmt_expr_id = 0
            try:
                if isinstance(stmt, actlang.statement.AssertStatement):
                    self.__addExpressionConstraints(stmt.api_getExpr(), info)
                elif isinstance(stmt, actlang.statement.EmptyStatement):
                    pass
                elif isinstance(stmt, actlang.statement.ExprStatement):
                    e = stmt.api_getExpr()
                    info.side_effect_exprs.add(e)
                    self.__addExpressionConstraints(e, info)
                elif isinstance(stmt, actlang.statement.SendStatement):
                    signame = stmt.api_getSignalName()
                    try:
                        nodes = self.__sigparm_nodes[signame]
                    except KeyError:
                        raise InputFormatError \
                              ("reference to an unknown signal encountered "
                               "(`%s')" % signame)
                    args = stmt.api_getArgList()
                    if len(nodes) != len(args):
                        raise InputFormatError \
                              ("invalid number of parameters for signal "
                               "`%s' (expected %d, got %d)"
                               % (signame, len(nodes), len(args)))
                    for (info.stmt_expr_id, e) in enumerate(args):
                        self.__addExpressionConstraints(e, info)
                        n = info.expr_node[e]
                        n.addTypeConstraint("<=", nodes[info.stmt_expr_id])
                        n.value_flow.append(nodes[info.stmt_expr_id])
                    info.stmt_expr_id = len(args)
                    self.__addExpressionConstraints(stmt.api_getTarget(), info)
                else:
                    raise InputFormatError \
                        ("unsupported type of action language statement "
                         "encountered (`%s')" % stmt.__class__.__name__)
            except abstraction.error.Error, e:
                raise InputFormatError \
                    (e.reason
                     + "\n    while processing statement `%s'" % str(stmt))

    def __generateConstraintCollector():
        """Generate a function object for collecting type constraints."""

        # Callbacks used on entry and exit to/from each subexpression.

        def recurse(expr, info):
            info.node = ExpressionNode \
                            (expr, info.transition, info.stmt_id,
                             info.stmt_expr_id, info.expr_id)
            info.expr_node[expr] = info.node
            info.nodes.add(info.node)
            info.expr_id += 1
            info.node_stack.append(info.node)

        def backtrack(expr, retval, info):
            raise ExpressionTraverser.Return(info.node_stack.pop())

        # Pre-callbacks.  These callbacks are used for "atomic" expressions,
        # i.e., those expressions that do not have any subexpressions.

        def boolLiteral(expr, info):
            info.node.addTypeConstraint("==", info.type_nodes['boolean'])
            raise ExpressionTraverser.Return()

        def intLiteral(expr, info):
            # Restrict types of integer literals using the _UNRESTRICTED_INT
            # pseudotype instead of 'int' to make default type assignment
            # possible during type finalization.
            info.node.addTypeConstraint \
                ("==", info.type_nodes[_UNRESTRICTED_INT])
            raise ExpressionTraverser.Return()

        def null(expr, info):
            info.node.addTypeConstraint("==", info.type_nodes[_NULL])
            raise ExpressionTraverser.Return()

        def identifier(expr, info):
            if isinstance(expr, actlang.names.Name):
                a = info.attribute_nodes[_resolveIdentifier \
                                             (expr, info.context)]
            # else `expr' is a PrimaryFieldAccessExpr; only primary field
            # accesses via `this' are supported
            elif not isinstance(expr.api_getReferenceExpr(),
                                actlang.expr.ThisExpr):
                raise InputFormatError \
                      ("unsupported form of field access (`%s')" % str(expr))
            else:
                a = info.attribute_nodes \
                    [_resolveIdentifier(expr.api_getFieldName(), info.context)]
            info.node.addTypeConstraint("==", a)
            if expr in info.lvalue_exprs:
                info.node.value_flow.append(a)
            else:
                a.value_flow.append(info.node)
            raise ExpressionTraverser.Return()

        def this(expr, info):
            info.node.addTypeConstraint \
                ("==", info.type_nodes[Utils.Models
                                       .get_qualified_name(info.context)])
            raise ExpressionTraverser.Return()

        def instanceCreation(expr, info):
            if not expr in info.side_effect_exprs:
                raise InputFormatError \
                      ("expression `%s' with a side effect encountered in a "
                       "context required to be free of side effects"
                       % str(expr))
            # Since the UML-to-Jumbala mapping does not define a clear
            # action language convention for representing names which might
            # be in different UML namespaces (as of 14 August, 2007), the
            # following code simply assumes classes to be referrable by
            # their names.
            cls = str(expr.api_getTypeName())
            try:
                clsnode = info.type_nodes[cls]
            except KeyError:
                raise InputFormatError \
                      ("reference to an unknown class in `new' expression "
                       "encountered (`%s')" % cls)
            if expr.api_getArgList():
                raise InputFormatError \
                      ("parameters in `new' expressions are not supported")
            info.node.addTypeConstraint("==", clsnode)
            raise ExpressionTraverser.Return()

        # Pre-callback for assignments.  The left-hand side subexpression of
        # the assignment has to be registered as an lvalue expression before
        # processing it; similarly, if the right-hand side subexpression is
        # a `new' expression, the side effect in the expression needs to be
        # allowed.

        def simpleAssignmentPre(expr, info):
            if not expr in info.side_effect_exprs:
                raise InputFormatError \
                      ("expression `%s' with a side effect encountered in a "
                       "context required to be free of side effects"
                       % str(expr))
            info.lvalue_exprs.add(expr.api_getLhsExpr())
            r = expr.api_getRhsExpr()
            if isinstance(r, actlang.expr.InstanceCreationExpr):
                info.side_effect_exprs.add(r)

        # Post-callbacks.  These callbacks will generally be used for all
        # compound expression types.

        def simpleAssignmentPost(expr, subnodes, info):
            node = info.node_stack[-1]
            (l, r) = subnodes
            r.addTypeConstraint("<=", l)
            r.value_flow.append(l)
            l.addTypeConstraint("==", node)
            l.value_flow.append(node)

        def unaryOperation(expr, subnodes, info):
            node = info.node_stack[-1]
            o = subnodes[0]
            node.addTypeConstraint("==", o)
            if isinstance(expr, actlang.expr.LogicalNotExpr):
                node.addTypeConstraint("==", info.type_nodes['boolean'])
            elif not isinstance(expr, actlang.expr.ParenExpr):
                # the only possible unary operations at this point are +, - and
                # ~; the operand of these operations must be an integer or an
                # abstracted integer
                o.addTypeConstraint(">=", info.type_nodes[_UNRESTRICTED_INT])
            o.value_flow.append(node)

        def condAndOr(expr, subnodes, info):
            node = info.node_stack[-1]
            (l, r) = subnodes
            node.addTypeConstraint("==", l)
            node.addTypeConstraint("==", r)
            node.addTypeConstraint("==", info.type_nodes['boolean'])
            l.value_flow.append(node)
            r.value_flow.append(node)

        def infix(expr, subnodes, info):
            node = info.node_stack[-1]
            (l, r) = subnodes
            o = expr.api_getOperator()
            if isinstance(o, (actlang.binaryop.NumericalOperator,
                              actlang.binaryop.Plus)):
                n = node
                # the operands of an arithmetic operation must be integers or
                # abstracted integers
                for s in subnodes:
                    s.addTypeConstraint \
                        (">=", info.type_nodes[_UNRESTRICTED_INT])
            elif isinstance(o, actlang.binaryop.NumericalBooleanOperator):
                n = node
                # the operands of a Boolean operation must be of primitive type
                # (int or boolean)
                for s in subnodes:
                    s.addTypeConstraint \
                        (">=", info.type_nodes[_UNRESTRICTED_PRIMITIVE])
            elif isinstance(o, (actlang.binaryop.EqualityOperator,
                                actlang.binaryop.NumericalComparison)):
                # Create an extra node encoding restrictions for the types
                # of the arguments of a test operation (the type of whose
                # result is always restricted to boolean).  This node will
                # be accessible by the attribute `argument_constraint'
                # added to the node representing the actual test
                # expression.
                n = ExpressionNode \
                        (expr, info.transition, info.stmt_id,
                         info.stmt_expr_id, node.expr_id)
                info.nodes.add(n)
                node.argument_constraint = n
                node.addTypeConstraint("==", info.type_nodes['boolean'])
                if isinstance(o, actlang.binaryop.NumericalComparison):
                    # the operands of a comparison operation must be integers
                    # or abstracted integers
                    for s in subnodes:
                        s.addTypeConstraint \
                            (">=", info.type_nodes[_UNRESTRICTED_INT])
            else:
                raise InputFormatError \
                      ("unsupported binary operator encountered (`%s')"
                       % str(o))
            n.addTypeConstraint(">=", l)
            n.addTypeConstraint(">=", r)
            l.value_flow.append(n)
            r.value_flow.append(n)

        def ternaryCond(expr, subnodes, info):
            node = info.node_stack[-1]
            (c, t, f) = subnodes
            c.addTypeConstraint("==", info.type_nodes['boolean'])
            node.addTypeConstraint(">=", t)
            node.addTypeConstraint(">=", f)
            t.value_flow.append(node)
            f.value_flow.append(node)

        e = actlang.expr
        n = actlang.names
        pre_callbacks = \
            (("entry", recurse),
             (e.BoolLiteralExpr, boolLiteral),
             (e.IntLiteralExpr, intLiteral),
             (e.NullExpr, null),
             ((n.SimpleName, n.QualifiedName, e.PrimaryFieldAccessExpr),
              identifier),
             (e.ThisExpr, this),
             (e.InstanceCreationExpr, instanceCreation),
             (e.SimpleAssignmentExpr, simpleAssignmentPre),
             ((e.ParenExpr, e.UnaryPlusExpr, e.UnaryMinusExpr, e.BitNotExpr,
               e.LogicalNotExpr, e.CondAndExpr, e.CondOrExpr, e.InfixExpr,
               e.TernaryCondExpr),
              ExpressionTraverser.ignore))
        post_callbacks = \
            (("exit", backtrack),
             (e.SimpleAssignmentExpr, simpleAssignmentPost),
             ((e.ParenExpr, e.UnaryPlusExpr, e.UnaryMinusExpr, e.BitNotExpr,
               e.LogicalNotExpr),
              unaryOperation),
             ((e.CondAndExpr, e.CondOrExpr), condAndOr),
             (e.InfixExpr, infix),
             (e.TernaryCondExpr, ternaryCond))
        collector = ExpressionTraverser.ExpressionTraverser()
        for (cls, f) in pre_callbacks:
            collector.register_pre_callback(cls, f)
        for (cls, f) in post_callbacks:
            collector.register_post_callback(cls, f)
        return collector

    __collectConstraints = __generateConstraintCollector()
    del __generateConstraintCollector

    def __addExpressionConstraints(self, expr, info):
        """Collect type constraints from an action language expression.

        Arguments:
            expr -- An action language expression.
            info -- A __ConstraintCollectionInfo object.
        """
        info.expr_id = 0
        info.node_stack = []
        try:
            try:
                self.__collectConstraints(expr, info)
            except ExpressionTraverser.UnregisteredCallbackError, e:
                raise InputFormatError \
                      ("subexpression `%s' is of an unsupported type (`%s')"
                       % (str(e.expr), e.expr.__class__.__name__))
        except abstraction.error.Error, e:
            raise InputFormatError \
                  (e.reason + "\n    while processing expression `%s'"
                   % str(expr))

    # -------------------------------------------------------------------------
    # Public API for adding and removing constraints

    def addExternalConstraint(self, constrainee, relation, constrainer):
        """Constrain the type of an abstractable attribute.

        Arguments:
            constrainee -- Identifies a type inference graph node whose
                           associated type should be constrained.  This
                           argument should be either
                             * a type inference graph node representing
                               an abstractable attribute, or
                             * a Coral object representing an
                               abstractable attribute of a class or a
                               signal in a UML model.
            relation    -- A constraint relation identifier in any of the
                           forms accepted by TypeConstraint.__init__.
            constrainer -- Identifies a type inference graph node acting
                           as the constraining node.  Similarly to the
                           first argument, this argument can be a node
                           in the type inference graph or a Coral object
                           representing an attribute (or an association
                           end), with the exception that the attribute
                           (associated with the node) does not necessarily
                           have to be abstractable in this case.
                           Additionally, this argument can also be the
                           name of a type in the type library.

        Returns the new constraint object.
        """
        constrainee = self.findNode(constrainee)
        constrainer = self.findNode(constrainer)
        if not isinstance(constrainee, AttributeNode) \
               or not constrainee.is_abstractable:
            raise ValueError \
                  ("first argument does not represent an abstractable "
                   "attribute in the model")
        return constrainee.addTypeConstraint(relation, constrainer, True)

    def removeExternalConstraint(self, constrainee, relation="all",
                                 constrainer=None):
        """Remove a constraint on a type of an attribute.

        Arguments:
            constrainee -- A node from which to remove a constraint, or
                           a TypeConstraint object corresponding to the
                           constraint to remove.
            relation    -- A constraint relation identifier in any of the
                           forms supported by TypeConstraint.__init__, or
                           the string "all", which acts as a wildcard
                           to represent any relation.
            constrainer -- A node acting as a constrainer (the default
                           value None acts as a wildcard to represent any
                           node).
        """
        if isinstance(constrainee, TypeConstraint):
            c = constrainee
            if not c.is_external:
                raise ValueError("argument is not an external constraint")
            if not c.constrainee in self.__nodes \
                   or not c.constrainer in self.__nodes:
                raise ValueError \
                      ("argument does not represent a valid constraint for "
                       "the type inference graph")
            if not isinstance(c.constrainee, AttributeNode) \
                   or not c.constrainee.is_abstractable:
                raise ValueError \
                      ("constrainee of constraint is not an abstractable "
                       "attribute in the model")
            c.constrainee.type_constraints.remove(c)
        else:
            constrainee = self.findNode(constrainee)
            if not isinstance(constrainee, AttributeNode) \
                   or not constrainee.is_abstractable:
                raise ValueError \
                      ("first argument does not identify an abstractable "
                       "attribute in the model")
            if not relation in ("all", TypeConstraint.CMP_LE,
                                TypeConstraint.CMP_EQ, TypeConstraint.CMP_GE):
                if isinstance(relation, str):
                    try:
                        relation = { "<=" : TypeConstraint.CMP_LE,
                                     "==" : TypeConstraint.CMP_EQ,
                                     ">=" : TypeConstraint.CMP_GE }[relation]
                    except KeyError:
                        raise ValueError("invalid second argument")
                else:
                    raise TypeError("second argument is of incorrect type")
            if constrainer:
                constrainer = self.findNode(constrainer)
            updated_constraints = []
            for c in constrainee.type_constraints:
                if not (c.is_external \
                        and (relation == "all" or c.relation == relation) \
                        and (not constrainer or constrainer is c.constrainer)):
                    updated_constraints.append(c)
            constrainee.type_constraints = updated_constraints

    def findNode(self, arg):
        """Return the type inference graph node identified by argument.

        Arguments:
            arg -- This argument should identify a node in the type
                   inference graph.  It can be one of the following:
                     * a node in the type inference graph (in which
                       case the function simply returns the node itself)
                     * a Coral object representing an attribute or an
                       association end in a UML model (in which case
                       the function returns the AttributeNode
                       corresponding to the attribute/association end),
                       or
                     * a name of a type in the type library (in which
                       case the function returns the TypeNode representing
                       the type).

        Returns the type inference graph node identified by the argument.
        """
        if isinstance(arg, InferenceNode):
            if not arg in self.__nodes:
                raise ValueError \
                      ("expected a node in the type inference graph as "
                       "argument")
            return arg
        elif isinstance(arg, coral.kernel.ElementPtr):
            # Try to find an AttributeNode corresponding to a Coral object
            try:
                return self.__attribute_nodes[arg]
            except KeyError:
                raise ValueError \
                      ("expected an attribute or an association end as "
                       "argument")
        elif isinstance(arg, str):
            # Try to find a TypeNode corresponding to a type name
            try:
                return self.__type_nodes[arg]
            except KeyError:
                raise ValueError \
                      ("type name argument (`%s') is not a name of any type "
                       "in the type library" % arg)
        else:
            raise TypeError \
                  ("expected a type inference graph node, a Coral object "
                   "representing an attribute or an association end, or a "
                   "type name as argument")

    def findAbstractableNodes(self):
        """Get a list of abstractable nodes in the type inference graph."""
        return [n for n in self.__attribute_nodes.itervalues()
                if n.is_abstractable]

    def findReachableNodes(self, node, filt=lambda node: True, prune=lambda node: False):
        """Find all nodes reachable from a given node via value flow.

        A node is reachable from itself only if there exists a non-empty
        path from the node to itself in the value flow relation.

        Arguments:
            node  -- Node from which to start the search.
            filt  -- An optional function for filtering the nodes to
                     be included in the result.  The function should
                     accept an InferenceNode as a parameter and return
                     True if the node should be included in the result.
                     By default, all reachable nodes are returned.
            prune -- An optional function for pruning the search.  The
                     function should accept an InferenceNode as a
                     parameter and return True if the search should
                     not be extended to the descendants of the node.
                     By default, no pruning is applied.

        Returns the set of nodes reachable from the given node (without
        passing through nodes for which `prune' returns True) which
        satisfy the condition `filt'.
        """
        return node.findReachableNodes(filt, prune)

    def attachTypeLibrary(self, typelibrary, retain_common_types=False):
        """(Re)attach a type library to the type inference system.

        Arguments:
            typelibrary         -- An abstraction.typelibrary.TypeLibrary
                                   object (called the "new" type library
                                   below).
            retain_common_types -- A flag for choosing whether to retain
                                   constraints referring to non-primitive
                                   types which are common to the type
                                   library originally associated with
                                   the type inference system and the "new"
                                   type library.

        This function has the following effects:
          * The type inference system will be associated with the given
            "new" type library.  More specifically:
                - Type nodes representing primitive types of the action
                  language or classes defined in the model (from which
                  the type inference system was initialized) will always
                  be preserved in the type inference graph.
                - The type inference graph will be updated to include
                  TypeNodes for each type (currently unknown to the type
                  inference system) defined in the "new" type library.
                - TypeNodes corresponding to types not defined in the
                  "new" type library will be removed.
          * All type constraints referring to types not defined in the
            "new" type library will be removed from the type inference
            system.  Note that these types include _all_ non-primitive
            types defined in the "old" type library unless
            `retain_common_types' is True, in which case constraints
            referring to types with identical names in the "old" and "new"
            type libraries will be preserved.

        Raises an InputFormatError if there are conflicts between names of
        types defined in the model and the names of types defined in the
        "new" type library, or if the "new" type library does not satisfy
        certain consistency requirements in order to be finalizable.  In
        these cases the type inference system will remain associated to the
        "old" type library.

        Changes to the type library object given as a parameter will not
        be automatically reflected in the type inference system.  To inform
        the type inference system about any changes to the type library,
        the type library has to be reattached to the type inference system.
        """
        if not isinstance(typelibrary, abstraction.typelibrary.TypeLibrary):
            raise TypeError \
                  ("expected abstraction.typelibrary.TypeLibrary as first "
                   "argument")
        if not isinstance(retain_common_types, bool):
            raise TypeError("expected boolean as second argument")
        type_nodes = {}
        retained_typenodes = set()
        # Preserve all TypeNodes corresponding to types defined in the model
        # together with the TypeNodes corresponding to the primitive action
        # language types
        for (t, n) in self.__type_nodes.iteritems():
            if not n.is_external:
                type_nodes[t] = n
                retained_typenodes.add(n)
        # Ensure that there are no name conflicts between types defined in the
        # model and types defined in the new type library.  Create a new
        # TypeNode for each new type defined in the new type library.
        for t in typelibrary.typenames:
            if not t in typelibrary.STD_TYPES:
                if t in type_nodes:
                    raise InputFormatError \
                          ("name of abstract type `%s' coincides with a class "
                           "or data type of the same name defined in the "
                           "model" % t)
                if not retain_common_types or not t in self.__type_nodes:
                    type_nodes[t] = TypeNode(t, True)
                else:
                    n = self.__type_nodes[t]
                    type_nodes[t] = n
                    retained_typenodes.add(n)
        # Remove all constraints referring to TypeNodes whose corresponding
        # types will get removed from the "old" type library
        constraints_to_remove = []
        for n in self.__nodes:
            for c in n.type_constraints:
                if ((isinstance(c.constrainee, TypeNode) 
                     and not c.constrainee in retained_typenodes)
                    or (isinstance(c.constrainer, TypeNode)
                        and not c.constrainer in retained_typenodes)):
                    constraints_to_remove.append(c)
        for c in constraints_to_remove:
            self.removeExternalConstraint(c)
        # Remove all of these TypeNodes
        for n in self.__type_nodes.itervalues():
            if not n in retained_typenodes:
                self.__nodes.remove(n)
        # Ensure that all new type nodes are included in the graph
        self.__nodes.update(type_nodes.values())
        # Update remaining attributes of `self'
        self.__type_nodes = type_nodes
        self.__typelibrary = copy.deepcopy(typelibrary)
        # Ensure that all type names (including types from the "old" type
        # library, class names from the model, and types from the "new" type
        # library) are known in the type hierarchy associated with the type
        # inference system, in order to make type inference work correctly.
        for t in self.__type_nodes.iterkeys():
            if not t in self.__typelibrary.type_ordering:
                self.__typelibrary.type_ordering.addType(t)
        # Ensure that the pseudotypes used for type inference are ordered
        # correctly.
        self.__orderPseudoTypesInTypeLibrary()

    def initExternalConstraintsFromFile(self, f="-"):
        """Initialize type constraints from a type association file.

        Argument:
            f -- A file name as a string.  The special file name "-"
                 refers to standard input.

        This function has the following effects:
          * The type inference system will be associated with a new type
            library built from the type specification files imported in
            the input.  This will include all side effects listed in the
            documentation of `attachTypeLibrary'.
          * All external constraints on types of nodes in the type
            inference system will be removed and replaced with those
            defined in the input.
        
        Returns the list of type library file names imported in the input.
        Raises an InputFormatError if an error occurs while processing.  In
        this case the type inference system will remain unchanged.
        """
        if f == "-":
            f = "(standard input)"
            i = sys.stdin
        else:
            try:
                i = open(f, "r")
            except IOError:
                raise InputFormatError("could not open `%s' for reading" % f)
        type_association = abstraction.typeassociation.TypeAssociation()
        try:
            type_association.read(i)
        except abstraction.typeassociation.InputFormatError, e:
            raise InputFormatError("%s:%s" % (f, str(e)))
        if i != sys.stdin:
            i.close()
        try:
            self.initConstraintsFromTypeAssociation(type_association)
        except InputFormatError, e:
            raise InputFormatError("%s:%s" % (f, str(e)))
        return type_association.libfiles

    def initConstraintsFromTypeAssociation(self, type_association):
        """Initialize type constraints from a TypeAssociation object.

        Argument:
            type_association -- A typeassociation.TypeAssociation object.

        This function has the following effects:
          * The type inference system will be associated with a new type
            library built from the type specification files listed in
            `type_association.libfiles'.  This will include all side
            effects listed in the documentation of `attachTypeLibrary'.
          * All external constraints on types of nodes in the type
            inference system will be removed and replaced with those
            defined in the input.

        Raises an InputFormatError if an error occurs while processing.
        In this case the type inference system will remain unchanged.
        """
        lib = abstraction.typelibrary.TypeLibrary()
        try:
            lib.read(type_association.libfiles)
        except abstraction.typelibrary.InputFormatError, e:
            raise InputFormatError(str(e))
        abstractable_attributes = {}
        for n in self.__attribute_nodes.itervalues():
            if n.is_abstractable:
                abstractable_attributes[n.coral_element.__XMIid__()] = n
        constraints = set()
        for (x, cs) in type_association.constraints.iteritems():
            try:
                n = abstractable_attributes[x]
            except KeyError:
                if cs and cs[0][2] != -1:
                    lineno_str = "%d:" % cs[0][2]
                else:
                    lineno_str = ""
                raise InputFormatError \
                      ("%s`%s' does not match the XMI id of any "
                       "abstractable class attribute or signal parameter"
                       % (lineno_str, x))
            for (op, t, l) in cs:
                if not t in lib.typenames:
                    if l != -1:
                        lineno_str = "%d:" % l
                    else:
                        lineno_str = ""
                    raise InputFormatError \
                          ("%sreference to an unknown type `%s' encountered"
                           % (lineno_str, t))
                constraints.add((n, op, t))
        self.attachTypeLibrary(lib)
        for n in self.__nodes:
            n.type_constraints = [c for c in n.type_constraints
                                  if not c.is_external]
        for (n, op, t) in constraints:
            self.addExternalConstraint(n, op, t)

    # -------------------------------------------------------------------------
    # Type inference
        
    def solveConstraints(self, default_type=None, f=None, ordering=None):
        """Perform type inference.

        Arguments:
            default_type -- Name of a default type to be assigned for
                            unconstrained attributes.
            f            -- An optional function which can be called with
                            one parameter.  During the type inference
                            process, this function will be called with
                            information about the steps made in the type
                            inference process.  Each of the steps is
                            represented by an object of (a subclass) of
                            the class InferenceStep.  The general
                            "protocol" of the type inference process
                            (i.e., the sequences of InferenceSteps with
                            which the function f can be called during type
                            inference) is as follows:
                                BeginTypeInference
                                MakeMoreAbstract*
                                (DetectInconsistency
                                      | EndTypeInference
                                        BeginTypeFinalization
                                        (InformAboutDefaultType
                                              | InformAboutConstraints)*
                                        EndTypeFinalization)
            ordering     -- An optional object which represents the
                            precision ordering between types.  If not
                            given, defaults to the ordering attached to
                            the type library associated with the type
                            inference system.  A custom `ordering' object
                            must implement the methods
                              compare(<type name>, <type name>)
                                which compares the precision of two
                                types, returning
                                  * abstraction.typehierarchy.
                                               .TypeHierarchy.CMP_EQ
                                    if the two types are equally precise
                                  * abstraction.typehierarchy
                                               .TypeHierarchy.CMP_LT
                                    if the first type is strictly "more
                                    precise" than the second
                                  * abstraction.typehierarchy
                                               .TypeHierarchy.CMP_GT
                                    if the first type is strictly "less
                                    precise" than the second
                                  * abstraction.typehierarchy.
                                               .TypeHierarchy.CMP_IC
                                    otherwise
                              join(<type name>, <type name>, ...)
                                which returns the name of a type which
                                is at least as abstract than any of the
                                types given as arguments, or None if no
                                such type exists.
                              meet(<type name>, <type name>, ...)
                                which returns the name of a type which
                                is at least as precise than any of the
                                types given as arguments, or None if no
                                such type exists.
                            These methods must also be able to be
                            applicable to the pseudotypes used for type
                            inference (see the comments near their
                            definition in this file), the primitive types
                            "int" and "boolean" of the Jumbala action
                            language, and all reference types (classes)
                            known by the TypeInferenceSystem.  An
                            example of a class which supports this
                            interface can be found in the
                            refinement.partition.py module.

        If the type constraint system has a solution, the results of
        type inference can be inspected after calling this function via
        the `inferred_type' attributes of the nodes in the type constraint
        graph.  A TypeInferenceError will be raised if the constraint
        system has no solution.  The `failure' attribute of the
        TypeInferenceError is a DetectInconsistency object whose type can
        be examined further to find the reason for the error.

        It is safe to call this function multiple times (for example,
        to try to search for a satisfying solution to the constraint
        system by adding or removing type constraints between nodes in
        the graph in between the function calls).
        """
        if default_type and not default_type in self.__typelibrary.typenames:
            raise ValueError("undefined default type `%s'" % default_type)
        if f != None:
            if not hasattr(f, "__call__"):
                raise ValueError("a function expected as `f' argument")
            f(TypeInferenceSystem.BeginTypeInference())
        if ordering == None:
            ordering = self.__typelibrary.type_ordering

        # Construct a dictionary of dependencies between nodes in the type
        # inference graph.  Intuitively, `deps[n]' will give the list of nodes,
        # the type constraints associated with which need to be rechecked
        # whenever the type associated with n is made more abstract during the
        # type inference process.
        deps = {}
        constraints_to_check = set()
        for n in self.__nodes:
            if not isinstance(n, ConstantNode):
                n.inferred_type = _UNRESTRICTED
            if not deps.has_key(n):
                deps[n] = []
            deps[n].extend(n.type_constraints)
            for c in n.type_constraints:
                if c.relation in (TypeConstraint.CMP_EQ,
                                  TypeConstraint.CMP_LE):
                    deps[n].extend(c.constrainer.type_constraints)
                if c.relation in (TypeConstraint.CMP_EQ,
                                  TypeConstraint.CMP_GE):
                    if not deps.has_key(c.constrainer):
                        deps[c.constrainer] = []
                    deps[c.constrainer].extend(n.type_constraints)
                constraints_to_check.add(c)
        # Type inference using a worklist of constraints to check
        while constraints_to_check:
            c = constraints_to_check.pop()
            if not c.test(ordering):
                lub = \
                    ordering.join \
                        (c.constrainee.inferred_type,
                         c.constrainer.inferred_type)
                if not lub:
                    failure = TypeInferenceSystem.DetectIncoercibility(c)
                    if f:
                        f(failure)
                    raise TypeInferenceError(failure)
                for n in (c.constrainee, c.constrainer):
                    if n.inferred_type != lub:
                        if isinstance(n, ConstantNode):
                            failure = TypeInferenceSystem \
                                      .DetectAttemptToChangeConstant(c, n, lub)
                            if f:
                                f(failure)
                            raise TypeInferenceError(failure)
                        if f:
                            f(TypeInferenceSystem.MakeMoreAbstract
                              (n, n.inferred_type, lub, c))
                        n.inferred_type = lub
                        constraints_to_check.update(deps[n])
        if f:
            f(TypeInferenceSystem.EndTypeInference())
            f(TypeInferenceSystem.BeginTypeFinalization())

        # Assign types for nodes whose types have not been constrained.
        inferred_type = {}
        typemap = {}
        for n in self.__nodes:
            if isinstance(n, ConstantNode):
                continue
            if n.inferred_type == _UNRESTRICTED:
                e = TypeInferenceError("")
                e.reason = "unexpected: some nodes left unconstrained after " \
                           "type inference"
                raise e
            if n.inferred_type == _UNRESTRICTED_INT:
                # Determine correct type for nodes which have integer type
                # in the concrete model, but whose type has not yet been
                # constrained.
                types = self.__findTypesOfNearestRestrictedContexts(n, typemap)
                for c in n.type_constraints:
                    if (c.relation == TypeConstraint.CMP_LE
                        and isinstance(c.constrainer, TypeNode)):
                        types.add(c.constrainer.inferred_type)
                if not types and default_type:
                    # value of node n does not flow into any
                    # type-restricted context, and a default type was
                    # specified
                    inferred_type[n] = default_type
                    if isinstance(n, AttributeNode) and f:
                        f(TypeInferenceSystem.InformAboutDefaultType(n))
                else:
                    # value of node n flows into zero or more
                    # type-restricted contexts
                    inferred_type[n] = 'int' # see comment above
                    if isinstance(n, AttributeNode) and f:
                        f(TypeInferenceSystem.InformAboutConstraints
                          (n, types, ordering))
        for (n, t) in inferred_type.iteritems():
            n.inferred_type = t
        if f:
            f(TypeInferenceSystem.EndTypeFinalization())

    def __findTypesOfNearestRestrictedContexts(self, node, typemap):
        """Find the types of contexts reachable from a node via value flow.

        Arguments:
            node    -- A node in the type inference graph.
            typemap -- Dictionary for caching the results.

        Returns a set of types representing the types of those contexts
        that are reachable from `node' solely via type-unrestricted nodes.
        The pseudotypes _UNRESTRICTED and _UNRESTRICTED_INT will never be
        included in the result.
        """
        if node in typemap:
            return typemap[node]
        result = set()
        typemap[node] = result
        if not node.inferred_type in (_UNRESTRICTED, _UNRESTRICTED_INT):
            result.add(node.inferred_type)
            return result
        types = set()
        for n in node.value_flow:
            try:
                types.update(typemap[n])
            except KeyError:
                types.update \
                    (self.__findTypesOfNearestRestrictedContexts(n, typemap))
        result.update(types)
        return result

    class InferenceStep:
        """Base class for steps in the type inference process."""
        pass

    class BeginTypeInference(InferenceStep):
        """Class representing the start of type inference."""

        def __str__(self):
            """Return the string representation of the object."""
            return "begin type inference"

    class DetectInconsistency(InferenceStep):
        """Base class for representing the detection of an inconsistency.

        Every object of this class has the attribute `constraint'
        representing a TypeConstraint object which cannot be satisfied in
        the type inference system.
        """

        def __init__(self, constraint):
            """Initialize a DetectInconsistency object."""
            self.constraint = constraint

    class DetectIncoercibility(DetectInconsistency):
        """Class for detecting failures due to type incoercibility."""
        
        def __init__(self, constraint):
            """Initialize a DetectIncoercibility object."""
            TypeInferenceSystem.DetectInconsistency.__init__(self, constraint)

        def __str__(self):
            """Return the string representation of the object."""
            return "type constraint %s failed: cannot coerce types assigned " \
                   "to the nodes %s and %s (currently %s and %s, " \
                   "respectively) to a common more abstract type" \
                   % (str(self.constraint),
                      str(self.constraint.constrainee),
                      str(self.constraint.constrainer),
                      self.constraint.constrainee.inferred_type,
                      self.constraint.constrainer.inferred_type)

    class DetectAttemptToChangeConstant(DetectInconsistency):
        """Class for detecting failures due to changes to constant nodes.

        In addition to the attributes common to all DetectInconsistency
        objects, every object of this class has the following additional
        attributes:
            node        -- A node in the type inference graph whose
                           associated type was attempted to be changed.
            target_type -- Name of the type which was tried to be assigned
                           to the node.
        """
        
        def __init__(self, constraint, node, target_type):
            """Initialize a DetectAttemptToChangeConstant object."""
            TypeInferenceSystem.DetectInconsistency.__init__(self, constraint)
            self.node = node
            self.target_type = target_type

        def __str__(self):
            """Return the string representation of the object."""
            return "type constraint %s failed: cannot assign node %s type " \
                   "%s since the type of this node (%s) must remain constant" \
                   % (str(self.constraint), str(self.node), self.target_type,
                      self.node.inferred_type)

    class MakeMoreAbstract(InferenceStep):
        """Class representing steps to change types associated with nodes.

        Every object of this class has the following attributes:
            node        -- A node in the type inference graph.
            source_type -- Name of the type associated with the node before
                           the change.
            target_type -- Name of the type that the node was associated
                           with in the inference step.
            constraint  -- A constraint representing the justification for
                           the change operation.
        """

        def __init__(self, node, source_type, target_type, constraint):
            """Initialize a MakeMoreAbstract object."""
            self.node = node
            self.source_type = source_type
            self.target_type = target_type
            self.constraint = constraint

        def __str__(self):
            """Return the string representation of the object."""
            srctype = (self.source_type == _UNRESTRICTED) and "UNRESTRICTED" \
                      or self.source_type
            return "changing the type assigned to node `%s' from %s to %s " \
                   "(due to constraint %s)" \
                   % (str(self.node), srctype, self.target_type,
                      self.constraint)

    class EndTypeInference(InferenceStep):
        """Class representing the completion of type inference."""

        def __str__(self):
            """Return the string representation of the object."""
            return "end of type inference"

    class BeginTypeFinalization(InferenceStep):
        """Class representing the start of type finalization."""

        def __str__(self):
            """Return the string representation of the object."""
            return "begin type finalization"

    class InformAboutDefaultType(InferenceStep):
        """Class for reporting attributes assigned the default type.

        Every object of the class has the attribute `node', which
        represents a node in the type inference graph which was assigned
        the default type at the end of the type inference process.
        """

        def __init__(self, node):
            """Initialize an InformAboutDefaultType object."""
            self.node = node

        def __str__(self):
            """Return the string representation of the object."""
            return "assigning default type to node `%s'" % self.node

    class InformAboutConstraints(InferenceStep):
        """Class for reporting transitively type-constrained attributes.

        Every object of the class has the following attributes:
            node  -- A node in the type inference graph whose associated
                     type is directly affected by the type inference
                     process, but whose value may flow into type-constrained
                     contexts.
            types -- A set of type names representing the types of contexts
                     that the value of the node flows into.  (Intuitively,
                     `node' could then be assigned any type which is at
                     least as precise as any of the types of the contexts in
                     which the value of `node' is used, without danger of
                     having to make the types associated with the contexts
                     more abstract.)
            bound -- The greatest lower bound of the types in `types' (or
                     None if such a lower bound does not exist). 
        """

        def __init__(self, node, types, type_ordering):
            """Initialize an InformAboutConstraints object."""
            self.node = node
            self.types = list(types)
            self.bound = type_ordering.meet(*self.types)

        def __str__(self):
            """Return the string representation of the object."""
            if not self.types:
                attribute_nodes = [str(n)
                                   for n in self.node.findReachableNodes()
                                   if isinstance(n, AttributeNode)
                                   and n != self.node]
                if not attribute_nodes:
                    return "type of node %s is unconstrained" % self.node
                else:
                    return "type of node %s is unconstrained, but the value " \
                           "of the node flows into attribute(s) {%s}" \
                           % (self.node, ", ".join(attribute_nodes))
            else:
                return "type of node %s is unconstrained, but the value " \
                       "of the node flows into type-constrained context(s) " \
                       "of type(s) {%s}%s" \
                       % (self.node, ", ".join(self.types),
                          (not self.bound and " (no common lower bound)") or
                          (len(self.types) > 1 and " (common lower bound: %s)"
                           % self.bound) or "")

    class EndTypeFinalization(InferenceStep):
        """Class representing the end of type finalization."""

        def __str__(self):
            """Return the string representation of the object."""
            return "end type finalization"

    # -------------------------------------------------------------------------
    # Model rewriting

    class __ExpressionRewriterInfo:
        """State information for rewriting action language expressions."""

        def __init__(self, typelibrary, expr_node, flattening_threshold,
                     undefined_ops):
            self.typelibrary = typelibrary
            self.expr_node = expr_node
            self.abstract_expr_to_expr_node = {}
            self.undefined_ops = undefined_ops
            self.context = ""
            self.choose_id = 0

            # reference to an UML Class
            self.uml_context = None
            # left hand sides of assignments
            self.lhs_exprs = set()
            # mapping from expressions to their input variables
            self.var_refs = {}
            # threshold for flattening
            self.flattening_threshold = flattening_threshold

        def getOperationsBySymbol(self, op):
            """Return a dictionary of operations based on an operator symbol.

            Argument:
                op -- An operator symbol.
            """
            if op in ("==", "!=", "<", "<=", ">", ">="):
                return self.typelibrary.tests
            return self.typelibrary.operators

        def updateAbstractCorrespondence(self, old_expr, new_expr, tname=None):
            """(Re)associate an expression with a RewriteInfo object.

            Arguments:
                old_expr -- An action language expression.
                new_expr -- An action language expression which should
                            replace `old_expr' (which may occur when
                            rewriting the model).
                tname    -- An optional name of the type of `new_expr'.
            """
            try:
                node = self.abstract_expr_to_expr_node[old_expr]
            except KeyError:
                # if `old_expr' is not present in
                # `self.abstract_expr_to_expr_node', assume that `old_expr'
                # is an expression from the concrete model
                node = self.expr_node[old_expr]
                tname = node.inferred_type
                node.rewrite_info.flattened = False
            self.abstract_expr_to_expr_node[new_expr] = node
            node.rewrite_info.expression = new_expr
            if tname != None:
                node.rewrite_info.typename = tname

        def operationExpression(self, dtype, rtype, operator, *arguments):
            """Construct an act. lang. expression for an abstract operation.

            Arguments:
                dtype      -- Name of the domain type of the operation.
                rtype      -- Name of the target type of the operation.
                operator   -- Operator symbol.
                arguments  -- Action language expressions for the arguments
                              of the operator.

            Returns an action language expression representing the abstract
            operation as a native method call of the form
               $$SMUML.operation(<domain type of operation>,
                                 <range type of operation>,
                                 <operator symbol>,
                                 <argument expressions>)
            """
            numdomain = self.typelibrary.numdomain[dtype]
            if not numdomain:
                raise RewriteError("domain of type `%s' is empty" % dtype)
            arity = len(arguments)
            cartesize = abstraction.cartesize.cartesize

            # check whether the operation is well-defined
            operations = self.getOperationsBySymbol(operator)
            try:
                op = operations[dtype][arity][operator]
                for args in cartesize(*(arity * [numdomain.keys()])):
                    if not op[tuple(args)]:
                        raise KeyError()
            except KeyError:
                if self.undefined_ops != None:
                    self.undefined_ops.add((dtype, operator, arity))
                    e = \
                      actlang.expr.NameMethodInvocationExpr \
                          (actlang.names.SimpleName("$$SMUML.error", (0, 0)),
                           [],
                           (0, 0))
                    return e
                raise UndefinedOperationError(dtype, arity, operator)

            args = \
                 [actlang.expr.StringLiteralExpr(dtype, (0, 0)),
                  actlang.expr.StringLiteralExpr(rtype, (0, 0)),
                  actlang.expr.StringLiteralExpr(operator, (0, 0))]
            args.extend(arguments)
            result = \
                actlang.expr.NameMethodInvocationExpr \
                    (actlang.names.SimpleName \
                         ("$$SMUML.operation", (0, 0)),
                     args,
                     (0, 0))
            return result

        def coercedExpression(self, expression, from_type, to_type):
            """Construct an action language expression for type coercion.

            Arguments:
                expression -- Expression to which to apply the coercion.
                from_type  -- Name of the source type of the coercion.
                to_type    -- Name of the target type of the coercion.

            Returns `expression' if `from_type' and `to_type' represent
            the same type, or `to_type' is not an abstract integer type.
            Otherwise,
              - if the coercion from `from_type' and `to_type' is not
                well-defined:
                  * if `self.undefined_ops' == None, raises an
                    UndefinedCoercion exception
                  * otherwise, adds the undefined coercion to
                    `self.undefined_ops' and returns an action language
                    expression representing a native method call of the
                    form $$SMUML.error()
              - if the coercion is well-defined, returns an action
                language expression representing the coercion operation
                as a native method call of the form
                  $$SMUML.coerce(<operand expression>,
                                 <name of source type>,
                                 <name of target type>)
                In this case, updates also the RewriteInfo object
                currently referred to by
                `self.abstract_expr_to_expr_node[expression].rewrite_info'
                with the returned expression (and the target type of the
                coercion), and associates the returned expression with
                this RewriteInfo object in
                `self.abstract_expr_to_expr_node[expression].rewrite_info'.
            """
            if from_type == to_type \
                   or not to_type in self.typelibrary.typenames \
                   or to_type in self.typelibrary.STD_TYPES:
                return expression

            for t in (from_type, to_type):
                if not t in self.typelibrary.STD_TYPES \
                       and not self.typelibrary.numdomain[t]:
                    raise RewriteError("domain of type `%s' is empty" % t)

            # check that the coercion is well-defined
            try:
                dfn = self.typelibrary.coercions[to_type][from_type]["Jumbala"]
                if not dfn:
                    raise KeyError()
                for (test, retvals) in dfn:
                    if not retvals:
                        raise KeyError()
            except KeyError:
                if self.undefined_ops != None:
                    self.undefined_ops.add \
                        (("coercion", from_type, to_type, "Jumbala"))
                    e = \
                      actlang.expr.NameMethodInvocationExpr \
                          (actlang.names.SimpleName("$$SMUML.error", (0, 0)),
                           [],
                           (0, 0))
                    return e
                raise UndefinedCoercionError(from_type, to_type)

            result = \
                actlang.expr.NameMethodInvocationExpr \
                    (actlang.names.SimpleName("$$SMUML.coerce", (0, 0)),
                     (expression,
                      actlang.expr.StringLiteralExpr(from_type, (0, 0)),
                      actlang.expr.StringLiteralExpr(to_type, (0, 0))),
                     (0, 0))
            self.updateAbstractCorrespondence(expression, result, to_type)
            return result

        def inlineOperationDefinition(self, domain_type, range_type, symbol, *arguments):
            """Construct an expression for an inlined abstract operation.

            Arguments:
                domain_type -- Name of the operand type of the operation.
                range_type  -- Name of the result type of the operation.
                symbol      -- Operation symbol (one of the operators or
                               tests in the Jumbala action language).
                arguments   -- A list of actlang.expr.Expr objects (the
                               positional arguments for the operator).

            Returns a pair (<expr>, <used arguments>), where <expr> is an
            actlang.expr.Expr object representing the definition of the
            inlined operation, and <used arguments> is a set of indices of
            those arguments in `arguments' which actually occur as
            subexpressions of the returned expression <expr>.
            """
            arity = len(arguments)
            op = self.getOperationsBySymbol(symbol)[domain_type][arity][symbol]
            cartesize = abstraction.cartesize.cartesize
            numdomain = sorted(self.typelibrary.numdomain[domain_type].keys())
            eq = actlang.binaryop.Equal((0, 0))
            case_expr = None
            args_referred_to_in_inlined_expr = set()
            for args in cartesize(*(arity * [numdomain])):
                if not case_expr:
                    case_expr = \
                        self.chooseExpression(op[tuple(args)], range_type)
                    continue
                test_case = None # will be initialized as the expression
                                 # /\{0<=i<arity} arguments[i] == val[i]
                args_referred_to_in_test_case = []
                for (i, (arg, val)) in enumerate(zip(arguments, args)):
                    # construct a term in test case: "arguments[i] == args[i]"
                    term = \
                        actlang.expr.InfixExpr \
                            (arg,
                             actlang.expr.IntLiteralExpr(val, (0, 0)),
                             eq)
                    # try to evaluate the term
                    result = self.evalConcreteExpr(term, "boolean")
                    if result == True:
                        # term vacuously true, no need to add the term to
                        # the test case
                        pass
                    elif result == False:
                        # term vacuously false => skip the entire test case
                        break
                    else:
                        args_referred_to_in_test_case.append(i)
                        test_case = \
                            (test_case and actlang.expr.CondAndExpr \
                                               (test_case, term)
                             or term)
                else:
                    ret_expr = \
                        self.chooseExpression(op[tuple(args)], range_type)
                    if test_case == None:
                        # test case vacuously true
                        case_expr = ret_expr
                        args_referred_to_in_inlined_expr = set()
                    else:
                        case_expr = \
                            actlang.expr.TernaryCondExpr \
                                (test_case, ret_expr, case_expr)
                        args_referred_to_in_inlined_expr.update \
                            (args_referred_to_in_test_case)
            if case_expr == None:
                case_expr = \
                    actlang.expr.NameMethodInvocationExpr \
                        (actlang.names.SimpleName("$$SMUML.error", (0, 0)))
            return (case_expr, args_referred_to_in_inlined_expr)

        def inlineCoercionDefinition(self, operand, from_type, to_type):
            """Construct an expression for an inlined coercion.

            Arguments:
                operand   -- An actlang.expr.Expr object (the expression
                             to be coerced).
                from_type -- Name of the source type for the coercion.
                to_type   -- Name of the target type for the coercion.

            Returns a pair (<expr>, <operand used?>), where <expr> is an
            actlang.expr.Expr object representing the definition of the
            inlined coercion, and <operand used?> is a truth value telling
            whether `operand' actually occurs as a subexpression of the
            returned expression <expr>.
            """
            dfn = self.typelibrary.coercions[to_type][from_type]["Jumbala"][:]
            case_expr = actlang.expr.IntLiteralExpr(-1, (0, 0))
            operand_referred_to_in_inlined_expr = False
            while dfn:
                (test, retvals) = dfn.pop()
                if not from_type in self.typelibrary.STD_TYPES:
                    # translate references to values in the domain of
                    # `from_type'
                    for (sym, val) in self.typelibrary.domain[from_type] \
                                          .iteritems():
                        test = test.replace("<%s>" % sym, str(val))
                # plug in the expression to be coerced
                test = test.replace("<>", "(x)")
                (test_case, substituted) = \
                    ExpressionTraverser.substitute_exprs \
                        (_parseJumbalaStatement \
                             ("assert (%s);" % test)[0].api_getExpr(),
                         {"x" : operand})
                ret_expr = self.chooseExpression(retvals, to_type)
                # try to evaluate the test case
                result = self.evalConcreteExpr(test_case, "boolean")
                if result == True:
                    # test case vacuously true
                    case_expr = ret_expr
                    operand_referred_to_in_inlined_expr = False
                elif result == False:
                    # test case vacuously false, skip it
                    pass
                else:
                    case_expr = \
                        actlang.expr.TernaryCondExpr \
                            (test_case, ret_expr, case_expr)
                    operand_referred_to_in_inlined_expr |= substituted
            return (case_expr, operand_referred_to_in_inlined_expr)

        def chooseExpression(self, vals, typename):
            """Construct an expression for choosing a value.

            Arguments:
                vals     -- An iterable of values to choose from (a sequence
                            of integers or booleans).
                typename -- Name of the type of the domain of the values.

            If there is only one value to choose from, returns an action
            language expression representing the value.

            If there are two or more values to choose from, returns an
            action language NameMethodInvocationExpr object representing a
            method call, which is of the form
            $$SMUML.choose_<type>(<values>, <type name>, <id>, <context>),
            where
              <type>      is "boolean" if `typename' is "boolean", or "int"
                          otherwise,
              <values>    is a string literal representing a ", "-separated
                          list of string representations (in Jumbala) of
                          values in the (numeric, if applicable) domain of
                          type `typename',
              <type name> is a string literal, the name of type `typename',
              <id>        is a nonnegative integer identifier given for the
                          `choose' method call,
              <context>   is a string literal representing the context of
                          the method call (see the documentation of
                          TypeInferenceSystem.rewriteModel).
            As a side effect, `self.choose_id' will be incremented in this
            case.
            """
            if typename != "boolean":
                uniq = actlang.expr.IntLiteralExpr
                method = "$$SMUML.choose_int"
                translate = str
            else:
                uniq = actlang.expr.BoolLiteralExpr
                method = "$$SMUML.choose_boolean"
                translate = lambda x: x and "true" or "false"
            if len(vals) == 1:
                return uniq(vals[0], (0, 0))
            self.choose_id += 1
            return actlang.expr.NameMethodInvocationExpr \
                       (actlang.names.SimpleName(method, (0, 0)),
                        (actlang.expr.StringLiteralExpr
                             (", ".join(map(translate, vals)), (0, 0)),
                         actlang.expr.StringLiteralExpr(typename, (0, 0)),
                         actlang.expr.IntLiteralExpr(self.choose_id, (0, 0)),
                         actlang.expr.StringLiteralExpr(self.context, (0, 0))),
                        (0, 0))

        def flattenedExpression(self, expr):
            """Construct a flattened abstract expression.

            Argument:
                expr -- An actlang.expr.Expr object.

            An expression can be flattened if (i) it is not the left hand
            side of an assignment expression, (ii) all of its variable
            references ("input variables") are references to attributes of
            an abstract type, and (iii) the product of the sizes of these
            attributes' abstract domains does not exceed the threshold in
            `self.flattening_threshold'.

            Flattening an action language expression means converting the
            expression into a case analysis which enumerates the admissible
            values of the expression for all combinations of values of the
            expression's input variables.

            This function should be called only after
            `self.abstract_expr_to_expr_node', `self.lhs_exprs', and
            `self.var_refs' have been initialized (by first applying
            TypeInferenceSystem.__substituteAbstractOperations and
            TypeInferenceSystem.__collectVarRefs to `expr').  Furthermore,
            `self.uml_context' must have been initialized to refer to a
            coral.kernel.ElementPtr object representing a UML Class which
            should be used to resolve variable references in `expr'.

            If the expression can be flattened, returns the flattened action
            language expression, otherwise returns the value None.
            """
            if expr in self.lhs_exprs:
                # `expr' is the left hand side of an assignment, do not flatten
                return None
            refs = self.var_refs[expr]
            if not refs:
                # `expr' is a constant expression, simply evaluate it
                retvals = self.evalAbstractExpr(expr, self, {})
                return self.chooseExpression \
                           (retvals.raw_contents, retvals.typename)
            # attributes actually referenced in the expression (the same
            # attribute may be referenced multiple times using different names)
            arg_attrs = []
            # numeric domains of the attributes
            arg_domains = []
            # mapping from attributes referenced in the expression to pairs
            # consisting of the name of the type of the attribute and a list of
            # expressions representing references to the attribute in `expr'
            arg_info = {}
            input_domain_size = 1
            for arg_expr in refs:
                arg_type = \
                    self.abstract_expr_to_expr_node[arg_expr].inferred_type
                if arg_type == "int" or \
                       not arg_type in self.typelibrary.typenames:
                    # `arg_expr' refers to a non-abstract attribute (whose type
                    # is not "boolean"), do not flatten `expr'
                    return None
                attr = _resolveIdentifier(str(arg_expr), self.uml_context)
                try:
                    (typename, arg_exprs) = arg_info[attr]
                except KeyError:
                    if arg_type == "boolean":
                        domain_values = [True, False]
                    else:
                        domain_values = \
                            sorted(self.typelibrary.domain[arg_type].values())
                    input_domain_size *= len(domain_values)
                    if self.flattening_threshold >= 0 \
                           and input_domain_size > self.flattening_threshold:
                        return None
                    arg_attrs.append(attr)
                    arg_domains.append(domain_values)
                    arg_info[attr] = (arg_type, [arg_expr])
                else:
                    if typename != arg_type:
                        raise RewriteError \
                                  ("unexpected: type of identifier `%s' "
                                   "conflicts with the type of another "
                                   "identifier referring to the same "
                                   "attribute in the expression"
                                   % str(arg_expr))
                    arg_exprs.append(arg_expr)

            # an environment in which to evaluate `expr'
            env = {}

            # construct the case expression
            case_expr = None

            for arg_vals in abstraction.cartesize.cartesize(*arg_domains):
                for (attr, val) in zip(arg_attrs, arg_vals):
                    for ref in arg_info[attr][1]:
                        env[ref] = val
                retvals = self.evalAbstractExpr(expr, self, env)
                test_expr = None
                for (attr, val) in zip(arg_attrs, arg_vals):
                    if isinstance(val, bool):
                        val_expr = actlang.expr.BoolLiteralExpr(val, (0, 0))
                    else:
                        val_expr = actlang.expr.IntLiteralExpr(val, (0, 0))
                    e = \
                      actlang.expr.InfixExpr \
                          (arg_info[attr][1][0], val_expr,
                           actlang.binaryop.Equal((0, 0)))
                    test_expr = \
                        (test_expr and actlang.expr.CondAndExpr(test_expr, e)
                         or e)
                choose_expr = \
                    self.chooseExpression \
                        (retvals.raw_contents, retvals.typename)
                case_expr = \
                    (case_expr
                     and actlang.expr.TernaryCondExpr \
                             (test_expr, choose_expr, case_expr)
                     or choose_expr)

            return case_expr

        def evalConcreteExpr(self, expr, typename):
            """Try to evaluate a concrete action language expression.

            Arguments:
                expr     -- An actlang.expr.Expr object.
                typename -- Type of the expression (should be "int" or
                            "boolean").

            Returns the value of the expression (as a value corresponding
            to type `typename'), or None if the action language interpreter
            failed to evaluate the expression.
            """
            code = "%s x; x = (%s);" % (typename, str(expr))
            env = actlang.basetypes.getInitialEnvironment()
            vm = actlang.basetypes.getInitialMachine()
            tree = actlang.grammar.parse(code)
            try:
                main = tree.translate(env)
                vm.initialize(main)
                vm.run()
                return vm.getAuto("x")
            except actlang.errors.Error:
                return None

        def __generateAbstractExpressionEvaluator():
            """Generate a func. obj. for evaluating an abstract expression."""

            ValueCollection = abstraction.typelibrary.ValueCollection

            # Pre-callbacks.  Each callback takes an actlang.expr.Expr object,
            # an __ExpressionRewriterInfo object, and an environment (a mapping
            # from attributes referenced in the expression to their current
            # values) as parameters and returns an
            # abstraction.TypeLibrary.ValueCollection object representing the
            # admissible values of the expression.

            def recurse(expr, info, *rest):
                node = info.abstract_expr_to_expr_node[expr]
                if isinstance(expr, actlang.expr.NameMethodInvocationExpr) \
                       and str(expr.api_getInvocationName()) \
                               == "$$SMUML.coerce":
                    t = node.rewrite_info.typename
                else:
                    t = node.inferred_type
                if t in info.typelibrary.STD_TYPES and not info.var_refs[expr]:
                    # `expr' is of a primitive type constant expression: try
                    # to evaluate it using the action language interpreter
                    result = info.evalConcreteExpr(expr, t)
                    if result != None:
                        raise ExpressionTraverser.Return \
                                  (ValueCollection \
                                       (info.typelibrary, t, result))

            def recurseAbstractOperation(expr, info, env):
                name = str(expr.api_getInvocationName())
                args = expr.api_getArgList()
                if name == "$$SMUML.operation":
                    (dtype, rtype, operator) = args[:3]
                    operands = \
                        [info.evalAbstractExpr(a, info, env) for a in args[3:]]
                    operator = operator.api_getValue()
                    if operator in ("==", "!=", "<", "<=", ">", ">="):
                        f = info.typelibrary.applyTest
                    else:
                        f = info.typelibrary.applyOperator
                    retvals = \
                        f(dtype.api_getValue(), operator, *operands)
                elif name == "$$SMUML.coerce":
                    (operand, source_type, target_type) = args
                    operand = info.evalAbstractExpr(operand, info, env)
                    retvals = \
                        info.typelibrary.applyCoercion \
                            (source_type.api_getValue(),
                             target_type.api_getValue(), operand)
                else:
                    raise RewriteError \
                              ("unexpected: cannot evaluate abstract "
                               "subexpression `%s' (unsupported method call "
                               "in expression)" % str(expr))
                raise ExpressionTraverser.Return(retvals)

            # Post-callbacks.  Each callback takes an actlang.expr.Expr object,
            # a list of subresults, an __ExpressionRewriterInfo object, and an
            # environment dictionary as parameters and returns an
            # abstraction.TypeLibrary.ValueCollection object representing the
            # admissible return values of the abstract operation.

            def nameref(expr, arg_vals, info, env):
                # return the value of an identifier from the environment
                return ValueCollection \
                           (info.typelibrary,
                            info.abstract_expr_to_expr_node[expr]
                                .inferred_type,
                            env[expr])

            def paren(expr, arg_vals, *rest):
                return arg_vals[0]

            def binaryCond(expr, arg_vals, *rest):
                (lhs, rhs) = arg_vals
                if not lhs.typename != "boolean" or rhs.typename != "boolean":
                    raise RewriteError("unexpected: incompatible input types")
                if isinstance(expr, actlang.expr.CondAndExpr):
                    test = lambda x, y: x and y
                else:
                    test = lambda x, y: x or y
                retvals = set()
                for (l, r) in abstraction.cartesize.cartesize \
                                  (lhs.raw_contents, rhs.raw_contents):
                    retvals.add(test(l, r))
                return ValueCollection(info.typelibrary, "boolean", retvals)

            def neg(expr, arg_vals, info):
                arg = arg_vals[0]
                if not arg.typename != "boolean":
                    raise RewriteError("unexpected: incompatible input types")
                retvals = set()
                for v in arg:
                    retvals.add(not v)
                return ValueCollection(info.typelibrary, "boolean", retvals)

            def ternaryCond(expr, arg_vals, info):
                (cond, true, false) = arg_vals
                if not cond.typename != "boolean" or \
                       true.typename != false.typename:
                    raise RewriteError("unexpected: incompatible input types")
                retvals = set()
                for v in cond:
                    if v:
                        retvals.update(true.raw_contents)
                    else:
                        retvals.update(false.raw_contents)
                return ValueCollection \
                           (info.typelibrary, true.typename, retvals)

            e = actlang.expr
            n = actlang.names
            pre_callbacks = \
                (("entry", recurse),
                 (e.NameMethodInvocationExpr, recurseAbstractOperation))
            post_callbacks = \
                (((n.SimpleName, n.QualifiedName, e.PrimaryFieldAccessExpr),
                  nameref),
                 (e.ParenExpr, paren),
                 ((e.CondAndExpr, e.CondOrExpr), binaryCond),
                 (e.LogicalNotExpr, neg),
                 (e.TernaryCondExpr, ternaryCond))
            evaluator = \
                ExpressionTraverser.ExpressionTraverser \
                    (ExpressionTraverser.ignore)
            for (cls, f) in pre_callbacks:
                evaluator.register_pre_callback(cls, f)
            for (cls, f) in post_callbacks:
                evaluator.register_post_callback(cls, f)
            return evaluator
                       
        evalAbstractExpr = __generateAbstractExpressionEvaluator()
        del __generateAbstractExpressionEvaluator

    def rewriteModel(self, flattening_threshold=0, rename_transitions=False, undefined_ops=None):
        """Rewrite act.lang. code in model based on results of type inference.

        Arguments:
            flattening_threshold -- Threshold value for converting abstract
                                    operations into case analyses.  If 0,
                                    operation flattening is disabled.  If
                                    negative, all admissible abstract
                                    operations are flattened.  If positive,
                                    only those operations are flattened
                                    whose input domains (i.e., the product
                                    of the abstract domains of the
                                    attributes referenced in an expression)
                                    have at most as many elements as the
                                    given threshold.
            rename_transitions   -- If True, rename the transitions in the
                                    model with the action language code
                                    associated with the transitions.
            undefined_ops        -- If not None, a set in which to collect
                                    operations with incomplete definitions
                                    (see below).  Each element added to
                                    the set is a tuple of the form
                                      (<type name>, <operator symbol>,
                                       <arity>)
                                    or
                                      ("coercion", <name of source type>,
                                       <name of target type>, "Jumbala").

        This function will replace the action language code appearing in
        guards and effects of transitions in state machines in the model
        from which the TypeInferenceSystem was initialized with abstract
        versions of the code.  This involves
            (i)  wrapping expressions and subexpressions occurring in the
                 code inside coercions to convert values used in a
                 particular context into the abstract type of the
                 context (determined by type inference), and
            (ii) inlining the definitions of operations between abstract
                 values in the code.  Abstract operations with multiple
                 return values (of which one should be chosen
                 nondeterministically) are translated into calls of
                 the special methods `$$SMUML.choose_boolean' and
                 `$$SMUML.choose_int', which are to be interpreted by a
                 back-end tool used to analyze the abstract models.
                 These function calls have four parameters (in order):
                   1. a string literal in the form of a comma-separated
                      list of boolean ($$SMUML.choose_boolean) or integer
                      ($$SMUML.choose_int) values in the Jumbala action
                      language, representing the set of values from which
                      one should be chosen nondeterministically;
                   2. a string literal encoding the name of the abstract
                      type of the boolean/integer values in the string
                      literal given in the first argument;
                   3. a nonnegative integer identifying the instance of a
                      nondeterministic operation that the `choose'
                      function represents within a guard or effect of a
                      transition in a state machine (*); and
                   4. a string literal of the form "guard:<uuid>" or
                      "effect:<uuid>" identifying whether the expression
                      in which the invocation of the `choose' function
                      occurs originates from a guard or an effect of a
                      transition in a state machine, where <uuid> is the
                      UUID of the Coral object representing the
                      transition.

        (*) The instances of `choose' method invocations are numbered in
            the abstract model such that a method invocation with a
            specific identifier represents a particular nondeterministic
            choice to be made when evaluating the expression containing
            the method invocation.  (It is possible for a `choose' method
            call to occur multiple times with the same identifier in an
            expression; in this case all instances of the method call
            represent the same nondeterministic choice.)  Each individual
            nondeterministic choice made in a guard or an effect of a
            transition has a unique identifier within the guard or effect;
            however, `choose' method invocations in two different guards
            or effects always represent different nondeterministic choices.

        If `undefined_ops' is None, the function will raise a
        RefinementError (more specifically, an UndefinedOperationError or
        an UndefinedCoercionError) if the model rewriting procedure fails
        due to a missing or incomplete definition of an operation which
        would need to be inlined in the rewritten model.  Otherwise, all
        undefined or partially defined operations will be collected into
        the `undefined_ops' set.  In both cases, the state of the model
        to which the rewriting procedure is applied will remain undefined
        until this function is called again successfully (after completing
        the definitions of any missing operations).
        """
        if not isinstance(flattening_threshold, int):
            raise TypeError \
                      ("an integer expected as `flattening_threshold' "
                       "argument")
        if undefined_ops != None and not isinstance(undefined_ops, set):
            raise TypeError("a set expected as `undefined_ops' argument")
        info = self.__ExpressionRewriterInfo \
                   (self.__typelibrary, self.__expr_node, flattening_threshold,
                    undefined_ops)
        # Introduce type coercions for class attribute initializer expressions
        for (a, n) in self.__attribute_nodes.iteritems():
            if n.is_abstractable and n.inferred_type != n.model_type \
                    and n.initializer_expression:
                result = \
                    self.__rewriteExpression \
                        (n.initializer_expression, info, n.inferred_type)
                # Check that the abstracted expression contains no method calls
                # (i.e., nondeterministic "choose" operations)
                def __raise(*args):
                    raise RewriteError \
                              ("initial value `%s' of attribute `%s' does not "
                               "abstract to a unique abstract value"
                               % (str(n.initializer_expression),
                                  Utils.Models.get_qualified_name(a)))
                check_no_methods = \
                    ExpressionTraverser.ExpressionTraverser \
                        (ExpressionTraverser.ignore,
                         ExpressionTraverser.ignore)
                check_no_methods.register_pre_callback \
                    (actlang.expr.NameMethodInvocationExpr, __raise)
                check_no_methods(result)
                if not a.initialValue:
                    TM = coral.getTransactionManager()
                    TM.beginModifyCmd()
                    a.initialValue = _UML.Expression()
                    TM.endModifyCmd()
                a.initialValue.body = str(result)
        # Rewrite transition guards and effects of state machines
        as = actlang.statement
        for statemachine in self.__statemachines:
            info.uml_context = statemachine.context
            for state in statemachine.top.subvertex:
                for t in state.outgoing:
                    if hasattr(t, 'guard') and t.guard:
                        info.choose_id = -1
                        info.context = "guard:%s" % t.__XMIid__()
                        e = self.__uml_to_statement[t.guard][0].api_getExpr()
                        t.guard.expression.body = \
                            str(self.__rewriteExpression(e, info))
                    if hasattr(t, 'effect') and t.effect:
                        info.choose_id = -1
                        info.context = "effect:%s" % t.__XMIid__()
                        stmts = self.__uml_to_statement[t.effect]
                        rewritten_stmts = []
                        for s in stmts:
                            if isinstance(s, (as.AssertStatement,
                                              as.ExprStatement)):
                                rewritten_stmts.append \
                                    (s.__class__ \
                                     (self.__rewriteExpression \
                                          (s.api_getExpr(), info),
                                      (0, 0)))
                            elif isinstance(s, as.EmptyStatement):
                                rewritten_stmts.append(s)
                            elif isinstance(s, as.SendStatement):
                                signame = s.api_getSignalName()
                                target = s.api_getTarget()
                                rewritten_parms = []
                                for (n, e) in \
                                        zip(self.__sigparm_nodes[signame],
                                            s.api_getArgList()):
                                    rewritten_parms.append \
                                        (self.__rewriteExpression
                                             (e, info, n.inferred_type))
                                rewritten_stmts.append \
                                    (as.SendStatement \
                                     (signame, rewritten_parms,
                                      self.__rewriteExpression(target, info),
                                      (0, 0)))
                            else:
                                raise RewriteError \
                                      ("unexpected: fall through "
                                       "(rewriteModel)")
                        t.effect.script.body = \
                            " ".join(map(str, rewritten_stmts))
                        self.__uml_to_rewritten_statement[t.effect] = \
                            rewritten_stmts
                    if rename_transitions:
                        name = []
                        if hasattr(t, 'trigger') and t.trigger:
                            name.extend((str(t.trigger.signal.name), '('))
                            name.extend([(hasattr(p, "name") and p.name
                                          and str(p.name)) or "_"
                                         for p in t.trigger.parameter])
                            name.append(') ')
                        if hasattr(t, 'guard') and t.guard:
                            name.extend \
                                (('[', str(t.guard.expression.body), '] / '))
                        if hasattr(t, 'effect') and t.effect:
                            name.append(str(t.effect.script.body))
                        t.name = "".join(name)
        if undefined_ops:
            # if it is not possible to rewrite action language expressions in
            # the model associated with `self' (in which case it is already in
            # an undefined state), it does not make sense to continue with
            # an analysis of the instantiated objects in the model
            return
        TM = coral.getTransactionManager()
        TM.beginModifyCmd()
        # If the model contains instantiated objects with AttributeLinks to
        # attributes of abstracted types, the values associated with these
        # links need to be coerced into the appropriate abstract types.
        # Replace each DataValue associated with an AttributeLink referring to
        # an abstracted attribute with a new DataValue representing the
        # abstracted initial value of the unabstracted attribute.
        obsolete_datavalues = []
        for (o, links) in self.__abstractable_object_attrlinks.iteritems():
            for (link, node, concrete_initial_value) in links:
                if node.model_type == node.inferred_type:
                    continue
                try:
                    result = \
                        self.__typelibrary.applyCoercion \
                            ("int", node.inferred_type, concrete_initial_value)
                except RuntimeError:
                    raise RewriteError \
                              ("unexpected: cannot coerce %d into abstract "
                               "type `%s': coercion from `int' to `%s' is not "
                               "well-defined for the Python back-end"
                               % (concrete_initial_value, node.inferred_type,
                                  node.inferred_type))
                coerced_vals = result.raw_contents
                if len(coerced_vals) != 1:
                    raise RewriteError \
                              ("unexpected: cannot coerce %d into abstract "
                               "type `%s': result of coercion is ambiguous "
                               "(for the Python back-end)"
                               % (val, node.inferred_type))
                if len(link.value.attributeLink) == 1:
                    # if replacing the value of the AttributeLink with a new
                    # DataValue will remove all references to the "old"
                    # DataValue, prepare to remove the "old" DataValue from the
                    # model
                    obsolete_datavalues.append(link.value)
                link.value = \
                    _UML.DataValue \
                        (name=str(coerced_vals[0]),
                         classifier=link.value.classifier,
                         owner=link.value.owner)
        TM.endModifyCmd()
        # remove all obsolete DataValues from the model
        TM.beginModifyCmd()
        for v in obsolete_datavalues:
            owner = v.owner
            if owner.oclIsKindOf(_UML.Object):
                owned_elements = owner.ownedInstance
            else:
                # owner is a UML Class, Model, or Package (guaranteed by
                # TypeInferenceSystem.__init__)
                owned_elements = owner.ownedElement
            owned_elements.remove(v)
        TM.endModifyCmd()

    def __generateAbstractOperationSubstitutor():
        """Generate a function object for substituting abstract operations."""

        inconsistency_msg =  \
            "unexpected type inconsistency while substituting abstract " \
            "operations"

        # Post-callbacks.  Each callback takes an actlang.expr.Expr object,
        # a list of subresults, and an __ExpressionRewriterInfo object as
        # parameters and returns a triple of the form
        #     (<expr>, <type name>, <rewritten>)
        # where <expr> is the result of substituting abstract operations into
        # the argument expression, <type name> is the name of the type
        # associated with <expr>, and <rewritten> is a Boolean value telling
        # whether any substitutions were actually made in the expression.

        def paren(expr, subexprs, info):
            expr_type = info.expr_node[expr].inferred_type
            (inner_expr, inner_expr_type, inner_expr_rewritten) = subexprs[0]
            coerced_inner_expr = \
                info.coercedExpression(inner_expr, inner_expr_type, expr_type)
            if not inner_expr_rewritten and coerced_inner_expr == inner_expr:
                return (expr, expr_type, False)
            return (actlang.expr.ParenExpr(coerced_inner_expr, (0, 0)),
                    expr_type, True)

        def unaryOperation(expr, subexprs, info):
            expr_type = info.expr_node[expr].inferred_type
            (argument, argument_type, argument_rewritten) = subexprs[0]
            # the ! operator is not abstracted
            if isinstance(expr, actlang.expr.LogicalNotExpr) \
                   and expr_type != "boolean":
                raise RewriteError(inconsistency_msg + " (unaryOperation)")
            # there are no implicit coercions between primitive types
            if expr_type in ("int", "boolean"):
                if argument_type != expr_type:
                    raise RewriteError(inconsistency_msg + " (unaryOperation)")
                if not argument_rewritten:
                    return (expr, expr_type, False)
                return (expr.__class__(argument, (0, 0)), expr_type, True)
            if isinstance(expr, actlang.expr.UnaryMinusExpr):
                operator = '-'
            elif isinstance(expr, actlang.expr.UnaryPlusExpr):
                operator = '+'
            elif isinstance(expr, actlang.expr.BitNotExpr):
                operator = '~'
            else:
                raise RewriteError \
                          ("unsupported subexpression while substituting "
                           "abstract operations (unaryOperation)]")
            return (info.operationExpression \
                        (expr_type, expr_type, operator,
                         info.coercedExpression \
                             (argument, argument_type, expr_type)),
                    expr_type,
                    True)

        def simpleAssignment(expr, subexprs, info):
            expr_type = info.expr_node[expr].inferred_type
            ((lhs, lhs_type, lhs_rewritten),
             (rhs, rhs_type, rhs_rewritten)) = subexprs
            if lhs_type != expr_type:
                raise RewriteError(inconsistency_msg + " (simpleAssignment)")
            coerced_rhs = info.coercedExpression(rhs, rhs_type, expr_type)
            if not rhs_rewritten and coerced_rhs == rhs:
                return (expr, expr_type, False)
            return (actlang.expr.SimpleAssignmentExpr \
                        (lhs, coerced_rhs, (0, 0)),
                    expr_type,
                    True)

        def condAndOr(expr, subexprs, info):
            expr_type = info.expr_node[expr].inferred_type
            ((lhs, lhs_type, lhs_rewritten),
             (rhs, rhs_type, rhs_rewritten)) = subexprs
            # the && and || operators are not abstracted
            for t in (expr_type, lhs_type, rhs_type):
                if t != 'boolean':
                    raise RewriteError(inconsistency_msg + " (condAndOr)")
            if not lhs_rewritten and not rhs_rewritten:
                return (expr, expr_type, False)
            return (expr.__class__(lhs, rhs), expr_type, True)

        def infix(expr, subexprs, info):
            expr_type = info.expr_node[expr].inferred_type
            ((lhs, lhs_type, lhs_rewritten),
             (rhs, rhs_type, rhs_rewritten)) = subexprs
            o = expr.api_getOperator()
            if isinstance(o, (actlang.binaryop.EqualityOperator,
                              actlang.binaryop.NumericalComparison)):
                if expr_type != "boolean":
                    raise RewriteError(inconsistency_msg + " (infix)")
                domain_type = \
                    info.expr_node[expr].argument_constraint.inferred_type
            else:
                domain_type = expr_type
            coerced_lhs = info.coercedExpression(lhs, lhs_type, domain_type)
            coerced_rhs = info.coercedExpression(rhs, rhs_type, domain_type)
            if not domain_type in info.typelibrary.typenames \
                   or domain_type in info.typelibrary.STD_TYPES:
                if not lhs_rewritten and not rhs_rewritten \
                       and coerced_lhs == lhs and coerced_rhs == rhs:
                    return (expr, expr_type, False)
                return (actlang.expr.InfixExpr(coerced_lhs, coerced_rhs, o),
                        expr_type, True)
            else:
                return (info.operationExpression \
                            (domain_type, expr_type, str(o), coerced_lhs,
                             coerced_rhs),
                        expr_type,
                        True)

        def ternaryCond(expr, subexprs, info):
            expr_type = info.expr_node[expr].inferred_type
            ((cond_expr, cond_type, cond_rewritten),
             (true_expr, true_type, true_rewritten),
             (false_expr, false_type, false_rewritten)) = subexprs
            if cond_type != "boolean":
                raise RewriteError(inconsistency_msg + " (ternaryCond)")
            coerced_true_expr = \
                info.coercedExpression(true_expr, true_type, expr_type)
            coerced_false_expr = \
                info.coercedExpression(false_expr, false_type, expr_type)
            if not cond_rewritten and not true_rewritten \
                   and not false_rewritten and coerced_true_expr == true_expr \
                   and coerced_false_expr == false_expr:
                return (expr, expr_type, False)
            return (actlang.expr.TernaryCondExpr \
                        (cond_expr, coerced_true_expr, coerced_false_expr),
                    expr_type,
                    True)

        def backtrack(expr, retval, info):
            # `expr' is a concrete subexpression: associate the abstracted
            # version of `expr' with the RewriteInfo of `expr'
            info.updateAbstractCorrespondence(expr, retval[0])

        e = actlang.expr
        n = actlang.names
        post_callbacks = \
            (((e.BoolLiteralExpr, e.IntLiteralExpr, n.SimpleName,
               n.QualifiedName, e.PrimaryFieldAccessExpr, e.ThisExpr,
               e.InstanceCreationExpr, e.NullExpr),
              lambda expr, subexprs, info:
                  (expr, info.expr_node[expr].inferred_type, False)),
             (e.ParenExpr, paren),
             ((e.UnaryPlusExpr, e.UnaryMinusExpr, e.BitNotExpr,
               e.LogicalNotExpr), unaryOperation),
             (e.SimpleAssignmentExpr, simpleAssignment),
             ((e.CondAndExpr, e.CondOrExpr), condAndOr),
             (e.InfixExpr, infix),
             (e.TernaryCondExpr, ternaryCond),
             ("exit", backtrack))
        substitutor = \
            ExpressionTraverser.ExpressionTraverser(ExpressionTraverser.ignore)
        for (cls, f) in post_callbacks:
            substitutor.register_post_callback(cls, f)
        return substitutor

    # Function object for substituting abstract operations and coercions in
    # an expression.
    __substituteAbstractOperations = __generateAbstractOperationSubstitutor()
    del __generateAbstractOperationSubstitutor

    def __generateDefinitionInliner():
        """Gen. a func. object for inlining defs. of abstract operations."""

        # Pre-callback for inlining definitions of abstract operations and
        # coercions between abstract types.  The pre-callback takes an
        # actlang.expr.Expr object and an __ExpressionRewriterInfo object as
        # parameters and raises an ExpressionTraverser.Return exception
        # initialized with a pair (expr, True), where `expr' is the
        # expression obtained from the input expression by (recursively)
        # inlining all definitions of abstract operations in the expression.

        def inlineAbstractOperationDefinition(expr, info):
            result = None
            name = str(expr.api_getInvocationName())
            if info.flattening_threshold != 0:
                result = info.flattenedExpression(expr)
                if result != None:
                    # record the fact that the abstracted expression is now
                    # flattened
                    info.abstract_expr_to_expr_node[expr].rewrite_info \
                        .flattened = True
            if result == None:
                args = expr.api_getArgList()
                if name == "$$SMUML.operation":
                    domain_type = args.pop(0).api_getValue()
                    range_type = args.pop(0).api_getValue()
                    operation_symbol = args.pop(0).api_getValue()
                    inlined_args = \
                        [TypeInferenceSystem.
                             _TypeInferenceSystem__inlineAbstractOpDefs
                             (a, info)[0]
                         for a in args]
                    (result, args_referred_to_in_result) = \
                        info.inlineOperationDefinition \
                            (domain_type, range_type, operation_symbol,
                             *inlined_args)
                    if len(args_referred_to_in_result) != len(inlined_args):
                        # if there is an argument expression not referenced in
                        # the result, mark the expression as flattened
                        info.abstract_expr_to_expr_node[expr].rewrite_info \
                            .flattened = True
                elif name == "$$SMUML.coerce":
                    inlined_arg = \
                        TypeInferenceSystem. \
                            _TypeInferenceSystem__inlineAbstractOpDefs \
                            (args[0], info)[0]
                    (result, result_refers_to_operand) = \
                        info.inlineCoercionDefinition \
                            (inlined_arg, args[1].api_getValue(),
                             args[2].api_getValue())
                    if not result_refers_to_operand:
                        # if the operand is not referenced in the coercion
                        # expression, mark the expression as flattened
                        info.abstract_expr_to_expr_node[expr].rewrite_info \
                            .flattened = True
                else:
                    raise RewriteError \
                              ("unsupported method call while inlining "
                               "abstract operations")
            raise ExpressionTraverser.Return((result, True))

        # Helper function for detecting whether a subexpression of an
        # expression was rewritten or not.

        def checkForRewrittenSubexprs(expr, subexprs):
            for (e, r) in subexprs:
                if r:
                    return
            raise ExpressionTraverser.Return((expr, False))

        # Post-callbacks.  Each callback takes an actlang.expr.Expr object,
        # a list of subresults, and an __ExpressionRewriterInfo object as
        # parameters and returns a pair of the form
        #     (<expr>, <rewritten>)
        # where <expr> is the result of inlining the definitions of any
        # abstract operations in the argument expression, and <rewritten>
        # is a Boolean value telling whether any rewriting was actually
        # applied to the expression.

        def defaultWithLineSpan(expr, subexprs, info):
            checkForRewrittenSubexprs(expr, subexprs)
            args = [s[0] for s in subexprs]
            args.append((0, 0))
            return (expr.__class__(*args), True)

        def defaultWithoutLineSpan(expr, subexprs, info):
            checkForRewrittenSubexprs(expr, subexprs)
            return (expr.__class__(*[s[0] for s in subexprs]), True)

        def infix(expr, subexprs, info):
            checkForRewrittenSubexprs(expr, subexprs)
            return (actlang.expr.InfixExpr \
                        (subexprs[0][0], subexprs[1][0],
                         expr.api_getOperator()),
                    True)

        def backtrack(expr, retval, info):
            (result, result_rewritten) = retval
            if result_rewritten:
                info.updateAbstractCorrespondence(expr, result)

        e = actlang.expr
        n = actlang.names
        post_callbacks = \
            (((e.BoolLiteralExpr, e.IntLiteralExpr, n.SimpleName,
               n.QualifiedName, e.PrimaryFieldAccessExpr, e.ThisExpr,
               e.InstanceCreationExpr, e.NullExpr),
              lambda expr, *rest: (expr, False)),
             ((e.ParenExpr, e.UnaryPlusExpr, e.UnaryMinusExpr, e.BitNotExpr,
               e.LogicalNotExpr, e.SimpleAssignmentExpr), defaultWithLineSpan),
             ((e.CondAndExpr, e.CondOrExpr, e.TernaryCondExpr),
              defaultWithoutLineSpan),
             (e.InfixExpr, infix),
             ("exit", backtrack))
        inliner = \
            ExpressionTraverser.ExpressionTraverser(ExpressionTraverser.ignore)
        inliner.register_pre_callback \
            (e.NameMethodInvocationExpr, inlineAbstractOperationDefinition)
        for (cls, f) in post_callbacks:
            inliner.register_post_callback(cls, f)
        return inliner

    # Function object for inlining the definitions of abstract operations and
    # coercions in an expression returned by __substituteAbstractOperations.
    __inlineAbstractOpDefs = __generateDefinitionInliner()
    del __generateDefinitionInliner

    def __generateVarReferenceCollector():
        """Generate a func. obj. for collecting var. refs. in an expression."""

        # Pre-callback for assignments to prevent adding their left-hand side
        # expressions to the result.  The callback will add the left hand side
        # of the assignment expression into the set `info.lhs_exprs'.

        def recurseAssignment(expr, info):
            info.lhs_exprs.add(expr.api_getLhsExpr())

        # Post-callbacks.  Each post-callback takes an actlang.expr.Expr
        # object, a list of lists of subresults for the immediate
        # subexpressions of the expression, and an __ExpressionRewriterInfo
        # object as parameters and returns a list of expressions representing
        # the variable references in the input expression.  As a side effect,
        # this list of expressions will also be associated with the input
        # expression in the `var_refs' dictionary of the
        # __ExpressionRewriterInfo object.

        def nameref(expr, subresults, info):
            if expr in info.lhs_exprs:
                return []
            return [expr]

        def method(expr, subresults, info):
            result = []
            for args in subresults[0]:
                result.extend(args)
            return result

        def default(expr, subresults, *rest):
            args = []
            for subargs in subresults:
                args.extend(subargs)
            return args

        def backtrack(expr, result, info, *rest):
            info.var_refs[expr] = result

        e = actlang.expr
        n = actlang.names
        post_callbacks = \
            (((e.BoolLiteralExpr, e.IntLiteralExpr, e.StringLiteralExpr,
               e.InstanceCreationExpr, e.ThisExpr, e.NullExpr),
              lambda *args: []),
             ((n.SimpleName, n.QualifiedName, e.PrimaryFieldAccessExpr),
              nameref),
             (e.NameMethodInvocationExpr, method),
             ((e.SimpleAssignmentExpr, e.CondAndExpr, e.CondOrExpr,
               e.LogicalNotExpr, e.InfixExpr, e.TernaryCondExpr, e.ParenExpr,
               e.UnaryPlusExpr, e.UnaryMinusExpr, e.BitNotExpr), default),
             ("exit", backtrack))
        ref_collector = \
            ExpressionTraverser.ExpressionTraverser \
                (ExpressionTraverser.ignore)
        ref_collector.register_pre_callback \
            (e.SimpleAssignmentExpr, recurseAssignment)
        for (cls, f) in post_callbacks:
            ref_collector.register_post_callback(cls, f)
        return ref_collector

    # Function object for collecting all variable references in an expression.
    __collectVarRefs = __generateVarReferenceCollector()
    del __generateVarReferenceCollector

    def __rewriteExpression(self, expr, info, target_type=None):
        """Rewrite an action language expression.

        Arguments:
            expr        -- An action language expression.
            info        -- An __ExpressionRewriterInfo object.
            target_type -- An optional target type in which to coerce
                           the rewritten expression (used for handling
                           attribute initializer expressions and signal
                           parameter expressions in `send' statements).

        Returns the rewritten action language expression.  As a side
        effect, updates the `rewrite_info' attributes of nodes
        corresponding to subexpressions of `expr' in the type inference
        system.
        """
        info.abstract_expr_to_expr_node = {}
        try:
            #print "expression: %s" % str(expr)
            result = self.__substituteAbstractOperations(expr, info)[0]
            if target_type != None:
                result = \
                    info.coercedExpression \
                        (result, info.expr_node[expr].inferred_type,
                         target_type)
            #print "    after substituting abstract operations: %s" \
            #      % str(result)
            if not info.undefined_ops:
                if info.flattening_threshold != 0:
                    info.lhs_expr = set()
                    info.var_refs = {}
                    self.__collectVarRefs(result, info)
                result = self.__inlineAbstractOpDefs(result, info)[0]
                #print "    after inlining the definitions of the operations:"\
                #      " %s" % str(result)
                #x = raw_input()
        except ExpressionTraverser.UnregisteredCallbackError, e:
            raise RewriteError \
                      ("unexpected error while rewriting model: %s" % str(e))
            
        return result

    # -------------------------------------------------------------------------

    def attachTypesToModel(self, type_library_files=None):
        """Update UML model with type associations from type inference.

        This function will update the model used to initialize the type
        inference system with tagged values to record the file names of
        type library files used for constructing an abstract type
        library, and an assignment of abstract data types (from the type
        library associated with the type inference system) to model
        attributes.  Furthermore, this function stores the correspondence
        between concrete and abstract subexpressions in the model using
        tagged values associated with guards and effects of transitions
        in the state machines of the model.

        Argument:
            type_library_files -- An optional list of file names to use
                                  as the files used to construct the
                                  type library associated with the type
                                  inference system.

        After calling this method,
            - self.__model.taggedValue will include exactly one object `v'
              satisfying
                  v.oclIsKindOf(coral.lang.UML14.TaggedValue) and
                      str(v.type.name) == "SMUML_typelibrary"
              Furthermore, str(v.dataValue) is a semicolon-separated list
              of file names.  All semicolons in the original file names
              have been replaced with the escape sequence `\;', and all
              backslashes with the escape sequence '\\'.  (NOTE: A simpler
              way to list the filenames would be to use a list as the
              dataValue, as apparently allowed by the UML 1.4
              specification.  However, Coral seems to accept only strings
              as dataValues.)
            - self.__model.taggedValue will include exactly one object `w'
              satisfying
                  w.oclIsKindOf(coral.lang.UML14.TaggedValue) and
                      str(w.type.name) == "SMUML_typeassignment"
              Furthermore, str(w.dataValue) is a semicolon-separated list
              of strings of the form
                  <UUID>==<typename>
              where
                  <UUID>     is the UUID of a model attribute, and
                  <typename> is a name of a type in the abstract type
                             library built from the library files listed in
                             the tagged value represented by the object
                             `v' described above.
            - every transition in a state machine of `self.__model' may
              have tagged values associated with its guard or effect as
              follows:
                  * If the transition t has a non-empty guard,
                    t.guard.taggedValue will include exactly one object
                    `d' satisfying
                        d.oclIsKindOf(coral.lang.UML14.TaggedValue) and
                            str(d.type.name) == "SMUML_expr_data"
                    Furthermore, str(d.dataValue) is a string which
                    consists of one or more `!'-separated segments of the
                    form <loc1>.<loc2>.<type name>.<flattened>, where
                      <loc1>      identifies a location of a
                                  subexpression of the guard expression
                                  in the original (unabstracted) model,
                      <loc2>      identifies the location of a
                                  subexpression of the corresponding guard
                                  expression in the abstract model such
                                  that the abstract subexpression at
                                  <loc2> is the "abstract version" of the
                                  concrete subexpression at <loc1>,
                      <type name> is the name of the abstract type of
                                  the subexpression identified by <loc2>
                                  in the abstract model (determined by
                                  rewriting the concrete model), and
                      <flattened> is 1 if the abstract subexpression at
                                  location <loc2> is flattened, and 0
                                  otherwise.
                    str(d.dataValue) contains a segment of the above form
                    for those subexpressions of the (concrete) guard
                    expression whose all ancestors in the syntax tree of
                    the expression have an unflattened corresponding
                    abstract subexpression.
                  * If the transition t has a non-empty effect,
                    t.effect.taggedValue will include exactly one object
                    `d' satisfying
                        d.oclIsKindOf(coral.lang.UML14.TaggedValue) and
                            str(d.type.name) == "SMUML_expr_data"
                    In this case, str(d.dataValue) is a string which
                    consists of one or more `|'-separated segments, one
                    for each simple statement in the effect of t.  The
                    segments are listed in the order of statements.
                    The segment corresponding to a simple statement `s'
                    is a `;'-separated string of substrings, one for
                    each top-level expression in the simple statement.
                    The substrings are listed in the order of the
                    top-level expressions as they occur from left to
                    right in the statement (currently only `send'
                    statements can have possibly more than one top-level
                    statement).  The substring corresponding to a
                    top-level expression `e' is a string of the form as
                    described above for guard expressions; this string
                    (a `!'-separated string of substrings) records the
                    correspondence between concrete and abstract
                    subexpressions of the top-level expression `e' in
                    the statement `s'.

        This method should be called only after `self.rewriteModel' has
        been called successfully.
        """
        tm = coral.getTransactionManager()
        tm.beginModifyCmd()

        # remove `SMUML_typelibrary' and `SMUML_typeassignment' tagged values
        # from model
        for v in \
            self.__model.taggedValue.select \
                (lambda v:
                     str(v.type.name) in ("SMUML_typelibrary",
                                          "SMUML_typeassignment")):
            self.__model.taggedValue.remove(v)

        # remove `SMUML_expr_data' tagged values from guards and effects of
        # transitions in the model
        for m in self.__statemachines:
            for s in m.top.subvertex:
                for t in s.outgoing:
                    for elem in ("guard", "effect"):
                        try:
                            e = getattr(t, elem)
                            for v in \
                                e.taggedValue.select \
                                    (lambda v:
                                     str(v.type.name) == "SMUML_expr_data"):
                                e.taggedValue.remove(v)
                        except AttributeError:
                            pass

        # remove tag definitions from model
        for e in \
            self.__model.ownedElement.select \
                (lambda e:
                     e.oclIsKindOf(_UML.TagDefinition)
                     and str(e.name) in ("SMUML_typelibrary",
                                         "SMUML_typeassignment",
                                         "SMUML_expr_data")):
            self.__model.ownedElement.remove(e)

        # create new tag definitions
        tagdefs = {}
        for tagname in ("SMUML_typelibrary", "SMUML_typeassignment",
                        "SMUML_expr_data"):
            mrange = _UML.MultiplicityRange(lower=1, upper=1)
            m = _UML.Multiplicity()
            m.range.add(mrange)
            tagdef = \
                _UML.TagDefinition \
                    (name=tagname, tagType="String", multiplicity=m)
            self.__model.ownedElement.add(tagdef)
            tagdefs[tagname] = tagdef

        # add the `SMUML_type_library_files' tagged value
        files = []
        for file in type_library_files:
            files.append(file.replace("\\", "\\\\").replace(";", "\\;"))
        value = \
            _UML.TaggedValue \
                (type=tagdefs["SMUML_typelibrary"],
                 name="SMUML_typelibrary", dataValue=";".join(files))
        self.__model.taggedValue.add(value)

        # add the `SMUML_type_assignment' tagged value
        assignments = []
        for n in self.findAbstractableNodes():
            assignments.append \
                ("%s==%s" % (n.coral_element.__XMIid__(), n.inferred_type))
        value = \
            _UML.TaggedValue \
                (type=tagdefs["SMUML_typeassignment"],
                 name="SMUML_typeassignment",
                 dataValue=";".join(assignments))
        self.__model.taggedValue.add(value)

        # add the `SMUML_expr_data' tagged values to guards and effects of
        # transitions
        assocStr = self.__expressionTypeAssociationString
        AS = actlang.statement
        for m in self.__statemachines:
            for s in m.top.subvertex:
                for t in s.outgoing:
                    for elem in ("guard", "effect"):
                        try:
                            e = getattr(t, elem)
                        except AttributeError:
                            continue
                        if not e:
                            continue
                        stmtstrs = []
                        if elem == "guard":
                            ce = self.__uml_to_statement[e][0].api_getExpr()
                            ae = self.__expr_node[ce].rewrite_info.expression
                            stmtstrs.append(assocStr(ce, ae))
                        else:
                            for (cs, as) in \
                                    zip(self.__uml_to_statement[e],
                                        self.__uml_to_rewritten_statement[e]):
                                if isinstance \
                                       (cs, (AS.AssertStatement,
                                             AS.ExprStatement)):
                                    stmtstrs.append \
                                        (assocStr
                                             (cs.api_getExpr(),
                                              as.api_getExpr()))
                                elif isinstance(cs, AS.EmptyStatement):
                                    # need to preserve correspondence between
                                    # statements
                                    stmtstrs.append("")
                                elif isinstance(cs, AS.SendStatement):
                                    exprstrs = []
                                    for (c_arg, a_arg) in \
                                            zip(cs.api_getArgList(),
                                                as.api_getArgList()):
                                        exprstrs.append(assocStr(c_arg, a_arg))
                                    exprstrs.append \
                                        (assocStr
                                             (cs.api_getTarget(),
                                              as.api_getTarget()))
                                    stmtstrs.append(";".join(exprstrs))
                        value = \
                            _UML.TaggedValue \
                                (type=tagdefs["SMUML_expr_data"],
                                 name="SMUML_expr_data",
                                 dataValue="|".join(stmtstrs))
                        e.taggedValue.add(value)

        tm.endModifyCmd()

    def __createExpressionTypeAssociator():
        """Construct a func. obj. for recording expression correspondence."""

        # Callback for recursing to a subexpression of a concrete expression.
        # As arguments, the callback receives (i) the concrete subexpression
        # `c_expr', (ii) the location `path' of `c_expr' in a top-level
        # concrete subexpression from which the search was started, (iii) a
        # top-level abstract subexpression `top_a_expr', (iv) a list `result'
        # in which to store the associations between locations of corresponding
        # concrete and abstract subexpressions, and (v) a dictionary
        # `expr_node' mapping concrete subexpressions into nodes in the type
        # inference system.  The callback returns no value, but adds into
        # `result' a string which records (i) an association between the
        # location `path' and the location of the abstract expression
        # corresponding to `c_expr' in `top_a_expr', (ii) the abstract type of
        # this abstract subexpression, and (iii) whether the abstract
        # subexpression is flattened or not.  If the abstract subexpression is
        # flattened, the search will not proceed to the subexpressions of
        # `c_expr'.
        def recurse(c_expr, path, top_a_expr, result, expr_node):
            info = expr_node[c_expr].rewrite_info
            abst_loc = \
                ExpressionTraverser.subexpression_to_path \
                    (top_a_expr, info.expression)
            result.append \
                ("%s.%s.%s.%s"
                 % (",".join(map(str, path)), ",".join(map(str, abst_loc)),
                    info.typename, info.flattened and "1" or "0"))
            path.append(0)
            if info.flattened:
                raise ExpressionTraverser.Return()

        # Callback used to update the current location of a concrete
        # subexpression during the search.
        def backtrack(c_expr, result, path, *rest):
            path.pop()
            if path:
                path[-1] += 1

        associator = \
            ExpressionTraverser.ExpressionTraverser \
                (ExpressionTraverser.ignore, ExpressionTraverser.ignore)
        associator.register_pre_callback("entry", recurse)
        associator.register_post_callback("exit", backtrack)
        return associator

    # Function object for constructing a string to record the correspondence
    # between concrete and abstract expressions.
    __associateTypes = __createExpressionTypeAssociator()
    del __createExpressionTypeAssociator

    def __expressionTypeAssociationString(self, top_c_expr, top_a_expr):
        """Construct a string representing associations between expressions.

        Arguments:
            top_c_expr -- An actlang.expr.Expr object (an expression from
                          concrete model).
            top_a_expr -- An actlang.expr.Expr object (the corresponding
                          expression from abstract model).

        Returns a `!'-separated string of strings of the form
        <loc1>.<loc2>.<type name>.<flattened>, where the components are
        described in the documentation of `attachTypesToModel'.
        """
        result = []
        self.__associateTypes \
            (top_c_expr, [], top_a_expr, result, self.__expr_node)
        return "!".join(result)

    def writeTypeAssociations(self, f="-", type_library_files=None):
        """Print the types of attributes in type association file format.

        Arguments:
            f                  -- Name of output file.  The name "-"
                                  refers to standard output.
            type_library_files -- An optional list of names of type
                                  library files for which to add "import"
                                  declarations to the output.
        """
        if f == "-":
            o = sys.stdout
        else:
            o = open(f, "w")
        if type_library_files:
            o.write("# Abstract type specification files\n")
            o.write("".join(['import "%s"\n'
                             % lf for lf in type_library_files]))
        for (t, f) in (("Class attributes",
                        lambda x: not isinstance(x, SignalParameterNode)),
                       ("Signal parameters",
                        lambda x: isinstance(x, SignalParameterNode))):
            attrs = []
            for n in self.__nodes:
                if isinstance(n, AttributeNode) and n.is_abstractable and f(n):
                    if isinstance(n, SignalParameterNode):
                        name = "%s::parameter_%d_%s" \
                               % (str(n.signal.name), n.position,
                                  str(n.coral_element.name))
                    else:
                        name = "%s::%s" \
                               % (Utils.Models.get_qualified_name \
                                  (n.coral_element.__parent__()),
                                  str(n.coral_element.name))
                    attrs.append \
                        ((name, n.coral_element.__XMIid__(), n.inferred_type))
            if attrs:
                attrs.sort(key=lambda x: x[0])
                o.write("# %s\n" % t)
                o.write("".join(["#   %s\n%s == %s\n" % a for a in attrs]))
        if o != sys.stdout:
            o.close()
