# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Abstract type library.

This module contains code for handling type libraries, and for reading
and writing type specification files.

Type specification files
------------------------

Abstract types can be defined from the contents of one or more type
specification files (see the documentation for the
abstraction.typelibrary.TypeLibrary.read function).

The basic structure of a type specification file is defined by the
grammar below.  Nonterminals in the grammar should be separated from other
sequences of non-whitespace characters by white space.  Strings in quotes
are to be interpreted literally, and the terminal symbols (can be assumed
to) have the following meaning:
    <EOL>: end of line
    <ID>: any string usable as a name for a Python identifier
    <OPERATOR-SYMBOL>: an arithmetic operator symbol in the Jumbala
                       action language
    <TEST-SYMBOL>: a Boolean operator symbol in the Jumbala action
                   language
    <ARITY>: "1" or "2"
    <STRING>: any string containing no newlines
    <INTEGER>: any integer

<type-specification-file> ::= <type-declaration>*

<type-declaration> ::= "type" <type-name> ":" <domain-specification> <EOL>
                       <type-definition>
<type-name> ::= <ID>
<domain-specification> ::= <symbolic-value-list>
                         | <symbolic-value-list> ":" <integer-value-list>
<type-definition> ::= (<operator-declaration> | <test-declaration>
                       | <coercion-declaration>)*

<operator-declaration> ::= "operator" <ARITY> <OPERATOR-SYMBOL> <EOL>
                           <operator-defs>
<operator-defs> ::= (<symbolic-value-list> ":" <symbolic-value-list> <EOL>)*

<test-declaration> ::= "test" <ARITY> <TEST-SYMBOL> <EOL>
                       <test-defs>
<test-defs> ::= (<symbolic-value-list> ":" <boolean-value-list> <EOL>)*

<coercion-declaration> ::= "coercion" <type-name> <backend-name> <EOL>
                           <coercion-defs>
<backend-name> ::= <ID>
<coercion-defs> ::= <STRING> ":" <symbolic-value-list> <EOL> <coercion-defs>
                  | "end" <EOL>

<symbolic-value-list> ::= <ID>
                        | <ID> "," <symbolic-value-list>
<integer-value-list> ::= <INTEGER>
                       | <INTEGER> "," <integer-value-list>
<boolean-value-list> ::= <boolean-value>
                       | <boolean-value> "," <boolean-value-list>
<boolean-value> ::= "true" | "false"

In the following, we use the term "type library" to refer to all types
defined in a given set of type specification files.

In brief, a type specification file contains definitions for abstract
types, operations and tests on the types, and coercions between the types.
The part following the first `#' of every input line (if this symbol is
present) is ignored.

The definition of an abstract type begins with the keyword "type"
followed by the name of the type (the names of primitive Jumbala types
"int" and "boolean" are reserved), and a comma-separated list of symbolic
names for the values in its domain separated from the type name with ":".
This list of symbolic names can optionally be followed by another ":" and
a comma-separated list of distinct integers to be used internally as the
numeric values of the corresponding elements of the symbolic domain of the
type.  (If this list is not specified, a suitable numeric domain will be
generated automatically.  The integers in the domain are guaranteed to
correspond to unique elements of the symbolic domain of the type.)

All operator, test, or coercion definitions before the next "type"
keyword in the file will be associated with this type ("the currently
active type").

An operator (test) for elements of the currently active type is declared
using the keyword "operator" ("test") followed by the arity n of the
operator or test (1 or 2), and the operator (test) symbol, which must be
an operator symbol (whose arity matches the specified one) in the Jumbala
action language. The possible value(s) yielded by the operator (test) shall
then be enumerated in the file separately for all n-tuples of input values.

A definition of a coercion from another type to the currently active type
T1 consists of the keyword "coercion" followed by the name of another type
T2 (a primitive type such as 'int' or 'boolean', or another type defined in
the type library) and the name of a "back-end" for which to define the
coercion.  This declaration is then followed by a sequence of lines
consisting of a string S containing no newlines, a ":", and a comma-separated
list L of values from the domain of the type T1. The definition of the
coercion is terminated by a line that contains the single keyword "end".

The string S is supposed to represent a condition under which a value of
type T2 will get coerced to (one of) the values in the list L.  For the
predefined back-ends "python" and "Jumbala", S must be a valid Boolean
expression template in the back-end language, i.e., a Boolean expression
in the language (with no references to variables), where the following
constructs have special semantics:
    - If the source type (T2) for the coercion is one of the primitive
      types of the Jumbala action language (int or boolean), all occurrences
      of <> will be replaced by the actual value to be coerced from T2 to
      T1.
    - If the source type (T2) for the coercion is another user-defined
      abstract type in the type library, all occurrences of <> will be
      replaced by the internal integer representation of the value (in
      the domain of T2) to be coerced from T2 to T1.  Furthermore, for
      any element E in the symbolic domain of T2, all occurrences of <E>
      will be replaced with the internal integer representation of the
      element E in the numeric domain of T2.
Furthermore, for these back-ends, the entire coercion has the following
semantics:
    - The tests of the coercion will be evaluated (after making the above
      substitutions) in their order of specification using `eval'.  (Thus
      every test should correspond to a valid Python expression after the
      substitutions.)
    - The result of the coercion will be the set of values associated with
      the first test which evaluates to True.
    - The result of the coercion is undefined if no test evaluates to True.
      The automatically generated code (see below) will raise a ValueError
      in this case.
For other back-ends, the string S and the coercion are not interpreted in
any way by the type library.  A user-defined back-end is thus free to
define its own semantics for the string S and the coercion.

The coercions defined between types in the type library should form a
partial order between the types such that more precise types can be coerced
only into types of (strictly) less precision, i.e., cyclic dependencies
between the precision of different types are not allowed.  The API provided
for defining coercions enforces this constraint automatically.  The API
treats the primitive types as maximally precise; using any of these
types as a target type of a coercion is not permitted.


Generating Python classes for abstract types
--------------------------------------------

The abstraction.typelibrary.TypeLibrary class provides a
`generatePythonClasses' method for generating Python classes from the
abstract types defined in a type library.  These classes will be added
to the abstraction.typelibrary module.

After calling `generatePythonClasses', the operators and tests of each
abstract type can be accessed either via the methods `operator' and
`test' in the corresponding Python class, or alternatively via the
module-level functions `abstraction.typelibrary.operator' and
`abstraction.typelibrary.test'.  (There exists also a module-level
function `abstraction.typelibrary.coercion' for accessing the
coercions defined in the abstract type library.)

Each instance of an abstract type actually represents a collection of
values in the domain of the abstract type.  This collection can be
accessed via the `contents' attribute of the instance.

The operators and tests differ in that the `operator' method of an
abstract type returns objects encoding values of the same abstract
type, but the `test' method returns tuples of Boolean literals.

An example of using the operations provided by the Python classes;
suppose that the abstract type library defines an abstract type
modeling the "sign" abstraction of unbounded integers:

>>> S = abstraction.typelibrary.Sign # Get a handle to an abstract type.
>>> S.domain.keys()                  # Get the domain of the abstract type.
['NEG', 'ZERO', 'POS']
>>> x = S(-100)                      # Abstract a few integers.
>>> y = S(100)
>>> x.contents                       # Get the "contents" of x...
['NEG']
>>> y.contents                       # ... and then y.
['POS']
>>> y = abstraction.typelibrary.coercion('Sign', 100)  # An alternative with
>>> y.contents                                         # the same effect.
['POS']
>>> w = S('POS')                     # Initialization with an abstract value
>>> v = S(['POS','ZERO'])            # ... or an iterable of values.
>>> w.contents
['POS']
>>> v.contents
['ZERO', 'POS']
>>> z = S.operator('+', x, y)        # Apply an operator.
>>> z.contents
['NEG', 'ZERO', 'POS']
>>> z = abstraction.typelibrary.operator('Sign', '+', x, y) # alternatively
>>> z.contents
['NEG', 'ZERO', 'POS']
>>> S.operator('*', z, S(0)).contents  # Another example.
['ZERO']
>>> S.test('<', x, y)                # Apply a test.
(True,)
>>> abstraction.typelibrary.test('Sign', '<', x, y)         # alternatively
(True,)

To access the numeric representation of the values in the domain of an
abstract type, use the `raw_contents' attribute of the abstract type
instance, e.g.,
  >>> v = abstraction.typelibrary.Sign('NEG')
  >>> v.raw_contents
  [0]
Constructing abstract value collections from these numeric values can be
done via the `numdomain' dictionary of the class representing the abstract
type, using the integer as the key.  For example:
  >>> abstraction.typelibrary.Sign(abstraction.typelibrary.Sign.numdomain[0]).contents
  ['NEG']
  >>> S(0).contents  # this will abstract the integer 0 to the Sign type!
  ['ZERO']

Author: Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

import copy
import os
import os.path
import sys
import re
import actlang
import error
from cartesize import cartesize
import typehierarchy
from Utils import ExpressionTraverser

# -----------------------------------------------------------------------------
# Exceptions.

class _EndOfInnerBlock(Exception):
    """Exception thrown at the end of a block when reading the type library."""

    pass

class _EndOfInput(Exception):
    """Exception thrown when finished reading the type library."""

    pass

class InputFormatError(error.Error):
    """An exception class for reporting errors while reading input."""

    pass

# -----------------------------------------------------------------------------
# Class used for reading abstract type specification files.

class _LineReader:
    """Line reader with a counter and an automatically updated line count."""

    def __init__(self, libfile, sloppy):
        """Initialize an input reader object.

        Argument:
            libfile -- A file object.  This file needs to be opened before
                       calling next().
            sloppy  -- Whether to accept only partially defined type
                       libraries.
        """

        self.file = libfile
        self.sloppy = sloppy
        self.line = ""
        self.line_num = 0

    def next(self):
        """Read the next nonempty line of input, skipping comments."""

        while True:
            self.line = self.file.readline()
            if self.line == "":
                raise _EndOfInput()
            self.line_num += 1
            pos = self.line.rfind('#')
            if pos != -1:
                self.line = self.line[:pos]
            self.line = self.line.strip()
            if self.line == "":
                continue
            return

# -----------------------------------------------------------------------------
# Base class for abstract types.

class AbstractType:
    """Base class for abstract types."""

    def __str__(self):
        """Return the contents of the object as a string."""
        return str(self.contents)

    def __getattr__(self, name):
        """Computed attributes.

        The following computed attributes are defined:
            contents     -- List of strings representing the contents
                            of the object as symbolic values from the
                            domain of the type.
            raw_contents -- List of integers representing the contents
                            of the object in the internal integer
                            representation.
        """
        if name == 'contents':
            return [self.numdomain[v] for v in self.value]
        elif name == 'raw_contents':
            return list(self.value)
        else:
            raise AttributeError \
                  ("%s instance has no attribute '%s'"
                   % (self.__class__.__name__, name))

    def operator(cls, operator, *arguments):
        """Map the Cartesian product of sets of elements in the domain of
        the type into a subset of the domain using an operator.

        Arguments:
            operator  -- Operator symbol.
            arguments -- Objects representing the product components.

        Returns an object representing the range of values obtained by
        applying the operator individually to each tuple in the Cartesian
        product of the sets of elements.
        """
        result = cls([])
        result.value = frozenset(AbstractType._AbstractType__map \
                                 (cls, cls.operators[len(arguments)][operator],
                                  arguments))
        return result

    def test(cls, operator, *arguments):
        """Map the Cartesian product of sets of elements in the domain of
        the type into a list of Booleans.

        Arguments:
            operator  -- Test symbol.
            arguments -- Objects representing the product components.

        Returns a tuple of Booleans representing the range of values
        obtained by applying the operator individually to each element in
        the Cartesian product of the sets of elements.
        """
        return tuple(AbstractType._AbstractType__map \
                     (cls, cls.tests[len(arguments)][operator], arguments))

    def __initByCoercion(value, testlist):
        """Initialize a collection of values via coercion from another type.

        Arguments:
            value    -- Object representing a collection of values to
                        coerce.
            testlist -- Definition of the type coercion as a list of pairs
                        of the form (Boolean expression, result of coercion).
        """
        result = set()
        if isinstance(value, AbstractType):
            for v in value.value:
                result.update(AbstractType._AbstractType__coerce(v, testlist))
        else:
            result.update(AbstractType._AbstractType__coerce(value, testlist))
        return frozenset(result)

    def __coerce(v, testlist):
        """Evaluate a coercion specification.

        Arguments:
            v        -- A (representation of a) value to be coerced.
            testlist -- Definition of the type coercion as a list of pairs
                        of the form (Boolean expression, result of
                        coercion).
        """
        for (test, retvals) in testlist:
            if eval(test.replace('<>', 'v')): # plug in argument of test
                return retvals
        raise ValueError("failed to coerce `" + str(v) + "'")

    def __map(cls, operation, arguments):
        for arg in arguments:
            if not isinstance(arg, cls):
                raise TypeError("%s arguments required" % cls.__name__)
        result = set()
        for v in cartesize(*[list(arg.value) for arg in arguments]):
            result.update(operation[tuple(v)])
        return result

    operator = classmethod(operator)
    test = classmethod(test)
    __initByCoercion = staticmethod(__initByCoercion)
    __coerce = staticmethod(__coerce)
    __map = staticmethod(__map)

# -----------------------------------------------------------------------------
# Class for representing collections of abstract values.

class ValueCollection:
    """Collections of abstract values."""
    
    def __init__(self, typelibrary, typename, values):
        """Initialize a ValueCollection.

        Arguments:
            typelibrary -- A TypeLibrary object.
            typename    -- A name of a type defined in `typelibrary'.
            values      -- A single value, or an iterable of values in
                           the symbolic or numeric domain of type
                           `typename' in `typelibrary'.
        """
        if not isinstance(typelibrary, TypeLibrary):
            raise TypeError("a TypeLibrary expected as first argument")
        if not isinstance(typename, str):
            raise TypeError("a type name expected as second argument")
        if not typename in typelibrary.typenames:
            raise ValueError("invalid type name `%s'" % typename)
        self.typelibrary = typelibrary
        self.typename = typename
        if not hasattr(values, "__iter__"):
            values = (values,)
        if typename in typelibrary.STD_TYPES:
            self.values = frozenset(values)
        else:
            self.values = \
                frozenset(typelibrary._TypeLibrary__normalizeValues \
                          (values, typelibrary.domain[typename],
                           typelibrary.numdomain[typename], True))

    def __getattr__(self, name):
        """Computed attributes.

        The following computed attributes are defined:
            contents     -- List of strings representing the contents
                            of the object as symbolic values from the
                            domain of the type associated with the
                            ValueCollection.
            raw_contents -- List of integers representing the contents
                            of the object as numeric values from the
                            domain of the type associated with the
                            ValueCollection.
        """
        if name == 'contents':
            if not self.typename in self.typelibrary.STD_TYPES:
                return [self.typelibrary.numdomain[self.typename][v]
                        for v in self.values]
            else:
                return list(self.values)
        elif name == 'raw_contents':
            return list(self.values)
        else:
            raise AttributeError \
                  ("%s instance has no attribute '%s'"
                   % (self.__class__.__name__, name))

    def __str__(self):
        """Return the string representation of the ValueCollection."""
        return "[%s] %s" % (self.typename, str(self.contents))

    def __iter__(self):
        """Return an iterable to the values in the ValueCollection."""
        return self.values.__iter__()

# -----------------------------------------------------------------------------
# Abstract type library.

class TypeLibrary:
    """Abstract type library.

    The class provides the class attribute STD_TYPES, which is a tuple of
    the names of the standard primitive types available in the Jumbala
    action language ('int' and 'boolean').  Similarly, the class attribute
    STD_BACKENDS gives a tuple of predefined coercion back-ends (i.e.,
    names of languages in which the coercion test can be written).

    Each instance of the class has the following public attributes:
        typenames      -- A set of names of types.
        domain         -- A dictionary of domains for the types.  For a type
                          t in typenames, `domain[t]' is a mapping from
                          strings (the symbolic elements in the domain of the
                          type) to integers (the corresponding numeric
                          values).
        numdomain      -- A dictionary of numeric domains for the types.  For
                          a type t in typenames, `numdomain[t]' is a mapping
                          from integers (the numeric elements in the domain
                          of the type) into strings (the corresponding
                          symbolic values).  This is the inverse mapping of
                          `domain'.
        operators      -- A dictionary of operators defined for the types.
                          For a type t in typenames, an arity n, and an
                          operator symbol s, `operators[t][n][s]' gives the
                          definition of the operator s for n-tuples of values
                          in the numeric domain of t, i.e.,
                          `operators[t][n][s][(v1,...,vn)]=[r1,...,rm]'
                          for some 1 <= m <= |domain(t)|, rj in domain(t) for
                          all j.
        tests          -- A dictionary of tests defined for the types.  For a
                          type t in typenames, an arity n, and an operator
                          symbol s, `tests[t][n][s]' gives the definition of
                          the test s for n-tuples of values in the numeric
                          domain of t, i.e.,
                          `tests[t][n][s][(v1,...,vn)] = [b1,...,bm]' for
                          some 1 <= m <= 2, bj in {True, False} for all j.
        coercions      -- A dictionary of coercions defined for the type.
                          For two types t1 and t2 in typenames and a back-end
                          b, `coercions[t1][t2][b]' gives the definition of
                          the coercion from a value of type t2 to a value of
                          type t1 for back-end b.  This definition is a list
                          of tuples with two elements (a test as a string and
                          a list of integers from the numeric domain of t1
                          representing the possible outcomes of the coercion
                          if the test evaluates to true).
        backends       -- A set of names for back-ends.
        type_ordering  -- Partial ordering of type names as a TypeHierarchy
                          object.
    """

    # Standard types of the Jumbala action language
    STD_TYPES = ('int', 'boolean')

    # Standard back-ends
    STD_BACKENDS = ('python', 'Jumbala')

    # Pattern for valid type names.
    __typename_re = re.compile("^[a-zA-Z_][a-zA-Z0-9_]*$")

    # Pattern for valid domain values.
    __domainvalue_re = re.compile("^[^:,\s]+$")
    
    def __init__(self):
        """Initialize an empty type library."""

        self.typenames = set(TypeLibrary.STD_TYPES)
        self.domain = {'boolean' : {'true' : True, 'false' :  False}}
        self.numdomain = {}
        self.operators = {}
        self.tests = {}
        self.coercions = {}
        self.backends = set(TypeLibrary.STD_BACKENDS)
        self.type_ordering = typehierarchy.TypeHierarchy()
        for t in TypeLibrary.STD_TYPES:
            self.type_ordering.addType(t)

    # -------------------------------------------------------------------------
    # Convenience public API for manipulating type libraries

    def addType(self, typename, domain=None, numdomain=None):
        """Add a new type to the type library.

        Arguments:
            typename  -- Name of the new type.
            domain    -- A sequence of strings representing the symbolic
                         domain of the type.
            numdomain -- A sequence of integers representing the numeric
                         domain of the type.  Can be specified only if
                         `domain' is also specified, in which case the
                         lengths of the sequences must match.  If not
                         specified, a numeric domain will be initialized
                         automatically.

        The function will add the new type to the type library.  The
        operators, tests and coercions defined for the type will initially
        be empty, and the type will initially be incomparable w.r.t. the
        precision ordering with every other type defined in the type
        library.
        """
        if not isinstance(typename, str):
            raise TypeError("expected a type name as first argument")
        if domain != None and not isinstance(domain, (list, tuple)):
            raise TypeError \
                  ("a sequence of strings expected as second argument")
        if numdomain != None:
            if domain == None:
                raise ValueError("missing definition of symbolic domain")
            if not isinstance(numdomain, (list, tuple)):
                raise TypeError \
                      ("a sequence of integers expected as third argument")
        if typename in self.typenames:
            raise ValueError("type `%s' has already been defined" % typename)
        if self.__typename_re.match(typename) == None:
            raise ValueError("`%s' is an invalid name for a type" % typename)
        if domain == None:
            domain = []
        if numdomain == None:
            numdomain = range(0, len(domain))
        d = {}
        n = {}
        self.__validateDomainValueLists(domain, numdomain)
        for (symval, numval) in zip(domain, numdomain):
            d[symval] = numval
            n[numval] = symval
        self.domain[typename] = d
        self.numdomain[typename] = n
        self.typenames.add(typename)
        self.operators[typename] = {}
        self.tests[typename] = {}
        self.coercions[typename] = {}
        if not typename in self.type_ordering:
            self.type_ordering.addType(typename)

    def removeType(self, typename):
        """Remove a non-primitive type from the type library.

        Argument:
            typename -- Name of the type to remove.

        The function will remove a type from the type library, all
        coercions from the type to other types in the library, and the
        precision ordering between the type and any other type in the
        type library.
        """
        if typename in self.STD_TYPES:
            raise ValueError("removal of primitive types is not allowed")
        self.typenames.remove(typename)
        del self.domain[typename]
        del self.numdomain[typename]
        del self.operators[typename]
        del self.tests[typename]
        del self.coercions[typename]
        for (totype, fromtypes) in self.coercions.iteritems():
            if typename in fromtypes:
                del fromtypes[typename]
        self.type_ordering.removeType(typename)

    def addDomainElements(self, typename, elements, numelements=None):
        """Extend the domain of a type.

        Arguments:
            typename    -- Name of a type in the type library.
            elements    -- New symbolic value to be added to the domain of
                           the type, or a container of such symbolic values.
            numelements -- An optional numeric value corresponding to the
                           symbolic one, or a container of such values.  If
                           not specified, the numeric values corresponding
                           to the new symbolic domain values will be
                           generated automatically.
        """
        if typename in self.STD_TYPES:
            raise ValueError("cannot extend the domains of primitive types")
        domain = self.domain[typename]
        numdomain = self.numdomain[typename]
        if not hasattr(elements, "__iter__"):
            elements = [elements]
        if numelements != None:
            if not hasattr(numelements, "__iter__"):
                numelements = [numelements]
        else:
            if not len(domain):
                val = 0
            else:
                val = max(numdomain.keys()) + 1
            numelements = range(val, val + len(elements))
        self.__validateDomainValueLists(elements, numelements)
        for (symval, numval) in zip(elements, numelements):
            if symval in domain:
                raise ValueError \
                      ("value `%s' is already included in the domain of "
                       "type `%s'" % (symval, typename))
            if numval in numdomain:
                raise ValueError \
                      ("value %d is already included in the numeric domain of "
                       "type `%s'" % (numval, typename))
        for (symval, numval) in zip(elements, numelements):
            domain[symval] = numval
            numdomain[numval] = symval
        for ops in (self.operators[typename], self.tests[typename]):
            for (arity, symbols) in ops.iteritems():
                for t in cartesize(*(arity * [numdomain.keys()])):
                    args = tuple(t)
                    for (symbol, dfn) in symbols.iteritems():
                        if not args in dfn:
                            dfn[args] = []

    def removeDomainElements(self, typename, elements, collect_incomplete=False):
        """Remove elements from the domain of a non-primitive type.

        Arguments:
            typename           -- Name of a non-primitive type in the type
                                  library.
            elements           -- Symbolic or numeric value of the element
                                  to be removed from the domain of the
                                  type, or a container of such values.
            collect_incomplete -- Whether to collect operators and
                                  coercions whose definitions become
                                  incomplete in the operation.  See the
                                  explanation below.

        The function will remove the given elements from the domain of the
        given type.  All definitions of the type's operators, tests and
        coercions which refer to the removed element will be updated
        accordingly.  Furthermore, all coercions to other types from the
        type will be updated such that every test referring to any of the
        removed domain elements will be removed.

        In a well-formed type library, every operator, test, or a test in
        a coercion defined for a type should have at least one possible
        return value.  Obviously, the removal of an element from the
        domain of the type may break this property for operators and
        coercions (because tests do not return values from the domain of
        the type, their results are not affected by the removal), and even
        the domain of the type may become empty.  How to react to such
        situations is controlled by the optional `collect_incomplete'
        argument: when True, the function will return a triple containing
        the elements
             - a truth value telling whether the domain of the type is
               still nonempty after the operation,
             - a list of triples of the form
                   (operator_symbol, arity, arguments),
               where
                   operator_symbol  is an operator symbol as a string,
                   arity            is the arity of the operator, and
                   arguments        is a tuple of values (of length
                                    `arity') in the numeric domain of
                                    the type such that the operator
                                    identified by `operator_symbol' has
                                    an empty list of return values on
                                    `arguments' after removing the given
                                    element from the domain of the type.
             - a list of quadruples of the form
                   (fromtype, totype, backend, n)
               such that the n'th test (counting from 0) in the coercion
               defined from `fromtype' to `totype' for back-end `backend'
               has no return values after removing the given element from
               the domain of the type.
        (If `collect_incomplete' is False, the function returns None.)
        """
        if typename in self.STD_TYPES:
            raise ValueError \
                  ("cannot remove elements from domains of primitive types")
        domain = self.domain[typename]
        numdomain = self.numdomain[typename]
        if not hasattr(elements, "__iter__"):
            elements = [elements]
        removed_sym_elems = set()
        removed_num_elems = set()
        for e in elements:
            if not isinstance(e, (str, int)):
                raise ValueError \
                      ("expected a string, integer, or a container of strings "
                       "and integers as second argument")
            if isinstance(e, str):
                (symval, numval) = (e, domain[e])
            else:
                (symval, numval) = (numdomain[e], e)
            removed_sym_elems.add(symval)
            if numval in removed_num_elems:
                raise ValueError \
                      ("value `%s' (%d) specified multiple times"
                       % (symval, numval))
            removed_num_elems.add(numval)
        if collect_incomplete:
            incomplete_operators = []
            incomplete_coercions = []
        type_operators = self.operators[typename]
        type_tests = self.tests[typename]
        type_coercions = self.coercions[typename]
        # Update operators and tests of the type
        for operations in (type_operators, type_tests):
            for (arity, operators) in operations.iteritems():
                for (operator, definition) in operators.iteritems():
                    args_to_undefine = []
                    for (args, vals) in definition.iteritems():
                        for e in removed_num_elems:
                            if e in args:
                                args_to_undefine.append(args)
                                break
                        if operations is type_operators:
                            for e in removed_num_elems:
                                while e in vals:
                                    vals.remove(e)
                            if collect_incomplete and not vals:
                                incomplete_operators.append \
                                    ((operator, arity, args))
                    for args in args_to_undefine:
                        del definition[args]
        # Update coercions to the type
        for (fromtype, backends) in type_coercions.iteritems():
            for (backend, definition) in backends.iteritems():
                for (i, (test, retvals)) in enumerate(definition):
                    for e in removed_num_elems:
                        while e in retvals:
                            retvals.remove(e)
                    if collect_incomplete and not retvals:
                        incomplete_coercions.append \
                            ((fromtype, totype, backend, i))
        # Update coercions from the type to other types
        for (totype, fromtypes) in self.coercions.iteritems():
            if totype != typename:
                for (fromtype, backends) in fromtypes.iteritems():
                    for (backend, definition) in backends.iteritems():
                        if backend in self.STD_BACKENDS:
                            tests_to_remove = []
                            for (i, (test, retvals)) in enumerate(definition):
                                for e in removed_sym_elems:
                                    if test.find("<" + e + ">") != -1:
                                        tests_to_remove.append(i)
                                        break
                            while tests_to_remove:
                                del definition[tests_to_remove.pop()]
        for e in removed_sym_elems:
            del domain[e]
        for e in removed_num_elems:
            del numdomain[e]
        if collect_incomplete:
            return (len(domain) > 0, incomplete_operators,
                    incomplete_coercions)
        else:
            return None

    def addOperator(self, typename, operator, arity):
        """Add an operator for a non-primitive type in the type library.

        Arguments:
            typename -- Name of a non-primitive type in the type library.
            operator -- An operator symbol as a string.
            arity    -- Arity of the operator.
        """
        self.__addOperation \
            (typename, operator, arity, self.operators[typename])

    def addTest(self, typename, operator, arity):
        """Add a test for a non-primitive type in the type library.

        Arguments:
            typename -- Name of a non-primitive type in the type library.
            operator -- A test operator symbol as a string.
            arity    -- Arity of the test.
        """
        self.__addOperation(typename, operator, arity, self.tests[typename])

    def __addOperation(self, typename, operator, arity, operations):
        """Add an operation for a non-primitive type in the type library.

        Arguments:
            typename   -- Name of a non-primitive type in the type library.
            operator   -- An operation symbol as a string.
            arity      -- Arity of the operation.
            operations -- Dictionary of operations.
        """
        if typename in self.STD_TYPES:
            raise ValueError("cannot define operators for primitive types")
        if not isinstance(operator, str):
            raise TypeError("an operator symbol expected as second argument")
        if not isinstance(arity, int):
            raise TypeError("an integer expected as third argument")
        if arity <= 0:
            raise ValueError("arity must be positive")
        dfn = {}
        try:
            operators = operations[arity]
            if operator in operators:
                raise ValueError \
                      ("operation `%s' of arity %d has already been defined "
                       "for type `%s'" % (operator, arity, typename))
            operators[operator] = dfn
        except KeyError:
            operations[arity] = {operator : dfn}
        for args in cartesize(*(arity * [self.numdomain[typename].keys()])):
            dfn[tuple(args)] = []

    def defineOperator(self, typename, operator, operands, values):
        """(Partially) (re)define an operator for a non-primitive type.

        Arguments:
            typename -- Name of a non-primitive type in the type library.
            operator -- An operator symbol as a string.
            operands -- A list or a tuple of values from the symbolic or
                        numeric domain of the type (the operands for the
                        operator).  The operand container may not be
                        empty.
            values   -- A collection of values from the symbolic or
                        numeric domain of the type (the possible return
                        values of the operator when applied to the
                        operands in `operands').

        The function adds the definition of the given operator for the
        given type and operands to the type library, or redefines the
        operator for the given type and operands if it had already been
        defined for these operands.  (Also the operator is added to the
        type library if it is has not been previously introduced.)
        """
        self.__defineOperation \
            (typename, operator, operands, values, self.operators[typename])

    def defineTest(self, typename, operator, operands, values):
        """(Partially) (re)define a test for a non-primitive type.

        Arguments:
            typename -- Name of a non-primitive type in the type library.
            operator -- A test operator symbol as a string.
            operands -- A list or a tuple of values from the symbolic or
                        numeric domain of the type (the operands for the
                        test).  The operand container may not be empty.
            values   -- A collection of values.  The possible outcomes
                        of the test will be obtained from these values by
                        converting them to Boolean values.

        The function adds the definition of the given test for the given
        type and operands to the type library, or redefines the test for
        the given type and operands if it had already been defined for
        these operands.  (Also the test is added to the type library if it
        has not been previously introduced.)
        """
        self.__defineOperation \
            (typename, operator, operands, values, self.tests[typename])

    def __defineOperation(self, typename, operator, operands, retvals,
                          operations):
        """(Partially) (re)define an operation for a non-primitive type.

        Arguments:
            typename   -- Name of a non-primitive type in the type library.
            operator   -- An operation symbol as a string.
            operands   -- A list or a tuple of values from the symbolic or
                          numeric domain of the type (the operands for
                          the operation).  The operand container must not
                          be empty.
            values     -- A collection of values representing the possible
                          outcomes of the operation on the given operands.
            operations -- A dictionary of operations.
        """
        if not isinstance(operands, (list, tuple)):
            raise TypeError \
                  ("a list or a tuple of values expected as third argument")
        if not operands:
            raise ValueError("operand container must be nonempty")
        if not hasattr(retvals, "__iter__"):
            raise TypeError("a container expected as fourth argument")
        type_domain = self.domain[typename]
        type_numdomain = self.numdomain[typename]
        operands = tuple(self.__normalizeValues
                         (operands, type_domain, type_numdomain))
        if operations is self.tests[typename]:
            vals = set()
            for val in retvals:
                vals.add((val and True) or False)
            retvals = list(vals)
        else:
            retvals = self.__normalizeValues \
                      (retvals, type_domain, type_numdomain, True)
        arity = len(operands)
        if not arity in operations or not operator in operations[arity]:
            self.__addOperation(typename, operator, arity, operations)
        operations[arity][operator][operands] = retvals

    def undefineOperator(self, typename, operator, operands):
        """Remove the definition of an operator for given operands.

        Arguments:
            typename -- Name of a non-primitive type in the type library.
            operator -- An operator symbol as a string.
            operands -- A list or a tuple of values from the symbolic or
                        numeric domain of the type (the operands for the
                        operator).

        This function will remove the definition of an operator for a
        type and a specific combination of operands.  The operator itself
        will remain in the type library even if it is not defined for
        any combination of operands after calling this function.
        """
        self.__undefineOperation \
            (typename, operator, operands, self.operators[typename])

    def undefineTest(self, typename, operator, operands):
        """Remove the definition of a test for given operands.

        Arguments:
            typename -- Name of a non-primitive type in the type library.
            operator -- A test operator symbol as a string.
            operands -- A list or a tuple of values from the symbolic or
                        numeric domain of the type (the operands for the
                        test).

        This function will remove the definition of an test for a type
        and a specific combination of operands.  The test itself will
        remain in the type library even if it is not defined for any
        combination of operands after calling this function.
        """
        self.__undefineOperation \
            (typename, operator, operands, self.tests[typename])

    def __undefineOperation(self, typename, operator, operands, operations):
        """Remove the definition of an operation for given operands.

        Arguments:
            typename -- Name of a non-primitive type in the type library.
            operator -- An operation symbol as a string.
            operands -- A list or a tuple of values from the symbolic or
                        numeric domain of the type (the operands for the
                        operation).
        """
        operands = tuple(self.__normalizeValues
                         (operands, self.domain[typename],
                          self.numdomain[typename]))
        del operations[len(operands)][operator][operands]

    def removeOperator(self, typename, operator, arity):
        """Remove the definition of an operator from the type library.

        Arguments:
            typename -- Name of a type in the type library.
            operator -- An operator symbol as a string.
            arity    -- Arity of the operator.

        This function will remove the definition of an operator and also
        the operator itself from the type library.
        """
        self.__removeOperation(operator, arity, self.operators[typename])

    def removeTest(self, typename, operator, arity):
        """Remove the definition of a test from the type library.

        Arguments:
            typename -- Name of a type in the type library.
            operator -- A test operator symbol as a string.
            arity    -- Arity of the test.

        This function will remove the definition of a test and also the
        the test itself from the type library.
        """
        self.__removeOperation(operator, arity, self.tests[typename])

    def __removeOperation(self, operator, arity, operations):
        """Remove the definition of an operation from the type library.

        Arguments:
            operator   -- An operation symbol as a string.
            arity      -- Arity of the operation.
            operations -- A dictionary of operations.
        """
        arities = operations[arity]
        del arities[operator]
        if not arities:
            del operations[arity]

    def __normalizeValues(self, values, domain, numdomain, is_set=False):
        """Normalize a collection of domain elements to numeric elements.

        Arguments:
            values    -- A container of values to normalize.
            domain    -- The symbolic domain.
            numdomain -- The numeric domain.
            is_set    -- If True, `values' is treated as a set, i.e.,
                         all duplicates in `values' will be removed;
                         furthermore, the returned list of values will
                         be sorted.  Otherwise the values will simply be
                         normalized one by one as they occur in the
                         container.

        Returns a list of numeric elements in the domain of the type.
        """
        if is_set:
            result = set()
            add = lambda x: x.add
        else:
            result = []
            add = lambda x: x.append
        for val in values:
            if val in domain:
                add(result)(domain[val])
            elif val in numdomain:
                add(result)(val)
            else:
                raise KeyError(str(val))
        if is_set:
            result = list(result)
            result.sort()
        return result

    def addCoercion(self, fromtype, totype):
        """Add a coercion between two types in the type library.

        Arguments:
            fromtype, totype -- Type names in the type library.  `totype'
                                must be a non-primitive type.

        Raises an abstraction.typehierarchy.ConsistencyError if the new
        coercion would make the type hierarchy associated with the type
        library cyclic.  In this case the type library will remain in the
        state before calling this function.
        """
        if not fromtype in self.typenames:
            raise KeyError(str(fromtype))
        if totype in self.STD_TYPES:
            raise ValueError("cannot define coercions to primitive types")
        fromtypes = self.coercions[totype]
        if fromtype in fromtypes:
            raise ValueError \
                  ("coercion from `%s' to `%s' has already been defined")
        self.type_ordering.orderTypes(fromtype, totype)
        fromtypes[fromtype] = {}

    def defineCoercion(self, fromtype, totype, backend, definition, validate=True):
        """(Re)defines a coercion from a type to another type for a back-end.

        Arguments:
            fromtype, totype -- Type names in the type library.  `totype'
                                must be a non-primitive type.
            backend          -- A name for a back-end.
            definition       -- A list (or a tuple) of two-element tuples
                                containing a string and a container of
                                values from the symbolic or numeric domain
                                of `totype'.
            validate         -- If True, check coercion test strings for
                                consistency for pre-defined back-ends.

        The function adds the definition of a coercion from `fromtype' to
        `totype' for back-end `backend' in the type library.  If the
        specified back-end is not known to the type library, it is
        automatically added to the type library.  Similarly, if no direct
        relationship between the precision of `fromtype' and `totype' has
        previously been introduced in the type library, the type library
        is updated with the information that `totype' is more abstract
        than `fromtype'.  This may cause an
        abstraction.typehierarchy.ConsistencyError if the addition of
        the relationship between `fromtype' and `totype' would make the
        precision ordering between types cyclic (in which case the type
        library remains unchanged).
        """
        type_domain = self.domain[totype]
        type_numdomain = self.numdomain[totype]
        type_coercions = self.coercions[totype]
        if not fromtype in type_coercions:
            self.addCoercion(fromtype, totype)
        if not backend in self.backends:
            self.addBackEnd(backend)
        dfn = []
        for (test, retvals) in definition:
            if not isinstance(test, str) or not hasattr(retvals, "__iter__"):
                raise ValueError("invalid coercion definition")
            if validate:
                self.__validateCoercionTest(test, fromtype, backend)
            dfn.append((test,
                        self.__normalizeValues
                        (retvals, type_domain, type_numdomain, True)))
        type_coercions[fromtype][backend] = dfn

    def undefineCoercion(self, fromtype, totype, backend):
        """Remove the definition of a coercion for a specific back-end.

        Arguments:
            fromtype, totype -- Type names in the type library.
            backend          -- A name for a back-end.

        This function will not alter the precision relation between the
        types in the type hierarchy, i.e., `totype' will remain more
        abstract than `fromtype'.
        """
        del self.coercions[totype][fromtype][backend]

    def removeCoercion(self, fromtype, totype):
        """Remove the definition of a coercion between two types.

        Arguments:
            fromtype, totype -- Type names in the type library.

        This function will remove the coercion defined from `fromtype'
        and `totype' for all back-ends in the type library.  Also the
        precision relation between `fromtype' and `totype' is removed.
        """
        del self.coercions[totype][fromtype]
        self.type_ordering.unorderTypes(fromtype, totype)

    def addBackEnd(self, backend):
        """Introduce a new back-end for coercions in the type library.

        Argument:
            backend -- A name for a back-end.
        """
        if not isinstance(backend, str):
            raise TypeError("expected a back-end name as argument")
        if backend in self.backends:
            raise ValueError \
                  ("back-end `%s' has already been defined" % backend)
        self.backends.add(backend)

    def removeBackEnd(self, backend):
        """Remove a back-end from the type library.

        Argument:
            backend -- A name for a back-end.

        The function will remove a back-end from the back-ends known to the
        type library and all coercion definitions for this back-end.
        """
        if backend in self.STD_BACKENDS:
            raise ValueError("cannot remove a predefined back-end")
        self.backends.remove(backend)
        for (totype, fromtypes) in self.coercions.iteritems():
            for (fromtype, backends) in fromtypes.iteritems():
                if backend in backends:
                    del backends[backend]

    # -------------------------------------------------------------------------
    # Convenience functions for accessing the operations in the type library

    def applyOperator(self, typename, operator, *operands):
        """Compute the result of an operation applied to given operands.

        Arguments:
            typename -- Name of a type in the type library.
            operator -- An operator symbol.
            operands -- Operands for the operator.  Each operand should be
                        one of the following:
                          - a symbolic value in the domain of type
                            `typename',
                          - a numeric value in the domain of type
                            `typename', or
                          - a ValueCollection object representing a set of
                            values in the domain of type `typename'.

        Returns a ValueCollection object which represents all possible
        return values of the operation when applied to the given operands.
        (For each operand of type ValueCollection, the result collects all
        possible return values of the operation for all operand values
        represented by the ValueCollection.)
        """
        return self.__applyOperation \
                   (self.operators[typename][len(operands)][operator],
                    typename, typename, operands)

    def applyTest(self, typename, test, *operands):
        """Compute the result of a test applied to given operands.

        Arguments:
            typename -- Name of a type in the type library.
            test     -- A test symbol.
            operands -- Operands for the test.  Each operand should be one
                        of the following:
                          - a symbolic value in the domain of type
                            `typename',
                          - a numeric value in the domain of type
                            `typename', or
                          - a ValueCollection object representing a set of
                            values in the domain of type `typename'.

        Returns a ValueCollection object which represents all possible
        return values of the test when applied to the given operands.
        (For each operand of type ValueCollection, the result collects all
        possible return values of the test for all operand values
        represented by the ValueCollection.)
        """
        return self.__applyOperation \
                   (self.tests[typename][len(operands)][test], typename,
                    "boolean", operands)

    def __applyOperation(self, op, dtype, rtype, operands):
        """Compute the result of an operation applied to given operands.

        Arguments:
            op       -- Definition of the operation as a dictionary (a
                        mapping from tuples of values in the numeric
                        domain of `dtype' to lists of return values in
                        the domain of `rtype').
            dtype    -- Name of the type of the operation's domain.
            rtype    -- Name of the type of the operation's range.
            operands -- A sequence of operands.  Each operand in the
                        sequence should be one of the following:
                          - a symbolic value in the domain of type
                            `dtype',
                          - a numeric value in the domain of type
                            `dtype', or
                          - a ValueCollection object representing a set
                            of values in the domain of type `dtype'.

        Returns a ValueCollection object which represents all possible
        return values of the operation when applied to the given operands.
        (For each operand of type ValueCollection, the result collects all
        possible return values of the operation for all operand values
        represented by the ValueCollection.)
        """
        arg_domains = []
        for o in operands:
            if isinstance(o, ValueCollection):
                if o.typelibrary != self:
                    raise ValueError("invalid operands")
                if o.typename != dtype:
                    raise ValueError("invalid types for operands")
                arg_domains.append(list(o.values))
            else:
                arg_domains.append \
                    (self.__normalizeValues \
                         ((o,), self.domain[dtype], self.numdomain[dtype]))
        result = set()
        for args in cartesize(*arg_domains):
            result.update(op[tuple(args)])
        return ValueCollection(self, rtype, result)

    def applyCoercion(self, source_type, target_type, operand):
        """Compute a result of a coercion between two types.

        Arguments:
            source_type -- Name of the source type of the coercion.
            target_type -- Name of the target type of the coercion.
            operand     -- A symbolic or a numeric value in the domain of
                           `source_type', or a ValueCollection object
                           representing a set of values in the domain of
                           `source_type'.

        Returns a ValueCollection object which represents all possible
        return values of the coercion when applied to the given operand.
        (If the operand is a ValueCollection, the result collects all
        possible return values of the coercion for all operand values
        represented by the ValueCollection.)

        This method is applicable only between types which have a coercion
        for the "python" back-end defined in the type library.  The method
        raises a RuntimeError if the coercion is not well-defined (i.e.,
        if one of the values represented by `operand' has no corresponding
        return values in the coercion definition).
        """
        coercion = self.coercions[target_type][source_type]["python"]
        if isinstance(operand, ValueCollection):
            if operand.typelibrary != self:
                raise ValueError("invalid operand")
            if operand.typename != source_type:
                raise ValueError("type of operand does not match source type")
            arg_domain = operand.values
        else:
            if not hasattr(operand, "__iter__"):
                arg_domain = (operand,)
            else:
                arg_domain = operand
            if not source_type in self.STD_TYPES:
                arg_domain = \
                    self.__normalizeValues \
                        (arg_domain, self.domain[source_type],
                         self.numdomain[source_type], True)
        result = set()
        for v in arg_domain:
            for (test, retvals) in coercion:
                if not source_type in self.STD_TYPES:
                    for (elem, val) in self.domain[source_type].iteritems():
                        test = test.replace("<" + elem + ">", str(val))
                if eval(test.replace("<>", str(v))):
                    result.update(retvals)
                    break
            else:
                raise RuntimeError("coercion is not well-defined")
        return ValueCollection(self, target_type, result)

    # -------------------------------------------------------------------------
    # Input/output

    def read(self, source="-", sloppy=False):
        """Initialize the type library from type specification files.

        Argument:
            source -- An open file object, a file name, or a sequence of
                      file names.  The special file name "-" stands for
                      standard input.  If a file whose name does not contain
                      any path separators is not found in the current
                      directory, it will be searched for also in the
                      directories listed in the SMUML_ABSTRACT_TYPE_PATH
                      environment variable.  The value of this variable (if
                      set) should be a list of directories separated by
                      colons.
            sloppy -- Accept type libraries having types that miss domain
                      specifications or have only partially defined
                      operations.

        An abstraction.typelibrary.InputFormatError will be raised if the
        input is not syntactically correct, or if the type library is
        inconsistent (e.g., if it defines a cyclic precision ordering
        between types).
        """

        # Delete all previously defined types from the library.
        self.__init__()

        # Operators and tests of the Jumbala action language.
        jmb_operators = {}
        jmb_operators[1] = set(('+', '-', '~'))
        jmb_operators[2] = set(('+', '-', '*', '/', '%', '&', '|', '^',
                                    '<<', '>>', '>>>'))
        jmb_tests = {}
        jmb_tests[1] = set(('!'))
        jmb_tests[2] = set(('&&', '||', '==', '!=', '<', '<=', '>', '>='))

        # Names of types introduced but not defined in the input.
        undefined_typenames = {}

        if isinstance(source, file):
            if source.closed or source.mode.find("r") == -1:
                raise ValueError("file object is not ready for reading")
            files = [source]
        else:
            files = source
            if not hasattr(files, "__iter__"):
                files = [files]
            for f in files:
                if not isinstance(f, str):
                    raise TypeError \
                          ("an open file, a file name, or an iterable of "
                           "names expected as first argument")
        for f in files:
            if isinstance(f, file):
                libfile = f
            else:
                absolute_filenames_to_try = [f]
                if f != "-" and not os.path.split(f)[0]:
                    paths = os.getenv("SMUML_ABSTRACT_TYPE_PATH")
                    if paths:
                        for path in paths.split(":"):
                            absolute_filenames_to_try.append \
                                (path + os.sep + f)
                try:
                    for name in absolute_filenames_to_try:
                        if name == "-":
                            name = "(standard input)"
                            libfile = sys.stdin
                            break
                        elif os.path.isfile(name):
                            libfile = open(name, "r")
                            break
                        else:
                            raise IOError()
                except IOError:
                    raise InputFormatError \
                          ("could not open abstract type specification file "
                           "`%s'" % f)

            waitForEndKeyword = False
            self.reader = _LineReader(libfile, sloppy)

            try:
                try:
                    self.reader.next()
                    while True:
                        (kw, rest) = self.__extractKeyword(self.reader.line)
                        if kw == "type":
                            # parse type name and domain
                            if rest == "":
                                raise InputFormatError("type name expected")
                            if rest.find(':') == -1:
                                raise InputFormatError \
                                    ("`:' expected after type name")
                            (typename, domainspec) = self.__splitAndStrip \
                                                         (rest, ':', 1)
                            if not domainspec:
                                if not sloppy:
                                    raise InputFormatError \
                                          ("domain specification for type "
                                           "`%s' expected" % typename)
                                else:
                                    domain = []
                                    numdomain = None
                            elif domainspec.find(':') != -1:
                                (domainspec, numdomainspec) = \
                                    self.__splitAndStrip(domainspec, ':', 1)
                                domain = self.__splitAndStrip(domainspec, ',')
                                if not numdomainspec:
                                    raise InputFormatError \
                                          ("numeric domain specification for "
                                           "type `%s' expected" % typename)
                                try:
                                    numdomain = [int(v) for v in
                                                 self.__splitAndStrip \
                                                 (numdomainspec, ',')]
                                except Exception, e:
                                    raise InputFormatError \
                                          ("invalid value `%s' for a numeric "
                                           "domain element for type `%s'"
                                           % (v, typename))
                            else:
                                domain = self.__splitAndStrip(domainspec, ',')
                                numdomain = None
                            try:
                                if typename in undefined_typenames:
                                    self.addDomainElements \
                                        (typename, domain, numdomain)
                                    del undefined_typenames[typename]
                                else:
                                    self.addType(typename, domain, numdomain)
                            except Exception, e:
                                raise InputFormatError \
                                      ("error processing type `%s':%s"
                                       % (typename, str(e)))
                            # parse list of operator/test/coercion definitions
                            self.reader.next()
                            while True:
                                (kw, rest) = self.__extractKeyword \
                                             (self.reader.line)
                                if kw in ("operator", "test"):
                                    (arity, operator) = \
                                        self.__extractKeyword(rest)
                                    try:
                                        arity = int(arity)
                                        if arity < 1:
                                            raise ValueError
                                    except ValueError:
                                        raise InputFormatError \
                                              ("positive arity for %s "
                                               "expected after `%s'"
                                               % (kw, kw))
                                    operator.strip()
                                    if operator == "":
                                        raise InputFormatError \
                                              ("%s symbol expected after arity"
                                               % kw)
                                    if kw == "operator":
                                        valid_operators = jmb_operators[arity]
                                        valuedomain = self.domain[typename]
                                        operations = self.operators[typename]
                                    elif kw == "test":
                                        valid_operators = jmb_tests[arity]
                                        valuedomain = self.domain['boolean']
                                        operations = self.tests[typename]
                                    if not operator in valid_operators:
                                        raise InputFormatError \
                                              ("unknown %s symbol of arity %d "
                                               "(`%s')" % (kw, arity,
                                                           operator))
                                    try:
                                        self.__readOperationSpecification \
                                            (typename, operator, arity,
                                             self.domain[typename],
                                             valuedomain, operations)
                                    except _EndOfInnerBlock:
                                        pass
                                elif kw == "coercion":
                                    if not rest:
                                        raise InputFormatError \
                                              ("type name expected after `%s'"
                                               % kw)
                                    (fromtype, backend) = \
                                        self.__extractKeyword(rest)
                                    try:
                                        if not fromtype in self.typenames \
                                                and not fromtype in \
                                                undefined_typenames:
                                            undefined_typenames[fromtype] \
                                                = (libfile.name,
                                                   self.reader.line_num)
                                            self.addType(fromtype)
                                        if not typename in self.coercions \
                                                or not fromtype in \
                                                self.coercions[typename]:
                                            self.addCoercion \
                                                (fromtype, typename)
                                    except Exception, e:
                                        raise InputFormatError(str(e))
                                    if not backend:
                                        raise InputFormatError \
                                              ("back end name expected after "
                                               "`%s'" % fromtype)
                                    self.backends.add(backend)
                                    waitForEndKeyword = True
                                    try:
                                        self.__readCoercionDefinition \
                                            (fromtype, typename, backend)
                                    except _EndOfInnerBlock:
                                        pass
                                    except Exception, e:
                                        raise InputFormatError(str(e))
                                    waitForEndKeyword = False
                                    self.reader.next()
                                else:
                                    break
                        else:
                            raise InputFormatError \
                                  ("unknown keyword (`%s')" % kw)
                except IOError:
                    raise InputFormatError \
                          ("I/O error when reading `%s'" % libfile.name)
                except InputFormatError, e:
                    raise InputFormatError \
                          ("%s:%d:%s" % (libfile.name, self.reader.line_num,
                                         e.reason))
                except _EndOfInput:
                    if waitForEndKeyword:
                        raise InputFormatError \
                              ("%s:unexpected end of input" % libfile.name)
            finally:
                if isinstance(source, file):
                    libfile.close()
                del self.reader

        # Check whether there are still any references to undefined types in
        # case the type library was initialized from an external file.
        if undefined_typenames:
            (tname, (fname, ln)) = tuple(undefined_typenames.iteritems())[0]
            raise InputFormatError \
                  ("%s:%d:reference to an undefined type `%s'"
                   % (fname, ln, tname))

        # Check that every abstract operation and test is completely
        # defined.
        if not sloppy:
            incomplete_operations = self.findIncompleteOperations()
            if incomplete_operations:
                (typename, operation, arity, arguments) = \
                    incomplete_operations[0]
                numdomain = self.numdomain[typename]
                raise InputFormatError \
                      ("definition for operation `%s' of arity %d is "
                       "incomplete for type %s: `%s(%s)' is undefined"
                       % (operation, arity, typename, operation,
                          ", ".join([numdomain[a] for a in arguments])))

        # Check that every coercion is completely defined.
        if not sloppy:
            for (fromtype, totype, backend, n) in \
                    self.findUndefinedCoercions():
                if n != -1:
                    raise InputFormatError \
                          ("coercion defined from `%s' to `%s' for back-end "
                           "`%s' includes a test with no return values",
                           (fromtype, totype, backend))

        # Validate syntax of coercion tests for predefined back-ends.
        self.validateCoercionTests()

    def write(self, target="-", typenames=None, operators=None, tests=None,
              coercions=None):
        """Output the type library in the format of a type specification file.

        Arguments:
            target    -- An open file object or a name of an output file.
                         The special filename "-" corresponds to standard
                         output.
            typenames -- Container of type names identifying the types to
                         include in the output (if None, output the
                         definitions of all types).
            operators -- Container of (arity, symbol) pairs identifying the
                         operators to include in the output (if None, output
                         the definitions of all operators).
            tests     -- Container of (arity, symbol) pairs identifying the
                         tests to include in the output (if None, output
                         the definitions of all tests).
            coercions -- Container of (source type, back-end) pairs
                         identifying the coercions to include in the output
                         (if None, output the definitions of all coercions).

        The operators and tests to be outputted must have been completely
        defined for all pairs of operands in the domain of each type to be
        outputted before calling this function.  Also not that the output
        will be ill-formed if an operator or test (or a coercion) has an
        empty set of admissible return values for any combination of
        operands (for any test defined in the coercion).
        """
        if isinstance(target, file):
            if target.closed or target.mode.find("w") == -1:
                raise ValueError("file object is not ready for writing")
            libfile = target
        else:
            if not isinstance(target, str):
                raise TypeError \
                      ("an open file object or a file name expected as "
                       "argument")
            if target == "-":
                libfile = sys.stdout
            else:
                libfile = open(target, "w")
        first = True
        if typenames == None:
            typenames = self.typenames
        for typename in typenames:
            if typename in self.STD_TYPES:
                continue
            if not first:
                libfile.write("\n")
            first = False
            numdomain = self.numdomain[typename]
            numdomain_keys = sorted(numdomain.keys())
            libfile.write("type %s :" % typename)
            if numdomain:
                libfile.write \
                    (" %s : %s"
                     % (", ".join([numdomain[v] for v in numdomain_keys]),
                        ", ".join([str(v) for v in numdomain_keys])))
            libfile.write("\n")
            operations = []
            if typename in self.operators:
                operations.append \
                    (("operator", self.operators[typename], operators,
                      numdomain))
            if typename in self.tests:
                operations.append \
                    (("test", self.tests[typename], tests,
                      {True : 'true', False : 'false'}))
            for (optype, defined_ops, included_ops, rng) in operations:
                for (arity, ops) in defined_ops.iteritems():
                    for (op, dfn) in ops.iteritems():
                        if included_ops != None \
                           and not (arity, op) in included_ops:
                            continue
                        libfile.write("  %s %d %s\n" % (optype, arity, op))
                        for args in cartesize(*(arity * [numdomain_keys])):
                            retvals = [rng[v] for v in dfn[tuple(args)]]
                            libfile.write \
                                ("    %s : %s\n"
                                 % (", ".join([numdomain[a] for a in args]),
                                    ((retvals and ", ".join(retvals)) or
                                     "# no admissible return values")))
            if typename in self.coercions:
                for (t, backends) in self.coercions[typename].iteritems():
                    for (backend, dfn) in backends.iteritems():
                        if coercions != None and not (t, backend) in coercions:
                            continue
                        libfile.write("  coercion %s %s\n" % (t, backend))
                        for (test, vals) in dfn:
                            libfile.write \
                                ("    %s : %s\n"
                                 % (test,
                                    ((vals
                                      and
                                      ", ".join([numdomain[v] for v in vals]))
                                     or
                                     "# no admissible return values")))
                        libfile.write("    end\n")
        if not isinstance(target, file) and target != "-":
            libfile.close()

    # -------------------------------------------------------------------------
    # Validation

    def __validateDomainValueLists(symvals, numvals):
        """Check consistency of symbolic and numeric domain specifications.

        Symbolic and (corresponding) numeric domain value specifications
        for a type are consistent if all of the folllowing conditions hold:
            - the domain specifications have the same length,
            - the symbolic domain specification is a sequence of valid
              domain value strings with no duplicates
            - the numeric domain value specification is a sequence of
              integers with no duplicates.

        Arguments:
            symvals  -- Sequence of symbolic domain values.
            numvals  -- Sequence of corresponding numeric domain values.
        """
        seen_sym_elems = set()
        seen_num_elems = set()
        if len(symvals) != len(numvals):
            raise ValueError \
                      ("lengths of symbolic and numeric domains must match")
        for (symval, numval) in zip(symvals, numvals):
            if not isinstance(symval, str):
                raise TypeError("type mismatch in symbolic domain value")
            if TypeLibrary._TypeLibrary__domainvalue_re.match(symval) == None:
                raise ValueError("`%s' is not a valid domain value" % symval)
            if not isinstance(numval, int):
                raise TypeError("type mismatch in numeric domain value")
            if symval in seen_sym_elems:
                raise ValueError \
                          ("duplicate value `%s' in symbolic domain "
                           "specification" % symval)
            if numval in seen_num_elems:
                raise ValueError \
                          ("duplicate value %d in numeric domain "
                           "specification" % numval)
            seen_sym_elems.add(symval)
            seen_num_elems.add(numval)

    __validateDomainValueLists = staticmethod(__validateDomainValueLists)

    # Callbacks for resolving and validating the type of an action language
    # expression template.

    def __createTypeResolver():
        """Build a function object for resolving the type of an expression."""

        def __error(expr):
            raise InputFormatError \
                      ("subexpression `%s' is not type-consistent"
                       % str(expr).replace("x", "<>"))

        def __bool(expr, subexprs):
            for s in subexprs:
                if s != "boolean":
                    __error(expr)
            return "boolean"

        def __int(expr, subexprs):
            for s in subexprs:
                if s != "int":
                    __error(expr)
            return "int"

        def __infix(expr, subexprs):
            if isinstance(expr.api_getOperator(),
                          (actlang.binaryop.EqualityOperator,
                           actlang.binaryop.NumericalComparison)):
                if subexprs[0] != "int" or subexprs[1] != "int":
                    __error(expr)
                return "boolean"
            elif isinstance(expr.api_getOperator(),
                            actlang.binaryop.NumericalBooleanOperator) \
                            and subexprs[0] == subexprs[1] == "boolean":
                return "boolean"
            return __int(expr, subexprs)

        def __ternaryCond(expr, subexprs):
            if subexprs[0] != "boolean" or subexprs[1] != subexprs[2]:
                __error(expr)
            return subexprs[1]

        def __simpleName(expr, subexprs):
            if str(expr) != "x":
                raise InputFormatError \
                      ("no references to variables allowed in expression "
                       "templates")
            return "int"

        resolver = \
            ExpressionTraverser.ExpressionTraverser(ExpressionTraverser.ignore)
        callbacks = (
            ((actlang.expr.BoolLiteralExpr, actlang.expr.CondAndExpr,
              actlang.expr.CondOrExpr, actlang.expr.LogicalNotExpr),
             __bool),
            ((actlang.expr.BitNotExpr, actlang.expr.UnaryPlusExpr,
              actlang.expr.UnaryMinusExpr, actlang.expr.IntLiteralExpr),
             __int),
            (actlang.expr.InfixExpr, __infix),
            (actlang.expr.ParenExpr, lambda expr, subexprs: subexprs[0]),
            (actlang.expr.TernaryCondExpr, __ternaryCond),
            (actlang.names.SimpleName, __simpleName))
        for (cls, f) in callbacks:
            resolver.register_post_callback(cls, f)
        return resolver

    # Function object for resolving the type (boolean or int) of an action
    # language expression template and validating the type consistency of the
    # template (i.e., checking that the types are not mixed inappropriately,
    # as, for instance, in the otherwise syntactically valid expression
    # "! (3 + (2 > 0))").
    #
    # The function object accepts an actlang.expr.Expr object as a parameter
    # and returns the type of the expression object as a string ("int" or
    # "boolean").
    #
    # An InputFormatError will be raised if the type structure of the
    # expression is not consistent as described above.
    #
    # All subexpressions of type actlang.names.SimpleName contained in the
    # expression template should correspond to the name "x" representing the
    # parameter in the expression template (no references to variables are
    # allowed in expression templates).  If this is not the case, an
    # InputFormatError will be raised.

    __resolveTypeOfJumbalaExpressionTemplate = __createTypeResolver()
    del __createTypeResolver

    def parseJumbalaCoercionTestTemplate(test):
        """Validate a coercion test template in the action language.

        Argument:
            test -- A string representing a Boolean expression template.
                    A valid Boolean expression template is a semantically
                    valid Boolean expression in the Jumbala action
                    language with no side-effects, and no references to
                    variables except for the template parameter
                    represented by the string literal "<>".  Because there
                    are no implicit automatic type conversions defined
                    between primitive types in the Jumbala action language,
                    the template expression must also be type-consistent,
                    i.e., syntactically correct expressions such as
                    "! (3 + (2 > 0))" are invalid.

        Raises an InputFormatError if `test' does not represent a valid
        Boolean expression template in the Jumbala action language.

        On success, returns an actlang.expr.Expr object representing the
        action language expression which corresponds to the template with
        all occurrences of the template parameter replaced with references
        to the variable "x".
        """
        if not isinstance(test, str):
            raise TypeError("string expected as argument")
        code = "assert(%s);" % test.replace("x", "").replace("<>", "x")
        try:
            try:
                e = actlang.grammar.parse(code).api_getTopStatements()[0] \
                    .api_getExpr()
            except actlang.errors.Error:
                raise InputFormatError("")
            if TypeLibrary. \
                    _TypeLibrary__resolveTypeOfJumbalaExpressionTemplate(e) \
                    != "boolean":
                raise InputFormatError("")
        except ExpressionTraverser.UnregisteredCallbackError:
            raise InputFormatError \
                  ("`%s' is not a valid Boolean expression template in the "
                   "Jumbala action language" % test)
        except InputFormatError, e:
            raise InputFormatError \
                  ("`%s' is not a valid Boolean expression template in the "
                   "Jumbala action language%s"
                   % (test, str(e) and (" [%s]" % str(e)) or ""))
        return e

    parseJumbalaCoercionTestTemplate \
        = staticmethod(parseJumbalaCoercionTestTemplate)

    def findIncompleteOperations(self):
        """Find incompletely defined operations in the library.

        Returns a list of quadruples of the form
            (typename, symbol, arity, arguments)
        where
            typename   is a type name,
            symbol     is an operator/test symbol,
            arity      is the arity of the operator, and
            arguments  is a tuple of arguments (`arity' elements from the
                       numeric domain of the type) for which the result of
                       the operation is either undefined, or is not a
                       nonempty collection of elements from the domain of
                       the type.
        """
        result = []
        for typename in self.typenames:
            if typename in self.STD_TYPES:
                continue
            d = self.numdomain[typename]
            for operations in (self.operators[typename], self.tests[typename]):
                for (arity, ops) in operations.iteritems():
                    for (op, dfn) in ops.iteritems():
                        for args in cartesize(*(arity * [d.keys()])):
                            args = tuple(args)
                            if not args in dfn or not dfn[args]:
                                result.append((typename, op, arity, args))
        return result

    def findUndefinedCoercions(self):
        """Find undefined coercions in the type library.

        A type library is allowed to leave some coercions between types
        undefined.  However, a complete type library should define a
        direct coercion from every type to each type which is strictly
        more abstract than the type itself, for each back-end known to
        the type library.  Furthermore, each test in any defined coercion
        should have a non-empty list of return values.

        Returns a list of quadruples of the form
            (fromtype, totype, backend, n)
        where
            fromtype  is a type name,
            totype    is a type name,
            backend   is a name of a back-end, and
            n         determines the type of error in the coercion from
                      `fromtype' to `totype' for back-end `backend'.
                      If -1, then the coercion has not been defined at all;
                      if nonnegative, indicates the number of a test in the
                      coercion (from `fromtype' to `totype' for back-end
                      `backend', counting the tests from zero) with an
                      empty list of return values.
        """
        result = []
        for totype in self.typenames:
            if totype in self.STD_TYPES:
                continue
            less_abstract_types = self.type_ordering.lessAbstractTypes(totype)
            for fromtype in less_abstract_types:
                for backend in self.backends:
                    try:
                        for (i, (test, retvals)) in \
                                enumerate(self.coercions
                                          [totype][fromtype][backend]):
                            if not retvals:
                                result.append((fromtype, totype, backend, i))
                    except KeyError:
                        result.append((fromtype, totype, backend, -1))
        return result

    def validateCoercionTests(self):
        """Validate syntax of coercion tests for predefined back-ends."""
        for (totype, fromtypes) in self.coercions.iteritems():
            for (fromtype, backends) in fromtypes.iteritems():
                for (backend, dfn) in backends.iteritems():
                    if backend in self.STD_BACKENDS:
                        for (test, retvals) in dfn:
                            try:
                                self.__validateCoercionTest \
                                    (test, fromtype, backend)
                            except Exception, e:
                                raise InputFormatError \
                                    (str(e)
                                     + "\n    in coercion from `%s' to `%s' "
                                       "defined for back-end `%s'"
                                     % (fromtype, totype, backend))

    def __validateCoercionTest(self, test, fromtype, backend):
        """Validate a single coercion test for the predefined back-ends.

        Arguments:
            test     -- A coercion test as a string.
            fromtype -- Name of the source type of the coercion.
            backend  -- A back-end name.
        """
        if not backend in self.STD_BACKENDS:
            # do not validate coercions for undefined backends
            return
        instantiated_test = test
        if not fromtype in self.STD_TYPES:
            # replace symbolic values from the domain of "fromtype" with their
            # numeric correspondents
            for val in self.domain[fromtype].iterkeys():
                instantiated_test = instantiated_test.replace \
                                        ("<" + val + ">", "0")
        if backend == "python":
            instantiated_test = instantiated_test.replace("<>", "0")
            # the try-block is used to translate also exceptions from `eval'
            # into ValueErrors
            try:
                if not isinstance \
                        (eval(instantiated_test,
                              {"__builtins__" :
                                   {'True' : True, 'False' : False}}),
                         bool):
                    raise Exception()
            except Exception:
                raise ValueError \
                      ("`%s' is not a valid Boolean expression template in "
                       "Python" % test)
        else: # backend == "Jumbala"
            self.parseJumbalaCoercionTestTemplate(instantiated_test)

    # -------------------------------------------------------------------------
    # Code generation

    def generatePythonClasses(self):
        """Generate Python classes from abstract types in the type library."""

        env = globals()
        env["_domain"] = self.domain
        env["_numdomain"] = self.numdomain
        env["_operators"] = self.operators
        env["_tests"] = self.tests
        env["_coercions"] = self.coercions
        env["_STD_TYPES"] = self.STD_TYPES
        from string import Template
        code = Template(r"""class ${Classname}(AbstractType):
    '''Collection of values of type ${Classname}.

    The class has the following attributes:
        domain    -- Symbolic domain of the type ${Classname}.
                     `${Classname}.domain[val]' gives the corresponding
                     numeric value of the symbolic domain element `val'.
        numdomain -- Numeric domain of the type ${Classname}.
                     `${Classname}.domain[val]' gives the symbolic value of
                     the integer `val' in the numeric domain element `val'.
        operators -- Operators defined for the type ${Classname}.  For an
                     integer n and an operator symbol s,
                     `${Classname}.operators[n][s]' gives the definition
                     of the operator s for n-tuples of values in the
                     numeric domain D of ${Classname}, i.e.,
                     `${Classname}.operators[n][s][(v1,...,vn)]=(r1,...,rm)'
                     for some 1 <= m <= |D|, rj in D for all j.
        tests     -- Tests defined for the type ${Classname}.  For an
                     integer n and an operator symbol s,
                     `${Classname}.tests[n][s]' gives the definition of
                     the test s for n-tuples of values in the numeric
                     domain D of ${Classname}, i.e.,
                     `${Classname}.tests[n][s][(v1,...,vn)] = (b1,...,bm)'
                     for some 1 <= m <= 2, bj in {True, False} for all j.
        coercions -- Coercions defined for the type ${Classname}.  For a
                     type name t, `${Classname}.coercions[t]' gives the
                     definition of the coercion from a value of type t to a
                     value of type ${Classname}.  This definition
                     is a list of tuples with two elements (a test template
                     as a string and a list of numeric domain elements
                     representing the possible outcomes of the coercion if
                     the test evaluates to true).
    '''
    domain = copy.deepcopy(_domain['${Classname}'])
    numdomain = copy.deepcopy(_numdomain['${Classname}'])
    operators = copy.deepcopy(_operators['${Classname}'])
    tests = copy.deepcopy(_tests['${Classname}'])
    coercions = {}

    def initCoercions(coercions):
        for (fromtype, backends) in _coercions['${Classname}'].iteritems():
            if 'python' in backends:
                dfn = []
                coercions[fromtype] = dfn
                for (test, retvals) in backends['python']:
                    if not fromtype in _STD_TYPES:
                        for (elem, val) in _domain[fromtype].iteritems():
                            test = test.replace('<' + elem + '>', str(val))
                    dfn.append((test, list(retvals)))

    initCoercions(coercions)
    del initCoercions

    def __init__(self, v):
        '''Initialize a collection of values from
        ${Classname}.domain.

        Argument:
            v -- An element or a container of elements from the symbolic
                 domain of ${Classname}, or an AbstractType object
                 representing a collection of values of a type that is
                 coercible to ${Classname}.
        '''
        if isinstance(v, str):
            try:
                self.value = frozenset((self.domain[v],))
            except ValueError:
                raise ValueError \
                      ("`%s' is not a valid element of ${Classname}.domain"
                       % v)
        elif hasattr(v, '__iter__'):
            result = set()
            for value in v:
                try:
                    result.add(self.domain[value])
                except ValueError:
                    raise ValueError \
                          ("`%s' is not a valid element of ${Classname}.domain"
                           % value)
            self.value = frozenset(result)
        elif isinstance(v, ${Classname}):
            self.value = frozenset(v.value)
        ${Coercions}
        else:
            raise TypeError \
                  ("cannot coerce %s into ${Classname}" % v.__class__.__name__)
""")#'
        coercion_test = Template(r"""elif isinstance(v, ${Fromtype}) and '${Fromtype}' in self.coercions:
            self.value = AbstractType._AbstractType__initByCoercion \
                             (v, self.coercions['${Fromtype}'])""")
        for t in self.typenames:
            if not t in TypeLibrary.STD_TYPES:
                l = []
                for fromtype in self.coercions[t].iterkeys():
                    l.append(coercion_test.substitute({'Fromtype' : fromtype}))
                c = code.substitute({'Classname' : t,
                                     'Coercions' : "\n        ".join(l)})
                #print c
                exec c in env
        del env["_domain"]
        del env["_numdomain"]
        del env["_operators"]
        del env["_tests"]
        del env["_coercions"]
        del env["_STD_TYPES"]

    # Utility functions for manipulating (lists of) strings.

    def __extractKeyword(s):
        """Extract the leftmost sequence of non-whitespace chars from a string.

        Argument:
            s -- String to process.

        Returns:
            A pair of strings corresponding to the leftmost maximal
            sequence of consecutive non-whitespace characters in s, and
            the remaining characters.
        """

        l = TypeLibrary.__splitAndStrip(s, " ", 1)
        if len(l) == 1:
            return (l[0], "")
        else:
            return (l[0], l[1])

    __extractKeyword = staticmethod(__extractKeyword)

    def __splitAndStrip(strlist, separator, maxsplit=-1):
        """Split a string using split, then strip whitespace from each part.

        Arguments:
            strlist   -- A string to split.
            separator -- Separator to use between parts.
            maxsplit  -- Upper bound for the number of splits (counting from
                         the left).

        Returns:
            A list of strings.
        """

        return [s.strip() for s in strlist.split(separator, maxsplit)]

    __splitAndStrip = staticmethod(__splitAndStrip)

    def __splitAndRstrip(strlist, separator, maxsplit=-1):
        """Split a string using rsplit, then strip whitespace from each part.

        Arguments:
            strlist   -- A string to split.
            separator -- Separator to use between parts.
            maxsplit  -- Upper bound for the number of splits (counting from
                         the right).

        Returns:
            A list of strings.
        """

        return [s.strip() for s in strlist.rsplit(separator, maxsplit)]

    __splitAndRstrip = staticmethod(__splitAndRstrip)

    # Utility functions for parsing the type library file.

    def __readOperationSpecification(self, t, operation, arity, d, r, ops):
        """Read an operation specification from the type library.

        Arguments:
            reader    -- A __LineReader object.
            t         -- Name of type with which the operation is to be
                         associated.
            operation -- Name of the operation.
            arity     -- Arity of the operation.
            d         -- Mapping from symbolic values in the domain of
                         the operation into integers.
            r         -- Mapping from symbolic values in the range of
                         the operation into range values.
            ops       -- Dictionary of operations associated with t.
        """

        try:
            self.__addOperation(t, operation, arity, ops)
        except Exception, e:
            raise InputFormatError(str(e))
        self.reader.next()
        defined = set()
        while True:
            (parms, vals) = self.__parseSpecification \
                                (arity, d.keys(), r.keys())
            if parms in defined:
                raise InputFormatError \
                      ("redefinition of operation `%s' for type `%s' with "
                       "arguments %s" % (operation, t, str(parms)))
            defined.add(parms)
            self.__defineOperation \
                (t, operation, parms, [r[v] for v in vals], ops)
            #print "%s.%s(%s) = %s" % (t, operation, str(parms), str(vals))
            self.reader.next()

    def __parseSpecification(self, numparms, domain, range):
        """Distribute an operation specification line into its components.

        An operation specification line is of the form
            LIST-OF-DOMAIN-VALUES : LIST-OF-RANGE-VALUES
        where LIST-OF-DOMAIN-VALUES and LIST-OF-RANGE-VALUES are comma-
        separated lists of strings representing values from the domain and
        the range of the operation, respectively.

        Arguments:
            numparms -- Number of arguments expected in
                        LIST-OF-DOMAIN-VALUES.
            domain   -- A list of strings representing the domain of the
                        operator.
            range    -- A list of strings representing the range of the
                        operator.

        Returns:
            (parms, vals), where parms and vals are Python lists of the
            elements in LIST-OF-DOMAIN-VALUES and LIST-OF-RANGE-VALUES,
            respectively.
        """

        if self.reader.line.find(':') == -1:
            raise _EndOfInnerBlock()
        l = self.__splitAndStrip(self.reader.line, ':')
        if len(l) > 2:
            raise InputFormatError("syntax error")
        parms = tuple(self.__splitAndStrip(l[0], ','))
        if l[1]:
            vals = tuple(self.__splitAndStrip(l[1], ','))
        else:
            vals = ()
        if len(parms) != numparms:
            raise InputFormatError \
                  ("list of %d values expected before `:'" % numparms)
        if not vals and not self.reader.sloppy:
            raise InputFormatError("list of values expected after `:'")
        self.__validateDomain(parms, domain)
        self.__validateDomain(vals, range)
        return (parms, vals)

    def __readCoercionDefinition(self, fromtype, totype, backend):
        """Read a coercion specification from the type library.

        Arguments:
            fromtype  -- Name of the source type for coercion.
            totype    -- Name of the target type for coercion.
            backend   -- Name of the backend for which the coercion is
                         defined.
        """
        if backend in self.coercions[totype][fromtype]:
            raise InputFormatError \
                  ("redefinition of coercion from `%s' to `%s' for "
                   "backend `%s'" % (fromtype, totype, backend))
        dfn = []
        domain = self.domain[totype]
        self.reader.next()
        while True:
            if self.reader.line == "end":
                try:
                    self.defineCoercion(fromtype, totype, backend, dfn, False)
                except Exception, e:
                    raise InputFormatError(str(e))
                raise _EndOfInnerBlock()
            elif self.reader.line.find(':') == -1:
                raise InputFormatError("syntax error")
            (test, retvals) = self.__splitAndRstrip(self.reader.line, ':', 1)
            if not retvals:
                if not self.reader.sloppy:
                    raise InputFormatError("list of values expected after `:'")
                retvals = []
            else:
                retvals = tuple(self.__splitAndStrip(retvals, ','))
            self.__validateDomain(retvals, domain.keys())
            dfn.append((test, retvals))
            self.reader.next()

    def __validateDomain(values, domain):
        """Check that every value in values belongs to domain."""

        for v in values:
            if not v in domain:
                raise InputFormatError \
                      ("`%s' is not a valid element in domain %s"
                       % (v, str(domain)))

    __validateDomain = staticmethod(__validateDomain)

# -----------------------------------------------------------------------------
# Functions for accessing the operations in abstract types by type name.

def operator(typename, operator, *arguments):
    """Apply an operator defined for a type to arguments.

    Arguments:
        typename  -- Name of a type.
        operator  -- Operator symbol.
        arguments -- Arguments for the operator.
    """
    return eval(typename).operator(operator, *[arg for arg in arguments])

def test(typename, operator, *arguments):
    """Apply a test defined for a type to arguments.

    Arguments:
        typename  -- Name of a type.
        operator  -- Test symbol.
        arguments -- Arguments for the test.
    """
    return eval(typename).test(operator, *[arg for arg in arguments])

def coercion(typename, value):
    """Coerce a value to an object of a type defined in the type library.

    Arguments:
        typename  -- Name of a type.
        value     -- Value to coerce.
    """
    return eval(typename)(value)
