#! /usr/bin/python2.4

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Usage: abstractor.py MODE [OPTION]... FILE
Extract information about abstractable attributes, perform type inference
using given constraints, or generate an abstract model using given
constraints from a UML 1.4 model contained in a project file in XMI 2.0
format.

The first parameter to the script specifies the mode in which the script
is to operate.  The following modes are available:
  -e, --extract              Read a project file FILE containing a UML 1.4
                             model and write information about the model's
                             abstractable class and signal attributes to
                             standard output [default].
  -h, --help                 Display this help and exit.
  -i, --infer                Read a project file FILE containing a UML 1.4
                             model, perform type inference using
                             constraints from the model and standard input,
                             and write a refined set of constraints to
                             standard output.
  -r, --rewrite              Generate an abstract model from the UML 1.4
                             model in a project file FILE using type
                             constraints from the model and standard input.
  --gui                      Start a graphical user interface for
                             abstracting models and editing type
                             libraries.  This option will ignore the FILE
                             given as argument; use the menus in the
                             graphical user interface to open a project
                             file.

The following command line options are supported (console mode only):
  -d, --default-type=TYPENAME
                             (to be used in combination with -i): Specify
                             the type to be assigned to those attributes
                             whose type is not otherwise constrained by
                             the type inference process.  If not given,
                             these attributes will keep their original
                             types from the model.
  -o, --output-file=FILENAME (to be used in combination with -r): Specify
                             the name for a project file in which to write
                             the abstracted model [default:
                             `abstracted.xmi'].
  -l, --expression-flattening-limit=VALUE
                             (to be used in combination with -r): Specify
                             an integer limit for flattening definitions
                             of inlined abstract operations [default: 0,
                             which disables expression flattening].  The
                             value "none" enables maximal flattening.
  -t, --rename-transitions   (to be used in combination with -r): Rename
                             transitions in the model using the action
                             language code which was substituted for
                             their effects when rewriting the model.
  -v, --verbose              Display verbose output about the type
                             inference process.

RESTRICTIONS ON INPUT MODELS
  This script supports only input models which satisfy the following
  constraints:
    * The model should define the primitive Jumbala data types "int" and
      "boolean" as UML Primitives.
    * Attributes and associations:
        - The name of every class, signal, every class and signal
          attribute, and every navigable association end defined in the
          model should be a valid identifier in the Jumbala action
          language.  There should be no duplicate definitions of classes,
          signals, attributes, or association ends by the same name.  (The
          names of classes and signals reside in separate namespaces, as
          do the attributes and association ends defined within each class
          or a signal.)
        - The type of each attribute should be one of the primitive data
          types "int" or "boolean", or a class defined in the model.  Only
          attributes of type "int" are abstractable.
        - Each class attribute may optionally specify an initializer
          (initialValue) expression, which must be a side-effect free
          expression in the Jumbala action language.
        - Each association between classes should be a 1..1 association.
    * State machines:
        - States:
            . Every state machine in the model should have exactly one
              composite state (its "top" state); all other non-
              pseudostates in the state machine should be either simple
              states or final states.
            . All pseudostates of a state machine in the model should be
              either initial or choice pseudostates.
            . Entry and exit actions and doActivities for states are
              ignored.
        - Transitions:
            . Each transition in a state machine should either have no
              trigger, or it should be triggered by a SignalEvent
              associated with a signal defined in the model.
            . The guard of each transition in a state machine should
              either be empty, or a side-effect free Boolean expression in
              the Jumbala action language.
            . The effect of each transition in a state machine should be
              an UninterpretedAction with a (possibly empty) sequence of
              statements in the Jumbala action language as its body.  Each
              statement in this sequence should be one of the following:
                (i)   an "assert" statement with a side-effect free
                      Boolean expression;
                (ii)  an empty statement (";")
                (iii) an assignment statement, whose right-hand side is
                      either an expression of the form "new C()" for a
                      class C defined in the model, or a side-effect free
                      expression; or
                (iv)  a "send" statement referring to a signal defined in
                      the model, with side-effect free parameter and
                      target expressions.
        - Side-effect free action language expressions:
            . Side-effect free action language expressions should be
              type-consistent expressions in the Jumbala action language.
              The expressions can contain any side-effect free arithmetic
              or Boolean operators, or the ternary conditional operator.
            . Every attribute reference in a side-effect free expression
              must be either of the form "x1.x2...xn" for some n >= 1, or
              "this.x1.x2...xn" for some n >= 0.  For n >= 1, the two
              forms are semantically equivalent.  If the reference occurs
              in action language code belonging to (a state machine of)
              class C, the "this" expression refers to the instance of
              the class C itself. For all 1 <= i <= n, "xi" should be the
              name of a navigable (outgoing) end of an association (or,
              alternatively, if i == n, an attribute) defined in the class
              which is reached from class C by tracking the (possibly
              empty) sequence of associations x1->x2->...->x(i-1).

TYPE CONSTRAINTS (console mode only)
  Type constraints for the -i and -r modes of operation can be specified
  by feeding the script (via standard input) a list of names of type
  specification files (to define the library of abstract types to be used
  for type inference and abstract model construction), together with a
  list of associations between UUID's of attributes in the UML model and
  abstract types defined in the type specification files.  Use the -e mode
  of operation to obtain an initial list of constraints.

  More specifically, the input to the script (in modes -i and -r) should
  consist of zero or more lines of the form
      import <filename>
  followed by zero or more lines of the form
      <uuid> <constraint-operator> <type-name>
  where
      <filename>            is the path name of a type specification file
                            to be used for building the abstract type
                            library (if the file name begins or ends with
                            white space, enclose it in single or double
                            quotes),
      <uuid>                is the UUID of a class or signal attribute in
                            the UML model,
      <constraint-operator> is one of the operators "==", "<=", or ">="
                            (excluding the quotes) separated from <uuid>
                            by white space, and
      <type-name>           is the name of an abstract type defined in one
                            of the imported abstract type specification
                            files.

  An abstract type specification file whose name does not contain any path
  separators will be searched for, in addition to the current working
  directory, in the colon-separated list of directories given by the value
  of the SMUML_ABSTRACT_TYPE_PATH environment variable.

  The constraint operators (for an attribute with UUID `id' and a type
  name `t', respectively) have the following semantics:
      id == t : type of attribute with UUID `id' must be equal to `t'
      id <= t : type of attribute with UUID `id' may not be more abstract
                than `t' (but is allowed to be any type that is at least
                as precise as `t', including the original type of the
                attribute)
      id >= t : type of attribute with UUID `id' must be at least as
                abstract as `t'
  If no type constraint is specified for the UUID of an attribute, the
  corresponding attribute will be left unconstrained.  In this case the
  type inference procedure will be allowed to assign the attribute any
  type that is at least as abstract as its original type when searching
  for a type assignment which satisfies all constraints.  For attributes
  with no type constraints generated in the type inference process, this
  type will be either the default type for unconstrained attributes (if
  specified using -d), or the original type of the attribute.

  The part following the first `#' of every input line (if present) is
  ignored.

This script was written by Heikki Tauriainen for the SMUML project
<http://www.tcs.hut.fi/Research/Logic/SMUML.shtml>.
"""

import sys
import getopt
import coral
from abstraction import error
from abstraction import typeinference

UML = coral.lang.UML14

class Modes:
    """Operating modes of the script."""
    (EXTRACT, INFER, REWRITE) = range(0, 3)

# -----------------------------------------------------------------------------
# The script

# Mode of operation (default: extract type information from a model)
mode = Modes.EXTRACT

# Default type to be associated with unconstrained attributes
default_type = ""

# Default name of the output file
outputfile = "abstracted.xmi"

# Whether to rename transitions with the action language code.
rename_transitions = False

# Flattening limit for abstract operations
expression_flattening_limit = 0

# For verbose output about type inference
printer = None

try:
    options, args = \
        getopt.getopt \
            (sys.argv[1:], 'd:ehil:o:rtv',
             ("default-type=", "expression-flattening-limit=", "extract",
              "gui", "help", "infer", "output-file=", "rewrite",
              "rename-transitions", "verbose"))
except getopt.GetoptError, e:
    error.abort(str(e), 2)

for (option, value) in options:
    if option in ("-d", "--default-type"):
        default_type = value
    elif option in ("-e", "--extract"):
        mode = Modes.EXTRACT
    elif option == "--gui":
        import gtk
        from gui import gui
        gui.Gui()
        gtk.main()
        sys.exit(0)
    elif option in ("-h", "--help"):
        sys.stdout.write(__doc__)
        sys.exit(0)
    elif option in ("-i", "--infer"):
        mode = Modes.INFER
    elif option in ("-l", "--expression-flattening-limit"):
        if value.lower() == "none":
            expression_flattening_limit = -1
        else:
            try:
                expression_flattening_limit = int(value)
                if expression_flattening_limit < 0:
                    raise ValueError()
            except ValueError:
                error.abort \
                    ("the value given for %s must be a nonnegative integer"
                     % option,
                     2)
    elif option in ("-o", "--output-file"):
        outputfile = value
    elif option in ("-r", "--rewrite"):
        mode = Modes.REWRITE
    elif option in ("-t", "--rename-transitions"):
        rename_transitions = True
    elif option in ("-v", "--verbose"):
        printer = lambda record: sys.stderr.write(str(record) + "\n")

if len(args) == 0:
    error.abort("model name expected", 2)
elif len(args) > 1:
    error.abort("extra arguments on command line", 2)

try:
    project = coral.loadDocument(args[0])
except Exception, e:
    error.abort(str(e))

model = None
for x in project.contents():
    if x.oclIsKindOf(UML.Model):
        if not model:
            model = x
        else:
            error.abort("multiple model definitions in input file")
if not model:
    error.abort("no model found in input file")

try:
    s = typeinference.TypeInferenceSystem(model)
    if mode == Modes.EXTRACT:
        for n in s.findAbstractableNodes():
            n.inferred_type = n.model_type
        s.writeTypeAssociations()
        
    elif mode == Modes.INFER or mode == Modes.REWRITE:
        type_library_files = s.initExternalConstraintsFromFile()
        s.solveConstraints(default_type, printer)
        if mode == Modes.REWRITE:
            s.rewriteModel(expression_flattening_limit, rename_transitions)
            s.attachTypesToModel(type_library_files)
            project.setFilename(outputfile)
            try:
                coral.saveDocument(project)
            except Exception, e:
                raise error.Error(str(e))
        s.writeTypeAssociations("-", type_library_files)

except error.Error, e:
    msg = str(e)
    if mode == Modes.REWRITE:
        msg += "\n    no abstract model generated"
    error.abort(msg)
