#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# __init__.py
#
# 2007-06-11  Jori Dubrovin:  Added documentation
# 2006-02-02  Jori Dubrovin:  Added documentation
# 2005-11-11  Jori Dubrovin:  Added toString() to enum types
# 2005-08-26  Jori Dubrovin:  Updating documentation
# 2005-08-09  Jori Dubrovin:  Added some documentation
# 2005-08-08  Jori Dubrovin:  Initial version
#

r"""The action language package.



Usage
=====


Simple Usage
------------

import actlang

# Obtain a static environment and a virtual machine.
env = actlang.basetypes.getInitialEnvironment()
vm = actlang.basetypes.getInitialMachine()
# Make an input program.
prog = 'System.out.println(12 + 7);'
# Parse the program into a syntax tree.
tree = actlang.grammar.parse(prog)
# Translate the syntax tree into internal code.
main = tree.translate(env)
# Load the code into the virtual machine.
vm.initialize(main)
# Run the top-level statements.
vm.run()


Incremental Usage with Error Handling
-------------------------------------

import actlang

# Obtain a static environment and a virtual machine.
env = actlang.basetypes.getInitialEnvironment()
vm = actlang.basetypes.getInitialMachine()
while True:
    # Make an incremental part of the input program.
    prog = ...
    try:
        vm.initialize(actlang.grammar.parse(prog).translate(env))
    except actlang.errors.StaticError, e:
        # Static translation failed.  The environment and the virtual
        # machine are guaranteed to be in the same state as before
        # this iteration.
        print e
        continue
    try:
        vm.run()
    except actlang.errors.RuntimeError_, e:
        # A run-time error occurred.  The virtual machine may be in an
        # unrecoverable state.
        print e
        break



Selected Features of the Language and the Implementation
========================================================


Classes
-------

- top-level classes:
    class C { }

- nested classes:
    class C {
      class D { }
    }
    C.D d = new C.D();

- abstract classes:
    abstract class Shape {
      abstract void draw(int x, int y);
    }

- final classes:
    final class C { }  // cannot be extended

- inheritance (base class must be completely defined first):
    class Base { }
    class Derived extends Base { }

- constructors:
    class Base {
      Base(int x) { }
    }
    class Derived extends Base {
      Derived(int x, int y) {
        super(x);
        this.y = y;
      }
      Derived() {
        super(12);
      }
      int y;
    }


Interfaces
----------

- top-level and nested interfaces:
    interface I {
      void draw();
    }
    class C {
      interface J { }
    }
    I var = null;

- interface inheritance:
    interface I { }
    interface J { }
    interface K extends I, J { }

- class inheritance:
    interface I { void turnLeft(); }
    interface J { void turnRight(); }
    abstract class V implements I, J {
      void turnLeft() { angle++; }
    }


Enums
-----

- top-level and nested enums:
    enum Color { yellow, mauve }
    class C {
      enum Flavor { strawberry, spam }
    }

- common base class Enum:
    enum Color { black }
    Enum c = Color.black;

- the 'values' method:
    enum Color { red, macintoshGrey }
    Color[] vals = Color.values();
    Color grey = vals[1];

- the 'toString' method:
    enum Color { red, macintoshGrey }
    assert(Color.red.toString().equals("red"));


Fields
------

- static variables:
    C.s = C.f + 12;
    class C {
      static int s;
      static final int f = 17;
    }
    interface I {
      int R = 9;    // implicitly static and final
    }

- instance variables:
    class C {
      final int x = 20;
      int i = x + 1;
    }
    int i = new C().i;

- default value is 0, false, or null


Methods
-------

- static methods:
    int i = C.six();
    class C {
      static int six() { return 2 + four(); }
      static int four() { return 4; }
    }

- instance methods:
    class C {
      void increment(final int amount) { x += amount; }
      int x = 14;
    }
    new C().increment(100);

- native methods:
    class C {
      native int draw(int x, int y);
      static native void reset();
    }

- abstract methods:
    interface I {
      I draw(int x, int y);  // implicitly abstract
    }
    abstract class C implements I {
      abstract C draw(int x, int y);
    }

- hiding of static methods with the same name and argument types:
    class Base {
      static int count() { return 13; }
      final static int value() { return 1; }  // cannot be hidden
    }
    class Derived extends Base {
      static int count() { return 100; }
    }
    Base b = new Derived();
    System.out.println(b.count());             // 13
    System.out.println(((Derived)b).count());  // 100

- overriding of instance methods with the same name and argument types:
    class Base {
      int count() { return 13; }
      final int value() { return 1; }  // cannot be overridden
    }
    class Derived extends Base {
      int count() { return 100; }
    }
    Base b = new Derived();
    System.out.println(b.count());             // 100
    System.out.println(((Derived)b).count());  // 100

- return value covariance:
    class Base { Base myself() { return this; } }
    class Derived extends Base { Derived myself() { return this; } }


Objects
-------

- primordial base class for all references:
    class C { }
    interface I { }
    Object obj;
    I x = null;
    obj = new C();
    obj = x;

- cloning:
    class Bad { }
    class Good implements Cloneable { }
    Bad b = (Bad) (new Bad().clone());    // run-time error
    Good g = (Good) (new Good().clone()); // ok

- equality:
    class C { }
    class D { boolean equals(Object obj) { return obj instanceof D; } }
    boolean notTrue = new C().equals(new C());
    boolean isTrue = new D().equals(new D());
    boolean alsoTrue = "foobar".equals("foo" + "bar");

- string conversion:
    class C { }
    Object o = new C();
    String s = o.toString();


Strings
-------

- comparison:
    boolean b1 = ("xy" == "xy");       // false
    boolean b2 = ("xy".equals("xy"));  // true

- length:
    int five = "quark".length();

- character access:
    String s = "quark";
    int[] a = new int[2];
    a[0] = s.charAt(1);  // charAt returns an int.
    a[1] = s.charAt(4);
    s = String(a);       // s is "uk"

- conversion from any type:
    // The static method valueOf of class String converts
    // a value to a string.
    class C { String toString() { return "cee"; } }
    String s;
    s = String.valueOf(-200);           // s is "-200"
    s = String.valueOf(true);           // s is "true"
    s = String.valueOf((Object) null);  // s is "null"
    s = String.valueOf(new C());        // s is "cee"

- printing of any type:
    // System.out.print prints a value to the screen.
    // System.out.println also prints a newline.
    class C { String toString() { return "cee"; } }
    System.out.print(12);
    System.out.println(12);
    System.out.print(false);
    System.out.println(false);
    System.out.print((Object) null);
    System.out.println((Object) null);
    System.out.print(new C());
    System.out.println(new C());


Arrays
------

- creation:
    int[] a = new int[2];              // initial values are zeroes
    boolean[] b = new boolean[2];      // initial values are false
    Cloneable[] c = new Cloneable[2];  // initial values are null
    int[] d = { 10, 15 };              // The last two are equivalent.
    int[] e = new int[] { 10, 15 };

- final field length:
    int[] a = new int[0];
    int zero = a.length;

- inheritance:
    Object[] a = new String[2];

- array store error:
    Object[] a = new String[2];
    a[0] = new Object();        // run-time error

- cloning:
    int[] a = new int[2];
    int[] b = a.clone();
    a[1] = 13;            // does not affect b


Statements
----------

- local variable declaration statement:
    int x, y = 17;
    final int pi = 3;
    Object[] a = new Object[10];
    y = x;  // run-time error, x uninitialized

- empty statement:
    ;

- expression statements:
    x = 4;
    x++;
    --x;
    System.out.println();
    new File();

- send statements:
    send ACK(-1) to terminal;
  is equivalent to the call
    (terminal).$$signal_ACK(-1);

- assert statements:
    assert x > 0;

- if statement:
    if (x < 0)
      y = -x;
    else
      y = x;

- switch statement:
    switch (n) {
    case 1:
      break;
    default:
      System.out.print("s");
    }

- while statement:
    while (i > 0)
      i--;

- do statement:
    do
      i--;
    while (i > 0);

- for statement:
    for (int i = 0; i < 10; i++)
      System.out.println(i);

- break statement:
    while (true)
      break;
    lab: {
      if (i == 3)
        break lab;
      i++;
    }

- continue statement:
    loop:
    for (int j = 0; j < 10; j++) {
      for (int i = 0; i < 10; i++) {
        if (i == 3)
          continue;
        if (i == 5 && j == 6)
          continue loop;
        System.out.println(i);
      }
    }

- return statement:
    return this;


Expressions
-----------

- integer literals (32-bit signed integers):
    -20000
    0xff

- boolean literals:
    false
    true

- string literals:
    ""
    "foo"

- null literal:
    null

- this:
    this

- parenthesized expressions:
    (a + b)

- class instance creation expressions:
    new String("foo")

- array creation expressions (only the first dimension may be specified):
    new int[2]
    new String[12][]
    new int[][] {{100, 200}, {13}}

- field access expressions and names:
    x
    System.out
    values().length
    super.field

- method invocation expressions:
    append(x, y)
    getName().length()
    super.draw()

- array access expressions:
    a[17]

- unary expressions:
    x++
    x--
    ++x
    --x
    +x
    -x
    ~x
    !b

- cast expressions:
    (Object)"foo"

- binary arithmetic expressions:
    x * y
    x / y
    x % y
    x + y
    x - y
    x << y
    x >> y
    x >>> y

- string concatenation expression (both arguments must be Strings):
    String.valueOf(7) + " samurai"

- relational expressions:
    x < y
    x > y
    x <= y
    x >= y

- type comparison expressions:
    obj instanceof String

- equality expressions:
    x == 17
    x != 17
    obj1 == obj2
    this != obj

- bitwise and logical expressions:
    x & 255
    x ^ 255
    x | 255
    b & true
    b ^ true
    b | false

- conditional and, conditional or expressions:
    y != 0 && x / y > 10
    x == 0 || x == 1

- conditional expressions:
    x > 0 ? x : -x

- assignment expressions:
    x = y
    x *= y
    x /= y
    x %= y
    x += y
    x -= y
    x <<= y
    x >>= y
    x >>>= y
    b &= false
    b ^= true
    b |= false


Excluded Java Features
----------------------

- enhanced for statement:
    for (int i : histogram.values())
      total += i;
  workaround, use a basic for statement

- multidimensional array initialization:
    a = new String[12][20]
  workaround, use an explicit for loop:
    a = new String[12][];
    for (int i = 0; i < 12; i++)
      a[i] = new String[20];

"""

import basetypes
import binaryop
import errors
import expr
import grammar
import icode
import lexical
import machine
import mytypes
import names
import result
import statement
import staticenv
import syntaxtree
import typedecl
