#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# basetypes.py
#
# 2005-11-11  Jori Dubrovin:  Added an error message to CloneNotSupportedError
# 2005-08-08  Jori Dubrovin:  Restructuring the interface
# 2005-07-29  Jori Dubrovin:  Initial version
#

"""Definition of builtin types such as Object and String."""


from sys import stdout

from errors import CloneNotSupportedError
from staticenv import StaticEnvironment
from machine import VirtualMachine
from grammar import parse


_initInput = r'''
class Object {

    native Object clone();

    boolean equals(Object obj) { return (obj == this); }

    String toString() { return "<object>"; }
}

interface ObjectInterface {

    Object clone();

    boolean equals(Object obj);

    String toString();
}

interface Cloneable {}

abstract class Enum {}

final class String {

    String() { _chars = new int[0]; }

    String(int[] value) { _chars = value.clone(); }

    String(String original) { _chars = original._chars; }

    int charAt(int index) { return _chars[index]; }

    boolean equals(Object anObject) {
        if (anObject != null && anObject instanceof String) {
            String s = (String) anObject;
            if (s._chars.length == _chars.length) {
                for (int i = 0; i < _chars.length; i++) {
                    if (s._chars[i] != _chars[i])
                        return false;
                }
                return true;
            }
        }
        return false;
    }

    int length() { return _chars.length; }

    int[] getChars() { return _chars.clone(); }

    String concat(String str) {
        int[] newChars = new int[_chars.length + str._chars.length];
        for (int i = 0; i < _chars.length; i++)
            newChars[i] = _chars[i];
        for (int i = 0; i < str._chars.length; i++)
            newChars[_chars.length + i] = str._chars[i];
        return new String(newChars);
    }

    String toString() { return this; }

    static String valueOf(boolean b) { return (b ? "true" : "false"); }

    static String valueOf(int i) {
        if (i == 0)
            return "0";
        if (i == -2147483648)
            return "-2147483648";
        String digits = "0123456789";
        String result = "";
        int n = (i > 0) ? i : -i;
        while (n > 0) {
            int[] d = new int[1];
            d[0] = digits.charAt(n % 10);
            result = new String(d) + result;
            n /= 10;
        }
        if (i < 0)
            result = "-" + result;
        return result;
    }

    static String valueOf(Object obj) {
        if (obj == null)
            return "null";
        else
            return obj.toString();
    }

    static String valueOf(int[] data) { return new String(data); }

    int[] _chars;
}

class System {

    static PrintStream out = new PrintStream();
}

class PrintStream {

    // Write a byte to standard output.
    native void write(int b);

    // Print a boolean.
    void print(boolean b) { print(String.valueOf(b)); }

    // Print an integer.
    void print(int i) { print(String.valueOf(i)); }

    // Print an array of characters.
    void print(int[] s) { print(String.valueOf(s)); }

    // Print an object.
    void print(Object obj) { print(String.valueOf(obj)); }

    // Print a string.
    void print(String s) {
        if (s == null)
            print(String.valueOf(s));
        else
            for (int i = 0; i < s.length(); i++)
                write(s.charAt(i));
    }

    // Print a line separator.
    void println() { print("\n"); }

    // Print a boolean and a line separator.
    void println(boolean b) { print(b); println(); }

    // Print an integer and a line separator.
    void println(int i) { print(i); println(); }

    // Print an array of characters and a line separator.
    void println(int[] s) { print(s); println(); }

    // Print an object and a line separator.
    void println(Object obj) { print(obj); println(); }

    // Print a string and a line separator.
    void println(String s) { print(s); println(); }
}
'''

def _initialize():
    """Initialize base objects.

    The following global objects are used or initialized:

    _initInput: definition of the action language base classes written
    in the language itself.

    _initEnv: a StaticEnvironment that will contain the base declarations.

    _initMain: a code routine (of type Instruction) that initializes
    all static data defined in _initInput.
    """
    global _initInput, _initEnv, _initMain
    tree = parse(_initInput)
    cloneableInterface = None
    def clone(state, this):
        """Clone an Instance."""
        if not this.getType().isSubtypeOf(cloneableInterface):
            raise CloneNotSupportedError('class does not implement Cloneable')
        return this.clone()
    def write(state, this, b):
        """Write a character to standard output."""
        if b >= 0 and b < 256:
            stdout.write(chr(b))
    _initEnv = StaticEnvironment()
    _initEnv.setNativeHooks({'Object.clone': clone,
                             'PrintStream.write#int': write})
    _initMain = tree.translate(_initEnv)
    cloneableInterface = _initEnv.getTopType('Cloneable')
    _initEnv.setNativeHooks({})

def getInitialEnvironment():
    """Return a StaticEnvironment initialized with the base types."""
    global _initEnv
    return _initEnv.copy()

def getInitialMachine():
    """Return a VirtualMachine initialized with the base types."""
    global _initMain
    vm = VirtualMachine()
    vm.initialize(_initMain)
    vm.run()
    return vm

_initialize()
