#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# binaryop.py
#
# 2007-06-10  Jori Dubrovin:  Adding recurseStr API
# 2005-11-11  Jori Dubrovin:  Reflected changes in mytypes.py
# 2005-08-30  Jori Dubrovin:  Fixed GreaterOrEqual
# 2005-08-03  Jori Dubrovin:  Added string concatenation
# 2005-06-14  Jori Dubrovin:  Improved coding style
# 2005-04-26  Jori Dubrovin:  Added error messages
# 2005-04-22  Jori Dubrovin:  Initial version
#

"""Binary operators for the action language."""

from mytypes import intType, boolType, nullType
from result import ValueResult
from syntaxtree import SyntaxNode
from errors import ErrorProxy
from icode import IntMultiply, IntDivide, IntRemainder, IntPlus, IntNegate
from icode import IntShiftLeft, IntSignedShiftRight, IntUnsignedShiftRight
from icode import IntAnd, IntXor, IntOr, LogicalNot, CallDynamic
from icode import IntLessThan, IntLessOrEqual, IntEqual, RefEqual


class BinaryOperator(SyntaxNode):
    """A base class for binary operators."""

    def __init__(self, linespan):
        SyntaxNode.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):  #pragma: no cover
        return '<op>'

    def translate(self, env, leftType, rightType):  #pragma: no cover
        """Perform static checking and generate internal code.

        Raise a StaticError if type checking fails.

        The generated code pops first the rightmost and then the
        leftmost operand from the operand stack and pushes the
        resulting value to the operand stack.

        Param env: the StaticEnvironment.

        Param leftType: ResultType of the left operand.

        Param rightType: ResultType of the right operand.

        Return value: a ValueResult.
        """
        raise NotImplementedError('Pure virtual method')


class Plus(BinaryOperator):
    """The addition or string concatenation operator +."""

    def __init__(self, linespan):
        BinaryOperator.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '+'

    def translate(self, env, leftType, rightType):
        if leftType is intType and rightType is intType:
            return ValueResult(intType, [IntPlus()])
        stringType = env.getTopType('String')
        if leftType is stringType and rightType is stringType:
            concat = stringType.getExactFunctions('concat', [stringType])[0]
            assert concat.getReturnType() is stringType
            code = [CallDynamic(concat.getSignature(), 2)]
            return ValueResult(stringType, code)
        ErrorProxy().finish('cannot apply operator + to ' + str(leftType) +
                            ' and ' + str(rightType),
                            self.linespan())


class NumericalOperator(BinaryOperator):
    """A base class for a binary operator that operates on integer types."""

    def __init__(self, linespan):
        BinaryOperator.__init__(self, linespan)

    def translate(self, env, leftType, rightType):
        if leftType is not intType or rightType is not intType:
            ErrorProxy().finish('cannot apply operator ' + str(self) +
                                ' to ' + str(leftType) + ' and ' +
                                str(rightType), self.linespan())
        return ValueResult(intType, self._makeCode())

    def _makeCode(self):  #pragma: no cover
        """Return internal code that evaluates the operator."""
        raise NotImplementedError('Pure virtual method')


class Multiply(NumericalOperator):
    """The multiplication operator *."""

    def __init__(self, linespan):
        NumericalOperator.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '*'

    def _makeCode(self):
        return [IntMultiply()]


class Divide(NumericalOperator):
    """The division operator /."""

    def __init__(self, linespan):
        NumericalOperator.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '/'

    def _makeCode(self):
        return [IntDivide()]


class Remainder(NumericalOperator):
    """The remainder operator %."""

    def __init__(self, linespan):
        NumericalOperator.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '%'

    def _makeCode(self):
        return [IntRemainder()]


class Minus(NumericalOperator):
    """The substraction operator -."""

    def __init__(self, linespan):
        NumericalOperator.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '-'

    def _makeCode(self):
        return [IntNegate(), IntPlus()]


class ShiftLeft(NumericalOperator):
    """The left shift operator <<."""

    def __init__(self, linespan):
        NumericalOperator.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '<<'

    def _makeCode(self):
        return [IntShiftLeft()]


class SignedShiftRight(NumericalOperator):
    """The signed (arithmetic) right shift operator >>."""

    def __init__(self, linespan):
        NumericalOperator.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '>>'

    def _makeCode(self):
        return [IntSignedShiftRight()]


class UnsignedShiftRight(NumericalOperator):
    """The unsigned (logical) right shift operator >>>."""

    def __init__(self, linespan):
        NumericalOperator.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '>>>'

    def _makeCode(self):
        return [IntUnsignedShiftRight()]


class NumericalBooleanOperator(BinaryOperator):
    """A base class for an operator that expects two integers or booleans."""

    def __init__(self, linespan):
        BinaryOperator.__init__(self, linespan)

    def translate(self, env, leftType, rightType):
        if leftType not in (intType, boolType) or leftType is not rightType:
            ErrorProxy().finish('cannot apply operator ' + str(self) +
                                ' to ' + str(leftType) + ' and ' +
                                str(rightType), self.linespan())
        return ValueResult(leftType, self._makeCode())

    def _makeCode(self):  #pragma: no cover
        """Return internal code that evaluates the operator."""
        raise NotImplementedError('Pure virtual method')


class BitLogicalAnd(NumericalBooleanOperator):
    """The bitwise or logical and operator &."""

    def __init__(self, linespan):
        NumericalBooleanOperator.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '&'

    def _makeCode(self):
        return [IntAnd()]


class BitLogicalXor(NumericalBooleanOperator):
    """The bitwise or logical exclusive or operator ^."""

    def __init__(self, linespan):
        NumericalBooleanOperator.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '^'

    def _makeCode(self):
        return [IntXor()]


class BitLogicalOr(NumericalBooleanOperator):
    """The bitwise or logical inclusive or operator |."""

    def __init__(self, linespan):
        NumericalBooleanOperator.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '|'

    def _makeCode(self):
        return [IntOr()]


class NumericalComparison(BinaryOperator):
    """A base class for an operator that compares integers, returns boolean."""

    def __init__(self, linespan):
        BinaryOperator.__init__(self, linespan)

    def translate(self, env, leftType, rightType):
        if leftType is not intType or rightType is not intType:
            ErrorProxy().finish('cannot compare ' + str(leftType) +
                                ' and ' + str(rightType) +
                                ' with operator ' + str(self), self.linespan())
        return ValueResult(boolType, self._makeCode())

    def _makeCode(self):  #pragma: no cover
        """Return internal code that evaluates the operator."""
        raise NotImplementedError('Pure virtual method')


class LessThan(NumericalComparison):
    """The less than operator <."""

    def __init__(self, linespan):
        NumericalComparison.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '<'

    def _makeCode(self):
        return [IntLessThan()]


class GreaterThan(NumericalComparison):
    """The greater than operator >."""

    def __init__(self, linespan):
        NumericalComparison.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '>'

    def _makeCode(self):
        return [IntLessOrEqual(), LogicalNot()]


class LessOrEqual(NumericalComparison):
    """The less or equal operator <=."""

    def __init__(self, linespan):
        NumericalComparison.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '<='

    def _makeCode(self):
        return [IntLessOrEqual()]


class GreaterOrEqual(NumericalComparison):
    """The greater or equal operator >=."""

    def __init__(self, linespan):
        NumericalComparison.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '>='

    def _makeCode(self):
        return [IntLessThan(), LogicalNot()]


class EqualityOperator(BinaryOperator):
    """A base class for an equality operator == or !=."""

    def __init__(self, linespan):
        BinaryOperator.__init__(self, linespan)

    def translate(self, env, leftType, rightType):
        # Castability is a symmetric relation.
        if not leftType.isCastableTo(rightType):
            ErrorProxy().finish('cannot compare ' + str(leftType) +
                                ' and ' + str(rightType) +
                                ' with operator ' + str(self), self.linespan())
        if leftType in (intType, boolType):
            code = [IntEqual()]
        elif ((leftType is nullType or leftType.isReference()) and
              (rightType is nullType or rightType.isReference())):
            code = [RefEqual()]
        else:
            raise NotImplementedError('unknown type')
        return ValueResult(boolType, code + self._makeSuffixCode())

    def _makeSuffixCode(self):  #pragma: no cover
        """Return internal code that corrects the evaluation.

        The base class EqualityOperator generates code for evaluating
        the operator ==.  A subclass must generate suffix code that,
        if necessary, flips the truth value of the top value in the
        operand stack.
        """
        raise NotImplementedError('Pure virtual method')


class Equal(EqualityOperator):
    """The equal operator ==."""

    def __init__(self, linespan):
        EqualityOperator.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '=='

    def _makeSuffixCode(self):
        return []


class NotEqual(EqualityOperator):
    """The not equal operator !=."""

    def __init__(self, linespan):
        EqualityOperator.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '!='

    def _makeSuffixCode(self):
        return [LogicalNot()]
