#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# errors.py
#
# 2005-08-26  Jori Dubrovin:  Added AssertionError
# 2005-08-03  Jori Dubrovin:  Added CloneNotSupportedError
# 2005-06-15  Jori Dubrovin:  Improved code quality
# 2005-04-26  Jori Dubrovin:  Added UninitializedError
# 2005-04-25  Jori Dubrovin:  Initial version
#

"""Static and runtime errors reported by the action language."""

import exceptions


class Error(exceptions.Exception):
    """A base class for action language errors."""

    pass


class StaticError(Error):
    """A base class for static (not runtime) errors."""

    pass


class IncompletenessError(StaticError):
    """An exception that indicates that the input string ended unexpectedly.

    It is possible that more lines can be added to the input string to
    obtain a valid program.

    Use the ErrorProxy class to raise an instance.
    """

    pass


class FatalStaticError(StaticError):
    """A fatal grammatical or semantic error in the input.
    
    Use the ErrorProxy class to raise an instance.
    """

    pass


class RuntimeError_(Error):
    """A base class for runtime errors."""

    pass


class ZeroDivisionError_(RuntimeError_):
    """Division by zero."""

    pass


class UninitializedError(RuntimeError_):
    """Using an uninitialized variable."""

    pass


class NegativeShiftError(RuntimeError_):
    """Shifting by a negative amount of bits."""

    pass


class MissingReturnError(RuntimeError_):
    """Completion of a non-void method by falling off the method body."""

    pass


class NullReferenceError(RuntimeError_):
    """Trying to access a member of the null value."""

    pass


class NegativeArraySizeError(RuntimeError_):
    """Trying to create an array whose length is less than zero."""

    pass


class ArrayIndexOutOfBoundsError(RuntimeError_):
    """Trying to access an invalid index of an array."""

    pass


class ArrayStoreError(RuntimeError_):
    """Trying to store to an array a value that is not of component type."""

    pass


class ReturnTypeError(RuntimeError_):
    """Native method return a value of a wrong type."""

    pass


class ClassCastError(RuntimeError_):
    """Explicit cast to a noncompatible type."""

    pass


class CloneNotSupportedError(RuntimeError_):
    """Trying to clone an object that does not implement Cloneable."""

    pass


class AssertionError_(RuntimeError_):
    """Failure caught by an assert statement."""

    pass


class ErrorProxy:
    """A class that accumulates static error messages."""

    def __init__(self):
        self.__messages = ''
        self.__fatal = False

    def add(self, message, linespan=None, fatal=True):
        """Add a message about an input error.

        Param message: a string containing the error message (no
        linefeeds).
        Param linespan: a 2-tuple containing the first and last line
        numbers in which the error appears.
        Param fatal: if this is true, the error is fatal and
        raiseIfErrors will raise a FatalStaticError.  Otherwise the
        error is about an incomplete input.
        """
        if fatal:
            self.__fatal = True
        if linespan is not None:
            if linespan[1] > linespan[0]:
                message = ('lines ' + str(linespan[0]) + '..' +
                           str(linespan[1]) + ': ' + message)
            else:
                message = 'line ' + str(linespan[0]) + ': ' + message
        self.__messages += '\n' + message

    def addException(self, error):
        """Add all the messages in a StaticError to the list."""
        if isinstance(error, FatalStaticError):
            self.__fatal = True
        elif not isinstance(error, IncompletenessError):
            raise TypeError('a StaticError expected')
        if not str(error):
            raise ValueError('a nonempty error message expected')
        self.__messages += str(error)

    def raiseIfErrors(self):
        """Raise a StaticError if any errors have been recorded.

        If no error messages have been added, just return.  If add has
        been called with fatal=true or if addException has been called
        with a FatalStaticError as an argument, a new FatalStaticError
        is raised.  Otherwise an IncompletenessError is raised.
        """
        if self.__messages:
            if self.__fatal:
                raise FatalStaticError(self.__messages)
            else:
                raise IncompletenessError(self.__messages)

    def finish(self, message, linespan=None, fatal=True):
        """Add a message and raise a StaticError.

        This is equivalent to calling add, then raiseIfErrors.
        """
        self.add(message, linespan, fatal)
        self.raiseIfErrors()
        assert False
