#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# expr.py
#
# 2007-10-09  Jori Dubrovin:  Optimized string literal creation
# 2007-09-03  Jori Dubrovin:  Fixing an error message
# 2007-06-10  Jori Dubrovin:  Adding recurseStr API
# 2006-09-04  Jori Dubrovin:  Removed an extra argument
# 2006-08-15  Jori Dubrovin:  Made str(stmt) output a valid statement
# 2005-11-14  Jori Dubrovin:  Renamed CondExpr to TernaryCondExpr
# 2005-08-26  Jori Dubrovin:  Adding syntax tree API
# 2005-08-08  Jori Dubrovin:  Writing syntax tree API
# 2005-06-17  Jori Dubrovin:  Simplified label construction
# 2005-04-26  Jori Dubrovin:  Fixed error messages
# 2005-04-22  Jori Dubrovin:  Initial version
#

"""Classes for handling expressions in the action language."""

from syntaxtree import SyntaxNode
from mytypes import intType, boolType, voidType, nullType
from result import ValueResult, VarResult
from errors import ErrorProxy, StaticError
from icode import LoadAuto, StoreAuto, Goto, IntXor, Literal, Insert, Duplicate
from icode import LogicalNot, IfGoto, IntNegate, CallStatic, CallDynamic, Label
from icode import Pop, NewInstance, NegativeArraySize, BadArrayIndex, InitArray
from icode import LoadField, StoreField, CheckNonNull, IntLessThan, RefEqual
from icode import LoadArray, StoreArray, CheckArrayStore, InstanceOf, BadCast
from icode import LoadStatic, StoreStatic, boolValue, null
from icode import Instance


class Expr(SyntaxNode):
    """A base class for expressions.

    Because Names are Exprs, and names are sometimes interpreted as
    types instead of expressions, an Expr instance can also represent
    a type.
    """

    def __init__(self, linespan):
        SyntaxNode.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):  #pragma: no cover
        return '<expr>'

    def translate(self, env):
        """Perform static type checking and generate internal code.

        Raise a StaticError if type checking fails.

        Param env: a StaticEnvironment for the expression.  This must
        not be mutated.

        Return value: a StaticResult that contains the ResultType and
        internal code for this expression.
        """
        raise NotImplementedError('Pure virtual method')

    def resolveAsType(self, env):
        """Resolve the expression as a reference type.

        Raise a StaticError if resolution fails.  If this is not a
        Name, resolution always fails.

        Param env: a StaticEnvironment in which the expression appears.

        Return value: the resolved SingleReferenceType.
        """
        # Default implementation always fails.
        ErrorProxy().finish("'" + str(self) + "' is not a type",
                            self.linespan())


class LeftHandSideExpr(Expr):
    """A base class for left hand side expressions.

    Left hand side expressions are those that can appear on the left
    hand side of an assignment.
    """

    def __init__(self, linespan):
        Expr.__init__(self, linespan)


class StatementExpr(Expr):
    """A base class for expressions that can be used as statements."""

    def __init__(self, linespan):
        Expr.__init__(self, linespan)


class ParenExpr(Expr):
    """An expression in parentheses.

    The only purpose of this class is to prevent '((Type)) x' from
    being accepted as a valid cast expression.
    """

    def __init__(self, innerExpr, linespan):
        """Construct a parenthesized expression.

        Param innerExpr: the Expr in parentheses.

        Param linespan: see SyntaxNode.
        """
        Expr.__init__(self, linespan)
        self.__innerExpr = innerExpr

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return strFunc(self.__innerExpr, *args, **kwargs)

    def api_getInnerExpr(self):
        """Return the contained Expr."""
        return self.__innerExpr

    def translate(self, env):
        return self.__innerExpr.translate(env)


class IntLiteralExpr(Expr):
    """An integer literal."""

    def __init__(self, value, linespan):
        """Construct an integer literal.

        Param value: a 32-bit signed integer.
        Param linespan: see SyntaxNode.
        """
        Expr.__init__(self, linespan)
        self.__value = value

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return intType.valueToString(self.__value)

    def api_getValue(self):
        """Return the value as an integer."""
        return self.__value

    def translate(self, env):
        return ValueResult(intType, [Literal(self.__value)])


class BoolLiteralExpr(Expr):
    """A boolean literal."""

    def __init__(self, value, linespan):
        """Construct a boolean literal.

        Param value: False or True.
        Param linespan: see SyntaxNode.
        """
        Expr.__init__(self, linespan)
        self.__value = value

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return boolType.valueToString(self.__value)

    def api_getValue(self):
        """Return the value as False or True."""
        return self.__value

    def translate(self, env):
        valueAsInt = boolValue[self.__value]
        return ValueResult(boolType, [Literal(valueAsInt)])


class NullExpr(Expr):
    """The null literal."""

    def __init__(self, linespan):
        """Construct a null expression.

        Param linespan: see SyntaxNode.
        """
        Expr.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return 'null'

    def translate(self, env):
        return ValueResult(nullType, [Literal(null)])


class StringLiteralExpr(Expr):
    """A string literal."""

    def __init__(self, string, linespan):
        """Construct a string literal.

        Param string: the contents of the literal as a Python string,
        not including quotation marks.

        Param linespan: see SyntaxNode.
        """
        Expr.__init__(self, linespan)
        self.__string = string

    def api_recurseStr(self, strFunc, *args, **kwargs):
        def encode(c):
            if c == '\\':
                return '\\\\'
            if c == '"':
                return '\\"'
            if c == '\n':
                return '\\n'
            if ord(c) >= 32 and ord(c) < 128:
                return c
            s = '00' + oct(ord(c))
            return '\\' + s[-3:]
        return '"' + ''.join([encode(c) for c in self.__string]) + '"'

    def api_getValue(self):
        """Return the value as a Python string."""
        return self.__string

    def translate(self, env):
        # Make an array of ints and use the constructor String(int[]).
        stringType = env.getTopType('String')
        intArrayType = intType.getArrayType(env)
        lengthSig = intArrayType.getFields('length')[0].getSignature()
        if False:
            constructor = stringType.getExactFunctions('#constructor',
                                                       [intArrayType])[0]
            code = [NewInstance(stringType),
                    Duplicate(),
                    NewInstance(intArrayType),
                    Duplicate(),
                    Literal(len(self.__string)),
                    Duplicate(2),
                    Duplicate(2),
                    StoreField(lengthSig),
                    Literal(intType.getDefaultValue()),
                    InitArray()]
            index = 0
            for char in self.__string:
                code += [Duplicate(),
                         Literal(index),
                         Literal(ord(char)),
                         StoreArray()]
                index += 1
            code += [CallStatic(stringType, constructor.getSignature(), 2),
                     Pop()]
        else:
            # Optimized string creation, relies on immutability of
            # String and the String implementation in basetypes.py.
            charsSig = stringType.getFields('_chars')[0].getSignature()
            contents = Instance(intArrayType)
            contents.setField(lengthSig, len(self.__string))
            contents.initializeComponents(len(self.__string), 0)
            for (index, char) in enumerate(self.__string):
                contents.setComponent(index, ord(char))
            strObj = Instance(stringType)
            strObj.setField(charsSig, contents)
            code = [Literal(strObj)]
        return ValueResult(stringType, code)


class ThisExpr(Expr):
    """The 'this' expression."""

    def __init__(self, linespan):
        """Construct a this expression.

        Param linespan: see SyntaxNode.
        """
        Expr.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return 'this'

    def translate(self, env):
        if not env.isInstanceContext():
            ErrorProxy().finish("cannot access 'this' from the context",
                                self.linespan())
        return ValueResult(env.getEnclosingTypes()[-1], [LoadAuto('this')])


class CastExpr(Expr):
    """A base class for cast expressions."""

    def __init__(self, operand, linespan):
        """Construct a cast expression.

        Param operand: the Expr to be cast.

        Param linespan: see SyntaxNode.
        """
        Expr.__init__(self, linespan)
        self.__operand = operand

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('(' + self._strType(strFunc, *args, **kwargs) + ') (' +
                strFunc(self.__operand, *args, **kwargs) + ')')

    def api_getOperand(self):
        """Return the operand being cast as an Expr."""
        return self.__operand

    def translate(self, env):
        ep = ErrorProxy()
        try:
            targetType = self._getTargetType(env)
        except StaticError, e:
            ep.addException(e)
        try:
            result = self.__operand.translate(env)
        except StaticError, e:
            ep.addException(e)
        ep.raiseIfErrors()
        sourceType = result.getType()
        if not sourceType.isCastableTo(targetType):
            ep.finish('cannot cast ' + str(sourceType) + ' to ' +
                      str(targetType), self.linespan())
        # Because of the properties of the type system and run-time
        # environment, an explicit cast conversion is never needed.
        code = result.getEvalCode()
        if (sourceType.isReference() and
            not sourceType.isSubtypeOf(targetType)):
            okLabel = Label()
            # Make an explicit run-time check.  First check if the
            # reference is null, because InstanceOf instruction needs
            # a real non-null instance.
            code += [Duplicate(),
                     Literal(null),
                     RefEqual(),
                     IfGoto(okLabel),
                     Duplicate(),
                     InstanceOf(targetType),
                     IfGoto(okLabel),
                     BadCast(),
                     okLabel]
        return ValueResult(targetType, code)

    def _strType(self, strFunc, *args, **kwargs):  #pragma: no cover
        """Return a string representation for the type."""
        raise NotImplementedError('Pure virtual method')

    def _getTargetType(self, env):  #pragma: no cover
        """Return the target Type of the cast, or raise a StaticError."""
        raise NotImplementedError('Pure virtual method')


class TypeCastExpr(CastExpr):
    """A cast expression using a TypeSymbol."""

    def __init__(self, typeSymbol, operand, linespan):
        """Construct a cast to a type represented as a TypeSymbol.

        Param typeSymbol: the TypeSymbol representing the target type.

        Param operand: the Expr to be cast.

        Param linespan: see SyntaxNode.
        """
        CastExpr.__init__(self, operand, linespan)
        self.__typeSymbol = typeSymbol

    def api_getTypeSymbol(self):
        """Return the target type as a TypeSymbol."""
        return self.__typeSymbol

    def _strType(self, strFunc, *args, **kwargs):
        return strFunc(self.__typeSymbol, *args, **kwargs)

    def _getTargetType(self, env):
        return self.__typeSymbol.resolve(env)


class NameCastExpr(CastExpr):
    """A cast expression using an Expr (a Name) to represent the type."""

    def __init__(self, typeExpr, operand, linespan):
        """Construct a cast to a type represented as an Expr.

        Param typeExpr: the Expr representing the target type.
        Because of grammar restrictions, typeExpr can be any kind of
        expression.  However, if it is not an instance of Name,
        translation always fails.

        Param operand: the Expr to be cast.

        Param linespan: see SyntaxNode.
        """
        CastExpr.__init__(self, operand, linespan)
        self.__typeExpr = typeExpr

    def api_getTypeExpr(self):
        """Return the target type as an Expr.

        If the return value is not an instance of Name, it does not
        denote a type, hence this cast is not valid.  The syntax still
        allows this.
        """
        return self.__typeExpr

    def _strType(self, strFunc, *args, **kwargs):
        return strFunc(self.__typeExpr, *args, **kwargs)

    def _getTargetType(self, env):
        return self.__typeExpr.resolveAsType(env)


class IncrementOrDecrementExpr(StatementExpr):
    """A base class for pre- or postfix increment or decrement expressions."""

    def __init__(self, operand, operator):
        """Construct an expression object.

        Param operand: the enclosed expression.
        
        Param operator: a binary operator, which must be Plus for
        pre- or postincrement, or Minus for pre- or postdecrement.
        """
        linespan = (min(operator.linespan()[0], operand.linespan()[0]),
                    max(operator.linespan()[1], operand.linespan()[1]))
        StatementExpr.__init__(self, linespan)
        self._operand = operand
        self._operator = operator

    def api_getOperand(self):
        """Return the operand as an Expr."""
        return self._operand

    def api_getOperator(self):
        """Return the BinaryOperator for this expression.

        Return value: an instance of Plus if this is a pre- or
        postincrement expression, or Minus if this is a pre- or
        postdecrement expression.
        """
        return self._operator

    def translate(self, env):
        operandResult = self._operand.translate(env)
        if not operandResult.isVar():
            ErrorProxy().finish("'" + str(self._operand) +
                                "' cannot be assigned to",
                                self.linespan())
        if operandResult.getType() is not intType:
            ErrorProxy().finish('cannot apply operator ' +
                                2*str(self._operator) + ' to ' +
                                str(operandResult.getType()), self.linespan())
        operatorResult = self._operator.translate(env, operandResult.getType(),
                                                   intType)
        iLen = operandResult.getIndirectionLen()
        if self._isPrefix:
            dupBefore = []
            dupAfter = [Insert(iLen + 1)]
        else:
            dupBefore = [Insert(iLen + 1)]
            dupAfter = []
        code = (operandResult.getIndirectionCode() +
                [Duplicate(iLen) for x in xrange(iLen)] +
                operandResult.getLoadCode() +
                dupBefore +
                [Literal(1)] +
                operatorResult.getEvalCode() +
                dupAfter +
                operandResult.getUncheckedStoreCode())
        return ValueResult(intType, code)


class PrefixExpr(IncrementOrDecrementExpr):
    """A preincrement or predecrement expression."""

    def __init__(self, operand, operator):
        """Construct a preincrement or predecrement expression.

        Param operand: the enclosed expression.
        
        Param operator: a binary operator, which must be Plus for
        preincrement or Minus for predecrement.
        """
        IncrementOrDecrementExpr.__init__(self, operand, operator)
        self._isPrefix = True

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return (2*strFunc(self._operator, *args, **kwargs) +
                '(' + strFunc(self._operand, *args, **kwargs) + ')')


class PostfixExpr(IncrementOrDecrementExpr):
    """A postincrement or postdecrement expression."""

    def __init__(self, operand, operator):
        """Construct a postincrement or postdecrement expression.

        Param operand: the enclosed expression.
        
        Param operator: a binary operator, which must be Plus for
        postincrement or Minus for postdecrement.
        """
        IncrementOrDecrementExpr.__init__(self, operand, operator)
        self._isPrefix = False

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('(' + strFunc(self._operand, *args, **kwargs) + ')' +
                2*strFunc(self._operator, *args, **kwargs))


class UnaryPlusExpr(Expr):
    """An expression whose main operator is the unary plus operator +."""

    def __init__(self, operand, linespan):
        """Construct a unary plus expression.

        Param operand: the enclosed expression.
        Param linespan: see SyntaxNode.
        """
        Expr.__init__(self, linespan)
        self.__operand = operand

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '+(' + strFunc(self.__operand, *args, **kwargs) + ')'

    def api_getOperand(self):
        """Return the operand as an Expr."""
        return self.__operand

    def translate(self, env):
        result = self.__operand.translate(env)
        if result.getType() is not intType:
            ErrorProxy().finish('cannot apply unary plus to ' +
                                str(result.getType()), self.linespan())
        return ValueResult(intType, result.getEvalCode())


class UnaryMinusExpr(Expr):
    """An expression whose main operator is the unary minus operator -."""

    def __init__(self, operand, linespan):
        """Construct a unary minus expression.

        Param operand: the enclosed expression.
        Param linespan: see SyntaxNode.
        """
        Expr.__init__(self, linespan)
        self.__operand = operand

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '-(' + strFunc(self.__operand, *args, **kwargs) + ')'

    def api_getOperand(self):
        """Return the operand as an Expr."""
        return self.__operand

    def translate(self, env):
        result = self.__operand.translate(env)
        if result.getType() is not intType:
            ErrorProxy().finish('cannot apply unary minus to ' +
                                str(result.getType()), self.linespan())
        code = result.getEvalCode() + [IntNegate()]
        return ValueResult(intType, code)


class BitNotExpr(Expr):
    """An expression whose main operator is the bitwise not operator ~."""

    def __init__(self, operand, linespan):
        """Construct a bitwise not expression.

        Param operand: the enclosed expression.
        Param linespan: see SyntaxNode.
        """
        Expr.__init__(self, linespan)
        self.__operand = operand

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '~(' + strFunc(self.__operand, *args, **kwargs) + ')'

    def api_getOperand(self):
        """Return the operand as an Expr."""
        return self.__operand

    def translate(self, env):
        result = self.__operand.translate(env)
        if result.getType() is not intType:
            ErrorProxy().finish('cannot apply operator ~ to ' +
                                str(result.getType()), self.linespan())
        code = result.getEvalCode() + [Literal(-1), IntXor()]
        return ValueResult(intType, code)


class LogicalNotExpr(Expr):
    """An expression whose main operator is the logical not operator !."""

    def __init__(self, operand, linespan):
        """Construct a logical not expression.

        Param operand: the enclosed expression.
        Param linespan: see SyntaxNode.
        """
        Expr.__init__(self, linespan)
        self.__operand = operand

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return '!(' + strFunc(self.__operand, *args, **kwargs) + ')'

    def api_getOperand(self):
        """Return the operand as an Expr."""
        return self.__operand

    def translate(self, env):
        result = self.__operand.translate(env)
        if result.getType() is not boolType:
            ErrorProxy().finish('cannot apply operator ! to ' +
                                str(result.getType()), self.linespan())
        code = result.getEvalCode() + [LogicalNot()]
        return ValueResult(boolType, code)


class InstanceOfExpr(Expr):
    """An instanceof expression."""

    def __init__(self, operand, typeSymbol, linespan):
        """Construct an instanceof expression.

        Param operand: the Expr whose type is being tested.

        Param typeSymbol: the TypeSymbol to test against.

        Param linespan: see SyntaxNode.
        """
        Expr.__init__(self, linespan)
        self.__operand = operand
        self.__typeSymbol = typeSymbol

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('(' + strFunc(self.__operand, *args, **kwargs) +
                ') instanceof ' + strFunc(self.__typeSymbol, *args, **kwargs))

    def api_getOperand(self):
        """Return the operand appearing on the left as an Expr."""
        return self.__operand

    def api_getTypeSymbol(self):
        """Return the reference type as a TypeSymbol."""
        return self.__typeSymbol

    def translate(self, env):
        ep = ErrorProxy()
        try:
            result = self.__operand.translate(env)
        except StaticError, e:
            ep.addException(e)
        try:
            targetType = self.__typeSymbol.resolve(env)
        except StaticError, e:
            ep.addException(e)
        ep.raiseIfErrors()
        sourceType = result.getType()
        if not sourceType.isCastableTo(targetType):
            ep.finish(str(sourceType) + ' is never an instance of ' +
                      str(targetType), self.linespan())
        code = result.getEvalCode()
        if sourceType.isSubtypeOf(targetType):
            # The result is true whenever the operand is not null.
            code += [Literal(null),
                     RefEqual(),
                     LogicalNot()]
        else:
            # Make an explicit test.  Null is never instanceof anything.
            falseLabel = Label()
            endLabel = Label()
            code += [Duplicate(),
                     Literal(null),
                     RefEqual(),
                     IfGoto(falseLabel),
                     InstanceOf(targetType),
                     Goto(endLabel),
                     falseLabel,
                     Pop(),
                     Literal(boolValue[False]),
                     endLabel]
        return ValueResult(boolType, code)


class InfixExpr(Expr):
    """A base class for expressions whose main operator is an infix operator.

    Conditional and and conditional or expressions are represented by
    their own classes, not InfixExpr.
    """

    def __init__(self, leftOperand, rightOperand, operator):
        """Construct an infix expression.

        Param leftOperand: the expression to the left of the operator.
        Param rightOperand: the expression to the right of the operator.
        Param operator: a BinaryOperator.
        """
        Expr.__init__(self,
                      (leftOperand.linespan()[0], rightOperand.linespan()[1]))
        self.__leftOperand = leftOperand
        self.__rightOperand = rightOperand
        self.__operator = operator

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('(' + strFunc(self.__leftOperand, *args, **kwargs) + ') ' +
                strFunc(self.__operator, *args, **kwargs) +
                ' (' + strFunc(self.__rightOperand, *args, **kwargs) + ')')

    def api_getLeftOperand(self):
        """Return the left operand as an Expr."""
        return self.__leftOperand

    def api_getRightOperand(self):
        """Return the right operand as an Expr."""
        return self.__rightOperand

    def api_getOperator(self):
        """Return the infix BinaryOperator."""
        return self.__operator

    def translate(self, env):
        ep = ErrorProxy()
        try:
            leftResult = self.__leftOperand.translate(env)
        except StaticError, e:
            ep.addException(e)
        try:
            rightResult = self.__rightOperand.translate(env)
        except StaticError, e:
            ep.addException(e)
        ep.raiseIfErrors()
        operatorResult = self.__operator.translate(env, leftResult.getType(),
                                                   rightResult.getType())
        return ValueResult(operatorResult.getType(),
                           leftResult.getEvalCode() +
                           rightResult.getEvalCode() +
                           operatorResult.getEvalCode())


class TernaryCondExpr(Expr):
    """An expression whose main operator is the conditional ? : operator."""

    def __init__(self, condition, trueExpr, falseExpr):
        """Construct a ternary conditional expression.

        Param condition: an expression representing the Boolean
        condition.

        Param trueExpr: an expression to be evaluated when the
        condition is true.

        Param falseExpr: an expression to be evaluated when the
        condition is false.
        """
        Expr.__init__(self, (condition.linespan()[0],
                             falseExpr.linespan()[1]))
        self.__condition = condition
        self.__trueExpr = trueExpr
        self.__falseExpr = falseExpr

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('(' + strFunc(self.__condition, *args, **kwargs) + ') ? (' +
                strFunc(self.__trueExpr, *args, **kwargs) + ') : (' +
                strFunc(self.__falseExpr, *args, **kwargs) + ')')

    def api_getCondition(self):
        """Return the boolean condition as an Expr."""
        return self.__condition

    def api_getTrueExpr(self):
        """Return the Expr that appears before the colon."""
        return self.__trueExpr

    def api_getFalseExpr(self):
        """Return the Expr that appears after the colon."""
        return self.__falseExpr

    def translate(self, env):
        ep = ErrorProxy()
        try:
            condResult = self.__condition.translate(env)
            if condResult.getType() is not boolType:
                ep.add('left side of operator ?: must be boolean, not ' +
                       str(condResult.getType()), self.linespan())
        except StaticError, e:
            ep.addException(e)
        try:
            trueResult = self.__trueExpr.translate(env)
        except StaticError, e:
            ep.addException(e)
        try:
            falseResult = self.__falseExpr.translate(env)
        except StaticError, e:
            ep.addException(e)
        ep.raiseIfErrors()
        trueType = trueResult.getType()
        falseType = falseResult.getType()
        if trueType is voidType:
            ep.finish('void result not allowed in conditional expression',
                      self.__trueExpr.linespan())
        if falseType is voidType:
            ep.finish('void result not allowed in conditional expression',
                      self.__falseExpr.linespan())
        if trueType.isSubtypeOf(falseType):
            genericType = falseType
        elif falseType.isSubtypeOf(trueType):
            genericType = trueType
        else:
            ep.finish('incompatible operand types ' + str(trueType) +
                      ' and ' + str(falseType) + ' in conditional expression',
                      self.linespan())
        trueLabel = Label()
        endLabel = Label()
        code = (condResult.getEvalCode() +
                [IfGoto(trueLabel)] +
                falseResult.getEvalCode() +
                [Goto(endLabel)] +
                [trueLabel] +
                trueResult.getEvalCode() +
                [endLabel])
        return ValueResult(genericType, code)


class CondAndExpr(Expr):
    """An expression whose main operator is the conditional and operator &&."""

    def __init__(self, leftOperand, rightOperand):
        """Construct a conditional and expression.

        Param leftOperand: the expression to the left of the operator.
        Param rightOperand: the expression to the right of the operator.
        """
        Expr.__init__(self,
                      (leftOperand.linespan()[0], rightOperand.linespan()[1]))
        self.__leftOperand = leftOperand
        self.__rightOperand = rightOperand

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('(' + strFunc(self.__leftOperand, *args, **kwargs) + ') && (' +
                strFunc(self.__rightOperand, *args, **kwargs) + ')')

    def api_getLeftOperand(self):
        """Return the left operand as an Expr."""
        return self.__leftOperand

    def api_getRightOperand(self):
        """Return the right operand as an Expr."""
        return self.__rightOperand

    def translate(self, env):
        ep = ErrorProxy()
        try:
            leftResult = self.__leftOperand.translate(env)
            if leftResult.getType() is not boolType:
                ep.add('left side of operator && must be boolean, ' +
                       'not ' + str(leftResult.getType()), self.linespan())
        except StaticError, e:
            ep.addException(e)
        try:
            rightResult = self.__rightOperand.translate(env)
            if rightResult.getType() is not boolType:
                ep.add('right side of operator && must be boolean, ' +
                       'not ' + str(rightResult.getType()), self.linespan())
        except StaticError, e:
            ep.addException(e)
        ep.raiseIfErrors()
        trueLabel = Label()
        endLabel = Label()
        code = (leftResult.getEvalCode() +
                [IfGoto(trueLabel)] +
                [Literal(boolValue[False])] +
                [Goto(endLabel)] +
                [trueLabel] +
                rightResult.getEvalCode() +
                [endLabel])
        return ValueResult(boolType, code)


class CondOrExpr(Expr):
    """An expression whose main operator is the conditional or operator ||."""

    def __init__(self, leftOperand, rightOperand):
        """Construct a conditional or expression.

        Param leftOperand: the expression to the left of the operator.
        Param rightOperand: the expression to the right of the operator.
        """
        Expr.__init__(self,
                      (leftOperand.linespan()[0], rightOperand.linespan()[1]))
        self.__leftOperand = leftOperand
        self.__rightOperand = rightOperand

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('(' + strFunc(self.__leftOperand, *args, **kwargs) + ') || (' +
                strFunc(self.__rightOperand, *args, **kwargs) + ')')

    def api_getLeftOperand(self):
        """Return the left operand as an Expr."""
        return self.__leftOperand

    def api_getRightOperand(self):
        """Return the right operand as an Expr."""
        return self.__rightOperand

    def translate(self, env):
        ep = ErrorProxy()
        try:
            leftResult = self.__leftOperand.translate(env)
            if leftResult.getType() is not boolType:
                ep.add('left side of operator || must be boolean, ' +
                       'not ' + str(leftResult.getType()), self.linespan())
        except StaticError, e:
            ep.addException(e)
        try:
            rightResult = self.__rightOperand.translate(env)
            if rightResult.getType() is not boolType:
                ep.add('right side of operator || must be boolean, ' +
                       'not ' + str(rightResult.getType()), self.linespan())
        except StaticError, e:
            ep.addException(e)
        ep.raiseIfErrors()
        trueLabel = Label()
        endLabel = Label()
        code = (leftResult.getEvalCode() +
                [IfGoto(trueLabel)] +
                rightResult.getEvalCode() +
                [Goto(endLabel)] +
                [trueLabel] +
                [Literal(boolValue[True])] +
                [endLabel])
        return ValueResult(boolType, code)


class SimpleAssignmentExpr(StatementExpr):
    """A simple assignment expression."""

    def __init__(self, lhsExpr, rhsExpr, linespan):
        """Construct an assignment expression.

        Param lhsExpr: a LeftHandSideExpr representing the variable to
        be assigned to.

        Param rhsExpr: an Expr that represents the right hand side.

        Param linespan: see SyntaxNode.
        """
        StatementExpr.__init__(self, linespan)
        self.__lhsExpr = lhsExpr
        self.__rhsExpr = rhsExpr

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return (strFunc(self.__lhsExpr, *args, **kwargs) + ' = (' +
                strFunc(self.__rhsExpr, *args, **kwargs) + ')')

    def api_getLhsExpr(self):
        """Return the left hand side as a LeftHandSideExpr."""
        return self.__lhsExpr

    def api_getRhsExpr(self):
        """Return the right hand side as an Expr."""
        return self.__rhsExpr

    def translate(self, env):
        ep = ErrorProxy()
        try:
            lhsResult = self.__lhsExpr.translate(env)
            if not lhsResult.isVar():
                ep.add("'" + str(self.__lhsExpr) + "' cannot be assigned to",
                       self.linespan())
        except StaticError, e:
            ep.addException(e)
        try:
            rhsResult = self.__rhsExpr.translate(env)
        except StaticError, e:
            ep.addException(e)
        ep.raiseIfErrors()
        if not rhsResult.getType().isSubtypeOf(lhsResult.getType()):
            ep.finish("cannot assign " + str(rhsResult.getType()) +
                      " to '" + str(self.__lhsExpr) + "'", self.linespan())
        code = (lhsResult.getIndirectionCode() +
                rhsResult.getEvalCode() +
                [Insert(lhsResult.getIndirectionLen() + 1)] +
                lhsResult.getCheckedStoreCode())
        return ValueResult(lhsResult.getType(), code)


class CompoundAssignmentExpr(StatementExpr):
    """A compound assignment expression.

    Compound assignment is a combination of a binary operator and
    assignment.
    """

    def __init__(self, lhsExpr, rhsExpr, operator, linespan):
        """Construct a compound assignment expression.

        Param lhsExpr: a LeftHandSideExpr representing the variable to
        be assigned to.

        Param rhsExpr: an Expr that represents the right hand side.

        Param operator: a BinaryOperator that combines the old value
        of the variable and the rhs expression.

        Param linespan: see SyntaxNode.
        """
        StatementExpr.__init__(self, linespan)
        self.__lhsExpr = lhsExpr
        self.__rhsExpr = rhsExpr
        self.__operator = operator

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return (strFunc(self.__lhsExpr, *args, **kwargs) + ' ' +
                strFunc(self.__operator, *args, **kwargs) + '= (' +
                strFunc(self.__rhsExpr, *args, **kwargs) + ')')

    def api_getLhsExpr(self):
        """Return the left hand side as a LeftHandSideExpr."""
        return self.__lhsExpr

    def api_getRhsExpr(self):
        """Return the right hand side as an Expr."""
        return self.__rhsExpr

    def api_getOperator(self):
        """Return the associated BinaryOperator."""
        return self.__operator

    def translate(self, env):
        ep = ErrorProxy()
        try:
            lhsResult = self.__lhsExpr.translate(env)
            if not lhsResult.isVar():
                ep.add("'" + str(self.__lhsExpr) + "' cannot be assigned to",
                       self.linespan())
        except StaticError, e:
            ep.addException(e)
        try:
            rhsResult = self.__rhsExpr.translate(env)
        except StaticError, e:
            ep.addException(e)
        ep.raiseIfErrors()
        opResult = self.__operator.translate(env, lhsResult.getType(),
                                             rhsResult.getType())
        # If there are several numeric primitive types, the following
        # condition may need to be revised.
        if opResult.getType() is not lhsResult.getType():
            ep.finish(str(opResult.getType()) + " is incompatible with '" +
                      str(self.__lhsExpr) + "'",
                      self.linespan())
        iLen = lhsResult.getIndirectionLen()
        code = (lhsResult.getIndirectionCode() +
                [Duplicate(iLen) for x in xrange(iLen)] +
                lhsResult.getLoadCode() +
                rhsResult.getEvalCode() +
                opResult.getEvalCode() +
                [Insert(lhsResult.getIndirectionLen() + 1)] +
                lhsResult.getUncheckedStoreCode())
        return ValueResult(lhsResult.getType(), code)


class PrimaryFieldAccessExpr(LeftHandSideExpr):
    """Access of a field in the form primary.identifier."""

    def __init__(self, referenceExpr, identifier, linespan):
        """Construct a field access expression.

        Param referenceExpr: the Expr appearing on the left side of the
        period.

        Param identifier: the string appearing on the right side of
        the period.

        Param linespan: see SyntaxNode.
        """
        LeftHandSideExpr.__init__(self, linespan)
        self.__referenceExpr = referenceExpr
        self.__identifier = identifier

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('(' + strFunc(self.__referenceExpr, *args, **kwargs) + ').' +
                self.__identifier)

    def api_getReferenceExpr(self):
        """Return the Expr that refers to the containing object."""
        return self.__referenceExpr

    def api_getFieldName(self):
        """Return the name of the field as a string."""
        return self.__identifier

    def translate(self, env):
        referenceResult = self.__referenceExpr.translate(env)
        result = translateField(referenceResult.getType(), self.__identifier,
                                referenceResult.getEvalCode(), self)
        if result is not None:
            return result
        ErrorProxy().finish("'" + self.__identifier +
                            "' is not a member variable of " +
                            str(referenceResult.getType()),
                            self.linespan())


class SuperFieldAccessExpr(LeftHandSideExpr):
    """Access of a field in the form super.identifier."""

    def __init__(self, identifier, linespan):
        """Construct a field access expression.

        Param identifier: the string appearing on the right side of
        the period.

        Param linespan: see SyntaxNode.
        """
        LeftHandSideExpr.__init__(self, linespan)
        self.__identifier = identifier

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return 'super.' + self.__identifier

    def api_getFieldName(self):
        """Return the name of the field as a string."""
        return self.__identifier

    def translate(self, env):
        if not env.isInstanceContext():
            ErrorProxy().finish("cannot access 'super' from the context",
                                self.linespan())
        enclosingType = env.getEnclosingTypes()[-1]
        assert enclosingType is not env.getTopType('Object')
        superclass = enclosingType.getDirectSupertypes()[0]
        result = translateField(superclass, self.__identifier,
                                [LoadAuto('this')], self)
        if result is not None:
            return result
        ErrorProxy().finish("'" + self.__identifier +
                            "' is not a member variable of " +
                            str(superclass), self.linespan())


def translateField(type_, identifier, instanceCode, node):
    """Find a field of a type and return a corresponding StaticResult.

    Raise a StaticError if the given type is not a reference type, if
    there are several fields with the given name, or if instanceCode
    is None and an instance variable is found.

    Param type_: a ResultType that should contain the field.

    Param identifier: a string containing the simple name for the
    field.

    Param instanceCode: a list of code instructions that evaluate an
    instance of the given type_.  This instance is used to refer to
    the field if it is an instance variable.  If instanceCode is None,
    the field cannot be an instance variable, or a StaticError occurs.

    Param node: a SyntaxNode that is referred to in case a StaticError
    needs to be raised.

    Return value: a StaticResult that represents the field.  If no
    field with the given name is found, the return value is None.
    """
    if not type_.isReference():
        ErrorProxy().finish(str(type_) + ' is not a reference type',
                            node.linespan())
    fields = type_.getFields(identifier)
    if fields:
        if len(fields) > 1:
            ErrorProxy().finish("ambiguous variable '" + identifier + "'",
                                node.linespan())
        if fields[0].isStatic():
            if instanceCode is None:
                code = []
            else:
                # Evaluate instance reference even when accessing a
                # static variable.
                code = instanceCode + [Pop()]
            if fields[0].isFinal():
                return ValueResult(fields[0].getType(),
                                   code +
                                   [LoadStatic(fields[0].getSignature())])
            else:
                return VarResult(fields[0].getType(), code, 0,
                                 [LoadStatic(fields[0].getSignature())],
                                 [StoreStatic(fields[0].getSignature())],
                                 [StoreStatic(fields[0].getSignature())])
        else:
            if instanceCode is None:
                ErrorProxy().finish("cannot access instance variable '" +
                                    identifier + "' from the context",
                                    node.linespan())
            if fields[0].isFinal():
                return ValueResult(fields[0].getType(),
                                   (instanceCode +
                                    [LoadField(fields[0].getSignature())]))
            else:
                return VarResult(fields[0].getType(), instanceCode, 1,
                                 [CheckNonNull(1),
                                  LoadField(fields[0].getSignature())],
                                 [CheckNonNull(2),
                                  StoreField(fields[0].getSignature())],
                                 [StoreField(fields[0].getSignature())])
    return None


class ArrayAccessExpr(LeftHandSideExpr):
    """Access of a component of an array."""

    def __init__(self, arrayExpr, indexExpr, linespan):
        """Construct an array access expression.

        Param arrayExpr: the Expr appearing to the left of the brackets.

        Param indexExpr: the Expr appearing within the brackets.

        Param linespan: see SyntaxNode.
        """
        LeftHandSideExpr.__init__(self, linespan)
        self.__arrayExpr = arrayExpr
        self.__indexExpr = indexExpr

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('(' + strFunc(self.__arrayExpr, *args, **kwargs) + ')[' +
                strFunc(self.__indexExpr, *args, **kwargs) + ']')

    def api_getArrayExpr(self):
        """Return the Expr that refers to the array."""
        return self.__arrayExpr

    def api_getIndexExpr(self):
        """Return the array index as an Expr."""
        return self.__indexExpr

    def translate(self, env):
        ep = ErrorProxy()
        try:
            arrayResult = self.__arrayExpr.translate(env)
            if not arrayResult.getType().isArray():
                ep.add(str(arrayResult.getType()) + ' is not an array type',
                       self.linespan())
        except StaticError, e:
            ep.addException(e)
        try:
            indexResult = self.__indexExpr.translate(env)
            if indexResult.getType() is not intType:
                ep.add('array index must be int', self.__indexExpr.linespan())
        except StaticError, e:
            ep.addException(e)
        ep.raiseIfErrors()
        compType = arrayResult.getType().getComponentType()
        lengthField = arrayResult.getType().getFields('length')[0]
        badIndexLabel = Label()
        goodIndexLabel = Label()
        loadCode = [CheckNonNull(2),             # arr, ind
                    Duplicate(),                 # arr, ind, ind
                    Literal(0),                  # arr, ind, ind, 0
                    IntLessThan(),               # arr, ind, ind<0
                    IfGoto(badIndexLabel),       # arr, ind
                    Duplicate(),                 # arr, ind, ind
                    Duplicate(3),                # arr, ind, ind, arr
                    LoadField(lengthField.getSignature()),
                                                 # arr, ind, ind, len
                    IntLessThan(),               # arr, ind, ind<len
                    IfGoto(goodIndexLabel),      # arr, ind
                    badIndexLabel,
                    BadArrayIndex(),
                    goodIndexLabel,
                    LoadArray()]                 # val
        # Make new labels for store code.
        badIndexLabel = Label()
        goodIndexLabel = Label()
        chkStoreCode = [CheckNonNull(3),         # arr, ind, val
                        Duplicate(2),            # arr, ind, val, ind
                        Literal(0),              # arr, ind, val, ind, 0
                        IntLessThan(),           # arr, ind, val, ind<0
                        IfGoto(badIndexLabel),   # arr, ind, val
                        Duplicate(2),            # arr, ind, val, ind
                        Duplicate(4),            # arr, ind, val, ind, arr
                        LoadField(lengthField.getSignature()),
                                                 # arr, ind, val, ind, len
                        IntLessThan(),           # arr, ind, val, ind<len
                        IfGoto(goodIndexLabel),  # arr, ind, val
                        badIndexLabel,
                        BadArrayIndex(),
                        goodIndexLabel]
        unchkStoreCode = []
        if compType.isReference():
            # Must check that the value being stored is assignable to
            # the component type of the array.
            chkStoreCode += [CheckArrayStore()]
            unchkStoreCode += [CheckArrayStore()]
        chkStoreCode += [StoreArray()]
        unchkStoreCode += [StoreArray()]
        return VarResult(compType,
                         arrayResult.getEvalCode() + indexResult.getEvalCode(),
                         2, loadCode, chkStoreCode, unchkStoreCode)


def _areSubtypes(x, y):
    """Return true if x[i] is a subtype of y[i] for all i."""
    for (xType, yType) in zip(x, y):
        if not xType.isSubtypeOf(yType):
            return False
    return True

def _isStrictlyMoreSpecific(f, g):
    """Return true if Function f is strictly more specific than Function g."""
    fTypes = [t for (p, t) in f.getParameters()]
    gTypes = [t for (p, t) in g.getParameters()]
    return (_areSubtypes(fTypes, gTypes) and not _areSubtypes(gTypes, fTypes))

def determineFunction(candidates, argTypes):
    """Find the most specific function matching the given attributes.

    Param candidates: the list of Functions to examine.

    Param argTypes: a list of Types representing the argument types.

    Return value: a list of the maximally specific Functions among the
    candidates that accept the given argument types.  If the list is
    empty, there are no Functions that match the arguments.  If the
    list has two or more elements, the invocation with given argument
    types is ambiguous (we assume here that candidates does not
    contain two Functions with the same argument types).
    """
    applicable = []
    for f in candidates:
        if len(f.getParameters()) == len(argTypes):
            paramTypes = [t for (p, t) in f.getParameters()]
            if _areSubtypes(argTypes, paramTypes):
                # Each argument type is a subtype of the corresponding
                # parameter type.
                applicable.append(f)
    maximallySpecific = []
    for f in applicable:
        for g in applicable:
            if _isStrictlyMoreSpecific(g, f):
                # g is strictly more specific than f, so
                # f cannot be maximally specific.
                break
        else:
            maximallySpecific.append(f)
    return maximallySpecific


class MethodInvocationExpr(StatementExpr):
    """A base class for method invocation expressions."""

    def __init__(self, argList, linespan):
        """Construct a method invocation expression.

        Param argList: a list of Exprs representing the arguments.

        Param linespan: see SyntaxNode.
        """
        StatementExpr.__init__(self, linespan)
        self.__argList = argList

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return (self._strName(strFunc, *args, **kwargs) + '(' +
                ', '.join([strFunc(e, *args, **kwargs) for
                           e in self.__argList]) + ')')

    def api_getArgList(self):
        """Return the arguments as a list of Exprs."""
        return self.__argList[:]

    def _strName(self, strFunc, *args, **kwargs):  #pragma: no cover
        """Return a string representing the part before the argument list."""
        raise NotImplementedError('Pure virtual method')

    def translate(self, env):
        ep = ErrorProxy()
        try:
            (containingType, name, targetCode) = self._translateName(env)
        except StaticError, e:
            ep.addException(e)
        argTypes = []
        argCode = []
        for a in self.__argList:
            try:
                result = a.translate(env)
                argTypes.append(result.getType())
                argCode += result.getEvalCode()
            except StaticError, e:
                ep.addException(e)
        ep.raiseIfErrors()
        methods = determineFunction(containingType.getFunctions(name),
                                    argTypes)
        if not methods:
            ep.finish("no method '" + name + "' in " + str(containingType) +
                      " matches argument types (" +
                      ', '.join([str(t) for t in argTypes]) +
                      ")",
                      self.linespan())
        if len(methods) > 1:
            ep.finish("ambiguous method invocation matches both '" +
                      str(methods[0]) + "' and '" + str(methods[1]) + "'",
                      self.linespan())
        if methods[0].isStatic():
            code = []
            if targetCode is not None:
                # Evaluate target but discard result.
                code += targetCode + [Pop()]
            code += argCode + [CallStatic(containingType,
                                          methods[0].getSignature(),
                                          len(self.__argList))]
        else:
            if targetCode is None:
                ep.finish("cannot call the instance method '" +
                          str(methods[0]) + "' from the context",
                          self.linespan())
            code = (targetCode + argCode +
                    self._instanceMethodCode(containingType, methods[0]))
        return ValueResult(methods[0].getReturnType(), code)

    def _translateName(self, env):  #pragma: no cover
        """Determine the name and other attributes for the method invocation.

        Raise a StaticError if static checking fails.

        Param env: a StaticEnvironment in which the invocation appears.

        Return value: a triple (type_, name, code), where type_ is the
        ReferenceType whose member the method should be, name is the
        simple name of the method being invoked, and code is a list of
        code instructions that evaluate the target reference for the
        method invocation.  If there is no target reference (the
        method must be a static method), code is None.
        """
        raise NotImplementedError('Pure virtual method')

    def _instanceMethodCode(self, containingType, method):
        """Return code that calls an instance method.

        Raise a StaticError if the method cannot be invoked.

        Param containingType: the Type that contains the method.

        Param method: the method as a Function.

        Return value: a list of code instructions.
        """
        # Default implementation
        return [CallDynamic(method.getSignature(),
                            len(method.getParameters()) + 1)]


class NameMethodInvocationExpr(MethodInvocationExpr):
    """A method invocation using a simple or qualified method name."""

    def __init__(self, name, argList, linespan):
        """Construct a method invocation expression using a method name.

        Param name: the Name used in the invocation.

        Param argList: a list of expressions that are used as arguments.

        Param linespan: see SyntaxNode.
        """
        MethodInvocationExpr.__init__(self, argList, linespan)
        self.__name = name

    def api_getInvocationName(self):
        """Return the Name used to refer to the method.

        The last part of the Name identifies the method.  The
        preceding parts denote the type or variable that contain the
        method.
        """
        return self.__name

    def _strName(self, strFunc, *args, **kwargs):
        return strFunc(self.__name, *args, **kwargs)

    def _translateName(self, env):
        return self.__name.resolveAsMethod(env)


class PrimaryMethodInvocationExpr(MethodInvocationExpr):
    """A method invocation using a primary."""

    def __init__(self, referenceExpr, identifier, argList, linespan):
        """Construct a primary method invocation expression.

        Param referenceExpr: the Expr appearing on the left side of
        the period.

        Param identifier: the string appearing on the right side of
        the period.

        Param argList: a list of expressions that are used as
        arguments.

        Param linespan: see SyntaxNode.
        """
        MethodInvocationExpr.__init__(self, argList, linespan)
        self.__referenceExpr = referenceExpr
        self.__identifier = identifier

    def api_getReferenceExpr(self):
        """Return the Expr used to refer to the object that has the method."""
        return self.__referenceExpr

    def api_getSimpleMethodName(self):
        """Return the string used to identify the simple name of the method."""
        return self.__identifier

    def _strName(self, strFunc, *args, **kwargs):
        return ('(' + strFunc(self.__referenceExpr, *args, **kwargs) + ').' +
                self.__identifier)

    def _translateName(self, env):
        result = self.__referenceExpr.translate(env)
        if not result.getType().isReference():
            ErrorProxy().finish(str(result.getType()) + " has no method '" +
                                self.__identifier + "'", self.linespan())
        return (result.getType(), self.__identifier, result.getEvalCode())


class SuperMethodInvocationExpr(MethodInvocationExpr):
    """A method invocation of the form super.method(args)."""

    def __init__(self, identifier, argList, linespan):
        """Construct a super method invocation expression.

        Param identifier: the string appearing on the right side of
        the period.

        Param argList: a list of expressions that are used as
        arguments.

        Param linespan: see SyntaxNode.
        """
        MethodInvocationExpr.__init__(self, argList, linespan)
        self.__identifier = identifier

    def api_getSimpleMethodName(self):
        """Return the string used to identify the simple name of the method."""
        return self.__identifier

    def _strName(self, strFunc, *args, **kwargs):
        return 'super.' + self.__identifier

    def _translateName(self, env):
        if not env.isInstanceContext():
            ErrorProxy().finish("cannot access 'super' from the context",
                                self.linespan())
        enclosingType = env.getEnclosingTypes()[-1]
        assert enclosingType is not env.getTopType('Object')
        superclass = enclosingType.getDirectSupertypes()[0]
        return (superclass, self.__identifier, [LoadAuto('this')])

    def _instanceMethodCode(self, containingType, method):
        if method.isAbstract():
            ErrorProxy().finish("cannot directly invoke '" + str(method) + "'",
                                self.linespan())
        # Bind to the method in superclass even if it is an instance method.
        return [CallStatic(containingType, method.getSignature(),
                           len(method.getParameters()) + 1)]


class InstanceCreationExpr(StatementExpr):
    """Class instance creation using new."""

    def __init__(self, typeName, argList, linespan):
        """Construct an instance creation expression.

        Param typeName: the Name appearing after the keyword 'new'.

        Param argList: a list of expressions that are used as arguments.

        Param linespan: see SyntaxNode.
        """
        StatementExpr.__init__(self, linespan)
        self.__typeName = typeName
        self.__argList = argList

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('new ' + strFunc(self.__typeName, *args, **kwargs) + '(' +
                ', '.join([strFunc(e, *args, **kwargs) for
                           e in self.__argList]) + ')')

    def api_getTypeName(self):
        """Return the name of the class as a Name."""
        return self.__typeName

    def api_getArgList(self):
        """Return the arguments as a list of Exprs."""
        return self.__argList[:]

    def translate(self, env):
        ep = ErrorProxy()
        code = []
        try:
            type_ = self.__typeName.resolveAsType(env)
            if type_.isEnum() or type_.isAbstract():
                ep.add(str(type_) + ' cannot be instantiated',
                       self.linespan())
            else:
                code += [NewInstance(type_), Duplicate()]
        except StaticError, e:
            ep.addException(e)
        argTypes = []
        for a in self.__argList:
            try:
                result = a.translate(env)
                argTypes.append(result.getType())
                code += result.getEvalCode()
            except StaticError, e:
                ep.addException(e)
        ep.raiseIfErrors()
        constructors = determineFunction(type_.getFunctions('#constructor'),
                                         argTypes)
        if not constructors:
            ep.finish("no constructor matches argument types (" +
                      ', '.join([str(t) for t in argTypes]) +
                      ")",
                      self.linespan())
        if len(constructors) > 1:
            ep.finish("ambiguous constructor invocation matches both '" +
                      str(constructors[0]) + "' and '" +
                      str(constructors[1]) + "'",
                      self.linespan())
        code += [CallStatic(type_, constructors[0].getSignature(),
                            len(self.__argList) + 1),
                 Pop()]
        return ValueResult(type_, code)


class ArrayCreationExpr(Expr):
    """Creation of an array."""

    def __init__(self, typeSymbol, lengthExprs, extraDims,
                 initializer, linespan):
        """Construct an array creation expression.

        Param typeSymbol: the TypeSymbol appearing after the keyword
        'new', representing the element type.

        Param lengthExprs: a list of Exprs inside brackets, each
        representing the length of a dimension of the array.  This
        must be an empty list if and only if initializer is not None.
        The length of this list must currently not exceed 1.

        Param extraDims: a list of None elements, one for each empty
        pair of brackets following the possible nonempty pairs.

        Param initializer: an ArrayInitializer, if present, or None.

        Param linespan: see SyntaxNode.
        """
        Expr.__init__(self, linespan)
        self.__typeSymbol = typeSymbol
        self.__lengthExprs = lengthExprs
        self.__extraDims = extraDims
        self.__initializer = initializer

    def api_recurseStr(self, strFunc, *args, **kwargs):
        s = 'new ' + strFunc(self.__typeSymbol, *args, **kwargs)
        s += ''.join(['[' + strFunc(e, *args, **kwargs) + ']' for
                      e in self.__lengthExprs])
        s += ''.join(['[]' for d in self.__extraDims])
        if self.__initializer is not None:
            s += ' ' + strFunc(self.__initializer, *args, **kwargs)
        return s

    def api_getElementType(self):
        """Return the element type of the array as a TypeSymbol.

        Notice that the component type of an array may be an array
        type, but the element type may not.
        """
        return self.__typeSymbol

    def api_getDimensions(self):
        """Return a list that specifies the dimensions of the array.

        The list has the form [expr_1, ..., expr_k, None, ..., None],
        where k is greater than or equal to 0, and the number of None
        elements may be 0 or more.  The expr_i is an Expr that
        specifies the length of the i:th dimension.  The None elements
        stand for dimensions whose length is unspecified or whose
        length is specified by the initializer.  The total number of
        dimensions in the array is the number of elements in this
        list.

        The array has an initializer if and only if k is 0.

        The current syntax restricts the value of k to be at most 1.
        """
        return self.__lengthExprs + self.__extraDims

    def api_getInitializer(self):
        """Return the ArrayInitializer if it exists or None otherwise."""
        return self.__initializer

    def translate(self, env):
        ep = ErrorProxy()
        t = self.__typeSymbol.resolve(env)
        for d in self.__extraDims:
            t = t.getArrayType(env)
        if self.__initializer is not None:
            assert len(self.__lengthExprs) == 0
            arrayType = t
            return ValueResult(arrayType,
                               self.__initializer.translate(env, arrayType))
        else:
            assert len(self.__lengthExprs) == 1
            componentType = t
            arrayType = t.getArrayType(env)
            result = self.__lengthExprs[0].translate(env)
            if result.getType() is not intType:
                ep.finish('length of an array must be int',
                          self.__lengthExprs[0].linespan())
            lengthField = arrayType.getFields('length')[0]
            negativeLabel = Label()
            nonnegativeLabel = Label()
            code = ([NewInstance(arrayType),      # arr
                     Duplicate()] +               # arr, arr
                    result.getEvalCode() +        # arr, arr, len
                    [Duplicate(),                 # arr, arr, len, len
                     Literal(0),                  # arr, arr, len, len, 0
                     IntLessThan(),               # arr, arr, len, len<0
                     IfGoto(negativeLabel),       # arr, arr, len
                     Goto(nonnegativeLabel),
                     negativeLabel,
                     NegativeArraySize(),
                     nonnegativeLabel,            # arr, arr, len
                     Duplicate(2),                # arr, arr, len, arr
                     Duplicate(2),                # arr, arr, len, arr, len
                     StoreField(lengthField.getSignature()),
                                                  # arr, arr, len
                     Literal(componentType.getDefaultValue()),
                                                  # arr, arr, len, val
                     InitArray()])                # arr
            return ValueResult(arrayType, code)


class VarInitializer(SyntaxNode):
    """A base class for variable initializers.

    A variable initializer is like an expression, but its type is
    determined by its context and not only by what is inside.
    """

    def __init__(self, linespan):
        SyntaxNode.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):  #pragma: no cover
        return '<initializer>'

    def translate(self, env, expectedType):
        """Perform static type checking and generate internal code.

        Raise a StaticError if type checking fails.

        Param env: a StaticEnvironment for the initializer.  This must
        not be mutated.

        Param expectedType: the Type of the variable that is being
        initialized.  If the type of the initializer does not match, a
        StaticError is raised.

        Return value: a list of internal code instructions that
        evaluate the initializer and push its value on
        top of the operand stack.
        """
        raise NotImplementedError('Pure virtual method')


class ExprInitializer(VarInitializer):
    """A variable initializer that is an expression."""

    def __init__(self, expr):
        """Construct a variable initializer from an expression.

        Param expr: the expression (an Expr).
        """
        VarInitializer.__init__(self, expr.linespan())
        self.__expr = expr

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return strFunc(self.__expr, *args, **kwargs)

    def api_getExpr(self):
        """Return the contained Expr."""
        return self.__expr

    def translate(self, env, expectedType):
        result = self.__expr.translate(env)
        if not result.getType().isSubtypeOf(expectedType):
            ErrorProxy().finish('cannot initialize ' +
                                str(expectedType) + ' with ' +
                                str(result.getType()),
                                self.linespan())
        return result.getEvalCode()


class ArrayInitializer(VarInitializer):
    """An array initializer."""

    def __init__(self, components, linespan):
        """Construct an array initializer.

        Param components: a list of VarInitializers that are used to
        initialize the components of the array.

        Param linespan: see SyntaxNode.
        """
        VarInitializer.__init__(self, linespan)
        self.__components = components

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('{' + ', '.join([strFunc(c, *args, **kwargs) for
                                 c in self.__components]) + '}')

    def api_getComponentInitializers(self):
        """Return the components as a list of VarInitializers."""
        return self.__components[:]

    def translate(self, env, expectedType):
        ep = ErrorProxy()
        if not expectedType.isArray():
            ep.finish('cannot initialize ' + str(expectedType) +
                      ' with an array',
                      self.linespan())
        componentType = expectedType.getComponentType()
        lengthField = expectedType.getFields('length')[0]
        code = [NewInstance(expectedType),           # arr
                Duplicate(),                         # arr, arr
                Literal(len(self.__components)),     # arr, arr, len
                Duplicate(2),                        # arr, arr, len, arr
                Duplicate(2),                        # arr, arr, len, arr, len
                StoreField(lengthField.getSignature()),
                                                     # arr, arr, len,
                Literal(componentType.getDefaultValue()),
                                                     # arr, arr, len, val
                InitArray()]                         # arr
        index = 0
        for c in self.__components:
            try:
                code += ([Duplicate(),               # arr, arr
                          Literal(index)] +          # arr, arr, ind
                         c.translate(env, componentType) +
                                                     # arr, arr, ind, val
                         [StoreArray()])             # arr
            except StaticError, e:
                ep.addException(e)
            index += 1
        ep.raiseIfErrors()
        return code
