#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# grammar.py
#
# 2007-11-09  Jori Dubrovin:  Added generateParenExpr flag
# 2007-06-11  Jori Dubrovin:  Removed debugging code
# 2006-07-29  Jori Dubrovin:  Eliminated unnecessary parser generation
# 2006-01-31  Jori Dubrovin:  Added modifiers to parameters and local variables
# 2005-11-14  Jori Dubrovin:  Renamed CondExpr to TernaryCondExpr
# 2005-09-01  Jori Dubrovin:  Fixed yacc.errok usage
# 2005-08-09  Jori Dubrovin:  Fixed path for parsetab.py
# 2005-06-17  Jori Dubrovin:  Now accepts an empty input, too
# 2005-05-02  Jori Dubrovin:  Rearranged the ForCondition rule
# 2005-04-22  Jori Dubrovin:  Initial version
#

"""Grammar for the action language.

Use the parse function to parse an input string.
"""

import sys
import os
import ply.yacc

import lexical
import syntaxtree
import statement
import typedecl
import expr
import binaryop
import mytypes
import names
import errors


def parse(input_, yaccdebug=0, generateParenExpr=True):
    """Parse an input string and return the syntax tree.

    Param input_: the input program.  It must not contain carriage
    returns, only linefeeds.

    Param yaccdebug: the debug parameter passed to yacc.Parser.parse.

    Param generateParenExpr: if set to False, no ParenExpr syntax
    nodes will be generated.  As a result, some (obscure) syntax
    errors will go undetected.

    A StaticError is raised if the input is grammatically incorrect.

    Return value: the root node of the syntax tree.
    """

    # Make a temporary global ErrorProxy object and generateParenExpr
    # flag that are used during this parse.
    global _errorProxy, _generateParenExpr
    try:
        _generateParenExpr = generateParenExpr
        _errorProxy = errors.ErrorProxy()
        try:
            # Add a space character to input_ to suppress an error from lex
            # when the input is an empty string.
            tree = _parser.parse(input_ + ' ', lexer=lexical.lexer(),
                                 debug=yaccdebug)
        except errors.StaticError, e:
            # Catch errors raised by lexical and add them to the list
            # together with the possible grammatical errors.
            _errorProxy.addException(e)
        _errorProxy.raiseIfErrors()
    finally:
        del _errorProxy, _generateParenExpr
    return tree


def _lines(p):
    """Return the linespan of all the symbols in the given yacc production."""
    return (p.linespan(1)[0], p.linespan(len(p)-1)[1])

# These rules return a SyntaxRoot

def p_SyntaxRoot(p):
    '''SyntaxRoot : TopStatementList'''
    p[0] = syntaxtree.SyntaxRoot(p[1], _lines(p))

# These rules return a list of TopStatements

def p_TopStatementList(p):
    '''TopStatementList :
                        | TopStatementList TopStatement'''
    if len(p) < 3:
        p[0] = []
    else:
        p[0] = p[1] + [p[2]]

# These rules return a TopStatement

def p_TopStatement(p):
    '''TopStatement : TopTypeDeclaration
                    | BlockStatement'''
    p[0] = p[1]

# These rules return a TopTypeDeclaration

def p_TopTypeDeclaration(p):
    '''TopTypeDeclaration : TypeDeclaration'''
    p[0] = typedecl.TopTypeDeclaration(p[1])

# These rules return a list of BlockStatements

def p_BlockStatementList(p):
    '''BlockStatementList :
                          | BlockStatementList BlockStatement'''
    if len(p) < 3:
        p[0] = []
    else:
        p[0] = p[1] + [p[2]]

# These rules return a BlockStatement

def p_BlockStatement(p):
    '''BlockStatement : LocalVarDeclarationStatement
                      | Statement'''
    p[0] = p[1]

def p_LocalVarDeclarationStatement_1(p):
    '''LocalVarDeclarationStatement : TypeY LocalVarDeclaratorList SEMICOLON'''
    p[0] = statement.LocalVarDeclarationStatement([], p[1], p[2], _lines(p))

def p_LocalVarDeclarationStatement_2(p):
    '''LocalVarDeclarationStatement : ModifierList TypeY \
                                          LocalVarDeclaratorList SEMICOLON'''
    p[0] = statement.LocalVarDeclarationStatement(p[1], p[2], p[3], _lines(p))

# These rules return a Statement

def p_Block(p):
    '''Block : LBRACE BlockStatementList RBRACE'''
    p[0] = statement.Block(p[2], _lines(p))

def p_Statement(p):
    '''Statement : StatementWithoutTrailingSubstatement
                 | LabeledStatement
                 | IfThenStatement
                 | IfThenElseStatement
                 | WhileStatement
                 | ForStatement'''
    p[0] = p[1]

def p_StatementWithoutTrailingSubstatement(p):
    '''StatementWithoutTrailingSubstatement : Block
                                            | EmptyStatement
                                            | ExprStatement
                                            | SendStatement
                                            | AssertStatement
                                            | SwitchStatement
                                            | DoStatement
                                            | ReturnStatement
                                            | BreakStatement
                                            | ContinueStatement'''
    p[0] = p[1]

def p_StatementNoShortIf(p):
    '''StatementNoShortIf : StatementWithoutTrailingSubstatement
                          | LabeledStatementNoShortIf
                          | IfThenElseStatementNoShortIf
                          | WhileStatementNoShortIf
                          | ForStatementNoShortIf'''
    p[0] = p[1]

def p_EmptyStatement(p):
    '''EmptyStatement : SEMICOLON'''
    p[0] = statement.EmptyStatement(_lines(p))

def p_ExprStatement(p):
    '''ExprStatement : StatementExpr SEMICOLON'''
    p[0] = statement.ExprStatement(p[1], _lines(p))

def p_SendStatement(p):
    '''SendStatement : SEND IDENTIFIER Arguments TO Expr SEMICOLON'''
    p[0] = statement.SendStatement(p[2], p[3], p[5], _lines(p))

def p_AssertStatement(p):
    '''AssertStatement : ASSERT Expr SEMICOLON'''
    p[0] = statement.AssertStatement(p[2], _lines(p))

def p_LabeledStatement(p):
    '''LabeledStatement : IDENTIFIER COLON Statement'''
    p[0] = statement.LabeledStatement(p[1], p[3], _lines(p))

def p_LabeledStatementNoShortIf(p):
    '''LabeledStatementNoShortIf : IDENTIFIER COLON StatementNoShortIf'''
    p[0] = statement.LabeledStatement(p[1], p[3], _lines(p))

def p_IfThenStatement(p):
    '''IfThenStatement : IF LPAREN Expr RPAREN Statement'''
    falsePart = statement.EmptyStatement((p.linespan(5)[1], p.linespan(5)[1]))
    p[0] = statement.IfStatement(p[3], p[5], falsePart, _lines(p))

def p_IfThenElseStatement(p):
    '''IfThenElseStatement : IF LPAREN Expr RPAREN \
                                 StatementNoShortIf ELSE Statement'''
    p[0] = statement.IfStatement(p[3], p[5], p[7], _lines(p))

def p_IfThenElseStatementNoShortIf(p):
    '''IfThenElseStatementNoShortIf : IF LPAREN Expr RPAREN \
                                          StatementNoShortIf ELSE \
                                          StatementNoShortIf'''
    p_IfThenElseStatement(p)  # Delegate action to the other rule

def p_SwitchStatement(p):
    '''SwitchStatement : SWITCH LPAREN Expr RPAREN \
                             LBRACE SwitchGroupList RBRACE'''
    p[0] = statement.SwitchStatement(p[3], p[6], _lines(p))

# These rules return a list of tuples (SwitchLabel, list of Statements)

def p_SwitchGroupList(p):
    '''SwitchGroupList :
                       | SwitchGroupList SwitchGroup'''
    if len(p) < 3:
        p[0] = []
    else:
        p[0] = p[1] + [p[2]]

# These rules return a tuple (SwitchLabel, list of Statements)

def p_SwitchGroup(p):
    '''SwitchGroup : SwitchLabel BlockStatementList'''
    p[0] = (p[1], p[2])

# These rules return a SwitchLabel

def p_SwitchLabel_1(p):
    '''SwitchLabel : CASE Expr COLON'''
    p[0] = statement.SwitchLabel(p[2], (p.linespan(1)[0], p.linespan(3)[1]))

def p_SwitchLabel_2(p):
    '''SwitchLabel : DEFAULT COLON'''
    p[0] = statement.SwitchLabel(None, (p.linespan(1)[0], p.linespan(2)[1]))

# These rules return a Statement

def p_WhileStatement(p):
    '''WhileStatement : WHILE LPAREN Expr RPAREN Statement'''
    p[0] = statement.WhileStatement(p[3], p[5], _lines(p))

def p_WhileStatementNoShortIf(p):
    '''WhileStatementNoShortIf : WHILE LPAREN Expr RPAREN StatementNoShortIf'''
    p_WhileStatement(p)  # Delegate action to the other rule

def p_DoStatement(p):
    '''DoStatement : DO Statement WHILE LPAREN Expr RPAREN SEMICOLON'''
    p[0] = statement.DoStatement(p[5], p[2], _lines(p))

def p_ForStatement_1(p):
    '''ForStatement : FOR LPAREN LocalVarDeclarationStatement \
                          ForCondition ForUpdate \
                          RPAREN Statement'''
    p[0] = statement.LocalVarForStatement(p[3], p[4], p[5], p[7], _lines(p))

def p_ForStatement_2(p):
    '''ForStatement : FOR LPAREN StatementExprList SEMICOLON \
                          ForCondition ForUpdate \
                          RPAREN Statement'''
    p[0] = statement.ExprListForStatement(p[3], p[5], p[6], p[8], _lines(p))

def p_ForStatementNoShortIf_1(p):
    '''ForStatementNoShortIf : FOR LPAREN LocalVarDeclarationStatement \
                                   ForCondition ForUpdate \
                                   RPAREN StatementNoShortIf'''
    p_ForStatement_1(p)  # Delegate action to the other rule

def p_ForStatementNoShortIf_2(p):
    '''ForStatementNoShortIf : FOR LPAREN StatementExprList SEMICOLON \
                                   ForCondition ForUpdate \
                                   RPAREN StatementNoShortIf'''
    p_ForStatement_2(p)  # Delegate action to the other rule

# These rules return a list of TypeBodyDeclarations

def p_ClassBodyDeclarationList(p):
    '''ClassBodyDeclarationList :
                                | ClassBodyDeclarationList \
                                      ClassBodyDeclaration'''
    if len(p) < 3:
        p[0] = []
    else:
        p[0] = p[1] + [p[2]]

def p_MemberDeclarationList(p):
    '''MemberDeclarationList :
                             | MemberDeclarationList MemberDeclaration'''
    if len(p) < 3:
        p[0] = []
    else:
        p[0] = p[1] + [p[2]]

def p_EnumConstantList(p):
    '''EnumConstantList : CommaOpt
                        | EnumConstantNonemptyList CommaOpt'''
    if len(p) < 3:
        p[0] = []
    else:
        p[0] = p[1]

def p_EnumConstantNonemptyList(p):
    '''EnumConstantNonemptyList : EnumConstant
                                | EnumConstantNonemptyList COMMA \
                                      EnumConstant'''
    if len(p) < 4:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[3]]

# These rules return a TypeBodyDeclaration

def p_ClassBodyDeclaration(p):
    '''ClassBodyDeclaration : MemberDeclaration
                            | ConstructorDeclaration'''
    p[0] = p[1]

def p_MemberDeclaration(p):
    '''MemberDeclaration : FieldDeclaration
                         | MethodDeclaration
                         | TypeDeclaration'''
    p[0] = p[1]

def p_TypeDeclaration(p):
    '''TypeDeclaration : ClassDeclaration
                       | EnumDeclaration
                       | InterfaceDeclaration'''
    p[0] = p[1]

def p_FieldDeclaration_1(p):
    '''FieldDeclaration : TypeX FieldDeclaratorList SEMICOLON'''
    p[0] = typedecl.FieldDeclaration([], p[1], p[2], _lines(p))

def p_FieldDeclaration_2(p):
    '''FieldDeclaration : ModifierList TypeX FieldDeclaratorList SEMICOLON'''
    p[0] = typedecl.FieldDeclaration(p[1], p[2], p[3], _lines(p))

def p_MethodDeclaration_1(p):
    '''MethodDeclaration : TypeX IDENTIFIER \
                               FormalParameters MethodBody
                         | VoidResultType IDENTIFIER \
                               FormalParameters MethodBody'''
    p[0] = typedecl.MethodDeclaration([], p[1], p[2], p[3], p[4], _lines(p))

def p_MethodDeclaration_2(p):
    '''MethodDeclaration : ModifierList TypeX IDENTIFIER \
                               FormalParameters MethodBody
                         | ModifierList VoidResultType IDENTIFIER \
                               FormalParameters MethodBody'''
    p[0] = typedecl.MethodDeclaration(p[1], p[2], p[3], p[4], p[5], _lines(p))

def p_ConstructorDeclaration(p):
    '''ConstructorDeclaration : IDENTIFIER FormalParameters \
                                    LBRACE ConstructorStatementList RBRACE'''
    p[0] = typedecl.ConstructorDeclaration(p[1], p[2], p[4], _lines(p))

def p_EnumDeclaration(p):
    '''EnumDeclaration : ENUM IDENTIFIER \
                             LBRACE EnumConstantList RBRACE'''
    p[0] = typedecl.EnumDeclaration(p[2], p[4], _lines(p))

def p_EnumConstant(p):
    '''EnumConstant : IDENTIFIER'''
    p[0] = typedecl.EnumConstantDeclaration(p[1], _lines(p))

def p_ClassDeclaration_1(p):
    '''ClassDeclaration : CLASS IDENTIFIER Superclass Implements \
                              LBRACE ClassBodyDeclarationList RBRACE'''
    p[0] = typedecl.ClassDeclaration([], p[2], p[3], p[4], p[6], _lines(p))

def p_ClassDeclaration_2(p):
    '''ClassDeclaration : ModifierList CLASS IDENTIFIER Superclass Implements \
                              LBRACE ClassBodyDeclarationList RBRACE'''
    p[0] = typedecl.ClassDeclaration(p[1], p[3], p[4], p[5], p[7], _lines(p))

def p_InterfaceDeclaration(p):
    '''InterfaceDeclaration : INTERFACE IDENTIFIER Superinterfaces \
                                  LBRACE MemberDeclarationList RBRACE'''
    p[0] = typedecl.InterfaceDeclaration(p[2], p[3], p[5], _lines(p))

# These rules return None

def p_CommaOpt(p):
    '''CommaOpt :
                | COMMA'''
    p[0] = None

# These rules return a list of Names

def p_Superclass(p):
    '''Superclass :
                  | EXTENDS NameX'''
    if len(p) < 3:
        p[0] = []
    else:
        p[0] = [p[2]]

def p_Implements(p):
    '''Implements :
                  | IMPLEMENTS NameListX'''
    if len(p) < 3:
        p[0] = []
    else:
        p[0] = p[2]

def p_Superinterfaces(p):
    '''Superinterfaces :
                       | EXTENDS NameListX'''
    if len(p) < 3:
        p[0] = []
    else:
        p[0] = p[2]

def p_NameListX(p):
    '''NameListX : NameX
                 | NameListX COMMA NameX'''
    if len(p) < 4:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[3]]

# These rules return a list of FieldDeclarators

def p_FieldDeclaratorList(p):
    '''FieldDeclaratorList : FieldDeclarator
                           | FieldDeclaratorList COMMA FieldDeclarator'''
    if len(p) < 4:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[3]]

# These rules return a FieldDeclarator

def p_FieldDeclarator(p):
    '''FieldDeclarator : IDENTIFIER
                       | IDENTIFIER EQ VarInitializer'''
    if len(p) < 4:
        p[0] = typedecl.FieldDeclarator(p[1], None, _lines(p))
    else:
        p[0] = typedecl.FieldDeclarator(p[1], p[3], _lines(p))

# These rules return a Statement or None

def p_MethodBody_1(p):
    '''MethodBody : Block'''
    p[0] = p[1]

def p_MethodBody_2(p):
    '''MethodBody : SEMICOLON'''
    p[0] = None

# These rules return a list of ConstructorStatements

def p_ConstructorStatementList(p):
    '''ConstructorStatementList :
                                | ConstructorStatementList \
                                      ConstructorStatement'''
    if len(p) < 3:
        p[0] = []
    else:
        p[0] = p[1] + [p[2]]

# These rules return a ConstructorStatement

def p_ConstructorStatement_1(p):
    '''ConstructorStatement : BlockStatement'''
    p[0] = typedecl.ConstructorBlockStatement(p[1])

def p_ConstructorStatement_2(p):
    '''ConstructorStatement : SUPER Arguments SEMICOLON'''
    p[0] = typedecl.SuperConstructorInvocation(p[2], _lines(p))

# These rules return a list of strings

def p_ModifierList(p):
    '''ModifierList : Modifier
                    | ModifierList Modifier'''
    if len(p) < 3:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[2]]

# These rules return a string

def p_Modifer_1(p):
    '''Modifier : STATIC'''
    p[0] = 'static'

def p_Modifier_2(p):
    '''Modifier : NATIVE'''
    p[0] = 'native'

def p_Modifier_3(p):
    '''Modifier : ABSTRACT'''
    p[0] = 'abstract'

def p_Modifier_4(p):
    '''Modifier : FINAL'''
    p[0] = 'final'

# These rules return a list of FormalParameters

def p_FormalParameters(p):
    '''FormalParameters : LPAREN RPAREN
                        | LPAREN FormalParameterList RPAREN'''
    if len(p) < 4:
        p[0] = []
    else:
        p[0] = p[2]

def p_FormalParameterList(p):
    '''FormalParameterList : FormalParameter
                           | FormalParameterList COMMA FormalParameter'''
    if len(p) < 4:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[3]]

# These rules return a FormalParameter

def p_FormalParameter_1(p):
    '''FormalParameter : TypeX IDENTIFIER'''
    p[0] = typedecl.FormalParameter([], p[1], p[2], _lines(p))

def p_FormalParameter_2(p):
    '''FormalParameter : ModifierList TypeX IDENTIFIER'''
    p[0] = typedecl.FormalParameter(p[1], p[2], p[3], _lines(p))

# These rules return an Expr

def p_ForCondition(p):
    '''ForCondition : SEMICOLON
                    | Expr SEMICOLON'''
    if len(p) < 3:
        p[0] = expr.BoolLiteralExpr(True, _lines(p))
    else:
        p[0] = p[1]

# These rules return a list of Exprs

def p_ForUpdate(p):
    '''ForUpdate : StatementExprList'''
    p[0] = p[1]

def p_StatementExprList(p):
    '''StatementExprList :
                         | StatementExprNonemptyList'''
    if len(p) < 2:
        p[0] = []
    else:
        p[0] = p[1]

def p_StatementExprNonemptyList(p):
    '''StatementExprNonemptyList : StatementExpr
                                 | StatementExprNonemptyList COMMA \
                                       StatementExpr'''
    if len(p) < 4:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[3]]

# These rules return a Statement

def p_ReturnStatement(p):
    '''ReturnStatement : RETURN SEMICOLON
                       | RETURN Expr SEMICOLON'''
    if len(p) < 4:
        expr = None
    else:
        expr = p[2]
    p[0] = statement.ReturnStatement(expr, _lines(p))

def p_BreakStatement(p):
    '''BreakStatement : BREAK SEMICOLON
                      | BREAK IDENTIFIER SEMICOLON'''
    if len(p) < 4:
        label = None
    else:
        label = p[2]
    p[0] = statement.BreakStatement(label, _lines(p))

def p_ContinueStatement(p):
    '''ContinueStatement : CONTINUE SEMICOLON
                         | CONTINUE IDENTIFIER SEMICOLON'''
    if len(p) < 4:
        label = None
    else:
        label = p[2]
    p[0] = statement.ContinueStatement(label, _lines(p))

# These rules return a list of LocalVarDeclarators

def p_LocalVarDeclaratorList(p):
    '''LocalVarDeclaratorList : LocalVarDeclarator
                              | LocalVarDeclaratorList COMMA \
                                    LocalVarDeclarator'''
    if len(p) < 4:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[3]]

# These rules return a LocalVarDeclarator

def p_LocalVarDeclarator(p):
    '''LocalVarDeclarator : IDENTIFIER
                          | IDENTIFIER EQ VarInitializer'''
    if len(p) < 4:
        p[0] = statement.LocalVarDeclarator(p[1], None, _lines(p))
    else:
        p[0] = statement.LocalVarDeclarator(p[1], p[3], _lines(p))

# These rules return a TypeSymbol

def p_VoidResultType(p):
    '''VoidResultType : VOID'''
    p[0] = names.PrimitiveTypeSymbol(mytypes.voidType, _lines(p))

def p_PrimitiveType_1(p):
    '''PrimitiveType : INT'''
    p[0] = names.PrimitiveTypeSymbol(mytypes.intType, _lines(p))

def p_PrimitiveType_2(p):
    '''PrimitiveType : BOOLEAN'''
    p[0] = names.PrimitiveTypeSymbol(mytypes.boolType, _lines(p))

# Grammar rules for names and named types are duplicated with suffixes
# X and Y, depending on the context, to prevent conflicts in SLR table
# generation.
#
# X  -  in type body declarations and after instanceof
# Y  -  in expressions

def p_TypeX(p):
    '''TypeX : PrimitiveType
             | ReferenceTypeX'''
    p[0] = p[1]

def p_ReferenceTypeX_1(p):
    '''ReferenceTypeX : NameX'''
    p[0] = names.NameTypeSymbol(p[1])

def p_ReferenceTypeX_2(p):
    '''ReferenceTypeX : ArrayTypeX'''
    p[0] = p[1]

def p_ArrayTypeX_1(p):
    '''ArrayTypeX : PrimitiveType LBRACKET RBRACKET
                  | ArrayTypeX LBRACKET RBRACKET'''
    p[0] = names.ArrayTypeSymbol(p[1], _lines(p))

def p_ArrayTypeX_2(p):
    '''ArrayTypeX : NameX LBRACKET RBRACKET'''
    p[0] = names.ArrayTypeSymbol(names.NameTypeSymbol(p[1]), _lines(p))

def p_TypeY(p):
    '''TypeY : PrimitiveType
             | ReferenceTypeY'''
    p_TypeX(p)                     # Delegate action to the other rule

def p_ReferenceTypeY_1(p):
    '''ReferenceTypeY : NameY'''
    p_ReferenceTypeX_1(p)          # Delegate action to the other rule

def p_ReferenceTypeY_2(p):
    '''ReferenceTypeY : ArrayTypeY'''
    p_ReferenceTypeX_2(p)          # Delegate action to the other rule

def p_ArrayTypeY_1(p):
    '''ArrayTypeY : PrimitiveType LBRACKET RBRACKET
                  | ArrayTypeY LBRACKET RBRACKET'''
    p_ArrayTypeX_1(p)              # Delegate action to the other rule

def p_ArrayTypeY_2(p):
    '''ArrayTypeY : NameY LBRACKET RBRACKET'''
    p_ArrayTypeX_2(p)              # Delegate action to the other rule

def p_NonArrayTypeY_1(p):
    '''NonArrayTypeY : PrimitiveType'''
    p[0] = p[1]

def p_NonArrayTypeY_2(p):
    '''NonArrayTypeY : NameY'''
    p[0] = names.NameTypeSymbol(p[1])


# These rules return a Name

def p_NameX_1(p):
    '''NameX : IDENTIFIER'''
    p[0] = names.SimpleName(p[1], _lines(p))

def p_NameX_2(p):
    '''NameX : NameX PERIOD IDENTIFIER'''
    p[0] = names.QualifiedName(p[1], p[3], _lines(p))

def p_NameY_1(p):
    '''NameY : IDENTIFIER'''
    p_NameX_1(p)                   # Delegate action to the other rule

def p_NameY_2(p):
    '''NameY : NameY PERIOD IDENTIFIER'''
    p_NameX_2(p)                   # Delegate action to the other rule

# These rules return a list of NoneType

def p_DimList(p):
    '''DimList :
               | DimList LBRACKET RBRACKET'''
    if len(p) < 4:
        p[0] = []
    else:
        p[0] = p[1] + [None]

# These rules return a list of VarInitializers

def p_VarInitializerList(p):
    '''VarInitializerList : CommaOpt
                          | VarInitializerNonemptyList CommaOpt'''
    if len(p) < 3:
        p[0] = []
    else:
        p[0] = p[1]

def p_VarInitializerNonemptyList(p):
    '''VarInitializerNonemptyList : VarInitializer
                                  | VarInitializerNonemptyList COMMA \
                                        VarInitializer'''
    if len(p) < 4:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[3]]

# These rules return a VarInitializer

def p_VarInitializer_1(p):
    '''VarInitializer : Expr'''
    p[0] = expr.ExprInitializer(p[1])

def p_VarInitializer_2(p):
    '''VarInitializer : ArrayInitializer'''
    p[0] = p[1]

def p_ArrayInitializer(p):
    '''ArrayInitializer : LBRACE VarInitializerList RBRACE'''
    p[0] = expr.ArrayInitializer(p[2], _lines(p))

# These rules return an Expr

def p_ArrayCreationExpr_1(p):
    '''ArrayCreationExpr : NEW NonArrayTypeY LBRACKET Expr RBRACKET DimList'''
    # Only one of the bracket pairs is allowed to contain an Expr (no
    # implicit creation of multidimensional arrays).
    p[0] = expr.ArrayCreationExpr(p[2], [p[4]], p[6], None, _lines(p))

def p_ArrayCreationExpr_2(p):
    '''ArrayCreationExpr : NEW NonArrayTypeY LBRACKET RBRACKET DimList \
                               ArrayInitializer'''
    p[0] = expr.ArrayCreationExpr(p[2], [], [None] + p[5], p[6], _lines(p))

def p_StatementExpr(p):
    '''StatementExpr : Assignment
                     | PreIncrementExpr
                     | PreDecrementExpr
                     | PostIncrementStatementExpr
                     | PostDecrementStatementExpr
                     | MethodInvocationExpr
                     | ClassInstanceCreationExpr'''
    p[0] = p[1]

def p_PrimaryExpr(p):
    '''PrimaryExpr : PrimaryNoCallExpr
                   | ArrayCreationExpr
                   | MethodInvocationExpr
                   | ClassInstanceCreationExpr'''
    p[0] = p[1]

def p_PrimaryNoNewArrayExpr(p):
    '''PrimaryNoNewArrayExpr : PrimaryNoCallExpr
                             | MethodInvocationExpr
                             | ClassInstanceCreationExpr'''
    p[0] = p[1]

def p_PrimaryNoCallExpr_1(p):
    '''PrimaryNoCallExpr : Literal
                         | FieldAccess
                         | ArrayAccess'''
    p[0] = p[1]

def p_PrimaryNoCallExpr_2(p):
    '''PrimaryNoCallExpr : THIS'''
    p[0] = expr.ThisExpr(_lines(p))

def p_PrimaryNoCallExpr_3(p):
    '''PrimaryNoCallExpr : LPAREN Expr RPAREN'''
    if _generateParenExpr:
        p[0] = expr.ParenExpr(p[2], _lines(p))
    else:
        p[0] = p[2]

def p_ClassInstanceCreationExpr(p):
    '''ClassInstanceCreationExpr : NEW NameY Arguments'''
    p[0] = expr.InstanceCreationExpr(p[2], p[3], _lines(p))

def p_Literal_1(p):
    '''Literal : INT_LITERAL'''
    p[0] = expr.IntLiteralExpr(p[1], _lines(p))

def p_Literal_2(p):
    '''Literal : INT_LITERAL_OVERFLOW'''
    _errorProxy.add('value ' + repr(p[1]) + ' is too large for an int',
                    _lines(p))
    p[0] = expr.Expr(_lines(p))  # Dummy expression

def p_Literal_3(p):
    '''Literal : FALSE'''
    p[0] = expr.BoolLiteralExpr(False, _lines(p))

def p_Literal_4(p):
    '''Literal : TRUE'''
    p[0] = expr.BoolLiteralExpr(True, _lines(p))

def p_Literal_5(p):
    '''Literal : NULL'''
    p[0] = expr.NullExpr(_lines(p))

def p_Literal_6(p):
    '''Literal : STRING_LITERAL'''
    p[0] = expr.StringLiteralExpr(p[1], _lines(p))

def p_FieldAccess_1(p):
    '''FieldAccess : PrimaryExpr PERIOD IDENTIFIER'''
    p[0] = expr.PrimaryFieldAccessExpr(p[1], p[3], _lines(p))

def p_FieldAccess_2(p):
    '''FieldAccess : SUPER PERIOD IDENTIFIER'''
    p[0] = expr.SuperFieldAccessExpr(p[3], _lines(p))

def p_ArrayAccess(p):
    '''ArrayAccess : NameY LBRACKET Expr RBRACKET
                   | PrimaryNoNewArrayExpr LBRACKET Expr RBRACKET'''
    p[0] = expr.ArrayAccessExpr(p[1], p[3], _lines(p))

def p_NameOrPrimaryExpr(p):
    '''NameOrPrimaryExpr : NameY
                         | PrimaryNoCallExpr
                         | ArrayCreationExpr'''
    p[0] = p[1]

# Postfix, PostIncrement and PostDecrement grammar symbols are
# duplicated to xxxInnerExpr and xxxStatementExpr.  The latter
# are used for statement expressions and the former for everything
# else.  This trick is needed to eliminate a reduce/reduce
# conflict in SLR parser generation.

def p_PostfixInnerExpr(p):
    '''PostfixInnerExpr : NameOrPrimaryExpr
                        | PostIncrementInnerExpr
                        | PostDecrementInnerExpr
                        | MethodInvocationExpr
                        | ClassInstanceCreationExpr'''
    p[0] = p[1]

def p_PostIncrementInnerExpr(p):
    '''PostIncrementInnerExpr : PostfixInnerExpr PLUSPLUS'''
    p[0] = expr.PostfixExpr(p[1], binaryop.Plus(p.linespan(2)))

def p_PostDecrementInnerExpr(p):
    '''PostDecrementInnerExpr : PostfixInnerExpr MINUSMINUS'''
    p[0] = expr.PostfixExpr(p[1], binaryop.Minus(p.linespan(2)))

def p_PostfixStatementExpr(p):
    '''PostfixStatementExpr : NameOrPrimaryExpr
                            | PostIncrementStatementExpr
                            | PostDecrementStatementExpr
                            | MethodInvocationExpr
                            | ClassInstanceCreationExpr'''
    p_PostfixInnerExpr(p)        # Delegate action to the other rule

def p_PostIncrementStatementExpr(p):
    '''PostIncrementStatementExpr : PostfixStatementExpr PLUSPLUS'''
    p_PostIncrementInnerExpr(p)  # Delegate action to the other rule

def p_PostDecrementStatementExpr(p):
    '''PostDecrementStatementExpr : PostfixStatementExpr MINUSMINUS'''
    p_PostDecrementInnerExpr(p)  # Delegate action to the other rule

def p_UnaryExpr_1(p):
    '''UnaryExpr : PLUS UnaryExpr'''
    p[0] = expr.UnaryPlusExpr(p[2], _lines(p))

def p_UnaryExpr_2(p):
    '''UnaryExpr : MINUS NegatableExpr'''
    p[0] = expr.UnaryMinusExpr(p[2], _lines(p))

def p_UnaryExpr_3(p):
    '''UnaryExpr : UnaryExprNotPlusMinus
                 | PreIncrementExpr
                 | PreDecrementExpr'''
    p[0] = p[1]

def p_NegatableExpr_1(p):
    '''NegatableExpr : UnaryExpr'''
    p[0] = p[1]

def p_NegatableExpr_2(p):
    '''NegatableExpr : INT_LITERAL_MUST_NEGATE'''
    p[0] = expr.IntLiteralExpr(p[1], _lines(p))

def p_PreIncrementExpr(p):
    '''PreIncrementExpr : PLUSPLUS UnaryExpr'''
    p[0] = expr.PrefixExpr(p[2], binaryop.Plus(p.linespan(1)))

def p_PreDecrementExpr(p):
    '''PreDecrementExpr : MINUSMINUS UnaryExpr'''
    p[0] = expr.PrefixExpr(p[2], binaryop.Minus(p.linespan(1)))

def p_UnaryExprNotPlusMinus_1(p):
    '''UnaryExprNotPlusMinus : PostfixInnerExpr
                             | CastExpr'''
    p[0] = p[1]

def p_UnaryExprNotPlusMinus_2(p):
    '''UnaryExprNotPlusMinus : TILDE UnaryExpr'''
    p[0] = expr.BitNotExpr(p[2], _lines(p))

def p_UnaryExprNotPlusMinus_3(p):
    '''UnaryExprNotPlusMinus : EXCLAM UnaryExpr'''
    p[0] = expr.LogicalNotExpr(p[2], _lines(p))

def p_CastExpr_1(p):
    '''CastExpr : LPAREN PrimitiveType RPAREN UnaryExpr
                | LPAREN ArrayTypeY RPAREN UnaryExpr'''
    p[0] = expr.TypeCastExpr(p[2], p[4], _lines(p))

def p_CastExpr_2(p):
    '''CastExpr : LPAREN Expr RPAREN UnaryExprNotPlusMinus'''
    # The Expr inside parentheses 
    p[0] = expr.NameCastExpr(p[2], p[4], _lines(p))

def p_MultiplicativeExpr_1(p):
    '''MultiplicativeExpr : UnaryExpr'''
    p[0] = p[1]

def p_MultiplicativeExpr_2(p):
    '''MultiplicativeExpr : MultiplicativeExpr ASTER UnaryExpr'''
    p[0] = expr.InfixExpr(p[1], p[3], binaryop.Multiply(p.linespan(2)))

def p_MultiplicativeExpr_3(p):
    '''MultiplicativeExpr : MultiplicativeExpr SLASH UnaryExpr'''
    p[0] = expr.InfixExpr(p[1], p[3], binaryop.Divide(p.linespan(2)))

def p_MultiplicativeExpr_4(p):
    '''MultiplicativeExpr : MultiplicativeExpr PERCENT UnaryExpr'''
    p[0] = expr.InfixExpr(p[1], p[3], binaryop.Remainder(p.linespan(2)))

def p_AdditiveExpr_1(p):
    '''AdditiveExpr : MultiplicativeExpr'''
    p[0] = p[1]

def p_AdditiveExpr_2(p):
    '''AdditiveExpr : AdditiveExpr PLUS MultiplicativeExpr'''
    p[0] = expr.InfixExpr(p[1], p[3], binaryop.Plus(p.linespan(2)))

def p_AdditiveExpr_3(p):
    '''AdditiveExpr : AdditiveExpr MINUS MultiplicativeExpr'''
    p[0] = expr.InfixExpr(p[1], p[3], binaryop.Minus(p.linespan(2)))

def p_ShiftExpr_1(p):
    '''ShiftExpr : AdditiveExpr'''
    p[0] = p[1]

def p_ShiftExpr_2(p):
    '''ShiftExpr : ShiftExpr LTLT AdditiveExpr'''
    p[0] = expr.InfixExpr(p[1], p[3], binaryop.ShiftLeft(p.linespan(2)))

def p_ShiftExpr_3(p):
    '''ShiftExpr : ShiftExpr GTGT AdditiveExpr'''
    p[0] = expr.InfixExpr(p[1], p[3], binaryop.SignedShiftRight(p.linespan(2)))

def p_ShiftExpr_4(p):
    '''ShiftExpr : ShiftExpr GTGTGT AdditiveExpr'''
    p[0] = expr.InfixExpr(p[1], p[3],
                          binaryop.UnsignedShiftRight(p.linespan(2)))

def p_RelationalExpr_1(p):
    '''RelationalExpr : ShiftExpr'''
    p[0] = p[1]

def p_RelationalExpr_2(p):
    '''RelationalExpr : RelationalExpr LT ShiftExpr'''
    p[0] = expr.InfixExpr(p[1], p[3], binaryop.LessThan(p.linespan(2)))

def p_RelationalExpr_3(p):
    '''RelationalExpr : RelationalExpr GT ShiftExpr'''
    p[0] = expr.InfixExpr(p[1], p[3], binaryop.GreaterThan(p.linespan(2)))

def p_RelationalExpr_4(p):
    '''RelationalExpr : RelationalExpr LTEQ ShiftExpr'''
    p[0] = expr.InfixExpr(p[1], p[3], binaryop.LessOrEqual(p.linespan(2)))

def p_RelationalExpr_5(p):
    '''RelationalExpr : RelationalExpr GTEQ ShiftExpr'''
    p[0] = expr.InfixExpr(p[1], p[3], binaryop.GreaterOrEqual(p.linespan(2)))

def p_RelationalExpr_6(p):
    '''RelationalExpr : RelationalExpr INSTANCEOF ReferenceTypeX'''
    p[0] = expr.InstanceOfExpr(p[1], p[3], _lines(p))

def p_EqualityExpr_1(p):
    '''EqualityExpr : RelationalExpr'''
    p[0] = p[1]

def p_EqualityExpr_2(p):
    '''EqualityExpr : EqualityExpr EQEQ RelationalExpr'''
    p[0] = expr.InfixExpr(p[1], p[3], binaryop.Equal(p.linespan(2)))

def p_EqualityExpr_3(p):
    '''EqualityExpr : EqualityExpr EXCLAMEQ RelationalExpr'''
    p[0] = expr.InfixExpr(p[1], p[3], binaryop.NotEqual(p.linespan(2)))

def p_AndExpr_1(p):
    '''AndExpr : EqualityExpr'''
    p[0] = p[1]

def p_AndExpr_2(p):
    '''AndExpr : AndExpr AMPER EqualityExpr'''
    p[0] = expr.InfixExpr(p[1], p[3], binaryop.BitLogicalAnd(p.linespan(2)))

def p_XorExpr_1(p):
    '''XorExpr : AndExpr'''
    p[0] = p[1]

def p_XorExpr_2(p):
    '''XorExpr : XorExpr CARET AndExpr'''
    p[0] = expr.InfixExpr(p[1], p[3], binaryop.BitLogicalXor(p.linespan(2)))

def p_OrExpr_1(p):
    '''OrExpr : XorExpr'''
    p[0] = p[1]

def p_OrExpr_2(p):
    '''OrExpr : OrExpr VERT XorExpr'''
    p[0] = expr.InfixExpr(p[1], p[3], binaryop.BitLogicalOr(p.linespan(2)))

def p_CondAndExpr_1(p):
    '''CondAndExpr : OrExpr'''
    p[0] = p[1]

def p_CondAndExpr_2(p):
    '''CondAndExpr : CondAndExpr AMPERAMPER OrExpr'''
    p[0] = expr.CondAndExpr(p[1], p[3])

def p_CondOrExpr_1(p):
    '''CondOrExpr : CondAndExpr'''
    p[0] = p[1]

def p_CondOrExpr_2(p):
    '''CondOrExpr : CondOrExpr VERTVERT CondAndExpr'''
    p[0] = expr.CondOrExpr(p[1], p[3])

def p_TernaryCondExpr_1(p):
    '''TernaryCondExpr : CondOrExpr'''
    p[0] = p[1]

def p_TernaryCondExpr_2(p):
    '''TernaryCondExpr : CondOrExpr QUESTION Expr COLON TernaryCondExpr'''
    p[0] = expr.TernaryCondExpr(p[1], p[3], p[5])

def p_AssignmentExpr(p):
    '''AssignmentExpr : TernaryCondExpr
                      | Assignment'''
    p[0] = p[1]

def p_Assignment_1(p):
    '''Assignment : LeftHandSide EQ AssignmentExpr'''
    p[0] = expr.SimpleAssignmentExpr(p[1], p[3], _lines(p))

def p_Assignment_2(p):
    '''Assignment : LeftHandSide CompoundAssignmentOperator AssignmentExpr'''
    p[0] = expr.CompoundAssignmentExpr(p[1], p[3], p[2], _lines(p))

def p_LeftHandSide(p):
    '''LeftHandSide : NameY
                    | FieldAccess
                    | ArrayAccess'''
    p[0] = p[1]

def p_Expr(p):
    '''Expr : AssignmentExpr'''
    p[0] = p[1]

def p_MethodInvocationExpr_1(p):
    '''MethodInvocationExpr : NameY Arguments'''
    p[0] = expr.NameMethodInvocationExpr(p[1], p[2], _lines(p))

def p_MethodInvocationExpr_2(p):
    '''MethodInvocationExpr : PrimaryExpr PERIOD IDENTIFIER Arguments'''
    p[0] = expr.PrimaryMethodInvocationExpr(p[1], p[3], p[4], _lines(p))

def p_MethodInvocationExpr_3(p):
    '''MethodInvocationExpr : SUPER PERIOD IDENTIFIER Arguments'''
    p[0] = expr.SuperMethodInvocationExpr(p[3], p[4], _lines(p))

# These rules return a list of Exprs

def p_Arguments(p):
    '''Arguments : LPAREN RPAREN
                 | LPAREN ArgumentList RPAREN'''
    if len(p) < 4:
        p[0] = []
    else:
        p[0] = p[2]

def p_ArgumentList(p):
    '''ArgumentList : Expr
                    | ArgumentList COMMA Expr'''
    if len(p) < 4:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[3]]

# These rules return a BinaryOperator

def p_CompoundAssignmentOperator_1(p):
    '''CompoundAssignmentOperator : ASTEREQ'''
    p[0] = binaryop.Multiply(_lines(p))

def p_CompoundAssignmentOperator_2(p):
    '''CompoundAssignmentOperator : SLASHEQ'''
    p[0] = binaryop.Divide(_lines(p))

def p_CompoundAssignmentOperator_3(p):
    '''CompoundAssignmentOperator : PERCENTEQ'''
    p[0] = binaryop.Remainder(_lines(p))

def p_CompoundAssignmentOperator_4(p):
    '''CompoundAssignmentOperator : PLUSEQ'''
    p[0] = binaryop.Plus(_lines(p))

def p_CompoundAssignmentOperator_5(p):
    '''CompoundAssignmentOperator : MINUSEQ'''
    p[0] = binaryop.Minus(_lines(p))

def p_CompoundAssignmentOperator_6(p):
    '''CompoundAssignmentOperator : LTLTEQ'''
    p[0] = binaryop.ShiftLeft(_lines(p))

def p_CompoundAssignmentOperator_7(p):
    '''CompoundAssignmentOperator : GTGTEQ'''
    p[0] = binaryop.SignedShiftRight(_lines(p))

def p_CompoundAssignmentOperator_8(p):
    '''CompoundAssignmentOperator : GTGTGTEQ'''
    p[0] = binaryop.UnsignedShiftRight(_lines(p))

def p_CompoundAssignmentOperator_9(p):
    '''CompoundAssignmentOperator : AMPEREQ'''
    p[0] = binaryop.BitLogicalAnd(_lines(p))

def p_CompoundAssignmentOperator_10(p):
    '''CompoundAssignmentOperator : CARETEQ'''
    p[0] = binaryop.BitLogicalXor(_lines(p))

def p_CompoundAssignmentOperator_11(p):
    '''CompoundAssignmentOperator : VERTEQ'''
    p[0] = binaryop.BitLogicalOr(_lines(p))

def p_error(t):
    """Handle unexpected tokens"""
    if t is None:
        # No more tokens
        _errorProxy.add('unexpected end of input', fatal=False)
        # Generate error token
        return
    elif t.type in lexical.reservedTokens:
        item = 'reserved word ' + repr(t.value)
    elif t.type in lexical.separatorTokens:
        item = 'separator ' + repr(t.value)
    elif t.type in lexical.operatorTokens:
        item = 'operator ' + repr(t.value)
    elif t.type == 'IDENTIFIER':
        item = 'identifier ' + repr(t.value)
    elif t.type in ('INT_LITERAL', 'INT_LITERAL_OVERFLOW'):
        item = 'integer ' + repr(t.value)
    elif t.type == 'INT_LITERAL_MUST_NEGATE':
        _errorProxy.add('value ' + str(abs(t.value)) +
                        ' may only appear after a minus sign',
                        (t.lineno, t.lineno))
        t.type = 'INT_LITERAL'
        # Recover by calling errok and returning a valid token
        ply.yacc.errok()
        return t
    elif t.type == 'CHARACTER_LITERAL':
        item = 'character literal ' + repr(t.value)
    elif t.type == 'STRING_LITERAL':
        item = 'string literal'
    else:
        assert False

    _errorProxy.add('unexpected ' + item, (t.lineno, t.lineno))
    # Generate error token by returning nothing


tokens = lexical.tokens

# Generate a parser using ply.yacc.  Use the directory of this module
# as the output directory for parse tables.  Also add that directory
# to sys.path so that the parse tables are found when yacc reads them
# using an import statement.  This way, yacc does not have to generate
# parse tables unnecessarily.
_myDir = os.path.dirname(__file__)
sys.path.append(_myDir)
_parser = ply.yacc.yacc(outputdir=_myDir, debug=0)
