#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# icode.py
#
# 2005-11-11  Jori Dubrovin:  Fixed a bug in IntSignedShiftRight
# 2005-08-26  Jori Dubrovin:  Added CheckAssert
# 2005-08-08  Jori Dubrovin:  Added virtual machine state to native calls
# 2005-06-17  Jori Dubrovin:  Fixed a typo
# 2005-06-03  Jori Dubrovin:  Initial version
#

"""The internal code language."""

from copy import copy
from sets import Set

from errors import UninitializedError, MissingReturnError
from errors import ZeroDivisionError_, NegativeShiftError, NullReferenceError
from errors import NegativeArraySizeError, ArrayIndexOutOfBoundsError
from errors import ArrayStoreError, ClassCastError, AssertionError_


# Integers 0 and 1 are used in internal code to represent boolean values.
boolValue = {False: 0, True: 1}


def _truncate32(i):
    """Make a value fit a 32-bit signed int.

    The integer i is truncated to a signed integer of given number of
    bits by discarding the most significant bits that do not fit.
    """
    i &= (1L<<32) - 1
    if i >= 1L<<31:
        i -= 1L<<32
    return int(i)


def buildRoutine(code):
    """Convert a list of instructions to an executable routine.

    Param code: nonempty list of internal code instructions.  These
    instructions are linked together to form the routine and cannot
    be reused to build another routine.

    Return value: entry point (Instruction) to the routine.
    """
    for i in xrange(len(code) - 1):
        code[i].setNext(code[i + 1])
    return code[0]


class _Uninitialized:
    """A class for the singleton uninitialized object."""

    def __str__(self):
        return '<uninitialized>'

uninitialized = _Uninitialized()


class _Null:
    """A class for the singleton null object."""

    def __str__(self):
        return 'null'

null = _Null()


class _Void:
    """A class for the singleton void object."""

    def __str__(self):
        return '<void>'

void = _Void()


class Instance:
    """A class representing object and array instances."""

    def __init__(self, type_):
        """Construct an instance.

        Param type_: the ReferenceType for this instance.
        """
        self.__type = type_
        self.__fields = {}
        self.__components = []   # Components are only used in arrays.

    def clone(self):
        """Return a clone of this instance.

        The clone has the same type, and its fields and array
        components refer to the same values as this instance.
        """
        other = Instance(self.__type)
        other.__fields = copy(self.__fields)
        other.__components = self.__components[:]
        return other

    def getType(self):
        """Return the ReferenceType of this instance."""
        return self.__type

    def getField(self, signature):
        """Return the value of the field with the given signature."""
        return self.__fields[signature]

    def setField(self, signature, value):
        """Set the value of the field with the given signature."""
        self.__fields[signature] = value

    def initializeComponents(self, count, value):
        """Initialize the component list.

        Param count: the number of elements.

        Param value: the value to be written to each component.
        """
        self.__components = [value for x in xrange(count)]

    def getComponent(self, index):
        """Return the value of the array component at the given index."""
        return self.__components[index]

    def setComponent(self, index, value):
        """Set the value of the array component at the given index."""
        self.__components[index] = value


class Instruction:
    """A base class for instructions."""

    def __init__(self):
        self._predecessors = []
        self._successors = []
        self.__labelName = None #$$

    def __str__(self):
        return self.indent('<instruction>')  # Dummy default

    def setNext(self, next):
        """Link this instruction to the next one in a list of instructions.

        Use this method to build a routine (a directed graph) from a
        list of unlinked instructions.

        Override this for instruction classes that do not always fall
        through to the next instruction when executed, e.g. goto,
        branch, return etc.

        Param next: the instruction that comes after this one in the list.
        """
        self._link(next)

    def _link(self, next):
        """Add a directed link from this to the given instruction."""
        self._successors.append(next)
        next._predecessors.append(self)

    __nextLabelNum = 0  #$$

    def listing(self):
        """Return the program listing as a string.""" #$$buggy?
        s = []
        queue = [self]
        queued = Set()
        queued.add(id(self))
        while queue:
            instr = queue.pop(0)
            if len(instr._predecessors) > 1 and instr.__labelName is None:
                #$$ Stupid hack
                Instruction.__nextLabelNum += 1
                instr.__labelName = 'L' + str(Instruction.__nextLabelNum)
                s.append(instr.__labelName + ':')
            s.append(str(instr))
            for i in instr._successors:
                if i not in queued:
                    queue.insert(0, i)
                    queued.add(i)
            if len(instr._successors) > 0:
                next = instr._successors[0]
                if len(queue) == 0 or queue[0] is not next:
                    s.append(self.indent('goto ' + str(next.__labelName)))
        return '\n'.join(s)

    def indent(s):
        """Indent all the lines of string s."""
        lines = s.split('\n')
        return '\n'.join(['  ' + ins for ins in lines])
    indent = staticmethod(indent)


class SourceRef(Instruction):
    """A pseudoinstruction that refers to a portion of source code text.

    Operand stack before:
    ...
    Operand stack after:
    ...
    """

    def __init__(self, syntaxNode):
        """Construct a source-reference.

        Param syntaxNode: a SyntaxNode of the AST that is being referred to.
        """
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('nop <source ref>')

    def execute(self, state):
        state.setPc(self._successors[0])


class Label(Instruction):
    """A label that marks a location in the code.

    This is a effectively a nop provided only for convenience when
    building lists of instructions.

    Operand stack before:
    ...
    Operand stack after:
    ...
    """

    def __init__(self):
        """Construct a label."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('nop <label>')

    def execute(self, state):
        state.setPc(self._successors[0])


class Goto(Instruction):
    """An instruction that jumps to a label unconditionally.

    Operand stack before:
    ...
    Operand stack after:
    ...
    """

    def __init__(self, label):
        """Construct a goto instruction.

        Param label: the destination Label.
        """
        Instruction.__init__(self)
        self.__destLabel = label

    def __str__(self):
        return self.indent('nop <goto>')

    def setNext(self, next):
        self._link(self.__destLabel)

    def execute(self, state):
        state.setPc(self.__destLabel)


class Pop(Instruction):
    """An instruction that discards the top value(s) of the operand stack.

    Operand stack before:
    ..., x
    Operand stack after:
    ...
    """

    def __init__(self):
        """Construct a pop instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('pop')

    def execute(self, state):
        state.popOperand()
        state.setPc(self._successors[0])


class Duplicate(Instruction):
    """An instruction that duplicates a value of the operand stack.

    Operand stack before:
    ..., x(1), ..., x(depth)
    Operand stack after:
    ..., x(1), ..., x(depth), x(1)
    """

    def __init__(self, depth=1):
        """Construct a duplicate instruction.

        Param count: the depth of the element to duplicate.
        """
        Instruction.__init__(self)
        self.__depth = depth

    def __str__(self):
        return self.indent('duplicate  ' + str(self.__depth))

    def execute(self, state):
        state.pushOperand(state.peekOperand(self.__depth))
        state.setPc(self._successors[0])


class Insert(Instruction):
    """An instruction that inserts a value to the operand stack.

    Operand stack before:
    ..., x(1), ..., x(skip-1), y
    Operand stack after:
    ..., y, x(1), ..., x(skip-1), y
    """

    def __init__(self, skip):
        """Construct an insert instruction.

        Param skip: the number of elements that will be above the new
        inserted value in the operand stack.
        """
        Instruction.__init__(self)
        self.__skip = skip

    def __str__(self):
        return self.indent('insert     ' + str(self.__skip))

    def execute(self, state):
        state.pushOperand(state.peekOperand(), self.__skip)
        state.setPc(self._successors[0])


class CallStatic(Instruction):
    """An instruction that calls a routine in a given class.

    Operand stack before:
    ..., x(1), ..., x(argCount)
    Operand stack after:
    ..., x(1), ..., x(argCount)
    Expected operand stack after routine returns:
    ..., return value
    """

    def __init__(self, containingType, signature, argCount):
        """Construct a call-static instruction.

        Param containingType: the ReferenceType that contains the
        routine.

        Param signature: a string containing the routine signature.

        Param argCount: number of arguments to the routine.
        """
        Instruction.__init__(self)
        self.__containingType = containingType
        self.__signature = signature
        self.__argCount = argCount

    def __str__(self):
        return self.indent('call static ' + str(self.__containingType) +
                           '.' + self.__signature +
                           ' (' + str(self.__argCount) + ' arguments)')

    def execute(self, state):
        state.setPc(self._successors[0])
        state.pushFrame()
        state.setPc(self.__containingType.getRoutine(self.__signature))


class CallDynamic(Instruction):
    """An instruction that calls a routine virtually.

    Operand stack before:
    ..., instance, x(2), ..., x(argCount)
    Operand stack after:
    ..., instance, x(2), ..., x(argCount)
    Expected operand stack after routine returns:
    ..., return value
    """

    def __init__(self, signature, argCount):
        """Construct a call-dynamic instruction.

        Param signature: a string containing the routine signature.

        Param argCount: number of arguments to the routine, including
        the 'this' reference.
        """
        Instruction.__init__(self)
        self.__signature = signature
        self.__argCount = argCount

    def __str__(self):
        return self.indent('call dynamic ' + str(self.__containingType) +
                           '.' + self.__signature +
                           ' (' + str(self.__argCount) + ' arguments)')

    def execute(self, state):
        state.setPc(self._successors[0])
        state.pushFrame()
        instance = state.peekOperand(self.__argCount)
        #$$create a new instruction for null testing...
        if instance is null:
            raise NullReferenceError('null reference in method invocation') #$$
        state.setPc(instance.getType().getRoutine(self.__signature))


class CallNative(Instruction):
    """An instruction that calls a Python function.

    Operand stack before:
    ..., x(1), ..., x(argCount)
    Operand stack after:
    ..., return value
    """

    def __init__(self, wrapper, argCount):
        """Construct a call-native instruction.

        Param wrapper: a NativeWrapper instance.

        Param argCount: number of arguments to the function.
        """
        Instruction.__init__(self)
        self.__wrapper = wrapper
        self.__argCount = argCount

    def __str__(self):
        return self.indent('call native ' + str(self.__wrapper) +
                           ' (' + str(self.__argCount) + ' arguments)')

    def execute(self, state):
        # Read arguments from operand stack.  Use peekOperand so that
        # this call is exception safe.
        args = [state.peekOperand(self.__argCount - argN) for
                argN in xrange(self.__argCount)]
        returnValue = self.__wrapper.execute(state, args)
        # Remove arguments.
        for argN in xrange(self.__argCount):
            state.popOperand()
        state.pushOperand(returnValue)
        state.setPc(self._successors[0])


class CheckNonNull(Instruction):
    """An instruction that checks that a value is not null.

    Raise a NullReferenceError upon execution if a value at a given
    position in the operand stack is null.

    Operand stack before:
    ..., value, x(2), ..., x(depth)
    Operand stack after:
    ..., value, x(2), ..., x(depth)
    """

    def __init__(self, depth):
        """Construct a check-non-null instruction."""
        Instruction.__init__(self)
        self.__depth = depth

    def __str__(self):
        return self.indent('check non-null ' + str(depth))

    def execute(self, state):
        value = state.peekOperand(self.__depth)
        if value is null:
            raise NullReferenceError('dereferencing a null reference')  #$$
        state.setPc(self._successors[0])


class CheckAssert(Instruction):
    """An instruction that checks for an assertion error.

    Remove the top value from the operand stack and raise an
    AssertionError_ if it was zero.

    Operand stack before:
    ..., integer x
    Operand stack after:
    ...
    """

    def __init__(self):
        """Construct a check-assert instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('check assert')

    def execute(self, state):
        value = state.popOperand()
        state.setPc(self._successors[0])
        if value == 0:
            raise AssertionError_('assertion error')  #$$


class BadCast(Instruction):
    """An instruction that raises a ClassCastError.

    Operand stack before:
    ...
    Operand stack after:
    ...
    """

    def __init__(self):
        """Construct a bad-cast instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('bad cast')

    def setNext(self, next):
        pass

    def execute(self, state):
        raise ClassCastError('incompatible run-time type')  #$$


class InstanceOf(Instruction):
    """An instruction that checks the type of the top operand.

    The top element must be an Instance of some type.

    Operand stack before:
    ..., instance
    Operand stack after:
    ..., integer 1 if instance is of (a subtype of) the given type, 0 otherwise
    """

    def __init__(self, type_):
        """Construct an instance-of instruction."""
        Instruction.__init__(self)
        self.__type = type_

    def __str__(self):
        return self.indent('instance of ' + str(self.__type))

    def execute(self, state):
        instance = state.popOperand()
        iType = instance.getType()
        state.pushOperand(boolValue[iType.isSubtypeOf(self.__type)])
        state.setPc(self._successors[0])


class NewInstance(Instruction):
    """An instruction that creates a class instance.

    The new instance does not have any fields assigned.

    Operand stack before:
    ...
    Operand stack after:
    ..., instance
    """

    def __init__(self, type_):
        """Construct a new-instance instruction.

        Param type_: the ReferenceType for the instance.
        """
        Instruction.__init__(self)
        self.__type = type_

    def __str__(self):
        return self.indent('new instance ' + str(self.__type))

    def execute(self, state):
        state.pushOperand(Instance(self.__type))
        state.setPc(self._successors[0])


class InitArray(Instruction):
    """Initialize an array.

    Set the component list of the array to contain 'count' copies of
    'value'.

    Operand stack before:
    ..., array instance, integer count, value
    Operand stack after:
    ...
    """

    def __init__(self):
        """Construct an init-array instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('init array')

    def execute(self, state):
        value = state.popOperand()
        count = state.popOperand()
        array = state.popOperand()
        array.initializeComponents(count, value)
        state.setPc(self._successors[0])


class NegativeArraySize(Instruction):
    """An instruction that raises a NegativeArraySizeError.

    Operand stack before:
    ...
    Operand stack after:
    ...
    """

    def __init__(self):
        """Construct a negative-array-size instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('negative array size')

    def setNext(self, next):
        pass

    def execute(self, state):
        raise NegativeArraySizeError('array size cannot be negative')  #$$


class BadArrayIndex(Instruction):
    """An instruction that raises an ArrayIndexOutOfBoundsError.

    Operand stack before:
    ...
    Operand stack after:
    ...
    """

    def __init__(self):
        """Construct a bad-array-index instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('bad array index')

    def setNext(self, next):
        pass

    def execute(self, state):
        raise ArrayIndexOutOfBoundsError('array index out of bounds')  #$$


class Halt(Instruction):
    """An instruction that halts the execution.

    Operand stack before:
    ...
    Operand stack after:
    ...
    """

    def __init__(self):
        """Construct a halt instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('halt')

    def setNext(self, next):
        pass

    def execute(self, state):
        state.halt()


class Return(Instruction):
    """An instruction that returns from a routine.

    This same instruction is used to return a value from the operand stack,
    and to return void (value void is expected in the operand stack).

    Operand stack before:
    ..., return value
    Operand stack after:
    ..., return value
    """

    def __init__(self):
        """Construct a return instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('return')

    def setNext(self, next):
        pass

    def execute(self, state):
        state.popFrame()


class MissingReturn(Instruction):
    """An instruction that raises a MissingReturnError.

    Operand stack before:
    ...
    Operand stack after:
    ...
    """

    def __init__(self):
        """Construct a missing-return instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('missing return')

    def setNext(self, next):
        pass

    def execute(self, state):
        raise MissingReturnError('method must return a value')  #$$


class IfGoto(Instruction):
    """An instruction that jumps to a label conditionally.

    If the integer value on top op the operand stack is nonzero,
    execution jumps to the given label.  Otherwise execution falls
    through to the next statement.

    Operand stack before:
    ..., integer condition
    Operand stack after:
    ...
    """

    def __init__(self, label):
        """Construct an if-goto instruction

        Param label: the destination Label.
        """
        Instruction.__init__(self)
        self.__destLabel = label

    def __str__(self):
        return self.indent('if goto    ') #$$

    def setNext(self, next):
        self._link(next)
        self._link(self.__destLabel)

    def execute(self, state):
        if state.popOperand() != 0:
            state.setPc(self._successors[1])
        else:
            state.setPc(self._successors[0])


class Literal(Instruction):
    """An instruction that pushes a literal to the operand stack.

    Operand stack before:
    ...
    Operand stack after:
    ..., literal value
    """

    def __init__(self, value):
        """Construct a literal instruction.

        Param value: the value to be pushed.
        """
        Instruction.__init__(self)
        self.__value = value

    def __str__(self):
        return self.indent('literal    ' + str(self.__value))

    def execute(self, state):
        state.pushOperand(self.__value)
        state.setPc(self._successors[0])


class Delete(Instruction):
    """An instruction that deletes an automatic variable.

    Operand stack before:
    ...
    Operand stack after:
    ...
    """

    def __init__(self, varName):
        """Construct a delete instruction.

        Param varName: a string containing the name of the variable.
        """
        Instruction.__init__(self)
        self.__varName = varName

    def __str__(self):
        return self.indent('delete     ' + self.__varName)

    def execute(self, state):
        state.deleteAuto(self.__varName)
        state.setPc(self._successors[0])


class LoadAuto(Instruction):
    """An instruction that retrieves the value of an automatic variable.

    Operand stack before:
    ...
    Operand stack after:
    ..., value
    """

    def __init__(self, varName):
        """Construct a load-auto instruction.

        Param varName: name of the variable.
        """
        Instruction.__init__(self)
        self.__varName = varName

    def __str__(self):
        return self.indent('load       ' + self.__varName)

    def execute(self, state):
        value = state.getAuto(self.__varName)
        if value is uninitialized:
            raise UninitializedError('uninitialized local variable') #$$
        state.pushOperand(value)
        state.setPc(self._successors[0])


class StoreAuto(Instruction):
    """An instruction that assigns a value to an automatic variable.

    Operand stack before:
    ..., value
    Operand stack after:
    ...
    """

    def __init__(self, varName):
        """Construct a store-auto instruction.

        Param varName: name of the destination variable.
        """
        Instruction.__init__(self)
        self.__varName = varName

    def __str__(self):
        return self.indent('store      ' + self.__varName)

    def execute(self, state):
        state.assignAuto(self.__varName, state.popOperand())
        state.setPc(self._successors[0])


class LoadArray(Instruction):
    """An instruction that retrieves a component from an array.

    Operand stack before:
    ..., array instance, integer index
    Operand stack after:
    ..., value
    """

    def __init__(self):
        """Construct a load-array instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('load array')

    def execute(self, state):
        index = state.popOperand()
        array = state.popOperand()
        value = array.getComponent(index)
        state.pushOperand(value)
        state.setPc(self._successors[0])


class StoreArray(Instruction):
    """An instruction that assigns a value to an array component.

    Operand stack before:
    ..., array instance, integer index, value
    Operand stack after:
    ...
    """

    def __init__(self):
        """Construct a store-array instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('store array')

    def execute(self, state):
        value = state.popOperand()
        index = state.popOperand()
        array = state.popOperand()
        array.setComponent(index, value)
        state.setPc(self._successors[0])


class CheckArrayStore(Instruction):
    """An instruction that raises an ArrayStoreError if necessary.

    The top value must be an Instance or null.  If the type of the
    value is not a subtype of the component type of the array
    instance, an ArrayStoreError is raised upon execution.

    Operand stack before:
    ..., array instance, index, value
    Operand stack after:
    ..., array instance, index, value
    """

    def __init__(self):
        """Construct a check-array-store instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('check array store')

    def execute(self, state):
        value = state.peekOperand(1)
        array = state.peekOperand(3)
        compType = array.getType().getComponentType()
        if (value is not null and not value.getType().isSubtypeOf(compType)):
            raise ArrayStoreError('bad component type')  #$$
        state.setPc(self._successors[0])


class LoadField(Instruction):
    """An instruction that retrieves the value of an instance variable.

    Operand stack before:
    ..., instance
    Operand stack after:
    ..., value
    """

    def __init__(self, signature):
        """Construct a load-field instruction.

        Param signature: signature of the field as a string.
        """
        Instruction.__init__(self)
        self.__signature = signature

    def __str__(self):
        return self.indent('load field  ' + self.__signature)

    def execute(self, state):
        instance = state.popOperand()
        #$$create a new instruction for null testing...
        if instance is null:
            raise NullReferenceError('null reference in field access') #$$
        value = instance.getField(self.__signature)
        state.pushOperand(value)
        state.setPc(self._successors[0])


class StoreField(Instruction):
    """An instruction that assigns the value of an instance variable.

    Operand stack before:
    ..., instance, value
    Operand stack after:
    ...
    """

    def __init__(self, signature):
        """Construct a store-field instruction.

        Param signature: signature of the field as a string.
        """
        Instruction.__init__(self)
        self.__signature = signature

    def __str__(self):
        return self.indent('store field ' + self.__signature)

    def execute(self, state):
        value = state.popOperand()
        instance = state.popOperand()
        #$$create a new instruction for null testing...
        if instance is null:
            raise NullReferenceError('null reference in field access') #$$
        instance.setField(self.__signature, value)
        state.setPc(self._successors[0])


class LoadStatic(Instruction):
    """An instruction that retrieves the value of a static variable.

    Operand stack before:
    ...
    Operand stack after:
    ..., value
    """

    def __init__(self, signature):
        """Construct a load-static instruction.

        Param signature: signature of the variable as a string.
        """
        Instruction.__init__(self)
        self.__signature = signature

    def __str__(self):
        return self.indent('load static ' + self.__signature)

    def execute(self, state):
        value = state.getStaticVar(self.__signature)
        state.pushOperand(value)
        state.setPc(self._successors[0])


class StoreStatic(Instruction):
    """An instruction that assigns the value of a static variable.

    Operand stack before:
    ..., value
    Operand stack after:
    ...
    """

    def __init__(self, signature):
        """Construct a store-static instruction.

        Param signature: signature of the variable as a string.
        """
        Instruction.__init__(self)
        self.__signature = signature

    def __str__(self):
        return self.indent('store static ' + self.__signature)

    def execute(self, state):
        value = state.popOperand()
        state.assignStaticVar(self.__signature, value)
        state.setPc(self._successors[0])


class LogicalNot(Instruction):
    """An instruction that logically negates an integer.

    Operand stack before:
    ..., integer x = 0 or 1
    Operand stack after:
    ..., integer 1 - x
    """

    def __init__(self):
        """Construct a logical-not instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('logical not')

    def execute(self, state):
        state.pushOperand(_truncate32(state.popOperand() ^ 1))
        state.setPc(self._successors[0])


class IntNegate(Instruction):
    """An instruction that swaps the sign of an integer.

    Operand stack before:
    ..., integer x
    Operand stack after:
    ..., integer -x
    """

    def __init__(self):
        """Construct an integer-negate instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('negate')

    def execute(self, state):
        state.pushOperand(_truncate32(-state.popOperand()))
        state.setPc(self._successors[0])


class IntMultiply(Instruction):
    """An instruction that multiplies two integers.

    Operand stack before:
    ..., integer x, integer y
    Operand stack after:
    ..., ingeter x * y
    """

    def __init__(self):
        """Construct an integer-multiply instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('multiply')

    def execute(self, state):
        state.pushOperand(_truncate32(state.popOperand() * state.popOperand()))
        state.setPc(self._successors[0])


class IntDivide(Instruction):
    """An instruction that divides two integers.

    Operand stack before:
    ..., integer x, integer y
    Operand stack after:
    ..., integer x / y
    """

    def __init__(self):
        """Construct an integer-divide instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('divide')

    def execute(self, state):
        y = state.popOperand()
        x = state.popOperand()
        if y == 0:
            raise ZeroDivisionError_('division by zero')
        if (x < 0) == (y < 0):
            value = x // y  # Use floor division //
        else:
            # Unlike Python, we round negative results towards zero
            value = -(x // -y)
        state.pushOperand(_truncate32(value))
        state.setPc(self._successors[0])


class IntRemainder(Instruction):
    """An instruction that divides two integers and stores the remainder.

    Operand stack before:
    ..., integer x, integer y
    Operand stack after:
    ..., integer x % y
    """

    def __init__(self):
        """Construct an integer-remainder instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('remainder')

    def execute(self, state):
        y = state.popOperand()
        x = state.popOperand()
        if y == 0:
            raise ZeroDivisionError_('modulo by zero')
        if (x < 0) == (y < 0):
            value = x % y
        else:
            # In Python, x % y has the same sign as y.  We want
            # it to have the same sign as x.
            value = x % -y
        state.pushOperand(_truncate32(value))
        state.setPc(self._successors[0])


class IntPlus(Instruction):
    """An instruction that adds two integers.

    Operand stack before:
    ..., integer x, integer y
    Operand stack after:
    ..., integer x + y
    """

    def __init__(self):
        """Construct an integer-plus instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('plus')

    def execute(self, state):
        state.pushOperand(_truncate32(state.popOperand() + state.popOperand()))
        state.setPc(self._successors[0])


class IntShiftLeft(Instruction):
    """An instruction that shifts an integer left.

    Operand stack before:
    ..., integer x, integer y
    Operand stack after:
    ..., integer x << y
    """

    def __init__(self):
        """Construct an integer-shift-left instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('shift left')

    def execute(self, state):
        y = state.popOperand()
        x = state.popOperand()
        if y < 0:
            raise NegativeShiftError('left shift by negative amount')
        if y >= 32:
            value = 0  # Overflow
        else:
            value = long(x) << y
        state.pushOperand(_truncate32(value))
        state.setPc(self._successors[0])


class IntSignedShiftRight(Instruction):
    """An instruction that shifts a integer right arithmetically.

    Operand stack before:
    ..., integer x, integer y
    Operand stack after:
    ..., integer x >> y
    """

    def __init__(self):
        """Construct an integer-signed-shift-right instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('signed shift right')

    def execute(self, state):
        y = state.popOperand()
        x = state.popOperand()
        if y < 0:
            raise NegativeShiftError('right shift by negative amount')
        if y >= 32:
            # Overflow
            if x < 0:
                value = -1
            else:
                value = 0
        else:
            value = x >> y
        state.pushOperand(value)
        state.setPc(self._successors[0])


class IntUnsignedShiftRight(Instruction):
    """An instruction that shifts a integer right logically.

    Operand stack before:
    ..., integer x, integer y
    Operand stack after:
    ..., integer x >>> y
    """

    def __init__(self):
        """Construct an integer-unsigned-shift-right instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('unsigned shift right')

    def execute(self, state):
        y = state.popOperand()
        x = state.popOperand()
        if y < 0:
            raise NegativeShiftError('unsigned right shift by negative amount')
        if y >= 32:
            value = 0  # Overflow
        else:
            if x < 0:
                x += 1L << 32
            value = x >> y
        state.pushOperand(_truncate32(value))
        state.setPc(self._successors[0])


class IntAnd(Instruction):
    """An instruction that computes the bitwise and of two integers.

    Operand stack before:
    ..., integer x, integer y
    Operand stack after:
    ..., integer x & y
    """

    def __init__(self):
        """Construct an integer-and instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('and')

    def execute(self, state):
        state.pushOperand(_truncate32(state.popOperand() & state.popOperand()))
        state.setPc(self._successors[0])


class IntXor(Instruction):
    """An instruction that computes the exclusive or of two integers.

    Operand stack before:
    ..., integer x, integer y
    Operand stack after:
    ..., integer x ^ y
    """

    def __init__(self):
        """Construct an integer-xor instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('xor')

    def execute(self, state):
        state.pushOperand(_truncate32(state.popOperand() ^ state.popOperand()))
        state.setPc(self._successors[0])


class IntOr(Instruction):
    """An instruction that computes the inclusive or of two integers.

    Operand stack before:
    ..., integer x, integer y
    Operand stack after:
    ..., integer x | y
    """

    def __init__(self):
        """Construct an integer-and instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('or')

    def execute(self, state):
        state.pushOperand(_truncate32(state.popOperand() | state.popOperand()))
        state.setPc(self._successors[0])


class IntLessThan(Instruction):
    """An instruction that computes the value of the less-than operator.

    Operand stack before:
    ..., integer x, integer y
    Operand stack after:
    ..., integer 1 if x < y, 0 otherwise
    """

    def __init__(self):
        """Construct an integer-less-than instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('less than')

    def execute(self, state):
        y = state.popOperand()
        x = state.popOperand()
        state.pushOperand(boolValue[x < y])
        state.setPc(self._successors[0])


class IntLessOrEqual(Instruction):
    """An instruction that computes the value of the less-or-equal operator.

    Operand stack before:
    ..., integer x, integer y
    Operand stack after:
    ..., integer 1 if x <= y, 0 otherwise
    """

    def __init__(self):
        """Construct an integer-less-or-equal instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('less or equal')

    def execute(self, state):
        y = state.popOperand()
        x = state.popOperand()
        state.pushOperand(boolValue[x <= y])
        state.setPc(self._successors[0])


class IntEqual(Instruction):
    """An instruction that tests if two integers are equal.

    Operand stack before:
    ..., integer x, integer y
    Operand stack after:
    ..., integer 1 if x == y, 0 otherwise
    """

    def __init__(self):
        """Construct an integer-equal instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('int equal')

    def execute(self, state):
        state.pushOperand(boolValue[state.popOperand() == state.popOperand()])
        state.setPc(self._successors[0])


class RefEqual(Instruction):
    """An instruction that tests if two references are equal.

    The references may be Instances or null.

    Operand stack before:
    ..., reference x, reference y
    Operand stack after:
    ..., integer 1 if x and y are the same object, 0 otherwise
    """

    def __init__(self):
        """Construct an reference-equal instruction."""
        Instruction.__init__(self)

    def __str__(self):
        return self.indent('ref equal')

    def execute(self, state):
        state.pushOperand(boolValue[state.popOperand() is state.popOperand()])
        state.setPc(self._successors[0])
