#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# lexical.py
#
# 2007-06-08  Jori Dubrovin:  Test branch for api_recurseStr
# 2005-09-01  Jori Dubrovin:  Encapsulating PLY
# 2005-08-26  Jori Dubrovin:  Added keyword assert
# 2005-07-27  Jori Dubrovin:  Added keywords
# 2005-06-16  Jori Dubrovin:  Added keywords send and to
# 2005-04-29  Jori Dubrovin:  Added some keywords
# 2005-04-22  Jori Dubrovin:  Initial version
#

"""Lexical structure for the action language.

Use the lexer function to obtain a Lexer object, which can break an
input string into tokens.  A LexicalError is raised if an invalid token
is discovered.  Lexer is defined in the lex module of PLY (Python
Lex-Yacc).

The runmain function of the lex module is invoked if this is run as a
script.
"""

import copy
import re
import ply.lex

import errors


def lexer():
    """Return a new copy of the Lexer object."""
    return copy.copy(_lexer)


# Reserved words (keywords and named literals)
reservedWords = [
    'abstract',
    'assert',
    'boolean',
    'break',
    'case',
    'class',
    'continue',
    'default',
    'do',
    'else',
    'enum',
    'extends',
    'false',
    'final',
    'for',
    'if',
    'implements',
    'instanceof',
    'int',
    'interface',
    'native',
    'new',
    'null',
    'return',
    'send',
    'static',
    'super',
    'switch',
    'this',
    'to',
    'true',
    'void',
    'while',
    ]
reservedTokens = [s.upper() for s in reservedWords]

# Separators
separatorTokens = [
    'LPAREN', 'RPAREN', 'LBRACE', 'RBRACE', 'LBRACKET', 'RBRACKET',
    'SEMICOLON', 'COMMA', 'PERIOD',
    ]
t_LPAREN = r'\('
t_RPAREN = r'\)'
t_LBRACE = r'{'
t_RBRACE = r'}'
t_LBRACKET = r'\['
t_RBRACKET = r'\]'
t_SEMICOLON = r';'
t_COMMA = r','
t_PERIOD = r'\.'

# Operators
operatorTokens = [
    'EQ', 'GT', 'LT', 'EXCLAM', 'TILDE', 'QUESTION', 'COLON',
    'EQEQ', 'LTEQ', 'GTEQ', 'EXCLAMEQ',
    'AMPERAMPER', 'VERTVERT', 'PLUSPLUS', 'MINUSMINUS',
    'PLUS', 'MINUS', 'ASTER', 'SLASH', 'AMPER', 'VERT',
    'CARET', 'PERCENT', 'LTLT', 'GTGT', 'GTGTGT',
    'PLUSEQ', 'MINUSEQ', 'ASTEREQ', 'SLASHEQ', 'AMPEREQ', 'VERTEQ',
    'CARETEQ', 'PERCENTEQ', 'LTLTEQ', 'GTGTEQ', 'GTGTGTEQ',
    ]
t_EQ = r'='
t_GT = r'>'
t_LT = r'<'
t_EXCLAM = r'!'
t_TILDE = r'~'
t_QUESTION = r'\?'
t_COLON = r':'
t_EQEQ = r'=='
t_LTEQ = r'<='
t_GTEQ = r'>='
t_EXCLAMEQ = r'!='
t_AMPERAMPER = r'&&'
t_VERTVERT = r'\|\|'
t_PLUSPLUS = r'\+\+'
t_MINUSMINUS = r'--'
t_PLUS = r'\+'
t_MINUS = r'-'
t_ASTER = r'\*'
t_SLASH = r'/'
t_AMPER = r'&'
t_VERT = r'\|'
t_CARET = r'\^'
t_PERCENT = r'%'
t_LTLT = r'<<'
t_GTGT = r'>>'
t_GTGTGT = r'>>>'
t_PLUSEQ = r'\+='
t_MINUSEQ = r'-='
t_ASTEREQ = r'\*='
t_SLASHEQ = r'/='
t_AMPEREQ = r'&='
t_VERTEQ = r'\|='
t_CARETEQ = r'\^='
t_PERCENTEQ = r'%='
t_LTLTEQ = r'<<='
t_GTGTEQ = r'>>='
t_GTGTGTEQ = r'>>>='

# Tokens must be defined as a global variable for the lex module
tokens = reservedTokens + separatorTokens + operatorTokens + [
    'IDENTIFIER', 'STRING_LITERAL',
    'INT_LITERAL', 'INT_LITERAL_MUST_NEGATE', 'INT_LITERAL_OVERFLOW',
    ]

# Another special variable for the lex module.
# Ignore spaces, tabs and form feeds.
t_ignore = ' \t\f'

def t_traditionalComment(t):
    # Match C-style comments, i.e. /* followed by 0 or more characters
    # including newlines, ending with the first */.  Nested comments
    # are not allowed.
    r'/\*(.|\n)*?\*/'
    t.lexer.lineno += t.value.count('\n')
    # Return nothing so that no token is produced.

def t_endOfLineComment(t):
    # Match C++-style comments, which start with // and end with a newline.
    r'//.*?\n'
    t.lexer.lineno += 1
    # Return nothing so that no token is produced.

# Unterminated comments must be checked after terminated ones.
def t_unterminatedTraditionalComment(t):
    # Match C-style comments that lack the terminating */.
    # This produces a static error.
    r'/\*(.|\n)*'
    oldLineno = t.lexer.lineno
    t.lexer.lineno += t.value.count('\n')
    errors.ErrorProxy().finish('unterminated comment',
                               (oldLineno, t.lexer.lineno), fatal=False)

def t_unterminatedEndOfLineComment(t):
    # Match C++-style comments at the end of input (no terminating newline).
    # This is accepted as a comment.
    r'//.*'
    # Return nothing so that no token is produced.

def t_word(t):
    # Match valid identifiers and reserved words.
    r'[A-Za-z\$_][0-9A-Za-z\$_]*'
    if t.value in reservedWords:
        t.type = reservedTokens[reservedWords.index(t.value)]
    else:
        t.type = 'IDENTIFIER'
    return t

def _handleHexOct(t, base):
    """Handle integer literals given in other than decimal format.

    Param t: the original token.  This will be modified.
    Param base: the number base in which the value is interpreted.
    Return value: a token of type INT_LITERAL or INT_LITERAL_OVERFLOW
    """
    num = int(t.value, base)  # Convert to int or long
    if num < 1L<<32:
        # Fits in 32 bits.  Handle as a signed integer.
        t.type = 'INT_LITERAL'
        if num >= 1L<<31:
            num -= 1L<<32
        t.value = int(num)
    else:
        t.type = 'INT_LITERAL_OVERFLOW'
        # Keep t.value a string
    return t

def t_hexadecimal(t):
    # Match a hexadecimal integer with the prefix 0X or 0x.
    r'0[Xx][0-9A-Fa-f]+'
    return _handleHexOct(t, 16)

def t_octal(t):
    # Match an octal integer with the prefix 0.
    r'0[0-7]+'
    return _handleHexOct(t, 8)

# Decimal integers must be checked after hexadecimal and octal
# so that the prefix 0 or 0x is not interpreted as a decimal zero.
def t_decimal(t):
    # Match a nonnegative decimal integer.
    r'(0|[1-9][0-9]*)'
    num = int(t.value)  # Convert to int or long
    if num < 1L<<31:
        # Nonnegative int
        t.type = 'INT_LITERAL'
        t.value = num
    elif num == 1L<<31:
        # This value may only appear immediately after a minus sign
        t.type = 'INT_LITERAL_MUST_NEGATE'
        t.value = int(-num)
    else:
        t.type = 'INT_LITERAL_OVERFLOW'
        # Keep t.value a string
    return t

def _evalString(s):
    """Evaluate a string literal.

    Param s: representation of a string literal enclosed in single or
    double quotes, respectively.

    Return value: the string whose representation is s.
    """
    # Just use the Python interpreter to evaluate the contents.
    # This is possible because Python recognizes all the escape
    # sequences that we allow.
    # The following hack is needed because Python evaluates the string
    # "\400" to a single null character instead of two characters
    # '\040' and '0', as we'd like.

    # Replace the null character by the escape sequence '\000',
    # because the Python eval function does not like the null
    # character.
    s = re.sub(r'\0', r'\\000', s)
    # Replace occurrences of '\\' temporarily by '\-'.  This is needed
    # to prevent the misinterpretation of the substring '\4' in '\\4'.
    s = re.sub(r'\\\\', r'\\-', s)
    # Replace '\4' by '\04' to make sure that an octal number
    # beginning with 4 has at most two significant digits.  Similarly
    # for 5, 6, and 7.
    s = re.sub(r'\\4', r'\\04', s)
    s = re.sub(r'\\5', r'\\05', s)
    s = re.sub(r'\\6', r'\\06', s)
    s = re.sub(r'\\7', r'\\07', s)
    # Convert '\-' back to '\\' and evaluate the result.
    s = re.sub(r'\\-', r'\\\\', s)
    return eval(s)

def t_STRING_LITERAL(t):
    # Matches a string literal, i.e. zero or more characters
    # enclosed in double quotes.  A character is one of
    # - x, where x is a character other than newline, " or \
    # - \x, where x is one of b, t, n, f, r, ", ', or \
    # - \ooo, where ooo is an octal number between 0 and 377
    r'''"([^\n"\\]|\\[btnfr"'\\]|\\[0-3]?[0-7]{1,2})*"'''
    t.value = _evalString(t.value)
    return t

def t_stringLiteralError(t):
    # Matches a double quote that does not begin a valid string literal.
    r'''"'''
    errors.ErrorProxy().finish('malformed string literal',
                               (t.lineno, t.lineno))

def t_error(t):
    """Handle lexical errors in the input by raising a LexicalError."""
    # Take the rest of the input line (up to 10 characters) and store it
    # as the context.
    context = t.value.split('\n', 1)[0]
    if len(context) > 10:
        context = context[:10] + '...'
    errors.ErrorProxy().finish("unexpected character " + repr(t.value[0]) +
                               " in '" + context + "'", (t.lineno, t.lineno))

def t_newline(t):
    # Consume linefeeds but keep track of the line number
    r'\n+'
    t.lexer.lineno += len(t.value)


# Cache a global Lexer object.
_lexer = ply.lex.lex()

if __name__ == '__main__':
    ply.lex.runmain()
