#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# machine.py
#
# 2005-07-23  Jori Dubrovin:  Added static variables
# 2005-06-17  Jori Dubrovin:  Improved __str__
# 2005-06-03  Jori Dubrovin:  Initial version
#

"""The virtual machine."""

from copy import copy


class VirtualMachine:

    class StackFrame:
        
        def __init__(self):
            """Construct an empty stack frame"""
            self.autoValues = {}
            self.pc = None

    def __init__(self):
        """Construct a virtual machine."""
        self.__callStack = [self.StackFrame()]
        self.__operandStack = []
        self.__staticVars = {}
        self.__halted = True

    def __str__(self):
        s = []
        s.append('Call stack:')
        for f in self.__callStack:
            s.append('')
            if not f.autoValues:
                loc = 'None'
            else:
                loc = ', '.join([name + '=' + str(f.autoValues[name]) for
                                 name in f.autoValues])
            s.append('  Locals: ' + loc)
            s.append('  Program counter: ' + str(f.pc))
        s.append('')
        s.append('Operand stack:')
        if not self.__operandStack:
            s.append('<empty>')
        else:
            s.append(', '.join([str(e) for e in self.__operandStack]))
        s.append('')
        s.append('Static variables:')
        if not self.__staticVars:
            s.append('<empty>')
        else:
            for v in self.__staticVars:
                s.append(v + '=' + str(self.__staticVars[v]))
        return '\n'.join(s)

    def initialize(self, entry):
        self.setPc(entry)
        self.__halted = False

    def setPc(self, instr):
        self.__callStack[-1].pc = instr

    def autoValuation(self):
        return copy(self.__callStack[-1].autoValues)

    def assignAuto(self, varName, value):
        """Assign an automatic variable in the topmost stack frame."""
        self.__callStack[-1].autoValues[varName] = value

    def getAuto(self, varName):
        """Return the value of an automatic variable.

        The variable must exist in the topmost stack frame.

        Param varName: string containing the name of the variable.

        Return value: value of the variable.
        """
        return self.__callStack[-1].autoValues[varName]

    def deleteAuto(self, varName):
        """Remove an automatic variable from the topmost stack frame."""
        del self.__callStack[-1].autoValues[varName]

    def assignStaticVar(self, varName, value):
        """Assign a static variable."""
        self.__staticVars[varName] = value

    def getStaticVar(self, varName):
        """Return the value of a static variable.

        The variable must exist.

        Param varName: string containing the name of the variable.

        Return value: value of the variable.
        """
        return self.__staticVars[varName]

    def popOperand(self):
        """Remove and return the topmost element of the operand stack."""
        return self.__operandStack.pop()

    def peekOperand(self, depth=1):
        """Return an element of the operand stack.

        The operand stack is not changed by this operation.

        Param depth: position of the element in the operand stack.  1
        means the topmost element, 2 the second topmost, and so on.

        Return value: the value of the element.
        """
        return self.__operandStack[-depth]

    def pushOperand(self, value, depth=0):
        """Push an element to the operand stack.

        Param value: the value to be pushed.

        Param depth: position of the element before which the new
        value is inserted.  If depth is 0, the value is pushed on top
        of the stack.  If depth is 1, the value is pushed between the
        topmost and the second topmost element, and so on.
        """
        self.__operandStack.insert(len(self.__operandStack)-depth, value)

    def pushFrame(self):
        """Push a new empty frame on top of the call stack."""
        self.__callStack.append(self.StackFrame())

    def popFrame(self):
        """Remove the topmost frame from the call stack."""
        self.__callStack.pop()

    def halt(self):
        self.__halted = True

    def step(self):
        """Execute one code instruction."""
        if self.__halted:
            raise RuntimeError('VirtualMachine cannot run ' +
                               'because it is halted')
        try:
            self.__callStack[-1].pc.execute(self)
        except:
            self.halt()
            raise

    def run(self):
        while not self.__halted:
            self.step()
