#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# mytypes.py
#
# 2007-06-10  Jori Dubrovin:  Adding recurseStr API
# 2007-06-06  Jori Dubrovin:  Changed string conversion of negative ints
# 2005-08-11  Jori Dubrovin:  Improved structure
# 2005-06-14  Jori Dubrovin:  Improved coding style
# 2005-04-26  Jori Dubrovin:  Changed value_to_string for int
# 2005-04-22  Jori Dubrovin:  Initial version
#

"""The type system for the action language."""

from sets import Set

from icode import CallStatic, Return, buildRoutine, boolValue, null


class ResultType:
    """A base class for types and the void pseudotype.

    All ResultTypes are supposed to created as singletons, so they can
    be compared using the operator is.
    """

    def __init__(self):
        pass

    def __str__(self):  #pragma: no cover
        return '<type>'

    def getCanonicalName(self):  #pragma: no cover
        """Return the canonical name for this result type as a string."""
        raise NotImplementedError('Pure virtual method')

    def isCastableTo(self, other):
        """Return true if self is castable to the other ResultType.

        This is a symmetric and transitive relation.  This would also
        be reflexive if voidType could be cast to itself.
        """
        if self is voidType:
            return False
        if self.isSubtypeOf(other) or other.isSubtypeOf(self):
            return True
        if (self.isReference() and not self.isArray() and
            other.isReference() and not other.isArray()):
            if self.isInterface() and not other.isFinal():
                return True
            if other.isInterface() and not self.isFinal():
                return True
        if self.isArray() and other.isArray():
            compType = self.getComponentType()
            return compType.isCastableTo(other.getComponentType())
        return False

    def isSubtypeOf(self, other):  #pragma: no cover
        """Return True if self a subtype of the other ResultType."""
        raise NotImplementedError('Pure virtual method')

    def isReturnSubstitutableFor(self, other):
        """Return True if self is return-type substitutable for the other."""
        # If primitive types could be subtypes of one another, we
        # would need to explicitly return False for two distinct
        # primitive types.
        return self.isSubtypeOf(other)

    def isReference(self):
        """Return True if this type is a reference type."""
        return False  # Default implementation

    def isArray(self):
        """Return True if this type is an array type."""
        return False  # Default implementation

    def isEnum(self):
        """Return True if this type is an enum type.

        The abstract class Enum is not an enum type.
        """
        return False  # Default implementation


class Type(ResultType):
    """A base class for types."""

    def __init__(self):
        ResultType.__init__(self)
        self.__arrayType = None

    def getDefaultValue(self):  #pragma: no cover
        """Return the default run-time value for this type."""
        raise NotImplementedError('Pure virtual method')

    def getArrayType(self, env):
        """Return the ArrayType whose component type self is.

        Param env: a StaticEnvironment (used for finding the class Object).
        """
        if self.__arrayType is None:
            self.__arrayType = ArrayType(self, env)
        return self.__arrayType


class PrimitiveType(Type):
    """A base class for primitive types."""

    def __init__(self):
        Type.__init__(self)

    def isSubtypeOf(self, other):
        return (self is other)

    def valueToString(self, value):  #pragma: no cover
        return '<value>'


class _VoidType(ResultType):
    """The void pseudotype."""

    def __init__(self):
        ResultType.__init__(self)

    def __str__(self):
        return 'void'

    def getCanonicalName(self):
        return 'void'

    def isSubtypeOf(self, other):
        return (other is voidType)

voidType = _VoidType()


class _IntType(PrimitiveType):
    """The int type."""

    def __init__(self):
        PrimitiveType.__init__(self)

    def __str__(self):
        return self.getCanonicalName()

    def valueToString(self, value):
        if value >= 0:
            return str(value)
        else:
            # Return negative values in hexadecimal to avoid problems
            # with the minus sign.
            return hex((1L<<32) + value).rstrip('L')

    def getDefaultValue(self):
        return 0

    def getCanonicalName(self):
        return 'int'

intType = _IntType()


class _BoolType(PrimitiveType):
    """The boolean type."""

    def __init__(self):
        PrimitiveType.__init__(self)

    def __str__(self):
        return self.getCanonicalName()

    def valueToString(self, value):
        if value == False:
            return 'false'
        elif value == True:
            return 'true'
        else:
            return '<boolean>'

    def getDefaultValue(self):
        return boolValue[False]

    def getCanonicalName(self):
        return 'boolean'

boolType = _BoolType()


class _NullType(Type):
    """The null type."""

    def __init__(self):
        Type.__init__(self)

    def __str__(self):
        return '<null type>'

    def isSubtypeOf(self, other):
        return (other is nullType or other.isReference())

    def getDefaultValue(self):  #pragma: no cover
        return null

    def getCanonicalName(self):
        raise RuntimeError('null type has no canonical name')

nullType = _NullType()


class ReferenceType(Type):
    """A base class for reference types."""

    def __init__(self):
        """Construct a reference type."""
        Type.__init__(self)
        self.__memberTypes = []
        self.__fields = []
        self.__functions = []
        self.__routines = {}

    def isReference(self):
        return True

    def getDefaultValue(self):
        return null

    def getMemberTypes(self, name):
        """Return a list of member types with the given simple name."""
        return [t for t in self.__memberTypes if t.getSimpleName() == name]

    def getAllMemberTypes(self):
        """Return a list of all member types (SingleReferenceTypes)."""
        return self.__memberTypes[:]

    def addMemberType(self, type_):
        """Add a reference type as a member to this type.

        Param type_: the SingleReferenceType to add.
        """
        self.__memberTypes.append(type_)

    def getFields(self, name):
        """Return a list of all fields with the given simple name."""
        return [f for f in self.__fields if f.getSimpleName() == name]

    def getAllFields(self):
        """Return a list of all fields."""
        return self.__fields[:]

    def addField(self, field):
        """Add a Field as a member to this type."""
        self.__fields.append(field)

    def getExactFunctions(self, name, types):
        """Return a list of functions with the given attributes.

        The function name, parameter arity, and parameter types must
        match exactly.

        Param name: string containing the simple name for the function.

        Param types: list of Types for the parameters.

        Return value: list of Methods.
        """
        found = []
        for f in self.__functions:
            fTypes = [t for (n, t) in f.getParameters()]
            if (f.getSimpleName() == name and len(fTypes) == len(types)):
                for (fType, givenType) in zip(fTypes, types):
                    if fType is not givenType:
                        break
                else:
                    found.append(f)
        return found

    def getFunctions(self, name):
        """Return a list of all functions with the given name."""
        return [f for f in self.__functions if f.getSimpleName() == name]

    def getAllFunctions(self):
        """Return a list of all functions in this class."""
        return self.__functions[:]

    def addFunction(self, function):
        """Add a Function to this type."""
        self.__functions.append(function)

    def setRoutine(self, signature, routine):
        """Bind a routine.

        Param signature: a string containing the signature used as a key.

        Param routine: the entry Instruction for the routine.
        """
        self.__routines[signature] = routine

    def getRoutine(self, signature):
        """Return the entry Instruction to the routine with given signature."""
        return self.__routines[signature]


class SingleReferenceType(ReferenceType):
    """A user-defined class, enum, or interface type."""

    def __init__(self, name, enclosingName, kind):
        """Construct a non-array reference type.

        Param name: string containing the simple name for the type.

        Param enclosingName: string containing the canonical name for
        the enclosing type, or an empty string if this is a top-level
        type.

        Param kind: one of 'concrete class', 'final class',
        'abstract class', 'interface', or 'enum'.
        """
        ReferenceType.__init__(self)
        assert(kind in ['concrete class', 'final class', 'abstract class',
                        'interface', 'enum'])
        self.__simpleName = name
        if enclosingName:
            self.__canonicalName = enclosingName + '.' + name
        else:
            self.__canonicalName = name
        self.__isAbstract = (kind in ['abstract class', 'interface'])
        self.__isInterface = (kind == 'interface')
        self.__isEnum = (kind == 'enum')
        self.__isFinal = (kind in ['enum', 'final class'])
        self.__directSupertypes = []
        self.__transitiveSupertypes = Set([self])
        self.__isReferable = False

    def __str__(self):
        if self.__isInterface:
            s = 'interface '
        elif self.__isEnum:
            s = 'enum '
        else:
            s = 'class '
        return s + self.__canonicalName

    def getCanonicalName(self):
        return self.__canonicalName

    def isEnum(self):
        return self.__isEnum

    def isFinal(self):
        return self.__isFinal

    def isAbstract(self):
        return self.__isAbstract

    def isInterface(self):
        return self.__isInterface

    def getSimpleName(self):
        return self.__simpleName

    def isSubtypeOf(self, other):
        return (other in self.__transitiveSupertypes)

    def getDirectSupertypes(self):
        """Return a list of direct supertypes (SingleReferenceTypes)."""
        return self.__directSupertypes[:]

    def setDirectSupertypes(self, supertypes):
        """Set the list of direct supertypes.

        Param supertypes: a list of SingleReferenceTypes.  They must
        already have their direct supertypes set.
        """
        self.__directSupertypes = supertypes[:]
        self.__transitiveSupertypes = Set([self])
        for s in supertypes:
            self.__transitiveSupertypes.update(s.__transitiveSupertypes)

    def setReferable(self):
        """Set the referable state of this type."""
        self.__isReferable = True

    def isReferable(self):
        """Return True if this type is referable.

        A type is referable when it has all member types added.  This
        is significant when declaring subtypes of reference types.
        """
        return self.__isReferable


class ArrayType(ReferenceType):
    """An array type."""

    def __init__(self, componentType, env):
        """Construct an array type.

        Only Type.getArrayType is allowed to call this constructor.

        Param componentType: the Type for array components.

        Param env: the StaticEnvironment in which the type is created.
        """
        ReferenceType.__init__(self)
        self.__componentType = componentType
        self.__cloneableInterface = env.getTopType('Cloneable')
        self.addField(Field('length', intType, '[]',
                            isStatic=False, isFinal=True))
        # Inherit all functions from Object, except 'clone'.
        objectClass = env.getTopType('Object')
        for f in objectClass.getAllFunctions():
            if f.getSimpleName() != '#constructor':
                if f.getSimpleName() == 'clone':
                    assert not f.getParameters()
                    # Make another clone method that just calls
                    # Object.clone, but whose return type is this
                    # array type.
                    m = Function('clone', [], self)
                    code = [CallStatic(objectClass, f.getSignature(), 1),
                            Return()]
                    m.setRoutine(buildRoutine(code))
                else:
                    m = f
                self.addFunction(m)
                self.setRoutine(m.getSignature(), m.getRoutine())

    def __str__(self):
        return str(self.__componentType) + '[]'

    def getCanonicalName(self):
        return self.__componentType.getCanonicalName() + '[]'

    def isArray(self):
        return True

    def isSubtypeOf(self, other):
        if self is other:
            return True
        if self.__cloneableInterface.isSubtypeOf(other):
            return True
        if self.__componentType.isReference() and other.isArray():
            return self.__componentType.isSubtypeOf(other.__componentType)
        return False

    def getComponentType(self):
        """Return the Type of array components."""
        return self.__componentType


class Field:
    """A field (member variable)."""

    def __init__(self, name, type_, enclosingName, isStatic, isFinal):
        """Construct a field.

        Param name: string containing the simple name for the variable.

        Param type_: the Type of the variable.

        Param enclosingName: string describing the class or interface
        in which the variable is declared.

        Param isStatic: True if this is a static variable.

        Param isFinal: True if this is a final variable.
        """
        self.__name = name
        self.__type = type_
        self.__signature = enclosingName + '.' + name
        self.__isStatic = isStatic
        self.__isFinal = isFinal

    def __str__(self):
        s = ''
        if self.__isFinal:
            s += 'final '
        if self.__isStatic:
            s += 'static '
        s += self.__type.getCanonicalName() + ' ' + self.__signature
        return s

    def getSimpleName(self):
        return self.__name

    def getType(self):
        return self.__type

    def getSignature(self):
        return self.__signature

    def isStatic(self):
        return self.__isStatic

    def isFinal(self):
        return self.__isFinal


class Function:
    """A function, i.e. a method or a constructor."""

    def __init__(self, name, params, returnType, isAbstract=False,
                 isStatic=False, isFinal=False, isNative=False):
        """Construct a function.

        Param name: a string containing the name of the function.  If
        name is '#constructor', this is a constructor.  Otherwise this
        is a method.

        Param params: a list of tuples (name, type), containing the
        parameter names and Types.

        Param returnType: the type returned by the method, or voidType.

        Param isAbstract: True if this is an abstract method.

        Param isStatic: True if this is a static method.

        Param isFinal: True if this is a final method.  Constructors
        are not final even though they cannot be overridden.

        Param isNative: True if this is a native method.
        """
        self.__name = name
        self.__parameters = params
        self.__returnType = returnType
        self.__isAbstract = isAbstract
        self.__isStatic = isStatic
        self.__isFinal = isFinal
        self.__isNative = isNative
        self.__routine = None

    def __str__(self):
        if self.__name == '#constructor':
            s = '<constructor>'
        else:
            s = ''
            if self.__isAbstract:
                s += 'abstract '
            if self.__isStatic:
                s += 'static '
            if self.__isFinal:
                s += 'final '
            if self.__isNative:
                s += 'native '
            s += self.__returnType.getCanonicalName() + ' ' + self.__name
        s += '('
        s += ', '.join([t.getCanonicalName() + ' ' + n for
                        (n, t) in self.__parameters])
        s += ')'
        return s

    def isAbstract(self):
        return self.__isAbstract

    def isStatic(self):
        return self.__isStatic

    def isFinal(self):
        return self.__isFinal

    def isNative(self):
        return self.__isNative

    def getSimpleName(self):
        return self.__name

    def getReturnType(self):
        return self.__returnType

    def getParameters(self):
        return self.__parameters[:]

    def getRoutine(self):
        return self.__routine

    def setRoutine(self, routine):
        self.__routine = routine

    def getSignature(self):
        return '#'.join([self.__name] + [t.getCanonicalName() for
                                         (n, t) in self.__parameters])
