#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# names.py
#
# 2007-06-10  Jori Dubrovin:  Adding recurseStr API
# 2006-07-30  Jori Dubrovin:  Added QualifiedName.api_getQualifyingName
# 2006-01-31  Jori Dubrovin:  Added final local variables
# 2005-11-09  Jori Dubrovin:  Added super field access and super method calls
# 2005-09-14  Jori Dubrovin:  Added API functions
# 2005-08-08  Jori Dubrovin:  Made Name a subclass of LeftHandSideExpr
# 2005-06-14  Jori Dubrovin:  Improved coding style
# 2005-06-06  Jori Dubrovin:  Initial version
#

"""Syntactic representation of names and types in the action language."""

from syntaxtree import SyntaxNode
from errors import ErrorProxy
from result import VarResult, ValueResult
from icode import LoadAuto, StoreAuto
from expr import LeftHandSideExpr, translateField


class TypeSymbol(SyntaxNode):
    """Base class for a source representation of a result type."""

    def __init__(self, linespan):
        SyntaxNode.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):  #pragma: no cover
        return '<type>'

    def resolve(self, env):  #pragma: no cover
        """Resolve the result type.

        Raise a StaticError if resolution fails.

        Param env: a StaticEnvironment in which the type symbol appears.

        Return value: the resolved ResultType.
        """
        raise NotImplementedError('Pure virtual method')


class PrimitiveTypeSymbol(TypeSymbol):
    """Source code representation of a primitive type or void."""

    def __init__(self, type_, linespan):
        """Construct a primitive type symbol.

        Param type_: the PrimitiveType or voidType.

        Param linespan: see SyntaxNode.
        """
        TypeSymbol.__init__(self, linespan)
        self.__type = type_

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return str(self.__type)

    def api_getType(self):
        """Return the represented type as a string.

        The return value is one of 'void', 'int', or 'boolean'.
        """
        return str(self.__type)

    def resolve(self, env):
        return self.__type


class NameTypeSymbol(TypeSymbol):
    """A class or interface type represented by a name."""

    def __init__(self, name):
        """Construct a name type symbol.

        Param name: the Name used in the representation.
        """
        TypeSymbol.__init__(self, name.linespan())
        self.__name = name

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return strFunc(self.__name, *args, **kwargs)

    def api_getName(self):
        """Return the associated Name.

        The return value is not a string, but an instance of class Name.
        """
        return self.__name

    def resolve(self, env):
        """Resolve the reference type.

        Raise a StaticError if resolution fails.

        Param env: a StaticEnvironment in which the type symbol appears.

        Return value: the resolved SingleReferenceType.
        """
        return self.__name.resolveAsType(env)


class ArrayTypeSymbol(TypeSymbol):
    """Source code representation of an array type."""

    def __init__(self, componentTypeSymbol, linespan):
        """Construct an array type symbol.

        Param componentTypeSymbol: a TypeSymbol representing the
        component type.

        Param linespan: see SyntaxNode.
        """
        TypeSymbol.__init__(self, linespan)
        self.__componentTypeSymbol = componentTypeSymbol

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return strFunc(self.__componentTypeSymbol, *args, **kwargs) + '[]'

    def api_getComponent(self):
        """Return the TypeSymbol representing the components of this type.

        The component type symbol may also be an ArrayTypeSymbol.
        """
        return self.__componentTypeSymbol

    def resolve(self, env):
        return self.__componentTypeSymbol.resolve(env).getArrayType(env)


class Name(LeftHandSideExpr):
    """A base class for names.

    A name can be a type name, an ambiguous name, or a method name.

    An ambiguous name is a simple of qualified name in the source code
    that might be a type name or an expression name.  It is
    reclassified as an expression name whenever possible.
    """

    def __init__(self, linespan):
        LeftHandSideExpr.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):  #pragma: no cover
        return '<name>'

    def api_getParts(self):  #pragma: no cover
        """Return a list containing the simple parts that make up this name.

        If self is a simple name, the list contains just one string.
        If self is a qualified name of the form 's1.s2.s3. ... .sn',
        the returned list contains the strings s1, s2, ..., sn.
        """
        raise NotImplementedError('Pure virtual method')

    def api_getLastPart(self):  #pragma: no cover
        """Return the last part of the name as a string.

        If self is 's1.s2.s3. ... .sn', then the string sn is
        returned.  This is equivalent to self.api_getParts()[-1].
        """
        raise NotImplementedError('Pure virtual method')

    def translate(self, env):  #pragma: no cover
        """Attempt to reclassify this name as an expression.

        Raise a StaticError if static checking fails.

        Param env: a StaticEnvironment in which the name appears.

        Return value: a StaticResult that represents the expression,
        which is necessarily a field.
        """
        raise NotImplementedError('Pure virtual method')

    def resolveAsType(self, env):  #pragma: no cover
        """Resolve the name as a reference type.

        Raise a StaticError if resolution fails.

        Param env: a StaticEnvironment in which the name appears.

        Return value: the resolved SingleReferenceType.
        """
        raise NotImplementedError('Pure virtual method')

    def resolveAsExprOrType(self, env):  #pragma: no cover
        """Resolve the name as an expression or reference type.

        Raise a StaticError if resolution fails.

        Param env: a StaticEnvironment in which the name appears.

        Return value: a pair (result, type).  If self can be resolved
        as an expression, result is the corresponding StaticResult and
        type is None.  Otherwise if self can be resolved as a type,
        result is None and type is the resolved SingleReferenceType.
        Otherwise a StaticError is raised.
        """
        raise NotImplementedError('Pure virtual method')

    def resolveAsMethod(self, env):  #pragma: no cover
        """Resolve the name as a method.

        Raise a StaticError if static checking fails.

        Param env: a StaticEnvironment in which the invocation appears.

        Return value: a triple (containingType, simpleName,
        targetCode), where containingType is the ReferenceType whose
        member the method should be, simpleName is a string containing
        the simple name of the method being invoked, and targetCode is
        a list of instructions that evaluate the target reference for
        the method invocation.  If there is no target reference (the
        method must be a static method), targetCode is None.
        """
        raise NotImplementedError('Pure virtual method')


class SimpleName(Name):
    """An unqualified name."""

    def __init__(self, identifier, linespan):
        """Construct a simple name.

        Param identifier: string containing the simple name used in the
        representation.

        Param linespan: see SyntaxNode.
        """
        Name.__init__(self, linespan)
        self.__identifier = identifier

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return self.__identifier

    def api_getParts(self):
        return [self.__identifier]

    def api_getLastPart(self):
        return self.__identifier

    def __tryAsExpr(self, env):
        if env.isLocalVar(self.__identifier):
            type_ = env.getLocalVarType(self.__identifier)
            if env.isLocalVarFinal(self.__identifier):
                return ValueResult(type_, [LoadAuto(self.__identifier)])
            else:
                return VarResult(type_, [], 0,
                                 [LoadAuto(self.__identifier)],
                                 [StoreAuto(self.__identifier)],
                                 [StoreAuto(self.__identifier)])
        # Traverse enclosing type definitions, starting from the innermost one.
        directlyEnclosing = True
        for c in env.getEnclosingTypes()[::-1]:
            if directlyEnclosing and env.isInstanceContext():
                instanceCode = [LoadAuto('this')]
            else:
                instanceCode = None
            result = translateField(c, self.__identifier, instanceCode, self)
            if result is not None:
                return result
            directlyEnclosing = False
        return None

    def __tryAsType(self, env):
        # Traverse enclosing type definitions, starting from the innermost one.
        found = None
        for c in env.getEnclosingTypes()[::-1]:
            if c.getSimpleName() == self.__identifier:
                found = c
                break
            types = c.getMemberTypes(self.__identifier)
            if types:
                if len(types) > 1:
                    ErrorProxy().finish("ambiguous type '" +
                                        self.__identifier + "'",
                                        self.linespan())
                found = types[0]
                break
        else:
            if env.isTopType(self.__identifier):
                found = env.getTopType(self.__identifier)
        if found is not None:
            if not found.isReferable():
                ErrorProxy().finish("illegal reference to type '"
                                    + self.__identifier + "'",
                                    self.linespan())
        return found

    def translate(self, env):
        result = self.__tryAsExpr(env)
        if result is not None:
            return result
        ErrorProxy().finish("undeclared variable '" + self.__identifier + "'",
                            self.linespan())

    def resolveAsType(self, env):
        type_ = self.__tryAsType(env)
        if type_ is not None:
            return type_
        ErrorProxy().finish("unresolved type '" + self.__identifier + "'",
                            self.linespan(), fatal=False)

    def resolveAsExprOrType(self, env):
        result = self.__tryAsExpr(env)
        if result is not None:
            return (result, None)
        type_ = self.__tryAsType(env)
        if type_ is not None:
            return (None, type_)
        ErrorProxy().finish("unresolved name '" + self.__identifier + "'",
                            self.linespan(), fatal=False)

    def resolveAsMethod(self, env):
        # Traverse enclosing type definitions, starting from the innermost one.
        directlyEnclosing = True
        for c in env.getEnclosingTypes()[::-1]:
            methods = c.getFunctions(self.__identifier)
            if methods:
                if env.isInstanceContext() and directlyEnclosing:
                    # Use 'this' as the target reference.
                    targetCode = [LoadAuto('this')]
                else:
                    targetCode = None
                return (c, self.__identifier, targetCode)
            directlyEnclosing = False
        ErrorProxy().finish("undefined method '" + self.__identifier + "'",
                            self.linespan())


class QualifiedName(Name):
    """A qualified name."""

    def __init__(self, outerName, identifier, linespan):
        """Construct a qualified name.

        Param outerName: the Name appearing to the left of the period.

        Param identifier: string containing the simple name to the
        right of the period.

        Param linespan: see SyntaxNode.
        """
        Name.__init__(self, linespan)
        self.__outerName = outerName
        self.__identifier = identifier

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return (strFunc(self.__outerName, *args, **kwargs) + '.' +
                self.__identifier)

    def api_getParts(self):
        return self.__outerName.api_getParts() + [self.__identifier]

    def api_getLastPart(self):
        return self.__identifier

    def api_getQualifyingName(self):
        """Returns the qualifying Name.

        If this name is 'IDN. ... ID3.ID2.ID1', then the qualifying
        name is 'IDN. ... .ID3.ID2'.
        """
        return self.__outerName

    def __tryAsExpr(self, outerResult, outerType):
        if outerResult is not None:
            return translateField(outerResult.getType(), self.__identifier,
                                  outerResult.getEvalCode(), self)
        else:
            return translateField(outerType, self.__identifier, None, self)

    def __tryAsType(self, outerType):
        types = outerType.getMemberTypes(self.__identifier)
        if types:
            if len(types) > 1:
                ErrorProxy().finish("ambiguous type '" + str(self) + "'",
                                    self.linespan())
            assert types[0].isReferable()
            return types[0]
        return None

    def translate(self, env):
        (outerResult, outerType) = self.__outerName.resolveAsExprOrType(env)
        result = self.__tryAsExpr(outerResult, outerType)
        if result is not None:
            return result
        ErrorProxy().finish("undeclared variable '" + self.__identifier + "'",
                            self.linespan())

    def resolveAsType(self, env):
        outerType = self.__outerName.resolveAsType(env)
        type_ = self.__tryAsType(outerType)
        if type_ is not None:
            return type_
        ErrorProxy().finish("'" + self.__identifier +
                            "' is not a member type of " + str(outerType),
                            self.linespan())

    def resolveAsExprOrType(self, env):
        (outerResult, outerType) = self.__outerName.resolveAsExprOrType(env)
        result = self.__tryAsExpr(outerResult, outerType)
        if result is not None:
            return (result, None)
        if outerType is not None:
            type_ = self.__tryAsType(outerType)
            if type_ is not None:
                return (None, type_)
        ErrorProxy().finish("unresolved name '" + self.__identifier + "'",
                            self.linespan())

    def resolveAsMethod(self, env):
        (outerResult, outerType) = self.__outerName.resolveAsExprOrType(env)
        if outerResult is not None:
            # The left part is an expression.
            return (outerResult.getType(), self.__identifier,
                    outerResult.getEvalCode())
        # The left part is a type name.
        return (outerType, self.__identifier, None)
