#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# result.py
#
# 2005-08-01  Jori Dubrovin:  Adding unchecked store code to var results
# 2005-06-14  Jori Dubrovin:  Improved coding style
# 2005-04-25  Jori Dubrovin:  Changed handling of void
# 2005-04-22  Jori Dubrovin:  Initial version
#

"""Classes for handling results of expressions in the action language."""

from mytypes import voidType


class StaticResult:
    """A base class for static results.

    A static result contains the information required to evaluate an
    expression and to interpret its value.

    All results carry a ResultType, which may be a real type or the
    void pseudotype.

    The evaluation code of a result is a list of internal code
    instructions that evaluate the expression and push its value on
    top of the operand stack.  A void result pushes void to the
    operand stack.

    Because code instructions are mutable and StaticResult does not
    make copies of the code, the code returned by getEvalCode can
    only be used once to build a routine.
    """

    def __init__(self, type_, evaluationCode):
        self.__type = type_
        self.__evaluationCode = evaluationCode

    def getType(self):
        """Return the ResultType of the result."""
        return self.__type

    def isVar(self):
        """Return True if this is a variable result."""
        return False

    def getEvalCode(self):
        """Return a reference to the evaluation code."""
        return self.__evaluationCode


class VoidResult(StaticResult):
    """The static result of a function returning void."""

    def __init__(self, evaluationCode):
        StaticResult.__init__(self, voidType, evaluationCode)


class ValueResult(StaticResult):
    """A result that carries a value."""

    def __init__(self, type_, evaluationCode):
        """Construct a value result.

        Param type_: the ResultType of this result.

        Param evaluationCode: a list of instructions that compute the value
        and place it on top of the operand stack.
        """
        StaticResult.__init__(self, type_, evaluationCode)


class VarResult(StaticResult):
    """A result that represents a variable that can be assigned to.

    Use a variable result in one of two ways.

    1. Like any StaticResult by calling evaluationCode and appending
       code that consumes the resulting value from the operand stack.

    or

    2. Call getIndirectionCode to obtain instructions that place zero
       or more indirection elements on top of the operand stack.
       These elements can be saved by a Duplicate instruction, if
       necessary.  To get the value of the result, call getLoadCode,
       which returns instructions that consume the indirection
       elements from top of the operand stack and push the value of
       the expression to the stack.  To store a value to the variable
       represented by the VarResult, the value must be on top of the
       operand stack and the indirection elements must be right below
       that.  Call getCheckedStoreCode or getUncheckedStoreCode to
       obtain instructions that consume the value and the indirection
       elements and store the value to the variable.  The code
       returned by getUncheckedStoreCode does not check for null
       dereferencing etc., assuming that getLoadCode has already done
       that.  You can call either getLoadCode, or getLoadCode and then
       getUncheckedStoreCode, or getCheckedStoreCode.  You are not
       supposed to call getIndirectionCode, getLoadCode,
       getUncheckedStoreCode, or getCheckedStoreCode more than once
       because they only return references to instructions, not
       copies.
    """

    def __init__(self, type_, indirectionCode, indirectionLen,
                 loadCode, checkedStoreCode, uncheckedStoreCode):
        """Construct a variable result.

        Param type_: the ResultType of this result.

        Param indirectionCode: a list of instructions that place
        indirection elements on top of the operand stack.

        Param indirectionLen: number of indirection elements.

        Param loadCode: a list of instructions that remove indirection
        elements from top of the operand stack and replace them with
        the value.

        Param checkedStoreCode: a list of instructions that remove a
        value and indirection elements from top of the operand stack
        and store the value.

        Param uncheckedStoreCode: a list of instructions similar to
        checkedStoreCode, except that no null dereferencing or array
        bounds checking is done.
        """
        StaticResult.__init__(self, type_, indirectionCode + loadCode)
        self.__indirectionCode = indirectionCode
        self.__indirectionLen = indirectionLen
        self.__loadCode = loadCode
        self.__checkedStoreCode = checkedStoreCode
        self.__uncheckedStoreCode = uncheckedStoreCode

    def isVar(self):
        return True

    def getIndirectionCode(self):
        """Return a reference to the indirection code."""
        return self.__indirectionCode

    def getIndirectionLen(self):
        """Return the number of indirection elements."""
        return self.__indirectionLen

    def getLoadCode(self):
        """Return a reference to the loading code."""
        return self.__loadCode

    def getCheckedStoreCode(self):
        """Return a reference to the storing code."""
        return self.__checkedStoreCode

    def getUncheckedStoreCode(self):
        """Return a reference to the storing code."""
        return self.__uncheckedStoreCode
