#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# statement.py
#
# 2007-06-10  Jori Dubrovin:  Adding recurseStr API
# 2006-08-15  Jori Dubrovin:  Made str(stmt) output a valid statement
# 2006-01-31  Jori Dubrovin:  Added final local variables
# 2005-11-11  Jori Dubrovin:  Allowed enums in switch statements
# 2005-08-26  Jori Dubrovin:  Adding syntax tree API
# 2005-08-08  Jori Dubrovin:  Adding syntax tree API
# 2005-06-17  Jori Dubrovin:  Added api definitions to SendStatement
# 2005-05-02  Jori Dubrovin:  Changed for loop strings, fixed switch evaluation
# 2005-04-22  Jori Dubrovin:  Initial version
#

"""Classes for handling statements in the action language."""

from syntaxtree import SyntaxNode, TopStatement
from errors import ErrorProxy, StaticError
from mytypes import boolType, intType, voidType
from binaryop import NotEqual
from expr import PrimaryMethodInvocationExpr
from icode import Literal, StoreAuto, Goto, IfGoto, Pop, Return, CheckAssert
from icode import Duplicate, SourceRef, LogicalNot, Halt, Label, CheckNonNull
from icode import uninitialized, void


class BlockStatement(TopStatement):
    """A base class for block statement nodes in the syntax tree."""

    def __init__(self, linespan):
        TopStatement.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):  #pragma: no cover
        return '<block statement>;'

    def translatePass(self, pass_, env):
        if pass_ == 5:
            return self.translate(env)
        else:
            # Passes 1 to 4 are only for top-level type declarations.
            return []

    def translate(self, env):  #pragma: no cover
        """Generate internal code for block statements.

        Raise a StaticError if type checking fails.

        Param env: the StaticEnvironment of the block statement.  This
        may be mutated by translate, even if a StaticError is raised.

        Return value: a list of code instructions.
        """
        raise NotImplementedError('Pure virtual method')


class Statement(BlockStatement):
    """A base class for statement nodes in the syntax tree."""

    def __init__(self, linespan):
        BlockStatement.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):  #pragma: no cover
        return '<statement>;'

    def translateWithLabel(self, labelName, env):
        """Perform static type checking and generate internal code.

        This is called when the Statement is immediately contained in
        a LabeledStatement.  In this case, nested ContinueStatements
        are handled differently is this is an iteration statement.

        Param labelName: the label name as a string.

        Param env: the StaticEnvironment of the statement.  This must
        be not mutated.

        Return value: a list of code instructions.
        """
        return self.translate(env)  # Default implementation

    def translate(self, env):  #pragma: no cover
        """Perform static type checking and generate internal code.

        Unlike BlockStatement.translate, Statement.translate is not
        allowed to mutate the parameter env.
        """
        raise NotImplementedError('Pure virtual method')


class LocalVarDeclarationStatement(BlockStatement):
    """A statement that declares local variable(s) of one type."""

    def __init__(self, modifiers, typeSymbol, declList, linespan):
        """Construct a local variable declaration statement.

        Param modifiers: a list of strings containing the given
        modifiers.

        Param typeSymbol: a TypeSymbol representing the type of the
        declared variables.
        
        Param declList: a nonempty list of LocalVarDeclarator instances.

        Param linespan: see SyntaxNode.
        """
        BlockStatement.__init__(self, linespan)
        self.__modifiers = modifiers
        self.__typeSymbol = typeSymbol
        self.__declList = declList

    def api_recurseStr(self, strFunc, *args, **kwargs):
        typeString = strFunc(self.__typeSymbol, *args, **kwargs)
        declStrings = [strFunc(d, *args, **kwargs) for d in self.__declList]
        return (' '.join(self.__modifiers +
                         [typeString, ', '.join(declStrings)]) +
                ';')

    def api_getModifiers(self):
        """Return the modifiers as a list of strings."""
        return self.__modifiers[:]

    def api_getTypeSymbol(self):
        """Return the TypeSymbol for the type of declared variables."""
        return self.__typeSymbol

    def api_getDeclarators(self):
        """Return the declared variables as a list of LocalVarDeclarators."""
        return self.__declList[:]

    def translate(self, env):
        myType = self.__typeSymbol.resolve(env)  # May raise.
        ep = ErrorProxy()
        isFinal = False
        if len(self.__modifiers) == 1 and self.__modifiers[0] == 'final':
            isFinal = True
        elif self.__modifiers:
            ep.add('local variable declaration has illegal modifiers',
                   self.linespan())
        code = []
        for d in self.__declList:
            try:
                code += d.translate(env, myType, isFinal)
            except StaticError, e:
                ep.addException(e)
        ep.raiseIfErrors()
        return code


class LocalVarDeclarator(SyntaxNode):
    """A local variable declarator."""

    def __init__(self, name, initial, linespan):
        """Construct a variable declarator.

        Param name: string containing the name of the variable.

        Param initial: a VarInitializer with which the variable is to be
        initialized.  Set initial=None to skip initialization.

        Param linespan: see SyntaxNode.
        """
        SyntaxNode.__init__(self, linespan)
        self.__name = name
        self.__initial = initial

    def api_recurseStr(self, strFunc, *args, **kwargs):
        if self.__initial is None:
            return self.__name
        else:
            return (self.__name + ' = ' +
                    strFunc(self.__initial, *args, **kwargs))

    def api_getName(self):
        """Return the name of the variable as a string."""
        return self.__name

    def api_getInitializer(self):
        """Return the VarInitializer.

        None is returned if the variable has no initializer.
        """
        return self.__initial

    def translate(self, env, type_, isFinal):
        """Perform static type checking and generate internal code.

        Raise a StaticError if type checking fails.

        Param env: the StaticEnvironment of the statement.  This is
        mutated by the operation.

        Param type_: the Type of the variable being declared.

        Param isFinal: True if a final variable is being declared.

        Return value: a list of code instructions.
        """
        if env.isLocalVar(self.__name):
            ErrorProxy().finish("redeclaration of variable '" + self.__name +
                                "'", self.linespan())
        env.addLocalVar(self.__name, type_, isFinal)
        code = [SourceRef(self), Literal(uninitialized),
                StoreAuto(self.__name)]
        if self.__initial is not None:
            code += (self.__initial.translate(env, type_) +
                     [StoreAuto(self.__name)])
        elif isFinal:
            ErrorProxy().finish("final variable '" + self.__name +
                                "' must be initialized", self.linespan())
        return code


class Block(Statement):
    """A block of block statements within braces."""

    def __init__(self, blockStatements, linespan):
        """Construct a block from a list of block statements."""
        Statement.__init__(self, linespan)
        self.__blockStatements = blockStatements

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('{\n' +
                '\n'.join([self.indent(strFunc(s, *args, **kwargs)) for
                           s in self.__blockStatements]) + '\n}')

    def api_getBlockStatements(self):
        """Return the contained BlockStatements in a list."""
        return self.__blockStatements[:]

    def translate(self, env):
        newEnv = env.makeSubscopedCopy()
        code = [SourceRef(self)]
        ep = ErrorProxy()
        for s in self.__blockStatements:
            try:
                code += s.translate(newEnv)
            except StaticError, e:
                ep.addException(e)
        ep.raiseIfErrors()
        code += newEnv.getEndOfScopeCode()
        return code


class LabeledStatement(Statement):
    """A labeled statement."""

    def __init__(self, labelName, containedStatement, linespan):
        """Construct a labeled statement.

        Param labelName: the label name as a string.

        Param containedStatement: the Statement following the label.

        Param linespan: see SyntaxNode.
        """
        Statement.__init__(self, linespan)
        self.__labelName = labelName
        self.__containedStatement = containedStatement

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return (self.__labelName + ':\n' +
                strFunc(self.__containedStatement, *args, **kwargs))

    def api_getLabelName(self):
        """Return the label name as a string."""
        return self.__labelName

    def api_getContainedStatement(self):
        """Return the contained Statement."""
        return self.__containedStatement

    def translate(self, env):
        newEnv = env.makeSubscopedCopy()
        ep = ErrorProxy()
        if newEnv.canRetreat('break' + self.__labelName):
            ep.add("redeclaration of label '" + self.__labelName + "'",
                   self.linespan())
        endLabel = Label()
        newEnv.setRetreat('break' + self.__labelName, endLabel)
        code = [SourceRef(self)]
        try:
            s = self.__containedStatement
            code += s.translateWithLabel(self.__labelName, newEnv)
        except StaticError, e:
            ep.addException(e)
        ep.raiseIfErrors()
        return code + [endLabel]


class EmptyStatement(Statement):
    """An empty statement (just a semicolon)."""

    def __init__(self, linespan):
        Statement.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ';'

    def translate(self, env):
        return [SourceRef(self)]


class ExprStatement(Statement):
    """An expression statement."""

    def __init__(self, expr, linespan):
        """Construct a statement from an existing expression.

        Param expr: a StatementExpr.

        Param linespan: see SyntaxNode.
        """
        Statement.__init__(self, linespan)
        self.__expr = expr

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return strFunc(self.__expr, *args, **kwargs) + ';'

    def api_getExpr(self):
        """Return the contained StatementExpr."""
        return self.__expr

    def translate(self, env):
        result = self.__expr.translate(env)
        return ([SourceRef(self)] +
                result.getEvalCode() +
                [Pop()])  # Ignore value


class SendStatement(Statement):
    """A send statement."""

    def __init__(self, signal, argList, target, linespan):
        """Construct a send statement.

        Param signal: a string containing the name of the signal.

        Param argList: a list of expressions that are used as arguments.

        Param target: an expression that represents the target.

        Param linespan: see SyntaxNode.
        """
        Statement.__init__(self, linespan)
        self.__signal = signal
        self.__argList = argList
        self.__target = target

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('send ' + self.__signal + '(' +
                ', '.join([strFunc(e, *args, **kwargs) for
                           e in self.__argList]) +
                ') to ' + strFunc(self.__target, *args, **kwargs) + ';')

    def api_getSignalName(self):
        """Return the signal name as a string."""
        return self.__signal

    def api_getArgList(self):
        """Return the arguments of the signal as a list of Exprs."""
        return self.__argList[:]

    def api_getTarget(self):
        """Return the target object as an Expr."""
        return self.__target

    def translate(self, env):
        # This is effectively a method invocation, so let a
        # PrimaryMethodInvocationExpr do the work.
        methodName = '$$signal_' + self.__signal
        invocation = PrimaryMethodInvocationExpr(self.__target, methodName,
                                                 self.__argList,
                                                 self.linespan())
        result = invocation.translate(env)
        return ([SourceRef(self)] +
                result.getEvalCode() +
                [Pop()])  # Ignore value


class AssertStatement(Statement):
    """An assert statement."""

    def __init__(self, expr, linespan):
        """Construct an assert statement.

        Param expr: the expression that is supposed to be true.

        Param linespan: see SyntaxNode.
        """
        Statement.__init__(self, linespan)
        self.__expr = expr

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return 'assert ' + strFunc(self.__expr, *args, **kwargs) + ';'

    def api_getExpr(self):
        """Return the assertion expression as an Expr."""
        return self.__expr

    def translate(self, env):
        result = self.__expr.translate(env)
        if result.getType() is not boolType:
            ErrorProxy().finish('assertion expression must be boolean, not ' +
                                str(result.getType()), self.linespan())
        return [SourceRef(self)] + result.getEvalCode() + [CheckAssert()]


class ReturnStatement(Statement):
    """A return statement."""

    def __init__(self, expr, linespan):
        """Construct a return statement.

        Param expr: the expression that denotes the return value, or
        None if there is no value.

        Param linespan: see SyntaxNode.
        """
        Statement.__init__(self, linespan)
        self.__expr = expr

    def api_recurseStr(self, strFunc, *args, **kwargs):
        if self.__expr is None:
            return 'return;'
        else:
            return 'return ' + strFunc(self.__expr, *args, **kwargs) + ';'

    def api_getExpr(self):
        """Return the return value as an Expr, or None if there is none."""
        return self.__expr

    def translate(self, env):
        if not env.canReturn():
            ErrorProxy().finish('return is not allowed here', self.linespan())
        code = [SourceRef(self)]
        if self.__expr is None:
            if not env.canReturnType(voidType):
                ErrorProxy().finish('method must return a value',
                                    self.linespan())
            code += [Literal(void)]
        else:
            result = self.__expr.translate(env)
            if not env.canReturnType(result.getType()):
                ErrorProxy().finish('cannot return ' + str(result.getType()),
                                    self.linespan())
            code += result.getEvalCode()
        return code + [Return()]


class BreakStatement(Statement):
    """A break statement with or without a label."""

    def __init__(self, labelName, linespan):
        """Construct a break statement.

        Param labelName: the label name in the statement as a string,
        or None if this is an unlabeled break;

        Param linespan: see SyntaxNode.
        """
        Statement.__init__(self, linespan)
        self.__labelName = labelName

    def api_recurseStr(self, strFunc, *args, **kwargs):
        if self.__labelName is None:
            return 'break;'
        else:
            return 'break ' + self.__labelName + ';'

    def api_getLabelName(self):
        """Return the label name as a string, or None if this is unlabeled."""
        return self.__labelName

    def translate(self, env):
        if self.__labelName is None:
            n = 'break'
            if not env.canRetreat(n):
                ErrorProxy().finish('break is allowed only in a switch, ' +
                                    'while, do, or for statement',
                                    self.linespan())
        else:
            n = 'break' + self.__labelName
            if not env.canRetreat(n):
                ErrorProxy().finish("break is not in the scope of label '" +
                                    self.__labelName + "'", self.linespan())
        return [SourceRef(self)] + env.getRetreatCode(n)


class ContinueStatement(Statement):
    """A continue statement with or without a label."""

    def __init__(self, labelName, linespan):
        """Construct a continue statement.

        Param labelName: the label name in the statement as a string,
        or None if this is an unlabeled continue;

        Param linespan: see SyntaxNode.
        """
        Statement.__init__(self, linespan)
        self.__labelName = labelName

    def api_recurseStr(self, strFunc, *args, **kwargs):
        if self.__labelName is None:
            return 'continue;'
        else:
            return 'continue ' + self.__labelName + ';'

    def api_getLabelName(self):
        """Return the label name as a string, or None if this is unlabeled."""
        return self.__labelName

    def translate(self, env):
        if self.__labelName is None:
            n = 'continue'
            if not env.canRetreat(n):
                ErrorProxy().finish('continue is allowed only in a while, ' +
                                    'do, or for statement', self.linespan())
        else:
            n = 'continue' + self.__labelName
            if not env.canRetreat(n):
                ErrorProxy().finish("continue is not in the scope of a " +
                                    "while, do, or for statement with " +
                                    "label '" + self.__labelName + "'",
                                    self.linespan())
        return [SourceRef(self)] + env.getRetreatCode(n)


class IfStatement(Statement):
    """An if-then-else statement.

    An IfStatement always has an else branch.  To simulate an if
    statement without else branch, an EmptyStatement can be used as
    the else part.
    """

    def __init__(self, condition, truePart, falsePart, linespan):
        """Construct an if-then-else statement.

        Param condition: the boolean expression.
        Param truePart: the first enclosed statement.
        Param falsePart: the else part (must not be None).
        Param linespan: see SyntaxNode.
        """
        Statement.__init__(self, linespan)
        self.__condition = condition
        self.__truePart = truePart
        self.__falsePart = falsePart

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('if (' + strFunc(self.__condition, *args, **kwargs) + ')\n' +
                self.indent(strFunc(self.__truePart, *args, **kwargs)) +
                '\nelse\n' +
                self.indent(strFunc(self.__falsePart, *args, **kwargs)))

    def api_getCondition(self):
        """Return the boolean condition as an Expr."""
        return self.__condition

    def api_getTrueStatement(self):
        """Return the Statement that appears before the keyword 'else'."""
        return self.__truePart

    def api_getFalseStatement(self):
        """Return the Statement that appears after the keyword 'else'."""
        return self.__falsePart

    def translate(self, env):
        ep = ErrorProxy()
        try:
            condResult = self.__condition.translate(env)
            if condResult.getType() is not boolType:
                ep.add('condition must be boolean',
                       self.__condition.linespan())
        except StaticError, e:
            ep.addException(e)
        try:
            trueCode = self.__truePart.translate(env)
        except StaticError, e:
            ep.addException(e)
        try:
            falseCode = self.__falsePart.translate(env)
        except StaticError, e:
            ep.addException(e)
        ep.raiseIfErrors()
        trueLabel = Label()
        endLabel = Label()
        return ([SourceRef(self.__condition)] +
                condResult.getEvalCode() +
                [IfGoto(trueLabel)] +
                falseCode +
                [Goto(endLabel)] +
                [trueLabel] +
                trueCode +
                [endLabel])


class WhileStatement(Statement):
    """A while statement."""

    def __init__(self, condition, body, linespan):
        """Construct a while statement.

        Param condition: the boolean expression.
        Param body: the enclosed statement.
        Param linespan: see SyntaxNode.
        """
        Statement.__init__(self, linespan)
        self.__condition = condition
        self.__body = body

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('while (' + strFunc(self.__condition, *args, **kwargs) +
                ')\n' + self.indent(strFunc(self.__body, *args, **kwargs)))

    def api_getCondition(self):
        """Return the boolean condition as an Expr."""
        return self.__condition

    def api_getBody(self):
        """Return the loop body as a Statement."""
        return self.__body

    def translate(self, env):
        return self.translateWithLabel(None, env)

    def translateWithLabel(self, labelName, env):
        newEnv = env.makeSubscopedCopy()
        testLabel = Label()
        endLabel = Label()
        newEnv.setRetreat('break', endLabel)
        newEnv.setRetreat('continue', testLabel)
        if labelName is not None:
            newEnv.setRetreat('continue' + labelName, testLabel)
        ep = ErrorProxy()
        try:
            condResult = self.__condition.translate(newEnv)
            if condResult.getType() is not boolType:
                ep.add('condition must be boolean',
                       self.__condition.linespan())
        except StaticError, e:
            ep.addException(e)
        try:
            bodyCode = self.__body.translate(newEnv)
        except StaticError, e:
            ep.addException(e)
        ep.raiseIfErrors()
        return ([SourceRef(self.__condition)] +
                [testLabel] +
                condResult.getEvalCode() +
                [LogicalNot()] +
                [IfGoto(endLabel)] +
                bodyCode +
                [Goto(testLabel)] +
                [endLabel])


class DoStatement(Statement):
    """A do-while statement."""

    def __init__(self, condition, body, linespan):
        """Construct a do-while statement.

        Param condition: the boolean expression.
        Param body: the enclosed statement.
        Param linespan: see SyntaxNode.
        """
        Statement.__init__(self, linespan)
        self.__condition = condition
        self.__body = body

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('do\n' + self.indent(strFunc(self.__body, *args, **kwargs)) +
                '\nwhile (' + strFunc(self.__condition, *args, **kwargs) +
                ');')

    def api_getCondition(self):
        """Return the boolean condition as an Expr."""
        return self.__condition

    def api_getBody(self):
        """Return the loop body as a Statement."""
        return self.__body

    def translate(self, env):
        return self.translateWithLabel(None, env)

    def translateWithLabel(self, labelName, env):
        newEnv = env.makeSubscopedCopy()
        bodyLabel = Label()
        testLabel = Label()
        endLabel = Label()
        newEnv.setRetreat('break', endLabel)
        newEnv.setRetreat('continue', testLabel)
        if labelName is not None:
            newEnv.setRetreat('continue' + labelName, testLabel)
        ep = ErrorProxy()
        try:
            # The body part cannot add new local variables to newEnv.
            bodyCode = self.__body.translate(newEnv)
        except StaticError, e:
            ep.addException(e)
        try:
            condResult = self.__condition.translate(newEnv)
            if condResult.getType() is not boolType:
                ep.add('condition must be boolean',
                       self.__condition.linespan())
        except StaticError, e:
            ep.addException(e)
        ep.raiseIfErrors()
        return ([SourceRef(self.__body)] +
                [bodyLabel] +
                bodyCode +
                [SourceRef(self.__condition)] +
                [testLabel] +
                condResult.getEvalCode() +
                [IfGoto(bodyLabel)] +
                [endLabel])


class ForStatement(Statement):
    """A base class for basic for statements.

    A ForStatement instance always has a valid Expr as the boolean
    condition.  The literal 'true' can be used if the condition is
    skipped in the input program.

    The initialization part of a basic for statement can be either a
    local variable declaration or a (possibly empty) list of statement
    expressions.  The two subclasses of ForStatement take care of
    these possibilities.
    """

    def __init__(self, condition, update, body, linespan):
        """Construct a for statement.

        Param condition: the condition Expr (the second part in
        parentheses).

        Param update: a list of StatementExprs (the third part in
        parentheses).

        Param body: the enclosed Statement.

        Param linespan: see SyntaxNode.
        """
        Statement.__init__(self, linespan)
        self.__condition = condition
        self.__update = update
        self.__body = body

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('for (' + self._strInitialization(strFunc, *args, **kwargs) +
                ' ' + strFunc(self.__condition, *args, **kwargs) + '; ' +
                ', '.join([strFunc(expr, *args, **kwargs) for
                           expr in self.__update]) +
                ')\n' + self.indent(strFunc(self.__body, *args, **kwargs)))

    def api_getCondition(self):
        """Return the boolean condition as an Expr."""
        return self.__condition

    def api_getUpdatePart(self):
        """Return the update part as a list of StatementExprs."""
        return self.__update

    def api_getBody(self):
        """Return the loop body as a Statement."""
        return self.__body

    def translate(self, env):
        return self.translateWithLabel(None, env)

    def translateWithLabel(self, labelName, env):
        newEnv = env.makeSubscopedCopy()
        ep = ErrorProxy()
        try:
            # Initialization part may add new local variables to newEnv.
            initCode = self._translateInitialization(newEnv)
        except StaticError, e:
            ep.addException(e)
        updateLabel = Label()
        testLabel = Label()
        endLabel = Label()
        newEnv.setRetreat('break', endLabel)
        newEnv.setRetreat('continue', updateLabel)
        if labelName is not None:
            newEnv.setRetreat('continue' + labelName, updateLabel)
        try:
            condResult = self.__condition.translate(newEnv)
            if condResult.getType() is not boolType:
                ep.add('condition must be boolean',
                       self.__condition.linespan())
        except StaticError, e:
            ep.addException(e)
        try:
            updateCode = []
            for expr in self.__update:
                result = expr.translate(newEnv)
                # Evaluate for side effects and ignore the value.
                updateCode += result.getEvalCode() + [Pop()]
        except StaticError, e:
            ep.addException(e)
        try:
            # The body part cannot add new local variables to newEnv.
            bodyCode = self.__body.translate(newEnv)
        except StaticError, e:
            ep.addException(e)
        ep.raiseIfErrors()
        return (initCode +
                [SourceRef(self.__condition)] +
                [testLabel] +
                condResult.getEvalCode() +
                [LogicalNot()] +
                [IfGoto(endLabel)] +
                bodyCode +
                [updateLabel] +
                updateCode +
                [Goto(testLabel)] +
                [endLabel] +
                newEnv.getEndOfScopeCode())

    def _strInitialization(self, strFunc, *args, **kwargs):
        """Return the string representation of the initialization part."""
        raise NotImplementedError('Pure virtual method')

    def _translateInitialization(self, env):
        """Return code for the initialization part.

        This method may mutate env.
        """
        raise NotImplementedError('Pure virtual method')


class LocalVarForStatement(ForStatement):
    """A for statement whose initialization part declares local variables."""

    def __init__(self, localVarInit, condition, update, body, linespan):
        """Construct a for statement.

        Param localVarInit: a LocalVarDeclarationStatement (the first
        part in parentheses).
        
        Param condition: the condition Expr (the second part in
        parentheses).

        Param update: a list of StatementExprs (the third part in
        parentheses).

        Param body: the enclosed statement.

        Param linespan: see SyntaxNode.
        """
        ForStatement.__init__(self, condition, update, body, linespan)
        self.__localVarInit = localVarInit

    def api_getInitialization(self):
        """Return the initialization part as a LocalVarDeclarationStatement."""
        return self.__localVarInit

    def _strInitialization(self, strFunc, *args, **kwargs):
        return strFunc(self.__localVarInit, *args, **kwargs)

    def _translateInitialization(self, env):
        return self.__localVarInit.translate(env)


class ExprListForStatement(ForStatement):
    """A for statement whose initialization part contains expressions."""

    def __init__(self, exprInit, condition, update, body, linespan):
        """Construct a for statement.

        Param exprInit: a list of StatementExprs (the first part in
        parentheses).

        Param condition: the condition Expr (the second part in
        parentheses).

        Param update: a list of StatementExprs (the third part in
        parentheses).

        Param body: the enclosed statement.

        Param linespan: see SyntaxNode.
        """
        ForStatement.__init__(self, condition, update, body, linespan)
        self.__exprInit = exprInit

    def api_getInitialization(self):
        """Return the initialization part as a list of StatementExprs."""
        return self.__exprInit[:]

    def _strInitialization(self, strFunc, *args, **kwargs):
        return ', '.join([strFunc(expr, *args, **kwargs) for
                          expr in self.__exprInit]) + ';'

    def _translateInitialization(self, env):
        initCode = []
        for expr in self.__exprInit:
            result = expr.translate(env)
            # Evaluate for side effects, ignore value.
            initCode += [SourceRef(expr)] + result.getEvalCode() + [Pop()]
        return initCode


class SwitchLabel(SyntaxNode):
    """A case of default label inside a switch statement."""

    def __init__(self, caseExpr, linespan):
        """Construct a switch label.

        Param caseExpr: the expression contained in the case label.
        If there is no case label but a default label instead, this is
        None.

        Param linespan: see SyntaxNode.
        """
        SyntaxNode.__init__(self, linespan)
        self.caseExpr = caseExpr

    def api_recurseStr(self, strFunc, *args, **kwargs):
        if self.caseExpr is None:
            return 'default:'
        else:
            return 'case ' + strFunc(self.caseExpr, *args, **kwargs) + ':'

    def api_getCaseExpr(self):
        """Return the enclosed case expression as an Expr.

        None is returned if this is a default label (default:).
        """
        return self.caseExpr


class SwitchStatement(Statement):
    """A switch statement."""

    def __init__(self, switchExpr, groups, linespan):
        """Construct a switch statement.

        Param switchExpr: the switch expression.

        Param groups: a list of tuples, whose first element is a
        SwitchLabel and second element is a list of BlockStatements.

        Param linespan: see SyntaxNode.
        """
        Statement.__init__(self, linespan)
        self.__switchExpr = switchExpr
        self.__groups = groups

    def api_recurseStr(self, strFunc, *args, **kwargs):
        strs = ['switch (' + strFunc(self.__switchExpr, *args, **kwargs) + ')',
                '{']
        for (label, statements) in self.__groups:
            strs.append(strFunc(label, *args, **kwargs))
            strs += [self.indent(strFunc(s, *args, **kwargs)) for
                     s in statements]
        strs.append('}')
        return '\n'.join(strs)

    def api_getSwitchExpr(self):
        """Return the switch expression as an Expr."""
        return self.__switchExpr

    def api_getSwitchGroups(self):
        """Return the list of switch groups.

        Each switch group is a pair (label, statements) corresponding
        to a case or default label followed by block statements.  Here
        label is the case label or default label of type SwitchLabel.
        The statements field contains a list of the BlockStatements
        that appear between the associated label and the next
        case/default label (or the end of the switch body).
        """
        return self.__groups[:]

    def translate(self, env):
        newEnv = env.makeSubscopedCopy()
        ep = ErrorProxy()
        switchResult = self.__switchExpr.translate(env)
        switchCode = ([SourceRef(self.__switchExpr)] +
                      switchResult.getEvalCode())
        if switchResult.getType().isEnum():
            switchCode += [CheckNonNull(1)]
        elif switchResult.getType() is not intType:
            ep.finish('switch expression cannot be ' +
                      str(switchResult.getType()),
                      self.__switchExpr.linespan())
        bodyCode = []
        defaultFound = 0
        endLabel = Label()
        newEnv.setRetreat('break', endLabel)
        goDefaultCode = [Goto(endLabel)]
        for (label, statements) in self.__groups:
            if label.caseExpr is None:
                # A default label
                if defaultFound == 1:
                    ep.add('only one default label allowed',
                           label.linespan())
                defaultFound += 1
                defaultLabel = Label()
                bodyCode += [SourceRef(label), defaultLabel]
                goDefaultCode = (newEnv.getEnterScopeCode() +
                                 [Goto(defaultLabel)])
            else:
                try:
                    caseResult = label.caseExpr.translate(env)
                    caseType = caseResult.getType()
                    if not caseType.isSubtypeOf(switchResult.getType()):
                        ep.add('incompatible case expression type ' +
                               str(caseType),
                               label.caseExpr.linespan())
                    else:
                        operator = NotEqual(label.linespan())
                        opResult = operator.translate(env, caseType,
                                                      switchResult.getType())
                        notLabel = Label()
                        caseLabel = Label()
                        switchCode += ([SourceRef(label.caseExpr)] +
                                       [Duplicate()] +
                                       caseResult.getEvalCode() +
                                       opResult.getEvalCode() +
                                       [IfGoto(notLabel)] +
                                       [Pop()] +
                                       newEnv.getEnterScopeCode() +
                                       [Goto(caseLabel)] +
                                       [notLabel])
                        bodyCode += [SourceRef(label), caseLabel]
                except StaticError, e:
                    ep.addException(e)
            for s in statements:
                try:
                    # The statements in this group may add new local variables.
                    bodyCode += s.translate(newEnv)
                except StaticError, e:
                    ep.addException(e)
        ep.raiseIfErrors()
        switchCode += [Pop()] + goDefaultCode
        bodyCode += newEnv.getEndOfScopeCode() + [endLabel]
        return switchCode + bodyCode
