#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# staticenv.py
#
# 2007-06-05  Jori Dubrovin:  Added labeled statements
# 2005-12-15  Jori Dubrovin:  Support for final local variables
# 2005-07-29  Jori Dubrovin:  Refactoring
# 2005-06-17  Jori Dubrovin:  Simplified design
# 2005-06-03  Jori Dubrovin:  Initial version
#

"""Static environments in the action language."""

from copy import copy

from icode import Delete, Goto, Literal, StoreAuto, Return, uninitialized
from mytypes import voidType


class StaticEnvironment:
    """A compile-time context for declarations, statements and expressions."""

    def __init__(self):
        self.__topTypes = {}
        self.__enclosingTypes = []
        self.__isInstanceContext = False
        # Names of local variables, ordered by decreasing scope
        self.__vars = []
        self.__innerScopeIndex = 0
        # Map local variable name to its type
        self.__varTypes = {}
        # Map local variable name to True iff the variable is final
        self.__varIsFinal = {}
        self.__returnType = None
        # Map label name to pair (Label, scope index)
        self.__retreats = {}
        self.__nativeHooks = {}

    def __str__(self):
        s = ['Environment:']
        if self.__varTypes:
            s.append(', '.join([name + ':' + str(self.__varTypes[name]) for
                                name in self.__varTypes]))
        else:
            s.append('<empty>')
        return '\n'.join(s)

    def copyFrom(self, other):
        """Make self a copy of the other StaticEnvironment."""
        self.__topTypes = copy(other.__topTypes)
        self.__enclosingTypes = other.__enclosingTypes[:]
        self.__isInstanceContext = other.__isInstanceContext
        self.__vars = other.__vars[:]
        self.__innerScopeIndex = other.__innerScopeIndex
        self.__varTypes = copy(other.__varTypes)
        self.__varIsFinal = copy(other.__varIsFinal)
        self.__returnType = other.__returnType
        self.__retreats = copy(other.__retreats)
        self.__nativeHooks = copy(other.__nativeHooks)

    def copy(self):
        """Return a copy of self."""
        other = StaticEnvironment()
        other.copyFrom(self)
        return other

    def makeSubscopedCopy(self):
        """Return a subscoped copy of this environment.

        The copy will have initially the same local variables and the
        same return, break, and continue behavior.  Calling
        enterScopeCode or endOfScopeCode for the copy will only create
        or delete those variables that were added to the copy after
        calling makeSubscopedCopy.
        """
        other = self.copy()
        other.__innerScopeIndex = len(other.__vars)
        return other

    def makeNestedCopy(self, enclosingType):
        """Return a nested copy of this environment.

        The copy will have the same top-level types, but no local
        variables and no return, break, and continue behavior.  The
        list of enclosing types will be augmented by the given type.

        Param enclosingType: a SingleReferenceType that immediately
        encloses the new environment.
        """
        other = StaticEnvironment()
        other.__topTypes = copy(self.__topTypes)
        other.__enclosingTypes = self.__enclosingTypes + [enclosingType]
        other.__nativeHooks = copy(self.__nativeHooks)
        return other

    def addTopType(self, type_):
        """Add a SingleReferenceType to the top level."""
        self.__topTypes[type_.getSimpleName()] = type_

    def isTopType(self, name):
        """Return True if there is a top-level type with the given name."""
        return (name in self.__topTypes)

    def getTopType(self, name):
        """Return the top-level SingleReferenceType with the given name."""
        return self.__topTypes[name]

    def isInstanceContext(self):
        """Return True if self is an instance context ('this' is valid)."""
        return self.__isInstanceContext

    def setInstanceContext(self):
        """Set the instance context flag on this environment."""
        self.__isInstanceContext = True

    def addLocalVar(self, varName, type_, isFinal=False):
        """Add a local variable to the scope.

        Param varName: a string containing the name of the variable.

        Param type: the Type of the new variable.

        Param isFinal: True if the variable is final.
        """
        self.__vars.append(varName)
        self.__varTypes[varName] = type_
        self.__varIsFinal[varName] = isFinal

    def getEnclosingTypes(self):
        """Return a list of SingleReferenceTypes that enclose the environment.

        The outermost types will be ordered first.
        """
        return self.__enclosingTypes[:]

    def isLocalVar(self, varName):
        """Return true if there exists a local variable with the given name."""
        return (varName in self.__varTypes)

    def getLocalVarType(self, varName):
        """Return the Type of the local variable with the given name."""
        return self.__varTypes[varName]

    def isLocalVarFinal(self, varName):
        """Return true if the local variable with the given name is final."""
        return self.__varIsFinal[varName]

    def getEnterScopeCode(self):
        """Return internal code for entering this scope.

        The returned list of instructions creates all local variables that
        have been added to this environment object so far.
        """
        code = []
        for i in xrange(self.__innerScopeIndex, len(self.__vars)):
            code += [Literal(uninitialized), StoreAuto(self.__vars[i])]
        return code

    def getEndOfScopeCode(self):
        """Return internal code for leaving this scope.

        The returned list of instructions deletes all local variables that
        have been added to this environment object so far.
        """
        code = []
        for i in xrange(len(self.__vars)-1, self.__innerScopeIndex-1, -1):
            code += [Delete(self.__vars[i])]
        return code

    def setReturnType(self, type_):
        """Enable the effect of a return statement.

        Param type_: the ResultType that this routine must return.
        """
        self.__returnType = type_

    def canReturn(self):
        """Return true if a return statement is allowed at all."""
        return (self.__returnType is not None)

    def canReturnType(self, type_):
        """Return true if a return with the given ResultType is allowed."""
        if not self.canReturn():
            return False
        # Type void is the only subtype of itself.
        return type_.isSubtypeOf(self.__returnType)

    def getReturnCode(self):
        """Return internal code for a return statement.

        Return statements must be allowed (see the canReturn method).
        The top element of the operand stack must contain the return value,
        or None if the return type is void.
        """
        return [Return()]

    def setRetreat(self, name, label):
        """Add a retreat.

        A retreat is a shortcut jump directly out of a scope.

        Param name: the name of the retreat as a string.  Old retreats
        with the same name are overwritten

        Param label: the Label to which execution should jump.
        """
        self.__retreats[name] = (label, len(self.__vars))

    def canRetreat(self, name):
        """Return true if there is a retreat with the given name.

        Param name: the name of the retreat as a string.
        """
        return (name in self.__retreats)

    def getRetreatCode(self, name):
        """Return internal code for a retrat.

        The returned list of instructions jumps to the label set by
        setRetreat after deleting the variables that fall out of
        scope.

        Param name: the name of the retreat as a string.
        """
        (label, scopeIndex) = self.__retreats[name]
        code = []
        for i in xrange(len(self.__vars)-1, scopeIndex-1, -1):
            code += [Delete(self.__vars[i])]
        return code + [Goto(label)]

    def setNativeHooks(self, hooks):
        """Initialize the set of native Python function objects.

        If your program declares any native methods, you must call
        this method before translation.

        Param hooks: a dictionary (or other mapping object) that maps
        the canonical names of a native method to a Python function
        that accepts the same number of arguments as the native
        method.
        """
        self.__nativeHooks = copy(hooks)

    def isNativeHook(self, name):
        """Return true if a Python function for the canonical name exists."""
        return (name in self.__nativeHooks)

    def getNativeHook(self, name):
        """Return the Python function for the given canonical method name."""
        return self.__nativeHooks[name]
