#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# syntaxtree.py
#
# 2007-06-10  Jori Dubrovin:  Adding recurseStr API
# 2005-08-08  Jori Dubrovin:  Removed implicit env from SyntaxTree.translate
# 2005-06-17  Jori Dubrovin:  Made SyntaxRoot.translate exception safe
# 2005-04-22  Jori Dubrovin:  Initial version
#

"""General definitions for the nodes of the syntax tree."""

from errors import ErrorProxy, StaticError
from icode import SourceRef, Halt
from icode import buildRoutine


class SyntaxNode:
    """A base class for nodes of the syntax tree."""

    def __init__(self, linespan):
        """Construct a syntax tree node.

        Param linespan: a tuple containing the first and last line
        numbers of this element.
        """
        self.__linespan = linespan

    def __str__(self):
        return self.api_recurseStr(SyntaxNode.__str__)

    def api_recurseStr(self, strFunc, *args, **kwargs):  #pragma: no cover
        return '<syntax node>'

    def linespan(self):
        """Return the source line numbers in which the expression appears.

        Return value: (first line number, last line number).
        """
        return self.__linespan

    def indent(s):
        """Indent all the lines of string s."""
        lines = s.split('\n')
        return '\n'.join(['  ' + st for st in lines])
    indent = staticmethod(indent)


class SyntaxRoot(SyntaxNode):
    """A list of top statements (root node of syntax tree)."""

    def __init__(self, topStatements, linespan):
        """Construct a syntax tree from a list of TopStatements."""
        SyntaxNode.__init__(self, linespan)
        self.__topStatements = topStatements

    def api_recurseStr(self, strFunc, *args, **kwargs):  #pragma: no cover
        return '\n'.join([strFunc(s, *args, **kwargs) for
                          s in self.__topStatements])

    def api_getTopStatements(self):
        """Return the contained TopStatements in a list."""
        return self.__topStatements[:]

    def translate(self, env):
        """Perform static checking and return the entry point of a routine.

        Raise a StaticError if type checking fails.

        Param env: the StaticEnvironment of the program.  This may be
        mutated by translate, but not if a StaticError is raised.

        Return value: entry point for the top-level routine.
        """
        workEnv = env.copy()  # Work with a copy to provide exception safety.
        code = [SourceRef(self)]
        ep = ErrorProxy()
        passCount = 5
        for pass_ in xrange(1, passCount + 1):
            for s in self.__topStatements:
                try:
                    code += s.translatePass(pass_, workEnv)
                except StaticError, e:
                    ep.addException(e)
            ep.raiseIfErrors()
        env.copyFrom(workEnv)
        code += [Halt()]
        return buildRoutine(code)


class TopStatement(SyntaxNode):
    """A base class for block statements and types in the syntax tree."""

    def __init__(self, linespan):
        SyntaxNode.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):  #pragma: no cover
        return '<top statement>;'

    def translatePass(self, pass_, env):  #pragma: no cover
        """Perform a translation pass.

        Param pass_: the number of the current pass, starting from 1.

        Param env: the current StaticEnvironment.

        Return value: a list of top-level code instructions to
        accumulate.
        """
        raise NotImplementedError('Pure virtual method')
