#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# test_expr.py
#
# 2005-11-29  Jori Dubrovin:  Added a preliminary test, not yet implemented
# 2005-08-30  Jori Dubrovin:  Added tests for integer relational operators
# 2005-07-21  Jori Dubrovin:  Reflected changes in testhelp.py
# 2005-06-16  Jori Dubrovin:  Added tests for unary plus etc.
# 2005-05-02  Jori Dubrovin:  Moved test_local_decl to test_statement.py
# 2005-04-26  Jori Dubrovin:  Initial version
#

"""Test for the semantics of expressions."""

import unittest

import errors
from testhelp import checkValues, semanticError, runtimeError


class ExprTest(unittest.TestCase):
    """Test expressions."""

    def test_intAssignment(self):
        # Test assignment to int.
        checkValues("""
        int i, j, k = -8;
        i = -100;
        i *= j = 10;  // i == -1000
        k /= 3;       // k == -2
        """, {'i': -1000, 'j': 10, 'k': -2})
        checkValues("""
        int p = 765432, q = 45, r = 99999;
        p %= -1000;   // p == 432
        q += 11;      // q == 56
        r -= r = 2;   // equivalent to r = (int) (r - (r = 2)), so r == 99997
        """, {'p': 432, 'q': 56, 'r': 99997})
        checkValues("""
        int i = 1<<30, j = i, k = 0xffff0fff, p = k, q = 12;
        i <<= 1;      // i == -(1<<31)
        j <<= 33;     // j == 0
        k >>= 5;      // k == 0xfffff87f
        p >>>= 5;     // p == 0x07fff87f
        q <<= q - q;  // q == 12
        """, {'i': int(-(1L<<31)), 'j': 0, 'k': int(0xfffff87fL - (1L<<32)),
              'p': 0x07fff87f, 'q': 12})
        checkValues("""
        int i = 0x77777777, j = i, k = 0xffff0fff, p = k;
        i &= k;    // i == 0x77770777
        j |= ++k;  // k == 0xffff1000, j == 0xffff7777
        p ^= k;    // p == 0x00001fff
        """, {'i': 0x77770777, 'j': int(0xffff7777L - (1L<<32)),
              'k': int(0xffff1000L - (1L<<32)), 'p': 0x00001fff})
        checkValues("""
        int i = 7, j = -1;
        j += i *= 3;
        """, {'i': 21, 'j': 20})
        runtimeError('int i = 1; i /= 0;', errors.ZeroDivisionError_)
        runtimeError('int i = 1; i %= 0;', errors.ZeroDivisionError_)
        runtimeError('int i = 1; i <<= -1;', errors.NegativeShiftError)
        runtimeError('int i = 1; i >>>= -32;', errors.NegativeShiftError)
        semanticError('int i = 1; i = false;')
        semanticError('int i = 1; i = null;')
        semanticError('int i = 1; i += i <= i;')
        semanticError('class C {} int i = new C();')
        semanticError('class C { static void m() {} } int i = C.m();')

    def test_boolAssignment(self):
        # Test assignment to boolean.
        checkValues("""
        boolean b, c;
        b = true;
        c = !b;
        """, {'b': True, 'c': False})
        checkValues('boolean b, c; c = b = false;',
                    {'b': False, 'c': False})
        checkValues("""
        boolean b = false, c = b, d = c, e = true, f = e, g = f;
        b &= false;
        c &= true;
        d |= b;
        e |= e;
        f ^= false;
        g ^= g;
        """, {'b': False, 'c': False, 'd': False,
              'e': True, 'f': True, 'g': False})
        semanticError('boolean b = false; b = 0;')
        semanticError('boolean b = false; b = null;')
        semanticError('boolean b = false; b += false;')
        semanticError('boolean b = false; b -= 1;')
        semanticError('boolean b = false; b >>= 0;')
        semanticError('class C {} boolean b = new C();')

    def test_classAssignment(self):
        # Test assignment to a variable of class type.
        checkValues("""
        class C {}
        C c;
        c = null;
        C d = new C();
        """, {'c': 'null', 'd': 'nonnull'})
        checkValues('class C {} C c = new C(); C d; d = c;', {'c == d': True})
        semanticError('class C {} C c = false;')
        semanticError('class C {} C c = new C(); c = 0;')
        semanticError('class C {} C c = new C(); c += 0;')
        semanticError('class C {} C c = new C(); c |= false;')
        semanticError('class C {} C c = new C(); c -= new C();')
        semanticError('class C {} class D {} C c; c = new D();')

    def test_condExpr(self):
        # Test conditional operators ?:, ||, &&.
        checkValues("""
        int a = 1 == 1 ? 1000 : 10;
        int p, q, r;
        p = false ? q = -7 : (r = -10);  // p = r = -10, q uninitialized
        int x, y, z;
        x = true ? y = 5 : (z = 100);    // x = y = 5, z uninitialized
        """, {'a': 1000, 'p': -10, 'q': None, 'r': -10,
              'x': 5, 'y': 5, 'z': None})
        checkValues("""
        boolean b = 5 < 5 ? 10 < 7 : 10 > 7;
        """, {'b': True})
        semanticError('int i = 100 ? 4 : 5;')
        semanticError('int i = true ? true : 5;')
        semanticError('class C {} int i = new C() ? 4 : 5;')
        checkValues("""
        int p, q, r, s;
        boolean b = 10 == (p = 20) && 30 == (q = 40);   // false && false
        boolean d = 10 == (r = 10) && 50 == (s = -50);  // true && false
        boolean e = b == b && d == d;                   // true && true
        """, {'p': 20, 'q': None, 'b': False, 'r': 10,
              's': -50, 'd': False, 'e': True})
        checkValues("""
        int p, q, r, s;
        boolean b = 10 == (p = 20) || 30 == (q = 40);   // false || false
        boolean d = 10 == (r = 10) || 50 == (s = -50);  // true || false
        boolean e = b != b || d == d;                   // false || true
        """, {'p': 20, 'q': 40, 'b': False, 'r': 10, 's': None,
              'd': True, 'e': True})
        semanticError('int i = false && false;')
        semanticError('boolean b = false && 3;')
        semanticError('boolean b = true || -3;')
        semanticError('class C {} boolean b = true || new C();')
        semanticError('class C {} boolean b = new C() && new C();')
        semanticError('class C {} C c = true || false;')

    def test_bitLogicalExpr(self):
        # Test bit and logical operators &, |, ^.
        checkValues("""
        int i = 0xf0f0 & 0xaaaa;  // i == 0xa0a0
        int j = 0xf0f0 | 0xaaaa;  // j == 0xfafa
        int k = i ^ j;            // k == 0x5a5a
        """, {'i': 0xa0a0, 'j': 0xfafa, 'k': 0x5a5a})
        checkValues("""
        boolean b, c, d, e, f;
        b = (false | false) & (c = true);  // false & true
        d = (true & true) | (e = false);   // true | false
        f = b ^ d;                         // false ^ true
        """, {'c': True, 'b': False, 'e': False, 'd': True, 'f': True})
        semanticError('boolean b = true & 0;')
        semanticError('int i = true | 0;')
        semanticError('class C {} int i = new C() | 0;')

    def test_relationalExpr(self):
        # Test equality and other relational operators ==, !=, <, >, <=, >=.
        domain = [-(1L<<31), -12048, -1, 0, 1, 573652, (1L<<31)-1]
        for x in domain:
            for y in domain:
                s1 = 'boolean b = ' + str(x)
                s2 = str(y) + ';'
                checkValues(s1 + '==' + s2, {'b': x == y})
                checkValues(s1 + '!=' + s2, {'b': x != y})
                checkValues(s1 + '<' + s2, {'b': x < y})
                checkValues(s1 + '>' + s2, {'b': x > y})
                checkValues(s1 + '<=' + s2, {'b': x <= y})
                checkValues(s1 + '>=' + s2, {'b': x >= y})
        checkValues("""
        boolean b = 4 < 6 && 0x7fffffff > 0x80000000;
        boolean c = 6 > 6 || 1 > 10;
        boolean d = 5-1 <= 5-2 || -2 >= 1;
        boolean e = 3 >= 2 && -2 <= 1;
        """, {'b': True, 'c': False, 'd': False, 'e': True})
        checkValues('boolean b = 0xfffffffe == -2;', {'b': True})
        checkValues('boolean c = 10 == 11;', {'c': False})
        checkValues('boolean d = true == true;', {'d': True})
        checkValues('boolean e = 5*6 != 30;', {'e': False})
        checkValues('boolean f = true != false;', {'f': True})
        checkValues('class C {} boolean b = new C() == new C();', {'b': False})
        checkValues('class C {} boolean b = new C() == null;', {'b': False})
        checkValues('class C {} C c = null; boolean b = c == null;',
                    {'b': True})
        checkValues('''
        class C {} C c = new C(), d; d = c; boolean b = c == d;
        ''', {'b': True})
        checkValues('class C {} boolean b = new C() != new C();', {'b': True})
        checkValues('class C {} boolean b = new C() != null;', {'b': True})
        checkValues('class C {} C c = null; boolean b = c != null;',
                    {'b': False})
        checkValues('''
        class C {} C c = new C(), d; d = c; boolean b = c != d;
        ''', {'b': False})
        semanticError('class C {} C c; boolean b = c == false;')
        semanticError('class C {} C c; boolean b = c == 1;')
        semanticError('class C {} C c; boolean b = c < c;')
        semanticError('class C {} C c; boolean b = c > 0;')
        semanticError('int x = 0; boolean b = x == false;')
        semanticError('int x = 0; boolean b = x != false;')
        runtimeError('boolean b = b == b;', errors.UninitializedError)

    def test_shiftExpr(self):
        # Test shift operators <<, >>, >>>.
        checkValues('int x = 1, y = 10 << x;', {'x': 1, 'y': 20})
        checkValues('int x = 1 << 32;', {'x': 0})
        checkValues('int x = (1 << 31) << 0;', {'x': (-1)<<31})
        checkValues('int x = -2 >> 3;', {'x': -1})
        checkValues('int x = 100 >> 3;', {'x': 12})
        checkValues('int x = -2 >>> 3;', {'x': 0x1fffffff})
        checkValues('int x = 100 >>> 3;', {'x': 12})
        semanticError('int x = 1 << false;')
        semanticError('int x = true >> false;')
        semanticError('class C {} C c; int x = c << 1;')
        runtimeError('int x = 1 << -5;', errors.NegativeShiftError)
        runtimeError('int x = 1 >> 0xffffffff;', errors.NegativeShiftError)
        runtimeError('int x = 1 >>> -1;', errors.NegativeShiftError)

    def test_additiveExpr(self):
        # Test additive operators +, -.
        checkValues('int x = 1, y = x + 10;', {'x': 1, 'y': 11})
        checkValues('int x = 2000000000 + 2100000000;', {'x': -194967296})
        checkValues('int x = 3 - 4;', {'x': -1})
        semanticError('class C {} C c = c + c;')
        semanticError('int x = 1 + false;')
        semanticError('int x = 1 - false;')
        semanticError('int x = false + false;')
        semanticError('int x = false - 0;')

    def test_multiplicativeExpr(self):
        # Test multiplicative operators *, /, %.
        checkValues('int x = 99, y = x / x;', {'x': 99, 'y': 1})
        checkValues('int x = 99, y = x * x;', {'x': 99, 'y': 9801})
        checkValues('int x = 99, y = x % x;', {'x': 99, 'y': 0})
        checkValues('int x = 2000000000*2000000000;', {'x': -1651507200})
        checkValues('int x = -4000 * 7000;', {'x': -28000000})
        checkValues('int x = 00 * 0x7fffffff;', {'x': 0})
        checkValues('int x = -0 / -1;', {'x': 0})
        checkValues('int x = -2147483648 / -1;', {'x': -2147483648})
        checkValues('int x =  17 /  6;', {'x':  2})
        checkValues('int x =  17 %  6;', {'x':  5})
        checkValues('int x = -17 /  6;', {'x': -2})
        checkValues('int x = -17 %  6;', {'x': -5})
        checkValues('int x =  17 / -6;', {'x': -2})
        checkValues('int x =  17 % -6;', {'x':  5})
        checkValues('int x = -17 / -6;', {'x':  2})
        checkValues('int x = -17 % -6;', {'x': -5})
        semanticError('class C {} C c = 0 * c;')
        semanticError('int x = 1 * false;')
        semanticError('int x = 1 / false;')
        semanticError('int x = false % true;')
        runtimeError('int x = 0 / 0;', errors.ZeroDivisionError_)
        runtimeError('int x = 4 / 0;', errors.ZeroDivisionError_)

    def test_unaryExpr(self):
        # Test unary prefix and postfix operators -, +, !, ~, ++, --.
        checkValues('int x = 5; x++;', {'x': 6})
        checkValues('int x = 0x7fffffff; ++(x);', {'x': -2147483648})
        checkValues('int x = 0x80000000; (x)--;', {'x': 0x7fffffff})
        checkValues('int x = 1; --x;', {'x': 0})
        checkValues('int x = 5; x++;', {'x': 6})
        checkValues('int x = 5, y = x++;', {'x': 6, 'y': 5})
        checkValues('int x = 5, y = ++x;', {'x': 6, 'y': 6})
        checkValues('int y = 3; y = y-- + --y;', {'y': 4})
        checkValues('int x = -(5-15);', {'x': 10})
        checkValues('int x = 4; x = +(x + 4);', {'x': 8})
        checkValues('int x = 0x80000000; x = -x;', {'x': -2147483648})
        checkValues('int x = 0x80000000; x = ~x;', {'x': 0x7fffffff})
        checkValues('boolean b = !false;', {'b': True})
        checkValues('''
        class C { int x; }
        C c = new C(), d = new C();
        c.x = 10;
        d.x = c.x++;
        ''', {'c.x': 11, 'd.x': 10})
        semanticError('5++;')
        semanticError('int a = 1; --(a+3);')
        semanticError('int a = 5; a = !a;')
        semanticError('int a = 5; a++++;')
        semanticError('int a = 5; --+a;')
        semanticError('boolean b = false; b = ~b;')
        semanticError('boolean b = false; b++;')
        semanticError('boolean b = false; b--;')
        semanticError('boolean b = false; ++b;')
        semanticError('boolean b = false; --b;')
        semanticError('class C {} C c = +c;')

    def test_paren(self):
        # Test parentheses in expressions.
        checkValues('int x = 2 + 3 * 4;', {'x': 14})
        checkValues('int x = (2 + 3) * 4;', {'x': 20})

    def test_this(self):
        # Test the 'this' expression.
        semanticError('boolean b = (this == this);')
        semanticError('''
        class C
        {
          static int m() { boolean b = (this == this); }
        }
        ''')
        checkValues('''
        class C
        {
          boolean m(C c) { return c == this; }
        }
        C c = new C();
        ''', {'c.m(null)': False, 'c.m(c)': True, 'c.m(new C())': False})

    def test_fieldAccess(self):
        # Test field access and qualified name expressions.
        semanticError('class C { class D {} int x; } C.D d; d.x = 3;')
        semanticError('class C { class D {} int x; } C.D.x = 3;')
        runtimeError('class C { int x; } C c; c.x = 3;',
                     errors.UninitializedError)
        runtimeError('class C { int x; } C c = null; c.x = 3;',
                     errors.NullReferenceError)
        checkValues('''
        class C
        {
          C get() { return this; }
          int x;
        }
        C c = new C();
        c.get().x = 3;
        c.get().x++;
        c.get().x *= c.get().x;
        ''', {'c.x': 16})

    def test_methodInvocation(self):
        # Test method invocation expressions.
        classDef = '''
        class C
        {
          static int meth0() { return 12; }
          static void meth1(boolean y) { return; }
          static boolean meth2(int x, int y) { return x < y; }
          int meth3(C obj) { return 17; }
        }
        '''
        checkValues(classDef + 'C.meth0();', {})
        checkValues(classDef + 'int i = C.meth0();', {'i': 12})
        checkValues(classDef + '''
        int y, x;
        boolean b = C.meth2(y = 20, x = 15);
        ''', {'y': 20, 'x': 15, 'b': False})
        runtimeError(classDef + 'C c; c.meth0();', errors.UninitializedError)
        checkValues(classDef + 'C c = null; c.meth0();', {})
        runtimeError(classDef + 'C c; c.meth3(null);',
                     errors.UninitializedError)
        runtimeError(classDef + 'C c = null; c.meth3(null);',
                     errors.NullReferenceError)
        semanticError(classDef + 'C c = new C(); c.meth3(0);')
        semanticError(classDef + 'C c = new C(); c.meth3(C);')
        checkValues(classDef + 'C c = new C(); c.meth3(null);', {})
        checkValues(classDef + 'C c = new C(); c.meth3(c);', {})
        semanticError(classDef + 'C c = new C(); class D {} D d; c.meth3(d);')
        semanticError(classDef + 'C.D.meth0();')
        semanticError(classDef + 'C.meth0(1);')
        semanticError(classDef + 'C.meth1();')
        semanticError(classDef + 'C.meth1(1);')
        semanticError(classDef + 'C.meth1(false, false);')
        semanticError(classDef + 'C.meth2(0, false);')
        semanticError(classDef + 'C.meth2(false, 0);')
        semanticError(classDef + 'int x = C.meth1(false);')
        semanticError(classDef + 'int x = C.meth2(1, 2);')
        semanticError(classDef + 'boolean x = C.meth0();')
        checkValues('''
        C c = new C();
        c.f = 10;
        int x = c.f;
        int y = C.smeth();
        int z = c.meth();
        class C
        {
          int meth() { return smeth() + f; }
          static int smeth() { return new C().fifty(); }
          int f;
          int fifty() { return 50; }
        }
        ''', {'x': 10, 'y': 50, 'z': 60})
        def check(s):
            pass # not implemented$$
        check('''
        abstract class Base {
          abstract int abst();
          int inst() { return 1; }
          static int clas() { return 2; }
        }
        class Derived extends Base {
          Derived good = this;
          Derived bad = null;

          int abst() { return 3; }
          int inst() { return 4; }
          static int clas() { return 5; }

          void test() {
            assert abst() == 3;
            assert good.abst() == 3;
            RERR:NullPointerError bad.abst();
            assert ((Base) this).abst() == 3;
            assert ((Base) good).abst() == 3;
            RERR:NullPointerError ((Base) bad).abst();
            CERR super.abst();
            CERR Derived.abst();

            assert inst() == 4;
            assert good.inst() == 4;
            RERR:NullPointerError bad.inst();
            assert ((Base) this).inst() == 4;
            assert ((Base) good).inst() == 4;
            RERR:NullPointerError ((Base) bad).inst();
            assert super.inst() == 1;
            CERR Derived.inst();

            assert clas() == 5;
            assert good.clas() == 5;
            assert bad.clas() == 5;
            assert ((Base) this).clas() == 2;
            assert ((Base) good).clas() == 2;
            assert ((Base) bad).clas() == 2;
            assert super.clas() == 2;
            assert Derived.clas() == 5;
          }
        }
        new Derived().test();
        ''')
        check('''
        interface I1 { }
        interface I2 { }
        class C implements I1, I2 {
          int m(I1 i) { return 1; }
          int m(I2 i) { return 2; }
        }
        CERR new C().m(new C());
        CERR new C().m((Object) new C());
        assert new C().m((I1) new C()) == 1;
        assert new C().m((I2) new C()) == 2;
        ''')
        check('''
        class C {
          int m(Object i, String j) { return 1; }
        }
        class D extends C {
          int m(String i, Object j) { return 2; }
        }
        CERR new D().m("", "");
        CERR new D().m(new Object(), new Object());
        assert new D().m((Object) "", "") == 1;
        assert new D().m("", (Object) "") == 2;
        assert new C().m("", "") == 1;
        ''')

if __name__ == '__main__':
    unittest.main()
