#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# test_grammar.py
#
# 2007-06-10  Jori Dubrovin:  Adding recurseStr API
# 2006-08-15  Jori Dubrovin:  Reflected changes in ExprStatement.__str__
# 2005-12-15  Jori Dubrovin:  Added modifiers to parameters and local variables
# 2005-11-09  Jori Dubrovin:  Added super field access and super method calls
# 2005-08-30  Jori Dubrovin:  Updated and added tests.
# 2005-07-29  Jori Dubrovin:  Added parentheses.  Needs update since 07-21.
# 2005-06-16  Jori Dubrovin:  Added tests for classes and methods etc.
# 2005-05-02  Jori Dubrovin:  Added tests for loop and condition statements
# 2005-04-22  Jori Dubrovin:  Initial version
#

"""Unit test for grammar.py."""

import unittest

import grammar
import lexical
import errors


class GrammarTest(unittest.TestCase):
    """Test parsing and grammar rules.

    The input string is given to the parser, which returns the root node
    of the syntax tree.  The tree is converted back to a string, whose
    contents are checked to be lexically equivalent to a reference string.
    """

    def __lexCompare(self, test, ref):
        """Verify that two strings are lexically equivalent.

        Use lexical.py to compare the two strings test (string under
        test) and ref (reference string).  If the returned tokens are
        not equivalent, or if test cannot be lexed, call self.__fail.
        Otherwise return.  Line numbers are ignored.
        """
        refLexer = lexical.lexer()
        refLexer.input(ref)
        testLexer = lexical.lexer()
        testLexer.input(test)
        while True:
            refToken = refLexer.token()
            try:
                testToken = testLexer.token()
            except errors.StaticError, e:
                self.fail('Test string cannot be lexed: ' + str(e))
            if refToken is None:
                if testToken is None:
                    break
                else:
                    self.fail('Expected end of tokens, found ' +
                              testToken.type)
            else:
                self.assertNotEqual(testToken, None,
                                    'Expected ' + refToken.type +
                                    ', found end of tokens')
                self.assertEqual(testToken.type, refToken.type,
                                 'Expected token type ' + refToken.type +
                                 ', found ' + testToken.type)
                self.assertEqual(testToken.value, refToken.value,
                                 'Expected token value ' +
                                 str(refToken.value) +
                                 ', found ' + str(testToken.value))

    def __fatal(self, input_):
        """Verify that the given input causes a fatal error when parsed."""
        self.assertRaises(errors.FatalStaticError, grammar.parse, input_)

    def __incomplete(self, input_):
        """Verify that the input causes an incompleteness error when parsed."""
        self.assertRaises(errors.IncompletenessError, grammar.parse, input_)

    def test_int(self):
        # Test integer literals.
        lc = self.__lexCompare
        p = grammar.parse
        lc(str(p('i = 0;')), 'i = (0);')
        lc(str(p('i = 1000000;')), 'i = (1000000);')
        lc(str(p('i = -200;')), 'i = (-(200));')
        lc(str(p('i = 0xffffffff;')), 'i = (0xffffffff);')
        lc(str(p('i = 037777777777;')), 'i = (0xffffffff);')
        lc(str(p('i = -0x7fffffff;')), 'i = (-(2147483647));')
        lc(str(p('i = 0400;')), 'i = (256);')
        lc(str(p('i = 2147483647;')), 'i = (2147483647);')
        lc(str(p('i = -2147483648;')), 'i = (-(0x80000000));')
        lc(str(p('i = -0x80000000;')), 'i = (-(0x80000000));')
        self.__fatal('i = 2147483648;')
        self.__fatal('i = 2147483649;')
        self.__fatal('i = 0x100000000;')
        self.__fatal('i = 040000000000;')

    def test_literals(self):
        # Test string literals, named literals and the this keyword.
        lc = self.__lexCompare
        p = grammar.parse
        lc(str(p('i = false;')), 'i = (false);')
        lc(str(p('i = true;')), 'i = (true);')
        lc(str(p('i = null;')), 'i = (null);')
        lc(str(p('i = this;')), 'i = (this);')
        lc(str(p('this++;')), '(this)++;')
        lc(str(p('true++;')), '(true)++;')
        lc(str(p('""++;')), '("")++;')
        lc(str(p('x = "foo";')), 'x = ("foo");')
        self.__fatal('false = 0;')
        self.__fatal('true = 0;')
        self.__fatal('null = 0;')
        self.__fatal('this = 0;')
        self.__fatal('x = a.null;')
        self.__fatal('x = a.this;')
        self.__fatal('x = this();')

    def test_localDecl(self):
        # Test local variable declarations.
        tree = grammar.parse("""
        boolean b;
        int i;
        boolean b2 = false;
        int i2, i3 = 10 + i, i4;
        oolean b3 = zooz;
        final int fi;
        final boolean fb = false, fb2;
        static final final abstract List l;
        """)
        self.__lexCompare(str(tree), """
        boolean b;
        int i;
        boolean b2 = false;
        int i2, i3 = (10) + (i), i4;
        oolean b3 = zooz;
        final int fi;
        final boolean fb = false, fb2;
        static final final abstract List l;
        """)
        self.__fatal('boolean C.x;')
        self.__fatal('boolean 5;')
        self.__fatal('boolean x,;')
        self.__fatal('boolean;')
        self.__fatal('boolean x = ;')
        self.__fatal('void x;')
        self.__fatal('hint;')

    def test_assignment(self):
        # Test assignment expressions.
        tree = grammar.parse("""
        i = 1;
        i *= j /= 10;
        i %= j + 100 * (k += 20 - i);
        i -= (-20);
        i <<= j >>= k >>>= i = 20;
        b &= false;
        b |= k++;
        b ^= b |= true;
        """)
        self.__lexCompare(str(tree), """
        i = (1);
        i *= (j /= (10));
        i %= ((j) + ((100) * (k += ((20) - (i)))));
        i -= (-(20));
        i <<= (j >>= (k >>>= (i = (20))));
        b &= (false);
        b |= ((k)++);
        b ^= (b |= (true));
        """)
        self.__fatal('10 = 10;')

    def test_condExpr(self):
        # Test conditional operators ?:, ||, &&.
        tree = grammar.parse("""
        a = condition ? true : condition2 ? x : y;
        b = (c = 10 ? d = 11 : b1 || b2);
        f = a || b && x || y ? c || d || e : c && d && e;
        k = a || b & c && d | e && f;
        """)
        self.__lexCompare(str(tree), """
        a = ((condition) ? (true) : ((condition2) ? (x) : (y)));
        b = (c = ((10) ? (d = (11)) : ((b1) || (b2))));
        f = ((((a) || ((b) && (x))) || (y)) ?
             (((c) || (d)) || (e)) : (((c) && (d)) && (e)));
        k = ((a) || ((((b) & (c)) && ((d) | (e))) && (f)));
        """)
        self.__fatal('b ? 1 : 2;')
        self.__fatal('b || c;')
        self.__fatal('b && c;')
        self.__fatal('b ? 1 : 2 = 5;')
        self.__fatal('b || c = 4;')
        self.__fatal('b && c = 4;')

    def test_bitlogicalExpr(self):
        # Test bit and logical operators &, |, ^.
        tree = grammar.parse("""
        a = x | y | z & p & q & r;
        b = x | y ^ z ^ p != r & q;
        """)
        self.__lexCompare(str(tree), """
        a = (((x) | (y)) | ((((z) & (p)) & (q)) & (r)));
        b = ((x) | (((y) ^ (z)) ^ (((p) != (r)) & (q))));
        """)
        self.__fatal('a & b;')
        self.__fatal('a | b;')
        self.__fatal('a ^ b;')
        self.__fatal('a & b = 0;')
        self.__fatal('a | b = 0;')
        self.__fatal('a ^ b = 0;')

    def test_relationalExpr(self):
        # Test equality and other relational operators ==, !=, <, >, <=, >=,
        # and instanceof.
        tree = grammar.parse("""
        a = b == c != d != e;
        a = b = c != d == e == f;
        a = b == c < d < e > f > g <= h <= i != k;
        a = b != c > d << 3 >= e;
        """)
        self.__lexCompare(str(tree), """
        a = ((((b) == (c)) != (d)) != (e));
        a = (b = ((((c) != (d)) == (e)) == (f)));
        a = (((b) == (((((((c) < (d)) < (e)) > (f)) > (g)) <= (h)) <= (i))) !=
             (k));
        a = ((b) != (((c) > ((d) << (3))) >= (e)));
        """)
        tree = grammar.parse("""
        x = b instanceof c;
        x = a < b instanceof c < d;
        x = b instanceof int[];
        """)
        self.__lexCompare(str(tree), """
        x = ((b) instanceof c);
        x = ((((a) < (b)) instanceof c) < (d));
        x = ((b) instanceof int[]);
        """)
        self.__fatal('x instanceof f;')
        self.__fatal('x = y instanceof int;')
        self.__fatal('x == 10;')
        self.__fatal('x != 10;')
        self.__fatal('x < 10;')
        self.__fatal('x > 10;')
        self.__fatal('x <= 10;')
        self.__fatal('x >= 10;')
        self.__fatal('x == 10 = false;')
        self.__fatal('x != 10 = false;')
        self.__fatal('x < 10 = false;')
        self.__fatal('x > 10 = false;')
        self.__fatal('x <= 10 = false;')
        self.__fatal('x >= 10 = false;')

    def test_shiftExpr(self):
        # Test shift operators <<, >>, >>>.
        tree = grammar.parse("""
        a = b << c << d;
        a = b >> c >> d;
        a = b >>> c >>> d;
        a = b << c >> d >>> e << f;
        a = b >> c + d;
        """)
        self.__lexCompare(str(tree), """
        a = (((b) << (c)) << (d));
        a = (((b) >> (c)) >> (d));
        a = (((b) >>> (c)) >>> (d));
        a = (((((b) << (c)) >> (d)) >>> (e)) << (f));
        a = ((b) >> ((c) + (d)));
        """)
        self.__fatal('a << b;')
        self.__fatal('a >> b;')
        self.__fatal('a >>> b;')
        self.__fatal('a << b = 1;')
        self.__fatal('a >> b = 1;')
        self.__fatal('a >>> b = 1;')

    def test_additiveExpr(self):
        # Test additive operators +, -.
        tree = grammar.parse("""
        a = b + c + d - e - f + g;
        a = 9 + 8 * 7 + 6;
        """)
        self.__lexCompare(str(tree), """
        a = ((((((b) + (c)) + (d)) - (e)) - (f)) + (g));
        a = (((9) + ((8) * (7))) + (6));
        """)
        self.__fatal('a - b;')
        self.__fatal('a + b;')
        self.__fatal('a - b = 1;')
        self.__fatal('a + b = 1;')

    def test_multiplicativeExpr(self):
        # Test multiplicative operators *, /, %.
        tree = grammar.parse("""
        a = b * c * d / e * f / g / h;
        a = b * --c % d * e % -f % ~g;
        """)
        self.__lexCompare(str(tree), """
        a = (((((((b) * (c)) * (d)) / (e)) * (f)) / (g)) / (h));
        a = ((((((b) * (--(c))) % (d)) * (e)) % (-(f))) % (~(g)));
        """)
        self.__fatal('a * b;')
        self.__fatal('a / b;')
        self.__fatal('a % b;')
        self.__fatal('a * b = 1;')
        self.__fatal('a / b = 1;')
        self.__fatal('a % b = 1;')

    def test_unaryExpr(self):
        # Test unary prefix and postfix operators -, +, !, ~, ++, --.
        tree = grammar.parse("""
        i++;
        (i+j)--;
        ++k++;
        --++X++--;
        a = ~b++ + -x * -y;
        c = ~-e--;
        d = !r++++;
        e = +3;
        """)
        self.__lexCompare(str(tree), """
        (i)++;
        ((i)+(j))--;
        ++((k)++);
        --(++(((X)++)--));
        a = ((~((b)++)) + ((-(x)) * (-(y))));
        c = (~(-((e)--)));
        d = (!(((r)++)++));
        e = (+(3));
        """)
        self.__fatal('~b;')
        self.__fatal('-b;')
        self.__fatal('!b;')
        self.__fatal('--a = 1;')
        self.__fatal('++a = 1;')
        self.__fatal('a-- = 1;')
        self.__fatal('a++ = 1;')
        self.__fatal('-a = 1;')
        self.__fatal('+a = 1;')
        self.__fatal('!a = 1;')
        self.__fatal('~a = 1;')

    def test_castExpr(self):
        # Test cast expressions.
        tree = grammar.parse(r"""
        x = (int) 7;
        x = (int) y++;
        x = (int) z * w;
        x = (int[]) -7;
        x = (boolean) +8;
        x = (Object) +10; // not a cast
        x = (Object) -10; // not a cast
        x = (Object) x;
        x = (int[]) y;
        x = (My.String[]) "foo";
        foo = (5 + 6) 7;
        """)
        self.__lexCompare(str(tree), """
        x = ((int) (7));
        x = ((int) ((y)++));
        x = (((int) (z)) * (w));
        x = ((int[]) (-(7)));
        x = ((boolean) (+(8)));
        x = ((Object) + (10));
        x = ((Object) - (10));
        x = ((Object) (x));
        x = ((int[]) (y));
        x = ((My.String[]) ("foo"));
        foo = (((5) + (6)) (7));
        """)

    def test_paren(self):
        # Test parentheses in expressions.
        tree = grammar.parse("""
        a = b = (c + (d = 10));
        a = ((10 + 5) * 11 - 3) << 2;
        x = (y) * -(y);
        (x + 100 * (y | 1) == !(t ? (r = 11) : (r = 11)))--;
        """)
        self.__lexCompare(str(tree), """
        a = (b = ((c) + (d = (10))));
        a = (((((10) + (5)) * (11)) - (3)) << (2));
        x = ((y) * (-(y)));
        (((x) + ((100) * ((y) | (1))))
         == (!((t) ? (r = (11)) : (r = (11)))))--;
        """)
        self.__fatal('(a) = 1;')

    def test_methodInvocation(self):
        # Test method invocation expressions.
        tree = grammar.parse("""
        Cl.meth();
        Cl.meth(1);
        f = 2 + Cl.meth(x, y=4);
        """)
        self.__lexCompare(str(tree), """
        Cl.meth();
        Cl.meth(1);
        f = ((2) + (Cl.meth(x, y = (4))));
        """)
        tree = grammar.parse("""
        int y = a().b();
        c().d().e();
        c.d.e(x.y, z);
        int z = c + d.f();
        int z = (c + d).f();
        """)
        self.__lexCompare(str(tree), """
        int y = (a()).b();
        ((c()).d()).e();
        c.d.e(x.y, z);
        int z = (c) + (d.f());
        int z = ((c) + (d)).f();
        """)
        self.__incomplete('Cl.meth()')
        self.__fatal('Cl.meth() = 4;')
        self.__fatal('Cl.();')
        self.__fatal('meth(int a);')
        self.__fatal('static meth(3);')
        self.__fatal('Cl.meth(a, );')
        self.__fatal('Cl.meth(, a);')
        self.__fatal('.meth();')
        self.__fatal('Cl.meth;')
        self.__fatal('Cl.meth((a, b));')
        self.__fatal('Cl.meth x);')

    def test_fieldAccess(self):
        # Test field access and qualified name expressions.
        tree = grammar.parse("""
        C.field = 3;
        C.f++;
        C.f.g--;
        --C.z;
        ++g().z;
        y = (1 + 2).x;
        this.i = that.j;
        (t++).i = ++t.i;
        """)
        self.__lexCompare(str(tree), """
        C.field = (3);
        (C.f)++;
        (C.f.g)--;
        --(C.z);
        ++((g()).z);
        y = (((1) + (2)).x);
        (this).i = (that.j);
        ((t)++).i = (++(t.i));
        """)
        tree = grammar.parse("""
        A.b.c *= 10;
        A().b.c *= 10;
        A.b().c *= 10;
        """)
        self.__lexCompare(str(tree), """
        A.b.c *= (10);
        ((A()).b).c *= (10);
        (A.b()).c *= (10);
        """)
        self.__fatal('A.b.c() *= 10;')

    def test_instanceCreation(self):
        # Test instance creation expressions.
        tree = grammar.parse("""
        new C();
        new C.D();
        c = new C();
        new C(1).x++;
        new C(new D(), 5 * 6);
        """)
        self.__lexCompare(str(tree), """
        new C();
        new C.D();
        c = (new C());
        ((new C(1)).x)++;
        new C(new D(), (5) * (6));
        """)
        self.__fatal('new C.D.E() = null;')
        self.__fatal('new C;')
        self.__fatal('new int();')

    def test_labeledStatements(self):
        # Test labeled statements.
        tree = grammar.parse("""
        lab: p = 0;
        lab: { }
        lab1: lab2: lab1: new T();
        if (true) _: x++;
        if (true) _: x++; else y: y++;
        _5: return 5;
        """)
        self.__lexCompare(str(tree), """
        lab: p = (0);
        lab: { }
        lab1: lab2: lab1: new T();
        if (true) _: (x)++; else ;
        if (true) _: (x)++; else y: (y)++;
        _5: return 5;
        """)
        self.__fatal('label { }')
        self.__fatal('label: int i;')
        self.__fatal('label: class C { }')
        self.__fatal('label: enum E { }')

    def test_simpleStatements(self):
        # Test empty, break and continue statements.
        tree = grammar.parse("""
        break;
        ;
        continue;
        """)
        self.__lexCompare(str(tree), """
        break;
        ;
        continue;
        """)
        tree = grammar.parse("""
        break myleg;
        continue here;
        """)
        self.__lexCompare(str(tree), """
        break myleg;
        continue here;
        """)
        self.__fatal('break (x);')
        self.__fatal('continue (x);')
        self.__fatal('break x y;')
        self.__fatal('break x + y;')
        self.__incomplete('break')
        self.__incomplete('continue')

    def test_returnStatements(self):
        # Test return statements.
        tree = grammar.parse("""
        return;
        return x;
        return ++x;
        { return; }
        """)
        self.__lexCompare(str(tree), """
        return;
        return x;
        return ++(x);
        { return; }
        """)
        self.__fatal('return = 0')
        self.__fatal('return void;')
        self.__fatal('return return;')
        self.__incomplete('return')

    def test_assertStatements(self):
        # Test assert statements.
        tree = grammar.parse("""
        assert x;
        assert 5 + 6;
        assert t = 7;
        { assert false; }
        """)
        self.__lexCompare(str(tree), """
        assert x;
        assert (5) + (6);
        assert t = (7);
        { assert false; }
        """)
        self.__fatal('assert;')
        self.__fatal('assert false assert;')
        self.__fatal('assert false : "message";')
        self.__fatal('z = assert false;')

    def test_sendStatements(self):
        # Test send statements.
        tree = grammar.parse("""
        send sig() to x;
        send more(1) to 3;
        send much(more, 5 + 6) to x = y;
        """)
        self.__lexCompare(str(tree), """
        send sig() to x;
        send more(1) to 3;
        send much(more, (5) + (6)) to x = (y);
        """)
        self.__fatal('send no_paren to x;')
        self.__fatal('send SomeClass.sig() to x;')
        self.__fatal('send 5() to x;')
        self.__fatal('send int() to x;')
        self.__fatal('send () to x;')
        self.__fatal('send sig();')
        self.__fatal('send to x;')
        self.__fatal('send sig() x;')
        self.__fatal('send sig() to;')
        self.__fatal('send sig() to int;')
        self.__incomplete('send sig() to x')

    def test_ifStatements(self):
        # Test if-then and if-then-else statements.
        tree = grammar.parse("""
        if (1 + 1 == 2)
          math = ok;
        else
          math = strange;
        i = 0;
        """)
        self.__lexCompare(str(tree), """
        if (((1) + (1)) == (2))
          math = (ok);
        else
          math = (strange);
        i = (0);
        """)
        tree = grammar.parse("""
        if (x) {
          two++;
          things++;
        }
        """)
        self.__lexCompare(str(tree), """
        if (x) {
          (two)++;
          (things)++;
        }
        else
          ;
        """)
        tree = grammar.parse("""
        if (outer)
          if (inner)
            misleading = 1;
        else {
          indentation = 1;
        }
        """)
        self.__lexCompare(str(tree), """
        if (outer)
          if (inner)
            misleading = (1);
          else {
            indentation = (1);
          }
        else
          ;
        """)
        tree = grammar.parse("""
        if (outer)
          if (inner)
            a = 1;
          else
            a = 2;
        else
          a = 3;
        """)
        self.__lexCompare(str(tree), """
        if (outer)
          if (inner)
            a = (1);
          else
            a = (2);
        else
          a = (3);
        """)
        self.__fatal('if ();')
        self.__fatal('if true ;')
        self.__fatal('if (x) else i++;')
        self.__fatal('if (x) i++; else j++; else z++;')
        self.__fatal('if (x) int j = 4;')

    def test_whileStatements(self):
        # Test while and do-while statements.
        tree = grammar.parse("""
        while (i > 0)
          i--;
        """)
        self.__lexCompare(str(tree), """
        while ((i) > (0))
          (i)--;
        """)
        tree = grammar.parse("""
        if (x)
          while (i > 0)
            i--;
        i++;
        """)
        self.__lexCompare(str(tree), """
        if (x)
          while ((i) > (0))
            (i)--;
        else
          ;
        (i)++;
        """)
        tree = grammar.parse("""
        if (x)
          while (i > 0)
            if (i > 10)
              misleading = 1;
        else
          indentation = 1;
        """)
        self.__lexCompare(str(tree), """
        if (x)
          while ((i) > (0))
            if ((i) > (10))
              misleading = (1);
            else
              indentation = (1);
        else
          ;
        """)
        tree = grammar.parse("""
        if (x)
          while (r--) {
            t++;
            u++;
          }
        else
          x = true;
        """)
        self.__lexCompare(str(tree), """
        if (x)
          while ((r)--) {
            (t)++;
            (u)++;
          }
        else
          x = (true);
        """)
        tree = grammar.parse("""
        do {
          s = null;
        } while (s);
        """)
        self.__lexCompare(str(tree), """
        do {
          s = (null);
        } while (s);
        """)
        self.__incomplete('while (i > 0)')
        self.__incomplete('do i--; while (i > 0)')
        self.__fatal('while () ;')
        self.__fatal('while true i = 2;')
        self.__fatal('do while (5); x = 1;')
        self.__fatal('do { r++; } while (r < 10) {}')

    def test_forStatements(self):
        # Test for statements.
        tree = grammar.parse("""
        for (int i = 10, j = 22; i < 20 && j < 100; i++, j++) {
          x += i + j;
        }
        """)
        self.__lexCompare(str(tree), """
        for (int i = 10, j = 22; ((i) < (20)) && ((j) < (100)); (i)++, (j)++) {
          x += ((i) + (j));
        }
        """)
        tree = grammar.parse("""
        for (i = 10, j = 22; false; )
          ;
        """)
        self.__lexCompare(str(tree), """
        for (i = (10), j = (22); false; )
          ;
        """)
        tree = grammar.parse("""
        if (x)
          for (int i; ; i = i + 19)
            x--;
        else
          x = 0;
        """)
        self.__lexCompare(str(tree), """
        if (x)
          for (int i; true; i = ((i) + (19)))
            (x)--;
        else
          x = (0);
        """)
        tree = grammar.parse("""
        if (x < 20)
          for (; x < 30; x++) {
            x--;
            x *= x;
          }
        else
          ;
        """)
        self.__lexCompare(str(tree), """
        if ((x) < (20))
          for (; (x) < (30); (x)++) {
            (x)--;
            x *= (x);
          }
        else
          ;
        """)
        self.__fatal('for (;) ;')
        self.__fatal('for (void x;;) ;')
        self.__fatal('for (x = 4; x; int y = x + 1) ;')
        self.__incomplete('for (;;)')
        self.__fatal('for (;;) int j = i++;')
        self.__fatal('for (x++, int y = x; x < 4; x++) ;')

    def test_switchStatements(self):
        # Test switch statements.
        tree = grammar.parse("""
        switch (a)
        {
        }
        """)
        self.__lexCompare(str(tree), """
        switch (a)
        {
        }
        """)
        tree = grammar.parse("""
        switch (z)
        {
        case z+1:
          strange++;
        case 0:
        }
        """)
        self.__lexCompare(str(tree), """
        switch (z)
        {
        case (z)+(1):
          (strange)++;
        case 0:
        }
        """)
        tree = grammar.parse("""
        switch (500+600*3)
        {
        default:
          int r;
          r = r + 1;
          break;
        case r:
        case 9:
          {
            r = 0;
          }
        case 0:
          break;
        }
        """)
        self.__lexCompare(str(tree), """
        switch ((500)+((600)*(3)))
        {
        default:
          int r;
          r = ((r) + (1));
          break;
        case r:
        case 9:
          {
            r = (0);
          }
        case 0:
          break;
        }
        """)
        self.__fatal('switch (if (x) i; else j;) { }')
        self.__fatal('switch () { }')
        self.__fatal('switch (o) {int y; case 2: break;}')
        self.__fatal('switch (o) {case: break;}')
        self.__fatal('switch (o) {case 4: { j++; case 5: i++; } }')
        self.__fatal('case 20: break;')
        self.__fatal('default: break;')

    def test_arrays(self):
        # Test array types.
        tree = grammar.parse("""
        T[] a, b = new T[3];
        b = (T[]) c + (int[]) c + (T[][]) c;
        int[][][][][] i = new int[-123][][][][];
        class C
        {
          P[][] p;
          boolean[] f(C[][][][] c);
        }
        """)
        self.__lexCompare(str(tree), """
        T[] a, b = new T[3];
        b = ((((T[]) (c)) + ((int[]) (c))) + ((T[][]) (c)));
        int[][][][][] i = new int[-(123)][][][][];
        class C
        {
          P[][] p;
          boolean[] f(C[][][][] c);
        }
        """)
        tree = grammar.parse("""
        i = new T[] { };
        i = new T[][] { { null }, null, null };
        T[] i = { };
        class C { T[][] i = { { null, null } }; }
        """)
        self.__lexCompare(str(tree), """
        i = (new T[] { });
        i = (new T[][] { { null }, null, null });
        T[] i = { };
        class C { T[][] i = { { null, null } }; }
        """)
        self.__fatal('class C extends D[] { }')
        self.__fatal('class C implements D[] { }')
        self.__fatal('interface C extends D[] { }')
        self.__fatal('i = new T[];')
        self.__fatal('i = new T[2][3];')
        self.__fatal('i = new T[][3];')
        self.__fatal('i = new T[2] { null };')

    def test_classDecl(self):
        # Test class declarations.
        tree = grammar.parse("""
        class C1 {}
        (x)++;
        class C2
        {
          static void foo()
          {
            t = (4);
          }
        }
        """)
        self.__lexCompare(str(tree), """
        class C1 {}
        (x)++;
        class C2
        {
          static void foo()
          {
            t = (4);
          }
        }
        """)
        tree = grammar.parse("""
        class Outer
        {
          class Nested1
          {
            class Deep
            {
            }
          }
          class Nested2
          {
          }
        }
        """)
        self.__lexCompare(str(tree), """
        class Outer
        {
          class Nested1
          {
            class Deep
            {
            }
          }
          class Nested2
          {
          }
        }
        """)  
        tree = grammar.parse("""
        class C
        {
          int y;
          Foo x;
          class D
          {
            native boolean r;
            static static D self;
          }
          Outer.Nested z;
        }
        """)
        self.__lexCompare(str(tree), """
        class C
        {
          int y;
          Foo x;
          class D
          {
            native boolean r;
            static static D self;
          }
          Outer.Nested z;
        }
        """)
        tree = grammar.parse("""
        class C extends D { }
        final class E implements F { }
        static final class G extends W implements X, Y, Z { }
        """)
        self.__lexCompare(str(tree), """
        class C extends D { }
        final class E implements F { }
        static final class G extends W implements X, Y, Z { }
        """)
        self.__fatal('(class C {})')
        self.__fatal('class int {}')
        self.__fatal('class void {}')
        self.__fatal('class C { class D; }')
        self.__fatal('class C { void d; }')
        self.__fatal('class C { Foo() x; }')
        self.__fatal('u = class C {}')
        self.__fatal('class C;')
        self.__fatal('class C static void foo() {}')
        self.__fatal('class { static void foo() {} }')
        self.__fatal('class C { ; }')
        self.__fatal('{ class C {} }')
        self.__fatal('if (false) { class C {} }')
        self.__fatal('while (true) { class C {} }')
        self.__fatal('do { class C {} } while (false);')
        self.__fatal('for (;;) { class C {} }')
        self.__fatal('switch (1) { class C {} }')
        self.__fatal('class C { static void foo() { class D {} } }')
        self.__fatal('class C extends D, E { }')
        self.__fatal('class C implements D extends E { }')
        self.__fatal('class C extends { }')
        self.__fatal('class C extends implements E { }')
        self.__fatal('class C implements { }')

    def test_interfaceDecl(self):
        # Test interface declarations.
        tree = grammar.parse("""
        interface I1 {}
        x++;
        interface I2
        {
          static void foo();
        }
        """)
        self.__lexCompare(str(tree), """
        interface I1 {}
        (x)++;
        interface I2
        {
          static void foo();
        }
        """)
        tree = grammar.parse("""
        interface Outer
        {
          class Nested1
          {
            interface Deep
            {
            }
          }
          interface Nested2
          {
          }
        }
        """)
        self.__lexCompare(str(tree), """
        interface Outer
        {
          class Nested1
          {
            interface Deep
            {
            }
          }
          interface Nested2
          {
          }
        }
        """)  
        tree = grammar.parse("""
        interface C
        {
          int y;
          Foo x;
          interface D
          {
            native boolean r;
            static static D self;
          }
          Outer.Nested z;
          int hasBody() { }
        }
        """)
        self.__lexCompare(str(tree), """
        interface C
        {
          int y;
          Foo x;
          interface D
          {
            native boolean r;
            static static D self;
          }
          Outer.Nested z;
          int hasBody() { }
        }
        """)
        tree = grammar.parse("""
        interface C extends D { }
        interface G extends W, Y, Z { }
        """)
        self.__lexCompare(str(tree), """
        interface C extends D { }
        interface G extends W, Y, Z { }
        """)
        self.__fatal('(interface C {})')
        self.__fatal('interface int {}')
        self.__fatal('interface void {}')
        self.__fatal('interface C { interface D; }')
        self.__fatal('interface C { void d; }')
        self.__fatal('interface C { Foo() x; }')
        self.__fatal('interface C { C() { } }')
        self.__fatal('u = interface C {}')
        self.__fatal('interface C;')
        self.__fatal('interface C static void foo();')
        self.__fatal('interface { static void foo(); }')
        self.__fatal('interface C { ; }')
        self.__fatal('{ interface C {} }')
        self.__fatal('if (false) { interface C {} }')
        self.__fatal('while (true) { interface C {} }')
        self.__fatal('do { interface C {} } while (false);')
        self.__fatal('for (;;) { interface C {} }')
        self.__fatal('switch (1) { interface C {} }')
        self.__fatal('class C { static void foo() { interface D {} } }')
        self.__fatal('interface C extends { }')
        self.__fatal('interface C implements')
        self.__fatal('static interface C { }')
        self.__fatal('abstract interface C { }')
        self.__fatal('native interface C { }')

    def test_enumDecl(self):
        # Test enum declarations.
        tree = grammar.parse("""
        enum E { x }
        class C { enum E { x, y, z } }
        interface I { enum E { x, y, z, } }
        enum E { }
        """)
        self.__lexCompare(str(tree), """
        enum E { x }
        class C { enum E { x, y, z } }
        interface I { enum E { x, y, z } }
        enum E { }
        """)
        self.__fatal('enum;')
        self.__fatal('enum { }')
        self.__fatal('{ enum E { x } }')
        self.__fatal('enum { x }')
        self.__fatal('enum E extends F { x }')
        self.__fatal('enum E { x; }')
        self.__fatal('enum E { , x }')
        self.__fatal('enum E { y,, x }')
        self.__fatal('enum E { x() }')
        self.__fatal('enum E.F { x }')
        self.__fatal('class C { void f() { enum E { x } } }')

    def test_fieldDecl(self):
        # Test member variable declarations.
        tree = grammar.parse("""
        class C
        {
          int i;
          F.G k = null;
          T[] a, b, c = 45, d;
          static final abstract native final boolean y;
        }
        """)
        self.__lexCompare(str(tree), """
        class C
        {
          int i;
          F.G k = null;
          T[] a, b, c = 45, d;
          static final abstract native final boolean y;
        }
        """)
        tree = grammar.parse("""
        interface C
        {
          int i;
          F.G k = null;
          T[] a, b, c = 45, d;
          static final abstract native final boolean y;
        }
        """)
        self.__lexCompare(str(tree), """
        interface C
        {
          int i;
          F.G k = null;
          T[] a, b, c = 45, d;
          static final abstract native final boolean y;
        }
        """)

    def test_methodDecl(self):
        # Test method declarations.
        tree = grammar.parse("""
        class C
        {
          static int m(int x)
          {
            int y = 4;
          }
          static void m(boolean x, final int y) {}
          abstract static native native boolean m()
          {
            return false;
          }
        }
        """)
        self.__lexCompare(str(tree), """
        class C
        {
          static int m(int x)
          {
            int y = 4;
          }
          static void m(boolean x, final int y) {}
          abstract static native native boolean m()
          {
            return false;
          }
        }
        """)
        tree = grammar.parse("""
        class C
        {
          class D
          {
            C meth(C a, C.D b);
            D meth() { x++; }
            void meth(int y) {}
            boolean meth();
          }
        }
        """)
        self.__lexCompare(str(tree), """
        class C
        {
          class D
          {
            C meth(C a, C.D b);
            D meth() { (x)++; }
            void meth(int y) {}
            boolean meth();
          }
        }
        """)
        self.__fatal('class C { static foo() {} }')
        self.__fatal('class C { void f() { super(); } }')
        self.__fatal('class C { static int () {} }')
        self.__fatal('class C { native int meth(void x); }')
        self.__fatal('class C { static int foo {} }')
        self.__fatal('class C { static int x.y() {} }')
        self.__fatal('class C { static int x() }')
        self.__fatal('class C { static void a(,) {} }')
        self.__fatal('class C { static void a(x) {} }')
        self.__fatal('class C { static void a(, int x) {} }')
        self.__fatal('class C { static void a(int x, ) {} }')
        self.__fatal('class C { static void a(int x, y) {} }')

    def test_constructorDecl(self):
        # Test constructor declarations.
        tree = grammar.parse('class C { foo() {} }')
        self.__lexCompare(str(tree), 'class C { foo() {} }')
        tree = grammar.parse("""
        class C
        {
          C() { int y; int z; }
          int x = 12;
          C(C c) { }
          C(final boolean x, String s) { super(); }
          C() { super(13, 15 + 6); x++; super(); }
        }
        """)
        self.__lexCompare(str(tree), """
        class C
        {
          C() { int y; int z; }
          int x = 12;
          C(C c) { }
          C(final boolean x, String s) { super(); }
          C() { super(13, (15) + (6)); (x)++; super(); }
        }
        """)
        self.__fatal('class C { D.C() {} }')
        self.__fatal('class C { static C() {} }')
        self.__fatal('class C { native C() {} }')
        self.__fatal('class C { abstract C() {} }')
        self.__fatal('interface I { I() {} }')
        self.__fatal('enum E { E() {} }')
        self.__fatal('C() {}')


if __name__ == '__main__':
    unittest.main()
