#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# test_lexical.py
#
# 2005-08-26  Jori Dubrovin:  Added keyword assert
# 2005-07-27  Jori Dubrovin:  Added keywords
# 2005-06-16  Jori Dubrovin:  Added keywords
# 2005-04-25  Jori Dubrovin:  Reflected changes in error and integer handling
# 2005-04-22  Jori Dubrovin:  Initial version
#

"""Unit test for lexical.py."""

import unittest

import lexical
import errors


class LexTest(unittest.TestCase):

    def setUp(self):
        # Get a fresh copy of the lexer
        self.__lexer = lexical.lexer()

    def tearDown(self):
        self.__lexer = None

    def __expect(self, type_, value=None, lineno=None):
        """Check that the next token is of given type.

        Also check the value, if given, and/or line number, if given.
        """
        token = self.__lexer.token()
        self.assertNotEqual(token, None,
                            'Expected ' + type_ + ', found end of tokens')
        self.assertEqual(token.type, type_,
                         'Expected token type ' + type_ +
                         ', found ' + token.type)
        if value is not None:
            self.assertEqual(token.value, value,
                             'Expected token value ' + repr(value) +
                             ', found ' + repr(token.value))
        if lineno is not None:
            self.assertEqual(token.lineno, lineno,
                             'Expected line number ' + str(lineno) +
                             ', found ' + str(token.lineno))

    def __expect_end(self):
        """Check that there are no more tokens."""
        token = self.__lexer.token()
        if token is not None:
            self.fail('Expected end of tokens, found ' + token.type)

    def __fatal(self, input_):
        """Check that the input immediately causes a FatalStaticError."""
        self.__lexer.input(input_)
        self.assertRaises(errors.FatalStaticError, self.__lexer.token)
        # Get a fresh copy of the lexer
        self.__lexer = lexical.lexer()

    def __incomplete(self, input_):
        """Check that the input immediately causes an IncompletenessError."""
        self.__lexer.input(input_)
        self.assertRaises(errors.IncompletenessError, self.__lexer.token)
        # Get a fresh copy of the lexer
        self.__lexer = lexical.lexer()

    def test_reserved(self):
        # Test all reserved words
        self.__lexer.input('''
        abstract assert boolean break case class continue default do
        else enum extends false final for if implements instanceof int
        interface native new null return send static super switch this
        to true void while
        ''')
        self.__expect('ABSTRACT')
        self.__expect('ASSERT')
        self.__expect('BOOLEAN')
        self.__expect('BREAK')
        self.__expect('CASE')
        self.__expect('CLASS')
        self.__expect('CONTINUE')
        self.__expect('DEFAULT')
        self.__expect('DO')
        self.__expect('ELSE')
        self.__expect('ENUM')
        self.__expect('EXTENDS')
        self.__expect('FALSE')
        self.__expect('FINAL')
        self.__expect('FOR')
        self.__expect('IF')
        self.__expect('IMPLEMENTS')
        self.__expect('INSTANCEOF')
        self.__expect('INT')
        self.__expect('INTERFACE')
        self.__expect('NATIVE')
        self.__expect('NEW')
        self.__expect('NULL')
        self.__expect('RETURN')
        self.__expect('SEND')
        self.__expect('STATIC')
        self.__expect('SUPER')
        self.__expect('SWITCH')
        self.__expect('THIS')
        self.__expect('TO')
        self.__expect('TRUE')
        self.__expect('VOID')
        self.__expect('WHILE')
        self.__expect_end()

    def test_separators(self):
        # Test all separators
        self.__lexer.input(' (){}[];,. ')
        self.__expect('LPAREN')
        self.__expect('RPAREN')
        self.__expect('LBRACE')
        self.__expect('RBRACE')
        self.__expect('LBRACKET')
        self.__expect('RBRACKET')
        self.__expect('SEMICOLON')
        self.__expect('COMMA')
        self.__expect('PERIOD')
        self.__expect_end()

    def test_operators(self):
        # Test all operators
        self.__lexer.input(r'''
        =><!~?:==<=>=!=&&||++--+-*/&|^%<<>> >>>+=-=*=/=&=|=^=%=<<=>>=>>>=
        ''')
        self.__expect('EQ')
        self.__expect('GT')
        self.__expect('LT')
        self.__expect('EXCLAM')
        self.__expect('TILDE')
        self.__expect('QUESTION')
        self.__expect('COLON')
        self.__expect('EQEQ')
        self.__expect('LTEQ')
        self.__expect('GTEQ')
        self.__expect('EXCLAMEQ')
        self.__expect('AMPERAMPER')
        self.__expect('VERTVERT')
        self.__expect('PLUSPLUS')
        self.__expect('MINUSMINUS')
        self.__expect('PLUS')
        self.__expect('MINUS')
        self.__expect('ASTER')
        self.__expect('SLASH')
        self.__expect('AMPER')
        self.__expect('VERT')
        self.__expect('CARET')
        self.__expect('PERCENT')
        self.__expect('LTLT')
        self.__expect('GTGT')
        self.__expect('GTGTGT')
        self.__expect('PLUSEQ')
        self.__expect('MINUSEQ')
        self.__expect('ASTEREQ')
        self.__expect('SLASHEQ')
        self.__expect('AMPEREQ')
        self.__expect('VERTEQ')
        self.__expect('CARETEQ')
        self.__expect('PERCENTEQ')
        self.__expect('LTLTEQ')
        self.__expect('GTGTEQ')
        self.__expect('GTGTGTEQ')
        self.__expect_end()
        self.__fatal('#')
        self.__fatal('@')
        self.__fatal('\\')
        self.__fatal('`')

    def test_identifiers(self):
        # Test identifiers
        self.__lexer.input('''
        a A var caps_lock integer Int Boolean BOOLEAN boolean_1
        _0 _ $ $$ $9xy five$ 2nd_ no"good" loop9876543210.2
        ''')
        self.__expect('IDENTIFIER', 'a')
        self.__expect('IDENTIFIER', 'A')
        self.__expect('IDENTIFIER', 'var')
        self.__expect('IDENTIFIER', 'caps_lock')
        self.__expect('IDENTIFIER', 'integer')
        self.__expect('IDENTIFIER', 'Int')
        self.__expect('IDENTIFIER', 'Boolean')
        self.__expect('IDENTIFIER', 'BOOLEAN')
        self.__expect('IDENTIFIER', 'boolean_1')
        self.__expect('IDENTIFIER', '_0')
        self.__expect('IDENTIFIER', '_')
        self.__expect('IDENTIFIER', '$')
        self.__expect('IDENTIFIER', '$$')
        self.__expect('IDENTIFIER', '$9xy')
        self.__expect('IDENTIFIER', 'five$')
        self.__expect('INT_LITERAL', 2)
        self.__expect('IDENTIFIER', 'nd_')
        self.__expect('IDENTIFIER', 'no')
        self.__expect('STRING_LITERAL', 'good')
        self.__expect('IDENTIFIER', 'loop9876543210')
        self.__expect('PERIOD')
        self.__expect('INT_LITERAL', 2)
        self.__expect_end()
        
    def test_decimal(self):
        # Test decimal integer literals
        self.__lexer.input('''
        0 1 1234567890 98765432109876543210 -10
        2147483647 2147483648 2147483649
        ''')
        self.__expect('INT_LITERAL', 0)
        self.__expect('INT_LITERAL', 1)
        self.__expect('INT_LITERAL', 1234567890)
        self.__expect('INT_LITERAL_OVERFLOW', '98765432109876543210')
        self.__expect('MINUS')
        self.__expect('INT_LITERAL', 10)
        self.__expect('INT_LITERAL', 2147483647)
        self.__expect('INT_LITERAL_MUST_NEGATE', -2147483648)
        self.__expect('INT_LITERAL_OVERFLOW', '2147483649')
        self.__expect_end()

    def test_octal(self):
        # Test octal integer literals
        self.__lexer.input('''
        01 00 030 0004 07654321 0123456701234567 055581 09
        017777777777 020000000000 020000000001 037777777777
        0000030203040306 040000000000
        ''')
        self.__expect('INT_LITERAL', 01)
        self.__expect('INT_LITERAL', 00)
        self.__expect('INT_LITERAL', 030)
        self.__expect('INT_LITERAL', 0004)
        self.__expect('INT_LITERAL', 07654321)
        self.__expect('INT_LITERAL_OVERFLOW', '0123456701234567')
        self.__expect('INT_LITERAL', 0555)
        self.__expect('INT_LITERAL', 81)  # 81 is decimal, not octal
        self.__expect('INT_LITERAL', 0)
        self.__expect('INT_LITERAL', 9)   # 9 is decimal, not octal
        self.__expect('INT_LITERAL', 017777777777)
        self.__expect('INT_LITERAL', 020000000000L - (1L<<32))
        self.__expect('INT_LITERAL', 020000000001L - (1L<<32))
        self.__expect('INT_LITERAL', -1)
        self.__expect('INT_LITERAL', 030203040306L - (1L<<32))
        self.__expect('INT_LITERAL_OVERFLOW', '040000000000')
        self.__expect_end()

    def test_hexadecimal(self):
        # Test hexadecimal integer literals
        self.__lexer.input('''
        0x0 0x 0x1234567 0x89abc 0xdefg 0xFED0CBA 0Xc00 -0X2d
        0x123456789abcdef  0x7fFfFfFf 0x000080000000 0x80000001
        0xa4054920 0xffffffff 0x100000000
        ''')
        self.__expect('INT_LITERAL', 0x0)
        self.__expect('INT_LITERAL', 0)
        self.__expect('IDENTIFIER', 'x')
        self.__expect('INT_LITERAL', 0x1234567)
        self.__expect('INT_LITERAL', 0x89abc)
        self.__expect('INT_LITERAL', 0xdef)
        self.__expect('IDENTIFIER', 'g')
        self.__expect('INT_LITERAL', 0xFED0CBA)
        self.__expect('INT_LITERAL', 0Xc00)
        self.__expect('MINUS')
        self.__expect('INT_LITERAL', 0X2d)
        self.__expect('INT_LITERAL_OVERFLOW', '0x123456789abcdef')
        self.__expect('INT_LITERAL', 0x7fffffff)
        self.__expect('INT_LITERAL', -0x80000000L)
        self.__expect('INT_LITERAL', 0x80000001L - (1L<<32))
        self.__expect('INT_LITERAL', 0xa4054920L - (1L<<32))
        self.__expect('INT_LITERAL', -1)
        self.__expect('INT_LITERAL_OVERFLOW', '0x100000000')
        self.__expect_end()

    def test_string(self):
        # Test string literals
        self.__lexer.input(r'''
        "0x9boolean )"  " "  " ."  ""  "''" "\b\t\n\f\r" "\"\'" "\\"
        "\0\00\000" "\1\01\377\74" "\0051\08\400" "\501\662\774\333\2"
        "\\4" "\\5" "\\6" "\\7" "\\8" "\\\4"
        "/**/"
        ''')
        self.__expect('STRING_LITERAL', '0x9boolean )')
        self.__expect('STRING_LITERAL', ' ')
        self.__expect('STRING_LITERAL', ' .')
        self.__expect('STRING_LITERAL', '')
        self.__expect('STRING_LITERAL', "''")
        self.__expect('STRING_LITERAL', chr(8)+chr(9)+chr(10)+chr(12)+chr(13))
        self.__expect('STRING_LITERAL', '\"\'')
        self.__expect('STRING_LITERAL', '\\')
        self.__expect('STRING_LITERAL', chr(00)+chr(000)+chr(0000))
        self.__expect('STRING_LITERAL', chr(01)+chr(001)+chr(0377)+chr(074))
        self.__expect('STRING_LITERAL',
                      chr(0005)+'1'+chr(00)+'8'+chr(040)+'0')
        self.__expect('STRING_LITERAL',
                      chr(050)+'1'+chr(066)+'2'+chr(077)+'4'+chr(0333)+chr(2))
        self.__expect('STRING_LITERAL', '\\4')
        self.__expect('STRING_LITERAL', '\\5')
        self.__expect('STRING_LITERAL', '\\6')
        self.__expect('STRING_LITERAL', '\\7')
        self.__expect('STRING_LITERAL', '\\8')
        self.__expect('STRING_LITERAL', '\\' + chr(4))
        self.__expect('STRING_LITERAL', '/**/')
        self.__expect_end()
        self.__fatal(r'"')
        self.__fatal(r'"foo\8"')
        self.__fatal(r'"\8"')
        self.__fatal(r'"\a"')
        self.__fatal('"contains a\nlinefeed"')
        self.__fatal(r"'")
        self.__fatal(r"''")
        self.__fatal(r"'b'")

    def test_space(self):
        # Test whitespace, comments and line numbering
        self.__lexer.input('''\
        1 // comment
        2
        3\f
        4 /*\tcomment */ \t5
        6 /*long
        multiline
        comment /* */7//
        8 /**/ /*//*/ 9
        10 // no_linefeed_at_the_end''')
        self.__expect('INT_LITERAL', 1, lineno=1)
        self.__expect('INT_LITERAL', 2, lineno=2)
        self.__expect('INT_LITERAL', 3, lineno=3)
        self.__expect('INT_LITERAL', 4, lineno=4)
        self.__expect('INT_LITERAL', 5, lineno=4)
        self.__expect('INT_LITERAL', 6, lineno=5)
        self.__expect('INT_LITERAL', 7, lineno=7)
        self.__expect('INT_LITERAL', 8, lineno=8)
        self.__expect('INT_LITERAL', 9, lineno=8)
        self.__expect('INT_LITERAL', 10, lineno=9)
        self.__expect_end()
        self.__incomplete(' /* comment does not \n end * /')


if __name__ == '__main__':
    unittest.main()

