#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# test_statement.py
#
# 2007-06-06  Jori Dubrovin:  Added labeled statements
# 2006-01-31  Jori Dubrovin:  Added final local variables
# 2005-07-21  Jori Dubrovin:  Started using testhelp.py
# 2005-06-16  Jori Dubrovin:  Added dirty tests for break, return etc.
# 2005-05-02  Jori Dubrovin:  Initial version
#

"""Test for the semantics of statements."""

import unittest

import errors
from testhelp import checkValues, semanticError, runtimeError


class StatementTest(unittest.TestCase):
    """Test statements."""

    def test_simple(self):
        # Some tests for break, continue, and return statements.
        semanticError('break;')
        semanticError('continue;')
        semanticError('return;')
        semanticError('{break;}')
        semanticError('{continue;}')
        semanticError('{break a;}')
        semanticError('{continue tr_;}')
        semanticError('tr_: {continue tr_;}')
        semanticError('{return;}')
        semanticError('for (;;) {return;}')
        semanticError('lab: lab: a = a;')
        semanticError('lab: { lab: a = a; }')
        semanticError('break lab;')
        semanticError('lab: ; break lab;')
        semanticError('lab: if (true) while (true) continue lab;')
        checkValues('lab: break lab;', {})
        checkValues('''
        int x = 0;
        lab:
        break lab;
        lab:
        x = 1;
        ''', {'x': 1})

    def test_localDecl(self):
        # Test local variable declarations.
        checkValues('''
        boolean b; int i;
        ''', {'b': None, 'i': None})
        checkValues('''
        final boolean b = false;
        int i = b ? -11 : 10;
        ''', {'b': False, 'i': 10})
        checkValues('''
        int i = 300*300, j, k = i - 1;
        ''', {'i': 90000, 'j': None, 'k': 89999})
        checkValues('int i = i = 1;', {'i': 1})
        checkValues('{ int i = 1; } int i = 2;', {'i': 2})
        semanticError('boolean b; bb = false;')
        semanticError('boolean b; b = 1;')
        semanticError('int i = j, j = 2;')
        semanticError('int i = 4; { int i = 6; }')
        semanticError('{ int i = 4; } i++;')
        semanticError('abstract int i = 4;')
        semanticError('static int i = 4;')
        semanticError('native int i = 4;')
        semanticError('final static int i = 4;')
        semanticError('final final int i = 4;')
        semanticError('final int i;')
        semanticError('final int i = 4; i = 5;')
        semanticError('final int i = 4; i++;')
        runtimeError('boolean b; int i = b ? -11 : 10;',
                     errors.UninitializedError)

    def test_if(self):
        # Test if-then and if-then-else statments.
        checkValues('''
        int i = 10;
        boolean gt5;
        if (i > 5)
          gt5 = true;
        else
          gt5 = false;
        i = 0;
        ''', {'i': 0, 'gt5': True})
        checkValues('''
        int i = 3;
        boolean gt5;
        if (i > 5)
          gt5 = true;
        else
          gt5 = false;
        i = 0;
        ''', {'i': 0, 'gt5': False})
        checkValues('''
        int i = 3;
        if (i > 5) {
          i++;
          i++;
        }
        ''', {'i': 3})
        checkValues('''
        int i = 30;
        if (i > 5) {
          i++;
          i++;
        }
        ''', {'i': 32})
        semanticError('int i; if (i = 10) ;')
        semanticError('''
        if (false) {
          int j = 10;
        }
        else {
          j = 11;
        }
        ''')
        semanticError('if (10 == 10) break;')
        semanticError('if (10 == 10) continue;')
        semanticError('lb: if (10 == 10) continue lb;')

    def test_while(self):
        # Test while and do-while statements.
        checkValues('''
        int i = 4, j = 0;
        while (i++ < 7)
          j += i * i;
        j--;
        ''', {'i': 8, 'j': 5*5 + 6*6 + 7*7 - 1})
        checkValues('''
        int i = 3, j = 0;
        while (++i < 7) {
          if (i == 5) {
            break;
          }
          j += i * i;
        }
        ''', {'i': 5, 'j': 4*4})
        checkValues('''
        int i = 3, j = 0;
        while (++i < 7) {
          if (i == 5) {
            continue;
          }
          j += i * i;
        }
        ''', {'i': 7, 'j': 4*4 + 6*6})
        checkValues('''
        int i = 3, j = 0;
        l1:
        while (++i < 7)
          if (i == 5) {
            continue l1;
          }
        ''', {'i': 7, 'j': 0})
        checkValues('''
        int i = 3, j = 0;
        lab:
        while (++i < 7) {
          lab2:
          if (i == 5) {
            lab3:
            continue lab;
          }
          j += i * i;
        }
        ''', {'i': 7, 'j': 4*4 + 6*6})
        checkValues('''
        int i = 3, j = 0;
        lab:
        while (++i < 7) {
          lab2: {
            if (i == 5) {
              break lab2;
            }
            j += i * i;
          }
        }
        ''', {'i': 7, 'j': 4*4 + 6*6})
        checkValues('''
        int i = 4;
        while (i < 7) {
          int j = 0;
          j += i * i;
          i++;
        }
        ''', {'i': 7})
        checkValues('''
        int i = 4, j = 0;
        while (i > 7) {
          j++;
          i++;
        }
        ''', {'i': 4, 'j': 0})
        checkValues('''
        int i = 4, j = 0;
        do {
          j++;
          i++;
        } while (i > 7);
        ''', {'i': 5, 'j': 1})
        checkValues('''
        int i = 4, j = 1;
        do {
          j *= i;
          i++;
          ;
        } while (i < 7);
        ''', {'i': 7, 'j': 4*5*6})
        checkValues('''
        int i = 5;
        int k = 0;
        while (i > 3) {
          int j = 1;
          do {
            if (i == 5 && j == 2)
              break;
            k += i / j;
          } while (++j < i);
          i--;
        }
        ''', {'i': 3, 'k': 5/1 + 4/1 + 4/2 + 4/3})
        checkValues('''
        int i = 5;
        int k = 0;
        while (i > 3) {
          int j = 1;
          do {
            if (i == 5 && j == 2)
              continue;
            k += i / j;
          } while (++j < i);
          i--;
        }
        ''', {'i': 3, 'k': 5/1 + 5/3 + 5/4 + 4/1 + 4/2 + 4/3})
        checkValues('''
        int i = 5;
        int k = 0;
        while (i > 3) {
          int j = 1;
          la:
          do {
            if (i == 5 && j == 2)
              continue la;
            k += i / j;
          } while (++j < i);
          i--;
        }
        ''', {'i': 3, 'k': 5/1 + 5/3 + 5/4 + 4/1 + 4/2 + 4/3})
        checkValues('''
        int i = 0;
        int k = 0;
        while (++i <= 2) {
          int j = 5;
          while (++j <= 7) {
            if (i == 1 && j == 6)
              continue;
            k += i * j;
          }
        }
        ''', {'i': 3, 'k': 1*7 + 2*6 + 2*7})
        checkValues('''
        int i = 0;
        int k = 0;
        a:
        while (++i <= 2) {
          int j = 5;
          while (++j <= 7) {
            if (i == 1 && j == 6)
              continue a;
            k += i * j;
          }
        }
        ''', {'i': 3, 'k': 2*6 + 2*7})
        checkValues('''
        int i = 0;
        int k = 0;
        a:
        while (++i <= 2) {
          int j = 5;
          while (++j <= 7) {
            if (i == 1 && j == 6)
              break a;
            k += i * j;
          }
        }
        ''', {'i': 1, 'k': 0})
        semanticError('int i = 2; while (i = 4) ;')
        semanticError('int i = 2; do ; while (3);')
        semanticError('int i = 2; while (++i < 4) {int i;}')

    def test_for(self):
        # Test for statements.
        checkValues('''
        int j = 1;
        for (int i = 4; i > 0; i--)
          j *= i;
        ''', {'j': 4 * 3 * 2 * 1})
        checkValues('''
        int j = 1;
        for (int i = 4, x = 20; ; i--)
          if (i > 0)
            j *= i;
          else
            break;
        ''', {'j': 4 * 3 * 2 * 1})
        checkValues('''
        int i, j;
        lab:
        for (i = 4, j = 1; ;) {
          if (i > 0)
            j *= i;
          else
            break lab;
          i--;
        }
        ''', {'i': 0, 'j': 4 * 3 * 2 * 1})
        checkValues('''
        int j;
        for (int i = 2; i < 7; i++) {
          if (i == 2) {
            j = 10;
            continue;
          }
          j += i;
        }
        ''', {'j': 10 + 3 + 4 + 5 + 6})
        checkValues('''
        int j;
        lab:
        for (int i = 2; i < 7; i++) {
          if (i == 2) {
            j = 10;
            continue lab;
          }
          j += i;
        }
        ''', {'j': 10 + 3 + 4 + 5 + 6})
        checkValues('''
        int i, j = 0, k = 0, c = 0;
        for (i = 0; i < 5; i++)
          for (; j < i; j++) {
            if (i == 2)
              break;
            k += i*j;
            c++;
          }
        ''', {'i': 5, 'j': 4, 'k': 1*0 + 3*1 + 3*2 + 4*3, 'c': 4})
        checkValues('''
        int j = 0;
        for (int i = 4; i < 6; i++)
          j += i;
        for (int i = 4; i < 6; i++)
          j += i;
        ''', {'j': 4 + 5 + 4 + 5})
        semanticError('int i, j; for (int i = 2; i < 5; i++) ;')
        semanticError('for (int i = 2; i < 5; i++) ; i = 0;')

    def test_switch(self):
        # Test switch statements.
        checkValues('''
        int x;
        switch (3)
        {
        case 2:
        }
        ''', {'x': None})
        checkValues('''
        int x;
        switch (3)
        {
        case 3:
          x = 100;
        }
        ''', {'x': 100})
        checkValues('''
        int x;
        switch (1)
        {
        case 1:
          int y = 120;
          // Fall through
        case 4:
          x = y;
        }
        ''', {'x': 120})
        checkValues('''
        int x;
        switch (1)
        {
        case 1:
          int y = 120;
          break;
        case 4:
          x = y;
        }
        ''', {'x': None})
        checkValues('''
        int x;
        lab:
        switch (1)
        {
        case 1:
          int y = 120;
          break lab;
        case 4:
          x = y;
        }
        ''', {'x': None})
        checkValues('''
        int x = 1;
        switch (x)
        {
        default:
          x = 10;
          break;
        case 2-1:
          x = 20;
          break;
        case -1*-1:
          x = 30;
          break;
        }
        ''', {'x': 20})
        checkValues('''
        int x = 5;
        switch (x)
        {
        default:
          x = 10;
          break;
        case 2-1:
          x = 20;
          break;
        case -1*-1:
          x = 30;
          break;
        }
        ''', {'x': 10})
        checkValues('''
        int x = 0;
        for (int i = 1; i <= 3; i++) {
          switch (i)
          {
          case 2:
            break;
          default:
            x += i;
            break;
          }
        }
        ''', {'x': 1 + 3})
        checkValues('''
        int x = 0;
        for (int i = 1; i <= 3; i++) {
          switch (i)
          {
          case 2:
            continue;
          default:
            x += i;
            break;
          }
        }
        ''', {'x': 1 + 3})
        checkValues('''
        int x = 0;
        a:
        for (int i = 1; i <= 3; i++) {
          switch (i)
          {
          case 2:
            break a;
          default:
            x += i;
            break;
          }
        }
        ''', {'x': 1})
        checkValues('''
        int x = 4;
        switch (x)
        {
        case 4:
          for (int i = 1; i <= 3; i++) {
            if (i == 2)
              break;
            x += i;
          }
          x += 100;
        }
        ''', {'x': 4 + 1 + 100})
        semanticError('switch (false) {}')
        semanticError('switch (4) {case 4: continue;}')
        semanticError('switch (4) {default: default: }')
        semanticError('switch (4) {case false: }')
        runtimeError('''
        int x;
        switch (4)
        {
        case 1:
          int y = 120;
        case 4:
          x = y;
        }
        ''', errors.UninitializedError)


if __name__ == '__main__':
    unittest.main()
