#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# test_typedecl.py
#
# 2006-01-31  Jori Dubrovin:  Final methods and parameters
# 2005-11-09  Jori Dubrovin:  Added super field access and super method calls
# 2005-07-22  Jori Dubrovin:  Initial version
#

"""Test for type declarations and names."""

import unittest

import errors
from testhelp import checkValues, semanticError, runtimeError


class TypeDeclTest(unittest.TestCase):
    """Test type declarations and names."""

    def test_duplicateNames(self):
        # Test overloading of a name in general.
        semanticError('class C {} class C {}')
        checkValues('class C { C c; }', {})
        checkValues('C c; class C { }', {})
        checkValues('class D { C c; } class C { }', {})
        semanticError('class C { int x; C x; }')
        checkValues('class C { int m(int x) {} int m(boolean x) {} }', {})
        semanticError('class C { int m(); C m(); }')
        semanticError('class C { int m(int x); C m(int x); }')
        semanticError('class C { int m(int x); static int m(int x); }')
        semanticError('class C { int m(int x); static void m(int x); }')
        semanticError('class C { class C {} }')
        semanticError('class D { class C { class C {} } }')
        semanticError('class C { class D { class C {} } }')
        checkValues('''
        int C = 12;
        class C
        {
          int C;

          int C()
          {
            int d = C;
            C++;
            C C = new C();
            C = this;
            if (C.C < 10)
              C();
            return d;
          }
        }
        C obj = new C();
        obj.C = 8;
        int d = obj.C();
        ''', {'C': 12, 'd': 8, 'obj.C': 10})
        checkValues('''
        class C
        {
          InC inC;
          InC.InInC inInC;
          int meth1()
          {
            inC = new InC();
            return InC.InInC.m() + inC.m() + inC.x;
          }
          int meth2() { d = new D(); return d.m() + d.x; }
          class InC
          {
            int x;
            int m() { x = 10; return 4; }
            class InInC
            {
              static int m() { return In2.m(); }
            }
          }
          class In2
          {
            static int m() { return -2; }
          }
          D d;
        }
        class D
        {
          int x;
          int m() { x = 5; return 1; }
        }
        int x1 = new C().meth1();
        int x2 = new C().meth2();
        ''', {'x1': 12, 'x2': 6})
        semanticError('class C { class D {} } C.E x;')

    def test_nestedNames(self):
        # Test accessing names from nested classes.
        checkValues('''
        class D { int m() { return 1; } }
        class C1
        {
          class D { int m() { return 2; } }
          class C2
          {
            class C3
            {
              D d;
              int m() { d = new D(); return d.m() + d.sm(); }
              class D { int m() { return 4; } static int sm() { return 14; } }
            }
            class D { int m() { return 3; } }
          }
        }
        C1.C2.C3 c = new C1.C2.C3();
        int y = c.m();
        ''', {'y': 18})
        checkValues('''
        class D { int m() { return 1; } }
        class C1
        {
          class D { int m() { return 2; } }
          class C2
          {
            class C3
            {
              D d;
              int m() { d = new D(); return d.m() + d.sm(); }
            }
            class D { int m() { return 4; } static int sm() { return 14; } }
          }
        }
        C1.C2.C3 c = new C1.C2.C3();
        int y = c.m();
        ''', {'y': 18})
        checkValues('''
        class C1
        {
          int m() { return 2; }
          class C2
          {
            class C3
            {
              int m() { return 17; }
              int f() { return m(); }
              static int sf() { return C2.m(); }
              int i() { C1 c1 = new C1(); return c1.m(); }
            }
            int f() { return m(); }
            static int sf() { return m(); }
            static int m() { return -14; }
          }
          int f() { return m(); }
        }
        C1 c1 = new C1();
        C1.C2 c2 = new C1.C2();
        C1.C2.C3 c3 = new C1.C2.C3();
        ''', {'c1.f()': 2, 'c2.f()': -14, 'c2.sf()': -14,
              'c3.f()': 17, 'c3.sf()': -14, 'c3.i()': 2})
        checkValues('''
        class C1
        {
          int m;
          class C2
          {
            class C3
            {
              int m;
              int f() { return m; }
            }
            int f() { return m; }
            int m;
          }
          int f() { return m; }
        }
        C1 c1 = new C1();
        c1.m = 1;
        C1.C2 c2 = new C1.C2();
        c2.m = 2;
        C1.C2.C3 c3 = new C1.C2.C3();
        c3.m = 3;
        ''', {'c1.f()': 1, 'c2.f()': 2, 'c3.f()': 3})
        semanticError('''
        class C1
        {
          static int m() { return 15; }
          class C2
          {
            static int m(int x) { return m(); }
          }
        }
        ''')

    def test_instanceContext(self):
        # Test accessing instance variables.
        semanticError('''
        class C1
        {
          int x;
          class C2
          {
            int m() { return x; }
          }
        }
        ''')
        semanticError('class C { int x; static int m() { return x; }}')
        semanticError('int x; class C { static int m() { return x; }}')
        semanticError('class C { int m() { return x; }}')
        semanticError('class C { int x; class D { int m() { return x; }}}')
        semanticError('class C { int m() { return this.x; }}')
        semanticError('Object x = this;')
        semanticError('''
        class B { int x; }
        class C extends B { static int m() { return super.x; }}
        ''')
        checkValues('''
        class C1
        {
          int x;
          class C2
          {
            int x;
            int m1() { return x; }
            int m2() { int x = 20; return x; }
            int m3() { int x = 20; return this.x; }
          }
          class x {}
        }
        C1.C2 c = new C1.C2();
        c.x = 170;
        int y1 = c.m1();
        int y2 = c.m2();
        int y3 = c.m3();
        ''', {'y1': 170, 'y2': 20, 'y3': 170})

    def test_methodDecl(self):
        # Test method declarations.
        semanticError('class C { void m(); }')
        semanticError('class C { static void m(); }')
        semanticError('class C { static static void m() {} }')
        semanticError('interface I { final void m(); }')
        semanticError('interface I { final void m() {} }')
        semanticError('''
        class C { static final void m() {} }
        class D extends C { static void m() {} }
        ''')
        semanticError('''
        class C { final void m() {} }
        class D extends C { void m() {} }
        ''')
        semanticError('''class C { void m(abstract int t) {} }''')
        semanticError('''class C { void m(static int t) {} }''')
        semanticError('''class C { void m(native int t) {} }''')
        semanticError('''class C { void m(final final int t) {} }''')
        semanticError('''class C { void m(final int t) { t++; } }''')
        checkValues('''
        class C { static int m(int x, final int y) { x += y; return x; } }
        int t = C.m(30, 40);
        ''', {'t': 70})

    def test_fieldDecl(self):
        # Test field declarations.
        semanticError('class C { native int m; }')
        semanticError('class C { static static int m; }')

    def test_constructorDecl(self):
        # Test constructor declarations.
        semanticError('class C { D() {} }')
        semanticError('class C { C(int x, int x) {} }')
        semanticError('class C { C(abstract int x) {} }')
        semanticError('class C { C(static int x) {} }')
        semanticError('class C { C(native int x) {} }')
        semanticError('class C { C(final final int x) {} }')
        semanticError('class C { C(final int x) { x = 3; } }')
        checkValues('''
        class C { C(int x, final int y) { x += y; z = x; } int z; }
        int t = new C(30, 40).z;
        ''', {'t': 70})

    def test_instances(self):
        # Test properties of object instances.
        checkValues('''
        class C { int x; }
        C c, d;
        c = new C();
        c.x = 1;
        d = new C();
        d.x = 2;
        ''', {'c.x': 1, 'd.x': 2})

if __name__ == '__main__':
    unittest.main()
