#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# testhelp.py
#
# 2007-06-06  Jori Dubrovin:  Added more info in an exception
# 2005-08-30  Jori Dubrovin:  Removed the fail function
# 2005-08-08  Jori Dubrovin:  Using updated basetypes.py
# 2005-07-19  Jori Dubrovin:  Initial version
#

"""Help functions for testing the interpreter."""

import grammar
import basetypes
import errors


# A flag that tells whether the exceptions catched during dirty testing
# should be printed to screen.
printErrors = False


def checkValues(input_, expectedValuation):
    """Evaluate input and check that the given values are produced.

    The input string is parsed, type checked, translated and evaluated.
    All expression-expected value pairs are then checked.

    Param input_: a string containing the input program.

    Param expectedValuation: a mapping from (side-effect free) action
    language expressions to expected values.  The expected value can
    be None (representing an uninitialized value), 'nonnull' (a
    reference value that is not null), 'null', False, True (for
    boolean values), or an integer (for int values).
    """
    env = basetypes.getInitialEnvironment()
    vm = basetypes.getInitialMachine()
    tree = grammar.parse('''
    class ValueChecker {
      native static void checkInt(int x);
    }
    ''')
    expectedInt = []
    def checkInt(state, x):
        if x != expectedInt[0]:
            raise AssertionError('got ' + str(x) +
                                 ', expected ' + str(expectedInt[0]))
    env.setNativeHooks({'ValueChecker.checkInt#int': checkInt})
    vm.initialize(tree.translate(env))
    vm.run()
    tree = grammar.parse(input_)
    vm.initialize(tree.translate(env))
    vm.run()
    for expr in expectedValuation:
        expected = expectedValuation[expr]
        if expected is None:
            # expr must be an uninitialized variable.
            expectedInt = [2]  # checkInt always fails if reached.
            tree = grammar.parse('ValueChecker.checkInt((' +
                                 expr + ') == (' + expr + ') ? 1 : 0);')
            vm.initialize(tree.translate(env))
            try:
                vm.run()
            except errors.UninitializedError, e:
                pass
        else:
            # Must use 'is' for comparison with False and True because
            # False == 0 and True == 1.  False and True are instances
            # of (a subtype of) type int in Python.
            if expected == 'nonnull':
                expectedInt = [0]
                tree = grammar.parse('ValueChecker.checkInt((' + expr +
                                     ') == null ? 1 : 0);')
            elif expected == 'null':
                expectedInt = [1]
                tree = grammar.parse('ValueChecker.checkInt((' + expr +
                                     ') == null ? 1 : 0);')
            elif expected is False:
                expectedInt = [0]
                tree = grammar.parse('ValueChecker.checkInt((' + expr +
                                     ') ? 1 : 0);')
            elif expected is True:
                expectedInt = [1]
                tree = grammar.parse('ValueChecker.checkInt((' + expr +
                                     ') ? 1 : 0);')
            elif isinstance(expected, int):
                expectedInt = [expected]
                tree = grammar.parse('ValueChecker.checkInt(' + expr + ');')
            else:
                raise TypeError()
            vm.initialize(tree.translate(env))
            vm.run()

def semanticError(input_):
    """Check that the input does not pass type checking.

    The input string must be grammatically correct, though.

    Param input_: a string containing the input program.
    """
    tree = grammar.parse(input_)
    try:
        tree.translate(basetypes.getInitialEnvironment())
        raise AssertionError
    except errors.FatalStaticError, e:
        if printErrors:
            print
            print 'Program:'
            print input_
            print 'Semantic error:', e
            print

def runtimeError(input_, errorClass):
    """Check that the input raises an error upon evaluation.

    The input string must be grammatically correct and pass type
    checks, though.  The raised error must be an instance of the
    given class.

    Param input_: a string containing the input program.

    Param errorClass: the class whose instance the raised exception
    must be.
    """
    tree = grammar.parse(input_)
    vm = basetypes.getInitialMachine()
    vm.initialize(tree.translate(basetypes.getInitialEnvironment()))
    try:
        vm.run()
        raise AssertionError
    except errorClass, e:
        if printErrors:
            print
            print 'Program:'
            print input_
            print 'Runtime error:', e
            print
