#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

# typedecl.py
#
# 2007-09-03  Jori Dubrovin:  Fixing an error message
# 2007-06-10  Jori Dubrovin:  Adding recurseStr API
# 2006-11-29  Jori Dubrovin:  String parameters in native methods
# 2006-01-31  Jori Dubrovin:  Final methods, final parameters
# 2005-11-11  Jori Dubrovin:  Reflected changes in mytypes.py
# 2005-09-01  Jori Dubrovin:  Fixed the env for super constructor arguments
# 2005-08-08  Jori Dubrovin:  Fixing bugs
# 2005-06-29  Jori Dubrovin:  Initial version
#

"""Source representation for type declarations in the action language."""

from syntaxtree import SyntaxNode, TopStatement
from mytypes import SingleReferenceType, Function, Field
from mytypes import voidType, intType, boolType
from errors import ErrorProxy, StaticError, ReturnTypeError
from expr import StringLiteralExpr, determineFunction
from icode import SourceRef, StoreAuto, Return, MissingReturn, Literal
from icode import LoadAuto, StoreField, StoreStatic, CallNative, Pop
from icode import CallStatic, NewInstance, Duplicate, InitArray, LoadStatic
from icode import StoreArray, Label, RefEqual, IfGoto, Goto, Instance
from icode import buildRoutine, void, null, boolValue
from icode import Instance


class TopTypeDeclaration(TopStatement):
    """A top-level class or interface declaration."""

    def __init__(self, typeDecl):
        """Construct a top-level type declaration.

        Param typeDecl: a TypeDeclaration instance that represents the
        class or interface declaration.
        """
        TopStatement.__init__(self, typeDecl.linespan())
        self.__typeDecl = typeDecl

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return strFunc(self.__typeDecl, *args, **kwargs) + '\n'

    def api_getTypeDeclaration(self):
        """Return the TypeDeclaration."""
        return self.__typeDecl

    def translatePass(self, pass_, env):
        if pass_ == 1:
            self.__typeDecl.translateDeclaredTypes(env, None)
            return []
        if pass_ == 2:
            self.__typeDecl.translateInheritedTypes(env, None)
            return []
        if pass_ == 3:
            # Return static variable creation code, ignore instance
            # variable creation code (which is necessarily empty since
            # __typeDecl is not a field declaration).
            return self.__typeDecl.translateMembers(env, None)[0]
        if pass_ == 4:
            # Return static variable initialization code, ignore instance
            # variable initialization code (which is necessarily empty
            # since __typeDecl is not a field declaration).
            return self.__typeDecl.translateBodies(env, None)[0]
        if pass_ == 5:
            # Pass 5 is only for top-level block statements.
            return []
        assert False


class TypeBodyDeclaration(SyntaxNode):
    """A base class for declarations that appear in a class/interface body."""

    def __init__(self, linespan):
        SyntaxNode.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):  #pragma: no cover
        return '<declaration>'

    def translateDeclaredTypes(self, env, enclosingType):
        """Generate the types declared in this declaration.

        If self is a type declaration, add the declared type as a
        member to the enclosing type (or to the environment as a
        top-level type if self is a top-level type declaration).  Do
        the same recursively for nested type declarations.

        If self is not a type declaration, do nothing.

        Raise a StaticError if static checking fails.

        Param env: the StaticEnvironment of the enclosing type, or the
        top-level environment if self is a top-level type declaration.
        The method may mutate env is self is a top-level type
        declaration.

        Param enclosingType: the SingleReferenceType whose declaration
        directly contains self.  The type must be incomplete in that
        it only contains the member types declared before this one,
        and no other kinds of members.  If self is a top-level type
        declaration, enclosingType is None.
        """
        # Default implementation does nothing.
        pass

    def translateInheritedTypes(self, env, enclosingType):
        """Add inherited member types to the type declared in self.

        If self is a type declaration, resolve the supertypes and add
        all inherited type members to the type.  Do the same
        recursively for nested type declarations.

        If self is not a type declaration, do nothing.

        Raise a StaticError if static checking fails.

        Param env: the current StaticEnvironment.

        Param enclosingType: the SingleReferenceType whose declaration
        directly contains self.  The type must be incomplete in that
        it contains all declared and inherited member types, but no
        other kinds of members.  If self is a top-level type
        declaration, enclosingType is None.
        """
        # Default implementation does nothing.
        pass

    def translateMembers(self, env, enclosingType):  #pragma: no cover
        """Add members to their enclosing types.

        If self is a field, method, or constructor declaration, add
        the declared entity to the directly enclosing reference type.

        If self is a type declaration, apply translateMembers
        recursively.  Do not add members to the enclosing type,
        because translateDeclaredTypes has already done so.

        Raise a StaticError if static checking fails.

        Param env: the StaticEnvironment of the enclosing type, or
        the top-level environment if self is a top-level type
        declaration.  The method must not mutate env.

        Param enclosingType: the SingleReferenceType whose declaration
        directly contains self.  The type must be incomplete in that
        it must contain all declared and inherited member types and
        possibly some declared other members, but not inherited
        members other than types.  If self is a top-level type
        declaration, enclosingType is None.

        Return value: a tuple (staticVarCreationCode,
        instanceVarCreationCode).  Here staticVarCreationCode is a
        list of code instructions that set all static variables that
        are declared within this declaration (including nested
        declarations, not including variables declared in supertypes)
        to their default values.  If this is an enum constant
        declaration, the instructions create the enum instance instead
        of storing the default value (which is null).  If this is an
        instance variable declaration, instanceVarCreationCode is a
        list of code instructions that set the variable to its default
        value.  Otherwise instanceVarCreationCode is an empty list.
        """
        raise NotImplementedError('Pure virtual method')

    def translateBodies(self, env, enclosingType):  #pragma: no cover
        """Generate internal code for methods, constructors and initializers.

        If self is a method/constructor declaration, generate code for
        the method/constructor body and attach it to the
        method/constructor of the enclosing type.

        If self is a field declaration, return code that
        initializes the variable.

        If self is a type declaration, apply translateBodies
        recursively and concatenate the staticVarInitCode part of the
        return values (instructions that initialize static variables).

        Raise a StaticError if static checking fails.

        Param env: the StaticEnvironment of the enclosing type, or
        the top-level environment if self is a top-level type
        declaration.  The method must not mutate env.

        Param enclosingType: the (complete) SingleReferenceType whose
        declaration directly contains self.  If self is a top-level
        type declaration, enclosingType is None.

        Return value: a tuple (staticVarInitCode,
        instanceVarInitCode).  Here staticVarInitCode is a list of
        code instructions that evaluate and store the values of the
        initialization expressions of all static variables that are
        declared within this declaration (including nested
        declarations, not including inherited variable) and have an
        initializer.  If this is an instance variable declaration that
        has an initializer, instanceVarCreationCode is a list of code
        instructions that evaluate the initialization expression of
        the variable and store the result.  Otherwise
        instanceVarCreationCode is an empty list.
        """
        raise NotImplementedError('Pure virtual method')


class TypeDeclaration(TypeBodyDeclaration):
    """A base class for class, interface, and enum declarations.

    These declarations may appear at top level as well as in class or
    interface bodies.
    """

    def __init__(self, name, bodyDecls, linespan):
        """Construct a type declaration.

        Param name: string containing the simple name of the type.

        Param bodyDecls: list of TypeBodyDeclarations that constitute
        the body.

        Param linespan: see SyntaxNode.
        """
        TypeBodyDeclaration.__init__(self, linespan)
        self._name = name
        self._bodyDecls = bodyDecls

    def api_getName(self):
        """Return the name of the declared type as a string."""
        return self._name

    def api_getBodyDeclarations(self):
        """Return the body as a list of BodyDeclarations."""
        return self._bodyDecls[:]

    def translateDeclaredTypes(self, env, enclosingType):
        ep = ErrorProxy()
        for c in env.getEnclosingTypes():
            if c.getSimpleName() == self._name:
                ep.finish("nested redeclaration of type '" + self._name +
                          "'", self.linespan())
        if enclosingType is None:
            # This is a top-level type declaration.
            if env.isTopType(self._name):
                ep.finish("redeclaration of top-level type '" + self._name +
                          "'", self.linespan())
            type_ = self._createType('')
            env.addTopType(type_)
        else:
            # This is a nested type declaration.
            if enclosingType.getMemberTypes(self._name):
                ep.finish("redeclaration of member type '" + self._name +
                          "'", self.linespan())
            type_ = self._createType(enclosingType.getCanonicalName())
            enclosingType.addMemberType(type_)
        newEnv = env.makeNestedCopy(type_)
        for d in self._bodyDecls:
            try:
                d.translateDeclaredTypes(newEnv, type_)
            except StaticError, e:
                ep.addException(e)
        ep.raiseIfErrors()

    def translateInheritedTypes(self, env, enclosingType):
        type_ = self.__getMyType(env, enclosingType)
        newEnv = env.makeNestedCopy(type_)
        supertypes = self._resolveSupertypes(newEnv, type_)
        type_.setDirectSupertypes(supertypes)
        inheritedTypes = []
        for stype in supertypes:
            assert stype.isReferable()
            for t in stype.getAllMemberTypes():
                if not type_.getMemberTypes(t.getSimpleName()):
                    if not t in inheritedTypes:
                        inheritedTypes.append(t)
        for t in inheritedTypes:
            type_.addMemberType(t)
        ep = ErrorProxy()
        for d in self._bodyDecls:
            try:
                d.translateInheritedTypes(newEnv, type_)
            except StaticError, e:
                ep.addException(e)
        ep.raiseIfErrors()
        type_.setReferable()

    def translateMembers(self, env, enclosingType):
        type_ = self.__getMyType(env, enclosingType)
        newEnv = env.makeNestedCopy(type_)
        ep = ErrorProxy()
        staticVarCreationCode = []
        instanceVarCreationCode = [SourceRef(self), StoreAuto('this')]
        for d in self._bodyDecls:
            try:
                sc, ic = d.translateMembers(newEnv, type_)
                staticVarCreationCode += sc
                instanceVarCreationCode += ic
            except StaticError, e:
                ep.addException(e)
        ep.raiseIfErrors()
        instanceVarCreationCode += [Literal(void), Return()]
        type_.setRoutine('#createvars', buildRoutine(instanceVarCreationCode))
        self._addImplicitMembers(env, type_)
        # Add a default constructor if no constructors declared.
        if not type_.isInterface() and not type_.getFunctions('#constructor'):
            constructor = Function('#constructor', [], voidType)
            type_.addFunction(constructor)
            code = [SourceRef(self),
                    StoreAuto('this'),
                    LoadAuto('this'),
                    CallStatic(type_, '#createvars', 1),
                    Pop()]
            if type_ is not env.getTopType('Object'):
                superclass = type_.getDirectSupertypes()[0]
                superCtors = superclass.getExactFunctions('#constructor', [])
                if not superCtors:
                    ep.finish('superclass constructor must be ' +
                              'called explicitly',
                              self.linespan())
                code += [LoadAuto('this'),
                         CallStatic(superclass,
                                    superCtors[0].getSignature(), 1),
                         Pop()]
            code += [LoadAuto('this'),
                     CallStatic(type_, '#initvars', 1),
                     Pop(),
                     Literal(void),
                     Return()]
            constructor.setRoutine(buildRoutine(code))
        # Inherit fields.
        inheritedFields = []
        for stype in type_.getDirectSupertypes():
            for v in stype.getAllFields():
                if not type_.getFields(v.getSimpleName()):
                    if not v in inheritedFields:
                        inheritedFields.append(v)
        for v in inheritedFields:
            type_.addField(v)
        # Inherit methods.
        potentialMethods = {}
        for stype in type_.getDirectSupertypes():
            for m in stype.getAllFunctions():
                if m.getSimpleName() != '#constructor':
                    paramTypes = [t for (n, t) in m.getParameters()]
                    if not type_.getExactFunctions(m.getSimpleName(),
                                                   paramTypes):
                        # A method with the same signature is not
                        # declared here.
                        sig = m.getSignature()
                        if sig in potentialMethods:
                            if m not in potentialMethods[sig]:
                                potentialMethods[sig].append(m)
                        else:
                            potentialMethods[sig] = [m]
        for sig in potentialMethods:
            methods = potentialMethods[sig]
            # Now methods is a list of the methods in direct
            # supertypes that have the same signature, such that no
            # method with that signature is declared in this type.  If
            # there is a static or non-abstact method in the list, it
            # is the first one because the list of supertypes begins
            # with the superclass followed by interfaces.
            if len(methods) == 1:
                type_.addFunction(methods[0])
            elif methods[0].isStatic():
                ep.add("inherited method '" + str(methods[0]) +
                       "' cannot hide a method in a superinterface",
                       self.linespan())
            else:
                if not methods[0].isAbstract():
                    candidates = [methods[0]]
                else:
                    candidates = methods[:]
                for c in candidates:
                    cType = c.getReturnType()
                    for m in methods:
                        mType = m.getReturnType()
                        if not cType.isReturnSubstitutableFor(mType):
                            break
                    else:
                        # Method c is return-type substitutable for
                        # all other potential methods with the same
                        # signature.  Inherit c.
                        type_.addFunction(c)
                        break
                else:
                    # No candidate method is return-type substitutable
                    # for all others.
                    ep.add("incompatible return type for method '" +
                           str(methods[0]) + "'", self.linespan())
        ep.raiseIfErrors()
        return (staticVarCreationCode, [])

    def translateBodies(self, env, enclosingType):
        type_ = self.__getMyType(env, enclosingType)
        newEnv = env.makeNestedCopy(type_)
        ep = ErrorProxy()
        staticVarInitCode = []
        instanceVarInitCode = [SourceRef(self), StoreAuto('this')]
        for d in self._bodyDecls:
            try:
                sc, ic = d.translateBodies(newEnv, type_)
                staticVarInitCode += sc
                instanceVarInitCode += ic
            except StaticError, e:
                ep.addException(e)
        ep.raiseIfErrors()
        instanceVarInitCode += [Literal(void), Return()]
        type_.setRoutine('#initvars', buildRoutine(instanceVarInitCode))
        abstractMethods = []
        for f in type_.getAllFunctions():
            if f.isAbstract():
                abstractMethods.append(f)
            else:
                type_.setRoutine(f.getSignature(), f.getRoutine())
        if abstractMethods and not type_.isAbstract():
            ep.finish("non-abstract class has a method '" +
                      str(abstractMethods[0]) + "'", self.linespan())
        return (staticVarInitCode, [])

    def __getMyType(self, env, enclosingType):
        """Return the SingleReferenceType declared here.

        Assume the type already exists in the enclosing type or top level.
        """
        if enclosingType is None:
            return env.getTopType(self._name)
        else:
            return enclosingType.getMemberTypes(self._name)[0]

    def _createType(self, enclosingName):  #pragma: no cover
        """Return a new SingleReferenceType.

        Raise a StaticError if the modifiers in the declaration are
        illegal.

        The returned type must be a class, enum, or interface,
        depending on the kind of the declaration.

        Param enclosingName: a string containing the canonical name
        for the enclosing type.  The string is empty if this is a top
        type.
        """
        raise NotImplementedError('Pure virtual method')

    def _resolveSupertypes(self, env, myType):  #pragma: no cover
        """Return the list of direct supertypes.

        Raise a StaticError if the list of supertypes contains illegal
        references or refers to the same type twice.

        Param env: a StaticEnvironment for the type.

        Param myType: this Type.
        """
        raise NotImplementedError('Pure virtual method')

    def _addImplicitMembers(self, env, myType):
        """Add implcit members to this type.

        The only implicit member is the method 'values' in enum types.

        Param env: a StaticEnvironment for the type.

        Param myType: this Type.
        """
        # Default implementation does nothing.
        pass


class ClassDeclaration(TypeDeclaration):
    """A class declaration."""

    def __init__(self, modifiers, name, superNames, interfaceNames,
                 bodyDecls, linespan):
        """Construct a class declaration.

        Param modifiers: a list of strings containing the given
        modifiers.

        Param name: string containing the simple name of the class.

        Param superNames: a list of one Name if a superclass is
        declared using the keyword extends.  Otherwise an empty list.

        Param interfaceNames: a list of Names representing the
        implemented interfaces.

        Param bodyDecls: list of TypeBodyDeclarations that constitute
        the class body.

        Param linespan: see SyntaxNode.
        """
        TypeDeclaration.__init__(self, name, bodyDecls, linespan)
        self.__modifiers = modifiers
        self.__superNames = superNames
        self.__interfaceNames = interfaceNames

    def api_recurseStr(self, strFunc, *args, **kwargs):
        s = ' '.join(self.__modifiers + ['class', self._name])
        if self.__superNames:
            s += ' extends ' + strFunc(self.__superNames[0], *args, **kwargs)
        if self.__interfaceNames:
            s += ' implements ' + ', '.join([strFunc(n, *args, **kwargs) for
                                             n in self.__interfaceNames])
        s += '\n{\n'
        s += '\n\n'.join([self.indent(strFunc(d, *args, **kwargs)) for
                          d in self._bodyDecls])
        s += '\n}'
        return s

    def api_getModifiers(self):
        """Return the modifiers as a list of strings."""
        return self.__modifiers[:]

    def api_getSuperclassName(self):
        """Return the superclass as a Name.

        If the declaration mentions no superclass, None is returned.
        """
        if self.__superNames:
            return self.__superNames[0]
        else:
            return None

    def api_getInterfaceNames(self):
        """Return the implemented interfaces as a list of Names."""
        return self.__interfaceNames[:]

    def _createType(self, enclosingName):
        ep = ErrorProxy()
        dup = _reportDuplicateModifiers(self.__modifiers)
        if dup:
            ep.add(dup, self.linespan())
        if 'abstract' in self.__modifiers:
            if 'final' in self.__modifiers:
                ep.add('abstract class cannot be final', self.linespan())
            kind = 'abstract class'
        else:
            if 'final' in self.__modifiers:
                kind = 'final class'
            else:
                kind = 'concrete class'
        if 'native' in self.__modifiers:
            ep.add('classes cannot be native', self.linespan())
        if 'static' in self.__modifiers:
            ep.add('all classes are implicitly static', self.linespan())
        ep.raiseIfErrors()
        return SingleReferenceType(self._name, enclosingName, kind)

    def _resolveSupertypes(self, env, myType):
        ep = ErrorProxy()
        supertypes = []
        if self.__superNames:
            superclass = self.__superNames[0].resolveAsType(env)
            if superclass.isInterface():
                ep.finish('cannot extend ' + str(superclass),
                          self.__superNames[0].linespan())
            if superclass.isSubtypeOf(env.getTopType('Enum')):
                ep.finish('cannot extend ' + str(superclass),
                          self.__superNames[0].linespan())
            if superclass.isFinal():
                ep.finish('cannot extend ' + str(superclass),
                          self.__superNames[0].linespan())
            supertypes.append(superclass)
        else:
            objectClass = env.getTopType('Object')
            if myType is not objectClass:
                supertypes.append(objectClass)
        for n in self.__interfaceNames:
            try:
                interface = n.resolveAsType(env)
                if not interface.isInterface():
                    ep.add(str(interface) + ' is not an interface',
                           n.linespan())
                elif interface in supertypes:
                    ep.add(str(interface) + ' implemented twice',
                           n.linespan())
                else:
                    supertypes.append(interface)
            except StaticError, e:
                ep.addException(e)
        ep.raiseIfErrors()
        return supertypes


class InterfaceDeclaration(TypeDeclaration):
    """An interface declaration."""

    def __init__(self, name, superNames, bodyDecls, linespan):
        """Construct a class declaration.

        Param name: string containing the simple name of the class.

        Param superNames: a list of Names representing the
        superinterfaces.

        Param bodyDecls: list of TypeBodyDeclarations that constitute
        the class body.  ConstructorDeclarations are not allowed in
        the list.

        Param linespan: see SyntaxNode.
        """
        TypeDeclaration.__init__(self, name, bodyDecls, linespan)
        self.__superNames = superNames

    def api_recurseStr(self, strFunc, *args, **kwargs):
        s = 'interface ' + self._name
        if self.__superNames:
            s += ' extends ' + ', '.join([strFunc(n, *args, **kwargs) for
                                          n in self.__superNames])
        s += '\n{\n'
        s += '\n\n'.join([self.indent(strFunc(d, *args, **kwargs)) for
                          d in self._bodyDecls])
        s += '\n}'
        return s

    def api_getSuperinterfaceNames(self):
        """Return the extended interfaces as a list of Names."""
        return self.__superNames[:]

    def _createType(self, enclosingName):
        return SingleReferenceType(self._name, enclosingName, 'interface')

    def _resolveSupertypes(self, env, myType):
        ep = ErrorProxy()
        supertypes = []
        if self.__superNames:
            for n in self.__superNames:
                try:
                    interface = n.resolveAsType(env)
                    if not interface.isInterface():
                        ep.add(str(interface) + ' is not an interface',
                               n.linespan())
                    elif interface in supertypes:
                        ep.add(str(interface) + ' extended twice',
                               n.linespan())
                    else:
                        supertypes.append(interface)
                except StaticError, e:
                    ep.addException(e)
        else:
            objectInterface = env.getTopType('ObjectInterface')
            if myType is objectInterface:
                supertypes.append(env.getTopType('Object'))
            else:
                supertypes.append(objectInterface)
        ep.raiseIfErrors()
        return supertypes


class EnumDeclaration(TypeDeclaration):
    """An enum declaration."""

    def __init__(self, name, bodyDecls, linespan):
        """Construct an enum declaration.

        Param name: string containing the simple name of the enum.

        Param bodyDecls: list of EnumConstantDeclarationos that
        constitute the body of the declaration.

        Param linespan: see SyntaxNode.
        """
        TypeDeclaration.__init__(self, name, bodyDecls, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):
        s = [strFunc(d, *args, **kwargs) for d in self._bodyDecls]
        return ('enum ' + self._name + '\n{\n' +
                self.indent(', '.join(s)) + '\n}')

    def _createType(self, enclosingName):
        return SingleReferenceType(self._name, enclosingName, 'enum')

    def _resolveSupertypes(self, env, myType):
        return [env.getTopType('Enum')]

    def _addImplicitMembers(self, env, myType):
        self.__addValuesMethod(env, myType)
        self.__addToStringMethod(env, myType)

    def __addValuesMethod(self, env, myType):
        # Create a static function values that takes no parameters and
        # returns an array that contains all the enum constants.
        arrayType = myType.getArrayType(env)
        method = Function('values', [], arrayType, isStatic=True)
        myType.addFunction(method)
        constants = myType.getAllFields()  # Get all enum constants.
        code = [SourceRef(self),
                NewInstance(arrayType),
                Duplicate(),
                Literal(len(constants)),
                Duplicate(2),
                Duplicate(2),
                StoreField(arrayType.getFields('length')[0].getSignature()),
                Literal(myType.getDefaultValue()),
                InitArray()]
        index = 0
        for c in constants:
            assert c.getType() is myType
            code += [Duplicate(),
                     Literal(index),
                     LoadStatic(c.getSignature()),
                     StoreArray()]
            index += 1
        code += [Return()]
        method.setRoutine(buildRoutine(code))

    def __addToStringMethod(self, env, myType):
        # Create an instance method toString that returns the declared
        # name of the enum constant as a string.
        stringType = env.getTopType('String')
        method = Function('toString', [], stringType)
        myType.addFunction(method)
        constants = myType.getAllFields()  # Get all enum constants.
        code = [SourceRef(self),
                StoreAuto('this')]
        endLabel = Label()
        for c in constants:
            # Compare 'this' to all enum constants in turn and return
            # a string if they are equal.  Use expr.StringLiteralExpr
            # to create a string for every enum constant.
            assert c.getType() is myType
            constantName = c.getSimpleName()
            stringResult = StringLiteralExpr(constantName, self).translate(env)
            equalLabel = Label()
            notEqualLabel = Label()
            code += ([LoadAuto('this'),
                      LoadStatic(c.getSignature()),
                      RefEqual(),
                      IfGoto(equalLabel),
                      Goto(notEqualLabel),
                      equalLabel] +
                     stringResult.getEvalCode() +
                     [Goto(endLabel),
                      notEqualLabel])
        # Return null if 'this' is not any enum constant (should not
        # happen).
        code += [Literal(null),
                 endLabel,
                 Return()]
        method.setRoutine(buildRoutine(code))


class EnumConstantDeclaration(TypeBodyDeclaration):
    """Declaration of an enum constant."""

    def __init__(self, name, linespan):
        """Construct an enum constant declaration.

        Param name: string containing the simple name of the constant.

        Param linespan: see SyntaxNode.
        """
        TypeBodyDeclaration.__init__(self, linespan)
        self.__name = name

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return self.__name

    def api_getName(self):
        """Return the name of the enum constant as a string."""
        return self.__name

    def translateMembers(self, env, enclosingType):
        ep = ErrorProxy()
        if enclosingType.getFields(self.__name):
            ep.finish("redeclaration of enum constant '" + self.__name + "'",
                      self.linespan())
        constant = Field(self.__name, enclosingType,
                         enclosingType.getCanonicalName(),
                         isStatic=True, isFinal=True)
        enclosingType.addField(constant)
        staticCode = [SourceRef(self),
                      NewInstance(enclosingType),
                      Duplicate(),
                      CallStatic(enclosingType, '#constructor', 1),
                      Pop(),
                      StoreStatic(constant.getSignature())]
        return (staticCode, [])

    def translateBodies(self, env, enclosingType):
        return ([], [])


class FieldDeclaration(TypeBodyDeclaration):
    """Declaration of a field."""

    def __init__(self, modifiers, typeSymbol, declarators, linespan):
        """Construct a field declaration.

        Param modifiers: a list of strings containing the given
        modifiers.

        Param typeSymbol: a TypeSymbol representing the type of the
        field.

        Param declarators: a nonempty list of FieldDeclarators.

        Param linespan: see SyntaxNode.
        """
        TypeBodyDeclaration.__init__(self, linespan)
        self.__modifiers = modifiers
        self.__typeSymbol = typeSymbol
        self.__declarators = declarators

    def api_recurseStr(self, strFunc, *args, **kwargs):
        typeString = strFunc(self.__typeSymbol, *args, **kwargs)
        declStrings = [strFunc(d, *args, **kwargs) for d in self.__declarators]
        return ' '.join(self.__modifiers +
                        [typeString, ', '.join(declStrings)]) + ';'

    def api_getModifiers(self):
        """Return the modifiers as a list of strings."""
        return self.__modifiers[:]

    def api_getTypeSymbol(self):
        """Return the field type as a TypeSymbol."""
        return self.__typeSymbol

    def api_getDeclarators(self):
        """Return the declared fields as a list of FieldDeclarators."""
        return self.__declarators[:]

    def translateMembers(self, env, enclosingType):
        ep = ErrorProxy()
        dup = _reportDuplicateModifiers(self.__modifiers)
        if dup:
            ep.add(dup, self.linespan())
        isStatic = ('static' in self.__modifiers or
                    enclosingType.isInterface())
        isFinal = ('final' in self.__modifiers or
                   enclosingType.isInterface())
        if 'native' in self.__modifiers:
            ep.add('variables cannot be native', self.linespan())
        if 'abstract' in self.__modifiers:
            ep.add('variables cannot be abstract', self.linespan())
        type_ = self.__typeSymbol.resolve(env)
        staticCode = []
        instanceCode = []
        for d in self.__declarators:
            if enclosingType.getFields(d.name):
                ep.add("redeclaration of field '" + d.name + "'",
                       self.linespan())
            else:
                field = Field(d.name, type_, enclosingType.getCanonicalName(),
                              isStatic, isFinal)
                enclosingType.addField(field)
                if isStatic:
                    staticCode += [SourceRef(d),
                                   Literal(type_.getDefaultValue()),
                                   StoreStatic(field.getSignature())]
                else:
                    instanceCode += [SourceRef(d),
                                     LoadAuto('this'),
                                     Literal(type_.getDefaultValue()),
                                     StoreField(field.getSignature())]
        ep.raiseIfErrors()
        return (staticCode, instanceCode)

    def translateBodies(self, env, enclosingType):
        ep = ErrorProxy()
        staticCode = []
        instanceCode = []
        for d in self.__declarators:
            field = enclosingType.getFields(d.name)[0]
            if field.isFinal() and d.initial is None:
                ep.add("final variable '" + d.name + "' must be initialized",
                       d.linespan())
            if d.initial is not None:
                if field.isStatic():
                    try:
                        staticCode += ([SourceRef(d)] +
                                       d.initial.translate(env,
                                                           field.getType()) +
                                       [StoreStatic(field.getSignature())])
                    except StaticError, e:
                        ep.addException(e)
                else:
                    newEnv = env.makeSubscopedCopy()
                    newEnv.setInstanceContext()
                    try:
                        instanceCode += ([SourceRef(d),
                                          LoadAuto('this')] +
                                         d.initial.translate(newEnv,
                                                             field.getType()) +
                                         [StoreField(field.getSignature())])
                    except StaticError, e:
                        ep.addException(e)
        ep.raiseIfErrors()
        return (staticCode, instanceCode)


class FieldDeclarator(SyntaxNode):
    """A field declarator (a single identifier and possibly an initializer)."""

    def __init__(self, name, initial, linespan):
        """Construct a field declarator.

        Param name: a string containing the simple name for the field.

        Param initial: a VarInitializer representing the initial value for the
        field, or None if initialization is skipped.

        Param linespan: see SyntaxNode.
        """
        SyntaxNode.__init__(self, linespan)
        self.name = name
        self.initial = initial

    def api_recurseStr(self, strFunc, *args, **kwargs):
        if self.initial is None:
            return self.name
        else:
            return self.name + ' = ' + strFunc(self.initial, *args, **kwargs)

    def api_getName(self):
        """Return the name of the field as a string."""
        return self.name

    def api_getInitializer(self):
        """Return the VarInitializer.

        None is returned if the field has no initializer.
        """
        return self.initial


class MethodDeclaration(TypeBodyDeclaration):
    """Declaration of a method."""

    def __init__(self, modifiers, returnTypeSymbol, name, formals,
                 body, linespan):
        """Construct a method declaration.

        Param modifiers: a list of strings containing the given
        modifiers.

        Param returnTypeSymbol: a TypeSymbol representing the result
        type.

        Param name: a string containing the simple name of the method.

        Param formals: list of FormalParameters.

        Param body: a Block that contains the body of the method, or
        None if there is just a semicolon in place of the body.

        Param linespan: see SyntaxNode.
        """
        TypeBodyDeclaration.__init__(self, linespan)
        self.__modifiers = modifiers
        self.__returnTypeSymbol = returnTypeSymbol
        self.__name = name
        self.__formals = formals
        self.__body = body

    def api_recurseStr(self, strFunc, *args, **kwargs):
        typeString = strFunc(self.__returnTypeSymbol, *args, **kwargs)
        formalStrings = [strFunc(f, *args, **kwargs) for f in self.__formals]
        if self.__body is None:
            bodyString = ';'
        else:
            bodyString = '\n' + strFunc(self.__body, *args, **kwargs)
        return (' '.join(self.__modifiers +
                         [typeString, self.__name]) +
                '(' + ', '.join(formalStrings) + ')' +
                bodyString)

    def api_getModifiers(self):
        """Return the modifiers as a list of strings."""
        return self.__modifiers[:]

    def api_getReturnTypeSymbol(self):
        """Return the return type as a TypeSymbol."""
        return self.__returnTypeSymbol

    def api_getName(self):
        """Return the name of the declared method as a string."""
        return self.__name

    def api_getParameters(self):
        """Return the formal parameters as a list of FormalParameters."""
        return self.__formals[:]

    def api_getBody(self):
        """Return the method body as a Block.

        None is returned if the declaration has no body.
        """
        return self.__body

    def translateMembers(self, env, enclosingType):
        ep = ErrorProxy()
        dup = _reportDuplicateModifiers(self.__modifiers)
        if dup:
            ep.add(dup, self.linespan())
        isStatic = 'static' in self.__modifiers
        isAbstract = ('abstract' in self.__modifiers or
                      enclosingType.isInterface())
        isNative = 'native' in self.__modifiers
        isFinal = 'final' in self.__modifiers
        if isAbstract and self.__body is not None:
            ep.add('abstract methods must not have a body',
                   self.linespan())
        if isNative and self.__body is not None:
            ep.add('native methods must not have a body',
                   self.linespan())
        if not isNative and not isAbstract and self.__body is None:
            ep.add('method must have a body', self.linespan())
        if isAbstract and isStatic:
            ep.add('abstract methods cannot be static', self.linespan())
        if isAbstract and isFinal:
            ep.add('abstract methods cannot be final', self.linespan())
        if isAbstract and isNative:
            ep.add('native methods cannot be abstract', self.linespan())
        try:
            returnType = self.__returnTypeSymbol.resolve(env)
        except StaticError, e:
            ep.addException(e)
        params = []
        for f in self.__formals:
            if len(f.modifiers) > 1 or (len(f.modifiers) == 1 and
                                        f.modifiers[0] != 'final'):
                ep.add("formal parameter '" + f.name +
                       "' has illegal modifiers", f.linespan())
            if f.name in [name for (name, type_) in params]:
                ep.add("redeclaration of formal parameter '" +
                       f.name + "'", f.linespan())
            try:
                paramType = f.typeSymbol.resolve(env)
                params.append((f.name, paramType))
            except StaticError, e:
                ep.addException(e)
        ep.raiseIfErrors()
        types = [type_ for (name, type_) in params]
        declared = enclosingType.getExactFunctions(self.__name, types)
        if declared:
            ep.finish("redefinition of method '" + str(declared[0]) + "'",
                      self.linespan())
        method = Function(self.__name, params, returnType, isAbstract,
                          isStatic, isFinal, isNative)
        for stype in enclosingType.getDirectSupertypes():
            for m in stype.getExactFunctions(self.__name, types):
                if isStatic and (not m.isStatic() or m.isFinal()):
                    ep.finish("method '" + str(method) +
                              "' cannot hide '" + str(m) + "' in " +
                              str(stype),
                              self.linespan())
                if not isStatic and (m.isStatic() or m.isFinal()):
                    ep.finish("method '" + str(method) +
                              "' cannot override '" + str(m) + "' in " +
                              str(stype),
                              self.linespan())
                if not returnType.isReturnSubstitutableFor(m.getReturnType()):
                    ep.finish("method '" + str(method) +
                              "' has an incompatible return type with '" +
                              str(m) + "' in " + str(stype),
                              self.linespan())
        enclosingType.addFunction(method)
        return ([], [])

    def translateBodies(self, env, enclosingType):
        types = [f.typeSymbol.resolve(env) for f in self.__formals]
        method = enclosingType.getExactFunctions(self.__name, types)[0]
        if not method.isAbstract():
            returnType = method.getReturnType()
            params = method.getParameters()
            newEnv = env.makeSubscopedCopy()
            newEnv.setReturnType(returnType)
            code = [SourceRef(self)]
            if method.isNative():
                canonicalName = (enclosingType.getCanonicalName() + '.' +
                                 method.getSignature())
                if not env.isNativeHook(canonicalName):
                    ErrorProxy().finish("cannot find native implementation " +
                                        "for '" + canonicalName + "'",
                                        self.linespan())
                nativeParamTypes = [type_ for (name, type_) in params]
                if not method.isStatic():
                    nativeParamTypes = [enclosingType] + nativeParamTypes
                wrapper = NativeWrapper(env, canonicalName,
                                        nativeParamTypes,
                                        returnType)
                code += [CallNative(wrapper, len(nativeParamTypes))]
                code += [Return()]
            else:
                areFinal = [('final' in f.modifiers) for f in self.__formals]
                # Load argument values from the operand stack in reverse order.
                for ((name, type_), isFinal) in zip(params, areFinal)[::-1]:
                    newEnv.addLocalVar(name, type_, isFinal)
                    code += [StoreAuto(name)]
                if not method.isStatic():
                    code += [StoreAuto('this')]
                    newEnv.setInstanceContext()
                code += self.__body.translate(newEnv)
                if returnType is voidType:
                    # Return from method if execution
                    # falls off the end of the body.
                    code += [Literal(void), Return()]
                else:
                    # Execution must not fall off the body.
                    # Raise a runtime error if it does.
                    code += [MissingReturn()]
            method.setRoutine(buildRoutine(code))
        return ([], [])


class FormalParameter(SyntaxNode):
    """Representation of a formal parameter in the source code."""
    
    def __init__(self, modifiers, typeSymbol, name, linespan):
        """Construct a formal parameter representation.

        Param modifiers: a list of strings containing the given
        modifiers.

        Param typeSymbol: TypeSymbol representing the Type of the parameter.

        Param name: string containing the name of the parameter.

        Param linespan: see SyntaxNode.
        """
        SyntaxNode.__init__(self, linespan)
        self.modifiers = modifiers
        self.typeSymbol = typeSymbol
        self.name = name

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ' '.join(self.modifiers +
                        [strFunc(self.typeSymbol, *args, **kwargs), self.name])

    def api_getModifiers(self):
        """Return the modifiers as a list of strings."""
        return self.modifiers[:]

    def api_getTypeSymbol(self):
        """Return the parameter type as a TypeSymbol."""
        return self.typeSymbol

    def api_getName(self):
        """Return the name of the parameter as a string."""
        return self.name


class ConstructorDeclaration(TypeBodyDeclaration):
    """Declaration of a constructor."""

    def __init__(self, name, formals, constructorStatements, linespan):
        """Construct a constructor declaration.

        Param name: a string containing the simple type name appearing
        in the header of the declaration.

        Param formals: list of FormalParameters.

        Param constructorStatements: a list of ConstructorStatements
        that constitute the constructor body.

        Param linespan: see SyntaxNode.
        """
        TypeBodyDeclaration.__init__(self, linespan)
        self.__name = name
        self.__formals = formals
        self.__ctorStatements = constructorStatements

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return (self.__name + '(' +
                ', '.join([strFunc(f, *args, **kwargs) for
                           f in self.__formals]) + ')\n{\n' +
                '\n'.join([self.indent(strFunc(s, *args, **kwargs)) for
                           s in self.__ctorStatements]) + '\n}')

    def api_getParameters(self):
        """Return the formal parameters as a list of FormalParameters."""
        return self.__formals[:]

    def api_getBody(self):
        """Return the constructor body as a list of ConstructorStatements."""
        return self.__ctorStatements[:]

    def translateMembers(self, env, enclosingType):
        assert not enclosingType.isInterface()
        ep = ErrorProxy()
        if self.__name != enclosingType.getSimpleName():
            ep.finish('missing return type in method declaration',
                      self.linespan())
        params = []
        for f in self.__formals:
            if len(f.modifiers) > 1 or (len(f.modifiers) == 1 and
                                        f.modifiers[0] != 'final'):
                ep.add("formal parameter '" + f.name +
                       "' has illegal modifiers", f.linespan())
            if f.name in [name for (name, type_) in params]:
                ep.add("redeclaration of formal parameter '" +
                       f.name + "'", f.linespan())
            try:
                paramType = f.typeSymbol.resolve(env)
                params.append((f.name, paramType))
            except StaticError, e:
                ep.addException(e)
        ep.raiseIfErrors()
        types = [type_ for (name, type_) in params]
        declared = enclosingType.getExactFunctions('#constructor', types)
        if declared:
            ep.finish("redeclaration of '" + str(declared[0]) + "'",
                      self.linespan())
        constructor = Function('#constructor', params, voidType)
        enclosingType.addFunction(constructor)
        return ([], [])

    def translateBodies(self, env, enclosingType):
        ep = ErrorProxy()
        types = [f.typeSymbol.resolve(env) for f in self.__formals]
        constructor = enclosingType.getExactFunctions('#constructor', types)[0]
        params = constructor.getParameters()
        newEnv = env.makeSubscopedCopy()
        code = [SourceRef(self)]
        areFinal = [('final' in f.modifiers) for f in self.__formals]
        # Load argument values from the operand stack in reverse order.
        for ((name, type_), isFinal) in zip(params, areFinal)[::-1]:
            newEnv.addLocalVar(name, type_, isFinal)
            code += [StoreAuto(name)]
        code += [StoreAuto('this'),
                 LoadAuto('this'),
                 CallStatic(enclosingType, '#createvars', 1),
                 Pop(),
                 LoadAuto('this')]
        superclass = enclosingType.getDirectSupertypes()[0]
        # hasExplicit is True if the body begins with an explicit
        # superclass constructor invocation.
        hasExplicit = (len(self.__ctorStatements) > 0 and
                       self.__ctorStatements[0].isConstructorInvocation())
        argTypes = []
        if not hasExplicit:
            # Try to generate implicit call super();
            superCtors = superclass.getExactFunctions('#constructor', [])
        else:
            code += [SourceRef(self.__ctorStatements[0])]
            for a in self.__ctorStatements[0].getConstructorArgs():
                try:
                    result = a.translate(newEnv)
                    argTypes.append(result.getType())
                    code += result.getEvalCode()
                except StaticError, e:
                    ep.addException(e)
            ep.raiseIfErrors()
            allSuperCtors = superclass.getFunctions('#constructor')
            superCtors = determineFunction(allSuperCtors, argTypes)
        if not superCtors:
            ep.finish("no superclass constructor matches argument types (" +
                      ', '.join([str(t) for t in argTypes]) +
                      ")",
                      self.linespan())
        if len(superCtors) > 1:
            ep.finish("ambiguous superclass constructor invocation " +
                      "matches both '" +
                      str(superCtors[0]) + "' and '" +
                      str(superCtors[1]) + "'",
                      self.linespan())
        code += [CallStatic(superclass, superCtors[0].getSignature(), 1),
                 Pop(),
                 LoadAuto('this'),
                 CallStatic(enclosingType, '#initvars', 1),
                 Pop()]
        newEnv = newEnv.makeSubscopedCopy()
        newEnv.setReturnType(void)
        # Allow the use of 'this' in constructor body, but not in
        # superclass constructor invocation.
        newEnv.setInstanceContext()
        if hasExplicit:
            restStatements = self.__ctorStatements[1:]
        else:
            restStatements = self.__ctorStatements
        for s in restStatements:
            try:
                code += s.translateBlockStatement(newEnv)
            except StaticError, e:
                ep.addException(e)
        ep.raiseIfErrors()
        code += newEnv.getEndOfScopeCode()
        code += [Literal(void), Return()]
        constructor.setRoutine(buildRoutine(code))
        return ([], [])


class ConstructorStatement(SyntaxNode):
    """A base class for constructor statements.

    A constructor statement is either an explicit constructor
    invocation or a block statement in a constructor body.
    """

    def __init__(self, linespan):
        SyntaxNode.__init__(self, linespan)

    def api_recurseStr(self, strFunc, *args, **kwargs):  #pragma: no cover 
        return '<ctor stmt>'

    def isConstructorInvocation(self):
        """Return True if self is an explicit constructor invocation."""
        return False  # Default behavior

    def getConstructorArgs(self):
        """Return the superclass constructor arguments.

        Self must be an explicit constructor invocation.  The return
        value is a list of Exprs.
        """
        assert False  # Default behavior

    def translateBlockStatement(self, env):
        """Treat self as a block statement and mimic its translate method."""
        raise NotImplementedError('Pure virtual method')


class ConstructorBlockStatement(ConstructorStatement):
    """A block statement inside a constructor body."""

    def __init__(self, blockStatement):
        ConstructorStatement.__init__(self, blockStatement.linespan())
        self.__blockStatement = blockStatement

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return strFunc(self.__blockStatement, *args, **kwargs)

    def api_getBlockStatement(self):
        """Return the BlockStatement."""
        return self.__blockStatement

    def translateBlockStatement(self, env):
        return self.__blockStatement.translate(env)


class SuperConstructorInvocation(ConstructorStatement):
    """An explicit constructor invocation using the keyword super."""

    def __init__(self, argList, linespan):
        """Construct a superclass constructor invocation.

        Param argList: a list of Exprs representing the arguments.

        Param linespan: see SyntaxNode.
        """
        ConstructorStatement.__init__(self, linespan)
        self.__argList = argList

    def api_recurseStr(self, strFunc, *args, **kwargs):
        return ('super(' +
                ', '.join([strFunc(e, *args, **kwargs) for
                           e in self.__argList]) + ');')

    def api_getArgList(self):
        """Return the arguments as a list of Exprs."""
        return self.__argList[:]

    def isConstructorInvocation(self):
        return True

    def getConstructorArgs(self):
        return self.__argList[:]

    def translateBlockStatement(self, env):
        ErrorProxy().finish('explicit superclass constructor invocation ' +
                            'must begin the constructor body',
                            self.linespan())


def _reportDuplicateModifiers(modifiers):
    """Check if a modifier appears twice and return an error message if so.

    Param modifiers: a list of strings.

    Return value: an error message if the list contains two strings
    that are equal.  Otherwise an empty string is returned.
    """
    for i in xrange(len(modifiers)):
        if modifiers[i] in modifiers[:i]:
            return 'modifier ' + modifiers[i] + ' appears twice'
    return ''


class NativeWrapper:
    """A wrapper for Python functions."""

    def __init__(self, env, name, paramTypes, returnType):
        """Construct a wrapper.

        Param env: a StaticEnvironment for obtaining the function object.

        Param name: a string containing the canonical name of the method.

        Param paramTypes: a list of Types for method parameters.

        Param returnType: the ResultType that should be returned.
        """
        self.__name = name
        self.__function = env.getNativeHook(name)
        self.__stringType = env.getTopType('String')
        self.__intArrayType = intType.getArrayType(env)
        self.__paramTypes = paramTypes
        self.__returnType = returnType

    def __str__(self):
        return self.__name

    def execute(self, state, args):
        """Run the function.

        Raise a ReturnTypeError if the Python function returns a bad
        value.

        Param state: the VirtualMachine instance passed to the function.

        Param args: the function arguments in the form used by the
        virtual machine.  These are converted to Python format before
        calling the function.

        Return value: the value returned by the Python function object
        converted to the format accepted by the virtual machine and
        compatible with the given return type.
        """
        pargs = [state]
        for (value, type_) in zip(args, self.__paramTypes):
            if value is null:
                pvalue = None
            elif type_ is boolType:
                if value == boolValue[True]:
                    pvalue = True
                else:
                    pvalue = False
            elif type_ is self.__stringType:
                assert value.getType() is self.__stringType
                # Convert String object to a Python string.  This
                # conversion is dependent on the String implementation
                # in basetypes.py, among other things.
                array = value.getField('String._chars')
                length = array.getField('[].length')
                chars = [array.getComponent(i) for i in xrange(length)]
                pvalue = ''.join([chr(c) for c in chars if c >= 0 and c < 256])
            else:
                pvalue = value
            pargs.append(pvalue)
        returnValue = self.__function(*pargs)
        if self.__returnType is boolType:
            if returnValue == True:
                return boolValue[True]
            if returnValue == False:
                return boolValue[False]
            raise ReturnTypeError('return value is not a boolean')
        if self.__returnType is intType:
            if (isinstance(returnValue, int) and
                returnValue >= -(1L<<31) and returnValue < 1L<<31):
                return returnValue
            raise ReturnTypeError('return value is not an int')
        if self.__returnType is voidType:
            return void
        if self.__returnType.isReference():
            if returnValue is None:
                return null
            if self.__returnType is self.__stringType:
                # Convert Python string to a String object.  This
                # conversion is dependent on the String implementation
                # in basetypes.py, among other things.
                components = [ord(c) for c in str(returnValue) if
                              ord(c) >= 0 and ord(c) < 256]
                array = Instance(self.__intArrayType)
                array.setField('[].length', len(components))
                array.initializeComponents(len(components), 0)
                for i in xrange(len(components)):
                    array.setComponent(i, components[i])
                string = Instance(self.__stringType)
                string.setField('String._chars', array)
                return string
            if (isinstance(returnValue, Instance) and
                returnValue.getType().isSubtypeOf(self.__returnType)):
                return returnValue
            raise ReturnTypeError('return value is not ' +
                                  str(self.__returnType))
        assert False  # Given return type is unknown
