#! /usr/bin/python2.4

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Usage: analyze.py [OPTION]... MODEL [BACKEND-OPTION]...

Check a UML 1.4 model contained in the project file MODEL (in XMI 2.0
format) for counterexamples.

General options:
  --configuration-file=FILENAME
                             Specify alternate configuration file
                             [default: paths.conf]
  -h, --help                 Display this help and exit.
  --gui                      Run the analyzer in graphical mode.
  --quiet, --silent          Suppress all messages when running in console
                             mode.
  -v, --verbose              Show verbose output (repeating the option
                             multiple times will increase output
                             verbosity).  Effective only when running in
                             console mode.

  In console mode, the script will check the given model for
  counterexamples.  If a counterexample is found, the script will print a
  generic counterexample trace to standard output (or, alternatively, save
  the trace to a file specified using the --counterexample-trace-file
  option).  In graphical mode, the script displays a GUI for choosing model
  analysis settings; this GUI will be displayed again between model
  checking runs.

Model preprocessing options:
  --flatten-hierarchy        Run the state machine hierarchy flattener
                             before model analysis.
  --no-hierarchy-flattening  Do not run the state machine hierarchy
                             flattener before model analysis [default].
  --flatten-action-language  Run the action language flattener before model
                             analysis.
  --flatten-action-language-split-sends
                             Run the action language flattener before model
                             analysis; break each transition whose effect
                             contains multiple `send' statements into a
                             sequence of transitions, each of which
                             contains at most one `send' statement in its
                             effect.
  --flatten-action-language-single-statement
                             Run the action language flattener before model
                             analysis; break each transition whose effect
                             contains multiple simple statements into a
                             sequence of transitions, each of which
                             contains at most one simple statement in its
                             effect.
  --no-action-language-flattening
                             Do not run the action language flattener
                             before model analysis [default].
  --slice-assertions         Try to simplify the model before analysis by
                             removing action language code which cannot
                             affect any assertions in the model.
  --no-slicing               Do not apply model slicing [default].
  --flatten-effect-expressions
                             Run the state machine transition effect
                             expression flattener before model analysis.
  --no-effect-expression-flattening
                             Do not run the state machine transition effect
                             expression flattener before model analysis
                             [default].

  The model preprocessors are always run in the order (i) state machine
  hierarchy flattener, (ii) action language flattener, (iii) model
  slicer, (iv) transition effect expression flattener.

Model analysis options:
  --backend=NAME             Specify a model checking back-end to use
                             [default: uboco]. The following back-ends are
                             supported:
                               proco
                                 Use the Proco back-end with Spin as the
                                 model checker.
                               suboco[:SMT-SOLVER]
                                 Use the Suboco back-end.  SMT-SOLVER
                                 selects between the SMT solvers available
                                 for use in bounded model checking (CVC3,
                                 MathSat, STP, Yices, STP, Z3).  The
                                 default is Yices.
                               uboco[:MODE[:BOUND]]
                                 Use the Uboco back-end with NuSMV as the
                                 model checker.  MODE can be either a name
                                 of one of the satisfiability solvers
                                 supported by NuSMV (ZChaff, MiniSat, Sim),
                                 in which case the model will be analyzed
                                 using bounded model checking with the
                                 chosen SAT solver (using BOUND as the
                                 bound for BMC), or "bdd", in which case
                                 the model will be analyzed exhaustively
                                 using BDD-based model checking.  The
                                 default value for MODE is ZChaff, and
                                 the default BOUND is 30.
  --counterexample-trace-file=FILE
                             If a real counterexample is found, write a
                             counterexample trace in generic format to FILE
                             [default: "-", which refers to standard
                             output].
  --simulate-counterexample  If a real counterexample is found, simulate it
                             (writing output to standard output).
  --no-counterexample-simulation
                             Do not simulate counterexamples [default].

  Any BACKEND-OPTIONs present at the end of the command line will be passed
  to the chosen back-end (Proco, Uboco, Suboco).

Abstraction refinement options:
  --refinement               Apply automatic abstraction refinement to the
                             model's integer attributes.
  --no-refinement            Do not apply automatic abstraction refinement
                             [default].
  --integer-semantics=SEMANTICS
                             Specify the integer semantics to use for
                             generating abstract types for abstracion
                             refinement.  SEMANTICS can be one of
                             "bounded", "unbounded", or "mixed" [default:
                             bounded].
  --abstract-type-smt-solver=SOLVER
                             Specify an SMT solver to use for generating
                             operations for abstract types under bounded
                             or mixed integer semantics.  SOLVER can be
                             one of CVC3, MathSAT, STP, Yices, or Z3
                             [default: Yices].
  --expression-flattening-limit=VALUE
                             Specify an integer limit for flattening
                             definitions of inlined abstract operations
                             [default: 0, which disables expression
                             flattening altogether].  Use the value "none"
                             to enable maximal flattening.

  Abstract arithmetic operations generated using bounded integer semantics
  provide safe approximations of the corresponding operations on 32-bit
  signed integers (with wraparound on overflow) used in the Jumbala action
  language.  Alternatively, the abstract operations can be made to
  approximate operations on unbounded integers by using unbounded integer
  semantics; however, in this case only models with basic arithmetic (+, -,
  *, /), logical (&&, ||, !) and comparison operations (==, !=, <, <=,
  >, >=), and the ternary conditional operator (?:) are supported as input.
  When using mixed semantics for abstract types, definitions of basic
  arithmetic and comparison operations (and coercions between abstract
  types) will be generated by approximating the corresponding operations on
  unbounded integers, and the remaining operators will be approximations of
  bounded integer operations.

This script was written by Heikki Tauriainen for the SMUML project
<http://www.tcs.hut.fi/Research/Logic/SMUML.shtml>.
"""

import sys, os, logging
from tempfile import mkdtemp
import coral
from frontend import configuration, backends, logutil, subtaskwrapper, \
     refinementcycle, checkutil
from abstraction import error
import flatten_state_machine_hierarchy as HierarchyFlattener
import flatten_state_machine_action_language as ActionLanguageFlattener
import flatten_transition_effects as TransitionEffectFlattener

UML = coral.lang.UML14

# -----------------------------------------------------------------------------
# Helper functions for file and directory access.

def loadProject(filename):
    """Load a Coral project containing a UML 1.4 model.

    Argument:
        filename -- Input file name.

    Returns a pair (project, model), where `project' is a coral.Project
    object, and `model' is a coral.kernel.ElementPtr object for which
    `model.oclIsKindOf(coral.lang.UML14.Model) == True' holds.

    Raises an error.Error if the project file cannot be processed, or
    if the project in the given file contains multiple models (or no
    model at all).
    """
    try:
        project = coral.loadDocument(filename)
    except Exception, e:
        raise error.Error(str(e))

    model = None
    for element in project.contents():
        if element.oclIsKindOf(UML.Model):
            if model:
                raise error.Error("multiple model definitions in input file")
            model = element
    if not model:
        raise error.Error("no model found in input file")

    return (project, model)

def saveProject(project, filename):
    """Saves a Coral project under a given name.

    Arguments:
        project  -- A coral.Project object.
        filename -- Name of output file.

    Raises an error.Error if an error occurs while saving.
    """
    try:
        project.setFilename(filename)
        coral.saveDocument(project)
    except Exception, e:
        raise error.Error(str(e))

def cleanDirectory(directory, files_to_exclude=None):
    """Delete files in a given directory.

    Argument:
        directory        -- Name of directory.
        files_to_exclude -- An optional list of file names to exclude from
                            deletion (by default, all files are deleted).
    """
    if files_to_exclude == None:
        files_to_exclude = []
    for filename in os.listdir(directory):
        name = os.path.join(directory, filename)
        if os.path.isfile(name) and not filename in files_to_exclude:
            try:
                os.remove(name)
            except:
                pass

# -----------------------------------------------------------------------------
# The script

# process and validate command line options; initialize a logger for output
# messages

try:
    config = configuration.Configuration(__doc__)
    if config.gui_mode:
        logging.setLoggerClass(logutil.GraphicalLogger)
        logger = logging.getLogger("log")
        logger.setLevel(1)
        log_handler = logutil.LogHandler(logger)
        log_handler.setLevel(0)
        from frontend.settings_gui import SettingsGui
        settings_gui = SettingsGui(config)
    else:
        logging.setLoggerClass(logutil.StderrLogger)
        logger = logging.getLogger("log")
        logger.setLevel(config.loglevel)
        log_handler = logutil.LogHandler(logger)
        log_handler.setLevel(config.loglevel)
        if not config.model_file_name:
            raise error.OptionError("model name expected")
    log_handler.setFormatter(logging.Formatter())
    logger.addHandler(log_handler)
except error.OptionError, e:
    error.abort(str(e), 2)

current_dir = os.getcwd()
delete_work_dir = True
workdir = None

try:
    while True:
        os.chdir(current_dir)
        if config.gui_mode:
            # run the settings GUI (using the directory in which the script was
            # started as the working directory)
            settings_gui.run()

        # prepare the logger to process messages
        logger.initDecorations()

        try:
            try:
                # finalize configuration, read the paths configuration file
                config.finalize()

                # clean or create a working directory
                if workdir:
                    logger.info("cleaning up working directory `%s'" % workdir)
                    cleanDirectory(workdir)
                else:
                    workdir = mkdtemp()
                    logger.info("using `%s' as working directory" % workdir)
                os.chdir(workdir)

                # load the model
                logger.info("loading model `%s'" % config.model_file_name)
                logger.startActivity("Loading model")
                try:
                    (project, model) = loadProject(config.model_file_name)
                finally:
                    logger.stopActivity()

                # model preprocessing
                if config.flatten_hierarchy \
                       or config.flatten_action_language \
                       or config.slice_assertions \
                       or config.flatten_transition_effects:
                    logger.info(("preparing model for analysis", [], True))
                    logger.startActivity("Preparing model for analysis")
                    try:
                        preprocessors = []

                        if config.flatten_hierarchy:
                            # flatten state machine hierarchy
                            preprocessors.append("H")
                            subtaskwrapper.runSubtask \
                                (HierarchyFlattener.flatten,
                                 args=(model,),
                                 msg="flattening state machine hierarchy",
                                 capture_prefix="flatten_hierarchy",
                                 stdout=subtaskwrapper.PIPE,
                                 stderr=subtaskwrapper.PIPE)
                            config.model_file_name = \
                                "model_%s.xmi" % "".join(preprocessors)
                            saveProject(project, config.model_file_name)

                        if config.flatten_action_language:
                            # flatten action language constructs
                            preprocessors.append("A")
                            subtaskwrapper.runSubtask \
                                (ActionLanguageFlattener.flatten,
                                 args=(project, config.flatten_actlang_all,
                                       config.flatten_actlang_sends),
                                 msg="flattening action language constructs",
                                 capture_prefix="flatten_action_language",
                                 stdout=subtaskwrapper.PIPE,
                                 stderr=subtaskwrapper.PIPE)
                            config.model_file_name = \
                                "model_%s.xmi" % "".join(preprocessors)
                            saveProject(project, config.model_file_name)

                        if config.slice_assertions:
                            # run model slicer:
                            # <slicer> -o <sliced model> -A <model>
                            preprocessors.append("S")
                            new_model_file_name = \
                                "model_%s.xmi" % "".join(preprocessors)
                            args = config.getPath("PYTHON")
                            args.extend(config.getPath("slicer"))
                            args.extend \
                                (("-o", new_model_file_name, "-A",
                                  config.model_file_name))
                            subtaskwrapper.runSubtask \
                                (args, "slicing model",
                                 capture_prefix="slicer",
                                 stdout=subtaskwrapper.PIPE,
                                 stderr=subtaskwrapper.PIPE)
                            config.model_file_name = new_model_file_name
                            (project, model) = \
                                loadProject(config.model_file_name)

                        if config.flatten_transition_effects:
                            # flatten expressions in effects of transitions
                            preprocessors.append("E")
                            subtaskwrapper.runSubtask \
                                (TransitionEffectFlattener.processModel,
                                 args=(model,),
                                 msg="flattening expressions in effects of "
                                     "transitions",
                                 capture_prefix="flatten_effect_expressions",
                                 stdout=subtaskwrapper.PIPE,
                                 stderr=subtaskwrapper.PIPE)
                            config.model_file_name = \
                                "model_%s.xmi" % "".join(preprocessors)
                            saveProject(project, config.model_file_name)
                    finally:
                        logger.dedent(logging.INFO)
                        logger.stopActivity()

                # model checking
                if not config.use_refinement:
                    # if not using abstraction refinement, simply run the
                    # back-end model checker on the model
                    trace_file = \
                        checkutil.analyzeModel \
                            (config.model_file_name, config.backend,
                             config.getPath)
                else:
                    # else start the abstraction refinement cycle
                    project.setFilename("abstract_model.xmi")
                    refiner = \
                        refinementcycle.ModelRefiner(project, model, config)
                    logger.info("Starting abstraction refinement cycle")
                    iteration = 1
                    while True:
                        logger.info(("Iteration %d" % iteration, [], True))
                        try:
                            refiner.abstract()
                            trace_file = \
                                checkutil.analyzeModel \
                                    ("abstract_model.xmi", config.backend,
                                     config.getPath)
                            if trace_file == None:
                                # no counterexample was found
                                break
                            refinement = \
                                checkutil.findRefinement \
                                    (trace_file, refiner, config)
                            if refinement == None:
                                # a real counterexample was found
                                break
                            refiner.refine(refinement)
                            cleanDirectory(workdir, [config.model_file_name])
                            iteration += 1
                        finally:
                            logger.dedent(logging.INFO)

                # counterexample reporting
                if trace_file == None:
                    logger.info \
                        ("no errors found in model %s"
                         % ((isinstance(config.backend, backends.Uboco)
                             and not config.backend.use_bmc
                             and "(exhaustive search)")
                            or "(bounded model checking)"))
                else:
                    logger.info(("a counterexample was found", [], True))
                    try:
                        checkutil.processCounterExampleTrace \
                            (config.model_file_name, trace_file,
                             config.trace_file_name,
                             config.simulate_counterexample, config.gui_mode,
                             config.getPath)
                    finally:
                        logger.dedent(logging.INFO)

                logger.finish("Finished")
                if config.gui_mode:
                    # if running in GUI mode, proceed to the next model
                    # checking run
                    continue

                break

            except SystemExit, e:
                raise
            except Exception, e:
                logger.error(str(e))
                logger.finish("Aborted")
                if config.gui_mode and isinstance(e, error.Error):
                    # if a non-fatal error occurs in GUI mode, proceed to the
                    # next model checking run
                    continue
                # otherwise exit the program
                if workdir:
                    delete_work_dir = \
                        logger.askQuestion \
                            ("Remove temporary working directory `%s'"
                             % workdir)
                if isinstance(e, error.Error):
                    # exit with a non-fatal error
                    sys.exit(1)
                else:
                    # exit due to an unexpected error
                    raise
        finally:
            logger.cleanUpDecorations()
finally:
    os.chdir(current_dir)
    if workdir and delete_work_dir:
        try:
            cleanDirectory(workdir)
            os.rmdir(workdir)
        except:
            pass
    if config.cfg_file_is_a_tempfile:
        try:
            os.remove(config.cfg_file_name)
        except:
            pass
