#! /usr/bin/python2.4
# coding: latin-1

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""SMUML Counterexample analysator. Analyzes abstract counterexamples
got from abstracted UML 1.4 models.

Author: Vesa Ojala <vojala@tcs.tkk.fi>
"""

import sys
import coral
import actlang
import Sim
import tracesim
from canal import absttrace
from canal import utils
from canal import error
import optparse

class Counterexample_Analysator_Attribute_Error(Exception):
    """Exception for situations where the parameters passed to feas.py
    during initialization are wrong type

    Fields:
        parameter -- parameter exception concerns (as string)
        typelist -- list of types it should be (as list of strings)
    """
    def __init__(self, parameter, typelist):
        """Initializes Counterexample_Analysator_Attribute_Error

        Arguments:
            parameter -- parameter exception concerns (as string)
            typelist -- list of types it should be (as list of strings)
        """
        self.parameter = parameter
        self.typelist = typelist
        
    def __str__(self):
        s = "" + self.__class__.__name__ + ": " + self.parameter + " should be "
        if len(typelist) >= 2:
            s += " one from the list: "
        s += ",".join(typelist)

class Internal_Error(error.Error):
    """Internal error in the feasibility analysis.

    Attributes:
        reason -- explanation for the cause of the exception (string)
    """
    def __init__(self, msg=""):
        """Initialize an Error with an optional error message.
        
        Arguments:
            msg  -- error message
        """
        self.reason = "Feasibility Analysis - Internal Error: " + msg

class Counterexample_Analysator:
    """Counterexample Analysator.

    """
    def __init__(self, trace, abst_model, conc_model, error_types=("ASSERT", "IMPL"),
                 stdout=sys.stdout, stderr=sys.stderr, verbose=1):
        """Initializes Counterexample_Analysator.

        Attributes:
            trace        -- Filename of a file containing a counterexample trace
                            as a SMUML generic trace format.
            abst_model   -- Abstract model associated with the counterexample trace.
                            Can be a filename of a file containing a Coral project, 
                            a Coral project object, or None. If None is given, a
                            filename of a file containing the abstract model is read
                            from the trace file.
            conc_model   -- Concrete model associated with the counterexample trace.
                            Can be a filename of a file containing a Coral project or
                            a Coral project object.
            error_types  -- a list of errors to be checked, may contain 'ASSERT' and 'IMPL'
            stdout       -- this can be used for overriding stdout as an output for messages
            stderr       -- this can be used for overriding stdout as an output for error
                            messages
            verbose      -- 0, 1, or 2 depending on how verbose output is preferred.
        
        """
        
        if not (type(trace) is str):
            raise Counterexample_Analysator_Attribute_Error("trace", ["string"])
        self.trace_filename = trace
        if not (type(abst_model) is str or
                abst_model == None or
                isinstance(abst_model, coral.kernel.ProjectPtr)):
            raise Counterexample_Analysator_Attribute_Error("abst_model",
                                                            ["string","None","Coral project"])
        self.abst_model = abst_model
        if not (type(conc_model) is str or
                isinstance(conc_model, coral.kernel.ProjectPtr)):
            raise Counterexample_Analysator_Attribute_Error("conc_model",
                                                            ["string","Coral project"])
        self.conc_model = conc_model

        for err in error_types:
            if not (type(err) is str):
                raise Counterexample_Analysator_Attribute_Error("error_types", ["list of string"])
        self.error_types = error_types

        self.stdout = stdout
        self.stderr = stderr
        if not verbose in (0,1,2):
            raise Counterexample_Analysator_Attribute_Error("verbose", ["0, 1, or 2"])
        self.verbose = verbose

        if 'ASSERT' in self.error_types:
            check_asserts = True
        else:
            check_asserts = False

        # trace simulator for concrete model
        self.conc_ts = tracesim.Trace_Simulator(self.trace_filename, check_asserts, self.conc_model)

        # trace simulator for abstract model
        if self.abst_model != None:
            self.abst_ts = tracesim.Trace_Simulator(self.trace_filename, check_asserts, self.abst_model)
        else:
            self.abst_ts = tracesim.Trace_Simulator(self.trace_filename, check_asserts)

        # when the differing place is recognized, following variables are initialized
        self.conc_sim_oid = None
        self.abst_sim_oid = None

    def _inform_activity(self, s):
        if self.verbose >= 1:
            print >> self.stdout, s

    def _inform_result(self, s):
        if self.verbose >= 1:
            print >> self.stdout, "" + s

    def _inform_relevant(self, ct):
        if self.verbose >= 2:
            print >> self.stdout, "Counterexample trace with relevant variables marked with star(*):"
            print >> self.stdout, str(ct)

    def _inform_refinement_hint(self, rh):
        if self.verbose >= 1:
            s = "Following variables are suggested to be refined:\n"
            for v, h in rh.refinements:
                s += "  " + v + ": " + ", ".join(map(str, h)) + "\n"
            self._inform_result(s)

    def analyze(self, mode='FULL'):
        """Analyses counterexample.

        Attributes:
            mode  -- Determines the extent of the analysis. Can be one of the
                     followings:
                       * 'FEAS'     -- Determines only feasibility of a
                                       counterexample.
                       * 'RELEVANT' -- Determines feasibility of a counterexample,
                                       analyzes relevant variables at the error
                                       point and calculates relevant variables
                                       in other points of the trace.
                       * 'FULL'     -- Determines feasibility of a counterexample,
                                       calculates relevant variables and gives
                                       a refinement hint.
        
        Returns:
          if mode is 'FEAS':
            * True if the counterexample is feasible and False if not.
          if mode is 'RELEVANT':
            * True if the counterexample is feasible and False if not.
          if mode is 'FULL':
            * Refinement_Hint object in the case of spurious counterexample
            * None in the case of true counterexample
        """
        i = 0

#        REL_MSG = "Counterexample trace with relevant variables marked with star(*):\n"

        self._inform_activity("Counterexample analysator is checking for following errors:")
        self._inform_activity("  * runtime errors (can not be turned off)")
        if 'ASSERT' in self.error_types:
            self._inform_activity("  * assertion errors")
        if 'IMPL' in self.error_types:
            self._inform_activity("  * implicit consumptions")
        self._inform_activity("")
        self._inform_activity("Analysing feasibility of the counterexample by simulating\n"+\
                              "counterexample step by step in the concrete model:")
        while True:
            # execute one step in the concrete simulator
            try:
                action = self.conc_ts.simulate_next_step()
                if action == None:
                    break
                elif action[0] == 'DEFER':
                    (dummy, sim_oid) = action
                    self._inform_activity("  Object "+str(sim_oid)+" defers a signal")
                elif action[0] == 'IMPL':
                    # true implicit consumption
                    self._inform_activity("  Object "+str(sim_oid)+" performs implicit consumption")
                    if action[0] in self.error_types:
                        self._inform_activity("")
                        if mode == 'FEAS':
                            return True
                        elif mode == 'RELEVANT':
                            return True
                        else:
                            return None
                    (dummy, sim_oid) = action
                elif action[0] == 'EXEC':
                    (dummy, sim_oid, uml_trans) = action
                    self._inform_activity("  Object "+str(sim_oid)+" moves from '"+\
                                          uml_trans.source.name+"' to '"+uml_trans.target.name+"'")
                elif action[0] == 'EffectChooseValue':
                    pass
                elif action[0] == 'GuardChooseValue':
                    pass

            except tracesim.TraceSimulatorError, e:
                #print "Exception occurred during simulation of the abstract trace in concrete model"
                if isinstance(e, (tracesim.DeferNotEnabledError, tracesim.ImplNotEnabledError)):
                    # recognize error type 'DEFER' or 'IMPL'
                    #print "The trace up to and including the current global state is:"
                    #self.conc_ts.sim.api_print_trace()
                    #print e
                    if isinstance(e, tracesim.DeferNotEnabledError):
                        self._inform_result("Signal defer in Object "+str(e.sim_oid)+" can not be executed.")
                        error_type = 'DEFER'
#                        error_msg = "Signal defer happened in the abstract"+\
#                                    "model can not happen in the concrete one."
                    elif isinstance(e, tracesim.ImplNotEnabledError):
                        self._inform_result("Implicit comsumption in Object "+str(e.sim_oid)+" can not be executed.")
#                        error_msg = "Implicit signal consumption happened in the abstract"+\
#                                    "model can not happen in the concrete one."
                        error_type = 'IMPL'

                    if mode == 'FEAS':
                        return False
                    
                    # check that abstract simulator can simulate the next step
                    abst_action = self.abst_ts.simulate_next_step()
                    assert abst_action[0] == error_type
                    self.abst_ts.backtrack_one_step()
                    self.abst_ts.process_next_chooses()

                    abst_action = self.abst_ts.get_next_action()
                    assert abst_action[0] == error_type
                        
                    # concete simulator object id
                    sim_oid = e.sim_oid

                    abst_gs = self.abst_ts.sim.get_working_global_state()
                    conc_gs = self.conc_ts.sim.get_working_global_state()
                    abst_obj = utils.get_object(abst_gs, abst_action[1])
                    conc_obj = utils.get_object(conc_gs, sim_oid)

                    assert abst_obj.get_input_queue_length() > 0
                    abst_msg = abst_obj.get_input_queue_message(0)
                    if conc_obj.get_input_queue_length() > 0:
                        conc_msg = conc_obj.get_input_queue_message(0)
                    else:
                        conc_msg = None
                    self._inform_activity("Calculating relevant variables")
                    # signal has gone in the from place (not same type of message in the both signal
                    # queues)
                    if (conc_msg == None or
                        conc_msg.get_signal().__XMIid__() != abst_msg.get_signal().__XMIid__()):
                        #print "message error"
                        ct = absttrace.Counterexample_Trace(self.conc_ts,
                                                            self.abst_ts,
                                                            "trigger")
                        ct.analyze_trigger_error()
                    # correct type of signal but the guard is not true in the abstract world
                    else:
                        #print "guard error"
                        ct = absttrace.Counterexample_Trace(self.conc_ts,
                                                            self.abst_ts,
                                                            "guard")
                        ct.analyze_guard_error()
                        #ct.analyze_error_point("guard")
                        #ct.make_slice2()

                    # common steps for both cases
                    #print ct
                    self._inform_relevant(ct)
                    if mode=='RELEVANT':
                        return False
                    self._inform_activity("Determining variables that are suggested to be refined")
                    ct.find_refinement_variables()
                    self._inform_refinement_hint(ct.get_refinement_hint())
                    return ct.get_refinement_hint()
                elif isinstance(e, tracesim.TransNotEnabledError):
#                    error_msg = "Transition in the trace is not enabled in the concrete model."
                    if mode=='FEAS':
                        return False
                    trans_str = ""
                    if e.uml_trans.trigger:
                        trans_str += str(e.uml_trans.trigger.signal.name)+"("
                        trans_str += ", ".join([str(e.uml_trans.trigger.parameter[i].name)
                                                for i in range(0, len(e.uml_trans.trigger.parameter))])
                        trans_str += ")"
                    if e.uml_trans.guard:
                        assert e.uml_trans.guard.expression
                        assert e.uml_trans.guard.expression.body
                        trans_str += "['"+str(e.uml_trans.guard.expression.body) + "']"
                    if e.uml_trans.effect:
                        assert e.uml_trans.effect.script
                        assert e.uml_trans.effect.script.body
                        trans_str += "/'" + str(e.uml_trans.effect.script.body) + "'"
                    if e.uml_trans.trigger and e.uml_trans.trigger.signal.name[0] != '$' and \
                       (len(self.conc_ts.global_state.objects[e.sim_oid].sm_input_queue) == 0 or \
                       self.conc_ts.global_state.objects[e.sim_oid].sm_input_queue[0].signal != \
                       e.uml_trans.trigger.signal):
                        self._inform_result("Transition "+e.uml_trans.name+": "+trans_str+\
                                            "can not be executed "+\
                                            "in Object "+str(e.sim_oid)+" because the trigger does not receive correct signal.")
#                        error_msg = error_msg[:-1] + ", because the trigger does not receive correct signal."
                        self.abst_ts.process_next_chooses()
                        self._inform_activity("Calculating relevant variables")
                        ct = absttrace.Counterexample_Trace(self.conc_ts,
                                                            self.abst_ts,
                                                            "trigger")
                        ct.analyze_trigger_error()
                    else:
                        self._inform_result("Transition "+e.uml_trans.name+": "+trans_str+" can not be executed "+\
                                            "in Object "+str(e.sim_oid)+" because the guard"+\
                                            " evaluates false.")
#                        error_msg = error_msg[:-1] + ", because the guard is evaluated false."
                        self.abst_ts.process_next_chooses()
                        self._inform_activity("Calculating relevant variables")
                        ct = absttrace.Counterexample_Trace(self.conc_ts,
                                                            self.abst_ts,
                                                            "guard")
                        ct.analyze_guard_error()
                        #ct.analyze_error_point("guard")
                        #ct.make_slice2()
                    self._inform_relevant(ct)
                    #common steps for both cases
                    #print ct
                    if mode=='RELEVANT':
                        return False
                    self._inform_activity("Determining variables that are suggested to be refined")
                    ct.find_refinement_variables()
                    self._inform_refinement_hint(ct.get_refinement_hint())
                    return ct.get_refinement_hint()
                elif isinstance(e, tracesim.ExecutionError):
                    #print "The trace up to and including the current global state is:"
                    #self.conc_ts.sim.api_print_trace()
                    #print e
                    #print "Aborting!"

                    sim_error = e.sim_error

                    # is the error in the simulator an expected one or not
                    if not isinstance(sim_error, Sim.Error.ActionLanguageError):
                        self._inform_result("Internal error during transition execution"+\
                                            "Reason: unknown")
                        if mode in ('FEAS', 'RELEVANT'):
                            return True
                        else:
                            return None
                        #raise Internal_Error("Internal error during transition execution"+\
                        #                     "Reason: unknown")
                    if not isinstance(sim_error.original_error, actlang.errors.AssertionError_):
                        self._inform_result("Internal error during transition execution"+\
                                            "Reason: action language error"+\
                                            str(sim_error.original_error))
                        if mode in ('FEAS', 'RELEVANT'):
                            return True
                        else:
                            return None
                        #raise Internal_Error("Internal error during transition execution"+\
                        #                     "Reason: action language error"+\
                        #                     str(sim_error.original_error))

                    self._inform_result("True counterexample found: assertion in transition "+\
                                        e.uml_trans.name+" fails in the concrete model.")
                    if mode == 'FEAS':
                        return True
                    elif mode == 'RELEVANT':
                        return True
                    else:
                        return None
                         
                elif isinstance(e, tracesim.UnknownActionError):
                    raise Internal_Error(str(e))
                    #print e
                    #sys.exit(1)

            # increase counter
            i += 1

            # execute one step in the abstract simulator
            try:
                action = self.abst_ts.simulate_next_step()
                
            #for assertion errors
            except tracesim.ExecutionError, e:
                #check that everything is like it should be
                sim_error = e.sim_error
                if not isinstance(sim_error, Sim.Error.ActionLanguageError):
                    raise Internal_Error("Execution error in simulator is unknown type: "+str(sim_error))
                if not isinstance(sim_error.original_error, actlang.errors.AssertionError_):
                    raise Internal_Error("Actlang error unknown type: "+\
                                         sim_error.original_error.__class__.__name__)

#                error_msg = "Assertion failing in the abstract model, holds in the concrete one."

                self._inform_result("Assertion in Object "+str(e.sim_oid)+"'s transition "+e.uml_trans.name+" holds in the concrete model but does not hold in the abstract model.")
                
                if mode == 'FEAS':
                    return False
                
                # in the concrete world there has to be atleast one transition that
                # was succesfully executed because otherwise we would have a true counterexample
                if not (len(self.conc_ts.sim.trace.states) >= 2 or
                        len(self.conc_ts.sim.trace.actions) >= 2):
                    raise Internal_Error("Trace simulated in the concrete world produces "+ \
                                         "less than 2 states or actions (like empty trace would)")
                # find correct place in the concrete model
                while self.conc_ts.trace_index > self.abst_ts.trace_index:
                    self.conc_ts.backtrack_one_step()
                    self.conc_ts.process_next_chooses()

                # check that the next transition to be executed in both simulators is the same one
                (dummy, self.conc_sim_oid, self.conc_uml_trans) = self.conc_ts.get_next_action()
                self.abst_uml_trans = e.uml_trans
                self.abst_sim_oid = e.sim_oid
                assert self.conc_uml_trans.__XMIid__() == self.abst_uml_trans.__XMIid__()

                # counterexample analysis
                self._inform_activity("Calculating relevant variables")
                ct = absttrace.Counterexample_Trace(self.conc_ts, self.abst_ts, "effect")
                ct.analyze_assertion_error()
                #ct.analyze_error_point("assert")
                #ct.make_slice2()
                self._inform_relevant(ct)

                if mode == 'RELEVANT':
                    return False
                self._inform_activity("Determining variables that are suggested to be refined")
                ct.find_refinement_variables()
                self._inform_refinement_hint(ct.get_refinement_hint())
                return ct.get_refinement_hint()

            #for other errors, these should not happen unless something is wrong
            except tracesim.TraceSimulatorError, e:
                raise Internal_Error("Unexpected error in the simulation of the abstract "+\
                                     "counterexample: " + str(e))

        raise error.Error("Error in model checking procedure: "+\
                          "Abstract counterexample trace simulatable without errors")
        
if __name__ == '__main__':
    #if not (len(sys.argv) in (3,)):
    #    print "Usage: " + sys.argv[0] + " trace conc_model"
    #    sys.exit(1)

    usage_str = "[options] TRACE_FILE CONC_MODEL"
    general_help = """
    
Checks whether an abstract counterexample trace corresponds to a real
counterexample in the concrete model. If it does not, relevant variables
in the trace can be calculated. Relevant variables in a given point of
trace are a set of variables are most likely the source of the spurious
counterexample. Also a hint for abstraction refinement can be given
(is given by default). A refinement hint consist of variable XMIids and of
values that should be abstracted precisely in the abstract type the variable
is. This is primarily intented to be used with interval abstractions
(used in SMUML automatic abstraction refinement procedure).

arguments:
    TRACE_FILE          Abstract counterexample trace.
    CONC_MODEL          Concrete model."""
    
    op = optparse.OptionParser(usage="%prog "+usage_str+general_help,
                               version='SMUML Counterexample Analysator 1.0')

    op.add_option('-a', '--check-assert', action="store_true", default=False,
                  help="Analyze assertion errors in counterexamples.")
    op.add_option('-i', '--check-implicit-consumption', action="store_true", default=False,
                  help="Analyze implicit consumptions as errors in counterexamples.")
    op.add_option('-r', '--only-relevant-variables', action="store_const",
                  const="RELEVANT", dest="mode", default="FULL",
                  help="Calculates only relevant variables in the case of spurious counterexample, "+\
                  "does not suggest variables for refinement.")
    op.add_option('-f', '--only-feasibility', action="store_const", const="FEAS",
                  dest="mode", default="FULL",
                  help="Check only feasibility of a counterexample.")
    op.add_option('-q', '--quiet', action="store_const", const=0, dest="verbose", default=1,
                  help="Suppress all messages.")
    op.add_option('-v', '--verbose', action="store_const", const=2, dest="verbose", default=1,
                  help="Show verbose output. When this option is used, trace simulated step by step in the abstract and in the concrete model are shown with relevant variables marked. (Requires the calculation of relevant variables)")

    (opt, args) = op.parse_args()

    # make sure that there is correct number of arguments
    if len(args) != 2:
        print >> sys.stderr, "Wrong numger of arguments"
        print >> sys.stderr, "Usage: "+sys.argv[0]+" "+usage_str
        sys.exit(1)

    # check what kind of errors we are checking from the model
    error_types = []
    if opt.check_assert:
        error_types.append('ASSERT')
    if opt.check_implicit_consumption:
        error_types.append('IMPL')

    # files needed
    trace_filename = args[0]
    model_filename = args[1]

    try:
        P = coral.loadDocument(model_filename)
        ca = Counterexample_Analysator(trace_filename, None, P, error_types, verbose=opt.verbose)
        result = ca.analyze(opt.mode)
        if opt.mode in ('FEAS', 'RELEVANT'):
            if result[0]:
                print "True counterexample"
            else:
                print "Spurious counterexample"
            print "Reason: ",
            print result[1]
        elif opt.mode=='FULL':
            f = open("canal.out", "w")
            if isinstance(result, absttrace.Refinement_Hint):
                f.write(str(result))
                #sys.exit(1)
            elif result == None:
                f.write("true counterexample\n")
            else:
                raise Internal_Error("Counterexample_Analysator.analyze() returned an invalid type"+\
                                  "value: " + result.__class__.__name__)
            f.close()
        else:
            raise Internal_Error("variable 'opt.mode': unknown value")
        sys.exit(0)
    except error.Error, e:
        print >> sys.stderr, e
        sys.exit(1)
