# coding: latin-1

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Tools for analyzing concrete parse tree against corresponding
abstract parse tree.

Author: Vesa Ojala <vojala@tcs.tkk.fi>
"""

import sys
import string
import coral
import actlang
import Sim
import Sim.Core
import Sim.Object
import Utils.String
import Utils.Models
from Utils.Models import get_qualified_name

import Sim.GlobalState
import tracesim
from abstraction import typelibrary

import types
import slicer.refdef
import slicer.slicer

import abstraction.typeassociation

import utils

UML14 = coral.meta.UML14
DIA20 = coral.meta.XMIDI20

class CompResult:
    """Class used internally to represent comparison results
    between the abstract and the concrete parse tree.
    """
    def __init__(self, a=None, b=None):
        self.a = a
        self.b = b

    def __nonzero__(self):
        if isinstance(self.a, Sim.Object.Object):
            assert isinstance(self.b, Sim.Object.Object)
            return self.a.idnum == self.b.idnum

        return self.a == self.b

    def __eq__(self, other):
        return self.__nonzero__() == other
        
    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "  a: " + str(self.a) + "\n  b: " + str(self.b) + "\n"
        
class Abstraction_Comparer:
    """Class for comparing concrete parse trees against abstract parse trees.
    
    Following fields has to be set manually before using the object of
    this class:
        conc_sim      -- concrete world simulator
        abst_sim      -- abstract world simulator
        conc_sim_oid  -- object id indicating the object where evaluation
                         takes place in the concrete world
        abst_sim_oid  -- object id indicating the object where evaluation
                         takes place in the abstract world
        coercion      -- coercion function for getting the corresponding
                         abstract values from concrete values
        absdb         -- Absdb object containing needed information about
                         the abstract expressions corresponding to the
                         concrete ones.
    """
    def __init__(self):
        self.conc_sim = None
        self.abst_sim = None
        self.conc_sim_oid = None
        self.abst_sim_oid = None
        self.coercion = None
        self.absdb = None

    def collect_vars(self, tree, conc_obj, abst_obj):
        """Collects all the variable names in the tree.

        Arguments:
            tree      -- Jumbala parse tree
            conc_obj  -- Concrete object acting as tree's context
            abst_obj  -- Abstract object acting as tree's context

        Returns a set of (oid, v) pairs where oid is the simulator
        object id indicating the object that owns a variable named v.
        """
        varlist = utils.collect_relevant_var_names(tree, conc_obj, abst_obj)
        rval = []
        for o, v in varlist:
            xmiid = o.get_class_def()._variable_defs[v].uml_reference.__XMIid__()
            ai = self.absdb.get_abst_type(o.get_class_def()._variable_defs[v].uml_reference.__XMIid__())
            rval.append((o.idnum, v))
        return set(rval)

    def are_exprs_equal(self, con_tree):
        """Evaluates expression in the concrete world and a corresponding
        abstract expression in the abstract world and returns a
        CompResult object indicating whether the concrete value is coerced
        to the correct abstract value.
        
        Arguments:
            con_tree  -- Jumbala parse tree containing the concrete
                         expression.
        """
        debug = False
        if debug:
            print "Con vs abs equal begins!"

        conc_sim = self.conc_sim
        abst_sim = self.abst_sim

        # get copy of the global states (shoud the working state be restored after operations)
        old_con_gs = conc_sim.get_working_global_state().api_copy()
        old_abs_gs = abst_sim.get_working_global_state().api_copy()

        #debug printings
        if debug:
            print "concrete vars:"
            print utils.get_object(old_con_gs, self.conc_sim_oid).variable_values

            print "conc expr: "
            print con_tree
            print "abst type: ",
            print self.absdb.get_abst_expr(con_tree).typename

        expr_type = utils.determine_return_type(con_tree, conc_sim.get_working_global_state(),
                                                self.conc_sim_oid)

        # execute codes in simulators
        abst_expr = str(self.absdb.get_abst_expr(con_tree).expr)
        conc_value = conc_sim.execute_code_in_object(self.conc_sim_oid, "return " + str(con_tree) + ";", expr_type)
        abst_value = abst_sim.execute_code_in_object(self.abst_sim_oid, "return " + abst_expr + ";", expr_type)

        #debug printing
        if debug:
            print "concrete value: ",
            print conc_value
            print "abstract value: ",
            print abst_value

            print "concrete working state:"
            conc_sim.get_working_global_state().api_print()
            print "abstract working state:"
            abst_sim.get_working_global_state().api_print()

        # construction of return value
        rval = CompResult(conc_value.get_value(), abst_value.get_value())
        if isinstance(expr_type, Sim.Core.SimType_Int):
            if self.absdb.get_abst_expr(con_tree).typename != 'int':
                # coerce concrete value to corresponding abstract one for return value
                conc_value = self.coercion(self.absdb.get_abst_expr(con_tree).typename, conc_value.get_value())
                rc = conc_value.raw_contents
                # coercion should return an exact abstract value (len = 1)
                assert len(rc) == 1
                rval.a = rc[0]
                rval.b = abst_value.get_value()
                
        #debug printing
        if debug:
            print "rvalue:"
            print "  a: ",
            print rval.a
            print "  b: ",
            print rval.b

        return rval

    def find_vars(self, con_tree):
        """Finds a set of variables that most likely caused inaccurate
        evaluation in the abstract world. This method is intented to be used
        for finding initial relevant variables.

        Arguments:
            con_tree  -- Jumbala parse tree from the concrete world
            
        Returns a set of (oid, v) pairs where oid is the simulator
        object id indicating the object that owns a variable named v.
        """
        debug = False
        
        assert self.conc_sim_oid != None
        assert self.abst_sim_oid != None
        
        conc_sim = self.conc_sim
        abst_sim = self.abst_sim

        #debug printing
        if debug:
            print "Find vars begins"
            print con_tree.__class__.__name__

        #old_con_gs = con_sim.get_working_global_state().api_copy()
        #old_abs_gs = abs_sim.get_working_global_state().api_copy()

        # assert that there is difference in how abstract and concrete values are evaluated
        # otherwise shouldn't be here
        foo = self.are_exprs_equal(con_tree)
        if debug:
            print foo
        assert not foo

        # if the abstract expression is flattened we can't process
        # further so just collect the variables right away
        if self.absdb.get_abst_expr(con_tree).flattened:
            return self.collect_vars(con_tree,
                                     utils.get_object(conc_sim.get_working_global_state(),
                                                      self.conc_sim_oid),
                                     utils.get_object(abst_sim.get_working_global_state(),
                                                      self.abst_sim_oid))
        
    #    assert not are_exprs_equal(con_sim, abs_sim, con_sim_oid, abs_sim_oid, \
    #                                con_tree, abs_map, coercion)

        # process each actlang-tree-type appropriately
        if isinstance(con_tree, actlang.statement.SendStatement):
            # not implemented yet
            assert False
            arglst = con_tree.api_getArgList()
            assert len(arglst)+1 == len(asses)

            for i, arg in enumerate(arglst):
                map_to_tree(arg, asses[i], dic)
            map_to_tree(con_tree.api_getTarget(), asses[-1], dic)
        elif isinstance(con_tree, actlang.expr.SimpleAssignmentExpr):
            # not implemented yet
            assert False
            assert len(asses) == 2
            map_to_tree(con_tree.api_getLhsExpr(), asses[0], dic)
            map_to_tree(con_tree.api_getRhsExpr(), asses[1], dic)
        elif isinstance(con_tree, actlang.expr.ParenExpr):
            return self.find_vars(con_tree.api_getInnerExpr())
        elif isinstance(con_tree, (actlang.expr.CondAndExpr,
                                   actlang.expr.CondOrExpr)):
            #print con_tree
            a = con_tree.api_getLeftOperand()
            b = con_tree.api_getRightOperand()
            # evaluation takes account short-circuit evaluation
            result_a = self.are_exprs_equal(a)
            #print result_a
            if result_a == False:
                #print "if"
                #print bool(result_a)
                #print result_a.__nonzero__()
                #print result_a.a
                #print result_a.b
                #print result_a.a == result_a.b
                return self.find_vars(a)
            else:
                result_b = self.are_exprs_equal(b)
                if result_b == False:
                    #print "else"
                    return self.find_vars(b)
                else:
                    assert False
        elif isinstance(con_tree, actlang.expr.InfixExpr):
            a = con_tree.api_getLeftOperand()
            b = con_tree.api_getRightOperand()

            result_a = self.are_exprs_equal(a)
            result_b = self.are_exprs_equal(b)
            if result_a == True and result_b == True:
                return self.collect_vars(con_tree,
                                         utils.get_object(conc_sim.get_working_global_state(), self.conc_sim_oid),
                                         utils.get_object(abst_sim.get_working_global_state(), self.abst_sim_oid))
            elif result_a == False and result_b == False:
                rst = self.find_vars(a)
                rst.update(self.find_vars(b))
                return rst
            else:
                if result_a == False:
                    return self.find_vars(a)
                elif result_b == False:
                    return self.find_vars(b)
                else:
                    assert False
        elif isinstance(con_tree, (actlang.expr.LogicalNotExpr,
                               actlang.expr.BitNotExpr,
                               actlang.expr.UnaryPlusExpr,
                               actlang.expr.UnaryMinusExpr)):
            return self.find_vars(con_tree.api_getOperand())
        elif isinstance(con_tree, actlang.expr.TernaryCondExpr):
            c = con_tree.api_getCondition()
            a = con_tree.api_getTrueExpr()
            b = con_tree.api_getFalseExpr()
            result_c = self.are_exprs_equal(c)
            if result_c == True:
                if result_c.a == True:
                    return self.find_vars(a)
                else:
                    return self.find_vars(b)
            else:
                return self.find_vars(c)        
        elif isinstance(con_tree, (actlang.expr.IntLiteralExpr,
                            actlang.expr.BoolLiteralExpr,
                            actlang.expr.ThisExpr)):
            return
        elif isinstance(con_tree, (actlang.names.SimpleName,
                                   actlang.names.QualifiedName,
                                   actlang.expr.PrimaryFieldAccessExpr)):
            return self.collect_vars(con_tree,
                                     utils.get_object(conc_sim.get_working_global_state(), self.conc_sim_oid),
                                     utils.get_object(abst_sim.get_working_global_state(), self.abst_sim_oid))
        elif isinstance(con_tree, actlang.expr.InstanceCreationExpr):
            raise Error()
        else:
            print con_tree.__class__.__name__
            raise Error()

    def find_abst_error(self, con_tree):
        """Finds first subexpression that is evaluated differently in the
        concrete and in the abstract world. Evaluating differently means
        here that when the concrete subexpression is evaluated, the result
        is coerced to the abstract type of the corresponding abstract
        subexpression, and if the coerced value and the value of the abstract
        subexpression evaluated in the abstract world are different, then
        the subexpression evaluated differently in the oncrete and in the
        abstract world.

        Arguments:
            con_tree  -- Jumbala parse tree from the concrete world.

        Returns a concrete parse tree containing the first differently
        evaluating subexpression.
        """
        # returns tree
        debug = False
        
        assert self.conc_sim_oid != None
        assert self.abst_sim_oid != None
        assert self.conc_sim != None
        assert self.abst_sim != None
        
        conc_sim = self.conc_sim
        abst_sim = self.abst_sim

        #debug printing
        if debug:
            print "Find vars begins"
            print con_tree.__class__.__name__

        #old_con_gs = con_sim.get_working_global_state().api_copy()
        #old_abs_gs = abs_sim.get_working_global_state().api_copy()

        # assert that there is difference in how abstract and concrete values are evaluated
        # otherwise shouldn't be here
        foo = self.are_exprs_equal(con_tree)
        if debug:
            print con_tree
            print foo
        assert not foo

        # if the abstract expression is flattened we can't process
        # further so just return the whole tree
        if self.absdb.get_abst_expr(con_tree).flattened:
            return con_tree

    #    assert not are_exprs_equal(con_sim, abs_sim, con_sim_oid, abs_sim_oid, \
    #                                con_tree, abs_map, coercion)

        # process each actlang-tree-type appropriately
        if isinstance(con_tree, actlang.statement.SendStatement):
            # not implemented yet
            assert False
        elif isinstance(con_tree, actlang.expr.SimpleAssignmentExpr):
            # not implemented yet
            assert False
        elif isinstance(con_tree, actlang.expr.ParenExpr):
            return self.find_abst_error(con_tree.api_getInnerExpr())
        elif isinstance(con_tree, (actlang.expr.CondAndExpr,
                                   actlang.expr.CondOrExpr)):
            a = con_tree.api_getLeftOperand()
            b = con_tree.api_getRightOperand()
            # evaluation takes account short-circuit evaluation
            result_a = self.are_exprs_equal(a)
            if result_a == False:
                return self.find_abst_error(a)
            else:
                result_b = self.are_exprs_equal(b)
                if result_b == False:
                    return self.find_abst_error(b)
                else:
                    assert False
        elif isinstance(con_tree, actlang.expr.InfixExpr):
            a = con_tree.api_getLeftOperand()
            b = con_tree.api_getRightOperand()

            result_a = self.are_exprs_equal(a)
            result_b = self.are_exprs_equal(b)
            if result_a == True and result_b == True:
                return con_tree
                #self.result.update(self.collect_vars(con_tree,
                #              conc_sim.get_working_global_state().get_object(self.conc_sim_oid)))
            elif result_a == False and result_b == False:
                return self.find_abst_error(a)
                #self.find_abst_error(b)
            else:
                if result_a == False:
                    return self.find_abst_error(a)
                elif result_b == False:
                    return self.find_abst_error(b)
                else:
                    assert False
        elif isinstance(con_tree, (actlang.expr.LogicalNotExpr,
                               actlang.expr.BitNotExpr,
                               actlang.expr.UnaryPlusExpr,
                               actlang.expr.UnaryMinusExpr)):
            return self.find_abst_error(con_tree.api_getOperand())
        elif isinstance(con_tree, actlang.expr.TernaryCondExpr):
            c = con_tree.api_getCondition()
            a = con_tree.api_getTrueExpr()
            b = con_tree.api_getFalseExpr()
            result_c = self.are_exprs_equal(c)
            if result_c == True:
                if result_c.a == True:
                    return self.find_abst_error(a)
                else:
                    return self.find_abst_error(b)
            else:
                return self.find_abst_error(c)        
        elif isinstance(con_tree, (actlang.expr.IntLiteralExpr,
                            actlang.expr.BoolLiteralExpr,
                            actlang.expr.ThisExpr)):
            return
        elif isinstance(con_tree, (actlang.names.SimpleName,
                                   actlang.names.QualifiedName,
                                   actlang.expr.PrimaryFieldAccessExpr)):
            #!!!
            return con_tree
            #self.result.update(self.collect_vars(con_tree,
            #                               conc_sim.get_working_global_state().get_object(self.conc_sim_oid)))
        elif isinstance(con_tree, actlang.expr.InstanceCreationExpr):
            raise Error()
        else:
            print con_tree.__class__.__name__
            raise Error()

