# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Tools for analyzing abstract counterexample traces (calculation of
relevant variables and refinement hint)

Author: Vesa Ojala <vojala@tcs.tkk.fi>
"""

import sys
import string
import coral
import actlang
import Sim
import Sim.Core
import Sim.Object
import Utils.String
import Utils.Models
from Utils.Models import get_qualified_name

import Sim.GlobalState
import tracesim
from abstraction import typelibrary

import types
import slicer.refdef
import slicer.slicer

import abstraction.typeassociation

from canal import abstcomp
from canal import utils
from canal import events

import error

UML14 = coral.meta.UML14
DIA20 = coral.meta.XMIDI20

class Absdb:
    """Abstraction database for getting the name of the abstract datatype
    by giving the xmiid of the variable.
    """
    def __init__(self, abst_model, conc_code_map):
        """Initializes abstraction database.

        Arguments:
            filename  -- Name of the file containing the type information.

        Returns initialixed abstraction database.
        """
        self.ta = abstraction.typeassociation.TypeAssociation()
        self.ta.initFromModel(abst_model, conc_code_map)
        self.libfiles = self.ta.libfiles
        self.m = self.ta.constraints

    def get_abst_type(self, xmiid):
        """Gives the name of the abstract type of a variable given.

        Arguments:
            xmiid  -- xmiid of the variable.

        Returns the name of the abstract type as a string.
        """
        if not self.m.has_key(xmiid):
            return None
        lst = self.m[xmiid]
        assert len(lst) == 1
        (rel, typ, line) = lst[0]
        assert rel == "=="
        return typ

    def get_abst_expr(self, conc_expr):
        """Gives abstract expression corresponding the concrete expression
        given.

        Arguments:
            conc_expr  -- Concrete expression as a Jumbala parse tree.

        Returns a data structure containing fields:
            expr       an actlang.expr.Expr object representing an
                       abstracted action language expression,
            typename   the name of the abstract type of the abstracted
                       expression, and
            flattened  a truth value telling whether the abstracted
                       expression is "flattened" or not.
        """
        if not self.ta.expr_correspondence.has_key(conc_expr):
            print conc_expr
        assert self.ta.expr_correspondence.has_key(conc_expr)
        data = self.ta.expr_correspondence[conc_expr]
        return data

class Counterexample_Trace:
    """Class for representing and analysing an abstract counterexample trace.
    """
    def __init__(self, conc_ts, abst_ts, mode):
        """Initializes Counterexample_Trace ojbect.

        Arguments:
            conc_ts  -- Trace_Simulator object used for simulating the
                        counterexample in the concrete world.
            abst_ts  -- Trace_Simulator object used for simulating the
                        counterexample in the abstract world.
            mode     -- should be one of the following strings:
                          guard   - guard is evaluated false in the concrete
                                    world but true in the abstract world
                          effect  - reason why transition could not be
                                    executed in the concrete world lies
                                    in the effect of the transition. For
                                    example assertion could not be executed.
                          trigger - trigger gets an incorrect message
        """
        assert (mode == "guard" or mode == "effect" or mode == "trigger")
        assert conc_ts.trace_index == abst_ts.trace_index

        self.debug = False
        self.events = []
        self.conc_ts = conc_ts
        self.abst_ts = abst_ts
        self.parse_tree_map = {}
        self.abst_sendermap = {}
        self.conc_sendermap = {}

        # collect transition lists
        conc_transition_list = []
        conc_statemachine_list = []
        self.conc_transition_map = {}
        slicer.slicer.collect_elements(conc_ts.model, UML14.StateMachine, conc_statemachine_list)
        for sm in conc_statemachine_list:
            conc_transition_list.extend(sm.transitions)
        for trans in conc_transition_list:
            self.conc_transition_map[trans.__XMIid__()] = trans
        abst_transition_list = []
        abst_statemachine_list = []
        self.abst_transition_map = {}
        slicer.slicer.collect_elements(abst_ts.model, UML14.StateMachine, abst_statemachine_list)
        for sm in abst_statemachine_list:
            abst_transition_list.extend(sm.transitions)
        for trans in abst_transition_list:
            self.abst_transition_map[trans.__XMIid__()] = trans
        if self.debug:
            print "Counterexample_Trace:"
            print "  transition_list"
            print conc_transition_list
            print "  transition_map"
            print conc_transition_map
            
        # process all trace actions before erroneous place
        # ('GuardChooseValue')
        # ('EffectChooseValue')
        # ('EXEC', sim_oid, uml_trans)
        # ('DEFER', sim_oid)
        # ('IMPL', sim_oid)

        #parse all jumbala stuff in the trace
        statement_map = {}
        for conc_uml_trans in self.conc_transition_map.values():
            if conc_uml_trans.guard:
                guard_str = str(conc_uml_trans.guard.expression.body)
                foo = actlang.grammar.parse("assert (" + guard_str + ");", generateParenExpr=False)
                stmts = foo.api_getTopStatements()
                statement_map[conc_uml_trans.guard.__XMIid__()] = stmts
                self.parse_tree_map[conc_uml_trans.guard.__XMIid__()] = stmts[0].api_getExpr()
            if conc_uml_trans.effect:
                prog = conc_uml_trans.effect.script.body
                treelist = actlang.grammar.parse(prog, generateParenExpr=False).api_getTopStatements()
                statement_map[conc_uml_trans.effect.__XMIid__()] = treelist
                self.parse_tree_map[conc_uml_trans.effect.__XMIid__()] = treelist

        self.absdb = Absdb(self.abst_ts.model, statement_map)
        self.typelibs = self.absdb.libfiles
        
        #backup current state of simulators
        bck_conc_gs = conc_ts.sim.get_working_global_state()
        bck_abst_gs = abst_ts.sim.get_working_global_state()
        bck_conc_chooses = conc_ts.get_choose_maps()
        bck_abst_chooses = abst_ts.get_choose_maps()
        gs_index = 0
        gs_conc = conc_ts.sim.trace.states[gs_index].api_copy()
        gs_abst = abst_ts.sim.trace.states[gs_index].api_copy()
        conc_ts.sim.set_working_global_state(gs_conc.api_copy())
        abst_ts.sim.set_working_global_state(gs_abst.api_copy())

        for action in conc_ts.trace[:conc_ts.trace_index]:
            if action[0] in ('GuardChooseValue', 'EffectChooseValue'):
                abst_ts.process_choose_list([action])
                
                #if event.chooses == None:
                #    event.chooses = []
                #event.chooses.append(action)
            elif action[0] == 'EXEC':
                # init simulators
                self.conc_ts.sim.set_working_global_state(gs_conc.api_copy())
                self.abst_ts.sim.set_working_global_state(gs_abst.api_copy())

                event = events.Counterexample_Exec(self, action)
                self.events.append(event)
                
                gs_index += 1
                gs_conc = conc_ts.sim.get_working_global_state().api_copy()
                gs_abst = abst_ts.sim.get_working_global_state().api_copy()
                #update statemachine configuration appropriately
                for i in range(conc_ts.sim.trace.states[gs_index].get_nof_objects()):
                    o = conc_ts.sim.trace.states[gs_index].get_object(i)
                    c = o.get_state_machine_configuration()
                    gs_conc.get_object(i).set_state_machine_configuration(c)
                for i in range(abst_ts.sim.trace.states[gs_index].get_nof_objects()):
                    o = abst_ts.sim.trace.states[gs_index].get_object(i)
                    c = o.get_state_machine_configuration()
                    gs_abst.get_object(i).set_state_machine_configuration(c)
                
                #assert that transition executed piece by piece had the same result
                #as the transition execution in the trace
                assert gs_conc.compare(conc_ts.sim.trace.states[gs_index]) == 0
                assert gs_abst.compare(abst_ts.sim.trace.states[gs_index]) == 0
            elif action[0] == 'IMPL':
                # init simulators
                self.conc_ts.sim.set_working_global_state(gs_conc.api_copy())
                self.abst_ts.sim.set_working_global_state(gs_abst.api_copy())

                event = events.Counterexample_Impl(self, action)
                self.events.append(event)
                
                gs_index += 1
                gs_conc = conc_ts.sim.get_working_global_state().api_copy()
                gs_abst = abst_ts.sim.get_working_global_state().api_copy()
                
                #assert that transition executed piece by piece had the same result
                assert gs_conc.compare(conc_ts.sim.trace.states[gs_index]) == 0
                assert gs_abst.compare(abst_ts.sim.trace.states[gs_index]) == 0
            elif action[0] in ('DEFER'):
                # init simulators
                self.conc_ts.sim.set_working_global_state(gs_conc.api_copy())
                self.abst_ts.sim.set_working_global_state(gs_abst.api_copy())

                event = events.Counterexample_Defer(self, action)
                self.events.append(event)
                
                gs_index += 1
                gs_conc = conc_ts.sim.get_working_global_state().api_copy()
                gs_abst = abst_ts.sim.get_working_global_state().api_copy()
                
                #assert that transition executed piece by piece had the same result
                assert gs_conc.compare(conc_ts.sim.trace.states[gs_index]) == 0
                assert gs_abst.compare(abst_ts.sim.trace.states[gs_index]) == 0
            else:
                raise error.Error("Unsupported feature")

        # process the last action, the one where the error occured
        action = conc_ts.trace[conc_ts.trace_index]
        self.conc_ts.sim.set_working_global_state(gs_conc.api_copy())
        self.abst_ts.sim.set_working_global_state(gs_abst.api_copy())

        if action[0] == 'EXEC':        
            event = events.Counterexample_Exec(self, action, ("limit", mode))
        elif action[0] == 'IMPL':
            event = events.Counterexample_Impl(self, action, ("limit", mode))
        elif action[0] == 'DEFER':
            event = events.Counterexample_Defer(self, action, ("limit", mode))
        self.events.append(event)

        #restore states of the simulators
        conc_ts.sim.set_working_global_state(bck_conc_gs)
        abst_ts.sim.set_working_global_state(bck_abst_gs)
        conc_ts.set_choose_maps(bck_conc_chooses)
        abst_ts.set_choose_maps(bck_abst_chooses)

        #init typelibrary
        self._init_typelibrary()
        
    def _init_typelibrary(self):
        """Initializes TypeLibrary object without generating python classes
        used for some functionality in TypeLibrary(). Those classes are not
        needed for functionality Counterexample_Trace uses.
        """
        assert self.typelibs != None
        tl = typelibrary.TypeLibrary()
        tl.read(self.typelibs)
        self.typelibrary = tl
        self.coercion = lambda t,v: self.typelibrary.applyCoercion("int", t, v)

    def _init_typelibrary_with_class_generation(self):
        """Initializes TypeLibrary object with generating python classes
        used for some functionality in TypeLibrary().
        """
        assert self.typelibs != None
        tl = typelibrary.TypeLibrary()
        tl.read(self.typelibs)
        tl.generatePythonClasses()
        self.typelibrary = tl
        self.coercion = typelibrary.coercion

    def get_refinement_hint(self):
        """Returns Refinement_Hint object indicating the last suggested
        refinement. Method find_refinement_variables has to be
        executed prior execution of this method.
        """
        return self.rh

    def analyze_guard_error(self):
        """Calculates heuristically initial relevant variables at the
        error point and based on those calculates relevant variables
        for remaining points in the trace. This function is used when
        the error has happened in the guard of transition.
        """
        event, item = self._get_last_item()
        if not isinstance(item, events.Counterexample_Guard):
            print item.__class__.__name__
        assert isinstance(item, events.Counterexample_Guard)
        tree = item.guard
        crit = self._find_initial_relevant_variables(item, tree)
        self._calculate_relevant_variables(crit)

    def analyze_assertion_error(self):
        """Calculates heuristically initial relevant variables at the
        error point and based on those calculates relevant variables
        for remaining points in the trace. This function is used when
        the error has happened in the effect(assertion) of transition.
        """
        event, item = self._get_last_item()
        assert isinstance(item, events.Counterexample_Effect)
        assert isinstance(item.code, actlang.statement.AssertStatement)
        tree = item.code.api_getExpr()
        crit = self._find_initial_relevant_variables(item, tree)
        self._calculate_relevant_variables(crit)
    
    def analyze_trigger_error(self):
        """Finds the first place where the messages are send to different
        targets in the abstract and in the concrete world so that in one
        of these worlds target was the object in which the trigger is
        enabled in one world but not enabled in the other. Based on the
        target expression in the differenting send statement calculates 
        heuristically initial relevant variables which are used for
        calculating relevant variables at other points of the
        countrerexample. This function is used when the error has happened
        in the trigger of transition.
        """
        event, item = self._get_last_item()
        items = self.get_item_list()
        conc_target = item.parent.conc_sim_oid
        abst_target = item.parent.abst_sim_oid
        abst_obj = utils.get_object(item.abst_state, abst_target)
        abst_msg = abst_obj.sm_input_queue[0]
        ce_send = self.abst_sendermap[abst_msg.idnum]
        assert ce_send == self.abst_sendermap[abst_msg.idnum]
        index = self._check_signal_correspondence(ce_send, conc_target, abst_target)
        assert index != -1
        # if signal send targets didn't correspond
        if index >= 0:
            item = items[index]
            assert isinstance(item, events.Counterexample_Send)
            crit = self._find_initial_relevant_variables(item, item.target_expr)
            self._calculate_relevant_variables(crit, index)
        else:
            raise error.Error("Invalid trigger error: No signal send targets "+\
                              "that did not correspond to each other")
        
    def _find_initial_relevant_variables(self, item, tree):
        """Finds relevant variables in the item given based on assumption
        that the item represents the error point.

        Arguments:
            item  -- Counterexample_Item representing the error point.
            tree  -- tree from which the variables are extracted.

        Returns a set of <oid, name> pairs where oid is the object id where
        the relevant variable is located and name is the name of the variable.
        """
        # backup simulator states
        bck_conc_chooses = self.conc_ts.get_choose_maps()
        bck_abst_chooses = self.abst_ts.get_choose_maps()
        bck_conc_gs = self.conc_ts.sim.get_working_global_state().api_copy()
        bck_abst_gs = self.abst_ts.sim.get_working_global_state().api_copy()

        # init abstraction comparer
        ac = abstcomp.Abstraction_Comparer()
        ac.coercion = self.coercion
        ac.absdb = self.absdb
        # set up simulator for comparing
        self.conc_ts.sim.set_working_global_state(item.conc_state.api_copy())
        self.abst_ts.sim.set_working_global_state(item.abst_state.api_copy())
        conc_sim = self.conc_ts.sim
        abst_sim = self.abst_ts.sim
        ac.conc_sim = conc_sim
        ac.abst_sim = abst_sim
        ac.conc_sim_oid = item.parent.conc_sim_oid
        ac.abst_sim_oid = item.parent.abst_sim_oid
        self.abst_ts.set_choose_maps(item.parent.chooses)
        
        # compare
        #print tree
        crit = ac.find_vars(tree)
        #print self.crit
        
        # revert old simulator states
        self.conc_ts.sim.set_working_global_state(bck_conc_gs)
        self.abst_ts.sim.set_working_global_state(bck_abst_gs)        
        self.conc_ts.set_choose_maps(bck_conc_chooses)
        self.abst_ts.set_choose_maps(bck_abst_chooses)

        return crit

    def find_refinement_variables(self):
        """Finds a suggestion of variables to be refined. Error has to be
        analyzed before using this function (one of the functions with
        analyze_ prefix executed before the usage of this function).

        Returns an Refinement_Hint object representing the refinement
        suggested.
        """
        ac = abstcomp.Abstraction_Comparer()
        ac.absdb = self.absdb
        ac.coercion = self.coercion
        bck_conc_chooses = self.conc_ts.get_choose_maps()
        bck_abst_chooses = self.abst_ts.get_choose_maps()
        bck_conc_gs = self.conc_ts.sim.get_working_global_state().api_copy()
        bck_abst_gs = self.abst_ts.sim.get_working_global_state().api_copy()
        event, item = self._find_first_abst_error()
        self.conc_ts.sim.set_working_global_state(item.conc_state.api_copy())
        self.abst_ts.sim.set_working_global_state(item.abst_state.api_copy())
        conc_sim = self.conc_ts.sim
        abst_sim = self.abst_ts.sim
        ac.conc_sim = conc_sim
        ac.abst_sim = abst_sim
        ac.conc_sim_oid = event.conc_sim_oid
        ac.abst_sim_oid = event.abst_sim_oid
        #self.abst_ts.process_choose_list(event.chooses)
        self.abst_ts.set_choose_maps(event.chooses)
        if isinstance(item, events.Counterexample_Send):
            if item.conc_target.idnum != item.abst_target.idnum:
                expr = item.code.api_getTarget()
            else:
                for arg in item.arglist:
                    result = ac.are_exprs_equal(arg)
                    if not result:
                        expr = arg
                        break

            error_tree = ac.find_abst_error(expr)
            se = Subexpression_Evaluator(conc_sim, event.conc_sim_oid, error_tree, self.absdb)
            se.collect()
            # init structure for refinement hint
            self.rh = Refinement_Hint()
            self.rh.init_basic(se.names, se.values)
            
        elif isinstance(item, events.Counterexample_Effect):
            if isinstance(item.code, actlang.statement.ExprStatement):
                expr = item.code.api_getExpr()
                if isinstance(expr, actlang.expr.SimpleAssignmentExpr):
                    # analyze lhs so we can add it to the variable set got from rhs
                    lhs_o = utils.get_object(conc_sim.get_working_global_state(), event.conc_sim_oid)
                    lhs_o, lhs_v = utils.get_obj_name_pair(expr.api_getLhsExpr(), lhs_o)
                    xmiid = lhs_o.get_class_def()._variable_defs[lhs_v].uml_reference.__XMIid__()
                    ai = self.absdb.get_abst_type(xmiid)
                    if ai != None and ai not in ("int", "boolean"):
                        lhs_set = set([xmiid])
                    else:
                        lhs_set = set()
                    # process rhs
                    #print self.abst_ts.get_choose_maps()
                    error_tree = ac.find_abst_error(expr.api_getRhsExpr())
                    se = Subexpression_Evaluator(conc_sim, event.conc_sim_oid, error_tree, self.absdb)
                    se.collect()
                    # init structure for refinement hint
                    self.rh = Refinement_Hint()
                    self.rh.init_basic(se.names | lhs_set, se.values)
                else:
                    raise error.Error()
            elif isinstance(item.code, actlang.statement.AssertStatement):
                expr = item.code.api_getExpr()
                error_tree = ac.find_abst_error(expr)
                se = Subexpression_Evaluator(conc_sim, event.conc_sim_oid, error_tree, self.absdb)
                se.collect()
                # init structure for refinement hint
                self.rh = Refinement_Hint()
                self.rh.init_basic(se.names, se.values)
                #print self.rh
            else:
                raise error.Error()
        elif isinstance(item, events.Counterexample_Guard):
            #print "ca_guard"
            expr = item.guard
            error_tree = ac.find_abst_error(expr)
            se = Subexpression_Evaluator(conc_sim, event.conc_sim_oid, error_tree, self.absdb)
            se.collect()
            # init structure for refinement hint
            self.rh = Refinement_Hint()
            self.rh.init_basic(se.names, se.values)
            #print self.rh            
        else:
            raise error.Error()
        self.conc_ts.sim.set_working_global_state(bck_conc_gs)
        self.abst_ts.sim.set_working_global_state(bck_abst_gs)
        self.conc_ts.set_choose_maps(bck_conc_chooses)
        self.abst_ts.set_choose_maps(bck_abst_chooses)

        return self.rh

    def _get_last_item(self):
        """Returns an Counterexample_Item object representing the last
        action in the trace.
        """
        lst = self.get_item_list()
        item = lst[-1]
        return (item.parent, item)

    def _find_first_abst_error(self):
        """Finds first point in the counterexample trace where some
        relevant variable is abstracted incorrectly.

        Returns a pair (event, item) where item is a Counterexample_Item
        where the first incorrectly abstracted relevant variable locates
        and event is a Counterexample_Event which owns the item.
        """
        last_relevant_event = None
        last_relevant_item = None
        previous_event = None
        previous_item = None

        items = self.get_item_list()
        for item in items:
            rval = self.__check_global_state_abst_error(item)
            if rval:
                return (previous_event, previous_item)
            if self._has_refinable_variables(item):
                last_relevant_event = item.get_event()
                last_relevant_item = item
            previous_event = item.get_event()
            previous_item = item
        if last_relevant_event == None or last_relevant_item == None:
            raise error.Error("There are no relevant variables in the trace")
        return (last_relevant_event, last_relevant_item)

    # NOTE! this method relies on the fact that message parameter and attribute where the
    # parameter is assigned in the trigger has to have a same abstract type.
    # If the condition above does not hold, then message parameters has to be taken account here
    def _has_refinable_variables(self, item):
        """Checks whether Counterexample_Item has variables that are both refinable and relevant.

        Arguments:
            item  -- item to be checked

        Returns boolean value as an result.
        """
        for key, value in item.slice_map.items():
            if value:
                if len(key) == 2:
                    idnum, vname = key
                    obj = utils.get_object(item.conc_state, idnum)
                    xmiid = obj.get_class_def()._variable_defs[vname].uml_reference.__XMIid__()
                    t = self.absdb.get_abst_type(xmiid)
                    if t != None and t not in ("int", "boolean"):
                        return True
                if len(key) == 3:
                    #signal parameters would be handled here, see comment above
                    pass
        return False

#    def get_refinement_hint(self, tree, obj):
#        varlist = self.collect_vars(tree, obj)
#        
    def __check_global_state_abst_error(self, item):
        """Checks whether the abstract state corresponds to a concrete state in a given item.

        Arguments:
            item  -- Counterexaple_Item to be checked.
        """
        conc_gs = item.conc_state
        abst_gs = item.abst_state
        mask = item.slice_map
        for conc_o, abst_o in zip(conc_gs.objects, abst_gs.objects):
            assert conc_o.idnum == abst_o.idnum
            conc_items = conc_o.variable_values.iteritems()
            abst_items = abst_o.variable_values.iteritems()
            assert len(conc_items) == len(abst_items)
            for n, cv in conc_items:
                if mask.has_key((conc_o.idnum, n)) and mask[(conc_o.idnum, n)]:
                    av = abst_o.variable_values[n]
                    xmiid = conc_o.get_class_def()._variable_defs[n].uml_reference.__XMIid__()
                    abst_type = self.absdb.get_abst_type(xmiid)
                    if abst_type and abst_type not in ("int", "boolean"):
                        rc = self.coercion(abst_type, cv.get_value()).raw_contents
                        # coercion should return an exact abstract value (len = 1)
                        assert len(rc) == 1
                        if av.get_value() != rc[0]:
                            return True
                    elif abst_type == None:
                        if cv.get_value() == "null" and av.get_value == "null":
                            pass
                        elif ((cv.get_value() == "null" and av.get_value != "null") or
                              (cv.get_value() != "null" and av.get_value == "null") or 
                              cv.get_value().idnum != av.get_value().idnum):
                            return True
                    else:
                        if cv.get_value().idnum != av.get_value().idnum:
                            return True
            for i, conc_msg in enumerate(conc_o.sm_input_queue):
                for j,p in enumerate(conc_msg.parameters):
                    if mask.has_key(("conc", conc_msg.idnum, i)) and mask[("conc", conc_msg.idnum, i)]:
                        abst_msg = abst_o.sm_input_queue[i]
                        # conc queue and abst queue has to match each other because otherwise
                        # we would have found the reason already (signal correspondence check)
                        #assert conc_msg.idnum == abst_msg.idnum
                        # for some reason corresponding signals do not have always the same idnum
                        cv = p
                        av = abst_msg.get_parameter(j)
                        uml_param = conc_msg.get_signal().feature[j]
                        assert uml_param.oclIsKindOf(UML14.Attribute)
                        xmiid = uml_param.__XMIid__()
                        abst_type = self.absdb.get_abst_type(xmiid)
                        if abst_type and abst_type not in ("int", "boolean"):
                            rc = self.coercion(abst_type, cv.get_value()).raw_contents
                            # coercion should return an exact abstract value (len = 1)
                            assert len(rc) == 1
                            if av.get_value() != rc[0]:
                                return True                        
                        elif abst_type == None:
                            if cv.get_value() == "null" and av.get_value == "null":
                                pass
                            elif ((cv.get_value() == "null" and av.get_value != "null") or
                                  (cv.get_value() != "null" and av.get_value == "null") or 
                                  cv.get_value().idnum != av.get_value().idnum):
                                return True
                        else:
                            if cv.get_value().idnum != av.get_value().idnum:
                                return True
        return False
        
    def __slice_step(self, oldmap, item):
        """Calculates relevant variables at the point just before
        execution of item.

        Arguments:
            oldmap  -- slice_map of the successor of item. (dictionary)
            item    -- Counterexample_Item to which slice_map we are
                       calculating.
        """
        objects = item.conc_state.objects
        newmap = item.slice_map
        exec_obj = utils.get_object(item.conc_state, item.parent.conc_sim_oid)
        if isinstance(item, events.Counterexample_Effect):
            code = item.code
        elif isinstance(item, events.Counterexample_Guard):
            code = item.guard

        # copy old values
        for obj in objects:
            for vname in obj.variable_values.keys():
                if oldmap.has_key((obj.idnum, vname)):
                    newmap[(obj.idnum, vname)] = oldmap[(obj.idnum, vname)]
                else:
                    if self.debug:
                        print "cannot copy slicemap for key: ("+obj.idnum+","+vname+")"
                    newmap[(obj.idnum, vname)] = False
            for msg in obj.sm_input_queue:
                for i,p in enumerate(msg.parameters):
                    if oldmap.has_key(("conc",msg.idnum,i)):
                        newmap[("conc",msg.idnum,i)] = oldmap[("conc",msg.idnum,i)]
                    else:
                        newmap[("conc",msg.idnum,i)] = False
        for obj in item.abst_state.objects:
            for msg in obj.sm_input_queue:
                for i,p in enumerate(msg.parameters):
                    if oldmap.has_key(("abst",msg.idnum,i)):
                        newmap[("abst",msg.idnum,i)] = oldmap[("abst",msg.idnum,i)]
                    else:
                        newmap[("abst",msg.idnum,i)] = False

        # update according to the new situation
        if isinstance(item, events.Counterexample_Trigger):
            rval = False
            if item.signal_name[0] == '$':
                return rval
            obj = utils.get_object(item.conc_state, item.parent.conc_sim_oid)
            msg = obj.sm_input_queue[0]
            for i,a in enumerate(item.assignments):
                if oldmap[(obj.idnum, a.variable)]:
                    newmap[(obj.idnum, a.variable)] = False
                    newmap[("conc",msg.idnum,i)] = True
                    rval = True
                else:
                    newmap[("conc",msg.idnum,i)] = False
            obj = utils.get_object(item.abst_state, item.parent.abst_sim_oid)
            msg = obj.sm_input_queue[0]
            for i,a in enumerate(item.assignments):
                if oldmap[(obj.idnum, a.variable)]:
                    newmap[("abst",msg.idnum,i)] = True
                else:
                    newmap[("abst",msg.idnum,i)] = False
            return rval
                
        elif isinstance(item, events.Counterexample_Send):
            for i,arg in enumerate(item.arglist):
                if oldmap[("conc",item.conc_msg_id,i)]:
                    rd = slicer.slicer.get_refdef(str(arg), "expression")
                    refs = rd.refs
                    assert len(rd.defs) == 0
                    for r in refs:
                        stmts = actlang.grammar.parse("x = (" + r + ");").api_getTopStatements()
                        assert len(stmts) == 1
                        tree = stmts[0].api_getExpr().api_getRhsExpr()
                        obj, r = utils.get_obj_name_pair(tree, exec_obj)
                        newmap[(obj.idnum, r)] = True
        elif isinstance(item, events.Counterexample_Impl_Item):
            # needs only copying of the variables in the old map,
            # which has already been done in the beginning of this function
            pass
        elif isinstance(item, events.Counterexample_Defer_Item):
            # needs only copying of the variables in the old map,
            # which has already been done in the beginning of this function
            pass
        else:
            if isinstance(item, events.Counterexample_Effect):
                rd = slicer.slicer.get_refdef(str(code), "statement")
            elif isinstance(item, events.Counterexample_Guard):
                rd = slicer.slicer.get_refdef(str(code), "expression")
            else:
                raise error.Error()
            refs = rd.refs
            defs = rd.defs
            if len(defs) > 0:
                assert len(defs) == 1
                d = defs.pop()
                stmts = actlang.grammar.parse("x = (" + d + ");").api_getTopStatements()
                assert len(stmts) == 1
                tree = stmts[0].api_getExpr().api_getRhsExpr()
                obj, d = utils.get_obj_name_pair(tree, exec_obj)
                if newmap[(obj.idnum, d)]:
                    newmap[(obj.idnum, d)] = False
                    for r in refs:
                        stmts = actlang.grammar.parse("x = (" + r + ");").api_getTopStatements()
                        assert len(stmts) == 1
                        tree = stmts[0].api_getExpr().api_getRhsExpr()
                        obj, r = utils.get_obj_name_pair(tree, exec_obj)
                        newmap[(obj.idnum, r)] = True

    def _calculate_relevant_variables(self, crit, ending_point=-1):
        """Calculates relevant variables based on the initial relevant
        variables given. After execution information about the relevant
        variables can be found in items' slice_maps.

        Arguments:
            crit          -- a set of (oid, name) pairs where oid is the
                             object id and name is the name of the variable
            ending_point  -- This is an index indicating the last point in
                             the trace that is taken account. In other
                             words the variable names in 'crit' are the
                             relevant variables in that point.
        """
        #self.crit = crit
        self.var_map = {}

        items = self.get_item_list()
        item = items[ending_point]

        for obj in item.conc_state.objects:
            for vname in obj.variable_values.keys():
                if (obj.idnum, vname) in crit:
                    item.slice_map[(obj.idnum, vname)] = True
                else:
                    item.slice_map[(obj.idnum, vname)] = False
            for msg in obj.sm_input_queue:
                for i,p in enumerate(msg.parameters):
                    item.slice_map[("conc",msg.idnum,i)] = False
        for obj in item.abst_state.objects:
            for msg in obj.sm_input_queue:
                for i,p in enumerate(msg.parameters):
                    item.slice_map[("abst",msg.idnum,i)] = False

        oldmap = item.slice_map
        for item in reversed(items[:ending_point]):
            if isinstance(item, events.Counterexample_Effect):
                self.__slice_step(oldmap, item)
            elif isinstance(item, events.Counterexample_Guard):
                self.__slice_step(oldmap, item)
            elif isinstance(item, events.Counterexample_Trigger):
                #print item.slice_map
                if self.__slice_step(oldmap, item):
                    #print item.signal_name
                    #print item.slice_map
                    conc_target = item.parent.conc_sim_oid
                    abst_target = item.parent.abst_sim_oid
                    conc_obj = utils.get_object(item.conc_state, conc_target)
                    conc_msg = conc_obj.sm_input_queue[0]
                    abst_obj = utils.get_object(item.abst_state, abst_target)
                    abst_msg = abst_obj.sm_input_queue[0]
                    conc_ce = self.conc_sendermap[conc_msg.idnum]
                    abst_ce = self.abst_sendermap[abst_msg.idnum]
                    earlier_ce = self._earlier_item(conc_ce, abst_ce)
                    index = self._check_signal_correspondence(earlier_ce, conc_target, abst_target)
                    # if signal send targets didn't correspond, find the place where targets are off
                    # and start analyzing from there
                    if index >= 0:
                        self._clear_relevant_variables()
                        item = items[index]
                        assert isinstance(item, events.Counterexample_Send)
                        crit = self._find_initial_relevant_variables(item, item.target_expr)
                        self._calculate_relevant_variables(crit, index)
                        return
            elif isinstance(item, events.Counterexample_Impl_Item):
                self.__slice_step(oldmap, item)
            elif isinstance(item, events.Counterexample_Defer_Item):
                self.__slice_step(oldmap, item)
            oldmap = item.slice_map

    def _earlier_item(self, item1, item2):
        """Returns item that appears first in the Counterexample_Trace.

        Arguments:
            item1  -- first Counterexample_Item object
            item2  -- second Counterexample_Item object
        """
        if item1 == item2:
            return item1
        items = self.get_item_list()
        for item in items:
            if item == item1:
                return item1
            if item == item2:
                return item2
        raise error.Error("Item not in items of the Counterexample_Trace")

    def _check_signal_correspondence(self, ending_point, conc_target, abst_target):
        """Checks whether all the sends to an object are done correspondingly in the
        abstract and in the concrete world.

        Arguments:
            ending_point  -- last item to be checked (Counterexample_Item)
            conc_target   -- concrete object id in simulator
            abst_target   -- abstract object id in simulator

        Returns -1 if sends are done correspondingly, otherwise an index for the item
        where signal sends differ.
        """
        items = self.get_item_list()
        for i, item in enumerate(items):
            if isinstance(item, events.Counterexample_Send):
                if not ((conc_target == item.conc_target.idnum and
                         abst_target == item.abst_target.idnum ) or
                        (conc_target != item.conc_target.idnum and
                         abst_target != item.abst_target.idnum )):
                    return i
            if item == ending_point:
                return -1

    def get_item_list(self):
        """Returns a list of Counterexample_Item objects contained in this trace.
        """
        lst = []
        for e in self.events:
            lst.extend(e.get_item_list())
        return lst

    def _clear_relevant_variables(self):
        """Clears slice_maps in every Counterexample_Item in this trace
        """
        items = self.get_item_list()
        for i in items:
            for k in i.slice_map.keys():
                i.slice_map[k] = False

    def __str__(self):
        s = ""
        indent = Utils.String.Indent()
        for event_i, event in enumerate(self.events):
            s += str(indent) + "Event " + str(event_i) + ": " + str(event)        
        return s
        
class Refinement_Hint:
    """Refinement hint for the refinement procedure.

    Fields:
        refinements -- a list of tuples (xmiid, values), where xmiid
                       is the xmiid of the variable to be refined and
                       values is a list of values that should be isolated
                       in the interval abstraction
    """
    def __init__(self):
        """Initializes an empty Refinement_Hint
        """
        self.refinements = []

    def add(self, var, hints):
        """Adds new variable and a set of values that should be isolated in the
        variables asbtract type

        Arguments:
            var    -- xmiid of the variable
            hints  -- a list of values
        """
        self.refinements.append((var, hints))

    def init_basic(self, varlist, hints):
        """Sets the refinement hint so that all the variables in varlist are
        refined with all the values in list of hints

        Arguments:
            varlist -- a list of xmiids
            hints   -- a list of values
        """
        self.refinements = []
        for v in varlist:
            self.add(v, list(hints))

    def get_refinements(self):
        """Returns a list of suggested refinements as a list containing tuples
        (xmiid, values), where xmiid is the xmiid of the variable to be refined
        and values is a list of values that should be isolated in the interval
        abstraction
        """
        return self.refinements

    def __str__(self):
        indent = Utils.String.Indent()
        return self.__str_indent(indent)

    def __str_indent(self, indent):
        assert isinstance(indent, Utils.String.Indent)
        s = ""
        s += str(indent) + "Refinement hint:\n"
        indent.push()
        for v, h in self.refinements:
            s += str(indent) + v + ": " + ", ".join(map(str, h)) + "\n"
        return s

class Subexpression_Evaluator:
    """Class for collecting all the subexpressions of the expression and
    evaluating the values of subexpressions collected.

    Fields:
        values  -- a set of values subexpressions evaluated to.
        names   -- a set of xmiids indicating the set of variables
                   subexpressions contained.
    """
    def __init__(self, sim, oid, tree, absdb):
        """Initializes Subexpression_Evaluator object.

        Arguments:
            sim    -- Simulator where evaluations are made.
            oid    -- Simulator object id indicating the context in which
                      the evaluations are made
            tree   -- expression as a Jumbala parse tree. The expression
                      has to be side effect free.
            absdb  -- Absdb ojbect containing information about
                      corresponding abstract subexpressions.
        """
        self.sim = sim
        self.oid = oid
        self.tree = tree
        self.absdb = absdb
        assert self.sim != None
        assert self.oid != None
        assert self.tree != None
        assert self.absdb != None
        self.values = set()
        self.names = set()

    def _collect_vars(self, tree, exec_obj):
        """Collects a set of variables' xmiids indicating the set of variables
        tree contains.

        Arguments:
            tree      -- Jumbala parse tree
            exec_obj  -- Object whose context is used as a context for tree
        """
        varlist = utils.collect_var_names(tree, exec_obj)
        rval = []
        for obj, v in varlist:
            xmiid = obj.get_class_def()._variable_defs[v].uml_reference.__XMIid__()
            ai = self.absdb.get_abst_type(xmiid)
            if ai != None and ai not in ("int", "boolean"):
                rval.append(xmiid)
        return rval

    def collect(self):
        """Collects variables' xmiids and subexpression results from a tree given at
        initialization. Stores the information to fields 'values' and 'names'.
        """
        self.values = self._collect_subexpr_evals(self.tree)
        obj = utils.get_object(self.sim.get_working_global_state(), self.oid)
        self.names = set(self._collect_vars(self.tree, obj))
    
    def _collect_subexpr_evals(self, tree):
        """Collects all integer values subexpressions of the tree are evaluated to.

        Arguments:
            tree  -- Jumbala parse tree.
        """
        debug = False
        values = set()

        # process each actlang-tree-type appropriately
        if isinstance(tree, actlang.expr.ParenExpr):
            return self._collect_subexpr_evals(tree.api_getInnerExpr())
        elif isinstance(tree, actlang.expr.CondAndExpr):
            a = tree.api_getLeftOperand()
            b = tree.api_getRightOperand()
            # evaluation takes account short-circuit evaluation
            
            result_a = self.__evaluate_tree(a)
            assert type(result_a) is bool
            if result_a == False:
                values.update(self._collect_subexpr_evals(a))
            else:
                values.update(self._collect_subexpr_evals(a))
                values.update(self._collect_subexpr_evals(b))
        elif isinstance(tree, actlang.expr.CondOrExpr):
            a = tree.api_getLeftOperand()
            b = tree.api_getRightOperand()
            # evaluation takes account short-circuit evaluation
            
            result_a = self.__evaluate_tree(a)
            assert type(result_a) is bool
            if result_a == True:
                values.update(self._collect_subexpr_evals(a))
            else:
                values.update(self._collect_subexpr_evals(a))
                values.update(self._collect_subexpr_evals(b))
        elif isinstance(tree, actlang.expr.InfixExpr):
            a = tree.api_getLeftOperand()
            b = tree.api_getRightOperand()

            result = self.__evaluate_tree(tree)
            if type(result) is int:
                values.add(result)
            values.update(self._collect_subexpr_evals(a))
            values.update(self._collect_subexpr_evals(b))
        elif isinstance(tree, (actlang.expr.LogicalNotExpr,
                               actlang.expr.BitNotExpr,
                               actlang.expr.UnaryPlusExpr,
                               actlang.expr.UnaryMinusExpr)):
            result = self.__evaluate_tree(tree)
            if type(result) is int:
                values.add(result)
            values.update(self._collect_subexpr_evals(tree.api_getOperand()))
        elif isinstance(tree, actlang.expr.TernaryCondExpr):
            c = tree.api_getCondition()
            a = tree.api_getTrueExpr()
            b = tree.api_getFalseExpr()
            result_c = self.__evaluate_tree(c)
            assert type(result_c) is bool
            values.update(self._collect_subexpr_evals(a))
            if result_c == True:
                values.update(self._collect_subexpr_evals(a))
            else:
                values.update(self._collect_subexpr_evals(b))
        elif isinstance(tree, (actlang.expr.BoolLiteralExpr,
                            actlang.expr.ThisExpr)):
            return
        elif isinstance(tree, (actlang.expr.IntLiteralExpr,
                               actlang.names.SimpleName,
                               actlang.names.QualifiedName)):
            result = self.__evaluate_tree(tree)
            if type(result) is int:
                values.add(result)
        elif isinstance(tree, (actlang.expr.InstanceCreationExpr,
                            actlang.expr.PrimaryFieldAccessExpr)):
            raise Error()
        else:
            print tree.__class__.__name__
            raise Error()

        return values

    def __evaluate_tree(self, tree):
        """Evaluates a tree.

        Arguments:
            tree  -- Jumbala parse tree.

        Returns a value to which tree evaluated.
        """
        # only expressions and names supported
        assert isinstance(tree, (actlang.expr.Expr, actlang.names.Name))

        # get copy of the global states (shoud the working state be restored after operations)
        old_gs = self.sim.get_working_global_state().api_copy()

        # get type information for execute_code_in_object-method
        typ = utils.determine_return_type(tree, self.sim.get_working_global_state(), self.oid)
        value = self.sim.execute_code_in_object(self.oid, "return " + str(tree) + ";", typ)
        return value.get_value()

