# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Classes for different events in the counterexample trace to be analyzed.
Subclasses of Counterexample_Event represent atomic events in the
counterexample (transiticion executions, implicit consumptions and signal
defers). Subclasses of Counterexample_Item are actions from which events are
constructed from. For example transition execution can contain message
receiving, guard evaluation and a list of statement executions.

Author: Vesa Ojala <vojala@tcs.tkk.fi>
"""

import coral
import actlang
import Sim
import Sim.Core
import Sim.Object
import Utils.String
import Utils.Models

import tracesim
from abstraction import typelibrary

#from canal import abstcomp
from canal import utils
import absttrace

import error

UML14 = coral.meta.UML14
DIA20 = coral.meta.XMIDI20

class Counterexample_Event(object):
    """Abstract base class for events in the counterexample.

    Fields:
        parent        -- Counterexample_Trace object in which this event
                         belongs to.
        conc_sim_oid  -- simulator object id representing the object in which
                         this action happens in the concrete world.
        abst_sim_oid  -- simulator object id representing the object in which
                         this action happens in the abstract world.
        chooses       -- Counterexample_Chooses object carrying the information
                         about the non-deterministic chooses made at this point
                         of the trace.
    """
    def __init__(self, ct):
        """Abstract init function

        Arguments:
            ct  -- Counterexample_Trace object in which the event belongs
        """
        assert isinstance(ct, absttrace.Counterexample_Trace)
        self.parent = ct
        self.conc_sim_oid = None
        self.abst_sim_oid = None
        self.chooses = ct.abst_ts.get_choose_maps()

    def __str__(self):
        """Returns a string representation of an event.
        """
        return self._str_indent()
        
    def _str_indent(self, indent=None):
        """Abstract method
        """
        raise error.Error("Method of abstract class")

class Counterexample_Chooses:
    """Class storing the state of the choose maps at the beginning of the
    event execution.

    Fields:
        effect_map  -- A dictionary for chooses in effects
        guard_map   -- A dictionary for chooses in guards
    """
    def __init__(self):
        self.effect_map = None
        self.guard_map = None

class Counterexample_Item(object):
    """Base class of the classes for different items in events:
    triggers, guards, effects, implicit consumptions and
    signal defers.

    Fields:
        parent      -- Counterexample_Event in which this item belongs to.
        slice_map   -- A dictionary containing the information about the
                       relevant variables at this point of execution.
        conc_state  -- Concrete state at this point of execution.
        abst_state  -- Abstract state at this point of execution.
    """
    def __init__(self, parent, conc_sim, abst_sim):
        """Initializes Counterexample_Item object.

        Fields:
            parent    -- Counterexample_Event in which this item belongs to.
            conc_sim  -- Concrete simulator in the state just before
                         execution of the action the item represents.
            abst_sim  -- Abstract simulator in the state just before
                         execution of the action the item represents.

        Returns an initialized Counterexample_Item object.
        """
        self.slice_map = {}
        self.parent = parent
        self.conc_state = conc_sim.get_working_global_state().api_copy()
        self.abst_state = abst_sim.get_working_global_state().api_copy()

    def __str__(self):
        return self._str_indent(None)

    def _str_indent(self, indent):
        assert indent == None or isinstance(indent, Utils.String.Indent)
        if indent == None:
            indent = Utils.String.Indent()

        s = ""
        s += str(indent) + "Concrete state:\n"
        indent.push()
        s += _str_global_state(indent, self.conc_state, self.slice_map, "conc")
        indent.pop()
        s += str(indent) + "Abstract state:\n"
        indent.push()
        s += _str_global_state(indent, self.abst_state, self.slice_map, "abst")
        indent.pop()
        
        return s

    def has_relevant_variables(self):
        """Returns True if the global state before the execution of the item
        has relevant variables, otherwise False
        """
        if True in self.slice_map.values():
            return True
        else:
            return False

    def get_event(self):
        return self.parent

class Counterexample_Impl(Counterexample_Event):
    """Implicit consumption event in the counterexample

    Fields:
        item     -- Counterexample_Impl_Item representing
                    the implicit signal consumption made
        trigger  -- Trigger which is not enabled in the abstract
                    world but is enabled in the concrete world
                    and thus prevents the implicit consumption
                    in the concrete world.
        guard    -- Guard which evaluates to false in the abstract
                    world but evaluates to true in the concrete world
                    thus preventing the implicit consumption in the
                    concrete world.
    """
    def __init__(self, ct, action, mode=("normal","effect")):
        """Initializes Counterexample_Impl object.

        Arguments:
            ct      -- Counterexample_Trace object where this object
                       belongs to.
            action  -- Trace action from tracesim, should be
                       ('IMPL', sim_oid) where sim_oid is simulator
                       object id
            mode    -- Defines whether the processing of this event
                       should be limited. Limited processing is used when
                       the action is the last action in the counterexample
                       trace. Could be one of the following:
                         * ('normal', dummy), where dummy is a arbitrary
                           string
                         * ('limit', 'guard') when the reason for the
                           implicit consumption happening in the abstract
                           world but not in the concrete world is a guard
                           evaluating differently between the abstract and
                           the concrete world
                         * ('limit', 'trigger') when the reason for the
                           implicit consumption happening in the abstract
                           world but not in the concrete world is a
                           missing or an extra message in a message queue
                           (messages sent to wrong place)

        Returns an initialized Counterexample_Impl object
        """
        super(Counterexample_Impl, self).__init__(ct)
        (dummy, mc_object) = action

        limit, limit_type = mode
            
        # init sim_oid
        self.conc_sim_oid = ct.conc_ts.mc_object_to_sim_oid_map[mc_object]
        self.abst_sim_oid = ct.abst_ts.mc_object_to_sim_oid_map[mc_object]

        self.item = Counterexample_Impl_Item(self, ct.conc_ts.sim, ct.abst_ts.sim)
        self.trigger = None
        self.guard = None
        
        if limit == 'limit' and limit_type == 'guard':
            # find enabled transition in the concrete world which
            # disables the possibility for implicit consumption
            enabled_sets,implicit_consumption_objects = ct.conc_ts.sim.get_enabled_in_working_global_state()
            trans_xmiid = None
            for (index, (o,set)) in enumerate(enabled_sets):
                if o.idnum != self.conc_sim_oid:
                    continue
                assert len(set) == 1
                (reason,sim_transition) = set[0]
                if (reason == "completion" or reason == "ATO" or
                    reason == "signal"):
                    pass
                elif reason == "defer":
                    continue
                else:
                    assert False and "Internal error: something not implemented"
                trans_xmiid = sim_transition.__XMIid__()
            assert trans_xmiid
            #from exec
            assert ct.conc_transition_map.has_key(trans_xmiid)
            assert ct.abst_transition_map.has_key(trans_xmiid)
            conc_uml_trans = ct.conc_transition_map[trans_xmiid]
            self.conc_uml_trans = conc_uml_trans
            abst_uml_trans = ct.abst_transition_map[trans_xmiid]
            # make Counterexample_Trigger
            if conc_uml_trans.trigger:
                self.trigger = Counterexample_Trigger(self, conc_uml_trans.trigger,
                                                      ct.conc_ts.sim,
                                                      ct.abst_ts.sim)
                self.trigger.execute(ct.conc_ts.sim, ct.abst_ts.sim)
            # make Counterexample_Guard
            if conc_uml_trans.guard and limit_type in ("effect","guard"):
                guard_tree = ct.parse_tree_map[conc_uml_trans.guard.__XMIid__()]
                self.guard = Counterexample_Guard(self, guard_tree, ct.conc_ts.sim, ct.abst_ts.sim)
        elif limit == 'limit' and limit_type == 'trigger':
            # doesn't need any special attention
            pass

    def get_item_list(self):
        """Returns list of Counterexample_Item objects containing all
        items in this object.
        """
        lst = [self.item]
        if self.trigger:
            lst.append(self.trigger)
        if self.guard:
            lst.append(self.guard)
        assert (len(lst) == 3 or len(lst) == 1)
        return lst

    def _str_indent(self, indent=None):
        assert indent == None or isinstance(indent, Utils.String.Indent)
        if indent == None:
            indent = Utils.String.Indent()

        s = "Implicit consumption\n"
        indent.push()
        s += str(indent) + "Concrete: Object " + str(self.conc_sim_oid) + "\n"
        s += str(indent) + "Abstract: Object " + str(self.abst_sim_oid) + "\n"
        s += self.item._str_indent(indent)
        if self.trigger:
            indent.push()
            s += "Trigger from transition enabled in the concrete side:\n"
            s += self.trigger._str_indent(indent)
            indent.pop()
        if self.guard:
            indent.push()
            s += "Guard from transition enabled in the concrete side:\n"
            s += self.guard._str_indent(indent)
            indent.pop()
        indent.pop()
        return s
    
class Counterexample_Impl_Item(Counterexample_Item):
    """Class for implicit consumptions happening, is located inside
    Counterexample_Impl object.
    """
    def __init__(self, parent, conc_sim, abst_sim):
        """Initializes Counterexample_Impl_Item object.

        Arguments:
            parent    -- Counterexample_Impl object in which this object
                         belongs to.
            conc_sim  -- Concrete simulator in the state just before
                         execution of the implicit consumption
            abst_sim  -- Abstract simulator in the state just before
                         execution of the implicit consumption
                         
        Returns initialized Counterexample_Impl_Item object.
        """
        #self.parent = parent
        #self.slice_map = {}
        #self.conc_state = conc_sim.get_working_global_state().api_copy()
        #self.abst_state = abst_sim.get_working_global_state().api_copy()
        super(Counterexample_Impl_Item, self).__init__(parent, conc_sim, abst_sim)

    def execute(self, conc_sim, abst_sim):
        """Executes this implicit consumption in the abstract and in
        the concrete world.

        Arguments:
            conc_sim  -- Simulator where the implicit consumption is
                         executed in the concrete world.
            abst_sim  -- Simulator where the implicit consumption is
                         executed in the abstract world.
        """
        for sim, sim_oid, error_str in [(conc_sim, self.parent.conc_sim_oid, "concrete"),
                                        (abst_sim, self.parent.abst_sim_oid, "abstract")]:
            enabled_sets,implicit_consumption_objects = sim.get_enabled_in_working_global_state()
            found_index = None
            for (index,o) in enumerate(implicit_consumption_objects):
                if o.idnum != sim_oid:
                    continue
                assert not found_index
                found_index = index

            if found_index == None:
                raise error.Error("Unexpected error in the " + error_str + " model simulation: " +
                                  "Error during execution of implicit consumption")
class Counterexample_Defer(Counterexample_Event):
    """Signal defer in the counterexample

    Fields:
        item     -- Counterexample_Defer_Item representing
                    the implicit signal consumption made
        trigger  -- Trigger which is not enabled in the abstract
                    world but is enabled in the concrete world
                    and thus prevents the signal defer in the
                    concrete world.
        guard    -- Guard which evaluates to false in the abstract
                    world but evaluates to true in the concrete world
                    thus preventing the signal defer in the concrete world.
    """
    def __init__(self, ct, action, mode=("normal","effect")):
        """Initializes Counterexample_Defer object.

        Arguments:
            ct      -- Counterexample_Trace object where this object
                       belongs to.
            action  -- Trace action from tracesim, should be
                       ('DEFER', sim_oid) where sim_oid is simulator
                       object id
            mode    -- Defines whether the processing of this event
                       should be limited. Limited processing is used when
                       the action is the last action in the counterexample
                       trace. Could be one of the following:
                         * ('normal', dummy), where dummy is a arbitrary
                           string
                         * ('limit', 'guard') when the reason for the
                           signal defer happening in the abstract world
                           but not in the concrete world is a guard evaluating
                           differently between the abstract and the
                           concrete world
                         * ('limit', 'trigger') when the reason for the
                           signal defer happening in the abstract world
                           but not in the concrete world is a missing or an extra
                           message in a message queue (messages sent to
                           wrong place)

        Returns an initialized Counterexample_Defer object
        """
        super(Counterexample_Defer, self).__init__(ct)
        (dummy, mc_object) = action

        limit, limit_type = mode
            
        # init sim_oid
        self.conc_sim_oid = ct.conc_ts.mc_object_to_sim_oid_map[mc_object]
        self.abst_sim_oid = ct.abst_ts.mc_object_to_sim_oid_map[mc_object]

        self.item = Counterexample_Defer_Item(self, ct.conc_ts.sim, ct.abst_ts.sim)
        self.trigger = None
        self.guard = None
        
        if limit == 'limit' and limit_type == 'guard':
            # find enabled transition in the concrete world which
            # disables the possibility for signal defer
            enabled_sets,implicit_consumption_objects = ct.conc_ts.sim.get_enabled_in_working_global_state()
            trans_xmiid = None
            for (index, (o,set)) in enumerate(enabled_sets):
                if o.idnum != self.conc_sim_oid:
                    continue
                assert len(set) == 1
                (reason,sim_transition) = set[0]
                if (reason == "completion" or reason == "ATO" or
                    reason == "signal"):
                    pass
                elif reason == "defer":
                    continue
                else:
                    assert False and "Internal error: something not implemented"
                trans_xmiid = sim_transition.__XMIid__()
            assert trans_xmiid
            #from exec
            assert ct.conc_transition_map.has_key(trans_xmiid)
            assert ct.abst_transition_map.has_key(trans_xmiid)
            conc_uml_trans = ct.conc_transition_map[trans_xmiid]
            self.conc_uml_trans = conc_uml_trans
            abst_uml_trans = ct.abst_transition_map[trans_xmiid]
            # make Counterexample_Trigger
            if conc_uml_trans.trigger:
                self.trigger = Counterexample_Trigger(self, conc_uml_trans.trigger,
                                                      ct.conc_ts.sim,
                                                      ct.abst_ts.sim)
                self.trigger.execute(ct.conc_ts.sim, ct.abst_ts.sim)
            # make Counterexample_Guard
            if conc_uml_trans.guard and limit_type in ("effect","guard"):
                guard_tree = ct.parse_tree_map[conc_uml_trans.guard.__XMIid__()]
                self.guard = Counterexample_Guard(self, guard_tree, ct.conc_ts.sim, ct.abst_ts.sim)
        elif limit == 'limit' and limit_type == 'trigger':
            pass

    def get_item_list(self):
        """Returns list of Counterexample_Item objects containing all
        items in this object.
        """
        lst = [self.item]
        if self.trigger:
            lst.append(self.trigger)
        if self.guard:
            lst.append(self.guard)
        assert (len(lst) == 3 or len(lst) == 1)
        return lst

    def _str_indent(self, indent=None):
        assert indent == None or isinstance(indent, Utils.String.Indent)
        if indent == None:
            indent = Utils.String.Indent()

        s = "Signal defer\n"
        indent.push()
        s += str(indent) + "Concrete: Object " + str(self.conc_sim_oid) + "\n"
        s += str(indent) + "Abstract: Object " + str(self.abst_sim_oid) + "\n"
        s += self.item._str_indent(indent)
        indent.pop()
        return s
    
class Counterexample_Defer_Item(Counterexample_Item):
    """Class for signal defers. Is located inside Counterexample_Defer object
    which is the event for signal defers.
    """
    def __init__(self, parent, conc_sim, abst_sim):
        """Initializes Counterexample_Defer_Item object.

        Arguments:
            parent    -- Counterexample_Defer object in which this object
                         belongs to.
            conc_sim  -- Concrete simulator in the state just before
                         execution of the signal defer
            abst_sim  -- Abstract simulator in the state just before
                         execution of the signal defer
                         
        Returns initialized Counterexample_Defer_Item object.
        """
        #self.parent = parent
        #self.slice_map = {}
        #self.conc_state = None
        #self.abst_state = None

        # save global states before code execution
        #self.conc_state = conc_sim.get_working_global_state().api_copy()
        #self.abst_state = abst_sim.get_working_global_state().api_copy()

        super(Counterexample_Defer_Item, self).__init__(parent, conc_sim, abst_sim)

    def execute(self, conc_sim, abst_sim):
        """Executes the signal defer in the abstract and in
        the concrete world.

        Arguments:
            conc_sim  -- Simulator where the signal defer is
                         executed in the concrete world.
            abst_sim  -- Simulator where the signal defer is
                         executed in the abstract world.
        """
        for sim, sim_oid, error_str in [(conc_sim, self.parent.conc_sim_oid, "concrete"),
                                        (abst_sim, self.parent.abst_sim_oid, "abstract")]:
            enabled_sets,implicit_consumption_objects = sim.get_enabled_in_working_global_state()
            found_index = None
            for (index,o) in enumerate(implicit_consumption_objects):
                if o.idnum != sim_oid:
                    continue
                assert not found_index
                found_index = index

            if found_index == None:
                raise error.Error("Unexpected error in the " + error_str + " model simulation: " +
                                  "Error during execution of implicit consumption")

class Counterexample_Exec(Counterexample_Event):
    """Transition execution event in the counterexample.

    Fields:
        trigger         -- Counterexample_Trigger object for the trigger
                           in this transition execution. None if transition
                           has no trigger.
        guard           -- Counterexample_Guard object for the guard
                           in this transition execution. None if transition
                           has no guard.
        effect          -- A list of Counterexample_Effect objects, one
                           for each statement in the effect of the transitio.
        conc_uml_trans  -- Transition in the concrete UML model.
    """
    def __init__(self, ct, action, mode=("normal","effect")):
        """Initializes Counterexample_Exec object.

        Arguments:
            ct      -- Counterexample_Trace object where this object
                       belongs to.
            action  -- Trace action from tracesim, should be
                       ('EXEC', mc_object, xmiid) where mc_object is the
                       model checking object id (this is later translated
                       to simulator object id), and xmiid is the xmiid of
                       the transition that was executed.
            mode    -- Defines whether the processing of this event
                       should be limited. Limited processing is used when
                       the event is the last event in the counterexample
                       trace. Could be one of the following:
                         * ('normal', dummy), where dummy is a arbitrary
                           string
                         * ('limit', 'trigger') when the reason for the
                           transition execution to fail a trigger that
                           does not receive correct signal.
                         * ('limit', 'guard') when the reason for the
                           transition execution to fail is a guard that
                           evaluates false.
                         * ('limit', 'effect') when assertion failure
                           happens in the effect of the transition.

        Returns an initialized Counterexample_Exec object
        """
        super(Counterexample_Exec, self).__init__(ct)
        
        #self.parent = ct
        self.trigger = None
        self.guard = None
        self.effect = []
        self.conc_uml_trans = None

        #assert isinstance(ct, Counterexample_Trace)
        (dummy, mc_object, trans_xmiid) = action

        limit, limit_type = mode
            
        # init sim_oid
        self.conc_sim_oid = ct.conc_ts.mc_object_to_sim_oid_map[mc_object]
        self.abst_sim_oid = ct.abst_ts.mc_object_to_sim_oid_map[mc_object]
        # get uml transitions
        assert ct.conc_transition_map.has_key(trans_xmiid)
        assert ct.abst_transition_map.has_key(trans_xmiid)
        conc_uml_trans = ct.conc_transition_map[trans_xmiid]
        self.conc_uml_trans = conc_uml_trans
        abst_uml_trans = ct.abst_transition_map[trans_xmiid]
        # make Counterexample_Trigger
        if conc_uml_trans.trigger:
            self.trigger = Counterexample_Trigger(self, conc_uml_trans.trigger,
                                                  ct.conc_ts.sim,
                                                  ct.abst_ts.sim)
            if not (limit == "limit" and limit_type == "trigger"):
                self.trigger.execute(ct.conc_ts.sim, ct.abst_ts.sim)
        # make Counterexample_Guard
        if conc_uml_trans.guard and limit_type in ("effect","guard"):
            guard_tree = ct.parse_tree_map[conc_uml_trans.guard.__XMIid__()]
            self.guard = Counterexample_Guard(self, guard_tree, ct.conc_ts.sim, ct.abst_ts.sim)
        # make Counterexample_Effects
        if conc_uml_trans.effect and limit_type == "effect":
            # get statements to be executed (parts of transition effect)
            conc_treelist = ct.parse_tree_map[conc_uml_trans.effect.__XMIid__()]
            abst_prog = abst_uml_trans.effect.script.body
            abst_treelist = actlang.grammar.parse(abst_prog).api_getTopStatements()
            assert len(conc_treelist) == len(abst_treelist)
            
            ziplist = zip(conc_treelist,abst_treelist)
            for conc_tree, abst_tree in ziplist:
                if isinstance(conc_tree, actlang.statement.SendStatement):
                    ce = Counterexample_Send(self, conc_tree, abst_tree, ct.conc_ts.sim, ct.abst_ts.sim)
                else:
                    ce = Counterexample_Effect(self, conc_tree, abst_tree, ct.conc_ts.sim, ct.abst_ts.sim)
                succesfully_executed = ce.execute(ct.conc_ts.sim, ct.abst_ts.sim)
                self.effect.append(ce)
                if not succesfully_executed:
                    assert limit == "limit"
                    break
            return

    def get_item_list(self):
        """Returns list of Counterexample_Item objects containing all
        items in this object.
        """
        lst = []
        if self.trigger != None:
            lst.append(self.trigger)
        if self.guard != None:
            lst.append(self.guard)
        lst.extend(self.effect)
        return lst

    def _str_indent(self, indent=None):
        assert indent == None or isinstance(indent, Utils.String.Indent)
        if indent == None:
            indent = Utils.String.Indent()

        s = "Transition execution\n"
        indent.push()
        s += str(indent) + "Concrete: Object " + str(self.conc_sim_oid) + "\n"
        s += str(indent) + "Abstract: Object " + str(self.abst_sim_oid) + "\n"
        s += str(indent) + "Trigger:\n"
        indent.push()
        if self.trigger != None:
            t = self.trigger
            s += t._str_indent(indent)
        else:
            s += str(indent) + "Trigger: no trigger\n"
        indent.pop()
        s += str(indent) + "Guard:\n"
        indent.push()
        if self.guard != None:
            g = self.guard
            s += g._str_indent(indent)
        else:
            s += str(indent) + "Guard: no guard\n"
        indent.pop()
        s += str(indent) + "Effect:\n"
        indent.push()
        if len(self.effect) == 0:
            s += str(indent) + "no effect\n"
        for effect_i, effect in enumerate(self.effect):
            s += str(indent) + "Part "+str(effect_i)+":\n"
            indent.push()
            s += effect._str_indent(indent)
            indent.pop()
        indent.pop()
        indent.pop()
        return s
    
class Counterexample_Trigger(Counterexample_Item):
    """Class for triggers executed in the counterexample.

    Fields:
        signal       -- UML signal from the concrete model
        signal_name  -- name of the signal as a string
        assignments  -- a list of Counterexample_Trigger_Assignment
                        objects, one for each message parameter.
    """
    def __init__(self, parent, trigger, conc_sim, abst_sim):
        """Initializes Counterexample_Trigger object.

        Arguments:
            parent    -- Counterexample_Exec object in which this object
                         belongs to.
            trigger   -- UML trigger which was triggered
            conc_sim  -- Concrete simulator in the state just before
                         execution of the trigger
            abst_sim  -- Abstract simulator in the state just before
                         execution of the trigger
                         
        Returns initialized Counterexample_Trigger object.
        """
        #self.parent = parent
        #self.conc_sim = conc_sim
        #self.abst_sim = abst_sim
        self.signal = None
        self.signal_name = None
        #self.slice_map = {}
        #self.code = None
        #self.sender = None
        self.assignments = []

        super(Counterexample_Trigger, self).__init__(parent, conc_sim, abst_sim)

        #self.conc_state = conc_sim.get_working_global_state().api_copy()
        #self.abst_state = abst_sim.get_working_global_state().api_copy()

        if not trigger.oclIsKindOf(UML14.SignalEvent):
            raise error.Error("unsupported type of trigger encountered")

        self.signal = trigger.signal
        self.signal_xmiid = trigger.signal.__XMIid__()
        self.signal_name = str(trigger.signal.name)
        for i in range(0, len(trigger.parameter)):
            name = str(trigger.parameter[i].name)
            assign = Counterexample_Trigger_Assignment(name)
            self.assignments.append(assign)
            # variables receiving signal parameters has to be simple names
            #assert check_simple_name(name)
            #param = Variable(this_class, name)
            #signal.add_parameter(param)

    def execute(self, conc_sim, abst_sim):
        """Executes the trigger in the abstract and in the concrete
        world. Execution of a trigger means that the values in a
        message the trigger consumes are assigned to appropriate variables
        in the model.

        Arguments:
            conc_sim  -- Simulator where the trigger is executed in the
                         concrete world.
            abst_sim  -- Simulator where the trigger is executed in the
                         abstract world.
        """
        #execute in concrete model
        self.__execute_in_gs(conc_sim.get_working_global_state(), self.parent.conc_sim_oid)
        #execute in abstract model
        self.__execute_in_gs(abst_sim.get_working_global_state(), self.parent.abst_sim_oid)

    def __execute_in_gs(self, gs, oid):
        # if external signal, no processing is needed
        if self.signal.name[0] == '$':
            return
        obj = utils.get_object(gs, oid)
        if obj.get_input_queue_length() < 1:
            raise error.Error("Counterexample_Trigger: no signal to receive")
        msg = obj.get_input_queue_message(0)
        msgsig = msg.get_signal()
        if self.signal.__XMIid__() != msgsig.__XMIid__():
            raise error.Error("Counterexample_Trigger: wrong message type")
        # do assignments (does not support object references yet!)
        for i, a in enumerate(self.assignments):
            #print msg.get_parameter(i)
            value = msg.get_parameter(i)
            name = a.variable
            obj.variable_values[name] = value
        # remove signal from the list
        del obj.sm_input_queue[0]

    def __str__(self):
        return self._str_indent(None)

    def _str_indent(self, indent):
        s = super(Counterexample_Trigger, self)._str_indent(indent)
        if indent != None:
            str_indent = str(indent)
        else:
            str_indent = ""
        s += str_indent + "Trigger: " + str(self.signal_name) +\
             "("+", ".join([a.variable for a in self.assignments])+")"+ "\n"
        return s

class Counterexample_Trigger_Assignment:
    """A class for assignments happening when message is received.

    Fields:
        variable  -- a variable name as a string
    """
    def __init__(self, variable):
        """Initializes Counterexample_Trigger_Assignment object.

        Arguments:
            variable  -- a variable name as a string

        Returns an initialized Counterexample_Trigger_Assignment object.
        """
        self.variable = variable

class Counterexample_Guard(Counterexample_Item):
    """Class for guards in Counterexample_Exec events. Guards evaluated in
    a counterexample trace are represented with these objects.

    Fields:
        guard_tree  -- Guard in a concrete world represented as a Jumbala
                       parse tree.
    """
    def __init__(self, parent, guard_tree, conc_sim, abst_sim):
        """Initializes Counterexample_Guard object.

        Arguments:
            parent      -- Counterexample_Exec object in which this object
                           belongs to.
            guard_tree  -- The guard as a Jumbala expression (parse tree).
            conc_sim    -- Concrete simulator in the state just before
                           execution of the guard.
            abst_sim    -- Abstract simulator in the state just before
                           execution of the guard.
                         
        Returns initialized Counterexample_Guard object.
        """
        #self.parent = parent
        #self.slice_map = {}
        #self.conc_state = None
        #self.abst_state = None
        #self.guard = None

        super(Counterexample_Guard, self).__init__(parent, conc_sim, abst_sim)
        self.guard = guard_tree
        #self.conc_state = conc_sim.get_working_global_state().api_copy()
        #self.abst_state = abst_sim.get_working_global_state().api_copy()

    def __str__(self):
        return self._str_indent(None)

    def _str_indent(self, indent):
        s = super(Counterexample_Guard, self)._str_indent(indent)
        if indent != None:
            str_indent = str(indent)
        else:
            str_indent = ""
        s += str_indent + "Guard: " + str(self.guard) + "\n"
        return s
    
class Counterexample_Effect(Counterexample_Item):
    """Class for effect statements in Counterexample_Exec events. Effect
    is formed from a list of these objects, one object for each statement.

    Fields:
        code       -- Statement in the concrete world represented as a
                      Jumbala parse tree.
        abst_tree  -- Statement in the abstract world represented as a
                      Jumbala parse tree.
    """
    def __init__(self, parent, conc_tree, abst_tree, conc_sim, abst_sim):
        """Initializes Counterexample_Effect object.

        Arguments:
            parent     -- Counterexample_Exec object in which this object
                          belongs to.
            conc_tree  -- Jumbala statement describing the effect in the
                          concrete world (parse tree)
            abst_tree  -- Jumbala statement describing the effect in the
                          abstract world (parse tree)
            conc_sim   -- Concrete simulator in the state just before
                          execution of the effect
            abst_sim   -- Abstract simulator in the state just before
                          execution of the effect
                         
        Returns initialized Counterexample_Effect object.
        """
        #self.parent = parent
        #self.slice_map = {}
        #self.conc_state = None
        #self.abst_state = None
        #self.code = None

        super(Counterexample_Effect, self).__init__(parent, conc_sim, abst_sim)
        self.code = conc_tree
        self.abst_tree = abst_tree
        # save global states before code execution
        #self.conc_state = conc_sim.get_working_global_state().api_copy()
        #self.abst_state = abst_sim.get_working_global_state().api_copy()

    def execute(self, conc_sim, abst_sim):
        """Executes the effect in the abstract and in the concrete
        world.

        Arguments:
            conc_sim  -- Simulator where the effect is executed in the
                         concrete world.
            abst_sim  -- Simulator where the effect is executed in the
                         abstract world.
        """
        try:
            abst_sim.execute_code_in_object(self.parent.abst_sim_oid, str(self.abst_tree), None)
        except actlang.errors.Error, e:
            if not isinstance(self.abst_tree, actlang.statement.AssertStatement):
                raise error.Error("Unexpected error in the abstract model simulation: "+\
                                  str(e))
            return False
        try:
            conc_sim.execute_code_in_object(self.parent.conc_sim_oid, str(self.code), None)        
        except actlang.errors.Error, e:
            raise error.Error("Unexpected error in the concrete model simulation: "+\
                              str(e))
        return True

    def __str__(self):
        return self._str_indent(None)

    def _str_indent(self, indent):
        s = super(Counterexample_Effect, self)._str_indent(indent)
        if indent != None:
            str_indent = str(indent)
        else:
            str_indent = ""
        s += str_indent + "Code: " + str(self.code) + "\n"
        return s

class Counterexample_Send(Counterexample_Effect):
    """Class for send statements in Counterexample_Exec events.

    Fields:
        name         -- Name of the signal as a string.
        target_expr  -- A target of the signal send represented as a Jumbala
                        parse tree.
        arglist      -- A list of Jumbala parse trees representing the arguments
                        of the message to be sended.
    """
    def __init__(self, parent, conc_tree, abst_tree, conc_sim, abst_sim):
        """Initializes Counterexample_Send object.

        Arguments:
            parent     -- Counterexample_Exec object in which this object
                          belongs to.
            conc_tree  -- Jumbala statement describing the effect in the
                          concrete world (parse tree)
            abst_tree  -- Jumbala statement describing the effect in the
                          abstract world (parse tree)
            conc_sim   -- Concrete simulator in the state just before
                          execution of the effect
            abst_sim   -- Abstract simulator in the state just before
                          execution of the effect
                         
        Returns initialized Counterexample_Send object.
        """
        assert isinstance(conc_tree, actlang.statement.SendStatement)
        #conc_gs = conc_sim.get_working_global_state()
        #abst_gs = abst_sim.get_working_global_state()
        #self.parent = parent
        self.name = conc_tree.api_getSignalName()
        self.target_expr = conc_tree.api_getTarget()
        self.arglist = conc_tree.api_getArgList()
        super(Counterexample_Send, self).__init__(parent, conc_tree, abst_tree, conc_sim, abst_sim)

    def execute(self, conc_sim, abst_sim):
        """Executes the signal send in the abstract and in the concrete
        world.

        Arguments:
            conc_sim  -- Simulator where the signal send is executed in the
                         concrete world.
            abst_sim  -- Simulator where the signal send is executed in the
                         abstract world.
        """
        value = conc_sim.execute_code_in_object(self.parent.conc_sim_oid,
                                                "return " + str(self.code.api_getTarget()) + ";",
                                                Sim.Core.SimType_ObjectRef())
        self.conc_target = value.get_value()
        value = abst_sim.execute_code_in_object(self.parent.abst_sim_oid,
                                                "return " + str(self.abst_tree.api_getTarget()) + ";",
                                                Sim.Core.SimType_ObjectRef())
        self.abst_target = value.get_value()
        rval = super(Counterexample_Send, self).execute(conc_sim, abst_sim)
        if rval:
            conc_msg_i = self.conc_target.get_input_queue_length()-1
            conc_msg = self.conc_target.get_input_queue_message(conc_msg_i)
            self.parent.parent.conc_sendermap[conc_msg.idnum] = self
            self.conc_msg_id = conc_msg.idnum
            abst_msg_i = self.abst_target.get_input_queue_length()-1
            abst_msg = self.abst_target.get_input_queue_message(abst_msg_i)
            self.parent.parent.abst_sendermap[abst_msg.idnum] = self
            self.abst_msg_id = abst_msg.idnum
        else:
            raise error.Error("Execution of signal send did not succeed")
        return rval

def _str_global_state(indent, gs, mask, simtype):
    """Returns string representation of a global state.

    Arguments:
    indent   -- instance of Utils.String.Indent or None
    gs       -- global state to be represented as a string
    mask     -- dictionary describing the relevant variables
    simtype  -- describes whether the global state is from the
                concrete world or from the abstract world.
                Valid values are 'conc' or 'abst'.

    Returns a string representation of a global state.
    """
    assert indent == None or isinstance(indent, Utils.String.Indent)
    if indent == None:
        indent = Utils.String.Indent()
    s = ""
    for o in gs.objects:
        s += _str_masked_obj(indent, o, mask, simtype)
    return s

def _str_masked_obj(indent, obj, mask, simtype):
    """Returns string representation of an object.

    Arguments:
    indent   -- instance of Utils.String.Indent.
    obj      -- object to be represented as a string
    mask     -- dictionary describing the relevant variables
    simtype  -- describes whether  the object is from the concrete
                world or from the abstract world. Valid values are
                'conc' or 'abst'

    Returns a string representation of an object.
    """
    assert isinstance(indent, Utils.String.Indent)
    assert isinstance(obj, Sim.Object.Object)
    assert type(mask) is dict

    s = ""

    s += (str(indent) + "Object " + str(obj.idnum) + " "
          + obj.get_uml_name() + "::" + obj.get_class_def().get_name() +
          "\n")
    indent.push()
    for (n,v) in obj.variable_values.iteritems():            
        if mask.has_key((obj.idnum, n)) and mask[(obj.idnum, n)]:
            s += str(indent) + "* " + n + " = " + v.get_str()
        else:
            s += str(indent) + n + " = " + v.get_str()
        s += "\n"
    if obj.get_class_def().uml_state_machine:
        # Print the input queue contents
        s += (str(indent) + "sm_input_queue = [")
        s += _str_msg_queue(obj.sm_input_queue, simtype, mask)
        s += ("]\n")
        # Print the defer queue contents
        s += (str(indent) + "sm_defer_queue = [")
        s += _str_msg_queue(obj.sm_defer_queue, simtype, mask)
        s += ("]\n")

        # Print the state configuration
        s += (str(indent) + "sm_state =\n")
        indent.push()
        s += _str_sm_state(indent, obj.sm_state)
        indent.pop()
    indent.pop()
    return s

def _str_msg_queue(queue, simtype, mask):
    """Returns string representation of a message queue.

    Arguments:
    queue    -- queue to be represented as a string
    simtype  -- describes whether the queue is from the
                concreteworld or from the abstract world.
                Valid values are 'conc' or 'abst'
    mask     -- dictionary describing the relevant variables

    Returns a string representation of a message queue.
    """
    s = ''
    mysep = ""
    for msg in queue:
        s += mysep + str(msg.signal.name) + "("
        mysep2 = ""
        for i,p in enumerate(msg.parameters):
            if mask.has_key((simtype, msg.idnum, i)) and mask[(simtype, msg.idnum, i)]:
                s += mysep2 + '*' + p.get_str()
            else:
                s += mysep2 + p.get_str()
            mysep2 = ','
        s += ")"
        mysep = ", "
    return s

def _str_sm_state(indent, sm_state):
    """Returns string representation of a state machine state.

    Arguments:
    indent    -- instance of Utils.String.Indent.
    sm_state  -- state machine state to be represented as a string
                 in Simulator format

    Returns a string representation of a state machine state.
    """
    s = ""
    if sm_state.uml_reference.name:
        s += str(indent) + sm_state.uml_reference.name + "\n"
    else:
        s += str(indent) + "<<anonymous>>\n"
    indent.push()
    for c in sm_state.children:
        s += _str_sm_state(indent, c)
    indent.pop()
    return s

