# coding: latin-1

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""A set of utilities for counterexample analysator.

Author: Vesa Ojala <vojala@tcs.tkk.fi>
"""

import actlang
import Sim
import Sim.Core
import types
import error

class Determine_Return_Type_Error(error.Error):
        def __init__(self, tree, reason):
            self.tree = tree
            self.reason = "determine_return_type: unsupported parse tree\n" +\
                          "  reason: "+reason+"\n" +\
                          "  parse tree type: " + str(tree.__class__.__name__) + "\n"

        def __str__(self):
            return self.reason

def determine_return_type(tree, gs, oid):
    """Determines the type of the expression.

    Arguments:
        tree  -- Jumbala parse tree which type is to be determined
        gs    -- GlobalState where the tree is evaluated
        oid   -- Object id of the object where the tree is evaluated

    Returns the type of the expression. It is one of the following:
        - Sim.Core.SimValue_ObjectRef or,
        - Sim.Core.SimType_Boolean() or,
        - Sim.Core.SimType_Int()

    Throws Determine_Return_Type_Error in the case error occurs.
    """
    if isinstance(tree, actlang.expr.InfixExpr):
        op = tree.api_getOperator()
        if isinstance(op, actlang.binaryop.NumericalOperator):
            return Sim.Core.SimType_Int()
        elif isinstance(op, (actlang.binaryop.NumericalComparison,
                             actlang.binaryop.EqualityOperator)):
            return Sim.Core.SimType_Boolean()

        elif isinstance(op, actlang.binaryop.NumericalBooleanOperator):
            lop = tree.api_getLeftOperand()
            rop = tree.api_getRightOperand()
            ltype = determine_return_type(lop, gs, oid)
            rtype = determine_return_type(rop, gs, oid)
            assert ltype.__class__ == rtype.__class__
            return ltype
        elif isinstance(op, actlang.binaryop.Plus):
            # does not support Jumbala strings because the simulator does not either
            return Sim.Core.SimType_Int()
        else:
            raise Error("unsupperted binaryop")
    elif isinstance(tree, (actlang.expr.CondAndExpr,
                           actlang.expr.CondOrExpr,
                           actlang.expr.BoolLiteralExpr,
                           actlang.expr.InstanceOfExpr,
                           actlang.expr.LogicalNotExpr)):
        return Sim.Core.SimType_Boolean()
    elif isinstance(tree, (actlang.expr.UnaryPlusExpr,
                           actlang.expr.UnaryMinusExpr,
                           actlang.expr.BitNotExpr,
                           actlang.expr.IntLiteralExpr)):
        return Sim.Core.SimType_Int()
    elif isinstance(tree, (actlang.expr.NullExpr)):
        return Sim.Core.SimType_ObjectRef()
    
    elif isinstance(tree, actlang.expr.TernaryCondExpr):
        return determine_return_type(tree.api_getTrueExpr(), gs, oid)
    elif isinstance(tree, actlang.expr.ParenExpr):
        return determine_return_type(tree.api_getInnerExpr(), gs, oid)
    elif isinstance(tree, actlang.names.SimpleName):
        v = get_object(gs, oid).get_variable_value(str(tree))
        return v.get_type()
    elif isinstance(tree, actlang.names.QualifiedName):
        last_id = tree.api_getLastPart()
        qn = tree.api_getQualifyingName()
        obj = follow_references(qn, get_object(gs, oid))
        return obj.get_variable_value(str(last_id)).get_type()
    elif isinstance(tree, actlang.expr.PrimaryFieldAccessExpr):
        last_id = tree.api_getFieldName()
        qn = tree.api_getReferenceExpr()
        obj = follow_references(qn, get_object(gs, oid))
        return obj.get_variable_value(str(last_id)).get_type()
    elif isinstance(tree, actlang.expr.ThisExpr):
        return Sim.Core.SimType_ObjectRef()

    #unsupported features
    elif isinstance(tree, (actlang.expr.SimpleAssignmentExpr,
                           actlang.expr.CompoundAssignmentExpr,
                           actlang.expr.PrefixExpr,
                           actlang.expr.PostfixExpr,
                           actlang.expr.InstanceCreationExpr,
                           actlang.expr.ArrayCreationExpr)):
        raise Determine_Return_Type_Error(tree, "parse tree has side effects")
    elif isinstance(tree, MethodInvocationExpr):
        #superclass for method invocations
        raise Determine_Return_Type_Error(tree, "method invocations not supported")
    elif isinstance(tree, actlang.expr.SuperFieldAccessExpr):
        raise Determine_Return_Type_Error(tree, "superclasses not supported")
    elif isinstance(tree, actlang.expr.ArrayAccessExpr):
        raise Determine_Return_Type_Error(tree, "arrays not supported")
    elif isinstance(tree, actlang.expr.StringLiteralExpr):
        raise Determine_Return_Type_Error(tree, "strings not supported")
    elif isinstance(tree, actlang.expr.CastExpr):
        raise Determine_Return_Type_Error(tree, "type casting not supported")
    elif isinstance(tree, actlang.expr.VarInitializer):
        #superclass for initializers
        raise Determine_Return_Type_Error(tree, "initializers not supported (side-effects)")
    else:
        raise Determine_Return_Type_Error(tree, "unknown tree type")

def get_obj_name_pair(tree, obj):
    """Returns object reference containing the variable in which tree is referring and
    an name of the attribute it is referring.

    Attributes:
        tree  -- Jumbala parse tree containing the variable accessor expression
        obj   -- Object where tree is evaluated

    Returns a tuple (<object>, <variable name>), where <object> is the object where the variable tree
    was referring locates and <variable name> is the name of the variable in <object> in
    which the tree was referring

    For example, if tree given contains the parse tree of Jumbala expression "a.x" and
    obj is the object 1 which has a variable a, the object reference to object 2, then this
    function retuns a tuple (<object 2>, "x")
    """
    if isinstance(tree, actlang.names.SimpleName):
        return (obj, tree.api_getLastPart())
    elif isinstance(tree, actlang.names.QualifiedName):
        obj = follow_references(tree.api_getQualifyingName(), obj)
        return (obj, tree.api_getLastPart())
    #elif isinstance(tree, actlang.expr.ThisExpr):
    #    return obj
    elif isinstance(tree, actlang.expr.PrimaryFieldAccessExpr):
        obj = follow_references(tree.api_getReferenceExpr(), obj)
        return (obj, tree.api_getFieldName())
    elif isinstance(tree, actlang.expr.ParenExpr):
        return get_obj_name_pair(tree.api_getInnerExpr(), obj)
    else:
        raise error.Error("get_obj_name_pair: \n")

def get_checked_obj_name_pair(tree, conc_obj, abst_obj):
    """Works like get_obj_name_pair except that if reference chains do not correspond between
    the abstract world and the concrete world then the variable returned is the mismatching
    reference variable and the object is the concrete object which own the reference variable.

    Attributes:
        tree      -- Jumbala parse tree containing the variable accessor expression
        conc_obj  -- Concrete object where tree is evaluated
        abst_obj  -- Abstract object where tree is evaluated

    Returns a tuple (<object>, <variable name>), where <object> is the concrete object where the
    variable tree was referring locates and <variable name> is the name of the variable in <object> in
    which the tree was referring

    For example, if tree given contains the parse tree of Jumbala expression "a.x" and
    obj is the object 1 which has a variable a, the object reference to object 2, then this
    function retuns a tuple (<object 2>, "x")
    """
    stack = []
    while True:
        if isinstance(tree, actlang.names.SimpleName):
            stack.append(tree)
            break
        elif isinstance(tree, actlang.names.QualifiedName):
            stack.append(tree)
            tree = tree.api_getQualifyingName()
        elif isinstance(tree, actlang.expr.ThisExpr):
            stack.append(tree)
            break
        elif isinstance(tree, actlang.expr.PrimaryFieldAccessExpr):
            stack.append(tree)
            tree = tree.api_getReferenceExpr()
        elif isinstance(tree, actlang.expr.ParenExpr):
            pass
        else:
            print tree.__class__.__name__
            raise error.Error("follow_references: invalid reference chain\n")

    varname = None
    for t in reversed(stack):
        if isinstance(t, actlang.names.SimpleName):
            varname = str(t.api_getLastPart())
        elif isinstance(t, actlang.expr.ThisExpr):
            pass
        elif isinstance(t, (actlang.names.QualifiedName,
                               actlang.expr.PrimaryFieldAccessExpr)):
            value = conc_obj.get_variable_value(varname)
            assert isinstance(value, Sim.Core.SimValue_ObjectRef)
            conc_obj_new = value.get_value()
            
            value = abst_obj.get_variable_value(varname)
            assert isinstance(value, Sim.Core.SimValue_ObjectRef)
            abst_obj_new = value.get_value()

            if conc_obj_new.idnum != abst_obj_new.idnum:
                return (conc_obj, varname)

            conc_obj = conc_obj_new
            abst_obj = abst_obj_new

            if isinstance(t, actlang.names.QualifiedName):
                varname = str(t.api_getLastPart())
            else:
                varname = str(t.api_getFieldName())
        else:
            print t.__class__.__name__
            raise error.Error("follow_references: invalid reference chain\n")
    return (conc_obj, varname)
    
def follow_references(tree, obj):
    """Returns object reference got by evaluating tree in the object obj
    
    Attributes:
        tree  -- Jumbala parse tree containing the object reference chain to be followed
        obj   -- The object which is the starting point

        Returns an object resulting from evaluating tree in the obj
    """
    if isinstance(tree, actlang.names.SimpleName):
        value = obj.get_variable_value(str(tree.api_getLastPart()))
        assert isinstance(value, Sim.Core.SimValue_ObjectRef)
        return value.get_value()
    elif isinstance(tree, actlang.names.QualifiedName):
        obj = follow_references(tree.api_getQualifyingName(), obj)
        value = obj.get_variable_value(str(tree.api_getLastPart()))
        assert isinstance(value, Sim.Core.SimValue_ObjectRef)
        return value.get_value()
    elif isinstance(tree, actlang.expr.ThisExpr):
        return obj
    elif isinstance(tree, actlang.expr.PrimaryFieldAccessExpr):
        obj = follow_references(tree.api_getReferenceExpr(), obj)
        value = obj.get_variable_value(str(tree.api_getFieldName()))
        assert isinstance(value, Sim.Core.SimValue_ObjectRef)
        return value.get_value()
    elif isinstance(tree, actlang.expr.ParenExpr):
        return follow_references(tree.api_getInnerExpr(), obj)
    else:
        print tree.__class__.__name__
        raise error.Error("follow_references: invalid reference chain\n")
    
def get_object(gs, oid):
    """Returns object in the global state given that has a ID given.

    Attributes:
        gs    -- Global state where the search is conducted
        oid   -- ID of the object that is search for

    Returns an object with a given ID, if not found None is returned.
    """
    for obj in gs.objects:
        if obj.idnum == oid:
            return obj
    return None

def collect_var_names(tree, obj):
    """Returns a set of all variables appearing in the tree. Variables in
    the set are represented by tuples containging an object that own the
    variable and a variable name of the variable.

    Attributes:
        tree  -- Jumbala parse tree where the variables are collected from
        obj   -- The context where tree is located

    Returns a set of tuples (<object>, <variable name>), where <object>
    is the object where the variable tree was referring locates and
    <variable name> is the name of the variable in <object> in which the
    tree was referring.
    """
    if subexpression_accessors.has_key(tree.__class__):
        r = set()
        for accessor in subexpression_accessors[tree.__class__]:
            r.update(collect_var_names(accessor(tree), obj))
        return r
    elif isinstance(tree, actlang.names.SimpleName):
        return set([(obj, tree.api_getLastPart())])
    elif isinstance(tree, actlang.names.QualifiedName):
        last_id = tree.api_getLastPart()
        qn = tree.api_getQualifyingName()
        obj = follow_references(qn, obj)
        return set([(obj, last_id)])
    elif isinstance(tree, actlang.expr.PrimaryFieldAccessExpr):
        last_id = tree.api_getFieldName()
        qn = tree.api_getReferenceExpr()
        obj = follow_references(qn, obj)
        return set([(obj, last_id)])
    else:
        raise error.Error("collect_var_names: accessor not found")
    
def collect_relevant_var_names(tree, conc_obj, abst_obj):
    """Works like collect_var_names function except makes sure that
    every reference in a reference chain leading to the actual variable
    takes to corresponding objects in the abstract and in the concrete
    worlds. If references do not match in these worlds, then the mismatching
    reference type variable is returned in the set instead of the one that
    wuold be returned in the collect_var_names. Variables in
    the set are represented by tuples containging an object that own the
    variable and a variable name of the variable.

    Attributes:
        tree  -- Jumbala parse tree where the variables are collected from
        obj   -- The context where tree is located

    Returns a set of tuples (<object>, <variable name>), where <object>
    is the object where the variable tree was referring locates and
    <variable name> is the name of the variable in <object> in which the
    tree was referring.
    """
    if subexpression_accessors.has_key(tree.__class__):
        r = set()
        for accessor in subexpression_accessors[tree.__class__]:
            r.update(collect_relevant_var_names(accessor(tree), conc_obj, abst_obj))
        return r
    elif isinstance(tree, (actlang.names.SimpleName,
                           actlang.names.QualifiedName,
                           actlang.expr.PrimaryFieldAccessExpr)):
        obj, vname = get_checked_obj_name_pair(tree, conc_obj, abst_obj)
        return set([(obj, vname)])
    else:
        raise error.Error("collect_var_names: accessor not found")
    
subexpression_accessors = {
    # not supported: ArrayCreationExpr
#    actlang.expr.ArrayCreationExpr : (),
    actlang.expr.BitNotExpr : (actlang.expr.BitNotExpr.api_getOperand,),
    actlang.expr.BoolLiteralExpr : (),

    # subclasses of CastExpr
    #actlang.expr.NameCastExpr :
    #(actlang.expr.NameCastExpr.api_getTypeExpr,),
    #actlang.expr.TypeCastExpr :
    #(actlang.expr.TypeCastExpr.api_getOperand,),
    # end subclasses of CastExpr

    actlang.expr.CondAndExpr :
    (actlang.expr.CondAndExpr.api_getLeftOperand,
     actlang.expr.CondAndExpr.api_getRightOperand),
    actlang.expr.CondOrExpr :
    (actlang.expr.CondOrExpr.api_getLeftOperand,
     actlang.expr.CondOrExpr.api_getRightOperand),
    actlang.expr.InfixExpr :
    (actlang.expr.InfixExpr.api_getLeftOperand,
     actlang.expr.InfixExpr.api_getRightOperand),
    actlang.expr.InstanceOfExpr :
    (actlang.expr.InstanceOfExpr.api_getOperand,),
    actlang.expr.IntLiteralExpr: (),

    # subclasses of LeftHandSideExpr
    #actlang.expr.ArrayAccessExpr :
    #(actlang.expr.ArrayAccessExpr.api_getArrayExpr,
    # actlang.expr.ArrayAccessExpr.api_getIndexExpr),
    
    # not supported: PrimaryFieldAccessExpr
    # actlang.expr.PrimaryFieldAccessExpr :
    # (actlang.expr.PrimaryFieldAccessExpr.api_getFieldName,),
    # not supported: SuperFieldAccessExpr
    # end subclasses of LeftHandSideExpr

    actlang.expr.LogicalNotExpr :
    (actlang.expr.LogicalNotExpr.api_getOperand,),
    actlang.expr.NullExpr : (),
    actlang.expr.ParenExpr : (actlang.expr.ParenExpr.api_getInnerExpr,),

    # subclasses of StatementExpr
    actlang.expr.CompoundAssignmentExpr :
    (actlang.expr.CompoundAssignmentExpr.api_getLhsExpr,
     actlang.expr.CompoundAssignmentExpr.api_getRhsExpr),
    # subclasses of IncrementOrDecrementExpr
    actlang.expr.PostfixExpr : (actlang.expr.PostfixExpr.api_getOperand,),
    actlang.expr.PrefixExpr : (actlang.expr.PrefixExpr.api_getOperand,),
    # end subclasses of IncrementOrDecrementExpr
    # not supported: InstanceCreationExpr
    # subclasses of MethodInvocationExpr
    # not supported: NameMethodInvocationExpr
    # not supported: PrimaryMethodInvocationExpr
    # not supported: SuperMethodInvocationExpr
    # end subclasses of MethodInvocationExpr
    actlang.expr.SimpleAssignmentExpr :
    (actlang.expr.SimpleAssignmentExpr.api_getLhsExpr,
     actlang.expr.SimpleAssignmentExpr.api_getRhsExpr),
    # end subclasses of StatementExpr

    actlang.expr.StringLiteralExpr : (),
    actlang.expr.TernaryCondExpr :
    (actlang.expr.TernaryCondExpr.api_getCondition,
     actlang.expr.TernaryCondExpr.api_getTrueExpr,
     actlang.expr.TernaryCondExpr.api_getFalseExpr),
    actlang.expr.ThisExpr : (),
    actlang.expr.UnaryMinusExpr :
    (actlang.expr.UnaryMinusExpr.api_getOperand,),
    actlang.expr.UnaryPlusExpr :
    (actlang.expr.UnaryPlusExpr.api_getOperand,)

    # subclasses of VarInitializer
    # ArrayInitializer: not supported
    # ExprInitializer: not supported
    # end subclasses of VarInitializer

}
    
    
