#! /usr/bin/python

"""
State machine action language flattener

Removes for and while loops from the action language statements appearing in
the state machines of a UML model.

Author: Tommi Junttila

Copyright (c) 2006-2007 Helsinki University of Technology
Released under the GNU General Public License version 2
"""

import sys
import optparse

import coral

import Sim
import ActionLanguageFlattener
import ActionLanguageFlattener.Main
import Sim.SMToDot
import Utils.Models
from Utils.Models import get_qualified_name

UML14 = coral.meta.UML14





def flatten(project,
            opt_single_statement = False,
            opt_atmost_one_send_statement = False,
            opt_dotty_filename = "",
            stdout = sys.stdout, stderr = sys.stderr):
    """
    Flatten the action language statements in all the state machines in
    the UML Model of the Coral project 'project'.

    If 'opt_single_statement' is True,
    then each transition will have only one statement in its effect.

    If 'opt_atmost_one_send_statement' is True,
    then each transition will have at most one send statement in its effect.

    If 'opt_dotty_filename' is a non-empty string,
    then the state machines are output in the graphviz dotty format in files
    'opt_dotty_filename'_smX.dot and 'opt_dotty_filename'_smX_alflattened.dot,
    where X is the number of the state machine.
    """
    assert isinstance(project, coral.kernel.Project)
    assert type(opt_single_statement) is bool
    assert type(opt_atmost_one_send_statement) is bool

    def _verb(s):
        if stdout:
            print >> stdout, s
    
    #
    # Find the model
    #
    models = [m for m in project.contents() if m.oclIsKindOf(UML14.Model)]
    if not models:
        raise RuntimeError("The project does not contain a UML Model")
    if len(models) > 1:
        raise RuntimeError("The project contains "+str(len(models))+
                           " UML Models instead of one")
    model = models[0]

    # Find all StateMachines in the model
    state_machines = Utils.Models.get_elements_of_type(model,
                                                       UML14.StateMachine)
    Utils.Models.xmi_sort(state_machines)

    TM = coral.getTransactionManager()
    TM.beginModifyCmd()

    for (index,sm) in enumerate(state_machines):
        if opt_dotty_filename:
            dottyfilename = opt_dotty_filename+"_sm"+str(index)+".dot"
            _verb("Producing dotty file '"+dottyfilename+"' "+
                  "for state machine '"+get_qualified_name(sm)+"'")
            f = open(dottyfilename, 'w')
            Sim.SMToDot.SMToDot(f, sm)
            f.close()

        if not sm.context.oclIsKindOf(UML14.Class):
            raise RuntimeError("StateMachine '"+get_qualified_name(sm)+"' "+
                               "is not in class context")
        if not sm in sm.context.behavior:
            raise RuntimeError("StateMachine '"+get_qualified_name(sm)+"' "+
                               "is not a behavior of the owning class")
        flat_sm = ActionLanguageFlattener.Main.flatten(sm,
                                                       opt_single_statement = opt_single_statement,
                                                       opt_atmost_one_send_statement = opt_atmost_one_send_statement)
        

        context = sm.context
        context.behavior.remove(sm)
        context.ownedElement.remove(sm)
        context.behavior.append(flat_sm)
        context.ownedElement.append(flat_sm)
    
        if opt_dotty_filename:
            dottyfilename = opt_dotty_filename+"_sm"+str(index)+\
                            "_alflattened.dot"
            _verb("Producing dotty file '"+dottyfilename+"' "+
                  "for the action language flattened version of "+
                  "state machine '"+get_qualified_name(flat_sm)+"'")
            f = open(dottyfilename, 'w')
            Sim.SMToDot.SMToDot(f, flat_sm)
            f.close()

    TM.endModifyCmd()



if __name__ == '__main__':
    usage = """python %prog [options] inputmodel.xmi outputmodel.xmi

Removes if, for, while, do, break, and continue constructs
from the action language statements appearing in
the state machines of the model in the file inputmodel.xmi.
Writes the resulting model in the file outputmodel.xmi."""

    p = optparse.OptionParser(usage=usage)
    p.add_option('--single-statement',
                 action='store_true', default=False,
                 help='extra flattening: '
                 'each transition will have at most one statement')
    p.add_option('--atmost-one-send-statement',
                 action='store_true', default=False,
                 help='extra flattening: '
                 'each transition will have at most one send statement')
    p.add_option('--dotty-file', type='string', default="", metavar='FILE',
                 help='print the state machines in dotty format '
                 'to files starting with FILE [default: no dotty output]')

    (opt, args) = p.parse_args()
    if len(args) != 2:
        p.error('wrong number of arguments')

    inputfilename = args[0]
    outputfilename = args[1]

    P = coral.loadDocument(inputfilename)

    flatten(P,
            opt_single_statement = opt.single_statement,
            opt_atmost_one_send_statement = opt.atmost_one_send_statement,
            opt_dotty_filename = opt.dotty_file)

    P.setFilename(outputfilename)
    coral.saveDocument(P)
