#! /usr/bin/python

"""
Flatten the hierarchy of all state machines appearing in a model.
In addition, the state machines are normalized
so that each simple state either has (i) no outgoing transitions,
(ii) only signal triggered outgoing transitions, or (iii) only
outgoing completion transitions with the additional quarantee that
at least one of them will always be enabled.

Author: Tommi Junttila

Copyright (c) 2006-2007 Helsinki University of Technology
Released under the GNU General Public License version 2
"""

import sys
import optparse

import coral

import Sim
import Sim.Flattener
import Sim.SMToDot
import Utils.Models
from Utils.Models import get_qualified_name

UML14 = coral.meta.UML14

def flatten(model,
            opt_dotty_filename = None,
            stdout=sys.stdout, stderr=sys.stderr):
    """
    Flattens the hierarchy of all state machines appearing in the UML Model
    'model'.
    In addition, the state machines are normalized so that
    each simple state either has
    (i) no outgoing transitions,
    (ii) only signal triggered outgoing transitions, or
    (iii) only outgoing completion transitions with the additional quarantee
    that at least one of them will always be enabled.

    If 'opt_dotty_filename' is a non-empty string,
    then the state machines are output in the graphviz dotty format in files
    'opt_dotty_filename'_smX.dot and 'opt_dotty_filename'_smX_flattened.dot,
    where X is the number of the state machine.
    """
    assert model.oclIsKindOf(UML14.Model)

    def _verb(s):
        if stdout:
            print >> stdout, s

    state_machines = [sm for sm in model.__getAllParts__()
                      if sm.oclIsKindOf(UML14.StateMachine)]

    TM = coral.getTransactionManager()
    TM.beginModifyCmd()

    for (index,sm) in enumerate(state_machines):
        if not sm.context.oclIsKindOf(UML14.Class):
            raise RuntimeError("StateMachine "+get_qualified_name(sm)+
                               " given in a non-class context")
        if not sm in sm.context.behavior:
            raise RuntimeError("StateMachine "+get_qualified_name(sm)+
                               " is not a behavior of the owning class")

        if opt_dotty_filename:
            dottyfilename = opt_dotty_filename+"_sm"+str(index)+".dot"
            _verb("Producing dotty file '"+dottyfilename+"' "+
                  "for state machine '"+get_qualified_name(sm)+"'")
            f = open(dottyfilename, 'w')
            Sim.SMToDot.SMToDot(f, sm)
            f.close()

        flat_sm = Sim.Flattener.flatten(sm, stdout=stdout, stderr=stderr)

        # Change sm to flat_sm in the context of sm
        context = sm.context
        context.behavior.remove(sm)
        context.ownedElement.remove(sm)
        context.behavior.append(flat_sm)
        context.ownedElement.append(flat_sm)

        if opt_dotty_filename:
            dottyfilename = opt_dotty_filename+"_sm"+str(index)+\
                            "_flattened.dot"
            _verb("Producing dotty file '"+dottyfilename+"' "+
                  "for the flattened version of "+
                  "state machine '"+get_qualified_name(flat_sm)+"'")
            f = open(dottyfilename, 'w')
            Sim.SMToDot.SMToDot(f, flat_sm)
            f.close()

    TM.endModifyCmd()



if __name__ == "__main__":
    usage = """python %prog [options] inputmodel.xmi outputmodel.xmi

Flattens the hierarchy of all state machines appearing in the model in
the file 'inputmodel.xmi'. In addition, the state machines are normalized
so that each simple state either has (i) no outgoing transitions,
(ii) only signal triggered outgoing transitions, or (iii) only
outgoing completion transitions with the additional quarantee that
at least one of them will always be enabled.

The resulting model is written in the file 'outputmodel.xmi'.
"""
    p = optparse.OptionParser(usage=usage)
    p.add_option('--dotty-file', type='string', default="", metavar='FILE',
                 help='print the state machines in dotty format '
                 'to files starting with FILE [default: no dotty output]')

    (opt, args) = p.parse_args()
    if len(args) != 2:
        p.error('wrong number of arguments')

    P = coral.loadDocument(args[0])

    #
    # Find the model
    #
    models = [e for e in P.contents() if e.oclIsKindOf(UML14.Model)]
    if len(models) != 1:
        raise RuntimeError("The project contains "+str(len(models))+
                           " models instead of one - aborting!")
    model = models[0]
    flatten(model,
            opt_dotty_filename = opt.dotty_file)
    #
    # Save the new model with flattened state machines
    #
    P.setFilename(args[1])
    coral.saveDocument(P)
