#! /usr/bin/python2.4

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Usage: flatten_transition_effects.py INFILE OUTFILE
Flatten action language expressions in effects of transitions of state
machines in a UML model contained in a XMI 2.0 project file INFILE, and
generate a new project file and model (with possibly new temporary
variables added to its classes) as OUTFILE.  The model contained in the
input file should first have been processed through the state machine
hierarchy and action language flatteners (if necessary).

"Flattening" an expression means translating the expression to a series
of assignments to temporary variables introduced for subexpressions of
the expression.  For example, the expression "((y + z) - w) * q" could
be flattened to a plain reference to a temporary variable "tmp_2"
initialized using the sequence of assignments
"tmp_0 = (y + z); tmp_1 = tmp_0 - w; tmp_2 = tmp_1 * q;".

This script will flatten only action language expressions of integer type;
for example, the topmost "Boolean part" of any expression is left intact
(integer subexpressions of Boolean expressions are still flattened, though).

Options:
  -h, --help                 Display this help and exit.

This script was written by Heikki Tauriainen for the SMUML project
<http://www.tcs.hut.fi/Research/Logic/SMUML.shtml>.
"""

import sys
import getopt
import coral
import actlang
from abstraction import error
from abstraction.typeinference import TypeInferenceSystem
from Utils.Models import get_elements_of_type
from Utils import ExpressionTraverser
UML = coral.lang.UML14

TMPVAR_PREFIX = "tmp_"

def __generateExpressionFlattener():
    """Generate a function object for flattening act. lang. expressions.

    The function object returned by this function can be called with the
    following parameters (in order)
        1. An actlang.expr.Expr object.
        2. A dictionary, which should be empty.
        3. A mapping from action language expression objects to nodes in
           a type inference graph (needed for determining the type of
           the expression given as the first parameter).
        4. A list to be filled with assignment statements needed for
           initializing temporary variables introduced while flattening
           (when calling the function object from outside, use an empty
           list).
        5. A list, the first element of which gives the smallest
           nonnegative integer which (or any larger integer) can be used
           as an identifier for a new temporary variable.

    The function object returns an actlang.expr.Expr object representing the
    flattened version of the action language expression given as the first
    parameter.  As a side effect, the list given as the fourth argument
    will be populated with a list of actlang.statement.ExprStatement
    objects representing the assignments needed to initialize the temporary
    variables referred to in the flattened expression.
    """

    def tmpvar(stmts, id, expr):
        """Introduce a new temporary variable for an expression.

        Arguments:
            stmts -- A list of statements in which to append an assignment
                     statement for initializing a new temporary variable
                     using the given expression.
            id    -- A list, the first element of which gives the smallest
                     nonnegative integer which can be used (along with any
                     larger integer) as an identifier for a new temporary
                     variable.
            expr  -- An actlang.expr.Expr object.

        Returns an actlang.names.SimpleName object representing a reference
        to the new temporary variable.  As side effects, (i) the `stmts'
        list will be updated with an assignment statement for initializing
        the new temporary variable using `expr', and (ii) the first element
        of `next_free_tmpvar_id' will be incremented.
        """
        var_expr = actlang.names.SimpleName(TMPVAR_PREFIX + str(id[0]), (0, 0))
        id[0] += 1
        stmts.append \
            (actlang.statement.ExprStatement \
                 (actlang.expr.SimpleAssignmentExpr(var_expr, expr, (0, 0)),
                  (0, 0)))
        return var_expr

    # Callbacks for flattening action language expressions.

    # Pre-callback for consulting the temporary variable cache.

    def recurse(expr, result_cache, *rest):
        try:
            raise ExpressionTraverser.Return(result_cache[str(expr)])
        except KeyError:
            pass

    # Pre-callback for atomic expressions.

    def atomic(expr, *rest):
        raise ExpressionTraverser.Return(expr)

    # Post-callbacks.

    def paren(expr, subexprs, *rest):
        return actlang.expr.ParenExpr(subexprs[0], (0, 0))

    def assignment(expr, subexprs, result_cache, expr_node, stmts, id):
        # an assignment expression is always the "outermost" expression
        # (guaranteed by initialization of the type inference system)
        #if stmts:
        #    # optimize away one temporary variable
        #    e = stmts[-1].api_getExpr()
        #    assert(subexprs[1] == e.api_getLhsExpr())
        #    subexprs[1] = e.api_getRhsExpr()
        #    stmts.pop()
        #    id[0] -= 1
        return actlang.expr.SimpleAssignmentExpr \
                   (subexprs[0], subexprs[1], (0, 0))

    def cond(expr, subexprs, *rest):
        return expr.__class__(*subexprs)

    def infix(expr, subexprs, result_cache, expr_node, stmts, id):
        e = actlang.expr.InfixExpr \
                (subexprs[0], subexprs[1], expr.api_getOperator())
        if expr_node[expr].model_type == "int":
            raise ExpressionTraverser.Return(tmpvar(stmts, id, e))
        return e

    def default(expr, subexprs, result_cache, expr_node, stmts, id):
        subexprs.append((0, 0))
        e = expr.__class__(*subexprs)
        if expr_node[expr].model_type == "int":
            raise ExpressionTraverser.Return(tmpvar(stmts, id, e))
        return e

    def backtrack(expr, retval, result_cache, *rest):
        result_cache[str(expr)] = retval

    e = actlang.expr
    n = actlang.names
    pre_callbacks = \
        (("entry", recurse),
         ((e.BoolLiteralExpr, e.IntLiteralExpr, n.SimpleName,
           n.QualifiedName, e.PrimaryFieldAccessExpr, e.ThisExpr,
           e.InstanceCreationExpr, e.NullExpr), atomic),)
    post_callbacks = \
        ((e.ParenExpr, paren),
         (e.SimpleAssignmentExpr, assignment),
         ((e.CondAndExpr, e.CondOrExpr, e.TernaryCondExpr), cond),
         (e.InfixExpr, infix),
         ((e.UnaryPlusExpr, e.UnaryMinusExpr, e.BitNotExpr, e.LogicalNotExpr),
          default),
         ("exit", backtrack))
    flattener = \
        ExpressionTraverser.ExpressionTraverser(ExpressionTraverser.ignore)
    for (cls, f) in pre_callbacks:
        flattener.register_pre_callback(cls, f)
    for (cls, f) in post_callbacks:
        flattener.register_post_callback(cls, f)
    return flattener

__flatten = __generateExpressionFlattener()
del __generateExpressionFlattener

def flatten(expr, expr_node, id):
    """Flatten an action language expression.

    Arguments:
        expr      -- An actlang.expr.Expr object.
        expr_node -- A mapping from action language expression objects to
                     nodes in a type inference graph (needed for
                     determining the type of `expr' and its subexpressions).
        id        -- A list, the first element of which gives the smallest
                     nonnegative integer which (or any larger integer) can
                     be used as an identifier for a new temporary variable.

    Returns a 2-tuple containing a list of assignment statements needed to
    initialize temporary variables in the flattened expression, and the
    flattened expression.  As a side effect, the first element of `id'
    will be updated to the next free identifier for a new temporary
    variable.
    """
    stmts = []
    expr = __flatten(expr, {}, expr_node, stmts, id)
    return (stmts, expr)

def processModel(model, stdout=sys.stdout, stderr=sys.stderr):
    """Flatten effects of transitions in model.

    Argument:
        model -- A coral.kernel.ElementPtr representing a UML Model.

    Raises an error.Error if there is a problem processing the model.
    """
    if not isinstance(model, coral.kernel.ElementPtr):
        raise TypeError("a coral.kernel.ElementPtr expected as argument")
    if not model.oclIsKindOf(UML.Model):
        raise ValueError("an UML Model expected as argument")

    # get a handle to the "int" UML Primitive in the model
    int_datatype = None
    datatypes = get_elements_of_type(model, UML.Primitive)
    for t in datatypes:
        if str(t.name) == "int":
            if int_datatype:
                raise error.Error \
                          ("multiple definitions of Primitive `int' in model")
            int_datatype = t

    # validate the model and determine the types of expressions in the model
    ti = TypeInferenceSystem(model)
    expr_node = ti._TypeInferenceSystem__expr_node

    # rewrite effects of transitions of state machines in the model

    tm = coral.getTransactionManager()
    tm.beginModifyCmd()

    S = actlang.statement

    for statemachine in ti._TypeInferenceSystem__statemachines:
        max_tmpvar_id = 0
        for state in statemachine.top.subvertex:
            for t in state.outgoing:
                try:
                    if hasattr(t, 'effect') and t.effect:
                        flattened_effect = []
                        stmts = \
                            ti._TypeInferenceSystem__uml_to_statement[t.effect]
                        max_tmpvar_id_within_effect = 0
                        for s in stmts:
                            tmpvar_id = [0]
                            if isinstance \
                                   (s, (S.AssertStatement, S.ExprStatement)):
                                (tmpvar_init_stmts, expr) = \
                                    flatten \
                                        (s.api_getExpr(), expr_node, tmpvar_id)
                                flattened_effect.extend(tmpvar_init_stmts)
                                flattened_effect.append \
                                    (s.__class__(expr, (0, 0)))
                            elif isinstance(s, S.EmptyStatement):
                                pass
                            elif isinstance(s, S.SendStatement):
                                (target_init_stmts, target_expr) = \
                                    flatten \
                                        (s.api_getTarget(), expr_node,
                                         tmpvar_id)
                                flattened_effect.extend(target_init_stmts)
                                parm_exprs = []
                                for e in s.api_getArgList():
                                    (parm_init_stmts, parm_expr) = \
                                        flatten(e, expr_node, tmpvar_id)
                                    flattened_effect.extend(parm_init_stmts)
                                    parm_exprs.append(parm_expr)
                                flattened_effect.append \
                                    (S.SendStatement
                                         (s.api_getSignalName(), parm_exprs,
                                          target_expr, (0, 0)))
                            else:
                                raise error.Error \
                                          ("unsupported type of statement `%s'"
                                           % str(s))
                            if max_tmpvar_id_within_effect < tmpvar_id[0]:
                                max_tmpvar_id_within_effect = tmpvar_id[0]
                            if max_tmpvar_id < tmpvar_id[0]:
                                max_tmpvar_id = tmpvar_id[0]
                        # reset all temporary variables
                        for i in xrange(0, max_tmpvar_id_within_effect):
                            flattened_effect.append \
                                (TMPVAR_PREFIX + "%d = 0;" % i)
                        t.effect.script.body = \
                            " ".join(map(str, flattened_effect))
                except error.Error, e:
                    msg = \
                        str(e) + \
                        "\n    while processing transition `%s' from state " \
                        "`%s' to state `%s'" \
                        % (str(t.name), str(t.source.name), str(t.target.name))
                    raise error.Error(msg)
        # introduce temporary variables in the class associated with the state
        # machine
        if not statemachine.context.oclIsKindOf(UML.Class):
            raise error.Error("context of state machine must be a UML Class")
        for i in xrange(0, max_tmpvar_id):
            if int_datatype == None:
                raise error.Error \
                          ("missing definition of the int Primitive in model")
            varname = TMPVAR_PREFIX + str(i)
            for f in statemachine.context.feature:
                if f.oclIsKindOf(UML.Attribute) and str(f.name) == varname:
                    raise error.Error \
                        ("name of variable `%s' in model conflicts with a "
                         "name of a temporary" % varname)
            statemachine.context.feature.append \
                (UML.Attribute
                     (name=varname,
                      multiplicity=UML.Multiplicity
                                       (range=[UML.MultiplicityRange
                                                   (lower=1, upper=1)]),
                      visibility="public",
                      changeability="changeable",
                      initialValue=UML.Expression(body="0"),
                      type=int_datatype))

    tm.endModifyCmd()
# -----------------------------------------------------------------------------
# The command-line script

if __name__ == "__main__":
    # process command line arguments
    try:
        (options, args) = getopt.getopt(sys.argv[1:], "h", ["help"])
    except getopt.GetoptError, e:
        error.abort(str(e), 2)

    for (option, value) in options:
        if option in ("-h", "--help"):
            sys.stdout.write(__doc__)
            sys.exit(0)

    if len(args) < 1:
        error.abort("input file name expected", 2)
    if len(args) < 2:
        error.abort("output file name expected", 2)
    if len(args) > 2:
        error.abort("extra arguments on command line", 2)

    # load the model
    try:
        project = coral.loadDocument(args[0])
    except Exception, e:
        error.abort(str(e))

    # get a handle to the UML Model in the project file
    model = None
    for x in project.contents():
        if x.oclIsKindOf(UML.Model):
            if not model:
                model = x
            else:
                error.abort("multiple model definitions in input file")
    if not model:
        error.abort("no model found in input file")

    try:
        processModel(model)
    except error.Error, e:
        error.abort(str(e))

    # save the modified model
    try:
        project.setFilename(args[1])
        coral.saveDocument(project)
    except Exception, e:
        error.abort(str(e))
