# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Model checking back-end specifications.

This module defines classes for storing information about model checking
back-ends.

Author: Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

from string import Template
from abstraction.smtlib_typegen import PySmt

class Backend:
    """Data specific to all back-ends."""

    def __init__(self):
        self.arguments = []

class Uboco(Backend):
    """Data specific to the Uboco back-end."""
    
    DEFAULT_BMC_BOUND = 30
    NUSMV_SAT_SOLVERS = ("ZChaff", "MiniSat", "Sim")

    def __init__(self):
        Backend.__init__(self)
        self.use_bmc = True
        self.sat_solver = self.NUSMV_SAT_SOLVERS[0]
        self.bmc_bound = self.DEFAULT_BMC_BOUND

    def getNuSmvInputScript(self):
        if self.use_bmc:
            input_template = self.NUSMV_input_with_bmc
            # use incremental mode for ZChaff
            check_bmc_command = \
                self.sat_solver == "ZChaff" \
                and "check_ltlspec_bmc_inc" or "check_ltlspec_bmc"
        else:
            input_template = self.NUSMV_input_without_bmc
            check_bmc_command = ""
        return input_template.substitute \
                   ({"NUSMV_BMC_BOUND" : str(self.bmc_bound),
                     "NUSMV_SAT_SOLVER" : self.sat_solver,
                     "NUSMV_CHECK_BMC" : check_bmc_command})

    NUSMV_input_with_bmc = Template("""set input_file model.smv
set on_failure_script_quits 1
set bmc_length ${NUSMV_BMC_BOUND}
set bmc_loopback X
set sat_solver ${NUSMV_SAT_SOLVER}
go_bmc
${NUSMV_CHECK_BMC}
echo -o nusmv.out no trace
show_traces -o nusmv.out
quit
""")

    NUSMV_input_without_bmc = Template("""set input_file model.smv
set on_failure_script_quits 1
go
check_ltlspec
echo -o nusmv.out no trace
show_traces -o nusmv.out
quit
""")

class Suboco(Backend):
    """Data specific to the Suboco back-end."""

    DEFAULT_BMC_MAX_BOUND = 999999
    SMT_SOLVERS = PySmt.SMT_SOLVERS

    def __init__(self):
        Backend.__init__(self)
        self.smt_solver = "Yices" # default solver: Yices

class Proco(Backend):
    """Data specific to the Proco back-end."""

    DEFAULT_MAX_SEARCH_DEPTH = 10000

    def __init__(self):
        Backend.__init__(self)
        self.max_search_depth = self.DEFAULT_MAX_SEARCH_DEPTH
