# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Helper functions for model checking and counterexample analysis.

This module provides helper functions for model checking and
counterexample analysis.

Author: Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

import os, sys
from os.path import isfile, abspath
import logging
from abstraction import error
from frontend import backends
from frontend.logutil import logCompletion, logFailure
from frontend.subtaskwrapper import PIPE, runSubtask
import uboco_trace_to_generic_trace

def analyzeModel(model_file_name, backend, getPath):
    """Run model checker on a model file, testing for a counterexample.

    This function acts as a wrapper to run model checking on a given model
    file using a given back-end model checker.

    Arguments:
        model_file_name -- Name of the file to analyze.
        backend         -- A back-end object (of type backends.Uboco,
                           backends.Suboco, or backends.Proco).
        getPath         -- Function for determining commands to run
                           external programs.

    If model checking is successful, returns None.  If there is a
    counterexample, returns the absolute path of a trace file which
    contains the counterexample in the generic counterexample format.

    Raises an error.Error if model checking fails due to an error.
    """
    logger = logging.getLogger("log")
    logger.info(("analyzing model", [], True))
    logger.startActivity("Analyzing model")
    try:    
        if isinstance(backend, backends.Uboco):
            # run Uboco:
            # <uboco> [arguments] <model> model.smv
            args = getPath("PYTHON")
            args.extend(getPath("uboco"))
            args.extend(backend.arguments)
            args.extend((model_file_name, "model.smv"))
            runSubtask \
                (args, "generating NuSMV model", capture_prefix="uboco",
                 stdout=PIPE, stderr=PIPE)

            # create a script of commands to run for NuSMV
            logger.debug("creating NuSMV input script...")
            try:
                nusmv_script_file = open("nusmv.scr", "w")
                try:
                    nusmv_script_file.write(backend.getNuSmvInputScript())
                finally:
                    nusmv_script_file.close()
            except Exception, e:
                logFailure(e)
                raise error.Error("aborted")
            logCompletion()

            # run NuSMV:
            # <NuSMV> -int -load nusmv.scr
            args = getPath("nusmv")
            args.extend(("-int", "-load", "nusmv.scr"))
            runSubtask \
                (args, "running NuSMV", capture_prefix="nusmv", stdout=PIPE,
                 stderr=PIPE)

            # check whether there is a counterexample trace
            logger.debug("checking whether there is a counterexample...")
            try:
                trace_file = open("nusmv.out")
                try:
                    line = trace_file.readline()
                finally:
                    trace_file.close()
                if line.find("no trace") != -1:
                    logCompletion("no counterexample found")
                    return None
                elif line.find("Trace number:") == -1:
                    raise Exception("error parsing output from NuSMV")
            except Exception, e:
                logFailure(e)
                raise error.Error("aborted")
            logCompletion("yes")

            # translate counterexample to generic format
            try:
                generic_trace_file = open("counterexample.trace", "w")
            except Exception, e:
                raise error.Error \
                          ("aborted: could not create file for generic "
                           "counterexample trace")
            try:
                runSubtask \
                    (uboco_trace_to_generic_trace.translate,
                     "translating counterexample to generic format",
                     capture_prefix="translate_nusmv_trace",
                     stdout=generic_trace_file, stderr=PIPE,
                     args=("model.smv", "nusmv.out"))
                return abspath("counterexample.trace")
            finally:
                generic_trace_file.close()

        elif isinstance(backend, backends.Suboco):
            # run Suboco:
            # <suboco> [arguments] --trace-file counterexample.trace
            #     --<smt solver> <smt solver path> <model>
            args = getPath("PYTHON")
            args.extend(getPath("suboco"))
            args.extend(backend.arguments)
            args.extend \
                (("--trace-file", "counterexample.trace",
                  "--%s" % backend.smt_solver, getPath(backend.smt_solver)[0],
                  model_file_name))
            runSubtask \
                (args, "running Suboco", capture_prefix="suboco", stdout=PIPE,
                 stderr=PIPE)

            # check whether there is a counterexample trace
            logger.debug("checking whether there is a counterexample...")
            if not isfile("counterexample.trace"):
                logCompletion("no counterexample found")
                return None
            logCompletion("yes")
            return abspath("counterexample.trace")

        elif isinstance(backend, backends.Proco):
            # run newproco:
            # <newproco> [arguments] --dont-print-cex --spin <spin path>
            #     --compiler <compiler path> --trace-filename
            #     counterexample.trace <model>
            args = getPath("PYTHON")
            args.extend(getPath("proco"))
            args.extend(backend.arguments)
            args.extend \
                (("--dont-print-cex", "--spin", " ".join(getPath("spin")),
                  "--compiler", " ".join(getPath("CC")), "--trace-filename",
                  "counterexample.trace", model_file_name))
            runSubtask \
                (args, "running Proco", capture_prefix="proco", stdout=PIPE,
                 stderr=PIPE, ignore_exit_status=True)

            # check whether there is a counterexample trace
            logger.debug("checking whether there is a counterexample...")
            if not isfile("counterexample.trace"):
                logCompletion("no counterexample found")
                return None
            logCompletion("yes")
            return abspath("counterexample.trace")

        else:
            raise error.Error("unsupported model checking back-end")
    finally:
        logger.dedent(logging.INFO)
        logger.stopActivity()

def findRefinement(trace_file_name, refiner, config):
    """Analyze the feasibility of a counterexample to find a refinement.

    This function runs the counterexample feasibility analyzer on a
    counterexample.  If the counterexample is spurious, the function
    suggests a refinement to rule out the counterexample.

    Arguments:
        trace_file_name     -- Name of a file containing a counterexample
                               trace (in generic format).
        refiner             -- A frontend.refinementcycle.ModelRefiner
                               object.
        config              -- A frontend.Configuration object.

    Returns None if the abstract counterexample in `trace_file_name' is
    feasible.  Otherwise, returns a list of pairs (<UUID>, <values>),
    where UUID is the XMI id of an attribute in the concrete model in
    file `config.model_file_name' and <values> is a list of integers with
    which to refine the abstract type of the attribute.
    """
    logger = logging.getLogger("log")
    logger.info(("analyzing counterexample", [], True))
    logger.startActivity("Analyzing counterexample")
    try:
        # save the type library
        # [note: the file name of the type library must be the same as the one
        # attached to the abstract model in "refiner.abstract"]
        logger.debug("saving type library...")
        try:
            refiner.partitions.getTypeLibrary().write("abstract_type_library")
        except Exception, e:
            logFailure(e)
            raise error.Error("aborted")
        logCompletion()

        # run counterexample analysis:
        # <c.ex.analyzer> <trace file> <concrete model>
        args = config.getPath("PYTHON")
        args.extend(config.getPath("analyze_counterexample"))
        args.extend(config.feasibility_analyzer_args)
        args.extend((trace_file_name, config.model_file_name))
        runSubtask \
            (args, "running counterexample analyzer",
             capture_prefix="feasibility_analysis", stdout=PIPE, stderr=PIPE)

        # process the refinement information from `canal.out'
        refinement = []
        logger.debug("checking whether the counterexample is feasible...")
        try:
            refinement_file = open("canal.out")
            try:
                for (i, l) in enumerate(refinement_file):
                    l = l.replace("\n", "")
                    if i == 0:
                        if l == "true counterexample":
                            logCompletion("yes")
                            return None
                        if l != "Refinement hint:":
                            raise Exception("error parsing `canal.out'")
                    else:
                        (uuid, vals) = map(str.strip, l.rsplit(":", 1))
                        vals = map(int, vals.split(","))
                        refinement.append((uuid, vals))
            finally:
                refinement_file.close()
        except Exception, e:
            logFailure(e)
            raise error.Error("aborted")
        logCompletion("no")
        return refinement
    finally:
        logger.dedent(logging.INFO)
        logger.stopActivity()

def processCounterExampleTrace(model_file, trace_file, output_file, simulate, gui_mode, getPath):
    """Display and (optionally) save and/or simulate a counterexample trace.

    Arguments:
        model_file  -- Name of the model file to which the counterexample
                       trace applies.
        trace_file  -- Name of a counterexample trace file.
        output_file -- The name of the file into which the counterexample
                       trace should be copied (in text mode, "-" refers to
                       standard output).  The value None will prevent
                       saving/outputting the counterexample trace.
        simulate    -- If True, run the simulator on the counterexample;
                       display results on standard output (text mode) or
                       the log (graphical mode).
        gui_mode    -- Whether the script is being run in graphical mode or
                       not.
        getPath     -- Function for determining commands to run external
                       programs.
    """
    logger = logging.getLogger("log")

    def toFile(f):
        """Get a function for writing a line of text to a file."""
        return lambda line: f.write(line)

    def toLog(line):
        """Pass a line of text to the logger."""
        logger.debug((line, ["small", "monospace"]))

    def mapFile(filename, operation):
        """Apply an operation to all lines of a file."""
        input_file = open(filename, "r")
        try:
            for line in input_file:
                operation(line)
        finally:
            input_file.close()

    # if running in graphical mode, add the counterexample trace to the log
    if gui_mode:
        logger.debug(("Counterexample trace", [], True))
        try:
            try:
                mapFile(trace_file, toLog)
            except Exception, e:
                logFailure(e, "error processing counterexample trace")
        finally:
            logger.dedent(logging.DEBUG)

    # save or output the generic counterexample trace if requested
    if output_file != None:
        logger.startActivity("Saving counterexample trace")
        try:
            try:
                if not gui_mode and output_file == "-":
                    # in text mode, the special filename "-" means stdout
                    logger.debug("outputting counterexample trace...")
                    output_file = sys.stdout
                else:
                    logger.debug \
                        ("saving counterexample trace to `%s'..."
                         % output_file)
                    output_file = open(output_file, "w")
                try:
                    mapFile(trace_file, toFile(output_file))
                finally:
                    if output_file != sys.stdout:
                        output_file.close()
                logCompletion()
            except Exception, e:
                logFailure(e)
        finally:
            logger.stopActivity()

    # simulate the counterexample if necessary
    if simulate:
        logger.startActivity("Simulating counterexample")
        try:
            args = getPath("PYTHON")
            args.extend(getPath("simulate_counterexample"))
            args.extend((trace_file, model_file))
            try:
                output_file = open("simulate_counterexample.stdout", "w")
                try:
                    runSubtask \
                        (args, "running counterexample simulator",
                         capture_prefix="simulate_counterexample",
                         ignore_exit_status=True, stdout=output_file.fileno(),
                         stderr=PIPE)
                finally:
                    if not output_file.closed:
                        output_file.close()
                if gui_mode:
                    # if running in graphical mode, add the output of the
                    # simulation to the log
                    logger.debug(("Simulated trace", [], True))
                    try:
                        try:
                            mapFile("simulate_counterexample.stdout", toLog)
                        except Exception, e:
                            logFailure \
                                (e,
                                 "failed to process output from "
                                 "counterexample simulator")
                    finally:
                        logger.dedent(logging.DEBUG)
                else:
                    # else write the output of simulator to standard output
                    try:
                        mapFile \
                            ("simulate_counterexample.stdout",
                             toFile(sys.stdout))
                    except Exception, e:
                        logger.error \
                            ("failed to process output from counterexample "
                             "simulator [%s]" % str(e))
            except Exception, e:
                logger.error("failed to simulate counterexample")
        finally:
            logger.stopActivity()
