# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Configuration information for the model analyzer front-end.

This module defines a class for storing configuration information for
the model analyzer front-end.

Author: Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

import sys
import os
from os.path import abspath
import logging
import getopt
import re
from abstraction import error
from abstraction.smtlib_typegen import PySmt
from frontend import backends, logutil

class Configuration:
    """Configuration data."""

    # Available semantics for generating operations for abstract types
    ABSTRACT_TYPE_SEMANTICS = ("Bounded", "Unbounded", "Mixed")
    # Available SMT solvers for generating abstract types under bounded/mixed
    # integer semantics
    ABSTRACT_TYPEGEN_SOLVERS = PySmt.SMT_SOLVERS

    def __init__(self, help_text=""):
        """Initialize a Configuration object from command-line arguments.

        Argument:
            help_text -- Text to display when the -h or --help option is
                         specified.

        Raises an error.OptionError if there are errors in the command-line
        arguments.
        """
        self.cfg_file_name = "paths.conf"
        self.cfg_file_is_a_tempfile = False
        self.model_file_name = None
        self.backend = None
        self.gui_mode = False
        self.integer_semantics = "bounded"
        self.use_refinement = False
        self.flatten_hierarchy = False
        self.flatten_action_language = False
        self.flatten_actlang_sends = False
        self.flatten_actlang_all = False
        self.slice_assertions = False
        self.flatten_transition_effects = False
        self.typegen_smt_solver = "Yices"
        self.flattening_limit = 0
        self.simulate_counterexample = False
        self.trace_file_name = None
        self.loglevel = logging.WARN
        self.paths = {}
        self.extra_arguments = []
        self.feasibility_analyzer_args = []

        try:
            (options, self.extra_arguments) = \
                getopt.getopt \
                    (sys.argv[1:], "hv",
                         ("abstract-type-smt-solver=", "backend=",
                          "configuration-file=", "counterexample-trace-file=",
                          "expression-flattening-limit=",
                          "flatten-action-language",
                          "flatten-action-language-split-sends",
                          "flatten-action-language-single-statement",
                          "flatten-hierarchy", "flatten-effect-expressions",
                          "gui", "help", "integer-semantics=",
                          "no-action-language-flattening",
                          "no-counterexample-simulation",
                          "no-effect-expression-flattening",
                          "no-hierarchy-flattening", "no-refinement",
                          "no-slicing", "quiet", "refinement",
                          "simulate-counterexample", "silent",
                          "slice-assertions", "verbose"))
        except getopt.GetoptError, e:
            raise error.OptionError(str(e))

        for (option, value) in options:
            if option == "--abstract-type-smt-solver":
                try:
                    self.typegen_smt_solver = \
                        self.ABSTRACT_TYPEGEN_SOLVERS \
                            [map(str.lower, self.ABSTRACT_TYPEGEN_SOLVERS) \
                                 .index(value.lower())]
                except ValueError:
                    raise error.OptionError \
                              ("unknown SMT solver `%s' specified for the "
                               "--abstract-type-smt-solver option" % value)
            elif option == "--backend":
                value_lower = value.lower()
                if value_lower == "proco":
                    self.backend = backends.Proco()
                elif value_lower == "uboco" or value_lower[0:6] == "uboco:":
                    self.backend = backends.Uboco()
                    if len(value_lower) > 5:
                        mode_and_optional_bound = value[6:].split(":")
                        if len(mode_and_optional_bound) > 1:
                            (mode, bound) = mode_and_optional_bound
                        else:
                            (mode, bound) = \
                                (mode_and_optional_bound[0],
                                 backends.Uboco.DEFAULT_BMC_BOUND)
                        mode_lower = mode.lower()
                        if mode_lower == "bdd":
                            self.backend.sat_solver = None
                            self.backend.use_bmc = False
                        else:
                            try:
                                self.backend.sat_solver = \
                                    backends.Uboco.NUSMV_SAT_SOLVERS \
                                        [map(str.lower,
                                             backends.Uboco.NUSMV_SAT_SOLVERS)
                                         .index(mode_lower)]
                                self.backend.use_bmc = True
                            except ValueError:
                                raise error.OptionError \
                                          ("unknown mode `%s' specified for "
                                           "the --backend:uboco option" % mode)
                        try:
                            self.backend.bmc_bound = int(bound)
                            if self.backend.bmc_bound <= 0:
                                raise ValueError()
                        except ValueError:
                            raise error.OptionError \
                                      ("invalid BMC bound `%s' specified "
                                       "for the --backend:uboco option"
                                       % bound)
                elif value_lower == "suboco" or value_lower[0:7] == "suboco:":
                    self.backend = backends.Suboco()
                    if len(value_lower) > 6:
                        solver = value[7:]
                        try:
                            self.backend.smt_solver = \
                                backends.Suboco.SMT_SOLVERS \
                                    [map(str.lower,
                                         backends.Suboco.SMT_SOLVERS)
                                     .index(solver.lower())]
                        except ValueError:
                            raise error.OptionError \
                                      ("unknown SMT solver `%s' specified for "
                                       "the --backend:suboco option" % solver)
                else:
                    raise error.OptionError \
                              ("unknown back-end `%s' specified for the "
                               "--backend option" % value)
            elif option == "--configuration-file":
                self.cfg_file_name = abspath(value)
            elif option == "--counterexample-trace-file":
                if value != "-":
                    self.trace_file_name = abspath(value)
                else:
                    self.trace_file_name = None
            elif option == "--expression-flattening-limit":
                if value.lower() == "none":
                    self.flattening_limit = -1
                else:
                    try:
                        self.flattening_limit = int(value)
                        if value < 0:
                            raise ValueError()
                    except ValueError:
                        raise error.OptionError \
                                  ("the value for "
                                   "--expression-flattening-limit must be a "
                                   "nonnegative integer")
            elif option == "--flatten-action-language":
                self.flatten_action_language = True
            elif option == "--flatten-action-language-split-sends":
                self.flatten_action_language = True
                self.flatten_actlang_sends = True
                self.flatten_actlang_all = False
            elif option == "--flatten-action-language-single-statement":
                self.flatten_action_language = True
                self.flatten_actlang_sends = False
                self.flatten_actlang_all = True
            elif option == "--no-action-language-flattening":
                self.flatten_action_language = False
            elif option == "--flatten-effect-expressions":
                self.flatten_transition_effects = True
            elif option == "--no-effect-expression-flattening":
                self.flatten_transition_effects = False
            elif option == "--flatten-hierarchy":
                self.flatten_hierarchy = True
            elif option == "--no-hierarchy-flattening":
                self.flatten_hierarchy = False
            elif option == "--gui":
                self.gui_mode = True
            elif option == "--integer-semantics":
                v = value.lower()
                if not v in map(str.lower, self.ABSTRACT_TYPE_SEMANTICS):
                    raise error.OptionError \
                              ("unknown integer semantics `%s' specified for "
                               "the --integer-semantics option" % value)
                self.integer_semantics = v
            elif option in ("-h", "--help"):
                sys.stdout.write(help_text)
                sys.exit(0)
            elif option == "--refinement":
                self.use_refinement = True
            elif option == "--no-refinement":
                self.use_refinement = False
            elif option == "--simulate-counterexample":
                self.simulate_counterexample = True
            elif option == "--no-counterexample-simulation":
                self.simulate_counterexample = False
            elif option in ("--quiet", "--silent"):
                self.loglevel = logging.CRITICAL
            elif option == "--slice-assertions":
                self.slice_assertions = True
            elif option == "--no-slicing":
                self.slice_assertions = False
            elif option in ("-v", "--verbose"):
                if self.loglevel == logging.ERROR:
                    self.loglevel = logging.WARN
                elif self.loglevel == logging.WARN:
                    self.loglevel = logging.INFO
                elif self.loglevel == logging.INFO:
                    self.loglevel = logging.DEBUG
                else:
                    self.loglevel = logutil.LOG_EXTRA_DEBUG
        if self.extra_arguments:
            self.model_file_name = abspath(self.extra_arguments.pop(0))

    def finalize(self):
        """Finalizes a Configuration object.

        This method finalizes a Configuration object by making sure that
        (i) a model checking back-end and at least one --check option for
        the back-end has been selected, and (ii) initializing the external
        programs known to the Configuration object by reading a paths
        configuration file currently registered with the Configuration
        object.  As a side effect, the methods logs some
        informative messages about the configuration using the root logger.

        A paths configuration file should consist of lines of the form
            <key>=<value>
        where <key> matches the regular expression [A-Za-z_][A-Za-z_0-9]*
        Comments should be preceded by the character `#'.  If <value>
        begins or ends with white space, it should be enclosed in single
        or double quotes.  <value> can refer to keys defined previously
        in the file or in the global environment by using the syntax
        $<key> or ${<key>}.  The values from the environment override
        values defined in the path configuration file.

        Raises an error.Error if there is an error processing the paths
        configuration file.
        """
        logger = logging.getLogger("log")

        if self.backend == None:
            logger.info("no back-end specified, using Uboco")
            self.backend = backends.Uboco()
        else:
            logger.info \
                ("using %s as model checking back-end"
                 % self.backend.__class__.__name__)

        supported_check_options = \
            ["--check-deadlock", "--check-implicit-consumption",
             "--check-assertions"]

        if isinstance(self.backend, backends.Proco):
            supported_check_options.append("--check-queue-overflows")
        else:
            supported_check_options.append("--check-runtime-errors")

        if self.use_refinement:
            for arg in self.extra_arguments:
                if len(arg) >= 9 and "--check-deadlock".startswith(arg):
                    logger.warn \
                        ("abstraction may introduce spurious deadlocking "
                         "computations which cannot be simulated")
                    break
        check_option_found = False
        self.feasibility_analyzer_args = []
        for arg in self.extra_arguments:
            if len(arg) >= 9:
                for option in supported_check_options:
                    if option.startswith(arg):
                        check_option_found = True
                        if option == "--check-assertions":
                            self.feasibility_analyzer_args.append \
                                ("--check-assert")
                        elif option == "--check-implicit-consumption":
                            self.feasibility_analyzer_args.append(option)
        if not check_option_found:
            logger.warn \
                ("no --check parameter supported by the chosen back-end "
                 "supplied, assuming --check-assertions")
            self.extra_arguments.append("--check-assertions")
            self.feasibility_analyzer_args.append("--check-assert")

        if isinstance(self.backend, backends.Proco):
            pass
        elif isinstance(self.backend, backends.Uboco):
            if self.backend.use_bmc:
                logger.info \
                    ("using %s as SAT solver for NuSMV"
                     % self.backend.sat_solver)
            else:
                logger.info("using BDD-based exhaustive model checking")
        else: # Suboco
            logger.info \
                ("using %s as SMT solver for Suboco" % self.backend.smt_solver)
            self.backend.smt_solver = self.backend.smt_solver.lower()

        if self.use_refinement:
            logger.info \
                ("using abstraction refinement with %s abstract type semantics"
                 % self.integer_semantics)
            if self.integer_semantics != "unbounded":
                logger.info \
                    ("using %s as solver for generating abstract operations"
                     % self.typegen_smt_solver)
            if self.flattening_limit < 0:
                logger.info("flattening all admissible abstract operations")
            if self.flattening_limit > 0:
                logger.info \
                    ("using %d as the limit for flattening abstract "
                     "operations" % self.flattening_limit)

        # if running in text mode and no trace file was specified, output
        # counterexample trace (if any) to standard output
        if not self.gui_mode and self.trace_file_name == None:
            self.trace_file_name = "-"

        self.backend.arguments = self.extra_arguments

        # read the path configuration file
        logger.info("using paths configuration file `%s'" % self.cfg_file_name)

        def translateValue(token):
            def retrieveDefinition(matchobj):
                varname = matchobj.group(1) or matchobj.group(2)
                try:
                    return os.environ[varname]
                except KeyError:
                    try:
                        return " ".join(self.paths[varname])
                    except KeyError:
                        return ""
    
            return re.sub \
                       ("\$([A-Za-z_][A-Za-z_0-9]*)|"
                        "\${([A-Za-z_][A-Za-z_0-9]*)}",
                        retrieveDefinition, token)

        self.paths = {}
        try:
            cfg_file = open(self.cfg_file_name, "r")
            try:
                for l in cfg_file:
                    l = l.split("#", 1)[0].strip()
                    if not l:
                        continue
                    tokens = map(str.strip, l.split("=", 1))
                    key = tokens[0]
                    if re.match("[A-Za-z_][A-Za-z_0-9]*", key) == None:
                        raise Exception("invalid configuration key `%s'" % key)
                    try:
                        self.paths[key] = os.environ[key].split()
                    except KeyError:
                        if len(tokens) > 1:
                            value = tokens[1]
                            if len(value) >= 2 and \
                                   ((value[0] == value[-1] == '"')\
                                    or (value[0] == value[-1] == "'")):
                                value = value[1:-1]
                        else:
                            value = ""
                        self.paths[key] = translateValue(value).split()
            finally:
                cfg_file.close()
        except Exception, e:
            raise error.Error(str(e))

    def getPath(self, key):
        """Find the command required to run an external program.

        Argument:
            key -- Key to query.

        Returns the command line for the program identified by `key' in
        the configuration as a list of strings (program executable
        and command-line arguments).  Raises an error.Error if there is
        no external program registered with key `key' in the configuration.
        """
        try:
            value = self.paths[key][:]
            if not value:
                raise KeyError()
            return value
        except KeyError:
            raise error.Error \
                      ("no path specified for program `%s' in the path "
                       "configuration file or in the environment" % key)
