# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Classes for logging output in various forms.

This module defines classes used for logging purposes.

Author: Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

import os, sys, logging, time
from datetime import datetime
from threading import Thread, Condition

# Additional logging levels
LOG_EXTRA_DEBUG = 5
LOG_DEBUG_SUCCESS = 11
LOG_DEBUG_FAILURE = 12

def logCompletion(description="done"):
    """Log the completion of an activity (with level LOG_DEBUG_SUCCESS).

    Argument:
        description -- A message indicating the completion of an
                       activity.
    """
    logging.getLogger("log").log(LOG_DEBUG_SUCCESS, " %s\n" % description)

def logFailure(reason, description="failed"):
    """Log the completion of an activity (with level LOG_DEBUG_FAILURE).

    Arguments:
        reason      -- A more specific reason for the failure.  Can be
                       any object which supports the __str__ method.
        description -- A message indicating the completion of an
                       activity.
    """
    logger = logging.getLogger("log")
    logger.log(LOG_DEBUG_FAILURE, " %s" % description)
    reason = str(reason)
    if reason:
        logger.debug(" [%s]" % reason)
    logger.debug("\n")

class Logger(logging.Logger):
    """Interface for loggers used in the front-end."""

    def __init__(self, name=""):
        """Initialize a Logger.

        Argument:
            name -- Name for the logger in the logger hierarchy.
        """
        logging.Logger.__init__(self, name)

    def extradebug(self, *args, **kwargs):
        """Method for logging messages at level LOG_EXTRA_DEBUG."""
        self.log(LOG_EXTRA_DEBUG, *args, **kwargs)

    def initDecorations(self, *args, **kwargs):
        """Method for preparing the logger for logging output."""
        pass

    def finish(self, *args, **kwargs):
        """Method for signaling the end of logging."""
        pass

    def cleanUpDecorations(self, *args, **kwargs):
        """Method for cleaning up any resources needed by the logger."""
        pass

    def startActivity(self, *args, **kwargs):
        """Method for signaling the start of an activity."""
        pass

    def stopActivity(self, *args, **kwargs):
        """Method for signaling the end of an activity."""
        pass

    def logMessage(self, *args, **kwargs):
        """For internal use.  Do not call this method directly."""
        pass

    def dedent(self, *args, **kwargs):
        """Method for decreasing the logger's indentation level."""
        pass

    def askQuestion(self, *args, **kwargs):
        """Method for asking a yes/no question from the user."""
        pass

class StderrLogger(Logger):
    """Logger for logging text messages into the standard error stream."""
    
    def __init__(self, name):
        """Initialize StderrLogger.

        Argument:
            name -- Name of the logger in the logger hierarchy.
        """
        Logger.__init__(self, name)
        # do not propagate anything up the logger hierarchy
        self.propagate = False
        # current indentation level
        self.indent = 0
        # whether the next message to be output should be printed with
        # added indentation
        self.begin_indented_block = False
        # whether the last outputted message contained a newline
        self.newline = False

    def logMessage(self, msg, lvl, tags, prepare_indented_block=False):
        """Log a message to sys.stderr.

        Arguments:
            msg                    -- Message to be logged.
            lvl                    -- Log level of the message.
            tags                   -- Dummy argument to make the method
                                      signature compatible with that
                                      expected by `LogHandler.emit'.
            prepare_indented_block -- If True, any text logged after `msg'
                                      (but before a call to `dedent')
                                      should be placed into a new indented
                                      block.
        """
        if lvl >= logging.INFO:
            msg += "\n"
            if lvl == logging.INFO:
                msg = "%s:info:%s" % (sys.argv[0], msg)
            elif lvl == logging.WARNING:
                msg = "%s:warning:%s" % (sys.argv[0], msg)
            elif lvl == logging.ERROR:
                msg = "%s:error:%s" % (sys.argv[0], msg)
            elif lvl == logging.CRITICAL:
                msg = "%s:critical:%s" % (sys.argv[0], msg)
            else:
                msg = "%s:log:%s" % (sys.argv[0], msg)
        try:
            if lvl >= logging.INFO:
                # never indent log messages with level >= logging.INFO;
                # these messages always end in a newline 
                self.indent = 0
                self.newline = True
                self.begin_indented_block = prepare_indented_block
                if prepare_indented_block:
                    self.indent += 4
                sys.stderr.write(msg)
                return

            if self.begin_indented_block:
                if not self.newline:
                    # if the previous outputted message does not end in a
                    # newline, start a new line
                    sys.stderr.write("\n")
                    self.newline = True
                self.begin_indented_block = False

            if self.newline:
                self.newline = False
                sys.stderr.write(" " * self.indent)

            if prepare_indented_block:
                # `msg' is the "title" of a new block of text to be (possibly)
                # indented
                sys.stderr.write(msg)
                self.newline = False
                self.begin_indented_block = True
                self.indent += 4
            else:
                if msg and msg[-1] == "\n":
                    # if `msg' ends in a newline, do not indent after the
                    # trailing newline (but indent all other text lines)
                    msg = \
                         ("\n" + " " * self.indent).join(msg[:-1].split("\n"))\
                         + "\n"
                    self.newline = True
                else:
                    msg = ("\n" + " " * self.indent).join(msg.split("\n"))
                    self.newline = False
                sys.stderr.write(msg)
        finally:
            sys.stderr.flush()

    def dedent(self, lvl, output=()):
        """Decrease the indentation level of log messages.

        Arguments:
            lvl    -- Log level of the indented text block to be closed.
            output -- A sequence of (text, tags) pairs to be appended
                      to the "title" of the text view to be closed.
                      Here, `text' is a string of text, and `tags' is
                      a list of tags defined in GraphicalLogger.__init__.
        """
        if not self.isEnabledFor(lvl):
            # if the logger's log level is higher than the log level of the
            # block to be closed, do nothing
            return
        self.indent -= 4
        if not self.begin_indented_block:
            # there was actual text inserted at the higher indent level
            pass
        else:
            # no text was actually inserted after the last indented block
            # "title"
            self.begin_indented_block = False
        if self.newline:
            sys.stderr.write(" " * self.indent)
        if output:
            for (text, tags) in output:
                sys.stderr.write(text)
                self.newline = text and text[-1] == "\n"
        if not self.newline:
            sys.stderr.write("\n")
        self.newline = True

    def askQuestion(self, question):
        """Ask a yes/no question (defaulting to yes).

        Argument:
            question -- The question as a string.

        Returns True iff the answer does not begin with `n' or `N'.
        """
        sys.stdout.write("%s? [y] " % question)
        answer = raw_input()
        return answer and not (answer[0] in ("n", "N"))

class GraphicalLogger(Logger):
    """Class for logging in graphical mode."""

    def __init__(self, name):
        """Initialize a GraphicalLogger.

        Argument:
            name -- Name of the logger w.r.t. the root logger.
        """
        global gobject, gtk, pygtk, pango
        import os, gobject, gtk, pygtk, pango
        Logger.__init__(self, name)
        self.propagate = False
        gobject.threads_init()
        # the GUI
        PADDING = 5
        self.dialog = gtk.Dialog("SMUML Model Analyzer", flags=0)
        self.dialog.set_default_size(640, 480)
        self.dialog.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_NORMAL)
        self.save_button = self.dialog.add_button("_Save log as...", 2)
        self.save_button.connect("clicked", self.__saveLog)
        self.close_button = \
            self.dialog.add_button(gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE)
        table = gtk.Table(3, 1)
        table.set_border_width(PADDING)
        table.set_col_spacings(PADDING)
        table.set_row_spacings(PADDING)
        subtable = gtk.Table(6, 6)
        subtable.set_border_width(PADDING)
        subtable.set_col_spacings(4 * PADDING)
        subtable.set_row_spacings(4 * PADDING)
        self.progress_label = gtk.Label()
        subtable.attach(self.progress_label, 1, 5, 1, 2, gtk.EXPAND, 0)
        self.progress_time = gtk.Label()
        subtable.attach(self.progress_time, 1, 5, 2, 3, gtk.EXPAND, 0)
        self.progress_bar = gtk.ProgressBar()
        subtable.attach \
            (self.progress_bar, 1, 5, 3, 5, gtk.EXPAND | gtk.FILL, 0)
        table.attach(subtable, 0, 1, 0, 1, gtk.EXPAND | gtk.FILL, gtk.FILL)
        expander = gtk.Expander()
        expander.set_use_markup(True)
        expander.set_label('<span size="small"><i>Show messages</i></span>')
        subtable = gtk.Table(2, 2)
        self.root_text_view = gtk.TextView()
        self.root_text_view.set_wrap_mode(False)
        self.root_text_view.set_editable(False)
        self.root_text_view.set_overwrite(False)
        self.root_text_view.set_cursor_visible(False)
        self.__initColors(self.root_text_view)
        self.root_text_view_buffer = self.root_text_view.get_buffer()
        self.root_text_view_buffer.create_tag("red", foreground="#FF0000")
        self.root_text_view_buffer.create_tag("orange", foreground="#F08010")
        self.root_text_view_buffer.create_tag("green", foreground="#008000")
        self.root_text_view_buffer.create_tag("blue", foreground="#0000FF")
        self.root_text_view_buffer.create_tag("gray", foreground="#808080")
        self.root_text_view_buffer.create_tag("bold", weight=pango.WEIGHT_BOLD)
        self.root_text_view_buffer.create_tag \
            ("italic", style=pango.STYLE_ITALIC)
        self.root_text_view_buffer.create_tag \
            ("underline", underline_set=True, underline=pango.UNDERLINE_SINGLE)
        self.root_text_view_buffer.create_tag \
            ("small", scale_set=True, scale=pango.SCALE_SMALL)
        self.root_text_view_buffer.create_tag \
            ("monospace", family_set=True, family="Monospace")
        self.root_text_view_buffer.create_mark \
            ("end", self.root_text_view_buffer.get_start_iter(), False)
        self.root_text_view_buffer.insert \
            (self.root_text_view_buffer.get_start_iter(), "\n")
        self.root_text_view_buffer.create_mark \
            ("end_text", self.root_text_view_buffer.get_start_iter(), False)
        self.text_view_window = gtk.ScrolledWindow()
        self.text_view_window.set_policy \
            (gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.text_view_window.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        self.text_view_window.add(self.root_text_view)
        subtable.attach(self.text_view_window, 0, 2, 0, 1)
        self.scroll_checkbox = gtk.CheckButton("Scroll on output")
        self.scroll_checkbox.set_active(True)
        subtable.attach(self.scroll_checkbox, 0, 1, 1, 2, gtk.FILL, 0)
        self.expand_checkbox = \
            gtk.CheckButton("Expand new branches automatically")
        self.expand_checkbox.set_active(False)
        self.expand_checkbox.connect("toggled", self.__expandBranch)
        subtable.attach(self.expand_checkbox, 1, 2, 1, 2, gtk.FILL, 0)
        expander.add(subtable)
        expander.connect \
            ("notify::expanded",
             lambda expander, param_spec:
                 expander.set_label
                     ('<span size="small"><i>%s messages</i></span>'
                      % (expander.get_expanded() and "Hide" or "Show")))
        expander.set_expanded(True)
        table.attach(expander, 0, 1, 1, 3)
        table.show_all()
        self.dialog.vbox.add(table)

        # GUI thread
        self.gui_thread = None
        # directory for file chooser used for saving log files
        self.log_directory = os.getcwd()
        # timer for pulsing the progress bar
        self.progress_bar_timer = None
        # progress bar pulser thread
        self.pulser_thread = None
        # whether the pulser should be stopped
        self.stop_pulser_thread = False
        # time of the first call to startActivity since the last call to
        # `initDecorations'
        self.total_activity_start_time = None
        # time of the latest call to startActivity
        self.activity_start_time = None
        # the text view hierarchy and current location
        self.text_view_hierarchy = {}
        self.text_view_hierarchy_location = [0]
        # the currently active text view and text buffer
        self.text_view = self.root_text_view
        self.text_view_buffer = self.root_text_view_buffer
        # stack for controlling indentation: each item in the stack is a pair
        # of references to a gtk.TextView and a gtk.TextBuffer representing a
        # text view and the title of an indented subblock contained in that
        # text view
        self.indentation_stack = []
        # whether a new indented text view should be created on the next call
        # to `logMessage'
        self.begin_indented_block = False
        # reference to a gtk.Table containing the most recently added text view
        # item which represents the title of an indented block of text
        self.indented_block_title = None
        # a list of gtk.Expanders representing the bottom branch in the tree
        # of text views
        self.bottom_branch = []
        # whether to save full or reduced logs
        self.save_reduced_log = False
        # condition variable for synchronizing the logger and GUI threads
        self.logger_cv = Condition()

    def initDecorations(self):
        """Prepare the graphical logger for logging output.

        This method resets the logger's internal variables, resets the
        sensitivities of widgets in the logger's GUI, and starts the GUI.
        """
        self.total_activity_start_time = None
        self.indentation_stack = []
        self.begin_indented_block = False
        self.indented_block_title = None
        self.bottom_branch = []
        self.root_text_view_buffer.delete \
            (self.root_text_view_buffer.get_start_iter(),
             self.root_text_view_buffer.get_iter_at_mark \
                 (self.root_text_view_buffer.get_mark("end_text")))
        self.text_view_hierarchy = {}
        self.text_view_hierarchy_location = [0]
        self.text_view = self.root_text_view
        self.text_view_buffer = self.root_text_view_buffer
        self.save_button.set_sensitive(False)
        self.close_button.set_sensitive(False)
        self.expand_checkbox.set_sensitive(True)
        self.scroll_checkbox.set_sensitive(True)
        self.gui_thread = Thread(target=self.__gui, args=())
        self.gui_thread.start()

    def finish(self, text="Finished"):
        """Method to be called at the end of any output to be logged.

        This method sets the logger GUI into a "final state" by displaying
        a message to indicate that all processing has been completed,
        showing the total time elapsed since the first call to
        `startActivity' after the latest call to `initDecorations',
        setting the progress bar to indicate that all processing has been
        completed, enabling the GUI's Save as... and Close buttons, and
        waiting for the GUI thread to finish (which occurs when the user
        selects the Close button).

        Argument:
            text -- Message to indicate the end of logging.
        """
        self.stopActivity()
        gobject.idle_add(self.__finish_in_gui_thread, text)
        self.gui_thread.join()

    def __finish_in_gui_thread(self, text):
        """Method used in `finish' to update GUI in the GUI thread."""
        self.progress_label.set_markup \
            ('<span size="large"><b>%s</b></span>' % text)
        self.progress_bar.set_fraction(1.0)
        for widget in (self.scroll_checkbox, self.expand_checkbox):
            widget.set_sensitive(False)
        for widget in (self.save_button, self.close_button):
            widget.set_sensitive(True)
        if self.total_activity_start_time:
            self.progress_time.set_text \
                (time.strftime
                     ("Total elapsed time: %H:%M:%S",
                      time.gmtime
                          (time.time() - self.total_activity_start_time)))

    def cleanUpDecorations(self):
        """Hide the logger's GUI.

        This method should be called after calling `finish'.  After
        calling this method, the graphical logger is ready to be
        started again by calling `initDecorations'.
        """
        gobject.idle_add(self.dialog.hide)

    def startActivity(self, text):
        """Update the logger's progress label with an activity description.

        Argument:
            text -- Message to display.

        This method initializes the label of the logger's progress bar
        with a user-defined message (to indicate the starting of some
        activity), records the start time for the activity, and starts
        a thread for pulsing the progress bar until the method
        `stopActivity' is called.
        """
        gobject.idle_add(self.__startActivity_in_gui_thread, text)
        self.activity_start_time = time.time()
        if not self.total_activity_start_time:
            # this is the first activity started since the last call to
            # `initDecorations'
            self.total_activity_start_time = self.activity_start_time
        # start a thread for pulsing the progress bar
        self.stop_pulser_thread = False
        self.pulser_thread = Thread(target=self.__pulse)
        self.pulser_thread.start()

    def __startActivity_in_gui_thread(self, text):
        """Method used in `startActivity' to update GUI in the GUI thread."""
        # update the progress label
        self.progress_label.set_markup \
            ('<span size="large"><b>%s</b></span>' % text)

    def stopActivity(self):
        """Reset the progress bar and its label to show no activity.

        This method resets the progress bar and its label to show no
        activity.  If there is a thread running which updates the
        progress bar, stop the thread.
        """
        if self.pulser_thread:
            # if there is a pulser thread currently active, stop it
            self.stop_pulser_thread = True
            self.pulser_thread.join()
            self.pulser_thread = None
        gobject.idle_add(self.__stopActivity_in_gui_thread)

    def __stopActivity_in_gui_thread(self):
        """Method used in `stopActivity' to update GUI in the GUI thread."""
        self.progress_label.set_text("")
        self.progress_time.set_text("")

    def logMessage(self, msg, lvl, tags, prepare_indented_block=False):
        """Log a message in the logger GUI.

        Arguments:
            msg                    -- Message to be logged.
            lvl                    -- Log level of the message.
            tags                   -- A list of tags to apply to the text
                                      (one or more of the tags defined in
                                      GraphicalLogger.__init__).
            prepare_indented_block -- If True, any text logged after `msg'
                                      (but before a call to `dedent')
                                      should be placed into a new indented
                                      block (in which case `msg' acts as
                                      a label for the indented block).
        """
        if lvl >= logging.INFO:
            msg = msg[0].upper() + msg[1:] + "\n"
        if lvl == LOG_DEBUG_FAILURE:
            tags.extend(("red", "italic"))
        elif lvl == LOG_DEBUG_SUCCESS:
            tags.extend(("green", "italic"))
        elif lvl == logging.INFO:
            tags.extend(("blue", "bold"))
        elif lvl == logging.WARN:
            tags.extend(("orange", "bold"))
            msg = "Warning: " + msg
        elif lvl == logging.ERROR:
            tags.extend(("red", "bold"))
        self.logger_cv.acquire()
        gobject.idle_add \
            (self.__logMessage_in_gui_thread, msg, tags,
             prepare_indented_block)
        self.logger_cv.wait()
        self.logger_cv.release()

    def __logMessage_in_gui_thread(self, msg, tags, prepare_indented_block):
        """Method used in `logMessage' to update the GUI in the GUI thread."""
        if self.begin_indented_block:
            # a new indented block should be constructed for the text to be
            # inserted
            self.begin_indented_block = False
            # create an expander which opens a new indented text view
            # (which replaces the currently active text view)
            expander = gtk.Expander()
            table = gtk.Table(1, 2)
            self.__initColors(table)
            l = gtk.Label()
            l.set_padding(expander.style_get_property("expander-size"), 0)
            table.attach(l, 0, 1, 0, 1, 0, 0)
            self.text_view_buffer = \
                gtk.TextBuffer(self.root_text_view_buffer.get_tag_table())
            self.text_view = gtk.TextView(self.text_view_buffer)
            self.text_view.set_editable(False)
            self.text_view.set_wrap_mode(False)
            self.text_view.set_cursor_visible(False)
            self.__initColors(self.text_view)
            table.attach(self.text_view, 1, 2, 0, 1)
            expander.add(table)
            expander.set_expanded(self.expand_checkbox.get_active())
            self.bottom_branch.append(expander)
            self.__initColors(expander)
            # replace the label of the most recent item which begins a new
            # indented block with the expander
            block_label = self.indented_block_title.get_children()[0]
            self.indented_block_title.remove(block_label)
            expander.set_label_widget(block_label)
            self.indented_block_title.add(expander)
            self.indented_block_title.show_all()

        # determine the insertion point in the currently active text buffer
        if self.text_view_buffer is self.root_text_view_buffer:
            insert_iter = \
                self.text_view_buffer.get_iter_at_mark \
                    (self.text_view_buffer.get_mark("end_text"))
        else:
            insert_iter = self.text_view_buffer.get_end_iter()

        if not prepare_indented_block:
            # base case: simply insert the text in the currently active
            # text buffer
            self.text_view_buffer.insert_with_tags_by_name \
                (insert_iter, unicode(msg), *tags)
        else:
            # the text to be inserted acts as the title of a new indented
            # block; prepare to actually construct the indented block on
            # the next call of this method (unless `dedent' is called)
            self.begin_indented_block = True
            anchor = self.text_view_buffer.create_child_anchor(insert_iter)
            self.text_view_hierarchy \
                [tuple(self.text_view_hierarchy_location)] = anchor
            self.text_view_hierarchy_location.append(0)
            while msg[-1] == "\n":
                msg = msg[:-1]
            label_buf = \
                gtk.TextBuffer(self.root_text_view_buffer.get_tag_table())
            label_buf.insert_with_tags_by_name \
                (label_buf.get_end_iter(), unicode(msg), *tags)
            label_view = gtk.TextView(label_buf)
            label_view.set_editable(False)
            label_view.set_wrap_mode(False)
            label_view.set_cursor_visible(False)
            self.__initColors(label_view)
            self.indented_block_title = gtk.Table(1, 1)
            self.indented_block_title.attach(label_view, 0, 1, 0, 1)
            self.indented_block_title.show_all()
            self.text_view.add_child_at_anchor \
                (self.indented_block_title, anchor)
            self.indentation_stack.append((self.text_view, label_buf))
        # flush all pending events before scrolling
        while gtk.events_pending():
            gtk.main_iteration()
        self.__scrollToEnd()
        gtk.gdk.flush()
        # allow the logger thread to proceed
        self.logger_cv.acquire()
        self.logger_cv.notify()
        self.logger_cv.release()

    def dedent(self, lvl, output=()):
        """Decrease the indentation level of log messages.

        This method closes the currently active indented text view and
        returns to the previous one such that further log messages will
        be added to that text view.  The method also allows appending
        new text to the "title" of the text view to be closed (the title
        resides in the text view to be returned to).

        Arguments:
            lvl    -- A dummy argument.
            output -- A sequence of (text, tags) pairs to be appended
                      to the "title" of the text view to be closed.
                      Here, `text' is a string of text, and `tags' is
                      a list of tags defined in GraphicalLogger.__init__.
        """
        self.logger_cv.acquire()
        gobject.idle_add(self.__dedent_in_gui_thread, lvl, output)
        self.logger_cv.wait()
        self.logger_cv.release()

    def __dedent_in_gui_thread(self, lvl, output):
        """Method used in `dedent' to update the GUI inside the GUI thread."""
        if not self.begin_indented_block:
            # there was actual text inserted to the currently active text
            # view
            i = self.text_view_buffer.get_end_iter()
            i.backward_char()
            last_char = \
                self.text_view_buffer.get_text \
                    (i, self.text_view_buffer.get_end_iter())
            if last_char == "\n":
                # if the last character in the text buffer to be closed is
                # a newline, remove it
                self.text_view_buffer.delete \
                    (i, self.text_view_buffer.get_end_iter())
            if self.bottom_branch:
                self.bottom_branch.pop()
        else:
            # no text was actually inserted after the last indented block
            # "title"
            self.begin_indented_block = False
        (self.text_view, label_buf) = self.indentation_stack.pop()
        self.text_view_buffer = self.text_view.get_buffer()
        self.text_view_hierarchy_location.pop()
        self.text_view_hierarchy_location[-1] += 1
        if output:
            # update the "title" of the closed text view
            for (text, tags) in output:
                while text and text[-1] == "\n":
                    text = text[:-1]
                label_buf.insert_with_tags_by_name \
                    (label_buf.get_end_iter(), unicode(text), *tags)
        # add a newline at the end of the now active text view
        if self.text_view_buffer is self.root_text_view_buffer:
            insert_iter = \
                self.text_view_buffer.get_iter_at_mark \
                    (self.text_view_buffer.get_mark("end_text"))
        else:
            insert_iter = self.text_view_buffer.get_end_iter()
        self.text_view_buffer.insert(insert_iter, u"\n")
        gtk.gdk.flush()
        # allow the logger thread to proceed
        self.logger_cv.acquire()
        self.logger_cv.notify()
        self.logger_cv.release()

    def __scrollToEnd(self):
        """Scroll to the end of the root text view (run in the GUI thread)."""
        if self.scroll_checkbox.get_active() \
               and not False in [e.get_expanded() for e in self.bottom_branch]:
            self.root_text_view.scroll_to_mark \
                (self.root_text_view_buffer.get_mark("end"), 0.0, True, 0.0,
                 1.0)
            self.root_text_view.queue_draw()

    def __expandBranch(self, checkbox):
        """Fully expand the newest branch in the indented text view tree.

        Argument:
            checkbox -- Dummy argument (needed for using the method as a
                        signal handler).
        """
        if self.expand_checkbox.get_active():
            for expander in self.bottom_branch:
                expander.set_expanded(True)
            self.__scrollToEnd()

    def __saveLog(self, button):
        """The "Save as..." dialog for saving log files.

        Argument:
            button -- Dummy argument (needed for using the method as a
                      signal handler).
        """
        file_dialog = \
            gtk.FileChooserDialog \
                ("Save log file", action=gtk.FILE_CHOOSER_ACTION_SAVE,
                 parent=self.dialog,
                 buttons=(gtk.STOCK_OK, gtk.RESPONSE_OK, gtk.STOCK_CANCEL,
                          gtk.RESPONSE_CANCEL))
        save_reduced_log_checkbox = \
            gtk.CheckButton("Save only currently open branches")
        save_reduced_log_checkbox.set_active(self.save_reduced_log)
        file_dialog.set_extra_widget(save_reduced_log_checkbox)
        file_dialog.set_current_folder(self.log_directory)
        filters = \
            (((".log"), "Log files (*.log)"), ( ("*",), "All files (*)"))
        for (extensions, description) in filters:
            file_filter = gtk.FileFilter()
            for e in extensions:
                file_filter.add_pattern(e)
            file_filter.set_name(description)
            file_dialog.add_filter(file_filter)
        try:
            while True:
                response = file_dialog.run()
                if response != gtk.RESPONSE_OK:
                    return
                self.log_directory = file_dialog.get_current_folder()
                self.save_reduced_log = save_reduced_log_checkbox.get_active()
                try:
                    log_file = open(file_dialog.get_filename(), "w")
                    try:
                        log_file.write \
                            ("---- SMUML model analyzer log created on %s "
                             "-----\n" % (str(datetime.now())))
                        self.__writeLog \
                            (log_file, self.root_text_view_buffer, [], 0)
                    finally:
                        log_file.close()
                    return
                except Exception, e:
                    error_dialog = \
                        gtk.MessageDialog \
                            (parent=file_dialog,
                             type=gtk.MESSAGE_ERROR,
                             flags=gtk.DIALOG_MODAL
                                       | gtk.DIALOG_DESTROY_WITH_PARENT,
                             buttons=gtk.BUTTONS_CLOSE,
                             message_format="Failed to write log file!")
                    error_dialog.format_secondary_text(str(e))
                    error_dialog.run()
                    error_dialog.destroy()
        finally:
            file_dialog.destroy()

    def __writeLog(self, log_file, buf, location, indent):
        """Write the log to a file.

        Arguments:
            log_file -- A stream opened for writing.
            buf      -- A gtk.TextBuffer object.
            location -- Identifier of the location of `buf' in the text
                        buffer hierarchy.
            indent   -- Indentation level.
        """
        if buf is self.root_text_view_buffer:
            end_iter = buf.get_iter_at_mark(buf.get_mark("end_text"))
        else:
            end_iter = buf.get_end_iter()
        slc = buf.get_slice(buf.get_start_iter(), end_iter, True)
        subbranch = 0
        for line in slc.split("\n"):
            if line == u"\uFFFC": # location of a text anchor
                location.append(subbranch)
                anchor = self.text_view_hierarchy[tuple(location)]
                w = anchor.get_widgets()[0].get_children()[0]
                if isinstance(w, gtk.TextView):
                    # subbranch contains only a text view (the label)
                    label_buffer = w.get_buffer()
                else:
                    # subbranch contains an expander (opening an indented
                    # block of text)
                    label_buffer = w.get_label_widget().get_buffer()
                label = \
                    label_buffer.get_text \
                        (label_buffer.get_start_iter(),
                         label_buffer.get_end_iter())
                while label and label[-1] == "\n":
                    label = label[:-1]
                log_file.write(indent * " " + label + "\n")
                if isinstance(w, gtk.Expander) and \
                       (not self.save_reduced_log or w.get_expanded()):
                    self.__writeLog \
                        (log_file,
                         w.get_child().get_children()[0].get_buffer(),
                         location, indent + 4)
                location.pop()
                subbranch += 1
            else:
                log_file.write(" " * indent + line + "\n")
        return

    def __gui(self):
        """Run the GUI dialog (in the GUI thread)"""
        while self.dialog.run() != gtk.RESPONSE_CLOSE:
            pass

    def __pulse(self):
        """Periodically update (pulse) the progress bar."""
        while not self.stop_pulser_thread:
            t = time.time()
            gobject.idle_add(self.progress_bar.pulse)
            gobject.idle_add \
                (self.progress_time.set_text,
                     ("Elapsed time: %s (Total: %s)"
                      % (time.strftime \
                             ("%H:%M:%S",
                              time.gmtime(t - self.activity_start_time)),
                         time.strftime \
                             ("%H:%M:%S",
                              time.gmtime \
                                  (t - self.total_activity_start_time)))))
            time.sleep(0.5)

    def __initColors(widget):
        """Set base and background colors of a widget to white.

        Argument:
            widget -- The widget.
        """
        for state in (gtk.STATE_NORMAL, gtk.STATE_ACTIVE, gtk.STATE_PRELIGHT,
                      gtk.STATE_INSENSITIVE):
            widget.modify_base(state, gtk.gdk.Color(65535, 65535, 65535, 0))
            widget.modify_bg(state, gtk.gdk.Color(65535, 65535, 65535, 0))

    __initColors = staticmethod(__initColors)

class LogHandler(logging.Handler):
    """Class for handling log messages using the Logger interface."""

    def __init__(self, logger):
        """Initialize a LogHandler.

        Argument:
            logger -- A Logger object.
        """
        if not isinstance(logger, Logger):
            raise TypeError \
                      ("a frontend.logutil.Logger object expected as argument")
        logging.Handler.__init__(self)
        self.logMessage = logger.logMessage
        self.createLock()

    def emit(self, record):
        """Emit a log message.

        Argument:
            record -- A LogRecord object.  If the record is a tuple, it
                      should contain two or three elements consisting of
                      a message string, a list of formatting tag strings,
                      and (optionally) a truth value indicating whether
                      the message string acts as the "title" of a block
                      of text to be indented.
        """
        if isinstance(record.msg, tuple):
            if len(record.msg) > 2:
                (msg, tags, prepare_indented_block) = record.msg[:3]
            else:
                (msg, tags) = record.msg[:2]
                prepare_indented_block = False
        else:
            (msg, tags, prepare_indented_block) = \
                  (record.getMessage(), [], False)
        self.logMessage \
            (msg, record.levelno, list(tags), prepare_indented_block)
