# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Abstraction refinement for integer attributes using SimpleIntPartitions.

This module implements the abstraction refinement cycle using
SimpleIntPartitions.

Author: Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

import sys, logging, random
import coral
from abstraction.typeinference import \
     TypeInferenceSystem, AttributeNode, SignalParameterNode, \
     TypeInferenceError, RewriteError
from abstraction.error import Error
from frontend.logutil import logCompletion, logFailure
from frontend import subtaskwrapper
from refinement import semantics
from refinement.partition import SimpleIntPartition, PartitionCollection
from Utils.Models import get_qualified_name

class ModelRefiner:
    """Operations for abstraction refinement."""
    
    def __init__(self, project, model, configuration, excluded_attributes=()):
        """Initialize a model refiner.

        Arguments:
            project             -- A coral.Project object.
            model               -- A coral.kernel.ElementPtr representing
                                   a UML Model.  `model' must be contained
                                   in `project'.
            configuration       -- A frontend.configuration.Configuration
                                   object.
            excluded_attributes -- A collection of UUID's of attributes
                                   to be excluded from automatic
                                   abstraction refinement.
        """
        if not isinstance(project, coral.Project):
            raise TypeError \
                      ("a coral.Project object expected as first argument")
        self.project = project
        if not isinstance(model, coral.kernel.ElementPtr):
            raise TypeError \
                      ("a coral.kernel.ElementPtr object expected as second "
                       "argument")
        if not model.oclIsKindOf(coral.lang.UML14.Model):
            raise ValueError \
                      ("second argument does not represent a UML Model")
        if not model in project.contents():
            raise ValueError \
                      ("the UML Model given as second argument is not "
                       "contained in the project given as first argument")

        if configuration.integer_semantics == "bounded":
            type_library_extender = \
                semantics.BoundedIntegerSemantics \
                    (configuration.typegen_smt_solver.lower(),
                     configuration.getPath
                         (configuration.typegen_smt_solver.lower())[0])
        elif configuration.integer_semantics == "unbounded":
            type_library_extender =  semantics.UnboundedIntegerSemantics()
        elif configuration.integer_semantics == "mixed":
            type_library_extender = \
                semantics.MixedIntegerSemantics \
                    (configuration.typegen_smt_solver.lower(),
                     configuration.getPath \
                         (configuration.typegen_smt_solver.lower())[0])
        else:
            raise ValueError \
                      ("`%s' does not match any known integer semantics"
                       % str(configuration.integer_semantics))

        if not isinstance(configuration.flattening_limit, int):
            raise TypeError("expression flattening limit must be an integer")
        self.flattening_limit = configuration.flattening_limit

        logger = logging.getLogger("log")
        logger.info("setting up type system for abstraction refinement")
        logger.startActivity \
            ("Setting up type system for abstraction refinement")
        try:
            # construct the initial type assignment, compute value flow
            # dependencies between model attributes
            self.partitions = PartitionCollection(type_library_extender)
            ANY_INTEGER = self.partitions.add(SimpleIntPartition())
            self.type_inference_system = \
                TypeInferenceSystem(model, self.partitions.getTypeLibrary())
            self.typeof = {}
            self.deps = {}
            self.attribute_by_uuid = {}
            self.attributes = \
                self.type_inference_system.findAbstractableNodes()
            for a in self.attributes:
                uuid = a.coral_element.__XMIid__()
                self.attribute_by_uuid[uuid] = a
                if uuid in excluded_attributes:
                    self.typeof[a] = "int"
                else:
                    self.typeof[a] = ANY_INTEGER
                if not a in self.deps:
                    self.deps[a] = []
                for node in a.findReachableNodes \
                                (filt=lambda n: isinstance(n, AttributeNode),
                                 prune=lambda n:
                                           n != a and
                                           isinstance(n, AttributeNode)):
                    try:
                        self.deps[node].append(a)
                    except KeyError:
                        self.deps[node] = [a]
                # when refining a signal parameter, all attributes to which
                # the value of this parameter may get assigned will also need
                # to be refined
                if isinstance(a, SignalParameterNode):
                    for succ in a.value_flow:
                        if isinstance(succ, AttributeNode):
                            self.deps[a].append(succ)
        finally:
            logger.stopActivity()

    def abstract(self):
        """Abstract the model associated with the model refiner.

        This method will (i) perform type inference (using the current
        type assignment of the ModelRefiner) to determine the abstract
        types of expressions in the model associated with the ModelRefiner,
        (ii) check whether abstracting the model using this type
        assignment for expressions requires any new abstract operations to
        be generated, (iii) generate all missing abstract operations
        required for model abstraction, (iv) abstract the model, (v)
        attach information about the types of abstracted expressions and
        attributes to the model (using the name "abstract_type_library"
        for the file name of a type library corresponding to the
        collection of abstract types currently defined in the type
        inference system associated with the ModelRefiner), and (vi) save
        the project file under its current file name.
        """
        logger = logging.getLogger("log")

        # perform type inference
        logger.info(("checking consistency of abstract type system", [], True))
        logger.startActivity("Checking consistency of abstract type system")
        try:
            self.type_inference_system.attachTypeLibrary \
                (self.partitions.getTypeLibrary())
            for (a, t) in self.typeof.iteritems():
                if t != "int":
                    t = t.name
                self.type_inference_system.addExternalConstraint(a, "==", t)
            try:
                self.type_inference_system.solveConstraints \
                    (ordering=self.partitions, f=self.__typeinferencelogger)
            except TypeInferenceError, e:
                raise Error("failure in type inference [%s]" % str(e.failure))
        finally:
            logger.dedent(logging.INFO)
            logger.stopActivity()

        # generate abstract model
        logger.info(("generating abstract model", [], True))
        logger.startActivity("Generating abstract model")
        try:
            # check whether any new operations need be generated (by trying
            # to abstract the model)
            logger.debug \
                (("checking whether any new abstract operations need to be "
                  "generated...", [], True))
            try:
                operations = set()
                self.type_inference_system.rewriteModel \
                    (flattening_threshold=self.flattening_limit,
                     undefined_ops=operations)
                result = operations and "yes" or "no"
                for op in operations:
                    if len(op) == 3:
                        (typename, operation, arity) = op
                        logger.debug \
                            ("need to define operation `%s' of arity %d for "
                             "type %s\n"
                             % (operation, arity,
                                self.__typestr \
                                    (self.partitions.find(typename))))
                    else:
                        (source_type, target_type) = (op[1], op[2])
                        logger.debug \
                            ("need to define a coercion from type %s to type "
                             "%s\n"
                             % (self.__typestr \
                                    (self.partitions.find(source_type)),
                                self.__typestr \
                                    (self.partitions.find(target_type))))
                logger.dedent \
                    (logging.DEBUG, ((" %s\n" % result, ["green", "italic"]),))
            except Exception, e:
                logger.dedent \
                    (logging.DEBUG,
                     ((" failed", ["red", "italic"]),
                      (" [%s]\n" % str(e), [])))
                raise Error("aborted")

            # generate new abstract operations if necessary, then re-abstract
            # the model
            if operations:
                subtaskwrapper.runSubtask \
                    (self.partitions.generateDefinitions,
                     "generating operations",
                     capture_prefix="generate_abstract_operations",
                     args=(operations,), stdout=subtaskwrapper.PIPE,
                     stderr=subtaskwrapper.PIPE)
                logger.debug("abstracting model...")
                try:
                    self.type_inference_system.attachTypeLibrary \
                        (self.partitions.getTypeLibrary())
                    self.type_inference_system.rewriteModel \
                        (flattening_threshold=self.flattening_limit)
                    logCompletion()
                except Exception, e:
                    logFailure(e)
                    raise Error("aborted")

            # attach type associations to the model
            self.type_inference_system.attachTypesToModel \
                (("abstract_type_library",))

            # save the abstract model
            logger.debug("saving abstract model...")
            try:
                coral.saveDocument(self.project)
                logCompletion()
            except Exception, e:
                logFailure(e)
                raise Error("aborted")                

        finally:
            logger.stopActivity()
            logger.dedent(logging.INFO)

    def refine(self, refinement):
        """Refine the abstraction using a given refinement.

        Given a specification for abstraction refinement, this method
        computes a set of attributes whose types should be refined and
        values with which to refine them.  The refined types will be
        then added to the ModelRefiner's internal type library, and
        the type assignment for attributes is updated.

        Argument:
            refinement -- The refinement specification as a sequence of
                          pairs (<UUID>, <values>).  Here, <UUID> should
                          be the UUID of an attribute in the model
                          associated with the ModelRefiner, and <values>
                          should be an iterable of integers which should
                          be added to the set of integers represented
                          exactly by the type of the attribute with UUID
                          <UUID>.  (<values> can include values already
                          represented exactly by the type of <UUID>; it
                          can also be empty.)
        """
        logger = logging.getLogger("log")
        logger.info(("applying refinement", [], True))
        logger.startActivity("Applying refinement")
        try:
            # Dictionary for collecting the types of refined attributes (as
            # sets of values which should be represented exactly by each
            # refined type).
            refined_types = {}

            # Determine the refined types of attributes by processing a
            # stack of (<attribute>, <value which should be modeled
            # exactly by the refined type>) pairs.  Initialize the stack
            # using the refinement specification.
            to_refine = []
            proper_refinement_found = False
            for (uuid, values) in refinement:
                try:
                    a = self.attribute_by_uuid[uuid]
                except KeyError:
                    raise Error \
                              ("error: UUID `%s' does not match that of any "
                               "abstractable model attribute" % uuid)
                to_refine.extend([(a, v) for v in values])
            while not proper_refinement_found:
                while to_refine:
                    (attribute, value) = to_refine.pop()
                    try:
                        current_values = refined_types[attribute]
                    except KeyError:
                        t = self.typeof[attribute]
                        if t == "int":
                            continue
                        current_values = t.getValues()
                    if not value in current_values:
                        # the (refined) type of `attribute' does not model
                        # `value' exactly: update the refined type of the
                        # attribute, and force refinement of also those
                        # attributes from which `attribute' is dependent
                        # via value flow
                        proper_refinement_found = True
                        current_values.add(value)
                        refined_types[attribute] = current_values
                        to_refine.extend \
                            ([(a, value) for a in self.deps[attribute]])
                if not proper_refinement_found:
                    # no proper refinement done, choose one at random
                    logger.debug \
                        ("no proper refinements suggested, choosing one at "
                         "random\n")
                    while True:
                        attribute = random.choice(self.attributes)
                        value = random.randint(-2147483648, 2147483647)
                        t = self.typeof[attribute]
                        if t != "int" and not t.modelsExactly(value):
                            to_refine.append((a, value))
                            break

            # Construct types for the refined attributes
            for (attribute, values) in refined_types.iteritems():
                t = self.typeof[attribute]
                refined_partition = self.partitions.add(t.refine(values))
                self.typeof[attribute] = refined_partition
                logger.debug \
                    ("refining type of `%s' from %s to %s\n"
                     % (get_qualified_name(attribute.coral_element),
                        self.__typestr(t),
                        self.__typestr(refined_partition)))
        finally:
            logger.dedent(logging.INFO)
            logger.stopActivity()

    def __typeinferencelogger(record):
        """Log an event in type inference.

        This method will log a description of a step in type inference
        (with level LOG_EXTRA_DEBUG) using the "log" logger.

        Argument:
            record -- An abstraction.typeinference.InferenceStep object.
        """
        logging.getLogger("log").extradebug(("%s\n" % str(record), ["small"]))

    __typeinferencelogger = staticmethod(__typeinferencelogger)

    def __typestr(partition):
        """Return the set representation of a partition as a string."""
        return "%s:{%s}" \
               % (partition.name,
                  ", ".join(map(str, sorted(partition.getValues()))))

    __typestr = staticmethod(__typestr)
