# This file is part of the SMUML toolset.
# Copyright (C) 2007, 2008 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Graphical user interface for preparing model analysis settings.

This module implements the graphical user interface for adjusting model
analysis settings.

Author: Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

import os, sys, subprocess
from os.path import abspath
from tempfile import mkstemp
from frontend import backends

class SettingsGui:
    """Graphical user interface for preparing model analysis settings."""
    
    def __init__(self, configuration):
        """Initialize the SettingsGui.

        Argument:
            configuration -- A Configuration object to be associated with
                             the GUI.  The GUI will be used to change
                             settings in this object.  The passed object
                             can be already populated with settings passed
                             on the command line for the script.
        """

        global gobject, gtk, pygtk, pango
        import gobject, gtk, pygtk, pango

        gobject.threads_init()
        
        PADDING = 5
        self.configuration = configuration

        table = gtk.Table(3, 1)
        table.set_border_width(PADDING)
        table.set_row_spacings(PADDING)
        table.set_col_spacings(PADDING)

        # model selection
        hbox = gtk.HBox()
        label = gtk.Label("_Model:")
        label.set_alignment(0.0, 0.5)
        self.model_name_entry = gtk.Entry()
        if configuration.model_file_name:
            self.model_name_entry.set_text(configuration.model_file_name)
        label.set_use_underline(True)
        label.set_mnemonic_widget(self.model_name_entry)
        hbox.pack_start(label, fill=False, expand=False, padding=PADDING)
        hbox.pack_start(self.model_name_entry, padding=PADDING)
        browse_model_button = gtk.Button("_Browse...", use_underline=True)
        hbox.pack_end(browse_model_button, fill=False, expand=False)
        hbox.show_all()
        table.attach(hbox, 0, 1, 0, 1, gtk.EXPAND | gtk.FILL, 0)

        notebook = gtk.Notebook()

        # model preprocessing options
        preprocessing_tab = gtk.Table(5, 1)
        preprocessing_tab.set_row_spacings(PADDING)
        preprocessing_tab.set_col_spacings(PADDING)
        preprocessing_tab.set_border_width(PADDING)
        self.flatten_hierarchy_button = \
            gtk.CheckButton("Flatten state machine hierarchy")
        self.flatten_hierarchy_button.set_active \
            (configuration.flatten_hierarchy)
        preprocessing_tab.attach \
            (self.flatten_hierarchy_button, 0, 1, 0, 1, gtk.FILL, 0)
        self.flatten_actlang_button = \
            gtk.CheckButton("Flatten action language control flow")
        self.flatten_actlang_button.set_active \
            (configuration.flatten_action_language)
        preprocessing_tab.attach \
            (self.flatten_actlang_button, 0, 1, 1, 2, gtk.FILL, 0)
        subtable = gtk.Table(3, 2)
        subtable.set_row_spacings(PADDING)
        subtable.set_col_spacings(PADDING)
        padding = gtk.Label()
        padding.set_padding \
            (self.flatten_actlang_button.style_get_property("indicator-size")
                 + self.flatten_actlang_button.style_get_property \
                       ("indicator-spacing"),
             0)
        subtable.attach(padding, 0, 1, 0, 1, 0, 0)
        self.flatten_actlang_control_flow_only = \
            gtk.RadioButton \
               (None,
                "Allow any number of simple statements in transition effects")
        subtable.attach \
            (self.flatten_actlang_control_flow_only, 1, 2, 0, 1, gtk.FILL, 0)
        self.flatten_actlang_sends = \
            gtk.RadioButton \
                (self.flatten_actlang_control_flow_only,
                 "At most one 'send' statement per transition effect")
        subtable.attach(self.flatten_actlang_sends, 1, 2, 1, 2, gtk.FILL, 0)
        self.flatten_actlang_all = \
            gtk.RadioButton \
                (self.flatten_actlang_control_flow_only,
                 "At most one simple statement per transition effect")
        subtable.attach(self.flatten_actlang_all, 1, 2, 2, 3, gtk.FILL, 0)
        subtable.show_all()
        preprocessing_tab.attach(subtable, 0, 1, 2, 3, gtk.FILL, 0)
        self.slice_assertions_button = \
            gtk.CheckButton("Slice model with respect to assertions")
        self.slice_assertions_button.set_active(configuration.slice_assertions)
        preprocessing_tab.attach \
            (self.slice_assertions_button, 0, 1, 3, 4, gtk.FILL, 0)
        self.flatten_transition_exprs_button = \
            gtk.CheckButton("Flatten expressions in effects of transitions")
        self.flatten_transition_exprs_button.set_active \
            (configuration.flatten_transition_effects)
        preprocessing_tab.attach \
            (self.flatten_transition_exprs_button, 0, 1, 4, 5, gtk.FILL, 0)
        label = gtk.Label("_Preprocessing")
        label.set_use_underline(True)
        label.set_mnemonic_widget(preprocessing_tab)
        notebook.append_page(preprocessing_tab, label)
        preprocessing_tab.show_all()

        # model checking options
        model_checking_tab = gtk.Table(3, 1)
        model_checking_tab.set_row_spacings(PADDING)
        model_checking_tab.set_col_spacings(PADDING)
        model_checking_tab.set_border_width(PADDING)
        label = gtk.Label("Model _checking")
        label.set_use_underline(True)
        label.set_mnemonic_widget(model_checking_tab)
        label.show()
        notebook.append_page(model_checking_tab, label)

        # a class used for collecting together widgets associated with a
        # specific back-end
        class Widgets:
            pass

        # back-ends
        self.uboco = Widgets()
        self.suboco = Widgets()
        self.proco = Widgets()
        self.backends = [self.uboco, self.suboco, self.proco]

        # back-end selection
        hbox = gtk.HBox()
        label = gtk.Label("Model checking back-end:")
        label.set_alignment(1.0, 0.5)
        hbox.pack_start(label, fill=False, expand=False, padding=PADDING)
        self.backend_combo_box = gtk.combo_box_new_text()
        for b in ("Uboco", "Suboco", "Proco"):
            self.backend_combo_box.append_text(b)
        if configuration.backend == None \
               or isinstance(configuration.backend, backends.Uboco):
            self.current_backend = self.uboco
        elif isinstance(configuration.backend, backends.Suboco):
            self.current_backend = self.suboco
        elif isinstance(configuration.backend, backends.Proco):
            self.current_backend = self.proco
        self.backend_combo_box.set_active \
            (self.backends.index(self.current_backend))
        hbox.pack_start(self.backend_combo_box, fill=False, expand=False)
        hbox.show_all()
        model_checking_tab.attach(hbox, 0, 1, 0, 1, gtk.FILL | gtk.EXPAND, 0)

        basic_options_size_group = gtk.SizeGroup(gtk.SIZE_GROUP_BOTH)
        additional_options_size_group = gtk.SizeGroup(gtk.SIZE_GROUP_BOTH)
        additional_option_name_width_size_group = \
            gtk.SizeGroup(gtk.SIZE_GROUP_HORIZONTAL)
        bound_options_size_group = gtk.SizeGroup(gtk.SIZE_GROUP_VERTICAL)
        bound_options_width_size_group = \
            gtk.SizeGroup(gtk.SIZE_GROUP_HORIZONTAL)

        # options for the Uboco back-end
        tab = self.uboco
        tab.table = gtk.Table(6, 2)
        tab.table.set_row_spacings(PADDING)
        tab.table.set_col_spacings(PADDING)
        subtable = gtk.Table(5, 1)
        subtable.set_row_spacings(PADDING)
        subtable.set_col_spacings(PADDING)
        subsubtable = gtk.Table(2, 1)
        subsubtable.set_col_spacings(PADDING)
        subsubtable.set_row_spacings(PADDING)
        tab.bounded_model_checking_checkbox = \
            gtk.CheckButton("Use bounded model checking")
        subsubtable.attach \
            (tab.bounded_model_checking_checkbox, 0, 1, 0, 1, gtk.FILL, 0)
        if isinstance(configuration.backend, backends.Uboco):
            tab.bounded_model_checking_checkbox.set_active \
                (configuration.backend.use_bmc)
        else:
            tab.bounded_model_checking_checkbox.set_active(True)
        tab.max_bmc_bound_hbox = gtk.HBox()
        padding = gtk.Label("")
        padding.set_padding(10, 0)
        tab.max_bmc_bound_hbox.pack_start(padding, fill=False, expand=False)
        label = gtk.Label("Maximum number of BMC steps:")
        label.set_alignment(1.0, 0.5)
        tab.max_bmc_bound_hbox.pack_start(label, fill=False, expand=False)
        subsubtable.attach(tab.max_bmc_bound_hbox, 0, 1, 1, 2, gtk.FILL, 0)
        bound_options_width_size_group.add_widget(tab.max_bmc_bound_hbox)
        tab.max_bound_entry = gtk.Entry()
        tab.max_bound_entry.set_width_chars(7)
        if isinstance(configuration.backend, backends.Uboco):
            tab.max_bound_entry.set_text(str(configuration.backend.bmc_bound))
        else:
            tab.max_bound_entry.set_text(str(backends.Uboco.DEFAULT_BMC_BOUND))
        subsubtable.attach(tab.max_bound_entry, 1, 2, 1, 2, gtk.FILL, 0)
        bound_options_size_group.add_widget(subsubtable)
        subtable.attach(subsubtable, 0, 1, 0, 2, gtk.FILL, 0)
        tab.check_assertions_checkbox = \
            gtk.CheckButton("Check for assertion violations")
        tab.check_assertions_checkbox.set_active(True)
        subtable.attach \
            (tab.check_assertions_checkbox, 0, 1, 2, 3, gtk.FILL, 0)
        tab.check_deadlocks_checkbox = \
            gtk.CheckButton("Check for the presence of deadlocks")
        subtable.attach \
            (tab.check_deadlocks_checkbox, 0, 1, 3, 4, gtk.FILL, 0)
        tab.check_implicit_consumption_checkbox = \
            gtk.CheckButton("Check for implicit consumption of events")
        subtable.attach \
            (tab.check_implicit_consumption_checkbox, 0, 1, 4, 5, gtk.FILL,
             0)
        tab.check_runtime_errors_checkbox = \
            gtk.CheckButton("Check for runtime errors")
        subtable.attach \
            (tab.check_runtime_errors_checkbox, 0, 1, 5, 6, gtk.FILL, 0)
        tab.table.attach(subtable, 0, 1, 0, 6, gtk.FILL, gtk.FILL)
        basic_options_size_group.add_widget(subtable)
        subtable = gtk.Table(6, 3)
        subtable.set_col_spacings(PADDING)
        subtable.set_row_spacings(PADDING)
        padding = gtk.Label()
        padding.set_padding(10, 0)
        subtable.attach(padding, 0, 1, 0, 1, 0, 0)
        tab.sat_solver_label = gtk.Label("SAT solver for NuSMV:")
        tab.sat_solver_label.set_alignment(1.0, 0.5)
        subtable.attach(tab.sat_solver_label, 1, 2, 0, 1, gtk.FILL, 0)
        additional_option_name_width_size_group.add_widget(label)
        tab.nusmv_sat_solver_combo_box = gtk.combo_box_new_text()
        for (i, solver) in enumerate(backends.Uboco.NUSMV_SAT_SOLVERS):
            tab.nusmv_sat_solver_combo_box.append_text(solver)
            if isinstance(configuration.backend, backends.Uboco) \
                   and configuration.backend.sat_solver \
                   and configuration.backend.sat_solver.lower() \
                           == solver.lower():
                tab.nusmv_sat_solver_combo_box.set_active(i)
        if tab.nusmv_sat_solver_combo_box.get_active() == -1:
            tab.nusmv_sat_solver_combo_box.set_active(0)
        hbox = gtk.HBox()
        hbox.pack_start \
            (tab.nusmv_sat_solver_combo_box, fill=False, expand=False)
        subtable.attach(hbox, 2, 3, 0, 1, gtk.FILL, 0)
        label = gtk.Label("Number of bits for the int type:")
        label.set_alignment(1.0, 0.5)
        subtable.attach(label, 1, 2, 1, 2, gtk.FILL, 0)
        additional_option_name_width_size_group.add_widget(label)
        tab.int_bits_entry = gtk.Entry()
        tab.int_bits_entry.set_width_chars(2)
        tab.int_bits_entry.set_max_length(2)
        tab.int_bits_entry.set_text("8")
        hbox = gtk.HBox()
        hbox.pack_start(tab.int_bits_entry, fill=False, expand=False)
        subtable.attach(hbox, 2, 3, 1, 2, gtk.FILL, 0)
        label = gtk.Label("Capacity of event queues:")
        label.set_alignment(1.0, 0.5)
        subtable.attach(label, 1, 2, 2, 3, gtk.FILL, 0)
        additional_option_name_width_size_group.add_widget(label)
        tab.event_queue_size_entry = gtk.Entry()
        tab.event_queue_size_entry.set_width_chars(4)
        tab.event_queue_size_entry.set_text("2")
        hbox = gtk.HBox()
        hbox.pack_start(tab.event_queue_size_entry, fill=False, expand=False)
        subtable.attach(hbox, 2, 3, 2, 3, gtk.FILL, 0)
        label = gtk.Label("Execution semantics:")
        label.set_alignment(1.0, 0.5)
        subtable.attach(label, 1, 2, 3, 4, gtk.FILL, 0)
        additional_option_name_width_size_group.add_widget(label)
        tab.execution_semantics_combo_box = gtk.combo_box_new_text()
        for s in ("Step (Dynamic)", "Step (Static)", "Interleaving"):
            tab.execution_semantics_combo_box.append_text(s)
        tab.execution_semantics_combo_box.set_active(0)
        hbox = gtk.HBox()
        hbox.pack_start \
            (tab.execution_semantics_combo_box, fill=False, expand=False)
        subtable.attach(hbox, 2, 3, 3, 4, gtk.FILL, 0)
        tab.allow_queue_overdraft = \
            gtk.CheckButton("Allow exceeding queue capacity")
        subtable.attach \
            (tab.allow_queue_overdraft, 2, 3, 4, 5, gtk.FILL, 0)
        label = gtk.Label("Extra options:")
        label.set_alignment(1.0, 0.5)
        subtable.attach(label, 1, 2, 5, 6, gtk.FILL, 0)
        additional_option_name_width_size_group.add_widget(label)
        tab.additional_options_entry = gtk.Entry()
        tab.additional_options_entry.set_text \
            (" ".join(configuration.extra_arguments))
        hbox = gtk.HBox()
        hbox.pack_start(tab.additional_options_entry, fill=False, expand=False)
        subtable.attach(hbox, 2, 3, 5, 6, gtk.FILL | gtk.EXPAND, 0)
        tab.additional_options_expander = gtk.Expander()
        tab.additional_options_expander.set_use_markup(True)
        tab.additional_options_expander.set_label \
            ('<span size="small"><i>Show additional options</i></span>')
        tab.additional_options_expander.add(subtable)
        additional_options_size_group.add_widget(subtable)
        tab.table.attach \
            (tab.additional_options_expander, 1, 2, 0, 6, gtk.FILL, gtk.FILL)
        model_checking_tab.attach(tab.table, 0, 1, 1, 2)

        # options for the Suboco back-end
        tab = self.suboco
        tab.table = gtk.Table(6, 2)
        tab.table.set_row_spacings(PADDING)
        tab.table.set_col_spacings(PADDING)
        subtable = gtk.Table(5, 1)
        subtable.set_row_spacings(PADDING)
        subtable.set_col_spacings(PADDING)
        subsubtable = gtk.Table(2, 2)
        subsubtable.set_col_spacings(PADDING)
        label = gtk.Label("Minimum number of BMC steps:")
        label.set_alignment(1.0, 0.5)
        subsubtable.attach(label, 0, 1, 0, 1, gtk.FILL, 0)
        bound_options_width_size_group.add_widget(label)
        tab.min_bound_entry = gtk.Entry()
        tab.min_bound_entry.set_width_chars(7)
        tab.min_bound_entry.set_text("0")
        subsubtable.attach(tab.min_bound_entry, 1, 2, 0, 1, gtk.FILL, 0)
        label = gtk.Label("Maximum number of BMC steps:")
        label.set_alignment(1.0, 0.5)
        subsubtable.attach(label, 0, 1, 1, 2, gtk.FILL, 0)
        bound_options_width_size_group.add_widget(label)
        tab.max_bound_entry = gtk.Entry()
        tab.max_bound_entry.set_width_chars(7)
        tab.max_bound_entry.set_text \
            (str(backends.Suboco.DEFAULT_BMC_MAX_BOUND))
        subsubtable.attach(tab.max_bound_entry, 1, 2, 1, 2, gtk.FILL, 0)
        bound_options_size_group.add_widget(subsubtable)
        subtable.attach(subsubtable, 0, 1, 0, 2, gtk.FILL, 0)
        tab.check_assertions_checkbox = \
            gtk.CheckButton("Check for assertion violations")
        tab.check_assertions_checkbox.set_active(True)
        subtable.attach \
            (tab.check_assertions_checkbox, 0, 1, 2, 3, gtk.FILL, 0)
        tab.check_deadlocks_checkbox = \
            gtk.CheckButton("Check for the presence of deadlocks")
        subtable.attach \
            (tab.check_deadlocks_checkbox, 0, 1, 3, 4, gtk.FILL, 0)
        tab.check_implicit_consumption_checkbox = \
            gtk.CheckButton("Check for implicit consumption of events")
        subtable.attach \
            (tab.check_implicit_consumption_checkbox, 0, 1, 4, 5, gtk.FILL, 0)
        tab.check_runtime_errors_checkbox = \
            gtk.CheckButton("Check for runtime errors")
        subtable.attach \
            (tab.check_runtime_errors_checkbox, 0, 1, 5, 6, gtk.FILL, 0)
        tab.table.attach(subtable, 0, 1, 0, 6, gtk.FILL, gtk.FILL)
        basic_options_size_group.add_widget(subtable)
        subtable = gtk.Table(6, 3)
        subtable.set_col_spacings(PADDING)
        subtable.set_row_spacings(PADDING)
        padding = gtk.Label()
        padding.set_padding(10, 0)
        subtable.attach(padding, 0, 1, 0, 1, 0, 0)
        label = gtk.Label("SMT solver for Suboco:")
        label.set_alignment(1.0, 0.5)
        subtable.attach(label, 1, 2, 0, 1, gtk.FILL, 0)
        additional_option_name_width_size_group.add_widget(label)
        tab.smt_solver_combo_box = gtk.combo_box_new_text()
        for (i, solver) in enumerate(backends.Suboco.SMT_SOLVERS):
            tab.smt_solver_combo_box.append_text(solver)
            if isinstance(configuration.backend, backends.Suboco) \
                   and configuration.backend.smt_solver.lower() \
                           == solver.lower():
                tab.smt_solver_combo_box.set_active(i)
        if tab.smt_solver_combo_box.get_active() == -1:
            tab.smt_solver_combo_box.set_active \
                (list(backends.Suboco.SMT_SOLVERS).index("Yices"))
        hbox = gtk.HBox()
        hbox.pack_start(tab.smt_solver_combo_box, fill=False, expand=False)
        subtable.attach(hbox, 2, 3, 0, 1, gtk.FILL, 0)
        label = gtk.Label("Number of bits for the int type:")
        label.set_alignment(1.0, 0.5)
        subtable.attach(label, 1, 2, 1, 2, gtk.FILL, 0)
        additional_option_name_width_size_group.add_widget(label)
        tab.int_bits_entry = gtk.Entry()
        tab.int_bits_entry.set_width_chars(2)
        tab.int_bits_entry.set_max_length(2)
        tab.int_bits_entry.set_text("8")
        hbox = gtk.HBox()
        hbox.pack_start(tab.int_bits_entry, fill=False, expand=False)
        subtable.attach(hbox, 2, 3, 1, 2, gtk.FILL, 0)
        label = gtk.Label("Capacity of event queues:")
        label.set_alignment(1.0, 0.5)
        subtable.attach(label, 1, 2, 2, 3, gtk.FILL, 0)
        additional_option_name_width_size_group.add_widget(label)
        tab.event_queue_size_entry = gtk.Entry()
        tab.event_queue_size_entry.set_width_chars(4)
        tab.event_queue_size_entry.set_text("2")
        hbox = gtk.HBox()
        hbox.pack_start(tab.event_queue_size_entry, fill=False, expand=False)
        subtable.attach(hbox, 2, 3, 2, 3, gtk.FILL, 0)
        label = gtk.Label("Execution semantics:")
        label.set_alignment(1.0, 0.5)
        subtable.attach(label, 1, 2, 3, 4, gtk.FILL, 0)
        additional_option_name_width_size_group.add_widget(label)
        tab.execution_semantics_combo_box = gtk.combo_box_new_text()
        for s in ("Step (Dynamic)", "Step (Static)", "Interleaving"):
            tab.execution_semantics_combo_box.append_text(s)
        tab.execution_semantics_combo_box.set_active(0)
        hbox = gtk.HBox()
        hbox.pack_start \
            (tab.execution_semantics_combo_box, fill=False, expand=False)
        subtable.attach(hbox, 2, 3, 3, 4, gtk.FILL, 0)
        tab.allow_queue_overdraft = \
            gtk.CheckButton("Allow exceeding queue capacity")
        subtable.attach(tab.allow_queue_overdraft, 2, 3, 4, 5, gtk.FILL, 0)
        label = gtk.Label("Extra options:")
        label.set_alignment(1.0, 0.5)
        subtable.attach(label, 1, 2, 5, 6, gtk.FILL, 0)
        additional_option_name_width_size_group.add_widget(label)
        tab.additional_options_entry = gtk.Entry()
        tab.additional_options_entry.set_text \
            (" ".join(configuration.extra_arguments))
        hbox = gtk.HBox()
        hbox.pack_start(tab.additional_options_entry, fill=False, expand=False)
        subtable.attach(hbox, 2, 3, 5, 6, gtk.FILL | gtk.EXPAND, 0)
        tab.additional_options_expander = gtk.Expander()
        tab.additional_options_expander.set_use_markup(True)
        tab.additional_options_expander.set_label \
            ('<span size="small"><i>Show additional options</i></span>')
        tab.additional_options_expander.add(subtable)
        additional_options_size_group.add_widget(subtable)
        tab.table.attach \
            (tab.additional_options_expander, 1, 2, 0, 6, gtk.FILL, gtk.FILL)
        model_checking_tab.attach(tab.table, 0, 1, 1, 2)

        # options for the Proco back-end
        tab = self.proco
        tab.table = gtk.Table(5, 2)
        tab.table.set_row_spacings(PADDING)
        tab.table.set_col_spacings(PADDING)
        subtable = gtk.Table(5, 1)
        subtable.set_row_spacings(PADDING)
        subtable.set_col_spacings(PADDING)
        subsubtable = gtk.Table(2, 2)
        subsubtable.set_col_spacings(PADDING)
        label = gtk.Label("Maximum search depth:")
        label.set_alignment(1.0, 0.5)
        subsubtable.attach(label, 0, 1, 0, 1, gtk.FILL, 0)
        bound_options_width_size_group.add_widget(label)
        tab.max_bound_entry = gtk.Entry()
        tab.max_bound_entry.set_width_chars(7)
        tab.max_bound_entry.set_text \
            (str(backends.Proco.DEFAULT_MAX_SEARCH_DEPTH))
        subsubtable.attach(tab.max_bound_entry, 1, 2, 0, 1, gtk.FILL, 0)
        bound_options_size_group.add_widget(subsubtable)
        subtable.attach(subsubtable, 0, 1, 0, 2, gtk.FILL, 0)
        tab.check_assertions_checkbox = \
            gtk.CheckButton("Check for assertion violations")
        tab.check_assertions_checkbox.set_active(True)
        subtable.attach \
            (tab.check_assertions_checkbox, 0, 1, 2, 3, gtk.FILL, 0)
        tab.check_deadlocks_checkbox = \
            gtk.CheckButton("Check for the presence of deadlocks")
        subtable.attach \
            (tab.check_deadlocks_checkbox, 0, 1, 3, 4, gtk.FILL, 0)
        tab.check_implicit_consumption_checkbox = \
            gtk.CheckButton("Check for implicit consumption of events")
        subtable.attach \
            (tab.check_implicit_consumption_checkbox, 0, 1, 4, 5, gtk.FILL, 0)
        tab.check_queue_overflows_checkbox = \
            gtk.CheckButton("Check for event queue overflows")
        subtable.attach \
            (tab.check_queue_overflows_checkbox, 0, 1, 5, 6, gtk.FILL, 0)
        tab.table.attach(subtable, 0, 1, 0, 6, gtk.FILL, gtk.FILL)
        basic_options_size_group.add_widget(subtable)
        subtable = gtk.Table(8, 3)
        subtable.set_col_spacings(PADDING)
        subtable.set_row_spacings(PADDING)
        padding = gtk.Label()
        padding.set_padding(10, 0)
        subtable.attach(padding, 0, 1, 0, 1, 0, 0)
        label = gtk.Label("Search mode:")
        label.set_alignment(1.0, 0.5)
        subtable.attach(label, 1, 2, 0, 1, gtk.FILL, 0)
        additional_option_name_width_size_group.add_widget(label)
        tab.search_mode_combo_box = gtk.combo_box_new_text()
        for p in ("Depth first", "Breadth first"):
            tab.search_mode_combo_box.append_text(p)
        tab.search_mode_combo_box.set_active(0)
        hbox = gtk.HBox()
        hbox.pack_start(tab.search_mode_combo_box, fill=False, expand=False)
        subtable.attach(hbox, 2, 3, 0, 1, gtk.FILL, 0)
        tab.minimize_cex_checkbox = \
            gtk.CheckButton("Minimize counterexamples for DFS")
        subtable.attach(tab.minimize_cex_checkbox, 2, 3, 1, 2, 0, 0)
        label = gtk.Label("Capacity of event queues:")
        label.set_alignment(1.0, 0.5)
        subtable.attach(label, 1, 2, 2, 3, gtk.FILL, 0)
        additional_option_name_width_size_group.add_widget(label)
        tab.event_queue_size_entry = gtk.Entry()
        tab.event_queue_size_entry.set_width_chars(4)
        tab.event_queue_size_entry.set_text("2")
        hbox = gtk.HBox()
        hbox.pack_start(tab.event_queue_size_entry, fill=False, expand=False)
        subtable.attach(hbox, 2, 3, 2, 3, gtk.FILL, 0)
        tab.use_collapse = gtk.CheckButton("Use collapse compression")
        subtable.attach(tab.use_collapse, 1, 3, 3, 4, 0, 0)
        label = gtk.Label("Extra options:")
        label.set_alignment(1.0, 0.5)
        subtable.attach(label, 1, 2, 4, 5, gtk.FILL, 0)
        additional_option_name_width_size_group.add_widget(label)
        tab.additional_options_entry = gtk.Entry()
        tab.additional_options_entry.set_text \
            (" ".join(configuration.extra_arguments))
        hbox = gtk.HBox()
        hbox.pack_start \
            (tab.additional_options_entry, fill=False, expand=False)
        subtable.attach(hbox, 2, 3, 4, 5, gtk.FILL | gtk.EXPAND, 0)
        tab.additional_options_expander = gtk.Expander()
        tab.additional_options_expander.set_use_markup(True)
        tab.additional_options_expander.set_label \
            ('<span size="small"><i>Show additional options</i></span>')
        tab.additional_options_expander.add(subtable)
        additional_options_size_group.add_widget(subtable)
        tab.table.attach \
            (tab.additional_options_expander, 1, 2, 0, 5, gtk.FILL, gtk.FILL)
        model_checking_tab.attach(tab.table, 0, 1, 1, 2)
        model_checking_tab.show()

        for tab in self.backends:
            tab.additional_options_expander.set_expanded(True)
            tab.table.show_all()
            tab.additional_options_expander.set_expanded(False)
            tab.table.hide()

        # abstraction refinement options
        abstraction_refinement_tab = gtk.Table(3, 1)
        abstraction_refinement_tab.set_row_spacings(PADDING)
        abstraction_refinement_tab.set_col_spacings(PADDING)
        abstraction_refinement_tab.set_border_width(PADDING)
        self.use_refinement_checkbox = \
            gtk.CheckButton("Apply automatic abstraction refinement")
        self.use_refinement_checkbox.set_active(configuration.use_refinement)
        abstraction_refinement_tab.attach \
            (self.use_refinement_checkbox, 0, 2, 0, 1, gtk.FILL, 0)
        self.abstraction_refinement_options = gtk.Table(2, 1)
        self.abstraction_refinement_options.set_col_spacings(PADDING)
        self.abstraction_refinement_options.set_row_spacings(PADDING)
        label = gtk.Label("Integer semantics for generating abstract types:")
        label.set_alignment(1.0, 0.5)
        self.abstraction_refinement_options.attach \
            (label, 0, 1, 0, 1, gtk.FILL, 0)
        self.type_semantics_combo_box = gtk.combo_box_new_text()
        self.type_semantics_combo_box.set_active(0)
        for (i, s) in enumerate(configuration.ABSTRACT_TYPE_SEMANTICS):
            self.type_semantics_combo_box.append_text(s)
            if s.lower() == configuration.integer_semantics.lower():
                self.type_semantics_combo_box.set_active(i)
        hbox = gtk.HBox()
        hbox.pack_start \
            (self.type_semantics_combo_box, fill=False, expand=False)
        self.abstraction_refinement_options.attach \
            (hbox, 1, 2, 0, 1, gtk.FILL, 0)
        self.typegen_solver_label = \
            gtk.Label("SMT solver for generating abstract types:")
        self.typegen_solver_label.set_alignment(1.0, 0.5)
        self.abstraction_refinement_options.attach \
            (self.typegen_solver_label, 0, 1, 1, 2, gtk.FILL, 0)
        self.typegen_solver_combo_box = gtk.combo_box_new_text()
        self.typegen_solver_combo_box.set_active(0)
        for (i, s) in enumerate(configuration.ABSTRACT_TYPEGEN_SOLVERS):
            self.typegen_solver_combo_box.append_text(s)
            if s.lower() == configuration.typegen_smt_solver.lower():
                self.typegen_solver_combo_box.set_active(i)
        for w in (self.typegen_solver_label, self.typegen_solver_combo_box):
            w.set_sensitive(self.type_semantics_combo_box.get_active() != 1)
        hbox = gtk.HBox()
        hbox.pack_start \
            (self.typegen_solver_combo_box, fill=False, expand=False)
        self.abstraction_refinement_options.attach \
            (hbox, 1, 2, 1, 2, gtk.FILL, 0)
        label = gtk.Label("Limit for flattening abstract operations:")
        label.set_alignment(1.0, 0.5)
        self.abstraction_refinement_options.attach \
            (label, 0, 1, 2, 3, gtk.FILL, 0)
        self.flattening_limit_entry = gtk.Entry()
        self.flattening_limit_entry.set_width_chars(7)
        if configuration.flattening_limit < 0:
            self.flattening_limit_entry.set_text("None")
        else:
            self.flattening_limit_entry.set_text \
                (str(configuration.flattening_limit))
        hbox = gtk.HBox()
        hbox.pack_start(self.flattening_limit_entry, fill=False, expand=False)
        self.abstraction_refinement_options.attach \
            (hbox, 1, 2, 2, 3, gtk.FILL, 0)
        abstraction_refinement_tab.attach \
            (self.abstraction_refinement_options, 0, 1, 1, 2, gtk.FILL, 0)
        abstraction_refinement_tab.show_all()
        label = gtk.Label("_Abstraction refinement")
        label.set_use_underline(True)
        label.set_mnemonic_widget(abstraction_refinement_tab)
        notebook.append_page(abstraction_refinement_tab, label)

        # counterexample simulation options
        cex_simulation_tab = gtk.Table(2, 1)
        cex_simulation_tab.set_row_spacings(PADDING)
        cex_simulation_tab.set_col_spacings(PADDING)
        cex_simulation_tab.set_border_width(PADDING)
        hbox = gtk.HBox(spacing=PADDING)
        self.output_trace_checkbox = \
            gtk.CheckButton("Save a counterexample trace (if found) to file")
        self.output_trace_checkbox.set_active(False)
        hbox.pack_start(self.output_trace_checkbox, fill=False, expand=False)
        self.trace_file_name_entry = gtk.Entry()
        hbox.pack_start(self.trace_file_name_entry, fill=True, expand=True)
        if configuration.trace_file_name != None:
            self.trace_file_name_entry.set_text(configuration.trace_file_name)
            self.output_trace_checkbox.set_active(True)
        self.browse_for_trace_file_button = gtk.Button("Browse...")
        hbox.pack_start \
            (self.browse_for_trace_file_button, fill=False, expand=False)
        cex_simulation_tab.attach(hbox, 0, 1, 0, 1, gtk.FILL | gtk.EXPAND, 0)
        self.simulate_cex_checkbox = gtk.CheckButton("Simulate counterexample")
        self.simulate_cex_checkbox.set_active \
            (configuration.simulate_counterexample)
        cex_simulation_tab.attach \
            (self.simulate_cex_checkbox, 0, 1, 1, 2, gtk.FILL, 0)
        cex_simulation_tab.show_all()
        label = gtk.Label("Counter_examples")
        label.set_use_underline(True)
        notebook.append_page(cex_simulation_tab, label)

        # configuration options
        configuration_tab = gtk.Table(3, 1)
        configuration_tab.set_row_spacings(PADDING)
        configuration_tab.set_col_spacings(PADDING)
        configuration_tab.set_border_width(PADDING)
        hbox = gtk.HBox()
        label = gtk.Label("Configuration file:")
        hbox.pack_start(label, fill=False, expand=False, padding=PADDING)
        self.cfg_file_entry = gtk.Entry()
        self.cfg_file_entry.set_text(configuration.cfg_file_name)
        hbox.pack_start \
            (self.cfg_file_entry, fill=True, expand=True, padding=PADDING)
        browse_cfg_file_button = gtk.Button("Browse...")
        hbox.pack_start(browse_cfg_file_button, fill=False, expand=False)
        configuration_tab.attach(hbox, 0, 1, 0, 1, gtk.FILL | gtk.EXPAND, 0)
        self.paths_view = gtk.TextView()
        self.paths_view.set_wrap_mode(False)
        paths_window = gtk.ScrolledWindow()
        paths_window.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        paths_window.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        paths_window.add(self.paths_view)
        configuration_tab.attach(paths_window, 0, 1, 1, 2)
        self.__readConfigurationFile()
        hbox = gtk.HBox()
        label = gtk.Label("")
        hbox.pack_start(label, fill=True, expand=True)
        self.save_cfg_file_button = gtk.Button("Save changes")
        self.save_cfg_file_button.set_sensitive(False)
        hbox.pack_start(self.save_cfg_file_button, fill=False, expand=False)
        configuration_tab.attach(hbox, 0, 1, 2, 3, gtk.FILL, 0)
        label = gtk.Label("C_onfiguration")
        label.set_use_underline(True)
        configuration_tab.show_all()
        notebook.append_page(configuration_tab, label)

        # the "about" tab
        about_tab = gtk.Table(5, 1)
        about_tab.set_row_spacings(PADDING)
        about_tab.set_col_spacings(PADDING)
        about_tab.set_border_width(PADDING * 10)
        label = gtk.Label()
        label.set_markup \
            ('<span size="x-large"><b>The SMUML Toolset</b></span>')
        about_tab.attach(label, 0, 1, 0, 1, gtk.EXPAND, gtk.EXPAND)
        URL = "http://www.tcs.hut.fi/Research/Logic/SMUML.html"
        label = gtk.Label(URL)
        attrs = pango.AttrList()
        attrs.insert(pango.AttrForeground(0, 0, 65535, 0, -1))
        attrs.insert(pango.AttrUnderline(pango.UNDERLINE_SINGLE, 0, -1))
        label.set_attributes(attrs)
        eventbox = gtk.EventBox()
        eventbox.set_visible_window(False)
        eventbox.add(label)
        eventbox.connect \
            ("enter-notify-event",
             lambda box, event:
                 box.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.HAND1)))
        eventbox.connect \
            ("leave-notify-event",
             lambda box, event: box.window.set_cursor(None))
        eventbox.connect \
            ("button-press-event",
             lambda box, event:
                 "BROWSER" in os.environ
                 and subprocess.Popen((os.environ["BROWSER"], URL)))
        about_tab.attach(eventbox, 0, 1, 1, 2, gtk.EXPAND, gtk.EXPAND)
        label = gtk.Label()
        label.set_markup \
            ('<span size="large">Copyright (C) 2005-2008 Helsinki University '
             'of Technology (TKK)</span>')
        about_tab.attach(label, 0, 1, 2, 3, gtk.EXPAND, gtk.EXPAND)
        label = gtk.Label()
        label.set_line_wrap(True)
        label.set_justify(gtk.JUSTIFY_CENTER)
        label.set_markup \
            ("<i>Written by Jori Dubrovin, Tommi Junttila, Toni Jussila, Timo "
             "Latvala, Sami Liedes, Vesa Ojala, Juhani Peltonen, Heikki "
             "Tauriainen</i>")
        about_tab.attach(label, 0, 1, 3, 4, gtk.EXPAND, gtk.EXPAND)
        label = gtk.Label()
        label.set_markup \
            ('<span size="small">Released under the GNU General Public '
             'License version 2.</span>')
        about_tab.attach(label, 0, 1, 4, 5, gtk.EXPAND, gtk.EXPAND)
        label = gtk.Label("Abo_ut")
        label.set_use_underline(True)
        about_tab.show_all()
        notebook.append_page(about_tab, label)
        
        table.attach(notebook, 0, 1, 1, 2)
        notebook.set_current_page(0)
        notebook.show()

        # the main dialog
        self.dialog = \
            gtk.Dialog \
                ("SMUML Model Analyzer Settings", flags=gtk.DIALOG_MODAL,
                 buttons=("_Start", gtk.RESPONSE_OK,
                          "_Quit", gtk.RESPONSE_CANCEL))
        self.dialog.set_default_size(640, 480)
        self.dialog.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_NORMAL)
        self.dialog.vbox.add(table)
        table.show()

        # current directories for file choosers
        self.current_model_directory = os.getcwd()
        self.current_trace_file_directory = os.getcwd()
        self.current_cfg_directory = os.getcwd()

        # associate widgets with signals
        browse_model_button.connect("clicked", self.__browseForModel)
        self.flatten_actlang_button.connect \
            ("toggled", self.__updateActlangFlatteningOptions)
        self.backend_combo_box.connect \
            ("changed", self.__updateModelCheckingTab)
        self.uboco.bounded_model_checking_checkbox.connect \
            ("toggled", self.__updateUbocoBmcOptions)
        self.proco.search_mode_combo_box.connect \
            ("changed", self.__updateProcoCexMinimizationCheckbox)
        for tab in self.backends:
            tab.additional_options_expander.connect \
                ("notify::expanded", self.__updateOptionExpanderLabel)
        for tab in (self.uboco, self.suboco):
            tab.execution_semantics_combo_box.connect \
                ("changed", self.__updateQueueOverdraftCheckbox, tab)
        self.use_refinement_checkbox.connect \
            ("toggled", self.__updateRefinementOptions)
        self.type_semantics_combo_box.connect \
            ("changed",
             lambda combo_box:
                 [w.set_sensitive(combo_box.get_active() != 1)
                  for w in (self.typegen_solver_label,
                            self.typegen_solver_combo_box)])
        self.output_trace_checkbox.connect \
            ("toggled", self.__updateTraceFileOptions)
        self.browse_for_trace_file_button.connect \
            ("clicked", self.__browseForTraceFile)
        browse_cfg_file_button.connect \
            ("clicked", self.__browseForConfigurationFile)
        self.cfg_file_entry.connect_object \
            ("changed", lambda w: w.set_sensitive(True),
             self.save_cfg_file_button)
        self.save_cfg_file_button.connect \
            ("clicked", self.__saveConfigurationFile,
             self.cfg_file_entry.get_text)
        self.paths_view.get_buffer().connect_object \
            ("changed", lambda w: w.set_sensitive(True),
             self.save_cfg_file_button)

        # set sensitivities of widgets based on initial choices
        self.__updateModelCheckingTab()
        self.__updateActlangFlatteningOptions(self.flatten_actlang_button)
        self.__updateUbocoBmcOptions \
            (self.uboco.bounded_model_checking_checkbox)
        self.__updateProcoCexMinimizationCheckbox \
            (self.proco.search_mode_combo_box)
        for tab in (self.uboco, self.suboco):
            self.__updateQueueOverdraftCheckbox(tab.allow_queue_overdraft, tab)
        self.__updateRefinementOptions(self.use_refinement_checkbox)
        self.__updateTraceFileOptions(self.output_trace_checkbox)

    def run(self):
        """Run the graphical user interface for setting analysis options.

        If the `Start' button was clicked, returns a sequence of command
        line options to be passed to a back-end script.  As a side effect,
        populates the Configuration object associated with `self' with user
        selections.

        Exits with status 0 if the `Quit' button was clicked or the
        interface dialog was closed.
        """
        while True:
            response = self.dialog.run()
            if response != gtk.RESPONSE_OK:
                self.dialog.destroy()
                sys.exit(0)

            # collect and validate settings
            try:
                # model file
                if not self.model_name_entry.get_text():
                    raise ValueError("Please specify a model file.")
                self.configuration.model_file_name = \
                    abspath(self.model_name_entry.get_text())
                if not os.path.isfile(self.configuration.model_file_name):
                    raise ValueError \
                              ("Model file `%s' does not exist."
                               % self.configuration.model_file_name)

                # action language flattening
                if self.flatten_actlang_button.get_active():
                    self.configuration.flatten_actlang_sends = \
                        self.flatten_actlang_sends.get_active()
                    self.configuration.flatten_actlang_all = \
                        self.flatten_actlang_all.get_active()

                # model preprocessors
                self.configuration.flatten_hierarchy = \
                    self.flatten_hierarchy_button.get_active()
                self.configuration.flatten_action_language = \
                    self.flatten_actlang_button.get_active()
                self.configuration.slice_assertions = \
                    self.slice_assertions_button.get_active()
                self.configuration.flatten_transition_effects = \
                    self.flatten_transition_exprs_button.get_active()

                # determine the chosen back-end, build argument list for it
                active_backend = self.backend_combo_box.get_active()
                if active_backend in (0, 1): # (S)uboco
                    if active_backend == 0: # Uboco
                        self.configuration.backend = backends.Uboco()
                        b = self.uboco
                        args = b.additional_options_entry.get_text().split()
                        self.configuration.backend.use_bmc = \
                            b.bounded_model_checking_checkbox.get_active()
                        if self.configuration.backend.use_bmc:
                            self.configuration.backend.bmc_bound = \
                                self.__validateInt \
                                    (b.max_bound_entry.get_text(),
                                     "Maximum number of BMC steps", 0, None)
                            self.configuration.backend.sat_solver = \
                                backends.Uboco.NUSMV_SAT_SOLVERS \
                                    [b.nusmv_sat_solver_combo_box.get_active()]
                    else: # Suboco
                        self.configuration.backend = backends.Suboco()
                        b = self.suboco
                        args = b.additional_options_entry.get_text().split()
                        self.configuration.backend.smt_solver = \
                            backends.Suboco.SMT_SOLVERS \
                                [b.smt_solver_combo_box.get_active()]
                        max_bound = \
                            self.__validateInt \
                                (b.max_bound_entry.get_text(),
                                 "Maximum number of BMC steps", 0, None)
                        args.extend(("--max-bound", str(max_bound)))
                        self.__validateInt \
                            (b.min_bound_entry.get_text(),
                             "Minimum number of BMC steps", 0, max_bound)
                        args.extend \
                            (("--min-bound", b.min_bound_entry.get_text()))
                    if b.check_assertions_checkbox.get_active():
                        args.append("--check-assertions")
                    if b.check_deadlocks_checkbox.get_active():
                        args.append("--check-deadlock")
                    if b.check_implicit_consumption_checkbox.get_active():
                        args.append("--check-implicit-consumption")
                    if b.check_runtime_errors_checkbox.get_active():
                        args.append("--check-runtime-errors")
                    self.__validateInt \
                        (b.int_bits_entry.get_text(),
                         "Number of bits for the int type", 2, 32)
                    args.extend(("--int-bits", b.int_bits_entry.get_text()))
                    self.__validateInt \
                        (b.event_queue_size_entry.get_text(),
                         "Capacity of event queues", 0, None)
                    args.extend \
                        (("--queue-size", b.event_queue_size_entry.get_text()))
                    semantics = b.execution_semantics_combo_box.get_active()
                    if semantics in (0, 1): # step semantics
                        if b.allow_queue_overdraft.get_active():
                            args.append("--allow-queue-overdraft")
                        if semantics == 1: # static step semantics
                            args.append("--encode-static")
                    elif semantics == 2: # interleaving semantics
                        args.append("--encode-interleaving")
                else: # Proco
                    self.configuration.backend = backends.Proco()
                    b = self.proco
                    args = b.additional_options_entry.get_text().split()
                    args.extend \
                        (("--max-search-depth",
                          str(self.__validateInt \
                                  (b.max_bound_entry.get_text(),
                                   "Maximum search depth", 0, None))))
                    if b.check_assertions_checkbox.get_active():
                        args.append("--check-assertions")
                    if b.check_deadlocks_checkbox.get_active():
                        args.append("--check-deadlock")
                    if b.check_implicit_consumption_checkbox.get_active():
                        args.append("--check-implicit-consumption")
                    if b.check_queue_overflows_checkbox.get_active():
                        args.append("--check-queue-overflows")
                    self.__validateInt \
                        (b.event_queue_size_entry.get_text(),
                         "Capacity of event queues", 1, None)
                    args.extend \
                        (("--queue-size",
                          b.event_queue_size_entry.get_text()))
                    if b.search_mode_combo_box.get_active() == 1:
                        args.append("--use-bfs")
                    elif b.minimize_cex_checkbox.get_active():
                        args.append("--use-cex-minimization")
                    if b.use_collapse.get_active():
                        args.append("--use-collapse")

                # abstraction refinement settings
                self.configuration.use_refinement = \
                    self.use_refinement_checkbox.get_active()
                self.configuration.integer_semantics = \
                    self.configuration.ABSTRACT_TYPE_SEMANTICS \
                        [self.type_semantics_combo_box.get_active()].lower()
                self.configuration.typegen_smt_solver = \
                    self.configuration.ABSTRACT_TYPEGEN_SOLVERS \
                        [self.typegen_solver_combo_box.get_active()]
                if self.flattening_limit_entry.get_text().strip().lower() \
                       == "none":
                    self.configuration.flattening_limit = -1
                else:
                    self.configuration.flattening_limit = \
                        self.__validateInt \
                            (self.flattening_limit_entry.get_text(),
                             "Limit for flattening abstract operations", 0,
                             None)

                # counterexample settings
                if self.output_trace_checkbox.get_active():
                    self.configuration.trace_file_name = \
                        abspath(self.trace_file_name_entry.get_text())
                else:
                    self.configuration.trace_file_name = None
                self.configuration.simulate_counterexample = \
                    self.simulate_cex_checkbox.get_active()

                # if the contents of the configuration file text buffer
                # have not been saved, save the contents of the buffer into
                # a temporary file and use that file as the configuration
                # file
                self.configuration.cfg_file_is_a_tempfile = \
                    self.save_cfg_file_button.get_property("sensitive")
                if not self.configuration.cfg_file_is_a_tempfile:
                    self.configuration.cfg_file_name = \
                        abspath(self.cfg_file_entry.get_text())
                if self.configuration.cfg_file_is_a_tempfile:
                    filename = self.__saveConfigurationFile(None)
                    if filename == None:
                        continue
                    self.configuration.cfg_file_name = abspath(filename)

                self.configuration.extra_arguments = args
                self.dialog.hide()
                return

            except ValueError, e:
                error_dialog = \
                    gtk.MessageDialog \
                        (self.dialog,
                         flags=gtk.DIALOG_MODAL
                               | gtk.DIALOG_DESTROY_WITH_PARENT,
                         type=gtk.MESSAGE_ERROR, buttons=gtk.BUTTONS_CLOSE,
                         message_format="Some of the options have invalid "
                                        "values!")
                error_dialog.format_secondary_text(str(e))
                error_dialog.run()
                error_dialog.destroy()

    def __browseForFile(self, title, filters, directory, dialog_type=None):
        """Open a file selection dialog for choosing a file name.

        Arguments:
            title       -- Title for the dialog window.
            filters     -- A list of pairs, each of which should consist of
                           a sequence of file name filters and their
                           description.
            directory   -- Name of the directory in which the file selection
                           dialog should be opened.
            dialog_type -- One of the gtk.FILE_CHOOSER_ACTION_ constants
                           determining the type of the dialog to be opened.

        If a file was selected, returns a pair (<file name>,
        <directory name>).  If the operation was cancelled, (None, None)
        will be returned.
        """
        if dialog_type == None:
            dialog_type = gtk.FILE_CHOOSER_ACTION_OPEN
        dialog = \
            gtk.FileChooserDialog \
                (title, parent=self.dialog, action=dialog_type,
                 buttons=(gtk.STOCK_OK, gtk.RESPONSE_OK, gtk.STOCK_CANCEL,
                          gtk.RESPONSE_CANCEL))
        dialog.set_current_folder(directory)
        filters.append((("*",), "All files (*)"),)
        for (extensions, description) in filters:
            file_filter = gtk.FileFilter()
            for e in extensions:
                file_filter.add_pattern(e)
            file_filter.set_name(description)
            dialog.add_filter(file_filter)
        try:
            response = dialog.run()
            if response == gtk.RESPONSE_OK:
                return (dialog.get_filename(), dialog.get_current_folder())
            return (None, None)
        finally:
            dialog.destroy()

    def __browseForModel(self, button):
        """Open a file selection dialog for choosing a model file.

        Argument:
            button -- Dummy argument required for using the method as a
                      handler for a signal emitted by clicking a button.
        """
        (filename, newdir) = \
            self.__browseForFile \
                ("Select model to analyze",
                 [(("*.xmi", "*.xml"), "Model files (*.xmi, *.xml)")],
                 self.current_model_directory)
        if filename != None:
            self.model_name_entry.set_text(filename)
            self.current_model_directory = newdir

    def __browseForTraceFile(self, button):
        """Open a file selection dialog for counterexample traces.

        Argument:
            button -- Dummy argument required for using the method as a
                      handler for a signal emitted by clicking a button.
        """
        (filename, newdir) = \
            self.__browseForFile \
                ("Select trace file",
                 [(("*.trace",), "Trace files (*.trace)")],
                 self.current_trace_file_directory,
                 dialog_type=gtk.FILE_CHOOSER_ACTION_SAVE)
        if filename != None:
            self.trace_file_name_entry.set_text(filename)
            self.current_trace_file_directory = newdir

    def __browseForConfigurationFile(self, button):
        """Open a file selection dialog for choosing a configuration file.

        Argument:
            button -- Dummy argument required for using the method as a
                      handler for a signal emitted by clicking a button.
        """
        (filename, newdir) = \
            self.__browseForFile \
                ("Select configuration file",
                 [(("*.conf",), "Configuration files (*.conf)")],
                 self.current_cfg_directory)
        if filename != None:
            self.cfg_file_entry.set_text(filename)
            self.__readConfigurationFile()
            self.current_cfg_directory = newdir

    def __readConfigurationFile(self):
        """Read the contents of a paths cfg. file into its text buffer."""
        filename = self.cfg_file_entry.get_text()
        buf = self.paths_view.get_buffer()
        buf.delete(buf.get_start_iter(), buf.get_end_iter())
        if os.path.isfile(filename):
            cfg_file = open(filename, "r")
            try:
                try:
                    for line in cfg_file:
                        buf.insert_at_cursor(unicode(line))
                finally:
                    cfg_file.close()
            except Exception:
                # ignore all exceptions while reading the file
                pass

    def __saveConfigurationFile(self, button, getFilename=None):
        """Save the contents of a paths configuration file.

        Arguments:
            button      -- If not None, this button will be made
                           insensitive after saving the file.
            getFilename -- A function for obtaining the name of the target
                           file, or None if a temporary output file should
                           be created.

        On success, returns the name of the target file.  On failure,
        displays an error dialog and returns None.
        """
        buf = self.paths_view.get_buffer()
        try:
            if getFilename == None:
                (fd, filename) = mkstemp()
                os.close(fd)
            else:
                filename = getFilename()
            cfg_file = open(filename, "w")
            try:
                cfg_file.write \
                    (buf.get_text(buf.get_start_iter(), buf.get_end_iter()))
                if button:
                    button.set_sensitive(False)
                return filename
            finally:
                cfg_file.close()
        except Exception, e:
            dialog = \
                gtk.MessageDialog \
                    (parent=self.dialog,
                     type=gtk.MESSAGE_ERROR,
                     flags=gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                     buttons=gtk.BUTTONS_CLOSE,
                     message_format="Failed to write configuration!")
            dialog.format_secondary_text(str(e))
            dialog.run()
            dialog.destroy()
            return None

    def __updateActlangFlatteningOptions(self, checkbox):
        """Update the sensitivities of action language flattening options.

        Argument:
            checkbox -- A gtk.CheckButton (representing the checkbox which
                        is used to toggle whether action language
                        flattening is in use or not).
        """
        # enable selecting the action language flattening mode only if action
        # language flattening is in use
        enabled = checkbox.get_active()
        for widget in (self.flatten_actlang_control_flow_only,
                       self.flatten_actlang_sends,
                       self.flatten_actlang_all):
            widget.set_sensitive(enabled)

    def __updateModelCheckingTab(self, widget=None):
        """Update the currently visible back-end in the model checking tab.

        Argument:
            widget -- A dummy argument required for using the method as
                      a signal handler.
        """
        old_backend = self.current_backend
        self.current_backend = \
            self.backends[self.backend_combo_box.get_active()]
        if old_backend:
            old_backend.table.hide()
            # copy the settings from the previously selected back-end to the
            # newly selected one
            for attr in dir(old_backend):
                if hasattr(self.current_backend, attr):
                    o_a = getattr(old_backend, attr)
                    c_a = getattr(self.current_backend, attr)
                    if isinstance(o_a, gtk.Entry):
                        c_a.set_text(o_a.get_text())
                    elif isinstance(o_a, (gtk.CheckButton, gtk.ComboBox)):
                        c_a.set_active(o_a.get_active())
                        c_a.set_sensitive(o_a.get_property("sensitive"))
                    elif isinstance(o_a, gtk.Expander):
                        c_a.set_expanded(o_a.get_expanded())
                        self.__updateOptionExpanderLabel(o_a, None)
        self.current_backend.table.show_all()

    def __updateOptionExpanderLabel(self, expander, param_spec):
        """Update the label of an 'Additional options' expander.

        Arguments:
            expander   -- A gtk.Expander.
            param_spec -- Dummy argument required for using the method as
                          a signal handler.
        """
        expander.set_label \
            ('<span size="small"><i>'
             + (expander.get_expanded() and "Hide" or "Show")
             + " additional options</i></span>")

    def __updateUbocoBmcOptions(self, checkbox):
        """Update BMC-related widget sensitivities for the Uboco back-end.

        Argument:
            checkbox -- A gtk.CheckButton (representing the checkbox which
                        is used to toggle whether to use bounded model
                        checking or not).
        """
        enable = checkbox.get_active()
        for w in (self.uboco.max_bmc_bound_hbox, self.uboco.max_bound_entry,
                  self.uboco.sat_solver_label,
                  self.uboco.nusmv_sat_solver_combo_box):
            w.set_sensitive(enable)

    def __updateProcoCexMinimizationCheckbox(self, combo_box):
        """Update widget sensitivities for the Proco back-end.

        Argument:
            combo_box -- A gtk.ComboBox (representing the combo box for
                         choosing the search mode).
        """
        # allow toggling counterexample minimization only if using DFS mode
        self.proco.minimize_cex_checkbox.set_sensitive \
            (combo_box.get_active() == 0)

    def __updateQueueOverdraftCheckbox(self, combo_box, backend_widgets):
        """Update queue overdraft checkbox sensitivity for (S)uboco.

        Arguments:
            combo_box       -- A gtk.ComboBox (representing the combo box
                               for choosing the execution semantics).
            backend_widgets -- An object containing widgets for a back-end.
        """
        # allow toggling queue overdraft checkbox only if using step semantics
        backend_widgets.allow_queue_overdraft.set_sensitive \
            (combo_box.get_active() != 2)

    def __updateRefinementOptions(self, checkbox):
        """Update widget sensitivities on the abstraction refinement tab.

        Arguments
            checkbox -- A gtk.CheckButton (representing the checkbox for
                        toggling whether to use abstraction refinement).
        """
        # allow setting abstraction refinement options only if abstraction
        # refinement is active
        self.abstraction_refinement_options.set_sensitive \
            (checkbox.get_active())

    def __updateTraceFileOptions(self, checkbox):
        """Update trace file name entry and browse button sensitivities.

        Argument:
            checkbox -- A gtk.CheckButton (representing the checkbox for
                        toggling whether counterexamples should be saved
                        to a file).
        """
        # allow modifying the name of the trace file only if trace files are to
        # be saved somewhere
        enabled = checkbox.get_active()
        for widget in (self.trace_file_name_entry,
                       self.browse_for_trace_file_button):
            widget.set_sensitive(enabled)

    def __validateInt(self, value, desc, min_val, max_val):
        """Check whether a value represents an integer between a given range.

        Arguments:
            value   -- Value to test.
            desc    -- Textual description of the value.
            min_val -- Lower bound of the accepted (inclusive) range, or
                       None if there is no lower bound.
            max_val -- Upper bound of the accepted (inclusive) range, or
                       None if there is no upper bound.

        On success, returns `value' as an integer.  On failure, raises a
        ValueError initialized with an error message.
        """
        try:
            result = int(value)
            if (min_val != None and result < min_val) \
                   or (max_val != None and result > max_val):
                raise ValueError()
            return result
        except ValueError:
            if min_val == None:
                if max_val == 0:
                    msg = "a negative integer"
                elif max_val == 1:
                    msg = "a non-positive integer"
                else:
                    msg = "an integer less than or equal to %d" % max_val
            elif max_val == None:
                if min_val == 0:
                    msg = "a non-negative integer"
                elif min_val == 1:
                    msg = "a positive integer"
                else:
                    msg = "an integer greater than or equal to %d" % min_val
            else:
                msg = \
                    "an integer between %d and %d (inclusive)" \
                    % (min_val, max_val)
            raise ValueError("%s must be %s." % (desc, msg))
