# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Helper function for capturing output from a subtask.

This module defines a function for running a subtask (an external
program or a Python function), capturing (and logging) its output.

Author: Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

import os
import sys
import subprocess
import logging
from threading import Thread, Lock
from traceback import format_tb
from frontend import logutil
from abstraction import error

PIPE = subprocess.PIPE

class __Reader(Thread):
    """Class encapsulating a thread for reading input."""

    # If there are multiple threads logging using the same logger, these
    # variables are used to ensure that only one copy of the message
    # indicating the start of output is logged.  (If no thread produces
    # actual logged output, the message will not be outputted, either.)

    # lock for controlling access to `output_active'
    begin_output_lock = Lock()

    # has any output has been produced yet?
    output_active = False

    # message to display at the start of output
    begin_output_msg = ""

    # read buffer size
    BUFSIZE = 1024

    def reset(cls, msg=""):
        """Reset the static variables.

        Argument:
            msg -- Message to send to the root logger at the start of
                   output the next time when using __Reader threads.
        """
        cls.output_active = False
        cls.begin_output_msg = msg

    reset = classmethod(reset)

    def __init__(self, in_fd, out_fd):
        """Initialize a __Reader object.

        Arguments:
            in_fd  -- File descriptor for the input stream (assumed to
                      be already opened for reading).
            out_fd -- File descriptor for the output stream (assumed to
                      be already opened for writing).  In addition to
                      this file descriptor, the input read from the
                      input stream will be sent also to the root logger
                      using loglevel logutil.LOG_EXTRA_DEBUG.
        """
        Thread.__init__(self)
        self.in_fd = in_fd
        self.out_fd = out_fd

    def run(self):
        """Read input from a file descriptor and write it to another."""
        logger = logging.getLogger("log")
        try:
            while True:
                buf = os.read(self.in_fd, self.BUFSIZE)
                if not buf:
                    return
                self.begin_output_lock.acquire()
                try:
                    if not self.output_active:
                        # display a log message to indicate the start of
                        # actual output
                        logger.extradebug(("----- ", ["gray"]))
                        logger.extradebug \
                            ((self.begin_output_msg,
                              ["small", "italic", "gray"]))
                        logger.extradebug((" -----\n", ["gray"]))
                        self.__class__.output_active = True
                finally:
                    self.begin_output_lock.release()
                os.write(self.out_fd, buf)
                logger.extradebug((buf, ["small", "monospace"]))
        except Exception, e:
            pass

def runSubtask(subtask, msg=None, capture_prefix="prg", ignore_exit_status=False, **kwargs):
    """Execute a subtask, optionally capturing and logging output.

    This function acts as a wrapper for (i) running external programs
    using subprocess.Popen, and (ii) invoking "independent" callable
    Python objects which should be treated as if they were external
    programs.  The standard output and error streams of the subtask can
    optionally be redirected to separate files and passed also to the
    root logger (if the subtask is a callable object, it must support
    the keyword arguments "stdout" and "stderr", in which the object
    can accept stream objects which it should then use as its standard
    output and standard error stream, respectively).

    Arguments:
        subtask            -- A sequence of strings representing the
                              parameters for an external program to run
                              (where the first element in the sequence
                              is the command for running the program
                              itself), or an object supporting the
                              "__call__" method.
        msg                -- A message string to display before
                              starting the subtask.
        capture_prefix     -- A string prefix for the file names used
                              for capturing output from the subtask.  If
                              capturing of output is enabled (via
                              keyword arguments for this function, see
                              `kwargs' below), the standard output
                              (standard error) stream of the subtask
                              will be redirected to
                              `<capture_prefix>.stdout'
                              (`<capture_prefix>.stderr',
                              respectively).
        ignore_exit_status -- If True, no exception will be raised even
                              if the subtask exits with non-zero exit
                              status.
        kwargs             -- Optional keyword arguments.  Setting one
                              (or both) of the keyword arguments
                              `stdout' and `stderr' to PIPE will cause
                              the corresponding stream of the subtask to
                              be redirected to a newly created file (and
                              logged using the "log" logger).  If
                              `subtask' represents an external program,
                              the arguments in `kwargs' will be passed
                              further to `subprocess.Popen' (which is
                              used to actually execute the command).
                              Otherwise, if `subtask' is a callable
                              Python object, `kwargs' should include the
                              key `args' initialized to the sequence of
                              positional arguments for the callable
                              object.  If `kwargs' contains further
                              keyword arguments, they will be passed to
                              the object as well (except possibly for
                              the `stdout' and `stderr' arguments, see
                              above).
    """
    if isinstance(subtask, (list, tuple)):
        if msg == None:
            msg = "running %s" % subtask[0]
        for arg in subtask:
            if not isinstance(arg, str):
                raise ValueError \
                          ("the sequence given as first argument is not a "
                           "sequence of strings")
        __Reader.reset("output from `%s'" % " ".join(subtask))
        subtask_is_external_process = True
    elif hasattr(subtask, "__call__"):
        if msg == None:
            msg = "calling %s" % subtask.__name__
        __Reader.reset("output from subtask")
        subtask_is_external_process = False
        try:
            args = list(kwargs["args"])
            del kwargs["args"]
        except KeyError:
            args = []
    else:
        raise TypeError \
                  ("a sequence of strings or a callable object expected as "
                   "the first argument")
    logger = logging.getLogger("log")
    logger.debug(("%s..." % msg, [], True))
    redirect_fds = {}
    reader_threads = []
    try:
        try:
            # create output files for streams which should be captured
            for suffix in ("stdout", "stderr"):
                try:
                    if kwargs[suffix] == PIPE:
                        redirect_fds[suffix] = \
                            os.open \
                                (capture_prefix + os.extsep + suffix,
                                 os.O_CREAT | os.O_WRONLY, 0600)
                        kwargs[suffix] = subprocess.PIPE
                except KeyError:
                    pass

            if subtask_is_external_process:
                # do not buffer output from the subprocess to be started
                kwargs["bufsize"] = 0
                # run the external program
                child = subprocess.Popen(subtask, **kwargs)
                # create and start threads for reading redirected output
                for (suffix, fd) in redirect_fds.iteritems():
                    thread = __Reader(getattr(child, suffix).fileno(), fd)
                    reader_threads.append(thread)
                    thread.start()
                # wait for the child process to run to completion, then wait
                # for the output reader threads to finish
                try:
                    status = child.wait()
                finally:
                    for thread in reader_threads:
                        thread.join()
                if not ignore_exit_status and status != 0:
                    # child exited abnormally
                    raise Exception \
                              ("`%s' exited with status %d"
                               % (" ".join(subtask), status))
            else:
                # create pipes for redirected streams of the subtask and
                # start threads for reading them
                for (suffix, fd) in redirect_fds.iteritems():
                    (read_fd, write_fd) = os.pipe()
                    thread = __Reader(read_fd, fd)
                    reader_threads.append(thread)
                    thread.start()
                    kwargs[suffix] = os.fdopen(write_fd, "w", 0)
                    if suffix == "stderr":
                        old_stderr = error.getErrorStream()
                # call the object representing the subtask
                try:
                    try:
                        subtask(*args, **kwargs)
                    except SystemExit, s:
                        # trap calls of sys.exit from the subtask
                        if not ignore_exit_status and s.code != 0:
                            raise Exception("subtask exited abnormally")
                    except:
                        # if the subtask caused an exception and the standard
                        # error stream of the subtask was redirected, write a
                        # stack trace to the redirected stream
                        if "stderr" in kwargs:
                            kwargs["stderr"].write \
                                ("".join(format_tb(sys.exc_info()[2])))
                        # re-raise the exception
                        raise
                finally:
                    # ensure that the output streams of the subtask are closed
                    for suffix in redirect_fds.iterkeys():
                        stream = kwargs[suffix]
                        if not stream.closed:
                            try:
                                stream.close()
                            except:
                                pass
                        if suffix == "stderr":
                            error.setErrorStream(old_stderr)
                    # wait for the output reader threads to finish
                    for thread in reader_threads:
                        thread.join()
        finally:
            for fd in redirect_fds.itervalues():
                try:
                    os.close(fd)
                except OSError:
                    pass
            if __Reader.output_active:
                logger.extradebug(("\n----- ", ["gray"]))
                logger.extradebug \
                    (("end of output", ["small", "italic", "gray"]))
                logger.extradebug((" -----", ["gray"]))
        logger.dedent(logging.DEBUG, ((" done\n", ["green", "italic"]),))
    except Exception, e:
        # in case of error, display output from the subtask if it was captured
        if not logger.isEnabledFor(logutil.LOG_EXTRA_DEBUG):
            for suffix in ("stdout", "stderr"):
                name = capture_prefix + os.extsep + suffix
                if os.path.isfile(name):
                    try:
                        logger.log \
                            (logutil.LOG_DEBUG_FAILURE,
                                 ("----- %s -----\n" % name,
                                  ["bold", "underline"]))
                        f = open(name, "r")
                        try:
                            for line in f:
                                logger.log \
                                    (logutil.LOG_DEBUG_FAILURE,
                                         (line, ["small", "monospace"]))
                        finally:
                            f.close()
                    except IOError, e:
                        logutil.logFailure(e, "I/O error")
        logger.dedent \
            (logging.DEBUG,
             ((" failed", ["red", "italic"]), (" [%s]\n" % str(e), [])))
        raise error.Error("aborted")

