#! /usr/bin/python2.4

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Usage: generate_abstract_type [OPTION]...
Generate abstract versions of the operations in the Jumbala action language
for user-defined abstract integer types.

Available options:
  -c, --coercion=FROM_TYPE:TO_TYPE
                             Generate a coercion from FROM_TYPE to TO_TYPE,
                             where FROM_TYPE and TO_TYPE are names of two
                             distinct types declared in the input.  By
                             default, no coercions between abstract types
                             are generated automatically.
  -h, --help                 Display this help and exit.
  -i, --inputfile=FILENAME   Specify input file name [default: "-", which
                             refers to standard input].
  -o, --outputfile=FILENAME  Specify output file name [default: "-", which
                             refers to standard output].
  -s, --solver=SOLVER:COMMAND
                             Specify the external tool to be used for
                             solving constraints when defining operations
                             for abstract types.  SOLVER can be one of
                             CVC3, MathSAT, STP, Yices, or Z3; COMMAND
                             should be the command to run the solver
                             executable.
  -I, --include-operators=OPERATOR-LIST
                             Generate abstract versions of only the
                             operations specified in OPERATOR-LIST
                             [default: generate abstract versions of all
                             abstractable operations].
  -E, --exclude-operators=OPERATOR-LIST
                             Do not generate abstract versions of the
                             operations specified in OPERATOR-LIST
                             [default: do not exclude any operations].
  -v, --verbose              Show verbose output.
  -x, --model-runtime-exceptions
                             If the generation of abstract versions of
                             division, remainder, or one of the bit shift
                             operations is requested, extend the abstract
                             domains with a special value INVALID to
                             represent the "outcome" of a possible runtime
                             exception (division or modulo by zero, or
                             shift by a negative number of bit positions).
                             Any arithmetic operation involving the abstract
                             value INVALID will again yield the value
                             INVALID; any Boolean operation involving this
                             value will yield either true or false.

  The -I and -E options require as a parameter a comma-separated
  OPERATOR-LIST of operator or test symbols supported by the Jumbala
  action language.  If present in OPERATOR-LIST, the symbol + (-) will
  cause both the unary and the binary variants of the + operator (-
  operator, respectively) to be included or excluded; to include or
  exclude only one variant of one of these operators, prepend the arity
  (1 or 2) of the desired variant to the operator symbol (for example,
  use the symbol `1-' to refer to the unary variant of the - operator).

USER-DEFINED ABSTRACT TYPES
  The script accepts input in the format used in the SMUML toolset for
  defining abstract type libraries.  The definition of each abstract type
  should begin with a line starting with the keyword "type" followed by
  the name of the type, a colon ":", and a non-empty comma-separated list
  of symbolic names for the values in the domain of the type.  This line
  should, at a minimum, be followed by a definition of an abstraction
  mapping from integers to values in the domain of the abstract type for
  the Jumbala back-end, i.e., a line "coercion int Jumbala" followed by
  one or more lines, each of which consists of
     - a Boolean expression template in the Jumbala action language
       (in which the construct "<>" refers to the integer parameter of the
       template),
     - a colon ":", and
     - a comma-separated list of values from the domain of the abstract
       type.
  The definition of the abstraction mapping is terminated by a line
  containing the single keyword "end".  The abstraction mapping is
  considered to be ill-defined if there is an integer for which none of
  the Boolean expressions obtained by substituting the integer for the
  template parameter evaluates to true.  The script is not in general
  able to enforce the well-definedness of the abstraction mapping
  (the ability to do this depends on the capabilities of the chosen
  external solver), so the ultimate responsibility of ensuring that no
  abstraction mapping is ill-defined remains with the user.  (This is
  easy by simply using "true" as the condition of the last test in each
  abstraction mapping.)

  If the input for the script already includes definitions of operators
  or tests (or coercions from other types or for other back-ends) for
  the abstract type, these definitions will override any automatically
  generated ones.  This allows, for example, the operations for an
  abstract type to be constructed incrementally by feeding the script
  its own output using different values for the -I option on each
  invocation of the script.

  The script accepts multiple abstract type definitions in the same
  input file.  In particular, using the -c option for automatically
  generating coercions between abstract types necessitates that the
  types are defined in the same input file.

  For example, the following input is sufficient for automatically
  generating the abstract operations for the "sign" abstraction.
  
      type Sign: NEG, ZERO, POS
        coercion int Jumbala
          <> < 0 : NEG
          <> == 0: ZERO
          true   : POS
          end

CAVEATS
  The PySMT interface used for generating abstract operations has no
  direct support for signed division ("/"), remainder ("%"), and the bit
  shift operators ("<<", ">>", ">>>") of the Jumbala action language due
  to limitations in the SMT-LIB format supported by many Satisfiability
  Modulo Theories as their input language.  Consequently, when generating
  abstract versions of these operations, the translation of the semantics
  of these operators into PySMT circuits only approximates the operators'
  actual action language semantics as follows:
    * When translating Boolean expression templates of abstraction
      mappings, it is assumed that the evaluation of a template using the
      action language interpreter will not cause a runtime exception for
      any concrete integer substituted for the template parameter.  That
      is, it is assumed that the divisor/modulus in every
      division/remainder expression is always nonzero, and that shift
      counts for bit shifts are always nonnegative.
    * By default, the above assumption is applied also to the definitions
      of the abstract operations.  That is, the definitions of abstract
      division, modulo, and bit shift operations are safe
      overapproximations of the corresponding concrete operations only for
      those concrete operand values for which the concrete operation is
      defined.
    * Instead of the default behavior, it is possible to define abstract
      operations which model the "result" of operations on invalid
      (concrete) operand combinations explicitly using a special abstract
      value INVALID.  This behavior is activated by the -x command line
      option.  When using this option, the result of each abstract
      operation on a combination of abstract operands will include the
      value INVALID if the concretization of the abstract operands covers
      at least one combination of concrete values for which the
      corresponding concrete operation is not defined.  The domains of
      the types will be automatically extended with the value INVALID
      if necessary.  Note, however, that the special semantics of the
      INVALID value is not currently supported by any model analysis
      tools in the SMUML toolset.

This script was written by Heikki Tauriainen for the SMUML project
<http://www.tcs.hut.fi/Research/Logic/SMUML.shtml>.
"""

import sys
import getopt
from abstraction import error
from abstraction import smtlib_typegen
from abstraction.typelibrary import TypeLibrary

# -----------------------------------------------------------------------------

lib = TypeLibrary()

try:
    options, args = \
        getopt.getopt \
            (sys.argv[1:], "c:E:I:i:ho:s:vx",
             ["coercion=", "exclude-operators=", "include-operators=",
              "inputfile=", "solver=", "outputfile=", "help",
              "model-runtime-exceptions", "verbose"])
except getopt.GetoptError, e:
    error.abort(str(e), 2)

try:
    input_file_name = "-"
    output_file_name = "-"
    include_ops = []
    exclude_ops = []
    solver = None
    model_runtime_exceptions = False
    verbose = False
    coercions_to_generate = set()

    # Process command line options
    for (option, value) in options:
        if option in ("-c", "--coercion"):
            types = value.split(":")
            if len(types) != 2:
                raise error.OptionError \
                          ("syntax error in argument `%s' for %s"
                           % (value, option))
            coercions_to_generate.add(tuple(types))
        elif option in ("-E", "--exclude-operators"):
            if include_ops or exclude_ops:
                raise error.OptionError \
                          ("only one -I or -E argument is allowed")
            exclude_ops = value.split(",")
        elif option in ("-I", "--include-operators"):
            if include_ops or exclude_ops:
                raise error.OptionError \
                          ("only one -I or -E argument is allowed")
            include_ops = value.split(",")
        if option in ("-i", "--inputfile"):
            input_file_name = value
        elif option in ("-h", "--help"):
            sys.stdout.write(__doc__)
            sys.exit(0)
        elif option in ("-o", "--outputfile"):
            output_file_name = value
        elif option in ("-s", "--solver"):
            args = value.split(':')
            solver = args[0].lower()
            if not solver in map(str.lower, smtlib_typegen.PySmt.SMT_SOLVERS):
                raise error.OptionError("unknown solver `%s'" % args[0])
            if len(args) < 2 or not args[1]:
                raise error.OptionError \
                          ("missing path to the solver executable")
            solver = smtlib_typegen.PySmt(solver, args[1], None, sys.stderr)
        elif option in ("-x", "--model-runtime-exceptions"):
            model_runtime_exceptions = True
        elif option in ("-v", "--verbose"):
            verbose = True
    if not solver:
        raise error.OptionError("use the -s option to specify a solver to use")

    lib.read(input_file_name)

    # Collect the operations for which to try to generate abstract versions.
    op_arity = {"1+" : ("+",  1), "1-"  : ("-",   1), "~"  : ("~", 1),
                "2+" : ("+",  2), "2-"  : ("-",   2), "*"  : ("*", 2),
                "/"  : ("/",  2), "%"   : ("%",   2), "&"  : ("&", 2),
                "|"  : ("|",  2), "^"   : ("^",   2), "<<" : ("<<", 2),
                ">>" : (">>", 2), ">>>" : (">>>", 2), "==" : ("==", 2),
                "!=" : ("!=", 2), "<"   : ("<",   2), "<=" : ("<=", 2),
                ">"  : (">",  2), ">="  : (">=",  2)}

    typenames = [t for t in lib.typenames if not t in lib.STD_TYPES]

    if not include_ops:
        include_ops = op_arity.keys()

    for ops in (include_ops, exclude_ops):
        if "+" in ops:
            ops.extend(["1+", "2+"])
        if "-" in ops:
            ops.extend(["1-", "2-"])

    operations_to_generate = []
    for (op, (symbol, arity)) in op_arity.iteritems():
        if not op in include_ops or op in exclude_ops:
            continue
        for t in typenames:
            try:
                # do not overwrite old definitions of operators and tests
                if symbol in lib.operators[t][arity] \
                        or symbol in lib.tests[t][arity]:
                    continue
            except KeyError:
                operations_to_generate.append((t, symbol, arity))
            try:
                # do not overwrite coercion from "int" for the "python" backend
                if lib.coercions[t]["int"]["python"]:
                    continue
            except KeyError:
                operations_to_generate.append(("coercion", "int", t, "python"))

    for (fromtype, totype) in coercions_to_generate:
        for backend in ("Jumbala", "python"):
            # do not overwrite old definitions of coercions
            if not totype in lib.coercions \
                    or not fromtype in lib.coercions[totype] \
                    or not backend in lib.coercions[totype][fromtype]:
                operations_to_generate.append \
                    (("coercion", fromtype, totype, backend))

    smtlib_typegen.generateOperations \
        (lib, operations_to_generate, solver, model_runtime_exceptions,
         verbose=verbose, stderr=sys.stderr)

    lib.write(output_file_name)

except error.OptionError, e:
    error.abort(str(e), 2)
except error.Error, e:
    error.abort(str(e))
