#! /usr/bin/python2.4

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Usage: generate_interval_abstraction [OPTION]... [PARTITION]...
Generate an abstract type library with abstract types corresponding to
finite partitions of the set of (unbounded) integers.  Each generated
abstract type will support the basic arithmetic (+, -, *, /) and test
(==, !=, <, >, <=, >=) operations of the Jumbala action language.  (As
in Jumbala, the result of integer division is always rounded towards
zero.)

Each PARTITION should be a type name followed by a colon (`:') and a
comma-separated list of intervals (with no spaces in between) of the
following form (in each case without the quotes):
    - "*" or "(-inf;inf)"
      all integers
    - "n" or "[n;n]"
      an interval which contains the single integer n
    - "..n" or "(-inf;n]"
      an interval extending from -infinity to the integer n (inclusive)
    - "n..m" or "[n;m]"
      an interval extending from the integer n to the integer m
      (inclusive)
    - "n.." or "[n;inf)"
      an interval starting from the integer n and extending to
      +infinity (inclusive)
The type name and the colon can be omitted (in which case a type name
will be generated automatically).  No interval in the list of intervals
should be empty; nor should any two intervals in the list overlap.  The
intervals should be listed in the increasing order of their endpoints.
It is not necessary for the intervals to form a partition of the set of
all integers; additional intervals will be generated automatically to
cover ranges which do not fall within any explicitly specified interval.

Additionally, the script accepts the following options:
  -c, --coercion=SOURCE-TYPE:TARGET-TYPE
                             Augment the generated type library with
                             the definition of a coercion from the
                             type SOURCE-TYPE to the type TARGET-TYPE.
                             Both SOURCE-TYPE and TARGET-TYPE must be
                             among the types listed in the PARTITION
                             specifications.  To generate multiple
                             coercions, specify this option separately
                             for each pair of source and target types.
                             By default, no coercions between abstract
                             types are generated automatically.
  -h, --help                 Display this help and exit.
  -o, --output-file=FILENAME
                             Specify output file name [default: `-',
                             which stands for standard output].
  -I, --include-operators=OPERATOR-LIST
                             Generate abstract versions of only the
                             operations specified in OPERATOR-LIST
                             [default: generate abstract versions of
                             all basic arithmetic and test operations].
  -E, --exclude-operators=OPERATOR-LIST
                             Do not generate abstract versions of the
                             operations specified in OPERATOR-LIST
                             [default: do not exclude any operations].
  -v, --verbose              Show verbose output.
  -x, --model-runtime-exceptions
                             If integer division (`/') belongs to the
                             set of included operators, extend the
                             abstract domain of every generated type
                             with a special value INVALID to represent
                             the "result" of division by zero.  Any
                             arithmetic operation involving the abstract
                             value INVALID will again yield the value
                             INVALID; any Boolean operation involving
                             this value may yield either one of the
                             two possible truth values.

  The -I and -E options require as a parameter a comma-separated
  OPERATOR-LIST of operator or test symbols supported by the Jumbala
  action language.  If present in OPERATOR-LIST, the symbol + (-) will
  cause both the unary and the binary variants of the + operator (-
  operator, respectively) to be included or excluded; to include or
  exclude only one variant of one of these operators, prepend the arity
  (1 or 2) of the desired variant to the operator symbol (for example,
  use the symbol `1-' to refer to the unary variant of the - operator).

CAVEATS
  The operations generated for the abstract type will approximate the
  corresponding concrete operations on unbounded signed integers
  (unless the -x option is used, the case of division by zero is
  considered to never occur), not 32-bit signed integers used in the
  Jumbala action language.  Consequently, the generated abstract
  operations are not sound approximations of the corresponding concrete
  operations in the Jumbala action language since they ignore the
  possibility of signed integer overflow.  Use the
  `generate_abstract_type.py' script to generate abstract types whose
  operations soundly approximate the semantics of the Jumbala action
  language.

  Because of the use of unbounded integers, generating an abstract type
  whose domain includes intervals, some endpoints of which are outside
  of the range of integers supported by the Jumbala action language
  (i.e., the interval [-2^31;(2^31)-1]) may result in undesired behavior
  if the generated coercion function from integers to abstract values is
  used for abstract model generation and verification.

  The special semantics of the INVALID value is not currently supported
  by any model analysis tools in the SMUML toolset.

This script was written by Heikki Tauriainen for the SMUML project
<http://www.tcs.hut.fi/Research/Logic/SMUML.shtml>.
"""

import sys
import re
import getopt
from abstraction import error

class Integer:
    """Representation of integers with positive and negative infinity.

    The intuitive semantics of the infinities is that the values (+inf)
    and (-inf) represent _any_ (finite) positive or negative integer,
    respectively.
    """
    
    INFINITY = (-1,)
    NEG_INFINITY = (1,)
    
    def __init__(self, i):
        """Initialize an integer.

        Argument:
            i -- A plain integer or another Integer object.
        """
        if isinstance(i, Integer):
            self.value = i.value
        elif isinstance(i, (int, long)) \
                 or i in (Integer.INFINITY, Integer.NEG_INFINITY):
            self.value = i
        else:
            raise TypeError("invalid type for initializer")

    def __str__(self):
        """Return the string representation of an Integer object."""
        if self.value == Integer.INFINITY:
            return "+inf"
        elif self.value == Integer.NEG_INFINITY:
            return "-inf"
        else:
            return str(self.value)

    def __add__(self, i):
        """Add two integers.

        Argument:
            i -- An Integer object.

        Returns:
            The sum of the integers represented by `self' and `i' as a
            new Integer object.
            The semantics of addition for infinities is
                `x + (+inf)' == `(+inf) + x' == (+inf)  [-inf < x <= +inf],
                `x + (-inf)' == `(-inf) + x' == (-inf)  [-inf <= x < +inf].
            Attempting to add (+inf) to (-inf) or vice versa will cause
            a ValueError to be raised.
        """
        if (self.value == Integer.INFINITY
            and i.value == Integer.NEG_INFINITY) \
            or (self.value == Integer.NEG_INFINITY
                and i.value == Integer.INFINITY):
            raise ValueError("cannot add +inf and -inf")
        if self.value in (Integer.INFINITY, Integer.NEG_INFINITY):
            return Integer(self.value)
        elif i.value in (Integer.INFINITY, Integer.NEG_INFINITY):
            return Integer(i.value)
        else:
            return Integer(self.value + i.value)

    def __sub__(self, i):
        """Subtract an integer from another one.

        Argument:
            i -- An Integer object.

        Returns:
            The difference of the integers represented by `self' and `i'
            as a new Integer object.
            The semantics of subtraction for infinities is
                `(+inf) - x' == (+inf)  [-inf <= x < +inf],
                `x - (-inf)' == (+inf)  [-inf < x <= +inf],
                `(-inf) - x' == (-inf)  [-inf < x <= +inf],
                `x - (+inf)' == (-inf)  [-inf <= x < inf].
            Attempting to subtract (+inf) or (-inf) from itself will
            cause a ValueError to be raised.
        """
        if (self.value == Integer.INFINITY
            and i.value == Integer.INFINITY) \
            or (self.value == Integer.NEG_INFINITY
                and i.value == Integer.NEG_INFINITY):
            raise ValueError("cannot subtract +-inf from itself")
        if self.value in (Integer.INFINITY, Integer.NEG_INFINITY):
            return Integer(self.value)
        elif i.value == Integer.INFINITY:
            return Integer(Integer.NEG_INFINITY)
        elif i.value == Integer.NEG_INFINITY:
            return Integer(Integer.INFINITY)
        else:
            return Integer(self.value - i.value)

    def __mul__(self, i):
        """Multiply an integer with another one.

        Argument:
            i -- An Integer object.

        Returns:
            The product of the integers represented by `self' and `i'
            as a new Integer object.
            The semantics of multiplication for infinities is
                `(+inf) * x' == `x * (+inf)' == (-inf)  [-inf <= x < 0]
                `(+inf) * x' == `x * (+inf)' == 0       [x == 0]
                `(+inf) * x' == `x * (+inf)' == (+inf)  [0 < x <= +inf]
                `(-inf) * x' == `x * (-inf)' == (+inf)  [-inf <= x < 0]
                `(-inf) * x' == `x * (-inf)' == 0       [x == 0]
                `(-inf) * x' == `x * (-inf)' == (-inf)  [0 < x <= +inf]
        """
        if self.value == 0 or i.value == 0:
            return Integer(0)
        neg = (self.ge(Integer(0)) != i.ge(Integer(0)))
        if self.value in (Integer.INFINITY, Integer.NEG_INFINITY) \
               or i.value in (Integer.INFINITY, Integer.NEG_INFINITY):
            return Integer((neg and Integer.NEG_INFINITY) or Integer.INFINITY)
        else:
            return Integer(self.value * i.value)

    def __floordiv__(self, i):
        """Integer division between two integers.

        Argument:
            i -- An Integer object.

        Returns:
            The result of integer division between the integers
            represented by `self' and `i' as a new Integer object.
            Note that the semantics of integer division reflect the
            semantics of the Jumbala action language in that the
            result of the division is always rounded towards zero
            and not towards -infinity (as in Python).
            The semantics of integer division for infinities is
                `x div (+inf)' == `x div (-inf)' == 0  [-inf < x < +inf]
                `(+inf) div x' == -inf                 [-inf <= x < 0]
                `(+inf) div x' == +inf                 [0 < x <= +inf]
                `(-inf) div x' == +inf                 [-inf <= x < 0]
                `(-inf) div x' == -inf                 [0 < x <= +inf]
            Division by zero will cause a ZeroDivisionError to be
            raised.
        """
        if i.value == 0:
            raise ZeroDivisionError("integer division or modulo by zero")
        elif self.value == 0:
            return Integer(0)
        neg = (self.ge(Integer(0)) != i.ge(Integer(0)))
        if self.value in (Integer.INFINITY, Integer.NEG_INFINITY):
            return Integer((neg and Integer.NEG_INFINITY) or Integer.INFINITY)
        elif i.value in (Integer.INFINITY, Integer.NEG_INFINITY):
            return Integer(0)
        else:
            result = abs(self.value) // abs(i.value)
            if neg:
                return Integer(-result)
            else:
                return Integer(result)

    def __neg__(self):
        """Negate an integer.

        Returns:
            The result of negating the integer represented by `self'.
        """
        return self * Integer(-1)

    def eq(self, i):
        """Test two Integer objects for equality.

        Argument:
            i -- An Integer object.

        Returns:
            True iff neither `self' or `i' represent one of the infinities
            and the integer values of `self' and `i' are equal.
        """
        if self.value in (Integer.INFINITY, Integer.NEG_INFINITY) \
           or i.value in (Integer.INFINITY, Integer.NEG_INFINITY):
            return False
        return (self.value == i.value)

    def neq(self, i):
        """Test two Integer objects for inequality.

        Argument:
            i -- An Integer object.

        Returns:
            True iff neither `self' or `i' represent one of the infinities
            and the integer values of `self' and `i' are not equal.
        """
        if self.value in (Integer.INFINITY, Integer.NEG_INFINITY) \
           or i.value in (Integer.INFINITY, Integer.NEG_INFINITY):
            return False
        return (self.value != i.value)

    def lt(self, i):
        """Test whether an integer is "less than" another Integer object.

        Argument:
            i -- An Integer object.

        Returns:
            True iff `self' represents an integer "less than" `i'.
            The semantics of "less than" comparison for infinities is:
                `x < -inf' == False  [-inf <= x <= +inf]
                `-inf < x' == True   [-inf < x <= +inf]
                `x < +inf' == True   [-inf <= x < +inf]
                `+inf < x' == False  [-inf <= x <= +inf]
        """
        if self.value == Integer.INFINITY or i.value == Integer.NEG_INFINITY:
            return False
        elif self.value == Integer.NEG_INFINITY or i.value == Integer.INFINITY:
            return True
        else:
            return (self.value < i.value)

    def le(self, i):
        """Test whether an integer is "less than or equal" to another.

        Argument:
            i -- An Integer object.

        Returns:
            True iff `self' represents an integer that is "less than or
            equal" to `i'.
        """
        return self.lt(i) or self.eq(i)

    def gt(self, i):
        """Test whether an integer is "greater than" another Integer object.

        Argument:
            i -- An Integer object.

        Returns:
            True iff `self' represents an integer "greater than" `i'.
            The semantics of "greater than" comparison for infinities is:
                `x > -inf' == True   [-inf < x <= +inf]
                `-inf > x' == False  [-inf <= x <= +inf]
                `x > +inf' == False  [-inf <= x <= +inf]
                `+inf > x' == False  [-inf <= x < +inf]
        """
        if self.value == Integer.NEG_INFINITY or i.value == Integer.INFINITY:
            return False
        elif self.value == Integer.INFINITY or i.value == Integer.NEG_INFINITY:
            return True
        else:
            return (self.value > i.value)

    def ge(self, i):
        """Test whether an integer is "greater than or equal" to another.

        Argument:
            i -- An Integer object.

        Returns:
            True iff `self' represents an integer that is "greater than or
            equal" to `i'.
        """
        return self.gt(i) or self.eq(i)

# -----------------------------------------------------------------------------

class Interval:
    """Intervals of integers."""
    
    def __init__(self, lo, hi):
        """Initialize an integer interval.

        Arguments:
            lo -- Lower bound of the interval (inclusive).
            hi -- Upper bound of the interval (inclusive).
        """
        if not isinstance(lo, Integer):
            lo = Integer(lo)
        if not isinstance(hi, Integer):
            hi = Integer(hi)
        self.lo = lo
        self.hi = hi

    def __str__(self):
        """Return the string representation of an interval."""
        return "".join([(self.lo.value == Integer.NEG_INFINITY and "(") or "[",
                        str(self.lo),
                        ";",
                        str(self.hi),
                        (self.hi.value == Integer.INFINITY and ")") or "]"])

    def __add__(self, interval):
        """Add two intervals.

        Argument:
            interval -- An Interval object.

        Returns:
            A one-element tuple containing a new Interval object which
            covers the sum of any two integers from the intervals
            represented by `self' and `interval'.
        """
        assert not self.empty() and not interval.empty()
        return (Interval(self.lo + interval.lo, self.hi + interval.hi),)

    def __sub__(self, interval):
        """Subtract an interval from an interval.

        Argument:
            interval -- An Interval object.

        Returns:
            A one-element tuple containing a new Interval object which
            covers the difference of any two integers from the intervals
            represented by `self' and `interval'.
        """        
        assert not self.empty() and not interval.empty()
        return (Interval(self.lo - interval.hi, self.hi - interval.lo),)

    def __mul__(self, interval):
        """Multiply the interval by another interval.

        Argument:
            interval -- An Interval object.

        Returns:
            A one-element tuple containing a new Interval object which
            covers the product of any two integers from the intervals
            represented by `self' and `interval'.
        """
        assert not self.empty() and not interval.empty()
        lo = None
        hi = None
        # find lower and upper bounds for the product: see explanation at the
        # end of this file
        # products of intervals involving infinities as endpoints should work
        # correctly (checked by hand, 3rd May 2007)
        for c in (self.lo * interval.lo, self.lo * interval.hi,
                  self.hi * interval.lo, self.hi * interval.hi):
            if not lo or c.lt(lo):
                lo = c
            if not hi or c.gt(hi):
                hi = c
        return (Interval(lo, hi),)

    def __floordiv__(self, interval):
        """Divide the interval by another interval.

        Argument:
            interval -- An Interval object.

        Returns:
            A tuple of Interval objects which cover the result of dividing
            any integer from the interval represented by `self' by an
            integer from the interval represented by `interval' using
            integer division.  If `interval' covers only the integer
            zero, the result is a one-element tuple that consists of an
            Interval object corresponding to an empty interval.  Otherwise,
            if the interval represented by `interval' has zero as its
            endpoint, the result that will be returned will be obtained by
            ignoring the case of division by zero (the divisor interval is
            adjusted such that it does not include the integer zero).
        """
        assert not self.empty() and not interval.empty()
        # adjust the divisor interval such that 0 is not an endpoint
        int_lo = interval.lo
        int_hi = interval.hi
        if int_lo.eq(Integer(0)) and int_hi.eq(Integer(0)):
            # divisor is the constant 0: return an empty interval
            return (Interval(1, 0),)
        elif int_lo.eq(Integer(0)):
            int_lo = Integer(1)
        elif int_hi.eq(Integer(0)):
            int_hi = Integer(-1)
        # find lower and upper bounds for the quotient: see explanation at
        # the end of this file
        # quotients of intervals involving infinities as endpoints should work
        # correctly (checked by hand, 3rd May 2007)
        if int_lo.ge(Integer(0)) == int_hi.ge(Integer(0)):
            # bounds of divisor interval have the same sign
            lo = None
            hi = None
            for c in (self.lo // int_lo, self.lo // int_hi,
                      self.hi // int_lo, self.hi // int_hi):
                if not lo or c.lt(lo):
                    lo = c
                if not hi or c.gt(hi):
                    hi = c
            return (Interval(lo, hi),)
        else:
            if self.lo.ge(Integer(0)):
                return (Interval(-self.hi, self.lo // int_lo),
                        Interval(self.lo // int_hi, self.hi))
            elif self.hi.le(Integer(0)):
                return (Interval(self.hi // int_lo, -self.lo),
                        Interval(self.lo, self.hi // int_hi))
            else:
                # self.lo < 0 and self.hi > 0
                return ((-self.lo.ge(self.hi)
                         and (Interval(self.lo, -self.lo),))
                        or (Interval(-self.hi, self.hi),))

    def empty(self):
        """Test whether the interval is empty."""
        return self.lo.gt(self.hi)

    def intersects(self, interval):
        """Test whether two intervals intersect.

        Argument:
            interval -- An Interval object.

        Returns:
            True iff the intersection of the integer intervals represented
            by `self' and `interval' is not empty.
        """
        return not self.empty() and not interval.empty() \
               and self.lo.le(interval.hi) and interval.lo.le(self.hi)

# -----------------------------------------------------------------------------

class ParseError(error.Error):
    """Exception for reporting parse errors of interval specifications."""

    def __init__(self, *args, **kwargs):
        error.Error.__init__(self, *args, **kwargs)

class IntPartition:
    """A partition of the set of integers into intervals."""

    # Regular expression for lists of intervals
    _spec_format = re.compile \
                   (r"""^\*(?:,|$)"""
                    # "*"
                    r"""|^((?:-|\+)?\d+)(?:,|$)"""
                    # group 1: "<1:integer>"
                    r"""|^(?:\.\.\.?)((?:-|\+)?\d+)(?:,|$)"""
                    # group 2: "...<2:integer>"
                    r"""|^((?:-|\+)?\d+)(?:\.\.\.?)((?:-|\+)?\d+)?(?:,|$)"""
                    # groups 3, 4: "<3:integer>...(<4:integer>)"
                    r"""|^\(-inf;((?:-|\+)?\d+)\](?:,|$)"""
                    # group 5: "(-inf;<5:integer>]"
                    r"""|^\[((?:-|\+)?\d+);((?:-|\+)?\d+)\](?:,|$)"""
                    # groups 6, 7: "[<6:integer>;<7:integer>]"
                    r"""|^\[((?:-|\+)?\d+);\+?inf\)(?:,|$)"""
                    # group 8: "[<8:integer>;(+)inf)"
                    r"""|^\(-inf;\+?inf\)(?:,|$)"""
                    # "(-inf;(+)inf)"
                    )

    def __init__(self, spec):
        """Initialize a partition of integers.

        Argument:
            spec -- An interval specification string.  This string
                    should be a comma-separated list of pairwise
                    disjoint integer intervals listed in the increasing
                    order of their endpoints.  See the beginning of the
                    file for a description of the format for specifying
                    intervals.
        """
        self.contents = []
        spec = spec.strip()
        old_hi = None
        while spec:
            result = IntPartition._spec_format.search(spec)
            if not result:
                raise ParseError \
                          ("error parsing interval specification `%s'" % spec)
            if result.group(1):
                lo = long(result.group(1))
                hi = lo
            elif result.group(2) or result.group(5):
                lo = Integer.NEG_INFINITY
                hi = long(result.group(2) or result.group(5))
            elif result.group(3) or result.group(6) or result.group(8):
                lo = long(result.group(3) or result.group(6)
                          or result.group(8))
                if result.group(4) or result.group(7):
                    hi = long(result.group(4) or result.group(7))
                else:
                    hi = Integer.INFINITY
            else:
                lo = Integer.NEG_INFINITY
                hi = Integer.INFINITY
            lo = Integer(lo)
            hi = Integer(hi)
            if lo.gt(hi):
                raise ParseError \
                      ("interval specification contains an empty interval `%s'"
                       % result.group(0))
            if old_hi:
                if lo.le(old_hi):
                    raise ParseError \
                          ("interval `%s' overlaps with a previous one"
                           % result.group(0))
                if (lo - Integer(1)).neq(old_hi):
                    self.contents.append \
                        (Interval(old_hi + Integer(1), lo - Integer(1)))
            elif lo.value != Integer.NEG_INFINITY:
                self.contents.append \
                    (Interval(Integer(Integer.NEG_INFINITY), lo - Integer(1)))
            self.contents.append(Interval(lo, hi))
            old_hi = hi
            spec = spec[len(result.group(0)):]
        if old_hi.value != Integer.INFINITY:
            self.contents.append(Interval(old_hi.value + 1, Integer.INFINITY))

# -----------------------------------------------------------------------------

def generate(args, stdout=sys.stdout, stderr=sys.stderr):
    """Generate abstract types.

    Arguments:
        args   -- A sequence of command-line arguments.
        stdout -- Stream for standard output.
        stderr -- Stream for standard error.

    Writes the generated types into `stdout' (or the file specified using
    the option `--output-file' in `args').  Raises an error.OptionError if
    there is an error processing the command-line arguments `args, or an
    error.Error if there is an error while generating abstract types.
    """
    error.setErrorStream(stderr)

    try:
        (options, args) = \
            getopt.getopt \
                (args, 'c:E:I:ho:vx',
                 ["coercion=", "exclude-operators=", "include-operators=",
                  "output-file=", "help", "model-runtime-exceptions",
                  "verbose"])
    except getopt.GetoptError, e:
        raise error.OptionError(str(e))

    verbose = False
    output_file_name = "-"
    include_ops = None
    exclude_ops = None
    partitions = {}
    coercions_to_generate = {}
    model_runtime_exceptions = False
    domain_includes_invalid = False
    for (option, value) in options:
        if option in ("-c", "--coercion"):
            pos = value.find(":")
            if pos == -1:
                raise error.OptionError \
                          ("error parsing coercion specification `%s'" % value)
            (fromtype, totype) = (value[:pos], value[pos+1:])
            if not totype in coercions_to_generate:
                coercions_to_generate[totype] = [fromtype]
            else:
                coercions_to_generate.append(fromtype)
        elif option in ("-E", "--exclude-operators"):
            if include_ops or exclude_ops:
                raise error.OptionError \
                          ("only one -I or -E argument is allowed")
            exclude_ops = set(value.split(","))
        elif option in ("-I", "--include-operators"):
            if include_ops or exclude_ops:
                raise error.OptionError \
                          ("only one -I or -E argument is allowed")
            include_ops = set(value.split(","))
        elif option in ("-h", "--help"):
            stdout.write(__doc__)
            return
        elif option in ("-o", "--output-file"):
            output_file_name = value
        elif option in ("-v", "--verbose"):
            verbose = True
        elif option in ("-x", "--model-runtime-exceptions"):
            model_runtime_exceptions = True

    if len(args) == 0:
        raise error.OptionError("one or more interval specifications expected")

    # make sure that `include_ops' explicitly contains all operators for
    # which to generate operations
    if not include_ops:
        include_ops = \
            set(["1+", "2+", "1-", "2-", "*", "/", "==", "!=", "<", ">", "<=",
                 ">="])
    else:
        if "+" in include_ops:
            include_ops.add("1+")
            include_ops.add("2+")
        if "-" in include_ops:
            include_ops.add("1-")
            include_ops.add("2-")
    if exclude_ops:
        if "+" in exclude_ops:
            exclude_ops.add("1+")
            exclude_ops.add("2+")
        if "-" in exclude_ops:
            exclude_ops.add("1-")
            exclude_ops.add("2-")
        for op in exclude_ops:
            include_ops.discard(op)
    # augment the domain with a special value INVALID if the division
    # operator is to be generated and division by zero is to be modeled
    if model_runtime_exceptions and "/" in include_ops:
        domain_includes_invalid = True

    partition_counter = 0
    for spec in args:
        pos = spec.find(':')
        if pos == -1:
            typename = "anon_partition_%d" % partition_counter
            error.inform \
                ("using `%s' as type name for type with specification `%s'"
                 % (typename, spec))
            partition_counter += 1
        else:
            (typename, spec) = (spec[:pos], spec[pos+1:])
        if typename in ("int", "boolean"):
            raise error.OptionError("type name `%s' is reserved" % typename)
        if typename in partitions:
            raise error.OptionError \
                      ("duplicate name for a partition (`%s')" % typename)
        try:
            partitions[typename] = IntPartition(spec).contents
        except ParseError:
            raise error.OptionError \
                      ("error parsing partition specification `%s'" % spec)

    for (totype, fromtypes) in coercions_to_generate.iteritems():
        if not totype in partitions:
            raise error.OptionError \
                      ("unknown target type for coercion (`%s')" % totype)
        for f in fromtypes:
            if not f in partitions:
                raise error.OptionError \
                          ("unknown source type for coercion (`%s')" % f)

    try:
        if output_file_name == "-":
            output_file = stdout
        else:
            try:
                output_file = open(output_file_name, "w")
            except Exception:
                raise error.Error \
                          ("could not open `%s' for writing"
                           % output_file_name)

        first = True
        for (typename, intervals) in partitions.iteritems():
            if not first:
                output_file.write("\n")
            first = False

            if verbose:
                error.inform("processing type `%s'" % typename)

            # abstract domain
            domain = map(str, intervals)
            if domain_includes_invalid:
                domain.append("INVALID")

            # generate the abstract types
            output_file.write \
                ("type %s : %s : %s\n"
                 % (typename,
                    ", ".join(domain),
                    ", ".join(map(str, range(0, len(domain))))))

            # unary plus and minus
            if "1+" in include_ops:
                if verbose:
                    stderr.write("    generating operator `+' of arity 1\n")
                output_file.write("  operator 1 +\n")
                for val in domain:
                    output_file.write("    %s : %s\n" % (val, val))
            if "1-" in include_ops:
                if verbose:
                    stderr.write("    generating operator `-' of arity 1\n")
                output_file.write("  operator 1 -\n")
                for (i1, val1) in zip(intervals, domain):
                    result = i1 * Interval(-1, -1)
                    retvals = []
                    for (i2, val2) in zip(intervals, domain):
                        for i in result:
                            if i2.intersects(i):
                                retvals.append(val2)
                                break
                    output_file.write \
                        ("    %s : %s\n" % (val1, ", ".join(retvals)))
                if domain_includes_invalid:
                    output_file.write("    INVALID : INVALID\n")

            # binary arithmetic operators
            for operator in ("+", "-", "*", "/"):
                if not operator in include_ops \
                       and not (operator == "+" and "2+" in include_ops) \
                       and not (operator == "-" and "2-" in include_ops):
                    continue
                if verbose:
                    stderr.write \
                        ("    generating operation `%s' of arity 2\n"
                         % operator)
                output_file.write("  operator 2 %s\n" % operator)
                for (i1, val1) in zip(intervals, domain):
                    for (i2, val2) in zip(intervals, domain):
                        retvals = []
                        if operator == "+":
                            result = i1 + i2
                        elif operator == "-":
                            result = i1 - i2
                        elif operator == "*":
                            result = i1 * i2
                        elif operator == "/":
                            result = i1 // i2
                            if domain_includes_invalid \
                                   and i2.intersects(Interval(0, 0)):
                                retvals.append("INVALID")
                        for (i3, val3) in zip(intervals, domain):
                            for i in result:
                                if i3.intersects(i):
                                    retvals.append(val3)
                                    break
                        if not retvals:
                            error.warn \
                                ("no admissible return values for operation "
                                 "`%s:%s(%s, %s)', using the maximal range"
                                 % (typename, operator, val1, val2))
                            retvals = domain
                        output_file.write \
                            ("    %s, %s : %s\n"
                             % (val1, val2, ", ".join(retvals)))
                if domain_includes_invalid:
                    for val in domain:
                        if val != "INVALID":
                            output_file.write \
                                ("    %s, INVALID : INVALID\n" % val)
                            output_file.write \
                                ("    INVALID, %s : INVALID\n" % val)
                    output_file.write("    INVALID, INVALID : INVALID\n")

            # binary test operations
            for test in ("==", "!=", "<", ">", ">=", "<="):
                if not test in include_ops:
                    continue
                if verbose:
                    stderr.write \
                        ("    generating operation `%s' of arity 2\n" % test)
                output_file.write("  test 2 %s\n" % test)
                for (inum1, i1) in enumerate(intervals):
                    for (inum2, i2) in enumerate(intervals):
                        if test in ("==", "!="):
                            if inum1 != inum2:
                                retvals = \
                                    (test == "==" and ["false"]) or ["true"]
                            elif i1.lo.value == i1.hi.value:
                                retvals = \
                                    (test == "==" and ["true"]) or ["false"]
                            else:
                                retvals = ["true", "false"]
                        elif test == "<" or test == "<=":
                            if inum1 < inum2:
                                retvals = ["true"]
                            elif inum1 > inum2:
                                retvals = ["false"]
                            elif i1.lo.value == i1.hi.value:
                                retvals = \
                                    (test == "<" and ["false"]) or ["true"]
                            else:
                                retvals = ["true", "false"]
                        elif test == ">" or test == ">=":
                            if inum1 < inum2:
                                retvals = ["false"]
                            elif inum1 > inum2:
                                retvals = ["true"]
                            elif i1.lo.value == i1.hi.value:
                                retvals = \
                                    (test == ">" and ["false"]) or ["true"]
                            else:
                                retvals = ["true", "false"]
                        output_file.write \
                            ("    %s, %s : %s\n"
                             % (domain[inum1], domain[inum2],
                                ", ".join(retvals)))
                if domain_includes_invalid:
                    for val in domain:
                        if val != "INVALID":
                            output_file.write \
                                ("    %s, INVALID : true, false\n" % val)
                            output_file.write \
                                ("    INVALID, %s : true, false\n" % val)
                    output_file.write("    INVALID, INVALID : true, false\n")

            # coercions
            if verbose:
                stderr.write \
                    ("    generating coercion from type `int'\n")
            for (b, t, a) in (("Jumbala", "true", "&&"),
                              ("python", "True", "and")):
                # from integers
                output_file.write("  coercion int %s\n" % b)
                for (i, val) in zip(intervals, domain):
                    if b == "Jumbala":
                        for v in (i.lo.value, i.hi.value):
                            if not v in (Integer.NEG_INFINITY,
                                         Integer.INFINITY) \
                                    and (v < -2147483648L or v >= 2147483648L):
                                error.warn \
                                    ("the value %d cannot be represented in "
                                     "Jumbala" % v)
                    if i.lo.value == Integer.NEG_INFINITY:
                        if i.hi.value == Integer.INFINITY:
                            output_file.write("    %s : %s\n" % (t, val))
                        else:
                            output_file.write \
                                ("    <> <= %s : %s\n" % (str(i.hi), val))
                    elif i.hi.value == Integer.INFINITY:
                        output_file.write \
                            ("    <> >= %s : %s\n" % (str(i.lo), val))
                    elif i.lo.value == i.hi.value:
                        output_file.write \
                            ("    <> == %s : %s\n" % (str(i.lo), val))
                    else:
                        output_file.write \
                            ("    (<> >= %s) %s (<> <= %s) : %s\n"
                             % (str(i.lo), a, str(i.hi), val))
                output_file.write("    end\n")

            # from user-specified partition abstractions
            if typename in coercions_to_generate:
                for fromtype in coercions_to_generate[typename]:
                    if verbose:
                        stderr.write \
                            ("    generating coercion from type `%s'\n"
                             % fromtype)
                    fromtype_intervals = partitions[fromtype]
                    for b in ("Jumbala", "python"):
                        output_file.write("  coercion %s %s\n" % (fromtype, b))
                        for i1 in fromtype_intervals:
                            retvals = \
                                [str(i) for i in intervals if i1.intersects(i)]
                            output_file.write \
                                ("    <> == <%s> : %s\n"
                                 % (str(i1), ", ".join(retvals)))
                        if domain_includes_invalid:
                            output_file.write \
                                ("    <> == <INVALID> : INVALID\n")
                        output_file.write("    end\n")

        if output_file != stdout:
            output_file.close()
    except IOException, e:
        raise error.Error("error while writing to output file [%s]" % str(e))

if __name__ == "__main__":
    try:
        generate(sys.argv[1:])
    except error.OptionError, e:
        error.abort(str(e), 2)
    except error.Error, e:
        error.abort(str(e), 1)

# -----------------------------------------------------------------------------
# Product of intervals
#
# Let x and y be integers such that  x_min <= x <= x_max  and
# y_min <= y <= y_max  hold.
#
# Assume that x_min <= x_max and y_min <= y_max hold.                     (1)
#
# Problem: Find integers  z_min  and  z_max  such that
#          z_min <= x * y <= z_max.
#
# There are the following cases:
#
#     1. x_min >= 0, y_min >= 0:
#        By assumption (1) and the properties of integer multiplication,
#        it follows that
#            x_min * y_min <= x * y <= x_max * y_max
#        holds in this case.
#
#     2. x_min <= 0, x_max <= 0, y_min >= 0:
#        In this case the product of  x  and  y  cannot be positive.
#        Because both x_min and x_max are not positive, |x_min| >= |x_max|
#        holds by assumption (1).  Thus the product of  x  and  y  will
#        fall in the range
#            [x_min * y_max, x_max * y_min]
#        in this case.
#
#     3. x_min >= 0, x_max >= 0, y_min <= 0, y_max <= 0:
#        Because multiplication is a commutative operation, this case can
#        be reduced to the previous one by switching the roles of  x  and
#        y.  Thus the product of  x  and  y  is in the range
#            [x_max * y_min, x_min * y_max]
#        in this case.
#
#     4. x_min <= 0, x_max <= 0, y_min <= 0, y_max <= 0:
#        Because no interval endpoint is positive, it follows that
#            x_max * y_max <= x * y <= x_min * y_min.
#
#     5. x_min <= 0, x_max >= 0, y_min >= 0, y_max >= 0:
#        This case can be reduced to a combination of previous cases by
#        splitting the interval [x_min, x_max] at 0.
#            5.1 x_min <= x <= 0  &&  y_min >= 0, y_max >= 0
#                => x_min * y_max <= x * y <= 0                      (case 2)
#            5.2 0 <= x <= x_max  &&  y_min >= 0, y_max >= 0
#                => 0 <= x * y <= x_max * y_max                      (case 1)
#        Combining the two subcases shows that the product of  x  and  y
#        will always be in the range
#            [x_min * y_max, x_max * y_max].
#
#     6. x_min <= 0, x_max >= 0, y_min <= 0, y_max <= 0:
#        Reducing this case to previous ones gives the following subcases:
#            6.1 x_min <= x <= 0  &&  y_min <= 0, y_max <= 0
#                => 0 <= x * y <= x_min * y_min                      (case 4)
#            6.2 0 <= x <= x_max  &&  y_min <= 0, y_max <= 0
#                => x_max * y_min <= x * y <= 0                      (case 3)
#        The combination of the subcases yields the range
#            [x_max * y_min, x_min * y_min].
#
#     7. x_min <= 0, x_max >= 0, y_min <= 0, y_max >= 0:
#        This case can be represented as the combination of the following
#        four subcases:
#            7.1 x_min <= x <= 0  &&  y_min <= y <= 0
#                => 0 <= x * y <= x_min * y_min                      (case 4)
#            7.2 x_min <= x <= 0  &&  0 <= y <= y_max
#                => x_min * y_max <= x * y <= 0                      (case 2)
#            7.3 0 <= x <= x_max  &&  y_min <= y <= 0
#                => x_max * y_min <= x * y <= 0                      (case 3)
#            7.4 0 <= x <= x_max  &&  0 <= y <= y_max
#                => 0 <= x * y <= x_max * y_max                      (case 1)
#        Combining the subcases gives the interval
#            [min{x_min*y_max, x_max*y_min}, max{x_min*y_min, x_max*y_max}].
#
#     8. x_min <= 0, x_max <= 0, y_min <= 0, y_max >= 0:
#        This is the same as case (6) with  x  and  y  switched: the result
#        is
#            [x_min * y_max, x_min * y_min].
#
#     9. x_min >= 0, x_max >= 0, y_min <= 0, y_max >= 0:
#        This is the same as case (5) with  x  and  y  switched: the result
#        is
#            [x_max * y_min, x_max * y_max].
#
# Combining all of these cases shows that the product of  x  and  y  will
# always be within the range
#     [min{x_min*y_min, x_min*y_max, x_max*y_min, x_max*y_max},
#      max{x_min*y_min, x_min*y_max, x_max*y_min, x_max*y_max}].
# The implementation of Interval.__mul__ chooses the endpoints of the range
# from this set of values.
#
# -----------------------------------------------------------------------------
# (Integer) division of intervals
#
# Let x and y be integers such that  x_min <= x <= x_max  and
# y_min <= y <= y_max  hold.
#
# Assume that x_min <= x_max and y_min <= y_max hold.                     (1)
#
# Problem: Find integers  z_min  and  z_max  such that
#          z_min <= x / y <= z_max,
#          assuming the divisor to be always nonzero.
#
# It is easy to see that, assuming that the divisor is a nonzero integer, the
# absolute value |x/y| satisfies the following inequality:
#       min{|z| | x_min<=z<=x_max} / max{|z| | y_min<=z<=y_max, z!=0}
#    <= |x/y|                                                             (2)
#    <= max{|z| | x_min<=z<=x_max} / min{|z| | y_min<=z<=y_max, z!=0}
#
# There are the following cases (the assumption (1) is assumed to hold):
#
#     1. x_min >= 0, x_max >= 0, y_min > 0, y_max > 0:
#        In this case
#            min{|z| | x_min<=z<=x_max} / max{|z| | y_min<=z<=y_max, z!= 0}
#              == |x_min| / |y_max| == x_min / y_max,  and
#            max{|z| | x_min<=z<=x_max} / min{|z| | y_min<=z<=y_max, z!= 0}
#              == |x_max| / |y_min| == x_max / y_min.
#        From (2) it follows that
#            x_min / y_max <= |x/y| == x/y <= x_max / y_min.
#
#     2. x_min >= 0, x_max >= 0, y_min < 0, y_max < 0:
#        In this case
#            min{|z| | x_min<=z<=x_max} / max{|z| | y_min<=z<=y_max, z!= 0}
#              == |x_min| / |y_min| == x_min / -y_min,  and
#            max{|z| | x_min<=z<=x_max} / min{|z| | y_min<=z<=y_max, z!= 0}
#              == |x_max| / |y_max| == x_max / -y_max.
#        From (2),
#            x_min / -y_min <= |x/y| == x/-y <= x_max / -y_max,  and thus
#            x_max / y_max <= x/y <= x_min / y_min.
#
#     3. x_min >= 0, x_max >= 0, y_min < 0, y_max > 0, y != 0:
#        This case can be split into two subcases, which reduce into one of
#        the above cases as follows:
#            3.1 x_min >= 0, x_max >= 0, y_min < 0, y_max == -1:
#                By case 2, in this case
#                    x_max / -1 == -x_max <= x/y <= x_min / y_min.
#            3.2 x_min >= 0, x_max >= 0, y_min == 1, y_max > 0:
#                In this case, it follows (by case 1) that
#                    x_min / y_max <= x/y <= x_max / 1 == x_max.
#        Combining the two subcases shows that  x/y  will be fall in either of
#        the ranges
#            [-x_max, x_min/y_min]  or  [x_min/y_max, x_max].
#
#     4. x_min <= 0, x_max <= 0, y_min > 0, y_max > 0:
#        In this case
#            min{|z| | x_min<=z<=x_max} / max{|z| | y_min<=z<=y_max, z!= 0}
#              == |x_max| / |y_max| == -x_max / y_max,  and
#            max{|z| | x_min<=z<=x_max} / min{|z| | y_min<=z<=y_max, z!= 0}
#              == |x_min| / |y_min| == -x_min / y_min.
#        From (2),
#            -x_max / y_max <= |x/y| == -x/y <= -x_min / y_min,  and thus
#            x_min / y_min <= x/y <= x_max / y_max.
#
#     5. x_min <= 0, x_max <= 0, y_min < 0, y_max < 0:
#        In this case
#            min{|z| | x_min<=z<=x_max} / max{|z| | y_min<=z<=y_max, z!= 0}
#              == |x_max| / |y_min| == x_max / y_min,  and
#            max{|z| | x_min<=z<=x_max} / min{|z| | y_min<=z<=y_max, z!= 0}
#              == |x_min| / |y_max| == x_min / y_max.
#        From (2) it follows that
#            x_max / y_min <= x/y <= x_min / y_max
#        holds in this case.
#
#     6. x_min <= 0, x_max <= 0, y_min < 0, y_max > 0, y != 0:
#        Similarly to case 3, this case can be split into two subcases,
#        which can then be solved using cases 4 and 5:
#            6.1 x_min <= 0, x_max <= 0, y_min < 0, y_max == -1:
#                By case 5,
#                    x_max / y_min <= x/y <= x_min / -1 == -x_min
#                holds in this case.
#            6.2 x_min <= 0, x_max <= 0, y_min == 1, y_max > 0:
#                By case 4,
#                    x_min / 1 == x_min <= x/y <= x_max / y_max.
#        It follows that  x/y  will fall in either of the ranges
#            [x_min, x_max/y_max]  or  [x_max/y_min, -xmin].
#
#     7. x_min <= 0, x_max >= 0, y_min > 0, y_max > 0:
#        Also this case can be split into two subcases:
#            7.1 x_min <= 0, x_max == 0, y_min > 0, y_max > 0:
#                By case 4,
#                    x_min / y_min <= x/y <= 0.
#            7.2 x_min == 0, x_max >= 0, y_min > 0, y_max > 0:
#                By case 1,
#                    0 <= x/y <= x_max / y_min.
#        From the subcases it follows that
#            x_min / y_min <= x <= x_max / y_min
#        holds in this case.
#
#     8. x_min <= 0, x_max >= 0, y_min < 0, y_max < 0:
#        There are the following subcases:
#            7.1 x_min <= 0, x_max == 0, y_min < 0, y_max < 0:
#                By case 5,
#                    0 <= x/y <= x_min / y_max.
#            7.2 x_min == 0, x_max >= 0, y_min < 0, y_max < 0:
#                By case 2,
#                    x_max / y_max <= x/y <= 0.
#        Consequently,
#            x_max / y_max <= x/y <= x_min / y_max.
#
#     9. x_min <= 0, x_max >= 0, y_min < 0, y_max > 0, y != 0:
#        In this case there are the following two subcases:
#            9.1 x_min <= 0, x_max == 0, y_min < 0, y_max > 0, y != 0:
#                Case 6 yields the ranges
#                    [x_min, 0]  and  [0, -x_min],
#                which can be combined into the range
#                    [x_min, -x_min].
#            9.2 x_min == 0, x_max >= 0, y_min < 0, y_max > 0, y != 0:
#                By case 3, x/y will fall in either of the ranges
#                    [-x_max, 0]  or  [0, x_max],
#                i.e., the combined range
#                    [-x_max, x_max].
#        It follows that  x/y  will be in the range
#            [min{x_min, -x_max}, max{-x_min, x_max}].
#
# From cases 1, 2, 4, 5, 7 and 8, it follows that if  y_min  and  y_max
# have the same sign, then both the lower bound and the upper bound for
# x/y  are included among the values
#    x_min/y_min,  x_min/y_max,  x_max/y_min  and  x_max/y_max.
