# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Common dialogs used in the model abstractor GUI.

This module defines some common dialogs used in the model abstractor GUI.

Authors: Juhani Peltonen
         Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

from os.path import exists, isfile
import gtk
import pygtk

class ErrorDialog:
    def __init__(self, root_window, primary_msg, secondary_msg):
        dialog = gtk.MessageDialog \
            (root_window,
             gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
             gtk.MESSAGE_ERROR, gtk.BUTTONS_NONE,
             primary_msg)
        dialog.set_title('Error')
        if secondary_msg:
            dialog.format_secondary_text(secondary_msg)
        dialog.set_default_response(gtk.RESPONSE_OK)
        dialog.add_buttons(gtk.STOCK_OK, gtk.RESPONSE_OK)
        dialog.show_all()
        dialog.run()
        dialog.destroy()

class SaveFileDialog:
    def __init__(self, root_window, title, filters=None):
        self.dialog = gtk.FileChooserDialog(title,
                                            root_window,
                                            gtk.FILE_CHOOSER_ACTION_SAVE,
                                            (gtk.STOCK_OK,
                                             gtk.RESPONSE_OK,
                                             gtk.STOCK_CANCEL,
                                             gtk.RESPONSE_CANCEL))
        self.dialog.set_default_response(gtk.RESPONSE_OK)
        if filters == None:
            return
        for (filter_name, patterns) in filters:
            filt = gtk.FileFilter()
            filt.set_name(filter_name)
            for p in patterns:
                filt.add_pattern(p)
            self.dialog.add_filter(filt)

    def run(self, directory):
        try:
            self.dialog.set_current_folder(directory)
            while True:
                resp = self.dialog.run()
                if resp != gtk.RESPONSE_OK:
                    return (None, None, None)
                filename = self.dialog.get_filename()

                # check whether the file exists
                if exists(filename):
                    if not isfile(filename):
                        ErrorDialog(self.dialog,
                                    'Invalid file name!'
                                    "`%s' exists but is not a file."
                                    % filename)
                        continue

                    dialog = gtk.MessageDialog \
                                 (self.dialog,
                                  gtk.DIALOG_MODAL
                                  | gtk.DIALOG_DESTROY_WITH_PARENT,
                                  gtk.MESSAGE_QUESTION,
                                  gtk.BUTTONS_NONE,
                                  'Overwrite existing file?')
                    dialog.set_default_response(gtk.RESPONSE_YES)
                    dialog.set_title('Confirmation')
                    dialog.format_secondary_text \
                        ("The old contents of file `%s' will be lost if "
                         'you continue with the operation.' % filename)
                    dialog.add_buttons \
                        ('Yes, overwrite', gtk.RESPONSE_YES,
                     'No, choose a different file name', gtk.RESPONSE_NO,
                     gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
                    resp = dialog.run()
                    dialog.destroy()
                    if resp == gtk.RESPONSE_YES:
                        return (filename, self.dialog.get_current_folder(),
                                self.dialog.get_filter())
                    elif resp == gtk.RESPONSE_CANCEL:
                        return (None, None, None)
                else:
                    return (filename, self.dialog.get_current_folder(),
                            self.dialog.get_filter())
        finally:
            self.dialog.destroy()
