#! /usr/bin/python2.4

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""GUI for the model abstractor.

This module implements the main GUI for the model abstractor.

Authors: Juhani Peltonen
         Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

import pygtk
import gtk, gobject
import copy
import sys
import os
from dialogs import ErrorDialog, SaveFileDialog
from xtypes import Model
from xtypes import Class
from xtypes import Signal
from xtypes import Attribute
from tlib import Tlib
from type_editor import Type_editor
from os import popen, remove, write, close
from popen2 import Popen3
from tempfile import mkstemp
from abstraction.error import Error
from abstraction.smtlib_typegen import PySmt

class FloatingWindow:
	def __init__(self, title, default_size=None):
		self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
		self.window.set_title(title)
		self.window.set_modal(False)
		self.window.set_keep_above(True)
		if hasattr(self.window, 'set_deletable'):
			self.window.set_deletable(False)
		self.window.connect('delete_event', self.handle_delete_event)
		if default_size:
			self.window.set_default_size(*default_size)
		self.position = None

	def handle_delete_event(self, window, event):
		self.hide()
		return False

	def is_visible(self):
		return self.window.get_property('visible')

	def get_window(self):
		return self.window

	def hide(self):
		if self.window.get_property('visible'):
			self.position = self.window.get_position()
			self.window.hide()

	def show(self):
		if not self.window.get_property('visible'):
			if self.position:
				self.window.move(*self.position)
			self.window.show_all()

class DotImage:
	def __init__(self):
		self.image = gtk.Image()
		self.image_loaded = False

	def get(self):
		return self.image

	def set(self, dot_command, prepare_input):
		if self.image_loaded:
			return
		(fd, path) = mkstemp()
		close(fd)
		try:
			child = Popen3 \
			            ("%s -Tpng -o %s" % (dot_command, path),
				     True)
			prepare_input(child.tochild)
			child.tochild.close()
			status = child.wait()
		except IOError:
			status = 1
		if status:
			msg = ["failed to execute dot"]
			msg.extend([line for line in child.childerr])
			if len(msg) > 1:
				msg[0] += ":"
			raise IOError("\n    ".join(msg))
		if hasattr(self.image, 'clear'):
			self.image.clear()
		self.image.set_from_file(path)
		remove(path)
		image_loaded = True

	def invalidate(self):
		if self.image_loaded:
			if hasattr(self.image, 'clear'):
				self.image.clear()
			self.image_loaded = False

class SharedState:
	# Available SMT solvers
	SMT_SOLVERS = PySmt.SMT_SOLVERS

	# Data shared between the main GUI and the type editor
	def __init__(self):
		# The model (None if not loaded)
		self.model = None
		# The current type library
		self.tlib = Tlib()
		# Default type for type inference
		self.default_type = None
		# Whether to run type inference automatically
		self.automatic_type_inference = False
		# Whether to model runtime exceptions in type generation
		self.model_rte = False
		# (Default) SMT solver to use
		self.solver = 'Yices'
		# (Default) path to the SMT solver to use
		self.solver_path = 'yices'
		# (Default) path to the Dot tool
		self.dot_path = 'dot'
		# Type hierarchy window
		self.type_hierarchy_window = FloatingWindow('Type hierarchy')
		self.type_hierarchy_image = DotImage()
		self.type_hierarchy_window.get_window().add \
		    (self.type_hierarchy_image.get())
		# Attribute dependency window
		self.attribute_dependency_window = \
		    FloatingWindow('Attribute dependencies', (200, 200))
		self.attribute_dependency_image = DotImage()
		self.attribute_dependency_window.get_window().add \
		    (self.attribute_dependency_image.get())
		# Default padding between widgets
		self.DEFAULT_PADDING = 5

class Gui:
	def __init__(self):
		# the columns in the attribute tree view (do not
		# change)
		self.COLUMN_ATTRIBUTE = 0
		self.COLUMN_ORIGINAL_TYPE = 1
		self.COLUMN_LAST_INFERRED_TYPE = 2
		self.COLUMN_CHANGED = 3

		# the pages in the constraint notebook (changing these
		# changes the order of the pages in the notebook)
		self.NB_PAGE_ATTRIBUTE = 0
		self.NB_PAGE_ALL = 1
		self.NB_PAGE_CONTRADICTIONS = 2

		self.STR_NO_DEFAULT_TYPE = '<None>'

		self.PREFERENCES_PATH = os.environ['HOME'] \
					+ '/.smuml_abst_gui_prefs'

		self.WIN_DEFAULT_WIDTH = 600
		self.WIN_DEFAULT_HEIGHT = 600
		self.WIN_DEFAULT_TITLE = \
		    'SMUML Model Abstractor and Type Editor'

		# positions of vertical and horizontal paned dividers
		# (these affect sizes of the frames in the window)
		self.V_PANED_DIV_POS = 250
		self.H_PANED_DIV_POS = 350

		# shared data between main gui and type editor
		self.shared_state = SharedState()

		# current directory for project files
		self.current_project_directory = os.getcwd()

		self.selected_attribute = None
		self.current_constraints_notebook_page = None

		# need to store reference to the acceleration group in
		# order to remove it
		self.accel_group = None

		# the picture that is shown in the changed column
		self.xpm_changed = [
			"16 16 2 1",
			"       c None",
			".      c #FF0000",
			"                ",
			"     ......     ",
			"   ..........   ",
			"  ............  ",
			"  ............  ",
			" .............. ",
			" .............. ",
			" .............. ",
			" .............. ",
			" .............. ",
			" .............. ",
			"  ............  ",
			"  ............  ",
			"   ..........   ",
			"     ......     ",
			"                "]

		self.pb_changed = gtk.gdk.pixbuf_new_from_xpm_data \
		    (self.xpm_changed)

		self.win = gtk.Window(gtk.WINDOW_TOPLEVEL)
		self.win.connect('delete_event', self.handle_delete_event)

		self.type_editor = Type_editor(self.shared_state)

		self.table = None

		self.load_preferences()
		self.init_main_window_widgets()
		self.update_attribute_tree_view()
		self.shared_state.type_hierarchy_window.get_window().connect \
		    ('delete_event',
		     self.handle_type_hierarchy_window_delete_event)
		self.shared_state.attribute_dependency_window.get_window() \
		    .connect \
		    ('delete_event',
		     self.handle_attribute_dependency_window_delete_event)
		self.update_widget_sensitivities()
		self.set_default_type(None)
		self.win.set_title(self.WIN_DEFAULT_TITLE)
		self.win.show_all()

		self.type_editor.root_dialog.set_transient_for(self.win)
		self.type_editor.root_dialog.set_destroy_with_parent(True)

	# add a constraint to the given attribute
	def add_constraint(self, attribute, constraint, constraining_type):
		attribute.add_constraint(constraint, constraining_type)
		self.update_attribute_constraints()
		self.update_all_constraints()
		if self.shared_state.automatic_type_inference:
			self.infer_types()
		self.update_buttons()

	# update the attribute constraints model in the constraints
	# notebook
	def update_attribute_constraints(self):
		model = self.attribute_constraints_tv.get_model()
		model.clear()
		attribute = self.get_selected_attribute()
		if attribute != None:
			full_name = attribute.get_full_name()
			for (c, t) in attribute.constraints:
				model.append((full_name, c, t))
	
	# update the all constraints model in the constraints notebook
	def update_all_constraints(self):
		model = self.all_constraints_tv.get_model()
		model.clear()
		if self.shared_state.model == None:
			return
		for attr in self.shared_state.model.get_attributes():
			for (constraint, type_name) in attr.constraints:
				model.append((attr.get_full_name(),
					      constraint, type_name))

	# update the contradictory constraints model in the
	# constraints notebook
	def update_contradictory_constraints(self):
		model = self.contradictory_constraints_tv.get_model()
		model.clear()
		if self.shared_state.model == None:
			return
		for attr in self.shared_state.model.get_contradicting_attributes():
			for (constraint, type_name) in attr.constraints:
				model.append((attr.get_full_name(),
					      constraint, type_name))

	# update the all the models in the constraint notebook
	def update_constraints(self):
		self.update_attribute_constraints()
		self.update_all_constraints()
		self.update_contradictory_constraints()

	def set_selected_attribute(self, attribute):
		self.selected_attribute = attribute

	def get_selected_attribute(self):
		return self.selected_attribute

	def handle_add_edit_constraint_click(self, button, tree_view):
		constraint_operators = self.shared_state.model.get_constraint_operators()
		num_operators = len(constraint_operators)

		dialog = gtk.Dialog('%s constraint'
				    % (tree_view and 'Edit' or 'Add'),
				    self.win,
				    gtk.DIALOG_DESTROY_WITH_PARENT \
				    | gtk.DIALOG_MODAL,
				    (gtk.STOCK_OK, gtk.RESPONSE_OK,
				     gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))

		table = gtk.Table(2 + num_operators, 2)
		table.set_border_width(self.shared_state.DEFAULT_PADDING)
		table.set_row_spacings(self.shared_state.DEFAULT_PADDING)
		table.set_col_spacings(self.shared_state.DEFAULT_PADDING)

		if tree_view: # "edit" operation
			# get selected constraint
			(model, it) = tree_view.get_selection().get_selected()
			assert(it != None)
			selected_attr = model.get_value(it, 0)
			old_constraint = model.get_value(it, 1)
			old_constraining_type = model.get_value(it, 2)
			old_attr = self.shared_state.model.get_attribute_with_full_name(selected_attr)
		else: # "add" operation
			selected_attr = self.get_selected_attribute()
			if selected_attr != None:
				selected_attr = selected_attr.get_full_name()
			old_constraint = None
			old_constraining_type = None
			old_attr = None

		l = gtk.Label('Attribute:')
		table.attach(l, 0, 1, 0, 1, gtk.FILL, gtk.FILL)

		cbox_attribute = gtk.combo_box_new_text()
		for (i, attr) in enumerate(self.shared_state.model.get_attributes()):
			full_name = attr.get_full_name()
			cbox_attribute.append_text(full_name)
			if selected_attr == full_name:
				cbox_attribute.set_active(i)
		table.attach(cbox_attribute, 1, 2, 0, 1, gtk.FILL, gtk.FILL)

		l = gtk.Label('Constraint:')
		table.attach(l, 0, 1, 1, 2, gtk.FILL, gtk.FILL)
		radio_buttons = []
		for (i, c) in enumerate(constraint_operators):
			radio_button = gtk.RadioButton(radio_buttons and radio_buttons[0] or None, c)
			if old_constraint == c:
				radio_button.set_active(True)
			radio_buttons.append(radio_button)
			table.attach(radio_button, 1, 2, 1 + i, 2 + i,
				     gtk.FILL, gtk.FILL)

		l = gtk.Label('Type:')
		table.attach(l, 0, 1, 2 + num_operators, 3 + num_operators,
			     gtk.FILL, gtk.FILL)
		cbox_type = gtk.combo_box_new_text()
		for (i, t) in enumerate(self.shared_state.tlib.get_int_types()):
			cbox_type.append_text(t)
			if t == old_constraining_type:
				cbox_type.set_active(i)
		table.attach(cbox_type, 1, 2, 2 + num_operators,
			     3 + num_operators, gtk.FILL, gtk.FILL)
		dialog.vbox.set_spacing(self.shared_state.DEFAULT_PADDING)
		dialog.vbox.pack_start(table)

		dialog.show_all()
		resp = dialog.run()
		dialog.destroy()

		if resp != gtk.RESPONSE_OK:
			return

		full_name = self.cbox_get_active(cbox_attribute)
		constraint = None
		for (b, o) in zip(radio_buttons, constraint_operators):
			if b.get_active():
				constraint = o
				break
		constraining_type = self.cbox_get_active(cbox_type)

		if full_name == None \
			    or constraint == None \
			    or constraining_type == None:
			return

		attr = self.shared_state.model.get_attribute_with_full_name(full_name)
		if attr.has_constraint(constraint, constraining_type):
			ErrorDialog(self.win, 'Constraint already defined!')
			return
		if old_attr:
			old_attr.remove_constraint(old_constraint,
						   old_constraining_type)
		attr.add_constraint(constraint, constraining_type)
		if self.shared_state.automatic_type_inference:
			self.infer_types()
		self.update_constraints()
		self.update_buttons()

	def update_attribute_tree_view(self):
		tm = gtk.TreeStore(gobject.TYPE_STRING,
				   gobject.TYPE_STRING,
				   gobject.TYPE_STRING,
				   gtk.gdk.Pixbuf)
		self.construct_tree_model(tm)
		self.attribute_tv.set_model(tm)
		self.attribute_tv.expand_all()

	# initialize main GUI
	def init_main_window_widgets(self):
		if self.table != None:
			self.win.remove(self.table)

		self.table = gtk.Table(3, 1, False)
		self.win.add(self.table)

		###########################################################
		# menu
		ui = '''
		<ui>
			<menubar name="Menubar">
				<menu action = "File">
					<menuitem action = "Open project"/>
					<separator/>
					<menuitem action = "Export abstract model"/>
					<separator/>
					<menuitem action = "Close"/>
					<menuitem action = "Quit"/>
				</menu>
				<menu action = "Edit">
					<menuitem action = "Types"/>
					<menuitem action = "Preferences"/>
				</menu>
				<menu action = "Type inference">
					<menuitem action = "Run"/>
					<menuitem action = "Change default type"/>
		                        <separator/>
                                        <menuitem action = "Run automatically"/>
				</menu>
				<menu action = "View">
					<menuitem action = "Type hierarchy"/>
					<menuitem action = "Attribute dependencies"/>
				</menu>
			</menubar>
			<toolbar name = "Toolbar">
				<toolitem action = "t_Open"/>
				<toolitem action = "t_Export"/>
				<separator/>
				<toolitem action = "t_Edit types"/>
				<separator/>
				<toolitem action = "t_Infer types"/>
			</toolbar>
		</ui>'''

		ui_manager = gtk.UIManager()
		accel_group = ui_manager.get_accel_group()
		self.win.add_accel_group(accel_group)
		self.accel_group = accel_group
		self.action_group = gtk.ActionGroup('menu')
		self.action_group.add_actions([
			('File', None, '_File'),
			('Open project', None, '_Open project...',
			 '<Ctrl>O', None, self.handle_menubar_selection),
			('Export abstract model', None, '_Export abstract model...',
			 None, None, self.handle_menubar_selection),
			('Close', None, '_Close',
			 '<Ctrl>W', None, self.handle_menubar_selection),
			('Quit', None, '_Quit',
			 '<Ctrl>Q', None, self.handle_menubar_selection),
			('Edit', None, '_Edit'),
			('Types', None, '_Types...',
			 None, None, self.handle_menubar_selection),
			('Preferences', None, '_Preferences...',
			 None, None, self.handle_menubar_selection),
			('Type inference', None, '_Type inference'),
			('Run', None, '_Run',
			 None, None, self.handle_menubar_selection),
			('Change default type', None, '_Change default type...',
			 None, None, self.handle_menubar_selection),
			('View', None, '_View'),
			('t_Open', gtk.STOCK_OPEN, '_Open',
			 None, None, self.handle_menubar_selection),
			('t_Export', gtk.STOCK_SAVE_AS, '_Export',
			 None, None, self.handle_menubar_selection),
			('t_Edit types', gtk.STOCK_EDIT, '_Edit types',
			 None, None, self.handle_menubar_selection),
			('t_Infer types', gtk.STOCK_EXECUTE, '_Infer types',
			 None, None, self.handle_menubar_selection)])

		self.action_group.add_toggle_actions([
			('Run automatically', None, 'Run _automatically',
			 None, None, self.update_widget_sensitivities),
			('Type hierarchy', None, '_Type hierarchy',
			 None, None, self.handle_menubar_selection),
			('Attribute dependencies', None, '_Attribute dependencies',
			 None, None, self.handle_menubar_selection)])

		self.menubar_funcs \
			= {'Open project': self.open_project,
			   'Export abstract model': self.export_model,
			   'Close': self.close_model,
			   'Quit': self.quit,
			   'Types': self.edit_types,
			   'Preferences': self.edit_preferences,
			   'Run': self.infer_types,
			   'Change default type': self.change_default_type,
			   'Type hierarchy': self.update_type_hierarchy_window,
			   'Attribute dependencies': self.update_attribute_dependency_window,
			   't_Open': self.open_project,
			   't_Export': self.export_model,
			   't_Edit types': self.edit_types,
			   't_Infer types': self.infer_types}

		infer_types_action = self.action_group.get_action('t_Infer types')
		infer_types_action.set_sensitive(True)

	        ui_manager.insert_action_group(self.action_group, 0)
	        ui_manager.add_ui_from_string(ui)
	        self.menubar = ui_manager.get_widget('/Menubar')
        	self.toolbar = ui_manager.get_widget('/Toolbar')

		###########################################################
		# attribute tree view
		self.attribute_tv = gtk.TreeView()

		cr = gtk.CellRendererText()
		col = gtk.TreeViewColumn('Attribute', cr)
		col.add_attribute(cr, 'text', self.COLUMN_ATTRIBUTE)
		col.set_resizable(True)
		self.attribute_tv.insert_column(col, self.COLUMN_ATTRIBUTE)

		cr = gtk.CellRendererText()
		col = gtk.TreeViewColumn('Original type', cr)
		col.add_attribute(cr, 'text', self.COLUMN_ORIGINAL_TYPE)
		col.set_resizable(True)
		self.attribute_tv.insert_column(col, self.COLUMN_ORIGINAL_TYPE)

		cr = gtk.CellRendererText()
		col = gtk.TreeViewColumn('Inferred type', cr)
		col.add_attribute(cr, 'text', self.COLUMN_LAST_INFERRED_TYPE)
		col.set_cell_data_func(cr, self.cdf_last_inferred_type)
		col.set_resizable(True)
		self.attribute_tv.insert_column(col,
						self.COLUMN_LAST_INFERRED_TYPE)

		cr = gtk.CellRendererPixbuf()
		cr.set_property('xalign', 0)
		col = gtk.TreeViewColumn('Changed', cr)
		col.add_attribute(cr, 'pixbuf', self.COLUMN_CHANGED)
		col.set_cell_data_func(cr, self.cdf_changed)
		col.set_resizable(True)
		self.attribute_tv.insert_column(col, self.COLUMN_CHANGED)

		sw = gtk.ScrolledWindow()
		sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
		sw.add_with_viewport(self.attribute_tv)
		self.attribute_tv.connect('button-press-event',
					  self.handle_tree_view_events)
		self.attribute_tv.connect('key-press-event',
					  self.handle_tree_view_events)
		selection = self.attribute_tv.get_selection()
		selection.connect('changed',
				  self.handle_selection_on_attribute_tv)
		self.attribute_sw = sw

		###########################################################
		# types
		frame = gtk.Frame('Types')
		sw = gtk.ScrolledWindow()
		sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
		self.default_type_cbox = gtk.combo_box_new_text()
		self.default_type_cbox_changed_handler = \
		    self.default_type_cbox.connect \
		        ('changed',
			 self.handle_default_type_selection)
		self.default_type_cbox.append_text(self.STR_NO_DEFAULT_TYPE)
		table = gtk.Table(1, 2)
		hbox = gtk.HBox()
		l = gtk.Label('Default type: ')
		hbox.pack_start(l, False)
		hbox.pack_start(self.default_type_cbox)
		table.attach(hbox, 0, 1, 0, 1,
			     gtk.FILL, gtk.FILL,
			     0, 5)
		table.attach(sw, 0, 1, 1, 2,
			     gtk.FILL | gtk.EXPAND,
			     gtk.FILL | gtk.EXPAND,
			     0, 5)
		frame.add(table)
		cr = gtk.CellRendererText()
		col = gtk.TreeViewColumn('', cr, text = 0)
		model = gtk.ListStore(gobject.TYPE_STRING)
		self.types_tv = gtk.TreeView(model)
		self.types_tv.append_column(col)
		self.types_tv.set_headers_visible(False)
		sw.add(self.types_tv)
		self.types_tv.connect('button-press-event',
				      self.handle_tree_view_events)
		self.types_tv.connect('key-press-event',
				      self.handle_tree_view_events)
		selection = self.types_tv.get_selection()
		selection.connect('changed',
				  self.handle_selection_on_types_tv)
		self.types_frame = frame
		for t in self.shared_state.tlib.get_int_types():
			self.default_type_cbox.append_text(t)
			model.append((t,))
		self.default_type_cbox.set_active(0)

		###########################################################
		# constraints
		self.constraint_notebook = gtk.Notebook()

		# attribute constraints page
		b_add = gtk.Button(None, gtk.STOCK_ADD)
		b_edit = gtk.Button(None, gtk.STOCK_EDIT)
		b_remove = gtk.Button(None, gtk.STOCK_REMOVE)

		b_add.set_sensitive(False)
		b_edit.set_sensitive(False)
		b_remove.set_sensitive(False)

		table = gtk.Table(2, 3)
		table.attach(b_add, 0, 1, 0, 1,
			     gtk.FILL, gtk.FILL, 10, 10)
		table.attach(b_edit, 1, 2, 0, 1,
			     gtk.FILL, gtk.FILL, 10, 10)
		table.attach(b_remove, 2, 3, 0, 1,
			     gtk.FILL, gtk.FILL, 10, 10)

		ls = gtk.ListStore(gobject.TYPE_STRING,
				   gobject.TYPE_STRING,
				   gobject.TYPE_STRING)
		tv = gtk.TreeView(ls)
		tv.connect('button-press-event',
			   self.handle_tree_view_events)
		tv.connect('key-press-event',
			   self.handle_tree_view_events)
		selection = tv.get_selection()
		selection.connect('changed',
				  self.handle_selection_on_attribute_constraints_tv)
		cr = gtk.CellRendererText()

		col = gtk.TreeViewColumn('Attribute', cr)
		col.add_attribute(cr, 'text', 0)
		col.set_resizable(True)
		tv.append_column(col)

		col = gtk.TreeViewColumn('Constraint', cr)
		col.add_attribute(cr, 'text', 1)
		col.set_resizable(True)
		tv.append_column(col)

		col = gtk.TreeViewColumn('Type', cr)
		col.add_attribute(cr, 'text', 2)
		col.set_resizable(True)
		tv.append_column(col)

		sw = gtk.ScrolledWindow()
		sw.set_policy(gtk.POLICY_AUTOMATIC,
			      gtk.POLICY_AUTOMATIC)
		sw.add(tv)
		table.attach(sw, 0, 3, 1, 2)

		self.constraint_notebook.insert_page \
		    (table, gtk.Label('Attribute'), self.NB_PAGE_ATTRIBUTE)

		b_add.connect('clicked',
			      self.handle_add_edit_constraint_click, None)
		b_edit.connect('clicked',
			       self.handle_add_edit_constraint_click, tv)
		b_remove.connect('clicked',
				 self.handle_remove_constraint_click, tv)

		self.b_add_constraint_attribute = b_add
		self.b_edit_constraint_attribute = b_edit
		self.b_remove_constraint_attribute = b_remove
		self.attribute_table = table
		self.attribute_constraints_tv = tv

		# all constraints page
		b_add = gtk.Button(None, gtk.STOCK_ADD)
		b_edit = gtk.Button(None, gtk.STOCK_EDIT)
		b_remove = gtk.Button(None, gtk.STOCK_REMOVE)

		table = gtk.Table(2, 3)
		table.attach(b_add, 0, 1, 0, 1,
			     gtk.FILL, gtk.FILL, 10, 10)
		table.attach(b_edit, 1, 2, 0, 1,
			     gtk.FILL, gtk.FILL, 10, 10)
		table.attach(b_remove, 2, 3, 0, 1,
			     gtk.FILL, gtk.FILL, 10, 10)

		ls = gtk.ListStore(gobject.TYPE_STRING,
				   gobject.TYPE_STRING,
				   gobject.TYPE_STRING)
		tv = gtk.TreeView(ls)
		tv.connect('button-press-event',
			   self.handle_tree_view_events)
		tv.connect('key-press-event',
			   self.handle_tree_view_events)
		selection = tv.get_selection()
		selection.connect('changed',
				  self.handle_selection_on_all_constraints_tv)
		cr = gtk.CellRendererText()

		col = gtk.TreeViewColumn('Attribute', cr)
		col.add_attribute(cr, 'text', 0)
		col.set_resizable(True)
		tv.append_column(col)

		col = gtk.TreeViewColumn('Constraint', cr)
		col.add_attribute(cr, 'text', 1)
		col.set_resizable(True)
		tv.append_column(col)

		col = gtk.TreeViewColumn('Type', cr)
		col.add_attribute(cr, 'text', 2)
		col.set_resizable(True)
		tv.append_column(col)

		sw = gtk.ScrolledWindow()
		sw.set_policy(gtk.POLICY_AUTOMATIC,
			      gtk.POLICY_AUTOMATIC)
		sw.add(tv)
		table.attach(sw, 0, 3, 1, 2)

		self.constraint_notebook.insert_page \
		    (table, gtk.Label('All'), self.NB_PAGE_ALL)

		b_add.connect('clicked',
			      self.handle_add_edit_constraint_click, None)
		b_edit.connect('clicked',
			       self.handle_add_edit_constraint_click, tv)
		b_remove.connect('clicked',
				 self.handle_remove_constraint_click, tv)

		self.b_add_constraint_all = b_add
		self.b_edit_constraint_all = b_edit
		self.b_remove_constraint_all = b_remove
		self.all_constraints_tv = tv

		# contradictions constraints page
		b_add = gtk.Button(None, gtk.STOCK_ADD)
		b_edit = gtk.Button(None, gtk.STOCK_EDIT)
		b_remove = gtk.Button(None, gtk.STOCK_REMOVE)

		table = gtk.Table(2, 3)
		table.attach(b_add, 0, 1, 0, 1,
			     gtk.FILL, gtk.FILL, 10, 10)
		table.attach(b_edit, 1, 2, 0, 1,
			     gtk.FILL, gtk.FILL, 10, 10)
		table.attach(b_remove, 2, 3, 0, 1,
			     gtk.FILL, gtk.FILL, 10, 10)

		ls = gtk.ListStore(gobject.TYPE_STRING,
				   gobject.TYPE_STRING,
				   gobject.TYPE_STRING)
		tv = gtk.TreeView(ls)
		tv.connect('button-press-event',
			   self.handle_tree_view_events)
		tv.connect('key-press-event',
			   self.handle_tree_view_events)
		selection = tv.get_selection()
		selection.connect('changed',
				  self.handle_selection_on_contradictory_constraints_tv)
		cr = gtk.CellRendererText()

		col = gtk.TreeViewColumn('Attribute', cr)
		col.add_attribute(cr, 'text', 0)
		col.set_resizable(True)
		tv.append_column(col)

		col = gtk.TreeViewColumn('Constraint', cr)
		col.add_attribute(cr, 'text', 1)
		col.set_resizable(True)
		tv.append_column(col)

		col = gtk.TreeViewColumn('Type', cr)
		col.add_attribute(cr, 'text', 2)
		col.set_resizable(True)
		tv.append_column(col)

		sw = gtk.ScrolledWindow()
		sw.set_policy(gtk.POLICY_AUTOMATIC,
			      gtk.POLICY_AUTOMATIC)
		sw.add(tv)
		table.attach(sw, 0, 3, 1, 2)

		self.constraint_notebook.insert_page \
		    (table, gtk.Label('Contradictions'),
		     self.NB_PAGE_CONTRADICTIONS)

		b_edit.connect('clicked',
			       self.handle_add_edit_constraint_click, tv)
		b_remove.connect('clicked',
				 self.handle_remove_constraint_click, tv)

		b_add.set_sensitive(False)
		self.b_edit_constraint_contradictions = b_edit
		self.b_remove_constraint_contradictions = b_remove
		self.contradictory_constraints_tv = tv

		frame = gtk.Frame('Type constraints')
		frame.add(self.constraint_notebook)
		self.constraints_frame = frame

		self.constraint_notebook.connect \
		    ('switch-page',
		     self.handle_constraints_notebook_page_change)
		self.constraint_notebook.set_current_page \
		    (self.NB_PAGE_ATTRIBUTE)
		self.current_constraints_notebook_page \
			= self.NB_PAGE_ATTRIBUTE

		###########################################################
		# attach widgets to the table
	        self.table.attach(self.menubar, 0, 1, 0, 1,
				  gtk.FILL | gtk.EXPAND,
				  gtk.FILL)
        	self.table.attach(self.toolbar, 0, 1, 1, 2,
				  gtk.FILL | gtk.EXPAND,
				  gtk.FILL)

		frame = gtk.Frame()
		frame.add(self.attribute_sw)
		v_paned = gtk.VPaned()
		h_paned = gtk.HPaned()
		v_paned.set_position(self.V_PANED_DIV_POS)
		h_paned.set_position(self.H_PANED_DIV_POS)
		v_paned.add1(frame)
		v_paned.add2(h_paned)
		self.table.attach(v_paned, 0, 1, 2, 3)
		h_paned.add1(self.constraints_frame)
		h_paned.add2(self.types_frame)

		self.selected_attribute = None
		self.current_constraints_notebook_page \
			= self.NB_PAGE_ATTRIBUTE

		self.win.set_default_size(self.WIN_DEFAULT_WIDTH,
					  self.WIN_DEFAULT_HEIGHT)
		self.win.show_all()

	def update_widget_sensitivities(self, data=None):
		# set the sensitivity of widgets which depend on whether a
		# model has been loaded
		widgets = [self.action_group.get_action(a)
			   for a in ('Export abstract model',
				     'Close',
				     'Run automatically',
				     'Attribute dependencies',
				     't_Export')]
		widgets.extend([self.attribute_tv, self.constraints_frame])
		for w in widgets:
			w.set_sensitive(self.shared_state.model != None)

		self.shared_state.automatic_type_inference = \
		    self.action_group.get_action('Run automatically') \
		        .get_proxies()[0].get_active()

		widgets = [self.action_group.get_action(a)
			   for a in ('Run', 't_Infer types')]
		for w in widgets:
			w.set_sensitive(self.shared_state.model != None
					and not self.shared_state.automatic_type_inference)

	def handle_menubar_selection(self, action):
		self.menubar_funcs[action.get_name()]()

	# handle default type selection from the default_type combo
	# box
	def handle_default_type_selection(self, cbox):
		dtype = self.cbox_get_active(cbox)
		if dtype == self.STR_NO_DEFAULT_TYPE:
			dtype = None
		self.set_default_type(dtype)

	def open_project(self):
		fc_dialog = gtk.FileChooserDialog('Open project',
			self.win,
			gtk.FILE_CHOOSER_ACTION_OPEN,
			(gtk.STOCK_OK, gtk.RESPONSE_OK,
			gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
		fc_dialog.set_default_response(gtk.RESPONSE_OK)
		fc_dialog.set_current_folder(self.current_project_directory)

		filt = gtk.FileFilter()
		filt.set_name('Project files (*.xmi, *.xml)')
		filt.add_pattern('*.xmi')
		filt.add_pattern('*.xml')
		fc_dialog.add_filter(filt)

		filt = gtk.FileFilter()
		filt.set_name('All files (*)')
		filt.add_pattern('*')
		fc_dialog.add_filter(filt)

		resp = fc_dialog.run()
		filename = fc_dialog.get_filename()
		try:
			if resp == gtk.RESPONSE_OK:
				self.current_project_directory = \
					fc_dialog.get_current_folder()
				self.load_model(filename)
		finally:
			fc_dialog.destroy()

        def quit(self):
		if not self.type_editor.confirm_changes_to_type_library \
			    ('quitting', self.win):
			return False
		self.close_model()
		gtk.main_quit()

	def edit_preferences(self):
		dialog = gtk.Dialog('Preferences', self.win,
				    gtk.DIALOG_DESTROY_WITH_PARENT,
				    (gtk.STOCK_SAVE, gtk.RESPONSE_OK,
				     gtk.STOCK_APPLY, gtk.RESPONSE_APPLY,
				     gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))

		notebook = gtk.Notebook()
		dialog.vbox.pack_start(notebook)

		str_ati = 'Automatic type inference'
		cb_ati = gtk.CheckButton(str_ati)
		if isinstance(self.shared_state.automatic_type_inference, bool):
			cb_ati.set_active(self.shared_state.automatic_type_inference)
		vbox_options = gtk.VBox()
		vbox_options.set_border_width(self.shared_state.DEFAULT_PADDING)
		vbox_options.set_spacing(self.shared_state.DEFAULT_PADDING)
		vbox_options.pack_start(cb_ati, False)

		str_model_rte = \
			'Model runtime exceptions in ' \
			+ 'automatic type generation'
		cb_model_rte = gtk.CheckButton(str_model_rte)
		if isinstance(self.shared_state.model_rte, bool):
			cb_model_rte.set_active(self.shared_state.model_rte)
		vbox_options.pack_start(cb_model_rte, False)

		table_paths = gtk.Table(3, 3)
		table_paths.set_border_width(self.shared_state.DEFAULT_PADDING)
		table_paths.set_row_spacings(self.shared_state.DEFAULT_PADDING)
		table_paths.set_col_spacings(self.shared_state.DEFAULT_PADDING)
		l = gtk.Label('SMT solver:')
		l.set_alignment(1.0, 0.5)
		table_paths.attach(l, 0, 1, 0, 1, gtk.FILL, gtk.FILL)
		cbox_solver = gtk.combo_box_new_text()
		for (i, solver) in enumerate(SharedState.SMT_SOLVERS):
			cbox_solver.append_text(solver)
			if solver == self.shared_state.solver:
				cbox_solver.set_active(i)
		table_paths.attach(cbox_solver, 1, 2, 0, 1,
				   gtk.FILL, gtk.FILL)
		l = gtk.Label('SMT solver path: ')
		l.set_alignment(1.0, 0.5)
		table_paths.attach(l, 0, 1, 1, 2,
				   gtk.FILL, gtk.FILL)
		solver_path_entry = gtk.Entry()
		table_paths.attach(solver_path_entry, 1, 2, 1, 2,
				   gtk.FILL | gtk.EXPAND, gtk.FILL)
		solver_path_entry.set_text(self.shared_state.solver_path)
		b_browse = gtk.Button('Browse...')
		b_browse.connect('clicked',
				 self.handle_file_selection_for_entry,
				 solver_path_entry)
		table_paths.attach(b_browse, 2, 3, 1, 2,
				   gtk.FILL, gtk.FILL,
				   2, 0)

		l = gtk.Label('Dot path: ')
		l.set_alignment(1.0, 0.5)
		table_paths.attach(l, 0, 1, 2, 3, gtk.FILL, gtk.FILL)
		dot_path_entry = gtk.Entry()
		table_paths.attach(dot_path_entry, 1, 2, 2, 3,
				   gtk.FILL | gtk.EXPAND, gtk.FILL)
		dot_path_entry.set_text(self.shared_state.dot_path)
		b_browse = gtk.Button('Browse...')
		b_browse.connect('clicked',
				 self.handle_file_selection_for_entry,
				 dot_path_entry)
		table_paths.attach(b_browse, 2, 3, 2, 3,
				   gtk.FILL, gtk.FILL,
				   2, 0)

		notebook.append_page(vbox_options, gtk.Label('Options'))
		notebook.append_page(table_paths, gtk.Label('External tools'))

		dialog.show_all()
		resp = dialog.run()
		dialog.destroy()

		if resp == gtk.RESPONSE_OK \
		       or resp == gtk.RESPONSE_APPLY:
			self.shared_state.automatic_type_inference = cb_ati.get_active()
			self.action_group.get_action('Run automatically') \
			    .get_proxies()[0].set_active(cb_ati.get_active())
			self.shared_state.model_rte = cb_model_rte.get_active()
			self.shared_state.solver = SharedState.SMT_SOLVERS[cbox_solver.get_active()]
			self.shared_state.solver_path = solver_path_entry.get_text()
			self.shared_state.dot_path = dot_path_entry.get_text()

			if resp != gtk.RESPONSE_OK:
				return

			try:
				f = open(self.PREFERENCES_PATH, 'w')
			except IOError, e:
				ErrorDialog(self.win,
					    'Failed to save preferences!',
					    str(e))
				return

			f.write(str_ati + ': ' \
				+ str(self.shared_state.automatic_type_inference) \
				+ '\n')
			f.write(str_model_rte + ': ' \
				+ str(self.shared_state.model_rte) \
				+ '\n')
			f.write('Solver: ' + self.shared_state.solver + '\n')
			f.write('Solver path: ' + self.shared_state.solver_path \
				+ '\n')
			f.write('Dot path: ' + self.shared_state.dot_path)

			f.close()

	def handle_file_selection_for_entry(self, action, entry):
		fc_dialog = gtk.FileChooserDialog('Browse',
						  self.win,
						  gtk.FILE_CHOOSER_ACTION_OPEN,
						  (gtk.STOCK_OK,
						   gtk.RESPONSE_OK,
						   gtk.STOCK_CANCEL,
						   gtk.RESPONSE_CANCEL))
		fc_dialog.set_default_response(gtk.RESPONSE_OK)
		resp = fc_dialog.run()
		if resp == gtk.RESPONSE_OK:
			entry.set_text(fc_dialog.get_filename())
		fc_dialog.destroy()

	def edit_types(self):
		while True:
			self.win.set_sensitive(False)
			self.type_editor.run()
			self.win.set_sensitive(True)
			if self.shared_state.model != None:
				try:
					self.shared_state.model \
					    .attach_type_library \
					    (self.shared_state.tlib
					     .get_type_library())
				except (TypeError, ValueError, Error), e:
					dialog = gtk.MessageDialog \
					    (self.win,
					     gtk.DIALOG_MODAL
					     | gtk.DIALOG_DESTROY_WITH_PARENT,
					     gtk.MESSAGE_ERROR,
					     gtk.BUTTONS_NONE,
					     'The type library is not '
					     'compatible with the currently '
					     'loaded model! Re-edit types to '
					     'fix this problem?')
					dialog.set_title('Error')
					dialog.format_secondary_text(str(e))
					dialog.add_button \
					    ('Yes, open type editor', 0)
					dialog.add_button \
					    ('No, close the model', 1)
					dialog.set_default_response(0)
					dialog.show_all()
					response = dialog.run()
					dialog.destroy()
					if response == 1:
						self.close_model()
						break
				else:
					break
			else:
				break

		self.update_types()
		self.update_constraints()
		self.update_buttons()
		self.action_group.get_action('Type hierarchy') \
		    .get_proxies()[0].set_active \
		    (self.shared_state.type_hierarchy_window.is_visible())

		# update the default type if it was deleted
		if not self.shared_state.default_type in \
			    self.shared_state.tlib.get_int_types():
			self.set_default_type(None)

	# handle the selection of a default type
	def change_default_type(self):
		dialog = gtk.Dialog('Change default type', self.win,
				    gtk.DIALOG_DESTROY_WITH_PARENT \
				    | gtk.DIALOG_MODAL,
				    (gtk.STOCK_OK, gtk.RESPONSE_OK,
				     gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))

		table = gtk.Table(2, 1)
		table.set_border_width(self.shared_state.DEFAULT_PADDING)
		table.set_row_spacings(self.shared_state.DEFAULT_PADDING)
		dialog.vbox.pack_start(table)

		l = gtk.Label("Select type in the list below:")
		table.attach(l, 0, 1, 0, 1)

		ls = gtk.ListStore(gobject.TYPE_STRING)
		ls.append((self.STR_NO_DEFAULT_TYPE,))
		selected = 0
		for (i, t) in enumerate(self.shared_state.tlib
					.get_int_types()):
			ls.append((t,))
			if t == self.shared_state.default_type:
				selected = i + 1

		cr = gtk.CellRendererText()
		col = gtk.TreeViewColumn(None, cr)
		col.add_attribute(cr, 'text', 0)
		tv = gtk.TreeView(ls)
		tv.set_headers_visible(False)
		tv.append_column(col)
		tv.get_selection().select_path((selected,))
		sw = gtk.ScrolledWindow()
		sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
		sw.add(tv)
		table.attach(sw, 0, 1, 1, 2)

		dialog.set_default_size(200, 200)
		dialog.show_all()
		resp = dialog.run()
		(model, it) = tv.get_selection().get_selected()
		dialog.destroy()

		if resp != gtk.RESPONSE_OK \
		       or it == None:
			return

		dtype = model.get_value(it, 0)
		if dtype == self.STR_NO_DEFAULT_TYPE:
			dtype = None
		self.set_default_type(dtype)

	def handle_type_hierarchy_window_delete_event(self, window, event):
		self.action_group.get_action('Type hierarchy') \
		    .get_proxies()[0].set_active(False)
		return True

	def update_type_hierarchy_window(self):
		checkbox = self.action_group.get_action('Type hierarchy') \
		               .get_proxies()[0]
		if checkbox.get_active():
			try:
				self.shared_state.type_hierarchy_image.set \
				    (self.shared_state.dot_path,
				     self.shared_state.tlib
				         .prepare_input_for_dot)
			except IOError, e:
				ErrorDialog \
				    (self.win, "Failed to show image!",
				     str(e))
				self.shared_state.type_hierarchy_window.hide()
				checkbox.set_active(False)
			else:
				self.shared_state.type_hierarchy_window.show()
		else:
			self.shared_state.type_hierarchy_window.hide()

	def handle_attribute_dependency_window_delete_event(self, win, event):
		self.action_group.get_action('Attribute dependencies') \
		    .get_proxies()[0].set_active(False)
		return True
				
	def update_attribute_dependency_window(self):
		checkbox = self.action_group \
		               .get_action('Attribute dependencies') \
			       .get_proxies()[0]
		if self.shared_state.model != None and checkbox.get_active():
			try:
				self.shared_state.attribute_dependency_image \
				    .set \
					(self.shared_state.dot_path,
					 self.shared_state.model
					 .prepare_input_for_dot)
			except IOError, e:
				ErrorDialog \
				    (self.win, "Failed to show image!",
				     str(e))
				self.shared_state.attribute_dependency_window \
				    .hide()
				checkbox.set_active(False)
			else:
				self.shared_state.attribute_dependency_window \
				    .show()
		else:
			self.shared_state.attribute_dependency_window.hide()

	def load_preferences(self):
		try:
			f = open(self.PREFERENCES_PATH, 'r')
		except IOError:
			return

		str_ati = 'Automatic type inference: '
		str_model_rte = 'Model runtime exceptions in ' \
				+ 'automatic type generation: '
		str_solver = 'Solver: '
		str_solver_path = 'Solver path: '
		str_dot_path = 'Dot path: '

		for line in f:
			if str_ati in line:
				if 'true' in line.lower():
					self.shared_state.automatic_type_inference = True
				else:
					self.shared_state.automatic_type_inference = False
			elif str_model_rte in line:
				if 'true' in line.lower():
					self.shared_state.model_rte = True
				else:
					self.shared_state.model_rte = False
			elif str_solver in line:
				line = map(str.strip, line.split())
				if len(line) > 1:
					try:
						self.shared_state.solver = SharedState.SMT_SOLVERS[map(str.lower, SharedState.SMT_SOLVERS).index(line[1].lower())]
					except ValueError:
						self.shared_state.solver = "Yices"
				else:
					self.shared_state.solver = "Yices"
			elif str_solver_path in line:
				i = line.find(str_solver_path) \
				    + len(str_solver_path)
				self.shared_state.solver_path = line[i:]
				self.shared_state.solver_path = self.shared_state.solver_path.rstrip('\n')
			elif str_dot_path in line:
				i = line.find(str_dot_path) \
				    + len(str_dot_path)
				self.shared_state.dot_path = line[i:]
				self.shared_state.dot_path = self.shared_state.dot_path.rstrip('\n')
		f.close()

	# update contents of the types frame
	def update_types(self):
		tv_model = self.types_tv.get_model()
		tv_model.clear()

		self.default_type_cbox.handler_block(self.default_type_cbox_changed_handler)
		cbox_model = self.default_type_cbox.get_model()
		cbox_model.clear()
		cbox_model.append((self.STR_NO_DEFAULT_TYPE,))

		index = 0
		for (i, t) in enumerate(self.shared_state.tlib
					.get_int_types()):
			tv_model.append((t,))
			cbox_model.append((t,))
			if t == self.shared_state.default_type:
				index = i + 1

		self.default_type_cbox.set_active(index)
		self.default_type_cbox.handler_unblock(self.default_type_cbox_changed_handler)

	# get the active item from the given combobox; return None if
	# there was no selection
	def cbox_get_active(self, cbox):
		i = cbox.get_active()
		if i < 0:
			return None
		model = cbox.get_model()
		return model[i][0]

	def set_default_type(self, dtype):
		value = dtype
		if dtype == None:
			value = self.STR_NO_DEFAULT_TYPE
			i = 0
		else:
			i = self.shared_state.tlib.get_int_types() \
			        .index(value) + 1
		self.default_type_cbox.set_active(i)
		self.shared_state.default_type = dtype

	def infer_types(self):
		try:
			self.shared_state.model.infer_types \
			    (self.shared_state.default_type)
		except (TypeError, ValueError, Error), e:
			ErrorDialog(self.win,
				    'Type inference was unsuccessful!',
				    str(e))
		self.update_contradictory_constraints()
		self.update_buttons()
		self.refresh()

	# update the buttons in the main window
	def update_buttons(self):
		page_num = self.current_constraints_notebook_page
		if page_num == self.NB_PAGE_ATTRIBUTE:
			attr = self.get_selected_attribute()
			value = True
			if attr == None:
				self.b_add_constraint_attribute.set_sensitive(False)
				value = False
			else:
				self.b_add_constraint_attribute.set_sensitive(True)
				selection = self.attribute_constraints_tv.get_selection()
				(model, it) = selection.get_selected()
				# there's no selected item
				if it == None:
					value = False
			self.b_edit_constraint_attribute.set_sensitive(value)
			self.b_remove_constraint_attribute.set_sensitive(value)
		elif page_num == self.NB_PAGE_ALL:
			selection = self.all_constraints_tv.get_selection()
			(model, it) = selection.get_selected()
			value = True
			# there's no selected item
			if it == None:
				value = False
			self.b_edit_constraint_all.set_sensitive(value)
			self.b_remove_constraint_all.set_sensitive(value)
		elif page_num == self.NB_PAGE_CONTRADICTIONS:
			selection = self.contradictory_constraints_tv.get_selection()
			(model, it) = selection.get_selected()
			value = True
			# there's no selected item
			if it == None:
				value = False
			self.b_edit_constraint_contradictions.set_sensitive(value)
			self.b_remove_constraint_contradictions.set_sensitive(value)

	def refresh(self):
		self.win.queue_draw()

	# cell data function for the column "last inferred type"
	def cdf_last_inferred_type(self, col, cell, model, iter):
		x = self.shared_state.model.get_with_path(model.get_path(iter))
		name = None
		if isinstance(x, Attribute):
			name = x.get_inferred_type()
		cell.set_property('text', name)

	# cell data function for the column "changed"
	def cdf_changed(self, col, cell, model, iter):
		x = self.shared_state.model.get_with_path(model.get_path(iter))
		if x == None or isinstance(x, Model):
			cell.set_property('pixbuf', None)
		elif x.changed:
			cell.set_property('pixbuf', self.pb_changed)
		else:
			cell.set_property('pixbuf', None)

	# called at least when the window is closed when the 'x'
	# (usually on the top right corner) is pressed
	def handle_delete_event(self, widget, event, data = None):
		self.quit()
		#return False

	# load a model from the given file
	def load_model(self, filename):
		model = Model()
		try:
			model.load(filename)
			model.attach_type_library \
			    (self.shared_state.tlib.get_type_library())
		except (TypeError, ValueError, Error), e:
			ErrorDialog(self.win, 'Cannot open project!', str(e))
			self.close_model()
			return
		self.shared_state.model = model
		self.win.set_title \
		    (self.shared_state.model.name + ': ' + filename)
		self.shared_state.attribute_dependency_image.invalidate()
		self.update_attribute_dependency_window()
		self.update_attribute_tree_view()
		self.update_constraints()
		self.update_widget_sensitivities()

	def close_model(self):
		if self.shared_state.model == None:
			return
		self.shared_state.model = None
		self.win.set_title(self.WIN_DEFAULT_TITLE)
		self.shared_state.attribute_dependency_image.invalidate()
		self.action_group.get_action('Attribute dependencies') \
		    .get_proxies()[0].set_active(False)
		self.update_attribute_dependency_window()
		self.update_attribute_tree_view()
		self.update_constraints()
		self.update_widget_sensitivities()

	# export the model; return True if export succeeded, False
	# otherwise
	def export_model(self):
		if self.shared_state.model == None:
			return False
		try:
			self.shared_state.model.can_save()
		except Error, e:
			ErrorDialog(self.win,
				    'The model cannot be exported!',
				    str(e))
			return False

		if self.shared_state.tlib.changed \
			    or not self.shared_state.tlib.tlib_files:
			dialog = gtk.MessageDialog \
			             (self.win,
				      gtk.DIALOG_MODAL
				      | gtk.DIALOG_DESTROY_WITH_PARENT,
				      gtk.MESSAGE_INFO,
				      gtk.BUTTONS_OK,
				      'Type library not exported!')
			dialog.format_secondary_text \
			    ('The current type library needs to be exported '
			     'before generating an abstract model. Click OK '
			     'to continue.')
			dialog.run()
			dialog.destroy()
			if not self.type_editor.export_file(self.win):
				return False

		projectfiles = 'Project files (*.xmi, *.xml)'
		dialog = SaveFileDialog \
		             (self.win,
			      'Select file name for exported model',
			      ((projectfiles, ('*.xmi', '*.xml')),
			       ('All files (*)', ('*',))))
		(filename, directory, filter) = \
			dialog.run(self.current_project_directory)
		if filename == None:
			return False

		self.current_project_directory = directory
		if filter.get_name() == projectfiles \
			    and not filename[-4:] in ('.xmi', '.xml'):
			filename += ".xmi"

		try:
			self.shared_state.model.save \
			    (filename,
			     self.shared_state.tlib.get_type_library(),
			     self.shared_state.tlib.tlib_files)
		except Exception, e:
			ErrorDialog(self.win,
				    'Failed to export abstract model!',
				    str(e))
			return False
		return True

	# construct a tree from the classes and signals of the model
	# to the given tree model (tree_store)
	def construct_tree_model(self, tree_store):
		if self.shared_state.model == None:
			return None
		if self.shared_state.model.has_classes():
			it = tree_store.append(None, ['Classes', None, None, None])
			for c in self.shared_state.model.get_classes():
				self.construct_class_subtree(tree_store, it, c)
		if self.shared_state.model.has_signals():
			it = tree_store.append(None, ['Signals', None, None, None])
			for s in self.shared_state.model.get_signals():
				self.construct_signal_subtree(tree_store, it, s)
		return tree_store

	# recursively construct a class subtree such that the given
	# class (cla) is a direct child of the given parent (it_parent
	# == parent's iterator)
	def construct_class_subtree(self, tree_store, it_parent, cla):
		it = tree_store.append(it_parent,
				       [cla.get_name(), None, None, None])
		for attr in cla.get_attributes():
			tree_store.append(it,
					  [attr.get_name(),
					   attr.get_original_type(),
					   None, None])
		for c in cla.classes:
			self.construct_class_subtree(tree_store, it, c)

	def construct_signal_subtree(self, tree_store, it_parent, sig):
		it = tree_store.append(it_parent,
				       [sig.get_name(), None, None, None])
		for attr in sig.get_attributes():
			tree_store.append(it,
					  [attr.get_name(),
					   attr.get_original_type(),
					   None, None])

	# handle page change on the constraints notebook (just record
	# the selected page and update the buttons)
	def handle_constraints_notebook_page_change(self, notebook, page, page_num):
		self.current_constraints_notebook_page = page_num
		self.update_buttons()

	def handle_remove_constraint_click(self, button, tree_view):
		selection = tree_view.get_selection()
		(model, it) = selection.get_selected()
		assert(it != None)
		full_name = model.get_value(it, 0)
		constraint = model.get_value(it, 1)
		type_name = model.get_value(it, 2)
		attr = self.shared_state.model.get_attribute_with_full_name(full_name)
		attr.remove_constraint(constraint, type_name)
		if self.shared_state.automatic_type_inference:
			self.infer_types()
		self.update_constraints()
		self.update_buttons()

	# handle selection from the attribute tree view
	def handle_selection_on_attribute_tv(self, selection):
		(model, it) = selection.get_selected()
		if it == None:
			self.set_selected_attribute(None)
		else:
			path = model.get_path(it)
			x = self.shared_state.model.get_with_path(path)
			if isinstance(x, Attribute):
				self.set_selected_attribute(x)
			else:
				self.set_selected_attribute(None)
		self.update_attribute_constraints()
		self.update_buttons()

	# handle selection from the attribute constraints tree view in
	# the constraints notebook
	def handle_selection_on_attribute_constraints_tv(self, selection):
		self.update_buttons()

	# handle selection from the all constraints tree view in the
	# constraints notebook
	def handle_selection_on_all_constraints_tv(self, selection):
		self.update_buttons()

	# handle selection from the contradictory constraints tree
	# view in the constraints notebook
	def handle_selection_on_contradictory_constraints_tv(self, selection):
		self.update_buttons()

	# handle selection from the types tree view (currently, no
	# selection can be made from the types tree view)
	def handle_selection_on_types_tv(self, selection):
		return

	# handle events (mouse button and key presses) in all tree
	# views
	def handle_tree_view_events(self, widget, event):
		if widget != self.attribute_tv \
		   and widget != self.attribute_constraints_tv \
		   and widget != self.all_constraints_tv \
		   and widget != self.contradictory_constraints_tv \
		   and widget != self.types_tv:
			return True

		if event.type != gtk.gdk.BUTTON_PRESS \
		       and event.type != gtk.gdk._2BUTTON_PRESS \
		       and event.type != gtk.gdk.KEY_PRESS:
			return True

		# no selection allowed in the types tree view
		if widget == self.types_tv:
			return True

		if event.type == gtk.gdk._2BUTTON_PRESS:
			# allow double click only with the left mouse
			# button
			if event.button != 1:
				return False
			if widget == self.attribute_tv:
				path = widget.get_path_at_pos(int(event.x),
							      int(event.y))
				if path == None:
					return True
				path = path[0]
				if isinstance(self.shared_state.model.get_with_path(path),
					      Attribute):
					self.b_add_constraint_attribute.emit('clicked')
			elif widget == self.attribute_constraints_tv:
				self.b_edit_constraint_attribute.emit('clicked')
			elif widget == self.all_constraints_tv:
				self.b_edit_constraint_all.emit('clicked')
			elif widget == self.contradictory_constraints_tv:
				self.b_edit_constraint_contradictions.emit('clicked')

		return False

if __name__ == '__main__':
	gui = Gui()
	gtk.main()
