# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Data structures for type editor GUI.

This module defines data structures used in the type editor GUI.

Authors: Juhani Peltonen
         Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

import abstraction.typelibrary
from abstraction import smtlib_typegen
from abstraction.typelibrary import TypeLibrary
from abstraction.typehierarchy import ConsistencyError
from abstraction.error import Error
from os import popen, remove, write, close
from tempfile import mkstemp
from Utils.ExpressionTraverser import actlang_to_python

class Tlib:
    def __init__(self):
        self.ID_TYPE = 0
        self.ID_OPERATOR = 1
        self.ID_TEST = 2
        self.ID_COERCION = 3

        self.STATE_UNDEFINED = 0
        self.STATE_PARTIALLY_DEFINED = 1
        self.STATE_COMPLETELY_DEFINED = 2

        self.tlib = TypeLibrary()
        self.tlib_files = []

        self.op_symbols = {}
        self.op_symbols[1] = ['+', '-', '~']
        self.op_symbols[2] = ['+', '-', '*', '/', '%', '&', '|', '^',
                              '<<', '>>', '>>>']
        self.all_op_symbols = ['+', '-', '*', '/', '%', '&', '|', '~',
                               '^', '<<', '>>', '>>>']
        self.test_symbols = {}
        self.test_symbols[1] = ['!']
        self.test_symbols[2] = ['&&', '||', '==', '!=', '<', '<=', '>',
                                '>=']
        self.all_test_symbols = ['&&', '||', '==', '!', '!=', '<', '<=',
                                 '>', '>=']

        self.non_abstractable_symbols = ['&&', '||', '!']

        self.selection_id = None
        self.selected_type = None
        self.selected_domain_element = None
        self.selected_operator = (None, None)
        self.selected_test = (None, None)
        self.selected_coercion = (None, None)
        self.selected_value = (None, None)

        self.changed = False

    def read(self, filenames):
        self.changed = False
        self.tlib = TypeLibrary()
        self.tlib.read(filenames, sloppy=True)
        try:
            for (to_type, fromtypes) in self.tlib.coercions.iteritems():
                for (from_type, backends) in fromtypes.iteritems():
                    if from_type == 'int':
                        continue
                    try:
                        dfn = backends['Jumbala']
                    except KeyError:
                        continue
                    for (test, results) in dfn:
                        if test == 'true':
                            continue
                        for val in self.tlib.domain[from_type].iterkeys():
                            if test == '<> == <%s>' % val:
                                break
                        else:
                            raise ValueError \
                                  ("unsupported coercion test `%s' in type "
                                   "library" % test)
            self.tlib_files = filenames[:]
        except Exception, e:
            self.tlib.__init__()
            self.tlib_files = []
            raise

    def write(self, filename, type_names = None, operators = None,
              tests = None, coercions = None):
        for (to_type, fromtypes) in self.tlib.coercions.iteritems():
            for (from_type, backends) in fromtypes.iteritems():
                try:
                    dfn = backends['Jumbala']
                except KeyError:
                    continue
                if from_type == 'int':
                    python_dfn = [(self.coercion_from_jumbala_to_python(case)
                                       .replace('x', '<>'), results)
                                  for (case, results) in dfn]
                else:
                    python_dfn = []
                    for (case, results) in dfn:
                        if case == 'true':
                            python_dfn.append(('True', results))
                        else:
                            python_dfn.append((case, results))
                self.tlib.defineCoercion \
                    (from_type, to_type, 'python', python_dfn)
        if filename == "-":
            filename = "./-"
        self.tlib.write(filename, type_names, operators, tests, coercions)
        self.changed = False
        self.tlib_files = [filename]

    def get_type_library(self):
        return self.tlib

    def get_types(self):
        return sorted([type_name for type_name in self.tlib.typenames
                       if not type_name in self.tlib.STD_TYPES],
                      key=lambda x: x.lower())

    def get_std_types(self):
        return sorted(self.tlib.STD_TYPES, key=lambda x:x.lower())

    def get_int_types(self):
        return sorted([type_name for type_name in self.tlib.typenames
                       if type_name != 'boolean'],
                      key=lambda x: x.lower())

    def get_domain(self, type_name):
        assert(type_name in self.tlib.domain)
        return sorted(self.tlib.domain[type_name].keys(),
                      key=lambda x: x.lower())

    def get_backends(self):
        return self.tlib.STD_BACKENDS

    # return operators of the given type as a list of tuples (op_name,
    # arity)
    def get_operators(self, type_name):
        assert(self.tlib.operators.has_key(type_name))
        lst = []
        for (arity, operations) in self.tlib.operators[type_name].iteritems():
            lst.extend([(op_name, arity) for op_name in operations.iterkeys()])
        lst.sort(key=lambda x: self.all_op_symbols.index(x[0].lower()))
        return lst

    # return tests of the given type as a list of tuples (test_name,
    # arity)
    def get_tests(self, type_name):
        assert(self.tlib.tests.has_key(type_name))
        lst = []
        for (arity, tests) in self.tlib.tests[type_name].iteritems():
            lst.extend([(test_name, arity) for test_name in tests.iterkeys()])
        lst.sort(key=lambda x: self.all_test_symbols.index(x[0].lower()))
        return lst

    # return coercions of the given type as a list of tuples
    # (from_type, backend)
    def get_coercions(self, to_type):
        try:
            return sorted([(from_type, 'Jumbala') for from_type
                           in self.tlib.coercions[to_type].iterkeys()],
                          key=lambda x: x[0].lower())
        except KeyError:
            return []

    # return undefined operators (those that are not found in the type
    # library) of the given type as list of tuples (op_name, arity)
    def get_undefined_operators(self, type_name):
        def_symbols = {}
        def_symbols[1] = []
        def_symbols[2] = []

        for (op_name, arity) in self.get_operators(type_name):
            def_symbols[arity].append(op_name)

        def_symbols[1] = set(def_symbols[1])
        def_symbols[2] = set(def_symbols[2])

        udef_symbols = {}
        udef_symbols[1] = list(set(self.op_symbols[1]) - def_symbols[1])
        udef_symbols[2] = list(set(self.op_symbols[2]) - def_symbols[2])

        op_symbols = self.get_all_abstractable_operator_symbols()
        lst = []
        for (op_name, arity) in op_symbols:
            if op_name in udef_symbols[arity]:
                lst.append((op_name, arity))

        return lst

    # return undefined tests (those that are not found in the type
    # library) of the given type as list of tuples (test_name, arity)
    def get_undefined_tests(self, type_name):
        def_symbols = {}
        def_symbols[1] = []
        def_symbols[2] = []

        for (test_name, arity) in self.get_tests(type_name):
            def_symbols[arity].append(test_name)

        def_symbols[1] = set(def_symbols[1])
        def_symbols[2] = set(def_symbols[2])

        udef_symbols = {}
        udef_symbols[1] = list(set(self.test_symbols[1]) - def_symbols[1])
        udef_symbols[2] = list(set(self.test_symbols[2]) - def_symbols[2])

        test_symbols = self.get_all_abstractable_test_symbols()
        lst = []
        for (test_name, arity) in test_symbols:
            if test_name in udef_symbols[arity]:
                lst.append((test_name, arity))
        
        return lst

    # return undefined coercions (those that are not found in the type
    # library) of the given type as list of tuples (from_type,
    # backend)
    def get_undefined_coercions(self, to_type):
        assert(to_type in self.tlib.coercions)
        lst = []
        for from_type in self.get_types():
            if from_type == to_type:
                continue
            if not from_type in self.tlib.coercions[to_type] \
                and self.coercion_can_define(from_type, to_type, 'Jumbala'):
                lst.append((from_type, 'Jumbala'))
        if not 'int' in self.tlib.coercions[to_type] \
               and self.coercion_can_define('int', to_type, 'Jumbala'):
            lst.append(('int', 'Jumbala'))
        lst.sort(key=lambda x: x[0].lower())
        return lst

    # return true if the given coercion (from_type -> to_type) can be
    # defined (i.e., to_type doesn't belong to std_types, there would
    # be no cycles, etc.), false otherwise
    def coercion_can_define(self, from_type, to_type, backend):
        if from_type == to_type or to_type in self.tlib.STD_TYPES:
            return False
        try:
            if backend in self.tlib.coercions[to_type][from_type]:
                return False
        except KeyError:
            pass
        return not from_type \
               in self.tlib.type_ordering.moreAbstractTypes(to_type)

    def set_selected_type(self, type_name):
        self.selection_id = self.ID_TYPE
        self.selected_type = type_name

    def set_selected_type_with_index(self, index):
        types = self.get_types()
        assert((index >= 0) and (index < len(types)))
        self.set_selected_type(types[index])

    def set_selected_domain_element(self, element):
        self.selected_domain_element = element

    def set_selected_domain_element_with_index(self, index):
        domain = self.get_domain(self.selected_type)
        assert((index >= 0) and (index < len(domain)))
        self.selected_domain_element = domain[index]

    def set_selected_operator(self, op_name, arity):
        self.selection_id = self.ID_OPERATOR
        self.selected_operator = (op_name, arity)

    def set_selected_operator_with_index(self, index):
        operators = self.get_operators(self.selected_type)
        assert((index >= 0) and (index < len(operators)))
        (op_name, arity) = operators[index]
        self.set_selected_operator(op_name, arity)

    def set_selected_test(self, test_name, arity):
        self.selection_id = self.ID_TEST
        self.selected_test = (test_name, arity)

    def set_selected_test_with_index(self, index):
        tests = self.get_tests(self.selected_type)
        assert((index >= 0) and (index < len(tests)))
        (test_name, arity) = tests[index]
        self.set_selected_test(test_name, arity)

    def set_selected_coercion(self, from_type, backend):
        self.selection_id = self.ID_COERCION
        self.selected_coercion = (from_type, backend)

    def set_selected_coercion_with_index(self, index):
        coercions = self.get_coercions(self.selected_type)
        assert((index >= 0) and (index < len(coercions)))
        (from_type, backend) = coercions[index]
        self.set_selected_coercion(from_type, backend)

    def set_selected_value(self, operands, results):
        if results != None:
            results.sort(lambda x, y: cmp(str(x).lower(),
                                          str(y).lower()))
        self.selected_value = (operands, results)

    def get_selected_type(self):
        return self.selected_type

    def get_selected_domain_element(self):
        return self.selected_domain_element

    def get_selected_operator(self):
        assert((self.selection_id == self.ID_OPERATOR)
               and (self.selected_type != None))
        return self.selected_operator

    def get_selected_test(self):
        assert((self.selection_id == self.ID_TEST)
               and (self.selected_type != None))
        return self.selected_test

    def get_selected_coercion(self):
        assert((self.selection_id == self.ID_COERCION) \
               and (self.selected_type != None))
        return self.selected_coercion

    def get_selected_value(self):
        return self.selected_value

    def get_selection_id(self):
        return self.selection_id

    def set_selection_id(self, selection_id):
        assert((selection_id == self.ID_TYPE) \
               or (selection_id == self.ID_OPERATOR) \
               or (selection_id == self.ID_TEST) \
               or (selection_id == self.ID_COERCION))
        self.selection_id = selection_id

    # return the operands and the results of the given operation as a list of
    # tuples (operands, results) where operands is a tuple with one or two
    # elements depending on the arity
    def get_operation_values(self, arity, dfn, tr_operand, tr_value):
        if arity == 1:
            return sorted([((tr_operand(operands[0]),),
                            [tr_value(v) for v in values])
                           for (operands, values) in dfn.iteritems()],
                          key=lambda x: x[0][0].lower())
        else:
            return sorted([((tr_operand(operands[0]), tr_operand(operands[1])),
                            [tr_value(v) for v in values])
                           for (operands, values) in dfn.iteritems()],
                          key=lambda x: (x[0][0].lower(), x[0][1].lower()))

    # return the operands and the results of the given operator as a
    # list of tuples (operands, results) where operands is a tuple
    # with one or two elements depending on the arity
    def get_operator_values(self, type_name, op_name, arity):
        assert(self.tlib.operators.has_key(type_name))
        assert(self.tlib.operators[type_name].has_key(arity))
        assert(self.tlib.operators[type_name][arity].has_key(op_name))
        return self.get_operation_values \
                   (arity, self.tlib.operators[type_name][arity][op_name],
                    lambda x: self.tlib.numdomain[type_name][x],
                    lambda x: self.tlib.numdomain[type_name][x])

    # return the operands and the results of the given test as a list
    # of tuples (operands, results) where operands is a tuple with one
    # or two elements depending on the arity
    def get_test_values(self, type_name, test_name, arity):
        assert(self.tlib.tests.has_key(type_name))
        assert(self.tlib.tests[type_name].has_key(arity))
        assert(self.tlib.tests[type_name][arity].has_key(test_name))
        return self.get_operation_values \
                   (arity, self.tlib.tests[type_name][arity][test_name],
                    lambda x: self.tlib.numdomain[type_name][x],
                    lambda x: x)

    # return the operand (or the test) and the results of the given
    # coercion as a list of tuples (operand, results)
    def get_coercion_values(self, from_type, to_type, backend):
        assert(self.tlib.coercions.has_key(to_type))
        assert(self.tlib.coercions[to_type].has_key(from_type))
        assert(self.tlib.coercions[to_type][from_type].has_key(backend))

        lst = []
        for (test, num_results) in self.tlib.coercions[to_type][from_type][backend]:
            sym_results = []
            for result in num_results:
                sym_results.append(self.tlib.numdomain[to_type][result])
            sym_results.sort(lambda x, y: cmp(x.lower(), y.lower()))
            lst.append((test, sym_results))

        return lst

    def add_type(self, type_name, domain):
        self.tlib.addType(type_name, domain)
        self.changed = True

    def remove_type(self, type_name):
        assert(type_name in self.tlib.typenames)
        self.tlib.removeType(type_name)
        self.changed = True

    def has_types(self):
        return len(self.tlib.typenames) != len(self.tlib.STD_TYPES)

    def add_domain_elements(self, type_name, elements):
        assert(self.tlib.domain.has_key(type_name))

        self.tlib.addDomainElements(type_name, elements)
        self.changed = True

        # redefine all operators and tests
        if self.tlib.operators.has_key(type_name):
            for element in elements:
                for arity in self.tlib.operators[type_name]:
                    for op_name in self.tlib.operators[type_name][arity]:
                        if arity == 1:
                            self.define_operator(type_name, op_name,
                                                 [element], [])
                        elif arity == 2:
                            for lhs in self.tlib.numdomain[type_name]:
                                self.define_operator(type_name, op_name,
                                                     [lhs, element], [])
                                for rhs in self.tlib.numdomain[type_name]:
                                    self.define_operator(type_name, op_name,
                                                         [element, rhs], [])

        if self.tlib.tests.has_key(type_name):
            for element in elements:
                for arity in self.tlib.tests[type_name]:
                    for test_name in self.tlib.tests[type_name][arity]:
                        if arity == 1:
                            self.define_test(type_name, test_name,
                                             [element], [])
                        elif arity == 2:
                            for lhs in self.tlib.numdomain[type_name]:
                                self.define_test(type_name, test_name,
                                                     [lhs, element], [])
                                for rhs in self.tlib.numdomain[type_name]:
                                    self.define_test(type_name, test_name,
                                                     [element, rhs], [])

    def remove_domain_elements(self, type_name, elements):
        assert(self.tlib.domain.has_key(type_name))
        for element in elements:
            assert(self.tlib.domain[type_name].has_key(element))

        self.tlib.removeDomainElements(type_name, elements)
        self.changed = True

    def has_domain(self, type_name):
        assert(self.tlib.domain.has_key(type_name))
        return len(self.tlib.domain[type_name]) != 0

    # add the given operator to the type library and generate all
    # possible operand combinations
    def add_operator(self, type_name, op_name, arity):
        assert(self.tlib.operators.has_key(type_name))

        self.tlib.addOperator(type_name, op_name, arity)
        self.changed = True
        domain = self.get_domain(type_name)

        if arity == 1:
            for value in domain:
                self.define_operator(type_name, op_name,
                                     [value], [])
        else:
            for lhs in domain:
                for rhs in domain:
                    self.define_operator(type_name, op_name,
                                         [lhs, rhs], [])

    def remove_operator(self, type_name, op_name, arity):
        assert(self.tlib.operators.has_key(type_name))
        assert(self.tlib.operators[type_name].has_key(arity))
        assert(self.tlib.operators[type_name][arity].has_key(op_name))

        self.tlib.removeOperator(type_name, op_name, arity)
        self.changed = True

    def define_operator(self, type_name, op_name, operands, results):
        assert(self.tlib.operators.has_key(type_name))

        assert(len(operands) > 0)
        arity = len(operands)
        assert((arity == 1) or (arity == 2))

        assert(self.tlib.operators[type_name].has_key(arity))
        assert(self.tlib.operators[type_name][arity].has_key(op_name))

        self.tlib.defineOperator(type_name, op_name,
                                 operands, results)
        self.changed = True

    def has_operators(self, type_name):
        assert(self.tlib.operators.has_key(type_name))
        return len(self.tlib.operators[type_name]) != 0

    def add_test(self, type_name, test_name, arity):
        assert(type_name in self.tlib.typenames)
        assert(self.tlib.tests.has_key(type_name))

        self.tlib.addTest(type_name, test_name, arity)
        self.changed = True
        domain = self.get_domain(type_name)
        
        if arity == 1:
            for value in domain:
                self.define_test(type_name, test_name,
                                 [value], [])
        else:
            for lhs in domain:
                for rhs in domain:
                    self.define_test(type_name, test_name,
                                     [lhs, rhs], [])

    def remove_test(self, type_name, test_name, arity):
        assert(self.tlib.tests.has_key(type_name))
        assert(self.tlib.tests[type_name].has_key(arity))
        assert(self.tlib.tests[type_name][arity].has_key(test_name))

        self.tlib.removeTest(type_name, test_name, arity)
        self.changed = True

    def define_test(self, type_name, test_name, operands, results):
        assert(self.tlib.tests.has_key(type_name))

        assert(len(operands) > 0)
        arity = len(operands)
        assert((arity == 1) or (arity == 2))

        assert(self.tlib.tests[type_name].has_key(arity))
        assert(self.tlib.tests[type_name][arity].has_key(test_name))

        self.tlib.defineTest(type_name, test_name,
                             operands, results)
        self.changed = True

    def has_tests(self, type_name):
        assert(self.tlib.tests.has_key(type_name))
        return len(self.tlib.tests[type_name]) != 0

    def add_coercion(self, from_type, to_type):
        assert(from_type in self.tlib.typenames)
        assert(to_type in self.tlib.typenames)

        self.tlib.addCoercion(from_type, to_type)
        try:
            self.tlib.defineCoercion(from_type, to_type,
                                     'Jumbala', [])
            self.changed = True
        except (TypeError, ValueError, Error):
            self.tlib.removeCoercion(from_type, to_type)
            raise
            
    def remove_coercion(self, from_type, to_type):
        assert(self.tlib.coercions.has_key(to_type))
        assert(self.tlib.coercions[to_type].has_key(from_type))

        self.tlib.removeCoercion(from_type, to_type)
        self.changed = True

    def remove_coercion_case_with_index(self, from_type, to_type, backend,
                                        index):
        assert(self.tlib.coercions.has_key(to_type))
        assert(self.tlib.coercions[to_type].has_key(from_type))
        values = self.tlib.coercions[to_type][from_type][backend]
        assert((index >= 0) and (index < len(values)))
        values.pop(index)
        self.changed = True

    def define_coercion(self, from_type, to_type, backend, values):
        assert(self.tlib.coercions.has_key(to_type))
        assert(self.tlib.coercions[to_type].has_key(from_type))
        assert(self.tlib.coercions[to_type][from_type].has_key(backend))

        self.tlib.defineCoercion(from_type, to_type, backend, values)
        self.changed = True

    def has_coercions(self, to_type):
        assert(self.tlib.coercions.has_key(to_type))
        return len(self.tlib.coercions[to_type]) != 0

    # return true if a coercion case can be moved upwards,
    # false otherwise
    def coercion_case_can_raise(self, treeview):
        (tree_model, selection_iter) = treeview.get_selection().get_selected()
        if not selection_iter:
            return False
        path = tree_model.get_path(selection_iter)
        return path[0] > 0

    def raise_coercion_case_with_index(self, from_type, to_type, backend, index):
        assert(self.tlib.coercions.has_key(to_type))
        assert(self.tlib.coercions[to_type].has_key(from_type))
        assert(self.tlib.coercions[to_type][from_type].has_key(backend))

        n_coercions = len(self.tlib.coercions[to_type][from_type][backend])

        assert((index >= 0) and (index < n_coercions))

        if n_coercions <= 1 or index == 0:
            return

        values = self.tlib.coercions[to_type][from_type][backend]
        tmp = values[index]
        values[index] = values[index - 1]
        values[index - 1] = tmp
        self.changed = True

    def coercion_case_can_lower(self, treeview):
        (tree_model, selection_iter) = treeview.get_selection().get_selected()
        if not selection_iter:
            return False
        return tree_model.iter_next(selection_iter) != None

    def lower_coercion_case_with_index(self, from_type, to_type, backend, index):
        assert(self.tlib.coercions.has_key(to_type))
        assert(self.tlib.coercions[to_type].has_key(from_type))
        assert(self.tlib.coercions[to_type][from_type].has_key(backend))

        n_coercions = len(self.tlib.coercions[to_type][from_type][backend])

        assert((index >= 0) and (index < n_coercions))

        if n_coercions <= 1 or index == (n_coercions - 1):
            return

        values = self.tlib.coercions[to_type][from_type][backend]
        tmp = values[index]
        values[index] = values[index + 1]
        values[index + 1] = tmp
        self.changed = True

    # return the arity of the given operator as a list of one or two
    # elements
    def get_operator_arity(self, op_name):
        lst = []
        for i in [1, 2]:
            if op_name in self.op_symbols[i]:
                lst.append(i)
        assert(len(lst) > 0)
        return lst

    def get_test_arity(self, test_name):
        lst = []
        for i in [1, 2]:
            if test_name in self.test_symbols[i]:
                lst.append(i)
        assert(len(lst) > 0)
        return lst

    def get_symbol_arity(self, symbol):
        lst = []

        for i in [1, 2]:
            if symbol in self.op_symbols[i]:
                lst.append(i)

        if len(lst) > 0:
            return lst

        for i in [1, 2]:
            if symbol in self.test_symbols[i]:
                lst.append(i)

        assert(len(lst) > 0)

        return lst

    def get_type_state(self, type_name):
        ret = 0
        n_ops = 0
        n_tests = 0
        n_coercions = 0
        n_defined_ops = 0
        n_defined_tests = 0
        n_defined_coercions = 0

        if not self.get_domain(type_name):
            return self.STATE_UNDEFINED

        ops = self.get_operators(type_name)
        for (op_name, arity) in ops:
            ret = self.get_operator_state(type_name, op_name, arity)
            if ret == self.STATE_PARTIALLY_DEFINED:
                return self.STATE_PARTIALLY_DEFINED
            if ret == self.STATE_COMPLETELY_DEFINED:
                n_defined_ops += 1

        tests = self.get_tests(type_name)
        for (test_name, arity) in tests:
            ret = self.get_test_state(type_name, test_name, arity)
            if ret == self.STATE_PARTIALLY_DEFINED:
                return self.STATE_PARTIALLY_DEFINED
            if ret == self.STATE_COMPLETELY_DEFINED:
                n_defined_tests += 1

        coercions = self.get_coercions(type_name)
        for (from_type, backend) in coercions:
            ret = self.get_coercion_state(from_type, type_name, backend)
            if ret == self.STATE_PARTIALLY_DEFINED:
                return self.STATE_PARTIALLY_DEFINED
            if ret == self.STATE_COMPLETELY_DEFINED:
                n_defined_coercions += 1

        n_ops = len(ops)
        n_tests = len(tests)
        n_coercions = len(coercions)

        if n_defined_ops == n_ops \
           and n_defined_tests == n_tests \
           and n_defined_coercions == n_coercions:
            return self.STATE_COMPLETELY_DEFINED

        return self.STATE_PARTIALLY_DEFINED

    # return the state of an operator
    def get_operator_state(self, type_name, op_name, arity):
        n_defined_ops = 0
        values = self.get_operator_values(type_name, op_name, arity)
        for (operands, results) in values:
            if len(results) > 0:
                n_defined_ops += 1
        if n_defined_ops == 0:
            return self.STATE_UNDEFINED
        if n_defined_ops == len(values):
            return self.STATE_COMPLETELY_DEFINED
        return self.STATE_PARTIALLY_DEFINED

    def get_test_state(self, type_name, test_name, arity):
        n_defined_tests = 0
        values = self.get_test_values(type_name, test_name, arity)
        for (operands, results) in values:
            if len(results) > 0:
                n_defined_tests += 1
        if n_defined_tests == 0:
            return self.STATE_UNDEFINED
        if n_defined_tests == len(values):
            return self.STATE_COMPLETELY_DEFINED
        return self.STATE_PARTIALLY_DEFINED

    def get_coercion_state(self, from_type, to_type, backend):
        n_defined_coercions = 0
        values = self.get_coercion_values(from_type, to_type, backend)
        for (test, results) in values:
            if len(results) > 0:
                n_defined_coercions += 1
        if n_defined_coercions == 0:
            return self.STATE_UNDEFINED
        if n_defined_coercions == len(values):
            return self.STATE_COMPLETELY_DEFINED
        return self.STATE_PARTIALLY_DEFINED

    # return all abstractable operators (not just the defined ones) as
    # a list of tuples (op_name, arity)
    def get_all_abstractable_operator_symbols(self):
        lst = []
        for op in self.all_op_symbols:
            if op in self.non_abstractable_symbols:
                continue
            if op in self.op_symbols[1]:
                lst.append((op, 1))
            if op in self.op_symbols[2]:
                lst.append((op, 2))

        return lst

    def get_all_abstractable_test_symbols(self):
        lst = []
        for test in self.all_test_symbols:
            if test in self.non_abstractable_symbols:
                continue
            if test in self.test_symbols[1]:
                lst.append((test, 1))
            if test in self.test_symbols[2]:
                lst.append((test, 2))

        return lst

    def get_all_abstractable_operation_symbols(self):
        lst = []
        for (symbol, arity) in self.get_all_abstractable_operator_symbols():
                if not symbol in self.non_abstractable_symbols:
                    lst.append((symbol, arity))
        for (symbol, arity) in self.get_all_abstractable_test_symbols():
                if not symbol in self.non_abstractable_symbols:
                    lst.append((symbol, arity))
        lst.sort(lambda x, y: cmp(x[0], y[0]))
        return lst

    # automatically generate/define the given operations (symbols) for
    # the given types
    def generate_operations(self, type_names, symbols, coercions,
                            solver, path_to_solver, model_rte):
        operations_to_generate = []
        for t in type_names:
            for (symbol, arity) in symbols:
                operations_to_generate.append((t, symbol, arity))
        for (from_type, to_type) in coercions:
            operations_to_generate.append \
                (('coercion', from_type, to_type, 'Jumbala'))
        try:
            smtlib_typegen.generateOperations \
                (self.tlib, operations_to_generate,
                 smtlib_typegen.PySmt(solver, path_to_solver),
                 model_rte)
        except Exception:
            raise

        self.changed = True

    # generate a type hierarchy graph; return a path to an image file
    # containing the graph (note, that the caller should remove the
    # created file if it's not needed anymore)
    def generate_type_hierarchy_graph(self, dot_path):
        dot_input = 'digraph G {'
        type_nodes = self.tlib.type_ordering.type_nodes
        edge = None
        for type_name in self.tlib.typenames:
            dot_input += '\n\t' + type_name
            for node in type_nodes[type_name].more_abstract:
                edge = type_name + ' -> ' + node.name
                if not edge in dot_input:
                    dot_input += '\n\t' + edge
        dot_input += '\n}'
        (fd_out, path_in) = mkstemp()
        write(fd_out, dot_input)
        close(fd_out)
        (fd_in, path_out) = mkstemp()
        close(fd_in)
        popen(dot_path + ' -Tpng ' + path_in + ' -o ' + path_out).close()
        remove(path_in)
        return path_out

    def coercion_case_is_valid(self, case):
        self.tlib.parseJumbalaCoercionTestTemplate(case)

    def coercion_from_jumbala_to_python(self, case):
        return actlang_to_python(self.tlib.parseJumbalaCoercionTestTemplate(case))

    def prepare_input_for_dot(self, file):
	file.write('digraph G {')
	type_nodes = self.tlib.type_ordering.type_nodes
	edge = None
	for type_name in self.tlib.typenames:
		if type_name == 'boolean':
			continue
		file.write('\n\t')
		file.write(type_name)
                edges = set()
		for node in type_nodes[type_name].more_abstract:
			if not (type_name, node.name) in edges:
				file.write("\n\t%s -> %s"
					   % (type_name, node.name))
				edges.add((type_name, node.name))
	file.write('\n}')
