# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Type editor GUI.

This module implements the GUI for the abstract type editor.

Authors: Juhani Peltonen
         Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

import pygtk
import gtk, gobject, pango
from abstraction.typelibrary import TypeLibrary
from abstraction.typelibrary import InputFormatError
from abstraction.error import Error
from os import remove, getcwd
from tlib import Tlib
from dialogs import ErrorDialog, SaveFileDialog

def alert():
    dialog = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_INFO,
                               gtk.BUTTONS_CLOSE, 'Here')
    dialog.run()
    dialog.destroy()

class Type_editor:
    def __init__(self, shared_state):
	# constants for rows in the operations column (do not change unless the
	# order of the rows are also changed)
	self.ROW_OPERATORS = 0
	self.ROW_TESTS = 1
	self.ROW_COERCIONS = 2

	self.WIN_DEFAULT_WIDTH = 800
	self.WIN_DEFAULT_HEIGHT = 400
	self.WIN_DEFAULT_TITLE = 'Type editor'

	self.RESULTS_SELECTION_DIALOG_DEFAULT_WIDTH = 200
	self.RESULTS_SELECTION_DIALOG_DEFAULT_HEIGHT = 200

        self.COLOR_RED = '#FF0000'
        self.COLOR_ORANGE = '#F08010'
        self.COLOR_BLACK = '#000000'
        self.COLOR_GRAY = '#808080'

	self.tooltips = gtk.Tooltips()

	self.shared_state = shared_state
	self.tlib = shared_state.tlib

	self.root_dialog = gtk.Window(gtk.WINDOW_TOPLEVEL)
	self.root_dialog.set_modal(True)
	self.root_dialog.set_title(self.WIN_DEFAULT_TITLE)
	self.root_dialog.connect('delete_event', self.delete_event)
	self.root_dialog.set_default_size \
	    (self.WIN_DEFAULT_WIDTH, self.WIN_DEFAULT_HEIGHT)
	vbox = gtk.VBox()
	self.root_dialog.add(vbox)
        self.root_dialog_position = None

        self.show_only_undef_ops_in_generation_dialog = False
        self.generate_operation_in_add_dialog = False

        self.current_typelibrary_directory = getcwd()

	# a table that holds all the the "main" widgets (menubar, scrolled
	# windows, etc.)
	self.table = gtk.Table(2, 4)
	self.table.set_row_spacings(self.shared_state.DEFAULT_PADDING)
	self.table.set_col_spacings(self.shared_state.DEFAULT_PADDING)
	vbox.pack_start(self.table)

	# tree view for available types
	self.ls_types = gtk.ListStore(str)
	cr = gtk.CellRendererText()
	col = gtk.TreeViewColumn('', cr)
	col.set_cell_data_func(cr, self.cdf_types)
        col.add_attribute(cr, 'text', 0)
	self.tv_types = gtk.TreeView(self.ls_types)
	self.tv_types.append_column(col)
	self.tv_types.set_headers_visible(False)

	self.tv_types.connect \
	    ('button-press-event', self.handle_tree_view_events)
	self.tv_types.connect('key-press-event', self.handle_tree_view_events)
	selection = self.tv_types.get_selection()
	selection.connect('changed', self.handle_selection_on_types_tv)

	# tree view for domain
	self.ls_domain = gtk.ListStore(str)
	cr = gtk.CellRendererText()
	col = gtk.TreeViewColumn('', cr)
        col.add_attribute(cr, 'text', 0)
	self.tv_domain = gtk.TreeView(self.ls_domain)
	self.tv_domain.append_column(col)
	self.tv_domain.set_headers_visible(False)

	self.tv_domain.connect \
	    ('button-press-event', self.handle_tree_view_events)
	self.tv_domain.connect('key-press-event', self.handle_tree_view_events)
	selection = self.tv_domain.get_selection()
	selection.connect('changed', self.handle_selection_on_domain_tv)

	# tree view for operations
	self.ts_operations = gtk.TreeStore(str, str, int) # displayed name,
                                                          # name, arity
	self.ts_operations.append(None, ('Operators', '', 0))
	self.ts_operations.append(None, ('Tests', '', 0))
	self.ts_operations.append(None, ('Coercions from', '', 0))
	cr = gtk.CellRendererText()
	col = gtk.TreeViewColumn('', cr)
	col.set_cell_data_func(cr, self.cdf_operations)
        col.add_attribute(cr, 'text', 0)
	self.tv_operations = gtk.TreeView(self.ts_operations)
	self.tv_operations.append_column(col)
	self.tv_operations.set_headers_visible(False)

	self.tv_operations.connect \
	    ('button-press-event', self.handle_tree_view_events)
	self.tv_operations.connect \
	    ('key-press-event', self.handle_tree_view_events)
	selection = self.tv_operations.get_selection()
	selection.connect('changed', self.handle_selection_on_operations_tv)

	# tree view for definition ("values"): depending on the arity of the
	# selected operator or test the view is appended with different
	# columns: with arity of one there's two columns (value (the operand)
	# and results) and with arity of two there's three columns (lhs, rhs
	# (the operands) and results)
	self.ls_values = gtk.ListStore(str, str, str)
	self.tv_values = gtk.TreeView(self.ls_values)
	cr = gtk.CellRendererText()
	self.col_value_lhs = gtk.TreeViewColumn('LHS', cr)
	self.col_value_lhs.set_cell_data_func(cr, self.cdf_operand)
	self.col_value_lhs.add_attribute(cr, 'text', 0)
	self.col_value_lhs.set_resizable(True)
	cr = gtk.CellRendererText()
	self.col_value_rhs = gtk.TreeViewColumn('RHS', cr)
	self.col_value_rhs.set_cell_data_func(cr, self.cdf_operand)
	self.col_value_rhs.add_attribute(cr, 'text', 1)
	self.col_value_rhs.set_resizable(True)
	cr = gtk.CellRendererText()
	self.col_results = gtk.TreeViewColumn('Result values', cr)
	self.col_results.add_attribute(cr, 'text', 2)
	self.col_results.set_resizable(True)
	self.tv_values.set_headers_visible(True)

	self.tv_values.connect \
	    ('button-press-event', self.handle_tree_view_events)
	self.tv_values.connect('key-press-event', self.handle_tree_view_events)
	selection = self.tv_values.get_selection()
	selection.connect('changed', self.handle_selection_on_values_tv)

        self.update_values_tree_view = False

	# types frame
	table = gtk.Table(2, 2)
	table.set_border_width(self.shared_state.DEFAULT_PADDING)
	table.set_row_spacings(self.shared_state.DEFAULT_PADDING)
	table.set_col_spacings(self.shared_state.DEFAULT_PADDING)
	frame = gtk.Frame('Types')
	frame.add(table)
	self.table.attach(frame, 0, 1, 1, 2)
	sw = gtk.ScrolledWindow()
	sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
	sw.add(self.tv_types)
	table.attach(sw, 0, 2, 1, 2)
	b_add = gtk.Button(None, gtk.STOCK_ADD)
	b_add.connect('clicked', self.handle_add_type_click)
	b_remove = gtk.Button(None, gtk.STOCK_REMOVE)
	b_remove.connect('clicked', self.handle_remove_type_click)
	table.attach(b_add, 0, 1, 0, 1, gtk.FILL, gtk.FILL)
	table.attach(b_remove, 1, 2, 0, 1, gtk.FILL, gtk.FILL)
	self.b_remove_type = b_remove

	# domain frame
	table = gtk.Table(2, 2)
	table.set_border_width(self.shared_state.DEFAULT_PADDING)
	table.set_row_spacings(self.shared_state.DEFAULT_PADDING)
	table.set_col_spacings(self.shared_state.DEFAULT_PADDING)
	frame = gtk.Frame('Domain')
	frame.add(table)
	self.table.attach(frame, 1, 2, 1, 2)
	sw = gtk.ScrolledWindow()
	sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
	sw.add(self.tv_domain)
	table.attach(sw, 0, 2, 1, 2)
	b_add = gtk.Button(None, gtk.STOCK_ADD)
	b_add.connect('clicked', self.handle_add_domain_click)
	b_remove = gtk.Button(None, gtk.STOCK_REMOVE)
	b_remove.connect('clicked', self.handle_remove_domain_click)
	table.attach(b_add, 0, 1, 0, 1, gtk.FILL, gtk.FILL)
	table.attach(b_remove, 1, 2, 0, 1, gtk.FILL, gtk.FILL)
	self.b_add_domain = b_add
	self.b_remove_domain = b_remove

	# operations frame
	table = gtk.Table(2, 2)
	table.set_border_width(self.shared_state.DEFAULT_PADDING)
	table.set_row_spacings(self.shared_state.DEFAULT_PADDING)
	table.set_col_spacings(self.shared_state.DEFAULT_PADDING)
	frame = gtk.Frame('Operations')
	frame.add(table)
	self.table.attach(frame, 2, 3, 1, 2)
	sw = gtk.ScrolledWindow()
	sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
	sw.add(self.tv_operations)
	table.attach(sw, 0, 2, 1, 2)
	b_add = gtk.Button(None, gtk.STOCK_ADD)
	b_add.connect('clicked', self.handle_add_operation_click)
	b_remove = gtk.Button(None, gtk.STOCK_REMOVE)
	b_remove.connect('clicked', self.handle_remove_operation_click)
	table.attach(b_add, 0, 1, 0, 1, gtk.FILL, gtk.FILL)
	table.attach(b_remove, 1, 2, 0, 1, gtk.FILL, gtk.FILL)
	self.b_add_operation = b_add
	self.b_remove_operation = b_remove

	# definition ("values") frame
	table = gtk.Table(2, 4)
	table.set_border_width(self.shared_state.DEFAULT_PADDING)
	table.set_row_spacings(self.shared_state.DEFAULT_PADDING)
	table.set_col_spacings(self.shared_state.DEFAULT_PADDING)
	frame = gtk.Frame('Operation definition')
	frame.add(table)
	self.table.attach(frame, 3, 4, 1, 2)
	sw = gtk.ScrolledWindow()
	sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
	sw.add(self.tv_values)
	table.attach(sw, 0, 4, 1, 2)
	b_add = gtk.Button(None, gtk.STOCK_ADD)
	b_add.set_sensitive(False)
	b_remove = gtk.Button(None, gtk.STOCK_REMOVE)
	b_remove.set_sensitive(False)
	b_up = gtk.Button(None, gtk.STOCK_GO_UP)
	b_up.connect('clicked', self.handle_up_click)
	b_up.set_sensitive(False)
	b_down = gtk.Button(None, gtk.STOCK_GO_DOWN)
	b_down.connect('clicked', self.handle_down_click)
	b_down.set_sensitive(False)
	table.attach(b_add, 0, 1, 0, 1, gtk.FILL, gtk.FILL)
	table.attach(b_remove, 1, 2, 0, 1, gtk.FILL, gtk.FILL)
	table.attach(b_up, 2, 3, 0, 1, gtk.FILL, gtk.FILL)
	table.attach(b_down, 3, 4, 0, 1, gtk.FILL, gtk.FILL)
	b_add.connect('clicked', self.handle_add_coercion_case_click)
	b_remove.connect('clicked', self.handle_remove_coercion_case_click)
	self.b_add_coercion_case = b_add
	self.b_remove_coercion_case = b_remove
	self.b_up = b_up
	self.b_down = b_down

	###########################################################
	# types popup menu
	self.types_pmenu_items = {}
	self.types_rmbutton_pmenu = gtk.Menu()
	for str_item in ('Remove', 'Generate operations...'):
            mitem = gtk.MenuItem(str_item)
	    self.types_rmbutton_pmenu.append(mitem)
	    mitem.connect \
		('activate', self.handle_types_rmbutton_pmenu_selection)
	    self.types_pmenu_items[str_item] = mitem
	self.types_rmbutton_pmenu.show_all()

	###########################################################
	# domain popup menu
	self.domain_pmenu_items = {}
	self.domain_rmbutton_pmenu = gtk.Menu()
	for str_item in ('Remove',):
            mitem = gtk.MenuItem(str_item)
	    self.domain_rmbutton_pmenu.append(mitem)
	    mitem.connect \
		('activate', self.handle_domain_rmbutton_pmenu_selection)
	    self.domain_pmenu_items[str_item] = mitem
	self.domain_rmbutton_pmenu.show_all()

	###########################################################
	# operations popup menu
	self.operations_pmenu_items = {}
	self.operations_rmbutton_pmenu = gtk.Menu()
	for str_item in ('Remove', 'Generate definition'):
            mitem = gtk.MenuItem(str_item)
	    self.operations_rmbutton_pmenu.append(mitem)
	    mitem.connect \
		('activate', self.handle_operations_rmbutton_pmenu_selection)
	    self.operations_pmenu_items[str_item] = mitem
	self.operations_rmbutton_pmenu.show_all()

	###########################################################
	# values popup menus
	self.values_pmenu_items = {}
	self.values_rmbutton_pmenu = gtk.Menu()
	for str_item in ('Edit...', 'Remove', 'Move up', 'Move down'):
            mitem = gtk.MenuItem(str_item)
	    self.values_rmbutton_pmenu.append(mitem)
	    mitem.connect \
		('activate', self.handle_values_rmbutton_pmenu_selection)
	    self.values_pmenu_items[str_item] = mitem
	self.values_rmbutton_pmenu.show_all()

	if self.tlib.has_types():
            model = self.tv_types.get_model()
	    for t in self.tlib.get_types():
                model.append((t,))
            self.tv_types.get_selection().select_path((0,))

	self.init_menu()
        self.update_generate_menu_options()
	self.update_buttons()

    def run(self):
        if self.root_dialog_position != None:
            self.root_dialog.move(*self.root_dialog_position)
        self.action_group.get_action('Type hierarchy').get_proxies()[0] \
	    .set_active(self.shared_state.type_hierarchy_window.is_visible())
	self.root_dialog.show_all()
	gtk.main()

    def get_abstraction_type_library(self):
        return self.tlib.tlib

    # initialize the main menu
    def init_menu(self):
	ui = '''<ui>
		  <menubar name="MenuBar">
	            <menu action = "Library">
	              <menuitem action = "Import"/>
	              <menuitem action = "Export"/>
	              <separator/>
	              <menuitem action = "Generate operations"/>
	              <separator/>
	              <menuitem action = "Close"/>
	            </menu>
	            <menu action = "View">
	              <menuitem action = "Type hierarchy"/>
	            </menu>
	          </menubar>
	        </ui>'''

	ui_manager = gtk.UIManager()
	accel_group = ui_manager.get_accel_group()
	self.root_dialog.add_accel_group(accel_group)
	self.action_group = gtk.ActionGroup('menu')
	self.action_group.add_actions \
	    ([('Library', None, '_Library'),
	      ('Import', None, '_Import...', None, None,
	       self.handle_menubar_selection),
	      ('Export', None, '_Export...', None, None,
	       self.handle_menubar_selection),
	      ('Generate operations', None, '_Generate operations...', None,
	       None, self.handle_menubar_selection),
	      ('Close', None, '_Close', '<Ctrl>W', None,
	       self.handle_menubar_selection),
	      ('View', None, '_View')])
	self.action_group.add_toggle_actions \
	    ([('Type hierarchy', None, '_Type hierarchy', None, None,
	       self.handle_menubar_selection)])
	ui_manager.insert_action_group(self.action_group, 0)
	ui_manager.add_ui_from_string(ui)
	self.menubar = ui_manager.get_widget('/MenuBar')
	self.table.attach(self.menubar, 0, 4, 0, 1, gtk.FILL, gtk.FILL)
	self.menubar_funcs = \
	    {'Import': self.import_file,
	     'Export': self.export_file,
	     'Close': self.quit,
	     'Generate operations': self.generate_operations,
	     'Type hierarchy': self.update_type_hierarchy_window}

    def handle_menubar_selection(self, action):
	self.menubar_funcs[action.get_name()]()

    # import a new type library
    def import_file(self):
        if not self.confirm_changes_to_type_library('importing a new one'):
            return

        fc_dialog = gtk.FileChooserDialog \
                        ('Select type library file(s) to import',
                         self.root_dialog,
                         gtk.FILE_CHOOSER_ACTION_OPEN,
                         (gtk.STOCK_OK, gtk.RESPONSE_OK,
                          gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
        fc_dialog.set_default_response(gtk.RESPONSE_OK)
        fc_dialog.set_select_multiple(True)
        fc_dialog.set_current_folder(self.current_typelibrary_directory)
        try:
            resp = fc_dialog.run()
            if resp == gtk.RESPONSE_OK:
                self.current_typelibrary_directory = \
                    fc_dialog.get_current_folder()
                try:
                    self.tlib.read(fc_dialog.get_filenames())
                except (TypeError, ValueError, Error), e:
                    ErrorDialog \
                        (fc_dialog, 'Failed to import type library!', str(e))
                    return
        finally:
            fc_dialog.destroy()

        self.tv_types.get_model().clear()
        self.tv_values.get_model().clear()
        self.tv_domain.get_model().clear()
        self.tv_operations.get_model().clear()

        if self.tlib.has_types():
            model = self.tv_types.get_model()
            for t in self.tlib.get_types():
                model.append((t,))
            self.tv_types.get_selection().select_path((0,))

        self.update_generate_menu_options()
        self.update_buttons()
        self.shared_state.type_hierarchy_image.invalidate()
        self.update_type_hierarchy_window()

    # export the current library
    def export_file(self, parent_window=None):
        if parent_window == None:
            parent_window = self.root_dialog
        dialog = SaveFileDialog \
	             (parent_window,
                      'Select file name for exported type library')
	(filename, directory, filter) = \
                dialog.run(self.current_typelibrary_directory)
	if filename == None:
            return
        self.current_typelibrary_directory = directory
	try:
            self.tlib.write(filename)
	    return True
	except Exception, e:
	    ErrorDialog \
		(self.root_dialog, 'Failed to export type library!', str(e))
	    return False

    # show a dialog which allows the user to export the type library before a
    # destructive operation (or cancel the operation); returns True iff it
    # is safe to continue with the operation
    def confirm_changes_to_type_library(self, msg, root_window=None):
        if not self.tlib.changed:
            return True
	if root_window == None:
            root_window = self.root_dialog
	while True:
            question = gtk.MessageDialog \
		           (root_window,
			    gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
			    gtk.MESSAGE_QUESTION,
			    gtk.BUTTONS_NONE,
			    'Export type library before %s?' % msg)
	    question.format_secondary_text \
		('The current type library has not been exported since the '
		 'latest changes. These changes will be lost if you continue '
		 'with the operation.')
	    question.set_default_response(gtk.RESPONSE_YES)
	    question.add_buttons \
		(gtk.STOCK_YES, gtk.RESPONSE_YES,
		 gtk.STOCK_NO, gtk.RESPONSE_NO,
		 gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
	    question.show_all()
	    resp = question.run()
	    question.destroy()
	    if resp == gtk.RESPONSE_NO:
                return True
	    elif resp == gtk.RESPONSE_CANCEL:
                return False
	    if self.export_file():
	        return True

    # dialog for selecting operations to generate automatically for a type (the
    # given type name (type_name) is the default selection)
    def generate_operations(self, type_name=None):
        # find types which have a coercion defined from int for the Jumbala
        # backend
        types = [t for t in self.tlib.get_types() if self.can_generate(t)]
        if not types:
            ErrorDialog \
                (self.root_dialog,
                 'Operation not available!',
                 "There are no types with a coercion from type `int' defined.")
            return

        if type_name == None:
            type_name = self.tlib.get_selected_type()
        if not type_name in types:
            if len(types) == 1:
                type_name = types[0]
            else:
                type_name = None

        # construct the selection dialog
        dialog = gtk.Dialog \
                     ('Select operations to generate',
                      self.root_dialog,
                      gtk.DIALOG_DESTROY_WITH_PARENT | gtk.DIALOG_MODAL)
        b_ok = dialog.add_button(gtk.STOCK_OK, gtk.RESPONSE_OK)
        dialog.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)

        table = gtk.Table(3, 1)
        table.set_border_width(self.shared_state.DEFAULT_PADDING)
        table.set_row_spacings(self.shared_state.DEFAULT_PADDING)
        table.set_col_spacings(self.shared_state.DEFAULT_PADDING)
        dialog.vbox.pack_start(table)

        hbox = gtk.HBox()

        l = gtk.Label('Type: ')
        l.set_alignment(1.0, 0.5)
        hbox.pack_start(l, False)

        cbox_type = gtk.combo_box_new_text()
        hbox.pack_start(cbox_type)
        table.attach(hbox, 0, 1, 0, 1, gtk.FILL, 0)

        cb_show_undef_ops_only = \
            gtk.CheckButton('Show only undefined operations')
        table.attach(cb_show_undef_ops_only, 0, 1, 1, 2, gtk.FILL, 0)
        cb_show_undef_ops_only.set_active \
            (self.show_only_undef_ops_in_generation_dialog)

        # fill in the type combo box
        active_index = -1
        b_ok.set_sensitive(False)
        for (i, t) in enumerate(types):
            if t == type_name:
                active_index = i
                b_ok.set_sensitive(True)
            cbox_type.append_text(t)
        cbox_type.set_active(active_index)

        # construct the tree view for operations
        paths = []
        ts = gtk.TreeStore(str, str, int, str) # displayed name, name, arity,
                                               # foreground color
        cr = gtk.CellRendererText()
        col = gtk.TreeViewColumn('', cr)
        col.add_attribute(cr, 'text', 0)
        col.add_attribute(cr, 'foreground', 3)
        col.set_cell_data_func(cr, self.cdf_select_results, paths)
        tv = gtk.TreeView(ts)
        tv.set_headers_visible(False)
        tv.append_column(col)
        sel = tv.get_selection()
        sel.set_mode(gtk.SELECTION_MULTIPLE)
        sel.set_select_function(self.sf_results_2, paths)

        sw = gtk.ScrolledWindow()
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        sw.add(tv)
        table.attach(sw, 0, 1, 2, 3)

        if type_name != None:
            self.fill_operations_to_generate \
                (type_name, self.show_only_undef_ops_in_generation_dialog, tv,
                 paths, types)

        # connect signals to objects
        cbox_type.connect \
            ('changed', self.handle_generated_type_selection,
             cb_show_undef_ops_only, tv, paths, types)
        cb_show_undef_ops_only.connect \
            ('toggled', self.handle_toggle_show_undef_ops_cb, cbox_type, tv,
             paths, types)
        cbox_type.connect_object \
            ('changed', lambda x: x.set_sensitive(True), b_ok)

        # show the selection dialog
        dialog.set_default_size(200, 500)
        dialog.show_all()
        resp = dialog.run()
        dialog.destroy()

        if resp != gtk.RESPONSE_OK:
            return

        self.show_only_undef_ops_in_generation_dialog = \
            cb_show_undef_ops_only.get_active()

        # no selected operations, return
        if len(paths) == 0:
            return

        symbols = []
        coercions = []
        type_name = self.cbox_get_active(cbox_type)

        # collect all selected operations
        for path in paths:
            if path[0] in (self.ROW_OPERATORS, self.ROW_TESTS):
                it = ts.get_iter(path)
                symbol = ts.get_value(it, 0)
                arity = 0
                # check if the selected symbol includes its arity
                pos = symbol.find('(1)')
                if pos != -1:
                    # remove ' (1)' (-1 for a white space)
                    symbol = symbol[:(pos-1)]
                    arity = 1
                else:
                    pos = symbol.find('(2)')
                    if pos != -1:
                        # remove ' (2)' (-1 for a white space)
                        symbol = symbol[:(pos-1)]
                        arity = 2
                    else:
                        arity = self.tlib.get_symbol_arity(symbol)
                        assert(len(arity) == 1)
                        arity = arity[0]
                symbols.append((symbol, arity))
            elif path[0] == self.ROW_COERCIONS:
                it = ts.get_iter(path)
                from_type = ts.get_value(it, 0)
                coercions.append((from_type, type_name))
            else:
                assert(False)

            if not self.can_generate_with_error_msg(type_name):
                return
            for (from_type, to_type) in coercions:
                if not self.can_generate_with_error_msg(from_type):
                    return

        # generate the operations
        try:
            self.tlib.generate_operations \
                ([type_name], symbols, coercions, self.shared_state.solver,
                 self.shared_state.solver_path, self.shared_state.model_rte)
        except Exception, e:
            ErrorDialog \
                (self.root_dialog, 'Failed to generate some operation(s)!',
                 str(e))

        # update the domain, operators and values tree views
        self.update_values_tree_view = True
        self.tv_types.get_selection().emit('changed')

    # check whether automatic type generation can be used for the
    # given type
    def can_generate(self, type_name):
        coercions = self.tlib.get_coercions(type_name)
        if not ('int', 'Jumbala') in coercions:
            return False

        values = self.tlib.get_coercion_values('int', type_name, 'Jumbala')
        if len(values) == 0:
            return False

        for (test, results) in values:
            if len(results) == 0:
                return False
        return True

    # check whether automatic type generation can be used for the given type
    # and display an error message if generation cannot be used: return True if
    # generation can be used, False otherwise
    def can_generate_with_error_msg(self, type_name):
        coercions = self.tlib.get_coercions(type_name)
        if not ('int', 'Jumbala') in coercions:
            ErrorDialog \
                (self.root_dialog,
                 'Cannot generate operation automatically!',
                 "To use this function, you must first define a coercion from "
                 "type `int' to type `%s'." % type_name)
            return False

        values = self.tlib.get_coercion_values('int', type_name, 'Jumbala')
        if not values:
            ErrorDialog \
                (self.root_dialog,
                 'Cannot generate operation automatically!',
                 "There are no tests defined in the coercion from type `int' "
                 "to type `%s'." % type_name)
            return False

        for (test, results) in values:
            if not results:
                ErrorDialog \
                    (self.root_dialog,
                     'Cannot generate operation automatically!',
                     "Coercion from type `int' to type `%s' contains a test "
                     "with no admissible return values." % type_name)
                return False

        return True

    # handle type selection from the generated type combo box in the generate
    # operations dialog
    def handle_generated_type_selection(self, cbox_type,
                                        cb_show_undef_ops_only, tree_view,
                                        paths, types_with_int_coercion):
        type_name = self.cbox_get_active(cbox_type)
        show_undef_ops_only = cb_show_undef_ops_only.get_active()
        self.fill_operations_to_generate \
            (type_name, show_undef_ops_only, tree_view, paths,
             types_with_int_coercion)

    # handle toggle of the show undefined operations check button
    def handle_toggle_show_undef_ops_cb(self, cb_show_undef_ops_only,
                                        cbox_type, tree_view, paths,
                                        types_with_int_coercion):
        type_name = self.cbox_get_active(cbox_type)
        show_undef_ops_only = cb_show_undef_ops_only.get_active()
        self.fill_operations_to_generate \
            (type_name, show_undef_ops_only, tree_view, paths,
             types_with_int_coercion)

    # fill the given tree view with operations
    def fill_operations_to_generate(self, type_name, show_undef_ops_only,
                                    tree_view, paths, types_with_int_coercion):
        colors = { self.tlib.STATE_UNDEFINED : self.COLOR_RED,
                   self.tlib.STATE_PARTIALLY_DEFINED : self.COLOR_ORANGE,
                   self.tlib.STATE_COMPLETELY_DEFINED : self.COLOR_BLACK }

        model = tree_view.get_model()
        selected_operations = set()
        while paths:
            iter = model.get_iter(paths.pop())
            if model.get_value(iter, 1) != 'coercion':
                selected_operations.add \
                    ((model.get_value(iter, 1), model.get_value(iter, 2)))
            else:
                selected_operations.add(model.get_value(iter, 0))
        model.clear()

        undefined_operators = self.tlib.get_undefined_operators(type_name)
        undefined_tests = self.tlib.get_undefined_tests(type_name)
        undefined_coercions = \
            frozenset([from_type for (from_type, backend)
                       in self.tlib.get_undefined_coercions(type_name)])

        coercions = []
        if show_undef_ops_only:
            ops = undefined_operators
            tests = undefined_tests
            if type_name in types_with_int_coercion:
                coercions = \
                    [from_type for from_type in undefined_coercions
                     if from_type in types_with_int_coercion]
        else:
            ops = self.tlib.get_all_abstractable_operator_symbols()
            tests = self.tlib.get_all_abstractable_test_symbols()
            if type_name in types_with_int_coercion:
                coercions = \
                    [from_type for from_type in types_with_int_coercion
                     if self.tlib.coercion_can_define \
                         (from_type, type_name, 'Jumbala')]

        for (title, operations, undefined, get_state) in \
                (('Operators', ops, undefined_operators,
                  self.tlib.get_operator_state),
                 ('Tests', tests, undefined_tests,
                  self.tlib.get_test_state)):
            if operations:
                parent = model.append(None, (title, '', 0, self.COLOR_BLACK))
                for op in operations:
                    (symbol, arity) = op
                    displayed_symbol = symbol
                    if len(self.tlib.get_symbol_arity(symbol)) > 1:
                        displayed_symbol += " (%d)" % arity
                    if op in undefined:
                        color = self.COLOR_GRAY
                    else:
                        state = get_state(type_name, symbol, arity)
                        color = colors[state]
                    child = model.append \
                                (parent,
                                 (displayed_symbol, symbol, arity, color))
                    if op in selected_operations:
                        paths.append(model.get_path(child))

        if coercions:
            parent = model.append \
                         (None, ('Coercions from', '', 0, self.COLOR_BLACK))
            for from_type in coercions:
                if from_type in undefined_coercions:
                    color = self.COLOR_GRAY
                else:
                    state = self.tlib.get_coercion_state \
                                (from_type, type_name, 'Jumbala')
                    color = colors[state]
                child = model.append \
                            (parent, (from_type, 'coercion', 0, color))
                if from_type in selected_operations:
                    paths.append(model.get_path(child))

        tree_view.expand_all()

    # quit the type editor
    def quit(self):
        self.root_dialog_position = self.root_dialog.get_position()
        self.root_dialog.hide()
        gtk.main_quit()

    def delete_event(self, widget, event, data=None):
        self.quit()
        return True

    # update sensitivity of menu options relating to automatic operation
    # generation
    def update_generate_menu_options(self, selected_type=None):
        can_generate = False
        if selected_type != None:
            can_generate = self.can_generate(selected_type)
            self.types_pmenu_items['Generate operations...'] \
                .set_sensitive(can_generate)
        if not can_generate:
            for t in self.tlib.get_types():
                if self.can_generate(t):
                    can_generate = True
                    break
        self.action_group.get_action('Generate operations') \
            .get_proxies()[0].set_sensitive(can_generate)

    # update the contents of the type hierarchy window
    def update_type_hierarchy_window(self):
        checkbox = self.action_group.get_action('Type hierarchy') \
                       .get_proxies()[0]
        if checkbox.get_active():
            try:
                self.shared_state.type_hierarchy_image.set \
                    (self.shared_state.dot_path,
                     self.shared_state.tlib.prepare_input_for_dot)
            except IOError, e:
                ErrorDialog(self.root_dialog, "Failed to show image!", str(e))
                self.shared_state.type_hierarchy_window.hide()
                checkbox.set_active(False)
            else:
                self.shared_state.type_hierarchy_window.show()
        else:
            self.shared_state.type_hierarchy_window.hide()

    # update buttons based on what is currently selected (the hide/show calls
    # are for getting aroung a bug in gtk: if the mouse cursor is on a button
    # and the button is set as sensitive, it cannot be clicked before the
    # cursor has re-entered its area)
    def update_buttons(self):
        self.b_remove_type.set_sensitive(False)
        self.b_add_domain.set_sensitive(False)
        self.b_remove_domain.set_sensitive(False)
        self.b_add_operation.set_sensitive(False)
        self.b_remove_operation.set_sensitive(False)
        self.b_add_coercion_case.set_sensitive(False)
        self.b_remove_coercion_case.set_sensitive(False)
        self.b_up.set_sensitive(False)
        self.b_down.set_sensitive(False)

        type_name = self.tlib.get_selected_type()
        if not self.tlib.has_types() or type_name == None:
            return

        self.b_remove_type.set_sensitive(True)
        self.b_remove_type.hide()
        self.b_remove_type.show()

        self.b_add_domain.set_sensitive(True)
        self.b_add_domain.hide()
        self.b_add_domain.show()

        if self.tree_view_has_selection(self.tv_domain):
            self.b_remove_domain.set_sensitive(True)
            self.b_remove_domain.hide()
            self.b_remove_domain.show()

        has_undefined_operations = \
            self.tlib.get_undefined_operators(type_name) \
            or self.tlib.get_undefined_tests(type_name) \
            or self.tlib.get_undefined_coercions(type_name)

        self.b_add_operation.set_sensitive(bool(has_undefined_operations))
        self.b_add_operation.hide()
        self.b_add_operation.show()

        selection_id = self.tlib.get_selection_id()

        if selection_id in (self.tlib.ID_OPERATOR, self.tlib.ID_TEST,
                            self.tlib.ID_COERCION):
            self.b_remove_operation.set_sensitive(True)
            self.b_remove_operation.hide()
            self.b_remove_operation.show()

            if selection_id == self.tlib.ID_COERCION:
                self.b_add_coercion_case.set_sensitive(True)
                self.b_add_coercion_case.hide()
                self.b_add_coercion_case.show()
                if not self.tree_view_has_selection(self.tv_values):
                    return
                self.b_remove_coercion_case.set_sensitive(True)
                self.b_remove_coercion_case.hide()
                self.b_remove_coercion_case.show()

                if self.tlib.coercion_case_can_raise(self.tv_values):
                    self.b_up.set_sensitive(True)
                    self.b_up.hide()
                    self.b_up.show()

                if self.tlib.coercion_case_can_lower(self.tv_values):
                    self.b_down.set_sensitive(True)
                    self.b_down.hide()
                    self.b_down.show()

    # check if the given tree view has a selection
    def tree_view_has_selection(self, tree_view):
        return tree_view.get_selection().get_selected()[1] != None

    # The "Types" frame -------------------------------------------------------

    # cell data function for types
    def cdf_types(self, col, cell, model, it):
        type_name = model.get_value(it, 0)
        state = self.tlib.get_type_state(type_name)
        self.cell_set_appearance(cell, state)

    def handle_selection_on_types_tv(self, selection):
        (model, iter) = selection.get_selected()
        if iter == None:
            type_name = None
        else:
            type_name = model.get_value(iter, 0)

        self.tlib.set_selected_type(type_name)

        m_domain = self.tv_domain.get_model()
        m_domain.clear()
        m_operations = self.tv_operations.get_model()
        m_operations.clear()
        self.tv_values.get_model().clear()
        for i in self.tv_values.get_columns():
            self.tv_values.remove_column(i)

        if type_name != None:
            for element in self.tlib.get_domain(type_name):
                m_domain.append((element,))

            for (title, row_id, get_ops) in \
                    (('Operators', self.ROW_OPERATORS,
                      self.tlib.get_operators),
                     ('Tests', self.ROW_TESTS, self.tlib.get_tests)):
                m_operations.append(None, (title, '', 0))
                it = m_operations.get_iter((row_id,))
                for (symbol, arity) in get_ops(type_name):
                    displayed_symbol = symbol
                    if len(self.tlib.get_symbol_arity(symbol)) > 1:
                        displayed_symbol += ' (%d)' % arity
                    m_operations.append(it, (displayed_symbol, symbol, arity))

            m_operations.append(None, ('Coercions from', '', 0))
            it = m_operations.get_iter((self.ROW_COERCIONS,))
            for (from_type, backend) in self.tlib.get_coercions(type_name):
                m_operations.append(it, (from_type, 'coercion', 0))

            self.tv_operations.expand_all()

            self.types_pmenu_items['Generate operations...'] \
                .set_sensitive(self.can_generate(type_name))

        self.update_buttons()

    # handle selection from the types tree view right mouse button popup menu
    def handle_types_rmbutton_pmenu_selection(self, item):
        sel = item.get_child().get_label()
        if sel == 'Remove':
            self.b_remove_type.emit('clicked')
        elif sel == 'Generate operations...':
            self.generate_operations(self.tlib.get_selected_type())

    # handle click of the add type button
    def handle_add_type_click(self, button):
        dialog = gtk.Dialog \
                     ('New type',
                      self.root_dialog,
                      gtk.DIALOG_DESTROY_WITH_PARENT | gtk.DIALOG_MODAL,
                      (gtk.STOCK_OK, gtk.RESPONSE_OK,
                       gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
        dialog.set_default_response(gtk.RESPONSE_OK)
        table = gtk.Table(4, 2)
        table.set_border_width(self.shared_state.DEFAULT_PADDING)
        table.set_row_spacings(self.shared_state.DEFAULT_PADDING)
        table.set_col_spacings(self.shared_state.DEFAULT_PADDING)
        dialog.vbox.pack_start(table)

        l = gtk.Label('Name: ')
        l.set_alignment(1.0, 0.5)
        table.attach(l, 0, 1, 0, 1, gtk.FILL, 0)
        entry_name = gtk.Entry()
        table.attach(entry_name, 1, 2, 0, 1, gtk.FILL, 0)

        l = gtk.Label()
        l.set_markup('<i>Enter a Jumbala identifier.</i>')
        l.set_alignment(0.0, 0.0)
        l.set_line_wrap(True)
        l.set_width_chars(30)
        l.set_justify(gtk.JUSTIFY_FILL)
        table.attach(l, 1, 2, 1, 2, gtk.FILL, 0)

        l = gtk.Label('Domain elements: ')
        l.set_alignment(1.0, 0.5)
        table.attach(l, 0, 1, 2, 3, gtk.FILL, 0)
        entry_domain = gtk.Entry()
        table.attach(entry_domain, 1, 2, 2, 3, gtk.FILL | gtk.EXPAND, 0)

        l = gtk.Label()
        l.set_markup('<i>Enter zero or more comma-separated values.</i>')
        l.set_alignment(0.0, 0.0)
        l.set_line_wrap(True)
        l.set_justify(gtk.JUSTIFY_FILL)
        l.set_width_chars(30)
        table.attach(l, 1, 2, 3, 4, gtk.FILL | gtk.EXPAND, 0)

        entry_name.connect_object \
            ('activate', lambda entry: entry.grab_focus(), entry_domain)
        entry_domain.set_activates_default(True)

        self.tooltips.set_tip \
            (entry_domain,
             "Enter zero or more comma-separated values, e.g., `VAL_1, VAL_2, "
             "VAL_3'.")

        dialog.show_all()

        # loop until valid values are given or the dialog is closed
        try:
            while True:
                resp = dialog.run()
                if resp != gtk.RESPONSE_OK:
                    return

                type_name = entry_name.get_text()
                str_domain = entry_domain.get_text()

                # get the elements of the given domain
                elems = self.from_csv_to_list(str_domain)
                if len(elems) == 1 and elems[0] == '':
                    elems = []
                try:
                    self.add_type(type_name, elems)
                    break
                except (TypeError, ValueError, Error), e:
                    ErrorDialog \
                        (dialog, 'Error in type specification!', str(e))
        finally:
            dialog.destroy()

    # add a new type to the type library, update the screen and set the
    # selected type to the new type
    def add_type(self, name, domain):
        self.tlib.add_type(name, domain)
        i = self.tlib.get_types().index(name)
        self.tv_types.get_model().insert(i, (name,))
        self.tv_types.get_selection().select_path((i,))
        self.shared_state.type_hierarchy_image.invalidate()
        self.update_type_hierarchy_window()
        self.update_buttons()

    def handle_remove_type_click(self, button):
        type_name = self.tlib.get_selected_type()

        try:
            self.tlib.remove_type(type_name)
        except (TypeError, ValueError, Error), e:
            ErrorDialog \
                (self.root_dialog, 'Failed to remove type!', str(e))
            return

        selection = self.tv_types.get_selection()
        (model, iter) = selection.get_selected()
        if model.remove(iter):
            selection.select_iter(iter)
        elif len(model):
            selection.select_path((len(model) - 1,))
        else:
            selection.unselect_all()
            self.tv_domain.get_model().clear()
            self.tv_operations.get_model().clear()
            self.tv_values.get_model().clear()

        self.update_buttons()
        self.shared_state.type_hierarchy_image.invalidate()
        self.update_type_hierarchy_window()

    # The "Domain" frame ------------------------------------------------------

    def handle_selection_on_domain_tv(self, selection):
        (model, iter) = selection.get_selected()
        if iter == None:
            element = None
        else:
            element = model.get_value(iter, 0)
        self.tlib.set_selected_domain_element(element)
        self.update_buttons()

    # handle selection from the domain tree view right mouse button popup menu
    def handle_domain_rmbutton_pmenu_selection(self, item):
        sel = item.get_child().get_label()
        if sel == 'Remove':
            self.b_remove_domain.emit('clicked')

    # handle click of the add domain (element) button
    def handle_add_domain_click(self, button):
        dialog = gtk.Dialog \
                     ('Add domain element(s)',
                      self.root_dialog,
                      gtk.DIALOG_DESTROY_WITH_PARENT | gtk.DIALOG_MODAL,
                      (gtk.STOCK_OK, gtk.RESPONSE_OK,
                       gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
        dialog.set_default_response(gtk.RESPONSE_OK)
        table = gtk.Table(2, 2)
        table.set_border_width(self.shared_state.DEFAULT_PADDING)
        table.set_row_spacings(self.shared_state.DEFAULT_PADDING)
        table.set_col_spacings(self.shared_state.DEFAULT_PADDING)
        dialog.vbox.pack_start(table)
        l = gtk.Label('Domain elements: ')
        l.set_alignment(1.0, 0.5)
        table.attach(l, 0, 1, 0, 1, gtk.FILL, 0)
        entry_name = gtk.Entry()
        entry_name.set_activates_default(True)
        table.attach(entry_name, 1, 2, 0, 1, gtk.FILL, 0)
        l = gtk.Label()
        l.set_markup('<i>Enter zero or more comma-separated values.</i>')
        l.set_alignment(0.0, 0.0)
        l.set_line_wrap(True)
        l.set_justify(gtk.JUSTIFY_FILL)
        l.set_width_chars(30)
        table.attach(l, 1, 2, 1, 2, gtk.FILL, 0)

        self.tooltips.set_tip \
            (entry_name,
             "Enter zero or more comma-separated values, e.g., `VAL_1, VAL_2, "
             "VAL_3'.")

        type_name = self.tlib.get_selected_type()
        dialog.show_all()

        try:
            while True:
                resp = dialog.run()
                if resp != gtk.RESPONSE_OK:
                    return
                elems = self.from_csv_to_list(entry_name.get_text())
                if len(elems) == 1 and elems[0] == '':
                    elems = []
                try:
                    self.tlib.add_domain_elements(type_name, elems)
                    domain = self.tlib.get_domain(type_name)
                    for element in elems:
                        self.tv_domain.get_model().insert \
                            (domain.index(element), (element,))
                    break
                except (TypeError, ValueError, Error), e:
                    ErrorDialog \
                        (dialog,
                         'Invalid domain element specification!',
                         str(e))
        finally:
            dialog.destroy()

        # make sure that the values tree view will be updated
        self.update_values_tree_view = True
        self.tv_operations.get_selection().emit('changed')

        self.update_buttons()
        self.tv_types.queue_draw()
        self.tv_operations.queue_draw()
        self.tv_values.queue_draw()

    # handle the remove domain button click
    def handle_remove_domain_click(self, button):
        type_name = self.tlib.get_selected_type()
        selected_domain_element = self.tlib.get_selected_domain_element()

        try:
            self.tlib.remove_domain_elements \
                (type_name, (selected_domain_element,))
        except (TypeError, ValueError, Error), e:
            ErrorDialog \
                (self.root_dialog, 'Failed to remove domain element!', str(e))
            return

        selection = self.tv_domain.get_selection()
        (model, iter) = selection.get_selected()
        if model.remove(iter):
            selection.select_iter(iter)
        elif len(model):
            selection.select_path((len(model) - 1,))
        else:
            selection.unselect_all()

        # make sure that the values tree view will be updated
        self.update_values_tree_view = True
        self.tv_operations.get_selection().emit('changed')

        self.update_buttons()
        self.tv_types.queue_draw()
        self.tv_operations.queue_draw()
        self.tv_values.queue_draw()

    # The "Operations" frame --------------------------------------------------

    # cell data function for operations
    def cdf_operations(self, col, cell, model, it):
        type_name = self.tlib.get_selected_type()
        path = model.get_path(it)
        if len(path) == 1:
            # top level items (the title for operators/tests/coercions)
            cell.set_property('foreground', None)
            cell.set_property('style', pango.STYLE_NORMAL)
            cell.set_property('weight', pango.WEIGHT_BOLD)
        else:
            # lower level items
            if path[0] == self.ROW_OPERATORS:
                (symbol, arity) = \
                    (model.get_value(it, 1), model.get_value(it, 2))
                state = self.tlib.get_operator_state(type_name, symbol, arity)
            elif path[0] == self.ROW_TESTS:
                (symbol, arity) = \
                    (model.get_value(it, 1), model.get_value(it, 2))
                state = self.tlib.get_test_state(type_name, symbol, arity)
            elif path[0] == self.ROW_COERCIONS:
                from_type = model.get_value(it, 0)
                state = self.tlib.get_coercion_state \
                    (from_type, type_name, 'Jumbala')
            self.cell_set_appearance(cell, state)

    def handle_selection_on_operations_tv(self, selection):
        (model, iter) = selection.get_selected()
        if iter == None:
            self.tv_values.get_model().clear()
            self.tlib.set_selection_id(self.tlib.ID_TYPE)
        else:
            path = model.get_path(iter)
            if len(path) == 1:
                # top level items
                self.tlib.set_selection_id(self.tlib.ID_TYPE)
                self.tv_values.get_model().clear()
                for item in self.operations_pmenu_items.itervalues():
                    item.set_sensitive(False)
            else:
                # other items
                optype = path[0]
                selected_type = self.tlib.get_selected_type()

                if optype == self.ROW_OPERATORS:
                    (op, data) = \
                        (model.get_value(iter, 1), model.get_value(iter, 2))
                    selection_id = self.tlib.ID_OPERATOR
                    f_get_selected = self.tlib.get_selected_operator
                    f_set_selected = self.set_selected_operator
                elif optype == self.ROW_TESTS:
                    (op, data) = \
                        (model.get_value(iter, 1), model.get_value(iter, 2))
                    selection_id = self.tlib.ID_TEST
                    f_get_selected = self.tlib.get_selected_test
                    f_set_selected = self.set_selected_test
                elif optype == self.ROW_COERCIONS:
                    (op, data) = (model.get_value(iter, 0), 'Jumbala')
                    selection_id = self.tlib.ID_COERCION
                    f_get_selected = self.tlib.get_selected_coercion
                    f_set_selected = self.set_selected_coercion
                else:
                    assert(False)

                # update the value tree view only if the selection differs
                # from the currently active one or if the update was explicitly
                # requested via the `self.update_values_tree_view' variable
                # (needed after editing/generating operation definitions)
                if selection_id != self.tlib.get_selection_id() \
                        or f_get_selected() != (op, data) \
                        or self.update_values_tree_view:
                    f_set_selected(op, data)
                self.update_values_tree_view = False

                # set sensitivity of the 'generate definition' right mouse
                # popup menu item
                can_generate_value = False
                if selection_id in (self.tlib.ID_OPERATOR, self.tlib.ID_TEST):
                    can_generate_value = self.can_generate(selected_type)
                elif selection_id == self.tlib.ID_COERCION:
                    (from_type, backend) = f_get_selected()
                    can_generate_value = \
                        not from_type in self.tlib.get_std_types() \
                        and self.can_generate(selected_type) \
                        and self.can_generate(op)

                self.operations_pmenu_items['Remove'].set_sensitive(True)
                self.operations_pmenu_items['Generate definition'] \
                    .set_sensitive(can_generate_value)

        self.update_buttons()

    # handle selection from the operations tree view right mouse button popup
    # menu
    def handle_operations_rmbutton_pmenu_selection(self, item):
        sel = item.get_child().get_label()
        if sel == 'Remove':
            self.b_remove_operation.emit('clicked')
        elif sel == 'Generate definition':
            selection_id = self.tlib.get_selection_id()
            selected_type = self.tlib.get_selected_type()
            try:
                (model, it) = self.tv_operations.get_selection().get_selected()
                path = model.get_path(it)
                if selection_id == self.tlib.ID_OPERATOR:
                    (op_name, arity) = self.tlib.get_selected_operator()
                    self.generate_definition \
                        (selected_type, op_name, arity, False)
                    self.set_selected_operator(op_name, arity)
                elif selection_id == self.tlib.ID_TEST:
                    (test_name, arity) = self.tlib.get_selected_test()
                    self.generate_definition \
                        (selected_type, test_name, arity, False)
                    self.set_selected_test(test_name, arity)
                elif selection_id == self.tlib.ID_COERCION:
                    (from_type, backend) = self.tlib.get_selected_coercion()
                    self.generate_definition \
                        (selected_type, from_type, selected_type, True)
                    self.set_selected_coercion(from_type, backend)
                self.tv_operations.get_selection().select_path(path)
            except Exception, e:
                ErrorDialog \
                    (self.root_dialog, 'Failed to generate operation!', str(e))

    # handle the add operation button click
    def handle_add_operation_click(self, button):
        type_name = self.tlib.get_selected_type()
        has_undefined_operators = \
            len(self.tlib.get_undefined_operators(type_name)) != 0
        has_undefined_tests = \
            len(self.tlib.get_undefined_tests(type_name)) != 0
        has_undefined_coercions = \
            len(self.tlib.get_undefined_coercions(type_name)) != 0

        if not (has_undefined_operators or has_undefined_tests
                or has_undefined_coercions):
            ErrorDialog \
                ("There are no undefined operators for type `%s'!" % type_name,
                 '')
            return

        dialog = gtk.Dialog \
                     ('New operation', self.root_dialog,
                      gtk.DIALOG_DESTROY_WITH_PARENT | gtk.DIALOG_MODAL)

        b_ok = dialog.add_button(gtk.STOCK_OK, gtk.RESPONSE_OK)
        dialog.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)

        op_table = gtk.Table(5, 2)
        op_table.set_border_width(self.shared_state.DEFAULT_PADDING)
        op_table.set_row_spacings(self.shared_state.DEFAULT_PADDING)
        op_table.set_col_spacings(self.shared_state.DEFAULT_PADDING)
        dialog.vbox.pack_start(op_table)

        id_operator = 0
        id_test = 1
        id_coercion = 2

        l_operation = gtk.Label('Operation: ')
        l_operation.set_alignment(1.0, 0.5)
        op_table.attach(l_operation, 0, 1, 0, 1, gtk.FILL, 0)
        operator_button = gtk.RadioButton(None, 'Operator')
        operator_button.set_sensitive(has_undefined_operators)
        op_table.attach(operator_button, 1, 2, 0, 1, gtk.FILL | gtk.EXPAND, 0)
        test_button = gtk.RadioButton(operator_button, 'Test')
        test_button.set_sensitive(has_undefined_tests)
        op_table.attach(test_button, 1, 2, 1, 2, gtk.FILL | gtk.EXPAND, 0)
        coercion_button = gtk.RadioButton(operator_button, 'Coercion')
        coercion_button.set_sensitive(has_undefined_coercions)
        op_table.attach(coercion_button, 1, 2, 2, 3, gtk.FILL | gtk.EXPAND, 0)

        l_symbol = gtk.Label('Symbol: ')
        l_symbol.set_alignment(1.0, 0.5)
        op_table.attach(l_symbol, 0, 1, 3, 4, gtk.FILL, 0)
        cbox_symbol = gtk.combo_box_new_text()
        op_table.attach(cbox_symbol, 1, 2, 3, 4, gtk.FILL, 0)

        l_from = gtk.Label('From: ')
        l_from.set_alignment(1.0, 0.5)
        op_table.attach(l_from, 0, 1, 3, 4, gtk.FILL, 0)
        cbox_from = gtk.combo_box_new_text()
        op_table.attach(cbox_from, 1, 2, 3, 4, gtk.FILL, 0)
        for (from_type, backend) in \
                self.tlib.get_undefined_coercions(type_name):
            cbox_from.append_text(from_type)

        checkbox_generate = gtk.CheckButton('Generate definition')
        checkbox_generate.set_active(self.generate_operation_in_add_dialog)
        op_table.attach(checkbox_generate, 0, 2, 4, 5, gtk.FILL, 0)

        operator_button.connect \
            ('toggled', self.handle_operation_selection, 'Operator', l_symbol,
             cbox_symbol, l_from, cbox_from, b_ok, checkbox_generate)
        test_button.connect \
            ('toggled', self.handle_operation_selection, 'Test', l_symbol,
             cbox_symbol, l_from, cbox_from, b_ok, checkbox_generate)
        coercion_button.connect \
            ('toggled', self.handle_operation_selection, 'Coercion', l_symbol,
             cbox_symbol, l_from, cbox_from, b_ok, checkbox_generate)
        cbox_symbol.connect \
            ('changed', lambda x: b_ok.set_sensitive(x.get_active() != -1))
        cbox_from.connect \
            ('changed', self.handle_coercion_selection, b_ok,
             checkbox_generate)

        l_operation.show()
        operator_button.show()
        test_button.show()
        coercion_button.show()
        checkbox_generate.show()

        op_table.show()

        # set the default selection to the operation combo box based on the
        # currently selected operation
        selection_id = self.tlib.get_selection_id()
        if selection_id == self.tlib.ID_OPERATOR and has_undefined_operators:
            activate_button = operator_button
        elif selection_id == self.tlib.ID_TEST and has_undefined_tests:
            activate_button = test_button
        elif selection_id == self.tlib.ID_COERCION and has_undefined_coercions:
            activate_button = coercion_button
        else:
            if has_undefined_operators:
                activate_button = operator_button
            elif has_undefined_tests:
                activate_button = test_button
            else:
                activate_button = coercion_button
        activate_button.set_active(True)
        activate_button.emit('toggled')

        dialog.show()
        resp = dialog.run()
        dialog.destroy()

        if resp != gtk.RESPONSE_OK:
            return

        self.generate_operation_in_add_dialog = checkbox_generate.get_active()

        type_name = self.tlib.get_selected_type()

        symbols_to_generate = []
        coercions_to_generate = []

        if coercion_button.get_active():
            from_type = self.cbox_get_active(cbox_from)
            try:
                self.add_coercion(from_type, type_name)
            except (TypeError, ValueError, Error), e:
                ErrorDialog \
                    (self.root_dialog, 'Failed to add coercion!', str(e))
                return
            coercions_to_generate.append((from_type, type_name))
            self.update_generate_menu_options(type_name)
            self.shared_state.type_hierarchy_image.invalidate()
            self.update_type_hierarchy_window()
        else:
            symbol = self.cbox_get_active(cbox_symbol)

            # get the arity of the selected operator/test from the string if
            # the string includes it
            i = symbol.find('(1)')
            if i != -1:
                # remove ' (1)' (-1 for a white space)
                symbol = symbol[:(i-1)]
                arity = 1
            else:
                i = symbol.find('(2)')
                if i != -1:
                    # remove ' (2)' (-1 for a white space)
                    symbol = symbol[:(i-1)]
                    arity = 2
                else:
                    arity = self.tlib.get_symbol_arity(symbol)
                    assert(len(arity) == 1)
                    arity = arity[0]

            if operator_button.get_active():
                try:
                    self.add_operator(type_name, symbol, arity)
                except (TypeError, ValueError, Error), e:
                    ErrorDialog \
                        (self.root_dialog, 'Failed to add operator!', str(e))
                    return
                symbols_to_generate.append((symbol, arity))
            elif test_button.get_active():
                try:
                    self.add_test(type_name, symbol, arity)
                except (TypeError, ValueError, Error), e:
                    ErrorDialog \
                        (self.root_dialog, 'Failed to add test!', str(e))
                    return
                symbols_to_generate.append((symbol, arity))

        self.tv_types.queue_draw()
        if checkbox_generate.get_property('sensitive') \
                and checkbox_generate.get_active():
            try:
                self.tlib.generate_operations \
                    ([type_name],
                     symbols_to_generate,
                     coercions_to_generate,
                     self.shared_state.solver,
                     self.shared_state.solver_path,
                     self.shared_state.model_rte)
            except Exception, e:
                ErrorDialog \
                    (self.root_dialog,
                     'Failed to generate a definition for the operation!',
                     str(e))

        # make sure that the values tree view will be updated
        self.update_values_tree_view = True
        self.tv_operations.get_selection().emit('changed')

    # handle operations selection from the operation combo box in the add
    # operation dialog
    def handle_operation_selection(self, button, op, l_symbol, cbox_symbol,
                                   l_from, cbox_from, b_ok,
                                   checkbox_generate):
        l_symbol.hide()
        cbox_symbol.hide()
        l_from.hide()
        cbox_from.hide()
        cbox_from.set_active(-1)

        type_name = self.tlib.get_selected_type()

        # an operator or a test was selected
        if op in ('Operator', 'Test'):
            cbox_symbol.get_model().clear()

            if op == 'Operator':
                undefined = self.tlib.get_undefined_operators(type_name)
            else:
                undefined = self.tlib.get_undefined_tests(type_name)

            for (symbol, arity) in undefined:
                text = symbol
                # append arity to the displayed text if the symbol has both
                # arities
                if len(self.tlib.get_symbol_arity(symbol)) > 1:
                    text += ' (' + str(arity) + ')'
                cbox_symbol.append_text(text)
            if len(undefined) == 1:
                cbox_symbol.set_active(0)
            else:
                cbox_symbol.set_active(-1)
            b_ok.set_sensitive(cbox_symbol.get_active() != -1)
            l_symbol.show()
            cbox_symbol.show()
            checkbox_generate.set_sensitive(self.can_generate(type_name))
        elif op == 'Coercion':
            l_from.show()
            cbox_from.show()
            if len(cbox_from.get_model()) == 1:
                cbox_from.set_active(0)
            else:
                cbox_from.set_active(-1)
            self.handle_coercion_selection(cbox_from, b_ok, checkbox_generate)

    def handle_coercion_selection(self, cbox_from, b_ok, checkbox_generate):
        if cbox_from.get_active() != -1:
            type_name = self.tlib.get_selected_type()
            from_type = cbox_from.get_active_text()
            checkbox_generate.set_sensitive \
                (self.can_generate(type_name)
                 and not from_type in self.tlib.tlib.STD_TYPES
                 and self.can_generate(from_type))
        else:
            checkbox_generate.set_sensitive(False)
        b_ok.set_sensitive(cbox_from.get_active() != -1)

    def add_operation(self, symbol, arity, row, index):
        displayed_symbol = symbol
        if symbol != 'coercion' and \
                len(self.tlib.get_symbol_arity(symbol)) > 1:
            displayed_symbol += ' (%d)' % arity
        model = self.tv_operations.get_model()
        model.insert \
            (model.get_iter((row,)), index, (displayed_symbol, symbol, arity))
        self.tv_operations.expand_to_path((row, index))
        self.tv_operations.get_selection().select_path((row, index))
        self.update_buttons()

    def add_operator(self, type_name, symbol, arity):
        self.tlib.add_operator(type_name, symbol, arity)
        i = self.tlib.get_operators(type_name).index((symbol, arity))
        self.add_operation(symbol, arity, self.ROW_OPERATORS, i)

    def add_test(self, type_name, symbol, arity):
        self.tlib.add_test(type_name, symbol, arity)
        i = self.tlib.get_tests(type_name).index((symbol, arity))
        self.add_operation(symbol, arity, self.ROW_TESTS, i)

    def add_coercion(self, from_type, to_type):
        self.tlib.add_coercion(from_type, to_type)
        i = self.tlib.get_coercions(to_type).index((from_type, 'Jumbala'))
        model = self.tv_operations.get_model()
        model.insert \
            (model.get_iter((self.ROW_COERCIONS,)), i,
             (from_type, 'coercion', 0))
        self.tv_operations.expand_to_path((self.ROW_COERCIONS, i))
        self.tv_operations.get_selection().select_path((self.ROW_COERCIONS, i))
        self.update_buttons()

    # generate definition (return values) for the given operation
    # (operation, arity_or_backend)
    def generate_definition(self, type_name, op_or_from_type,
                            arity_or_to_type, is_coercion):
        lst0 = []
        lst1 = []
        if is_coercion:
            lst1.append((op_or_from_type, arity_or_to_type))
        else:
            lst0.append((op_or_from_type, arity_or_to_type))
        self.tlib.generate_operations \
            ([type_name], lst0, lst1, self.shared_state.solver,
             self.shared_state.solver_path, self.shared_state.model_rte)
        # make sure that the domain, operations, and values tree views get
        # updated
        self.update_values_tree_view = True
        self.tv_types.get_selection().emit('changed')

    # handle remove operation button click: the operation can be an operator, a
    # test or a coercion
    def handle_remove_operation_click(self, button):
        selection_id = self.tlib.get_selection_id()
        type_name = self.tlib.get_selected_type()

        if selection_id == self.tlib.ID_OPERATOR:
            (op_name, arity) = self.tlib.get_selected_operator()
            try:
                self.tlib.remove_operator(type_name, op_name, arity)
            except (TypeError, ValueError, Error), e:
                ErrorDialog \
                    (self.root_dialog, 'Failed to remove operator!', str(e))
                return
        elif selection_id == self.tlib.ID_TEST:
            (test_name, arity) = self.tlib.get_selected_test()
            try:
                self.tlib.remove_test(type_name, test_name, arity)
            except (TypeError, ValueError, Error), e:
                ErrorDialog \
                    (self.root_dialog, 'Failed to remove test!', str(e))
                return
        elif selection_id == self.tlib.ID_COERCION:
            (from_type, backend) = self.tlib.get_selected_coercion()
            try:
                self.tlib.remove_coercion(from_type, type_name)
            except (TypeError, ValueError, Error), e:
                ErrorDialog \
                    (self.root_dialog, 'Failed to remove coercion!', str(e))
                return
            self.update_generate_menu_options(type_name)
            self.shared_state.type_hierarchy_image.invalidate()
            self.update_type_hierarchy_window()

        selection = self.tv_operations.get_selection()
        (model, iter) = selection.get_selected()
        path = model.get_path(iter)
        if model.remove(iter):
            selection.select_iter(iter)
        elif path[1] > 0:
            selection.select_path((path[0], path[1] - 1))
        else:
            self.tv_values.get_model().clear()
            selection.unselect_all()

        self.update_buttons()
        self.tv_types.queue_draw()

    # The "Operation definition" frame ----------------------------------------

    # cell data function for an operand of an operation (or test of a coercion)
    def cdf_operand(self, col, cell, model, it):
        results = model.get_value(it, 2)
        if not results:
            self.cell_set_appearance(cell, self.tlib.STATE_UNDEFINED)
        else:
            self.cell_set_appearance(cell, self.tlib.STATE_COMPLETELY_DEFINED)

    def handle_selection_on_values_tv(self, selection):
        value_edit = False
        value_remove = False
        value_move_up = False
        value_move_down = False
        (model, it) = selection.get_selected()
        if it == None:
            self.tlib.set_selected_value(None, None)
        else:
            path = model.get_path(it)
            row = path[0]
            selection_id = self.tlib.get_selection_id()
            selected_type = self.tlib.get_selected_type()
            values = None
            value_edit = True
            if selection_id == self.tlib.ID_OPERATOR:
                (op_name, arity) = self.tlib.get_selected_operator()
                values = self.tlib.get_operator_values \
                             (selected_type, op_name, arity)
            elif selection_id == self.tlib.ID_TEST:
                (test_name, arity) = self.tlib.get_selected_test()
                values = self.tlib.get_test_values \
                             (selected_type, test_name, arity)
            elif selection_id == self.tlib.ID_COERCION:
                (from_type, backend) = self.tlib.get_selected_coercion()
                values = self.tlib.get_coercion_values \
                             (from_type, selected_type, backend)
                to_type = self.tlib.get_selected_type()
                (case, results) = values[row]
                if self.tlib.coercion_case_can_raise(self.tv_values):
                    value_move_up = True
                if self.tlib.coercion_case_can_lower(self.tv_values):
                    value_move_down = True
                value_remove = True
            else:
                assert(False)

            self.tlib.set_selected_value(values[row][0], values[row][1])

            mitem = self.values_pmenu_items['Edit...']
            mitem.set_sensitive(value_edit)
            mitem = self.values_pmenu_items['Remove']
            mitem.set_sensitive(value_remove)
            mitem = self.values_pmenu_items['Move up']
            mitem.set_sensitive(value_move_up)
            mitem = self.values_pmenu_items['Move down']
            mitem.set_sensitive(value_move_down)
            self.update_buttons()

    # handle selection from the values tree view right mouse button popup menu
    def handle_values_rmbutton_pmenu_selection(self, item):
        sel = item.get_child().get_label()
        selection_id = self.tlib.get_selection_id()
        if sel == 'Remove':
            self.b_remove_coercion_case.emit('clicked')
        elif sel == 'Edit...':
            if selection_id == self.tlib.ID_OPERATOR:
                self.create_operator_results_selection_dialog()
            elif selection_id == self.tlib.ID_TEST:
                self.create_test_results_selection_dialog()
            elif selection_id == self.tlib.ID_COERCION:
                (from_type, backend) = self.tlib.get_selected_coercion()
                (case, results) = self.tlib.get_selected_value()
                self.edit_coercion_case(case)
        elif sel == 'Move up':
            self.b_up.emit('clicked')
        elif sel == 'Move down':
            self.b_down.emit('clicked')

    # handle the add coercion case button click
    def handle_add_coercion_case_click(self, button):
        self.edit_coercion_case(None)

    # handle the remove coercion case click
    def handle_remove_coercion_case_click(self, button):
        type_name = self.tlib.get_selected_type()
        (from_type, backend) = self.tlib.get_selected_coercion()

        selection = self.tv_values.get_selection()
        (model, iter) = selection.get_selected()
        selected = model.get_path(iter)[0]
        if model.remove(iter):
            selection.select_iter(iter)
        elif len(model):
            selection.select_path((len(model) - 1,))
        else:
            selection.unselect_all()

        self.tlib.remove_coercion_case_with_index \
            (from_type, type_name, backend, selected)

        self.update_generate_menu_options(type_name)

        self.update_buttons()
        self.tv_types.queue_draw()
        self.tv_operations.queue_draw()
        self.tv_values.queue_draw()

    # handle the up button click (the up button raises the selected coercion
    # case one step upwards if it's not the topmost one)
    def handle_up_click(self, button):
        type_name = self.tlib.get_selected_type()
        (from_type, backend) = self.tlib.get_selected_coercion()
        selection = self.tv_values.get_selection()
        (tree_model, tree_iter) = selection.get_selected()
        pos = tree_model.get_path(tree_iter)[0]
        assert(pos > 0)
        self.tlib.raise_coercion_case_with_index \
            (from_type, type_name, backend, pos)
        self.set_selected_coercion(from_type, backend)
        self.tv_values.get_selection().select_path((pos-1,))

    # handle the down button click (the down button lowers the selected
    # coercion case one step downwards if it's not the lowest one)
    def handle_down_click(self, button):
        type_name = self.tlib.get_selected_type()
        (from_type, backend) = self.tlib.get_selected_coercion()
        selection = self.tv_values.get_selection()
        (tree_model, tree_iter) = selection.get_selected()
        pos = tree_model.get_path(tree_iter)[0]
        assert(tree_model.iter_next(tree_iter))
        self.tlib.lower_coercion_case_with_index \
            (from_type, type_name, backend, pos)
        self.set_selected_coercion(from_type, backend)
        self.tv_values.get_selection().select_path((pos+1,))

    # edit the given coercion case (if the given case (old_case) is None, a new
    # case is created)
    def edit_coercion_case(self, old_case):
        type_name = self.tlib.get_selected_type()
        (from_type, backend) = self.tlib.get_selected_coercion()
        is_coercion_from_int = from_type in self.tlib.tlib.STD_TYPES
        if not is_coercion_from_int:
            source_domain = self.tlib.get_domain(from_type)
        target_domain = self.tlib.get_domain(type_name)

        dialog = gtk.Dialog \
                     (old_case and 'Edit coercion case' or 'New coercion case',
                      self.root_dialog,
                      gtk.DIALOG_DESTROY_WITH_PARENT | gtk.DIALOG_MODAL)
        dialog.set_default_size \
            (self.RESULTS_SELECTION_DIALOG_DEFAULT_WIDTH,
             self.RESULTS_SELECTION_DIALOG_DEFAULT_HEIGHT)
        b_ok = dialog.add_button(gtk.STOCK_OK, gtk.RESPONSE_OK)
        b_ok.set_sensitive(True)
        dialog.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
        dialog.set_default_response(gtk.RESPONSE_OK)

        table = gtk.Table(5, 3)
        table.set_border_width(self.shared_state.DEFAULT_PADDING)
        table.set_row_spacings(self.shared_state.DEFAULT_PADDING)
        table.set_col_spacings(self.shared_state.DEFAULT_PADDING)
        dialog.vbox.pack_start(table)

        l = gtk.Label(is_coercion_from_int and 'Test:' or 'Operand:')
        l.set_alignment(1.0, 0.5)
        table.attach(l, 0, 1, 0, 1, gtk.FILL, 0)

        if old_case != None:
            (test, results) = self.tlib.get_selected_value()
        else:
            (test, results) = ('', [])

        if is_coercion_from_int:
            entry_case = gtk.Entry()
            entry_case.set_activates_default(True)
            table.attach(entry_case, 1, 2, 0, 1, gtk.FILL | gtk.EXPAND, 0)
            self.tooltips.set_tip \
                (entry_case,
                 'Enter a Boolean expression template in\nthe Jumbala action '
                 'language. Use <> to\nrefer to the template parameter.')

            def insert_template_parameter(entry):
                entry.delete_selection()
                entry.emit('insert_at_cursor', '<>')
                pos = entry.get_position()
                entry.grab_focus()
                entry.set_position(pos)

            b = gtk.Button('<>')
            b.connect_object('clicked', insert_template_parameter, entry_case)
            table.attach(b, 3, 4, 0, 1, 0, 0)

            l = gtk.Label('')
            l.set_alignment(0, 0)
            l.set_line_wrap(True)
            l.set_width_chars(40)
            l.set_justify(gtk.JUSTIFY_FILL)
            l.set_markup \
                ('<i>Enter a Boolean expression template in the Jumbala '
                 'action language.</i>')
            table.attach(l, 1, 3, 1, 2, gtk.FILL | gtk.EXPAND, 0)
        else:
            test_cbox = gtk.combo_box_new_text()
            test_cbox.append_text("<any>")
            active_selection = -1
            for (i, v) in enumerate(source_domain):
                if test == "<> == <%s>" % v:
                    active_selection = i + 1
                test_cbox.append_text("%s" % v)
            if test == "true":
                active_selection = 0
            test_cbox.set_active(active_selection)
            b_ok.set_sensitive(active_selection != -1)
            test_cbox.connect_object \
                ('changed', lambda x: x.set_sensitive(True), b_ok)
            table.attach(test_cbox, 1, 3, 0, 1, gtk.FILL, 0)
            
        sw = gtk.ScrolledWindow()
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)

        ls = gtk.ListStore(gobject.TYPE_STRING)
        for val in target_domain:
            ls.append((val,))

        if old_case == None:
            if is_coercion_from_int:
                entry_case.set_text('')
            paths = []
        else:
            if is_coercion_from_int:
                entry_case.set_text(old_case)
            paths = [(i,) for (i, val) in enumerate(target_domain)
                     if val in results]

        table.attach(gtk.HSeparator(), 0, 4, 2, 3, gtk.FILL, 0)

        l = gtk.Label('Select the result values in the list below:')
        l.set_alignment(0, 0.5)
        table.attach(l, 0, 4, 3, 4, gtk.FILL, 0)

        cr = gtk.CellRendererText()
        col = gtk.TreeViewColumn('', cr)
        col.add_attribute(cr, 'text', 0)
        col.set_cell_data_func(cr, self.cdf_select_results, paths)
        tv = gtk.TreeView(ls)
        tv.set_headers_visible(False)
        tv.append_column(col)
        sel = tv.get_selection()
        sel.set_mode(gtk.SELECTION_MULTIPLE)
        sel.set_select_function(self.sf_results, paths)
        sw.add(tv)
        table.attach(sw, 0, 4, 4, 5)

        dialog.show_all()

        try:
            while True:
                resp = dialog.run()
                if resp != gtk.RESPONSE_OK:
                    return
                if is_coercion_from_int:
                    case = entry_case.get_text()
                    try:
                        self.tlib.coercion_case_is_valid(case)
                        break
                    except InputFormatError, e:
                        ErrorDialog \
                            (dialog,
                             'Invalid Boolean expression template!',
                             str(e))
                else:
                    selection = test_cbox.get_active()
                    if selection == 0:
                        case = "true"
                    else:
                        case = "<> == <%s>" % source_domain[selection - 1]
                    break
        finally:
            dialog.destroy()

        results = [ls.get_value(ls.get_iter(path), 0) for path in paths]
        dfn = self.tlib.get_coercion_values(from_type, type_name, backend)

        if old_case != None:
            (model, iter) = self.tv_values.get_selection().get_selected()
            selected = model.get_path(iter)[0]
            dfn[selected] = (case, results)
        else:
            dfn.append((case, results))

        try:
            self.tlib.define_coercion(from_type, type_name, backend, dfn)
        except (TypeError, ValueError, Error), e:
            ErrorDialog \
                (self.root_dialog, 'Failed to define coercion!', str(e))
            return

        self.tlib.set_selected_value(case, results)

        if old_case == None:
            m = self.tv_values.get_model()
            m.append((case, "", ", ".join(results)))
            self.tv_values.get_selection().select_path((len(m) - 1,))
        else:
            (m, it) = self.tv_values.get_selection().get_selected()
            m[m.get_path(it)[0]] = ((case, "", ", ".join(results)))

        self.update_generate_menu_options(type_name)
        self.tv_types.queue_draw()
        self.tv_operations.queue_draw()
        self.tv_values.queue_draw()

    # get the active item from the given combo box
    def cbox_get_active(self, cbox):
        model = cbox.get_model()
        i = cbox.get_active()
        if i < 0:
            return None
        return model[i][0]

    def update_values_columns_for_arity(self, arity):
        columns = self.tv_values.get_columns()
        # no need to change columns if there's already a right amount of them
        if len(columns) != (arity + 1):
            if not self.col_value_lhs in columns:
                self.tv_values.insert_column(self.col_value_lhs, 0)
            if arity == 1:
                if self.col_value_rhs in columns:
                    self.tv_values.remove_column(self.col_value_rhs)
            elif arity == 2:
                if not self.col_value_rhs in columns:
                    self.tv_values.insert_column(self.col_value_rhs, 1)
            if not self.col_results in columns:
                self.tv_values.append_column(self.col_results)

    def set_selected_operator(self, op_name, arity):
        self.tlib.set_selected_operator(op_name, arity)
        type_name = self.tlib.get_selected_type()
        values = self.tlib.get_operator_values(type_name, op_name, arity)
        model = self.tv_values.get_model()
        model.clear()
        for (operands, results) in values:
            if arity == 1:
                model.append((operands[0], "", ", ".join(results)))
            else:
                model.append((operands[0], operands[1], ", ".join(results)))
        if arity == 1:
            self.col_value_lhs.set_title('Operand')
        else:
            self.col_value_lhs.set_title('LHS')
        self.update_values_columns_for_arity(arity)

    def set_selected_test(self, test_name, arity):
        self.tlib.set_selected_test(test_name, arity)
        type_name = self.tlib.get_selected_type()
        values = self.tlib.get_test_values(type_name, test_name, arity)
        model = self.tv_values.get_model()
        model.clear()
        for (operands, results) in values:
            text_results = []
            if True in results:
                text_results.append("true")
            if False in results:
                text_results.append("false")
            text_results = ", ".join(text_results)
            if arity == 1:
                model.append((operands[0], "", text_results))
            else:
                model.append((operands[0], operands[1], text_results))
        if arity == 1:
            self.col_value_lhs.set_title('Operand')
        else:
            self.col_value_lhs.set_title('LHS')
        self.update_values_columns_for_arity(arity)

    def set_selected_coercion(self, from_type, backend):
        self.tlib.set_selected_coercion(from_type, backend)
        type_name = self.tlib.get_selected_type()
        model = self.tv_values.get_model()
        model.clear()
        for (test, results) in self.tlib.get_coercion_values \
                                   (from_type, type_name, backend):
            model.append((test, "", ", ".join(results)))
        self.col_value_lhs.set_title('Test')
        self.update_values_columns_for_arity(1)

    # cell data function for toggle like selection
    def cdf_select_results(self, col, cell, model, it, selected_paths):
        path = model.get_path(it)
        if path in selected_paths:
            cell.set_property('cell-background', 'grey')
        else:
            cell.set_property('cell-background', 'white')
        cell.set_property('text', model.get_value(model.get_iter(path), 0))

    # change the appearance of the given cell based on the value of its state
    def cell_set_appearance(self, cell, state):
        cell.set_property('weight', pango.WEIGHT_NORMAL)
        if state == self.tlib.STATE_UNDEFINED:
            cell.set_property('foreground', self.COLOR_RED)
            cell.set_property('style', pango.STYLE_ITALIC)
        elif state == self.tlib.STATE_PARTIALLY_DEFINED:
            cell.set_property('foreground', self.COLOR_ORANGE)
            cell.set_property('style', pango.STYLE_ITALIC)
        elif state == self.tlib.STATE_COMPLETELY_DEFINED:
            cell.set_property('foreground', None)
            cell.set_property('style', pango.STYLE_NORMAL)

    # toggle like selection function for results
    def sf_results(self, path, selected_paths):
        t = (path[0],)
        if t in selected_paths:
            selected_paths.remove(t)
        else:
            selected_paths.append(t)
        return False

    # toggle like selection function for items in a tree view with depth of two
    def sf_results_2(self, path, selected_paths):
        if len(path) != 2:
            return False
        t = (path[0], path[1])
        if t in selected_paths:
            selected_paths.remove(t)
        else:
            selected_paths.append(t)
        return False

    # return a list from the given string of comma separated
    # values and discard leading and trailing white spaces, e.g.,
    # 'a, b ,c, , d f ' -> ['a', 'b', 'c', '', 'd f']
    def from_csv_to_list(self, str_csv):
        return [s.strip() for s in str_csv.split(',')]

    def create_results_selection_dialog(self, valuedomain, selected_paths):
        dialog = gtk.Dialog \
                     ('Select result values',
                      self.root_dialog,
                      gtk.DIALOG_DESTROY_WITH_PARENT | gtk.DIALOG_MODAL,
                      (gtk.STOCK_OK, gtk.RESPONSE_OK,
                       gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
        dialog.set_default_size \
            (self.RESULTS_SELECTION_DIALOG_DEFAULT_WIDTH,
             self.RESULTS_SELECTION_DIALOG_DEFAULT_HEIGHT)
        dialog.set_default_response(gtk.RESPONSE_OK)

        table = gtk.Table(2, 1)
        table.set_row_spacings(self.shared_state.DEFAULT_PADDING)
        table.set_border_width(self.shared_state.DEFAULT_PADDING)
        l = gtk.Label("Select the result values in the list below:")
        table.attach(l, 0, 1, 0, 1, gtk.FILL, 0)

        ls = gtk.ListStore(gobject.TYPE_STRING)
        for element in valuedomain:
            ls.append((element,))
            
        cr = gtk.CellRendererText()
        col = gtk.TreeViewColumn('', cr)
        col.add_attribute(cr, 'text', 0)
        col.set_cell_data_func(cr, self.cdf_select_results, selected_paths)
        tv = gtk.TreeView(ls)
        tv.set_headers_visible(False)
        tv.append_column(col)
        sel = tv.get_selection()
        sel.set_mode(gtk.SELECTION_MULTIPLE)
        sel.set_select_function(self.sf_results, selected_paths)
        sw = gtk.ScrolledWindow()
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        sw.add(tv)
        table.attach(sw, 0, 1, 1, 2)

        dialog.vbox.pack_start(table)
        dialog.show_all()
        resp = dialog.run()
        dialog.destroy()

        return resp

    def create_operator_results_selection_dialog(self):
        type_name = self.tlib.get_selected_type()
        (op_name, arity) = self.tlib.get_selected_operator()
        values = self.tlib.get_operator_values(type_name, op_name, arity)
        valuedomain = self.tlib.get_domain(type_name)
        (operands, results) = self.tlib.get_selected_value()
        paths = [(valuedomain.index(v),) for v in results]
        if self.create_results_selection_dialog(valuedomain, paths) \
                != gtk.RESPONSE_OK:
            return
        results = [valuedomain[path[0]] for path in sorted(paths)]
        try:
            self.tlib.define_operator(type_name, op_name, operands, results)
        except (TypeError, ValueError, Error), e:
            ErrorDialog(self.root_dialog, 'Failed to define operator!', str(e))
        (model, iter) = self.tv_values.get_selection().get_selected()
        model[model.get_path(iter)[0]][2] = ", ".join(results)
        self.tv_types.queue_draw()
        self.tv_operations.queue_draw()
        self.tv_values.queue_draw()

    def create_test_results_selection_dialog(self):
        type_name = self.tlib.get_selected_type()
        (test_name, arity) = self.tlib.get_selected_test()
        values = self.tlib.get_test_values(type_name, test_name, arity)
        valuedomain = [True, False]
        (operands, results) = self.tlib.get_selected_value()
        paths = [(valuedomain.index(v),) for v in results]
        if self.create_results_selection_dialog(('true', 'false'), paths) \
                != gtk.RESPONSE_OK:
            return
        results = [valuedomain[path[0]] for path in sorted(paths)]
        try:
            self.tlib.define_test(type_name, test_name, operands, results)
        except (TypeError, ValueError, Error), e:
            ErrorDialog(self.root_dialog, 'Failed to define test!', str(e))
        text_results = []
        if True in results:
            text_results.append("true")
        if False in results:
            text_results.append("false")
        (model, iter) = self.tv_values.get_selection().get_selected()
        model[model.get_path(iter)[0]][2] = ", ".join(text_results)
        self.tv_types.queue_draw()
        self.tv_operations.queue_draw()
        self.tv_values.queue_draw()

    # handle events on tree views
    def handle_tree_view_events(self, widget, event):
        if widget != self.tv_types \
                and widget != self.tv_domain \
                and widget != self.tv_operations \
                and widget != self.tv_values:
            return True

        # the handle method isn't called if the selection wasn't changed (for
        # instance, when the user selected the same row again) but in the
        # values tree view, return values might have been changed even though
        # the selection remained the same; emit the "changed" signal to update
        # values
        if event.type == gtk.gdk.BUTTON_PRESS and widget == self.tv_values:
            selection = widget.get_selection()
            selection.emit('changed')

        if event.type != gtk.gdk.BUTTON_PRESS \
                and event.type != gtk.gdk._2BUTTON_PRESS \
                and event.type != gtk.gdk.KEY_PRESS:
            return True

        if event.type == gtk.gdk._2BUTTON_PRESS and event.button == 1:
            if widget != self.tv_values:
                return False
            selection_id = self.tlib.get_selection_id()
            if selection_id == self.tlib.ID_OPERATOR:
                self.create_operator_results_selection_dialog()
            elif selection_id == self.tlib.ID_TEST:
                self.create_test_results_selection_dialog()
            elif selection_id == self.tlib.ID_COERCION:
                (case, results) = self.tlib.get_selected_value()
                self.edit_coercion_case(case)

        if event.type != gtk.gdk.BUTTON_PRESS or event.button != 3:
            return False

        path = widget.get_path_at_pos(int(event.x), int(event.y))
        if path == None:
            return True
        path = path[0]

        selection = widget.get_selection()
        selection.select_path(path)

        menu = None
        if widget == self.tv_types:
            menu = self.types_rmbutton_pmenu
        elif widget == self.tv_domain:
            menu = self.domain_rmbutton_pmenu
        elif widget == self.tv_operations:
            menu = self.operations_rmbutton_pmenu
        elif widget == self.tv_values:
            menu = self.values_rmbutton_pmenu
        if menu != None:
            menu.popup(None, None, None, event.button, 0)

        return False
