# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Data structures for model abstractor GUI.

This module defines classes used in the main model abstraction GUI.

Authors: Juhani Peltonen
         Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

import coral
import Utils.Models
import Utils.UML14
import copy
from abstraction.typeinference import TypeInferenceSystem
from abstraction.typelibrary import TypeLibrary
from abstraction.typeinference import AttributeNode
from abstraction.typeinference import ExpressionNode
from abstraction.typeinference import TypeInferenceError
from abstraction.typeinference import TypeConstraint
from abstraction.error import Error
from os import popen, remove, write, close
from tempfile import mkstemp

class Model:
	CONSTRAINT_LE = 'must not be more abstract than'
	CONSTRAINT_EQ = 'must be of type'
	CONSTRAINT_GE = 'must be at least as abstract as'

	def __init__(self):
		self.name = ''
		self.classes = []
		self.signals = []
		self.type_inference_history = []
		self.contradicting_attributes = []
		self.lst_attributes = []

		# a dictionary of attributes: index with a full name
		# of an attribute
		self.dict_attributes = {}
		
		# a dictionary, that holds all classes, signals and
		# attributes of the model (index with a path (a tuple
		# of integers, e.g., a path of(0, 3, 1) refers to the
		# second attribute of the fourth class))
		self.dict_model = {}

		# type inference system nodes
		self.tis_nodes = {}

	def init(self):
		classes = {}
		signals = {}
		coral_classes \
			= Utils.Models.get_elements_of_type(self.coral_model,
							    coral.lang.UML14.Class)
		coral_signals \
			=  Utils.Models.get_elements_of_type(self.coral_model,
							     coral.lang.UML14.Signal)

		# construct the class hierarchy
		for coral_class in coral_classes:
			c_key = str(coral_class.__XMIid__())

			if classes.has_key(c_key):
				continue

			c_name = str(coral_class.name)
			c = Class()
			c.set_name(c_name)
			c.set_key(c_key)
			classes[c_key] = c

			parent = coral_class.__parent__()
			parent_name = str(parent.name)
			parent_key = str(parent.__XMIid__())

			# put the class to its parent and, if the
			# parent doesn't exist, create it and put it
			# to its parent and so forth until an existing
			# parent or the model is found
			while True:
				if parent.oclIsKindOf(coral.lang.UML14.Model):
					# check whether the class
					# being added is already there
					found = False
					for tmp_c in self.classes:
						if tmp_c.get_key() == c.get_key():
							found = True
							break
					if not found:
						self.add_class(c)
					break
				elif parent.oclIsKindOf(coral.lang.UML14.Class):
					if classes.has_key(parent_key):
						p_class = classes[parent_key]
						p_class.add_class(c)
						break
					else:
						tmp_c = Class()
						tmp_c.set_name(parent_name)
						tmp_c.set_key(parent_key)
						tmp_c.add_class(c)
						c = tmp_c
						classes[parent_key] = c
						parent = parent.__parent__()
						parent_name = str(parent.name)
						parent_key = str(parent.__XMIid__())
				else:
					assert(False)

		# all signals go the model
		for coral_signal in coral_signals:
			s_name = str(coral_signal.name)
			s_key = str(coral_signal.__XMIid__())

			s = Signal()
			s.set_name(s_name)

			if not signals.has_key(s_key):
				signals[s_key] = s
				self.add_signal(s)

		# put attributes to classes and signals
		for node in self.tis.findAbstractableNodes():
			parent = node.coral_element.__parent__()
			parent_key = str(parent.__XMIid__())

			attr = Attribute()
			attr.name = str(node.coral_element.name)
			attr.full_name \
				= self.get_full_name_of_coral_element(node.coral_element)
			attr.original_type = str(node.model_type)
			attr.abstract_type = attr.original_type
			attr.inferred_type = None

			if parent.oclIsKindOf(coral.lang.UML14.Signal):
				sig = signals[parent_key]
				sig.add_attribute(attr)
			elif parent.oclIsKindOf(coral.lang.UML14.Class):
				c = classes[parent_key]
				c.add_attribute(attr)
			else:
				assert(False)

			self.tis_nodes[attr.full_name] = node

		self.sort()
		self.collect_attributes()
		self.construct_dict()

	# construct the model dictionary (attributes in the dictionary
	# are accessed with a path)
	def construct_dict(self):
		path = []
		path.append(0)
		self.dict_model[tuple(path)] = None
		path.append(0)
		for c in self.classes:
			self.dict_model[tuple(path)] = c
			self.construct_class_dict(c, path)
			path[-1] += 1
		path.pop()
		path[-1] += 1
		self.dict_model[tuple(path)] = None
		path.append(0)
		for s in self.signals:
			self.dict_model[tuple(path)] = s
			path.append(0)
			for attr in s.attributes:
				self.dict_model[tuple(path)] = attr
				path[-1] += 1
			path.pop()
			path[-1] += 1

	def construct_class_dict(self, cla, parent_path):
		path = parent_path
		path.append(0)
		for attr in cla.attributes:
			self.dict_model[tuple(path)] = attr
			path[-1] += 1
		for c in cla.classes:
			self.dict_model[tuple(path)] = c
			self.construct_class_dict(c, path)
			path[-1] += 1
		path.pop()

	def add_class(self, cla):
		assert(cla not in self.classes)
		for c in self.classes:
			assert(c.get_name() != cla.get_name())
		self.classes.append(cla)

	def add_signal(self, sig):
		assert(sig not in self.signals)
		for s in self.signals:
			assert(s.get_name() != sig.get_name())
		self.signals.append(sig)

	def get_classes(self):
		return self.classes

	def get_signals(self):
		return self.signals

	def get_attributes(self):
		return self.lst_attributes

	# construct a list and a dictionary from all class and signal
	# attributes
	def collect_attributes(self):
		self.lst_attributes = []
		self.dict_attributes = {}
		for c in self.classes:
			self.collect_class_attributes(self.lst_attributes, c)
		for s in self.signals:
			for attr in s.attributes:
				self.lst_attributes.append(attr)
		self.lst_attributes.sort(lambda x, y: cmp(x.full_name.lower(),
							  y.full_name.lower()))
		for attr in self.lst_attributes:
			self.dict_attributes[attr.full_name] = attr

	# recursively collect all class attributes to the given list
	def collect_class_attributes(self, lst, cla):
		for attr in cla.attributes:
			lst.append(attr)
		for c in cla.classes:
			self.collect_class_attributes(lst, c)

	# load a model from the given file
	def load(self, filename):
		try:
			coral_project = coral.loadDocument(filename)
		except Exception, e:
			raise Error, e

		found = False
		for coral_model in coral_project.contents():
			if coral_model.oclIsKindOf(coral.lang.UML14.Model):
				found = True
				break

		if not found:
			raise Error, 'no model found'

		self.tis = TypeInferenceSystem(coral_model)
		self.name = str(coral_model.name)
		self.coral_model = coral_model
		self.coral_project = coral_project

		self.init()
		self.is_consistent()
		self.set_inferred_types()
		# set inferred types a second time to clear the "changed"
		# status
		self.set_inferred_types()

	def can_save(self):
		try:
			self.is_consistent()
			self.tis.rewriteModel()
		except Exception, e:
			raise Error, e

	def save(self, filename, type_library, tlib_filenames):
		self.is_consistent()
		self.tis.rewriteModel()
		self.tis.attachTypesToModel(tlib_filenames)
		self.coral_project.setFilename(filename)
		coral.saveDocument(self.coral_project)

	# get an item with a path (the path is a tuple of integers);
	# path[0] == 0 refers to classes and path[0] == 1 to signals
	def get_with_path(self, path):
		return self.dict_model[path]

	# record the history of type inference (callback function for
	# solveConstraints)
	def solve_constraints_callback(self, c):
		if isinstance(c, TypeInferenceSystem.MakeMoreAbstract):
			self.type_inference_history.append(c)

	# perform type inference
	def infer_types(self, default_type):
		self.set_constraints()
		self.type_inference_history = []
		self.contradicting_attributes = []

		try:
			self.tis.solveConstraints(default_type,
						  self.solve_constraints_callback)
		except TypeInferenceError, e:
			if isinstance(e.failure,
				      TypeInferenceSystem.DetectIncoercibility) \
				or isinstance(e.failure,
					      TypeInferenceSystem.DetectAttemptToChangeConstant):
				self.type_inference_history.reverse()
				conflicting_nodes = []
				self.solve_contradicting_attributes(0,
								    e.failure.constraint,
								    self.contradicting_attributes)
			raise Error, e
		self.set_inferred_types()

	# solve contradicting attributes based on type inference
	# history (start the search at history_index)
	def solve_contradicting_attributes(self, history_index, constraint,
					   contradicting_attributes):
		if history_index >= len(self.type_inference_history):
			return
		for i in range(history_index, len(self.type_inference_history)):
			event = self.type_inference_history[i]
			if event.node != constraint.constrainee \
			       and event.node != constraint.constrainer:
				continue
			if event.constraint.is_external:
				full_name = self.get_full_name_of_coral_element(event.node.coral_element)
				attr = self.get_attribute_with_full_name(full_name)
				if not attr in contradicting_attributes:
					contradicting_attributes.append(attr)
			self.solve_contradicting_attributes(i + 1,
							    event.constraint,
							    contradicting_attributes)

	# check if the model is consistent but don't set inferred
	# types
	def is_consistent(self):
		self.set_constraints()
		self.tis.solveConstraints()

	# get the list of contradicting attributes (those attributes
	# whose constraints introduced contradictions)
	def get_contradicting_attributes(self):
		return self.contradicting_attributes

	# set inferred types and changed flags of classes and signals
	def set_inferred_types(self):
		for c in self.classes:
			self.set_class_inferred_types(c)
		for s in self.signals:
			self.set_signal_inferred_types(s)

	# set inferred types of the attributes of the given class and
	# its subclasses and set the changed flag if necessary
	def set_class_inferred_types(self, cla):
		cla.changed = False
		for attr in cla.attributes:
			tis_node = self.tis_nodes[attr.full_name]
			if attr.inferred_type != tis_node.inferred_type:
				cla.changed = True
				attr.changed = True
				attr.inferred_type = tis_node.inferred_type
			else:
				attr.changed = False

		for c in cla.classes:
			ret = self.set_class_inferred_types(c)
			if c.changed == False:
				c.changed = ret
			if cla.changed == False:
				cla.changed = ret

		return cla.changed

	# set inferred types of the attributes of the given signal
	def set_signal_inferred_types(self, sig):
		sig.changed = False
		for attr in sig.attributes:
			tis_node = self.tis_nodes[attr.full_name]
			if attr.inferred_type != tis_node.inferred_type:
				sig.changed = True
				attr.changed = True
				attr.inferred_type = tis_node.inferred_type
			else:
				attr.changed = False

	# set constraints of classes and signals
	def set_constraints(self):
		for c in self.classes:
			self.set_class_constraints(c)
		for s in self.signals:
			self.set_signal_constraints(s)

	def set_constraint(self, tis_node, constraint, constraining_type):
		operator = None
		if constraint == self.CONSTRAINT_LE:
			operator = TypeConstraint.CMP_LE
		elif constraint == self.CONSTRAINT_EQ:
			operator = TypeConstraint.CMP_EQ
		elif constraint == self.CONSTRAINT_GE:
			operator = TypeConstraint.CMP_GE
		else:
			return
		self.tis.addExternalConstraint(tis_node,
					       operator,
					       constraining_type)

	# set constraints of the attributes of the given class and
	# its subclasses
	def set_class_constraints(self, cla):
		for attr in cla.attributes:
			tis_node = self.tis_nodes[attr.full_name]
			self.tis.removeExternalConstraint(tis_node)
			for (constraint, constraining_type) in attr.constraints:
				self.set_constraint(tis_node,
						    constraint,
						    constraining_type)

		for c in cla.classes:
			self.set_class_constraints(c)

	def set_signal_constraints(self, sig):
		for attr in sig.attributes:
			tis_node = self.tis_nodes[attr.full_name]
			self.tis.removeExternalConstraint(tis_node)
			for (constraint, constraining_type) in attr.constraints:
				self.set_constraint(tis_node,
						    constraint,
						    constraining_type)

	def compare(self, other):
		if len(self.classes) != other.classes:
			return False
		for i in range(0, len(self.classes)):
			c0 = self.classes[i]
			c1 = other.classes[i]
			if c0.compare(c1) == False:
				return False
		return True

	# attach the current type library
	def attach_type_library(self, type_library):
		self.tis.attachTypeLibrary(type_library)
		# remove constraints that have a constraining type
		# that doesn't belong to the type library anymore
		for attr in self.lst_attributes:
			for (c, t) in attr.constraints[:]:
				if not t in type_library.typenames:
					attr.remove_constraint(c, t)


	def has_classes(self):
		return len(self.classes) != 0

	def has_signals(self):
		return len(self.signals) != 0

	def sort(self):
		self.sort_classes(self.classes)
		self.signals.sort(lambda x, y: cmp(x.name.lower(),
						   y.name.lower()))
		for s in self.signals:
			s.attributes.sort(lambda x, y: cmp(x.name.lower(),
							   y.name.lower()))

	def sort_classes(self, classes):
		classes.sort(lambda x, y: cmp(x.name.lower(),
					      y.name.lower()))
		for c in classes:
			c.attributes.sort(lambda x, y: cmp(x.name.lower(),
							   y.name.lower()))
			self.sort_classes(c.classes)

	def get_constraint_operators(self):
		lst = []
		lst.append(self.CONSTRAINT_LE)
		lst.append(self.CONSTRAINT_EQ)
		lst.append(self.CONSTRAINT_GE)
		return lst

	def get_attribute_with_full_name(self, name):
		assert(self.dict_attributes.has_key(name))
		return self.dict_attributes[name]

	def prepare_input_for_dot(self, file):
		file.write('digraph G {')
		for node in self.tis.findAbstractableNodes():
			if not isinstance(node, AttributeNode):
				continue
			from_node = self.get_full_name_of_coral_element(node.coral_element)

			# uncomment if all nodes should be shown in
			# the graph instead of just those ones whose
			# value depends on another node
			#dot_input += '\n\t\"' + from_node + '\"'

			descendants = node.findReachableNodes(filt=lambda n: isinstance(n, AttributeNode), prune=lambda n: n != node and isinstance(n, AttributeNode))
			for r_node in descendants:
				to_node = self.get_full_name_of_coral_element(r_node.coral_element)
				file.write('"%s" -> "%s"\n\t' % (from_node,
								 to_node))
		file.write('\n}')

	# get the (modified) full name of the given coral element
	# (coral_element)
	def get_full_name_of_coral_element(self, coral_element):
		full_name = Utils.Models.get_qualified_name(coral_element)

		# remove the name of the model from the full_name
		i = full_name.find('::')
		full_name = full_name[(i + 2):]

		# remove unneccessary stuff from the name in case the
		# parent is a signal (leave only the name of the
		# signal and the attribute)
		if coral_element.__parent__().oclIsKindOf(coral.lang.UML14.Signal):
			i = full_name.rfind('::')
			j = full_name[:(i - 2)].rfind('::')
			if j != -1:
				full_name = full_name[(j + 2):]

		return full_name

class Class:
	def __init__(self):
		self.name = ''
		self.key = ''
		self.classes = []
		self.attributes = []
		self.changed = False

	def get_name(self):
		return self.name

	def set_name(self, name):
		self.name = name

	def set_key(self, key):
		self.key = key

	def get_key(self):
		return self.key

	def add_class(self, new_class):
		self.classes.append(new_class)

	def add_attribute(self, attribute):
		self.attributes.append(attribute)

	def get_attributes(self):
		return self.attributes

	def compare(self, other):
		if self.name != other.name \
		       or self.changed != other.changed \
		       or len(self.classes) != len(other.classes) \
		       or len(self.attributes) != len(other.attributes):
			return False
		for attr0, attr1 in zip(self.attributes, other.attributes):
			if not attr0.compare(attr1):
				return False
		for c0, c1 in zip(self.classes, other.classes):
			if not c0.compare(c1):
				return False
		return True

class Signal:
	def __init__(self):
		self.name = ''
		self.key = ''
		self.attributes = []
		self.changed = False

	def set_name(self, name):
		self.name = name

	def get_name(self):
		return self.name

	def set_key(self, key):
		self.key = key

	def get_key(self):
		return self.key

	def add_attribute(self, attribute):
		self.attributes.append(attribute)

	def get_attributes(self):
		return self.attributes

	def compare(self, other):
		if (self.name != other.name) \
		   or (self.changed != other.changed) \
		   or len(self.attributes) != len(other.attributes):
			return False
		for (attr0, attr1) in zip(other.attributes, self.attributes):
			if not attr0.compare(attr1):
				return False
		return True


class Attribute:
	def __init__(self):
		self.name = ''
		self.full_name = ''
		self.original_type = ''
		self.abstract_type = ''
		self.inferred_type = ''
		self.constraints = []

		# store whether the attribute was changed
		self.changed = False

	def get_name(self):
		return self.name

	def set_name(self, name):
		self.name = name

	def get_full_name(self):
		return self.full_name

	def get_original_type(self):
		return self.original_type
	def get_abstract_type(self):
		return self.abstract_type

	def get_inferred_type(self):
		return self.inferred_type

	def is_constrained(self):
		return len(self.constraints) != 0

	def add_constraint(self, constraint, type_name):
		if self.has_constraint(constraint, type_name):
			raise Error, 'Constraint already defined!'
		self.constraints.append((constraint, type_name))

	def change_constraint(self,
			      old_constraint, old_type,
			      new_constraint, new_type):
		for i in range(0, len(self.constraints)):
			(constraint, type_name) = self.constraints[i]
			if constraint == old_constraint \
			       and type_name == old_type:
				self.constraints[i] = (new_constraint,
						       new_type)
				return

	def remove_constraint(self, constraint, type_name):
		if (constraint, type_name) in self.constraints:
			self.constraints.remove((constraint, type_name))

	def has_constraint(self, constraint, type_name):
		if (constraint, type_name) in self.constraints:
			return True
		return False

	def get_constraints(self):
		return self.constraints

	def get_constraint_with_index(self, index):
		assert((index >= 0) and (index < len(self.constraints)))
		return self.constraints[index]

	def compare(self, other):
		if self.name != other.name \
		       or self.full_name != other.full_name \
		       or self.original_type != other.original_type \
		       or self.abstract_type != other.abstract_type \
		       or self.inferred_type != other.inferred_type \
		       or len(self.constraints) != len(other.constraints):
			return False

		for c0, c1 in zip(self.constraints, other.constraints):
			if c0[0] != c1[0] or c0[1] != c1[1]:
			    return False

		return True
