#! /usr/bin/python

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Usage: jumbala_to_uml.py [OPTION...] INPUT-FILE OUTPUT-FILE
Construct a UML 1.4 model from one or more simple sequential Jumbala
programs representing concurrent processes communicating via shared
variables.

The following command line option is supported:
  -h, --help                 Display this help and exit.

The input for the script should consist of definitions of process types
and processes.  A process type represents a simple sequential Jumbala
program with variables (each of which can be of type "int", "boolean", or
a reference to another process type); each process type will be
represented by a unique active class in the constructed model.  A process
is an instantiation of a process type and corresponds to an object in the
constructed model.  There can be multiple instantiations of the same
process type.

In short, each process type definition should begin with a line of the
form "#proctype <PROCTYPE-ID>" (where <PROCTYPE-ID> is a valid name for
a non-primitive type in the Jumbala action language) followed by a
newline-separated list of variable declarations (in the usual form,
e.g., "int x;") and a fragment of Jumbala action language code, which may
refer to the variables declared in the process type definition, but may
not contain any further variable, class, or function declarations.  The
type of a variable can be either "int", "boolean", or the name of a
process type defined in the same input file.  The name of a variable
should be a valid identifier in the Jumbala action language.

Each process definition should begin with a line of the form
"#process <PROCESS-ID> <PROCTYPE-ID>", where <PROCESS-ID> is a valid name
for an identifier in the Jumbala action language, and <PROCTYPE-ID> is a
name of a process type defined in the same input file.  If the process
type of the process declares references to other process types, all such
references have to be initialized for the process by following the
"#process ..." line with a newline-separated list of assignments of the
form "<REF-ID> = <PROCESS-ID>;", where <REF-ID> is the name of a
reference to a process type declared in <PROCTYPE-ID>, and <PROCESS-ID>
is the name of a process instatiated using this process type in the same
input file.  Primitive type variables of processes can (but do not have
to) be initialized (with a constant value of the appropriate type) using
a similar syntax "<VARIABLE-ID> = <VALUE>;".

Formally, the input for the script should conform to a string generable
from the nonterminal <definitions> using the following grammar.
(Nonterminals in the grammar should be separated from other sequences
of non-whitespace characters by white space.  Strings in quotes should
be interpreted literally without the quotes; <EOL> represents a newline.
Comments are enclosed in /* ... */.)

<definitions> ::= <definition> <definitions> | /* empty */

<definition> ::= <proctype-definition> | <process-definition>

<proctype-definition> ::= <proctype-declaration> <var-declarations>
                          <ACTION-LANGUAGE-CODE>

<proctype-declaration> ::= "#proctype" <PROCTYPE-ID> <EOL>

<var-declarations> ::= <var-declaration> <var-declarations> | /* empty */

<var-declaration> ::= ("int" | "boolean" | <PROCTYPE-ID>) <VARIABLE-ID> ";"
                      <EOL>

<process-definition> ::= <process-declaration> <initializations>

<process-declaration> ::= "#process" <PROCESS-ID> <PROCTYPE-ID> <EOL>

<initializations> ::= <initialization> <initializations>
                    | /* empty */

<initialization> ::= <var-initialization> | <ref-initialization>

<var-initialization> ::= <VARIABLE-ID> "=" <VALUE> ";" <EOL>

<ref-initialization> ::= <REF-ID> "=" <PROCESS-ID> ";" <EOL>

Note that this simple helper script will not check <ACTION-LANGUAGE-CODE>
for any errors.  Additionally, the generated model may need to be
processed through the action language flattener included in the SMUML
toolset before it is ready to be analyzed.

This script was written by Heikki Tauriainen for the SMUML project
<http://www.tcs.hut.fi/Research/Logic/SMUML.shtml>.
"""
#"

import sys
import re
import coral
import getopt

# -----------------------------------------------------------------------------
# Classes and helper functions

class ProcType:
    """Class for process types."""

    def __init__(self):
        self.vars = {}
        self.code = []
        self.cls = None
        self.attributes = {}
        self.associations = {}

class Process:
    """Class for process instances."""

    def __init__(self, proctype):
        self.proctype = proctype
        self.initialvalues = {}
        self.object = None

def abort(msg):
    """Abort with an error message.

    Argument:
        msg -- The error message.
    """
    sys.stderr.write("%s:%s\n" % (sys.argv[0], msg))
    sys.exit(1)

def read_next_line(input_file):
    """Read next non-empty line of input.

    Argument:
        input_file -- An file object opened for reading.

    Returns the next non-empty line of input with leading and trailing white
    space removed, or None on EOF.
    """
    while True:
        line = input_file.readline()
        if not line:
            return None
        line = line.strip()
        if line:
            return line

def get_default_attribute_multiplicity():
    """Helper function for constructing UML multiplicities."""
    return UML.Multiplicity(range=[UML.MultiplicityRange(lower=1, upper=1)])

def get_default_association_end_multiplicity():
    """Helper function for constructing UML multiplicities."""
    return UML.Multiplicity(range=[UML.MultiplicityRange(lower=0, upper=1)])

# -----------------------------------------------------------------------------
# The script

# some constants

UML = coral.lang.UML14
XMIDI = coral.lang.XMIDI20

IDENTIFIER_RE = "[a-zA-Z_][a-zA-Z_0-9]*"
PROCTYPE_RE = re.compile("^\s*#\s*proctype\s+(" + IDENTIFIER_RE + ")\s*$")
PROCESS_RE = \
    re.compile \
        ("^\s*#\s*process\s+(" + IDENTIFIER_RE + ")\s+("
         + IDENTIFIER_RE + ")\s*$")
VARDECL_RE = \
    re.compile \
        ("^\s*(int|boolean|" + IDENTIFIER_RE + ")"
         "\s+(" + IDENTIFIER_RE + ")\s*;\s*$")
VARDEF_RE = re.compile("^\s*(" + IDENTIFIER_RE + ")\s*=\s*(.*)\s*;\s*$")

# process command line options

try:
    options, args = getopt.getopt(sys.argv[1:], "h", ("help",))
except getopt.GetoptError, e:
    abort(str(e))

for (option, value) in options:
    if option in ("-h", "--help"):
        sys.stdout.write(__doc__)
        sys.exit(0)

if not args:
    abort("input file name expected")
if len(args) < 2:
    abort("output file name expected")
elif len(args) > 2:
    abort("extra arguments on command line")

input_file = open(args[0], "r")

# parse the input file

proctypes = {}
processes = {}
line = read_next_line(input_file)
while line != None:
    m = PROCTYPE_RE.match(line)
    if m:
        # "proctype" definition
        name = m.group(1)
        if name in ("int", "boolean"):
            abort("type name `%s' is reserved" % name)
        if name in proctypes:
            abort("duplicate definition of process type `%s'" % name)
        p = ProcType()
        proctypes[name] = p
        # collect variable declarations from input
        line = read_next_line(input_file)
        while line != None:
            m = VARDECL_RE.match(line)
            if m:
                p.vars[m.group(2)] = m.group(1)
                line = read_next_line(input_file)
            else:
                break
        # read in the rest of the code
        while line != None:
            if PROCTYPE_RE.match(line) or PROCESS_RE.match(line):
                if not p.code:
                    # if the process has no code associated with it, use a
                    # single empty statement as the code so that the model
                    # works with the simulator
                    p.code.append(";")
                break
            p.code.append(line)
            line = read_next_line(input_file)
        continue
    m = PROCESS_RE.match(line)
    if m:
        # "process" definition
        name = m.group(1)
        if name in processes:
            abort("duplicate definition of process `%s'" % name)
        p = Process(m.group(2))
        processes[name] = p
        # collect variable initializations from input
        line = read_next_line(input_file)
        while line != None:
            m = VARDEF_RE.match(line)
            if m:
                p.initialvalues[m.group(1)] = m.group(2)
                line = read_next_line(input_file)
            else:
                break
        continue
    abort("syntax error: `%s'" % line)

input_file.close()

if not processes:
    abort("no processes defined in input")

TM = coral.getTransactionManager()
TM.beginModifyCmd()

# create a Coral project with a UML model
project = coral.Project()
#project.setIOFormat("XMI 2.0")
model = UML.Model(name="Model")
project.addModel(model)

# create primitive data types
int_datatype = UML.Primitive(name="int")
boolean_datatype = UML.Primitive(name="boolean")
model.ownedElement.append(int_datatype)
model.ownedElement.append(boolean_datatype)

# create classes for the proctypes
for (proctypename, proctype) in proctypes.iteritems():
    proctype.cls = UML.Class(name=proctypename, isActive=True)
    model.ownedElement.append(proctype.cls)

for (proctypename, proctype) in proctypes.iteritems():
    # create attributes and associations for the process type
    for (varname, vartype) in proctype.vars.iteritems():
        if vartype in ("int", "boolean"):
            # add primitive type variables as attributes to the class
            attribute = \
                UML.Attribute \
                    (name=varname,
                     owner=proctype.cls,
                     multiplicity=get_default_attribute_multiplicity(),
                     visibility="public",
                     changeability="changeable",
                     initialValue=UML.Expression \
                                      (body=((vartype == "int" and "0")
                                             or "false")),
                     type=((vartype == "int" and int_datatype)
                           or boolean_datatype))
            proctype.attributes[varname] = attribute
            proctype.cls.feature.append(attribute)
        elif vartype in proctypes:
            # add an association between classes
            source_end = \
                UML.AssociationEnd \
                    (isNavigable=False,
                     ordering="unordered",
                     aggregation="none",
                     changeability="changeable",
                     visibility="public",
                     participant=proctype.cls)
            target_end = \
                UML.AssociationEnd \
                    (name=varname,
                     isNavigable=True,
                     ordering="unordered",
                     aggregation="none",
                     multiplicity=get_default_association_end_multiplicity(),
                     changeability="changeable",
                     visibility="public",
                     participant=proctypes[vartype].cls)
            proctype.associations[varname] = (source_end, target_end)
            proctype.cls.ownedElement.append \
                (UML.Association \
                     (name="%s:%s" % (proctypename, vartype),
                      connection=[source_end, target_end]))
        else:
            abort \
                ("unknown type `%s' specified for variable `%s' in process "
                 "type `%s'" % (vartype, varname, proctypename))

    # create a state machine for the process type
    init_state = UML.Pseudostate(name="init", kind="initial")
    start_compute_state = UML.SimpleState(name="start")
    do_compute_state = UML.SimpleState(name="run")
    end_state = UML.FinalState(name="stop")
    statemachine = \
        UML.StateMachine \
            (name="state machine",
             context=proctype.cls,
             top=UML.CompositeState \
                     (name="top",
                      subvertex=[init_state, start_compute_state,
                                 do_compute_state, end_state]),
             transitions=[UML.Transition \
                              (name="initialize", source=init_state,
                               target=start_compute_state),
                          UML.Transition \
                              (name="execute code", source=start_compute_state,
                               target=do_compute_state,
                               effect=UML.UninterpretedAction
                                          (script=UML.ActionExpression \
                                                      (body=" ".join
                                                           (proctype.code)))),
                          UML.Transition \
                              (name="terminate", source=do_compute_state,
                               target=end_state)])
    proctype.cls.ownedElement.append(statemachine)
    proctype.cls.behavior.append(statemachine)

# create instances of processes
for (processname, process) in processes.iteritems():
    try:
        proctype = proctypes[process.proctype]
    except KeyError:
        abort \
            ("unknown process type specified `%s' for process `%s'"
             % (process.proctype, processname))
    process.object = UML.Object(classifier=[proctype.cls], name=processname)
    model.ownedElement.append(process.object)

# initialize variables of processes
for (processname, process) in processes.iteritems():
    proctype = proctypes[process.proctype]
    # check that there are no initializations of unknown variables
    for varname in process.initialvalues.iterkeys():
        if not varname in proctype.vars:
            abort \
                ("initialization of unknown variable `%s' for process `%s'"
                 % (varname, processname))
    # process initializations of initial values, instantiate associations
    for (varname, vartype) in proctype.vars.iteritems():
        if vartype in ("int", "boolean"):
            try:
                value = process.initialvalues[varname]
            except KeyError:
                # primitive type variables can be left uninitialized
                pass
            else:
                try:
                    if vartype == "int":
                        typename = "integer"
                        val = long(value)
                        if val < -2147483648 or val > 2147483647:
                            raise ValueError()
                        datatype = int_datatype
                    else:
                        typename = "Boolean"
                        if not value in ("true", "false"):
                            raise ValueError()
                        datatype = boolean_datatype
                except ValueError:
                    abort \
                        ("initial value `%s' for attribute `%s' of process "
                         "`%s' is not a valid %s constant in the Jumbala "
                         "action language"
                         % (value, varname, processname, typename))
                datavalue = UML.DataValue(name=value,owner=process.object)
                datavalue.classifier.append(datatype)
                link = \
                    UML.AttributeLink \
                        (value=datavalue,
                         attribute=proctype.attributes[varname])
                process.object.slot.append(link)
        else:
            try:
                value = process.initialvalues[varname]
            except KeyError:
                # references must always be initialized
                abort \
                    ("missing initialization of reference variable `%s' for "
                     "process `%s'" % (varname, processname))
            try:
                ref_process = processes[value]
            except KeyError:
                abort \
                    ("unknown process name `%s' specified as initial value "
                     "of reference variable `%s' for process `%s'"
                     % (value, varname, processname))
            if ref_process.proctype != vartype:
                abort \
                    ("process type mismatch in initial value `%s' of "
                     "reference variable `%s' of process `%s'"
                     % (value, varname, processname))
            (source_end, target_end) = proctype.associations[varname]
            link = \
                UML.Link \
                    (association=source_end.association,
                     connection=[UML.LinkEnd 
                                     (associationEnd=source_end,
                                      instance=process.object),
                                 UML.LinkEnd
                                     (associationEnd=target_end,
                                      instance=ref_process.object)])
            process.object.ownedLink.append(link)

TM.endModifyCmd()

# save the model

project.setFilename(args[1])
coral.saveDocument(project)
