#!/usr/bin/python

"""
A translator from UML to promela

Translates UML1.4 models to the input language of
the model checking tool spin.
Can also run spin on the result and parse the counter-example back to
a SMUML trace file format.

Author: Tommi Junttila

Copyright (c) 2007 Helsinki University of Technology
Released under the GNU General Public License version 2
"""

import sys
from sets import Set
import string
import optparse
import os
import re

import coral

import actlang
import Sim
import Sim.Core
import Utils.String

from Sim.ActLangUtils import is_expr_side_effect_free

from Utils.String import Indent
from Utils.String import Writer
from Utils.Models import get_qualified_name
from Utils.Models import get_elements_of_type
from Utils.Models import xmi_sort
from Utils.StateMachines import has_outgoing_completion_transitions
from Utils.StateMachines import get_outgoing_completion_transitions
from Utils.StateMachines import is_completion_sensitive

#from Utils.ExprChooseEliminator import eliminate_chooses

import simulate_generic_trace
import proco_trace_to_generic_trace

UML14 = coral.meta.UML14

__version__ = "2.00"



def _strip(s):
    """
    Convert a string into a form that will be ok in promela.
    """
    assert type(s) is str
    r = ""
    for c in s:
        if c.isalnum():
            r += c
        elif c in ["_"]:
            r += c
        else:
            r += "x"
    return r


class _Allocator:
    """
    A small helper class for collecting a union of multi-sets of types.
    """
    def __init__(self):
        self.type_to_free_index = {}
        self.type_to_allocated_index = {}
        self.index_to_type = {}
        self.current_index = 0
    def allocate(self, t):
        if not t in self.type_to_free_index:
            assert not t in self.type_to_allocated_index
            index = self.current_index
            self.current_index += 1
            self.type_to_free_index[t] = []
            self.type_to_allocated_index[t] = [index]
            self.index_to_type[index] = t
            return index
        if not self.type_to_free_index[t]:
            index = self.current_index
            self.current_index += 1
            self.type_to_allocated_index[t].append(index)
            self.index_to_type[index] = t
            return index
        index = self.type_to_free_index[t].pop()
        self.type_to_allocated_index[t].append(index)
        return index
    def deallocate_all(self):
        for (t,l) in self.type_to_allocated_index.iteritems():
            while l:
                self.type_to_free_index[t].append(l.pop())
    def get_type(self, index):
        assert 0 <= index and index < self.current_index
        return self.index_to_type[index]
    def get_nof_indices(self):
        return self.current_index



def _generate_queue_flush(w, signal_parameter_allocator):
    """
    Generate promela code for flushing the defer queue in
    the front of the input queue.
    """
    assert isinstance(w, Writer)
    assert isinstance(signal_parameter_allocator, _Allocator)
    params = ""
    for i in xrange(0, signal_parameter_allocator.get_nof_indices()):
        params += ",param"+str(i)
    w.ln("/* Flush the defer queue in front of the input queue */")
    w.ln("inline queue_flush(input, defer) {")
    w.ln("  if")
    w.ln("  :: empty(defer) -> skip;")
    w.ln("  :: nempty(defer) ->")
    w.ln("     qcj = len(input);")
    w.ln("     do")
    w.ln("     :: empty(defer) -> break;")
    w.ln("     :: nempty(defer) ->")
    w.ln("        defer?trigger"+params+";")
    w.ln("        if")
    w.ln("        :: full(input) -> overflowed = true; goto end_queue_overflow;")
    w.ln("        :: nfull(input) -> input!trigger"+params+";")
    w.ln("        fi;")
    w.ln("     od;")
    w.ln("     do")
    w.ln("     :: (qcj > 0) ->")
    w.ln("        input?trigger"+params+";")
    w.ln("        input!trigger"+params+";")
    w.ln("        qcj = qcj - 1;")
    w.ln("     :: (qcj == 0) -> break;")
    w.ln("     od;")
    w.ln("  fi;")
    w.ln("}")
    w.ln("")



def _generate_defer_code(w, signal_parameter_allocator):
    """
    Generate promela code for deferring a message.
    """
    assert isinstance(w, Writer)
    assert isinstance(signal_parameter_allocator, _Allocator)
    w.ln("/* Defer the first message in the input queue that has")
    w.ln(" * already been read to the local variables trigger and param */")
    w.ln("inline defer_message(defer) {")
    w.ln("  if")
    w.ln("  :: full(defer) -> overflowed = true; goto end_queue_overflow;")
    w.ln("  :: nfull(defer) -> skip;")
    w.ln("  fi;")
    params = ""
    for i in xrange(0, signal_parameter_allocator.get_nof_indices()):
        params += ",param"+str(i)
    w.ln("  defer!trigger"+params+";")
    w.ln("}")
    w.ln("")



def _generate_send_code(w, signal_parameter_allocator):
    """
    Generate promela code for sending a message.
    """
    assert isinstance(w, Writer)
    assert isinstance(signal_parameter_allocator, _Allocator)
    N = signal_parameter_allocator.get_nof_indices()
    w.ln("/* Send a message */")
    w.ln("#define send_message(target,signal"+
         ("".join([",t"+str(i) for i in range(0, N)]))+
         ") \\")
    w.ln("  if\\")
    w.ln("  :: full(input[target]) -> overflowed = true; goto end_queue_overflow;\\")
    w.ln("  :: nfull(input[target]) -> input[target]!signal"+
         ("".join([",t"+str(i) for i in range(0, N)]))+
         ";\\")
    w.ln("  fi;")
    w.ln("")



def translate(sim,
              f = sys.stdout,
              opt_queue_size = 2,
              opt_check_assertions = False,
              opt_check_all_implicit_consumptions = False,
              opt_check_implicit_consumption_on_classes = Set(),
              opt_check_queue_overflows = False):
    """
    Translate the UML Model that has been read in the Sim.Core.Simulator
    'sim' into promela code and output it in the file 'f'.
    """
    assert isinstance(sim, Sim.Core.Simulator)

    indent = Indent()

    option_spin_max_live_object_id = 2

    uml_model = sim.get_uml_model()

    nof_classes = len(sim.get_class_defs())

    sim_initial_state = sim.trace_get_last_global_state()

    def _size_to_domain(i):
        assert type(i) is int or type(i) is long
        assert i >= 0
        if i < 256:
            return "byte"
        if i < pow(2, 31):
            return "int"
        assert False and "Too large domain"

    spin_object_id_type = _size_to_domain(len(sim.get_class_defs()))
    spin_max_live_object_id = max(option_spin_max_live_object_id,
                                  sim_initial_state.get_nof_objects())
    if spin_max_live_object_id >= 255:
        raise runtimeError("At most 254 objects can be handled")
    spin_null_object_id = spin_max_live_object_id + 1
    spin_largest_object_id = spin_null_object_id
    spin_queue_size_type = _size_to_domain(opt_queue_size)

    def _sim_type_to_spin_type(st):
        assert isinstance(st, Sim.Core.SimType)
        if isinstance(st, Sim.Core.SimType_Boolean):
            return "bool"
        elif isinstance(st, Sim.Core.SimType_Int):
            return "int"
        elif isinstance(st, Sim.Core.SimType_ObjectRef):
            return _size_to_domain(spin_largest_object_id)
        else:
            assert False
    def _sim_type_to_spin_default_value(st):
        assert isinstance(st, Sim.Core.SimType)
        if isinstance(st, Sim.Core.SimType_Boolean):
            return "false"
        elif isinstance(st, Sim.Core.SimType_Int):
            return "0"
        elif isinstance(st, Sim.Core.SimType_ObjectRef):
            return str(spin_null_object_id)
        else:
            assert False
    def _spin_type_to_spin_default_value(st):
        assert isinstance(st, str)
        if st == "bool":
            return "false"
        elif st == "byte":
            return "0"
        elif st == "int":
            return "0"
        else:
            #print st
            raise RuntimeError("Internal error")

    w = Writer(f)

    # Global defines
    w.ln("#define QSZ "+str(opt_queue_size))
    w.ln("#define MAX_IDS "+str(spin_max_live_object_id))

    w.ln(spin_object_id_type+" id = 0;")

    w.ln("")
    w.ln("/* A flag to catch that a queue has overflowed */")
    w.ln("bool overflowed = false;")
    
    # Print definitions for signal names
    w.ln("")
    w.ln("/* Signal definitions */")
    signal_to_index = {}
    def signal_to_spin_name(signal):
        """
        Given 'signal' that is either the name of a signal,
        a SignalDef object, or a UML Signal in the model,
        return the corresponding name in promela.
        """
        if type(signal) is str:
            signal_def = sim.get_signal_def(signal)
            return signal_to_spin_name(signal_def)
        s = "sig"+str(signal_to_index[signal])+"_"
        if isinstance(signal, Sim.Core.SignalDef):
            s += _strip(signal.get_name())
        else:
            s += _strip(str(signal.name))
        return s
    opt_use_mtype = True
    if opt_use_mtype:
        if len(sim.get_signal_defs()) > 256:
            raise RuntimeError("Too many signals, at most 256 allowed")
        spin_trigger_type = "mtype"
        w.ln("mtype {")
        w.push()
        sep = ""
        for (index,signal_def) in enumerate(sim.get_signal_defs()):
            signal_to_index[signal_def] = index
            signal_to_index[signal_def.get_uml_signal()] = index
            w.ln(sep+signal_to_spin_name(signal_def))
            sep = ","
        if sep == "":
            # No signals, print a dummy one
            w.ln("sig_dummy")
        w.pop()
        w.ln("}")
    else:
        spin_trigger_type = _size_to_domain(len(sim.get_signal_defs()))
        w.ln("/* Signal trigger constants */")
        for (index,signal_def) in enumerate(sim.get_signal_defs()):
            signal_to_index[signal_def] = index
            signal_to_index[signal_def.get_uml_signal()] = index
            w.ln("#define "+signal_to_spin_name(signal_def)+" "+str(index))
    w.ln("")

    # Allocate enough spin types to hold all possible signal parameters,
    # share as much as possible
    signal_parameter_allocator = _Allocator()
    signal_parameter_to_spin_parameter = {}
    for signal_def in sim.get_signal_defs():
        uml_signal = signal_def.get_uml_signal()
        for i in xrange(0, signal_def.get_nof_parameters()):
            sim_param_type = signal_def.get_parameter_type(i)
            spin_param_type = _sim_type_to_spin_type(sim_param_type)
            spin_index = signal_parameter_allocator.allocate(spin_param_type)
            signal_parameter_to_spin_parameter[(signal_def,i)] = spin_index
            signal_parameter_to_spin_parameter[(uml_signal,i)] = spin_index
        signal_parameter_allocator.deallocate_all()
    def _get_spin_signal_param_name(signal_def, param_index):
        assert isinstance(signal_def, Sim.Core.SignalDef)
        assert 0 <= param_index
        assert param_index < signal_def.get_nof_parameters()
        return "param"+str(signal_parameter_to_spin_parameter[(signal_def,
                                                               param_index)])

    # Compute a flag that tells whether any state machine in the model
    # has a state that declares a deferrable event
    model_has_deferrable_events = len([s for s in get_elements_of_type(uml_model, UML14.State) if s.deferrableEvent]) > 0

    s = spin_trigger_type
    for i in xrange(0, signal_parameter_allocator.get_nof_indices()):
        s += ", "+signal_parameter_allocator.get_type(i)
    w.ln("/* An input queue for each object */")
    w.ln("chan input[MAX_IDS] = [QSZ] of {"+s+"};")
    if model_has_deferrable_events:
        w.ln("/* A deferred message queue for each object */")
        w.ln("chan defer[MAX_IDS] = [QSZ] of {"+s+"};")

    # Associate each ClassDef with a unique integer index
    class_def_to_index = {}
    for (index,cd) in enumerate(sim.get_class_defs()):
        class_def_to_index[cd] = index
    def _class_def_spin_name(cd):
        """Get the promela name of the ClassDef cd."""
        assert isinstance(cd, Sim.Core.ClassDef)
        return "class"+str(class_def_to_index[cd])+"_"+_strip(cd.get_name())


    # Allocate enough spin types to hold all possible class attributes,
    # share as much as possible
    vardef_allocator = _Allocator()
    vardef_to_spin_index = {}
    for cd in sim.get_class_defs():
        var_defs = cd.get_variable_defs()
        for (var_name,var_def) in var_defs.items():
            sim_type = var_def.get_type()
            spin_type = _sim_type_to_spin_type(sim_type)
            spin_index = vardef_allocator.allocate(spin_type)
            vardef_to_spin_index[var_def] = spin_index
        vardef_allocator.deallocate_all()
    def _vardef_to_spin_name(var_def):
        assert isinstance(var_def, Sim.Core.VariableDef)
        return "v"+str(vardef_to_spin_index[var_def])
    if vardef_allocator.get_nof_indices() > 0:
        w.ln("")
        w.ln("/* Space to hold attributes of classes */")
        w.ln("typedef Attrs {")
        w.push()
        for i in xrange(0, vardef_allocator.get_nof_indices()):
            w.ln(vardef_allocator.get_type(i)+" v"+str(i)+";")
        w.pop()
        w.ln("}");
        w.ln("Attrs attrs[MAX_IDS];")
        w.ln("")


    if model_has_deferrable_events:
        #
        # Generate code for flushing the defer queue in front of input queue
        #
        _generate_queue_flush(w, signal_parameter_allocator)
        #
        # Generate code for deferring the first message in the input queue
        #
        _generate_defer_code(w, signal_parameter_allocator)

    _generate_send_code(w, signal_parameter_allocator)

    #
    # Collect some state machine information of active UML Class
    #
    uml_state_to_index = {}
    uml_composite_state_to_index_type = {}
    def _uml_statevertex_to_spin_name(s, quiescent = False):
        """Get the spin name of the UML StateVertex 's'."""
        assert s.oclIsKindOf(UML14.StateVertex)
        assert isinstance(quiescent, bool)
        spin_name = "s"+str(uml_state_to_index[s])
        if quiescent:
            spin_name += "q"
        spin_name += "_"+_strip(str(s.name))
        return spin_name
    for cd in sim.get_class_defs():
        uml_sm = cd.get_uml_state_machine()
        if not uml_sm:
            continue
        assert uml_sm.top.oclIsKindOf(UML14.CompositeState)
        assert uml_sm.top.stateMachine == uml_sm
        assert not uml_sm.top.container
        spin_composite_state_index = 0
        # Find non-concurrent composite states
        uml_nccss = [s for s in get_elements_of_type(uml_sm, UML14.CompositeState) if not s.isConcurrent]
        for uml_nccs in uml_nccss:
            uml_composite_state_to_index_type[uml_nccs] = (spin_composite_state_index, _size_to_domain(len(uml_nccs.subvertex)))
            spin_composite_state_index += 1
            w.ln("/* Substates of "+get_qualified_name(uml_nccs)+" */")
            spin_index = 0
            for s in uml_nccs.subvertex:
                uml_state_to_index[s] = len(uml_state_to_index.keys())
                w.ln("#define "+_uml_statevertex_to_spin_name(s, False)+
                     " "+str(spin_index))
                spin_index += 1
                if is_completion_sensitive(s):
                    w.ln("#define "+_uml_statevertex_to_spin_name(s, True)+
                         " "+str(spin_index))
                    spin_index += 1
            w.ln("")



    def _trans_effect(cd, node):
        """
        Translate the Jumbala effect statement 'node' into Promela code.
        The context (class) is given by the ClassDef 'cd'.
        """
        assert isinstance(cd, Sim.Core.ClassDef)
        #print "Translating:", node


        if isinstance(node, actlang.statement.SendStatement):
            signal_name = node.api_getSignalName()
            signal_def = sim.get_signal_def(signal_name)
            spin_signal_name = signal_to_spin_name(signal_name)
            target_ast = node.api_getTarget()
            if not is_expr_side_effect_free(target_ast):
                raise RuntimeError("Cannot translate to promela because "+
                                   "target expression has side effects: "+
                                   str(node))
            (target_sim_type,target_spin_code) = _trans_effect(cd, target_ast)
            assert isinstance(target_sim_type, Sim.Core.SimType_ObjectRef)
            s = ""
            s += "input["+target_spin_code+"]!"+spin_signal_name
            spin_param_index_to_arg_code = {}
            for (arg_index,arg_ast) in enumerate(node.api_getArgList()):
                (arg_sim_type,arg_spin_code) = _trans_effect(cd, arg_ast)
                assert arg_sim_type.compare(signal_def.get_parameter_type(arg_index)) == 0
                spin_param_index_to_arg_code[signal_parameter_to_spin_parameter[(signal_def,arg_index)]] = arg_spin_code
            for i in xrange(0, signal_parameter_allocator.get_nof_indices()):
                if spin_param_index_to_arg_code.has_key(i):
                    s += ","+spin_param_index_to_arg_code[i]
                else:
                    s += ","+_spin_type_to_spin_default_value(signal_parameter_allocator.get_type(i))
            s += ";"
            # A new version
            s = "send_message("+target_spin_code+","+spin_signal_name
            for i in xrange(0, signal_parameter_allocator.get_nof_indices()):
                if spin_param_index_to_arg_code.has_key(i):
                    s += ","+spin_param_index_to_arg_code[i]
                else:
                    s += ","+_spin_type_to_spin_default_value(signal_parameter_allocator.get_type(i))
            s += ");"
            return (None, s)


        elif isinstance(node, actlang.expr.ParenExpr):
            return _trans_effect(cd, node.api_getInnerExpr())


        elif isinstance(node, actlang.expr.IntLiteralExpr):
            val = node.api_getValue()
            assert -pow(2,31) <= val and val < pow(2,31)
            return (Sim.Core.SimType_Int(), str(val))


        elif isinstance(node, actlang.expr.BoolLiteralExpr):
            val = node.api_getValue()
            assert isinstance(val, bool)
            if val:
                return (Sim.Core.SimType_Boolean(), "true")
            return (Sim.Core.SimType_Boolean(), "false")


        elif isinstance(node, actlang.expr.NullExpr):
            return (Sim.Core.SimType_ObjectRef(), str(spin_null_object_id))


        elif isinstance(node, actlang.expr.StringLiteralExpr):
            raise RuntimeError("Strings are not handled in "+
                               "the promela translation")


        elif isinstance(node, actlang.expr.ThisExpr):
            return (Sim.Core.SimType_ObjectRef(cd.get_name()), "oid")


        elif isinstance(node, (actlang.expr.CastExpr
                               ,actlang.expr.TypeCastExpr
                               ,actlang.expr.NameCastExpr)):
            raise RuntimeError("Cast expressions are not handled in "+
                               "the promela translation: "+str(node))


        elif isinstance(node, actlang.expr.PostfixExpr):
            raise RuntimeError("Cannot translate to promela: "+str(node))
            #(t1, s1) = _trans_effect(cd, node.api_getOperand())
            #op = node.api_getOperator()
            #if isinstance(t1, Sim.Core.SimType_Int):
            #    if isinstance(op, actlang.binaryop.Plus):
            #        return (t1, "("+s1+")++")
            #    elif isinstance(op, actlang.binaryop.Minus):
            #        return (t1, "("+s1+")--")
            #    raise RuntimeError("Internal error")
            #raise RuntimeError("Cannot translate to promela: "+str(node))


        elif isinstance(node, actlang.expr.PrefixExpr):
            raise RuntimeError("Cannot translate to promela: "+str(node))


        elif isinstance(node, actlang.expr.UnaryPlusExpr):
            (t1, s1) = _trans_effect(cd, node.api_getOperand())
            if isinstance(t1, Sim.Core.SimType_Int):
                return (t1, "("+s1+")")
            raise RuntimeError("Applying unary plus to a non-integer "+
                               "expression: "+str(node))


        elif isinstance(node, actlang.expr.UnaryMinusExpr):
            (t1, s1) = _trans_effect(cd, node.api_getOperand())
            if isinstance(t1, Sim.Core.SimType_Int):
                return (t1, "-("+s1+")")
            raise RuntimeError("Applying unary minus to a non-integer "+
                               "expression: "+str(node))


        elif isinstance(node, actlang.expr.BitNotExpr):
            (t1, s1) = _trans_effect(cd, node.api_getOperand())
            if isinstance(t1, Sim.Core.SimType_Int):
                return (t1, "~("+s1+")")
            raise RuntimeError("Applying bitwise not to a non-integer "+
                               "expression: "+str(node))


        elif isinstance(node, actlang.expr.LogicalNotExpr):
            (t1, s1) = _trans_effect(cd, node.api_getOperand())
            if isinstance(t1, Sim.Core.SimType_Boolean):
                return (t1, "!("+s1+")")
            raise RuntimeError("Applying logical not to a non-Boolean "+
                               "expression: "+str(node))


        elif isinstance(node, actlang.expr.InstanceOfExpr):
            raise RuntimeError("Cannot translate to promela: "+str(node))


        elif isinstance(node, actlang.statement.AssertStatement):
            if opt_check_assertions:
                (t,s) = _trans_effect(cd, node.api_getExpr())
                return (None, "assert("+s+");")
            return (None, "")


        elif isinstance(node, actlang.expr.NameMethodInvocationExpr):
            n = node.api_getInvocationName()
            if str(n) == "$$SMUML.choose_int":
                raise RuntimeError("Non-deterministic choose-operations "
                                   "are not yet supported in the Promela "
                                   "translation")
                args = node.api_getArgList()
                assert len(args) >= 1
                value_list = args[0]
                assert isinstance(value_list, actlang.expr.StringLiteralExpr)
                value_list = str(value_list)[1:-1].split(',')
                value_list = [int(v) for v in value_list]
                for value in value_list:
                    assert -pow(2,31) <= value and value < pow(2,31)
                sys.stderr.write("Warning: chooses not yet supported\n")
                return (Sim.Core.SimType_Int(), str(value_list[-1]))
            elif str(n) == "$$SMUML.choose_boolean":
                raise RuntimeError("Non-deterministic choose-operations "
                                   "are not yet supported in the Promela "
                                   "translation")
                args = node.api_getArgList()
                assert len(args) >= 1
                value_list = args[0]
                assert isinstance(value_list, actlang.expr.StringLiteralExpr)
                value_list = str(value_list)[1:-1].split(',')
                value_list = [v.strip() for v in value_list]
                for value in value_list:
                    assert value in ["false", "true"]
                sys.stderr.write("Warning: chooses not yet supported\n")
                return (Sim.Core.SimType_Boolean(), str(value_list[-1]))
            raise RuntimeError("Method invocations are not supported "
                               "in the Promela translation")


        elif isinstance(node, actlang.statement.ExprStatement):
            (t, s) = _trans_effect(cd, node.api_getExpr())
            return (t, s+";")


        elif isinstance(node,  actlang.expr.TernaryCondExpr):
            (cond_t, cond_c) = _trans_effect(cd, node.api_getCondition())
            (true_t, true_c) = _trans_effect(cd, node.api_getTrueExpr())
            (false_t, false_c) = _trans_effect(cd, node.api_getFalseExpr())
            assert true_t.compare(false_t) == 0
            return (true_t, "(("+cond_c+") -> ("+true_c+") : ("+false_c+"))")


        elif isinstance(node, actlang.expr.CondOrExpr):
            (left_t, left_c) = _trans_effect(cd, node.api_getLeftOperand())
            (right_t, right_c) = _trans_effect(cd, node.api_getRightOperand())
            return (Sim.Core.SimType_Boolean(), "("+left_c+")||("+right_c+")")


        elif isinstance(node, actlang.expr.CondAndExpr):
            (left_t, left_c) = _trans_effect(cd, node.api_getLeftOperand())
            (right_t, right_c) = _trans_effect(cd, node.api_getRightOperand())
            return (Sim.Core.SimType_Boolean(), "("+left_c+")&&("+right_c+")")


        elif isinstance(node, actlang.expr.SimpleAssignmentExpr):
            (lt, ls) = _trans_effect(cd, node.api_getLhsExpr())
            (rt, rs) = _trans_effect(cd, node.api_getRhsExpr())
            if lt.compare(rt) != 0:
                raise RuntimeError("The types of the left and right hand sides of the assignment are incompatible: "+str(node))
            return (lt, ls+" = ("+rs+")")


        elif isinstance(node, actlang.expr.InstanceCreationExpr):
            class_name = node.api_getTypeName()
            assert isinstance(class_name, actlang.names.SimpleName)
            class_name = str(class_name)
            class_def = sim.get_class_def(class_name)
            raise RuntimeError("'new' statements are not yet supported "
                               "in the Promela translation")
            if node.api_getArgList():
                raise RuntimeError("Arguments in 'new' statements "+
                                   "are not supported "
                                   "in the Promela translation")
            raise RuntimeError("'new' statements not supported "
                               "in the Promela translation")
            s = "TO DO"
            return (Sim.Core.SimType_ObjectRef(class_def.get_name()), s)


        elif isinstance(node, actlang.expr.InfixExpr):
            (t1, s1) = _trans_effect(cd, node.api_getLeftOperand())
            (t2, s2) = _trans_effect(cd, node.api_getRightOperand())
            op = node.api_getOperator()
            if(isinstance(t1, Sim.Core.SimType_Int) and
               isinstance(t2, Sim.Core.SimType_Int)):
                if isinstance(op, (actlang.binaryop.NumericalComparison,
                                   actlang.binaryop.EqualityOperator)):
                    return (Sim.Core.SimType_Boolean(),
                            "("+s1+") "+str(op)+" ("+s2+")")
                if isinstance(op, (actlang.binaryop.Plus,
                                   actlang.binaryop.Minus,
                                   actlang.binaryop.Multiply,
                                   actlang.binaryop.Divide,
                                   actlang.binaryop.BitLogicalOr,
                                   actlang.binaryop.BitLogicalAnd,
                                   actlang.binaryop.BitLogicalXor)):
                    return (Sim.Core.SimType_Int(),
                            "("+s1+") "+str(op)+" ("+s2+")")
            elif(isinstance(t1, Sim.Core.SimType_Boolean) and
                 isinstance(t2, Sim.Core.SimType_Boolean)):
                if isinstance(op, (actlang.binaryop.EqualityOperator)):
                    return (Sim.Core.SimType_Boolean(),
                            "("+s1+") "+str(op)+" ("+s2+")")
            elif(isinstance(t1, Sim.Core.SimType_ObjectRef) and
                 isinstance(t2, Sim.Core.SimType_ObjectRef)):
                if isinstance(op, (actlang.binaryop.EqualityOperator)):
                    return (Sim.Core.SimType_Boolean(),
                            "("+s1+") "+str(op)+" ("+s2+")")
            raise RuntimeError("Cannot translate to promela: "+str(node))


        elif isinstance(node, actlang.expr.CompoundAssignmentExpr):
            if not is_expr_side_effect_free(node.api_getLhsExpr()):
                raise RuntimeError("Cannot translate to promela because "+
                                   "left-hand side has side effects: "+
                                   str(node))
            (t1, s1) = _trans_effect(cd, node.api_getLhsExpr())
            (t2, s2) = _trans_effect(cd, node.api_getRhsExpr())
            op = node.api_getOperator()
            if(isinstance(t1, Sim.Core.SimType_Int) and
               isinstance(t2, Sim.Core.SimType_Int)):
                if isinstance(op, (actlang.binaryop.Plus,
                                   actlang.binaryop.Minus,
                                   actlang.binaryop.Multiply,
                                   actlang.binaryop.Divide,
                                   actlang.binaryop.BitLogicalOr,
                                   actlang.binaryop.BitLogicalAnd,
                                   actlang.binaryop.BitLogicalXor)):
                    return (Sim.Core.SimType_Int(),
                            s1+" = (("+s1+") "+str(op)+" ("+s2+"))")
            elif(isinstance(t1, Sim.Core.SimType_Boolean) and
                 isinstance(t2, Sim.Core.SimType_Boolean)):
                if isinstance(op, (actlang.binaryop.BitLogicalOr,
                                   actlang.binaryop.BitLogicalAnd,
                                   actlang.binaryop.BitLogicalXor)):
                    return (Sim.Core.SimType_Boolean(),
                            s1+" = (("+s1+") "+str(op)+" ("+s2+"))")
            raise RuntimeError("Cannot translate to promela: "+str(node))


        elif isinstance(node, actlang.names.QualifiedName):
            context = cd
            parts = node.api_getParts()
            assert isinstance(parts, list) and parts != []
            prefix = []
            sim_type = None
            spin_code = "oid"
            while parts:
                part = parts.pop(0)
                prefix.append(part)
                assert isinstance(part, str)
                # Resolve: try context class 'cd' class attributes
                var_defs = context.get_variable_defs()
                if var_defs.has_key(part):
                    var_def = var_defs[part]
                    sim_type = var_def.get_type()
                    if parts:
                        if not isinstance(sim_type,
                                          Sim.Core.SimType_ObjectRef):
                            raise RuntimeError((".".join(prefix))+
                                               " is not of reference type")
                        target_name = sim_type.get_target_name()
                        context = sim.get_class_def(target_name)
                    spin_code = "attrs["+spin_code+"]."+\
                                _vardef_to_spin_name(var_def)
                    continue
                raise RuntimeError("Cannot bind variable '"+part+"' in the Class "+get_qualified_name(context.get_uml_class()))
            return (sim_type, spin_code)

        elif isinstance(node, actlang.names.SimpleName):
            name = node.api_getLastPart()
            assert isinstance(name, str)
            # Resolve: try context class 'cd' class attributes
            var_defs = cd.get_variable_defs()
            if var_defs.has_key(name):
                var_def = var_defs[name]
                sim_type = var_def.get_type()
                #spin_code = var_def.get_name()
                spin_code = "attrs[oid]."+_vardef_to_spin_name(var_def)
                return (sim_type, spin_code)
            raise RuntimeError("Cannot bind variable '"+name+"' in the Class "+
                               get_qualified_name(cd.get_uml_class()))

        elif isinstance(node, actlang.expr.PrimaryFieldAccessExpr):
            # expr.field
            
            # First evaluate the expression the context cd
            expr = node.api_getReferenceExpr()
            (expr_t, expr_s) = _trans_effect(cd, expr)
            if not isinstance(expr_t, Sim.Core.SimType_ObjectRef):
                raise RuntimeError("Type error: "+str(expr)+
                                   " is not of reference type")
            expr_class_name = expr_t.get_target_name()
            expr_cd = sim.get_class_def(expr_class_name)

            # Find the field in the class referenced by expr
            field_name = node.api_getFieldName()
            var_defs = expr_cd.get_variable_defs()
            if not var_defs.has_key(field_name):
                raise RuntimeError("Cannot bind variable '"+name+"' "+\
                                   "in the Class "+
                                   get_qualified_name(expr_cd.get_uml_class()))
            var_def = var_defs[field_name]
            sim_type = var_def.get_type()
            spin_code = "attrs["+expr_s+"]."+_vardef_to_spin_name(var_def)
            return (sim_type, spin_code)

        #print node.__class__
        raise RuntimeError("Not supported in transition effect: "+str(node))



    def _trans_guard(cd, trigger, node):
        """
        Translate the Jumbala guard expression 'node' into Promela code.
        The context (class) is given by the ClassDef 'cd'.
        If not None, 'trigger' gives the SignalEvent context,
        describing the which attributes the signal parameters
        are to be instantiated (the guard is evaluated as if this
        instantiation was already done).
        """
        assert isinstance(cd, Sim.Core.ClassDef)
        assert (trigger is None) or trigger.oclIsKindOf(UML14.SignalEvent)
        assert isinstance(node, actlang.expr.Expr)
        #print "Translating:", node

        if isinstance(node, actlang.expr.ParenExpr):
            return _trans_guard(cd, trigger, node.api_getInnerExpr())

        elif isinstance(node, actlang.expr.IntLiteralExpr):
            val = node.api_getValue()
            assert -pow(2,31) <= val and val < pow(2,31)
            return (Sim.Core.SimType_Int(), str(val))
            
        elif isinstance(node, actlang.expr.BoolLiteralExpr):
            val = node.api_getValue()
            assert isinstance(val, bool)
            if val:
                return (Sim.Core.SimType_Boolean(), "true")
            return (Sim.Core.SimType_Boolean(), "false")

        elif isinstance(node, actlang.expr.NullExpr):
            return (Sim.Core.SimType_ObjectRef(), str(spin_null_object_id))

        elif isinstance(node, actlang.expr.StringLiteralExpr):
            raise RuntimeError("Strings are not handled in "+
                               "the promela translation")

        elif isinstance(node, actlang.expr.ThisExpr):
            return (Sim.Core.SimType_ObjectRef(cd.get_name()), "oid")

        elif isinstance(node, (actlang.expr.CastExpr
                               ,actlang.expr.TypeCastExpr
                               ,actlang.expr.NameCastExpr)):
            raise RuntimeError("Cast expressions are not handled in "+
                               "the promela translation: "+str(node))
 
        elif isinstance(node, actlang.expr.UnaryPlusExpr):
            (t1, s1) = _trans_guard(cd, trigger, node.api_getOperand())
            if isinstance(t1, Sim.Core.SimType_Int):
                return (t1, "("+s1+")")
            raise RuntimeError("Applying unary plus to a non-integer "+
                               "expression: "+str(node))

        elif isinstance(node, actlang.expr.UnaryMinusExpr):
            (t1, s1) = _trans_guard(cd, trigger, node.api_getOperand())
            if isinstance(t1, Sim.Core.SimType_Int):
                return (t1, "-("+s1+")")
            raise RuntimeError("Applying unary minus to a non-integer "+
                               "expression: "+str(node))

        elif isinstance(node, actlang.expr.BitNotExpr):
            (t1, s1) = _trans_guard(cd, trigger, node.api_getOperand())
            if isinstance(t1, Sim.Core.SimType_Int):
                return (t1, "~("+s1+")")
            raise RuntimeError("Applying bitwise not to a non-integer "+
                               "expression: "+str(node))

        elif isinstance(node, actlang.expr.LogicalNotExpr):
            (t1, s1) = _trans_guard(cd, trigger, node.api_getOperand())
            if isinstance(t1, Sim.Core.SimType_Boolean):
                return (t1, "!("+s1+")")
            raise RuntimeError("Applying logical not to a non-Boolean "+
                               "expression: "+str(node))

        elif isinstance(node, actlang.expr.InfixExpr):
            (t1, s1) = _trans_guard(cd, trigger, node.api_getLeftOperand())
            (t2, s2) = _trans_guard(cd, trigger, node.api_getRightOperand())
            op = node.api_getOperator()
            if(isinstance(t1, Sim.Core.SimType_Int) and
               isinstance(t2, Sim.Core.SimType_Int)):
                if isinstance(op, (actlang.binaryop.NumericalComparison,
                                   actlang.binaryop.EqualityOperator)):
                    return (Sim.Core.SimType_Boolean(),
                            "("+s1+") "+str(op)+" ("+s2+")")
                if isinstance(op, (actlang.binaryop.Plus,
                                   actlang.binaryop.Minus,
                                   actlang.binaryop.Multiply,
                                   actlang.binaryop.Divide,
                                   actlang.binaryop.BitLogicalOr,
                                   actlang.binaryop.BitLogicalAnd,
                                   actlang.binaryop.BitLogicalXor)):
                    return (Sim.Core.SimType_Int(),
                            "("+s1+") "+str(op)+" ("+s2+")")
            elif(isinstance(t1, Sim.Core.SimType_Boolean) and
                 isinstance(t2, Sim.Core.SimType_Boolean)):
                if isinstance(op, (actlang.binaryop.EqualityOperator)):
                    return (Sim.Core.SimType_Boolean(),
                            "("+s1+") "+str(op)+" ("+s2+")")
            elif(isinstance(t1, Sim.Core.SimType_ObjectRef) and
                 isinstance(t2, Sim.Core.SimType_ObjectRef)):
                if isinstance(op, (actlang.binaryop.EqualityOperator)):
                    return (Sim.Core.SimType_Boolean(),
                            "("+s1+") "+str(op)+" ("+s2+")")
            raise RuntimeError("Cannot translate to promela: "+str(node))

        elif isinstance(node,  actlang.expr.TernaryCondExpr):
            (ct, cs) = _trans_guard(cd, trigger, node.api_getCondition())
            (tt, ts) = _trans_guard(cd, trigger, node.api_getTrueExpr())
            (ft, fs) = _trans_guard(cd, trigger, node.api_getFalseExpr())
            assert tt.compare(ft) == 0
            return (tt, "(("+cs+") -> ("+ts+") : ("+fs+"))")
        
        elif isinstance(node, actlang.expr.CondAndExpr):
            (lt, lc) = _trans_guard(cd, trigger, node.api_getLeftOperand())
            (rt, rc) = _trans_guard(cd, trigger, node.api_getRightOperand())
            assert isinstance(lt, Sim.Core.SimType_Boolean)
            assert isinstance(rt, Sim.Core.SimType_Boolean)
            return (Sim.Core.SimType_Boolean(), "("+lc+")&&("+rc+")")

        elif isinstance(node, actlang.expr.CondOrExpr):
            (lt, lc) = _trans_guard(cd, trigger, node.api_getLeftOperand())
            (rt, rc) = _trans_guard(cd, trigger, node.api_getRightOperand())
            assert isinstance(lt, Sim.Core.SimType_Boolean)
            assert isinstance(rt, Sim.Core.SimType_Boolean)
            return (Sim.Core.SimType_Boolean(), "("+lc+")||("+rc+")")

        elif isinstance(node, actlang.expr.NameMethodInvocationExpr):
            n = node.api_getInvocationName()
            if str(n) == "$$SMUML.choose_int":
                raise RuntimeError("Non-deterministic choose-operations "
                                   "are not yet supported in the Promela "
                                   "translation")
                args = node.api_getArgList()
                assert len(args) >= 1
                value_list = args[0]
                assert isinstance(value_list, actlang.expr.StringLiteralExpr)
                value_list = str(value_list)[1:-1].split(',')
                value_list = [int(v) for v in value_list]
                for value in value_list:
                    assert -pow(2,31) <= value and value < pow(2,31)
                sys.stderr.write("Warning: chooses not yet supported\n")
                return (Sim.Core.SimType_Int(), str(value_list[-1]))
            elif str(n) == "$$SMUML.choose_boolean":
                raise RuntimeError("Non-deterministic choose-operations "
                                   "are not yet supported in the Promela "
                                   "translation")
                args = node.api_getArgList()
                assert len(args) >= 1
                value_list = args[0]
                assert isinstance(value_list, actlang.expr.StringLiteralExpr)
                value_list = str(value_list)[1:-1].split(',')
                value_list = [v.strip() for v in value_list]
                for value in value_list:
                    assert value in ["false", "true"]
                sys.stderr.write("Warning: chooses not yet supported\n")
                return (Sim.Core.SimType_Boolean(), str(value_list[-1]))
            raise RuntimeError("Method invocations are not supported "
                               "in the Promela translation")


        elif isinstance(node, actlang.names.QualifiedName):
            context = cd
            parts = node.api_getParts()
            assert isinstance(parts, list) and parts != []
            prefix = []
            sim_type = None
            spin_code = "oid"
            while parts:
                part = parts.pop(0)
                prefix.append(part)
                assert type(part) is str
                # Resolve name: first try trigger parameters
                if trigger:
                    for (param_index, param) in enumerate(trigger.parameter):
                        if str(param.name) != part:
                            continue
                        assert False and "Not yet implemented"
                        uml_signal = trigger.signal
                        signal_def = sim.get_signal_def(uml_signal)
                        spin_param_name = _get_spin_signal_param_name(signal_def,
                                                                      param_index)
                        sim_param_type = signal_def.get_parameter_type(param_index)
                        if parts:
                            if not isinstance(sim_param_type,
                                              Sim.Core.SimType_ObjectRef):
                                raise RuntimeError((".".join(prefix))+
                                                   " is not of reference type")
                            target_name = sim_param_type.get_target_name()
                        assert False and "Not yet implemented"
                        continue
                # Resolve name: next try context class 'cd' class attributes
                var_defs = context.get_variable_defs()
                if var_defs.has_key(part):
                    var_def = var_defs[part]
                    sim_type = var_def.get_type()
                    if parts:
                        if not isinstance(sim_type,
                                          Sim.Core.SimType_ObjectRef):
                            raise RuntimeError((".".join(prefix))+
                                               " is not of reference type")
                        target_name = sim_type.get_target_name()
                        context = sim.get_class_def(target_name)
                    spin_code = "attrs["+spin_code+"]."+\
                                _vardef_to_spin_name(var_def)
                    continue
                raise RuntimeError("Cannot bind variable '"+part+"' in the Class "+get_qualified_name(context.get_uml_class()))
            return (sim_type, spin_code)

        elif isinstance(node, actlang.names.SimpleName):
            name = node.api_getLastPart()
            assert isinstance(name, str)
            # Resolve name: first try trigger parameters
            if trigger:
                for (param_index, param) in enumerate(trigger.parameter):
                    if str(param.name) != name:
                        continue
                    uml_signal = trigger.signal
                    signal_def = sim.get_signal_def(uml_signal)
                    spin_param_name = _get_spin_signal_param_name(signal_def,
                                                                  param_index)
                    sim_param_type = signal_def.get_parameter_type(param_index)
                    return (sim_param_type, spin_param_name)
            # Resolve name: next try context class 'cd' class attributes
            var_defs = cd.get_variable_defs()
            if var_defs.has_key(name):
                var_def = var_defs[name]
                sim_type = var_def.get_type()
                #spin_code = var_def.get_name()
                spin_code = "attrs[oid]."+_vardef_to_spin_name(var_def)
                return (sim_type, spin_code)
            raise RuntimeError("Cannot bind variable '"+name+"' in the Class "+
                               get_qualified_name(cd.get_uml_class()))

        raise RuntimeError("Not supported in transition guard: "+str(node))



    def _get_guard_code(cd, guard, trigger):
        """
        Get the promela code for
        the UML Guard 'guard'
        in the context 'cd'
        under the SignalEvent 'trigger' (that can be None).
        """
        assert isinstance(cd, Sim.Core.ClassDef)
        assert guard.oclIsKindOf(UML14.Guard)
        assert guard.expression
        assert guard.expression.body
        assert str(guard.expression.body) != ""
        assert (trigger is None) or trigger.oclIsKindOf(UML14.SignalEvent)

        guard_code = str(guard.expression.body)

        #sys.stderr.write(guard_code+"\n")
        #guard_code = eliminate_chooses(guard_code)
        #sys.stderr.write(guard_code+"\n")

        # Build the parse tree of the guard expression
        tmp = "x = "+guard_code+";"
        tree = actlang.grammar.parse(tmp)
        assert len(tree.api_getTopStatements()) == 1
        ast_node = tree.api_getTopStatements()[0]
        assert isinstance(ast_node, actlang.statement.ExprStatement)
        assert isinstance(ast_node.api_getExpr(),
                          actlang.expr.SimpleAssignmentExpr)
        ast_node = ast_node.api_getExpr().api_getRhsExpr()
        # Actually translate the guard to promela code
        if not is_expr_side_effect_free(ast_node):
            raise RuntimeError("The guard of '"+
                               get_qualified_name(guard.__parent__())+"' "
                               "is not side-effect free: "+guard_code)
        (guard_t,guard_s) = _trans_guard(cd, trigger, ast_node)
        assert isinstance(guard_t, Sim.Core.SimType_Boolean)
        return "("+guard_s+")"



    def _gen_trigger_parameter_instantiation(f, indent, cd, trigger):
        """
        Generate promela code that copies the values of trigger parameters
        from temporary to real variables.
        """
        assert type(f) is file
        assert isinstance(indent, Indent)
        assert isinstance(cd, Sim.Core.ClassDef)
        assert trigger.oclIsKindOf(UML14.SignalEvent)

        if not trigger.parameter:
            return

        uml_signal = trigger.signal
        signal_def = sim.get_signal_def(uml_signal)

        # Add a comment in the promela code
        f.write(str(indent)+"/* Instantiate trigger '"+
                signal_def.get_name()+"(")
        f.write(",".join([str(p.name) for p in trigger.parameter]))
        f.write(")' */\n")
        
        for (param_index, param) in enumerate(trigger.parameter):
            param_name = str(param.name)
            if param_name == "":
                # parameters with no name are ignored
                continue
            spin_param_name = _get_spin_signal_param_name(signal_def,
                                                          param_index)
            sim_param_type = signal_def.get_parameter_type(param_index)
            var_defs = cd.get_variable_defs()
            if not var_defs.has_key(param_name):
                raise RuntimeError("Class '"+cd.get_name()+
                                   "' has no attribute or association '"+
                                   param_name+"'")
            var_def = var_defs[param_name]
            assert var_def.get_type().compare(sim_param_type) == 0
            spin_var = "attrs[oid]."+_vardef_to_spin_name(var_def)
            f.write(str(indent)+spin_var+" = "+spin_param_name+";\n")
            f.write(str(indent)+spin_param_name+" = "+
                    _sim_type_to_spin_default_value(sim_param_type)+";\n")



    def _gen_effect_code(f, indent, cd, effect):
        """
        Translate the Jumbala code in 'effect' (if any) into Promela code.
        The context (class) is given by the ClassDef 'cd' and
        'effect' should be either None, an UML 1.4 Action or
        an UML1.4 UninterpretedAction.
        """
        assert type(f) is file
        assert isinstance(indent, Indent)
        assert isinstance(cd, Sim.Core.ClassDef)
        if not effect:
            return
        assert (effect.oclIsKindOf(UML14.Action) or
                effect.oclIsKindOf(UML14.UninterpretedAction))
        assert effect.script
        assert effect.script.body

        # Build the parse tree and translate to promela code
        tree = actlang.grammar.parse(str(effect.script.body))
        spin_effects = []
        for ast_node in tree.api_getTopStatements():
            (dummy,s) = _trans_effect(cd, ast_node)
            spin_effects.append(s)
        f.write(str(indent)+"/* The effect '"+str(effect.script.body)+"' */\n")
        for spin_effect in spin_effects:
            f.write(str(indent)+spin_effect+"\n")



    def _gen_state_configuration_code(f, indent, state_set):
        assert type(f) is file
        assert isinstance(indent, Indent)
        assert isinstance(state_set, Set)

        for r in [r for r in state_set if
                  r.oclIsKindOf(UML14.CompositeState) and not r.isConcurrent]:
            (spin_region_index,spin_region_type) = uml_composite_state_to_index_type[r]
            found = False
            for s in r.subvertex:
                if not s in state_set:
                    continue
                assert not found
                found = True
                f.write(str(indent)+"state"+str(spin_region_index)+" = "+
                        _uml_statevertex_to_spin_name(s, False)+";\n");



    def _gen_successor_configuration_code(f, indent, t):
        assert isinstance(f, file)
        assert isinstance(indent, Indent)
        assert t.oclIsKindOf(UML14.Transition)

        lca = Utils.StateMachines.get_least_common_ancestor(t.source, t.target)
        assert lca.oclIsKindOf(UML14.CompositeState)
        if lca.isConcurrent:
            assert lca.container
            assert lca.container.oclIsKindOf(UML14.CompositeState)
            assert not lca.container.isConcurrent
            lca = lca.container
        assert lca.oclIsKindOf(UML14.CompositeState)
        assert not lca.isConcurrent

        state_set = Set()
        # Upwards closure: include target and its containers up to and
        # including lca to the state_set
        s = t.target
        while s != lca:
            state_set.add(s)
            assert s.container
            s = s.container
        state_set.add(lca)
        # Downwards closure, add missing initial states
        for cs in [cs for cs in state_set if
                   cs.oclIsKindOf(UML14.CompositeState) and cs.isConcurrent]:
            for r in cs.subvertex:
                assert r.oclIsKindOf(UML14.CompositeState)
                assert not r.isConcurrent
                if r in state_set:
                    continue
                state_set.add(r)
                initial_states = [s for s in r.subvertex if
                                  s.oclIsKindOf(UML14.Pseudostate) and
                                  s.kind == "initial"]
                assert len(initial_states) == 1
                initial_state = initial_states[0]
                assert not (initial_state in state_set)
                state_set.add(initial_state)
        # For each region in the state set, check whether a subvertex is also
        # included and include the default initial state if none was.
        for r in [r for r in state_set if
                  r.oclIsKindOf(UML14.CompositeState) and not r.isConcurrent]:
             existing_states = [s for s in r.subvertex if
                                s in state_set]
             assert len(existing_states) <= 1
             if existing_states:
                 continue
             initial_states = [s for s in r.subvertex if
                               s.oclIsKindOf(UML14.Pseudostate) and
                               s.kind == "initial"]
             assert len(initial_states) == 1
             initial_state = initial_states[0]
             assert not (initial_state in state_set)
             state_set.add(initial_state)
           
        f.write(str(indent)+"/* Make the successor state configuration */\n")
        f.write(str(indent)+"d_step {\n")
        indent.push()
        _gen_state_configuration_code(f, indent, state_set)

        # Reset regions that were exited but not re-entered.
        # Not really needed but may reduce the promela state space.
        for r in [r for r in get_elements_of_type(lca, UML14.CompositeState) if
                  (not r.isConcurrent) and not (r in state_set)]:
            (spin_region_index,spin_region_type) = uml_composite_state_to_index_type[r]
            f.write(str(indent)+"state"+str(spin_region_index)+" = 0;\n")
        indent.pop()
        f.write(str(indent)+"}\n")



    def _gen_defer_code(f, indent, cd, state):
        """
        Generate code for deferring events in the State 'state'.
        Returns True if any code was generated (i.e. if there were deferrable
        events).
        'cd' is the ClassDef of the StateMachine of 'state'.
        """
        assert isinstance(f, file)
        assert isinstance(indent, Indent)
        assert isinstance(cd, Sim.Core.ClassDef)
        assert state.oclIsKindOf(UML14.State)

        if not state.deferrableEvent:
            return False
        f.write(str(indent)+"if\n")
        for defer_event in state.deferrableEvent:
            assert defer_event.oclIsKindOf(UML14.SignalEvent)
            assert defer_event.signal
            f.write(str(indent)+":: (trigger == "+
                    signal_to_spin_name(defer_event.signal)+") ->\n")
            indent.push()
            f.write(str(indent)+"/* Defer event */\n")
            f.write(str(indent)+"defer_message(defer_chan);\n")
            f.write(str(indent)+"printf(\"proco : DEFER "+
                    " class "+str(cd.get_uml_class().__XMIid__())+
                    " obj %d\\n\", oid);\n")
            f.write(str(indent)+"goto eval_triggered;\n")
            indent.pop()
        f.write(str(indent)+":: else -> skip;\n")
        f.write(str(indent)+"fi;\n")
        return True



    def _trans_state_triggered(f, indent, cd, vertex):
        """
        Generate code for triggered transitions.
        """
        assert isinstance(f, file)
        assert isinstance(indent, Indent)
        assert isinstance(cd, Sim.Core.ClassDef)
        assert vertex.oclIsKindOf(UML14.StateVertex)

        if vertex.oclIsKindOf(UML14.CompositeState):
            if not vertex.isConcurrent:
                (spin_region_index,X) = uml_composite_state_to_index_type[vertex]
                f.write(str(indent)+"if\n")
                for subvertex in vertex.subvertex:
                    if subvertex.oclIsKindOf(UML14.Pseudostate):
                        continue
                    f.write(str(indent)+":: (state"+str(spin_region_index)+
                            " == "+
                            _uml_statevertex_to_spin_name(subvertex, is_completion_sensitive(subvertex))+
                            ") ->\n")
                    indent.push()
                    _trans_state_triggered(f, indent, cd, subvertex)
                    indent.pop()
                f.write(str(indent)+":: else -> skip;\n")
                f.write(str(indent)+"fi;\n")
            else:
                for subvertex in vertex.subvertex:
                    assert subvertex.oclIsKindOf(UML14.CompositeState)
                    assert not subvertex.isConcurrent
                    _trans_state_triggered(f, indent, cd, subvertex)

            # Generate code for outgoing triggered transitions, if any
            outgoing_triggered = [t for t in vertex.outgoing if t.trigger]
            if outgoing_triggered:
                f.write(str(indent)+"if\n")
                for t in outgoing_triggered:
                    f.write(str(indent)+":: (trigger == "+
                            signal_to_spin_name(t.trigger.signal)+")")
                    if t.guard:
                        f.write(" && "+_get_guard_code(cd, t.guard, t.trigger))
                    f.write(" -> \n");
                    indent.push()
                    f.write(str(indent)+"/* Execute transition "+
                            str(t.name)+" */\n")
                    f.write(str(indent)+"printf(\"proco : EXEC "+
                            str(t.__XMIid__())+" class "+
                            str(cd.get_uml_class().__XMIid__())+
                            " obj %d\\n\", oid);\n")
                    _gen_trigger_parameter_instantiation(f, indent,
                                                         cd, t.trigger)
                    _gen_effect_code(f, indent, cd, t.effect)
                    _gen_successor_configuration_code(f, indent, t)
                    f.write(str(indent)+"goto eval_completion;\n")
                    indent.pop()
                f.write(str(indent)+":: else -> skip;\n")
                f.write(str(indent)+"fi;\n")
            # Deferrable events
            _gen_defer_code(f, indent, cd, vertex)


        elif(vertex.oclIsKindOf(UML14.CompositeState) and
             vertex.isConcurrent):
            assert False and "Concurrent composite states not yet handled"
            
        elif vertex.oclIsKindOf(UML14.SimpleState):
            has_some_effect = False
            # Generate code for outgoing triggered transitions, if any
            outgoing_triggered = [t for t in vertex.outgoing if t.trigger]
            if outgoing_triggered:
                has_some_effect = True
                f.write(str(indent)+"if\n")
                for t in outgoing_triggered:
                    uml_signal = t.trigger.signal
                    signal_def = sim.get_signal_def(uml_signal)
                    f.write(str(indent)+":: (trigger == "+
                            signal_to_spin_name(uml_signal)+")")
                    if t.guard:
                        f.write(" && "+_get_guard_code(cd, t.guard, t.trigger))
                    f.write(" -> \n");
                    indent.push()
                    f.write(str(indent)+"/* Execute transition "+
                            str(t.name)+" */\n")
                    f.write(str(indent)+"printf(\"proco : EXEC "+
                            str(t.__XMIid__())+" class "+
                            str(cd.get_uml_class().__XMIid__())+
                            " obj %d\\n\", oid);\n")
                    _gen_trigger_parameter_instantiation(f, indent,
                                                         cd, t.trigger)
                    _gen_effect_code(f, indent, cd, t.effect)
                    _gen_successor_configuration_code(f, indent, t)
                    f.write(str(indent)+"goto eval_completion;\n")
                    indent.pop()
                f.write(str(indent)+":: else -> skip;\n")
                f.write(str(indent)+"fi;\n")
            # Deferrable events
            has_defer_code = _gen_defer_code(f, indent, cd, vertex)
            has_some_effect = has_some_effect or has_defer_code
            if not has_some_effect:
                f.write(str(indent)+"skip;\n")

        elif vertex.oclIsKindOf(UML14.FinalState):
            assert not vertex.outgoing
            assert not vertex.deferrableEvent
            f.write(str(indent)+"/* Signal events cannot be consumed "+
                    "in a final state, do nothing */\n")
            f.write(str(indent)+"skip;\n")

        else:
            f.write(str(indent)+str(vertex)+"TO DO!!!\n")
            assert False





    def _trans_state_completion1(f, indent, cd, vertex, condprefix = []):
        """
        Generate the promela code for transitions leaving pseudostates,
        that is, those that can be enabled in a state configuration
        that is 'in compound transition'.
        """
        assert isinstance(f, file)
        assert isinstance(indent, Indent)
        assert isinstance(cd, Sim.Core.ClassDef)
        assert vertex.oclIsKindOf(UML14.StateVertex)
        assert isinstance(condprefix, list)

        if(vertex.oclIsKindOf(UML14.CompositeState) and
           not vertex.isConcurrent):
            (spin_region_index,X) = uml_composite_state_to_index_type[vertex]
            for subvertex in vertex.subvertex:
                if subvertex.oclIsKindOf(UML14.SimpleState):
                    continue
                if subvertex.oclIsKindOf(UML14.FinalState):
                    continue
                condprefix.append("(state"+str(spin_region_index)+" == "+
                                  _uml_statevertex_to_spin_name(subvertex,
                                                                False)+")")
                _trans_state_completion1(f, indent, cd, subvertex, condprefix)
                condprefix.pop()

        elif(vertex.oclIsKindOf(UML14.CompositeState) and
           vertex.isConcurrent):
            for subvertex in vertex.subvertex:
                assert subvertex.oclIsKindOf(UML14.CompositeState)
                assert not subvertex.isConcurrent
                _trans_state_completion1(f, indent, cd, subvertex, condprefix)

        elif(vertex.oclIsKindOf(UML14.Pseudostate) and
             vertex.kind in ["initial","choice"]):
            f.write(str(indent)+":: "+(" && ".join(condprefix))+" -> \n")
            indent.push()            
            f.write(str(indent)+"if\n")
            assert vertex.outgoing
            for t in vertex.outgoing:
                assert not t.trigger
                f.write(str(indent)+":: ")
                if t.guard:
                    f.write(_get_guard_code(cd, t.guard, None))
                else:
                    f.write("true")
                f.write(" -> \n");
                indent.push()
                f.write(str(indent)+"/* Execute transition "+
                        str(t.name)+" */\n")
                f.write(str(indent)+"printf(\"proco : EXEC "+
                        str(t.__XMIid__())+" class "+
                        str(cd.get_uml_class().__XMIid__())+
                        " obj %d\\n\", oid);\n")
                _gen_effect_code(f, indent, cd, t.effect)
                _gen_successor_configuration_code(f, indent, t)
                f.write(str(indent)+"goto eval_completion;\n")
                indent.pop()
            f.write(str(indent)+":: else ->\n")
            indent.push()
            f.write(str(indent)+"/* Error: no completion transition "
                    "enabled in a pseudo-state */\n")
            f.write(str(indent)+"assert(0);\n")
            indent.pop()
            f.write(str(indent)+"fi;\n")
            indent.pop()

        elif vertex.oclIsKindOf(UML14.SimpleState):
            pass

        elif vertex.oclIsKindOf(UML14.FinalState):
            pass

        else:
            f.write(str(indent)+str(vertex)+"TO DO!!!\n")
            assert False




    def _get_initial_state(region):
        """Get the initial Pseudostate of 'region'."""
        assert region.oclIsKindOf(UML14.CompositeState)
        assert not region.isConcurrent
        initial_states = [s for s in region.subvertex if
                          s.oclIsKindOf(UML14.Pseudostate) and
                          s.kind == "initial"]
        assert len(initial_states) == 1
        return initial_states[0]


    def _get_final_states(region):
        """Get the FinalStates of 'region'."""
        assert region.oclIsKindOf(UML14.CompositeState)
        assert not region.isConcurrent
        final_states = [s for s in region.subvertex if
                        s.oclIsKindOf(UML14.FinalState)]
        return final_states


    def _gen_quiescing_code(f, indent, cd, vertex):
        assert(vertex.oclIsKindOf(UML14.SimpleState) or
               vertex.oclIsKindOf(UML14.CompositeState))
        assert vertex.container
        print >> f, str(indent)+\
              "/* No completion transition enabled, quiesce */"
        print >> f, str(indent)+\
              "printf(\"proco : QUIESCE "+\
              "class "+str(cd.get_uml_class().__XMIid__())+" "+\
              "obj %d "+\
              "state "+str(vertex.__XMIid__())+\
              "\\n\", oid);"
        (spin_region_index,X) = uml_composite_state_to_index_type[vertex.container]
        print >> f, str(indent)+\
              "state"+str(spin_region_index)+" = "+\
              _uml_statevertex_to_spin_name(vertex, True)+";"
        print >> f, str(indent)+"goto eval_completion;"


    def _trans_state_completion2(f, indent, cd, vertex, condprefix = []):
        assert isinstance(indent, Indent)
        assert isinstance(cd, Sim.Core.ClassDef)
        assert vertex.oclIsKindOf(UML14.StateVertex)
        assert type(condprefix) is list

        if(vertex.oclIsKindOf(UML14.CompositeState) and
           not vertex.isConcurrent):
            (spin_region_index,X) = uml_composite_state_to_index_type[vertex]
            for subvertex in vertex.subvertex:
                if subvertex.oclIsKindOf(UML14.Pseudostate):
                    continue
                if subvertex.oclIsKindOf(UML14.FinalState):
                    assert not subvertex.outgoing
                    continue
                if(subvertex.oclIsKindOf(UML14.SimpleState) and
                   not has_outgoing_completion_transitions(subvertex)):
                    continue
                condprefix.append("(state"+str(spin_region_index)+" == "+
                                  _uml_statevertex_to_spin_name(subvertex,
                                                                False)+
                                  ")")
                _trans_state_completion2(f, indent, cd, subvertex, condprefix)
                condprefix.pop()

            outgoing_completion = get_outgoing_completion_transitions(vertex)

            if not outgoing_completion:
                assert not is_completion_sensitive(vertex)
                return

            final_states = _get_final_states(vertex)
            if not final_states:
                return
            pred = "false"
            for fs in final_states:
                pred += " || (state"+str(spin_region_index)+" == "+\
                        _uml_statevertex_to_spin_name(fs, False)+")"
            pred = (" && ".join(condprefix)) + " && ("+pred+")"

            f.write(str(indent)+":: "+pred+" ->\n")
            indent.push()

            f.write(str(indent)+"if\n")
            for t in outgoing_completion:
                assert not t.trigger
                f.write(str(indent)+":: ")
                if t.guard:
                    f.write(_get_guard_code(cd, t.guard, None))
                else:
                    f.write("true")
                f.write(" -> \n");
                indent.push()
                f.write(str(indent)+"/* Execute transition "+
                        str(t.name)+" */\n")
                f.write(str(indent)+"printf(\"proco : EXEC "+
                        str(t.__XMIid__())+" class "+
                        str(cd.get_uml_class().__XMIid__())+
                        " obj %d\\n\", oid);\n")
                _gen_effect_code(f, indent, cd, t.effect)
                _gen_successor_configuration_code(f, indent, t)
                f.write(str(indent)+"goto eval_completion;\n")
                indent.pop()
            # Code for the quiescing step
            f.write(str(indent)+":: else ->\n")
            indent.push()
            if is_completion_sensitive(vertex):
                _gen_quiescing_code(f, indent, cd, vertex)
            else:
                f.write(str(indent)+"/* Should never get here */\n")
                f.write(str(indent)+"assert(0);\n")
            indent.pop()
            f.write(str(indent)+"fi;\n")


        elif(vertex.oclIsKindOf(UML14.CompositeState) and
             vertex.isConcurrent):
            for region in vertex.subvertex:
                assert region.oclIsKindOf(UML14.CompositeState)
                assert not region.isConcurrent
                _trans_state_completion2(f, indent, cd, region, condprefix)

            outgoing_completion = get_outgoing_completion_transitions(vertex)

            if not outgoing_completion:
                assert not is_completion_sensitive(vertex)
                return

            pred = "true"
            for region in vertex.subvertex:
                assert region.oclIsKindOf(UML14.CompositeState)
                assert not region.isConcurrent
                (spin_region_index,X) = uml_composite_state_to_index_type[region]
                final_states = _get_final_states(region)
                pred += "&& (false"
                for fs in final_states:
                    pred += " || (state"+str(spin_region_index)+" == "+\
                            _uml_statevertex_to_spin_name(fs, False)+")"
                pred += ")"
            pred = (" && ".join(condprefix)) + " && ("+pred+")"

            f.write(str(indent)+":: "+pred+" ->\n")
            indent.push()

            f.write(str(indent)+"if\n")
            for t in outgoing_completion:
                assert not t.trigger
                f.write(str(indent)+":: ")
                if t.guard:
                    f.write(_get_guard_code(cd, t.guard, None))
                else:
                    f.write("true")
                f.write(" -> \n");
                indent.push()
                f.write(str(indent)+"/* Execute transition "+
                        str(t.name)+" */\n")
                f.write(str(indent)+"printf(\"proco : EXEC "+
                        str(t.__XMIid__())+" class "+
                        str(cd.get_uml_class().__XMIid__())+
                        " obj %d\\n\", oid);\n")
                _gen_effect_code(f, indent, cd, t.effect)
                _gen_successor_configuration_code(f, indent, t)
                f.write(str(indent)+"goto eval_completion;\n")
                indent.pop()
            # Code for the quiescing step
            f.write(str(indent)+":: else ->\n")
            if is_completion_sensitive(vertex):
                indent.push()
                _gen_quiescing_code(f, indent, cd, vertex)
            else:
                print >> f, str(indent)+"/* Should never get here */"
                print >> f, str(indent)+"assert(0);"
            indent.pop()
            f.write(str(indent)+"fi;\n")


        elif(vertex.oclIsKindOf(UML14.Pseudostate) and
             vertex.kind in ["initial","choice"]):
            pass


        elif vertex.oclIsKindOf(UML14.SimpleState):
            outgoing_completion = [t for t in vertex.outgoing if not t.trigger]
            if not outgoing_completion:
                assert not is_completion_sensitive(vertex)
                f.write(str(indent)+"skip;\n")
                return

            f.write(str(indent)+":: "+(" && ".join(condprefix))+" ->\n")
            indent.push()
            f.write(str(indent)+"if\n")
            for t in outgoing_completion:
                assert not t.trigger
                f.write(str(indent)+":: ")
                if t.guard:
                    f.write(_get_guard_code(cd, t.guard, None))
                else:
                    f.write("true")
                f.write(" -> \n");
                indent.push()
                f.write(str(indent)+"/* Execute transition "+
                        str(t.name)+" */\n")
                f.write(str(indent)+"printf(\"proco : EXEC "+
                        str(t.__XMIid__())+" class "+
                        str(cd.get_uml_class().__XMIid__())+
                        " obj %d\\n\", oid);\n")
                _gen_effect_code(f, indent, cd, t.effect)
                _gen_successor_configuration_code(f, indent, t)
                f.write(str(indent)+"goto eval_completion;\n")
                indent.pop()
            # Code for the quiescing step
            f.write(str(indent)+":: else ->\n")
            indent.push()
            if is_completion_sensitive(vertex):
                _gen_quiescing_code(f, indent, cd, vertex)
            else:
                print >> f, str(indent)+"/* Should never get here */"
                print >> f, str(indent)+"assert(0);"
            indent.pop()
            f.write(str(indent)+"fi;\n")
            indent.pop()

        elif vertex.oclIsKindOf(UML14.FinalState):
            assert not vertex.outgoing
            pass

        else:
            f.write(str(indent)+str(vertex)+"TO DO!!!\n")
            assert False




    def _get_state_predicate(vertex):
        assert vertex.oclIsKindOf(UML14.StateVertex)
        assert vertex.container

        s = ""
        sep = ""
        while vertex.container:
            (spin_region_index,X) = uml_composite_state_to_index_type[vertex.container]
            s += sep + "(state"+str(spin_region_index)+" == "+\
                 _uml_statevertex_to_spin_name(vertex, False)+")"
            sep = " && "
            vertex = vertex.container
            while vertex.container and vertex.container.isConcurrent:
                vertex = vertex.container
        return s


    def _gen_external_signal_reception_code(f, indent, class_def):
        """
        Generate promela code for non-deterministic reception of
        external signals.
        """
        assert isinstance(indent, Indent)
        assert isinstance(class_def, Sim.Core.ClassDef)

        uml_sm = class_def.get_uml_state_machine()

        signal_defs = sim.get_signal_defs()
        signal_defs.sort(lambda a,b: cmp(a.get_name(),b.get_name()))
        for signal_def in sim.get_signal_defs():
            if not signal_def.is_external():
                continue
            uml_signal = signal_def.get_uml_signal()

            # Get transitions triggered with the external signal
            uml_trans = [t for t
                         in get_elements_of_type(uml_sm,UML14.Transition)
                         if (t.trigger and t.trigger.signal == uml_signal)]
            xmi_sort(uml_trans)

            for t in uml_trans:
                spin_guard = ""
                if t.guard:
                    spin_guard = " && "+\
                                 _get_guard_code(class_def, t.guard, t.trigger)
                f.write(str(indent)+":: !overflowed && "+
                        "("+_get_state_predicate(t.source)+")"+
                        spin_guard+" ->\n")
                indent.push()
                f.write(str(indent)+"/* Non-deterministic reception of "+
                        "external signal "+signal_def.get_name()+"*/\n")
                f.write(str(indent)+"trigger = "+
                        signal_to_spin_name(signal_def)+";\n")
                if signal_def.get_nof_parameters() > 0:
                    raise RuntimeError("External signals with parameters "+
                                       "not yet handled")
                indent.pop()

    def _generate_external_signal_reception_code_OLD(f, indent, class_def):
        """
        Generate promela code for non-deterministic reception of
        external signals.
        """
        assert isinstance(indent, Indent)
        assert isinstance(class_def, Sim.Core.ClassDef)

        uml_sm = class_def.get_uml_state_machine()

        for signal_def in sim.get_signal_defs():
            if not signal_def.is_external():
                continue
            uml_signal = signal_def.get_uml_signal()
            state_vertices = [sv for sv in
                              get_elements_of_type(uml_sm,UML14.StateVertex) if
                              len([t for t in sv.outgoing if (t.trigger and t.trigger.signal == uml_signal)]) > 0]
            if not state_vertices:
                continue
            f.write(str(indent)+":: !overflowed && ("+
                    " || ".join(["("+_get_state_predicate(sv)+")" for sv in
                                 state_vertices])+") ->\n")
            indent.push()
            f.write(str(indent)+"/* Non-deterministic reception of "+
                    "external signal "+signal_def.get_name()+"*/\n")
            f.write(str(indent)+"trigger = "+
                    signal_to_spin_name(signal_def)+";\n")
            if signal_def.get_nof_parameters() > 0:
                raise RuntimeError("External signals with parameters not handled")
            indent.pop()

    #
    # Generate a proctype for each active UML Class
    #
    for class_def in sim.get_class_defs():
        # Is this Class active?
        uml_sm = class_def.get_uml_state_machine()
        if not uml_sm:
            continue
        
        sm_has_deferrable_events = len([s for s in get_elements_of_type(uml_sm, UML14.State) if s.deferrableEvent]) > 0

        f.write(str(indent)+"proctype "+_class_def_spin_name(class_def)+
                "("+spin_object_id_type+" oid) {\n")
        indent.push()

        f.write(str(indent)+"chan input_chan = input[oid];\n")
        #f.write(str(indent)+"xr input_chan;\n")
        if sm_has_deferrable_events:
            f.write(str(indent)+"chan defer_chan = defer[oid];\n")
            f.write(str(indent)+"xr defer_chan;\n")
            f.write(str(indent)+"xs defer_chan;\n")
            f.write(str(indent)+"local "+spin_queue_size_type+" qcj;\n")
        f.write(str(indent)+"local "+spin_trigger_type+" trigger;\n")
        f.write(str(indent)+"/* Temp variables for signal parameters */\n")
        for i in xrange(0, signal_parameter_allocator.get_nof_indices()):
            f.write(str(indent)+
                    "local "+signal_parameter_allocator.get_type(i)+
                    " param"+str(i)+";\n")

        # Add a local state variable for each region
        uml_nccss = [s for s in get_elements_of_type(uml_sm, UML14.CompositeState) if not s.isConcurrent]
        for uml_nccs in uml_nccss:
            (spin_region_index, spin_region_type) = uml_composite_state_to_index_type[uml_nccs]
            f.write(str(indent)+spin_region_type+" state"+
                    str(spin_region_index)+";\n")
        # Make initial configuration
        assert uml_sm.top.oclIsKindOf(UML14.CompositeState)
        assert not uml_sm.top.isConcurrent
        state_set = Set()
        state_set.add(uml_sm.top)
        uml_initial_states = [s for s in uml_sm.top.subvertex if
                              s.oclIsKindOf(UML14.Pseudostate) and
                              s.kind == "initial"]
        assert len(uml_initial_states) == 1
        state_set.add(uml_initial_states[0])
        _gen_state_configuration_code(f, indent, state_set)

        indent.pop()
        f.write(str(indent)+"eval_completion:\n")
        indent.push()
        f.write(str(indent)+"atomic {\n")
        indent.push()

        f.write(str(indent)+"/* If a queue has overflowed in some process,\n")
        f.write(str(indent)+" * go to the end state */\n")
        f.write(str(indent)+"if\n")
        f.write(str(indent)+":: overflowed -> goto end_queue_overflow;\n")
        f.write(str(indent)+":: else -> skip;\n")
        f.write(str(indent)+"fi;\n")

        f.write(str(indent)+"/* Translation of completion transitions "+
                "leaving from pseudostates */\n")
        f.write(str(indent)+"if\n")
        _trans_state_completion1(f, indent, class_def, uml_sm.top)
        f.write(str(indent)+":: else;\n")
        f.write(str(indent)+"fi;\n")

        f.write(str(indent)+"/* Translation of completion transitions "+
                "leaving from non-pseudostates */\n")
        f.write(str(indent)+"if\n")
        _trans_state_completion2(f, indent, class_def, uml_sm.top)
        f.write(str(indent)+":: else;\n")
        f.write(str(indent)+"fi;\n")

        if sm_has_deferrable_events:
            indent.pop()
            f.write(str(indent)+"flush_defer_queue:\n")
            indent.push()
            f.write(str(indent)+"queue_flush(input_chan, defer_chan);\n")

        indent.pop()
        f.write(str(indent)+"}\n")



        indent.pop()
        f.write(str(indent)+"eval_triggered:\n")
        indent.push()
        f.write(str(indent)+"/* Invariant: the state configuration is stable */\n")
        f.write(str(indent)+"atomic {\n")
        indent.push()
        
        f.write(str(indent)+"/* If a queue has overflowed in some process,\n")
        f.write(str(indent)+" * go to the end state */\n")
        f.write(str(indent)+"if\n")
        f.write(str(indent)+":: overflowed -> goto end_queue_overflow;\n")
        f.write(str(indent)+":: else -> skip;\n")
        f.write(str(indent)+"fi;\n")

        f.write(str(indent)+"/* Event dispatching (message consumption) */\n")
        f.write(str(indent)+"if\n")
        f.write(str(indent)+":: overflowed -> goto end_queue_overflow;\n")
        # Code for reception of a message from the input queue
        f.write(str(indent)+":: (!overflowed && nempty(input_chan)) -> input_chan?trigger")
        for i in xrange(0, signal_parameter_allocator.get_nof_indices()):
            f.write(",param"+str(i))
        f.write(";\n")
        _gen_external_signal_reception_code(f, indent, class_def)
        f.write(str(indent)+"fi;\n")
        _trans_state_triggered(f, indent, class_def, uml_sm.top)
        f.write(str(indent)+"/* Implicit consumption */\n")
        f.write(str(indent)+"printf(\"proco : IMPL class "+
                str(class_def.get_uml_class().__XMIid__())+
                " obj %d\\n\", oid);\n")
        if(opt_check_all_implicit_consumptions or
           class_def.get_name() in opt_check_implicit_consumption_on_classes):
            f.write(str(indent)+"printf(\"proco : ERROR IMPLICITCONSUMPTION\\n\");")
            f.write(str(indent)+"assert(false);\n")
        f.write(str(indent)+"goto eval_triggered;\n")
        indent.pop()
        f.write(str(indent)+"}\n")

        f.write(str(indent)+"goto valid_end;\n")
        f.write(str(indent)+"end_queue_overflow:\n")
        if opt_check_queue_overflows:
            f.write(str(indent)+"  printf(\"proco : ERROR QUEUEOVERFLOW\\n\");")
            f.write(str(indent)+"  assert(false);\n")
        else:
            f.write(str(indent)+"  false;\n")
        f.write(str(indent)+"valid_end: skip;\n")

        indent.pop()
        f.write(str(indent)+"}\n")
        f.write(str(indent)+"\n")


    #
    # Make the init process in promela.
    # It creates all the objects that are alive in
    # the initial state configuration of the model.
    #
    initial_state = sim.trace_get_last_global_state()

    w.ln("init {")
    w.push()
    w.ln("atomic {")
    w.push()
    w.ln("id = 0;") 
    sim_object_to_spin_id = {}
    for object_index in xrange(0, initial_state.get_nof_objects()):
        o = initial_state.get_object(object_index)
        sim_object_to_spin_id[o] = object_index
    for object_index in xrange(0, initial_state.get_nof_objects()):
        o = initial_state.get_object(object_index)
        cd = o.get_class_def()
        uml_object = o.get_uml_object()
        assert uml_object
        w.ln("printf(\"proco : class "+str(cd.get_uml_class().__XMIid__())+
             " obj %d xmi_id "+str(uml_object.__XMIid__())+"\\n\", id);")
        w.ln("assert(id < MAX_IDS);")
        w.ln("run "+_class_def_spin_name(cd)+"(id);")
        for (var_name,var_def) in cd.get_variable_defs().items():
            value = o.get_variable_value(var_name)
            spin_value = None
            if isinstance(value, Sim.Core.SimValue_ObjectRef):
                assert value.is_defined()
                if value.get_value() == "null":
                    spin_value = str(spin_null_object_id)
                else:
                    spin_value = str(sim_object_to_spin_id[value.get_value()])
            elif isinstance(value, Sim.Core.SimValue_Int):
                assert value.is_defined()
                spin_value = str(value.get_value())
            elif isinstance(value, Sim.Core.SimValue_Boolean):
                assert value.is_defined()
                assert value.get_value() in [False, True]
                if value.get_value():
                    spin_value = "true"
                else:
                    spin_value = "false"
            else:
                #print value
                raise RuntimeError("Internal error")
            assert not spin_value is None
            w.ln("attrs[id]."+_vardef_to_spin_name(var_def)+
                 " = "+spin_value+";")
        w.ln("id = id + 1;")
    w.pop()
    w.ln("}")
    w.pop()
    w.ln("}")


class Options:
    """
    The options that can be passed to the analyzer.

    c_compiler
       the name of the C compiler executable [default: cc]
    spin
       the name of the spin executable [default: spin]
    check_assertions
       if True, assertion violations are checked [default: True]
    check_deadlock
       if True, absense of deadlocks is checked [default: True]
    check_implicit_consumption
       if True, implicit consumption of messages is checked
       on all objects [default: False]
    check_implicit_consumption_on_class
       a set of class names on which implicit consumption of messages
       is checked [default: an empty set]
    check_queue_overflows
       if True, queue overflows are checked [default: False]
    queue_size
       the input queue size of objects [default: 2]
    max_search_depth
       the maximum search depth of spin [default: 1000000]
    use_bfs
       if True, spin will use breadth-first search instead of depth-first
       [default: False]
    use_collapse
       if True, spin will use the collapse state compression [default: True]
    use_cex_minimization
       if True, spin will trye to minimize the found counter-example
       [default: False]
    print_cex
       if True, print the counter-example in stdout
    trace_filename
       if a non-empty string, save the counter-example in the SMUML trace
       format in this filename
    """
    def __init__(self):
        self.c_compiler = "cc"
        self.spin = "spin"
        self.check_assertions = True
        self.check_deadlock = True
        self.check_implicit_consumption = False
        self.check_implicit_consumption_on_class = Set()
        self.check_queue_overflows = False
        self.queue_size = 2
        self.max_search_depth = 1000000
        self.use_bfs = False
        self.use_collapse = True
        self.use_cex_minimization = False
        self.print_cex = True
        self.trace_filename = ""
    def __str__(self):
        keys = self.__dict__.keys()
        keys.sort()
        return "\n".join([k+" = "+str(getattr(self, k)) for k in keys])



def analyze(model_filename,
            opts,
            stdout = sys.stdout, stderr = sys.stderr):
    """
    Analyze the UML model in the file 'model_filename' by using spin.
    
    Argument 'opts' should be an instance of Options.

    Returns 0 if no error was found, 1 otherwise.

    Can raise RuntimeErrors.
    """
    assert type(model_filename) is str
    assert isinstance(opts, Options)

    project = coral.loadDocument(model_filename)

    def _error(msg):
        raise RuntimeError(msg)

    def _verb(s):
        if stdout:
            print >> stdout, s

    #
    # Find the model
    #
    models = [m for m in project.contents() if m.oclIsKindOf(UML14.Model)]
    if not models:
        _error("No Model found")
    if len(models) > 1:
        _error("Multiple Models found")
    model = models[0]
    
    #
    # Build a simulator for the model.
    # The simulator is used because it builds handy data structures
    # out of the model.
    #
    sim =  Sim.Core.Simulator(model)

    # Check validity of some options
    existing_class_names = [cd.get_name() for cd in sim.get_class_defs()]
    for class_name in opts.check_implicit_consumption_on_class:
        if not class_name in existing_class_names:
            _error("Option check-implicit-consumption-on-class contains "+
                   "a class name '"+class_name+"' not existing in the model")
    if opts.queue_size < 1:
        _error('Option queue-size must be at least 1')
    if opts.max_search_depth < 1:
        _error('Option max-search-depth must be at least 1')
    if(opts.trace_filename and
       not re.match("^[a-zA-Z0-9_][a-zA-Z0-9_\-\.]*$", opts.trace_filename)):
       _error("Illegal trace file name; does not match the regexp "
              "[a-zA-Z0-9_][a-zA-Z0-9_-.]*")


    #
    # Do the translation.
    #
    _verb("Translating the model to promela file tmp_op...")
    f = open("tmp_op", "w")
    translate(sim,
              f,
              opts.queue_size,
              opts.check_assertions,
              opts.check_implicit_consumption,
              opts.check_implicit_consumption_on_class,
              opts.check_queue_overflows)
    f.close()

    #
    # Apply spin
    #
    _verb("Running spin on tmp_op to produce pan.c...")
    result = os.system(opts.spin+" -a tmp_op")
    if result != 0:
        _error("Running spin failed")

    #
    # Compile the actual verifier pan.c
    #
    compile_options = ""
    if opts.use_bfs:
        compile_options += "-DBFS "
    if opts.use_collapse:
        compile_options += "-DCOLLAPSE "
    if opts.use_cex_minimization:
        compile_options += "-DREACH "

    _verb("Compiling pan.c...")
    compile_command = opts.c_compiler+" -o pan -DSAFETY -DVECTORSZ=4096 "+\
                      compile_options+"pan.c"
    _verb("Executing: "+compile_command)
    result = os.system(compile_command)
    if result != 0:
        _error("Compiling pan.c failed")

    #
    # Run the compiled verifier
    #
    pan_options = ""
    if not opts.check_deadlock:
        pan_options += "-E "
    if opts.use_cex_minimization:
        pan_options += "-i "
    pan_options += "-m"+str(opts.max_search_depth)+" "

    # Remove the old spin trail file
    os.system("rm -f tmp_op.trail")

    redirect = "> tmp_op_pan.out 2>&1"
    if stdout:
        redirect = "2>&1 | tee tmp_op_pan.out"
    pan_command = "./pan "+pan_options
    _verb("Running pan...")
    _verb("Executing: "+pan_command)
    result = os.system("/usr/bin/time -o tmp_op.time "+pan_command+" "+
                       redirect)
    if result != 0:
        _error("Running pan failed")

    #
    # Output time that was consumed by the verifier
    #
    f = open("tmp_op.time","r")
    for l in f:
        _verb(l[:-1])
    f.close()

    #
    # Analyze what happened
    #
    
    # No trail -> no error -> exit
    if not os.path.exists("tmp_op.trail"):
        _verb("Spin found no errors")
        return 0

    # A trail -> an error -> simulate the trace
    _verb("An error was found")
    
    _verb("Simulating the spin counter-example...")
    result = os.system(opts.spin+" -t tmp_op > tmp_op.spin")

    _verb("Translating the spin counter-example to a SMUML trace...")
    try:
        out = open("tmp_op.gtrace", "w")
        nof_actions = proco_trace_to_generic_trace.translate(model_filename,
                                                             "tmp_op.spin",
                                                             out)
        out.close()
        _verb("The trace has %d actions" % nof_actions)
    except Exception, e:
        _error("Translating the spin counter-example to "+
               "a SMUML trace failed: "+ str(e))

    #result = os.system("python proco_trace_to_generic_trace.py "+
    #                   model_filename+" tmp_op.spin > tmp_op.gtrace")
    #if result != 0:
    #    _error("Translating the spin counter-example to generic form failed")

    if opts.trace_filename:
        _verb("Saving the SMUML trace in file '"+opts.trace_filename+"'")
        os.system("cp tmp_op.gtrace "+opts.trace_filename)

    if opts.print_cex:
        _verb("Simulating the counter-example in the SMUML simulator...")
        try:
            sim = simulate_generic_trace.simulate("tmp_op.gtrace",
                                                  stdout = stdout,
                                                  stderr = stderr)
            assert sim
            _verb("The counter-example is:")
            sim.api_print_trace(f = sys.stdout)
        except Exception, e:
            pass #_error("Simulating the counter-example failed: "+str(e))


    return 1



if __name__ == '__main__':
    p = optparse.OptionParser(usage='python %prog [options] infile.xmi',
                              version='newproco ' + __version__)
    p.add_option('--check-assertions', action='store_true', default=False,
                 help='check Jumbala assertion errors')
    p.add_option('--check-deadlock', action='store_true', default=False,
                 help='check deadlocks')
    p.add_option('--check-implicit-consumption', action='store_true',
                 default=False, help='check implicit consumption')
    p.add_option('--check-implicit-consumption-on-class', action='append',
                 default=[], metavar='CLASSNAME',
                 help='check whether an instance of a class "CLASSNAME" '
                 'can perform implicit consumption')
    p.add_option('--check-queue-overflows', action='store_true', default=False,
                 help='check queue overlows')
    p.add_option('--queue-size', type='int', default=2, metavar='N',
                 help='size of event queues [default: 2]')
    p.add_option('--max-search-depth', type='long', default=1000000, metavar='N',
                 help='the maximum search depth of spin [default: 1000000]')
    p.add_option('--use-bfs', action='store_true',
                 default=False, help='use breadth first search in spin')
    p.add_option('--use-collapse', action='store_true',
                 default=False, help='use the collapse mode of spin')
    p.add_option('--use-cex-minimization', action='store_true',
                 default=False, help='use the counter-example minimization '+
                 'option -i with spin (only relevant when --use-bfs '+
                 'is not used)')
    p.add_option('--spin', type='string', default="spin", metavar='FILE',
                 help='the spin executable [default: spin]')
    p.add_option('--compiler', type='string', default="cc", metavar='FILE',
                 help='the C compiler executable [default: cc]')
    p.add_option('--trace-filename', type='string', default=None,
                 metavar='FILE',
                 help='SMUML trace output filename')
    p.add_option('--dont-print-cex', action='store_true',
                 default=False, help="do not print the counter-example")
    p.add_option('--verbose', action='store_true',
                 default=False, help="produce some verbose output")
    (opts, args) = p.parse_args()
    if len(args) != 1:
        p.error('wrong number of arguments')
    if opts.queue_size < 1:
        p.error('Option queue-size must be at least 1')
    if opts.max_search_depth < 1:
        p.error('Option max-search-depth must be at least 1')

    model_filename = args[0]

    options = Options()
    options.c_compiler = opts.compiler
    options.spin = opts.spin
    options.check_assertions = opts.check_assertions
    options.check_deadlock = opts.check_deadlock
    options.check_implicit_consumption = opts.check_implicit_consumption
    options.check_implicit_consumption_on_class = opts.check_implicit_consumption_on_class
    options.check_queue_overflows = opts.check_queue_overflows
    options.queue_size = opts.queue_size
    options.max_search_depth = opts.max_search_depth
    options.use_bfs = opts.use_bfs
    options.use_collapse = opts.use_collapse
    options.use_cex_minimization = opts.use_cex_minimization
    options.print_cex = not opts.dont_print_cex
    options.trace_filename = opts.trace_filename

    verbose_file = None
    if opts.verbose:
        verbose_file = sys.stdout

    if verbose_file:
        def _v(s): print >> verbose_file, s
        _v("Checking for:")
        _v("  Runtime errors (null pointer references etc)")
        if opts.check_assertions:
            _v("  Assertion violations")
        if opts.check_deadlock:
            _v("  Deadlocks")
        if opts.check_implicit_consumption:
            _v("  Implicit consumption on all classes")
        elif opts.check_implicit_consumption_on_class:
            _v("  Implicit consumption on classes: "+
               (", ".join(opts.check_implicit_consumption_on_class)))
        if opts.check_queue_overflows:
            _v("  Queue overflows")

    try:
        result = analyze(model_filename,
                         options,
                         stdout = verbose_file)
    except RuntimeError, e:
        print "Error:", e
        sys.exit(2)

    assert result in [0,1]
    if result == 0:
        print "No error was found"
        print "  (either the model is correct, no property was specified,"
        print "   or spin exceeded its limits,"
        print "   inspect the output with --verbose flag turned on)"
    sys.exit(result)

