#!/usr/bin/python

"""
A translator from traces produced by proco to the generic trace format.

Author: Tommi Junttila

Copyright (c) 2007 Helsinki University of Technology
Released under the GNU General Public License version 2
"""

import re
import sys
import string
import Utils.String
import Utils.Models

def translate(modelfilename, procotracefilename, outfile):
    """
    Translate the trace in the file 'procotracefilename' produced by
    running proco on the UML model 'modelfilename' to
    the generic trace format.
    The generic trace is output to the file stream 'outfile'.
    """
    assert type(modelfilename) is str
    assert type(procotracefilename) is str
    assert type(outfile) is file

    spin_trace = open(procotracefilename, 'r')

    print >> outfile, "Model "+modelfilename

    #
    # Get the initial configuration map from spin_id to object's xmi_id
    #

    line = spin_trace.readline()
    while line != "":
        split_line = string.split(line)

        # Is this line produced by proco?
        if not (split_line[0] == "proco" and split_line[1] == ":"):
            line = spin_trace.readline()
            continue
        split_line.pop(0)
        split_line.pop(0)

        if split_line[0] == "class":
            assert len(split_line) == 6
            assert split_line[2] == "obj"
            assert Utils.String.IsInt(split_line[3])
            assert split_line[4] == "xmi_id"
            spin_object = split_line[1]+"#"+split_line[3]
            uml_object = split_line[5]

            print >> outfile,\
                  "InitialConfiguration object "+spin_object+\
                  " umlobject "+uml_object
        else:
            break

        line = spin_trace.readline()

    #
    # Translate spin actions to generic trace actions
    #
    end_reason = None
    nof_actions = 0
    while line != "":
        split_line = string.split(line)

        if re.match("^spin: .* Error: .*", line):
            end_reason = "RUNTIMEERROR"
            break

        # Is this line produced by proco?
        if not (split_line[0] == "proco" and split_line[1] == ":"):
            line = spin_trace.readline()
            continue
        split_line.pop(0)
        split_line.pop(0)

        if split_line[0] == "EXEC":
            # Execution of a UML transition
            # The line should be of form
            # 'EXEC' trans_xmi_id 'class' class_xmi_id 'obj' n
            assert len(split_line) == 6
            assert split_line[2] == "class"
            assert split_line[4] == "obj"
            assert Utils.String.IsInt(split_line[5])
            transition_xmi_id = split_line[1]
            spin_id = split_line[3], int(split_line[5])

            print >> outfile,\
                  "TransitionExecution "+\
                  "object "+split_line[3]+"#"+split_line[5]+" "+\
                  "transition "+transition_xmi_id
            nof_actions += 1

        elif split_line[0] == "IMPL":
            # Implicit consumption of a message
            # The line should be of form
            # 'IMPL' 'class' class_xmi_id 'obj' n
            assert len(split_line) == 5
            assert split_line[1] == "class"
            assert split_line[3] == "obj"
            assert Utils.String.IsInt(split_line[4])
            spin_id = split_line[2], int(split_line[4])
            
            print >> outfile,\
                  "ImplicitConsumption object "+split_line[2]+"#"+split_line[4]
            nof_actions += 1

        elif split_line[0] == "DEFER":
            # Deferring of the first message in the input queue
            # The line should be of form
            # 'DEFER' 'class' class_xmi_id 'obj' n
            assert len(split_line) == 5
            assert split_line[1] == "class"
            assert split_line[3] == "obj"
            assert Utils.String.IsInt(split_line[4])
            spin_id = split_line[2], int(split_line[4])

            print >> outfile,\
                  "EventDefer object "+split_line[2]+"#"+split_line[4]
            nof_actions += 1

        elif split_line[0] == "QUIESCE":
            # Quiescing
            # The line should be of form
            # 'QUIESCE' 'class' class_xmi_id 'obj' n 'state' state_xmi_id
            assert len(split_line) == 7
            assert split_line[1] == "class"
            assert split_line[3] == "obj"
            assert split_line[5] == "state"
            assert Utils.String.IsInt(split_line[4])
            spin_id = split_line[2], int(split_line[4])

            print >> outfile,\
                  "Quiesce "+\
                  "object "+split_line[2]+"#"+split_line[4]+" "+\
                  "state "+split_line[6]
            nof_actions += 1

        elif split_line[0] == "ERROR":
            assert len(split_line) == 2            
            end_reason = split_line[1]
            break

        else:
            raise RuntimeError("Unknown proco output line: " + line)

        line = spin_trace.readline()

    if not end_reason:
        end_reason = "DEADLOCK"

    #print >> outfile, "Error "+end_reason
    return nof_actions
   

__usage__ = """\
Usage: %s model proco_trace

Translates the trace 'proco_trace' produced by proco on the UML model 'model'
to the generic trace format.
The results is output into the standard output stream."""

if __name__ == '__main__':
    if not (len(sys.argv) == 3):
        print __usage__ % sys.argv[0]
        sys.exit(1)

    modelfilename = sys.argv[1]
    procotracefilename = sys.argv[2]

    translate(modelfilename, procotracefilename, sys.stdout)
    sys.exit(0)
