# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Partition types for abstraction refinement.

This module provides functionality for handling type hierarchies which
consist of abstract integer types represented using sets of integers to
be modeled exactly by a type.

Author: Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

import sys
from abstraction.typelibrary import TypeLibrary
from abstraction.typehierarchy import TypeHierarchy
from abstraction.typeinference import \
     _UNRESTRICTED, _UNRESTRICTED_INT, _UNRESTRICTED_PRIMITIVE, \
     _UNRESTRICTED_REFERENCE, _NULL

# Additional pseudotypes used for comparing the precision of types, and for
# computing upper and lower bounds of collections of types.  Internally, a
# PartitionCollection maintains a type hierarchy of the pseudotypes imported
# from abstraction.typeinference and the following pseudotypes to perform
# comparisons and computing upper and lower bounds during the type inference
# procedure: comparisons and upper and lower bounds corresponding to types in
# the following classes are handled separately.
_SOME_REFERENCE = "<SMUML_some_reference_type>"
_SOME_PARTITION = "<SMUML_some_partition>"

class SimpleIntPartition:
    """Partitions of integers with refinement by individual values."""

    def __init__(self, name="", vals=None):
        """Initialize a SimpleIntPartition.

        Argument:
            name -- An optional name for the partition.
            vals -- An iterable of integers representing the values which
                    should be represented exactly in the partition.  By
                    default, a partition representing the entire set of
                    integers will be constructed.
        """
        self.values = set()
        if not isinstance(name, str):
            raise TypeError("a string expected as `name' argument")
        self.name = name
        if vals == None:
            return
        if not hasattr(vals, "__iter__"):
            raise TypeError \
                      ("an iterable of integers expected as `vals' argument")
        for v in vals:
            if not isinstance(v, int):
                raise ValueError \
                          ("an iterable of integers expected as `vals' "
                           "argument")
            self.values.add(v)

    def __str__(self):
        """Return the string representation of a SimpleIntPartition.

        The returned string is suitable for use as an integer partition
        specification, for example, to be given as input to the
        generate_interval_abstraction.py script.
        """
        if not self.values:
            return "(-inf;+inf)"
        lo = None
        result = []
        for v in sorted(self.values):
            if lo == None:
                result.append("(-inf;%d]" % (v - 1))
            elif v - 1 >= lo:
                result.append("[%d;%d]" % (lo, v - 1))
            result.append("[%d;%d]" % (v, v))
            lo = v + 1
        result.append("[%d;+inf)" % lo)
        return ",".join(result)

    def getValues(self):
        """Return the values modeled exactly by the SimpleIntPartition."""
        return set(self.values)

    def getKey(self):
        """Return a normalized representation of a SimpleIntPartition.

        The value returned by the function is guaranteed to be identical
        for all SimpleIntPartition objects representing the "same"
        partition.  The return value can be used, e.g., as a key in a
        dictionary (hence the name of the function).
        """
        return tuple(sorted(self.values))

    def refine(self, vals):
        """Refine a SimpleIntPartition by one or more values.

        Argument:
            vals -- An iterable of integers.

        The function returns a new SimpleIntPartition object which refines
        the partition `self' by making the integers included in `vals' have
        exact representations.
        """
        if not hasattr(vals, "__iter__"):
            raise TypeError("an iterable of integers expected as argument")
        for v in vals:
            if not isinstance(v, int):
                raise ValueError \
                          ("an iterable of integers expected as argument")
        return SimpleIntPartition("", self.values.union(vals))

    def join(self, *partitions):
        """Find a partition "more abstract" than `self' and some partitions.

        Argument:
            partitions -- SimpleIntPartition objects.

        The function returns a new SimpleIntPartition object which
        represents a partition in which an integer is represented exactly
        iff it is represented exactly in `self' and all of the partitions
        in `partitions'.
        """
        result = set(self.values)
        for partition in partitions:
            if not isinstance(partition, SimpleIntPartition):
                raise TypeError("all arguments must be SimpleIntPartitions")
            result.intersection_update(partition.values)
        return SimpleIntPartition("", result)

    def meet(self, *partitions):
        """Find a partition "more precise" than `self' and some partitions.

        Argument:
            partitions -- SimpleIntPartition objects.

        The function returns a new SimpleIntPartition object which
        represents a partition in which an integer is represented exactly
        iff it is represented exactly in `self' or any of the partitions
        in `partitions'.
        """
        result = set(self.values)
        for partition in partitions:
            if not isinstance(partition, SimpleIntPartition):
                raise TypeError("all arguments must be SimpleIntPartitions")
            result.update(partition.values)
        return SimpleIntPartition("", result)

    def modelsExactly(self, value):
        """Test whether the SimpleIntPartition models a given integer exactly.

        Argument:
            value -- An integer to test.

        Returns True if the SimpleIntPartition models the given value exactly,
        False otherwise.
        """
        if not isinstance(value, int):
            raise TypeError("an integer expected as argument")
        return value in self.values

    def isLessPrecise(self, partition):
        """Test whether a partition is less precise than another one.

        Argument:
            partition -- A SimpleIntPartition object.

        Returns True iff the set of integers represented exactly by
        `self' form a subset of the integers represented exactly by
        `partition'.
        """
        if not isinstance(partition, SimpleIntPartition):
            raise TypeError("a SimpleIntPartition expected as argument")
        return self.values.issubset(partition.values)

    def isMorePrecise(self, partition):
        """Test whether a partition is more precise than another one.

        Argument:
            partition -- A SimpleIntPartition object.

        Returns True iff the set of integers represented exactly by
        `self' form a superset of the integers represented exactly by
        `partition'.
        """
        return self.values.issuperset(partition.values)

class PartitionCollection:
    """Collection of partitions."""

    def __init__(self, solver):
        """Initialize a PartitionCollection.

        Argument:
            solver -- An object supporting the `generateDefinitions' method
                      which can be used to generate definitions of
                      operations for types in a type library.
        """
        self.partitions_by_spec = {}
        self.partitions_by_name = {}
        self.typelibrary = TypeLibrary()
        # An ordering of pseudotypes and primitive types (mimics the ordering
        # used in type inference systems).
        self.type_ordering = TypeHierarchy()
        for t in (_UNRESTRICTED, _UNRESTRICTED_INT, _UNRESTRICTED_PRIMITIVE,
                  _UNRESTRICTED_REFERENCE, _NULL, _SOME_REFERENCE,
                  _SOME_PARTITION, "int", "boolean"):
            self.type_ordering.addType(t)
        # Order pseudotypes as they are ordered in
        # TypeInference.__orderPseudoTypesInTypeLibrary (in module
        # abstraction.typeinference); handle also the additional pseudotypes
        # _SOME_PARTITION and _SOME_REFERENCE.
        for (t1, t2) in ((_UNRESTRICTED, _UNRESTRICTED_PRIMITIVE),
                         (_UNRESTRICTED, _UNRESTRICTED_REFERENCE),
                         (_UNRESTRICTED_REFERENCE, _NULL),
                         (_UNRESTRICTED_PRIMITIVE, "boolean"),
                         (_UNRESTRICTED_PRIMITIVE, _UNRESTRICTED_INT),
                         (_UNRESTRICTED_INT, "int"),
                         ("int", _SOME_PARTITION),
                         (_UNRESTRICTED_REFERENCE, _SOME_REFERENCE),
                         (_NULL, _SOME_REFERENCE)):
            self.type_ordering.orderTypes(t1, t2)
        self.solver = solver

    def __iter__(self):
        """Return an iterator to the collection of partitions."""
        return self.partitions_by_name.itervalues()

    def getTypeLibrary(self):
        """Return the type library associated with the partition collection."""
        return self.typelibrary

    def find(self, arg):
        """Find a partition in the collection.

        Argument:
            arg -- A name of a partition as a string, or a
                   SimpleIntPartition object.

        If `arg' is a string, returns the partition with name `arg' in the
        collection (if such a partition exists).  If `arg' is a
        SimpleIntPartition, returns the partition which is equivalent to
        `arg' (i.e., represents the same integers exactly) in the
        collection (if it exists).
        """
        if isinstance(arg, str):
            return self.partitions_by_name[arg]
        elif isinstance(arg, SimpleIntPartition):
            return self.partitions_by_spec[arg.getKey()]
        else:
            raise TypeError \
                      ("a string or a SimpleIntPartition expected as argument")

    def __getPseudoTypeName(self, typename):
        """Return the name of the pseudo type of a type name.

        Argument:
            typename -- Type name to test.

        Returns:
            - _SOME_PARTITION if `typename' refers to one of the partitions
              in the PartitionCollection
            - _SOME_REFERENCE if `typename' does not represent any
              pseudotype or primitive type known by `self.type_ordering'
              or a partition in the PartitionCollection
            - `typename' (which is assumed in this case to be one of
              the types known by `self.type_ordering') otherwise.
        """
        if typename in self.partitions_by_name:
            return _SOME_PARTITION
        if typename in (_UNRESTRICTED, _UNRESTRICTED_REFERENCE, _NULL,
                        _UNRESTRICTED_PRIMITIVE, "boolean",
                        _UNRESTRICTED_INT, "int"):
            return typename
        return _SOME_REFERENCE

    # PartitionCollection objects are used for dynamic type inference.  The
    # `compare', `join' and `meet' methods support the interface required
    # of an object used as the "ordering" parameter for the function
    # abstraction.typeinference.TypeInferenceSystem.solveConstraints: the
    # difference between these methods and the default methods provided by
    # type hierarchies associated with type inference systems is that these
    # functions add new partitions to the type system when needed (i.e.,
    # when computing least upper bounds and greatest lower bounds of
    # collections of partitions).

    def compare(self, p1, p2):
        """Compare two type names according to their precision.

        Arguments:
            p1, p2 -- Names of two types.

        Returns:
            TypeHierarchy.CMP_EQ  if `p1' and `p2' are the same type
            TypeHierarchy.CMP_LT  if `p1' is strictly "more precise" than
                                  `p2'
            TypeHierarchy.CMP_GT  if `p1' is strictly "more abstract" then
                                  `p2'
            TypeHierarchy.CMP_EQ  otherwise.
        """
        if not isinstance(p1, str):
            raise TypeError("a type name expected as first argument")
        if not isinstance(p2, str):
            raise TypeError("a type name expected as second argument")
        if p1 == p2:
            return TypeHierarchy.CMP_EQ
        t1 = self.__getPseudoTypeName(p1)
        t2 = self.__getPseudoTypeName(p2)
        if t1 == _SOME_PARTITION and t2 == _SOME_PARTITION:
            # handle comparison between the precision of two partitions
            # separately
            p1 = self.partitions_by_name[p1]
            p2 = self.partitions_by_name[p2]
            if p1.isMorePrecise(p2):
                return TypeHierarchy.CMP_LT
            if p2.isMorePrecise(p1):
                return TypeHierarchy.CMP_GT
            return TypeHierarchy.CMP_IC
        if t1 == _SOME_REFERENCE and t2 == _SOME_REFERENCE:
            # two unequal reference types are always considered to be
            # incomparable
            return TypeHierarchy.CMP_IC
        return self.type_ordering.compare(t1, t2)

    def __separateTypenames(self, typenames):
        """Separate typenames into pseudonames, partitions and reference types.

        Argument:
            typenames -- An iterable of (pseudo) type names.

        Returns a triple with the following elements (in order):
            1. a set of names of pseudotypes which are represented by a
               type name in `typenames',
            2. a list of the actual SimpleIntPartitions referred to in the
               iterable of type names, and
            3. a set of type names which could not be recognized as
               pseudotypes or partition names.  (All such type names are
               assumed to name reference types, i.e., classes.)
        """
        pseudonames = set()
        partitions = []
        reference_types = set()
        for t in typenames:
            if not isinstance(t, str):
                raise TypeError("all arguments must be strings")
            pseudoname = self.__getPseudoTypeName(t)
            pseudonames.add(pseudoname)
            if pseudoname == _SOME_PARTITION:
                partitions.append(self.partitions_by_name[t])
            elif pseudoname == _SOME_REFERENCE:
                reference_types.add(t)
        return (pseudonames, partitions, reference_types)

    def join(self, *typenames):
        """Return a type "more abstract" than a given set of types.

        Arguments:
            typenames -- Type names.

        Returns the name of the type "more abstract" than the given set of
        types, or None if no such type exists.
        """
        (pseudonames, partitions, reference_types) = \
            self.__separateTypenames(typenames)
        join = self.type_ordering.join(pseudonames)
        if join == None:
            return None
        if join == _SOME_PARTITION:
            # the least upper bound of the types is a partition type; compute
            # the precise bound here
            return self.add(partitions.pop().join(*partitions)).name
        if join == _SOME_REFERENCE:
            if len(reference_types) > 1:
                # two different reference types are always assumed to be
                # incomparable
                return None
            return reference_types.pop()
        return join

    def meet(self, *typenames):
        """Return a type "more precise" than a given set of types.

        Arguments:
            typenames -- Type names.

        Returns the name of the type "more precise" than the given set of
        types, or None if no such type exists.
        """
        (pseudonames, partitions, reference_types) = \
            self.__separateTypenames(typenames)
        meet = self.type_ordering.meet(pseudonames)
        if meet == None:
            return None
        if meet == _SOME_PARTITION:
            # the greatest lower bound of the types is a partition type;
            # compute the precise bound here
            return self.add(partitions.pop().meet(*partitions)).name
        if meet == _SOME_REFERENCE:
            if len(reference_types) > 1:
                # two different reference types are always assumed to be
                # incomparable
                return None
            return reference_types.pop()
        return meet

    def add(self, partition):
        """Add a new partition to the collection.

        Argument:
            partition -- A SimpleIntPartition to be added to the collection.

        If there already exists a partition equivalent to `partition' (i.e.,
        a partition which represents the same integers exactly), returns
        this partition from the collection.  Otherwise returns `partition'.
        """
        if not isinstance(partition, SimpleIntPartition):
            raise TypeError("a SimpleIntPartition expected as argument")
        spec = partition.getKey()
        try:
            return self.partitions_by_spec[spec]
        except KeyError:
            partition.name = "SMUML_type_%d" % len(self.partitions_by_spec)
            domain = str(partition).split(",")
            self.typelibrary.addType(partition.name, domain)
            # define a coercion from `int' to the new type for the
            # "Jumbala" an "python" back-ends
            for (backend, andsym, truesym) in (("Jumbala", "&&", "true"),
                                               ("python", "and", "True")):
                dfn = []
                for (element, value) in \
                        [(e, self.typelibrary.domain[partition.name][e])
                         for e in domain]:
                    (lo, hi) = element[1:-1].split(";")
                    if lo == "-inf":
                        if hi == "+inf":
                            dfn.append((truesym, [value]))
                        else:
                            dfn.append(("<> <= %s" % hi, [value]))
                    else:
                        if hi == "+inf":
                            dfn.append(("<> >= %s" % lo, [value]))
                        else:
                            if lo == hi:
                                dfn.append(("<> == %s" % lo, [value]))
                            else:
                                dfn.append \
                                    (("(<> >= %s) %s (<> <= %s)"
                                      % (lo, andsym, hi), [value]))
                self.typelibrary.defineCoercion \
                    ("int", partition.name, backend, dfn)
            self.partitions_by_spec[spec] = partition
            self.partitions_by_name[partition.name] = partition
            return partition

    def generateDefinitions(self, operations, stdout=sys.stdout, stderr=sys.stderr):
        """Generate definitions for an iterable of operations.

        Arguments:
            operations -- An iterable of operations to generate.  Each
                          element in the iterable should be either of the
                          form
                              (<type name>, <operation symbol>, <arity>)
                          or
                              ("coercion", <name of source type>,
                               <name of target type>, <back-end>),
                          where
                              - <type name>, <name of source type>, and
                                <name of target type> are names of
                                partitions in the PartitionCollection,
                              - <operation symbol> is an operator or test
                                symbol in the Jumbala action language,
                              - <arity> is the arity of the operator or
                                test, and
                              - <back-end> is one of the strings
                                "python" or "Jumbala".
            stdout     -- Standard output stream (unused).
            stderr     -- Stream for error messages and warnings.
        """
        self.solver.generateDefinitions \
            (self.typelibrary, operations, stdout, stderr)

# -----------------------------------------------------------------------------
# class Integer:
#     def __init__(self, val):
#         self.val = val

#     def __lt__(self, i):
#         if self.val == NEG_INF:
#             return i.val != NEG_INF
#         if self.val == POS_INF or i.val == NEG_INF:
#             return False
#         if i.val == POS_INF:
#             return True
#         return self.val < i.val

#     def __gt__(self, i):
#         if self.val == POS_INF:
#             return i.val != POS_INF
#         if self.val == NEG_INF or i.val == POS_INF:
#             return False
#         if i.val == NEG_INF:
#             return True
#         return self.val > i.val

#     def __eq__(self, i):
#         if self.val in (NEG_INF, POS_INF) or i.val in (NEG_INF, POS_INF):
#             return False
#         return self.val == i.val

#     def __ne__(self, i):
#         if self.val in (NEG_INF, POS_INF) or i.val in (NEG_INF, POS_INF):
#             return False
#         return self.val != i.val

#     def __le__(self, i):
#         return self.__lt__(i) or self.__eq__(i)

#     def __ge__(self, i):
#         return self.__gt__(i) or self.__eq__(i)

# class NEG_INF:
#     pass

# class POS_INF:
#     pass

# class Interval:
#     def __init__(self, lo, hi):
#         self.lo = Integer(lo)
#         self.hi = Integer(hi)

#     def __str__(self):
#         result = []
#         if self.lo.val == NEG_INF:
#             result.append("(-inf;")
#         else:
#             result.append("[%d;" % self.lo.val)
#         if self.hi.val == POS_INF:
#             result.append("+inf)")
#         else:
#             result.append("%d]" % self.hi.val)
#         return "".join(result)

#     def __eq__(self, interval):
#         return ((self.lo.val == interval.lo.val == NEG_INF)
#                 or (self.lo.val == interval.lo.val == POS_INF)
#                 or (self.lo.val == interval.lo.val)) \
#                 and ((self.hi.val == interval.hi.val == NEG_INF)
#                      or (self.hi.val == interval.hi.val == POS_INF)
#                      or (self.hi.val == interval.hi.val))
    
#     def __ne__(self, interval):
#         return not self.__eq__(interval)

#     def contains(self, val):
#         return self.lo <= Integer(val) <= self.hi

# class IntPartition:
#     def __init__(self, spec=None):
#         if spec == None:
#             self.spec = [Interval(NEG_INF, POS_INF)]
#             return
#         if not isinstance(spec, (list, tuple)):
#             raise TypeError("a sequence expected as argument")
#         if len(spec) == 0:
#             raise ValueError("argument sequence must be non-empty")
#         lo = NEG_INF
#         for interval in spec:
#             if not isinstance(interval, Interval):
#                 raise ValueError("sequence of intervals expected as argument")
#             if lo != interval.lo.val:
#                 raise ValueError("invalid integer partition specification")
#             if interval.hi.val != POS_INF:
#                 lo = interval.hi.val + 1
#         if interval.hi.val != POS_INF:
#             raise ValueError("invalid integer partition specification")
#         self.spec = spec

#     def __str__(self):
#         return ",".join([str(i) for i in self.spec])

#     def __iter__(self):
#         return self.spec.__iter__()

#     def get_key(self):
#         return tuple([(i.lo.val, i.hi.val) for i in self.spec])

#     def refine(self, splitvals):
#         if not hasattr(splitvals, "__iter__"):
#             splitvals = [splitvals]
#         result = []
#         sorted_splitvals = list(set(splitvals))
#         sorted_splitvals.sort()
#         intervals = self.spec.__iter__()
#         splitvals = sorted_splitvals.__iter__()
#         interval = intervals.next()
#         more_splitvals = True
#         try:
#             splitval = splitvals.next()
#         except StopIteration:
#             more_splitvals = False
#         try:
#             while True:
#                 if more_splitvals and interval.contains(splitval):
#                     if interval.lo <= Integer(splitval - 1):
#                         result.append(Interval(interval.lo.val, splitval - 1))
#                     result.append(Interval(splitval, splitval))
#                     if interval.hi >= Integer(splitval + 1):
#                         interval = Interval(splitval + 1, interval.hi.val)
#                     else:
#                         interval = intervals.next()
#                     try:
#                         splitval = splitvals.next()
#                     except StopIteration:
#                         more_splitvals = False
#                 else:
#                     result.append(Interval(interval.lo.val, interval.hi.val))
#                     interval = intervals.next()
#         except StopIteration:
#             pass
#         return IntPartition(result)

#     def join(self, partition):
#         if not isinstance(partition, IntPartition):
#             raise TypeError("IntPartition expected as argument")
#         i1s = self.spec.__iter__()
#         i2s = partition.spec.__iter__()
#         i1 = i1s.next()
#         i2 = i2s.next()
#         lo = i1.lo.val
#         result = []
#         try:
#             while True:
#                 if i1.hi.val == i2.hi.val:
#                     result.append(Interval(lo, i1.hi.val))
#                     i1 = i1s.next()
#                     i2 = i2s.next()
#                     lo = i1.lo.val
#                 elif i1.hi < i2.hi:
#                     i1 = i1s.next()
#                 else:
#                     i2 = i2s.next()
#         except StopIteration:
#             return IntPartition(result)
