# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Wrapper module for generating abstract types using various semantics.

This module provides a common interface to use for extending type
libraries with automatically generated abstract operations.

Author: Heikki Tauriainen <heikki.tauriainen@tkk.fi>
"""

import os
from tempfile import mkstemp
import sys
import subprocess
import generate_interval_abstraction as IntervalGenerator
from abstraction import error
from abstraction.typelibrary import TypeLibrary
from abstraction import smtlib_typegen

class UnboundedIntegerSemantics:
    """Abstract operation generator using unbounded integer semantics."""
    
    def __init__(self):
        """Initialize an UnboundedIntegerSemantics object."""
        pass

    def generateDefinitions(self, typelibrary, operations, stdout=sys.stdout, stderr=sys.stderr):
        """Add definitions of new operations to a type library.

        Arguments:
            typelibrary -- The abstraction.typelibrary.TypeLibrary object
                           to be extended with new operations.
            operations  -- An iterable of operations to generate.  Each
                           element in the iterable should be either
                           a triple of the form
                               (<type name>, <operator symbol>, <arity>),
                           or a quadruple of the form
                               ("coercion", <name of source type>,
                                <name of target type>, <name of back-end>),
                           where <type name>, <name of source type>, and
                           <name of target type> are type names in
                           `typelibrary', <operator symbol> is a basic
                           arithmetic ("+", "-", "*", "/") or test
                           ("==", "!=", "<", ">", "<=", ">=") operator
                           symbol in the Jumbala action language, and
                           <arity> is one of the values 1 or 2.
                           <name of back-end> is unused.
            stdout      -- Standard output stream (unused).
            stderr      -- Standard error stream (for error messages and
                           warnings emitted while generating the
                           operations).

        This function extends the given type library with definitions of
        operators in `operations' (old definitions of these operators, if
        present in `typelibrary', will be destroyed).  The abstract
        operations are generated by approximating the corresponding
        concrete operations on unbounded integers.
        """
        # collect operations and coercions to generate
        operations_to_generate = []
        coercions_to_generate = []
        typenames = set()
        for op in operations:
            if op[0] == "coercion":
                coercions_to_generate.append((op[1], op[2]))
                typenames.update((op[1], op[2]))
            else:
                typenames.add(op[0])
                if op[1] in ("+", "-"):
                    operations_to_generate.append("%d%s" % (op[2], op[1]))
                elif not op[1] in ("*", "/", "==", "!=", "<", ">", "<=", ">="):
                    raise error.Error \
                              ("unsupported operator (`%s') in model" % op[1])
                else:
                    operations_to_generate.append(op[1])

        # build the command line for the type generator module:
        #     --verbose -I <operator/test symbols>
        #     [-c <source type of coercion>:<target type of coercion>]*
        #     [<type name>:<domain specification as a list of intervals>]+
        
        args = ["--verbose", "-I", ",".join(operations_to_generate)]
        for (fromtype, totype) in coercions_to_generate:
            args.extend(("-c", "%s:%s" % (fromtype, totype)))
        # the domain elements must be listed in the command line in the
        # increasing order of the lower bounds of intervals
        args.extend \
            (["%s:%s"
              % (t, ",".join([element for (element, value)
                              in sorted \
                                     (typelibrary.domain[t].items(),
                                      key=lambda x: x[1])]))
              for t in typenames])

        # generate a temporary type library containing (only) the types and
        # operations which should be generated
        (fd, tl_file_name) = mkstemp()
        tl = TypeLibrary()
        try:
            tl_file = os.fdopen(fd, "w")
            IntervalGenerator.generate(args, tl_file, stderr)
            tl_file.close()
            tl.read(tl_file_name)
        finally:
            os.remove(tl_file_name)

        # copy the generated operations from the temporary type library to
        # `typelibrary'
        for t in tl.typenames:
            if t in tl.STD_TYPES:
                continue
            for (element, value) in tl.domain[t].iteritems():
                if typelibrary.domain[t][element] != value:
                    raise error.Error \
                              ("inconsistency in domain of generated type")
            for (arity, symbols) in tl.operators[t].iteritems():
                if not arity in typelibrary.operators[t]:
                    typelibrary.operators[t][arity] = {}
                for (symbol, dfn) in symbols.iteritems():
                    typelibrary.operators[t][arity][symbol] = dfn
            for (arity, symbols) in tl.tests[t].iteritems():
                if not arity in typelibrary.tests[t]:
                    typelibrary.tests[t][arity] = {}
                for (symbol, dfn) in symbols.iteritems():
                    typelibrary.tests[t][arity][symbol] = dfn
            for (fromtype, backends) in tl.coercions[t].iteritems():
                if fromtype in typelibrary.STD_TYPES:
                    continue
                if not fromtype in typelibrary.coercions[t]:
                    typelibrary.addCoercion(fromtype, t)
                for (backend, dfn) in backends.iteritems():
                    typelibrary.coercions[t][fromtype][backend] = dfn

class BoundedIntegerSemantics:
    """Abstract operation generator using bounded integer semantics."""

    def __init__(self, solver, path):
        """Initialize a BoundedIntegerSemantics object.

        Arguments:
            solver -- One of the names of SMT solvers supported by PySMT.
            path   -- Command to run the solver executable.
        """
        self.solver = smtlib_typegen.PySmt(solver, path)

    def generateDefinitions(self, typelibrary, operations, stdout=sys.stdout, stderr=sys.stderr):
        """Add definitions of new operations to a type library.

        Arguments:
            typelibrary -- The abstraction.typelibrary.TypeLibrary object
                           to be extended with new operations.
            operations  -- A set of operations to generate.  Each element
                           in the set should be either a triple of the
                           form
                               (<type name>, <operator symbol>, <arity>),
                           or a quadruple of the form
                               ("coercion", <name of source type>,
                                <name of target type>, "Jumbala"),
                           where <type name>, <name of source type>, and
                           <name of target type> are type names in
                           `typelibrary', <operator symbol> is any
                           arithmetic or test operator symbol in the
                           Jumbala action language, and <arity> is one of
                           the values 1 or 2.
            stdout      -- Standard output stream (unused).
            stderr      -- Standard error stream (for error messages and
                           warnings emitted while generating the
                           operations).

        This function extends the given type library with definitions of
        operators in `operations' (old definitions of these operators, if
        present in `typelibrary', will be destroyed).  The abstract
        operations are generated by approximating the corresponding
        concrete operations on bounded integers.
        """
        # ensure that also Python versions of coercions are generated
        operations.update \
            ([("coercion", op[1], op[2], "python") for op in operations
              if op[0] == "coercion"])
        smtlib_typegen.generateOperations \
            (typelibrary, operations, self.solver, stdout=stdout,
             stderr=stderr, verbose=True)

class MixedIntegerSemantics(BoundedIntegerSemantics, UnboundedIntegerSemantics):
    """Abstract operation generator using mixed integer semantics."""

    def __init__(self, solver, path):
        """Initialize a BoundedIntegerSemantics object.

        Arguments:
            solver -- One of the names of SMT solvers supported by PySMT.
            path   -- Command to run the solver executable.
        """
        BoundedIntegerSemantics.__init__(self, solver, path)
        UnboundedIntegerSemantics.__init__(self)

    def generateDefinitions(self, typelibrary, operations, stdout=sys.stdout, stderr=sys.stderr):
        """Add definitions of new operations to a type library.

        Arguments:
            typelibrary -- The abstraction.typelibrary.TypeLibrary object
                           to be extended with new operations.
            operations  -- An iterable of operations to generate.  Each
                           element in the iterable should be either
                           a triple of the form
                               (<type name>, <operator symbol>, <arity>),
                           or a quadruple of the form
                               ("coercion", <name of source type>,
                                <name of target type>, <name of back-end>),
                           where <type name>, <name of source type>, and
                           <name of target type> are type names in
                           `typelibrary', <operator symbol> is any
                           arithmetic or test operator symbol in the
                           Jumbala action language, and <arity> is one of
                           the values 1 or 2.  <name of back-end> should
                           be "Jumbala".
            stdout      -- Standard output stream (unused).
            stderr      -- Standard error stream (for error messages and
                           warnings emitted while generating the
                           operations).

        This function extends the given type library with definitions of
        operators in `operations' (old definitions of these operators, if
        present in `typelibrary', will be destroyed).  The abstract basic
        arithmetic ("+", "-", "*", "/") and test ("==", "!=", "<", ">",
        "<=", ">=") operators are constructed by approximating the
        corresponding concrete operations on unbounded integers, and the
        remaining operations are constructed by approximating the
        corresponding concrete operations on bounded integers.
        """
        # filter operations and coercions to generate
        bounded_operations_to_generate = set()
        unbounded_operations_to_generate = []
        for op in operations:
            if op[0] == "coercion" or \
                   op[1] in ("+", "-", "*", "/", "==", "!=", "<", ">", "<=",
                             ">="):
                unbounded_operations_to_generate.append(op)
            else:
                bounded_operations_to_generate.add(op)
        if unbounded_operations_to_generate:
            UnboundedIntegerSemantics.generateDefinitions \
                (self, typelibrary, unbounded_operations_to_generate, stdout,
                 stderr)
        if bounded_operations_to_generate:
            BoundedIntegerSemantics.generateDefinitions \
                (self, typelibrary, bounded_operations_to_generate, stdout,
                 stderr)
