import sys
import string
import coral
import Sim
import Sim.Core
import Utils.String
import Utils.Models
from Utils.Models import get_qualified_name

"""
Simulate the generic trace given as the argument.

The format for the trace file is as follows.
- The first line should be of form
  'Model f'
  where 'f' is the UML Model file name.
- After the first line, lines starting with '#' are ignored as comments.
- The next lines of form
  'InitialConfiguration object o umlobject u'
  describe the correspondence between the analysis tool object 'o'
  (a string with no white space) and the UML Object 'u' in the UML Model
  ('u' is the XMI id of the UML Object).

- The rest of the lines describe the actual trace and are of form:
  - 'TransitionExecution object o transition t'
    describing that the analysis tool object 'o' executes the UML Transition
    't' ('t' is the XMI id of the UML Transition).
  - 'EventDefer object o'
    describing that the analysis tool object 'o' defers the first message
    in its input queue.
  - 'ImplicitConsumption object o'
    describing that the analysis tool object 'o' implicitly consumes
    the first message in its input queue.
  - 'GuardChooseValue object o transition t index i value v'
    describing the next value 'v' of the 'i':th choose-function in the guard
    of the transition 't' in the analysis tool object 'o'.
  - 'EffectChooseValue object o transition t index i value v'
    describing the next value 'v' of the 'i':th choose-function in the effect
    of the transition 't' in the analysis tool object 'o'.
"""

UML14 = coral.meta.UML14



def simulate(trace_filename, stdout = sys.stdout, stderr = sys.stderr):
    """
    Simulate the trace in the file 'trace_filename' in the SMUML simulator.
    Return a simulator with the executed trace.
    """
    assert type(trace_filename) is str
    
    def _verbose(msg):
        if stdout:
            print >> stdout, msg
        
    #
    # Open the trace file
    #
    _verbose("Processing the trace file " + trace_filename)
    try:
        trace_file = open(trace_filename, 'r')
    except IOError,e:
        raise RuntimeError("Cannot open trace file '"+trace_filename+"': "+
                           str(e))


    #
    # Read the first line giving the model name
    #
    l = trace_file.readline()
    l_split = string.split(l)
    if len(l_split) != 2 or l_split[0] != "Model":
        raise RuntimeError("Error at trace file line: "+l)
    model_filename = l_split[1]


    #
    # Read the initial global configuration correspondence
    #
    map_mc_object_to_uml_object = {}
    map_uml_object_to_mc_object = {}
    while True:
        l = trace_file.readline()
        if not l:
            break
        if l.startswith('#'):
            continue
        l_split = string.split(l)
        if l_split[0:2] == ['InitialConfiguration','object']:
            assert len(l_split) == 5
            assert l_split[3] == 'umlobject'
            map_mc_object_to_uml_object[l_split[2]] = l_split[4]
            map_uml_object_to_mc_object[l_split[4]] = l_split[2]
        else:
            break


    #
    # Read trace actions
    #
    trace = []
    stats_nof_actions = 0
    while l:
        l_split = string.split(l)
        if l_split[0:2] == ['TransitionExecution','object']:
            assert len(l_split) == 5
            assert l_split[3] == 'transition'
            trace.append(('EXEC', l_split[2], l_split[4]))
            stats_nof_actions += 1
        elif l_split[0:2] == ['EventDefer','object']:
            assert len(l_split) == 3
            trace.append(('DEFER', l_split[2]))
            stats_nof_actions += 1
        elif l_split[0:2] == ['ImplicitConsumption','object']:
            assert len(l_split) == 3
            trace.append(('IMPL', l_split[2]))
            stats_nof_actions += 1
        elif l_split[0] == 'GuardChooseValue':
            assert len(l_split) == 9
            assert l_split[1] == 'object'
            mc_object = l_split[2]
            assert l_split[3] == 'transition'
            uml_trans_xmiid = l_split[4]
            assert l_split[5] == 'index'
            index = int(l_split[6])
            assert l_split[7] == 'value'
            value = int(l_split[8])
            trace.append(('GuardChooseValue',mc_object,uml_trans_xmiid,
                          index,value))
        elif l_split[0] == 'EffectChooseValue':
            assert len(l_split) == 9
            assert l_split[1] == 'object'
            mc_object = l_split[2]
            assert l_split[3] == 'transition'
            uml_trans_xmiid = l_split[4]
            assert l_split[5] == 'index'
            index = int(l_split[6])
            assert l_split[7] == 'value'
            value = int(l_split[8])
            trace.append(('EffectChooseValue',mc_object,uml_trans_xmiid,
                          index,value))
        elif l.startswith('#'):
            # A comment, ignored
            pass
        else:
            assert False and "Internal error: something not yet implemented"

        l = trace_file.readline()

    _verbose("The trace has %d actions" % stats_nof_actions)


    #
    # Load project
    #
    _verbose("Loading model " + model_filename)
    try:
        P = coral.loadDocument(model_filename)
    except RuntimeError,e:
        raise RuntimeError("Error: cannot load UML model '"+model_filename+
                           "': "+str(e))


    #
    # Find the model
    #
    models = [e for e in P.contents() if e.oclIsKindOf(UML14.Model)]
    if not models:
        _error("No Model found")
    if len(models) > 1:
        _error("Multiple Models found")
    model = models[0]

    #
    # Create the simulator
    #
    sim =  Sim.Core.Simulator(model)



    #
    # Compute a map from XMI ids to transitions
    #
    xmi_id_to_uml_transition = {}
    transitions = Utils.Models.get_elements_of_type(model, UML14.Transition)
    for t in transitions:
        xmi_id = t.__XMIid__()
        assert not xmi_id in xmi_id_to_uml_transition
        xmi_id_to_uml_transition[xmi_id] = t


        
    #
    # Get correspondence between the simulator and mc initial global state
    #
    _verbose("Getting initial global configuration correspondence...")
    global_state = sim.trace.get_last_global_state()
    mc_object_to_sim_oid_map = {}
    for (mc_object,uml_object_xmiid) in map_mc_object_to_uml_object.items():
        sim_oid = None
        for o in global_state.objects:
            if o.get_uml_object():
                if o.get_uml_object().__XMIid__() == uml_object_xmiid:
                    assert sim_oid == None
                    sim_oid = o.idnum
        assert sim_oid != None
        assert not mc_object in mc_object_to_sim_oid_map
        mc_object_to_sim_oid_map[mc_object] = sim_oid


    #
    # Set up hooks that give values to the non-deterministic choose functions
    #
    guard_choose_values = {}
    effect_choose_values = {}
    def _choose_boolean_hook(dummy, sim, obj, uml_transition, where, index):
        dictionary = None
        if where == 'guard':
            dictionary = guard_choose_values
        elif where == 'effect':
            dictionary = effect_choose_values
        else:
            assert False and "Internal error"
        mc_object = None
        for (mc_obj,sim_oid) in mc_object_to_sim_oid_map.iteritems():
            if sim_oid == obj.idnum:
                mc_object = mc_obj
                break
        assert mc_object
        value = dictionary[(mc_object,uml_transition.__XMIid__(),index)]
        assert value == 0 or value == 1
        return value == 1
    def _choose_int_hook(dummy, sim, obj, uml_transition, where, index,
                         int_list):
        dictionary = None
        if where == 'guard':
            dictionary = guard_choose_values
        elif where == 'effect':
            dictionary = effect_choose_values
        else:
            assert False and "Internal error"
        mc_object = None
        for (mc_obj,sim_oid) in mc_object_to_sim_oid_map.iteritems():
            if sim_oid == obj.idnum:
                mc_object = mc_obj
                break
        assert mc_object
        value = dictionary[(mc_object,uml_transition.__XMIid__(),index)]
        assert value in int_list
        return value
    sim.set_choose_boolean_hook(_choose_boolean_hook, None)
    sim.set_choose_int_hook(_choose_int_hook, None)



    #
    # Simulate actions one-by-one in the simulator
    #
    _verbose("Simulating trace...")
    step_num = 1
    for action in trace:
        if action[0] == 'GuardChooseValue':
            # Update guard choose function value map
            (dummy,mc_object,uml_trans_xmiid,index,value) = action
            guard_choose_values[(mc_object,uml_trans_xmiid,index)] = value
            continue

        elif action[0] == 'EffectChooseValue':
            # Update effect choose function value map
            (dummy,mc_object,uml_trans_xmiid,index,value) = action
            effect_choose_values[(mc_object,uml_trans_xmiid,index)] = value
            continue

        elif action[0] == 'EXEC':
            # Execution of a single transition
            # The structure should be of form
            # ('EXEC',mc_object,uml_trans_xmi_id)
            (dummy,mc_object,uml_trans_xmiid) = action
            assert mc_object_to_sim_oid_map.has_key(mc_object)
            sim_oid = mc_object_to_sim_oid_map[mc_object]
            assert xmi_id_to_uml_transition.has_key(uml_trans_xmiid)
            uml_trans = xmi_id_to_uml_transition[uml_trans_xmiid]
            _verbose(" Object "+str(sim_oid)+" moves from '"+\
                     uml_trans.source.name+"' to '"+uml_trans.target.name+"'")
        
            # Get the current global state
            global_state = sim.trace.get_last_global_state().api_copy()

            #
            # Handle ATOs by adding the ATO signal in front of the input queue
            #
            if uml_trans.trigger and uml_trans.trigger.signal.name[0] == '$':
                parameters = []
                if uml_trans.trigger.signal.feature:
                    assert False and \
                           "External signals with parameters not handled"
                    # TODO: Decode possible ATO signal parameters
                # Insert ATO signal to the beginning of the queue of the object
                message = Sim.Message.Message(uml_trans.trigger.signal,
                                              parameters)
                tmp = [message]
                tmp.extend(global_state.objects[sim_oid].sm_input_queue)
                global_state.objects[sim_oid].sm_input_queue = tmp
            
            sim.set_working_global_state(global_state)
            enabled_sets,implicit_consumption_objects = sim.get_enabled_in_working_global_state()
            found_index = None
            for (index, (o,set)) in enumerate(enabled_sets):
                if o.idnum != sim_oid:
                    continue
                assert len(set) == 1
                (reason,sim_transition) = set[0]
                if (reason == "completion" or reason == "ATO" or
                    reason == "signal"):
                    pass
                elif reason == "defer":
                    continue
                else:
                    assert False and "Internal error: something not implemented"
                if sim_transition != uml_trans:
                    continue
                assert not found_index
                found_index = index
            if found_index == None:
                if stderr:
                    print >> stderr, "The trace up to and including "\
                          "the current global state is:"
                    sim.api_print_trace(stderr)
                    print >> stderr,\
                          "Internal error at step " + str(step_num)+": "+\
                          "object " + str(sim_oid)+" cannot execute "+\
                          "transition '"+get_qualified_name(uml_trans)+"' "+\
                          "from '"+get_qualified_name(uml_trans.source)+"' "+\
                          "to '"+get_qualified_name(uml_trans.target)+"'"
                raise RuntimeError("The trace cannot be simulated")

            try:
                sim_action_list = sim.execute_transition_set_in_working_global_state(found_index)
            except Sim.Error.SimError, e:
                if stderr:
                    print >> stderr, "The trace up to and including "\
                          "the current global state is:"
                    sim.api_print_trace(stderr)
                    print >> stderr, e
                    print >> stderr, "Aborting!"
                raise RuntimeError("The trace cannot be simulated")
                
            for sim_action in sim_action_list:
                sim.trace.add_action(sim_action)
            sim.trace.add_global_state(global_state)



        elif action[0] == "DEFER":
            # Signal deferring
            # The structure should be of form ('DEFER',mc_object)
            (dummy,mc_object) = action
            assert mc_object in mc_object_to_sim_oid_map
            sim_oid = mc_object_to_sim_oid_map[mc_object]
            _verbose(" Object "+str(sim_oid)+" defers a signal")
            
            # Get the current global state
            global_state = sim.trace.get_last_global_state().api_copy()

            sim.set_working_global_state(global_state)
            enabled_sets,implicit_consumption_objects = sim.get_enabled_in_working_global_state()
            found_index = None
            for (index, (o,set)) in enumerate(enabled_sets):
                if o.idnum != sim_oid:
                    continue
                assert len(set) == 1
                (reason,sim_transition) = set[0]
                if reason == "defer":
                    assert not found_index
                    found_index = index
                    break
            if found_index == None:
                if stderr:
                    print >> stderr, "The trace up to and including "\
                          "the current global state is:"
                    sim.api_print_trace(stderr)
                    print >> stderr, \
                          "Error at step "+str(step_num)+": "+\
                          "object "+str(sim_oid)+" cannot defer a signal"
                raise RuntimeError("The trace cannot be simulated")


            sim_action_list = sim.execute_transition_set_in_working_global_state(found_index)
            for sim_action in sim_action_list:
                sim.trace.add_action(sim_action)
            sim.trace.add_global_state(global_state)



        elif action[0] == "IMPL":
            # Implicit consumption
            # The structure should be of form ('IMPL',mc_object)
            (dummy,mc_object) = action
            assert mc_object in mc_object_to_sim_oid_map
            sim_oid = mc_object_to_sim_oid_map[mc_object]
            _verbose(" Object "+str(sim_oid)+" performs implicit consumption")
            
            # Get the current global state
            global_state = sim.trace.get_last_global_state().api_copy()
            sim.set_working_global_state(global_state)
            enabled_sets,implicit_consumption_objects = sim.get_enabled_in_working_global_state()
            found_index = None
            for (index,o) in enumerate(implicit_consumption_objects):
                if o.idnum != sim_oid:
                    continue
                assert not found_index
                found_index = index

            if found_index == None:
                if stderr:
                    print >> stderr, "The trace up to and including "\
                          "the current global state is:"
                    sim.api_print_trace(stderr)
                    print >> stderr, \
                          "Error at step "+str(step_num)+": "+\
                          "object "+str(sim_oid)+" cannot "+\
                          "do implicit consumption"
                raise RuntimeError("The trace cannot be simulated")

            action_list = sim.execute_implicit_consumption_in_working_global_state(found_index)
            for action in action_list:
                sim.trace.add_action(action)
            sim.trace.add_global_state(global_state)


        else:
            #print str(action)
            assert False and "Internal error: unknown trace action"

        step_num += 1

    #
    # Succesfully simulated the trace
    #
    return sim



if __name__ == '__main__':
    if not (len(sys.argv) == 2):
        print "Usage: " + sys.argv[0] + " trace"
        sys.exit(1)

    sim = simulate(sys.argv[1])

    if sim:
        print "Printing trace..."
        sim.api_print_trace()

