#! /usr/bin/python

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""
Usage: slice_state_machines.py [OPTIONS] FILE

Slices UML state machines in the UML model in FILE.

Options:
  -o, --outputfile=FILENAME  Specifies the name for the file in which to
                             write the sliced UML model. If not specified,
                             original name will be used with '.sliced' added
                             into it. For example, a sliced version of
                             'model.xmi' would be 'model.sliced.xmi'.
                             If the name of a model to be sliced does not
                             end with postfix '.xmi' or '.xml', a sliced
                             model will have the name of the original model
                             with a '.sliced.xmi' postfix added.

  -h, --help                 Displays this text.

  -C, --cfg                  Writes dot files describing control flow graph
                             (CFG) used in slicing.

  -I, --inputsm              Writes dot files containing the original state
                             machines represented by graphs in DOT language.

  -O, --outputsm             Writes dot files containing the sliced state
                             machines represented by graphs in DOT language.

  -D  --dependencies=OPTIONS Specifies what dependencies are drawn to the
                             dot-file representing CFG (if --cfg flag is used)
                             Possibilities are:
                             d = data dependencies
                             n = non-termination sensitive control dependencies
                             o = decisive control dependencies
                             i = interference dependencies
                             s = signal sending dependencies

                             For example -Ddno prints data dependencies,
                             non-termination sensitive control dependencies,
                             and decisive control dependencies.

                             If this flag is not given, all dependencies are
                             drawn as a default.

  -N, --nosliced             Writes no sliced UML model.

  -A  --sliceassert          Adds all assert statements to slicing criterion.


Slicing criterion is defined in the UML model given by user as an
input.  It is a set of transitions that contain all the transitions
from the original UML model that have a taggedValue named
'slicingCrit'. The actual value of taggedValue 'slicingCrit' have no
meaning.

You can also add all assert-statements to slising criterion by using
the correct command line option.

Dot files are text files describing a graph in a DOT language. Graphs
described in DOT can be viewed with a small program called 'dotty' which is
part of a Graphviz package.

For more information about the theory behind the slicer, see
http://www.tcs.hut.fi/Studies/T-79.5001/reports/ojal07sliuml.pdf

This script was written by Vesa Ojala for the SMUML project
<http://www.tcs.hut.fi/Research/Logic/SMUML.shtml>.

"""

import sys
import types
import coral
import getopt
from slicer import slicer
from slicer import printer

def abort(msg="aborted"):
    raise ("%s: " % sys.argv[0]) + msg

class CmdOption:
    def __init__(self, ltr, kw, takeargs):
        assert isinstance(ltr, types.StringType)
        assert len(ltr) == 1
        assert isinstance(kw, types.StringType)
        assert len(kw) >= 1
        assert isinstance(takeargs, types.BooleanType)
        self.letter = ltr
        self.keyword = kw
        self.takes_args = takeargs
        self.arg = None
        self.value = False

    def get_letter(self):
        return self.letter

    def get_keyword(self):
        return self.keyword

    def get_getopt_short(self):
        if self.takes_args:
            return self.letter + ':'
        else:
            return self.letter

    def get_getopt_long(self):
        if self.takes_args:
            return self.keyword + '='
        else:
            return self.keyword

    def get_all_cmdopts(self):
        return ('-'+self.letter, '--'+self.keyword)

    def get_value(self):
        return self.value

    def get_arg(self):
        return self.arg

    def get_take_args(self):
        return self.takes_args

    def set_value(self, nv):
        assert isinstance(nv, types.BooleanType)
        self.value = nv

    def set_arg(self, na):
        assert isinstance(na, types.StringType)
        self.arg = na

    def add_to_dict(self, dictionary):
        dictionary['-'+self.letter] = self
        dictionary['--'+self.keyword] = self
        dictionary[self.keyword] = self

UML = coral.lang.UML14

cmd_options = {}
#cmd_options['outputfile'] =
CmdOption('o', 'outputfile', True).add_to_dict(cmd_options)
#cmd_options['cfg'] =
CmdOption('C', 'cfg', False).add_to_dict(cmd_options)
#cmd_options['inputsm'] =
CmdOption('I', 'inputsm', False).add_to_dict(cmd_options)
#cmd_options['outputsm'] =
CmdOption('O', 'outputsm', False).add_to_dict(cmd_options)
#cmd_options['help'] =
CmdOption('h','help', False).add_to_dict(cmd_options)
#cmd_options['nosliced'] =
CmdOption('N', 'nosliced', False).add_to_dict(cmd_options)
#cmd_options['dependencies'] =
CmdOption('D', 'dependencies', True).add_to_dict(cmd_options)
CmdOption('A', 'sliceassert', False).add_to_dict(cmd_options)


# parse command line
try:
    cmdopt, cmdargs = getopt.getopt(sys.argv[1:],
                                    ''.join([v.get_getopt_short() for v in cmd_options.values()]),
                                    [v.get_getopt_long() for v in cmd_options.values()])
except getopt.GetoptError, e:
    abort(str(e))

# process command line options
for (option, value) in cmdopt:
    temp = cmd_options[option]
    temp.set_value(True)
    if temp.get_take_args():
        temp.set_arg(value)

if cmd_options['help'].get_value():
    sys.stdout.write(__doc__)
    sys.exit(0)

if cmd_options['sliceassert'].get_value():
    slicer.slice_assert = True

# set names correctly
if len(cmdargs) == 0:
    abort("model name expected")
elif len(cmdargs) > 1:
    abort("extra arguments on command line")
modelname = cmdargs[0]

# find basename
extension = modelname[-4:]
if extension == ".xml" or extension == ".xmi":
    basename = modelname[:-4]
else:
    basename = modelname

# set rest of the file names correctly
if cmd_options['outputfile'].get_value():
    outputname = cmd_options['outputfile'].get_arg()
else:
    if extension == ".xml" or extension == ".xmi":
        outputname = basename + ".sliced" + extension
    else:
        outputname = basename + ".sliced.xmi"
inputsmname = basename + ".original_sm.dot"
outputsmname = basename + ".sliced_sm.dot"
cfgname = basename + ".cfg.dot"

# load the project
project = coral.loadDocument(modelname)

# find the model in the project file
model = None
for m in project.contents():
    #print "C: " + str(m)
    if m.oclIsKindOf(UML.Model):
        if not model:
            model = m
        else:
            abort("multiple models in file")
if not model:
    abort("no model in file")

cfg = slicer.CFG(model)
print "%d statemachine(s)" % (len(cfg.list_of_cfgs),)

# print original state machines to dot file
if cmd_options['inputsm'].get_value():
    printer.print_state_machines_to_file(inputsmname, model)
    printer.print_state_machines_to_separate_files(inputsmname, model)
    print "state machines in model wrote to file '" + inputsmname + "'"

# slice model
cfg.gen_in_out_defs()
print "reaching definitions calculated"
cfg.gen_data_deps()
print "data dependencies calculated"
cfg.gen_ntscd()
print "non-termination sensitive control dependencies calculated"
cfg.gen_id()
print "interference dependencies calculated"
cfg.gen_dod()
print "decisive control dependencies calculated"
cfg.gen_slice()
print "control flow graph slice calculated"
cfg.get_statemachine_slice()
print "sliced UML model generated"

# print CFG to dot file
if cmd_options['cfg'].get_value():
    printer_options = printer.PrinterOptions()
    if cmd_options['dependencies'].get_value():
        printer_options.dod = False
        printer_options.dd = False
        printer_options.ntscd = False
        printer_options.ssd = False
        printer_options.id = False
        string = cmd_options['dependencies'].get_arg()
        for char in string:
            if char == 'd':
                printer_options.dd = True
            elif char == 'n':
                printer_options.ntscd = True
            elif char == 'o':
                printer_options.dod = True
            elif char == 's':
                printer_options.ssd = True
            elif char == 'i':
                printer_options.id = True
            else:
                print "unknown dependency flag: " + char
    else:
        printer_options.dod = True
        printer_options.dd = True
        printer_options.ntscd = True
        printer_options.ssd = True
        printer_options.id = True
    printer_options.separate_files = False
    printer.print_to_file(cfgname, cfg, printer_options)
    print "graph wrote to file '" + cfgname + "'"

# print sliced state machines to dot file
if cmd_options['outputsm'].get_value():
    printer.print_state_machines_to_file(outputsmname, model)
    printer.print_state_machines_to_separate_files(outputsmname, model)
    print "state machines in model wrote to file '" + outputsmname + "'"

# save sliced model
if not cmd_options['nosliced'].get_value():
    project.setFilename(outputname)
    coral.saveDocument(project)
    print "sliced model wrote to file '" + outputname + "'"

print
print "done"
