# -*- coding: iso-8859-1 -*-

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Tools for printing state machines and control flow graphs to dot format.

Author: Vesa Ojala <vojala@tcs.tkk.fi>
"""

import sys
import re
import coral
import slicer
import copy

#
# Prints a list of control-flow-graphs to dot format
#

class PrinterOptions:
    def __init__(self):
        self.dd = True
        self.ntscd = True
        self.dod = True
        self.ssd = True
        self.id = True
        self.in_out_defs = True
        self.separate_files = False

# escape string to format that can be printed as dot record label
def escape_to_dot_format(string):
    """ escapes dot format record labels """
    if string:
        #print string
        string = str(string)
        string = re.sub(r'<', r'\<', string)
        string = re.sub(r'>', r'\>', string)
        string = re.sub(r'\|', r'\|', string)
        string = re.sub(r'\}', r'\}', string)
        string = re.sub(r'\{', r'\{', string)
    return string

def escape_xml(string):
    """ escapes dot format html-like labels, not in use currently"""
    if string:
        string = re.sub(r'&', r'&amp;', string)
        string = re.sub(r'<', r'&lt;', string)
        string = re.sub(r'>', r'&gt;', string)
        string = re.sub(r'"', r'&quot;', string)
        string = re.sub(r"'", r'&apos;', string)
    return string

def print_CFGNode(file, node):
    """ Prints a CFGNode type node of CFG to dot format. Makes one node in
        dot-file but no arcs"""
    assert(isinstance(node, slicer.CFGNode) or isinstance(node, slicer.CFGSend))
    attribute_string = ""
    label = str(node.script)

    line = '  node%s [label="{<entry>%s' % (getid(node),
                                            escape_to_dot_format(label))
    line += '|refs: %s\\ndefs: %s|<exit>I: %s\\nO: %s}"]' % (
        ', '.join(node.refs),
        ', '.join(node.defs),
        ', '.join(['(%s, %s)' % (str(id(p) % 10), d) for p, d in node.defs_in]),
        ', '.join(['(%s, %s)' % (str(id(p) % 10), d) for p, d in node.defs_out]))
    if node.in_slice:
        line += '[color=blue]'
    line += ';'

    print >> file, line
    
def print_CFGBranch(file, node):
    """ Prints a CFGBranch type node of CFG to dot format. Makes one node to
        dot-file but no arcs"""
    assert(isinstance(node, slicer.CFGBranch))
    size = len(node.branches)

    line = '  node%s [label="{<entry>I: %s| ' % (getid(node),
                                                 ', '.join(['(%s, %s)' %
                                                            (str(id(p) % 10), d)
                                                            for p, d
                                                            in node.defs_in]))
    cells = '|'.join(['{<sig%s>T: %s\\nslice: %s|G: %s|refs: %s\\ndefs: %s|<exit%s>O: %s}'
                      % (getid(b.signal),
                         escape_to_dot_format(b.signal),
                         ', '.join([a.definition for a in b.signal.assignments if a.in_slice]),
                         escape_to_dot_format(b.guard),
                         ', '.join(b.refs),
                         ', '.join(b.defs),
                         getid(b),
                         ', '.join(['(%s, %s)' % (str(id(p) % 10), d)
                                    for p, d in b.defs_out]))
                      for b in node.branches])
    line += '{%s}}"]' % (cells,)
    if node.is_in_slice():
        line += '[color=blue]'
    line += ';'

    print >> file, line
            

def print_nodes(file, node, counter=0, visited=None):
    """ prints nodes in cfg starting from node given """
    if not visited:
        visited = {}

    visited[node] = 1

    if isinstance(node, slicer.CFGNode) or isinstance(node, slicer.CFGSend):
        print_CFGNode(file, node)
    elif isinstance(node, slicer.CFGBranch):
        print_CFGBranch(file, node)
    else:
        assert False
    counter += 1
    for t in node.get_outgoing():
        if not visited.has_key(t):
            counter = print_nodes(file, t, counter, visited)
    return counter

def print_transitions(file, node, counter=0, visited=None):
    """ prints transitions in cfg starting from node given """
    if not visited:
        visited = {}

    visited[node] = 1

    if isinstance(node, slicer.CFGNode) or isinstance(node, slicer.CFGSend):
        for t in node.get_outgoing():
            print >> file, '    node%s:%s -> node%s;' % (getid(node),
                                                            'exit',
                                                            getid(t))
    elif isinstance(node, slicer.CFGBranch):
        for b in node.branches:
            target = b.outgoing
            print >> file, '    node%s:%s -> node%s;' % (getid(node),
                                                            'exit%s'
                                                            % getid(b),
                                                            getid(target))
    else:
        assert False
    
    counter += 1
    for t in node.get_outgoing():
        if not visited.has_key(t):
            counter = print_transitions(file, t, counter, visited)
    return counter

# prints datadependency arcs in dot-file
def print_data_deps(file, nodes):
    """ prints data-dependency arcs to dot file from list of nodes given """
    for n in nodes:
        # solmunlaajuiset datariippuvuudet
        for d in n.deps_dd:
            if isinstance(d, slicer.CFGNode):
                print >> file, '  node%s -> node%s;' % (getid(n),
                                                        getid(d))
##             elif isinstance(d, slicer.Branch_element):
##                 assert False
##                 print >> file, '  node%s -> node%s:gugu%s;' % (str(id(n)),
##                                                            str(id(d.parent)),
##                                                            str(id(d.signal)))
            elif isinstance(d, slicer.Assignment):
                print >> file, '  node%s -> node%s:sig%s [label="%s"];' % (getid(n),
                                        getid(d.parent.parent.parent),
                                        getid(d.parent),
                                        d.definition)
            else:
                assert(False)

        # signaalien parametrien generoimat datariippuvuudet
        if isinstance(n, slicer.CFGBranch):
            for b in n.branches:
                for a in b.signal.assignments:
                    for d in a.deps:
                        if isinstance(d, slicer.CFGNode):
                            print >> file, '  node%s:sig%s -> node%s [label="%s"];' % (
                                getid(n),
                                getid(b.signal),
                                getid(d),
                                a.definition)
                        elif isinstance(d, slicer.Assignment):
                            print >> file, '  node%s:sig%s -> node%s:sig%s [label="%s"];' % (
                                getid(n),
                                getid(b.signal),
                                getid(d.parent.parent.parent),
                                getid(d.parent),
                                a.definition)
                        else:
                            assert False

# prints datadependency arcs in dot-file
def print_id(file, nodes):
    """ prints data-dependency arcs to dot file from list of nodes given """
    for n in nodes:
        # solmunlaajuiset datariippuvuudet
        for d in n.deps_id:
            if isinstance(d, slicer.CFGNode):
                print >> file, '  node%s -> node%s;' % (getid(n),
                                                        getid(d))
##             elif isinstance(d, slicer.Branch_element):
##                 assert False
##                 print >> file, '  node%s -> node%s:gugu%s;' % (str(id(n)),
##                                                            str(id(d.parent)),
##                                                            str(id(d.signal)))
            elif isinstance(d, slicer.Assignment):
                print >> file, '  node%s -> node%s:sig%s [label="%s"];' % (getid(n),
                                        getid(d.parent.parent.parent),
                                        getid(d.parent),
                                        d.definition)
            else:
                assert(False)

        # signaalien parametrien generoimat datariippuvuudet
        # !!!!!!!!!!!!!!!!!!!! disabloitu ->
        # sama koodi kuin yll datariippuvuuksien kohdalla, johtuu siit ett signaaliparametrien tapauksessa
        # interferenssiriippuvuudet laitetaan samaan joukkoon kuin normaalit datariippuvuudet
        if isinstance(n, slicer.CFGBranch) and False:
            for b in n.branches:
                for a in b.signal.assignments:
                    for d in a.deps:
                        if isinstance(d, slicer.CFGNode):
                            print >> file, '  node%s:sig%s -> node%s [label="%s"];' % (
                                getid(n),
                                getid(b.signal),
                                getid(d),
                                a.definition)
                        elif isinstance(d, slicer.Assignment):
                            print >> file, '  node%s:sig%s -> node%s:sig%s [label="%s"];' % (
                                getid(n),
                                getid(b.signal),
                                getid(d.parent.parent.parent),
                                getid(d.parent),
                                a.definition)
                        else:
                            assert False

# prints ntscd arcs in dot-file 
def print_ntscd(f, nodes):
    """ prints ntscd arcs to dot file from list of nodes given """
    for n in nodes:
        for d in n.deps_ntscd:
            print >> f, '  node%s -> node%s' % (getid(n),
                                                getid(d))

# prints dod arcs in dot-file
def print_dod(f, nodes):
    for n in nodes:
        for d in n.deps_dod:
            print >> f, '  node%s -> node%s [color=sienna4]' %(getid(n),
                                                                getid(d[0]))
            print >> f, '  node%s -> node%s [color=sienna3]' %(getid(n),
                                                                  getid(d[1]))

def print_incoming_signals(f, nodes):
    for node in [n for n in nodes if isinstance(n, slicer.CFGBranch)]:
        for b in node.branches:
            for i in b.signal.incoming:
                print >> f, '  node%s -> node%s' % (getid(node),
                                                    getid(i))

def print_to_file(fname, cfg_object, options):
    """ Prints a list of cfgs in dot format to file """
    if (not options.dd) and (not options.ntscd) and (not options.dod) and (not options.ssd) and (not options.id) and options.separate_files:
        options.separate_files = False
        for i, item in enumerate(cfg_object.list_of_cfgs):
            temp_cfg = copy.copy(cfg_object)
            temp_cfg.list_of_cfgs = [item]
            print_to_file(fname+"."+str(i), temp_cfg, options)
        return

    f = open(fname, "w")
    print >> f, "digraph g {"
    print >> f, "  node [shape=record];"

    list_of_cfgs = cfg_object.list_of_cfgs

    for i, cfg in enumerate(list_of_cfgs):
        print >> f, "subgraph cluster_cfg%d {" % i
        print >> f, "  edge [color=black];"
        print_nodes(f, cfg)
        print_transitions(f, cfg)
        print >> f, "}"

    if options.dd:
        print >> f, "subgraph datadeps {"
        print >> f, "  edge [color=red];"
        print_data_deps(f, cfg_object.get_nodes())
        print >> f, "}"

    if options.ntscd:
        print >> f, "subgraph ntscd {"
        print >> f, "  edge [color=green];"
        print_ntscd(f, cfg_object.get_nodes())
        print >> f, "}"

    if options.ssd:
        print >> f, "subgraph ssd {"
        print >> f, "  edge [color=orchid];"
        print_incoming_signals(f, cfg_object.get_nodes())
        print >> f, "}"

    if options.dod:
        print >> f, "subgraph dod {"
        print_dod(f, cfg_object.get_nodes())
        print >> f, "}"

    if options.id:
        print >> f, "subgraph id {"
        print >> f, "  edge [color=yellow2];"
        print_id(f, cfg_object.get_nodes())
        print >> f, "}"

    print >> f, "}"
    f.close()

def getid(node):
    temp = id(node) + (sys.maxint+1)
    return "%u" % temp

def collect_elements(element, type, elements):
    if element.oclIsKindOf(type):
        #print 'Found "' + str(element.name) + '" (' + type.getName() + ')'
        elements.append(element)
    if element.oclIsKindOf(UML.Namespace):
        for e in element.ownedElement:
            collect_elements(e, type, elements)

UML = coral.lang.UML14

#
# Statemachine printer
#
def print_state_machines_to_file(fname, model):

    state_machines = []
    collect_elements(model, UML.StateMachine, state_machines)
    
    """ Prints simple state machines in coral model to dot-file """
    f = open(fname, "w")
    print >> f, "digraph g {"
    print >> f, "  node [shape=box];"
    for sm in state_machines:
        for v in sm.top.subvertex:
            print_state(f, v)
    print >> f, ""
    for sm in state_machines:
        for v in sm.top.subvertex:
            print_outgoing(f, v)
    print >> f, "}"
    f.close()

def print_state_machines_to_separate_files(fname, model):

    state_machines = []
    collect_elements(model, UML.StateMachine, state_machines)
    
    """ Prints simple state machines in coral model to dot-file """
    for i, sm in enumerate(state_machines):
        f = open(fname + "." + str(i), "w")
        print >> f, "digraph g {"
        print >> f, "  node [shape=box];"
        for v in sm.top.subvertex:
            print_state(f, v)
        print >> f, ""
        for v in sm.top.subvertex:
            print_outgoing(f, v)
        print >> f, "}"
        f.close()

def print_state(f, vertex):
    assert vertex.oclIsKindOf(UML.StateVertex)

    if vertex.oclIsKindOf(UML.CompositeState):
        raise "CompositeStates not handled"
    elif vertex.oclIsKindOf(UML.SubmachineState):
        raise "Submachines not handled"
    elif vertex.oclIsKindOf(UML.Pseudostate):
        if vertex.kind == "initial":
            print >> f, "  " + str(id(vertex)) + " [shape=circle, fillcolor=black,",
            print >> f, "style=filled, label=\"" + str(vertex.name) + "\"];"
        elif vertex.kind == "choice":
            print >> f, "  " + str(id(vertex)) + " [",
            print >> f, "style=\"rounded\", label=\"" + str(vertex.name) + "\"];"
    elif vertex.oclIsKindOf(UML.FinalState):
        print >> f, "  " + str(id(vertex)) + " [shape=doublecircle, fillcolor=black,",
        print >> f, "style=filled, label=\"" + str(vertex.name) + "\"];"
    elif vertex.oclIsKindOf(UML.SimpleState):
        print >> f, "  " + str(id(vertex)) + " [label=\"" + str(vertex.name) + "\"];"
    else:
        print vertex
        assert False and "Vertex type not handled"

def print_outgoing(f, vertex):
    assert vertex.oclIsKindOf(UML.StateVertex)
    
    if vertex.oclIsKindOf(UML.CompositeState):
        raise "Vertex type not handled"
    for t in vertex.outgoing:
        assert t.source
        assert t.target

        assert not t.source.oclIsKindOf(UML.CompositeState)
        assert not t.target.oclIsKindOf(UML.CompositeState)

        f.write("\"" + str(id(t.source)) + "\"")
        f.write(" -> ")
        f.write("\"" + str(id(t.target)) + "\" ")

        f.write("[")

        f.write("label=\"")
        if t.trigger:
            assert t.trigger.oclIsKindOf(UML.SignalEvent)
            f.write(str(t.trigger.signal.name))
            f.write("(")
            f.write(", ".join(
                [str(t.trigger.parameter[i].name)
                 for i in range(0, len(t.trigger.parameter))]))
            f.write(")")
        if t.guard:
            assert t.guard.expression
            assert t.guard.expression.body
            f.write("['"+str(t.guard.expression.body) + "']")
        if t.effect:
            assert t.effect.script
            assert t.effect.script.body
            f.write("/'" + str(t.effect.script.body) + "'")
        f.write("\"")

        f.write("];\n")

