#!/usr/bin/python

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""Find variables defined and referenced in a simple Jumbala statement.

To use as a script, run with a Jumbala statement as a parameter.

Limitations:
   - Only assertion statements, empty statements, expression statements and
     send statements are supported.
   - Heap allocation ("new"), method invocations, "super"
     references, and array or expression initializers are not supported.

"""

import sets
import actlang

class Error(Exception):

    """Exception with a string message.

    Attributes:
        reason -- explanation for the cause of the exception.

    """

    def __init__(self, msg=""):
        """Initialize an Error with an optional message."""
        self.reason = msg

class UnsupportedFeatureError(Error):

    """An exception class for reporting unsupported features."""

    def __init__(self, msg=""):
        """Initialize an UnsupportedFeatureError with an optional message."""
        Error.__init__(self, msg)

class ConstraintError(Error):

    """An exception class for reporting unsatisfied constraints."""

    def __init__(self, msg=""):
        """Initialize a ConstraintError with an optional message."""
        Error.__init__(self, msg)

class DefRef:

    """Variables def'd and ref'd in a simple Jumbala expression or statement.

    Instance attributes:
    defs -- sets.Set of string identifiers of defined variables.
    refs -- sets.Set of string identifiers of referenced variables.

    """

    def __init__(self, stmt_or_expr):
        """Initialize a def-ref pair from a Jumbala expression or statement.

        Constraints:
          stmt_or_expr must be one of
            actlang.expr.Expr
            actlang.statement.AssertStatement
            actlang.statement.EmptyStatement
            actlang.statement.ExprStatement
            actlang.statement.SendStatement

        """
        #print "rd: " + str(stmt_or_expr)

        self.defs = sets.Set()
        self.refs = sets.Set()
        self.signal_name = ""
        self.signal_params = []
        self.signal_target = ""

        #print stmt_or_expr
        #print stmt_or_expr.__class__.__name__
        if isinstance(stmt_or_expr, actlang.expr.Expr):
            exprs = [stmt_or_expr]
        elif isinstance(stmt_or_expr, actlang.names.Name):
            exprs = [stmt_or_expr]
        elif isinstance(stmt_or_expr, actlang.statement.AssertStatement) or \
             isinstance(stmt_or_expr, actlang.statement.ExprStatement):
            exprs = [stmt_or_expr.api_getExpr()]
        elif isinstance(stmt_or_expr, actlang.statement.EmptyStatement):
            return
        elif isinstance(stmt_or_expr, actlang.statement.SendStatement):
            self.signal_name = stmt_or_expr.api_getSignalName()
            exprs = []
            #exprs = stmt_or_expr.api_getArgList()
            for arg in stmt_or_expr.api_getArgList():
                #if isinstance(arg, actlang.names.Name):
                self.signal_params.append(str(arg))
                #else:
                #    raise UnsupportedFeatureError("only SimpleNames supprted"
                #                                  "as signal parameters")
            self.signal_target = stmt_or_expr.api_getTarget()
            #exprs.append(stmt_or_expr.api_getTarget())
        else:
            raise ConstraintError("DefRef: unsupported statement type ("
                                  + stmt_or_expr.__class__.__name__ + ")")

        for e in exprs:
            self.collectDefsAndRefs(e)

    def collectDefsAndRefs(self, expr):
        """Collect identifiers defined or referenced in an expression."""

        # 'new' expressions can't have parameters but are otherwise supported
        if isinstance(expr, actlang.expr.InstanceCreationExpr):
            assert not expr.api_getArgList() # parameters not supported
            return # no need to do anything
            
        # method invocations are not supported
        if isinstance(expr, actlang.expr.MethodInvocationExpr):
            raise UnsupportedFeatureError("method invocation expressions not "
                                          "supported")

        if DefRef.subexpression_accessors.has_key(expr.__class__):
            # collect defs and refs of subexpressions
            defrefs = []
            for accessor in self.subexpression_accessors[expr.__class__]:
                defrefs.append(DefRef(accessor(expr)))

            # simple and qualified names
            if isinstance(expr, (actlang.names.Name,
                                 actlang.expr.PrimaryFieldAccessExpr)):
                self.refs.add(str(expr))

            # array creations
            elif isinstance(expr, actlang.expr.ArrayCreationExpr):
                assert expr.api_getInitializer() == None # new int[3] {a, b, c, d}; not supported 
                for dim in expr.api_getDimensions():
                    defrefs.append(DefRef(dim))
                for dr in defrefs:
                    self.defs.update(dr.defs)
                    self.refs.update(dr.refs)

            # expressions with side effects
            elif isinstance(expr, actlang.expr.SimpleAssignmentExpr) or \
                 isinstance(expr, actlang.expr.CompoundAssignmentExpr) or \
                 isinstance(expr, actlang.expr.IncrementOrDecrementExpr):
                if not isinstance(expr, actlang.expr.IncrementOrDecrementExpr):
                    # defrefs contains a single DefRef for the rhs of the expr
                    self.defs.update(defrefs[0].defs)
                    self.refs.update(defrefs[0].refs)
                    lvalue = expr.api_getLhsExpr()
                else:
                    lvalue = expr.api_getOperand()
                if isinstance(lvalue, actlang.names.Name):
                    self.defs.add(str(lvalue))
                    if not isinstance(expr, actlang.expr.SimpleAssignmentExpr):
                        self.refs.add(str(lvalue))
                elif isinstance(lvalue, actlang.expr.ArrayAccessExpr):
                    dr = DefRef(lvalue.api_getArrayExpr())
                    assert(len(dr.defs) == 0)
                    assert(len(dr.refs) == 1)
                    self.defs.update(dr.refs)
                    self.refs.update(dr.refs)
                    dr = DefRef(lvalue.api_getIndexExpr())
                    self.defs.update(dr.defs)
                    self.refs.update(dr.refs)
                else:
                    raise UnsupportedFeatureError('expression "' + str(expr) +
                                                  '" is of the unsupported '
                                                  'type '
                                                  + lvalue.__class__.__name__)

            # other (supported) expressions
            else:
                for dr in defrefs:
                    self.defs.update(dr.defs)
                    self.refs.update(dr.refs)
        else:
            raise UnsupportedFeatureError('expression "' + str(expr) +
                                          '" is of the unsupported type ' +
                                          expr.__class__.__name__)

    # hash table for mapping (supported) expression types into tuples of
    # functions for accessing their subexpressions
    subexpression_accessors = {
        # not supported: ArrayCreationExpr
        actlang.expr.ArrayCreationExpr : (),
        actlang.expr.BitNotExpr : (actlang.expr.BitNotExpr.api_getOperand,),
        actlang.expr.BoolLiteralExpr : (),

        # subclasses of CastExpr
        actlang.expr.NameCastExpr :
        (actlang.expr.NameCastExpr.api_getTypeExpr,),
        actlang.expr.TypeCastExpr :
        (actlang.expr.TypeCastExpr.api_getOperand,),
        # end subclasses of CastExpr

        actlang.expr.CondAndExpr :
        (actlang.expr.CondAndExpr.api_getLeftOperand,
         actlang.expr.CondAndExpr.api_getRightOperand),
        actlang.expr.CondOrExpr :
        (actlang.expr.CondOrExpr.api_getLeftOperand,
         actlang.expr.CondOrExpr.api_getRightOperand),
        actlang.expr.InfixExpr :
        (actlang.expr.InfixExpr.api_getLeftOperand,
         actlang.expr.InfixExpr.api_getRightOperand),
        actlang.expr.InstanceOfExpr :
        (actlang.expr.InstanceOfExpr.api_getOperand,),
        actlang.expr.IntLiteralExpr: (),

        # subclasses of LeftHandSideExpr
        actlang.expr.ArrayAccessExpr :
        (actlang.expr.ArrayAccessExpr.api_getArrayExpr,
         actlang.expr.ArrayAccessExpr.api_getIndexExpr),
        # not supported: PrimaryFieldAccessExpr
        # actlang.expr.PrimaryFieldAccessExpr :
        # (actlang.expr.PrimaryFieldAccessExpr.api_getFieldName,),
        # not supported: SuperFieldAccessExpr
        # end subclasses of LeftHandSideExpr

        actlang.expr.LogicalNotExpr :
        (actlang.expr.LogicalNotExpr.api_getOperand,),
        actlang.expr.NullExpr : (),
        actlang.expr.ParenExpr : (actlang.expr.ParenExpr.api_getInnerExpr,),

        # subclasses of StatementExpr
        actlang.expr.CompoundAssignmentExpr :
        (actlang.expr.CompoundAssignmentExpr.api_getRhsExpr,),
        # subclasses of IncrementOrDecrementExpr
        actlang.expr.PostfixExpr : (),
        actlang.expr.PrefixExpr : (),
        # end subclasses of IncrementOrDecrementExpr
        # not supported: InstanceCreationExpr
        # subclasses of MethodInvocationExpr
        # not supported: NameMethodInvocationExpr
        # not supported: PrimaryMethodInvocationExpr
        # not supported: SuperMethodInvocationExpr
        # end subclasses of MethodInvocationExpr
        actlang.expr.SimpleAssignmentExpr :
        (actlang.expr.SimpleAssignmentExpr.api_getRhsExpr,),
        # end subclasses of StatementExpr

        actlang.expr.StringLiteralExpr : (),
        actlang.expr.TernaryCondExpr :
        (actlang.expr.TernaryCondExpr.api_getCondition,
         actlang.expr.TernaryCondExpr.api_getTrueExpr,
         actlang.expr.TernaryCondExpr.api_getFalseExpr),
        actlang.expr.ThisExpr : (),
        actlang.expr.UnaryMinusExpr :
        (actlang.expr.UnaryMinusExpr.api_getOperand,),
        actlang.expr.UnaryPlusExpr :
        (actlang.expr.UnaryPlusExpr.api_getOperand,),

        # subclasses of VarInitializer
        # ArrayInitializer: not supported
        # ExprInitializer: not supported
        # end subclasses of VarInitializer

        actlang.names.SimpleName : (),
        actlang.names.QualifiedName : () }



if __name__ == "__main__":
    import sys
    if len(sys.argv) == 2:
        env = actlang.basetypes.getInitialEnvironment()
        tree = actlang.grammar.parse(sys.argv[1])
        stmts = tree.api_getTopStatements()
        try:
            if len(stmts) != 1:
                raise UnsupportedFeatureError("compound statements not "
                                              "supported")

            dr = DefRef(stmts[0])
            print "Defs: " + ", ".join(dr.defs)
            print "Refs: " + ", ".join(dr.refs)
        except Error, e:
            print ("%s: " % sys.argv[0]) + e.reason
