# coding: latin-1

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""State machine slicer.

Author: Vesa Ojala <vojala@tcs.tkk.fi>
"""

import coral
import types
import refdef
import actlang
import sets
import copy

def abort(msg="aborted"):
    raise msg

UML = coral.lang.UML14
slice_assert = False


def collect_elements(element, type, elements):
    """build a list of elements of a given type in the model
    """
    if element.oclIsKindOf(type):
        #print 'Found "' + str(element.name) + '" (' + type.getName() + ')'
        elements.append(element)
    if element.oclIsKindOf(UML.Namespace):
        for e in element.ownedElement:
            collect_elements(e, type, elements)

#
# Layer between the UML model and the CFG
#

class State:
    """Class linking UML state to corresponding elements in CFG
    """
    def __init__(self, original, cls):
        self.original = original
        self.cfg_node = CFGBranch(cls)
        self.transitions = {}
        self.associated_class = cls

    def get_associated_class(self):
        return self.associated_class

    def add_transition(self, original):
        this_class = original.stateMachine.context

        signal = Signal()
        if original.trigger:
            if not original.trigger.oclIsKindOf(UML.SignalEvent):
                abort("unsupported type of trigger encountered")
            signal.name = str(original.trigger.signal.name)
            for i in range(0, len(original.trigger.parameter)):
                name = str(original.trigger.parameter[i].name)
                # variables receiving signal parameters has to be simple names
                assert check_simple_name(name)
                param = Variable(this_class, name)
                signal.add_parameter(param)

        if original.guard:
            guard = original.guard.expression.body
        else:
            guard = ""

        new_element = self.cfg_node.add(signal, guard, None, this_class)

        if original.effect:
            effects = expand_effect_to_list(str(original.effect.script.body))
        else :
            effects = []
        
        new_trans = Transition(self, original, new_element, effects)
        self.transitions[original] = new_trans
        return new_trans

    def add_selfloop(self):
        new_element = self.cfg_node.add(Signal(), "", None, self.get_associated_class())
        new_element.outgoing = self.cfg_node
        self.cfg_node.incoming.append(new_element)

class Transition:
    """Class linking UML transition to corresponding elements in CFG
    """
    def __init__(self, parent, original, element, effects=None):
        assert isinstance(parent, State)
        self.parent = parent
        self.original = original
        self.element = element
        self.slicing_crit = False

        # slicing criterion true?
        for tg in original.taggedValue:
            if tg.name == "slicingCrit":
                self.slicing_crit = True
                self.parent.cfg_node.in_slice = True #whole cfg node in slice

        # construct nodes for effect
        if not effects:
            self.effects = []
        else:
            self.effects = [Effect(self, e, self.slicing_crit) for e in effects]
            
            #connect Branch_element and first effect node
            self.element.outgoing = self.get_effect_entrypoint()
            self.get_effect_entrypoint().incoming = self.element

            
            for a, b in [(self.effects[i], temp)
                         for i, temp in enumerate(self.effects[1:])]:
                a.get_exitpoint().outgoing = b.get_entrypoint()
                b.get_entrypoint().incoming = a.get_exitpoint()

        assert parent != None
        assert element != None

    def make_and_add_effect(self, effect):
        assert isinstance(slising_crit, types.BooleanType)
        old_exitpoint = self.get_exitpoint()
        old_successor = self.get_exitpoint().outgoing
        new_effect = self.effects.append(Effect(self, effect, self.slicing_crit))

        # update arcs in graph
        new_effect.get_entrypoint().incoming = old_exitpoint
        new_effect.get_exitpoint().outgoing = old_successor
        old_exitpoint.outgoing = new_effect.get_entrypoint()
        old_successor.incoming = new_effect.get_exitpoint()

    def get_effect_entrypoint(self):
        if len(self.effects) > 0:
            return self.effects[0].get_entrypoint()
        else:
            return None

    def get_effect_exitpoint(self):
        if len(self.effects) > 0:
            return self.effects[-1].get_exitpoint()
        else:
            return None

    def get_exitpoint(self):
        if len(self.effects) > 0:
            return self.get_effect_exitpoint()
        else:
            return self.element

    def get_associated_class(self):
        return self.parent.get_associated_class()

    def get_send_nodes(self):
        return [e.effect for e in self.effects if isinstance(e.effect, CFGSend)]

class Effect:
    """Class linking UML transitions effect to corresponding elements in CFG
    """
    def __init__(self, parent, effect, slicing_crit, predecessor=None, successor=None):
        assert isinstance(parent, Transition)
        assert isinstance(slicing_crit, types.BooleanType)
        self.parent = parent
        self.effect = None
        self.entrypoint = None
        self.exitpoint = None

        rd = get_refdef(effect, "statement")

        # Send statements
        if rd.signal_name:
            #pred = new_branch
            param_nodes = []
            # construct evaluator for signal target

            this_class = self.get_associated_class()
            to_node = CFGNode(this_class) # THIS CLASS???

            self.entrypoint = to_node
            to_node.script = str(rd.signal_target)

            # refs & defs
            rd2 = get_refdef(to_node.script, "expression")
            rd2lists = expand_refdef_to_lists(rd2)
            refs = sets.Set([Variable(this_class, r) for r in rd2lists[0]])
            to_node.refs = refs
            defs = sets.Set([Variable(this_class, d) for d in rd2lists[1]])
            to_node.defs = defs

            # predecessor so that the next node can be updated
            pred = to_node

            # construct evaluators for signal parameters
            for param in rd.signal_params:
                param_node = CFGNode(this_class)
                param_nodes.append(param_node)

                # link between nodes
                param_node.incoming = pred
                pred.outgoing = param_node

                param_node.script = str(param)

                # refs & defs
                rd2 = get_refdef(param_node.script, "expression")
                rd2lists = expand_refdef_to_lists(rd2)
                refs = sets.Set([Variable(this_class, r) for r in rd2lists[0]])
                param_node.refs = refs
                defs = sets.Set([Variable(this_class, d) for d in rd2lists[1]])
                param_node.defs = defs

                # predecessor so that the next node can be updated
                pred = param_node

            # finally the actual send node
            send_node = CFGSend(this_class)
            send_node.signal = rd.signal_name
            send_node.target = to_node
            send_node.params = param_nodes

            # link between nodes
            pred.outgoing = send_node
            send_node.incoming = pred

            send_node.script = effect

            # exitpoint & effect updated
            self.exitpoint = send_node
            self.effect = send_node

        # Other actions 
        else:
            this_class = self.get_associated_class()
            node = CFGNode(this_class)
            node.script = effect
            # if slice_assert is True, add node to slice
            if slice_assert:
                if check_assert(node.script):
                    node.in_slice = True
            
            # refs & defs
            rdlists = expand_refdef_to_lists(rd)
            node.refs = sets.Set([Variable(this_class, r) for r in rdlists[0]])
            node.defs = sets.Set([Variable(this_class, d) for d in rdlists[1]])

            self.effect = node
            self.entrypoint = node
            self.exitpoint = node

        assert self.effect != None
        assert self.entrypoint != None
        assert self.exitpoint != None

        # add nodes to slice if they belong to slicing criterion
        if slicing_crit:
            if isinstance(self.effect, CFGSend):
                for p in self.effect.params:
                    p.in_slice = True
                self.effect.target.in_slice = True
            self.effect.in_slice = True
        
        # set incoming/outgoing correctly if possible
        if predecessor:
            self.entrypoint.incoming = predecessor
        if successor:
            self.entrypoint.outgoing = successor
        
    def get_entrypoint(self):
        return self.entrypoint

    def get_exitpoint(self):
        return self.exitpoint

    def get_associated_class(self):
        return self.parent.get_associated_class()

#
# Control Flow Graph constructions
#

class Variable(str):
    """Class representing one occurrence of certain variable, ref and def -sets
    are made up of these
    """
    def __new__(cls, belongs, variable):
        assert belongs.oclIsKindOf(UML.Class)
        assert isinstance(variable, types.StringType)

        tokens = variable.split('.')
        variable = tokens.pop(-1)

        if len(tokens) == 0:
            local = True
        else:
            local = False
            for token in tokens:
                for as in belongs.association:
                    assert len(as.association.connection) == 2 #only 2-way associations
                    for connection in as.association.connection:
                        if not id(as) == id(connection):
                            if connection.name == token:
                                belongs = connection.participant
            
        new_object = str.__new__(cls, str(belongs.name) + "." + variable)
        new_object.local = local
        new_object.belongs = belongs
        new_object.variable = variable

        return new_object

    def __init__(self, belongs, variable):
        pass

    def __str__(self):
        return str(self.belongs.name) + "." + self.variable

    def __cmp__(self, other):
        if self.belongs != other.belongs:
            return self.belongs.name.__cmp__(other.belongs.name)
        if self.variable != other.variable:
            if self.variable < other.variable:
                return -1
            else:
                return 1
        return 0

class CFGSend:
    """Node representing signal sends in CFG
    """
    def __init__(self, owner):
        self.in_slice = False
        self.signal = ""
        self.target = None
        self.params = [] # a list of CFGNodes
        self.script = ""
        self.incoming = None
        self.outgoing = None
        
        self.refs = sets.Set()
        self.defs = sets.Set()

        self.defs_in = sets.Set()
        self.defs_out = sets.Set()
        self.deps_dd = sets.Set()
        self.deps_ntscd = sets.Set()
        self.deps_dod = sets.Set()
        self.deps_id = sets.Set()

        self.owner = owner

    def get_outgoing(self):
        return [self.outgoing]

    def get_incoming(self):
        return [self.incoming]

    def is_in_slice(self):
        return self.in_slice

    def is_param_or_target_in_slice(self):
        status = False
        for param in self.params:
            if param.is_in_slice():
                status = True
                
        if self.target.is_in_slice():
            status = True

        return status

class CFGNode:
    """Node representing simple actions in CFG
    """
    def __init__(self, cls, script=None, incoming=None, outgoing=None):
        self.in_slice = False
        
        if script != None:
            self.script = script
        else:
            self.script = ""
            
        self.incoming = incoming
        self.outgoing = outgoing

        if self.script:
            rd = get_refdef(script, "statement")
            rdlists = expand_refdef_to_lists(rd)
            self.refs = sets.Set([Variable(cls, r) for r in rdlists[0]])
            self.defs = sets.Set([Variable(cls, d) for d in rdlists[1]])
            self.sending = rd.signal_name
            self.sending_params = rd.signal_params
        else:
            self.sending = ""
            self.sending_params = []
            self.refs = sets.Set()
            self.defs = sets.Set()

        self.defs_in = sets.Set()
        self.defs_out = sets.Set()
        self.deps_dd = sets.Set()
        self.deps_ntscd = sets.Set()
        self.deps_dod = sets.Set()
        self.deps_id = sets.Set()

        self.owner = cls
        
    def get_outgoing(self):
        return [self.outgoing]

    def get_incoming(self):
        return [self.incoming]

    def get_refs(self):
        return self.refs

    def get_defs(self):
        return self.defs

    def is_in_slice(self):
        return self.in_slice

class Assignment:
    """represents basically statement 'definition = reference;'
    used to link reference to definition over signal sendings
    part of Signal class
    """
    def __init__(self, parent):
        self.parent = parent
        self.deps = sets.Set() # contains also interference dependencies
        self.definition = None
        self.in_slice = False

class Signal:
    """A trigger class, part of Branch_element
    """
    def __init__(self):
        self.parent = None
        self.incoming = []
        self.name = ""
        self.assignments = sets.Set()

    def add_parameter(self, param):
        a = Assignment(self)
        a.definition = param
        self.assignments.add(a)

    def add_incoming(self, node):
        pass

    def get_defs(self):
        return [a.definition for a in self.assignments]

    def __str__(self):
        if self.name:
            return self.name + '(' + ', '.join(self.get_defs()) + ')'
        else:
            return ""

    def __nonzero__(self):
        if self.name:
            return 1
        else:
            return 0

class Branch_element:
    """A part of CFGBranch representing one outgoing transition
    """
    def __init__(self, parent):
        assert(isinstance(parent, CFGBranch))
        self.parent = parent
        self.signal = None
        self.guard = ""
        self.outgoing = None # for getting outgoing arcs in cfg
#        self.action = None # for getting action of specific transition,
#                           # can jump over some nodes in cfg!!
        self.refs = sets.Set()
        self.defs_out = sets.Set()

    def get_defs(self):
        return self.signal.get_defs()

class CFGBranch:
    """CFG node for branches
    """
    def __init__(self, owner, incoming=None):
        if incoming != None:
            self.incoming = incoming
        else:
            self.incoming = []

        self.in_slice = False
        self.defs_in = sets.Set()
        self.deps_dd = sets.Set()
        self.deps_ntscd = sets.Set()
        self.deps_dod = sets.Set()
        self.deps_id = sets.Set()
        self.branches = []

        self.owner = owner

    # add new Branch_element to this branch (for new outgoing transition)
    def add(self, signal, guard, target, belongs):
        assert isinstance(signal, Signal)
        
        b = Branch_element(self)

        signal.parent = b
        b.signal = signal
        b.guard = guard
        b.outgoing = target

        if guard != "":
            rd = get_refdef(str(guard), "expression")
            rdlists = expand_refdef_to_lists(rd)
            b.refs = sets.Set([Variable(belongs, r) for r in rdlists[0]])
            b.defs = sets.Set([Variable(belongs, d) for d in rdlists[1]])
        else:
            b.refs = sets.Set()
            b.defs = sets.Set()

        self.branches.append(b)
        return b

    def is_in_slice(self):
        if self.in_slice:
            return True
        for b in self.branches:
            for a in b.signal.assignments:
                if a.in_slice:
                    return True
        return False

    def get_outgoing(self):
        return [b.outgoing for b in self.branches]

    def get_incoming(self):
        return self.incoming

class CFG:
    """Control flow graph.
    """
    def __init__(self, model):
        assert model.oclIsKindOf(UML.Model)
        self.model = model
        
        state_machines = []
        collect_elements(model, UML.StateMachine, state_machines)
        self.state_machines = state_machines

        mapping = {}
        lst = []
        self.sig_param_datatypes = collect_sig_param_types(model)
        self.senders = {}
        for s in state_machines:
            lst.append(self.make_cfg(s, mapping))
        self.mapping = mapping
        self.list_of_cfgs = lst

        self.gen_signal_sources()

    # construction of the cfg
    def make_cfg(self, sm, map2):
        initial_state = None
        for s in sm.top.subvertex:
            if s.oclIsKindOf(UML.CompositeState):
                abort('"' + str(sm.name) + '" is not a flat state machine')
            if s.oclIsKindOf(UML.Pseudostate) and s.kind == "initial":
                if not initial_state:
                    initial_state = s
                else:
                    abort('"' + str(sm.name) + '" has multiple initial states')
        if not initial_state:
            abort('"' + str(s.name) + '" has no initial state')

        assert sm.context.oclIsKindOf(UML.Class)
        temp = self.traverse_statemachine(initial_state, map2, sm.context)
        return temp

    def traverse_statemachine(self, state, map2, cls, visited=None):
        # check whether there is a State instance representing 'state'
        if map2.has_key(state):
            new_s = map2[state]
        else:
            new_s = State(state, cls)
            map2[state] = new_s

        if not visited:
            visited = {}

        visited[state] = 1

        # Create instance of Transition class and a CFBranch 
        if len(state.outgoing):
            # selfloop variable tracks whether we need to add selfloop to CFGBranch
            if (state.oclIsKindOf(UML.Pseudostate) and state.kind == "choice"):
                selfloop = False
            else:
                selfloop = True

            this_class = None
            # handle each outgoing transition
            for t in state.outgoing:
                new_trans = new_s.add_transition(t)
                                
                if not (t.trigger or t.guard):
                    selfloop = False
                
                # record signal sends
                for send in new_trans.get_send_nodes():
                    if self.senders.has_key(send.signal):
                        self.senders[send.signal].append(send)
                    else:
                        self.senders[send.signal] = [send]

                # create State instance representing the next UML state
                if not map2.has_key(t.target):
                    temp = State(t.target, cls)
                    map2[t.target] = temp
                new_trans.get_exitpoint().outgoing = map2[t.target].cfg_node
                map2[t.target].cfg_node.incoming.append(new_trans.get_exitpoint())

            # add selfloop to CFGBranch if needed
            if selfloop:
                new_s.add_selfloop()

        # continue recursively to the next UML state
        for t in state.outgoing:
            if not visited.has_key(t.target):
                self.traverse_statemachine(t.target, map2, cls, visited)

        return new_s.cfg_node

    # returns all nodes in the cfg
    def get_nodes(self, roots=None, visited=None, nodes=None):
        if roots == None:
            roots = self.list_of_cfgs
        if not visited:
            visited = {}
        if not nodes:
            nodes = []

        for root in roots:
            if not visited.has_key(root):
                visited[root] = True
                nodes.append(root)

                succs = [n for n in root.get_outgoing() if not visited.has_key(n)]
                self.get_nodes(succs, visited, nodes)

        return nodes

    #
    # data dependencies
    #
    def gen_in_out_defs(self):
        nodes = self.get_nodes()

        while True:
            break_condition = True

            for n in nodes:
                # update defs_in
                n.defs_in = sets.Set()
                for i in n.get_incoming():
                    n.defs_in.update(i.defs_out)

                # update defs_out for CFGNodes and CFGSends
                if isinstance(n, CFGNode) or isinstance(n, CFGSend):
                    old = n.defs_out
                    n.defs_out = sets.Set()

                    # defs from previous nodes that are not killed in this node
                    for d, v in n.defs_in:
                        if v not in n.defs:
                            n.defs_out.add((d, v))

                    # add defs from this node
                    for v in n.defs:
                        n.defs_out.add((n, v))

                    # update break condition
                    if old != n.defs_out:
                        break_condition = False

                # update defs_out for CFGBranches
                elif isinstance(n, CFGBranch):
                    for b in n.branches:
                        old = b.defs_out
                        b.defs_out = sets.Set()

                        # defs from previous nodes
                        for d, v in n.defs_in:
                            if v not in b.get_defs():
                                b.defs_out.add((d, v))

                        # new defs (they come from signal parameters)
                        for assignment in b.signal.assignments:
                            b.defs_out.add((assignment, assignment.definition)) 

                        # update break condition
                        if old != b.defs_out:
                            break_condition = False
                else:
                    assert(False)

            if break_condition:
                break

    # calculate data dependencies in cfg
    def gen_data_deps(self):
        nodes = self.get_nodes()
        for n in nodes:
            if isinstance(n, CFGNode):
                n.deps_dd.union_update([m for m, v in n.defs_in if v in n.refs])
            elif isinstance(n, CFGBranch):
                for b in n.branches:
                    new_deps = [m for m, v in n.defs_in if
                                (v in b.refs) and (not v in b.get_defs())]
                    n.deps_dd.union_update(new_deps)

                    # node is dependent on signal parameters that define variables referenced in the guard
                    for assign in b.signal.assignments:
                        if assign.definition in b.refs:
                            n.deps_dd.add(assign)

                    # assignments made by signal receiving are dependent on their evaluators on the sending side
                    for i, assign in enumerate(b.signal.assignments):
                        for source in b.signal.incoming:
                            assert isinstance(source, CFGSend)
                            source_deps = [source.params[i]]
                            assign.deps.union_update(source_deps)

    # finds all definitions of the variable where either the variable or the other definition is not local
    def __find_defs(self, variable):
        assert isinstance(variable, Variable)
        deps = []
        for m in self.get_nodes():
            if isinstance(m, CFGNode):
                for d in m.defs:
                    assert isinstance(d, Variable)
                    if variable == d and (variable.local == False or d.local == False):
                        deps.append(m)

            elif isinstance(m, CFGBranch):
                for b in m.branches:
                    for a in b.signal.assignments:
                        assert isinstance(a.definition, Variable)
                        if variable == a.definition and (variable.local == False or a.definition.local == False):
                            deps.append(a)
        return deps

    # calculates interference dependencies
    def gen_id(self):
        for n in self.get_nodes():
            if isinstance(n, CFGNode):
                for ref in n.refs:
                    n.deps_id.update(self.__find_defs(ref))
            elif isinstance(n, CFGBranch):
                for b in n.branches:
                    for ref in b.refs:
                        n.deps_id.update(self.__find_defs(ref))

    # fill senders dictionary with all possible senders for each signal type
    def gen_signal_sources(self):
        for node in self.get_nodes():
            if isinstance(node, CFGBranch):
                for branch in node.branches:
                    if (branch.signal and
                        self.senders.has_key(branch.signal.name) and
                        self.senders[branch.signal.name]):
                        branch.signal.incoming = self.senders[branch.signal.name]

    #
    # control dependencies
    #

    # calculation of non-termination sensitive control dependencies
    def gen_ntscd(self):
        nodes = self.get_nodes()
        workbag = sets.Set()
        s = {}
        cond_nodes = []

        for cond in [n for n in nodes if isinstance(n, CFGBranch)]:
            if len(cond.branches) > 1:
                cond_nodes.append(cond)
                for node in nodes:
                    s[(node, cond)] = sets.Set()
                for branch in cond.branches:
                    succ = branch.outgoing
                    assert s.has_key((succ, cond))
                    s[(succ, cond)].add((cond, succ))
                    workbag.add(succ)

        while len(workbag):
            node = workbag.pop()
            succs = node.get_outgoing()
            
            # one successor
            if (len(succs) == 1) and (id(succs[0]) != id(node)):
                succ = succs[0]
                for cond in cond_nodes:
                    assert s.has_key((node, cond))
                    if len(s[(node, cond)] - s[(succ, cond)]):
                        s[(succ, cond)].union_update(s[(node, cond)])
                        workbag.add(succ)

            # two or more successors
            elif (len(succs) > 1):
                for m in nodes:
                    if len(s[m, node]) == len(node.get_outgoing()):
                        for p in [c for c in cond_nodes if id(c) != id(node)]:
                            if len(s[(node, p)] - s[(m, p)]):
                                s[(m, p)].union_update(s[(node, p)])
                                workbag.add(m)

        for n in nodes:
            for m in cond_nodes:
                if 0 < len(s[(n, m)]) and len(s[(n, m)]) < len(m.get_outgoing()):
                    n.deps_ntscd.add(m)

    # calculation of decisive order dependencies
    def gen_dod(self):
        reachability = {}

        def dependence(n, m, p, nodes):
            color = {}
            for node in nodes:
                color[node] = "uncolored"
            color[m] = "white"
            color[p] = "black"

            visited = sets.Set()
            visited.add(m)
            visited.add(p)
            colored_dag(nodes, n, color, visited)

            white_child = False
            black_child = False
            for succ in n.get_outgoing():
                if color[succ] == "white":
                    white_child = True
                elif color[succ] == "black":
                    black_child = True
            return white_child and black_child

        def colored_dag(nodes, n, color, visited):
            if not n in visited:
                visited.add(n)
                succs = n.get_outgoing()
                if len(succs) > 0:
                    for succ in succs:
                        colored_dag(nodes, succ, color, visited)
                    c = color[succs[0]]
                    for succ in succs:
                        if color[succ] != c:
                            c = "uncolored"
                            break
                    color[n] = c
            return

        def all_reach(nodes):
            workbag = sets.Set(nodes)
            result = {}

            for n in nodes:
                result[n] = sets.Set([n])

            while len(workbag) > 0:
                node = workbag.pop()
                outgoing = node.get_outgoing()
                new_set = None

                if outgoing:
                    for out in outgoing:
                        if new_set == None:
                            new_set = sets.Set(result[out])
                        else:
                            new_set &= result[out]

                    current_set = result[node]
                    if not (current_set >= new_set):
                        current_set |= new_set
                        incoming = copy.copy(node.get_incoming())
                        for i, inc in enumerate(incoming):
                            if isinstance(inc, Branch_element):
                                incoming[i] = inc.parent
                        workbag.update(incoming)

            return result

        nodes = self.get_nodes()

        # init reachability
        reachability = all_reach(nodes)

        for cond in [n for n in nodes if isinstance(n, CFGBranch)]:
            for m in nodes:
                for p in [n for n in nodes if id(m) != id (n)]:

                    if ((m in reachability[p])
                        and (p in reachability[m])
                        and dependence(cond, p, m, nodes)):
                        #lis dod-relaatioon
                        p.deps_dod.add((cond, m))
                        m.deps_dod.add((cond, p))
                    
    #
    # calculation of slice in the CFG
    #
    def gen_slice(self):
        nodes = self.get_nodes()
        while True:
            break_condition = True
            for node in nodes:
                if node.is_in_slice():
                    # data dependencies, ntscd, and id
                    for dep in node.deps_dd | node.deps_ntscd | node.deps_id:
                        if isinstance(dep, Branch_element):
                            assert False and "Branch_element as dd, ntscd, or id"
                            dep = dep.parent
                        if not dep.in_slice:
                            dep.in_slice = True
                            break_condition = False
                    # dod
                    for d in node.deps_dod:
                        if d[1].is_in_slice(): #the other node in slice also?
                            if not d[0].in_slice:
                                d[0].in_slice = True #add the branch to the slice
                                break_condition = False
                    # dependencies from branch to signal senders
                    if isinstance(node, CFGBranch):
                        for b in node.branches:
                            for i in b.signal.incoming:
                                assert isinstance(i, CFGSend)
                                if not i.in_slice:
                                    i.in_slice = True
                                    i.target.in_slice = True # node evaluating the target has to be added to slice too
                                    break_condition = False
                # data dependencies from signal parameters
                if isinstance(node, CFGBranch):
                    for branch in node.branches:
                        for assign in branch.signal.assignments:
                            if assign.in_slice:
                                for dep in assign.deps:
                                    if not dep.in_slice:
                                        # signal parameters are handled in CFGNodes
                                        assert isinstance(dep, CFGNode)
                                        dep.in_slice = True
                                        break_condition = False
            if break_condition:
                break

    #
    # removal of parts not in slice from the UML model
    #
    def get_statemachine_slice(self):
        TM = coral.getTransactionManager()
        TM.beginModifyCmd()
        
        for state in self.mapping.values():
            sm_state = state.original
            # remove triggers and guards if they are not in slice
            if not state.cfg_node.is_in_slice():
                for sm_trans in sm_state.outgoing:
                    if sm_trans.trigger:
                        sm_trans.trigger = None
                    if sm_trans.guard:
                        sm_trans.guard = None
            # write new effect scripts with elements not in slice removed
            for sm_trans in sm_state.outgoing:
                trans = state.transitions[sm_trans]
                if sm_trans.effect:
                    effect_script = ""
                    for effect in trans.effects:
                        # handling of non-sending effects
                        if isinstance(effect.effect, CFGNode):
                            if effect.effect.in_slice: # included in slice, cannot remove
                                effect_script += effect.effect.script
                        # handling of signal send effects
                        elif isinstance(effect.effect, CFGSend):
                            # write effect if cannot be removed
                            if effect.effect.in_slice or effect.effect.is_param_or_target_in_slice():
                                param_list = []
                                param_datatypes = self.sig_param_datatypes[effect.effect.signal]
                                for i, p_node in enumerate(effect.effect.params):
                                    if p_node.in_slice:
                                        param_list.append(p_node.script)
                                    else:
                                        # replace unused parameter by dummy value specific to the datatype
                                        if param_datatypes[i] == 'int':
                                            param_list.append('0')
                                        elif param_datatypes[i] == 'boolean':
                                            param_list.append('true')
                                        elif param_datatypes[i] == 'object':
                                            param_list.append('null')
                                        else:
                                            assert False and "no correct parameter types"
                                s = ' send ' + effect.effect.signal + '('
                                s += ', '.join(param_list)
                                s += ') to ' + str(effect.effect.target.script) + ';'
                                effect_script += s
                        else:
                            assert False

                    # if new effect script is not empty, write it, otherwise remove whole effect
                    if effect_script:
                        sm_trans.effect.script.body = effect_script
                    else:
                        sm_trans.effect = None

        TM.endModifyCmd()


def get_refdef(s, mode):
    """Returns references and definitions in a statement of expression.

    Arguments:
        s     -- statement or expression as a string
        mode  -- string "statement" or "expression"
    """
    env = actlang.basetypes.getInitialEnvironment()
    if mode == "statement":
        tree = actlang.grammar.parse(s)
        stmts = tree.api_getTopStatements()
        try:
            if len(stmts) != 1:
                raise refdef.UnsupportedFeatureError("compound statements not "
                                                     "supported")

            return refdef.DefRef(stmts[0])
        except refdef.Error, e:
            print e.reason
    elif mode == "expression":
        tree = actlang.grammar.parse("x = (" + s + ");")
        stmts = tree.api_getTopStatements()
        try:
            if len(stmts) != 1:
                raise refdef.UnsupportedFeatureError("compound statements not "
                                                     "supported")

            return refdef.DefRef(stmts[0].api_getExpr().api_getRhsExpr())
        except refdef.Error, e:
            print e.reason
    else:
        assert False and "illegal option"

# Parameters:
#   rd: DefRef instance to be expanded
#   return: a tuple (refs, defs)
def expand_refdef_to_lists(rd):
    """Parameters:
      rd: DefRef instance to be expanded
      return: a tuple (refs, defs)
    """
    delim = '.'
    refs = []
    defs = []

    assert isinstance(rd, refdef.DefRef)
    
    # expand refs
    for variable in rd.refs:
        tokens = variable.split(delim)
        current_name = ''
        for token in tokens:
            if current_name:
                current_name += delim
            current_name += token
            refs.append(current_name)
        assert variable == current_name
            
    # expand defs
    for variable in rd.defs:
        tokens = variable.split(delim)
        current_name = ''
        for token in tokens:
            if current_name:
                current_name += delim
            current_name += token
            if variable == current_name:
                defs.append(current_name)
            else:
                refs.append(current_name)
        assert variable == current_name

    return (refs, defs)

def expand_effect_to_list(effect):
    """Makes a list of statements from a compound statement
    """
    #env = actlang.basetypes.getInitialEnvironment()
    tree = actlang.grammar.parse(effect)
    lst = []
    for stmt in tree.api_getTopStatements():
        lst.append(str(stmt))
    return lst;

def check_simple_name(string):
    """Checks whether a variable name is a simple name
    """
    tree = actlang.grammar.parse("x = " + string + ";")
    stmts = tree.api_getTopStatements()
    assert len(stmts) == 1
    stmt = stmts[0]
    assert isinstance(stmt, actlang.statement.ExprStatement)
    expr = stmt.api_getExpr()
    assert isinstance(expr, actlang.expr.SimpleAssignmentExpr)
    rhs = expr.api_getRhsExpr()
    if isinstance(rhs, actlang.names.SimpleName):
        return True
    else:
        return False

def collect_sig_param_types(model):
    """
    Finds parameter types for each signal in the model
    model: Coral model
    returns a dictonary where:
    key: signal name
    value: a list of signa parameter types
           parameter type is one of the following:
           'int', 'boolean', or 'object'
    """
    signals = []
    collect_elements(model, UML.Signal, signals)
    sig_param_datatypes = {}

    for s in signals:
        signame = str(s.name)
        assert not sig_param_datatypes.has_key(signame)

        lst = []
        for feat in s.feature:
            if feat.oclIsKindOf(UML.StructuralFeature):
                datatype = feat.type.name
                if datatype == 'int':
                    lst.append('int')
                elif datatype == 'boolean':
                    lst.append('boolean')
                elif datatype == 'object':
                    lst.append('object')
                else:
                    assert False
        sig_param_datatypes[signame] = lst

    return sig_param_datatypes
                
def check_assert(string):
    try:
        tree = actlang.grammar.parse(string)
        stmts = tree.api_getTopStatements()
        assert len(stmts) == 1
        if isinstance(stmts[0], actlang.statement.AssertStatement):
            return True
        else:
            return False
    except actlang.errors.Error:
        return False
