#! /usr/bin/python2.4
# coding: latin-1

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

""" Trace simulator for simulating SMUML generic traces

Author: Vesa Ojala <vojala@tcs.tkk.fi>
"""

import sys
import string
import coral
import actlang
import Sim
import Sim.Core
import Utils.String
import Utils.Models
from Utils.Models import get_qualified_name

UML14 = coral.meta.UML14
DIA20 = coral.meta.XMIDI20

TRACESIM_DEBUG = False

class Trace_Simulator:
    """Trace simulator for simulating SMUML generic traces.
    """
    def __init__(self, trace_filename, opt_evaluate_jumbala_assertions=True, project=None):
        """ Trace simulator class for simulating SMUML generic traces
        
        Arguments:
            trace_filename   -- filename of a generic trace to be simulated
            opt_evaluate_jumbala_assertions
                             -- boolean value specifying whether assertions
                                in effects are evaluated. Default is true.
            project          -- optional attribute that can be used to
                                override the Coral project specified in
                                the generic trace file. The attribute may
                                be a filenane represented as a string,
                                A Corel project, or None.
    """
        # check parameter types
        if not (type(trace_filename) is str):
            raise TraceSimulatorAttributeError("trace_filename", ["string"])
        self.trace_filename = trace_filename
        
        if type(project) is str or project==None:
            self.project_filename = project
            self.project = None
        elif isinstance(project, coral.kernel.ProjectPtr):
            self.project_filename = project.getFilename()
            self.project = project
        else:
            raise TraceSimulatorAttributeError("project", ["string", "None", "Coral project"])

        # other fields
        self.printing = False

        self.map_mc_object_to_uml_object = {}
        self.map_uml_object_to_mc_object = {}
        self.trace = []

        self.diagram = None
        self.model = None
        self.xmi_id_to_uml_transition = {}

        self.global_state = None
        self.sim = None
        self.mc_object_to_sim_oid_map = {}

        self.guard_choose_values = {}
        self.effect_choose_values = {}

        self.step_num = 0
        # trace_index contains the element in trace list which is to be executed next.
        # in the case of error the index of the action that caused the error
        # is in the trace_index.
        self.trace_index = 0

        self._load_trace()
        self._load_model()
        self._create_simulator()

    def set_printing(self, val):
        """Specifies whether trace simulator prints information about
        its actions to the standard output or not.

        Arguments:
            val -- Boolean value, True enables printing, False disables it
        """
        self.printing = val

    def get_printing(self):
        """Returns whether trace simulator is printing information about its
        actions to the standard output or not
        """
        return self.printing
        
    def _load_trace(self):
        """Loads generic trace file, must be done before simulation of a trace
        """
        #
        # Open the trace file
        #
        if self.printing:
            print "Processing the trace file " + self.trace_filename
        trace_file = open(self.trace_filename, 'r')

        #
        # Read the first line giving the model name
        #
        l = trace_file.readline()
        l_split = string.split(l)
        if len(l_split) != 2 or l_split[0] != "Model":
            print "Error at line: "+l
            sys.exit(1)
        # Use modelname only if it is not already set
        if not self.project_filename:
            self.project_filename = l_split[1]

        #
        # Read the initial configuration correspondence
        #
        self.map_mc_object_to_uml_object = {}
        self.map_uml_object_to_mc_object = {}
        l = trace_file.readline()
        while l:
            l_split = string.split(l)
            if l_split[0:2] == ['InitialConfiguration','object']:
                assert len(l_split) == 5
                assert l_split[3] == 'umlobject'
                self.map_mc_object_to_uml_object[l_split[2]] = l_split[4]
                self.map_uml_object_to_mc_object[l_split[4]] = l_split[2]
            elif l.strip()[0] == "#":
                pass
            else:
                break
            l = trace_file.readline()

        #
        # Read trace actions, strips chooses from the end of the trace so that
        # the last action(s) are not choose(s)
        #
        self.trace = []
        chooses = []
        while l:
            l_split = string.split(l)
            if l_split[0:2] == ['TransitionExecution','object']:
                assert len(l_split) == 5
                assert l_split[3] == 'transition'
                self.trace.extend(chooses)
                chooses = []
                self.trace.append(('EXEC', l_split[2], l_split[4]))
            elif l_split[0:2] == ['EventDefer','object']:
                assert len(l_split) == 3
                self.trace.extend(chooses)
                chooses = []
                self.trace.append(('DEFER', l_split[2]))
            elif l_split[0:2] == ['ImplicitConsumption','object']:
                assert len(l_split) == 3
                self.trace.extend(chooses)
                chooses = []
                self.trace.append(('IMPL', l_split[2]))
            elif l_split[0] == 'GuardChooseValue':
                assert len(l_split) == 9
                assert l_split[1] == 'object'
                mc_object = l_split[2]
                assert l_split[3] == 'transition'
                uml_trans_xmiid = l_split[4]
                assert l_split[5] == 'index'
                index = int(l_split[6])
                assert l_split[7] == 'value'
                value = int(l_split[8])
                chooses.append(('GuardChooseValue',mc_object,uml_trans_xmiid,index,value))
            elif l_split[0] == 'EffectChooseValue':
                assert len(l_split) == 9
                assert l_split[1] == 'object'
                mc_object = l_split[2]
                assert l_split[3] == 'transition'
                uml_trans_xmiid = l_split[4]
                assert l_split[5] == 'index'
                index = int(l_split[6])
                assert l_split[7] == 'value'
                value = int(l_split[8])
                chooses.append(('EffectChooseValue',mc_object,uml_trans_xmiid,index,value))
            elif l.strip()[0] == "#":
                pass
            else:
                assert False and "Internal error: something not yet implemented"
            l = trace_file.readline()

    def _load_model(self):
        """Loads Coral project where the model to be simulated lies.
        Must be done before simulation of the model.
        """
        #
        # Load project
        #
        if self.printing:
            print "Loading model " + self.project_filename

        if self.project:
            P = self.project
        else:
            try:
                P = coral.loadDocument(self.project_filename)
            except RuntimeError,e:
                print "Error: cannot load UML model "+self.project_filename+": "+str(e)
                sys.exit(1)

        #
        # Find the model
        #
        self.diagram = None
        self.model = None
        for m in P.contents():
            if m.oclIsKindOf(UML14.Model):
                if self.model != None:
                    raise "Two models found!"
                self.model = m
            
        #
        # Compute a map from XMI ids to transitions
        #
        self.xmi_id_to_uml_transition = {}
        transitions = Utils.Models.get_elements_of_type(self.model, UML14.Transition)
        for t in transitions:
            xmi_id = t.__XMIid__()
            assert not self.xmi_id_to_uml_transition.has_key(xmi_id)
            self.xmi_id_to_uml_transition[xmi_id] = t

    def _create_simulator(self):
        """Create Simulator for trace simulation, must be done before
        simulation of a trace
        """
        #
        # Create the simulator
        #
        if self.model==None:
            raise "No model found!"
        #if self.diagram==None:
        #    raise "No deployment diagram found!"

        self.sim =  Sim.Core.Simulator(self.model)

        self.sim.set_choose_boolean_hook(self._choose_boolean_hook, None)
        self.sim.set_choose_int_hook(self._choose_int_hook, None)
        
        #
        # Get correspondence between the simulator and mc initial global state
        #
        if self.printing:
            print "Getting initial configuration correspondence..."
        self.global_state = self.sim.trace.get_last_global_state()
        self.mc_object_to_sim_oid_map = {}
        for mc_object in self.map_mc_object_to_uml_object.keys():
            uml_object = self.map_mc_object_to_uml_object[mc_object]
            sim_oid = None
            for o in self.global_state.objects:
                if o.get_uml_object():
                    if o.get_uml_object().__XMIid__() == uml_object:
                        assert sim_oid == None
                        sim_oid = o.idnum
            assert sim_oid != None
            assert not self.mc_object_to_sim_oid_map.has_key(mc_object)
            self.mc_object_to_sim_oid_map[mc_object] = sim_oid

    #
    # Set up hooks that give values to the non-deterministic choose functions
    #
    def _choose_boolean_hook(self, dummy, sim, obj, uml_transition, where, index):
        """A callback function for choosing the same choose values as in
        the simulatable trace. This callback is for boolean type chooses.
        """
        dictionary = None
        if where == 'guard':
            dictionary = self.guard_choose_values
        elif where == 'effect':
            dictionary = self.effect_choose_values
        else:
            assert False and "Internal error"
        mc_object = None
        for (mc_obj,sim_oid) in self.mc_object_to_sim_oid_map.iteritems():
            if sim_oid == obj.idnum:
                mc_object = mc_obj
                break
        assert mc_object
        value = dictionary[(mc_object,uml_transition.__XMIid__(),index)]
        assert value == 0 or value == 1
        return value == 1
    def _choose_int_hook(self, dummy, sim, obj, uml_transition, where, index, int_list):
        """A callback function for choosing the same choose values as in
        the simulatable trace. This callback is for integer type chooses.
        """
        dictionary = None
        if where == 'guard':
            dictionary = self.guard_choose_values
        elif where == 'effect':
            dictionary = self.effect_choose_values
        else:
            assert False and "Internal error"
        mc_object = None
        for (mc_obj,sim_oid) in self.mc_object_to_sim_oid_map.iteritems():
            if sim_oid == obj.idnum:
                mc_object = mc_obj
                break
        assert mc_object
        value = dictionary[(mc_object,uml_transition.__XMIid__(),index)]
        assert value in int_list
        return value

    def get_choose_maps(self):
        """Gives a typle containing a copy of the current choose maps.
        """
        return (self.guard_choose_values.copy(), self.effect_choose_values.copy())

    def set_choose_maps(self, maptuple):
        """Sets choose maps as a current choose maps.

        Arguments:
            maptuple -- a tuple (guardmap, effectmap) containing choosemap for
                        guards and a choosemap for effects. The first element
                        must be a choosemap for guards.
        """
        self.guard_choose_values, self.effect_choose_values = maptuple

    def process_choose_list(self, lst):
        """Updates choosemaps to correspond chooses in the list given.

        Arguments:
           lst -- a list containing a set of chooses. Element in a list
                  has to be a tuple containing the following information:
                  (type,mc_object,uml_trans_xmiid,index,value)
                  type            = 'GuardChooseValue' or 'EffectChooseValue'
                  mc_object       = Object identifier in the model checker
                  uml_trans_xmiid = xmiid of the transition the choose is bound
                  index           = choose index
                  value           = choose value
        """
        if not lst:
            return
        for action in lst:
            if action[0] == 'GuardChooseValue':
                # Update guard choose function value map
                (dummy,mc_object,uml_trans_xmiid,index,value) = action
                self.guard_choose_values[(mc_object,uml_trans_xmiid,index)] = value
                if TRACESIM_DEBUG:
                    print "GuardChooseValue:" + str(action)

            elif action[0] == 'EffectChooseValue':
                # Update effect choose function value map
                (dummy,mc_object,uml_trans_xmiid,index,value) = action
                self.effect_choose_values[(mc_object,uml_trans_xmiid,index)] = value
                if TRACESIM_DEBUG:
                    print "EffectChooseValue:" + str(action)
            else:
                break

    def process_next_chooses(self):
        """Processes the next set of chooses in the trace.
        """
        while True:
            if self.trace_index >= len(self.trace):
                return
            action = self.trace[self.trace_index]
            if action[0] == 'GuardChooseValue':
                # Update guard choose function value map
                (dummy,mc_object,uml_trans_xmiid,index,value) = action
                self.guard_choose_values[(mc_object,uml_trans_xmiid,index)] = value
                if TRACESIM_DEBUG:
                    print "GuardChooseValue:" + str(action)

            elif action[0] == 'EffectChooseValue':
                # Update effect choose function value map
                (dummy,mc_object,uml_trans_xmiid,index,value) = action
                self.effect_choose_values[(mc_object,uml_trans_xmiid,index)] = value
                if TRACESIM_DEBUG:
                    print "EffectChooseValue:" + str(action)
            else:
                break
            self.trace_index += 1

    # Return values for succesful simulation
    # ('GuardChooseValue')
    # ('EffectChooseValue')
    # ('EXEC', sim_oid, uml_trans)
    # ('DEFER', sim_oid)
    # ('IMPL', sim_oid)
    # If there is no action left to be simulated None is returned
    def simulate_next_step(self):
        """Executes next action in the trace.
        If there are chooses before the next action, these are also processed.
        """
        #
        # Simulate actions one-by-one in the simulator
        #

        self.process_next_chooses()
        if self.trace_index >= len(self.trace):
            return None
        
        rvalue = None
        self.step_num += 1        
        action = self.trace[self.trace_index]
        
        if action[0] == 'EXEC':
            # Execution of a single transition
            # The structure should be of form ('EXEC',mc_object,uml_trans_xmi_id)
            (dummy,mc_object,uml_trans_xmiid) = action
            assert self.mc_object_to_sim_oid_map.has_key(mc_object)
            sim_oid = self.mc_object_to_sim_oid_map[mc_object]
            assert self.xmi_id_to_uml_transition.has_key(uml_trans_xmiid)
            uml_trans = self.xmi_id_to_uml_transition[uml_trans_xmiid]

            # Get the current global state
            self.global_state = self.sim.trace.get_last_global_state().api_copy()

            #
            # Handle ATOs by adding the ATO signal in front of the input queue
            #
            if uml_trans.trigger and uml_trans.trigger.signal.name[0] == '$':
                parameters = []
                if uml_trans.trigger.signal.feature:
                    assert False and "ATOs with parameters not yet handled"
                    # TODO: Decode possible ATO signal parameters
                # Insert ATO signal to the beginning of the queue of the object
                message = Sim.Message.Message(uml_trans.trigger.signal, parameters)
                tmp = [message]
                tmp.extend(self.global_state.objects[sim_oid].sm_input_queue)
                self.global_state.objects[sim_oid].sm_input_queue = tmp

            self.sim.set_working_global_state(self.global_state)
            enabled_sets,implicit_consumption_objects = self.sim.get_enabled_in_working_global_state()
            found_index = None
            for (index, (o,set)) in enumerate(enabled_sets):
                if o.idnum != sim_oid:
                    continue
                assert len(set) == 1
                (reason,sim_transition) = set[0]
                if (reason == "completion" or reason == "ATO" or
                    reason == "signal"):
                    pass
                elif reason == "defer":
                    continue
                else:
                    assert False and "Internal error: something not implemented"
                if sim_transition != uml_trans:
                    continue
                assert not found_index
                found_index = index
            if found_index == None:
                raise TransNotEnabledError(self.step_num, sim_oid, uml_trans)

            try:
                sim_action_list = self.sim.execute_transition_set_in_working_global_state(found_index)
            except Sim.Error.SimError, e:
                raise ExecutionError(self.step_num, sim_oid, uml_trans, e)

            for sim_action in sim_action_list:
                self.sim.trace.add_action(sim_action)
            self.sim.trace.add_global_state(self.global_state)

            rvalue = ('EXEC', sim_oid, uml_trans)

        elif action[0] == "DEFER":
            # Signal deferring
            # The structure should be of form ('DEFER',mc_object)
            (dummy,mc_object) = action
            assert self.mc_object_to_sim_oid_map.has_key(mc_object)
            sim_oid = self.mc_object_to_sim_oid_map[mc_object]

            # Get the current global state
            self.global_state = self.sim.trace.get_last_global_state().api_copy()

            self.sim.set_working_global_state(self.global_state)
            enabled_sets,implicit_consumption_objects = self.sim.get_enabled_in_working_global_state()
            found_index = None
            for (index, (o,set)) in enumerate(enabled_sets):
                if o.idnum != sim_oid:
                    continue
                assert len(set) == 1
                (reason,sim_transition) = set[0]
                if reason == "defer":
                    assert not found_index
                    found_index = index
                    break
            if found_index == None:
                raise DeferNotEnabledError(self.step_num, sim_oid)

            sim_action_list = self.sim.execute_transition_set_in_working_global_state(found_index)
            for sim_action in sim_action_list:
                self.sim.trace.add_action(sim_action)
            self.sim.trace.add_global_state(self.global_state)

            rvalue = ('DEFER', sim_oid)


        elif action[0] == "IMPL":
            # Implicit consumption
            # The structure should be of form ('IMPL',mc_object)
            (dummy,mc_object) = action
            assert self.mc_object_to_sim_oid_map.has_key(mc_object)
            sim_oid = self.mc_object_to_sim_oid_map[mc_object]

            # Get the current global state
            self.global_state = self.sim.trace.get_last_global_state().api_copy()
            self.sim.set_working_global_state(self.global_state)
            enabled_sets,implicit_consumption_objects = self.sim.get_enabled_in_working_global_state()
            found_index = None
            for (index,o) in enumerate(implicit_consumption_objects):
                if o.idnum != sim_oid:
                    continue
                assert not found_index
                found_index = index

            if found_index == None:
                raise ImplNotEnabledError(self.step_num, sim_oid)
                
            action_list = self.sim.execute_implicit_consumption_in_working_global_state(found_index)
            for action in action_list:
                self.sim.trace.add_action(action)
            self.sim.trace.add_global_state(self.global_state)

            rvalue = ('IMPL', sim_oid)

        else:
            raise UnknownActionError(self.step_num, action)

        self.trace_index += 1
        return rvalue

    def backtrack_one_step(self):
        """Backtracks simulation one action backwards.
        After execution, the trace_index is pointing to the first choose value update
        after last executed action when the backtracking operation is completed.
        """
        # check that operation can be done
        if self.sim.trace.length() < 1:
            return False
        if self.trace_index < 1:
            return False
        # update simulator trace (history)
        self.sim.trace.backtrack_one_step()
        # update current global state to correspond the trace
        self.global_state = self.sim.trace.get_last_global_state().api_copy()
        self.sim.set_working_global_state(self.global_state)
        # update the location in the generic trace (guide)
        flag = False
        while True:
            action = self.trace[self.trace_index-1]
            if not action[0] in ('GuardChooseValue', 'EffectChooseValue'):
                if not flag:
                    flag = True
                else:
                    return True
            assert self.trace_index > 0
            self.trace_index -= 1
        return False

    def get_next_action(self):
        """Returns the next action to be simulated.

        Return value is a tuple of a form:
          * ('EXEC', sim_oid, uml_trans) if the next action is a transition
            execution or,
          * ('DEFER', sim_oid) if the next action is a signal defer or,
          * ('IMPL', sim_oid) if the next action is a implicit consumption
        where sim_oid is a object id in the simulator and uml_trans a UML
        transition to be executed
        """
        action = self.trace[self.trace_index]
        if action[0] == 'EXEC':
            # Execution of a single transition
            # The structure should be of form ('EXEC',mc_object,uml_trans_xmi_id)
            (dummy,mc_object,uml_trans_xmiid) = action
            assert self.mc_object_to_sim_oid_map.has_key(mc_object)
            sim_oid = self.mc_object_to_sim_oid_map[mc_object]
            assert self.xmi_id_to_uml_transition.has_key(uml_trans_xmiid)
            uml_trans = self.xmi_id_to_uml_transition[uml_trans_xmiid]
            return (action[0], sim_oid, uml_trans)
        elif action[0] == "DEFER":
            # Signal deferring
            # The structure should be of form ('DEFER',mc_object)
            (dummy,mc_object) = action
            assert self.mc_object_to_sim_oid_map.has_key(mc_object)
            sim_oid = self.mc_object_to_sim_oid_map[mc_object]
            return (action[0], sim_oid)
        elif action[0] == "IMPL":
            # Implicit consumption
            # The structure should be of form ('IMPL',mc_object)
            (dummy,mc_object) = action
            assert self.mc_object_to_sim_oid_map.has_key(mc_object)
            sim_oid = self.mc_object_to_sim_oid_map[mc_object]
            return (action[0], sim_oid)
        else:
            raise UnknownActionError(self.step_num, action)

class TraceSimulatorAttributeError(Exception):
    """Exception for situations where the parameters passed to tracesim
    during initialization are wrong type

    Fields:
        parameter -- parameter exception concerns (as string)
        typelist -- list of types it should be (as list of strings)
    """
    def __init__(self, parameter, typelist):
        """Initializes TraceSimulatorAttributeError.
        
        Fields:
            parameter -- parameter exception concerns (as string)
            typelist -- list of types it should be (as list of strings)
        """
        self.parameter = parameter
        self.typelist = typelist
        
    def __str__(self):
        s = "" + self.__class__.__name__ + ": " + self.parameter + " should be "
        if len(typelist) >= 2:
            s += " one from the list: "
        s += ",".join(typelist)

class TraceSimulatorError(Exception):
    """Base class for TraceSimulator errors happening during trace processing

    Fields:
        step_num  -- Number of the step when the error happened
    """
    def __init__(self, step_num):
        """Initialize a.TraceSimulatorError

        Arguments:
            step_num  -- Number of the step when the error happened
        """
        self.step_num = step_num

    def __str__(self):
        return 'Error at step '+self.step_num

class UnknownActionError(TraceSimulatorError):
    """Action which was not recognized was tried to be executed.

    Fields:
        action  -- action tried to be executed.
    """
    def __init__(self, step_num, action):
        """Initialize a.TraceSimulatorError

        Arguments:
            step_num  -- Number of the step when the error happened
            action    -- action tried to be executed.
        """
        self.action = action
        TraceSimulatorError.__init__(self, step_num)
#        super(UnknownActionError, self).__init__(step_num)

    def __str__(self):
        return 'Internal error at step '+self.step_num+': unknown trace action '+str(self.action)

class ExecutionError(TraceSimulatorError):
    """Transition executed caused error in the simulator.
    
    Fields:
        sim_oid    -- simulator object id indicating the object in which
                      the transition was tried to be executed.
        uml_trans  -- UML transition tried to be executed
        sim_error  -- exception thrown by the simulator
    """
    def __init__(self, step_num, sim_oid, uml_trans, sim_error):
        """Initialize an ExecutionError.

        Arguments:
            step_num   -- Number of the step when the error happened
            sim_oid    -- simulator object id indicating the object in which
                          the transition was tried to be executed.
            uml_trans  -- UML transition tried to be executed
            sim_error  -- exception thrown by the simulator
        """
        self.sim_oid = sim_oid
        self.uml_trans = uml_trans
        self.sim_error = sim_error
        TraceSimulatorError.__init__(self, step_num)
#        super(ExecutionError, self).__init__(step_num)

    def __str__(self):
        return "Internal error at step "+str(self.step_num)+": object "+str(self.sim_oid) +\
               " during transition execution'"+get_qualified_name(self.uml_trans)+"' from '" +\
               get_qualified_name(self.uml_trans.source) + "' to '" + \
               get_qualified_name(self.uml_trans.target) + "': " +\
               str(self.sim_error)
        
class NotEnabledError(TraceSimulatorError):
    """Action to be simulated was not enabled.
    
    Fields:
        sim_oid    -- simulator object id indicating the object in which
                      the action tried to be executed was not enabled.
    """
    def __init__(self, step_num, sim_oid):
        """Initialize a NotEnabledError.

        Arguments:
            step_num  -- Number of the step when the error happened
            sim_oid   -- simulator object id indicating the object in which
                         the action tried to be executed was not enabled.
        """
        self.sim_oid = sim_oid
        TraceSimulatorError.__init__(self, step_num)
#        super(NotEnabledError, self).__init__(step_num)

    def __str__(self):
        return "Error at step "+str(self.step_num)+": object "+str(self.sim_oid)
        
class TransNotEnabledError(NotEnabledError):
    """Transition to be simulated was not enabled.

    Fields:
        uml_trans  -- UML transition tried to be executed
    """
    def __init__(self, step_num, sim_oid, uml_trans):
        """Initialize an TransNotEnabledError.

        Arguments:
            step_num   -- Number of the step when the error happened
            sim_oid    -- simulator object id indicating the object in which
                          the transition was tried to be executed.
            uml_trans  -- UML transition tried to be executed
        """
        self.uml_trans = uml_trans
        NotEnabledError.__init__(self, step_num, sim_oid)
#        super(TransNotEnabledError, self).__init__(step_num, sim_oid)
        
    def __str__(self):
        return "Error at step " + str(self.step_num) + ": object " + str(self.sim_oid) +\
               " cannot execute transition '"+get_qualified_name(self.uml_trans)+"' from '" +\
               get_qualified_name(self.uml_trans.source) + "' to '" + \
               get_qualified_name(self.uml_trans.target) + "'"
        
class ImplNotEnabledError(NotEnabledError):
    """Implicit consumption to be simulated was not enabled.
    """
    def __init__(self, step_num, sim_oid):
        """Initialize a ImplNotEnabledError.

        Arguments:
            step_num  -- Number of the step when the error happened
            sim_oid   -- simulator object id indicating the object in which
                         the implicit concumption tried to be executed was
                         not enabled.
        """
        NotEnabledError.__init__(self, step_num, sim_oid)
#        super(ImplNotEnabledError, self).__init__(step_num, sim_oid)
        
    def __str__(self):
        return "Error at step "+str(self.step_num)+": object "+str(self.sim_oid)+\
               " cannot do implicit consumption"
        
class DeferNotEnabledError(NotEnabledError):
    """Defer to be simulated was not enabled.
    """
    def __init__(self, step_num, sim_oid):
        """Initialize a DeferNotEnabledError.

        Arguments:
            step_num  -- Number of the step when the error happened
            sim_oid   -- simulator object id indicating the object in which
                         the defer tried to be executed was not enabled.
        """
        NotEnabledError.__init__(self, step_num, sim_oid)
#        super(DeferNotEnabledError, self).__init__(step_num, sim_oid)
        
    def __str__(self):
        return "Error at step "+str(self.step_num)+": object "+str(self.sim_oid)+\
               " cannot defer a signal"
        

if __name__ == '__main__':
    if not (len(sys.argv) in (2,3)):
        print "Usage: " + sys.argv[0] + " trace [model]"
        sys.exit(1)

    trace_filename = sys.argv[1]
    if len(sys.argv) == 3:
        project_filename = sys.argv[2]
        ts = Trace_Simulator(trace_filename, True, project_filename)
    else:
        project_filename = ''
        ts = Trace_Simulator(trace_filename)

    print "Simulating trace..."    
    while True:
        try:
            action = ts.simulate_next_step()
            if action == None:
                break
            elif action[0] == 'DEFER':
                (dummy, sim_oid) = action
                print " Object "+str(sim_oid)+" defers a signal"
            elif action[0] == 'IMPL':
                (dummy, sim_oid) = action
                print " Object "+str(sim_oid)+" performs implicit consumption"
            elif action[0] == 'EXEC':
                (dummy, sim_oid, uml_trans) = action
                print " Object "+str(sim_oid)+" moves from '"+\
                      uml_trans.source.name+"' to '"+uml_trans.target.name+"'"
            elif action[0] == 'EffectChooseValue':
                pass
            elif action[0] == 'GuardChooseValue':
                pass

        except TraceSimulatorError, e:
            if isinstance(e, NotEnabledError):
                print "The trace up to and including the current global state is:"
                ts.sim.api_print_trace()
                print e
                sys.exit(1)
            elif isinstance(e, ExecutionError):
                print "The trace up to and including the current global state is:"
                ts.sim.api_print_trace()
                print e
                print "Aborting!"
                sys.exit(1)
                
            elif isinstance(e, UnknownActionError):
                print e
                sys.exit(1)


    #
    # Succesfull, print the trace
    #
    print "Printing trace..."
    ts.sim.api_print_trace()

