#!/usr/bin/python

# This file is part of the SMUML toolset.
# Copyright (C) 2007 Helsinki University of Technology.
#
# The SMUML toolset is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2,
# as published by the Free Software Foundation.
#
# The SMUML toolset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the SMUML toolset.  If not, see <http://www.gnu.org/licenses/>.

"""
Given a NuSMV model file and an output file containing a counterexample,
translates the counterexample into the 'SMUML generic trace format'.

Author: Tommi Junttila
"""

import sys
import string
import types

def translate(smv_filename, trace_filename, stdout=sys.stdout, stderr=sys.stderr):
    """
    Translate a NuSMV counterexample trace (in 'trace_filename') for a
    NuSMV model in 'smv_filename' into generic format. The result is
    written to 'stdout'.
    """
    #
    #
    # Get model filename and NuSMV/UML class/object correspondences from log
    #
    #
    #print >> stdout, "Processing the log info in NuSMV file " + smv_filename
    smv_file = open(smv_filename, 'r')
    l = smv_file.readline()
    if not l.startswith("-- NuSMV input file generated by Uboco"):
        raise "Error in NuSMV file  "+smv_filename

    # Read the model file name in log line of form "-- Input file: model.xmi"
    l = smv_file.readline()
    l_split = string.split(l)
    if len(l_split) != 4 or l_split[0:3] != ['--','Input','file:']:
        raise "Error in NuSMV file "+smv_filename+" on line: "+l
    model_filename = l_split[3]

    # Read'n'skip the "-- Created on ..." line
    l = smv_file.readline()
    l_split = string.split(l)
    if l_split[0:3] != ['--','Created','on']:
        raise "Error in NuSMV file "+smv_filename+"on line: "+l

    # Find the beginning of the "-- $Map" section in the NuSMV file
    found_map = False
    l = smv_file.readline()
    while l:
        l_split = string.split(l)
        if l_split[0:2] == ['--$Map','Model']:
            found_map = True
            assert l_split[2] == model_filename
            break
        l = smv_file.readline()
    if not found_map:
        raise "No $Map info in NuSMV file  "+smv_filename

    map_nusmv_object_to_uml_object = {}
    map_uml_object_to_nusmv_object = {}
    uml_class_to_uml_trans_to_nusmv_trans = {}
    uml_trans_to_nusmv_trans = {}
    uml_class_to_uml_trans_to_guard_expr_index = {}
    uml_class_to_uml_trans_to_effect_expr_index = {}
    l = smv_file.readline()
    while l:
        l_split = string.split(l)
        if l_split[0:2] == ['--$Map','Object']:
            if len(l_split) == 4:
                map_nusmv_object_to_uml_object[l_split[2]] = (l_split[3],None)
            elif len(l_split) == 5:
                map_nusmv_object_to_uml_object[l_split[2]] = (l_split[3],l_split[4])
                map_uml_object_to_nusmv_object[l_split[4]] = l_split[2]
            else:
                raise "Error in NuSMV file "+smv_filename+" on line '"+l+"'"
        elif l_split[0:2] == ['--$Map','StateVertex']:
            assert len(l_split) == 5
            # TODO: handle state vertices later when necessary
        elif l_split[0:2] == ['--$Map','Trans']:
            assert len(l_split) == 5
            if not uml_class_to_uml_trans_to_nusmv_trans.has_key(l_split[2]):
                uml_class_to_uml_trans_to_nusmv_trans[l_split[2]] = {}
            uml_class_to_uml_trans_to_nusmv_trans[l_split[2]][l_split[3]] = l_split[4]
            assert not uml_trans_to_nusmv_trans.has_key(l_split[3])
            uml_trans_to_nusmv_trans[l_split[3]] = l_split[4]

        elif l_split[0:2] == ['--$Map','TransGuardChoose']:
            # The line should be of form
            # --$Map TransGuardChoose uml_class_xmi uml_trans_xmi nusmv_expr index
            assert len(l_split) == 6
            if not uml_class_to_uml_trans_to_guard_expr_index.has_key(l_split[2]):
                uml_class_to_uml_trans_to_guard_expr_index[l_split[2]] = {}
            uml_trans_to_guard_expr_index = uml_class_to_uml_trans_to_guard_expr_index[l_split[2]]
            if not uml_trans_to_guard_expr_index.has_key(l_split[3]):
                uml_trans_to_guard_expr_index[l_split[3]] = {}
            assert not uml_trans_to_guard_expr_index[l_split[3]].has_key(l_split[5])
            uml_trans_to_guard_expr_index[l_split[3]][l_split[5]] = l_split[4]

        elif l_split[0:2] == ['--$Map','TransEffectChoose']:
            # The line should be of form
            # --$Map TransEffectChoose uml_class_xmi uml_trans_xmi nusmv_expr index
            assert len(l_split) == 6
            if not uml_class_to_uml_trans_to_effect_expr_index.has_key(l_split[2]):
                uml_class_to_uml_trans_to_effect_expr_index[l_split[2]] = {}
            uml_trans_to_effect_expr_index = uml_class_to_uml_trans_to_effect_expr_index[l_split[2]]
            if not uml_trans_to_effect_expr_index.has_key(l_split[3]):
                uml_trans_to_effect_expr_index[l_split[3]] = {}
            assert not uml_trans_to_effect_expr_index[l_split[3]].has_key(l_split[5])
            uml_trans_to_effect_expr_index[l_split[3]][l_split[5]] = l_split[4]

        else:
            break
        l = smv_file.readline()

    #
    # Read the static ordering between transition instances
    #
    nusmv_transinst_list = []
    nusmv_transinst_to_priority = {}
    priority = 0
    while l:
        l_split = string.split(l)
        if l_split[0:2] == ['--$Map','TransInstance']:
            # Append a pair (NuSMV object,NuSMV trans)
            nusmv_transinst_list.append((l_split[2],l_split[3]))
            nusmv_transinst_to_priority[(l_split[2],l_split[3])] = priority
            priority += 1
        else:
            assert l_split[0:2] == ['--$Map','End']
            break
        l = smv_file.readline()


    #
    # Read the first trace to memory
    #
    #print >> stdout, "Loading NuSMV trace " + trace_filename
    NuSMV_trace = open(trace_filename, 'r')


    #
    # Find trace beginning by looking for line "Trace Type: Counterexample"
    #
    trace_found = False
    line = NuSMV_trace.readline()
    while line != "":
        split_line = string.split(line)
        if split_line == ['Trace', 'Type:', 'Counterexample']:
            trace_found = True
            break
        line = NuSMV_trace.readline()

    if not trace_found:
        raise "No trace found!"


    trace = []
    trace_num = None
    state_num = 1
    line = NuSMV_trace.readline()
    state_input_pair = [{},{}]
    while line:
        split_line = string.split(line)
        if split_line == ['NuSMV','>']:
            # NuSMV prompt -> end of trace
            break
        if split_line == ['--','Loop','starts','here']:
            #print >> stdout, "Warning: loopback points in NuSMV trace are ignored"
            line = NuSMV_trace.readline()
            continue

        # We should see a line of form "-> State: x.y <-"
        if not(split_line[0] == '->' and split_line[1] == 'State:' and
               split_line[3] == '<-'):
            raise "Trace syntax error on line '"+line+"'"
        # Check that x.y is consistent with previous information about the trace
        state_id = string.split(split_line[2], ".")
        if trace_num == None:
            trace_num = state_id[0]
        else:
            assert state_id[0] == trace_num
        assert int(state_id[1]) == state_num

        #print >> stdout, "Reading NuSMV state " + str(state_num)
        # Copy the state variable values from the previous time instance as
        # NuSMV trace may only include the changed state variable values
        state_input_pair = [state_input_pair[0].copy(),{}]
        trace.append(state_input_pair)
        # Read state variable value changes
        while True:
            line = NuSMV_trace.readline()
            if not line:
                break
            split_line = string.split(line)
            if split_line[1] != '=':
                break
            if state_input_pair[0].has_key(split_line[0]):
                assert state_num > 1
            state_input_pair[0][split_line[0]] = split_line[2]

        state_num += 1

        # If there are input variables,
        # we should see a line of form "-> Input: x.y <-"
        if not(split_line[0] == '->' and split_line[1] == 'Input:' and
               split_line[3] == '<-'):
            continue

        # Check that x.y is consistent with previous information about the trace
        state_id = string.split(split_line[2], ".")
        if trace_num == None:
            trace_num = state_id[0]
        else:
            assert state_id[0] == trace_num
        assert int(state_id[1]) == state_num

        # Read input variable values
        #print >> stdout, "Reading NuSMV input " + str(state_num)
        while True:
            line = NuSMV_trace.readline()
            assert line
            split_line = string.split(line)
            if split_line[1] != '=':
                break
            if state_input_pair[1].has_key(split_line[0]):
                assert state_num > 1
            state_input_pair[1][split_line[0]] = split_line[2]

    #print >> stdout, trace

    def __get_value(state, inputs, name):
        """Given two dictionaries containing the (name,value) pairs of state and
        input variables, fetch the value of variable 'name'"""
        assert name
        assert isinstance(name, types.StringType)
        if state.has_key(name):
            assert not inputs.has_key(name)
            return state[name]
        assert inputs.has_key(name)
        return inputs[name]


    #print >> stdout, "Interpreting NuSMV trace"
    interpreted_trace = []
    for state_input_pair in trace:
        state = state_input_pair[0]
        inputs = state_input_pair[1]
        if inputs == {}:
            continue
        actions = []
        #
        # Iterate over all objects in the NuSMV model
        #
        for nusmv_object in map_nusmv_object_to_uml_object.keys():
            (uml_class,uml_object) = map_nusmv_object_to_uml_object[nusmv_object]
            #print >> stdout, nusmv_object, uml_class
            class_uml_trans_to_nusmv_trans = uml_class_to_uml_trans_to_nusmv_trans[uml_class]

            # Is the object scheduled?
            if inputs[nusmv_object+'.scheduled'] == '0':
                # No it was not, do nothing but check some assertions
                assert __get_value(state,inputs,nusmv_object+'.dispatchSys') == '0'
                assert __get_value(state,inputs,nusmv_object+'.dispatchExt') == '0'
                assert __get_value(state,inputs,nusmv_object+'.dequeue') == '0'
                assert __get_value(state,inputs,nusmv_object+'.implicitConsumeSys') == '0'
                assert __get_value(state,inputs,nusmv_object+'.deferSys') == '0'
                for (uml_trans_xmi,nusmv_trans) in class_uml_trans_to_nusmv_trans.iteritems():
                    assert inputs[nusmv_object+'.fire#'+nusmv_trans] == '0'
                continue

            action = None

            # Check if object performs implicit consumption
            if __get_value(state,inputs,nusmv_object+'.implicitConsumeSys') == '1':
                #print >> stdout, "  Object "+nusmv_object+" performs implicit consumption"
                action = ('IMPL',nusmv_object)

            # Check if object fires a transition
            for (uml_trans_xmi,nusmv_trans) in class_uml_trans_to_nusmv_trans.iteritems():
                if inputs[nusmv_object+'.fire#'+nusmv_trans] == '1':
                    #print >> stdout, "  Object "+nusmv_object+" executes transition "+uml_trans_xmi
                    assert action == None
                    action = ('EXEC',nusmv_object,uml_trans_xmi)

            # Check if object defers a signal
            if __get_value(state,inputs,nusmv_object+'.deferSys') == '1':
                #print >> stdout, "  Object "+nusmv_object+" defers a signal"
                # TO DO: find the deferred signal name
                action = ('DEFER',nusmv_object)

            if not action:
                raise "Internal error: object "+nusmv_object+\
                      " is scheduled but cannot interprete what it does"
            if action:
                actions.append(action)
        interpreted_trace.append(actions)


    #
    # Find values of non-deterministic choose-functions in the trace
    #
    guard_choose_values_in_trace = []
    effect_choose_values_in_trace = []
    for state_input_pair in trace:
        (state, inputs) = state_input_pair
        if inputs == {}:
            continue
        guard_choose_values = []
        effect_choose_values = []
        #
        # Iterate over all objects in the NuSMV model
        #
        for nusmv_object in map_nusmv_object_to_uml_object.keys():
            (uml_class,uml_object) = map_nusmv_object_to_uml_object[nusmv_object]

            if uml_class_to_uml_trans_to_guard_expr_index.has_key(uml_class):
                # uml_class has transitions with guard involving choose exprs
                for uml_trans in uml_class_to_uml_trans_to_guard_expr_index[uml_class]:
                    for (index,expr) in uml_class_to_uml_trans_to_guard_expr_index[uml_class][uml_trans].iteritems():
                        guard_choose_values.append((nusmv_object,
                                                    uml_trans,
                                                    index,
                                                    inputs[nusmv_object+"."+expr]))


            if uml_class_to_uml_trans_to_effect_expr_index.has_key(uml_class):
                # uml_class has transitions with effect involving choose exprs
                for uml_trans in uml_class_to_uml_trans_to_effect_expr_index[uml_class]:
                    for (index,expr) in uml_class_to_uml_trans_to_effect_expr_index[uml_class][uml_trans].iteritems():
                        effect_choose_values.append((nusmv_object,
                                                     uml_trans,
                                                     index,
                                                     inputs[nusmv_object+"."+expr]))
        guard_choose_values_in_trace.append(guard_choose_values)
        effect_choose_values_in_trace.append(effect_choose_values)




    #
    # Linearize step semantics
    # Sort trace actions according to the transition instance map
    # (the static ordering between transition instances given in the NuSMV file)
    #
    new_interpreted_trace = []
    for actions in interpreted_trace:
        new_actions = []
        processed_actions = actions
        remaining_actions = []
        # Do implicit consumption first
        for act in processed_actions:
            if act[0] == 'IMPL':
                new_actions.append(act)
            else:
                remaining_actions.append(act)
        # Then event defers
        processed_actions = remaining_actions
        remaining_actions = []
        for act in processed_actions:
            if act[0] == 'DEFER':
                new_actions.append(act)
            else:
                remaining_actions.append(act)
        # Finally, sort the real transition firings according to the static order
        processed_actions = remaining_actions
        remaining_actions = []
        for act in processed_actions:
            assert act[0] == 'EXEC'

        def firing_compare(f1, f2):
            """Comparison operator needed to sort transition instances"""
            assert f1[0] == 'EXEC'
            assert f2[0] == 'EXEC'
            (c1,nusmv_object1,uml_trans_xmi1) = f1
            (c2,nusmv_object2,uml_trans_xmi2) = f2
            p1 = nusmv_transinst_to_priority[(nusmv_object1,uml_trans_to_nusmv_trans[uml_trans_xmi1])]
            p2 = nusmv_transinst_to_priority[(nusmv_object2,uml_trans_to_nusmv_trans[uml_trans_xmi2])]
            return cmp(p1,p2)

        processed_actions.sort(firing_compare)
        new_actions.extend(processed_actions)

        new_interpreted_trace.append(new_actions)

    #print >> stdout, interpreted_trace
    #print >> stdout, new_interpreted_trace
    interpreted_trace = new_interpreted_trace

    #
    #
    # Print the generic trace
    #
    #

    #
    # Print model name
    #
    print >> stdout, "Model "+model_filename

    #
    # Print initial configuration information
    #
    for nusmv_object in map_nusmv_object_to_uml_object.keys():
        (uml_class,uml_object) = map_nusmv_object_to_uml_object[nusmv_object]
        if uml_object:
            print >> stdout, "InitialConfiguration object "+nusmv_object+\
                  " umlobject "+uml_object

    import re
    def _nusmv_number_to_int(n):
        if re.compile("[0-9]+").match(n).end() == len(n):
            return int(n)
        if re.compile("0d[0-9]+_[0-9]+").match(n).end() == len(n):
            return int(n.split("_")[1])
        raise "Error: cannot parse nusmv number "+n


    #
    # Print trace actions
    #
    for (actions,guard_chooses,effect_chooses) in zip(interpreted_trace,guard_choose_values_in_trace,effect_choose_values_in_trace):
        for (nusmv_object,uml_trans,expr_index,expr_value) in guard_chooses:
            print >> stdout,\
                  "GuardChooseValue"+\
                  " object "+nusmv_object+\
                  " transition "+uml_trans+\
                  " index "+expr_index+\
                  " value "+str(_nusmv_number_to_int(expr_value))
        for (nusmv_object,uml_trans,expr_index,expr_value) in effect_chooses:
            print >> stdout,\
                  "EffectChooseValue"+\
                  " object "+nusmv_object+\
                  " transition "+uml_trans+\
                  " index "+expr_index+\
                  " value "+str(_nusmv_number_to_int(expr_value))
        for action in actions:
            if action[0] == "EXEC":
                (dummy,nusmv_object,uml_trans_xmiid) = action
                print >> stdout,\
                      "TransitionExecution object "+nusmv_object+\
                      " transition "+uml_trans_xmiid

            elif action[0] == "DEFER":
                (dummy, nusmv_object) = action
                print >> stdout, "EventDefer object "+nusmv_object

            elif action[0] == "IMPL":
                (dummy, nusmv_object) = action
                print >> stdout, "ImplicitConsumption object "+nusmv_object

            else:
                print >> stdout, "Unknown interpreted trace action : " + str(action)
                assert False

if __name__ == "__main__":
    if not (len(sys.argv) == 3):
        print "Usage: " + sys.argv[0] + " NuSMVfile NuSMVoutputfile"
        sys.exit(-1)
    translate(sys.argv[1], sys.argv[2])
