"""
A circuit based Python front-end for some SMT solvers and the SMT-LIB format.

Author: Tommi Junttila

Copyright (c) 2007 Helsinki University of Technology
Copyright (c) 2008 Tommi Junttila
Released under the GNU General Public License version 2
"""

import os
import sets
import string
import sys
import tempfile
import weakref
import re

def _type_error(s):
    raise RuntimeError(s)

def _internalerror_snyi():
    """Just an abbreviation."""
    raise RuntimeError("Internal error: something not yet implemented")

def _is_str_int(s):
    """
    Does the given string 's' represent an integer?
    """
    try:
        num = int(s)
    except ValueError:
        return False
    return True


#
#
# The type library
#
#
class _Type:
    """
    An abstract base class for PySMT types.
    """
    def __init__(self):
        raise "Abstract base class error"
    def __str__(self):
        raise "Abstract base class error"
    def get_a_value(self):
        """Return a fixed element in the domain of the type."""
        raise "Abstract base class error"
    def is_bool(self):
        return False
    def is_int(self):
        return False
    def is_rat(self):
        return False
    def is_sint(self):
        return False
    def is_uint(self):
        return False
    def is_tuple(self):
        return False
    def is_enum(self):
        return False
    def is_array(self):
        return False

class _Type_bool(_Type):
    """
    The Boolean type.
    """
    def __init__(self):
        pass
    def __str__(self):
        return "bool"
    def get_a_value(self):
        """Return a fixed element in the domain of the type."""
        return False
    def is_bool(self):
        return True

class _Type_int(_Type):
    """
    The unbounded integer number type.
    """
    def __init__(self):
        pass
    def __str__(self):
        return "int"
    def get_a_value(self):
        """Return a fixed element in the domain of the type."""
        return long(0)
    def is_int(self):
        return True

class _Type_rat(_Type):
    """
    The rational number type.
    """
    def __init__(self):
        pass
    def __str__(self):
        return "rat"
    def get_a_value(self):
        """Return a fixed element in the domain of the type."""
        raise RuntimeError("Not yet implemented")
    def is_rat(self):
        return True

class _Type_uint(_Type):
    """
    The bounded unsigned integer type.
    """
    def __init__(self, width):
        assert type(width) is int
        assert width >= 0
        self.width = width
    def __str__(self):
        return "uint"+str(self.width)
    def get_a_value(self):
        """Return a fixed element in the domain of the type."""
        return long(0)
    def is_uint(self):
        return True
    def get_width(self):
        return self.width

class _Type_sint(_Type):
    """
    The bounded signed integer type.
    """
    def __init__(self, width):
        assert type(width) is int
        assert width >= 1
        self.width = width
    def __str__(self):
        return "sint"+str(self.width)
    def get_a_value(self):
        """Return a fixed element in the domain of the type."""
        return long(0)
    def is_sint(self):
        return True
    def get_width(self):
        return self.width

class _Type_enum(_Type):
    """
    Enumeration primitive type.
    """
    def __init__(self, literals, name = None):
        """
        Create a new enumeration primitive type.
        'literals' should be a non-empty list or set of disjoint literals,
        a literal being a string that matches the regular expression
        [a-zA-Z][a-zA-Z0-9]*.
        Enumeration types act as namespaces in the sense that
        two literals with the same name but in different enumeration types
        are different.

        The name of the type, 'name', has no semantic meaning but
        is only used for verbose output and debugging purposes.
        """
        assert isinstance(literals, list) or isinstance(literals, sets.Set)
        assert len(literals) > 0
        self.literals = []
        self.literal_to_index = {}
        for lit in literals:
            assert type(lit) is str
            assert lit
            if not re.match("^[a-zA-Z][a-zA-Z0-9]*$", lit):
                raise RuntimeError("Illegal enumeration literal name "
                                   "'"+lit+"'")
            if lit in self.literal_to_index:
                raise RuntimeError("Literal '"+lit+"' occurs twice")
            self.literal_to_index[lit] = 0
            self.literals.append(lit)
        self.literals.sort()
        for (index,lit) in enumerate(self.literals):
            self.literal_to_index[lit] = index
        self.name = name
    def __str__(self):
        if self.name is None:
            return "enum"+str(id(self))+"{"+(",".join(self.literals))+"}"
        return str(self.name)
    def get_a_value(self):
        """Return a fixed element in the domain of the type."""
        return self.literals[0]
    def is_enum(self):
        return True
    def get_domain(self):
        """Return the list of literals in the type."""
        return self.literals[:]
    def has_literal(self, literal):
        """Check whether 'literal' belongs to the type."""
        assert type(literal) is str
        return literal in self.literal_to_index
    def get_literal_index(self, literal):
        """Map the literal 'literal' in the type into its index."""
        assert type(literal) is str
        assert literal in self.literal_to_index
        return self.literal_to_index[literal]
    def get_literal(self, index):
        """
        Get the literal at 'index'.
        Required: 0 <= 'index' < N, where N is the number of literals
                                    in the type.
        """
        assert (type(index) is int) or (type(index) is long)
        assert 0 <= index
        assert index < len(self.literals)
        return self.literals[index]

class _Type_tuple(_Type):
    """
    The tuple compound type.
    """
    def __init__(self, fields):
        """
        'fields' is a non-empty list of pairs of form (n,t),
        where n is the name of the field and 't' is its type.
        """
        assert type(fields) is list
        assert len(fields) >= 1
        self.fields = []
        self.name_to_field_index = {}
        for (n,t) in fields:
            assert type(n) is str
            assert n != ""
            assert isinstance(t, _Type)
            if self.name_to_field_index.has_key(n):
                raise RuntimeError("Field with name '"+n+"' defined twice")
            self.name_to_field_index[n] = len(self.fields)
            self.fields.append((n,t))
    def __str__(self):
        return "tuple("+(",".join([n+":"+str(t) for (n,t) in self.fields]))+")"
    def get_a_value(self):
        """Return a fixed element in the domain of the type."""
        l = [t.get_a_value() for (n,t) in self.fields]
        return tuple(l)
    def is_tuple(self):
        return True
    def get_field_name(self, i):
        """
        Get the name of the 'i':th field.
        """
        assert 0 <= i and i <  len(self.fields)
        (n,t) = self.fields[i]
        return n
    def get_field_type(self, x):
        """
        Get the type of the field 'x',
        where 'x' is either the index or the name of the field.
        """
        if type(x) is int:
            assert 0 <= x and x < len(self.fields)
            (n,t) = self.fields[x]
            return t
        assert type(x) is str
        assert self.name_to_field_index.has_key(x)
        i = self.name_to_field_index[x]

class _Type_array(_Type):
    """
    The array compound type.
    """
    def __init__(self, index_type, range_type):
        """
        Create a new array compound type.
        """
        assert isinstance(index_type, _Type)
        assert isinstance(range_type, _Type)
        self.index_type = index_type
        self.range_type = range_type
    def __str__(self):
        return "array "+str(self.index_type)+" of "+str(self.range_type)
    def get_a_value(self):
        """Return a fixed element in the domain of the type."""
        return {}
    def is_array(self):
        return True
    def get_index_type(self):
        return self.index_type
    def get_range_type(self):
        return self.range_type

class _Operand:
    """
    An abstract base class for PySMT operands.
    """
    def __init__(self):
        raise "Abstract base class error"
    def __str__(self):
        raise "Abstract base class error"
    def get_range_type(self):
        raise "Abstract base class error"
    def is_uif(self):
        return False

class _Operand_uif(_Operand):
    """
    The uninterpreted function operand.
    """
    def __init__(self, domain, range, name):
        """
        Create a new uninterpreted function (UIF) with signature
        D1 * ... * Dn -> 'range',
        where
        'domain' is a non-empty list [D1,...,Dn] of domain types, and
        'range' is the range type.

        The name of the UIF, 'name', has no semantic meaning but
        is only used for verbose output and debugging purposes.
        """
        assert type(domain) is list
        assert len(domain) >= 1
        for dt in domain:
            assert isinstance(dt, _Type)
        assert isinstance(range, _Type)
        self.domain = domain[:]
        self.range = range
        self.name = name
    def __str__(self):
        if self.name is None:
            n = "uif"+str(id(self))
        else:
            n = str(self.name)
        return n+" : "+\
               (" * ".join([str(dt) for dt in self.domain]))+\
               " -> "+str(self.range)
    def is_uif(self):
        return True
    def get_domain(self):
        """
        Get the domain.
        """
        d = self.domain[:]
        return d
    def get_domain_size(self):
        """
        Get number of types in the domain.
        """
        return len(self.domain)
    def get_domain_type(self, i):
        """
        Get the type of the 'i':th domain type.
        """
        assert type(i) is int
        assert 0 <= i and i < len(self.domain)
        return self.domain[i]
    def get_range_type(self):
        return self.range


# Predefined types
__pre_bool = _Type_bool()
__pre_int = _Type_int()
__pre_rat = _Type_rat()
__typemap = weakref.WeakValueDictionary()
__typemap["bool"] = __pre_bool
__typemap["int"] = __pre_int
__typemap["rat"] = __pre_rat

#
# Public API for PySMT types
#
def get_bool():
    """
    Return the boolean type.
    """
    return __pre_bool
def get_int():
    """
    Return the integer type.
    """
    return __pre_int
def get_rat():
    """
    Return the rational type.
    """
    return __pre_rat
def get_uint(W):
    """
    Return the W-bits bounded unsigned integer type.
    W must be a positive integer.
    """
    assert type(W) is int
    assert W >= 1
    name = "uint"+str(W)
    if name in __typemap:
        return __typemap[name]
    t = _Type_uint(W)
    __typemap[name] = t
    return t
def get_sint(W):
    """
    Return the W-bits bounded signed integer type.
    W must be a positive integer.
    """
    assert type(W) is int
    assert W >= 1
    name = "sint"+str(W)
    if name in __typemap:
        return __typemap[name]
    t = _Type_sint(W)
    __typemap[name] = t
    return t
def new_enum(literals, name = None):
    """
    Introduce a new enumeration primitive type.
    'literals' should be a non-empty list or set of disjoint names,
    name being a string that matches the regular expression
    [a-zA-Z][a-zA-Z0-9]*.
    Note that if you call this function two times with the same list
    of literals, you will get two distinct types, not the same one.

    The name of the type, 'name', has no semantic meaning but
    is only used for verbose output and debugging purposes.
    """
    return _Type_enum(literals, name)

def new_array(index, range):
    """
    Introduce a new array type from 'index' type to the 'range' type.
    Note that if you call this function two times with the same index and
    range types, you will get the same array type.
    """
    assert isinstance(index, _Type)
    assert isinstance(range, _Type)
    name = "array "+str(index)+" "+str(range)
    if name in __typemap:
        return __typemap[name]
    t = _Type_array(index, range)
    __typemap[name] = t
    return t

def new_uif(domain, range, name = None):
    """
    Introduce a new uninterpreted function (UIF) with signature
    D1 * ... * Dn -> 'range',
    where
    'domain' is a non-empty list [D1,...,Dn] of domain types, and
    'range'  is the range type.
    Note that if you call this function two times with the same domain and
    range types, you will get two distinct UIFs, not the same one.

    The name of the UIF, 'name', has no semantic meaning but
    is only used for verbose output and debugging purposes.
    """
    return _Operand_uif(domain, range, name)



#
#
# SMT solver wrappers
#
#
class Solver:
    """
    An abstract base class for SMT solver front-ends.
    """
    def _solve_smtlib(self, filename, return_solution = False):
        raise RuntimeError("Abstract base class method called")

    def _interpret_solution(self, partial_solution,
                            map_var_to_smt_name, map_smt_name_to_var):
        raise RuntimeError("Abstract base class method called")



class Z3(Solver):
    """
    A front-end for Z3.
    """
    def __init__(self, solver_binary = "./z3"):
        """
        Create a front-end for Z3.
        'solver_binary' is the file name for a Z3 executable binary.
        """
        assert type(solver_binary) is str
        self.binary = solver_binary


    def _solve_smtlib(self, filename, return_solution = False):
        """
        Use Z3 to solve the SMT-LIB problem in the file 'filename'.
        """
        tempfilename = "tmp_z3.out"
        #tempfilename = tempfile.mktemp()
        result = "unknown"
        solution = {}

        if os.system(self.binary+" < "+filename+
                     " > "+tempfilename+" 2>&1") != 0:
            raise RuntimeError("An error occurred when calling Z3, see "+
                               tempfilename+" for Z3 output")

        # Parse the result file produced by Z3
        f = open(tempfilename, "r")
        l = _chop(f.readline())
        if l == "sat":
            result = "sat"
            if return_solution:
                # The following lines contain the solution, interpret it
                for l in f:
                    l = string.split(l)
                    if l == []:
                        continue
                    if l[0] == '(=':
                        assert len(l) == 3
                        var = l[1]
                        value = l[2]
                        assert value[-1] == ')'
                        value = value[0:-1]
                        assert var not in solution
                        solution[var] = value
        elif l == "unsat":
            result = "unsat"
        else:
            #print "'"+l+"'"
            raise RuntimeError("Internal error: something not implemented")
        f.close()
        #os.remove(tempfilename)
        return (result, solution)


    def _solve_native(self, filename, return_solution = False):
        """
        Use Z3 to solve the problem in the file 'filename'.
        """

        tempfilename = "tmp_z3.out"
        #tempfilename = tempfile.mktemp()
        result = "unknown"
        solution = {}

        if os.system(self.binary+" -z3 -v:100 -st "+filename+
                     " > "+tempfilename+" 2>&1") != 0:
            print "An error occurred when calling Z3, see "+tempfilename+\
                  " for Z3 output"
            return ("error", solution)

        # Parse the result file produced by Z3
        result = "error"
        f = open(tempfilename, "r")
        while True:
            l = f.readline()
            if l == "":
                print "An error occurred when parsing Z3 output, "\
                      "see "+tempfilename+" for Z3 output"
                return ("error", solution)
            l = string.split(_chop(l))
            if l == ["sat"]:
                result = "sat"
                break
            if l == ["unsat"]:
                result = "unsat"
                break

        if result == "sat" and return_solution:
            raise RuntimeError("Producing models when using Z3 as the solver "
                               "is not currently supported")

        f.close()
        #os.remove(tempfilename)
        return (result, solution)


    def _interpret_solution(self,
                            partial_solution,
                            map_var_to_solvervar,
                            map_solvervar_to_var,
                            map_opname_to_op):
        raise RuntimeError("Producing models when using Z3 as the solver "
                           "is not currently supported")



class MathSAT(Solver):
    """
    A front-end for MathSAT.
    """
    def __init__(self, solver_binary="./mathsat"):
        """
        Create a front-end for MathSAT.
        'solver_binary' is the file name for a MathSAT executable binary.
        """
        assert type(solver_binary) is str
        self.binary = solver_binary


    def _solve_smtlib(self, filename, return_solution = False):
        """
        Use MathSAT to solve the SMT-LIB problem in the file 'filename'.
        """
        tempfilename = "tmp_mathsat.out"
        #tempfilename = tempfile.mktemp()
        result = "unknown"
        solution = {}

        if os.system(self.binary+" -input=smt "+filename+
                     " > "+tempfilename+" 2>&1") != 0:
            print "An error occurred when calling MathSAT, see "+tempfilename+\
                  " for MathSAT output"
            return ("error", solution)

        # Parse the result file produced by MathSAT
        result = "error"
        f = open(tempfilename, "r")
        while True:
            l = f.readline()
            if l == "":
                print "An error occurred when parsing MathSAT output, "\
                      "see "+tempfilename+" for MathSAT output"
                return ("error", solution)
            l = string.split(_chop(l))
            if l == ["Result", "=", "0"]:
                result = "unsat"
                break
            if l == ["Result", "=", "1"]:
                result = "sat"
                break

        if result == "sat" and return_solution:
            # Search the beginning of the solution
            while True:
                l = f.readline()
                if l == "":
                    print "No solution found when parsing MathSAT output, "\
                          "see "+tempfilename+" for MathSAT output"
                    return ("error", solution)
                l = string.split(_chop(l))
                if l == ["Assignment:"]:
                    break
            # The following lines contain the solution, interpret it
            while True:
                l = f.readline()
                if l == "":
                    print "EOF encountered when parsing the solution, "\
                          "see "+tempfilename+" for MathSAT output"
                    return ("error", solution)
                l = string.split(_chop(l))
                if l == []:
                    # End of solution
                    break
                if len(l) >= 3 and l[1] == '=':
                    var = l[0]
                    value = l[2]
                    assert var not in solution
                    solution[var] = value
        f.close()
        #os.remove(tempfilename)
        return (result, solution)


    def _solve_native(self, filename, return_solution = False):
        """
        Use MathSAT to solve the problem in the file 'filename'.
        """
        tempfilename = "tmp_mathsat.out"
        #tempfilename = tempfile.mktemp()
        result = "unknown"
        solution = {}

        if os.system(self.binary+" -input=msat "+filename+
                     " > "+tempfilename+" 2>&1") != 0:
            print "An error occurred when calling MathSAT, see "+tempfilename+\
                  " for MathSAT output"
            return ("error", solution)

        # Parse the result file produced by MathSAT
        result = "error"
        f = open(tempfilename, "r")
        while True:
            l = f.readline()
            if l == "":
                print "An error occurred when parsing MathSAT output, "\
                      "see "+tempfilename+" for MathSAT output"
                return ("error", solution)
            l = string.split(_chop(l))
            if l == ["Result", "=", "0"]:
                result = "unsat"
                break
            if l == ["Result", "=", "1"]:
                result = "sat"
                break

        if result == "sat" and return_solution:
            # Search the beginning of the solution
            while True:
                l = f.readline()
                if l == "":
                    print "No solution found when parsing MathSAT output, "\
                          "see "+tempfilename+" for MathSAT output"
                    return ("error", solution)
                l = string.split(_chop(l))
                if l == ["Assignment:"]:
                    break
            # The following lines contain the solution, interpret it
            while True:
                l = f.readline()
                if l == "":
                    print "EOF encountered when parsing the solution, "\
                          "see "+tempfilename+" for MathSAT output"
                    return ("error", solution)
                l = string.split(_chop(l))
                if l == []:
                    # End of solution
                    break
                if len(l) >= 3 and l[1] == '=':
                    var = l[0]
                    value = l[2]
                    assert var not in solution
                    solution[var] = value
        f.close()
        #os.remove(tempfilename)
        return (result, solution)


    def _interpret_solution(self,
                            partial_solution,
                            map_var_to_solvervar,
                            map_solvervar_to_var,
                            map_opname_to_op):
        solution = {}
        for (solvervar,value) in partial_solution.iteritems():
            assert map_solvervar_to_var.has_key(solvervar)
            gate = map_solvervar_to_var[solvervar]
            assert gate not in solution
            assert not gate.children
            assert not gate.is_constant()
            assert gate.operand == 'VAR'
            if gate.is_boolean():
                if value == 'false':
                    solution[gate] = False
                elif value == 'true':
                    solution[gate] = True
                else:
                    assert False
            elif gate.is_sint():
                assert len(value) > 2
                assert value[-2:] == '_2'
                value = value[0:-2]
                assert len(value) == gate.get_sint_width()
                solution[gate] = bin2sint(value)
            elif gate.is_uint():
                assert len(value) > 2
                assert value[-2:] == '_2'
                value = value[0:-2]
                assert len(value) == gate.get_uint_width()
                solution[gate] = bin2uint(value)
            elif gate.is_int():
                solution[gate] = long(value)
            elif gate.is_enum():
                if _is_str_int(value):
                    solution[gate] = gate.get_type().get_literal(int(value))
                elif value.startswith("enum"):
                    while value[0] != '_':
                        value = value[1:]
                    assert gate.get_type().has_literal(value)
                    solution[gate] = value
                else:
                    _internalerror_snyi()
            else:
                #print gate, value
                _internalerror_snyi()

        return solution


class CVC3(Solver):
    """
    A front-end for CVC3.
    """
    def __init__(self, solver_binary = "./cvc3"):
        """
        Create a front-end for CVC3.
        'solver_binary' is the file name for a CVC3 executable binary.
        """
        assert type(solver_binary) is str
        self.binary = solver_binary


    def _solve_smtlib(self, filename, return_solution = False):
        """
        Use CVC3 to solve the SMT-LIB problem in the file 'filename'.
        """
        tempfilename = "tmp_cvc3.out"
        #tempfilename = tempfile.mktemp()
        result = "unknown"
        solution = {}

        if os.system(self.binary+" -lang smtlib "+filename+
                     " > "+tempfilename+" 2>&1") != 0:
            print "An error occurred when calling CVC3, see "+tempfilename+\
                  " for CVC3 output"
            return ("error", solution)

        # Parse the result file produced by CVC3
        f = open(tempfilename, "r")
        l = _chop(f.readline())
        if l == "sat":
            result = "sat"
            raise RuntimeError("Building the model is not supported "
                               "when using CVC3 with the SMT-LIB format, "
                               "aborting")
            if return_solution:
                # The following lines contain the solution, interpret it
                for line in f:
                    l = string.split(line)
                    if l == []:
                        continue
                    if l[0:4] == ['Current','scope','level','is']:
                        continue
                    if l[0:4] == ['%Satisfiable','Variable','Assignment:','%']:
                        continue
                    if len(l) == 2 and l[0] == 'ASSERT':
                        # ASSERT var;
                        assert l[1][-1] == ';'
                        var = l[1][0:-1]
                        assert var not in solution
                        solution[var] = 'true'
                        continue
                    if len(l) == 3 and l[0:2] == ['ASSERT','NOT']:
                        # ASSERT NOT var;
                        assert l[2][-1] == ';'
                        var = l[2][0:-1]
                        assert var not in solution
                        solution[var] = 'false'
                        continue
                    if len(l) == 4 and l[0] == 'ASSERT' and l[2] == '=':
                        # ASSERT (x = v);
                        assert l[1][0] == '('
                        var = l[1][1:]
                        assert l[3].endswith(');')
                        value = l[3][0:-2]
                        assert var not in solution
                        solution[var] = value
                        continue
                    raise RuntimeError("Parsing CVC3 solution failed "+
                                       "on line '"+line+"'")
        elif l == "unsat":
            result = "unsat"
        else:
            #print "'"+l+"'"
            _internalerror_snyi()
        f.close()
        #os.remove(tempfilename)
        return (result, solution)


    def _solve_native(self, filename, return_solution = False):
        """
        Use CVC3 to solve the problem in the file 'filename'.
        """
        tempfilename = "tmp_cvc3.out"
        #tempfilename = tempfile.mktemp()
        result = "unknown"
        solution = {}

        if os.system(self.binary+" < "+filename+
                     " > "+tempfilename+" 2>&1") != 0:
            print "An error occurred when calling CVC3, see "+tempfilename+\
                  " for CVC3 output"
            return ("error", solution)

        # Parse the result file produced by CVC3
        f = open(tempfilename, "r")
        line = _chop(f.readline())
        if line == "Satisfiable.":
            result = "sat"
            if return_solution:
                # The following lines contain the solution, interpret it
                for line in f:
                    if not line:
                        continue
                    if line.startswith("cvc_"):
                        # Don't know what these lines are for, throw 'em away
                        continue
                    if line.startswith("IN "):
                        # Don't know what these lines are for, throw 'em away
                        continue
                    if re.search("^ASSERT \(t_.*;\n", line):
                        continue
                    if re.search("^ASSERT \(o_VAR[0-9]+ = \(ARRAY.*;\n", line):
                        continue
                    m = re.search("^ASSERT \(([a-zA-Z0-9_]*)\(([a-zA-Z0-9\-_\,\ ]*)\) = ([a-zA-Z0-9\-_]*)\);\n", line)
                    if m:
                        # Values for uninterpreted functions
                        # ASSERT (f(a1,...,an) = value);
                        name = m.groups(0)[0]
                        args = [arg.strip() for arg in
                                m.groups(0)[1].split(",")]
                        value = m.groups(0)[2]
                        key = [name]
                        key.extend(args)
                        key = tuple(key)
                        assert key not in solution
                        solution[key] = value
                        continue
                    m = re.search("^ASSERT \((o_VAR[0-9]+)\[([a-zA-Z0-9\-_]*)\] = ([a-zA-Z0-9\-_]+)\);\n", line)
                    if m:
                        # Values for uninterpreted functions
                        # ASSERT (o_VAR[index] = value);
                        name = m.groups(0)[0]
                        index = m.groups(0)[1]
                        value = m.groups(0)[2]
                        key = (name,index)
                        assert key not in solution
                        solution[key] = value
                        continue
                    l = string.split(line)
                    if l == []:
                        continue
                    if l[0:4] == ['Current','scope','level','is']:
                        continue
                    if l[0:4] == ['%Satisfiable','Variable','Assignment:','%']:
                        continue
                    if len(l) == 2 and l[0] == 'ASSERT':
                        # ASSERT var;
                        assert l[1].endswith(';')
                        var = l[1][0:-1]
                        assert var not in solution
                        solution[var] = 'true'
                        continue
                    if len(l) == 3 and l[0:2] == ['ASSERT','NOT']:
                        # ASSERT NOT var;
                        assert l[2].endswith(';')
                        var = l[2][0:-1]
                        assert var not in solution
                        solution[var] = 'false'
                        continue
                    if len(l) == 4 and l[0] == 'ASSERT' and l[2] == '=':
                        # ASSERT (var = value);
                        assert l[1][0] == '('
                        var = l[1][1:]
                        if var.startswith('_'):
                            continue
                        assert l[3].endswith(');')
                        value = l[3][0:-2]
                        assert var not in solution
                        solution[var] = value
                        continue
                    if(len(l) >= 4 and l[0] == 'ASSERT' and l[2] == '=' and
                       l[3] == '(LAMBDA'):
                        # ASSERT (var = (LAMBDA ...
                        continue
                    raise RuntimeError("Parsing CVC3 solution failed "+
                                       "on line '"+_chop(line)+"'")
        elif line == "Unsatisfiable.":
            result = "unsat"
        else:
            raise RuntimeError("Parsing CVC3 output failed "+
                               "on line '"+line+"'")
        f.close()
        #os.remove(tempfilename)
        #print solution
        return (result, solution)


    def _interpret_solution(self, partial_solution,
                            map_var_to_solvervar, map_solvervar_to_var,
                            map_opname_to_op):
        solution = {}
        def _v2v(t, value):
            """
            Convert a primitive CVC3 value to primitive PySMT value.
            """
            assert isinstance(t, _Type)
            assert type(value) is str
            if t.is_bool():
                if value == 'false':
                    return False
                elif value == 'true':
                    return True
                else:
                    assert False
            elif t.is_sint():
                assert value.startswith('0bin')
                value = value[len('0bin'):]
                assert len(value) == t.get_width()
                return bin2sint(value)
            elif t.is_uint():
                assert value.startswith('0bin')
                value = value[len('0bin'):]
                assert len(value) == t.get_width()
                return bin2uint(value)
            elif t.is_int():
                return long(value)
            elif t.is_enum():
                assert value.startswith("enum_")
                value = value[len("enum_"):]
                while value[0] != "_":
                    value = value[1:]
                value = value[1:]
                assert t.has_literal(value)
                return value
            else:
                #print t, value
                _internalerror_snyi()


        for (solvervar,value) in partial_solution.iteritems():
            if type(solvervar) is tuple and solvervar[0].startswith("o_VAR"):
                assert len(solvervar) == 2
                (name,index) = solvervar
                assert name in map_solvervar_to_var
                gate = map_solvervar_to_var[name]
                assert not gate.children
                assert not gate.is_constant()
                assert gate.operand == 'VAR'
                if gate not in solution:
                    solution[gate] = {}
                index_value = _v2v(gate.get_type().get_index_type(), index)
                value = _v2v(gate.get_type().get_range_type(), value)
                solution[gate][index_value] = value
                continue
            
            # Handle operand valuations
            if type(solvervar) is tuple:
                opname = solvervar[0]
                arg_values = solvervar[1:]
                op = map_opname_to_op[opname]
                assert op.is_uif()
                real_arg_values = []
                if type(arg_values) is tuple:
                    for (index,arg_value) in enumerate(arg_values):
                        dtype = op.get_domain_type(index)
                        real_arg_values.append(_v2v(dtype, arg_value))
                else:
                    dtype = op.get_domain_type(0)
                    real_arg_values.append(_v2v(dtype, arg_values))

                assert type(value) is str
                real_value = _v2v(op.get_range_type(), value)
                    
                solution[(op, tuple(real_arg_values))] = real_value
                continue
            
            if solvervar.startswith("t_"):
                continue
            #print solvervar
            assert solvervar in map_solvervar_to_var
            gate = map_solvervar_to_var[solvervar]
            assert gate not in solution
            assert not gate.children
            assert not gate.is_constant()
            assert gate.operand == 'VAR'
            solution[gate] = _v2v(gate.get_type(), value)

        return solution


class Yices(Solver):
    """
    A front-end for Yices.
    """
    def __init__(self, solver_binary="./yices"):
        """
        Create a front-end for Yices.
        'solver_binary' is the file name of the Yices executable binary.
        """
        assert type(solver_binary) is str
        self.binary = solver_binary


    def _parse_solution(self, f):
        """A small helper for parsing the solution."""
        assert type(f) is file
        solution = {}
        while True:
            l = f.readline()
            if not l:
                break
            if l.strip() == "":
                continue
            if l.startswith("(= "):
                l = string.split(l)[1:]
                var_list = []
                parens_seen = 0
                while True:
                    if not l:
                        l = f.readline()
                        l = l.split()
                        continue
                    item = l.pop(0)
                    if item.startswith("("):
                        assert parens_seen == 0
                        parens_seen += 1
                        item = item[1:]
                        assert not item.startswith("(")
                    if item.endswith(")"):
                        assert parens_seen == 1
                        parens_seen -= 1
                        item = item[:-1]
                        assert not item.endswith(")")
                    if not item:
                        continue
                    assert "(" not in item
                    assert ")" not in item
                    var_list.append(item)
                    if parens_seen == 0:
                        break
                value_list = []
                while True:
                    if not l:
                        l = f.readline()
                        l = l.split()
                        continue
                    item = l.pop(0)
                    assert item.endswith(")")
                    item = item[:-1]
                    value_list.append(item)
                    assert not l
                    break
                if len(var_list) > 1:
                    k = tuple(var_list)
                else:
                    k = var_list[0]
                if len(value_list) > 1:
                    v = tuple(value_list)
                else:
                    v = value_list[0]
                assert k not in solution
                solution[k] = v
                continue
            assert False
        return solution

    def _solve_smtlib(self, filename, return_solution = False):
        """
        Use Yices to solve the SMT-LIB problem in the file 'filename'.
        """
        tempfilename = "tmp_yices.out"
        #tempfilename = tempfile.mktemp()
        result = "unknown"
        solution = {}

        if os.system(self.binary+" --smtlib --evidence "+filename+
                     " > "+tempfilename+" 2>&1") != 0:
            print "An error occurred when calling yices, see "+tempfilename+\
                  " for yices output"
            return ("error", solution)

        # Parse the result file produced by yices
        f = open(tempfilename, "r")
        l = _chop(f.readline())
        if l == "sat":
            result = "sat"
            if return_solution:
                solution = self._parse_solution(f)
        elif l == "unsat":
            result = "unsat"
        else:
            #print "'"+l+"'"
            _internalerror_snyi()
        f.close()
        #os.remove(tempfilename)
        return (result, solution)


    def _solve_native(self, filename, produce_model = False):
        """
        Use Yices to solve the problem in the file 'filename'.
        """
        tempfilename = "tmp_yices.out"
        #tempfilename = tempfile.mktemp()
        result = "unknown"
        model = {}

        if os.system(self.binary+" < "+filename+
                     " > "+tempfilename+" 2>&1") != 0:
            print "An error occurred when calling yices, see "+tempfilename+\
                  " for yices output"
            return ("error", model)

        # Parse the result file produced by yices
        f = open(tempfilename, "r")
        l = _chop(f.readline())
        if l == "sat":
            result = "sat"
            if produce_model:
                model = self._parse_solution(f)
        elif l == "unsat":
            result = "unsat"
        else:
            #print "'"+l+"'"
            _internalerror_snyi()
        f.close()
        #print model
        #os.remove(tempfilename)
        return (result, model)


    def _interpret_solution(self,
                            partial_solution,
                            map_var_to_smt_name,
                            map_smt_name_to_var,
                            map_opname_to_op):
        def _v2v(t, value):
            """
            Convert a primitive Yices value to primitive PySMT value.
            """
            assert isinstance(t, _Type)
            assert type(value) is str
            if t.is_bool():
                if value == 'false':
                    return False
                elif value == 'true':
                    return True
                else:
                    assert False
            elif t.is_sint():
                assert value[0:2] == '0b'
                value = value[2:]
                assert len(value) == t.get_width()
                return bin2sint(value)
            elif t.is_uint():
                assert value[0:2] == '0b'
                value = value[2:]
                assert len(value) == t.get_width()
                return bin2uint(value)
            elif t.is_int():
                return long(value)
            elif t.is_enum():
                assert value.startswith("enum_")
                value = value[len("enum_"):]
                while value[0] != "_":
                    value = value[1:]
                value = value[1:]
                assert t.has_literal(value)
                return value
            else:
                #print t, value
                _internalerror_snyi()


        solution = {}
        for (solvervar,value) in partial_solution.iteritems():
            if(type(solvervar) is tuple and
               solvervar[0].startswith("t_")):
                continue
            if(type(solvervar) is tuple and
               (solvervar[0].startswith("o_VAR") or
                solvervar[0].startswith("t_"))):
                # An array gate
                assert len(solvervar) == 2
                (name,index) = solvervar
                #print solvervar,value
                #print map_smt_name_to_var
                assert name in map_smt_name_to_var
                gate = map_smt_name_to_var[name]
                assert not gate.children
                assert not gate.is_constant()
                assert gate.operand == 'VAR'
                if gate not in solution:
                    solution[gate] = {}
                index_value = _v2v(gate.get_type().get_index_type(), index)
                value = _v2v(gate.get_type().get_range_type(), value)
                solution[gate][index_value] = value
                continue

            # Handle operand valuations
            if type(solvervar) is tuple:
                opname = solvervar[0]
                arg_values = solvervar[1:]
                op = map_opname_to_op[opname]
                assert op.is_uif()
                real_arg_values = []
                if type(arg_values) is tuple:
                    for (index,arg_value) in enumerate(arg_values):
                        dtype = op.get_domain_type(index)
                        real_arg_values.append(_v2v(dtype, arg_value))
                else:
                    dtype = op.get_domain_type(0)
                    real_arg_values.append(_v2v(dtype, arg_values))

                assert type(value) is str
                real_value = _v2v(op.get_range_type(), value)
                    
                solution[(op, tuple(real_arg_values))] = real_value
                continue
            
            if solvervar.startswith("t_"):
                continue
            assert map_smt_name_to_var.has_key(solvervar)
            gate = map_smt_name_to_var[solvervar]
            assert gate not in solution
            assert not gate.children
            assert not gate.is_constant()
            assert gate.operand == 'VAR'
            solution[gate] = _v2v(gate.get_type(), value)

        return solution


class STP(Solver):
    """
    A front-end for STP.
    """
    def __init__(self, solver_binary="./stp"):
        """
        Create a front-end for STP.
        'solver_binary' is the file name for an STP executable binary.
        """
        assert type(solver_binary) is str
        self.binary = solver_binary


    def _solve_smtlib(self, filename, return_solution = False):
        """
        Use STP to solve the SMT-LIB problem in the file 'filename'.
        """
        tempfilename = "tmp_stp.out"
        #tempfilename = tempfile.mktemp()
        result = "unknown"
        solution = {}

        if os.system(self.binary+" -c -d -p "+filename+
                     " > "+tempfilename+" 2>&1") != 0:
            print "An error occurred when calling STP, see "+tempfilename+\
                  " for STP output"
            return ("error", solution)

        # Parse the result file produced by STP
        f = open(tempfilename, "r")
        l = _chop(f.readline())
        if l == "Invalid.":
            result = "sat"
            if return_solution:
                # The following lines contain the solution, interpret it
                for l in f:
                    l = string.split(l)
                    if l == []:
                        continue
                    if l[0] == 'ASSERT(':
                        assert len(l) == 5
                        var = l[1]
                        assert l[2] == '='
                        value = l[3]
                        assert l[4] == ');'
                        assert var not in solution
                        solution[var] = value
                    else:
                        #print l
                        _internalerror_snyi()
        elif l == "Valid.":
            result = "unsat"
        else:
            print "'"+l+"'"
            _internalerror_snyi()
        f.close()
        #os.remove(tempfilename)
        return (result, solution)


    def _solve_native(self, filename, return_solution = False):
        """
        Use STP to solve the problem in the file 'filename'.
        """
        tempfilename = "tmp_stp.out"
        #tempfilename = tempfile.mktemp()
        result = "unknown"
        solution = {}

        if os.system(self.binary+" "+filename+
                     " > "+tempfilename+" 2>&1") != 0:
            print "An error occurred when calling STP, see "+tempfilename+\
                  " for STP output"
            return ("error", solution)

        # Parse the result file produced by STP
        f = open(tempfilename, "r")
        l = _chop(f.readline())
        if l == "Invalid.":
            result = "sat"
            if return_solution:
                # The following lines contain the solution, interpret it
                while True:
                    l = f.readline()
                    if not l:
                        break
                    l = string.split(l)
                    if l == []:
                        continue
                    if(len(l) == 5 and l[0] == 'ASSERT(' and l[2] == '=' and
                       l[4] == ');'):
                        var = l[1]
                        value = l[3]
                        assert var not in solution
                        solution[var] = value
                    elif l[0] == 'ASSERT(':
                        while l == [] or l[-1] != ');':
                            l = f.readline()
                            assert l and "Error while parsing STP output"
                            l = string.split(l)
                    else:
                        #print l
                        _internalerror_snyi()
        elif l == "Valid.":
            result = "unsat"
        else:
            #print "'"+l+"'"
            _internalerror_snyi()
        f.close()
        #os.remove(tempfilename)
        return (result, solution)


    def _interpret_solution(self, partial_solution,
                            map_var_to_smt_name, map_smt_name_to_var):
        solution = {}
        for (smt_name,value) in partial_solution.iteritems():
            if smt_name.startswith("t_"):
                continue
            assert map_smt_name_to_var.has_key(smt_name)
            gate = map_smt_name_to_var[smt_name]
            assert gate not in solution
            assert not gate.children
            assert not gate.is_constant()
            assert gate.operand == 'VAR'
            if gate.is_boolean():
                if value == 'FALSE':
                    solution[gate] = False
                elif value == 'TRUE':
                    solution[gate] = True
                else:
                    #assert False
                    pass
            elif gate.is_sint():
                if value[0:4] == '0bin':
                    value = value[4:]
                    width = len(value)
                    assert width == gate.get_sint_width()
                    v = long(0)
                    for i in range(0,width):
                        if value[i] == '1':
                            v += pow(2, width-1-i)
                    if v >= pow(2,width-1):
                        v = -pow(2,width)+v
                    solution[gate] = v
                elif value[0:4] == '0hex':
                    value = value[4:]
                    width = len(value)*4
                    assert width == gate.get_sint_width()
                    v = long(0)
                    trans = {'0':0,'1':1,'2':2,'3':3,'4':4,'5':5,'6':6,'7':7,
                             '8':8,'9':9,'A':10,'B':11,'C':12,'D':13,'E':14,
                             'F':15}
                    for i in range(0,len(value)):
                        assert value[i] in trans
                        v += trans[value[i]] * pow(2, (len(value)-1-i)*4)
                    if v >= pow(2,width-1):
                        v = -pow(2,width)+v
                    solution[gate] = v
                else:
                    #assert False and "Not yet implemented"
                    pass
            elif gate.is_uint():
                if value[0:4] == '0bin':
                    value = value[4:]
                    width = len(value)
                    assert width == gate.get_uint_width()
                    v = long(0)
                    for i in range(0,width):
                        if value[i] == '1':
                            v += pow(2, width-1-i)
                    solution[gate] = v
                elif value[0:4] == '0hex':
                    value = value[4:]
                    width = len(value)*4
                    assert width == gate.get_uint_width()
                    v = long(0)
                    trans = {'0':0,'1':1,'2':2,'3':3,'4':4,'5':5,'6':6,'7':7,
                             '8':8,'9':9,'A':10,'B':11,'C':12,'D':13,'E':14,
                             'F':15}
                    for i in range(0,len(value)):
                        assert value[i] in trans
                        v += trans[value[i]] * pow(2, (len(value)-1-i)*4)
                    solution[gate] = v
                else:
                    #assert False and "Not yet implemented"
                    pass
            else:
                _internalerror_snyi()

        return solution
    


class Gate:
    """
    Just a handle front-end, the actual gates are not visible to the user.
    Use Circuit.add_* functions to create gates.
    """
    def _get_gate(self):
        assert self._gate
        return self._gate
    def __str__(self):
        return str(self._get_gate())
    def __del__(self):
        real_gate = self._get_gate()
        if not real_gate.has_handles():
            real_gate.get_circuit()._garbage_collect(real_gate)
    def get_type(self):
        """Return the type of the gate."""
        return self._get_gate().get_type()



class Circuit:
    """
    An extended Boolean circuit.
    """
    class _RealGate:
        def __init__(self, circuit, tyype, operand, opparams, children):
            self.constrained_value = None
            self.handle = None
            self._handles = weakref.WeakKeyDictionary()
            self.circuit = circuit
            self.tyype = tyype
            self.operand = operand
            self.opparams = opparams
            self.children = []
            self.parents = sets.Set()
            self.index = None
            for child in children:
                assert isinstance(child, Circuit._RealGate)
                assert child.has_index()
                assert circuit._gates[child.get_index()] is child
                assert child.get_circuit() is circuit
                self.children.append(child)
        def get_circuit(self):
            return self.circuit
        def set_index(self, index):
            """
            Set the index of this gate in the list of gates
            of the owning circuit.
            """
            assert type(index) is int
            assert self.get_circuit()._gates[index] is self
            self.index = index
        def has_index(self):
            return self.index is not None
        def get_index(self):
            assert self.index is not None
            return self.index

        def get_structural_id(self):
            return (self.get_type(),
                    self.operand,
                    self.opparams,
                    tuple([child.get_index() for child in self.children]))

        def __str__(self):
            s = 'g'+str(self.get_index())+' := '
            s += str(self.get_type())+' '
            s += str(self.operand)
            if self.opparams != None:
                s += str(self.opparams)
            s += '('
            s += ','.join(['g'+str(c.get_index()) for c in self.children])
            s += ');'
            return s


        def is_constrained(self):
            return self.constrained_value != None
            
        def has_handles(self):
            return len(self._handles) > 0

        def is_boolean(self):
            return self.get_type().is_bool()

        def is_enum(self):
            return self.get_type().is_enum()

        def is_array(self):
            return self.get_type().is_array()

        def is_int(self):
            return self.get_type().is_int()

        def is_rat(self):
            return self.get_type().is_rat()

        def is_sint(self):
            return self.get_type().is_sint()

        def get_sint_width(self):
            assert self.is_sint()
            return self.get_type().get_width()

        def is_uint(self):
            return self.get_type().is_uint()

        def get_uint_width(self):
            assert self.is_uint()
            return self.get_type().get_width()

        def get_bitwidth(self):
            if self.is_sint():
                return self.get_sint_width()
            if self.is_uint():
                return self.get_uint_width()
            assert False and "Internal error"

        def get_type(self):
            return self.tyype

        def is_commutative(self):
            if self.operand in ['OR','AND','PLUS','TIMES','EQ']:
                return True
            return False

        def sort_children(self):
            if self.is_commutative():
                self.children.sort(lambda a,b : cmp(a.index, b.index))

        def is_constant(self):
            if self.operand in ['FALSE','TRUE','CONST']:
                return True
            return False

        def is_var(self):
            """
            Is this gate a variable?
            """
            if self.operand == 'VAR':
                assert not self.children
                return True
            return False

        def to_dot(self, f):
            param_str = ''
            if self.opparams is not None:
                param_str = str(self.opparams)
            i = self.get_index()
            f.write('n'+str(i)+' ')
            f.write('[label="g'+str(i))
            f.write(' '+str(self.operand)+param_str)
            if self.is_constrained():
                f.write(' '+str(self.constrained_value))
            f.write('"];\n')
            for (index,child) in enumerate(self.children):
                f.write('n'+str(i)+' -> n'+str(child.get_index())+
                        ' [label="'+str(index)+'"];\n')
            if False:
                for (parent_index,dummy) in self.parents:
                    f.write('n'+str(parent_index)+' -> n'+str(i)+
                            ' [color=red];\n')

        def _to_cvc3(self, cache, defines, variables, extra_ops, is_root):
            """
            - 'extra_ops' is a dictionary into which encountered UIFs and
              enumeration data types are stored.
            """
            assert type(extra_ops) is dict
            assert type(is_root) is bool

            def _nyi(s):
                raise RuntimeError("Not yet implemented in CVC3 "
                                   "translation: "+str(s)+", aborting")

            tyype = self.get_type()
            param_str = ''
            if self.opparams != None:
                param_str = str(self.opparams)
            s = str(self.operand)+param_str

            # Index and store encoundered UIFs
            if isinstance(self.operand, _Operand):
                if self.operand not in extra_ops:
                    extra_ops[self.operand] = "uif_"+str(len(extra_ops))
            # Index and store encountered enumeration data types
            if self.get_type().is_enum():
                if self.get_type() not in extra_ops:
                    extra_ops[self.get_type()] = "enum_"+str(len(extra_ops))
            # Index and store encountered array data types
            if self.get_type().is_array():
                if self.get_type() not in extra_ops:
                    extra_ops[self.get_type()] = "A"+str(len(extra_ops))

            if not self.children:
                if self.operand == 'CONST':
                    if self.is_int():
                        return str(self.opparams)
                    elif self.is_sint():
                        width = self.get_sint_width()
                        return sint2bin(width, self.opparams, '0bin')
                    elif self.is_uint():
                        width = self.get_uint_width()
                        return uint2bin(width, self.opparams, '0bin')
                    elif self.is_enum():
                        return extra_ops[self.get_type()]+"_"+self.opparams
                    else:
                        _nyi("constants for type '"+str(self.get_type())+"'")
                elif self.operand == 'FALSE':
                    return 'FALSE'
                elif self.operand == 'TRUE':
                    return 'TRUE'
                assert not self.is_constant()
                assert self.operand == 'VAR'
                s = 'o_'+s
                variables.append((self,s))
                return s

            if self in cache:
                return cache[self]

            translated_children = []
            for child in self.children:
                translated_children.append(child._to_cvc3(cache,defines,
                                                          variables, extra_ops,
                                                          False))

            if self.operand == 'AND':
                assert len(self.children) >= 2
                s = '('+(' AND '.join(translated_children))+')'
            elif self.operand == 'OR':
                assert len(self.children) >= 2
                s = '('+(' OR '.join(translated_children))+')'
            elif self.operand == 'IFF':
                assert len(self.children) == 2
                s = '('+translated_children[0]+' <=> '+\
                    translated_children[1]+')'
            elif self.operand == 'NOT':
                assert len(self.children) == 1
                s = '(NOT '+translated_children[0]+')'
            elif self.operand == 'ITE':
                assert len(self.children) == 3
                s = '(IF '+\
                    translated_children[0]+' THEN '+\
                    translated_children[1]+' ELSE '+\
                    translated_children[2]+' ENDIF)'
            # Bitwise operations
            elif self.operand == 'BVOR':
                assert len(self.children) == 2
                s = '('+(' | '.join(translated_children))+')'
            elif self.operand == 'BVAND':
                assert len(self.children) == 2
                s = '('+(' & '.join(translated_children))+')'
            elif self.operand == 'BVXOR':
                assert len(self.children) == 2
                s = 'BVXOR('+(','.join(translated_children))+')'
            elif self.operand == 'BVNOT':
                assert len(self.children) == 1
                s = '(~ '+translated_children[0]+')'
            # Arithmetic operations
            elif self.operand == 'PLUS':
                assert len(self.children) >= 2
                if self.is_sint():
                    width = self.get_sint_width()
                    s = 'BVPLUS('+str(width)+','+\
                        (','.join(translated_children))+')'
                elif self.is_uint():
                    width = self.get_uint_width()
                    s = 'BVPLUS('+str(width)+','+\
                        (','.join(translated_children))+')'
                elif self.is_int() or self.is_rat():
                    s = '('+(' + '.join(translated_children))+')'
                else:
                    _nyi("addition for type '"+str(self.get_type())+"'")
            elif self.operand == 'MINUS':
                assert len(self.children) == 2
                if self.is_sint():
                    width = self.get_sint_width()
                    s = 'BVSUB('+str(width)+','+\
                        (','.join(translated_children))+')'
                elif self.is_uint():
                    width = self.get_uint_width()
                    s = 'BVSUB('+str(width)+','+\
                        (','.join(translated_children))+')'
                elif self.is_int() or self.is_rat():
                    s = '('+(' - '.join(translated_children))+')'
                else:
                    _nyi("subtraction for type '"+str(self.get_type())+"'")
            elif self.operand == 'NEG':
                assert len(self.children) == 1
                if self.is_sint() or self.is_uint():
                    s = 'BVUMINUS('+translated_children[0]+')'
                elif self.is_int() or self.is_rat():
                    s = '(- '+translated_children[0]+')'
                else:
                    _nyi("negation for type '"+str(self.get_type())+"'")
            elif self.operand == 'TIMES':
                assert len(self.children) == 2
                if self.is_sint():
                    width = self.get_sint_width()
                    s = 'BVMULT('+str(width)+','+\
                        (','.join(translated_children))+')'
                elif self.is_uint():
                    width = self.get_uint_width()
                    s = 'BVMULT('+str(width)+','+\
                        (','.join(translated_children))+')'
                elif self.is_int() or self.is_rat():
                    s = '('+(' * '.join(translated_children))+')'
                else:
                    _nyi("multiplication for type '"+str(self.get_type())+"'")
            # Comparison predicates
            elif self.operand == 'EQ':
                assert len(self.children) == 2
                child1 = self.children[0]
                if(child1.is_int() or child1.is_rat() or
                   child1.is_sint() or child1.is_uint() or
                   child1.is_enum()):
                    s = '('+translated_children[0]+' = '+\
                        translated_children[1]+')'
                else:
                    _nyi("equality comparison for type '"+
                         str(child1.get_type())+"'")
            elif self.operand == 'LE':
                assert len(self.children) == 2
                child1 = self.children[0]
                if child1.is_int() or child1.is_rat():
                    s = '('+translated_children[0]+' <= '+\
                        translated_children[1]+')'
                elif child1.is_sint():
                    s = 'BVSLE('+translated_children[0]+','+\
                        translated_children[1]+')'
                elif child1.is_uint():
                    s = 'BVLE('+translated_children[0]+','+\
                        translated_children[1]+')'
                else:
                    _nyi("less-or-equal comparison for type '"+
                         str(child1.get_type())+"'")
            elif self.operand == 'LT':
                assert len(self.children) == 2
                child1 = self.children[0]
                if child1.is_int() or child1.is_rat():
                    s = '('+translated_children[0]+' < '+\
                        translated_children[1]+')'
                elif child1.is_sint():
                    s = 'BVSLT('+translated_children[0]+','+\
                        translated_children[1]+')'
                elif child1.is_uint():
                    s = 'BVLT('+translated_children[0]+','+\
                        translated_children[1]+')'
                else:
                    _nyi("less-than comparison for type '"+
                         str(child1.get_type())+"'")

            # Array operations
            elif self.operand == 'READ':
                assert len(self.children) == 2
                s = '('+translated_children[0]+'['+\
                        translated_children[1]+'])'
            elif self.operand == 'WRITE':
                assert len(self.children) == 3
                s = '('+translated_children[0]+' '+\
                    'WITH ['+translated_children[1]+'] '+\
                    ':= '+translated_children[2]+')'
                is_root = True

            elif isinstance(self.operand, _Operand_uif):
                s = extra_ops[self.operand]+"("+\
                    (",".join(translated_children))+")"
            else:
                _nyi("operand '"+str(self.operand)+"'")

            if is_root or len(self.parents) >= 2:
                # This gate is a root or shared by at least two gates,
                # add a definition for it
                if self.is_boolean():
                    t = 't_'+str(self.get_index())
                else:
                    t = 't_'+str(self.get_index())
                cache[self] = t
                defines.append((self,t,s))
                return t
            return s


        def _to_smt(self, cache, defines, variables, extra_ops,
                    is_root,
                    opt_use_lets):
            """
            - 'extra_ops' is a dictionary into which encountered UIFs and
              enumeration data types are stored.
            - 'opt_use_lets' is a Boolean flag that tells whether
              the shared gates are later to be defined with
              (f)let constructions or with equality :assumptions.
            """
            assert type(cache) is dict
            assert type(extra_ops) is dict
            assert type(is_root) is bool
            assert type(opt_use_lets) is bool

            def _nyi(s):
                raise RuntimeError("Not yet implemented in the SMT-LIB "
                                   "translation: "+str(s)+", aborting")

            tyype = self.get_type()
            param_str = ''
            if self.opparams != None:
                param_str = str(self.opparams)
            s = str(self.operand)+param_str

            # Index and store encoundered UIFs
            if isinstance(self.operand, _Operand):
                if self.operand not in extra_ops:
                    extra_ops[self.operand] = "uif_"+str(len(extra_ops))
            # Index and store encountered enumeration data types
            if self.get_type().is_enum():
                if self.get_type() not in extra_ops:
                    extra_ops[self.get_type()] = "enum_"+str(len(extra_ops))

            if not self.children:
                if self.operand == 'CONST':
                    if self.is_int():
                        return str(self.opparams)
                    elif self.is_sint():
                        width = self.get_sint_width()
                        if self.opparams >= 0:
                            s = "bv"+str(self.opparams)
                        else:
                            s = "(bvneg bv"+str(-self.opparams)+")"
                        # Should not be necessary if width=32 but some
                        # solvers seem to require it even then
                        s = "(extract["+str(width-1)+":0] "+s+")"
                        return s
                    elif self.is_uint():
                        width = self.get_uint_width()
                        s = "bv"+str(self.opparams)
                        # Should not be necessary if width=32 but some
                        # solvers seem to require it even then
                        s = "(extract["+str(width-1)+":0] "+s+")"
                        return s
                    else:
                        _nyi("constants of type '"+str(self.get_type())+"'")
                elif self.operand == 'FALSE':
                    return 'false'
                elif self.operand == 'TRUE':
                    return 'true'
                assert not self.is_constant()
                assert self.operand == 'VAR'
                s = 'o_'+s
                variables.append((self,s))
                return s

            if self in cache:
                return cache[self]

            s = '('
            if self.operand == 'AND':
                assert len(self.children) >= 2
                s += 'and'
            elif self.operand == 'OR':
                assert len(self.children) >= 2
                s += 'or'
            elif self.operand == 'IFF':
                assert len(self.children) == 2
                s += 'iff'
            elif self.operand == 'NOT':
                assert len(self.children) == 1
                s += 'not'
            elif self.operand == 'ITE':
                if self.is_boolean():
                    assert len(self.children) == 3
                    s += 'if_then_else'
                elif(self.is_int() or self.is_rat() or
                     self.is_sint() or self.is_uint()):
                    assert len(self.children) == 3
                    s += 'ite'
                else:
                    _nyi("ite on type '"+str(self.get_type())+"'")
            # Bitwise operations
            elif self.operand == 'BVOR':
                assert len(self.children) == 2
                s += 'bvor'
            elif self.operand == 'BVAND':
                assert len(self.children) == 2
                s += 'bvand'
            elif self.operand == 'BVXOR':
                assert len(self.children) == 2
                _nyi("bitwise xor")
            elif self.operand == 'BVNOT':
                assert len(self.children) == 1
                s += 'bvnot'
            # Arithmetic operations
            elif self.operand == 'PLUS':
                assert len(self.children) >= 2
                if self.is_sint() or self.is_uint():
                    s += 'bvadd'
                else:
                    _nyi("addition on type '"+str(self.get_type())+"'")
            elif self.operand == 'MINUS':
                assert len(self.children) == 2
                if self.is_sint() or self.is_uint():
                    s += 'bvsub'
                else:
                    _nyi("subtraction on type '"+str(self.get_type())+"'")
            elif self.operand == 'NEG':
                assert len(self.children) == 1
                if self.is_sint() or self.is_uint():
                    s += 'bvneg'
                elif self.is_int() or self.is_rat():
                    s += '~'
                else:
                    _nyi("negation on type '"+str(self.get_type())+"'")
            elif self.operand == 'TIMES':
                assert len(self.children) >= 2
                if self.is_sint() or self.is_uint():
                    s += 'bvmul'
                else:
                    _nyi("multiplication on type '"+str(self.get_type())+"'")
            elif self.operand == 'EQ':
                assert len(self.children) == 2
                child1 = self.children[0]
                if(child1.is_int() or child1.is_rat() or
                   child1.is_sint() or child1.is_uint()):
                    s += '='
                else:
                    _nyi("equality comparison on type '"+
                         str(child1.get_type())+"'")
            elif self.operand == 'LE':
                assert len(self.children) == 2
                child1 = self.children[0]
                if child1.is_int() or child1.is_rat():
                    s += "<="
                elif child1.is_sint():
                    s += 'bvsleq'
                elif child1.is_uint():
                    s += 'bvleq'
                else:
                    _nyi("less-or-equal comparison on type '"+
                         str(child1.get_type())+"'")
            elif self.operand == 'LT':
                assert len(self.children) == 2
                child1 = self.children[0]
                if child1.is_int() or child1.is_rat():
                    s += "<"
                elif child1.is_sint():
                    s += 'bvslt'
                elif child1.is_uint():
                    s += 'bvlt'
                else:
                    _nyi("less-than comparison on type '"+
                         str(child1.get_type())+"'")
            elif isinstance(self.operand, _Operand_uif):
                s += extra_ops[self.operand]
            else:
                #print self
                _internalerror_snyi()
            s += ' '
            s += ' '.join([child._to_smt(cache, defines, variables,
                                         extra_ops, False, opt_use_lets)
                           for child in self.children])
            s += ')'
            
            if is_root or len(self.parents) >= 2:
                # This gate is a root or shared by at least two gates,
                # add a definition for it
                if self.is_boolean():
                    t = 't_'+str(self.get_index())
                    if opt_use_lets:
                        t = '$'+t
                else:
                    t = 't_'+str(self.get_index())
                    if opt_use_lets:
                        t = '?'+t
                cache[self] = t
                defines.append((self,t,s))
                return t
            return s



        def _to_smt2(self, cache, defines, variables, extra_ops,
                     opt_use_lets):
            """
            - 'extra_ops' is a dictionary into which encountered UIFs and
              enumeration data types are stored.
            - 'opt_use_lets' is a Boolean flag that tells whether
              the shared gates are later to be defined with
              (f)let constructions or with equality :assumptions.
            """
            assert type(cache) is dict
            assert type(extra_ops) is dict
            assert type(opt_use_lets) is bool

            def _nyi(s):
                raise RuntimeError("Not yet implemented in the SMT-LIB "
                                   "translation: "+str(s)+", aborting")

            gatestack = [self]
            while gatestack:
                g = gatestack[-1]

                # Already translated?
                if g in cache:
                    gatestack.pop()
                    continue

                # Check whether all children have been translated
                translated_children = []
                has_untranslated_children = False
                for child in g.children:
                    if child not in cache:
                        has_untranslated_children = True
                        gatestack.append(child)
                    else:
                        translated_children.append(cache[child])
                if has_untranslated_children:
                    continue

                gtype = g.get_type()

                # Index and store encoundered UIFs
                if isinstance(g.operand, _Operand):
                    if g.operand not in extra_ops:
                        extra_ops[g.operand] = "uif_"+str(len(extra_ops))
                # Index and store encountered enumeration data types
                if gtype.is_enum():
                    if gtype not in extra_ops:
                        extra_ops[gtype] = "enum_"+str(len(extra_ops))

                if not g.children:
                    if g.operand == 'CONST':
                        if g.is_int():
                            s = str(g.opparams)
                        elif g.is_sint():
                            width = g.get_sint_width()
                            if g.opparams >= 0:
                                s = "bv"+str(g.opparams)
                            else:
                                s = "(bvneg bv"+str(-g.opparams)+")"
                            # Should not be necessary if width=32 but some
                            # solvers seem to require it even then
                            s = "(extract["+str(width-1)+":0] "+s+")"
                        elif g.is_uint():
                            width = g.get_uint_width()
                            s = "bv"+str(g.opparams)
                            # Should not be necessary if width=32 but some
                            # solvers seem to require it even then
                            s = "(extract["+str(width-1)+":0] "+s+")"
                        else:
                            _nyi("constants of type '"+str(gtype)+"'")
                    elif g.operand == 'FALSE':
                        s = 'false'
                    elif g.operand == 'TRUE':
                        s = 'true'
                    else:
                        assert not g.is_constant()
                        assert g.operand == 'VAR'
                        s = 'o_'+str(g.operand)+str(g.opparams)
                        variables.append((g,s))
                    cache[g] = s
                    gatestack.pop()
                    continue

                if g.operand == 'AND':
                    assert len(g.children) >= 2
                    s = '(and '+(' '.join(translated_children))+')'
                elif g.operand == 'OR':
                    assert len(g.children) >= 2
                    s = '(or '+(' '.join(translated_children))+')'
                elif g.operand == 'IFF':
                    assert len(g.children) == 2
                    s = '(iff '+(' '.join(translated_children))+')'
                elif g.operand == 'NOT':
                    assert len(g.children) == 1
                    s = '(not '+(' '.join(translated_children))+')'
                elif g.operand == 'ITE':
                    assert len(g.children) == 3
                    if g.is_boolean():
                        s = '(if_then_else '+\
                            (' '.join(translated_children))+')'
                    elif(g.is_int() or g.is_rat() or
                         g.is_sint() or g.is_uint()):
                        s = '(ite '+(' '.join(translated_children))+')'
                    else:
                        _nyi("ite on type '"+str(g.get_type())+"'")
                # Bitwise operations
                elif g.operand == 'BVOR':
                    assert len(g.children) == 2
                    s = '(bvor '+(' '.join(translated_children))+')'
                elif g.operand == 'BVAND':
                    assert len(g.children) == 2
                    s = '(bvand '+(' '.join(translated_children))+')'
                elif g.operand == 'BVXOR':
                    assert len(g.children) == 2
                    _nyi("bitwise xor")
                elif g.operand == 'BVNOT':
                    assert len(g.children) == 1
                    s = '(bnot '+(' '.join(translated_children))+')'
                    s += 'bvnot'
                # Arithmetic operations
                elif g.operand == 'PLUS':
                    assert len(g.children) >= 2
                    if g.is_sint() or g.is_uint():
                        s = '(bvadd '+(' '.join(translated_children))+')'
                    elif g.is_int() or g.is_rat():
                        s = '(+ '+(' '.join(translated_children))+')'
                    else:
                        _nyi("addition on type '"+str(g.get_type())+"'")
                elif g.operand == 'MINUS':
                    assert len(g.children) == 2
                    if g.is_sint() or g.is_uint():
                        s = '(bvsub '+(' '.join(translated_children))+')'
                    elif g.is_int() or g.is_rat():
                        s = '(- '+(' '.join(translated_children))+')'
                    else:
                        _nyi("subtraction on type '"+str(g.get_type())+"'")
                elif g.operand == 'NEG':
                    assert len(g.children) == 1
                    if g.is_sint() or g.is_uint():
                        s = '(bvneg '+(' '.join(translated_children))+')'
                    elif g.is_int() or g.is_rat():
                        s = '(~ '+(' '.join(translated_children))+')'
                    else:
                        _nyi("negation on type '"+str(g.get_type())+"'")
                elif g.operand == 'TIMES':
                    assert len(g.children) >= 2
                    if g.is_sint() or g.is_uint():
                        s = '(bvmul '+(' '.join(translated_children))+')'
                    elif g.is_int() or g.is_rat():
                        s = '(* '+(' '.join(translated_children))+')'
                    else:
                        _nyi("multiplication on type '"+
                             str(g.get_type())+"'")
                elif g.operand == 'EQ':
                    assert len(g.children) == 2
                    child1 = g.children[0]
                    if(child1.is_int() or child1.is_rat() or
                       child1.is_sint() or child1.is_uint()):
                        s = '(= '+(' '.join(translated_children))+')'
                    else:
                        _nyi("equality comparison on type '"+
                             str(child1.get_type())+"'")
                elif g.operand == 'LE':
                    assert len(g.children) == 2
                    child1 = g.children[0]
                    if child1.is_int() or child1.is_rat():
                        s = '(<= '+(' '.join(translated_children))+')'
                    elif child1.is_sint():
                        s = '(bvsleq '+(' '.join(translated_children))+')'
                    elif child1.is_uint():
                        s = '(bvleq '+(' '.join(translated_children))+')'
                    else:
                        _nyi("less-or-equal comparison on type '"+
                             str(child1.get_type())+"'")
                elif g.operand == 'LT':
                    assert len(g.children) == 2
                    child1 = g.children[0]
                    if child1.is_int() or child1.is_rat():
                        s = '(< '+(' '.join(translated_children))+')'
                    elif child1.is_sint():
                        s = '(bvslt '+(' '.join(translated_children))+')'
                    elif child1.is_uint():
                        s = '(bvlt '+(' '.join(translated_children))+')'
                    else:
                        _nyi("less-than comparison on type '"+
                             str(child1.get_type())+"'")
                elif isinstance(g.operand, _Operand_uif):
                    s = '('+extra_ops[g.operand]+' '+\
                        (' '.join(translated_children))+')'
                else:
                    #print g
                    _internalerror_snyi()

                if (g is self) or len(g.parents) >= 2 or g.is_constrained():
                    # This gate is a root, shared by at least two gates, or
                    # constrained ->
                    # add a definition for it
                    if g.is_boolean():
                        t = 't_'+str(g.get_index())
                        if opt_use_lets:
                            t = '$'+t
                    else:
                        t = 't_'+str(g.get_index())
                        if opt_use_lets:
                            t = '?'+t
                    cache[g] = t
                    defines.append((g,t,s))
                else:
                    cache[g] = s
                    
                gatestack.pop()

            return cache[self]



        def _to_mathsat(self,
                        cache,
                        defines,
                        variables,
                        extra_ops,
                        is_root,
                        opt_enum_translation = 'int'):
            """
            - 'extra_ops' is a dictionary into which encountered UIFs and
              enumeration data types are stored.
            - 'opt_enum_translation' describes how enumeration data types
              are translated, should be either 'int' or 'object'.
            """
            assert type(extra_ops) is dict
            assert type(is_root) is bool
            assert opt_enum_translation in ['int', 'object']

            def _nyi(s):
                raise RuntimeError("Not yet implemented in MathSAT "
                                   "translation: "+str(s)+", aborting")

            # Index and store encoundered UIFs
            if isinstance(self.operand, _Operand):
                if self.operand not in extra_ops:
                    extra_ops[self.operand] = "uif_"+str(len(extra_ops))

            # Index and store encountered enumeration data types
            if self.is_enum():
                if self.get_type() not in extra_ops:
                    extra_ops[self.get_type()] = "enum_"+str(len(extra_ops))

            tyype = self.get_type()
            param_str = ''
            if self.opparams != None:
                param_str = str(self.opparams)
            s = str(self.operand)+param_str

            #
            # First, translate constants and input gates (variables)
            #
            if not self.children:
                if self.operand == 'CONST':
                    if self.is_int():
                        return str(self.opparams)
                    elif self.is_sint():
                        width = self.get_sint_width()
                        return '0b'+str(width)+'_'+\
                               sint2bin(width, self.opparams, '')
                    elif self.is_uint():
                        width = self.get_uint_width()
                        return '0b'+str(width)+'_'+\
                               uint2bin(width, self.opparams, '')
                    elif self.is_enum():
                        lit = self.opparams
                        if opt_enum_translation == 'int':
                            return str(self.get_type().get_literal_index(lit))
                        elif opt_enum_translation == 'object':
                            return extra_ops[self.get_type()]+"_"+lit
                        raise RuntimeError("Internal error")
                    else:
                        _nyi("constants for type '"+
                             str(self.get_type())+"'")
                elif self.operand == 'FALSE':
                    return 'false'
                elif self.operand == 'TRUE':
                    return 'true'
                assert not self.is_constant()
                assert self.operand == 'VAR'
                s = 'o_'+s
                variables.append((self,s))
                return s

            if self in cache:
                return cache[self]

            #
            # Recursively translate child gates first
            #
            translated_children = []
            for child in self.children:
                translated_child = child._to_mathsat(cache,
                                                     defines,
                                                     variables,
                                                     extra_ops,
                                                     False,
                                                     opt_enum_translation)
                translated_children.append(translated_child)

            if self.operand == 'AND':
                assert len(self.children) >= 2
                s = '('+(' & '.join(translated_children))+')'
            elif self.operand == 'OR':
                assert len(self.children) >= 2
                s = '('+(' | '.join(translated_children))+')'
            elif self.operand == 'IFF':
                assert len(self.children) == 2
                s = '('+translated_children[0]+' <-> '+\
                    translated_children[1]+')'
            elif self.operand == 'NOT':
                assert len(self.children) == 1
                s = '(! '+translated_children[0]+')'
            elif self.operand == 'ITE':
                assert len(self.children) == 3
                s = 'ite('+\
                    translated_children[0]+','+\
                    translated_children[1]+','+\
                    translated_children[2]+')'

            # Bitwise operations
            elif self.operand == 'BVOR':
                assert len(self.children) == 2
                s = '('+(' | '.join(translated_children))+')'
            elif self.operand == 'BVAND':
                assert len(self.children) == 2
                s = '('+(' & '.join(translated_children))+')'
            elif self.operand == 'BVXOR':
                assert len(self.children) == 2
                s = '('+(' xor '.join(translated_children))+')'
            elif self.operand == 'BVNOT':
                assert len(self.children) == 1
                s = '(! '+translated_children[0]+')'

            # Arithmetic operations
            elif self.operand == 'PLUS':
                assert len(self.children) >= 2
                if self.is_sint() or self.is_uint():
                    s = '('+(' + '.join(translated_children))+')'
                elif self.is_int() or self.is_rat():
                    s = '('+(' + '.join(translated_children))+')'
                else:
                    #print self
                    _internalerror_snyi()
            elif self.operand == 'MINUS':
                assert len(self.children) == 2
                if self.is_sint() or self.is_uint():
                    s = '('+(' - '.join(translated_children))+')'
                elif self.is_int() or self.is_rat():
                    s = '('+(' - '.join(translated_children))+')'
                else:
                    #print self
                    _internalerror_snyi()
            elif self.operand == 'NEG':
                assert len(self.children) == 1
                if self.is_sint() or self.is_uint():
                    s = '(- '+translated_children[0]+')'
                elif self.is_int() or self.is_rat():
                    s = '(- '+translated_children[0]+')'
                else:
                    #print self
                    _internalerror_snyi()
            elif self.operand == 'TIMES':
                assert len(self.children) >= 2
                if self.is_sint() or self.is_uint():
                    s = '('+(' * '.join(translated_children))+')'
                elif self.is_int() or self.is_rat():
                    s = '('+(' * '.join(translated_children))+')'
                else:
                    #print self
                    _internalerror_snyi()

            # Comparison predicates
            elif self.operand == 'EQ':
                assert len(self.children) == 2
                child1 = self.children[0]
                if(child1.is_int() or child1.is_rat() or
                   child1.is_sint() or child1.is_uint() or
                   child1.is_enum()):
                   s = '('+translated_children[0]+' = '+\
                        translated_children[1]+')'
                else:
                    _nyi("equality comparison for type '"+
                         str(child1.get_type())+"'")

            elif self.operand == 'LE':
                assert len(self.children) == 2
                child1 = self.children[0]
                if child1.is_int() or child1.is_rat():
                    s = '('+translated_children[0]+' <= '+\
                        translated_children[1]+')'
                elif child1.is_sint():
                    _nyi("signed comparison for bitvectors")
                    s = 'bvsle('+translated_children[0]+','+\
                        translated_children[1]+')'
                elif child1.is_uint():
                    s = '('+translated_children[0]+' <= '+\
                        translated_children[1]+')'
                else:
                    _nyi("less-or-equal comparison for type '"+
                         str(child1.get_type())+"'")

            elif self.operand == 'LT':
                assert len(self.children) == 2
                child1 = self.children[0]
                if child1.is_int() or child1.is_rat():
                    s = '('+translated_children[0]+' < '+\
                        translated_children[1]+')'
                elif child1.is_sint():
                    _nyi("signed comparison for bitvectors")
                    s = 'bvslt('+translated_children[0]+','+\
                        translated_children[1]+')'
                elif child1.is_uint():
                    s = '('+translated_children[0]+' < '+\
                        translated_children[1]+')'
                else:
                    _nyi("less-than comparison for type '"+
                         str(child1.get_type())+"'")

            elif isinstance(self.operand, _Operand_uif):
                if self.is_enum():
                    # Add a definition for the gate so that it can be
                    # constrained to have a value in the domain of the type
                    is_root = True
                s = extra_ops[self.operand]+\
                    "("+(','.join(translated_children))+')'

            elif self.operand in ['READ','WRITE']:
                raise RuntimeError("Arrays are not supported in the MathSAT "
                                   "translation, aborting")

            else:
                #print self
                _nyi("operand '"+str(self.operand)+"'")

            if is_root or len(self.parents) >= 2:
                # This gate is a root or shared by at least two gates,
                # add a definition for it
                definition_name = 't_'+str(self.get_index())
                cache[self] = definition_name
                defines.append((self,definition_name,s))
                return definition_name

            return s



        def _to_yices(self, cache, defines, variables, extra_ops, is_root):
            """
            - 'extra_ops' is a dictionary into which encountered UIFs and
              enumeration data types are stored.
            """
            assert type(extra_ops) is dict
            assert type(is_root) is bool

            def _nyi(s):
                raise RuntimeError("Not yet implemented in the Yices "
                                   "translation: "+str(s)+", aborting")

            tyype = self.get_type()
            param_str = ''
            if self.opparams != None:
                param_str = str(self.opparams)
            s = str(self.operand)+param_str

            # Index and store encoundered UIFs
            if isinstance(self.operand, _Operand):
                if self.operand not in extra_ops:
                    extra_ops[self.operand] = "uif_"+str(len(extra_ops))
            # Index and store encountered enumeration data types
            if self.get_type().is_enum():
                if self.get_type() not in extra_ops:
                    extra_ops[self.get_type()] = "enum_"+str(len(extra_ops))

            if not self.children:
                if self.operand == 'CONST':
                    if self.is_int():
                        return str(self.opparams)
                    elif self.is_sint():
                        width = self.get_sint_width()
                        if self.opparams >= 0:
                            s = "(mk-bv "+str(width)+" "+str(self.opparams)+")"
                        else:
                            s = "(bv-neg (mk-bv "+str(width)+" "+\
                                str(-self.opparams)+"))"
                        return s
                    elif self.is_uint():
                        width = self.get_uint_width()
                        s = "(mk-bv "+str(width)+" "+str(self.opparams)+")"
                        return s
                    elif self.is_enum():
                        return extra_ops[self.get_type()]+"_"+self.opparams
                    else:
                        _internalerror_snyi()
                elif self.operand == 'FALSE':
                    return 'false'
                elif self.operand == 'TRUE':
                    return 'true'
                assert not self.is_constant()
                assert self.operand == 'VAR'
                s = 'o_'+s
                variables.append((self,s))
                return s
            
            if self in cache:
                return cache[self]

            #
            # Recursively translate child gates first
            #
            translated_children = [child._to_yices(cache,
                                                   defines,
                                                   variables,
                                                   extra_ops,
                                                   False)
                                   for child in self.children]

            # Logical connectives
            if self.operand == 'AND':
                assert len(self.children) >= 2
                s = '(and '+(' '.join(translated_children))+')'
            elif self.operand == 'OR':
                assert len(self.children) >= 2
                s = '(or '+(' '.join(translated_children))+')'
            elif self.operand == 'IFF':
                assert len(self.children) == 2
                s = '(= '+(' '.join(translated_children))+')'
            elif self.operand == 'NOT':
                assert len(self.children) == 1
                s = '(not '+(' '.join(translated_children))+')'
            elif self.operand == 'ITE':
                assert len(self.children) == 3
                s = '(ite '+(' '.join(translated_children))+')'
            # Bitwise operations
            elif self.operand == 'BVOR':
                assert len(self.children) == 2
                s = '(bv-or '+(' '.join(translated_children))+')'
            elif self.operand == 'BVAND':
                assert len(self.children) == 2
                s = '(bv-and '+(' '.join(translated_children))+')'
            elif self.operand == 'BVXOR':
                assert len(self.children) == 2
                s = '(bv-xor '+(' '.join(translated_children))+')'
            elif self.operand == 'BVNOT':
                assert len(self.children) == 1
                s = '(bv-not '+(' '.join(translated_children))+')'
            # Arithmetic operations
            elif self.operand == 'PLUS':
                assert len(self.children) == 2
                if self.is_sint() or self.is_uint():
                    s = '(bv-add '+(' '.join(translated_children))+')'
                elif self.is_int() or self.is_rat():
                    s = '(+ '+(' '.join(translated_children))+')'
                else:
                    _nyi("addition for type '"+str(self.get_type())+"'")
            elif self.operand == 'MINUS':
                assert len(self.children) == 2
                if self.is_sint() or self.is_uint():
                    s = '(bv-sub '+(' '.join(translated_children))+')'
                elif self.is_int() or self.is_rat():
                    s = '(- '+(' '.join(translated_children))+')'
                else:
                    _nyi("subtraction for type '"+str(self.get_type())+"'")
            elif self.operand == 'NEG':
                assert len(self.children) == 1
                if self.is_sint() or self.is_uint():
                    s = '(bv-neg '+(' '.join(translated_children))+')'
                elif self.is_int() or self.is_rat():
                    s = '(- '+(' '.join(translated_children))+')'
                else:
                    _nyi("negation for type '"+str(self.get_type())+"'")
            elif self.operand == 'TIMES':
                assert len(self.children) >= 2
                if self.is_sint() or self.is_uint():
                    s = '(bv-mul '+(' '.join(translated_children))+')'
                elif self.is_int() or self.is_rat():
                    s = '(* '+(' '.join(translated_children))+')'
                else:
                    _nyi("multiplication for type '"+str(self.get_type())+"'")
            # Comparison predicates
            elif self.operand == 'EQ':
                assert len(self.children) == 2
                child1 = self.children[0]
                if(child1.is_int() or child1.is_rat() or
                   child1.is_sint() or child1.is_uint() or
                   child1.is_enum()):
                    s = '(= '+(' '.join(translated_children))+')'
                else:
                    _nyi("equality comparison for type '"+
                         str(child1.get_type())+"'")
            elif self.operand == 'LE':
                assert len(self.children) == 2
                child1 = self.children[0]
                if child1.is_int() or child1.is_rat():
                    s = '(<= '+(' '.join(translated_children))+')'
                elif child1.is_sint():
                    s = '(bv-sle '+(' '.join(translated_children))+')'
                elif child1.is_uint():
                    s = '(bv-le '+(' '.join(translated_children))+')'
                else:
                    _nyi("less-or-equal comparison for type '"+
                         str(child1.get_type())+"'")
            elif self.operand == 'LT':
                assert len(self.children) == 2
                child1 = self.children[0]
                if child1.is_int() or child1.is_rat():
                    s = '(< '+(' '.join(translated_children))+')'
                elif child1.is_sint():
                    s = '(bv-slt '+(' '.join(translated_children))+')'
                elif child1.is_uint():
                    s = '(bv-lt '+(' '.join(translated_children))+')'
                else:
                    _nyi("less-than comparison for type '"+
                         str(child1.get_type())+"'")

            # Array operations
            elif self.operand == 'READ':
                assert len(self.children) == 2
                s = '('+(' '.join(translated_children))+')'
            elif self.operand == 'WRITE':
                assert len(self.children) == 3
                s = '(update '+translated_children[0]+' '+\
                    '('+translated_children[1]+')'+\
                    translated_children[2]+')'
                is_root = True

            elif isinstance(self.operand, _Operand_uif):
                s = '('+extra_ops[self.operand]+' '+\
                    (' '.join(translated_children))+')'
            else:
                _nyi("operand '"+str(self.operand)+"'")

            if is_root or len(self.parents) >= 2:
                # This gate is a root or shared by at least two gates,
                # add a definition for it
                definition_name = 't_'+str(self.get_index())
                cache[self] = definition_name
                defines.append((self,definition_name,s))
                return definition_name

            return s


        def _to_z3(self, cache, defines, variables, extra_ops, consts,
                   is_root):
            """
            - 'extra_ops' is a dictionary into which encountered UIFs and
              enumeration data types are stored.
            """
            assert type(cache) is dict
            assert type(extra_ops) is dict
            assert type(consts) is dict
            assert type(is_root) is bool

            def _nyi(s):
                raise RuntimeError("Not yet implemented in the Z3 "
                                   "translation: "+str(s)+", aborting")

            tyype = self.get_type()
            param_str = ''
            if self.opparams != None:
                param_str = str(self.opparams)
            s = str(self.operand)+param_str

            # Index and store encoundered UIFs
            if isinstance(self.operand, _Operand):
                if self.operand not in extra_ops:
                    extra_ops[self.operand] = "uif_"+str(len(extra_ops))
            # Index and store encountered enumeration data types
            if self.get_type().is_enum():
                _nyi("enumeration data types")
                if self.get_type() not in extra_ops:
                    extra_ops[self.get_type()] = "enum_"+str(len(extra_ops))

            if not self.children:
                if self.operand == 'CONST':
                    if self.is_int():
                        consts["intc_"+str(self.opparams)] = self.opparams
                        return str(self.opparams)
                    elif self.is_sint():
                        width = self.get_sint_width()
                        s = "sbvc_"+sint2bin(width, self.opparams)
                        consts[s] = (str(self.opparams), "sbv"+str(width))
                        consts["T sbv"+str(width)]="bv ["+str(width)+"]"
                        return s
                    elif self.is_uint():
                        width = self.get_uint_width()
                        s = "ubvc_"+uint2bin(width, self.opparams)
                        consts[s] = (str(self.opparams), "ubv"+str(width))
                        consts["T ubv"+str(width)]="bv ["+str(width)+"]"
                        return s
                    elif self.is_enum():
                        _nyi("enumeration data types")
                        return extra_ops[self.get_type()]+"_"+self.opparams
                    else:
                        _internalerror_snyi()
                elif self.operand == 'FALSE':
                    return 'false'
                elif self.operand == 'TRUE':
                    return 'true'
                assert not self.is_constant()
                assert self.operand == 'VAR'
                s = 'o_'+s
                variables.append((self,s))
                return s
            
            if self in cache:
                return cache[self]

            #KIIK
            #
            # Recursively translate child gates first
            #
            translated_children = [child._to_z3(cache,
                                                defines,
                                                variables,
                                                extra_ops,
                                                consts,
                                                False)
                                   for child in self.children]

            # Logical connectives
            if self.operand == 'AND':
                assert len(self.children) >= 2
                s = 'and '+(' '.join(translated_children))
            elif self.operand == 'OR':
                assert len(self.children) >= 2
                s = 'or '+(' '.join(translated_children))
            elif self.operand == 'IFF':
                assert len(self.children) == 2
                s = 'iff '+(' '.join(translated_children))
            elif self.operand == 'NOT':
                assert len(self.children) == 1
                s = 'not '+(' '.join(translated_children))
            elif self.operand == 'ITE':
                assert len(self.children) == 3
                s = 'ite '+(' '.join(translated_children))
            # Bitwise operations
            elif self.operand == 'BVOR':
                assert len(self.children) == 2
                s = 'bvor '+(' '.join(translated_children))
            elif self.operand == 'BVAND':
                assert len(self.children) == 2
                s = 'bvand '+(' '.join(translated_children))
            elif self.operand == 'BVXOR':
                assert len(self.children) == 2
                s = 'bvxor '+(' '.join(translated_children))
            elif self.operand == 'BVNOT':
                assert len(self.children) == 1
                s = 'bvnot '+(' '.join(translated_children))
            # Arithmetic operations
            elif self.operand == 'PLUS':
                assert len(self.children) == 2
                if self.is_sint() or self.is_uint():
                    s = 'bvadd '+(' '.join(translated_children))
                elif self.is_int() or self.is_rat():
                    s = '+ '+(' '.join(translated_children))
                else:
                    _nyi("addition for type '"+str(self.get_type())+"'")
            elif self.operand == 'MINUS':
                assert len(self.children) == 2
                if self.is_sint() or self.is_uint():
                    s = 'bvsub '+(' '.join(translated_children))
                elif self.is_int() or self.is_rat():
                    s = '- '+(' '.join(translated_children))
                else:
                    _nyi("subtraction for type '"+str(self.get_type())+"'")
            elif self.operand == 'NEG':
                assert len(self.children) == 1
                if self.is_sint() or self.is_uint():
                    s = 'bvneg '+(' '.join(translated_children))
                elif self.is_int() or self.is_rat():
                    s = '~ '+(' '.join(translated_children))
                else:
                    _nyi("negation for type '"+str(self.get_type())+"'")
            elif self.operand == 'TIMES':
                assert len(self.children) >= 2
                if self.is_sint() or self.is_uint():
                    s = 'bvmul '+(' '.join(translated_children))
                elif self.is_int() or self.is_rat():
                    s = '* '+(' '.join(translated_children))
                else:
                    _nyi("multiplication for type '"+str(self.get_type())+"'")
            # Comparison predicates
            elif self.operand == 'EQ':
                assert len(self.children) == 2
                child1 = self.children[0]
                if(child1.is_int() or child1.is_rat() or
                   child1.is_sint() or child1.is_uint() or
                   child1.is_enum()):
                    s = '= '+(' '.join(translated_children))
                else:
                    _nyi("equality comparison for type '"+
                         str(child1.get_type())+"'")
            elif self.operand == 'LE':
                assert len(self.children) == 2
                child1 = self.children[0]
                if child1.is_int() or child1.is_rat():
                    s = '<= '+(' '.join(translated_children))
                elif child1.is_sint():
                    s = 'bvsle '+(' '.join(translated_children))
                elif child1.is_uint():
                    s = 'bvule '+(' '.join(translated_children))
                else:
                    _nyi("less-or-equal comparison for type '"+
                         str(child1.get_type())+"'")
            elif self.operand == 'LT':
                assert len(self.children) == 2
                child1 = self.children[0]
                if child1.is_int() or child1.is_rat():
                    s = '< '+(' '.join(translated_children))
                elif child1.is_sint():
                    s = 'bvslt '+(' '.join(translated_children))
                elif child1.is_uint():
                    s = 'bvult '+(' '.join(translated_children))
                else:
                    _nyi("less-than comparison for type '"+
                         str(child1.get_type())+"'")

            # Array operations
            elif self.operand == 'READ':
                assert len(self.children) == 2
                s = 'select '+(' '.join(translated_children))
            elif self.operand == 'WRITE':
                assert len(self.children) == 3
                s = 'store '+(' '.join(translated_children))

            elif isinstance(self.operand, _Operand_uif):
                s = ' '+extra_ops[self.operand]+' '+\
                    (' '.join(translated_children))
            else:
                _nyi("operand '"+str(self.operand)+"'")

            astid = 't'+str(self.get_index())
            if isinstance(self.operand, _Operand_uif):
                s = "Fun "+astid+" "+s
            else:
                s = "App "+astid+" "+s

            defines.append((astid,s))
            cache[self] = astid
            return astid



        def _to_stp(self, cache, defines, variables, is_root):
            assert type(is_root) is bool

            def _nyi(s):
                raise RuntimeError("Not yet implemented in the STP "
                                   "translation: "+str(s)+", aborting")

            def _stp_type_error():
                raise RuntimeError("Types other than bit vectors are "
                                   "not supported by the STP translation")

            tyype = self.get_type()
            param_str = ''
            if self.opparams != None:
                param_str = str(self.opparams)
            s = str(self.operand)+param_str
            if not self.children:
                if self.operand == 'CONST':
                    if self.is_sint():
                        width = self.get_sint_width()
                        return sint2bin(width, self.opparams, '0bin')
                    elif self.is_uint():
                        width = self.get_uint_width()
                        return uint2bin(width, self.opparams, '0bin')
                    else:
                        _stp_type_error()
                elif self.operand == 'FALSE':
                    return 'FALSE'
                elif self.operand == 'TRUE':
                    return 'TRUE'
                assert not self.is_constant()
                assert self.operand == 'VAR'
                s = 'o_'+s
                variables.append((self,s))
                return s

            if self in cache:
                return cache[self]

            translated_children = []
            for child in self.children:
                translated_children.append(child._to_stp(cache,defines,
                                                         variables,False))

            if self.operand == 'AND':
                assert len(self.children) >= 2
                s = '('+(' AND '.join(translated_children))+')'
            elif self.operand == 'OR':
                assert len(self.children) >= 2
                s = '('+(' OR '.join(translated_children))+')'
            elif self.operand == 'IFF':
                assert len(self.children) == 2
                s = '('+translated_children[0]+' <=> '+\
                    translated_children[1]+')'
            elif self.operand == 'NOT':
                assert len(self.children) == 1
                s = '(NOT '+translated_children[0]+')'
            elif self.operand == 'ITE':
                assert len(self.children) == 3
                s = '(IF '+translated_children[0]+\
                    ' THEN '+translated_children[1]+\
                    ' ELSE '+translated_children[2]+' ENDIF)'
            # Bitwise operations
            elif self.operand == 'BVOR':
                assert len(self.children) == 2
                s = '('+(' | '.join(translated_children))+')'
            elif self.operand == 'BVAND':
                assert len(self.children) == 2
                s = '('+(' & '.join(translated_children))+')'
            elif self.operand == 'BVXOR':
                assert len(self.children) == 2
                s = 'BVXOR('+translated_children[0]+' '+\
                     translated_children[1]+')'
            elif self.operand == 'BVNOT':
                assert len(self.children) == 1
                s = '(~'+translated_children[0]+')'
            # Arithmetic operations
            elif self.operand == 'PLUS':
                assert len(self.children) >= 2
                if self.is_sint():
                    width = self.get_sint_width()
                    s = 'BVPLUS('+str(width)+','+\
                        (','.join(translated_children))+')'
                elif self.is_uint():
                    width = self.get_uint_width()
                    s = 'BVPLUS('+str(width)+','+\
                        (','.join(translated_children))+')'
                else:
                    _stp_type_error()
            elif self.operand == 'MINUS':
                assert len(self.children) == 2
                if self.is_sint():
                    width = self.get_sint_width()
                    s = 'BVSUB('+str(width)+','+\
                        (','.join(translated_children))+')'
                elif self.is_uint():
                    width = self.get_uint_width()
                    s = 'BVSUB('+str(width)+','+\
                        (','.join(translated_children))+')'
                else:
                    _stp_type_error()
            elif self.operand == 'NEG':
                assert len(self.children) == 1
                if self.is_sint() or self.is_uint():
                    s = 'BVUMINUS('+translated_children[0]+')'
                else:
                    _stp_type_error()
            elif self.operand == 'TIMES':
                assert len(self.children) >= 2
                if self.is_sint():
                    width = self.get_sint_width()
                    s = 'BVMULT('+str(width)+','+\
                        (','.join(translated_children))+')'
                elif self.is_uint():
                    width = self.get_uint_width()
                    s = 'BVMULT('+str(width)+','+\
                        (','.join(translated_children))+')'
                else:
                    _stp_type_error()

            # Comparison predicates
            elif self.operand == 'EQ':
                assert len(self.children) == 2
                child1 = self.children[0]
                if child1.is_sint() or child1.is_uint():
                    s = '('+translated_children[0]+' = '+\
                        translated_children[1]+')'
                else:
                    _stp_type_error()
            elif self.operand == 'LE':
                assert len(self.children) == 2
                child1 = self.children[0]
                if child1.is_sint():
                    s = 'SBVLE('+(','.join(translated_children))+')'
                elif child1.is_uint():
                    s = 'BVLE('+(','.join(translated_children))+')'
                else:
                    _stp_type_error()
            elif self.operand == 'LT':
                assert len(self.children) == 2
                child1 = self.children[0]
                if child1.is_sint():
                    s = 'SBVLT('+(','.join(translated_children))+')'
                elif child1.is_uint():
                    s = 'BVLT('+(','.join(translated_children))+')'
                else:
                    _stp_type_error()

            # Array operations
            elif self.operand == 'READ':
                assert len(self.children) == 2
                s = '('+translated_children[0]+'['+\
                        translated_children[1]+'])'
            elif self.operand == 'WRITE':
                assert len(self.children) == 3
                s = '('+translated_children[0]+' '+\
                    'WITH ['+translated_children[1]+'] '+\
                    ':= '+translated_children[2]+')'
                is_root = True

            # Uninterpreted functions
            elif isinstance(self.operand, _Operand_uif):
                raise RuntimeError("Uninterpreted functions are not "+
                                   "supported in the STP translation")

            else:
                #print self
                _nyi("operand '"+str(self.operand)+"'")

            if is_root or len(self.parents) >= 2:
                # This gate is a root or shared by at least two gates,
                # add a definition for it
                if self.is_boolean():
                    t = 't_'+str(self.get_index())
                else:
                    t = 't_'+str(self.get_index())
                cache[self] = t
                defines.append((self,t,s))
                return t

            return s




        def evaluate(self, solution):
            """
            Evaluate the value of the gate in the partial solution and
            extend the solution accordingly.
            Assumes that all input gates already have values.
            """

            def _nyi(s):
                raise RuntimeError("Internal error. Not yet implemented: "+
                                   str(s)+", aborting")

            # Already done?
            if solution.has_key(self):
                return solution[self]
            # No, evaluate children first and collect their values in a list
            child_values = []
            for child in self.children:
                child_values.append(child.evaluate(solution))
            # Evaluate the gate
            if self.operand == 'ITE' and self.is_boolean():
                assert len(child_values) == 3
                (if_value, then_value, else_value) = child_values
                assert type(if_value) is bool
                assert type(then_value) is bool
                assert type(else_value) is bool
                value = ((if_value and then_value) or
                         ((not if_value) and else_value))
                solution[self] = value
                return value
            if(self.operand == 'ITE' and
               (self.is_int() or self.is_sint() or self.is_uint())):
                assert len(child_values) == 3
                (if_value, then_value, else_value) = child_values
                assert type(if_value) is bool
                assert type(then_value) is long
                assert type(else_value) is long
                if if_value:
                    value = then_value
                else:
                    value = else_value
                solution[self] = value
                return value
            elif self.operand == 'ITE' and self.is_enum():
                assert len(child_values) == 3
                (if_value, then_value, else_value) = child_values
                assert type(if_value) is bool
                assert type(then_value) is str
                assert type(else_value) is str
                if if_value:
                    value = then_value
                else:
                    value = else_value
                solution[self] = value
                return value
            elif self.operand == 'ITE' and self.is_array():
                assert len(child_values) == 3
                (if_value, then_value, else_value) = child_values
                assert type(if_value) is bool
                assert type(then_value) is dict
                assert type(else_value) is dict
                if if_value:
                    value = then_value
                else:
                    value = else_value
                solution[self] = value
                return value
            elif self.operand == 'OR':
                value = False
                for child_value in child_values:
                    assert type(child_value) is bool
                    value = value or child_value
                solution[self] = value
                return value
            elif self.operand == 'AND':
                value = True
                for child_value in child_values:
                    assert type(child_value) is bool
                    value = value and child_value
                solution[self] = value
                return value
            elif self.operand == 'IFF':
                assert len(child_values) == 2
                (child_value1, child_value2) = child_values
                assert type(child_value1) is bool
                assert type(child_value2) is bool
                value = (child_value1 == child_value2)
                solution[self] = value
                return value
            elif self.operand == 'NOT':
                assert len(child_values) == 1
                child_value = child_values[0]
                assert type(child_value) is bool
                value = not child_value
                solution[self] = value
                return value
            elif self.operand == 'BVOR':
                assert self.is_sint() or self.is_uint()
                assert len(child_values) >= 2
                value =  child_values[0]
                assert type(value) is long
                for child_value in child_values[1:]:
                    assert type(child_value) is long
                    if self.is_sint():
                        value = _sint_or(self.get_sint_width(),
                                         value, child_value)
                    elif self.is_uint():
                        value = _uint_or(self.get_uint_width(),
                                         value, child_value)
                    else:
                        _internalerror_snyi()
                solution[self] = value
                return value
            elif self.operand == 'BVAND':
                assert self.is_sint() or self.is_uint()
                assert len(child_values) >= 2
                value =  child_values[0]
                assert type(value) is long
                for child_value in child_values[1:]:
                    assert type(child_value) is long
                    if self.is_sint():
                        value = _sint_and(self.get_sint_width(),
                                          value, child_value)
                    elif self.is_uint():
                        value = _uint_and(self.get_uint_width(),
                                          value, child_value)
                    else:
                        _internalerror_snyi()
                solution[self] = value
                return value
            elif self.operand == 'BVXOR':
                assert self.is_sint() or self.is_uint()
                assert len(child_values) >= 2
                value =  child_values[0]
                assert type(value) is long
                for child_value in child_values[1:]:
                    assert type(child_value) is long
                    if self.is_sint():
                        value = _sint_xor(self.get_sint_width(),
                                          value, child_value)
                    elif self.is_uint():
                        value = _uint_xor(self.get_uint_width(),
                                          value, child_value)
                    else:
                        _internalerror_snyi()
                solution[self] = value
                return value
            elif self.operand == 'BVNOT':
                assert self.is_sint() or self.is_uint()
                assert len(child_values) == 1
                value =  child_values[0]
                assert type(value) is long
                if self.is_sint():
                    value = _sint_not(self.get_sint_width(), value)
                elif self.is_uint():
                    value = _uint_not(self.get_uint_width(), value)
                else:
                    _internalerror_snyi()
                solution[self] = value
                return value
            elif self.operand == 'EQ':
                assert len(child_values) == 2
                (child_value1, child_value2) = child_values
                value = (child_value1 == child_value2)
                solution[self] = value
                return value
            elif self.operand == 'LE':
                assert len(child_values) == 2
                (child_value1, child_value2) = child_values
                value = (child_value1 <= child_value2)
                solution[self] = value
                return value
            elif self.operand == 'LT':
                assert len(child_values) == 2
                (child_value1, child_value2) = child_values
                value = (child_value1 < child_value2)
                solution[self] = value
                return value
            elif self.operand == 'PLUS':
                if self.is_int() or self.is_sint() or self.is_uint():
                    assert len(child_values) >= 2
                    value = long(0)
                    for child_value in child_values:
                        assert type(child_value) is long
                        if self.is_sint():
                            value = _sint_plus(self.get_sint_width(),
                                               value, child_value)
                        elif self.is_uint():
                            value = _uint_plus(self.get_uint_width(),
                                               value, child_value)
                        elif self.is_int():
                            value = value + child_value
                        else:
                            assert False
                else:
                    _nyi("addition on type '"+str(self.get_type())+"'")
                solution[self] = value
                return value
            elif self.operand == 'MINUS':
                if self.is_int() or self.is_sint() or self.is_uint():
                    assert len(child_values) == 2
                    (child_value1, child_value2) = child_values
                    if self.is_sint():
                        value = _sint_minus(self.get_sint_width(),
                                            child_value1, child_value2)
                    elif self.is_uint():
                        value = _uint_minus(self.get_uint_width(),
                                            child_value1, child_value2)
                    elif self.is_int():
                        assert type(child_value1) is long
                        assert type(child_value2) is long
                        value = child_value1 - child_value2
                    else:
                        assert False
                else:
                    _nyi("subtraction on type '"+str(self.get_type())+"'")
                solution[self] = value
                return value
            elif self.operand == 'NEG':
                assert len(child_values) >= 1
                child_value = child_values[0]
                if self.is_sint():
                    value = _sint_neg(self.get_sint_width(), child_value)
                elif self.is_uint():
                    value = _uint_neg(self.get_uint_width(), child_value)
                elif self.is_int():
                    assert type(child_value) is long
                    value = -child_value;
                else:
                    _nyi("negation on type '"+str(self.get_type())+"'")
                solution[self] = value
                return value
            elif self.operand == 'TIMES':
                if self.is_int() or self.is_sint() or self.is_uint():
                    assert len(child_values) >= 2
                    value = long(1)
                    for child_value in child_values:
                        assert type(child_value) is long
                        if self.is_int():
                            value = value * child_value
                        elif self.is_sint():
                            value = _sint_times(self.get_sint_width(),
                                                value, child_value)
                        elif self.is_uint():
                            value = _uint_times(self.get_uint_width(),
                                                value, child_value)
                        else:
                            assert False
                else:
                    _nyi("multiplication on type '"+str(self.get_type())+"'")
                solution[self] = value
                return value
            #
            # Operations on arrays
            #
            elif self.operand == 'READ':
                assert len(child_values) == 2
                if child_values[1] in child_values[0]:
                    value = child_values[0][child_values[1]]
                else:
                    value = self.get_type().get_range_type().get_a_value()
                solution[self] = value
                return value
            elif self.operand == 'WRITE':
                assert len(child_values) == 3
                assert type(child_values[0]) is dict
                value = child_values[0].copy()
                value[child_values[1]] = child_values[2]
                solution[self] = value
                return value
            #
            # Uninterpreted functions
            #
            elif isinstance(self.operand, _Operand_uif):
                assert len(child_values) >= 1
                key = (self.operand, tuple(child_values))
                if key in solution:
                    value = solution[key]
                else:
                    value = self.get_type().get_a_value()
                    solution[key] = value
                value = solution[key]
                solution[self] = value
                return value
            else:
                print self
                _internalerror_snyi()

    def _make_handle(self, gate):
        assert isinstance(gate, Circuit._RealGate)
        handle = Gate()
        handle._gate = gate
        gate._handles[handle] = 1
        return handle
    
    def _install_gate(self, gate):
        """
        The only way to really add gates to the circuit.
        """
        assert isinstance(gate, Circuit._RealGate)
        # Sort children of commutative gate operands
        gate.sort_children()
        # Insert into gate list if not already there
        structural_id = gate.get_structural_id()
        if self._gate_lookup.has_key(structural_id):
            return self._make_handle(self._gates[self._gate_lookup[structural_id].get_index()])

        self._invalidate_solution()
        if self.free_gate_indices:
            gate_index = self.free_gate_indices.pop()
            assert self._gates[gate_index] is None
        else:
            gate_index = len(self._gates)
            self._gates.append(None)
        self._gates[gate_index] = gate
        gate.set_index(gate_index)
        self._gate_lookup[structural_id] = gate
        for (index,child) in enumerate(gate.children):
            child.parents.add((gate_index,index))
        self.nof_gates += 1
        return self._make_handle(gate)


    def _garbage_collect(self, gate):
        assert isinstance(gate, Circuit._RealGate)
        if((not gate.has_handles()) and (not gate.is_constrained()) and
           (not gate.parents)):
            #print "GC:",gate
            gate_index = gate.get_index()
            structural_id = gate.get_structural_id()
            del self._gate_lookup[structural_id]
            for (index,child) in enumerate(gate.children):
                child.parents.remove((gate_index, index))
                self._garbage_collect(child)
            self._gates[gate_index] = None
            self.free_gate_indices.append(gate_index)
            self.nof_gates -= 1


    def _unique(self, l):
        """
        A small helper that removes duplicate elements from the list l
        with hashable elements.
        """
        u = {}
        for e in l:
            u[e] = 1
        return u.keys()


    def _add_binary_comparison(self, op, child1, child2):
        """
        Add and return a new comparison gate.
        The types of the two children should match and be either
        boolean (equality comparison only), integer, rational,
        or bounded integers of same width.
        """
        assert op in ['LT','LE','EQ']
        # Get the children behind the handles
        child1 = self._get_real_gate(child1)
        child2 = self._get_real_gate(child2)
        # Check that the types are the same
        if child1.get_type() != child2.get_type():
            _type_error("Comparison of gates with incompatible types")
        # Type check
        if op == 'EQ' and child1.is_boolean():
            return self.add_iff(child1, child2)
        if op == 'EQ' and child1.is_enum():
            pass
        elif (child1.is_int() or child1.is_rat() or
              child1.is_sint() or child1.is_uint()):
            pass
        else:
            _type_error("Cannot compare gates of type "+
                        "'"+str(child1.get_type())+"'")
        return self._add_gate(get_bool(), op, None, [child1,child2])


    def _invalidate_solution(self):
        self.solution = None


    def _deduce_smtlib_logic_class(self, gate_list):
        """
        Return the SMT-LIB logic category of the circuit.
        This is a very inaccurate overapproximation.
        In addition, it is possible to define circuits that do not match any
        current SMT-LIB logic.
        """
        def _cantdeduce():
            raise RuntimeError("Cannot deduce the SMT-LIB logic category "
                               "of the circuit.")

        logic = "QF_UF"
        appearing_types = sets.Set()
        has_uifs = False
        for gate in gate_list:
            if gate is None:
                continue
            if not gate.is_boolean():
                appearing_types.add(gate.get_type())
            if isinstance(gate.operand, _Operand):
                has_uifs = True
        if not appearing_types:
            return "QF_UF"
        bounded_integer_types = []
        remaining_types = []
        for t in appearing_types:
            if t.is_sint() or t.is_uint():
                bounded_integer_types.append(t)
            else:
                remaining_types.append(t)
        if bounded_integer_types:
            if remaining_types:
                _cantdeduce()
            if has_uifs:
                return "QF_UFBV"
            return "QF_BV"
        if remaining_types == [get_int()]:
            if has_uifs:
                return "QF_UFLIA"
            return "QF_LIA"
        if remaining_types == [get_rat()]:
            if has_uifs:
                return "QF_UFLRA"
            return "QF_LRA"
        _cantdeduce()
        return ""


    def _to_yices(self, f, produce_model = True):
        """
        Output the circuit in Yices format to the file object 'f'.
        Returns maps between input gates and Yices format var names.
        """
        assert type(produce_model) is bool
        
        # Print a comment
        f.write("; A Yices problem generated by PySMT\n")

        # Find the constrained Boolean gates
        constrained_gates = [g for g in self._gates if
                             (g and g.is_constrained())]

        cache = {} #sets.Set()
        defines = []
        variables = []
        map_op_to_opname = {}
        trans = []
        for g in constrained_gates:
            trans.append((g, g._to_yices(cache, defines, variables,
                                         map_op_to_opname, True)))
        #print cache
        #print defines
        #print variables
        #print trans
        map_opname_to_op = {}
        for (op, name) in map_op_to_opname.items():
            map_opname_to_op[name] = op

        def _t2t(t):
            """Convert a PySMT type to the corresponding Yices type."""
            if t.is_bool():
                return "bool"
            elif t.is_int():
                return "int"
            elif t.is_rat():
                return "real"
            elif t.is_sint() or t.is_uint():
                return "(bitvector "+str(t.get_width())+")"
            elif t.is_enum():
                return map_op_to_opname[t]
            elif t.is_array():
                return "(-> "+_t2t(t.get_index_type())+" "+\
                       _t2t(t.get_range_type())+")"
            _internalerror_snyi()

        # Define the encountered enumeration data types
        for (opname,op) in map_opname_to_op.items():
            if isinstance(op, _Type_enum):
                f.write("(define-type "+opname+" (scalar ")
                f.write(" ".join([opname+"_"+str(l) for l in op.get_domain()]))
                f.write("))\n")

        # Define the encountered UIFs
        for (opname,op) in map_opname_to_op.items():
            if isinstance(op, _Operand_uif):
                f.write("(define "+opname+"::(-> ")
                for dtype in op.get_domain():
                    f.write(_t2t(dtype)+" ")
                f.write(_t2t(op.get_range_type()))
                f.write("))\n")
            elif isinstance(op, _Type_enum):
                pass
            else:
                _internalerror_snyi()

        # Output variable names and types
        map_var_to_smt_name = {}
        map_smt_name_to_var = {}
        for (gate,smt_name) in self._unique(variables):
            map_var_to_smt_name[gate] = smt_name
            map_smt_name_to_var[smt_name] = gate
            f.write("(define "+smt_name+"::"+_t2t(gate.get_type())+")\n")

        if True:
            # Add defines
            for (g,t,d) in defines:
                f.write("(define "+t+"::"+_t2t(g.get_type())+")\n")
                f.write("(assert (= "+t+" "+d+"))\n")
            # Add the conjunction of the constrained gates
            for (g,t) in trans:
                assert g.is_boolean()
                assert g.is_constrained()
                if g.constrained_value == False:
                    f.write('(assert (not '+t+'))\n')
                elif g.constrained_value == True:
                    f.write('(assert '+t+')\n')
                else:
                    assert False
        else:
            #
            # Output the formula
            #
            f.write("(assert ")
            # Add defines, i.e. the let and flet constructions
            for (g,t,d) in defines:
                f.write('(let (('+t+' '+d+')) ')
            # Add the conjunction of the constrained gates
            for (g,t) in trans:
                assert g.is_boolean()
                assert g.is_constrained()
                if g.constrained_value == False:
                    f.write('(and (not '+t+') ')
                elif g.constrained_value == True:
                    f.write('(and '+t+' ')
                else:
                    assert False
            f.write('true')
            # Add all the missing parentheses
            for (g,t) in trans:
                f.write(')')
            for (g,t,d) in defines:
                f.write(')')
            #f.write("\n");

            # The end
            f.write(")\n");

        #
        if produce_model:
            f.write("(set-evidence! true)\n")
        f.write("(check)\n")

        return (map_var_to_smt_name, map_smt_name_to_var, map_opname_to_op)


    def to_yices(self, f, produce_model = True):
        """
        Output the circuit in Yices format to the file object f.
        If 'produce_model' is True, then the 'set-evidence!' flag is set
        to true in the generated Yices problem file.
        """
        assert type(f) is file
        assert type(produce_model) is bool
        self._to_yices(f, produce_model)



    def _to_z3(self, f):
        """
        Output the circuit in Z3 format to the file object 'f'.
        Returns maps between input gates and Z3 format var names.
        """

        # Print a comment
        f.write("; A Z3 problem generated by PySMT\n")

        # Find the constrained Boolean gates
        constrained_gates = [g for g in self._gates if
                             (g and g.is_constrained())]
        
        cache = {} #sets.Set()
        defines = []
        variables = []
        map_op_to_opname = {}
        consts = {}
        trans = []
        for g in constrained_gates:
            trans.append((g, g._to_z3(cache, defines, variables,
                                      map_op_to_opname, consts, True)))
        #print cache
        #print defines
        #print variables
        #print trans
        map_opname_to_op = {}
        for (op, name) in map_op_to_opname.items():
            map_opname_to_op[name] = op

        def _t2t(t):
            """Convert a PySMT type to the corresponding Z3 type."""
            if t.is_bool():
                return "BOOL"
            elif t.is_int():
                return "INT"
            elif t.is_rat():
                return "REAL"
            elif t.is_sint():
                return "sbv_"+str(t.get_width())
            elif t.is_uint():
                return "ubv_"+str(t.get_width())
            elif t.is_enum():
                raise RuntimeError("Enumeration data types not supported in "
                                   "the Z3 translation")
                return map_op_to_opname[t]
            elif t.is_array():
                assert False and "Not yet implemented"
                return "(-> "+_t2t(t.get_index_type())+" "+\
                       _t2t(t.get_range_type())+")"
            _internalerror_snyi()

        # Declare types
        f.write("Solver LIA\n")
        f.write("Type INT int\n")
        f.write("Type BOOL bool\n")
        
        # Declare the encountered primitive types
        for (opname,op) in map_opname_to_op.items():
            if isinstance(op, _Type_enum):
                raise RuntimeError("Enumeration data types not supported in "
                                   "the Z3 translation") 
                f.write("(define-type "+opname+" (scalar ")
                f.write(" ".join([opname+"_"+str(l) for l in op.get_domain()]))
                f.write("))\n")
        for (name,definition) in consts.items():
            if name.startswith("T "):
                f.write("Type "+str(name[2:])+" "+definition+"\n")


        # Declare the encountered primitive constants
        for (name,definition) in consts.items():
            if name.startswith("intc_"):
                f.write("Num "+str(definition)+" "+str(definition)+" INT\n")
            elif name.startswith("ubvc_"):
                f.write("Num "+opname+" "+definition[0]+" "+definition[1]+"\n")
            elif name.startswith("sbvc_"):
                f.write("Num "+opname+" "+definition[0]+" "+definition[1]+"\n")
            elif name.startswith("T "):
                pass
            else:
                _internalerror_snyi()
            #KIIK

        # Define the encountered UIFs
        for (opname,op) in map_opname_to_op.items():
            if isinstance(op, _Operand_uif):
                f.write("Dec "+opname+" "+opname+" ")
                for dtype in op.get_domain():
                    f.write(_t2t(dtype)+" ")
                f.write(_t2t(op.get_range_type()))
                f.write("\n")
            elif isinstance(op, _Type_enum):
                pass
            else:
                _internalerror_snyi()

        # Output variable names and types
        map_var_to_smt_name = {}
        map_smt_name_to_var = {}
        for (gate,smt_name) in self._unique(variables):
            map_var_to_smt_name[gate] = smt_name
            map_smt_name_to_var[smt_name] = gate
            f.write("Const "+smt_name+" "+smt_name+" "+
                    _t2t(gate.get_type())+"\n")

        # Add defines
        for (astid,body) in defines:
            f.write(body+"\n")
            
        for (g,t) in trans:
            assert g.is_boolean()
            assert g.is_constrained()
            if g.constrained_value == False:
                f.write('App neg_'+t+' not '+t+'\n')
                f.write('Assert neg_'+t+'\n')
            elif g.constrained_value == True:
                f.write('Assert '+t+'\n')
            else:
                assert False

        f.write("Check\n")

        return (map_var_to_smt_name, map_smt_name_to_var, map_opname_to_op)


    def to_z3(self, f):
        """
        Output the circuit in Z3 native format to the file object f.
        """
        assert type(f) is file
        self._to_yices(f)



    def _to_stp(self, f, produce_model = True):
        """
        Output the circuit in STP format to the file object 'f'.
        Returns maps between input gates and STP format var names.
        """
        # Print a comment
        f.write("% An STP problem generated by PySMT\n")

        # Find the constrained Boolean gates
        constrained_gates = [g for g in self._gates if
                             (g and g.is_constrained())]

        cache = {} #sets.Set()
        defines = []
        variables = []
        trans = []
        for g in constrained_gates:
            trans.append((g, g._to_stp(cache, defines, variables, True)))
        #print cache
        #print defines
        #print variables
        #print trans

        def _t2t(t):
            """Convert a PySMT type to the corresponding STP type."""
            if t.is_bool():
                return "BOOLEAN"
            elif t.is_int():
                raise RuntimeError("Unbounded integers are not supported "
                                   "in the STP translation")
            elif t.is_rat():
                raise RuntimeError("Rational numbers are not supported "
                                   "in the STP translation")
            elif t.is_sint() or t.is_uint():
                return "BITVECTOR("+str(t.get_width())+")"
            elif t.is_enum():
                raise RuntimeError("Enumeration data types are not supported "
                                   "in the STP translation")
            elif t.is_array():
                return "ARRAY "+_t2t(t.get_index_type())+\
                       " OF "+_t2t(t.get_range_type())
            else:
                _internalerror_snyi()
            
        # Output variable names and types
        map_var_to_smt_name = {}
        map_smt_name_to_var = {}
        for (gate,smt_name) in self._unique(variables):
            map_var_to_smt_name[gate] = smt_name
            map_smt_name_to_var[smt_name] = gate
            f.write(smt_name+" : "+_t2t(gate.get_type())+";\n") 

        if False:
            #
            # An ASSERT-based translation
            # Does not seem to work when array are involved
            #
            # Add defines, i.e. the let and flet constructions
            for (g,t,d) in defines:
                if g.is_boolean():
                    f.write(t+" : "+_t2t(g.get_type())+";\n")
                    f.write("ASSERT("+t+" <=> "+d+");\n")
                else:
                    f.write(t+" : "+_t2t(g.get_type())+";\n")
                    f.write("ASSERT("+t+" = "+d+");\n")

            # Add the conjunction of the constrained gates
            for (g,t) in trans:
                assert g.is_boolean()
                assert g.is_constrained()
                if g.constrained_value == False:
                    f.write('ASSERT(NOT('+t+'));\n')
                elif g.constrained_value == True:
                    f.write('ASSERT('+t+');\n')
                else:
                    assert False
            f.write('QUERY(FALSE);\n')
            if produce_model:
                f.write('COUNTEREXAMPLE;\n')
        else:
            f.write("QUERY(NOT(")
            # Add defines, i.e. the let and flet constructions
            for (g,t,d) in defines:
                f.write('LET '+t+' = '+d+' IN (\n')
            # Add the conjunction of the constrained gates
            for (g,t) in trans:
                assert g.is_boolean()
                assert g.is_constrained()
                if g.constrained_value == False:
                    f.write('(NOT '+t+') AND\n')
                elif g.constrained_value == True:
                    f.write(''+t+' AND\n')
                else:
                    assert False
            f.write('TRUE')
            # Add all the missing parentheses
            for (g,t,d) in defines:
                f.write(')')
            #f.write("\n");
            # The end
            f.write("));\n");
            if produce_model:
                f.write('COUNTEREXAMPLE;\n')

        # 
        #f.write("(set-evidence! true)\n")
        #f.write("(check)\n")

        return (map_var_to_smt_name, map_smt_name_to_var)


    def to_stp(self, f, produce_model = True):
        """
        Output the circuit in STP format to the file object 'f'.
        If 'produce_model' is True, then the 'COUNTEREXAMPLE' command
        is inserted in the generated STP problem file.
        """
        assert type(f) is file
        assert type(produce_model) is bool
        self._to_stp(f, produce_model)



    def _to_cvc3(self, f, produce_model = True):
        """
        Output the circuit in CVC3 format to the file object 'f'.
        Returns maps between input gates and CVC3 format var names.
        """
        assert type(f) is file
        assert type(produce_model) is bool

        # Print a comment
        f.write("% A CVC3 problem generated by PySMT\n")

        # Find the constrained Boolean gates
        constrained_gates = [g for g in self._gates if
                             (g and g.is_constrained())]

        cache = {} #sets.Set()
        defines = []
        variables = []
        map_op_to_opname = {}
        trans = []
        for g in constrained_gates:
            trans.append((g, g._to_cvc3(cache, defines, variables,
                                        map_op_to_opname, True)))
        #print cache
        #print defines
        #print variables
        #print trans
        map_opname_to_op = {}
        for (op, name) in map_op_to_opname.items():
            map_opname_to_op[name] = op

        def _t2t(t):
            """Convert a PySMT type to the corresponding CVC3 type."""
            if t.is_bool():
                return "BOOLEAN"
            elif t.is_int():
                return "INT"
            elif t.is_rat():
                return "REAL"
            elif t.is_sint() or t.is_uint():
                return "BITVECTOR("+str(t.get_width())+")"
            elif t.is_enum():
                return map_op_to_opname[t]
            elif t.is_array():
                return map_op_to_opname[t]
            assert False

        # Define the encountered enumeration data types
        for (opname,op) in map_opname_to_op.items():
            if isinstance(op, _Type_enum):
                f.write("DATATYPE\n")
                f.write("  "+opname+" = ")
                f.write(" | ".join([opname+"_"+str(l) for
                                    l in op.get_domain()])+"\n")
                f.write("END;\n")

        # Define the encountered UIFs
        for (opname,op) in map_opname_to_op.items():
            if isinstance(op, _Operand_uif):
                f.write(opname+": ")
                domain = op.get_domain()
                if len(domain) > 1:
                    f.write("(")
                f.write(", ".join([_t2t(dt) for dt in domain]))
                if len(domain) > 1:
                    f.write(")")
                f.write(" -> "+_t2t(op.get_range_type())+";\n")
            elif isinstance(op, _Type_enum):
                pass
            elif isinstance(op, _Type_array):
                pass
            else:
                _internalerror_snyi()

        # Define the encountered array data types
        for (opname,op) in map_opname_to_op.items():
            if isinstance(op, _Type_array):
                f.write(opname+": TYPE = ARRAY "+_t2t(op.get_index_type())+
                        " OF "+_t2t(op.get_range_type())+";\n")


        # Output variable names and types
        map_var_to_smt_name = {}
        map_smt_name_to_var = {}
        for (gate,smt_name) in self._unique(variables):
            map_var_to_smt_name[gate] = smt_name
            map_smt_name_to_var[smt_name] = gate
            f.write(smt_name+" : "+_t2t(gate.get_type())+";\n")

        #
        # Output the formula
        #
        # Add defines, i.e. the let and flet constructions
        for (g,t,d) in defines:
            f.write(t+': '+_t2t(g.get_type())+' = '+d+';\n')
        # Add the conjunction of the constrained gates
        for (g,t) in trans:
            assert g.is_boolean()
            assert g.is_constrained()
            if g.constrained_value == False:
                f.write('ASSERT (NOT '+t+');\n')
            elif g.constrained_value == True:
                f.write('ASSERT '+t+';\n')
            else:
                assert False
        # Add all the missing parentheses
        #for (g,t) in trans:
        #    f.write(')')

        # The end
        f.write("CHECKSAT TRUE;\n");


        if False:
            #
            # Output the formula
            #
            f.write("CHECKSAT ")
            # Add defines, i.e. the let and flet constructions
            for (g,t,d) in defines:
                print g
                f.write('(LET '+t+' = '+d+' IN ')
            # Add the conjunction of the constrained gates
            f.write('(TRUE')
            for (g,t) in trans:
                assert g.is_boolean()
                assert g.is_constrained()
                if g.constrained_value == False:
                    f.write(' AND (NOT '+t+') ')
                elif g.constrained_value == True:
                    f.write(' AND '+t+' ')
                else:
                    assert False
            f.write(')')
            # Add all the missing parentheses
            #for (g,t) in trans:
            #    f.write(')')
            for (g,t,d) in defines:
                f.write(')')
            #f.write("\n");

            # The end
            f.write(";\n");


        if produce_model:
            f.write("COUNTERMODEL;\n")

        return (map_var_to_smt_name, map_smt_name_to_var, map_opname_to_op)


    def to_cvc3(self, f, produce_model):
        """
        Output the circuit in CVC3 format to the file object 'f'.
        If 'produce_model' is True, then the 'COUNTERMODEL' command
        is inserted in the generated CVC3 problem file.
        """
        assert type(f) is file
        assert type(produce_model) is bool
        self._to_cvc3(f, procude_model)


        
    def _to_mathsat(self, f, opt_enum_translation = 'int'):
        """
        Output the circuit in MathSAT format to the file object 'f'.
        Returns maps between input gates and MathSAT format variable names.
        """
        assert type(f) is file
        assert opt_enum_translation in ['int','object']
        # Print a comment
        f.write("# A MathSAT problem generated by PySMT\n")

        # Find the constrained Boolean gates
        constrained_gates = [g for g in self._gates if
                             (g and g.is_constrained())]

        cache = {} #sets.Set()
        defines = []
        variables = []
        map_op_to_opname = {}
        trans = []
        for g in constrained_gates:
            trans.append((g, g._to_mathsat(cache,
                                           defines,
                                           variables,
                                           map_op_to_opname,
                                           True,
                                           opt_enum_translation)))
        #print cache
        #print defines
        #print variables
        #print trans

        map_opname_to_op = {}
        for (op, name) in map_op_to_opname.items():
            map_opname_to_op[name] = op

        def _t2t(t):
            """Convert a PySMT type to the corresponding MathSAT type."""
            if t.is_bool():
                return "BOOLEAN"
            elif t.is_int():
                return "INTEGER"
            elif t.is_rat():
                return "REAL"
            elif t.is_sint() or t.is_uint():
                return "WORD["+str(t.get_width())+"]"
            elif t.is_enum():
                if opt_enum_translation == 'int':
                    return "INTEGER"
                elif opt_enum_translation == 'object':
                    return "OBJECT"
                else:
                    raise RuntimeError("Internal error")
            _internalerror_snyi()

        # Define the encountered enumeration data types
        for (opname,op) in map_opname_to_op.items():
            if isinstance(op, _Type_enum):
                if opt_enum_translation == 'object':
                    for literal in op.get_domain():
                        f.write("CONST "+opname+"_"+literal+"\n")

        # Define the encountered UIFs
        for (opname,op) in map_opname_to_op.items():
            if isinstance(op, _Operand_uif):
                f.write("VAR "+opname+" : "+
                        (" * ".join([_t2t(dt) for dt in op.get_domain()]))+
                        " -> "+_t2t(op.get_range_type())+"\n")
            elif isinstance(op, _Type_enum):
                pass
            else:
                _internalerror_snyi()

        def _get_domain_constraint(enum_type, varname):
            """
            Generate a constraint ensuring that an enumeration variable
            'varname' has its value belonging to the domain of
            the enumeration type 'enum_type'.
            """
            if opt_enum_translation == 'int':
                s = " | ".join(["("+varname+" = "+
                                str(enum_type.get_literal_index(literal))+")"
                                for literal in enum_type.get_domain()])
                #s = "("+varname+" >= 0) & ("+varname+" < "+\
                #    str(len(enum_type.get_domain()))+")"
            elif opt_enum_translation == 'object':
                opname = map_op_to_opname[enum_type]
                s = " | ".join(["("+varname+" = "+opname+"_"+literal+")"
                                for literal in enum_type.get_domain()])
            else:
                raise RuntimeError("Internal error")
            return s


        # Output variable names and types
        map_var_to_smt_name = {}
        map_smt_name_to_var = {}
        for (gate,smt_name) in self._unique(variables):
            map_var_to_smt_name[gate] = smt_name
            map_smt_name_to_var[smt_name] = gate
            f.write("VAR "+smt_name+" : "+_t2t(gate.get_type())+"\n")

            # Add a constraint that an enumeration variable should have
            # its value belonging to the domain of the enumeration type
            if gate.is_enum():
                f.write("FORMULA "+
                        _get_domain_constraint(gate.get_type(), smt_name)+"\n")

        #
        # Output the formula
        #
        # Add defines
        for (g,t,d) in defines:
            if g.is_boolean():
                f.write("DEFINE "+t+" : BOOLEAN := "+d+"\n")
            elif g.is_int():
                f.write("DEFINE "+t+" : INTEGER := "+d+"\n")
            elif g.is_rat():
                f.write("DEFINE "+t+" : REAL := "+d+"\n")
            elif g.is_sint():
                width = g.get_sint_width()
                assert width >= 1
                f.write("DEFINE "+t+" : WORD["+str(width)+"] := "+d+"\n")
            elif g.is_uint():
                width = g.get_uint_width()
                assert width >= 1
                f.write("DEFINE "+t+" : WORD["+str(width)+"] := "+d+"\n")
            elif g.is_enum():
                if opt_enum_translation == 'int':
                    f.write("DEFINE "+t+" : INTEGER := "+d+"\n")
                elif opt_enum_translation == 'object':
                    f.write("DEFINE "+t+" : OBJECT := "+d+"\n")
                else:
                    raise RuntimeError("Internal error")
                f.write("FORMULA "+
                        _get_domain_constraint(g.get_type(), t)+"\n")
            else:
                _internalerror_snyi()

        # Add the conjunction of the constrained gates
        for (g,t) in trans:
            assert g.is_boolean()
            assert g.is_constrained()
            if g.constrained_value is False:
                f.write('FORMULA !('+t+')\n')
            elif g.constrained_value is True:
                f.write('FORMULA '+t+'\n')
            else:
                raise RuntimeError("Internal error")

        return (map_var_to_smt_name,
                map_smt_name_to_var,
                map_opname_to_op)


    def to_mathsat(self, f):
        """
        Output the circuit in MathSAT format to the file object 'f'.
        """
        assert type(f) is file
        self._to_mathsat(f)


        
    def _to_smtlib(self, f):
        """
        Output the circuit in SMT-LIB format to the file object f.
        Returns maps between input gates and smtlib format var names.
        """

        #
        # A flag that indicates whether (f)lets are used in the translation
        #
        OPT_USE_LETS = True
        OPT_USE_LETS = False

        # Print the header
        f.write("(benchmark PySMT\n")
        logic = self._deduce_smtlib_logic_class(self._get_relevant_gates())
        f.write(" :logic "+logic+"\n")

        # Find the constrained Boolean gates
        constrained_gates = [g for g in self._gates if
                             (g and g.is_constrained())]
        cache = {} #sets.Set()
        defines = []
        variables = []
        trans = []
        map_op_to_opname = {}
        for g in constrained_gates:
            trans.append((g, g._to_smt2(cache, defines, variables,
                                        map_op_to_opname, OPT_USE_LETS)))
            #trans.append((g, g._to_smt(cache, defines, variables,
            #                           map_op_to_opname, True, OPT_USE_LETS)))
        #print cache
        #print defines
        #print variables
        #print trans
        map_opname_to_op = {}
        for (op, name) in map_op_to_opname.items():
            map_opname_to_op[name] = op

        def _t2t(t):
            """Convert a PySMT type to the corresponding SMT-LIB type."""
            if t.is_bool():
                assert False
            elif t.is_int():
                return "Int"
            elif t.is_rat():
                return "Real"
            elif t.is_sint() or t.is_uint():
                return "BitVec["+str(t.get_width())+"]"
            elif t.is_enum():
                raise RuntimeError("enumeration data types not supported in "
                                   "the SMT-LIB translation")
                return map_op_to_opname[t]
            assert False

        # Define the encountered UIFs
        for (opname,op) in map_opname_to_op.items():
            if isinstance(op, _Operand_uif):
                domain = op.get_domain()
                if op.get_range_type().is_bool():
                    f.write(" :extrapreds (("+opname+" ")
                    f.write(" ".join([_t2t(dt) for dt in domain]))
                    f.write("))\n")
                else:
                    f.write(" :extrafuns (("+opname+" ")
                    f.write(" ".join([_t2t(dt) for dt in domain]))
                    f.write(" "+_t2t(op.get_range_type())+"))\n")
            elif isinstance(op, _Type_enum):
                pass
            else:
                _internalerror_snyi()

        # Output variable names and types
        map_var_to_smt_name = {}
        map_smt_name_to_var = {}
        for (gate,smt_name) in self._unique(variables):
            map_var_to_smt_name[gate] = smt_name
            map_smt_name_to_var[smt_name] = gate
            if gate.is_boolean():
                f.write(" :extrapreds (("+smt_name+"))\n")
            else:
                f.write(" :extrafuns (("+smt_name+" "+
                        _t2t(gate.get_type())+"))\n")

        if OPT_USE_LETS:
            #
            # Option 1: use lets
            #
            # Outputs one massive :formula with shared subterms defined
            #   by using let and flet constructs
            #
            f.write(" :formula ")

            # Add defines, i.e. the let and flet constructions
            for (g,t,d) in defines:
                if g.is_boolean():
                    f.write('(flet ('+t+' '+d+') ')
                else:
                    f.write('(let ('+t+' '+d+') ')
            # Add the conjunction of the constrained gates
            for (g,t) in trans:
                assert g.is_boolean()
                assert g.is_constrained()
                if g.constrained_value == False:
                    f.write('(and (not '+t+') ')
                elif g.constrained_value == True:
                    f.write('(and '+t+' ')
                else:
                    assert False
            f.write('true')
            # Add all the missing parentheses
            for (g,t) in trans:
                f.write(')')
            for (g,t,d) in defines:
                f.write(')')
            f.write("\n");
        else:
            #
            # Option 2: use assumptions
            #
            # Outputs one smaller formula while shared subterms are defined
            #   by using :assumptions
            #
            # Add assumptions
            for (g,t,d) in defines:
                if g.is_boolean():
                    f.write(" :extrapreds (("+t+"))\n")
                    f.write(" :assumption (iff "+t+" "+d+")\n")
                else:
                    f.write(" :extrafuns (("+t+" "+_t2t(g.get_type())+"))\n")
                    f.write(" :assumption (= "+t+" "+d+")\n")
            # Add the conjunction of the constrained gates
            f.write(" :formula (and true")
            for (g,t) in trans:
                assert g.is_boolean()
                assert g.is_constrained()
                if g.constrained_value == False:
                    f.write(' (not '+t+')')
                elif g.constrained_value == True:
                    f.write(' '+t)
                else:
                    assert False
            f.write(")\n")

        # The end
        f.write(")\n");

        return (map_var_to_smt_name, map_smt_name_to_var, map_opname_to_op)


    def to_smtlib(self, f):
        """
        Output the circuit in SMT-LIB format to the file object f.
        """
        assert type(f) is file
        self._to_smtlib(f)



    def _extend_solution(self, solution,
                         map_var_to_smt_name, map_smt_name_to_var,
                         map_opname_to_op):

        """
        Extend the partial solution 'solution' to all gates in the circuit
        by assigning unassigned input gates a default value and
        then evaluating the non-input gates.
        Returns the extended solution.
        """
        #
        # Assert that all variable names in the SMT-LIB format problem have
        # a value by now
        #
        #for (gate,smt_name) in map_var_to_smt_name.iteritems():
        #    if not solution.has_key(gate):
        #        print gate
        #        assert False and "Internal error: an incomplete solution"
        
        #
        # Assign an arbitrary default value to input gates not yet assigned
        #
        for gate in [g for g in self._gates if (g and not g.children)]:
            if solution.has_key(gate):
                continue
            tyype = gate.get_type()
            if gate.operand == 'FALSE':
                solution[gate] = False
            elif gate.operand == 'TRUE':
                solution[gate] = True
            elif gate.operand == 'VAR':
                if gate.is_boolean():
                    solution[gate] = False
                elif gate.is_int():
                    solution[gate] = long(0)
                elif gate.is_sint():
                    solution[gate] = long(0)
                elif gate.is_uint():
                    solution[gate] = long(0)
                elif gate.is_enum():
                    solution[gate] = gate.get_type().get_a_value()
                else:
                    #print gate
                    _internalerror_snyi()
            elif gate.operand == 'CONST':
                if gate.is_sint():
                    solution[gate] = gate.opparams
                elif gate.is_uint():
                    solution[gate] = gate.opparams
                elif gate.is_int():
                    solution[gate] = gate.opparams
                elif gate.is_enum():
                    solution[gate] = gate.opparams
                else:
                    #print gate
                    _internalerror_snyi()
            else:
                print gate
                assert False

        if False: # or True:
            for (gate,value) in solution.iteritems():
                print str(gate)+" = "+str(value)

        for gate in self._gates:
            if gate:
                gate.evaluate(solution)

        if False: # or True:
            for (gate,value) in solution.iteritems():
                print str(gate)+" = "+str(value)

        return solution

    def _get_real_gate(self, gate):
        if isinstance(gate, Gate):
            gate = gate._get_gate()
        else:
            assert isinstance(gate, Circuit._RealGate)
        assert gate.get_circuit() is self
        assert gate.has_index()
        assert self._gates[gate.get_index()] is gate
        return gate

    def _get_relevant_gates(self):
        relevant = [g for g in self._gates if g and g.is_constrained()]
        found = sets.Set(relevant)
        i = 0
        while i < len(relevant):
            gate = relevant[i]
            for child in gate.children:
                if child not in found:
                    relevant.append(child)
                    found.add(child)
            i += 1
        return relevant

    def _type_error(self, s):
        assert type(s) is str
        raise RuntimeError("Type error: "+s)

    #
    #
    # The API starts here
    #
    #
    def __init__(self):
        """
        Create an empty circuit.
        """
        self._gate_lookup = {}
        self._gates = []
        self.free_gate_indices = []
        self.nof_gates = 0
        self._var_num = 0L
        self.solution = None
        self.is_consistent = True


    def get_nof_gates(self):
        """
        Get the number of gates in the circuit.
        """
        return self.nof_gates

    def get_nof_relevant_gates(self):
        """
        Get the number of gates in the circuit that are constrained
        gates or descendants of constrained gates.
        """
        return len(self._get_relevant_gates())

    def get_type(self, gate):
        """
        Return the type of the gate 'gate'.
        """
        assert isinstance(gate, Gate)
        gate = self._get_real_gate(gate)
        return gate.get_type()


    def is_constant_true(self, gate):
        """
        Returns True iff 'gate' corresponds to the Boolean constant True
        and the circuit is not inconsistent.
        """
        assert isinstance(gate, Gate)
        gate = self._get_real_gate(gate)
        return ((not self.is_inconsistent()) and
                gate.is_boolean() and
                gate.operand == "TRUE")

        
    def is_constant_false(self, gate):
        """
        Returns True iff 'gate' corresponds to the Boolean constant False
        and the circuit is not inconsistent.
        """
        assert isinstance(gate, Gate)
        gate = self._get_real_gate(gate)
        return ((not self.is_inconsistent()) and
                gate.is_boolean() and
                gate.operand == "FALSE")

        
    def is_inconsistent(self):
        """
        Returns True if it has been detected that the circuit is
        permanently inconsistent.
        """
        return self.is_consistent == False


    def constrain(self, gate, value):
        """
        Constrain the value of a Boolean gate to False or True.
        """
        assert isinstance(gate, Gate) or isinstance(gate, Circuit._RealGate)
        assert type(value) is bool
        gate = self._get_real_gate(gate)
        if not gate.is_boolean():
            _type_error("Cannot constrain a non-boolean gate")        
        if gate.is_constrained():
            if gate.constrained_value != value:
                self.is_consistent = False
                self._invalidate_solution()
        else:
            gate.constrained_value = value
            self._invalidate_solution()


    def get_value(self, gate):
        """
        Return the value of the gate 'gate' in the latest solution.
        Thus the previous solve_* API call to this circuit must have
        returned 'sat'.
        The solution is invalidated when any add_* or constrain_* function
        is called.
        """
        assert isinstance(gate, Gate)
        assert self.solution
        gate = self._get_real_gate(gate)
        assert self.solution.has_key(gate)
        return self.solution[gate]


    def add_var(self, t = None):
        """
        Add and return a new variable of type 't'.
        If 't' is None, a Boolean variable is returned.
        """
        if t is None:
            t = get_bool()
        assert isinstance(t, _Type)
        return self._add_gate(t, 'VAR', None, [])


    def add_sint_var(self, w):
        """
        Add and return a new bounded signed integer variable of width w.
        Required: 1 <= w <= 32.
        """
        assert type(w) is int
        assert 1 <= w and w <= 32
        return self._add_gate(get_sint(w), 'VAR', None, [])


    def add_sint_const(self, w, c):
        """
        Add and return a new bounded signed integer constant c of width w,
        1 <= w <= 32 and -2^{w-1} <= c <= 2^{w-1}-1
        """
        assert type(w) is int
        assert 1 <= w and w <= 32
        c = long(c)
        assert c >= -pow(2,w-1)
        assert c <= pow(2,w-1)-1
        return self._add_gate(get_sint(w), 'CONST', c, [])

    
    def add_uint_var(self, w):
        """
        Add and return a new bounded unsigned integer variable of width w.
        Required: 0 <= w.
        """
        assert type(w) is int
        assert 0 <= w
        if w == 0:
            return self.add_uint_const(0, 0)
        return self._add_gate(get_uint(w), 'VAR', None, [])


    def add_uint_const(self, w, c):
        """
        Add and return a new bounded unsigned integer constant c of width w.
        Required: 0 <= w and 0 <= c < 2^w.
        """
        assert type(w) is int
        assert 0 <= w
        c = long(c)
        assert 0 <= c
        assert c < pow(2,w)
        return self._add_gate(get_uint(w), 'CONST', c, [])

    
    def add_int_var(self):
        """
        Add and return a new unbounded integer variable.
        """
        return self._add_gate(get_int(), 'VAR', None, [])


    def add_int_const(self, c):
        """
        Add and return a new unbounded integer constant c.
        """
        assert (type(c) is int) or (type(c) is long)
        return self._add_gate(get_int(), 'CONST', long(c), [])


    def add_enum_var(self, t):
        """
        Add and return a variable of enumeration type 't'.
        """
        assert isinstance(t, _Type_enum)
        return self._add_gate(t, 'VAR', None, [])


    def add_enum_const(self, t, c):
        """
        Add and return a new constant 'c' of enumeration type 't'.
        'c' should be a literal in the domain of 't'.
        """
        assert type(c) is str
        assert isinstance(t, _Type_enum)
        assert t.has_literal(c)
        return self._add_gate(t, 'CONST', c, [])


    def add_array_var(self, t):
        """
        Add and return a variable of an array type 't'.
        """
        assert isinstance(t, _Type_array)
        return self._add_gate(t, 'VAR', None, [])

    def add_array_read(self, array, index):
        """
        Add and return a gate computing the value 'array'['index'],
        where 'array' is a gate of array type and 'index' is a gate of
        the index type of the array type.
        """
        array = self._get_real_gate(array)
        index = self._get_real_gate(index)
        if not isinstance(array.get_type(), _Type_array):
            _type_error("Read on a non-array type gate")
        if index.get_type() is not array.get_type().get_index_type():
            _type_error("Index type '"+str(index.get_type())+"' "+
                        "does not match the index type of the array '"+
                        str(array.get_type())+"'")
        return self._add_gate(array.get_type().get_range_type(),'READ',None,[array,index])

    def add_array_write(self, array, index, value):
        """
        Add and return a new array gate equal to 'array' except that
        'array'['index'] = 'value',
        where 'array' is a gate of array type,
        'index' is a gate of the index type of the array type, and
        'value' is a gate of the range type of the array type.
        """
        array = self._get_real_gate(array)
        index = self._get_real_gate(index)
        value = self._get_real_gate(value)
        if not isinstance(array.get_type(), _Type_array):
            _type_error("Write on a non-array type gate")
        if index.get_type() is not array.get_type().get_index_type():
            _type_error("Index type '"+str(index.get_type())+"' "+
                        "does not match the index type of the array '"+
                        str(array.get_type())+"'")
        if value.get_type() is not array.get_type().get_range_type():
            _type_error("Value type '"+str(value.get_type())+"' "+
                        "does not match the range type of the array '"+
                        str(array.get_type())+"'")
        return self._add_gate(array.get_type(),'WRITE',None,
                              [array,index,value])


    def add_eq(self, child1, child2):
        """
        Add and return a new equality gate.
        The types of the two children should match and be either Boolean,
        enumeration, integer, rational, or bounded integers of same width.
        The gate evaluates to true iff the children evaluate to the same
        value.
        """
        return self._add_binary_comparison('EQ',child1,child2)


    def add_neq(self, child1, child2):
        """
        Add and return a new negated equality gate.
        The types of the two children should match and be either Boolean,
        enumeration, integer, rational, or bounded integers of same width.
        The gate evaluates to true iff the children evaluate to the different
        values.
        """
        return self.add_not(self.add_eq(child1, child2))


    def add_lt(self, child1, child2):
        """
        Add and return a new less-than gate.
        The types of the two children should match and be either integer,
        rational, or bounded integers of same width.
        """
        return self._add_binary_comparison('LT',child1,child2)


    def add_le(self, child1, child2):
        """
        Add and return a new less-or-equal gate.
        The types of the two children should match and be either integer,
        rational, or bounded integers of same width.
        """
        return self._add_binary_comparison('LE',child1,child2)


    def add_ge(self, child1, child2):
        """
        Add and return a new greater-or-equal gate.
        The types of the two children should match and be either integer,
        rational, or bounded integers of same width.
        """
        return self.add_le(child2,child1)


    def add_gt(self, child1, child2):
        """
        Add and return a new greater-than gate.
        The types of the two children should match and be either integer,
        rational, or bounded integers of same width.
        """
        return self.add_lt(child2,child1)


    def add_plus(self, *children):
        """
        Add and return a new addition gate over the children.
        The types of the children should match and be either integer,
        rational, or bounded integers of same width.
        """
        assert type(children) is tuple
        assert len(children) >= 1
        if len(children) == 1 and (type(children[0]) is list):
            children = children[0]
        # Get the children behind the handles, type check
        real_children = []
        tyype = None
        for (index,child) in enumerate(children):
            child = self._get_real_gate(child)
            if tyype == None:
                tyype = child.get_type()
                if not (child.is_int() or child.is_rat() or
                        child.is_sint() or child.is_uint()):
                    _type_error("Addition on a gate of type '"+
                                str(tyype)+"'")
            else:
                if tyype != child.get_type():
                    _type_error("Addition on gates of incompatible types")
            real_children.append(child)
        assert tyype
        return self._add_gate(tyype, 'PLUS', None, real_children)


    def add_minus(self, child1, child2):
        """
        Add and return a new gate for the subtraction child1 - child2.
        The types of the children should match and be either integer,
        rational, or bounded integers of same width.
        """
        # Get the children behind the handles, type check
        child1 = self._get_real_gate(child1)
        child2 = self._get_real_gate(child2)
        if child1.get_type() != child2.get_type():
            _type_error("Subtraction on gates of incompatible types")
        if not (child1.is_int() or child1.is_rat() or
                child1.is_sint() or child1.is_uint()):
            _type_error("Subtraction on a gate of type '"+
                        str(child1.get_type())+"'")
        return self._add_gate(child1.get_type(),'MINUS',None,[child1,child2])


    def add_neg(self, child):
        """
        Add and return a new gate for the negation of child.
        The type of child must be either integer, rational, or bounded integer.
        """
        # Get the child behind the handle, type check
        child = self._get_real_gate(child)
        if not (child.is_int() or child.is_rat() or
                child.is_sint() or child.is_uint()):
            _type_error("Negation on a gate of type '"+
                        str(child.get_type())+"'")
        return self._add_gate(child.get_type(),'NEG',None,[child])


    def add_times(self, *children):
        """
        Add and return a new multiplication gate over the children.
        The types of the children should match and be either (i) integer,
        (ii) rational, or (iii) bounded integers of same width.
        """
        assert type(children) is tuple
        if len(children) == 1 and (type(children[0]) is list):
            children = children[0]
        assert len(children) >= 2
        # Get the children behind the handles, type check
        real_children = []
        tyype = None
        for (index,child) in enumerate(children):
            child = self._get_real_gate(child)
            if tyype == None:
                if not (child.is_int() or child.is_rat() or
                        child.is_sint() or child.is_uint()):
                    _type_error("Multiplication on a gate of type '"+
                                str(child.get_type())+"'")
                tyype = child.get_type()
            elif tyype != child.get_type():
                _type_error("Multiplication on gates of incompatible types")
            real_children.append(child)
        assert tyype
        return self._add_gate(tyype, 'TIMES', None, real_children)


    def add_bool_var(self):
        """
        Add and return a new boolean variable.
        """
        return self.add_var(get_bool())


    def add_false(self):
        """
        Get the boolean false constant gate.
        """
        g = self._RealGate(self,get_bool(),'FALSE',None,[])
        g = self._install_gate(g)
        self.constrain(g, False)
        return g


    def add_true(self):
        """
        Get the boolean true constant gate.
        """
        g = self._RealGate(self,get_bool(),'TRUE',None,[])
        g = self._install_gate(g)
        self.constrain(g, True)
        return g


    def add_not(self, child):
        """
        Add and return a new boolean not-gate over the child.
        """
        # Get the child behind the handle
        child = self._get_real_gate(child)
        if not child.is_boolean():
            _type_error("Cannot make a 'not' of a non-boolean gate")
        return self._add_gate(get_bool(),'NOT',None,[child])


    def add_and(self, *children):
        """
        Add and return a new boolean and-gate over the children.
        The gate evaluates to true iff all child gates evaluate to true.
        If no children are given, return the boolean true constant gate.
        If the children list is empty, the gate is equivalent to true.
        """
        assert type(children) is tuple
        if len(children) == 1 and (type(children[0]) is list):
            children = children[0]
        # Get the children behind the handles, type check
        real_children = []
        for (index,child) in enumerate(children):
            child = self._get_real_gate(child)
            if not child.is_boolean():
                _type_error("The argument "+str(index+1)+" is not boolean")
            real_children.append(child)
        return self._add_gate(get_bool(),'AND',None,real_children)


    def add_or(self, *children):
        """
        Add and return a new boolean or-gate over the children.
        The gate evaluates to true iff at least one of the child gates
        evaluates to true.
        If no children are given, return the boolean false constant gate.
        If the children list is empty, the gate is equivalent to false.
        """
        assert type(children) is tuple
        if len(children) == 1 and (type(children[0]) is list):
            children = children[0]
        # Get the children behind the handles, type check
        real_children = []
        for (index,child) in enumerate(children):
            child = self._get_real_gate(child)
            if not child.is_boolean():
                _type_error("The argument "+str(index+1)+" is not boolean")
            real_children.append(child)
        return self._add_gate(get_bool(),'OR',None,real_children)


    def add_implies(self, child1, child2):
        """
        Add and return a new boolean implies-gate over the boolean child gates.
        The gate evaluates to true iff (a) child1 evaluates to false or (b)
        child2 evaluates to true.
        """
        return self.add_or(self.add_not(child1), child2)


    def add_iff(self, child1, child2):
        """
        Add and return a new boolean iff-gate over the boolean child gates.
        The gate evaluates to true iff the child gates evaluate to
        the same value.
        """
        # Get the children behind the handles, type check
        real_children = []
        for (index,child) in enumerate([child1,child2]):
            child = self._get_real_gate(child)
            if not child.is_boolean():
                _type_error("The argument "+str(index+1)+" is not boolean")
            real_children.append(child)
        return self._add_gate(get_bool(),'IFF',None,real_children)


    def add_xor(self, child1, child2):
        """
        Add and return a new boolean xor-gate over the boolean child gates.
        """
        return self.add_not(self.add_iff(child1, child2))


    def add_bitwise_or(self, child1, child2):
        """
        Add and return a new gate for the bitwise or of child1 and child2.
        The types of the children should be bounded integers of same width.
        """
        # Get the children behind the handles, type check
        child1 = self._get_real_gate(child1)
        child2 = self._get_real_gate(child2)
        if child1.get_type() != child2.get_type():
            _type_error("Bitwise or on gates of incompatible types")
        if not (child1.is_sint() or child1.is_uint()):
            _type_error("Bitwise or on a gate of type '"+
                        str(child1.get_type())+"'")
        return self._add_gate(child1.get_type(),'BVOR',None,[child1,child2])


    def add_bitwise_and(self, child1, child2):
        """
        Add and return a new gate for the bitwise and of child1 and child2.
        The types of the children should be bounded integers of same width.
        """
        # Get the children behind the handles, type check
        child1 = self._get_real_gate(child1)
        child2 = self._get_real_gate(child2)
        if child1.get_type() != child2.get_type():
            _type_error("Bitwise and on gates of incompatible types")
        if not (child1.is_sint() or child1.is_uint()):
            _type_error("Bitwise and on a gate of type '"+
                        str(child1.get_type())+"'")
        return self._add_gate(child1.get_type(),'BVAND',None,[child1,child2])


    def add_bitwise_xor(self, child1, child2):
        """
        Add and return a new gate for the bitwise xor of child1 and child2.
        The types of the children should be bounded integers of same width.
        """
        # Get the children behind the handles, type check
        child1 = self._get_real_gate(child1)
        child2 = self._get_real_gate(child2)
        if child1.get_type() != child2.get_type():
            _type_error("Bitwise xor on gates of incompatible types")
        if not (child1.is_sint() or child1.is_uint()):
            _type_error("Bitwise xor on a gate of type '"+
                        str(child1.get_type())+"'")
        return self._add_gate(child1.get_type(),'BVXOR',None,[child1,child2])


    def add_bitwise_not(self, child):
        """
        Add and return a new gate for the bitwise not of child.
        The type of the child should be a bounded integer.
        """
        # Get the child behind the handle, type check
        child = self._get_real_gate(child)
        if not (child.is_sint() or child.is_uint()):
            _type_error("Bitwise not on a gate of type '"+
                        str(child.get_type())+"'")
        return self._add_gate(child.get_type(),'BVNOT',None,[child])


    def add_ite(self, if_child, then_child, else_child):
        """
        Add and return a new if-then-else-gate
        if-then-else(if_child, then_child, else_child), where
        (i) if_child must be Boolean, and
        (ii) then_child and else_child are of same type, the type being
             boolean, enumeration, integer, rational, array,
             bounded signed integer, or bounded unsigned integer.
        The type of the gate is the same as the type of the then and else
        childs.
        The gate evaluates to 
        (i) the value of then_child, if if_child is true, and
        (ii) the value of else_child, if_child is false.
        """
        if_child = self._get_real_gate(if_child)
        then_child = self._get_real_gate(then_child)
        else_child = self._get_real_gate(else_child)
        if not if_child.is_boolean():
            _type_error("The if-child of an ite is not boolean")
        if then_child.get_type() != else_child.get_type():
            _type_error("The then- and else-children of an ite are of "+
                        "incompatible types")
        if not (then_child.is_boolean() or
                then_child.is_int() or then_child.is_rat() or
                then_child.is_sint() or then_child.is_uint() or
                then_child.is_enum() or then_child.is_array()):
            _type_error("Then- and else-children of an ite cannot be "+
                        "of type '"+str(then_child.get_type())+"'")
        return self._add_gate(then_child.get_type(),'ITE',None,
                              [if_child,then_child,else_child])


    def add_uif(self, uif, *children):
        """
        Add and return a new uninterpreted function application gate
        over the children.
        The types of the children must match the argument types of the
        uninterpreted function 'uif'.
        The type of the returned gate is the range type of 'uif',
        and its value is uif(value(child1),...,value(childn)).
        """
        assert isinstance(uif, _Operand_uif)
        assert type(children) is tuple
        if len(children) == 1 and (type(children[0]) is list):
            children = children[0]
        # Get the children behind the handles, type check
        if uif.get_domain_size() != len(children):
            _type_error("The number of children does not match the number "
                        "of arguments of the UIF '"+str(uif)+"'")
        real_children = []
        for (index,child) in enumerate(children):
            uif_type = uif.get_domain_type(index)
            child = self._get_real_gate(child)
            if not (uif_type is child.get_type()):
                _type_error("The argument "+str(index+1)+" is not of type '"+
                            str(uif_type)+"' as expected by the UIF '"+
                            str(uif)+"'")
            real_children.append(child)
        return self._add_gate(uif.get_range_type(),uif,None,real_children)


    def _add_gate(self, tyype, operand, opparams, children):
        """
        The real code for adding new gates,
        includes some trivial simplifications.
        """
        # Some consistency checks
        assert tyype
        assert isinstance(operand, _Operand) or (type(operand) is str)
        assert type(children) is list
        for child in children:
            assert isinstance(child, Circuit._RealGate)
            assert child.get_circuit() is self
            assert child.has_index()
            assert self._gates[child.get_index()] is child

        if operand == 'VAR':
            assert len(children) == 0
            if opparams == None:
                g = self._RealGate(self, tyype, 'VAR', self._var_num, [])
                self._var_num += 1
            else:
                assert type(opparams) is long
                g = self._RealGate(self, tyype, 'VAR', opparams, [])
            return self._install_gate(g)

        elif operand == 'CONST':
            assert opparams != None
            assert len(children) == 0
            g = self._RealGate(self, tyype, 'CONST', opparams, [])
            return self._install_gate(g)

        elif operand == 'FALSE':
            assert tyype.is_bool()
            assert not children
            assert opparams == None
            g = self._RealGate(self,tyype,'FALSE',None,[])
            return self._install_gate(g)

        elif operand == 'TRUE':
            assert tyype.is_bool()
            assert not children
            assert opparams == None
            g = self._RealGate(self,tyype,'TRUE',None,[])
            return self._install_gate(g)

        elif operand == 'NOT':
            assert tyype.is_bool()
            assert len(children) == 1
            child = children[0]
            assert isinstance(child, self._RealGate)
            assert child.is_boolean()
            if child.operand == 'TRUE':
                return self.add_false()
            if child.operand == 'FALSE':
                return self.add_true()
            # Reduce NOT(NOT(x)) to x
            if child.operand == 'NOT':
                assert child.has_index()
                assert self._gates[child.get_index()] is child
                assert len(child.children) == 1
                grandchild = child.children[0]
                assert grandchild.has_index()
                assert self._gates[grandchild.get_index()] is grandchild
                return self._install_gate(grandchild)
            g = self._RealGate(self,get_bool(),'NOT',None,[child])
            return self._install_gate(g)

        elif operand == 'AND':
            assert tyype.is_bool()
            assert opparams == None
            # Remove duplicate and constant children
            reduced_children = []
            for child in self._unique(children):
                assert child.is_boolean()
                if child.operand == 'TRUE':
                    continue
                elif child.operand == 'FALSE':
                    return self.add_false()
                reduced_children.append(child)
            if reduced_children == []:
                return self.add_true()
            if len(reduced_children) == 1:
                child = reduced_children[0]
                return self._install_gate(child)
            # Reduce AND(x,!x,y) to false
            # This is quite inefficient (quadratic) at the moment
            for child in reduced_children:
                if child.operand == 'NOT':
                    grandchild = child.children[0]
                    if grandchild in reduced_children:
                        return self.add_false()
            g = self._RealGate(self,tyype,'AND',opparams,reduced_children)
            return self._install_gate(g)
            
        elif operand == 'OR':
            assert tyype.is_bool()
            assert opparams == None
            # Remove duplicate and constant children
            reduced_children = []
            for child in self._unique(children):
                assert child.is_boolean()
                if child.operand == 'FALSE':
                    continue
                elif child.operand == 'TRUE':
                    return self.add_true()
                reduced_children.append(child)
            if reduced_children == []:
                return self.add_false()
            if len(reduced_children) == 1:
                child = reduced_children[0]
                return self._install_gate(child)
            # Reduce OR(x,!x,y) to true
            # This is quite inefficient (quadratic) at the moment
            for child in reduced_children:
                if child.operand == 'NOT':
                    grandchild = child.children[0]
                    if grandchild in reduced_children:
                        return self.add_true()
            g = self._RealGate(self,tyype,'OR',opparams,reduced_children)
            return self._install_gate(g)
            
        elif operand == 'IFF':
            assert tyype.is_bool()
            assert opparams == None
            assert len(children) == 2
            child1 = children[0]
            child2 = children[1]
            assert child1.get_type() == tyype
            assert child2.get_type() == tyype
            # Remove constant children
            if child1.operand == 'FALSE':
                return self.add_not(child2)
            if child1.operand == 'TRUE':
                return self._install_gate(child2)
            if child2.operand == 'FALSE':
                return self.add_not(child1)
            if child2.operand == 'TRUE':
                return self._install_gate(child1)
            if child1 == child2:
                return self.add_true()
            g = self._RealGate(self,tyype,'IFF',None,[child1, child2])
            return self._install_gate(g)

        elif operand == 'ITE':
            assert opparams == None
            assert len(children) == 3
            (if_child,then_child,else_child) = children
            assert if_child.is_boolean()
            assert then_child.get_type() == tyype
            assert then_child.get_type() == else_child.get_type()
            assert (then_child.is_boolean() or then_child.is_enum() or
                    then_child.is_int() or then_child.is_rat() or
                    then_child.is_sint() or then_child.is_uint() or
                    then_child.is_array())
            # Remove constant children
            if if_child.operand == 'FALSE':
                # ite(False,t,e) == e
                return self._install_gate(else_child)
            if if_child.operand == 'TRUE':
                # ite(True,t,e) == t
                return self._install_gate(then_child)
            if then_child.is_boolean():
                if then_child.operand == 'FALSE':
                    # ite(i,False,e) == !i & e
                    return self.add_and(self.add_not(if_child),else_child)
                if then_child.operand == 'TRUE':
                    # ite(i,True,e) == i | e
                    return self.add_or(if_child,else_child)
                if else_child.operand == 'FALSE':
                    # ite(i,t,False) == i & t
                    return self.add_and(if_child,then_child)
                if else_child.operand == 'TRUE':
                    # ite(i,t,True) == ~i | t
                    return self.add_or(self.add_not(if_child),then_child)
            g = self._RealGate(self,tyype,'ITE',None,
                               [if_child,then_child,else_child])
            return self._install_gate(g)

        elif operand == 'BVOR':
            assert tyype.is_sint() or tyype.is_uint()
            assert opparams == None
            assert len(children) >= 2
            # Type check
            for child in children:
                assert tyype == child.get_type()
            g = self._RealGate(self,tyype,'BVOR',None,children)
            return self._install_gate(g)

        elif operand == 'BVAND':
            assert tyype.is_sint() or tyype.is_uint()
            assert opparams == None
            assert len(children) >= 2
            # Type check
            for child in children:
                assert tyype == child.get_type()
            g = self._RealGate(self,tyype,'BVAND',None,children)
            return self._install_gate(g)

        elif operand == 'BVXOR':
            assert tyype.is_sint() or tyype.is_uint()
            assert opparams == None
            assert len(children) >= 2
            # Type check
            for child in children:
                assert tyype == child.get_type()
            g = self._RealGate(self,tyype,'BVXOR',None,children)
            return self._install_gate(g)

        elif operand == 'BVNOT':
            assert tyype.is_sint() or tyype.is_uint()
            assert opparams == None
            assert len(children) == 1
            # Type check
            for child in children:
                assert tyype == child.get_type()
            g = self._RealGate(self,tyype,'BVNOT',None,children)
            return self._install_gate(g)

        elif operand == 'PLUS':
            assert (tyype.is_int() or tyype.is_rat() or
                    tyype.is_sint() or tyype.is_uint())
            assert opparams == None
            assert len(children) >= 1
            # Type check and collect constant gates
            constant_gates = []
            nonconstant_gates = []
            for child in children:
                assert tyype == child.get_type()
                if child.is_constant():
                    constant_gates.append(child)
                else:
                    nonconstant_gates.append(child)
            # Simplify constants
            if len(constant_gates) > 1:
                if tyype.is_int():
                    value = constant_gates.pop().opparams
                    while constant_gates:
                        value += constant_gates.pop().opparams
                    new_constant = self.add_int_const(value)
                elif tyype.is_sint():
                    width = tyype.get_width()
                    value = constant_gates.pop().opparams
                    while constant_gates:
                        value = _sint_plus(width, value,
                                           constant_gates.pop().opparams)
                    new_constant = self.add_sint_const(width, value)
                elif tyype.is_uint():
                    width = tyype.get_width()
                    value = constant_gates.pop().opparams
                    while constant_gates:
                        value = _uint_plus(width, value,
                                           constant_gates.pop().opparams)
                    new_constant = self.add_uint_const(width, value)
                else:
                    _internalerror_snyi()
                children = nonconstant_gates
                children.append(self._get_real_gate(new_constant))
            # Unary plus is interpreted as identity
            if len(children) == 1:
                return self._install_gate(children[0])
            g = self._RealGate(self,tyype,'PLUS',None,children)
            return self._install_gate(g)

        elif operand == 'MINUS':
            assert (tyype.is_int() or tyype.is_rat() or
                    tyype.is_sint() or tyype.is_uint())
            assert opparams == None
            assert len(children) == 2
            (child1,child2) = children
            assert child1.get_type() == tyype
            assert child2.get_type() == tyype
            # Some simplifications
            if child1 == child2:
                # x - x = 0
                if real_child1.is_int():
                    return self.add_int_const(0);
                elif real_child1.is_sint():
                    return self.add_sint_const(child1.get_sint_width(), 0);
                elif real_child1.is_uint():
                    return self.add_uint_const(child1.get_uint_width(), 0);
                else:
                    _internalerror_snyi()
            g = self._RealGate(self,tyype,'MINUS',None,[child1,child2])
            return self._install_gate(g)

        elif operand == 'NEG':
            assert (tyype.is_int() or tyype.is_rat() or
                    tyype.is_sint() or tyype.is_uint())
            assert opparams == None
            assert len(children) == 1
            child = children[0]
            assert child.get_type() == tyype
            # Some simplifications
            if child.is_constant():
                child_value = child.opparams
                if child.is_int():
                    assert type(child_value) is long
                    return self.add_int_const(-child_value)
                elif child.is_sint():
                    assert type(child_value) is long
                    width = child.get_sint_width()
                    return self.add_sint_const(width,
                                               _sint_neg(width, child_value))
                elif child.is_uint():
                    assert type(child_value) is long
                    width = child.get_uint_width()
                    return self.add_sint_const(width,
                                               _uint_neg(width, child_value))
                else:
                    _internalerror_snyi()
            g = self._RealGate(self,tyype,'NEG',None,[child])
            return self._install_gate(g)

        elif operand == 'TIMES':
            assert (tyype.is_int() or tyype.is_rat() or
                    tyype.is_sint() or tyype.is_uint())
            assert opparams == None
            assert len(children) >= 2
            # Type check and collect constant gates
            constant_gates = []
            nonconstant_gates = []
            for child in children:
                assert tyype == child.get_type()
                if child.is_constant():
                    constant_gates.append(child)
                else:
                    nonconstant_gates.append(child)
            # Simplify constants
            if len(constant_gates) > 1:
                if tyype.is_int():
                    value = constant_gates.pop().opparams
                    while constant_gates:
                        value *= constant_gates.pop().opparams
                    new_constant = self.add_int_const(value)
                elif tyype.is_sint():
                    width = tyype.get_width()
                    value = constant_gates.pop().opparams
                    while constant_gates:
                        value = _sint_times(width, value,
                                            constant_gates.pop().opparams)
                    new_constant = self.add_sint_const(width, value)
                elif tyype.is_uint():
                    width = tyype.get_width()
                    value = constant_gates.pop().opparams
                    while constant_gates:
                        value = _uint_times(width, value,
                                            constant_gates.pop().opparams)
                    new_constant = self.add_uint_const(width, value)
                else:
                    _internalerror_snyi()
                children = nonconstant_gates
                children.append(self._get_real_gate(new_constant))
            # Unary times is interpreted as identity
            if len(children) == 1:
                return self._install_gate(children[0])
            g = self._RealGate(self,tyype,'TIMES',None,children)
            return self._install_gate(g)

        elif operand in ['LT','LE','EQ']:
            assert tyype.is_bool()
            assert len(children) == 2
            assert opparams == None
            (child1,child2) = children
            assert child1.get_type() == child2.get_type()
            # Handle enums first
            if child1.is_enum():
                assert operand == 'EQ'
                if child1.is_constant() and child2.is_constant():
                    if child1.opparams == child2.opparams:
                        return self.add_true()
                    return self.add_false()
                g = self._RealGate(self, tyype, operand, opparams, children)
                return self._install_gate(g)
            # Handle arithmetic data types
            if not (child1.is_int() or
                    child1.is_rat() or
                    child1.is_sint() or
                    child1.is_uint()):
                print child1, child2
                raise RuntimeError("Type error")
            if (child1 == child2) and (operand in ['LE','EQ']):
                # (x = x) <-> (x <= x) <-> True
                return self.add_true()
            if child1 == child2 and (operand in ['LT']):
                # (x < x) <-> False
                return self.add_false()
            if child1.is_constant() and child2.is_constant():
                if child1.is_int() or child1.is_uint() or child1.is_sint():
                    if operand == 'EQ':
                        if child1.opparams == child2.opparams:
                            return self.add_true()
                        return self.add_false()
                    elif operand == 'LE':
                        if child1.opparams <= child2.opparams:
                            return self.add_true()
                        return self.add_false()
                    elif operand == 'LT':
                        if child1.opparams < child2.opparams:
                            return self.add_true()
                        return self.add_false()                        
            g = self._RealGate(self, tyype, operand, opparams, children)
            return self._install_gate(g)

        elif operand == 'READ':
            assert opparams is None
            assert len(children) == 2
            (array_child,index_child) = children
            assert array_child.get_type().is_array()
            assert array_child.get_type().get_index_type() is index_child.get_type()
            assert array_child.get_type().get_range_type() is tyype
            g = self._RealGate(self,tyype,'READ',None,
                               [array_child,index_child])
            return self._install_gate(g)

        elif operand == 'WRITE':
            assert opparams is None
            assert len(children) == 3
            (array_child,index_child,value_child) = children
            assert array_child.get_type().is_array()
            assert array_child.get_type().get_index_type() is index_child.get_type()
            assert array_child.get_type().get_range_type() is value_child.get_type()
            assert array_child.get_type() is tyype
            g = self._RealGate(self,tyype,'WRITE',None,
                               [array_child,index_child,value_child])
            return self._install_gate(g)

        elif isinstance(operand, _Operand_uif):
            g = self._RealGate(self, tyype, operand, opparams, children)
            return self._install_gate(g)

        else:
            #print tyype, operand, opparams
            _internalerror_snyi()



    def add_subcircuit(self, subcircuit, input_map, tracked_gates):
        """
        Instantiate the circuit 'subcircuit' into this circuit.
        'subcircuit' may not equal to 'self'.
        'input_map' should be a dictionary from (a subset of) the variable
        gates of 'subcircuit' to the gates already in this circuit.
        If a variable 'v' gate is a key in 'input_map', then it is
        instantiated as 'input_map[v]', otherwise it is instantiated
        as a new variable.        
        'tracked_gates' should be a list of gates in 'subcircuit';
        this method returns a dictionary from them to
        the corresponding gates instantiated into this circuit.
        """
        assert isinstance(subcircuit, Circuit)
        assert type(input_map) is dict
        assert type(tracked_gates) is list
        assert subcircuit is not self
        if not subcircuit.is_consistent:
            self.is_consistent = False
        self._invalidate_solution()
        seen_input_gates = sets.Set()
        real_input_map = {}
        for (subgate,mygate) in input_map.iteritems():
            assert isinstance(subgate, Gate)
            assert isinstance(mygate, Gate)
            subgate = subgate._get_gate()
            mygate = mygate._get_gate()
            assert subgate.circuit == subcircuit
            assert mygate.circuit == self
            assert subgate.is_var()
            if subgate.get_type() != mygate.get_type():
                raise RuntimeError("A gate pair of different types in input_map")
            assert subgate not in seen_input_gates
            seen_input_gates.add(subgate)
            real_input_map[subgate] = mygate

        instantiation_map = {}
        temp_handles = []
        for subroot in subcircuit._gates:
            if subroot is None:
                continue
            if instantiation_map.has_key(subroot):
                continue
            stack = [subroot]
            while stack:
                subgate = stack[-1]
                if instantiation_map.has_key(subgate):
                    stack.pop()
                    continue
                has_unprocessed_children = False
                new_children = []
                for subchild in subgate.children:
                    if instantiation_map.has_key(subchild):
                        new_children.append(instantiation_map[subchild])
                    else:
                        has_unprocessed_children = True
                        stack.append(subchild)
                if has_unprocessed_children:
                    continue
                stack.pop()
                #print "Adding",subgate,"to",self
                if subgate.is_var():
                    if real_input_map.has_key(subgate):
                        mygate = self._install_gate(real_input_map[subgate])
                    else:
                        mygate = self._add_gate(subgate.tyype, subgate.operand,
                                                None, new_children)
                else:
                    mygate = self._add_gate(subgate.tyype, subgate.operand,
                                            subgate.opparams, new_children)
                assert isinstance(mygate, Gate)
                temp_handles.append(mygate)
                mygate = self._get_real_gate(mygate)
                #print "Result",mygate

                # Handle constraints
                if subgate.is_constrained():
                    assert subgate.is_boolean()
                    self.constrain(mygate, subgate.constrained_value)
                    
                instantiation_map[subgate] = mygate

        tracked_map = {}
        for subgate in tracked_gates:
            assert isinstance(subgate, Gate)
            real_subgate = subgate._get_gate()
            assert real_subgate.circuit == subcircuit
            assert instantiation_map.has_key(real_subgate)
            mygate = instantiation_map[real_subgate]
            tracked_map[subgate] = self._install_gate(mygate)

        temp_handles = []

        return tracked_map


    def to_dot(self, f, only_relevant = False):
        """
        Print the circuit in the graphviz dotty format to
        the file object f.
        If 'only_relevant' is True, only the constrained gates and
        their descendants are printed.
        """
        assert type(f) is file
        assert type(only_relevant) is bool

        if only_relevant:
            gates = self._get_relevant_gates()
        else:
            gates = [g for g in self._gates if g]
        f.write('digraph problem {\n')
        for gate in gates:
            gate.to_dot(f)
        f.write('}\n')
        return

    def to_smtlib(self, f):
        """
        Output the circuit as an SMT-LIB format problem to
        the file object f.
        """
        self._to_smtlib(f)


    def solve_smtlib(self, solver):
        """
        Solve the circuit by using Solver solver
        with the SMT-LIB input format.

        Returns either 'error', 'unsat', or 'sat'.

        If 'sat' is returned, the satisfying solution can be queried by
        calling the get_value function.
        """
        assert isinstance(solver, Solver)
        # Remove the previous solution (if any)
        self._invalidate_solution()
        # Is already inconsistent?
        if not self.is_consistent:
            return 'unsat'
        
        # Output the circuit in SMT-LIB format into a temporary file
        #tempfilename = tempfile.mktemp()
        tempfilename = 'tmp_solver.in'
        assert tempfilename
        f = open(tempfilename, 'w')
        (map_var_to_smt_name, map_smt_name_to_var, map_opname_to_op) = self._to_smtlib(f)
        f.close()

        (result,sol) = solver._solve_smtlib(tempfilename, True)

        # Remove the temporary file
        #os.remove(tempfilename)

        if result == 'unsat':
            return 'unsat'

        if result == 'sat':
            sol = solver._interpret_solution(sol,
                                             map_var_to_smt_name,
                                             map_smt_name_to_var,
                                             map_opname_to_op)
            sol = self._extend_solution(sol,
                                        map_var_to_smt_name,
                                        map_smt_name_to_var,
                                        map_opname_to_op)
            self.solution = sol

            # Check that constraints are respected
            for gate in [g for g in self._gates if
                         (g and g.is_constrained())]:
                assert gate.is_boolean()
                if gate.constrained_value != sol[gate]:
                    raise RuntimeError("Internal consistency error")

            return 'sat'

        assert result == 'error'
        return result


    def solve(self, solver, mode = "native", produce_model = True):
        """
        Solve whether the circuit is satisfiable by using the Solver 'solver'.
        The 'mode' should be either 'native' or 'smtlib'.
        If 'mode' is 'native', use the native input format of 'solver',
        otherwise the SMT-LIB format is used.
        
        Returns either 'error', 'unsat', or 'sat'.

        If 'produce_model' is True and 'sat' is returned,
        the satisfying model can be queried by calling the get_value function.
        """
        assert isinstance(solver, Solver)
        assert type(produce_model) is bool
        assert mode in ['native', 'smtlib']
        if mode == "native":
            return self.solve_native(solver, produce_model)
        return self.solve_smtlib(solver)


    def solve_native(self, solver, produce_model = True):
        """
        Solve whether the circuit is satisfiable by using the Solver 'solver'.
        with its native input format.
        
        Returns either 'error', 'unsat', or 'sat'.

        If 'produce_model' is True and 'sat' is returned,
        the satisfying model can be queried by calling the get_value function.
        """
        assert isinstance(solver, Solver)
        assert type(produce_model) is bool
        if isinstance(solver, CVC3):
            return self.solve_cvc3(solver.binary, produce_model)
        elif isinstance(solver, MathSAT):
            return self.solve_mathsat(solver.binary, produce_model)
        elif isinstance(solver, STP):
            return self.solve_stp(solver.binary, produce_model)
        elif isinstance(solver, Yices):
            return self.solve_yices(solver.binary, produce_model)
        elif isinstance(solver, Z3):
            return self.solve_z3(solver.binary, produce_model)
        raise RuntimeError("Solving with the native format is not supported "+
                           "for the solver '"+str(solver.__class__)+"'")


    def solve_yices(self, yices_binary="./yices", produce_model = True):
        """
        Solve the circuit by using Yices and its native input format.
        Returns either 'error', 'unsat', or 'sat'.
        If 'produce_model' is True and 'sat' is returned,
        the satisfying model can be queried by calling the get_value function.
        """
        assert type(yices_binary) is str
        assert type(produce_model) is bool

        # Remove the previous solution (if any)
        self._invalidate_solution()
        # Is already inconsistent?
        if not self.is_consistent:
            return 'unsat'
        
        # Output the circuit in Yices format into a temporary file
        #tempfilename = tempfile.mktemp()
        tempfilename = 'tmp_solver.in'
        assert tempfilename
        f = open(tempfilename, 'w')
        (map_var_to_smt_name, map_smt_name_to_var, map_opname_to_op) = self._to_yices(f, produce_model)
        f.close()
        #print tempfilename

        solver = Yices(yices_binary)

        (result,sol) = solver._solve_native(tempfilename, produce_model)

        # Remove the temporary file
        #os.remove(tempfilename)

        if result == 'unsat':
            return 'unsat'

        if result == 'sat':
            if produce_model:
                sol = solver._interpret_solution(sol,
                                                 map_var_to_smt_name,
                                                 map_smt_name_to_var,
                                                 map_opname_to_op)
                sol = self._extend_solution(sol,
                                            map_var_to_smt_name,
                                            map_smt_name_to_var,
                                            map_opname_to_op)
                self.solution = sol

                # Check that constraints are respected
                for gate in [g for g in self._gates if
                             (g and g.is_constrained())]:
                    assert gate.is_boolean()
                    if gate.constrained_value != sol[gate]:
                        raise RuntimeError("Internal consistency error")
            return 'sat'

        assert result == 'error'
        return result



    def solve_z3(self, z3_binary="./z3", produce_model = True):
        """
        Solve the circuit by using Z3 and its native input format.
        Returns either 'error', 'unsat', or 'sat'.
        If 'produce_model' is True and 'sat' is returned,
        the satisfying model can be queried by calling the get_value function.
        """
        assert type(z3_binary) is str
        assert type(produce_model) is bool

        # Remove the previous solution (if any)
        self._invalidate_solution()
        # Is already inconsistent?
        if not self.is_consistent:
            return 'unsat'
        
        # Output the circuit in Z3 format into a temporary file
        #tempfilename = tempfile.mktemp()
        tempfilename = 'tmp_solver.in'
        assert tempfilename
        f = open(tempfilename, 'w')
        (map_var_to_smt_name, map_smt_name_to_var, map_opname_to_op) = self._to_z3(f)
        f.close()
        #print tempfilename

        solver = Z3(z3_binary)

        (result,sol) = solver._solve_native(tempfilename, produce_model)

        # Remove the temporary file
        #os.remove(tempfilename)

        if result == 'unsat':
            return 'unsat'

        if result == 'sat':
            if produce_model:
                sol = solver._interpret_solution(sol,
                                                 map_var_to_smt_name,
                                                 map_smt_name_to_var,
                                                 map_opname_to_op)
                sol = self._extend_solution(sol,
                                            map_var_to_smt_name,
                                            map_smt_name_to_var,
                                            map_opname_to_op)
                self.solution = sol

                # Check that constraints are respected
                for gate in [g for g in self._gates if
                             (g and g.is_constrained())]:
                    assert gate.is_boolean()
                    if gate.constrained_value != sol[gate]:
                        raise RuntimeError("Internal consistency error")
            return 'sat'

        assert result == 'error'
        return result



    def solve_cvc3(self, cvc3_binary="./cvc3", produce_model = True):
        """
        Solve the circuit by using CVC3 and its native input format.
        Returns either 'error', 'unsat', or 'sat'.
        If 'produce_model' is True and 'sat' is returned,
        the satisfying model can be queried by calling the get_value function.
        """
        assert type(cvc3_binary) is str
        assert type(produce_model) is bool

        # Remove the previous solution (if any)
        self._invalidate_solution()
        # Is already inconsistent?
        if not self.is_consistent:
            return 'unsat'
        
        # Output the circuit in CVC3 format into a temporary file
        #tempfilename = tempfile.mktemp()
        tempfilename = 'tmp_solver.in'
        assert tempfilename
        f = open(tempfilename, 'w')
        (map_var_to_smt_name, map_smt_name_to_var, map_opname_to_op) = self._to_cvc3(f)
        f.close()
        #os.system("cat "+tempfilename)
        #print tempfilename

        solver = CVC3(cvc3_binary)

        (result,sol) = solver._solve_native(tempfilename, True)

        # Remove the temporary file
        #os.remove(tempfilename)

        if result == 'unsat':
            return 'unsat'

        if result == 'sat':
            if produce_model:
                sol = solver._interpret_solution(sol, map_var_to_smt_name,
                                                 map_smt_name_to_var,
                                                 map_opname_to_op)
                sol = self._extend_solution(sol, map_var_to_smt_name,
                                            map_smt_name_to_var,
                                            map_opname_to_op)
                self.solution = sol

                # Check that constraints are respected
                for gate in [g for g in self._gates if
                             (g and g.is_constrained())]:
                    assert gate.is_boolean()
                    if gate.constrained_value != sol[gate]:
                        raise RuntimeError("Internal consistency error")

            return 'sat'

        assert result == 'error'
        return result



    def solve_mathsat(self, mathsat_binary="./mathsat_model",
                      produce_model = True):
        """
        Solve the circuit by using MathSAT and its native input format.
        Returns either 'error', 'unsat', or 'sat'.
        If 'produce_model' is True and 'sat' is returned,
        the satisfying model can be queried by calling the get_value function.
        """
        assert type(mathsat_binary) is str
        assert type(produce_model) is bool

        opt_enum_translation = 'int'
        assert opt_enum_translation in ['int','object']

        # Remove the previous solution (if any)
        self._invalidate_solution()
        # Is already inconsistent?
        if not self.is_consistent:
            return 'unsat'
        
        # Output the circuit in MathSAT format into a temporary file
        #tempfilename = tempfile.mktemp()
        tempfilename = 'tmp_solver.in'
        assert tempfilename
        f = open(tempfilename, 'w')
        (map_var_to_solvervar,
         map_solvervar_to_var,
         map_opname_to_op) = self._to_mathsat(f, opt_enum_translation)
        f.close()
        #print tempfilename

        solver = MathSAT(mathsat_binary)

        (result,sol) = solver._solve_native(tempfilename, True)

        # Remove the temporary file
        #os.remove(tempfilename)

        if result == 'unsat':
            return 'unsat'

        if result == 'sat':
            if produce_model:
                if [uif for uif in map_opname_to_op.values() if
                    isinstance(uif, _Operand_uif)]:
                    raise RuntimeError("Cannot yet reproduce solutions "+
                                       "when solving with MathSAT and "+
                                       "the problem involves UIFs")

                sol = solver._interpret_solution(sol,
                                                 map_var_to_solvervar,
                                                 map_solvervar_to_var,
                                                 map_opname_to_op)
                sol = self._extend_solution(sol,
                                            map_var_to_solvervar,
                                            map_solvervar_to_var,
                                            map_opname_to_op)
                self.solution = sol

                # Check that constraints are respected
                for gate in [g for g in self._gates if
                             (g and g.is_constrained())]:
                    assert gate.is_boolean()
                    if gate.constrained_value != sol[gate]:
                        raise RuntimeError("Internal consistency error")

            return 'sat'

        assert result == 'error'
        return result



    def solve_stp(self, stp_binary="./stp", produce_model = True):
        """
        Solve the circuit by using STP and its native input format.
        Returns either 'error', 'unsat', or 'sat'.
        If 'produce_model' is True and 'sat' is returned,
        the satisfying model can be queried by calling the get_value function.
        """
        assert type(stp_binary) is str
        assert type(produce_model) is bool

        # Remove the previous solution (if any)
        self._invalidate_solution()
        # Is already inconsistent?
        if not self.is_consistent:
            return 'unsat'
        
        # Output the circuit in STP format into a temporary file
        #tempfilename = tempfile.mktemp()
        tempfilename = 'tmp_solver.in'
        assert tempfilename
        f = open(tempfilename, 'w')
        (map_var_to_stp_name, map_stp_name_to_var) = self._to_stp(f)
        f.close()
        #print tempfilename

        solver = STP(stp_binary)

        (result,sol) = solver._solve_native(tempfilename, True)

        # Remove the temporary file
        #os.remove(tempfilename)

        if result == 'unsat':
            return 'unsat'

        if result == 'sat':
            if produce_model:
                sol = solver._interpret_solution(sol,
                                                 map_var_to_stp_name,
                                                 map_stp_name_to_var)
                sol = self._extend_solution(sol,
                                            map_var_to_stp_name,
                                            map_stp_name_to_var,
                                            {})
                self.solution = sol

                # Check that constraints are respected
                for gate in [g for g in self._gates if
                             (g and g.is_constrained())]:
                    assert gate.is_boolean()
                    if gate.constrained_value != sol[gate]:
                        raise RuntimeError("Internal consistency error")

            return 'sat'

        assert result == 'error'
        return result



def sint2bin(width, n, prefix = '0b'):
    """
    Converts a signed, 'width'-bit integer 'n' (given as int or long)
    to binary string.
    Required: width >= 1
    Required: -2^{width-1} <= n < 2^{width-1}
    """
    assert type(width) is int
    assert (type(n) is int) or (type(n) is long)
    assert 1 <= width
    assert -pow(2, width-1) <= n and n < pow(2, width-1)
    assert type(prefix) is str
    s = ''
    for i in xrange(0, width):
        s = str(n % 2) + s
        n = n / 2
    return prefix+s


def sint2hex(width, n, prefix = '0h'):
    """
    Converts a signed, 'width'-bit integer 'n' (given as int or long)
    to hexadecimal string.
    Required: width >= 1 and is a multiple of four
    Required: -2^{width-1} <= n < 2^{width-1}
    """
    assert type(width) is int
    assert (type(n) is int) or (type(n) is long)
    assert 1 <= width and width % 4 == 0
    assert -pow(2, width-1) <= n and n < pow(2, width-1)
    assert type(prefix) is str
    s = ''
    a = ['0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f']
    for i in xrange(0, width / 4):
        s = a[n % 16] + s
        n = n / 16
    return prefix+s


def uint2bin(width, n, prefix = '0b'):
    """
    Converts an unsigned, 'width'-bit integer 'n' (given as int or long)
    to binary string.
    Required: width >= 1
    Required: 0 <= n < 2^{width}
    """
    assert type(width) is int
    assert (type(n) is int) or (type(n) is long)
    assert 1 <= width
    assert 0 <= n and n < pow(2, width)
    assert type(prefix) is str
    s = ''
    for i in xrange(0, width):
        s = str(n % 2) + s
        n = n / 2
    return prefix+s


def uint2hex(width, n, prefix = '0h'):
    """
    Converts an unsigned, 'width'-bit integer 'n' (given as int or long)
    to hexadecimal string.
    Required: width >= 1 and is a multiple of four
    Required: -2^{width-1} <= n < 2^{width-1}
    """
    assert type(width) is int
    assert (type(n) is int) or (type(n) is long)
    assert 1 <= width and width % 4 == 0
    assert 0 <= n and n < pow(2, width)
    assert type(prefix) is str
    s = ''
    a = ['0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f']
    for i in xrange(0, width / 4):
        s = a[n % 16] + s
        n = n / 16
    return prefix+s

def bin2sint(s):
    """
    Converts a non-empty string of 0's and 1's in two's complement notation
    to a long integer, the first bit is interpreted as a sign bit.
    """
    assert type(s) is str
    assert s
    width = len(s)
    v = long(0)
    for (i,b) in enumerate(s):
        assert b == '0' or b == '1'
        if b == '1':
            v += pow(2, width-1-i)
    if v >= pow(2,width-1):
        v = -pow(2,width)+v
    return v

def bin2uint(s):
    """
    Converts a non-empty string of 0's and 1's to a long integer.
    """
    assert type(s) is str
    assert s
    width = len(s)
    v = long(0)
    for (i,b) in enumerate(s):
        assert b == '0' or b == '1'
        if b == '1':
            v += pow(2, width-1-i)
    return v




# Arithmetic operations for bounded signed ints represented with longs
def _sint_plus(width, v1, v2):
    """Compute v1+v2 of bounded signed integers v1 and v2."""
    assert (type(width) is int) or (type(width) is long)
    assert 1 <= width
    assert type(v1) is long
    assert type(v2) is long
    assert v1 >= -pow(2, width-1) and v1 < pow(2, width-1)
    assert v2 >= -pow(2, width-1) and v2 < pow(2, width-1)
    value = v1 + v2
    value = value % pow(2,width)
    if value >= pow(2,width-1):
        value = -pow(2,width)+value
    return value

def _sint_neg(width, v):
    """Compute -v of a bounded signed integer v."""
    assert (type(width) is int) or (type(width) is long)
    assert 1 <= width
    assert type(v) is long
    assert v >= -pow(2, width-1) and v < pow(2, width-1)
    value = -v
    value = value % pow(2,width)
    if value >= pow(2,width-1):
        value = -pow(2,width)+value
    return value

def _sint_minus(width, v1, v2):
    """Compute v1-v2 of bounded signed integers v1 and v2."""
    assert (type(width) is int) or (type(width) is long)
    assert 1 <= width
    assert type(v1) is long
    assert type(v2) is long
    assert v1 >= -pow(2, width-1) and v1 < pow(2, width-1)
    assert v2 >= -pow(2, width-1) and v2 < pow(2, width-1)
    value = v1 - v2
    value = value % pow(2,width)
    if value >= pow(2,width-1):
        value = -pow(2,width)+value
    return value

def _sint_times(width, v1, v2):
    """Compute v1*v2 of bounded signed integers v1 and v2."""
    assert (type(width) is int) or (type(width) is long)
    assert 1 <= width
    assert type(v1) is long
    assert type(v2) is long
    assert v1 >= -pow(2, width-1) and v1 < pow(2, width-1)
    assert v2 >= -pow(2, width-1) and v2 < pow(2, width-1)
    value = v1 * v2
    value = value % pow(2,width)
    if value >= pow(2,width-1):
        value = -pow(2,width)+value
    return value

def _sint_or(width, v1, v2):
    """
    Compute the bitwise or of width-bit signed integers v1 and v2.
    v1, v2, and the result are represented as longs
    in the range [-2^(width-1),2^(width-1)).
    """
    assert (type(width) is int) or (type(width) is long)
    assert 1 <= width
    assert type(v1) is long
    assert type(v2) is long
    assert v1 >= -pow(2, width-1) and v1 < pow(2, width-1)
    assert v2 >= -pow(2, width-1) and v2 < pow(2, width-1)
    value = v1 | v2
    assert value >= -pow(2, width-1) and value < pow(2, width-1)
    return value

def _sint_and(width, v1, v2):
    """
    Compute the bitwise and of width-bit signed integers v1 and v2.
    v1, v2, and the result are represented as longs
    in the range [-2^(width-1),2^(width-1)).
    """
    assert (type(width) is int) or (type(width) is long)
    assert 1 <= width
    assert type(v1) is long
    assert type(v2) is long
    assert v1 >= -pow(2, width-1) and v1 < pow(2, width-1)
    assert v2 >= -pow(2, width-1) and v2 < pow(2, width-1)
    value = v1 & v2
    assert value >= -pow(2, width-1) and value < pow(2, width-1)
    return value

def _sint_xor(width, v1, v2):
    """
    Compute the bitwise xor of width-bit signed integers v1 and v2.
    v1, v2, and the result are represented as longs
    in the range [-2^(width-1),2^(width-1)).
    """
    assert (type(width) is int) or (type(width) is long)
    assert 1 <= width
    assert type(v1) is long
    assert type(v2) is long
    assert v1 >= -pow(2, width-1) and v1 < pow(2, width-1)
    assert v2 >= -pow(2, width-1) and v2 < pow(2, width-1)
    value = v1 ^ v2
    assert value >= -pow(2, width-1) and value < pow(2, width-1)
    return value

def _sint_not(width, v):
    """
    Compute the bitwise not of width-bit signed integer v.
    v and the result are represented as longs
    in the range [-2^(width-1),2^(width-1)).
    """
    assert (type(width) is int) or (type(width) is long)
    assert 1 <= width
    assert type(v) is long
    assert v >= -pow(2, width-1) and v < pow(2, width-1)
    value = ~v
    assert value >= -pow(2, width-1) and value < pow(2, width-1)
    return value



# Arithmetic operations for bounded unsigned ints represented with longs
def _uint_plus(width, v1, v2):
    """Compute v1+v2 of width-bit unsigned integers v1 and v2.
    v1, v2, and the result are represented as longs
    in the range [0,2^width)."""
    assert (type(width) is int) or (type(width) is long)
    assert 0 <= width
    assert type(v1) is long
    assert type(v2) is long
    assert v1 >= 0 and v1 < pow(2, width)
    assert v2 >= 0 and v2 < pow(2, width)
    return (v1 + v2) % pow(2,width)

def _uint_neg(width, v):
    """Compute -v of a bounded unsigned integer v.
    v and the result are represented as longs in the range [0,2^width)."""
    assert (type(width) is int) or (type(width) is long)
    assert 0 <= width
    assert type(v) is long
    assert v >= 0 and v < pow(2, width)
    if v == 0:
        return long(0)
    return pow(2, width) - v

def _uint_minus(width, v1, v2):
    """Compute v1-v2 of width-bit unsigned integers v1 and v2.
    v1, v2, and the result are represented as longs
    in the range [0,2^width)."""
    assert (type(width) is int) or (type(width) is long)
    assert 0 <= width
    assert type(v1) is long
    assert type(v2) is long
    assert v1 >= 0 and v1 < pow(2, width)
    assert v2 >= 0 and v2 < pow(2, width)
    value = (v1 - v2)
    if value < 0:
        value += pow(2,width)
    return value

def _uint_times(width, v1, v2):
    """Compute v1*v2 of width-bit unsigned integers v1 and v2.
    v1, v2, and the result are represented as longs
    in the range [0,2^width)."""
    assert (type(width) is int) or (type(width) is long)
    assert 0 <= width
    assert type(v1) is long
    assert type(v2) is long
    assert v1 >= 0 and v1 < pow(2, width)
    assert v2 >= 0 and v2 < pow(2, width)
    return (v1 * v2) % pow(2,width)

def _uint_div(width, v1, v2):
    """Compute v1/v2 of width-bit unsigned integers v1 and v2.
    The result is the largest integer x such that x*v2 <= v1.
    v1, v2, and the result are represented as longs
    in the range [0,2^width).
    Throws a ZeroDivisionError is v2 is 0.
    """
    assert (type(width) is int) or (type(width) is long)
    assert 0 <= width
    assert type(v1) is long
    assert type(v2) is long
    assert v1 >= 0 and v1 < pow(2, width)
    assert v2 >= 0 and v2 < pow(2, width)
    return v1 / v2

def _uint_mod(width, v1, v2):
    """Compute v1 % v2 of width-bit unsigned integers v1 and v2.
    v1, v2, and the result are represented as longs
    in the range [0,2^width).
    Throws a ZeroDivisionError is v2 is 0.
    """
    assert (type(width) is int) or (type(width) is long)
    assert 0 <= width
    assert type(v1) is long
    assert type(v2) is long
    assert v1 >= 0 and v1 < pow(2, width)
    assert v2 >= 0 and v2 < pow(2, width)
    return v1 % v2

def _uint_or(width, v1, v2):
    """
    Compute the bitwise or of width-bit unsigned integers v1 and v2.
    v1, v2, and the result are represented as longs
    in the range [0,2^width).
    """
    assert (type(width) is int) or (type(width) is long)
    assert 0 <= width
    assert type(v1) is long
    assert type(v2) is long
    assert v1 >= 0 and v1 < pow(2, width)
    assert v2 >= 0 and v2 < pow(2, width)
    return v1 | v2

def _uint_and(width, v1, v2):
    """
    Compute the bitwise and of width-bit unsigned integers v1 and v2.
    v1, v2, and the result are represented as longs
    in the range [0,2^width).
    """
    assert (type(width) is int) or (type(width) is long)
    assert 0 <= width
    assert type(v1) is long
    assert type(v2) is long
    assert v1 >= 0 and v1 < pow(2, width)
    assert v2 >= 0 and v2 < pow(2, width)
    return v1 & v2

def _uint_xor(width, v1, v2):
    """
    Compute the bitwise xor of width-bit unsigned integers v1 and v2.
    v1, v2, and the result are represented as longs
    in the range [0,2^width).
    """
    assert (type(width) is int) or (type(width) is long)
    assert 0 <= width
    assert type(v1) is long
    assert type(v2) is long
    assert v1 >= 0 and v1 < pow(2, width)
    assert v2 >= 0 and v2 < pow(2, width)
    return v1 ^ v2

def _uint_not(width, v):
    """
    Compute the bitwise not of width-bit unsigned integer v.
    v and the result are represented as longs
    in the range [0,2^width).
    """
    assert (type(width) is int) or (type(width) is long)
    assert 0 <= width
    assert type(v) is long
    assert v >= 0 and v < pow(2, width)
    result = 0
    for i in xrange(0, width):
        if v & pow(2, i) == 0:
            result += pow(2, i)
    return result

# Auxiliary
def _chop(line):
    """
    A Perl-like chop, removes the newlines from the end of the line.
    """
    assert (line is None) or (type(line) is str)
    while line and line[-1] == '\n':
        line = line[:-1]
    return line
